/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : sim-ctl-tx.c
 *
 * Description  :
 * Simulates Agile Mirror (AM) controller.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

#include <stdlib.h>         // exit
#include <stdio.h>          // fprintf
#include <netdb.h>          // gethostbyname
#include <string.h>         // rindex
#include <sys/socket.h>     // inet_ntoa
#include <netinet/in.h>     // inet_ntoa
#include <arpa/inet.h>      // inet_ntoa

#define FS_MAXNAMELEN   255  // For simulation only

#include "exports/evm-api.h"
#include "exports/fs-api.h"
#include "exports/anpssc-api.h"
#include "agent-api.h"

/*------------------------------------------------------------------------------
 * Constants.
 *------------------------------------------------------------------------------
 */

/* Wait for EVM/EFS to respond (10s).
 */
static struct timeval waitEvmTo = {10, 0};

/*------------------------------------------------------------------------------
 * Data structures definition
 *------------------------------------------------------------------------------
 */

/* Basic message structure. */
typedef struct
{
    char    buf[500];  // optimize!!!
    uint32  len;
} am_msg_t;

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

/* The name of this program. */
static char * myName;

/*------------------------------------------------------------------------------
 * Prototype of local functions
 *------------------------------------------------------------------------------
 */

static void
print_usage(void);

/*------------------------------------------------------------------------------
 * Name         : print_usage()
 *
 * Description  : Shows the correct usage of this program.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 02-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
print_usage(void)

{
	fprintf(stderr,"Usage:\n");

	fprintf(stderr," %s server\n", myName);
	fprintf(stderr," %s client\n", myName);
	fprintf(stderr,"\n");
}

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : It all begins here.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 02-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

int
main(int argc, char *argv[])

{
	char *      actionStr;
	char *      objectStr;  // source of target
    am_msg_t    msg;
    struct hostent * hostptr;
    uint32      myIpAddr = 0;
    
    /* Initialize my own name.
     * Discard path if any.
     */
	myName = argv[0];
	if (rindex(myName, '/'))
    {
		strcpy(myName, rindex(myName, '/') + 1);
	}

    if ((hostptr = gethostbyname("localhost")) == NULL)
    {
        perror("gethostbyname()");
    }
    else
    {
        if (hostptr->h_addrtype == AF_INET)
        {
            struct in_addr * ptr;

            ptr = (struct in_addr *) * hostptr->h_addr_list;
			ptr->s_addr = 0x1408030a; // 10.3.8.20
			ptr->s_addr = 0x6306040a; // 10.4.6.99
			ptr->s_addr = 0x0101040a; // 10.4.1.1
            myIpAddr = (uint32) ptr->s_addr; // network byte order
            printf("\tInternet address: %s\n", inet_ntoa(*ptr));
        }
        else
        {
            exit(NFX_ERR);
        }
    }
    
    /* Make sure there's at least the action and the object strings.
     * Get the command and the object on which to execute the command.
     */
	if (argc < 2)
    {
		print_usage();
		exit(NFX_ERR);
	}
	objectStr = argv[1];
	//actionStr = argv[2];
      
    /* Each "command" has it's own action function.
     */
	if (!strcasecmp(objectStr, "server"))
    {
        am_listenReq_t *req = (am_listenReq_t *)&(msg.buf);
        
        req->hdr.msgType = AM_AGENT_LISTEN_REQ;
        req->hdr.amApplHdl = 0; // not used
        req->hdr.ctrlHdl  = 0; // don't care for test
            
        req->lclAddr = 0;    // local IP @ (server)
        req->lclAddr = myIpAddr;    // local IP @ (server)
        req->lclPort = htons(11000);    // local port (server)

		printf("ctltx server lclPort is %d\n", ntohs(req->lclPort));

        msg.len = sizeof(am_listenReq_t);
    }
	else if (!strcasecmp(objectStr, "client"))
    {
        am_startMirrorReq_t *req = (am_startMirrorReq_t *) &(msg.buf);
        
        req->hdr.msgType = AM_START_MIRROR_REQ;
        req->hdr.amApplHdl = 0; // not used
        req->hdr.ctrlHdl = 0;  // don't care for test
            
        req->lclAddr = myIpAddr;    // local IP @ (client)
        // req->lclPort = htons(11001);    // local port (client)
        req->rmtAddr = myIpAddr;  
        req->rmtPort = htons(11000);
		
		printf("ctltx client rmtPort is %d\n", ntohs(req->rmtPort));

        // get volume id
        {
            evm_MsgGetId_t      evmReq;
            am_msg_t            evmMsg;
            evm_MsgStat_t *     evmRsp = (evm_MsgStat_t *) evmMsg.buf;

            /* Reset memory for request and response.
             */
            bzero((void *) &evmReq, sizeof(evmReq));
            bzero((void *) &evmMsg, sizeof(evmMsg));
            
            /* Fill request.
             */
            evmReq.hdr.msgType = EVM_VOL_GET_ID;
            evmReq.hdr.clientPrivate = 0;   // not used
            strcpy(evmReq.name, "vol1");

            /* Send request to EVM.
             */
            evmMsg.len = sendAgileMsg(-1,
                                      &evmReq, sizeof(evmReq), 
                                      "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                      &waitEvmTo,
                                      evmMsg.buf, sizeof(evmMsg.buf));

            /* Receive response from EVM.
             * Copy the returned volume id in control block.
             */
            NFX_ASSERT(evmMsg.len >= sizeof(evm_msgHdr_t));
            NFX_ASSERT(evmRsp->hdr.msgType == EVM_MSG_STAT);
            
            if (evmRsp->stat == NFX_OK)
            {
                memcpy((void *) &(req->amVolId),
                       (const void *) &(evmRsp->payload.lVolId),
                       sizeof(evm_volId_t));
            }
            else
            {
                exit(NFX_ERR);
            }
            printf("my volume id = %lld\n", req->amVolId.whole_thing);
        }
#ifdef IN_CLUSTER_ONLY
        // get target volume id
        {
            evm_MsgGetId_t      evmReq;
            am_msg_t            evmMsg;
            evm_MsgStat_t *     evmRsp = (evm_MsgStat_t *) evmMsg.buf;

            /* Reset memory for request and response.
             */
            bzero((void *) &evmReq, sizeof(evmReq));
            bzero((void *) &evmMsg, sizeof(evmMsg));
            
            /* Fill request.
             */
            evmReq.hdr.msgType = EVM_VOL_GET_ID;
            evmReq.hdr.clientPrivate = 0;   // not used
            strcpy(evmReq.name, "vol2");

            /* Send request to EVM.
             */
            evmMsg.len = sendAgileMsg(-1,
                                      &evmReq, sizeof(evmReq), 
                                      "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                      &waitEvmTo,
                                      evmMsg.buf, sizeof(evmMsg.buf));

            /* Receive response from EVM.
             * Copy the returned volume id in control block.
             */
            NFX_ASSERT(evmMsg.len >= sizeof(evm_msgHdr_t));
            NFX_ASSERT(evmRsp->hdr.msgType == EVM_MSG_STAT);
            
            if (evmRsp->stat == NFX_OK)
            {
                memcpy((void *) &(req->tgtVolId),
                       (const void *) &(evmRsp->payload.lVolId),
                       sizeof(evm_volId_t));
            }
            else
            {
                exit(NFX_ERR);
            }
            printf("target volume id = %lld\n", req->amVolId.whole_thing);
        }
#else /* IN_CLUSTER_ONLY */
		strcpy(req->tgtFsysName, "vol2");
#endif /* IN_CLUSTER_ONLY */

        msg.len = sizeof(am_startMirrorReq_t);
    }
	else
    {
		fprintf(stderr, "Invalid command: %s\n",actionStr);
		print_usage();
		exit(NFX_ERR);
	}

    /* Send message to AM agent.
     * In simulation environment, N-NIM is on slot 1.
     */
    printf("sending...\n");
    if (sendAgileMsg(-1,
                     &msg.buf, msg.len,
                     "am", NFX_NFP_FP2, 1, MGMTPLANE,
                     NULL, NULL, 0) != 0)
    {
        perror("sendAgileMsg() failed\n");
        exit(NFX_ERR);
    }
		
	exit(NFX_OK);
}

/*----------------------------------------------------------------------------*/
