/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: Agile Network Protocol on the SSC (anpssc-api.h)
 *
 * RCS:  $Id: anpssc-api.h,v 1.17 2002/11/15 21:50:22 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: This defines the Agile Network message layer API for
 *              applications running on the SSC.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 1 June 2001
 *
 *-----------------------------------------------------------------
 */
#if !defined(__anpssc_api_h)
#define __anpssc_api_h

#if defined(NFX_MOD_SSC)	// Only valid for SSC compiles
#include <inttypes.h>		// For ANSI types
#include <sys/types.h>		// For other types
#include <sys/time.h>		// For struct timeval

/*
	This defines the abstraction applications on the SSC use to
	send and receive messages via the Agile network message passing protocol 
	(part of EEE on the embedded side). These two interfaces provide the
	following abstractions:

		1) Register to receive messages addressed to an "application" on the
		   SSC. What will be created here is an open file descriptor that
		   the application should pass to the message receiving function.

		2) Send messages (with option to block for a response) to an
		   "application" somewhere within the NFX over either the
		   management plane or the data plane. The addressing of these
		   messages incorporates the Agile network addressing concept
		   of (application, slot, CPU) in an easy to use form.

	Nothing in this abstraction ensures that messages sent are actually
	received by the addressee. There are facilities to specify timeouts
	(in the quanta of the OS - 10 ms) to wait for a response. This will
	allow the application to retransmit, etc, if it so desires.

	This is a classic client-server model. The server is an application that
	registers itself via regAgileApp() and listens for messages via
	recvAgileMsg(). The clients send messages to the server via 
	sendAgileMsg(). The mechanics of this abstraction allows the server
	to uniquely identify the source of the message and send a response via
	sendAgileMsg() (without expecting a response).

	This abstraction is agnostic as to the contents of the messages.
 */

/* The reception abstraction:

	First, you register to receive the messages for your application and
	you get a file descriptor onto which those messages will show up.

	Once you have that open file descriptor, you can do any of the things
	UNIX applications do with file descriptors. It can be set be non-blocking
	(via fcntl). You can have it send you a signal when data becomes ready
	(again fcntl). Etc. etc. etc.

	Better yet, you can use the provided reception function to wait for
	a message to arrive. If you're impatient, you can specify a timeout
	period after which it will just return and tell you that you got
	nothing.

	Optional arguments are set to NULL to disable that option.
	For the "timeout" argument, NULL says wait forever.
 */

int regAgileApp (char *appName);			// Which application are you
											// Bcast/Mcast will have own
											// "names".

void unRegAgileApp (int fd);

// Returns an open file descriptor
// Errors get you -1 and errno set appropriately

int recvAgileMsg (int fd, 					// From regAgileApp()
                  void *msgBuff, 			// Where to put the message
                  ssize_t msgLen, 			// and how much room there is
                  struct timeval *timeout,	// (opt) How long to wait
                  char *appName,			// (opt) From which application
                  uint32_t *cpuId,			// (opt) and where
                  uint32_t *slotId,			// (opt) it is on in the NFX
                  uint32_t *rPlane);		// (opt) On what plane did it arrive

// The source of the message can be obtained in via the appName, cpuId,
// slotId triplet. The appName string will contain a string that can be
// passed to sendAgileMsg() and likewise with cpuID and slotID to send
// a response from this message.

// Returns the number of bytes copied to msgBuff
// Timeout gets you 0
// Errors get you -1 and errno set appropriately from the recvfrom()


/* The transmission abstraction:

	Quite simply, send this message to that application over yonder and,
	if I tell you, wait for his response and put it where I told you.

	As for the management vs. data plane vs. the other CPU(s) on the
	SSC, it only makes a difference when the message is sent to another
	card.

	Optional arguments are set to NULL to disable that option.
	For the "timeout" argument, NULL says wait forever.

	To send a broadcast or multicast messages, the slotId field will be
	overloaded using -1 to -15. The Bcast/Mcast field in the Agile header
	will be the absolute value of this argument.

	Optionally, "sock" specifies an open socket fd. This is for use of
	registered applications allowing responses on their registered port.
	Setting "sock" to -1 tells sendAgileMsg() to create a socket, with
	system assigned port, to send the message and optionally wait for
	the response.
 */

int sendAgileMsg (int sock,					// Use this socket
                  void *msgBuff,			// The message
                  ssize_t msgLen,			// How many bytes it be
                  char *appName,			// To which application
                  uint32_t cpuId,			// and where
                  uint32_t slotId,			// it is on in the NFX
                  uint32_t whichPlane,		// 0=mgmt; 1=data
                  struct timeval *timeout,	// (opt) How long to wait for resp
                  void *respBuff,			// (opt) Where to put resp msg
                  ssize_t respLen);			// (opt) how big it is

// Returns the number of bytes copied to respBuff.
// No response requested gets you 0 and you don't block waiting for one.
// Errors get you -1 and errno set appropriately from sendto().
// If you provide a bogus "appName", you'll get ESRCH.

/*
   Checks to see if the specified CPU on the specified slot is
   in run time or not. If the CPU is in run time this returns
   NFX_OK else it returns a NFX_ERR. This API synchronously
   talks to the chassisd daemon also running on the SSC with a 
   5 second timeout to determine the state of the desired CPU
*/
 
int checkAgileDest(uint32_t cpu, uint32_t slot);

/*
 * Connect the socket to destination. The destination application
 * will receive the EEE_MSG_WINDOW_UPDATE message when the process
 * pulls a message from its receive queue
 */
int eee_connect(int sock, int slot, int cpu, int app);

#define AGILE_BASE 50000		// Start of Agile ports
#define MGMTPLANE 2
#define DATAPLANE 1


// Not part of the API so avert your eyes

/* for passing inter-slot message to SSC daemon */
typedef struct _ans_ssc_msghdr {
	int slot;
	int cpu;
	int appid;
	int plane;
} ans_ssc_msghdr_t;

/*
	And the "fake" EEE translation functions that show up in this
	library
 */
#include "nfx-incl.h"
extern int32 eee_getApplicationId (uchar8 *name);
extern uchar8 *eee_getAppNameFromId (int32 id);
extern int bindEEESocket(int s, int slot, int cpu, int app);

/* use only 1 definition of slot types as defined in nfx-defs.h */
#define SSC_SLOT            NFX_SLOTTYPE_SSC 

#define AGILE_SO_RCVBUF_LEN	262144
#define AGILE_SO_SNDBUF_LEN	262144

int setSocketOpt(int sock);
/*
 * Given an open EEE socket, tell me the application name bound to it
 * Returns 0 on success. -1 on error. See the man page for getsockname
 * for errno explanations.
 */
extern int app_name_from_socket (int s, char *appName);

#endif

#endif
