/* vi:ts=4
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include "anpssc-api.h"
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/uio.h>

#include "nfx-types.h"
#include "nfx-defs.h"
#include "nfx-error.h"
#include "exports/eee-desc.h"
#include "neteee/eee.h"
#include "exports/eee-error.h"
#include "exports/cm-api.h"

static const char AgileName[] = "Agile";
static const size_t AgileNameLen = sizeof(AgileName) - 1;


int
eee_connect(int fd, int slot, int cpu, int app)
{
	struct sockaddr_eee addr;
	addr.seee_len = sizeof(addr);
	addr.seee_family = AF_EEE;
	addr.seee_slot = slot;
	addr.seee_cpu = cpu;
	addr.seee_app = app;
	addr.seee_bplane = EEE_CONTROLBPLANE;
	return connect(fd, (struct sockaddr *)&addr, sizeof(addr));
}

int
bindEEESocket(int s, int slot, int cpu, int app)
{
	struct sockaddr_eee addr;
	int rc;
	socklen_t len __attribute__((unused)) = sizeof(addr);
	assert(0 <= app && app <= EDESC2_APP_NUM && 0 <= slot 
		   && slot < EDESC2_SLOT_NUM && 0 <= cpu
		   && cpu < EDESC2_CPU_NUM);
	addr.seee_family = AF_EEE;
	addr.seee_slot = slot;
	addr.seee_cpu = cpu;
	addr.seee_app = app;
	addr.seee_bplane = EEE_CONTROLBPLANE;
	rc = bind(s, (struct sockaddr *)&addr, sizeof(addr)) == 0;

	/* Check if the address we got is the address that was requested */
	assert(rc == 0 ||
		   (getsockname(s, (struct sockaddr *)&addr, &len) == 0
			&& (addr.seee_slot == EEE_NULLSLOT || addr.seee_slot == slot)
			&& (addr.seee_cpu == EEE_NULLCPU || addr.seee_cpu == cpu)
			&& (app == 0 || addr.seee_app == app)));

	return rc;
}

static int
bindEEESocketForApp(int s, char *app)
{
	int appId;
	return (appId = eee_getApplicationId(app)) != EEE_ERR_INVALID_APP_NAME
		&& bindEEESocket(s, SSC_SLOT, NFX_SSC_UNIX, appId);
}


int
regAgileApp(char *appName)
{
	int s;
	if ((s = socket(PF_EEE, SOCK_DGRAM, 0)) >= 0
		&& !bindEEESocketForApp(s, appName)) {
		close(s);
		s = -1;
	}

	if (s >= 0)
		setSocketOpt(s);

	return s;
}


void
unRegAgileApp(int fd)
{
	close(fd);
}


static int
validMsgLen(ssize_t len)
{
	return len < EDESC_TOTAL_LEN;
}


static int
waitData(int s, struct timeval *timeout)
{
	fd_set fds;
	struct timeval localTimeout; 
	struct timeval start, present, deltat;
	int rewait;
	int rc;
	
	FD_ZERO(&fds);
	FD_SET(s, &fds);

	assert(timeout != 0);
	localTimeout = *timeout;

	gettimeofday (&start, NULL);
	do {	// Just continue through signal delivery
		rc = select(s + 1, &fds, 0, 0, &localTimeout);

		rewait = (rc < 0) && (errno == EINTR);

		if (rewait) {
			gettimeofday (&present, NULL);
			timersub (&present, &start, &deltat);
			timersub (timeout, &deltat, &localTimeout);
			if ((localTimeout.tv_sec < 0) || (localTimeout.tv_usec < 0)) {
				rc = 0;
				break;		// timed out AND got interrupted at the same time
			}
		}
	} while (rewait);
	return rc;
}


static int
getResponse(int s, void *respBuff, ssize_t respLen,
			 struct timeval *timeout)
{
	int rc;

	if (!timeout || (rc = waitData(s, timeout)) > 0)
		rc = recv(s, respBuff, respLen, 0);

	return rc;
}


static int
getResponseFrom(int s, void *respBuff, ssize_t respLen,
				struct timeval *timeout, struct sockaddr_eee *from)
{
	int rc;
	int salen = sizeof(*from);
	
	if (!timeout || (rc = waitData(s, timeout)) > 0)
		rc = recvfrom(s, respBuff, respLen, 0, (struct sockaddr *)from,
					  &salen);

	return rc;
}


static int
getClientAppId(char *appName, int *id)
{
	char *endptr;

	assert(!strncmp(appName, AgileName, AgileNameLen));
	*id = strtol(&appName[AgileNameLen], &endptr, 10);
	return appName[AgileNameLen] != 0 && *endptr == 0;
}


static int
getRealAppId(char *appName, int *id)
{
	int rc;
	
	if (strncmp(appName, AgileName, strlen(AgileName)))
		rc = (*id = eee_getApplicationId(appName)) !=
			EEE_ERR_INVALID_APP_NAME;
	else
		rc = getClientAppId(appName, id);
	
	return rc;
}


int
recvAgileMsg (int fd,                   // From regAgileApp()
			  void *msgBuff,            // Where to put the message
			  ssize_t msgLen,           // and how much room there is
			  struct timeval *timeout,  // (opt) How long to wait
			  char *appName,			// (opt) From which application
			  uint32_t *cpuId,			// (opt) and where
			  uint32_t *slotId, 		// (opt) it is on in the NFX
			  uint32_t *rPlane)			// (opt) On which plane it came?
{
	int rc;
	struct sockaddr_eee seee;

	if ((rc = getResponseFrom(fd, msgBuff, msgLen, timeout, &seee)) > 0) {
		if (appName) {
			char *gotAppName;
			if ((gotAppName = eee_getAppNameFromId(seee.seee_app)))
				strcpy(appName, gotAppName);
			else				/* Not a registered application */
				sprintf(appName, "Agile%d", seee.seee_app);
		}
		if (cpuId) *cpuId = seee.seee_cpu;
		if (slotId) *slotId = seee.seee_slot;
		if (rPlane) *rPlane = seee.seee_bplane;
	}

	return rc;
}


static int
newSocket()
{
	int sock;
	struct sockaddr_eee seee;

	seee.seee_family = AF_EEE;
	seee.seee_bplane = 0;
	seee.seee_slot = SSC_SLOT;
	seee.seee_cpu = NFX_SSC_UNIX;
	seee.seee_app = 0;
	if ((sock = socket(PF_EEE, SOCK_DGRAM, 0)) >= 0) {
		if (bind(sock, (struct sockaddr *)&seee, sizeof(seee)) < 0) {
			close(sock);
			sock = -1;
		}
	}
	
	return sock;
}


static int
makeSocket(int *sock)
{
	if (*sock == -1) {
		*sock = newSocket();
		if (*sock >= 0)
			setSocketOpt(*sock);
	}

	return *sock != -1;
}

int
setSocketOpt(int sock)
{
	int rc;
	int rcvbuf, rcvbuf_len;
	int sndbuf, sndbuf_len;

	rcvbuf_len = sizeof(rcvbuf);
	rcvbuf = AGILE_SO_RCVBUF_LEN;
	rc = setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &rcvbuf, rcvbuf_len);
#if AS_IF_WE_CARE
	if (rc == -1)
		perror("setSocketOpt: setsockopt failed");
	rc = getsockopt(sock, SOL_SOCKET, SO_RCVBUF, &rcvbuf, &rcvbuf_len);
	/* printf("setSocketOpt: Setting SO_RCVBUF to %d\n", rcvbuf); */
#endif
	
	sndbuf = AGILE_SO_SNDBUF_LEN;
	sndbuf_len = sizeof(sndbuf);
	rc = setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &sndbuf, sndbuf_len);
#if AS_IF_WE_CARE
	if (rc == -1)
		perror("setSocketOpt: setsockopt failed");
	rc = getsockopt(sock, SOL_SOCKET, SO_SNDBUF, &sndbuf, &sndbuf_len);
	/* printf("setSocketOpt: Setting SO_SNDBUF to %d\n", sndbuf); */
#endif
	
	return(rc != -1);
}

 
static int
sendAgileMsgTo(int sock, void *msg, ssize_t len, char *appName,
			   uint32_t cpuId, uint32_t slotId, uint32_t whichPlane)
{
	int rc;
	int valid_app;

	if (appName) {
		struct sockaddr_eee seee;
		int appid;
		
		seee.seee_family = AF_EEE;
#ifdef notyet
		seee.seee_bplane = whichPlane;
#else
		/* Use bplane 1 to send to toshiba, 2 to all others */
		seee.seee_bplane = (cpuId == 1 && slotId == 0) ? 1 : 2;
#endif			
		seee.seee_slot = slotId;
		seee.seee_cpu = cpuId;
		valid_app = getRealAppId(appName, &appid);
		if (valid_app == EEE_ERR_INVALID_APP_NAME) {
			errno = ESRCH;
			rc = -1;
		}
		else {
			seee.seee_app = appid;
			rc = sendto(sock, msg, len, 0, (struct sockaddr *)&seee,
					sizeof(seee));
		}
	}
	else
		rc = send(sock, msg, len, 0);

	return rc;
}


int
sendAgileMsg (int regSock,				// Use this socket please
			  void *msgBuff,            // The message
			  ssize_t msgLen,           // How many bytes it be
			  char *appName,            // To which application
			  uint32_t cpuId,           // and where
			  uint32_t slotId,          // it is on in the NFX
			  uint32_t whichPlane,      // 0=mgmt; 1=data
			  struct timeval *timeout,  // (opt) How long to wait for resp
			  void *respBuff,           // (opt) Where to put resp msg
			  ssize_t respLen)          // (opt) how big it is
{
	int rc = -1;
	int oldSock = regSock;

	if (validMsgLen(msgLen)) {
		if (makeSocket(&regSock)
			&& (rc = sendAgileMsgTo(regSock, msgBuff, msgLen, appName,
									cpuId, slotId, whichPlane)) >= 0) {
			if (respBuff) {
				rc = getResponse(regSock, respBuff, respLen, timeout);
			}
			else {
				rc = 0;	// won't receive a thing so...
			}
		}
	}
	else {
		errno = EMSGSIZE;
		rc = -1;
	}

	if (regSock != oldSock && regSock != -1)
		close(regSock);

	return rc;
}

int checkAgileDest (uint32_t cpu, uint32_t slot)
{
	chassisdMsgInfoReq_t	cmReq;
	chassisdMsgInfoResp_t	cmRsp;
    int32               	rc = NFX_OK;
	int32					status = NFX_OK;

	struct timeval timeOut = {5, 0};

    /* 
	 * Initialize the structures.
     */
	bzero((void *)&cmRsp, sizeof(chassisdMsgInfoResp_t));
	bzero((void *)&cmReq, sizeof(chassisdMsgInfoReq_t));

	/*
	 * send the request to Chassisd
	 */
	cmReq.hdr.mid = CHASSISD_GET_CHASSIS_INFO_REQ;

	rc = sendAgileMsg(-1, &cmReq, sizeof(chassisdMsgInfoReq_t),
	                        "cm", 0, 0, 0,
						  &timeOut, &cmRsp, sizeof(chassisdMsgInfoResp_t));

	if (rc != sizeof(chassisdMsgInfoResp_t))
	{
		status = NFX_ERR;
	}
	else if (cmRsp.present[slot] == 0)
	{
		status = NFX_ERR;
	}
	else if (cmRsp.cpu_state[slot][cpu] != CM_SLICE_STATE_RUNTIME)
	{
		status = NFX_ERR;
	}
	else
		status = NFX_OK;

	return status;
}

/*
 * Given an open EEE socket, tell me its application name.
 */
extern int
app_name_from_socket (int s, char *appName)
{
	char *gotAppName;
	struct sockaddr_eee seee;
	socklen_t seee_len = sizeof (struct sockaddr_eee);
	int rc = 0;

	rc = getsockname (s, (struct sockaddr *)&seee, &seee_len);

	if (!rc) {
		if ((gotAppName = eee_getAppNameFromId(seee.seee_app))) {
			strcpy(appName, gotAppName);
		}
		else {				/* Not a registered application */
			sprintf(appName, "Agile%d", seee.seee_app);
		}
	}
	return rc;
}
