/* vi:ts=4  
 *-----------------------------------------------------------------
 *
 * Name: bsd-arpsniff.c
 *
 * RCS:  $Id: bsd-arpsniff.c,v 1.4 2002/09/20 22:10:19 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: arp sniffing to build layer2 connectivity 
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/24/02
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <nfx-ports.h>
#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-sys/errno.h"
#include "../sm-bsd-sys/macros.h"
#include "../sm-bsd-sys/socket.h"
#include "../sm-bsd-sys/syslog.h"

#include "../sm-bsd-net/if.h"
#include "../sm-bsd-net/if_types.h"

#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/in_var.h"
#include "../sm-bsd-netinet/if_ether.h"

#include "bsd-arpsniff.h"
#include "bsd-api.h"
#include "exports/ipm-api.h"
#include "exports/cm-api.h"

#if defined(ARP_SNIFFING) 

#define		ARPS_PRINTF(a)  	

arps_cfg_t	arpsCfg, *arpsCfgPtr;


/* 
 *-----------------------------------------------------------------
 *
 * Name: arpS_Init
 *
 * Description: 
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/28/2002
 *
 *-----------------------------------------------------------------
 */
void
arpS_Init()
{
	bzero(&arpsCfg, sizeof(arps_cfg_t));
	arpsCfgPtr = &arpsCfg;
	arpsCfgPtr->enable = 0;
	return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: arpS_FindEntry
 *
 * Description: Find an entry in connection table for a given IP,
 *				Entries are arranged in accending IP order.
 *				IP is in network order.
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/28/2002
 *
 *-----------------------------------------------------------------
 */
l2conn_ent_t*
arpS_FindEntry(uint32 ip)
{
	l2conn_ent_t		*connEntPtr = 0;
	
	for (connEntPtr = arpsCfgPtr->l2ConnTblHead; connEntPtr; 
		 connEntPtr = connEntPtr->l2ConnEnt_next) {
		if (ip == connEntPtr->l2ConnEnt_ip) 
			break;
	}
	
	return connEntPtr;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: arpS_AddEntry
 *
 * Description: Add an entry in connection table for a given IP,
 *				Entries are arranged in accending IP order. 
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/28/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
arpS_AddEntry(void	*inIfPtr,
			  void	*inArpPtr)
	
{
	char				ifName[ARPS_STR_SIZE];
	l2conn_ent_t		*connEntPtr, *prevConnEntPtr = 0;
	int					find = 0;
	arps_stat_t			*arpsStat = &(arpsCfgPtr->stat);
	struct ifnet		*ifPtr = (struct ifnet*)inIfPtr;
	struct ether_arp	*arpPtr= (struct ether_arp*)inArpPtr;
	struct in_ifaddr	*ia;
	uint32				srcIp, dstIp, mask;
	

	if (!arpsCfgPtr->enable)
		return NFX_ERR;
	
	/* 
	 * Sanity checks.
	 */
	if (!ifPtr || !arpPtr) {
		bsd_log(LOG_ERR, "arpS_AddEntry: empty ptr(s), intf=%lx, arp=%lx\n",
				(address_t)ifPtr, (address_t)arpPtr);
		return NFX_ERR;
	}
	
	if (ifPtr->if_unit >= IPM_LINK_MAX) {
		bsd_log(LOG_ERR, "arpS_AddEntry: bad interfaces, id=%d\n",
				ifPtr->if_unit);
		return NFX_ERR;
	}
	

	if (formPortName(NULL, bt_fp, nfxMySlotId, ifPtr->if_unit,
					 ifName)) {
		ifName[ARPS_STR_SIZE-1] ='\0';
		bsd_log(LOG_ERR, "arpS_AddEntry: bad interfaces name=%s\n",
				ifName);
		return NFX_ERR;			/* unrecognized ifName */
	}
	
	
	
	/* 
	 * filtering to ensure arp is from another nfx system
	 * Connectivity arps are always broadcast ARP request, 
	 * with target IP setto itself.
	 */
	memcpy(&srcIp, arpPtr->arp_spa, 4);
	memcpy(&dstIp, arpPtr->arp_tpa,4);
	if (((arpPtr->arp_op) != htons(ARPOP_REQUEST)) ||
		srcIp != dstIp) {
		ARPS_PRINTF(("arp not qualify %lx %lx, opt %x\n", (address_t)arpPtr->arp_spa,
					(address_t)arpPtr->arp_tpa,arpPtr->arp_op));
		return NFX_ERR;
	}

	/*
	 * Ignore entry: if an IP subnet already configured 
	 * for an interface A, and the arp belongs to this 
	 * IP subnet but received from another interface B.
	 * This is to enforce the correctness of the network
	 * configuration.
	 */
	for (ia = in_ifaddrhead.tqh_first; ia; ia = ia->ia_link.tqe_next) {
		/* reuse dstIp */
		dstIp = ia->ia_addr.sin_addr.s_addr;
		mask = ia->ia_sockmask.sin_addr.s_addr;
		if ((dstIp & mask) == (srcIp & mask)) {
			if (ia->ia_ifp != ifPtr) {
				ARPS_PRINTF(("arp %x rcv from %d, should be %d\n",
					   srcIp, ifPtr->if_unit, ia->ia_ifp->if_unit));
				return NFX_ERR;
			}
			else {
				ARPS_PRINTF(("arp %x intf matches, %d\n",
							 srcIp, ia->ia_ifp->if_unit));
				break;
			}
		}
	}
		

	/*
	 * Insert or find existing entry
	 */
	for (connEntPtr = arpsCfgPtr->l2ConnTblHead; connEntPtr; 
		 connEntPtr = connEntPtr->l2ConnEnt_next) {
		if (srcIp == connEntPtr->l2ConnEnt_ip) {
			find = 1;
			break;				/* existing entry found */
		}
		else if (srcIp < connEntPtr->l2ConnEnt_ip) {
			break;				/* insert after prevConnEntPtr */
		}
		else {
			prevConnEntPtr = connEntPtr;
			continue;
		}
	}
	
	if (find) {					/* updates the content */
		connEntPtr->l2ConnEnt_expire = ARPS_TIME_LIVE;
		/* same arp can in fact be coming from 
		 * multiple interfaces if they are 
		 * connected to the same physical network.
		 * We only keep one copy of the arp, since 
		 * the remote host is reachable through any 
		 * these interfaces*/
		memcpy(connEntPtr->l2ConnEnt_ifname, ifName, ARPS_STR_SIZE);
	}
	else {						/* adding an new entry  */
		
		connEntPtr = (l2conn_ent_t*)eee_ramAlloc(sizeof(l2conn_ent_t));
		arpsStat->alloc++;
		arpsCfgPtr->count++;
		
		if (!connEntPtr) {
			arpsStat->allocFail++;
			return NFX_ERR;
		}

		connEntPtr->l2ConnEnt_ip = srcIp;
		connEntPtr->l2ConnEnt_expire = ARPS_TIME_LIVE;
		memcpy(connEntPtr->l2ConnEnt_ifname, ifName, ARPS_STR_SIZE);

		/* insert to table */
		if (prevConnEntPtr == NULL) { /* insert at head */
			connEntPtr->l2ConnEnt_next = arpsCfgPtr->l2ConnTblHead;
			arpsCfgPtr->l2ConnTblHead = connEntPtr;
		} else {				/* insert in middle or tail */
			connEntPtr->l2ConnEnt_next = prevConnEntPtr->l2ConnEnt_next;
			prevConnEntPtr->l2ConnEnt_next = connEntPtr;
		}
	}
	
	return NFX_OK;
}


/* 
 *-----------------------------------------------------------------
 *
 * Name: arpS_Timer
 *
 * Description: remove any time expired entry from l2ConnTbl.
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/28/2001
 *
 *-----------------------------------------------------------------
 */
void arpS_Timer() 
{
	l2conn_ent_t	*connEntPtr, *prevConnEntPtr, *nxtConnEntPtr;
	struct ifnet	*ifp;
	struct ifaddr	*ifa;  
	struct arpcom	*ac;
	arps_stat_t	    *arpsStat = &(arpsCfgPtr->stat);

	/*
	 * remove time expired entries from l2 connectivity table 
	 */
	connEntPtr = arpsCfgPtr->l2ConnTblHead;
	prevConnEntPtr = 0;
	while (connEntPtr) {
		if((--connEntPtr->l2ConnEnt_expire) == 0) {	/* need to be removed */
			
			if (prevConnEntPtr == NULL) { /* remove from head */
				arpsCfgPtr->l2ConnTblHead = connEntPtr->l2ConnEnt_next;
			}
			else {				/* remove from middle or tail */
				prevConnEntPtr->l2ConnEnt_next = connEntPtr->l2ConnEnt_next;
			}

			/* free current connEnt */
			nxtConnEntPtr = connEntPtr->l2ConnEnt_next;
			eee_ramDealloc(connEntPtr);
			connEntPtr = nxtConnEntPtr;
			arpsStat->dealloc++;
			arpsCfgPtr->count--;
		}
		else {
			prevConnEntPtr = connEntPtr;
			connEntPtr = connEntPtr->l2ConnEnt_next;
		}
	}


	if (!arpsCfgPtr->enable)
		return;
	
	/* send periodical arp updates for all IP interfaces*/
	for (ifp = ifnet.tqh_first; ifp; ifp = ifp->if_link.tqe_next) {
								/* for each interface */
		ac = (struct arpcom*)ifp;
		
		for (ifa = ifp->if_addrhead.tqh_first; 
			 ifa; ifa = ifa->ifa_link.tqe_next) {
								/* for each address assigned to this interface */
			struct sockaddr_in *ip_addr;

			ip_addr = (struct sockaddr_in*)ifa->ifa_addr;
			if (ip_addr->sin_family != AF_INET)
				continue;
			
			/* broadcast arp for this IP interface */
			arprequest(ac, &IA_SIN(ifa)->sin_addr,
					   &IA_SIN(ifa)->sin_addr, ac->ac_enaddr); 
		}
	}
	return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: arpS_ShowL2ConnTbl
 *
 * Description: show connection table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/28/2001
 *
 *-----------------------------------------------------------------
 */
void
arpS_ShowL2ConnTbl()
{
	l2conn_ent_t		*connEntPtr = 0;
	
	printf("Layer2 Connection Table: Entry %d\n",arpsCfgPtr->count );
	printf("------------------------------------------------------\n");
	printf("IP                 Interface        Expire\n");
	printf("------------------------------------------------------\n");

	for (connEntPtr = arpsCfgPtr->l2ConnTblHead; connEntPtr; 
		 connEntPtr = connEntPtr->l2ConnEnt_next) 
		printf("%-19s%-17s%u\n", bsdPrintIpa(connEntPtr->l2ConnEnt_ip),
			   connEntPtr->l2ConnEnt_ifname, 
			   connEntPtr->l2ConnEnt_expire);

	return;
}

#endif /* ARP_SNIFFING */
