/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: bsd-vif.c
 *
 * RCS:  $Id: bsd-vif.c
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: virtual interface functions on txrx.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
#if !defined(PROM)
#include <nfx-incl.h>
#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-sys/errno.h"
#include "../sm-bsd-sys/macros.h"
#include "../sm-bsd-sys/mbuf.h"
#include "../sm-bsd-sys/socket.h"
#include "../sm-bsd-sys/socketvar.h"
#include "../sm-bsd-sys/sockio.h"
#include "../sm-bsd-sys/protosw.h"
#include "../sm-bsd-sys/syslog.h"

#include "../sm-bsd-net/if.h"
#include "../sm-bsd-net/if_types.h"
#include "../sm-bsd-net/route.h"
#include "../sm-bsd-net/if_dl.h"
#include "../sm-bsd-net/if_arp.h"

#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/in_var.h"
#include "../sm-bsd-netinet/in_pcb.h"
#include "../sm-bsd-netinet/ip_var.h"

#include "exports/cm-api.h"
#include "exports/bsd-api.h"
#include "exports/bsdnet-api.h"
#include "exports/malloc-api.h"
#include "exports/ipm-api.h"
#include "../sm-ipm/ipm.h"
#include "bsd-vif.h"

extern void lsi_set_mac_addr(int unit, uchar8 *mac_addr);


#if defined(LINK_AGG)
#define BSD_VIF_PRINTF(a)	printf a

#define LINK_UP(_ifp)		\
	((_ifp->if_flags & IFF_RUNNING) && (_ifp->if_flags & IFF_UP))

bsd_apiCb_t		*bsd_apiCbPtr;

/* 
 *-----------------------------------------------------------------
 *
 * Name: bsd_FindUpLinkVif
 *
 * Description: find a up link from vif.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/28/2002
 *
 *----------------------------------------------------------------
 */
static struct ifnet*
bsd_FindUpLinkVif(bsd_vif_t	*myVifp)
{
	struct ifnet*	tmpLink = NULL;
	int				i;
	
	for (i=0; i<IPM_MAXLINK_VIF; i++) {
		tmpLink = myVifp->bsd_vifLinks[i];
		if (tmpLink && LINK_UP(tmpLink))
			break;
	}
	
	return tmpLink;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: bsd_ChgeLinkMac
 *
 * Description: change Mac address of vif, link aggregation only
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
static void
bsd_ChgeLinkMac(struct ifnet *ifp, char *mac)
{
	struct ifaddr		*ifa;
	struct sockaddr_dl	*sdl;
	
	if (!ifp)
		return;
	
	memcpy(((struct arpcom*)ifp)->ac_enaddr, mac, 6);
			
	ifa = ifnet_addrs[ifp->if_index - 1];
	sdl = (struct sockaddr_dl *)ifa->ifa_addr;
	memcpy(LLADDR(sdl), mac, 6);			
	
	lsi_set_mac_addr(ifp->if_unit, mac);
	
	return;
}


/* 
 *-----------------------------------------------------------------
 *
 * Name: bsd_UpdtVifMac
 *
 * Description: change Mac address of vif, link aggregation only
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
static void
bsd_UpdtVifMac(bsd_vif_t *vifp, char	*mac)
{
	struct ifnet		*tmpLink;
	int					i;
	
	if ((!vifp) || !(vifp->bsd_vifFlags & IPM_VIF_MODE_LA))
		return;

	/* copy vif mac */
	memcpy(vifp->bsd_vifMac, mac, 6);

	for (i=0; i<IPM_MAXLINK_VIF; i++) {
		tmpLink = vifp->bsd_vifLinks[i];
		bsd_ChgeLinkMac(tmpLink, vifp->bsd_vifMac);
		
	}
	return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: bsd_InitApi
 *
 * Description: Init all global varibles.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
void
bsd_InitApi()
{
	bsd_apiCbPtr = eee_ramAlloc(sizeof(bsd_apiCb_t));
	VifHead = NULL;
	VifStats = (bsd_vifStats_t*)eee_ramAlloc(sizeof(bsd_vifStats_t));
	bzero(VifStats, sizeof(*VifStats));
	return;
}



/*-----------------------------------------------------------------
 *
 * Name: bsd_FindVifByName
 *
 * Description: find vif.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
bsd_vif_t*
bsd_FindVifByName(char	*vifName) 
{
	bsd_vif_t	*vifp;
	
	vifp = VifHead;
    while (vifp) {
		if (strcmp(vifp->bsd_vifName, vifName) == 0)
			return vifp;
		vifp = vifp->bsd_vifNxt;
	}
	return NULL;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_VifLAGetActLink
 *
 * Description: this function retunes the active link (i.e where
 *				ip address is assigned.) It only apply to link
 *				aggregation vif.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
void*
bsd_VifLAGetUpLink(void	*in_ifp)
{
	struct ifnet*	ifp = (struct ifnet*)in_ifp;
	bsd_vif_t		*myVifp;
	
	myVifp = ifp->if_vif;
	if (myVifp && (myVifp->bsd_vifFlags & IPM_VIF_MODE_LA))
		return myVifp->bsd_vifUpLink;

	return NULL;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_VifLAGetActLink
 *
 * Description: this function retunes the active link (i.e where
 *				ip address is assigned.) It only apply to link
 *				aggregation vif.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
void*
bsd_VifLAGetActLink(void	*in_ifp)
{
	struct ifnet*	ifp = (struct ifnet*)in_ifp;
	bsd_vif_t		*myVifp;
	
	myVifp = ifp->if_vif;
	if (myVifp && (myVifp->bsd_vifFlags & IPM_VIF_MODE_LA))
		return myVifp->bsd_vifActLink;

	return NULL;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_IsLinkVifLA
 *
 * Description: is the link part of link aggregation vif. 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/28/2002
 *
 *-----------------------------------------------------------------
 */
int32 					/*  */
bsd_IsLinkVifLA (void	*in_ifp)
{
	struct ifnet*	ifp = (struct ifnet*)in_ifp;
	bsd_vif_t		*myVifp;
	
	if (!ifp)
		return 0;
	
	myVifp = ifp->if_vif;
	if (myVifp && (myVifp->bsd_vifFlags & IPM_VIF_MODE_LA))
		return 1;
	else
		return 0;
}


/*-----------------------------------------------------------------
 *
 * Name: bsd_FindVifByLink
 *
 * Description: find vif by given link
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
	
bsd_vif_t*
bsd_FindVifByLink(void	*in_ifp) 
{
	struct	ifnet*	ifp = (struct ifnet*)in_ifp;

	return (bsd_vif_t*)ifp->if_vif;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_VifHasIP
 *
 * Description: does interface has IP. If has, return link
 *				that has the IP.
 *				
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
void*							/* ptr to link or NULL */
bsd_VifHasIP(uint32 ip, bsd_vif_t *vifp)
{
	struct ifnet	*myLink;
	
	if (!vifp)
		return NULL;

	/* find link has given ip */
	if ((myLink = bsd_FindIntfByIp(ip)) == NULL)
		return NULL;
	
	/* does link belongs to vif */
	if (vifp == bsd_FindVifByLink(myLink))
		return myLink;

	return NULL;
}

typedef struct tmp_ip 
{
	uint32	ip;
	uint32	mask;
	uint32	nhop;
	struct tmp_ip	*next;
} tmp_ip_t;

static tmp_ip_t *rtStart, *ipStart;

static int tmp_QueRt(struct radix_node *r_node, void* ifp)
{
	
	struct rtentry* rt = (struct rtentry*)r_node;
	tmp_ip_t		*tmpRtPtr;

	if ((rt->rt_ifp == ifp) && 
		(rt->rt_gateway) && 
		(rt->rt_gateway->sa_family == AF_INET)) {
		
		tmpRtPtr = eee_ramAlloc(sizeof(tmp_ip_t));

		
		if (!tmpRtPtr) {
			VifStats->vif_allocErr ++;
			return 0;
		}

		VifStats->vif_alloc++;
		
		tmpRtPtr->ip = ((struct sockaddr_in*)rt_key(rt))->sin_addr.s_addr;
		tmpRtPtr->mask = ((struct sockaddr_in*)rt_mask(rt))->sin_addr.s_addr;
		tmpRtPtr->nhop = ((struct sockaddr_in*)rt->rt_gateway)->sin_addr.s_addr;

		/* link */
		if (rtStart) {
			tmpRtPtr->next = rtStart;
			rtStart = tmpRtPtr;
		}
		else {
			rtStart = tmpRtPtr;
			tmpRtPtr->next = 0;
		}
	}
	
	return  0;
}
				 
/*-----------------------------------------------------------------
 *
 * Name: bsd_SwitchLinkIP
 *
 * Description: switch ip addresses from one link to the other.

 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
void
bsd_SwitchLinkIP(void *in_srcLink,  void *in_dstLink)
{
	struct ifnet	*srcLink = (struct ifnet*)in_srcLink;
	struct ifnet	*dstLink = (struct ifnet*)in_dstLink;
	tmp_ip_t		*tmpIpPtr, *tmpRtPtr;
	struct ifaddr			*ifa;
	struct sockaddr_in		*ip_addr, *ip_mask;
	struct radix_node_head	*rnh;
	
	rtStart = NULL, ipStart=NULL;

	/* dskLink exist */
	if ((!srcLink) || (!dstLink))
		return;
	
	/* gather IP info */
	for (ifa = srcLink->if_addrhead.tqh_first; ifa;
		 ifa = ifa->ifa_link.tqe_next) {

		ip_addr = (struct sockaddr_in*)ifa->ifa_addr;
		ip_mask = (struct sockaddr_in*)ifa->ifa_netmask;

		if (ip_addr->sin_family == AF_INET) {
			tmpIpPtr = eee_ramAlloc(sizeof(tmp_ip_t));
			
			if (!tmpIpPtr) {
				VifStats->vif_allocErr++;
				goto out;
			}
			VifStats->vif_alloc++;
			
			tmpIpPtr->ip = ip_addr->sin_addr.s_addr;
			tmpIpPtr->mask = ip_mask->sin_addr.s_addr;
			
			/* link */
			if (ipStart) {
				tmpIpPtr->next = ipStart;
				ipStart = tmpIpPtr;
			}
			else {
				ipStart = tmpIpPtr;
				tmpIpPtr->next = 0;
			}
		}
	}
	
	/* IP switch interfaces */
	bsd_DeleteIpAll(srcLink);
	for (tmpIpPtr = ipStart; tmpIpPtr; tmpIpPtr = tmpIpPtr->next) 
		bsd_AddIpByIntf(tmpIpPtr->ip, tmpIpPtr->mask, dstLink);


	/* gather route info */
	if ((rnh = rt_tables[AF_INET]) == 0) 
		goto out;
	rnh->rnh_walktree(rnh, tmp_QueRt, srcLink);
	
	
	/* rt switch interface */
	for (tmpRtPtr = rtStart; tmpRtPtr; tmpRtPtr = tmpRtPtr->next) {
		bsd_ModRoute(tmpRtPtr->ip, tmpRtPtr->mask, 
					 tmpRtPtr->nhop, 0, BSD_DEL, 0, 0);
		bsd_ModRoute(tmpRtPtr->ip, tmpRtPtr->mask, 
					 tmpRtPtr->nhop, 0, BSD_ADD, 0, 0);
	}

out:
	/* free chained nodes */
	while ((tmpIpPtr = ipStart)) {
		ipStart = tmpIpPtr->next;
		tmpIpPtr->next = 0;
		VifStats->vif_dealloc++;
		eee_ramDealloc(tmpIpPtr);
	}

	while ((tmpRtPtr = rtStart)) {
		rtStart = tmpRtPtr->next;
		tmpRtPtr->next = 0;
		VifStats->vif_dealloc++;
		eee_ramDealloc(tmpRtPtr);
	}
	
	return;
}

/*-----------------------------------------------------------------
 * 
 * Name: bsd_DestoryVif
 *
 * Descrition:	destory everything related to vif, including ip
 *				assigned.
 *
 * Created By: Su Wang
 *
 * Date Created: 10/18/2002
 *
 *-----------------------------------------------------------------
 */
void
bsd_DestoryVif(char	*vifName, void	*linkList, int count)
{
	struct ifnet			**linkPtr;
	int						i;
	struct radix_node_head	*rnh;
	tmp_ip_t				*tmpRtPtr;
	bsd_vif_t				*myVifp;


	linkPtr = (struct ifnet**)linkList;
	for (i = 0; i< count; i++) {
		
		/* remove all interface associated
		 * with link
		 */
		bsd_DeleteIpAll(*linkPtr);
		
		/* remove all rt assoiciated with
		 * link
		 */
		rnh = rt_tables[AF_INET];
		rnh->rnh_walktree(rnh, tmp_QueRt, *linkPtr);
	
		
		for (tmpRtPtr = rtStart; tmpRtPtr; tmpRtPtr = tmpRtPtr->next) {
			bsd_ModRoute(tmpRtPtr->ip, tmpRtPtr->mask, 
						 tmpRtPtr->nhop, 0, BSD_DEL, 0, 0);
		}

		while ((tmpRtPtr = rtStart)) {
			rtStart = tmpRtPtr->next;
			tmpRtPtr->next = 0;
			VifStats->vif_dealloc++;
			eee_ramDealloc(tmpRtPtr);
		}
		
		linkPtr ++;
	}

	myVifp = bsd_FindVifByName(vifName);
	if (myVifp) {
		myVifp->bsd_vifIpCnt = 0;
		bsd_DelVif(vifName);
	}
	
	return;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_AddLinkToVif
 *
 * Description: add specified link to vif
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK, NFX_ERR */
bsd_AddLinkToVif(void* in_ifp, bsd_vif_t *vifp) 
{

	bsd_vif_t		*myVifp;
	int				i, myLink;
	
	if (!in_ifp || !vifp)
		return NFX_ERR;
	
	/* does this link used by other virtual interface */
	myVifp = bsd_FindVifByLink(in_ifp);
	
	if (myVifp) {
		if (myVifp == vifp)		{/* link belongs to virtual interface already */
			bsd_log(BSD_WARN_INTF, "%s: link %s already in vif %s\n",
					__FUNCTION__, bsdPrintIntf(in_ifp), myVifp->bsd_vifName);
			return NFX_OK;
		}
		else	{				/* link belongs to another virtual interface */
			bsd_log(BSD_ERR_INTF, "%s: link %s error in vif %s\n",
					__FUNCTION__, bsdPrintIntf(in_ifp), myVifp->bsd_vifName);
			return NFX_ERR;
		}
	}
  
	myVifp = vifp;
	
	/* does this link has IP assigned to it */
	if (bsd_FindIpByIntf(in_ifp)) {
		bsd_log(BSD_ERR_INTF, "%s: link %s has ip assigned\n",
				__FUNCTION__, bsdPrintIntf(in_ifp));
		return NFX_ERR;
	}
	
	if (vifp->bsd_vifLinkCnt >= IPM_MAXLINK_VIF) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s has exceeded linkcnt %d\n",
				__FUNCTION__, myVifp->bsd_vifName);
		return NFX_ERR;
	}

	/* find a slot */
	myLink = -1;
	for (i=0; i < IPM_MAXLINK_VIF; i++) {
		if (myVifp->bsd_vifLinks[i] == 0) {
			myLink = i;
			break;
		}
	}

	/* something bad happend,vif_links and vif_linkCnt don't agree */
	if (myLink == -1 ) {
		bsd_log(BSD_ERR_INTF, "%s: vif linkcnt %d mismatch content\n",
				__FUNCTION__, myVifp->bsd_vifLinkCnt);
		return NFX_ERR;
	}

    /* establish link */
	myVifp->bsd_vifLinks[myLink] = in_ifp;
	((struct ifnet*)in_ifp)->if_vif = (void*)myVifp;
	myVifp->bsd_vifLinkCnt++;

#if 0
	/* first link by default will be prefered link */
	if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_FO) &&
		(myVifp->bsd_vifLinkCnt == 1)) {
		myVifp->bsd_vifPrefLink = in_ifp;
		myVifp->bsd_vifActLink = in_ifp;
	}
#endif /* 0 */
	return NFX_OK;
}


/*-----------------------------------------------------------------
 *
 * Name: bsd_RmLinkFromVif
 *
 * Description: remove specified link from vif
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_RmLinkFromVif(void	*in_ifp, bsd_vif_t *vifp)
{
	int		i;
	
	if ((!in_ifp) || (!vifp)) 
		return NFX_ERR;
	
	for (i=0; i<IPM_MAXLINK_VIF; i++) {
		if (vifp->bsd_vifLinks[i] == in_ifp) {
			vifp->bsd_vifLinks[i] = 0;
			((struct ifnet*)in_ifp)->if_vif = 0;
			vifp->bsd_vifLinkCnt--;
			return NFX_OK;
		}
	}
	return NFX_ERR;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_AddVif
 *
 * Description: create a VIF
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_AddVif(char	*vifName, uint32 flags, void *linkList, int count)
{
	struct ifnet	**linkPtr;
	bsd_vif_t		*myVifp = NULL;
	int32			rc = NFX_OK;
	int				i;

	/* destory existing vif of same name
	 * this happens during mcp restart (restoring config).
	 * any other case, ipmd would have detectd duplication
	 * and return to application without forwarding to txrx.
	 */
	bsd_DestoryVif(vifName, linkList, count);
	
	if (bsd_FindVifByName(vifName)) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s already exist\n",
				__FUNCTION__, vifName);
		rc =  NFX_ERR;
		goto out;
	}

	if (VifStats->vif_cnt >= IPM_MAX_VIF) {
		bsd_log(BSD_ERR_INTF, "%s: number reaches maximum\n",
				__FUNCTION__);
		rc =  NFX_ERR;
		goto out;
	}
	
	if ((myVifp = eee_ramAlloc(sizeof(bsd_vif_t))) == NULL) {
		bsd_log(BSD_ERR_INTF, "%s: vif alloc failed", __FUNCTION__);
		rc =  NFX_ERR;
		goto out;
	}
	bzero(myVifp, sizeof(*myVifp));
	
	/* add link to virtual interface structure,
	 * any failure to add link will cause entire
	 * operation to fail.
	 */
	linkPtr = (struct ifnet**)linkList;
	for (i = 0; i< count; i++) {
		rc = bsd_AddLinkToVif((*linkPtr), myVifp);
		if (rc != NFX_OK)
			goto out;
		
		linkPtr ++;
	}
	
	/* copy link and flags, vif created on txrx */
	memcpy(myVifp->bsd_vifName, vifName, IPM_STR_SIZE);
	myVifp->bsd_vifFlags = flags;

out:
	if (rc != NFX_OK) {			/* free */
		if (myVifp) 
			eee_ramDealloc(myVifp);
		VifStats->vif_addErr ++;
	}
	else	{					/* link */
		myVifp->bsd_vifNxt = VifHead;
		VifHead = myVifp;
		VifStats->vif_cnt ++;

		/* link aggregation, find the active link,
		 * and set mac
		 */
		if (myVifp->bsd_vifFlags & IPM_VIF_MODE_LA) {

			/* update active link and mac */
			myVifp->bsd_vifActLink = myVifp->bsd_vifLinks[0];
			bsd_UpdtVifMac(myVifp,
						   ((struct arpcom*)myVifp->bsd_vifActLink)->ac_enaddr);

			/* find a up Link */
			myVifp->bsd_vifUpLink = bsd_FindUpLinkVif(myVifp);
			if (!myVifp->bsd_vifUpLink)
				myVifp->bsd_vifUpLink = myVifp->bsd_vifActLink;
		}
	}
		
	return rc;
}


/*-----------------------------------------------------------------
 *
 * Name: bsd_DelVif
 *
 * Description: delete a VIF
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_DelVif(char *vifName)
{
	bsd_vif_t	*myVifp, *tmpVifp;
	int			i;
	int32		rc = NFX_OK;
	struct ifnet *tmpLink;

	/* no virtual interface */
	if ((myVifp = bsd_FindVifByName(vifName)) == NULL) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not exist\n",
				__FUNCTION__, vifName);
		rc = NFX_ERR;
		goto out;
	}

	/* make sure there is no IP associated with it */
	if (myVifp->bsd_vifIpCnt) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s still has ip, cnt %d\n",
				__FUNCTION__, vifName, myVifp->bsd_vifIpCnt);
		rc =  NFX_ERR;
		goto out;
	}
		
	/* clear all links */
	for (i = 0; i < IPM_MAXLINK_VIF; i++) {
		tmpLink = myVifp->bsd_vifLinks[i];
		bsd_RmLinkFromVif(tmpLink, myVifp);

		/* link aggregation, reset to old mac */
		if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_LA))
			bsd_ChgeLinkMac(tmpLink, 
							bsd_GetIntfMac(tmpLink));
	}
	
	/* unlink and free */
	tmpVifp = VifHead;
	if (tmpVifp == myVifp) {	/* first */
		VifHead = myVifp->bsd_vifNxt;
		myVifp->bsd_vifNxt = 0;
	}
	else {
		while (tmpVifp) {
			if (tmpVifp->bsd_vifNxt == myVifp)
				break;
		}

		tmpVifp->bsd_vifNxt = myVifp->bsd_vifNxt;
		myVifp->bsd_vifNxt = 0;
	}
	
	eee_ramDealloc(myVifp);
	VifStats->vif_cnt --;

 out:
	if (rc != NFX_OK)
		VifStats->vif_delErr ++;
	
	return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_PrefLinkVif
 *
 * Description: define prefered link in a virtual interface.
 *	
 *	1) same prefLink, do nothing.
 *
 *  2) a)	change prefered link
 *	   b)   if prefLink is active, change actLink,
 *	   c)	if preLink and actLink both inactive, change actLink
 *	   d)	if actLink not changed, let ssc know.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_ERR, NFX_OK */
bsd_PrefLinkVif(char *vifName, void *in_ifp)
{
	bsd_vif_t		*myVifp;
	struct ifnet	*ifp = (struct ifnet*)in_ifp;
	int32			rc = NFX_OK;
	
	if (!ifp) {
		rc = NFX_ERR;
		goto out;
	}
	
	/* make sure link and virtual interface matches */
	myVifp = bsd_FindVifByLink(ifp);
	if (!myVifp || strcmp(myVifp->bsd_vifName, vifName)) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not have link %s\n",
				__FUNCTION__, vifName, bsdPrintIntf(ifp));
		
		rc =  NFX_ERR;
		goto out;
	}

	/* only apply to fail-over links */
	if (!(myVifp->bsd_vifFlags & IPM_VIF_MODE_FO)) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s has bad mode %x\n",
				__FUNCTION__, vifName, myVifp->bsd_vifFlags);
		rc = NFX_ERR;
		goto out;
	}
		

	/* no need to change prefLink changed */
	if ((ifp == myVifp->bsd_vifPrefLink) &&
		(ifp == myVifp->bsd_vifActLink))
		return NFX_OK;
	
	myVifp->bsd_vifPrefLink = ifp;
	
	/* vif is intialiazing, give prefered link 
	 * to act link, if it is up; else
	 * need to find an active link
	 */
	if (!(myVifp->bsd_vifActLink)) {
		if (LINK_UP(ifp)) {
			myVifp->bsd_vifActLink = ifp;
		}
		else {
			/* find any active link */
			int				i;
			struct ifnet	*tmpLink = NULL;
			for (i=0; i<IPM_MAXLINK_VIF; i++) {
				tmpLink = myVifp->bsd_vifLinks[i];
				if (tmpLink && LINK_UP(tmpLink))
					break;
			}
			
			/* there is up link, assign as active link
			 * else use prefered link
			 */
			if (tmpLink)
				myVifp->bsd_vifActLink = tmpLink;
			else
				myVifp->bsd_vifActLink = myVifp->bsd_vifPrefLink;
		}
	}
	
	
	/* need to change actLink */
	if (ifp != myVifp->bsd_vifActLink) {
		if (LINK_UP(ifp) ||
			((!LINK_UP(ifp)) && (!LINK_UP(myVifp->bsd_vifActLink)))) {
			/* change active link, if prefered link is up
			 * or current active link is down
			 */
			bsd_SwitchLinkIP(myVifp->bsd_vifActLink, ifp);
			myVifp->bsd_vifActLink = ifp;
		}
		else	{
			/* act has not changed, ssc need to know,
			 * This is mimic the sequency where prefLink
			 * at first become active and become inactive
			 */
			ipm_mvif_t	vifMsg;

			bzero(&vifMsg, sizeof(ipm_mvif_t));
			memcpy(vifMsg.vif_name, myVifp->bsd_vifName, IPM_STR_SIZE);
			sprintf(vifMsg.vif_prefLink, "%s%d.%d", 
					ifp->if_name, nfxMySlotId, ifp->if_unit);
			sprintf(vifMsg.vif_actLink, "%s%d.%d", 
					myVifp->bsd_vifActLink->if_name, nfxMySlotId, 
					myVifp->bsd_vifActLink->if_unit);
			ipm_SendMsgByApp(IPM_MSG_VIF_EVNT, nfxSscSlotId, 
							 NFX_SSC_UNIX, &vifMsg, sizeof(vifMsg), 0);
		}
	}
	
out:
	if (rc != NFX_OK)
		VifStats->vif_prefErr ++;
	
	return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_AddVifLink
 *
 * Description: add link to vif.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_AddVifLink(char	*vifName, void*	 linkStart, int count)
{
	bsd_vif_t		*myVifp, *tmpVifp;
	struct ifnet	**linkPtr;
	int				i;
    int32			rc = NFX_OK;
	
	/* no virtual interface */
	if ((myVifp = bsd_FindVifByName(vifName)) == NULL) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not exist\n",
				__FUNCTION__, vifName);
		rc = NFX_ERR;
		goto out;
	}

	/* err checking, make sure link does not belong to
	 * other interface.
	 */
	linkPtr = (struct ifnet**)linkStart;
	for (i=0; i < count; i++) {
		tmpVifp = bsd_FindVifByLink((*linkPtr));
		if (tmpVifp && (tmpVifp != myVifp)) {
			rc = NFX_ERR;
			goto out;
		}
		linkPtr ++;
	}

	/* add link to virtual interface,
	 */
	linkPtr = (struct ifnet**)linkStart;
	for (i=0; i < count; i++) {
		bsd_AddLinkToVif((*linkPtr), myVifp);
		
		/* change mac address, if link aggregation */
		if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_LA))
			bsd_ChgeLinkMac((*linkPtr), myVifp->bsd_vifMac);
		
		linkPtr ++;
	}
	
out:
	if (rc != NFX_OK)
		VifStats->vif_addLinkErr++;
	
	return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_RmVifLink
 *
 * Description: remove  links from vif.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_RmVifLink(char *vifName, void*	 linkStart, int count)
{
	bsd_vif_t		*myVifp;
	struct ifnet	**linkPtr;
	int				i;
	int32			rc = NFX_OK;
	
	/* no virtual interface */
	if ((myVifp = bsd_FindVifByName(vifName)) == NULL) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not exist\n",
				__FUNCTION__, vifName);
		rc =  NFX_ERR;
		goto out;
	}

	/* for fail-over, removed link cannot be 
	 * active link
	 */
	linkPtr = (struct ifnet**)linkStart;
	for (i=0; i < count; i++) {
		if ((*linkPtr) == myVifp->bsd_vifActLink) {
			bsd_log(BSD_ERR_INTF, "%s: vif %s attempts to remvoe active link\n",
					__FUNCTION__, vifName);
			rc = NFX_ERR;
			goto out;
		}
		linkPtr ++;
	}
	

	/* remove links from virtual interface */
	linkPtr = (struct ifnet**)linkStart;
	for (i=0; i < count; i++) {
		bsd_RmLinkFromVif((*linkPtr), myVifp);

		/* link aggregation, reset to old mac */
		if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_LA)) {
			bsd_ChgeLinkMac((*linkPtr), 
							bsd_GetIntfMac(*linkPtr));
			
			/* if deleted link is up link,
			 * elect a new up link
			 */
			if (*linkPtr == myVifp->bsd_vifUpLink) {
				myVifp->bsd_vifUpLink = bsd_FindUpLinkVif(myVifp);
				if (!myVifp->bsd_vifUpLink)
					myVifp->bsd_vifUpLink = myVifp->bsd_vifActLink;
			}
		}
		
		/* if prefLink removed, assign new prefLink to actLink */
		if (myVifp->bsd_vifPrefLink == *linkPtr)
			myVifp->bsd_vifPrefLink = myVifp->bsd_vifActLink;
		linkPtr ++;
	}
	
out:
	if (rc != NFX_OK)
		VifStats->vif_delLinkErr ++;
	
	return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_AddVifIP
 *
 * Description: assign IP to virtual link
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_AddVifIP(char *vifName, uint32 ip, uint32 mask)
{
	bsd_vif_t		*myVifp;
	struct ifnet	*myLink;
	int32			rc = NFX_OK;

	/* vif exist and has links */
	if ((myVifp=bsd_FindVifByName(vifName)) == NULL) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not exist\n",
				__FUNCTION__, vifName);
		
		rc = NFX_ERR;
		goto out;
	}
	if (!myVifp->bsd_vifLinkCnt) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s has no link\n",
				__FUNCTION__, vifName);
		rc = NFX_ERR;
		goto out;
	}
	
	/* chck for dup ip */
	if (bsd_FindIntfByIp(ip)) {
		if (bsd_VifHasIP(ip, myVifp) == NULL) {
			bsd_log(BSD_ERR_INTF, "%s: ip %x already exist\n",
					__FUNCTION__, ip);
			rc = NFX_ERR;
			goto out;
		}

		/* vif already has the interface */
		return NFX_OK;
	}
	
	
	/* add IP address to the active link,
	 */
	myLink = myVifp->bsd_vifActLink;

	if (!myLink) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not has an active link\n",
				__FUNCTION__, vifName);
		rc = NFX_ERR;
		goto out;
	}	
	
	/* add  now */
	bsd_AddIpByIntf(ip, mask, myLink);
	myVifp->bsd_vifIpCnt++;

out:
	if (rc != NFX_OK)
		VifStats->vif_addIpErr++;
	
	return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_DelVifIP
 *
 * Description: delete IP to virtual link
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_DelVifIP(char *vifName, uint32 ip)
{
	bsd_vif_t		*myVifp;
	struct ifnet	*myLink = NULL;
	int32			rc = NFX_OK;

	/* vif exist  */
	if ((myVifp = bsd_FindVifByName(vifName)) == NULL) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not exist\n",
				__FUNCTION__, vifName);
		rc = NFX_ERR;
		goto out;
	}

	/* ip belongs to this virtual interface */
	if ((myLink =bsd_VifHasIP(ip, myVifp)) == NULL) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s does not has ip %x\n",
				__FUNCTION__, vifName, ip);
		rc = NFX_ERR;
		goto out;
	}

	/* fail over, ip should be on active link;
	 * if not, something is wrong. 
	 */
	if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_FO) &&
		myLink != myVifp->bsd_vifActLink) {
		bsd_log(BSD_ERR_INTF, "%s: vif %s ip %x not on active link %s\n",
				__FUNCTION__, vifName, ip, bsdPrintIntf(myLink));
		rc = NFX_ERR;
		goto out;
	}
	
	bsd_DeleteIpByIntf(ip, myLink);
	myVifp->bsd_vifIpCnt--;

out:
	if (rc != NFX_OK)
		VifStats->vif_delIpErr ++;
	
	return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_vifLinkChange
 *
 * Description: link state change has occured.
 *				the code handles fail-over case only now.
 *
 * Fail-Over
 * active link changes if:
 *		1: link is up and there is no active link.
 *		2: act link is down and there is active link ready to 
 *		   to taken over. 
 *
 * Link-Aggre:
 * Up link changes if:
 *		1. old up link is down and incoming link is up
 *		2. incoming link is down and its old up link
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
void
bsd_VifLinkChange(void*	in_ifp)
{
	struct ifnet	*ifp = (struct ifnet*)in_ifp;
	bsd_vif_t		*myVifp;
	struct ifnet	*oldActLink = NULL, *newActLink = NULL;
	ipm_mvif_t		vifMsg;
	
	
	/* find the vif that link belongs */
	if ((myVifp = bsd_FindVifByLink(ifp)) == NULL)
		return;

	/* link change state does not affect active link
	 * for link aggregation. The vif uses the same 
	 * active link and its mac for its entire
	 * exsitence
	 */
	if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_LA)) {
		if (!(LINK_UP(myVifp->bsd_vifUpLink)) && (LINK_UP(ifp))) {
			myVifp->bsd_vifUpLink = ifp;
			VifStats->vif_upLinkChange++;
		}
		else if (!LINK_UP(ifp) && (ifp == myVifp->bsd_vifUpLink)) {
			if (!(myVifp->bsd_vifUpLink = bsd_FindUpLinkVif(myVifp)))
				myVifp->bsd_vifUpLink = ifp;
			else
				VifStats->vif_upLinkChange++;
		}
		return;
	}
	
	/* prefered link has not been initilized in fail-over mode*/
	if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_FO) &&
		!(myVifp->bsd_vifPrefLink))
		return;
	
#if 0
	if (LINK_UP(ifp) && (myVifp->bsd_vifPrefLink == ifp)) 	{
		newActLink = myVifp->bsd_vifPrefLink;
		oldActLink = myVifp->bsd_vifActLink;
	}
	else
#endif /* 0 */ 
	if (LINK_UP(ifp) && (!LINK_UP(myVifp->bsd_vifActLink))) {
		newActLink = ifp;
		oldActLink = myVifp->bsd_vifActLink;
	}
	else if (!LINK_UP(ifp) && (myVifp->bsd_vifActLink == ifp)) {
		if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_FO) &&
			(LINK_UP(myVifp->bsd_vifPrefLink))) {
			/* is prefered link up */
			oldActLink = myVifp->bsd_vifActLink;
			newActLink = myVifp->bsd_vifPrefLink;
		}
		else {
			/* find any active link */
			int				i;
			struct ifnet	*tmpLink;
			for (i=0; i<IPM_MAXLINK_VIF; i++) {
				tmpLink = myVifp->bsd_vifLinks[i];
				if (tmpLink && LINK_UP(tmpLink)) {
					newActLink = tmpLink;
					oldActLink = myVifp->bsd_vifActLink;
					break;
				}
			}
		}
	}
	

	/* need to change active link */
	if (oldActLink != newActLink) {
		
		/* update local changes */
		myVifp->bsd_vifActLink = newActLink;
		bsd_SwitchLinkIP(oldActLink, newActLink);
		
		/* send update msg to ssc */
		bzero(&vifMsg, sizeof(ipm_mvif_t));
		memcpy(vifMsg.vif_name, myVifp->bsd_vifName, IPM_STR_SIZE);
		sprintf(vifMsg.vif_prefLink, "%s%d.%d", 
				oldActLink->if_name, nfxMySlotId, oldActLink->if_unit);
		sprintf(vifMsg.vif_actLink, "%s%d.%d", 
				newActLink->if_name, nfxMySlotId, newActLink->if_unit);
		ipm_SendMsgByApp(IPM_MSG_VIF_EVNT, nfxSscSlotId,
						 NFX_SSC_UNIX, &vifMsg, sizeof(vifMsg), 0);

		VifStats->vif_actLinkChange ++;
	}
	return;
}

/*-----------------------------------------------------------------
 *
 * Name: bsd_PrintVif
 *
 * Description: print virtual link
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/10/2002
 *
 *-----------------------------------------------------------------
 */
void
bsd_PrintVif(void*	in_vif)
{
	bsd_vif_t		*myVifp = (bsd_vif_t*)in_vif;
	int				i;
	struct ifnet	*tmpLink;
	
	if (!myVifp)
		return;
	
	printf("Virtual Interface %s:\n", myVifp->bsd_vifName);
	
	printf("Current Links (%d): ", myVifp->bsd_vifLinkCnt);
	for (i = 0; i < IPM_MAXLINK_VIF; i++ ) {
		tmpLink = myVifp->bsd_vifLinks[i];
		if (tmpLink) 
			printf("%s ", bsdPrintIntf(tmpLink));
		
	}
	printf("\n");
	
	printf("Flags %x, IP count %d\n", 
		   myVifp->bsd_vifFlags, myVifp->bsd_vifIpCnt);
	if ((myVifp->bsd_vifFlags & IPM_VIF_MODE_FO)) {
		printf("Prefered Link %s,", 
			   myVifp->bsd_vifPrefLink ? bsdPrintIntf(myVifp->bsd_vifPrefLink): "NULL");
		printf("Active Link %s\n", 
			   myVifp->bsd_vifActLink ? bsdPrintIntf(myVifp->bsd_vifActLink): "NULL");
	}
	else {
		printf("Up Link %s,", 
			   myVifp->bsd_vifUpLink ? bsdPrintIntf(myVifp->bsd_vifUpLink): "NULL");
		printf("Active Link %s\n", 
			   myVifp->bsd_vifActLink ? bsdPrintIntf(myVifp->bsd_vifActLink): "NULL");
		printf("Mac address: %s\n", bsdPrintMac(myVifp->bsd_vifMac));
	}
	printf("\n");
	
	return;
}

#endif /* LINK_AGG */

#endif /* PROM */
