/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	bsdippkt-api.c
 *
 * RCS:  $Id: bsdippkt-api.c,v 1.20 2002/10/30 22:18:28 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Contains all data path APIs to BSD TCP/IP.
 *
 * Created by:	Su Wang
 *
 * Date Created:	7/25/01
 *
 *-----------------------------------------------------------------
 */
 
#include <nfx-incl.h>
#include "exports/pkt-api.h"
#include "exports/bsdnet-api.h"
#include "exports/queue-api.h"
#include "exports/timer-api.h"
#include "exports/cm-api.h"

#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-sys/errno.h"
#include "../sm-bsd-sys/macros.h"
#include "../sm-bsd-sys/mbuf.h"
#include "../sm-bsd-sys/socket.h"
#include "../sm-bsd-sys/socketvar.h"
#include "../sm-bsd-sys/sockio.h"
#include "../sm-bsd-sys/protosw.h"

#include "../sm-bsd-net/if.h"
#include "../sm-bsd-net/route.h"
#include "../sm-bsd-net/if_dl.h"
#include "../sm-bsd-net/ethernet.h"
#include "../sm-bsd-net/if_types.h"

#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/if_ether.h"
#include "../sm-bsd-netinet/ip.h"
#include "../sm-bsd-netinet/udp.h"
#include "../sm-bsd-netinet/tcp.h"
#include "../sm-bsd-netinet/in_var.h"
#include "../sm-bsd-netinet/in_pcb.h"
#include "../sm-bsd-netinet/ip_var.h"
#include "../sm-bsd-netinet/ip_icmp.h"
#include "../sm-bsd-netinet/icmp_var.h"
#include "bsdrelay.h"
#include "bsd-arpsniff.h"
#include "exports/luc-api.h"
#include "bsd-vif.h"

#define senderr(e) { error = (e); goto bad;}

int32 pkt_xmtSlowPath(void *, pkt_desc_t *, char *, short );
void  in_arpinput(struct mbuf *);

// #define REDIRECT_IP	1


/*
 *-----------------------------------------------------------------------
 *
 * Name:   bsd_XmtPkt
 *
 * Description: convert IP/ARP in mbuf to pkt descriptor. It also initiate an
 *              arp request if dl addr not avaiable. Any IP packet without
 *              valid dl addr is droped by this function. 
 *              It is called by ip_output from bsd TCP/IP stack, and in 
 *              turn calls pkt_xmtSlowPath to add layer2 header. 
 *              Note: make sure to BSD compatible error code.
 * 
 * Created by : Su Wang
 *
 * Date Created:  8/6/01
 *------------------------------------------------------------------------
 */
int32 
bsd_XmtPkt(  struct ifnet		*ifp,	
								/* ptr to interface */
			 struct mbuf		*m,	
								/* ip packet */
			 struct sockaddr	*dst,  
								/*	nexthop address, is alway an IP */
			 struct rtentry		*rt0   
								/* dst route, can be direct or indirect */
								/* is 0 for ARP request/reply */
			 )
{
  short type;
  int   error = 0;
  u_char edst[6];
  register struct rtentry *rt;
  struct arpcom *ac = (struct arpcom *)ifp;
  char* nh_mac; 
  int len;
  pkt_desc_t    *pdesc;
  struct ip     *ipHdr;

#if defined(LINK_AGG)
  struct ifnet	*vifIfp;
  
  /* For link aggregation vif, change to up link. As active link
   * maybe down
   */
  vifIfp = (struct ifnet*)bsd_VifLAGetUpLink(ifp);
  if (vifIfp)
	  ifp = vifIfp;
#endif /* LINK_AGG */
  
  if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
    senderr(ENETDOWN);

  rt = rt0;
  if (rt) 
    {
      if ((rt->rt_flags & RTF_UP) == 0) 
		{
		  rt0 = rt = rtalloc1(dst, 1, 0UL);
		  if (rt0)
			rt->rt_refcnt--;
		  else
			senderr(EHOSTUNREACH);
		}

      if (rt->rt_flags & RTF_GATEWAY) 
		{
		  if (rt->rt_gwroute == 0)
			goto lookup;
		  if (((rt = rt->rt_gwroute)->rt_flags & RTF_UP) == 0) 
			{
			  rtfree(rt); rt = rt0;
      
lookup: 
			  rt->rt_gwroute = rtalloc1(rt->rt_gateway, 1,
										0UL);
			  if ((rt = rt->rt_gwroute) == 0)
				senderr(EHOSTUNREACH);
			}
		}

      if (rt->rt_flags & RTF_REJECT)
		if (rt->rt_rmx.rmx_expire == 0 ||
			nfx_time_second < rt->rt_rmx.rmx_expire)
		  senderr(rt == rt0 ? EHOSTDOWN : EHOSTUNREACH);
    }

  
  switch (dst->sa_family)
    {
    case AF_INET:
      if (!arpresolve(ac, rt, m, dst, edst, rt0))
		{
		  return 0;
		}

      nh_mac = edst;
      type = ETHERTYPE_IP;
      break;
      
    case AF_UNSPEC:
      nh_mac = ((struct ether_header *)dst->sa_data)->ether_dhost;
      type = ETHERTYPE_ARP;
      break;
      
    default:
      printf("%s%d: can't handle af%d\n", ifp->if_name, ifp->if_unit,
			 dst->sa_family);
      senderr(EAFNOSUPPORT);
      break;
    }


  /* convert mbuf to pkt descriptor */
  len = m->m_pkthdr.len; 
  if (((m->m_flags & M_PKTHDR) == 0) ||
	 !(len > 0 && len <= ETHERMTU))
	{
	/* assuming support stardard ethernet only */
	  bsd_log(BSD_WARN_INTF, "bsdippkt-api: bad mbuf format or size(%d)\n",
			  len);
	  senderr(0);
	}

  /* m already has a descriptor, this happens when mbuf is processed */
  /* within BSD stack without going though application */
  if (m->m_pktdesc)
  {
	PKT_FREE_PACKET((pkt_desc_t*)m->m_pktdesc);
	m->m_pktdesc = 0;
  } 

  /* if len > 2048, I want to have pkt layer provid apis to copy to */
  /* mulitiple pkt descriptor, not implemented */
      
  if ((pdesc = PKT_ALLOC_PACKET(PKT_USR_PKT)) == 0)
    senderr(0);

  pdesc->hdrSize = len;
  pdesc->pIpHdr  = pdesc->pHdr;
  pdesc->link = ifp->if_unit;
  m_copydata(m, 0, len, pdesc->pIpHdr);

  if (type == ETHERTYPE_IP)
	{
	  
	  ipHdr  = (struct ip *)(pdesc->pIpHdr);
	  pdesc->laddr   = ipHdr->ip_src.s_addr;
	  pdesc->raddr   = ipHdr->ip_dst.s_addr;
	  pdesc->pTplHdr = ((char *)ipHdr  + (ipHdr->ip_hl << 2));

	  /* this is ip pkt, enable ip chksum */
	  pdesc->flags   |= PKT_IP_CSUM_ENA;

	  if (ipHdr->ip_p == IPPROTO_UDP)
		{
		  struct udphdr  *udpHdr;
		  
		  /*-------------------------------
		   * Extract UDP Dst & Src ports.
		   * Setup App. hdr ptr.
		   *-------------------------------
		   */
		  udpHdr = (struct udphdr *)pdesc->pTplHdr;
		  pdesc->lport = (udpHdr->uh_sport);
		  pdesc->rport = (udpHdr->uh_dport);
		  pdesc->pHdr  = pdesc->pTplHdr + sizeof(struct udphdr);
		  pdesc->tplProt= IPPROTO_UDP;
		  
		  /* chksum could be overwritten, if fragment */
		  pdesc->flags   |= PKT_UDP_CSUM_ENA;
		}
	  else if (ipHdr->ip_p == IPPROTO_TCP)
		{
		  struct tcphdr  *tcpHdr;
		  
		  /*-------------------------------
		   * Extract TCP Dst & Src ports.
		   * Setup App. hdr ptr.
		   *-------------------------------
		   */
		  tcpHdr = (struct tcphdr *)pdesc->pTplHdr;
		  pdesc->lport = (tcpHdr->th_sport);
		  pdesc->rport = (tcpHdr->th_dport);
		  pdesc->pHdr  = pdesc->pTplHdr + (tcpHdr->th_off << 2);
		  pdesc->tplProt= IPPROTO_TCP;

		  /* chksum could be overwritten, if fragment */
		  pdesc->flags   |= PKT_TCP_CSUM_ENA;
		}
	  else
		pdesc->pHdr = pdesc->pTplHdr;


	  /* Exclude IP & TCP headers for Simulation */
	  pdesc->pktSize = len;
	  pdesc->hdrSize = (len - (pdesc->pHdr - pdesc->pIpHdr));
	}
  else							/* arp packet */
	{
	  pdesc->pktSize = len;
	  pdesc->hdrSize = 0;

	}
  
  if (pkt_xmtSlowPath(ifp, pdesc, nh_mac, type) != NFX_OK)
	PKT_FREE_PACKET(pdesc);

 
  /* pkt freed by lower layer */
  /* fall through */
 bad:
  if (m)
    m_freem(m);
  return (error);
} 


/*
 *-----------------------------------------------------------------------
 *
 * Name:   bsd_IpRcvPkt
 *
 * Description: Called by pkt_pollRcvQueue to convert IP in pkt-descr to mbuf. 
 *              It send packet to BSD TCP/IP stack by calling ip_input(). 
 *              packets are assumed to have layer2 header processed already.
 *
 *
 * Created by : Su Wang
 *
 * Date Created:  8/6/01
 *------------------------------------------------------------------------
 */
#define PKT_GET_TCPIP_HLEN(a)  MHLEN

int32  /* NFX_ERR, NFX_OK */
bsd_RcvIpPkt(void* pktDesc,
			 struct ifnet* ifp,
			 void* param1,
			 void* param2)
{
  pkt_desc_t	*pdesc = (pkt_desc_t*)pktDesc;
  struct mbuf	*m;
  int			len;
  struct ip		*hdrptr;
  ushort16		hlen;
  rl_addr_t		rla;
  int			ret;
#if defined(LINK_AGG)
  struct ifnet*	vifIfp;
#endif /* LINK_AGG */

  if ((!ifp) ||
	  (ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
    return PKT_FREE_PKT;

#if defined(LINK_AGG)
  /* convert rcving ifp to the active ifp where
   * IP and route are assigned. 
   */
  vifIfp = (struct ifnet*)bsd_VifLAGetActLink(ifp);
  if (vifIfp)
	  ifp = vifIfp;
#endif /* LINK_AGG */
  
  if (!pdesc)
    return PKT_FREE_PKT;

  if ((len = pdesc->pktSize - (pdesc->pIpHdr - pdesc->pMac))
	  == 0)
    return PKT_FREE_PKT;

  if ((hdrptr = (struct ip*)pdesc->pIpHdr) == 0)
    return PKT_FREE_PKT;
 

  /* should we relay rcved pkt */
  bzero(&rla, sizeof(rla)); 
  ret = bsdrl_DstPkt(pdesc, &rla, (void*)ifp);
  if (ret == BSD_RELAY)			/* send to another slot */
	{
	  
	  if (bsdrl_RelayPkt(pdesc, &rla, 0) == NFX_OK)
		{
		  ipstat.rl_ipfwd++;
		}
	  else
		{
		  ipstat.rl_ipfwderr++;
		  return PKT_FREE_PKT;
		}
	}
  else if (ret == BSD_NRELAY)	/* process locally */
	{
	  /* TCP or UDP, we can just */
	  /* copy one mbuf */
	  if ((hdrptr->ip_p == IPPROTO_TCP || hdrptr->ip_p == IPPROTO_UDP))
		hlen = pdesc->pktSize - (pdesc->pIpHdr - pdesc->pMac);
	  else
		hlen = len;
	  
	  if ((m = m_devget((char*)hdrptr, hlen, 0, ifp, 0)) == 0)
		return PKT_FREE_PKT;
	  
	  m->m_pktdesc = pdesc;
	  
	  ip_input(m);
	}
  else							/* err has occured, drop pkt */
	return PKT_FREE_PKT;
  
  return PKT_HOLD_PKT;
}

 
/*
 *-----------------------------------------------------------------------
 *
 * Name:   bsd_RcvArpPkt
 *
 * Description: Called by pkt_pollRcvQueue to convert ARP in pkt-descr to mbuf. 
 *              It send packet to BSD TCP/IP stack by calling arp_input(). 
 *              packets are assumed to have layer2 header processed already.
 *
 *
 * Created by : Su Wang
 *
 * Date Created:  8/6/01
 *------------------------------------------------------------------------
 */
int32							/* NFX_ERR, NFX_OK */
bsd_RcvArpPkt(void*		pktDesc,
			  struct	ifnet* ifp,
			  void*		param1,
			  void*		param2)
{
  pkt_desc_t* pdesc = (pkt_desc_t*)pktDesc;
  struct mbuf *m;
  int len;
  struct arphdr *ar;
  rl_addr_t		rla;
#if defined(LINK_AGG)
  struct ifnet*	vifIfp;
#endif /* LINK_AGG */
 
  if ((!ifp) || 
	  (ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
    return PKT_FREE_PKT;

#if defined(LINK_AGG)
  /* convert rcving ifp to the active ifp where
   * IP and route are assigned. Note this is 
   * done after itf chk where pkt received 
   */
  vifIfp = (struct ifnet*)bsd_VifLAGetActLink(ifp);
  if (vifIfp)
	  ifp = vifIfp;
#endif /* LINK_AGG */

  if (!pdesc)
    return PKT_FREE_PKT;
	
  if ((len = pdesc->pktSize - (pdesc->pIpHdr - pdesc->pMac)) == 0)
    return PKT_FREE_PKT;

  if ((ar = (struct arphdr*)pdesc->pIpHdr) == 0)
    return PKT_FREE_PKT;

  /* some sanity check, copied from arpintr */
  if (len < sizeof(struct arphdr) + 2*ar->ar_hln + 2*ar->ar_pln)
    return PKT_FREE_PKT;

  if (ntohs(ar->ar_hrd) != ARPHRD_ETHER)
    return PKT_FREE_PKT;
 

  if ((m = m_devget((char*)ar, len, 0, ifp, 0)) == 0)
	return PKT_FREE_PKT;
 
  m->m_pktdesc = pdesc;			  
  
  switch (ntohs(ar->ar_pro)) 
    {
    case  ETHERTYPE_IP:
	  /* disassociate pdesc from mbuf */
	  m->m_pktdesc = 0;
      in_arpinput(m);

	  /* broadcast to system, if arp received on */
	  /* local interface */
	  if (ifp->if_slot == nfxMySlotId &&
		  ifp->if_cpu  == nfxMyCPUId)
		{
#if defined(ARP_SNIFFING)
		  /* insert entry to layer2 connection table */
		  arpS_AddEntry(ifp, (struct ether_arp*)ar);
#endif /* ARP_SNIFFING */		  
			
		  bzero(&rla, sizeof(rla));
		  /* set rla to multicast, */
		  /* and send to all ssc and nfp */
		  rla.rl_ctrl |= BSD_RL_PKT_BCAST;
		  rla.rl_ifp  = ifp; 
		  if (bsdrl_RelayPkt(pdesc, &rla, 0) == NFX_OK)
			{
			  ipstat.rl_arpfwd++;
			}
		  else	
			{					
			  ipstat.rl_arpfwderr++;
			  return PKT_FREE_PKT;
			}
		}
	  else /* arp from bk plane, simply free pdesc */
		{
		  return PKT_FREE_PKT;
		}
      break;
	  
	default:
	  m_freem(m);
    }

  return PKT_HOLD_PKT;

}


int32 
pkt_xmtSlowPath(void *i_ifp, pkt_desc_t *pdesc, char *nh_mac, short type)
{
  struct ifnet	*ifp = (struct ifnet*)i_ifp;
  struct arpcom *ac = (struct arpcom*)ifp;
  int			addlen = 0;
 
  switch(ac->ac_if.if_type)
    {
    case IFT_ETHER:
      addlen = PKT_ETHER_HDR_SIZE;
      break;


    default:
		return NFX_ERR;
    }

  pdesc->pktSize += addlen;

  pdesc->pMac = pdesc->pIpHdr - PKT_ETHER_HDR_SIZE;

  /* fill in layer 2 header */
  switch(ac->ac_if.if_type)
    {
    case IFT_ETHER:
      memcpy((char*)((pkt_ethHdr_t*)pdesc->pMac)->srcMac, ac->ac_enaddr, 6);
      memcpy((char*)((pkt_ethHdr_t*)pdesc->pMac)->dstMac, nh_mac, 6);
      ((pkt_ethHdr_t*)pdesc->pMac)->type = htons(type);
      break;

    default:
		return NFX_ERR;
    }

  if (ifp->if_slot == nfxMySlotId &&
	  ifp->if_cpu == nfxMyCPUId)
	{
	  /* interface is local, send to luc driver */
#if defined(REDIRECT_IP) && defined(LINUX_TEST)
	  bsd_RedirIP(pdesc, ifp);
#else
	  /* put onto xmt queue */
	  luc_xmtPktReq( ifp->if_unit, pdesc );
#endif /* LINUX TEST */
	}
  else
	{
	  /* interface is remote, send to back plane */
	  rl_addr_t	rla;
	  
	  rla.rl_slot = ifp->if_slot;
	  rla.rl_cpu  = ifp->if_cpu;
	  rla.rl_ifp  = ifp;
	  rla.rl_ctrl = BSD_RL_PKT_OWN;
	 
	  return bsdrl_RelayPkt(pdesc, &rla, 0);

	}

  return NFX_OK;
}

                  

q_headNodePtr_t tst_pkt_queue;
static int32 tst_q_timer;
extern void *loopConn1;
extern void *loopConn2;

static int32
bsdProcTstQ(uint32 handle, void* param1, void* param2, void* param3, void* param4)
{
#if defined(LINUX_TEST)
  pkt_desc_t* pdesc = 0;
  struct ifnet* ifp = 0;
  eeeRTC_t now;

  
  // printf("sim queue called at %u\n",eee_timerGetCurrTic());

  /* process any loop bck pkt */
  ipintr(0);

  while (q_deQueue(tst_pkt_queue, (void *)&pdesc, (void *)&ifp) != Q_ERR_EMPTY_QUEUE)
    {

	  eee_rtc(&now);
	  printf("sim queue called at %llu\n", eee_rtcToMS(&now));
	  if ((ifp ==loopConn1) || (ifp == loopConn2))
		{
		  /* fastpath tcp loopbck test */
			  bsd_RcvTcpPkt((pkt_desc_t*)pdesc, ifp, 0, 0); 
		}
	  else if (*pdesc->pIpHdr == 0x45) /* ip */
		{
			bsd_RcvIpPkt(pdesc, ifp, 0, 0);
		}
      else if (*(ushort*)(pdesc->pIpHdr) == htons(ARPHRD_ETHER))
		bsd_RcvArpPkt(pdesc, ifp, 0, 0);
      else
		PKT_FREE_PACKET(pdesc);
    }
#endif /* LINUX_TEST */

  return 1;
}

/*-----------------------------------------------------------------
 * Name: bsd_qcompare
 *
 * Description: This is the queue comparision routine
 *				We enqueued pointers and this routine is given
 *				pointers to the type we enqueued.
 *
 * Created by:	Jonathan Goldick
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
static int32
bsd_qcompare(
	const void *i1,
	const void *i2
)
{
	void *item1 = *(void **)i1;
	void *item2 = *(void **)i2;

	if (item1 > item2) {
		return 1;
	}
	if (item1 < item2) {
		return -1;
	}
	return 0;
}

void
bsd_TstPktInit()
{
  if ((tst_pkt_queue = q_createQueue(sizeof(void *), sizeof(void *),
									 bsd_qcompare, bsd_qcompare)) == NULL)
    {
      printf("bsd_TstPktInit: tst_pkt_queue failed\n");
      return;
    }

  if ((tst_q_timer = eee_timerCreate((200), EEE_TIMER_REPEAT)) ==
      TIMER_ERR_NO_TIMER_AVAILABLE)
    {
      printf("bsd_TstPktInit: tst_q_timer fail to creat\n");
      return;      
    }

  if (eee_timerStart(tst_q_timer, bsdProcTstQ, 0,0,0,0) != NFX_OK)
    {
      printf("bsd_TstPktInit: tst_q_timer fail to start\n");
      return;          
    }
  
  printf(" pkt process timer started, queue  %lx, timer %u....\n",
		 (address_t)tst_pkt_queue, tst_q_timer);
  return;
}

#if defined(LINUX_TEST)
extern struct ifnethead ifnet;
void bsd_TstUDP(ipa_t src, ipa_t dest, u_short sport, u_short dport)
{
  static char ip_buf[] = { 0x45, 0x00, 0x00, 0x7c, 0xa7, 0x25, 0x00, 0x00, 
						   0x1e, 0x11, 0xdf, 0x0d, 0x0a, 0x01, 0x01, 0x09,
						   0x0a, 0x01, 0x01, 0x34, 0x03, 0x1f, 0x08, 0x01,
						   0x00, 0x68, 0x06, 0x06, 0x13, 0xa0, 0x74, 0x45,
						   0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 
						   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
						   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
						   0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x40, 0x3f,
						   0x00, 0x00, 0x00, 0x0b, 0x00, 0x00, 0xea, 0x61,
						   0x00, 0x00, 0xff, 0xfe, 0x00, 0x00, 0x02, 0x00,
						   0x00, 0x00, 0x20, 0x00, 0x80, 0x00, 0x0c, 0x01,
						   0x00, 0x00, 0x00, 0x10, 0x00, 0x06, 0x00, 0x00,
						   0x00, 0x00, 0x09, 0x00, 0x3b, 0x73, 0x95, 0xa8,
						   0x00, 0x0a, 0xae, 0x7c, 0x3b, 0x4b, 0x44, 0xb9,
						   0x00, 0x03, 0x0d, 0x41, 0x3a, 0xae, 0x85, 0x4d,
						   0x00, 0x07, 0xa1, 0x20};

  struct ip* ip_ptr = (struct ip*)ip_buf;
  struct udphdr* udp_ptr;
  int iplen = sizeof(ip_buf), udplen; 
  pkt_desc_t* pdesc;
  
  ip_ptr->ip_len = htons(iplen);
  if (src)
    ip_ptr->ip_src.s_addr = src;
  
  if (dest)
    ip_ptr->ip_dst.s_addr = dest;

  udplen = iplen - (ip_ptr->ip_hl <<2);
  udp_ptr = (struct udphdr*)(ip_buf + (ip_ptr->ip_hl <<2));

  udp_ptr->uh_ulen = htons(udplen);
  
  if (sport)
    udp_ptr->uh_sport = htons(sport);

  if (dport)
    udp_ptr->uh_dport = htons(dport);


  if ((pdesc = PKT_ALLOC_PACKET(PKT_USR_PKT)) == 0)
    return;
  pdesc->pHdr += PKT_HDR_INDENT;

  pdesc->pktSize = iplen;
  pdesc->hdrSize = iplen - 
	(PKT_IP_HDR_SIZE + PKT_UDP_HDR_SIZE);

  pdesc->pIpHdr = pdesc->pHdr; 
  memcpy(pdesc->pIpHdr, ip_buf, iplen);
  pdesc->pHdr += PKT_IP_HDR_SIZE + PKT_UDP_HDR_SIZE;
  pdesc->pMac =  pdesc->pIpHdr - PKT_ETHER_HDR_SIZE;
  pdesc->pTplHdr = pdesc->pIpHdr + PKT_IP_HDR_SIZE;

  bsd_RcvIpPkt(pdesc, ifnet.tqh_first, 0, 0);

  return;
}

static void
bsd_ansARP(pkt_desc_t* pdesc, struct ifnet* ifp)
{
  struct ether_arp* ea;
  uint32 addr;

  ea = (struct ether_arp*)pdesc->pIpHdr;
  if (ntohs(ea->arp_op) != ARPOP_REQUEST)
    {
      PKT_FREE_PACKET(pdesc);
      return ;
    } /* need to operation to replay */
  else
	ea->arp_op = htons(ARPOP_REPLY);

  /* switching IP address */
  addr = *((uint32*)ea->arp_spa);
  *((uint32*)ea->arp_spa) = *((uint32*)ea->arp_tpa);
  *((uint32*)ea->arp_tpa) = addr;

  /* get mac addr */
  memcpy(ea->arp_tha, ea->arp_sha, 6);
  *((uint32*)ea->arp_sha) = htonl(nfx_time_second);
  ea->arp_sha[4] = 11;
  ea->arp_sha[5] = 22;

  /* queue to input */
  debugPrintf(("bsd_ansARP enqueue %u items %lx %lx\n", 
			  tst_pkt_queue, pdesc, ifp));
  q_enQueue(tst_pkt_queue, (void *)&pdesc, (void *)&ifp);
  
  return;
}

static void
bsd_ansIP(pkt_desc_t* pdesc, struct ifnet* ifp)
{
  struct ip* ip_ptr;
  struct icmp* icmp_ptr;
  uint32 addr;
  uint32 csum;

  ip_ptr = (struct ip*)pdesc->pIpHdr;

  if (ip_ptr->ip_p == IPPROTO_ICMP)
    {
      icmp_ptr = (struct icmp*)(ip_ptr + 1);
      if (icmp_ptr->icmp_type == ICMP_ECHO)
		{
		  icmp_ptr->icmp_type = ICMP_ECHOREPLY;
		  addr = ip_ptr->ip_src.s_addr;
		  ip_ptr->ip_src.s_addr = ip_ptr->ip_dst.s_addr;
		  ip_ptr->ip_dst.s_addr = addr;
	  
		  ip_ptr->ip_sum = 0;
		  ip_ptr->ip_sum = in_cksum_hdr(ip_ptr);

		  /* type field has changed, adjust cksum accordingly */
		  csum = ~icmp_ptr->icmp_cksum & 0xffff;
		  csum -= 0x8;			/* not always right ?? */
		  csum += (csum >> 16) & 0xffff;
		  icmp_ptr->icmp_cksum = ~csum & 0xffff;
			
	
			//icmp_ptr->icmp_cksum = 0;
		   //icmp_ptr->icmp_cksum = bsd_ChkSum(icmp_ptr, (pdesc->hdrSize - 20)); 
		  debugPrintf(("bsd_ansIP enqueue %u items %lx %lx\n", 
					  tst_pkt_queue, pdesc, ifp));
		  q_enQueue(tst_pkt_queue, (void *)&pdesc, (void *)&ifp);
		  return; 
		}
	  else						/* added to test fragmentation */
		{
		  addr = ip_ptr->ip_src.s_addr;
		  ip_ptr->ip_src.s_addr = ip_ptr->ip_dst.s_addr;
		  ip_ptr->ip_dst.s_addr = addr;
		  
		  debugPrintf(("bsd_ansIP enqueue %u items %lx %lx\n", 
					  tst_pkt_queue, pdesc, ifp));
		  q_enQueue(tst_pkt_queue, (void *)&pdesc, (void *)&ifp);
		  return; 
		}
	  
    }
  PKT_FREE_PACKET(pdesc);
  return;
}


int
bsd_RedirIP(pkt_desc_t* pdesc, struct ifnet* ifp)
{
  struct ether_header *eh; 
  ushort16 ether_type;

  eh = (struct ether_header*)pdesc->pMac;
  ether_type = ntohs(eh->ether_type);

  switch (ether_type)
    {
    case ETHERTYPE_IP:
      bsd_ansIP(pdesc, ifp);
      break;

    case ETHERTYPE_ARP:
      bsd_ansARP(pdesc, ifp);
      break;

    default:
      PKT_FREE_PACKET(pdesc);
    }
  return;
}
#endif /* UNIX_TEST */
