/*  File Name  :    agilediskluns.c
 *
 *  Description:    HTTPD action functions for displaying free/used lun list of a given pool.
 *
 *  Created    :    3/15/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <anpssc-api.h>

#include <evm_com_defs.h>
#include <scsi-api.h>
#include <sdm-api.h>
#include <spm-api.h>

#include "agilepost.h"
#include "agileutil.h"

#define INVALID_TAG			0
#define LL_ARGCTRL			1
#define LL_ARGRAID			2
#define LL_ARGTYPE			3
#define LL_SID				4
#define LL_LUN				5
#define LL_DEV				6
#define LL_SIZE				7
#define LL_TYPE				8
#define LL_RANGE			9
#define LL_PREVBTN			10
#define LL_NEXTBTN			11
#define LL_FOPREVBTN		12
#define LL_FONEXTBTN		13
#define LL_COLSPAN			14
#define NUM_OF_TAGS			14

#define NUM_MAXIPAGE_ENTRIES	10
#define NUM_PAGE_ENTRIES		10

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{LL_ARGCTRL,	"args.ctrl"},
	{LL_ARGRAID,	"args.rl"},
	{LL_ARGTYPE,	"args.type"},
	{LL_SID,		"spm_poolLunList_t.sid"},
	{LL_LUN,		"spm_poolLunList_t.lun"},
	{LL_DEV,		"spm_poolLunList_t.dev"},
	{LL_SIZE,		"spm_poolLunList_t.size"},
	{LL_TYPE,		"lunlist.type"},
	{LL_RANGE,		"lunlist.range"},
	{LL_PREVBTN,	"lunlist.prevbtn"},
	{LL_NEXTBTN,	"lunlist.nextbtn"},
	{LL_FOPREVBTN,	"lunlist.foprevbtn"},
	{LL_FONEXTBTN,	"lunlist.fonextbtn"},
	{LL_COLSPAN,	"lunlist.colspan"},
};


#define PARAM_TYPE			1
#define PARAM_CTRL			2
#define PARAM_RAID			3
#define PARAM_ACT			4
#define PARAM_LUN			5
#define PARAM_START			6
#define PARAM_COUNT			7
#define NUM_OF_PARAMS		7

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_TYPE,	"type"},
	{PARAM_CTRL,	"ctrl"},
	{PARAM_LUN,		"lun"},
	{PARAM_RAID,	"rl"},
	{PARAM_ACT,		"act"},
	{PARAM_START,	"start"},
	{PARAM_COUNT,	"count"},
};

/* HERE ARE VARIABLES FOR LUN LIST INFORMATION */
static int dataCount;
static int dataCurr;
static spm_poolLunList_t *stList;	// table contains entries for current page
static spm_poolLunList_t stData;

static int pageRowCount;
static uint64 lastSID;
static int pageNum;
static uint64 nextStartLunSID;
static int totalEntries;
static char inType[5];
static char inCtrl[SDM_DEV_NAME_LEN+1];
static int32 inLun;
static evm_raid_t inRaid;
static char inAct[10];


/***********************************************************
 * routines to find entries for current page
 * also handle input parameters
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	inRaid = EVM_RAID_ANY;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_TYPE:
			strcpy(inType, argval[argind]);
			break;
		case PARAM_CTRL:
			strcpy(inCtrl, argval[argind]);
			break;
		case PARAM_RAID:
			sscanf(argval[argind], "%d", (int *) (&inRaid));
			break;
		case PARAM_ACT:
			strcpy(inAct, argval[argind]);
			break;
		case PARAM_LUN:
			sscanf(argval[argind], "%d", &inLun);
			break;
		case PARAM_START:
			sscanf(argval[argind], "%lld", &lastSID);
			break;
		case PARAM_COUNT:
			//sscanf(argval[argind], "%d", &pageRowCount);
			sscanf(argval[argind], "%d", &pageNum);
			break;
		}
	}

	return 0;
}

static int getLunType()
{
	if (!strcmp(inType, "u"))
		return SPM_LUN_USED_LIST;
	else if (!strcmp(inType, "f"))
		return SPM_LUN_FREE_LIST;
	else if (!strcmp(inType, "fo"))
		return SPM_LUN_FOREIGN_LIST;
	else if (!strcmp(inType, "oc"))
		return SPM_LUN_OUTCLUSTER_LIST;
	else
		return SPM_LUN_FOREIGN_LIST;
}

static int getTotalRowCount()
{
    int						stat;
    spm_MsgTotalLunReq_t    req;
    spm_MsgTotalLunRsp_t    rsp;
	struct timeval			timeout = {30, 0};
            
    //Fill the request. 
    memset((void *) &req, 0, sizeof(spm_MsgTotalLunReq_t));
    req.hdr.msgType = SPM_GET_TOTAL_LUNS_REQ;
	strcpy(req.hdr.node, "\0");
	strcpy(req.ctrl, inCtrl);
	req.rL = inRaid;
	req.lunType = getLunType();

    // Send message to SPM. 
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout, &rsp, sizeof(spm_MsgTotalLunRsp_t));

    // Check response from SPM. 
	if (stat < 0) 
		return -1;
	else if (!stat)
		return -1;

    if (rsp.hdr.msgType != SPM_GET_TOTAL_LUNS_RSP)
		return -1;
//	if (rsp.numLun <= 0)
	if (rsp.numLun < 0)
		return -1;

	return rsp.numLun;
}

static void getPageTable()
{
    int						stat;
    char					respbuff[10240];
    spm_MsgLunListReq_t		req;
    spm_MsgLunListRsp_t		*rsp;
	struct timeval			timeout = {30, 0};
            
    //Fill the request. 
    bzero((void *) &req, sizeof(spm_MsgLunListReq_t));
    req.hdr.msgType = SPM_GET_LUN_LIST_REQ;
    strcpy(req.hdr.node, "local");
	strcpy(req.ctrl, inCtrl);
	req.rL = inRaid;
	req.lunType = getLunType();
    req.numReq = pageRowCount;
	req.lastSID = lastSID;
	req.numReq = NUM_PAGE_ENTRIES;

    // Send message to SPM. 
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &timeout, respbuff, sizeof(respbuff));

    // Check response from SPM. 
	if (stat <= 0) 
		return;

    rsp = (spm_MsgLunListRsp_t *)respbuff;
    if (rsp->hdr.msgType != SPM_GET_LUN_LIST_RSP)
		return;

	if (stList != NULL)
		free(stList);
	stList = malloc((rsp->numLuns) * sizeof(spm_poolLunList_t));
	if (stList == NULL)
		return;
	memcpy(stList, rsp->lunList, (rsp->numLuns) * sizeof(spm_poolLunList_t));

	if (rsp->numLuns < pageRowCount)
		pageRowCount = rsp->numLuns;
	nextStartLunSID = rsp->lunList[rsp->numLuns-1].sid;
}


static int DoResetLun()
{
    int                       stat;
    uint32                    rc=0;
    char                	  respbuff[1024];
    sdm_getSdmHandleReqMsg_t  hmsg;
    sdm_getSdmHandleRspMsg_t *hrsp;
    sdm_devResetReqMsg_t	  msg;
    sdm_devResetRspMsg_t     *rsp;
	uint32 				      sdmHandle;

	/* get SDM handle */
	hmsg.hdr.mid = SDM_MID_GET_SDMHANDLE_REQ;
#if 0
    hmsg.wwn = inWwn;
#endif
    hmsg.lun = inLun;

    stat = sendAgileMsg(-1, &hmsg, sizeof(hmsg),
						"sdm", NFX_SSC_UNIX, SSC_SLOT, 0,
                        NULL, respbuff, sizeof(respbuff));

	if (stat < 0) 
		return -1;
	else if (!stat)
		return -1;

    hrsp = (sdm_getSdmHandleRspMsg_t *)respbuff;

    if( hrsp->hdr.mid != SDM_MID_GET_SDMHANDLE_RSP ) 
		return -1;
    else if( hrsp->rCode == SDM_FAIL ) 
		return -1;
	else 
		sdmHandle = hrsp->sdmHandle;
	if (sdmHandle == 0)
		return -1;

	/* reset device */
	bzero(&msg, sizeof(sdm_devResetReqMsg_t));
	msg.hdr.mid = SDM_MID_DEV_RESET_REQ;
	msg.handle  = sdmHandle;

	rc = sendAgileMsg(-1, &msg, sizeof(sdm_devResetReqMsg_t),
					  "sdm", NFX_SSC_UNIX, SSC_SLOT, 0,
					  NULL, respbuff, sizeof(respbuff));
	if (rc < 0)
		return -1;
	else 
	if (rc == 0)
		return -1;

	rsp = (sdm_devResetRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_DEV_RESET_RSP)
		return -1;
    else if (rsp->eCode != SDM_OK) 
		return -1;
#if 0
	else 
		return -1;
#endif
	return 1;
}

static int DoLabelLun()
{
    int							stat;
    spm_MsgMarkAgileLunReq_t	req;
    spm_MsgMarkAgileLunRsp_t	rsp;
	struct timeval				timeout = {10, 0};
            
    //Fill the request. 
    memset(&req, 0, sizeof(spm_MsgMarkAgileLunReq_t));
    req.hdr.msgType = SPM_MARK_LUN_AS_AGILE_REQ;
	strcpy(req.hdr.node, "\0");
    req.devCnt = 1;
	sprintf(req.dev[0], "%s_%d", inCtrl, inLun);

    // Send message to SPM. 
    stat = sendAgileMsg(-1, &req, sizeof(req),
				"spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
				&timeout, &rsp, sizeof(spm_MsgMarkAgileLunRsp_t));

    // Check response from SPM. 
	if (stat < 0) 
		return -1;
	else if (!stat)
		return -1;

    if (rsp.hdr.msgType != SPM_MARK_LUN_AS_AGILE_RSP)
		return -1;
	if (rsp.status != SPM_OK)
		return -1;

	return 1;
}

static int HandleLunAction(char *newfname, int count, char **argname, char **argval)
{
	int res;

	getInputParams(count, argname, argval);
	res = 1;
	if (!strcmp(inAct, "reset"))
	{
		res = DoResetLun();
		if (res <= 0)
			strcpy(newfname, "error_resetstgdev.html");
	}
	else if (!strcmp(inAct, "label"))
	{
		res = DoLabelLun();
		if (res <= 0)
			strcpy(newfname, "error_labellun.html");
	}
	return res;
}

static void	HandleDisplayPage(char *newfname, int count, char **argname, char **argval)
{
	int res;
//	startIndex = -1;
	lastSID = 0;
	pageNum = 0;
	pageRowCount = -1;

	if (count > 0)
	{
		getInputParams(count, argname, argval);
		res = HandleLunAction(newfname, count, argname, argval);
		if (res <= 0)
			return;
	}

	// get count
	totalEntries = getTotalRowCount();

	// get start index and row count
	/*
	if (startIndex < 0)
	{
		startIndex = 0;
		if (totalEntries > NUM_MAXIPAGE_ENTRIES)
			pageRowCount = NUM_PAGE_ENTRIES;
		else
			pageRowCount = totalEntries;
	}
	endIndex = startIndex + pageRowCount - 1;
	if (endIndex >= totalEntries)
		endIndex = totalEntries - 1;
	*/
	if (totalEntries > NUM_MAXIPAGE_ENTRIES)
		pageRowCount = NUM_PAGE_ENTRIES;
	else
		pageRowCount = totalEntries;
	// get row data for page table
	if (totalEntries > 0)
		getPageTable();
}


/***********************************************************
 * routines to retrieve disk/lun information 
 **********************************************************/

static int getTableRowCount(void)
{
	return pageRowCount;
}

static int getTableRowData(int *index, spm_poolLunList_t *data)
{
	if (*index < 1 || *index > dataCount)
		return -1;
	if (stList == NULL)
		return -1;
	
	memcpy(data, &stList[(*index)-1], sizeof(spm_poolLunList_t));
	return 0;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType, priv, type;
	int startIdx, endIdx;

	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}

	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType)
	{
	case LL_TYPE:
		type = getLunType();
		if (type == SPM_LUN_USED_LIST)
			strcpy(value, "Used");
		else if (type == SPM_LUN_FREE_LIST)
			strcpy(value, "Free");
		else if (type == SPM_LUN_FOREIGN_LIST)
			strcpy(value, "Foreign");
		else if (type == SPM_LUN_OUTCLUSTER_LIST)
			strcpy(value, "Out-Cluster");
		return 1;

	case LL_RANGE:
		if (pageRowCount < totalEntries)
		{
			startIdx = pageNum * NUM_PAGE_ENTRIES + 1;
			endIdx = startIdx + NUM_PAGE_ENTRIES - 1;
			if (endIdx > totalEntries)
				endIdx = totalEntries;
			sprintf(value, ": %d - %d of %d", startIdx, endIdx, totalEntries);
		}
		else
			strcpy(value, "\0");
		return 1;

	case LL_ARGCTRL:
		sprintf(value, "%s", inCtrl);
		return 1;

	case LL_ARGRAID:
		sprintf(value, "%d", inRaid);
		return 1;
	case LL_ARGTYPE:
		strcpy(value, inType);
		return 1;
	}

	if ((dataCurr <= 0) || (dataCurr > dataCount))
		return -1;

	switch (tagType) 
	{
	case LL_SID:
		sprintf(value, "%lld", stData.sid);
		break;
	case LL_LUN:
		sprintf(value, "%d", stData.lun);
		break;
	case LL_SIZE:
		sprintf(value, "%lld", GetMBfromBlock(stData.size));
		break;
	/*
	case LL_MAKE:
		strcpy(value, make);
		break;
	case LL_MODEL:
		strcpy(value, model);
		break;
	*/
	case LL_PREVBTN:
		if (pageRowCount < totalEntries)	// more than one pages
		{
			sprintf(value, "<a href=\"/agilerequestprefix/DiskLuns?count=0&ctrl=%s&rl=%d&type=%s\"> ", inCtrl, inRaid, inType);
			strcat(value, "<img src=\"/AgileView/images/button_first.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	case LL_NEXTBTN:
		if ((pageRowCount < totalEntries) &&	// more than one pages
			(pageNum*NUM_PAGE_ENTRIES+pageRowCount < totalEntries))	// not last page
		{
			sprintf(value, "<a href=\"/agilerequestprefix/DiskLuns?start=%lld&count=%d&ctrl=%s&rl=%d&type=%s\"> ", nextStartLunSID, pageNum+1, inCtrl, inRaid, inType);
			strcat(value, "<img src=\"/AgileView/images/button_next.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	case LL_FOPREVBTN:
		if (pageRowCount < totalEntries)	// more than one pages
		{
			sprintf(value, "<a href=\"/agilerequestprefix/DiskFOLuns?count=0&ctrl=%s&rl=%d&type=%s\"> ", inCtrl, inRaid, inType);
			strcat(value, "<img src=\"/AgileView/images/button_first.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	case LL_FONEXTBTN:
		if ((pageRowCount < totalEntries) &&	// more than one pages
			(pageNum*NUM_PAGE_ENTRIES+pageRowCount < totalEntries))	// not last page
		{
			sprintf(value, "<a href=\"/agilerequestprefix/DiskFOLuns?start=%lld&count=%d&ctrl=%s&rl=%d&type=%s\"> ", nextStartLunSID, pageNum+1, inCtrl, inRaid, inType);
			strcat(value, "<img src=\"/AgileView/images/button_next.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	case LL_COLSPAN:
		priv = GetUserPrivilege();
		if (priv >= PRIV_DIAG)
			sprintf(value, "%d", 2);
		else if (priv == PRIV_FULL)
			sprintf(value, "%d", 1);
		break;
	default:
		return -1;
	}

	if (value && !value[0])
		strcpy(value, "&nbsp");
	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		/* TABLE_BEGIN tag, need to see if there is any data */
		dataCount = getTableRowCount();
		if (dataCount > 0) {
			/* need to retrieve first table's data */
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) >= 0)
				return 1;
		}

		dataCurr = 0;
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		/* TABLE_END tag, check if end of data */
		if ((dataCount <= 0) || (dataCurr <= 0) || (dataCurr >= dataCount))
			return -1;

		dataCurr++;
		if (getTableRowData(&dataCurr, &stData) < 0)
			return -1;
		else {
			return 1;
		}
	}
	else {
		/* other tags */
		/*
		if (dataCount <= 0) {
			dataCount = getTableRowCount();
			if (dataCount <= 0)
				return -1;
		}

		if ((dataCurr < 1) || (dataCurr > dataCount)) {
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) < 0) {
				return -1;
			}
		}
		*/

		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionDiskLuns(void *r, int count, char **argname, char **argval)
{
	char newfname[128];

	inAct[0] = '\0';
	memset(newfname, 0, sizeof(newfname));

	/* Find entries for the current displaying page */
	/* Also handle reset, label LUN action */
	stList = NULL;
	HandleDisplayPage(newfname, count, argname, argval);

	/* initialize lun list information variables */
	dataCount = 0;
	dataCurr = 0;

	prepare_agile_template(r, newfname);

	agile_scan_and_send(r, actRptTagFunc);

	if (stList != NULL)
		free(stList);
	return 1;
}
