/*  File Name  :    agilelundetail.c 
 *
 *  Description:    HTTPD action functions for displaying lun details.
 *
 *  Created    :    2/6/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>

#include <nfx-defs.h>
#include <nfx-types.h>
#include <eee-desc.h>
#include <scsi-api.h>
#include <sdm-api.h>

#include "agilepost.h"
#include "agileutil.h"

#define INVALID_TAG			0
#define STGDEV_INDEX		1
#define STGDEV_WWNSTR		2
#define STGDEV_WWN			3
#define STGDEV_LUN			4
#define STGDEV_TYPE			5
#define STGDEV_VENDOR		6
#define STGDEV_MODEL		7
#define STGDEV_STATE		8
#define STGDEV_NAME			9
#define STGDEV_BLKCNT		10
#define STGDEV_BLKSIZE		11
#define STGDEV_PATHCNT		12
#define STGDEV_ACTPATH		13
#define STGDEV_PATHALL		14
#define STGDEV_HANDLECNT	15
#define STGDEV_HANDLELST	16
#define STGDEV_ARGSCSIID	17
#define STGDEV_ARGTYPE		18
#define STGDEV_TITLE		19
#define NUM_OF_TAGS			19

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{STGDEV_INDEX,		"args.index"},
	{STGDEV_WWNSTR,		"wwnstr"},
	{STGDEV_WWN,		"sdm_devInfoRspMsg_t.wwn"},
	{STGDEV_LUN,		"sdm_devInfoRspMsg_t.lun"},
	{STGDEV_TYPE,		"sdm_devInfoRspMsg_t.devType"},
	{STGDEV_VENDOR,		"sdm_devInfoRspMsg_t.vendorName"},
	{STGDEV_MODEL,		"sdm_devInfoRspMsg_t.modelName"},
	{STGDEV_STATE,		"sdm_devInfoRspMsg_t.state"},
	{STGDEV_NAME,		"sdm_devInfoRspMsg_t.devName"},
	{STGDEV_BLKCNT,		"sdm_devInfoRspMsg_t.blockCnt"},
	{STGDEV_BLKSIZE,	"sdm_devInfoRspMsg_t.blockSize"},
	{STGDEV_PATHCNT,	"sdm_devStatusRspMsg.pathCnt"},
	{STGDEV_ACTPATH,	"sdm_devStatusRspMsg.activePath"},
	{STGDEV_PATHALL,	"sdm_devStatusRspMsg.paths.all"},
	{STGDEV_HANDLECNT,	"handleCnt"},
	{STGDEV_HANDLELST,	"handleList"},
	{STGDEV_ARGTYPE,	"args.type"},
	{STGDEV_TITLE,		"title.path"},
};


#define PARAM_SCSIID		1
#define PARAM_TYPE			2
#define NUM_OF_PARAMS		2

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_SCSIID,	"scsiid"},
	{PARAM_TYPE,	"type"},
};


/* HERE ARE VARIABLES FOR STORAGE DEVICE DETAIL INFORMATION */
static scsi_id_t inScsiId;
static char inType[10];
static int res;
static sdm_devStatusRspMsg_t stData;

static char *devType_str[] = {
	"Disk",		"Tape",
	"printer",	"processor",
	"worm",		"cdrom",
	"scanner",	"optical_disk",
	"Changer",	"comm_dev",
	"printer1",	"printer2",
	"raid",		"ses",
	"unknown",	"unknown1"
};

static char *devState_str[] = {"Closed", "Opened", "Closing", "Opening" };


/***********************************************************
 * routines to retrieve storage information thru anpssc
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_SCSIID:
			sscanf(argval[argind], "%lld", &inScsiId);
			break;
		case PARAM_TYPE:
			strcpy(inType, argval[argind]);
			break;
		}
	}

	return 0;
}

static void GetDeviceData()
{
    int			stat;
    sdm_devStatusReqMsg_t	stsmsg;
	struct timeval			timeout = {60, 0};

	res = -1;

	memset(&stData, 0, sizeof(stData));

	/* get device status data */
    stsmsg.hdr.mid = SDM_MID_DEV_STATUS_REQ;
    stsmsg.id      = inScsiId;

	memset(&stData, 0, sizeof(sdm_devStatusRspMsg_t));
    stat = sendAgileMsg(-1, &stsmsg, sizeof(stsmsg),
                        "sdm", NFX_SSC_UNIX, SSC_SLOT, 0, &timeout,
                        &stData, sizeof(sdm_devStatusRspMsg_t));

	if (stat < 0) 		
		return;
	else if (!stat)
		return;

    else if(stData.eCode == SDM_FAIL) 
		return;

	res = 1;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
    char   pathbuf[64];

	if (res <= 0)
		return -1;

	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i ++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case STGDEV_WWNSTR:
		GetWwnStr(stData.wwn, value);
		break;
	case STGDEV_WWN:
		sprintf(value, "%lld", stData.wwn);
		break;
	case STGDEV_LUN:
		sprintf(value, "%d", stData.lun);
		break;
	case STGDEV_TYPE:
		strcpy(value, devType_str[stData.devType]);
		break;
	case STGDEV_VENDOR:
		strcpy(value, stData.st.vendor);
		break;
	case STGDEV_MODEL:
		strcpy(value, stData.st.model);
		break;
	case STGDEV_STATE:
		strcpy(value, devState_str[stData.state]);
		break;
	case STGDEV_NAME:
		strcpy(value, stData.st.name);
		break;
	case STGDEV_BLKCNT:
		sprintf(value, "%d", stData.blockCnt);
		break;
	case STGDEV_BLKSIZE:
		sprintf(value, "%d", stData.blockSize);
		break;
	case STGDEV_PATHCNT:
		sprintf(value, "%d", stData.pathCnt);
		break;
	case STGDEV_ACTPATH:
		if (stData.state == SDM_DEV_STATE_OPENED)
		{
			sprintf(value, "Slot %d, Port %d", 
				SCSI_PATH_GET_SLOT(stData.activePath), 
				SCSI_PATH_GET_PORT(stData.activePath));
		}
		else
			strcpy(value, "");
		break;
	case STGDEV_PATHALL:
		strcpy(value, "");
		for(i = 0; i < stData.pathCnt; i ++)
		{
			sprintf(pathbuf, "0x%08x", stData.paths[i].path);
			strcat(value, "<tr><td class=\"tabledata\" nowrap> ");
			strcat(value, pathbuf);
			strcat(value, " </td></tr>\n");
		}
		break;
	case STGDEV_HANDLECNT:
		sprintf(value, "%d", stData.handleCnt);
		break;
	case STGDEV_HANDLELST:
		strcpy(value, "");
		for(i = 0; i < stData.handleCnt; i ++)
		{
			sprintf(pathbuf, "0x%08x", stData.handles[i]);
			strcat(value, "<tr><td class=\"tabledata\" nowrap> ");
			strcat(value, pathbuf);
			strcat(value, " </td></tr>\n");
		}
		break;
	case STGDEV_ARGTYPE:
		strcpy(value, inType);
		break;
	case STGDEV_TITLE:
		if (inType[0] == 'v')
			strcpy(value, "Storage -> Volume -> LUN Detail");
		else
			strcpy(value, "Storage -> Disk -> LUN Detail");
		break;
	default:
		return -1;
	}

	if (value && (value[0] == '\0'))
		strcpy(value, "&nbsp");

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		return -1;
	}
	else {
		/* other tags */
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionLunDetail(void *r, int count, char **argname, char **argval)
{
	/* initialize storage information variables */
	getInputParams(count, argname, argval);
	GetDeviceData();

	prepare_agile_template(r, NULL);

	agile_scan_and_send(r, actRptTagFunc);

	return 1;
}
