/*  File Name  :	agilenfsmain.c
 *
 *  Description:	HTTPD action functions for nfs/share main page.
 *
 *  Created	:	3/22/2002
 *
 *  Author	 :	Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>
#include <unistd.h>
#include <anpssc-api.h>

#include <evm_com_defs.h>
#include <evm-api.h>
#include <fs-api.h>
#include <rpc.h>
#include <mount.h>

#include "agilepost.h"
#include "agileutil.h"

#define INVALID_TAG			0
#define NFS_PRIVCOL			1
#define NFS_VOLID			2
#define NFS_VOLNAME			3
#define NFS_PATH			4
#define NFS_IP				5
#define NFS_HOSTALL			6
#define NFS_PERMALL			7
#define NFS_RSALL			8
#define NUM_OF_TAGS			8

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{NFS_PRIVCOL,		"priv.colspan"},
	{NFS_VOLID,			"args.vol_id"},
	{NFS_VOLNAME,		"args.vol_name"},
	{NFS_PATH,			"nfs_exportResp_t.path_name"},
	{NFS_IP,			"nfs_exportResp_t.share_ip"},
	{NFS_HOSTALL,		"nfs_exportResp_t.host_mask.all"},
	{NFS_PERMALL,		"nfs_exportResp_t.permission.all"},
	{NFS_RSALL,			"nfs_exportResp_t.root_squash.all"},
};


#define PARAM_ACT			1
#define PARAM_ID			2
#define PARAM_NAME			3
#define PARAM_PATH			4
#define PARAM_IP			5
#define NUM_OF_PARAMS		5

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_ID,		"id"},		// volume id
	{PARAM_NAME,	"vn"},		// volume name
	{PARAM_PATH,	"path"},	// path name
	{PARAM_IP,		"ip"},		// local ip
	{PARAM_ACT,		"act"},
};

static char bufact[10];
static int64 inId;
static char bufname[VOL_MAX_NAME+1];
static char bufpath[MAX_NAME_LEN+1];
static uint32 inIP;


/* HERE ARE VARIABLES FOR NFS INFORMATION */
#define EXPORT_COUNT		20
#define HOST_COUNT			20

typedef struct _export_host_t
{
	char			hostIP[20];
	char			hostName[64];
	char			subnetMask[20];
	int				permission;	// 0 - RO, 1 - RW
	int				rootSquash; // 0 - no, 1 - yes
} export_host_t;

typedef struct _export_entry_t
{
	char			path[MAX_NAME_LEN+1];
	uint32			localIP;
	int				numClients;
	int				allocClients;
	export_host_t	*clients;
} export_entry_t;

static int entryCount;
static int dataCount;
static int dataCurr;
static export_host_t *stHostTable;
static export_entry_t *stNfsTable;
static export_entry_t stData;


/***********************************************************
 * routines to retrieve nfs information 
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_ACT:
			strcpy(bufact, argval[argind]);
			break;
		case PARAM_ID:
			sscanf(argval[argind], "%lld", &inId);
			break;
		case PARAM_NAME:
			strcpy(bufname, argval[argind]);
			break;
		case PARAM_PATH:
			strcpy(bufpath, argval[argind]);
			break;
		case PARAM_IP:
			inIP = GetIpAddressFromStr(argval[argind]);
			break;
		}
	}

	return 0;
}

static int AddHostEntry(export_entry_t* nfs, mnt_infoEntry_t *entry)
{
	export_host_t* newTable;
	int index;
	char *p, *tmpbuf;

	if (nfs->numClients >= nfs->allocClients)
	{
		newTable = realloc(nfs->clients, (nfs->allocClients+HOST_COUNT) * sizeof(export_host_t));
		if (newTable == NULL)
			return -1;
		nfs->allocClients += HOST_COUNT;
		nfs->clients = newTable;
	}

	index = nfs->numClients;

#if 0
/* SY: host IP is actually embedded in export info */
/* So, we can't use hostName below...			  */
/* The correct way (currently) to do this is in	*/
/* the #if 1, #endif block below...				*/
	/* host name */
	p = entry->pathName + entry->pathNameLen + entry->exportInfoLen;
	if (*p == '\0')
		return 0;

	strncpy(nfs->clients[index].hostName, p, entry->hostNameLen);
	nfs->clients[index].hostName[entry->hostNameLen] = '\0';

	GetIpAddressStr(entry->hostAddr, nfs->clients[index].hostIP);
#endif

	nfs->clients[index].subnetMask[0] = '\0';

	/* export info */
	tmpbuf = (char *) malloc(entry->exportInfoLen+1);
	if (tmpbuf)
	{
		p = entry->pathName + entry->pathNameLen;
		strncpy(tmpbuf, p, entry->exportInfoLen);
		tmpbuf[entry->exportInfoLen] = '\0';

		p = tmpbuf;
#if 1
/* Need to update in case Joseph update his mountd API */
		if (isdigit(*p))		/* start with digit, an IP addr */
		{
			while (isdigit(*p) || (*p == '.'))
				p++;
			if (*p == '(')
			{
				*p++ = '\0';
				strcpy(nfs->clients[index].hostName, tmpbuf);
				strcpy(nfs->clients[index].hostIP, tmpbuf);
			}
			else
			{
				/* something wrong with the data */
				p = tmpbuf;
			}
		}
		else if (*p == '(')
		{
			nfs->clients[index].hostName[0] = '\0';
			strcpy(nfs->clients[index].hostIP, "0.0.0.0");
		}
#endif

		nfs->clients[index].permission = GetNfsPermission(p);
		nfs->clients[index].rootSquash = GetNfsRootSquash(p);
		free(tmpbuf);
	}
	nfs->numClients++;
	return 1;
}

static int IsExist(export_entry_t* table, int count, mnt_infoEntry_t *entry)
{
	int i, res;
	char tmpbuf[1024];

	strncpy(tmpbuf, entry->pathName, entry->pathNameLen);
	tmpbuf[entry->pathNameLen] = '\0';

	for (i = 0; i < count; i ++)
	{
		if (!strcmp(table[i].path, tmpbuf) &&
			(table[i].localIP == entry->localIpAddr))
		{
			// Add Host Entry
			res = AddHostEntry(&table[i], entry);
			if (res < 0)
				return -1;
			return 1;
		}
	}
	return 0;
}

static int AddNfsEntry(int count, mnt_infoEntry_t *entry)
{
	export_entry_t* newTable;
	char *p, tmpbuf[512];

	if (count >= entryCount)
	{
		newTable = realloc(stNfsTable, (entryCount+EXPORT_COUNT) * sizeof(export_entry_t));
		if (newTable == NULL)
			return -1;
		entryCount += EXPORT_COUNT;
		stNfsTable = newTable;
	}
	
	strncpy(stNfsTable[count].path, entry->pathName, entry->pathNameLen);
	stNfsTable[count].path[entry->pathNameLen] = '\0';
//	stNfsTable[count].localIP = ntohl(entry->localIpAddr);
	stNfsTable[count].localIP = entry->localIpAddr;
	stNfsTable[count].numClients = 1;
	stNfsTable[count].allocClients = HOST_COUNT;
	stNfsTable[count].clients = malloc(HOST_COUNT * sizeof(export_host_t));
	if (stNfsTable[count].clients == NULL)
		return -1;

#if 0
/* Joseph's API doesn't have hostName and hostAddr... */
/* see below for the possible host address embedded in exportInfo */
	GetIpAddressStr(entry->hostAddr, stNfsTable[count].clients[0].hostIP);
fprintf(stderr, "  hostIp=%x\n", entry->hostAddr);

	p = entry->pathName + entry->pathNameLen + entry->exportInfoLen;
	strncpy(stNfsTable[count].clients[0].hostName, p, entry->hostNameLen);
	stNfsTable[count].clients[0].hostName[entry->hostNameLen] = '\0'; 
fprintf(stderr, "  len=%d hostName=%s\n", entry->hostNameLen, stNfsTable[count].clients[0].hostName);
#endif

	stNfsTable[count].clients[0].subnetMask[0] = '\0';

	p = entry->pathName + entry->pathNameLen;
	strncpy(tmpbuf, p, entry->exportInfoLen);
	tmpbuf[entry->exportInfoLen]='\0';

	p = tmpbuf;

#if 1
/* Need to update in case Joseph update his mountd API */
	if (isdigit(*p))		/* start with digit, an IP addr */
	{
		while (isdigit(*p) || (*p == '.'))
			p++;
		if (*p == '(')
		{
			*p++ = '\0';
			strcpy(stNfsTable[count].clients[0].hostName, tmpbuf);
			strcpy(stNfsTable[count].clients[0].hostIP, tmpbuf);
		}
		else
		{
			/* something wrong with the data */
			p = tmpbuf;
		}
	}
	else if (*p == '(')
	{
		stNfsTable[count].clients[0].hostName[0] = '\0';
		strcpy(stNfsTable[count].clients[0].hostIP, "0.0.0.0");
	}
#endif

	stNfsTable[count].clients[0].permission = GetNfsPermission(p);
	stNfsTable[count].clients[0].rootSquash = GetNfsRootSquash(p);
	return 1;
}

static int getTableRowCount(void)
{
	int					stat, i, res, count, addhostres;
	char 				respbuff[4096];
	mnt_infoReq_t   	req;
	mnt_infoResp_t		*rsp;
	struct timeval		timeout = {5, 0};
	mnt_infoEntry_t		*mtnInfo;

	stNfsTable = malloc(EXPORT_COUNT * sizeof(export_entry_t));
	if (stNfsTable == NULL)
		return -1;
	
	entryCount = EXPORT_COUNT;

	/* Fill the request. */
	memset(&req, 0, sizeof(mnt_infoReq_t));

	req.msgHdr.msgType = MNT_EXPORT_SHOW;
	req.volId.whole_thing = inId;
	req.nameLen = 0;		// All Paths for the given volume
	req.cookie = 0;	/* this is the starting index */

	count = 0;
	while (1)
	{
		memset(respbuff, 0, sizeof(respbuff));

		/* Send message to FS. */
		stat = sendAgileMsg(-1, &req, sizeof(req),
   						 "mountd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &timeout,
   						 respbuff, sizeof(respbuff));

		/* Check response from FS. */
		if (stat <= 0) 
			return -1;

		rsp = (mnt_infoResp_t *) respbuff;
		if (rsp->msgHdr.status != NFX_OK)
			return -1;

		/* no more entries? */
		if ((rsp->numOfEntries == 0) ||
			(req.cookie == rsp->cookie))
			break;

		/* 
		Note: need the following handling because mountd does not return
		hosts per path. Should remove the following extra handling code
		when there is a better API from mountd.
		*/

		/* Construct export_entry_t table, also calculate num of entries */
		mtnInfo = rsp->mtnInfoEntry;
		for (i = 0; i < rsp->numOfEntries; i ++)
		{
			addhostres = IsExist(stNfsTable, count, mtnInfo);
			if (addhostres == 0)
			{
				res = AddNfsEntry(count, mtnInfo);
				if (res <= 0)
				{
					free(stNfsTable);
					stNfsTable = NULL;
					count = 0;
					return -1;
				}
				count++;
			}
			else if (addhostres == -1)
			{
				free(stNfsTable);
				stNfsTable = NULL;
				count = 0;
				return -1;
			}

			MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfo);
		}

		req.cookie = rsp->cookie;
	}
	return count;
}

static int getTableRowData(int *index, export_entry_t *data)
{
	if (*index < 1 || *index > dataCount)
		return -1;
	if (stNfsTable == NULL)
		return -1;

	memcpy(data, &stNfsTable[(*index)-1], sizeof(export_entry_t));
	stHostTable = data->clients;
	return 0;
}


/***********************************************************
 *  Main routines to handle Add/Del/Mod actions
 **********************************************************/

static int DoDelShare()
{
	int					stat, i;
	char				reqbuff[1024];
	char				respbuff[4096];
	mnt_infoReq_t		*req;
	mnt_infoResp_t		*rsp;
	mnt_infoResp_t		unsh_rsp;
	struct timeval		timeout = {5, 0};
	mnt_infoEntry_t		*mtnInfo;

	while (1)
	{
		/* Fill the request. */
		memset(reqbuff, 0, sizeof(reqbuff));
		req = (mnt_infoReq_t *) reqbuff;
		req->msgHdr.msgType = MNT_EXPORT_SHOW;
		req->volId.whole_thing = inId;
		req->nameLen = 0;		// All Paths for the given volume
		req->cookie = 0;

		memset(respbuff, 0, sizeof(respbuff));	/* rsp point to this */
		rsp = (mnt_infoResp_t *) respbuff;

		/* Send message to FS. */
		stat = sendAgileMsg(-1, req, sizeof(mnt_infoReq_t),
							"mountd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &timeout,
							respbuff, sizeof(respbuff));

		/* Check response from FS. */
		if (stat <= 0) 
			return -1;

		rsp = (mnt_infoResp_t *) respbuff;
		if (rsp->msgHdr.status != NFX_OK)
			break;

		/* no more entries? */
		if ((rsp->numOfEntries == 0) ||
			(req->cookie == rsp->cookie))
			break;

		/* now we have all share information */
		/* need to pick out only the shares we want to unshare */

		mtnInfo = rsp->mtnInfoEntry;
		for (i = 0; i < rsp->numOfEntries; i++)
		{
			char *p, tmpbuf[512];

			strncpy(tmpbuf, mtnInfo->pathName, mtnInfo->pathNameLen);
			tmpbuf[mtnInfo->pathNameLen] = '\0';

			/* does this entry have same path and local ip? */
			if (!strcmp(tmpbuf, &bufpath[1]) &&
				(inIP == mtnInfo->localIpAddr) )
			{
				/* Yup, we need to unshare this entry */
				char clientip[32];
				int j = 0;

				/* find the client first... */
				p = mtnInfo->pathName + mtnInfo->pathNameLen;

				while (isdigit(*p) || (*p == '.'))
					clientip[j++] = *p++;
				clientip[j] = '\0';

				if (j == 0)
					strcpy(tmpbuf, bufpath);
				else
					sprintf(tmpbuf, "%s:%s", clientip, bufpath);

				/* Fill the request. */
				memset((void *) reqbuff, 0, sizeof(reqbuff));
				req = (mnt_infoReq_t *) reqbuff;
				req->msgHdr.msgType = MNT_UNEXPORT_REQ;
				req->volId.whole_thing = inId;
				req->nameLen = strlen(tmpbuf);
				strcpy(req->pathName, tmpbuf);

				if (inIP)
					req->localIpAddr = inIP;

				/* Send message to FS. */
				stat = sendAgileMsg(-1,
						reqbuff, sizeof(mnt_infoReq_t)+req->nameLen+1,
						"mountd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &timeout,
						&unsh_rsp, sizeof(unsh_rsp));
				/* don't care about the result as we */
				/* still need to do other clients	*/
			}
			MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfo);
		}
	}

	return 1;
}

static int DoAddShare()
{
	// Handled in agilenfsadd.c
	return 1;
}

static int DoModShare()
{
	// Handled in agilenfsadd.c
	return 1;
}

static int HandleNfsAction(char *newfname, int count, char **argname, char **argval)
{
	int res = 1;
	getInputParams(count, argname, argval);

	// Add Share
	if (!strcmp(bufact, "add"))
	{
		res = DoAddShare();
		if (res <= 0)
			strcpy(newfname, "error_addnfs.html");
	}
	// Unshare
	else if (!strcmp(bufact, "del"))
	{
		res = DoDelShare();
		if (res <= 0)
			strcpy(newfname, "error_delnfs.html");
	}
	// Modify Share
	else if (!strcmp(bufact, "mod"))
	{
		res = DoModShare();
		if (res <= 0)
			strcpy(newfname, "error_modnfs.html");
	}
	return res;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static void GetPermossionStr(int perm, char* buf)
{
	if (perm == 0)
		strcpy(buf, "RO");
	else if (perm == 1)
		strcpy(buf, "RW");
	else
		strcpy(buf, "Unknown");
}

static void GetRootSquashStr(int rs, char* buf)
{
	if (rs == 0)
		strcpy(buf, "No");
	else if (rs == 1)
		strcpy(buf, "Yes");
	else
		strcpy(buf, "Unknown");
}

static int getTagValue(tagCallbackParams_t *tp)
{
	void *req = tp->req;
	char *tag = tp->tag;
	char *value = tp->value;

	int i, tagType, pri;
	char tempbuf[256];
	
	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case NFS_PRIVCOL:
		pri = GetUserPrivilege();
		if (pri >= PRIV_FULL)	// del, mod
			strcpy(value, "2");
		else
			strcpy(value, "0");
		return 1;
	case NFS_VOLID:
		sprintf(value, "%lld", inId);
		return 1;
	case NFS_VOLNAME:
		strcpy(value, bufname);
		return 1;
	}

	if ((dataCurr <= 0) || (dataCurr > dataCount))
		return -1;

	switch (tagType) 
	{
	case NFS_PATH:
		sprintf(value, "/%s", stData.path);
		// strcpy(value, stData.path);
		// convertSpecCharToUrlChar(value);
		break;
	case NFS_IP:
		GetIpAddressStr(stData.localIP, value);
		break;
	case NFS_HOSTALL:
		if (stData.numClients <= 0)
		{
			strcpy(value, "<tr><td class=\"tabledata\" nowrap>&nbsp;</td></tr>\n");
		}
		else
		{
			tp->bytesAlreadyDumped = 0;
			strcpy(value, "");
			for(i = 0; i < stData.numClients; i ++)
			{
				int valsize;
				
				strcat(value, "<tr><td class=\"tabledata\" nowrap>");
				strcat(value, stHostTable[i].hostIP);
				if (stHostTable[i].subnetMask[0] != '\0')
				{
					strcat(value, "/");
					strcat(value, stHostTable[i].subnetMask);
				}
				strcat(value, "</td></tr>\n");

				valsize = strlen(value);
				if (valsize > (tp->valueSize >> 1))
				{
					if (dump_to_client(req, value, valsize) < 0)
					{
						// connection closed, return immediately
						strcpy(value, "");
						return 1;

					}
					tp->bytesAlreadyDumped += valsize;
					strcpy(value, "");
				}
			}
		}
		break;
	case NFS_PERMALL:
		if (stData.numClients <= 0)
		{
			strcpy(value, "<tr><td class=\"tabledata\" nowrap>&nbsp;</td></tr>\n");
		}
		else
		{
			tp->bytesAlreadyDumped = 0;
			strcpy(value, "");
			for(i = 0; i < stData.numClients; i ++)
			{
				int valsize;
				
				strcat(value, "<tr><td class=\"tabledata\" nowrap>");
				GetPermossionStr(stHostTable[i].permission, tempbuf);
				strcat(value, tempbuf);
				strcat(value, "</td></tr>\n");

				valsize = strlen(value);
				if (valsize > (tp->valueSize >> 1))
				{
					if (dump_to_client(req, value, valsize) < 0)
					{
						// connection closed, return immediately
						strcpy(value, "");
						return 1;

					}
					tp->bytesAlreadyDumped += valsize;
					strcpy(value, "");
				}
			}
		}
		break;
	case NFS_RSALL:
		if (stData.numClients <= 0)
		{
			strcpy(value, "<tr><td class=\"tabledata\" nowrap>&nbsp;</td></tr>\n");
		}
		else
		{
			strcpy(value, "");
			for(i = 0; i < stData.numClients; i++)
			{
				int valsize;
				
				strcat(value, "<tr><td class=\"tabledata\" nowrap>");
				GetRootSquashStr(stHostTable[i].rootSquash, tempbuf);
				strcat(value, tempbuf);
				strcat(value, "</td></tr>\n");

				valsize = strlen(value);
				if (valsize > (tp->valueSize >> 1))
				{
					if (dump_to_client(req, value, valsize) < 0)
					{
						// connection closed, return immediately
						strcpy(value, "");
						return 1;

					}
					tp->bytesAlreadyDumped += valsize;
					strcpy(value, "");
				}
			}
		}
		break;
	default:
		return -1;
	}

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		/* TABLE_BEGIN tag, need to see if there is any data */
		dataCount = getTableRowCount();
		if (dataCount > 0) {
			/* need to retrieve first table's data */
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) >= 0)
				return 1;
		}

		dataCurr = 0;
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		/* TABLE_END tag, check if end of data */
		if ((dataCount <= 0) || (dataCurr <= 0) || (dataCurr >= dataCount))
			return -1;

		dataCurr++;
		if (getTableRowData(&dataCurr, &stData) < 0)
			return -1;
		else
			return 1;
	}
	else {
		int rc;
		/* other tags */
		rc = getTagValue(tp);
		return rc;
	}

	/* shouldn't reach here... */
	return -1;
}


int actionNfsMain(void *r, int count, char **argname, char **argval)
{
	char newfname[128];

	inId = 0;
	inIP = 0;
	bufact[0] = '\0';
	bufname[0] = '\0';
	bufpath[0] = '\0';
	memset(newfname, 0, sizeof(newfname));

	/* Handle Add/Del/Mod actions */
	if (count > 0)
		HandleNfsAction(newfname, count, argname, argval);

	prepare_agile_template(r, newfname);

	/* initialize nfs information variables */
	dataCount = 0;
	dataCurr = 0;
	stNfsTable = NULL;
	stHostTable = NULL;

	agile_scan_and_send(r, actRptTagFunc);

	if (stNfsTable != NULL)
	{
		free(stNfsTable);
		stNfsTable = NULL;
	}
	return 1;
}
