/*  File Name  :    agilenwconnfp.c
 *
 *  Description:    HTTPD action functions for displaying TCP/UDP connections for FP.
 *
 *  Created    :    3/1/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>

#include <sys/socket.h>
#include <net/route.h>
#include <netinet/tcp_fsm.h>
#include <ipmd-lib.h>
#include "agilepost.h"
#include "agileutil.h"


#define INVALID_TAG			0
#define NWCO_TRANGE			1
#define NWCO_TPREVBTN		2
#define NWCO_TNEXTBTN		3
#define NWCO_TLADDR			4
#define NWCO_TLPORT			5
#define NWCO_TRADDR			6
#define NWCO_TRPORT			7
#define NWCO_TSTATE			8
#define NWCO_URANGE			9
#define NWCO_UPREVBTN		10
#define NWCO_UNEXTBTN		11
#define NWCO_ULADDR			12
#define NWCO_ULPORT			13
#define NWCO_URADDR			14
#define NWCO_URPORT			15
#define NWCO_USTATE			16
#define NUM_OF_TAGS			16

#define NUM_MAXIPAGE_ENTRIES	40
#define NUM_PAGE_ENTRIES		20

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{NWCO_TRANGE,	"tcp.range"},
	{NWCO_TPREVBTN,	"tcp.prevbtn"},
	{NWCO_TNEXTBTN,	"tcp.nextbtn"},
	{NWCO_TLADDR,	"tcp.ipm_conn_t.conn_lAddr"},
	{NWCO_TLPORT,	"tcp.ipm_conn_t.conn_lPort"},
	{NWCO_TRADDR,	"tcp.ipm_conn_t.conn_rAddr"},
	{NWCO_TRPORT,	"tcp.ipm_conn_t.conn_rPort"},
	{NWCO_TSTATE,	"tcp.ipm_conn_t.conn_state"},
	{NWCO_URANGE,	"udp.range"},
	{NWCO_UPREVBTN,	"udp.prevbtn"},
	{NWCO_UNEXTBTN,	"udp.nextbtn"},
	{NWCO_ULADDR,	"udp.ipm_conn_t.conn_lAddr"},
	{NWCO_ULPORT,	"udp.ipm_conn_t.conn_lPort"},
	{NWCO_URADDR,	"udp.ipm_conn_t.conn_rAddr"},
	{NWCO_URPORT,	"udp.ipm_conn_t.conn_rPort"},
	{NWCO_USTATE,	"udp.ipm_conn_t.conn_state"},
};


#define PARAM_TSTART		1
#define PARAM_TCOUNT		2
#define PARAM_USTART		3
#define PARAM_UCOUNT		4
#define NUM_OF_PARAMS		4

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_TSTART,	"tcpstart"},
	{PARAM_TCOUNT,	"tcpcount"},
	{PARAM_USTART,	"udpstart"},
	{PARAM_UCOUNT,	"udpcount"},
};


static char tcpstates[TCP_NSTATES][32] =
{
	{"closed"},
	{"listen"},
	{"syncSent"},
	{"syncReceived"},
	{"established"},
	{"closeWait"},
	{"finWait1"},
	{"closing"},
	{"lastAck"},
	{"finWait2"},
	{"timeWait"}
};


/* HERE ARE VARIABLES FOR CONNECTION LIST INFORMATION */
static int dataCount;
static int dataCurr;
static ipm_conn_t *stTcpConnTable;	// table contains entries for current page
static ipm_conn_t *stUdpConnTable;	// table contains entries for current page
static ipm_conn_t stData;

static int tcpPageRowCount;
static int udpPageRowCount;
static int tcpStartIndex;
static int udpStartIndex;
static int tcpEndIndex;
static int udpEndIndex;
static int tcpTotalEntries;
static int udpTotalEntries;


/***********************************************************
 * routines to find entries for current page
 * also handle input parameters
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_TSTART:
			sscanf(argval[argind], "%d", &tcpStartIndex);
			break;
		case PARAM_TCOUNT:
			sscanf(argval[argind], "%d", &tcpPageRowCount);
			break;
		case PARAM_USTART:
			sscanf(argval[argind], "%d", &udpStartIndex);
			break;
		case PARAM_UCOUNT:
			sscanf(argval[argind], "%d", &udpPageRowCount);
			break;
		}
	}

	return 0;
}

static int getTcpTotalRowCount()
{
	int32 res;
	int count, need;
	ipm_conn_t sscconn;

	// get system wide Tcp count
	count = 0;
	res = ipmd_GetTcpConnTable(&sscconn, sizeof(ipm_conn_t), 
							0, 10000, &count, &need);
	if (res == IPM_GET_ERR)
	{
		return 0;
	}
	else if ((res == IPM_GET_NO_BUF) || (res == IPM_CMDERR_APPNOBUF))
	{
		count = need / sizeof(ipm_conn_t);
	}
	return count;
}

static int getUdpTotalRowCount()
{
	int32 res;
	int count, need;
	ipm_conn_t sscconn;

	// get system wide Udp count
	count = 0;
	res = ipmd_GetUdpConnTable(&sscconn, sizeof(ipm_conn_t), 
							0, 10000, &count, &need);
	if (res == IPM_GET_ERR)
	{
		return 0;
	}
	else if ((res == IPM_GET_NO_BUF) || (res == IPM_CMDERR_APPNOBUF))
	{
		count = need / sizeof(ipm_conn_t);
	}
	return count;
}

static void getTcpPageTable()
{
	int32	res;
	int count, need;

	stTcpConnTable = (ipm_conn_t *) malloc(tcpPageRowCount * sizeof(ipm_conn_t));
	if (stTcpConnTable == NULL)
		return;

	res = ipmd_GetTcpConnTable(stTcpConnTable, tcpPageRowCount*sizeof(ipm_conn_t),
							tcpStartIndex, tcpPageRowCount, &count, &need);
	if ((res == IPM_GET_ERR) || (res == IPM_GET_NO_BUF) || (res == IPM_CMDERR_APPNOBUF))
	{
		free(stTcpConnTable);
		stTcpConnTable = NULL;
		tcpPageRowCount = 0;
		tcpStartIndex = -1;
	}
}

static void getUdpPageTable()
{
	int32	res;
	int count, need;

	stUdpConnTable = (ipm_conn_t *) malloc(udpPageRowCount * sizeof(ipm_conn_t));
	if (stUdpConnTable == NULL)
		return;

	res = ipmd_GetUdpConnTable(stUdpConnTable, udpPageRowCount*sizeof(ipm_conn_t),
							udpStartIndex, udpPageRowCount, &count, &need);
	if ((res == IPM_GET_ERR) || (res == IPM_GET_NO_BUF) || (res == IPM_CMDERR_APPNOBUF))
	{
		free(stUdpConnTable);
		stUdpConnTable = NULL;
		udpPageRowCount = 0;
		udpStartIndex = -1;
	}
}

static void	HandleDisplayPage(char *newfname, int count, char **argname, char **argval)
{
	tcpStartIndex = -1;
	tcpPageRowCount = -1;
	udpStartIndex = -1;
	udpPageRowCount = -1;

	if (count > 0)
	{
		getInputParams(count, argname, argval);
	}

	//------------------------- TCP -----------------------------------------
	// get total tcp count
	tcpTotalEntries = getTcpTotalRowCount();

	// get start index and row count
	if (tcpStartIndex < 0)
	{
		tcpStartIndex = 0;
		if (tcpTotalEntries > NUM_MAXIPAGE_ENTRIES)
			tcpPageRowCount = NUM_PAGE_ENTRIES;
		else
			tcpPageRowCount = tcpTotalEntries;
	}
	tcpEndIndex = tcpStartIndex + tcpPageRowCount - 1;
	if (tcpEndIndex >= tcpTotalEntries)
		tcpEndIndex = tcpTotalEntries - 1;

	// get row data for page table
	getTcpPageTable();

	//------------------------- UDP -----------------------------------------
	// get total tcp count
	udpTotalEntries = getUdpTotalRowCount();

	// get start index and row count
	if (udpStartIndex < 0)
	{
		udpStartIndex = 0;
		if (udpTotalEntries > NUM_MAXIPAGE_ENTRIES)
			udpPageRowCount = NUM_PAGE_ENTRIES;
		else
			udpPageRowCount = udpTotalEntries;
	}
	udpEndIndex = udpStartIndex + udpPageRowCount - 1;
	if (udpEndIndex >= udpTotalEntries)
		udpEndIndex = udpTotalEntries - 1;

	// get row data for page table
	getUdpPageTable();
}


/***********************************************************
 * routines to retrieve network information 
 **********************************************************/

static int getTcpTableRowCount(void)
{
	return tcpPageRowCount;
}

static int getTcpTableRowData(int *index, ipm_conn_t *data)
{
	if (*index < 1 || *index > dataCount)
		return -1;
	if (stTcpConnTable == NULL)
		return -1;
	
	memcpy(data, &stTcpConnTable[(*index)-1], sizeof(ipm_conn_t));
	return 0;
}

static int getUdpTableRowCount(void)
{
	return udpPageRowCount;
}

static int getUdpTableRowData(int *index, ipm_conn_t *data)
{
	if (*index < 1 || *index > dataCount)
		return -1;
	if (stUdpConnTable == NULL)
		return -1;
	
	memcpy(data, &stUdpConnTable[(*index)-1], sizeof(ipm_conn_t));
	return 0;
}

/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType, prev, next;
	
	if ((dataCurr <= 0) || (dataCurr > dataCount))
		return -1;

	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case NWCO_TLADDR:
	case NWCO_ULADDR:
		GetIpAddressStr(stData.conn_lAddr, value);
		break;
	case NWCO_TLPORT:
	case NWCO_ULPORT:
		sprintf(value, "%d", stData.conn_lPort);
		break;
	case NWCO_TRADDR:
	case NWCO_URADDR:
		GetIpAddressStr(stData.conn_rAddr, value);
		break;
	case NWCO_TRPORT:
	case NWCO_URPORT:
		sprintf(value, "%d", stData.conn_rPort);
		break;
	case NWCO_TSTATE:
	case NWCO_USTATE:
		strcpy(value, tcpstates[stData.conn_state]);
		break;
	case NWCO_TRANGE:
		if (tcpPageRowCount >= tcpTotalEntries)
			strcpy(value, "\0");
		else
			sprintf(value, ": %d - %d of %d", tcpStartIndex, tcpEndIndex, tcpTotalEntries);
		break;
	case NWCO_URANGE:
		if (udpPageRowCount >= udpTotalEntries)
			strcpy(value, "\0");
		else
			sprintf(value, ": %d - %d of %d", udpStartIndex, udpEndIndex, udpTotalEntries);
		break;
	case NWCO_TPREVBTN:
		if (tcpStartIndex > 0)
		{
			prev = tcpStartIndex - NUM_PAGE_ENTRIES;
			if (prev < 0)
				prev = 0;
			sprintf(value, "<a href=\"/agilerequestprefix/NetworkConnFP?tcpstart=%d&tcpcount=%d&udpstart=%d&udpcount=%d\">",
					prev, NUM_PAGE_ENTRIES, udpStartIndex, udpPageRowCount);
			strcat(value, "<img src=\"/AgileView/images/button_prev.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	case NWCO_UPREVBTN:
		if (udpStartIndex > 0)
		{
			prev = udpStartIndex - NUM_PAGE_ENTRIES;
			if (prev < 0)
				prev = 0;
			sprintf(value, "<a href=\"/agilerequestprefix/NetworkConnFP?tcpstart=%d&tcpcount=%d&udpstart=%d&udpcount=%d\">",
					tcpStartIndex, tcpPageRowCount, prev, NUM_PAGE_ENTRIES);
			strcat(value, "<img src=\"/AgileView/images/button_prev.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	case NWCO_TNEXTBTN:
		if (tcpEndIndex < (tcpTotalEntries-1))
		{
			next = tcpEndIndex + 1;
			sprintf(value, "<a href=\"/agilerequestprefix/NetworkConnFP?tcpstart=%d&tcpcount=%d&udpstart=%d&udpcount=%d\">",
					next, NUM_PAGE_ENTRIES, udpStartIndex, udpPageRowCount);
			strcat(value, "<img src=\"/AgileView/images/button_next.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	case NWCO_UNEXTBTN:
		if (udpEndIndex < (udpTotalEntries-1))
		{
			next = udpEndIndex + 1;
			sprintf(value, "<a href=\"/agilerequestprefix/NetworkConnFP?tcpstart=%d&tcpcount=%d&udpstart=%d&udpcount=%d\">",
					tcpStartIndex, tcpPageRowCount, next, NUM_PAGE_ENTRIES);
			strcat(value, "<img src=\"/AgileView/images/button_next.gif\" width=\"78\" height=\"20\" border=0></a>");
		}
		else
			strcpy(value, "\0");
		break;
	default:
		return -1;
	}

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN_TCP")) {
		/* TABLE_BEGIN tag, need to see if there is any data */
		dataCount = getTcpTableRowCount();
		if (dataCount > 0) {
			/* need to retrieve first table's data */
			dataCurr = 1;
			if (getTcpTableRowData(&dataCurr, &stData) >= 0)
				return 1;
		}

		dataCurr = 0;
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END_TCP")) {
		/* TABLE_END tag, check if end of data */
		if ((dataCount <= 0) || (dataCurr <= 0) || (dataCurr >= dataCount))
			return -1;

		dataCurr++;
		if (getTcpTableRowData(&dataCurr, &stData) < 0)
			return -1;
		else {
			return 1;
		}
	}
	else if (!strcmp(tag, "TABLE_BEGIN_UDP")) {
		/* TABLE_BEGIN tag, need to see if there is any data */
		dataCount = getUdpTableRowCount();
		if (dataCount > 0) {
			/* need to retrieve first table's data */
			dataCurr = 1;
			if (getUdpTableRowData(&dataCurr, &stData) >= 0)
				return 1;
		}

		dataCurr = 0;
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END_UDP")) {
		/* TABLE_END tag, check if end of data */
		if ((dataCount <= 0) || (dataCurr <= 0) || (dataCurr >= dataCount))
			return -1;

		dataCurr++;
		if (getUdpTableRowData(&dataCurr, &stData) < 0)
			return -1;
		else {
			return 1;
		}
	}
	else {
		/* other tags */
		/*
		if (dataCount <= 0) {
			dataCount = getTableRowCount();
			if (dataCount <= 0)
				return -1;
		}

		if ((dataCurr < 1) || (dataCurr > dataCount)) {
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) < 0) {
				return -1;
			}
		}
		*/

		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionNwConnFp(void *r, int count, char **argname, char **argval)
{
char newfname[128];

	memset(newfname, 0, sizeof(newfname));

	/* Find entries for the current displaying page */
	stTcpConnTable = NULL;
	stUdpConnTable = NULL;
	HandleDisplayPage(newfname, count, argname, argval);

	prepare_agile_template(r, newfname);

	/* initialize network information variables */
	dataCount = 0;
	dataCurr = 0;

	agile_scan_and_send(r, actRptTagFunc);

	if (stTcpConnTable != NULL)
		free(stTcpConnTable);
	if (stUdpConnTable != NULL)
		free(stUdpConnTable);
	return 1;
}
