/*
 *  UDP MIB group implementation - udp.c
 *
 */

#include <config.h>

#include <string.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <sys/queue.h>
#include <sys/socketvar.h>
#include <netinet/ip_var.h>
#include <netinet/in_pcb.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>

#include "tools.h"
#include "kernel.h"

#include "mibincl.h"
#include "util_funcs.h"
#include "system.h"
#include "asn1.h"
#include "snmp_debug.h"
#include "tools.h"

#include "auto_nlist.h"
#include "udp.h"
#include "udpTable.h"
#include "sysORTable.h"

#include <sys/sysctl.h>

#include <ipmd-lib.h>
#include <snmpmsgapi.h>

#include "ipifhelp.h"

#define UDP_STATS_CACHE_TIMEOUT	MIB_STATS_CACHE_TIMEOUT
marker_t udp_stats_cache_marker = NULL;

// static struct timeval msg_timeout = {3, 0};  // We'll wait 3 sec


	/*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/


	/*********************
	 *
	 *  Initialisation & common implementation functions
	 *
	 *********************/

struct variable8 udp_variables[] = {
    {UDPINDATAGRAMS, ASN_COUNTER, RONLY, var_udp, 1, {1}},
    {UDPNOPORTS, ASN_COUNTER, RONLY, var_udp, 1, {2}},
    {UDPINERRORS, ASN_COUNTER, RONLY, var_udp, 1, {3}},
    {UDPOUTDATAGRAMS, ASN_COUNTER, RONLY, var_udp, 1, {4}},
    {UDPLOCALADDRESS, ASN_IPADDRESS, RONLY, var_udpEntry, 3, {5, 1, 1}},
    {UDPLOCALPORT, ASN_INTEGER, RONLY, var_udpEntry, 3, {5, 1, 2}}
};

/* Define the OID pointer to the top of the mib tree that we're
   registering underneath, and the OID for the MIB module */
oid udp_variables_oid[] = { SNMP_OID_MIB2,7 };
oid udp_module_oid[]    = { SNMP_OID_MIB2,50 };

void init_udp(void)
{

  /* register ourselves with the agent to handle our mib tree */
  REGISTER_MIB("mibII/udp", udp_variables, variable8, udp_variables_oid);
  REGISTER_SYSOR_ENTRY( udp_module_oid,
		"The MIB module for managing UDP implementations");
}


	/*********************
	 *
	 *  System specific implementation functions
	 *
	 *********************/

long read_udp_stat(ipm_udpstat_t *, int);

u_char *
var_udp(struct variable *vp,
	oid *name,
	size_t *length,
	int exact,
	size_t *var_len,
	WriteMethod **write_method)
{
    static ipm_udpstat_t udpstat;
    static long ret_value;

    if (header_generic(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
		return NULL;

    ret_value = read_udp_stat(&udpstat, vp->magic);
    if ( ret_value < 0 )
		return NULL;

    switch (vp->magic) {
	case UDPINDATAGRAMS:
		ret_value = udpstat.udps_ipackets;
		return (u_char *) &ret_value;
	case UDPNOPORTS:
		ret_value = udpstat.udps_noport;
		return (u_char *) &ret_value;
	case UDPOUTDATAGRAMS:
		ret_value = udpstat.udps_hdrops + udpstat.udps_badsum +
						udpstat.udps_fullsock + udpstat.udps_badlen;
		return (u_char *) &ret_value;
	case UDPINERRORS:
		ret_value = udpstat.udps_opackets;
		return (u_char *) &ret_value;

	default:
	    DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_udp\n", vp->magic));
    }
    return NULL;
}


	/*********************
	 *
	 *  Internal implementation functions
	 *
	 *********************/


long
read_udp_stat(ipm_udpstat_t *udpstat, int magic)
{
    long ret_value = -1;
	int gbeslots[32], numofgbeslots;
	ipm_udpstat_t tmpstat;
	struct udpstat sscudpstat;
	int count, needed;

    if (udp_stats_cache_marker &&
		(!atime_ready( udp_stats_cache_marker, UDP_STATS_CACHE_TIMEOUT*1000)))
		return 0;

    if (udp_stats_cache_marker)
		atime_setMarker(udp_stats_cache_marker);
    else
		udp_stats_cache_marker = atime_newMarker();

	/* Get UDP stats for SSC ports */
	if (auto_nlist(UDPSTAT_SYMBOL, (char *)&sscudpstat, sizeof (*udpstat)))
		ret_value = 0;
	else
		goto finished_readudpstat;

	/* Get UDP stats for GbE ports */
	numofgbeslots = 32;
	if (IpInterface_Get_GbE_Slot(gbeslots, &numofgbeslots, NULL) >= 0) {
		int rval = -1;

		/* for now, we assume we have only one GbE slot */
		count = 1;
		needed = 0;
		rval = ipmd_GetUdpStats(gbeslots[0],
				&tmpstat, sizeof(ipm_udpstat_t), &count, &needed);
		if ((rval == IPM_GET_NO_BUF) || (rval == IPM_CMDERR_APPNOBUF)) {
			/* need to do something here.... */
		}
		else if (rval == IPM_GET_ERR) {
			/* something wrong here... */
		}
		else {
			/* assume there is only one GbE slot */
			/* if more than one, then we need to */
			/* add all the stats together. But   */
			/* we don't support that yet...      */
			memcpy(udpstat, &tmpstat, sizeof(ipm_udpstat_t));
			rval = 0;
		}
	}

	/* now combine stats from SSC and GbE */
	udpstat->udps_ipackets    += sscudpstat.udps_ipackets;
	udpstat->udps_hdrops      += sscudpstat.udps_hdrops;
	udpstat->udps_badsum      += sscudpstat.udps_badsum;
	udpstat->udps_badlen      += sscudpstat.udps_badlen;
	udpstat->udps_noport      += sscudpstat.udps_noport;
	udpstat->udps_noportbcast += sscudpstat.udps_noportbcast;
	udpstat->udps_fullsock    += sscudpstat.udps_fullsock;
	udpstat->udps_pcbhashmiss += sscudpstat.udps_pcbhashmiss;
	udpstat->udps_opackets    += sscudpstat.udps_opackets;

finished_readudpstat:
    if (ret_value == -1) {
		free(udp_stats_cache_marker);
		udp_stats_cache_marker = NULL;
    }
    return ret_value;
}
