/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2002, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc-encdec.c,v 1.1.1.1 2002/04/23 17:47:36 laurentb Exp $
 * 
 * Name			: cifs-rpc-encdec.c
 *
 * Description	: Contains CIFS DCE/RPC encode & decode tools. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/04/2002
 *
 *-----------------------------------------------------------------
 */

#ifdef FS_STANDALONE
#include "../sm-cifs/cifs-api.h"
#include "../sm-cifs/cifs-utils.h"
#include "../sm-cifs/cifs-unicode.h"
#else
#include <exports/cifs-api.h>
#include <exports/cifs-utils.h>
#include <exports/cifs-unicode.h>
#endif

#include "cifs-rpc.x"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)
#define DBGPRINT    printf


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDinit()
 *
 * Description  : CIFS RPC encode / decode init
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/04/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDinit(cifsrpc_ed *edp, 
				boolean encoding,
				uint32 size,
				cifsrpc_memList *mlp)
{

	memset(edp, 0, sizeof(cifsrpc_ed));
	
	edp->enc   = encoding;
	edp->align = CIFSRPC_ALIGNT;
	edp->memLp = mlp;
	if (size)
	{
		edp->size = size;
		CIFS_RAM_ALLOC(edp->bufp, uchar8, size);
		if (edp->bufp == NULL)
		{
		   	CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpcEDinit: "
						"malloc of %d bytes failed", size));
			return (NFX_ERR);
		}
		edp->freeMem = TRUE;
	}
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDassignData()
 *
 * Description  : Assign data to an RPC encode / decode structure
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
void
cifsrpcEDassignData(	cifsrpc_ed	*edp, 
						uchar8		*datp,
						uint32		size,
						boolean		freeMem)
{
	edp->bufp	 = datp;
	edp->size	 = size;
	edp->freeMem = freeMem;
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDreleaseData()
 *
 * Description  : Release buffer data
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
void
cifsrpcEDreleaseData(cifsrpc_ed *edp)
{
	if (edp->freeMem)
		RELEASE(edp->bufp);
	edp->size	 = 0;
	edp->offset	 = 0;
	edp->freeMem = FALSE;
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDcheckMemAvailability()
 *
 * Description  : CIFS RPC check whether memory is available to encode
 * 					or decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDcheckMemAvailability(cifsrpc_ed *edp, uint32 size)
{
//	CIFS_PRINT(CIFS_DBGLV_DBG3) DBGPRINT("checkMemAvailability: "
//				"offset=%u, addsize=%u, size=%u", 
//				edp->offset, size, edp->size);
	
	if (edp->offset + size <= edp->size)
		return (NFX_OK);
	if (CIFSRPC_ENCODING(edp))
	{	//The PDU does not fit in the PDU fragment, we have to copy in a
		//buffer of the maximum PDU size we support
		CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsrpcEDcheckMemAvail: "
					"need more memory to encode"));
		if (edp->freeMem == TRUE)
		{	//We have already allocated the dynamic buffer of the max PDU size
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpcEDcheckMemAvail: "
						"no more memory to encode"));
		} else {
			char *bufp    = edp->bufp;
			uint32 offset = edp->offset;

			//Allocate a dynamic buffer of the maximum size
			if (cifsrpcEDinit(edp, TRUE, CIFSRPC_S_PDU, edp->memLp) != NFX_OK)
				return (NFX_ERR);
			//Copy what has already been encoded in the PDU fragment
			memcpy(edp->bufp, bufp, offset);
			//restore the offset
			edp->offset = offset;

			return (NFX_OK);
		}
	}
	return (NFX_ERR);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDalign()
 *
 * Description  : move the data offset to the next alignment boundary
 * 				  if possible
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDalign(cifsrpc_ed *edp)
{
	uint32 alignOffset;
	
	if (edp->align == 0)
		return (NFX_OK);
	alignOffset = edp->align - (edp->offset & (edp->align-1));
	if (alignOffset != edp->align)
	{
		if (cifsrpcEDcheckMemAvailability(edp, alignOffset) != NFX_OK)
			return (NFX_ERR);
		memset(edp->bufp+edp->offset, 0, alignOffset);
		edp->offset += alignOffset;
	}
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDaddData()
 *
 * Description  : Add data to an encode / decode structure
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/10/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDaddData(cifsrpc_ed *edp, char *datp, uint32 size)
{
	if (cifsrpcEDcheckMemAvailability(edp, size) != NFX_OK)
		return (NFX_ERR);
	memcpy(edp->bufp + edp->offset, datp, size);
	edp->offset += size;
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDsetOffset()
 *
 * Description  : set the offset of an encode / decode structure
 * 				  check if possible
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDsetOffset(cifsrpc_ed *edp, uint32 offset)
{
	if (offset > edp->offset
	 && cifsrpcEDcheckMemAvailability(edp, offset - edp->offset) != NFX_OK)
		return (NFX_ERR);
	
	edp->offset = offset;
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuchar8()
 *
 * Description  : CIFS RPC uchar8 encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuchar8(cifsrpc_ed *edp, uchar8 *uchar8p, char *label, uchar8 lv)
{
	if (cifsrpcEDcheckMemAvailability(edp, sizeof(uchar8)) != NFX_OK)
		return (NFX_ERR);
	
	if (CIFSRPC_ENCODING(edp))
		PCAR(edp->bufp, edp->offset, *uchar8p);
	else
		*uchar8p = GCAR(edp->bufp, edp->offset);
	
	CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s: %02x", cifs_spaces(lv), label, *uchar8p));

	edp->offset += sizeof(uchar8);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuchar8Ar()
 *
 * Description  : CIFS RPC uchar8 array encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuchar8Ar(cifsrpc_ed *edp, uchar8 *uchar8p, uint32 uchar8n, 
					uint32 string, char *label, uchar8 lv)
{
	if (cifsrpcEDcheckMemAvailability(edp, uchar8n) != NFX_OK)
		return (NFX_ERR);
	
	if (CIFSRPC_ENCODING(edp))
		memcpy(edp->bufp + edp->offset, uchar8p, uchar8n);
	else
		memcpy(uchar8p, edp->bufp + edp->offset, uchar8n);
	
	IFDBG(CIFS_DBGLV_DBG3)
	{
		DBGPRINT("%s%s: ", cifs_spaces(lv), label);
		if (string)
			DBGPRINT("%s\n", uchar8p);
		else
			memDumplight(uchar8p, uchar8n);
	}

	edp->offset += uchar8n;
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDushort16()
 *
 * Description  : CIFS RPC ushort16 encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDushort16(cifsrpc_ed *edp, ushort16 *ushort16p, char*label, uchar8 lv)
{
	if (cifsrpcEDcheckMemAvailability(edp, sizeof(ushort16)) != NFX_OK)
		return (NFX_ERR);
	
	if (CIFSRPC_ENCODING(edp))
		PUTXE2(edp->bufp, edp->offset, *ushort16p, edp->bigEndian);
	else
		*ushort16p = GETXE2(edp->bufp, edp->offset, edp->bigEndian);
	
	CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s: %04x", cifs_spaces(lv), label, *ushort16p));

	edp->offset += sizeof(ushort16);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDushort16Ar()
 *
 * Description  : CIFS RPC ushort16 array encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDushort16Ar(cifsrpc_ed *edp, ushort16 *ushort16p, uint32 ushort16n,
					uint32 string, char *label, ushort16 lv)
{
	uint32 i;
	char *datp;
	
	if (cifsrpcEDcheckMemAvailability(edp, sizeof(ushort16) * ushort16n)
					!= NFX_OK)
		return (NFX_ERR);
	
	datp = edp->bufp + edp->offset;
	if (CIFSRPC_ENCODING(edp))
	{
		if (edp->bigEndian)
		{
			for (i=0; i<ushort16n; i++)
				PUTBE2(datp, i*2, ushort16p[i]);
		} else {
			for (i=0; i<ushort16n; i++)
				PUTLE2(datp, i*2, ushort16p[i]);
		}
	} else {
		if (edp->bigEndian)
		{
			for (i=0; i<ushort16n; i++)
				ushort16p[i] = GETBE2(datp, i*2);
		} else {
			for (i=0; i<ushort16n; i++)
				ushort16p[i] = GETLE2(datp, i*2);
		}
	}
	
	IFDBG(CIFS_DBGLV_DBG3)
	{
		DBGPRINT("%s%s: ", cifs_spaces(lv), label);
		if (string) 
		{
			uniprintf("", ushort16p);
			printf("\n");
		}
		else
			memDumplight(ushort16p, sizeof(ushort16) * ushort16n);
	}

	edp->offset += sizeof(ushort16) * ushort16n;
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuint32()
 *
 * Description  : CIFS RPC uint32 encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuint32(cifsrpc_ed *edp, uint32 *uint32p, char *label, uchar8 lv)
{
	if (cifsrpcEDcheckMemAvailability(edp, sizeof(uint32)) != NFX_OK)
		return (NFX_ERR);
	
	if (CIFSRPC_ENCODING(edp))
		PUTXE4(edp->bufp, edp->offset, *uint32p, edp->bigEndian);
	else
		*uint32p = GETXE4(edp->bufp, edp->offset, edp->bigEndian);
	
	CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s: %08x", cifs_spaces(lv), label, *uint32p));

	edp->offset += sizeof(uint32);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuint32Ar()
 *
 * Description  : CIFS RPC uint32 array encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuint32Ar(cifsrpc_ed *edp, uint32 *uint32p, uint32 uint32n,	
					char *label, uint32 lv)
{
	uint32 i;
	char *datp;
	
	if (cifsrpcEDcheckMemAvailability(edp, sizeof(uint32) * uint32n)
					!= NFX_OK)
		return (NFX_ERR);
	
	datp = edp->bufp + edp->offset;
	if (CIFSRPC_ENCODING(edp))
	{
		if (edp->bigEndian)
		{
			for (i=0; i<uint32n; i++)
				PUTBE4(datp, i*2, uint32p[i]);
		} else {
			for (i=0; i<uint32n; i++)
				PUTLE4(datp, i*2, uint32p[i]);
		}
	} else {
		if (edp->bigEndian)
		{
			for (i=0; i<uint32n; i++)
				uint32p[i] = GETBE4(datp, i*2);
		} else {
			for (i=0; i<uint32n; i++)
				uint32p[i] = GETLE4(datp, i*2);
		}
	}
	
	IFDBG(CIFS_DBGLV_DBG3)
	{
		DBGPRINT("%s%s: ", cifs_spaces(lv), label);
		for (i=0; i<uint32n; i++)
			DBGPRINT("%u-", uint32p[i]);
		DBGPRINT("\n");
	}

	edp->offset += sizeof(uint32) * uint32n;
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDunistruc()
 *
 * Description  : CIFS RPC unicode structure encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/14/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuniStruc(cifsrpc_ed *edp, cifsrpc_uniStruc *sp, char *label, uchar8 lv)
{

    CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s (uniStruc)", cifs_spaces(lv), label));
	lv++;

    if (cifsrpcEDalign (edp)								!= NFX_OK
	 || cifsrpcEDuint32(edp, &sp->maxLen,  "maxLen ", lv)	!= NFX_OK
     || cifsrpcEDuint32(edp, &sp->unknown, "unknown", lv)	!= NFX_OK
     || cifsrpcEDuint32(edp, &sp->len,	   "length ", lv)	!= NFX_OK)
        return (NFX_ERR);

	if (sp->len != 0)
	{
		uint32 strSize = sp->len * sizeof(uni16);
		if (CIFSRPC_DECODING(edp))
		{
			sp->str = (uni16*)cifsrpc_allocMemItem(edp->memLp, strSize);
			if (sp->str == NULL)
				return (NFX_ERR);
		}
		return (cifsrpcEDushort16Ar(edp, sp->str, sp->len, 1, label, lv));
	}
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDenumHd()
 *
 * Description  : CIFS RPC enum handle encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/15/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDenumHd(cifsrpc_ed *edp, cifsrpc_enumHandle *ehp, char *label,uchar8 lv)
{

    CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s", cifs_spaces(lv), label));
	lv++;

    if (cifsrpcEDalign (edp)								!= NFX_OK
	 || cifsrpcEDuint32(edp, &ehp->handlep,	"handlep", lv)	!= NFX_OK
	 || (ehp->handlep 
	  && cifsrpcEDuint32(edp, &ehp->handle,	"handle ", lv)	!= NFX_OK))
        return (NFX_ERR);
	
	return (NFX_OK);
}



/*-----------------------------------------------------------------
 * Name         : cifsrpcED()
 *
 * Description  : CIFS RPC encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01//2002
 *
 *-----------------------------------------------------------------
 */
//int32	//NFX_OK or NFX_ERR
//cifsrpcED(cifsrpc_ed *edp, 



/*-----------------------------------------------------------------
 * Name         : cifsrpcED()
 *
 * Description  : CIFS RPC encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01//2002
 *
 *-----------------------------------------------------------------
 */
//int32	//NFX_OK or NFX_ERR
//cifsrpcED(cifsrpc_ed *edp, 



/*-----------------------------------------------------------------
 * Name         : cifsrpcED()
 *
 * Description  : CIFS RPC encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01//2002
 *
 *-----------------------------------------------------------------
 */
//int32	//NFX_OK or NFX_ERR
//cifsrpcED(cifsrpc_ed *edp, 



/*-----------------------------------------------------------------
 * Name         : cifsrpcED()
 *
 * Description  : CIFS RPC encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01//2002
 *
 *-----------------------------------------------------------------
 */
//int32	//NFX_OK or NFX_ERR
//cifsrpcED(cifsrpc_ed *edp, 


