/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2002, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc-lsarpc.c,v 1.2 2002/12/10 18:13:23 laurentb Exp $
 *
 * Name			: cifs-rpc-lsarpc.c
 *
 * Description	: Contains CIFS DCE/RPC server services api functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/11/2002
 *
 *-----------------------------------------------------------------
 */

#ifdef FS_STANDALONE
#include "../sm-cifs/cifs-api.h"
#include "../sm-cifs/cifs-unicode.h"
#include "../sm-cifs/cifs-utils.h"
#include "../sm-cifs/cifs-err.h"
#else
#include <exports/cifs-api.h>
#include <exports/cifs-utils.h>
#include <exports/cifs-unicode.h>
#include <exports/cifs-err.h>
#endif

#include "cifs-rpc-utils.h"
#include "cifs-rpc.x"
#include "cifs-rpc-utils.h"
#include "cifs-rpc-lsarpc.h"
#include "cifs-rpc-ed-lsarpc.h"


//Unicode strings


sec_accessGenericMap cifs_policyAccessGenericMap = {
	SEC_GENR_POLICY_READ,
	SEC_GENR_POLICY_WRITE,
	SEC_GENR_POLICY_EXECUTE,
	SEC_GENR_POLICY_ALL
};



/*-----------------------------------------------------------------
 * Name			: newPolicyHandle
 *
 * Description	: generate a new policy handle
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/19/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
newPolicyHandle(cifsrpc_pipe *rpip, cifsrpc_lsaPolHnd *polHndp, void *datap)
{
	static uint64	polCnt = 0;
		
	polCnt++;

	
	polHndp->zero		= 0;
	polHndp->uuid.d1	= (uint32)	(polCnt);	
	polHndp->uuid.d2	= (ushort16)(polCnt >> 32);	
	polHndp->uuid.d3	= (ushort16)(polCnt >> 48);	
	memcpy(polHndp->uuid.d4, &rpip, 8);
	
	return (CIFSTATUS_OK);	
}


/*-----------------------------------------------------------------
 * Name			: lsaOpenPolicy1 & 2
 *
 * Description	: open policy & open policy 2
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/19/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
lsaOpenPolicy(	cifsrpc_pipe		*rpip,
				sec_accessMask		*desiredAccess,
				cifsrpc_lsaPolHnd	*polp)
{

	//Map generic access bits to lsa policy specific access bits
	cifs_secMapGenericAccess(desiredAccess,
							&cifs_policyAccessGenericMap);
						
	//We just allow execute access for now
	if (*desiredAccess & SEC_SPECR_ACCESS_MASK & ~SEC_GENR_POLICY_EXECUTE)
		return (CIFSTATUS_ACCESS_DENIED);

	newPolicyHandle(rpip, polp, NULL/*TODO*/);

		
	return (CIFSTATUS_OK);	
}


/*-----------------------------------------------------------------
 * Name			: lsarpc_openPolicy
 *
 * Description	: LSARPC open policy 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/19/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
lsarpc_openPolicy(cifsrpc_pipe *rpip)
{
	lsaOpenPolicyCall	openPolCall;
	lsaOpenPolicyResp	openPolResp;

	memset(&openPolCall, 0, sizeof(lsaOpenPolicyCall));
	memset(&openPolResp, 0, sizeof(lsaOpenPolicyResp));
	
	//Decode lsaOpenPolicy function call parameters
	if (cifsrpcEDlsaOpenPolicyCall(&rpip->inPdu.dec, &openPolCall, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call lsaOpenPolicy
	openPolResp.returnValue = lsaOpenPolicy(rpip,
											&openPolCall.desiredAccess,
											&openPolResp.pol);
	
	//Encode lsaOpenPolicy function response parameters
	if (cifsrpcEDlsaOpenPolicyResp(&rpip->outPdu.enc, &openPolResp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: lsarpc_openPolicy2
 *
 * Description	: LSARPC open policy 2
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/17/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
lsarpc_openPolicy2(cifsrpc_pipe *rpip)
{
	lsaOpenPolicy2Call	openPol2Call;
	lsaOpenPolicy2Resp	openPol2Resp;

	memset(&openPol2Call, 0, sizeof(lsaOpenPolicy2Call));
	memset(&openPol2Resp, 0, sizeof(lsaOpenPolicy2Resp));
	
	//Decode lsaOpenPolicy2 function call parameters
	if (cifsrpcEDlsaOpenPolicy2Call(&rpip->inPdu.dec, &openPol2Call, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call lsaOpenPolicy2
	openPol2Resp.returnValue = lsaOpenPolicy(	rpip,
												&openPol2Call.desiredAccess,
												&openPol2Resp.pol);
	
	//Encode lsaOpenPolicy2 function response parameters
	if (cifsrpcEDlsaOpenPolicy2Resp(&rpip->outPdu.enc, &openPol2Resp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}




/*-----------------------------------------------------------------
 * Name			: fillDomQuery
 *
 * Description	: fill domain query
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/22/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
fillDomQuery(domInfo *dip, uni16 *domName, sec_cifsSid *domSidp)
{
	uint32 domNameSize;

	if (domName != NULL) 
	{
		domNameSize		= unistrsize(domName);
		dip->uniLen		= domNameSize - 2;	//without ending 0
		dip->uniMaxLen	= domNameSize;		//with ending 0
		dip->domNamep	= RPC_DUMMY_PTR;
		fillUniStrucWithoutAlloc(&dip->domName, domName);
	} else {
		dip->uniMaxLen	= 0;
		dip->uniLen		= 0;
		dip->domNamep	= 0;
	}

	if (domSidp != NULL)
	{
		dip->domSidp			= RPC_DUMMY_PTR;
		dip->domSid.nbSubAuthy	= domSidp->nbSubAuthy;
		dip->domSid.sid			= *domSidp;
	} else
		dip->domSidp =  0;

	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: lsaQueryInfoPolicy
 *
 * Description	: LSARPC queryInfoPolicy
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/20/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
lsaQueryInfoPolicy(	cifsrpc_pipe *rpip,
					lsaQueryInfoPolicyCall *qInfoPolcp,
					lsaQueryInfoPolicyResp *qInfoPolrp)
{
	sec_cifsSid sid;
	
	switch (qInfoPolcp->infoClass)
	{
		case 2:
		{
			domInfo2 *cl2p = &qInfoPolrp->info.cl2;
			cl2p->auditEnabled = FALSE;
			cl2p->count1 = 0;
			cl2p->count2 = 0;
			//TODO see if it enough
			break;
		}

		case 3:
			fillDomQuery(&qInfoPolrp->info.cl3,
						rpip->clp->serverp->cfg.uniDomainName,
						&sid);	//FIXME we need the domain SID here
			break;

		case 5:
			fillDomQuery(&qInfoPolrp->info.cl3,
						rpip->clp->serverp->cfg.uniName,
						&sid);	//FIXME we need the machine SID here
			break;
	
		case 6:
			qInfoPolrp->info.cl6.serverRole = 2;
			break;
	}

	return (CIFSTATUS_OK);	
}


/*-----------------------------------------------------------------
 * Name			: lsarpc_queryInfoPolicy
 *
 * Description	: LSARPC queryInfoPolicy
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/19/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
lsarpc_queryInfoPolicy(cifsrpc_pipe *rpip)
{
	lsaQueryInfoPolicyCall	queryInfoPolCall;
	lsaQueryInfoPolicyResp	queryInfoPolResp;

	memset(&queryInfoPolCall, 0, sizeof(lsaQueryInfoPolicyCall));
	memset(&queryInfoPolResp, 0, sizeof(lsaQueryInfoPolicyResp));
	
	//Decode lsaQueryInfoPolicy function call parameters
	if (cifsrpcEDlsaQueryInfoPolicyCall(&rpip->inPdu.dec, &queryInfoPolCall, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call lsaQueryInfoPolicy
	queryInfoPolResp.returnValue = lsaQueryInfoPolicy(rpip,
													&queryInfoPolCall,
													&queryInfoPolResp);
	
	//Encode lsaQueryInfoPolicy function response parameters
	if (cifsrpcEDlsaQueryInfoPolicyResp(&rpip->outPdu.enc, &queryInfoPolResp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: lsaClose
 *
 * Description	: LSARPC close
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/19/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
lsaClose(cifsrpc_pipe *rpip, lsaCloseCall *closecp,	lsaCloseResp *closerp)
{
	
	//TODO Find the policy by policy handle and free it
	
	return (CIFSTATUS_OK);	
}


/*-----------------------------------------------------------------
 * Name			: lsarpc_close
 *
 * Description	: LSARPC close
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/19/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
lsarpc_close(cifsrpc_pipe *rpip)
{
	lsaCloseCall	closeCall;
	lsaCloseResp	closeResp;

	memset(&closeCall, 0, sizeof(lsaCloseCall));
	memset(&closeResp, 0, sizeof(lsaCloseResp));
	
	//Decode lsaClose function call parameters
	if (cifsrpcEDlsaCloseCall(&rpip->inPdu.dec, &closeCall, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call lsaClose
	closeResp.returnValue = lsaClose(rpip, &closeCall, &closeResp);
	
	//Encode lsaClose function response parameters
	if (cifsrpcEDlsaCloseResp(&rpip->outPdu.enc, &closeResp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}



static cifsrpc_function lsarpc_functions[] = {
{ LSARPC_OPENPOLICY,		"OpenPolicy",			lsarpc_openPolicy },
{ LSARPC_OPENPOLICY2,		"OpenPolicy2",			lsarpc_openPolicy2 },
{ LSARPC_QUERYINFOPOLICY,	"QueryInfoPolicy",		lsarpc_queryInfoPolicy },
{ LSARPC_CLOSE,				"Close",				lsarpc_close },
{ 0,						NULL,					NULL }
};


/*-----------------------------------------------------------------
 * Name			: cifsrpc_lsarpc
 *
 * Description	: process LSA services pipe requests
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/17/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_lsarpc(cifsrpc_pipe *rpip)
{
	return (cifsrpc_callFunction(rpip, lsarpc_functions));	
}


