/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-filereqs.c,v 1.2 2002/08/23 18:03:59 shaun Exp $
 *
 * Name			: cifs-filereqs.c
 *
 * Description	: Contains CIFS file requests to fs and fssf functions.
 *
 * Created by	: Laurent Brard
 * 				  Xiaobing Zhang
 *
 * Date Created	: 03/19/2002
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-err.h"
#include "cifs-string.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-encdec.h"
#include "cifs-intmsg.h"
#include "cifs-filetools.h"
#include "cifs-filereqs.h"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)

#ifdef LINUX_TEST	
#define	FS_CASE_INSENSITIVE		1
#define	FS_BACKSLASH_SEPARATOR	2
#endif

//Unicode strings


/*-----------------------------------------------------------------
 * Name			: cifs_fsMCLReq
 *
 * Description	: Make a MCL request to the file system
 * 				  if startDirp == NULL, use share handle
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/20/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fsMCLReq(	cifs_operation *opp,
				//TODO remove this ? (not used)
				fs_Handle *startDirp,	//handle of the directory to start from
				uchar8 *smbp,       	//source SMB message pointer
				uni16 *path,			//path
				fs_AttrType attrType)	//attributes to request
{
	uchar8 reqBuf[sizeof(fs_MsgMclReq)+FS_PATH_MAX + 10];	
	fs_MsgMclReq *reqMsgp = (fs_MsgMclReq*)reqBuf;
	uint32 rc, reqLen;

	cifs_msgBuildFSHdr(&reqMsgp->hdr, FS_REQ_MCL, attrType, 
						SMB_TESTFLAGS2BIT(smbp, SMB_FLGS_PATHCASELESS),
						opp->threadId); 
	if (startDirp)
		reqMsgp->parentHandle = *startDirp;
	else
		reqMsgp->parentHandle = opp->siup->sharep->handle;

	//Convert path name to UTF-8
	rc = strUni2Utf8(reqMsgp->objName,
					path,
					unistrlen(path),
					&reqMsgp->objLen,
					FS_PATH_MAX);
	if (rc != CIFSTATUS_OK)
		return (rc);

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_fsMCLReq path:[%s]", reqMsgp->objName));
	
	reqLen = sizeof(fs_MsgMclReq) + reqMsgp->objLen + 1;
	
	if (cifs_msgSend(CIFS_INTMSG_TO_FS, reqBuf, reqLen, FALSE) == NFX_OK)
		return (CIFSTATUS_OK);
	else 
		return (CIFSTATUS_UNSUCCESSFUL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_fsMCLReqFromMsg
 *
 * Description	: Make a MCL request to the file system, retreiving
 * 				  the request directly from the CIFS message and 
 * 				  returning last elem of the internal path if elem != NULL
 * 				  if startDirp == NULL, use share handle
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/22/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fsMCLReqFromMsg(	cifs_operation *opp,
						fs_Handle *startDirp,	//the directory to start from
						uchar8 *smbp,       	//source SMB message pointer
						uchar8 *spath,			//source path
						uni16  *elem,			//uni16 last element
						fs_AttrType attrTyp)	//attributes to request
{
	uni16 path[CIFS_S_PATHSTR+1];
    uint32 rc;
			
	rc = cifs_Smb2UniPath(smbp, spath, path);
	if (rc != CIFSTATUS_OK)
		return (rc);
	
	//get last element if requested
	if (elem != NULL)
	{
		rc = cifs_extractPathLastElement(path, elem, CIFS_S_FILESTR, FALSE);
		if (rc != CIFSTATUS_OK)
			return (rc);
	}
	
	return (cifs_fsMCLReq(opp, startDirp, smbp, path, attrTyp));
}


/*-----------------------------------------------------------------
 * Name			: cifs_fsMkdirReq
 *
 * Description	: Make a MkDir request to the file system
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/21/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fsMkdirReq(cifs_operation	*opp,
				uchar8			*smbp,       	//source SMB message pointer
				fs_Handle		*parentHandlep,	//handle of parent directory
				uni16			*dir,			//directory to create
				fs_Attr			*dirAttrp,		//attributes of directory
				fs_AttrType		attrType)		//attributes to request back
{
	uchar8 reqBuf[sizeof(fs_MsgMkdirReq)+FS_NAME_MAX+10];	
	fs_MsgMkdirReq *reqMsgp = (fs_MsgMkdirReq*)reqBuf;
	uint32 rc, reqLen;

	cifs_msgBuildFSHdr(&reqMsgp->hdr,
						FS_REQ_MKDIR,
						attrType,
						SMB_TESTFLAGS2BIT(smbp, SMB_FLGS_PATHCASELESS),
						opp->threadId); 
	
	reqMsgp->parentHandle = *parentHandlep;
	memcpy(&reqMsgp->dirAttr, dirAttrp, sizeof(fs_Attr));

	//Convert directory name to UTF-8
	rc = strUni2Utf8(reqMsgp->dirName,
					dir,
					unistrlen(dir),
					&reqMsgp->nameLen,
					FS_NAME_MAX);
	if (rc != CIFSTATUS_OK)
		return (rc);

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_fsMkdirReq dir:[%s]", reqMsgp->dirName));
	
	reqLen = sizeof(fs_MsgMkdirReq) + reqMsgp->nameLen + 1;
	
	if (cifs_msgSend(CIFS_INTMSG_TO_FS, reqBuf, reqLen, FALSE) == NFX_OK)
		return (CIFSTATUS_OK);
	else 
		return (CIFSTATUS_UNSUCCESSFUL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_fsReaddirplusReq
 *
 * Description  : Make a READDIRPLUS request to the file system
 *
 * Created by   : Xiaobing Zhang
 *
 * Date Created : 03/26/2002
 *
 *-----------------------------------------------------------------
 */
uint32			//CIFSTATUS
cifs_fsReaddirplusReq(cifs_operation	*opp,
					  cifs_search		*searchp,
					  fs_AttrType		attrType,
					  uint32			maxBytes)
{
	int32					rc;
	uchar8					reqBuf[sizeof(fs_MsgReadDirPlusReq)+10];
	fs_MsgReadDirPlusReq	*reqMsgp = (fs_MsgReadDirPlusReq*)reqBuf;
	uint32					reqLen;

	cifs_msgBuildFSHdr(&reqMsgp->hdr, FS_REQ_READDIRPLUS, attrType,
					SMB_TESTFLAGS2BIT(opp->reqMsg.smbp, SMB_FLGS_PATHCASELESS),
					opp->threadId);
	
	memcpy(&reqMsgp->fileHandle, &searchp->parentHandle, sizeof(fs_Handle));
	reqMsgp->cookieVerif = searchp->cookieVerif;
	reqMsgp->cookie = searchp->cookie;
	reqMsgp->maxDirBytes = maxBytes;
	reqMsgp->maxBytes = maxBytes;

	rc = cifs_msgSend(CIFS_INTMSG_TO_FS,
					  reqBuf,
					  sizeof(fs_MsgReadDirPlusReq),
					  FALSE);
	if (rc == NFX_OK)
		return (CIFSTATUS_OK);
	else
		return (CIFSTATUS_UNSUCCESSFUL);
} /* cifs_fsReaddirplusReq() */




//**************************************************************************
//**************************************************************************
//**************************************************************************



/*-----------------------------------------------------------------
 * Name			: cifs_fssfRemoveReq
 *
 * Description	: Make a Remove request to the file system statefull
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/02/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fssfRemoveReq(	cifs_operation	*opp,
					uchar8			*smbp,     		//source SMB message pointer
					fs_Handle		*parentHandlep,	//handle of parent directory
					uni16			*file)			//file/directory to remove
{
	uchar8 reqBuf[sizeof(fssf_removeReq_t)+FS_NAME_MAX+10];	
	fssf_removeReq_t *reqMsgp = (fssf_removeReq_t*)reqBuf;
	uint32 rc, reqLen;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_REMOVE_REQ, opp); 
	
	reqMsgp->parentHandle	= *parentHandlep;
	reqMsgp->nameFlags		= FS_BACKSLASH_SEPARATOR;
	if (SMB_TESTFLAGS2BIT(smbp, SMB_FLGS_PATHCASELESS))
		reqMsgp->nameFlags |= FS_CASE_INSENSITIVE;

	//Convert file name to UTF-8
	rc = strUni2Utf8(reqMsgp->name,
					file,
					unistrlen(file),
					&reqMsgp->nameLen,
					FS_NAME_MAX);
	if (rc != CIFSTATUS_OK)
		return (rc);

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_fssfRemoveReq name:[%s]",reqMsgp->name));
	
	reqMsgp->hdr.len = sizeof(fssf_removeReq_t) + reqMsgp->nameLen + 1;

	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE) 
			== NFX_OK)
		return (CIFSTATUS_OK);
	else 
		return (CIFSTATUS_UNSUCCESSFUL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_fssfOpenReq
 *
 * Description	: Make an open request to the file system statefull
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/31/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fssfOpenReq(cifs_operation		*opp,
				fs_Handle			*parentHandlep,	//handle of parent directory
				uni16				*file,			//file to open
				fssf_action_t		openAction,
				fs_Attr				*attrp,			//see fssf-api.h
				fssf_reqAccess_t	reqAccess,
				fssf_shareMode_t	shareMode,
				uint32				openFlags)
{
	uchar8 reqBuf[sizeof(fssf_openReq_t)+FS_NAME_MAX+10];	
	fssf_openReq_t *reqMsgp = (fssf_openReq_t*)reqBuf;
	uint32 rc;
	uchar8 *smbp = opp->reqMsg.smbp;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_OPEN_REQ, opp); 
	
	reqMsgp->owner			= cifs_buildFssfClientId(opp);
	reqMsgp->cnxId			= (address_t)opp->clp;
	reqMsgp->action			= openAction;
	reqMsgp->u.attr			= *attrp;
	reqMsgp->reqAccess		= reqAccess;
	reqMsgp->shareMode		= shareMode;
	reqMsgp->flags			= openFlags;
	if (cifs_uniIsExecutable(file))
		reqMsgp->flags		|= FSSF_FLAG_EXE;
	reqMsgp->parentHandle	= *parentHandlep;
	reqMsgp->nameFlags		= FS_BACKSLASH_SEPARATOR;
	if (SMB_TESTFLAGS2BIT(smbp, SMB_FLGS_PATHCASELESS))
		reqMsgp->nameFlags |= FS_CASE_INSENSITIVE;

	//Convert file name to UTF-8
	rc = strUni2Utf8(reqMsgp->name,
					file,
					unistrlen(file),
					&reqMsgp->nameLen,
					FS_NAME_MAX);
	if (rc != CIFSTATUS_OK)
		return (rc);

	IFDBG(CIFS_DBGLV_DBG2) {
		CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_fssfOpenReq name(f=0x%x):[%s]", 
					reqMsgp->nameFlags, reqMsgp->name));
		CIFS_PRINT(CIFS_DBGLV_DBG2, 
			("\towner=0x%llx/0x%llx action=%d-%d reqAcc=0x%x shMode=0x%x",
			reqMsgp->owner, reqMsgp->cnxId, reqMsgp->action.ifExist, 
			reqMsgp->action.ifNotExist, reqMsgp->reqAccess,reqMsgp->shareMode));
		CIFS_PRINT(CIFS_DBGLV_DBG2, ("\tflags=0x%x", reqMsgp->flags));
	}
	
	reqMsgp->hdr.len = sizeof(fssf_openReq_t) + reqMsgp->nameLen + 1;

	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE) 
			== NFX_OK)
		return (CIFSTATUS_OK);
	else 
		return (CIFSTATUS_UNSUCCESSFUL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_fssfOpendirReq
 *
 * Description	: Make an open directory request to the file system
 * 				  statefull
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/06/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fssfOpendirReq(cifs_operation		*opp,
					fs_Handle			*parentHandlep,	//handle of parent dir
					uni16				*dir,			//directory to open
					fssf_shareMode_t	shareMode,
					uint32				openFlags)
{
	uchar8 reqBuf[sizeof(fssf_opendirReq_t)+FS_NAME_MAX+10];	
	fssf_opendirReq_t *reqMsgp = (fssf_opendirReq_t*)reqBuf;
	uint32 rc;
	uchar8 *smbp = opp->reqMsg.smbp;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_OPENDIR_REQ, opp); 
	
	reqMsgp->owner			= cifs_buildFssfClientId(opp);
	reqMsgp->shareMode		= shareMode;
	reqMsgp->flags			= openFlags;
	reqMsgp->parentHandle	= *parentHandlep;
	reqMsgp->nameFlags		= FS_BACKSLASH_SEPARATOR;
	if (SMB_TESTFLAGS2BIT(smbp, SMB_FLGS_PATHCASELESS))
		reqMsgp->nameFlags |= FS_CASE_INSENSITIVE;

	//Convert file name to UTF-8
	rc = strUni2Utf8(reqMsgp->name,
					dir,
					unistrlen(dir),
					&reqMsgp->nameLen,
					FS_NAME_MAX);
	if (rc != CIFSTATUS_OK)
		return (rc);

	IFDBG(CIFS_DBGLV_DBG2) {
		CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_fssfOpendirReq dir(f=0x%x):[%s]", 
					reqMsgp->nameFlags, reqMsgp->name));
		CIFS_PRINT(CIFS_DBGLV_DBG2, 
			("\towner=0x%llx shMode=0x%x flags=0x%x",
			reqMsgp->owner, reqMsgp->shareMode, reqMsgp->flags));
	}
	
	reqMsgp->hdr.len = sizeof(fssf_opendirReq_t) + reqMsgp->nameLen + 1;

	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE) 
			== NFX_OK)
		return (CIFSTATUS_OK);
	else 
		return (CIFSTATUS_UNSUCCESSFUL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_fssfCloseReq
 *
 * Description	: Make a close request to the file system statefull
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/08/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fssfCloseReq(	cifs_operation	*opp,
					fssf_stateid_t	stateid)		//statid to close
{
	uchar8 reqBuf[sizeof(fssf_closeReq_t)];	
	fssf_closeReq_t *reqMsgp = (fssf_closeReq_t*)reqBuf;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_CLOSE_REQ, opp); 
	
	reqMsgp->nbState	= 1;
	reqMsgp->stateid[0] = stateid;

	CIFS_PRINT(CIFS_DBGLV_DBG2, 
				("cifs_fssfCloseReq: stateid=0x%llx", stateid.id));
	
	reqMsgp->hdr.len = sizeof(fssf_closeReq_t);

	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE) 
			== NFX_OK)
		return (CIFSTATUS_OK);
	else 
		return (CIFSTATUS_UNSUCCESSFUL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_fssfSetattrByIdReq
 *
 * Description	: Make a set attributes request to the file system statefull
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/10/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_fssfSetattrByIdReq(cifs_operation		*opp,
						fssf_stateid_t		stateid,
						fssf_attr_t			*attrp)
{
	uchar8 reqBuf[sizeof(fssf_setattrReq_t)];	
	fssf_setattrReq_t *reqMsgp = (fssf_setattrReq_t*)reqBuf;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_SETATTR_REQ, opp); 
	
	reqMsgp->stateid	= stateid;
	reqMsgp->attr		= *attrp;
	reqMsgp->hdr.len = sizeof(fssf_setattrReq_t);
	
	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE) 
			== NFX_OK)
		return (CIFSTATUS_OK);
	else 
		return (CIFSTATUS_UNSUCCESSFUL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_fssfGetAttrReqByFid
 *
 * Description  : Make a GETATTR request to an open file system stateful
 *
 * Created by   : Xiaobing Zhang
 *
 * Date Created : 04/05/2002
 *
 *-----------------------------------------------------------------
 */
uint32			//CIFSTATUS
cifs_fssfGetAttrReqByFid(cifs_operation	*opp,
						cifs_fileOpen	*fop,
						uint64			bitmap)
{
	uchar8				reqBuf[sizeof(fssf_getattrReq_t)];
	fssf_getattrReq_t	*reqMsgp = (fssf_getattrReq_t *)reqBuf;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_GETATTR_REQ, opp);
	reqMsgp->hdr.len = sizeof(fssf_getattrReq_t);

	reqMsgp->stateid = fop->stateid;
	reqMsgp->bitmap = bitmap;

	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE)
			== NFX_OK)
		return (CIFSTATUS_OK);
	else
		return (CIFSTATUS_UNSUCCESSFUL);
} /* cifs_fsGetAttrReqByFid() */

/*-----------------------------------------------------------------
 * Name			: cifs_fssfReadReq
 *
 * Description  : Make a READ request to the file system stateful
 *
 * Created by   : Xiaobing Zhang
 *
 * Date Created : 04/08/2002
 *
 *-----------------------------------------------------------------
 */
uint32			//CIFSTATUS
cifs_fssfReadReq(cifs_operation		*opp,
				 cifs_fileOpen		*fop,
				 uint64				offset,
				 uint32				maxCnt)
{
	uchar8			reqBuf[sizeof(fssf_readReq_t)];
	fssf_readReq_t	*reqMsgp = (fssf_readReq_t *)reqBuf;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_READ_REQ, opp);
	reqMsgp->hdr.len = sizeof(fssf_readReq_t);

	reqMsgp->stateid = fop->stateid;
	reqMsgp->pos = FSSF_SEEK_SET;
	reqMsgp->offset = (fs_Offset) offset;
	reqMsgp->len = maxCnt;

	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE)
			== NFX_OK)
		return (CIFSTATUS_OK);
	else
		return (CIFSTATUS_UNSUCCESSFUL);
} /* cifs_fssfReadReq() */

/*-----------------------------------------------------------------
 * Name			: cifs_fssfReadReq
 *
 * Description  : Make a READ request to the file system stateful
 *
 * Created by   : Xiaobing Zhang
 *
 * Date Created : 04/13/2002
 *
 *-----------------------------------------------------------------
 */
uint32			//CIFSTATUS
cifs_fssfWriteReq(cifs_operation	*opp,
				 cifs_fileOpen		*fop,
				 uint64				offset,
				 uint32				dataLen,
				 uint32				nbrUsedCod,
				 nfx_cod_entry_t	*cod,
				 ushort16			writeMode)
{
	int32			rc, i;
	uchar8			reqBuf[sizeof(fssf_writeReq_t)];
	fssf_writeReq_t	*reqMsgp = (fssf_writeReq_t *)reqBuf;
	fs_WriteStableHow	stableHow;

	cifs_msgBuildFssfHdr(&reqMsgp->hdr, FSSF_WRITE_REQ, opp);
	reqMsgp->hdr.len = sizeof(fssf_writeReq_t);

	reqMsgp->stateid = fop->stateid;
	reqMsgp->pos = FSSF_SEEK_SET;
	reqMsgp->offset = (fs_Offset) offset;
	reqMsgp->len = dataLen;
	reqMsgp->nbrUsedCod = nbrUsedCod;

	// Fill in COD for write buffer
	for (i = 0; i < nbrUsedCod; i++)
	{
		reqMsgp->cod[i].cod = (cod + i)->cod;
		reqMsgp->cod[i].offset = (cod + i)->offset;
		reqMsgp->cod[i].length = (cod + i)->length;
	}
	// Convert writeMode to stableHow
	reqMsgp->stableHow = cifs_writeMode2StableHow(writeMode);
	
	if (cifs_msgSend(CIFS_INTMSG_TO_FSSF, reqBuf, reqMsgp->hdr.len, FALSE)
			== NFX_OK)
		return (CIFSTATUS_OK);
	else
		return (CIFSTATUS_UNSUCCESSFUL);
} /* cifs_fssfWriteReq() */




//**************************************************************************
//**************************************************************************
//**************************************************************************




/*-----------------------------------------------------------------
 * Name			: cifs_openFile
 *
 * Description	: open a file
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/31/2002
 *
 *-----------------------------------------------------------------
 */
cifs_fileOpen *	//NULL if failed
cifs_openFile(	cifs_operation		*opp,
				fs_Handle			*parentHandlep,	//handle of parent directory
				uni16				*path,			//full path to open
				fssf_action_t		openAction,
				fssf_reqAccess_t	reqAccess,		//see fssf-api.h
				fssf_shareMode_t	shareMode,
				uint32				openFlags,
				fs_Attr				*attrp,			//attrs of existing file if
													//exist, if not attrs for
													//the file to create.
													//if open succeeds, filled
													//with attrs of the open
													//file
				uint32				*createAction)	//return fssf action 
{
	uni16 file[CIFS_S_FILESTR+1];	//file/directory to "open"		
	cifs_fileOpen *fop;
	uint32 rc;
	eee_desc_t *eDescp;

	fop = cifs_addFileOpen(opp->siup);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_openFile: no avail file open obj"));
		cifs_errResp(opp, 0, 0, CIFSTATUS_TOO_MANY_OPENED_FILES);
		return (NULL);
	}
	
	//extract file name from path
	rc = cifs_extractPathLastElement(path, file, CIFS_S_FILESTR, FALSE);
	if (rc != CIFSTATUS_OK)
	{   cifs_errResp(opp, 0, 0, rc); goto openError; }
	
	if (attrp->type == FS_FTYPE_DIR)
	{	//try to open the directory
		fssf_opendirRsp_t *opendirResp;

		rc = cifs_fssfOpendirReq(opp,
								parentHandlep,	
								file,
								shareMode,
								openFlags);
		if (rc != CIFSTATUS_OK) 
		{	cifs_errResp(opp, 0, 0, rc); goto openError; }
	
		//Wait for the answer
		rc = cifs_waitFssfResp(FSSF_OPENDIR_RSP, (void**)&opendirResp, &eDescp);
		if (rc != CIFSTATUS_OK)
		{	cifs_errResp(opp, 0, 0, CIFSTATUS_INTERNAL_ERROR); goto openError; }
		if (opendirResp->status != FS_ERR_NONE)
		{	
			cifs_returnFsError(opp, opendirResp->status); 
			eee_freePacket(eDescp); 
			goto openError;
		}
		fop->stateid	= opendirResp->stateid;
		fop->flags		|= CIFS_FOP_ISDIR;
		*attrp			= opendirResp->attr;
		*createAction	= FSSF_ACTION_OPEN;
	} 
	else
	{	//try to open the file
		fssf_openRsp_t *openResp;
		
		rc = cifs_fssfOpenReq(	opp,
								parentHandlep,	
								file,
								openAction,
								attrp,
								reqAccess,
								shareMode,
								openFlags);
		if (rc != CIFSTATUS_OK) 
		{	cifs_errResp(opp, 0, 0, rc); goto openError; }
	
		//Wait for the answer
		rc = cifs_waitFssfResp(FSSF_OPEN_RSP, (void**)&openResp, &eDescp);
		if (rc != CIFSTATUS_OK)
		{	cifs_errResp(opp, 0, 0, CIFSTATUS_INTERNAL_ERROR); goto openError; }
		if (openResp->status != FS_ERR_NONE)
		{	
			cifs_returnFsError(opp, openResp->status); 
			eee_freePacket(eDescp); 
			goto openError;
		}
		fop->stateid	= openResp->stateid;
		*attrp			= openResp->attr;
		*createAction	= openResp->action;
	}
	
	//store file/directory path
	CIFS_RAM_ALLOC_ALIGNED(fop->pathName, uni16, unistrsize(path));
	unistrcpy(fop->pathName, path);
	
	//free Open response
	eee_freePacket(eDescp); 
	
	return (fop);

openError:
	cifs_closeShareInUseFileOpen(opp->siup, fop);
	return (NULL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_readFile
 *
 * Description  : Read a file
 *
 * Created by   : Xiaobing Zhang
 *
 * Date Created : 04/08/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifs_readFile(	cifs_operation	*opp,
				cifs_fileOpen	*fop,
				uint64			offset,		// Offset from start of file
				uint32			maxCnt,		// Max number of bytes to read
				uint32			*dataCnt)	// Number of bytes read
{
	int32			rc;
	eee_desc_t		*eDescp;
	fssf_readRsp_t	*readResp;
	int32			codIdx;
	pkt_desc_t		*xmtPktDescp;

	//--------------------------
	// Send read request to FSSF
	//--------------------------
	rc = cifs_fssfReadReq(opp, fop, offset, maxCnt);
	if (rc != CIFSTATUS_OK)
		return (rc);

	//----------------------------
	// Wait for response from FSSF
	//----------------------------
	rc = cifs_waitFssfResp(FSSF_READ_RSP, (void**)&readResp, &eDescp);
	if (rc != CIFSTATUS_OK)
	{
		return (CIFSTATUS_UNSUCCESSFUL);
	}

	if (readResp->status != FS_ERR_NONE)
	{
		// There is error from FS
		uchar8 class = 0;
		ushort16 err = 0;

		rc = cifs_fsToCifsErrMapping(readResp->status, &class, &err);
		eee_freePacket(eDescp);
		return (rc);
	}

	//---------------------------
	// Process response from FSSF
	//---------------------------

	// User data in COD
	xmtPktDescp = opp->respMsg.tplPktDescp;

	NFX_ASSERT(readResp->nbrUsedCod <= NFX_MAX_COD_MSG);

	xmtPktDescp->numOfCod = readResp->nbrUsedCod;
	for (codIdx = 0; codIdx < xmtPktDescp->numOfCod; codIdx++)
	{
		xmtPktDescp->codEntry[codIdx] = readResp->cod[codIdx];
	}
	xmtPktDescp->dataSize = readResp->bytesRead;

	*dataCnt = readResp->bytesRead;

	// Store fsPrivate and set flag in operation object
	opp->fsPrivateForCod = readResp->fsPrivate;
	opp->flags |= CIFS_OPFLAG_READ_REQ;

	eee_freePacket(eDescp);

	return (CIFSTATUS_OK);
} /* cifs_readFile() */

/*-----------------------------------------------------------------
 * Name			: cifs_writeFile
 *
 * Description  : Write a file
 *
 * Created by   : Xiaobing Zhang
 *
 * Date Created : 04/12/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifs_writeFile(	cifs_operation	*opp,
				cifs_fileOpen	*fop,
				uchar8			*bp,			// Source data to write
				uint64			offset,			// Offset from start of file
				uint32			dataLen,		// Number of bytes to write
				ushort16		writeMode,		// Write mode
				uint32			*writtenDataLen)	// Number of bytes written
{
	int32				rc;
	uint32				nbrUsedCod;				// Number of used COD's
	nfx_cod_entry_t		cod[NFX_MAX_COD_MSG];	// Data to write
	fssf_writeRsp_t		*writeResp;
	eee_desc_t			*eDescp;

	// Sends data to write to GMC with COD
#ifndef FS_STANDALONE	//TODO Fix this
#ifdef SBM_REV2
	nbrUsedCod = luc_sndGmcData(opp->reqMsg.tplPktDescp, bp, dataLen, 0, &cod[0]);
#else
	nbrUsedCod = luc_sndGmcData(opp->reqMsg.tplPktDescp, bp, dataLen, &cod[0]);
#endif
#endif	

	if ((nbrUsedCod > NFX_MAX_COD_MSG) || (nbrUsedCod <= 0))
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_writeFile: "
					"Invalid COD number (%d)", nbrUsedCod));
#ifndef FS_STANDALONE	//TODO Fix this
		luc_panic();
#endif		
		return (CIFSTATUS_UNSUCCESSFUL);
	}

	//---------------------
	// Send request to FSSF
	//---------------------
	rc = cifs_fssfWriteReq(opp, fop, offset, dataLen, 
							nbrUsedCod, cod, writeMode);
	if (rc != CIFSTATUS_OK)
		return (rc);

	//----------------------------
	// Wait for response from FSSF
	//----------------------------
	rc = cifs_waitFssfResp(FSSF_WRITE_RSP, (void**)&writeResp, &eDescp);
	if (rc != CIFSTATUS_OK)
	{
		return (CIFSTATUS_UNSUCCESSFUL);
	}

	if (writeResp->status != FS_ERR_NONE)
	{
		// There is error from FS
		uchar8 class = 0;
		ushort16 err = 0;

		rc = cifs_fsToCifsErrMapping(writeResp->status, &class, &err);
		eee_freePacket(eDescp);
		return (rc);
	}

	*writtenDataLen = writeResp->bytesWritten;

	eee_freePacket(eDescp);

	return (CIFSTATUS_OK);
} /* cifs_writeFile() */

