/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-msg-xz.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 *
 * Name			: cifs-msg.c
 *
 * Description	: Contains CIFS encoding and decoding functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-transport.h"
#include "cifs-encdec.h"
#include "cifs-err.h"
#include "cifs-msg.h"
#include "cifs-ipc.h"
#include "cifs-intmsg.h"
#include "cifs-filetools.h"
#include "cifs-filereqs.h"


/*-----------------------------------------------------------------
 * Name			: cifsm_readNX
 *
 * Description	: Read a file (0x2E)
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/07/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_readNX(	cifs_client		*clp,
					cifs_session	*ssp,
					cifs_shareInUse	*siup,
					cifs_operation	*opp,
					uchar8			*icmdp,
					uchar8			*ocmdp)
{
	int32			rc;
	ushort16		fid;
	uint64			offset;
	uint32			maxReplySize, maxCnt, dataCnt;
	ushort16		minCnt;
	boolean			largeReadSupport;
	cifs_fileOpen	*fop;

	uchar8			*bp;
	ushort16		pad;
	pkt_desc_t		*xmtPktDescp;

	// Check if share is an IPC
	if (siup->flags & CIFS_SIU_ISIPC)
	{
		//TODO
		CIFS_PRINT(CIFS_DBGLV_WARN, ("cifsm_readNX: IPC read is not implemented!\n"));
		return (CIFS_SENDRESPNOW);
	}

	// Check large read support
	largeReadSupport = clp->serverp->cfg.capabilities & CAP_LARGE_READX;


	//--------------------------------------
	// Decode word fields in request message
	//--------------------------------------

	fid = GETLE2(icmdp, SMBCO_READNX_FID);
	fop = cifs_findFileOpenByFid(opp->siup, fid);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_readNX: Bad Fid %d", fid));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}

	offset = GETLE4(icmdp, SMBCO_READNX_OFFSET_LOW);
	if (GCAR(icmdp, SMBO_WCNT) == 12)
	{
		// Larege offset (64 bit) read
		uint64	offsetHigh = GETLE4(icmdp, SMBCO_READNX_OFFSET_HIGH);

		if (offsetHigh != 0)
		{
			if (largeReadSupport == TRUE)
			{
				// The CIFS server support large read
				offset |= (offsetHigh << 32);
			}
			else 
			{
				CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_readNX: "
							"Server does not support CAP_LARGE_READX"));
				cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_UNSUCCESSFUL);
			}
		}
	}

	maxCnt = GETLE2(icmdp, SMBCO_READNX_MAX_CNT_LOW);
	if (largeReadSupport == TRUE)
	{
		maxCnt |= (GETLE4(icmdp, SMBCO_READNX_MAX_CNT_HIGH) << 16);
	}

	minCnt = GETLE2(icmdp, SMBCO_READNX_MIN_CNT);

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsm_readNX: "
				"fid = 0x%x, maxCnt = 0x%x, minCnt = 0x%x, offset = 0x%llx",
				fid, maxCnt, minCnt, offset));

	//-----------------------------------------------
	// Get pointer to data in response message buffer
	//-----------------------------------------------

	// Set word counter in response message
	cifs_setCmdWordCnt(&opp->respMsg, 12, TRUE);

	// Align data bytes on a 32-bit boundary 
	//TODO: see if we want to go to 64-bit alignment on Sibyte CPU
	bp = SMBCP_BPAR(ocmdp);			// Start of SMB byte parameters
	pad = (address_t)bp % CIFS_ALIGN_32;
	if (pad)
		pad = CIFS_ALIGN_32 - pad;
	bp += pad;						// Start of Data read from resource

	//----------
	// Read file
	//----------
	
	if (largeReadSupport == TRUE)
	{
		//In this case, we don't check the request's maxCnt against
		//the negotiated buffer size. (See Section 4.2.4 in CIFS spec 0.99)
	}
	else
	{
		//Data in each reponse SMB is expected to be truncated to the maximum 
		//number of 512 byte blocks which will fit (starting at a 32 bit 
		//maxBytes) in the negociated buffer size with the odd bytes 
		//remaining (if any) in the final buffer (CIFS 0.99, paragraph 3.14)
		//maxReplySize is the maximum size of SMB message
		maxReplySize = opp->clp->maxBufSize;
		maxReplySize -= maxReplySize % 512;

		maxCnt = MIN(maxReplySize - SMB_DIFF(bp, opp->respMsg.smbp), maxCnt);
	}

	rc = cifs_readFile(opp, fop, offset, maxCnt, &dataCnt);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//-------------------
	// Build the response
	//-------------------
	cifs_initNXCmdHeader(ocmdp);	
	PUTLE2(ocmdp, SMBCO_WP2, -1);					//Remaining, must be -1
	PUTLE2(ocmdp, SMBCO_WP3, 0);					//DataCompactionMode
	PUTLE2(ocmdp, SMBCO_WP4, 0);					//Reserved
	PUTLE2(ocmdp, SMBCO_WP5, dataCnt & 0xFF);		//Number of data bytes
	PUTLE2(ocmdp, SMBCO_WP6, SMB_DIFF(bp, opp->respMsg.smbp));	
										//Offset from header start to data
	PUTLE2(ocmdp, SMBCO_WP7, (dataCnt >> 16) & 0xFF);
										//High 16 bits of number of data bytes

	PUTLE8(ocmdp, SMBCO_WP8, 0);								//Reserved
	cifs_setCmdByteCnt(&opp->respMsg, pad + dataCnt, FALSE);	//Byte count 

	opp->respMsg.smbLen -= dataCnt;		//Don't include COD user data count
										//in hdrSize in the packet descriptor

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsm_readNX: "
				"actual bytes of data read = 0x%x", dataCnt));

	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
	{
		NFX_ASSERT(FALSE);				//TODO - handle batched close request
		return cifs_processNextCmd(opp, ocmdp);
	}
	else
		return (CIFS_SENDRESPNOW);
} /* cifsm_readNX() */


/*-----------------------------------------------------------------
 * Name			: cifsm_read
 *
 * Description	: Read a file (0x0A)
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/11/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_read(cifs_client		*clp,
				 cifs_session		*ssp,
				 cifs_shareInUse	*siup,
				 cifs_operation		*opp,
				 uchar8				*icmdp,
				 uchar8				*ocmdp)
{
	int32			rc;
	ushort16		fid;
	ushort16		maxCnt;			// Number of bytes of data to read
	uint32			dataCnt;		// Number of bytes of data actually read
	uint32			offset;
	cifs_fileOpen	*fop;

	uchar8			*bp;
	ushort16		pad;
	pkt_desc_t		*xmtPktDescp;

	//--------------------------------------
	// Decode word fields in request message
	//--------------------------------------
	
	fid = GETLE2(icmdp, SMBCO_WP0);
	fop = cifs_findFileOpenByFid(opp->siup, fid);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_read: Bad Fid %d", fid));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}

	maxCnt = GETLE2(icmdp, SMBCO_WP1);
	offset = GETLE4(icmdp, SMBCO_WP2);

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsm_read: "
				"fid = 0x%x, byteCnt = 0x%x, offset = 0x%x",
				fid, maxCnt, offset));

	//-----------------------------------------------
	// Get pointer to data in response message buffer
	//-----------------------------------------------

	// Set word counter in response message
	cifs_setCmdWordCnt(&opp->respMsg, 5, TRUE);

	// Align data bytes on a 32-bit boundary
	bp = SMBCP_BPAR(ocmdp);			// Start of SMB byte parameters
	pad = 3;
	bp += pad;						// Start of Data read from resource

	//----------
	// Read file
	//----------

	// Mare sure that the amount of data requested will fit in 
	// the negotiated maximum buffer size
	maxCnt = MIN(opp->clp->maxBufSize - SMB_DIFF(bp, opp->respMsg.smbp), maxCnt);

	rc = cifs_readFile(opp, fop, offset, maxCnt, &dataCnt);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//-------------------
	// Build the response
	//-------------------
	dataCnt &= 0xFF;
	PUTLE2(ocmdp, SMBCO_WP0, dataCnt);			//Number of bytes of data read
	PUTLE8(ocmdp, SMBCO_WP1, 0);								//Reserved
	cifs_setCmdByteCnt(&opp->respMsg, pad + dataCnt, FALSE);	//Byte count

	opp->respMsg.smbLen -= dataCnt;		//Don't include COD user data count
										//in hdrSize in the packet descriptor

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsm_read: "
				"actual bytes of data read = 0x%x", dataCnt));

	return (CIFS_SENDRESPNOW);
} /* cifsm_read() */

/*-----------------------------------------------------------------
 * Name			: cifsm_writeNX
 *
 * Description	: Write a file (0x2F)
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/10/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_writeNX(cifs_client		*clp,
					cifs_session	*ssp,
					cifs_shareInUse	*siup,
					cifs_operation	*opp,
					uchar8			*icmdp,
					uchar8			*ocmdp)
{
	int32			rc;
	ushort16		fid;
	uint64			offset;
	uint32			dataLen, writtenDataLen;
	ushort16		writeMode, remaining, dataOffset, byteCnt;
	boolean			largeWriteSupport;
	uchar8			*bp;
	cifs_fileOpen	*fop;

	// Check if share is an IPC
	if (siup->flags & CIFS_SIU_ISIPC)
	{
		//TODO
		CIFS_PRINT(CIFS_DBGLV_WARN, ("cifsm_writeNX: IPC write is not implemented!\n"));
		return (CIFS_SENDRESPNOW);
	}

	// Check large write support
	largeWriteSupport = clp->serverp->cfg.capabilities & CAP_LARGE_WRITEX;

	//--------------------------------------
	// Decode word fields in request message
	//--------------------------------------

	fid = GETLE2(icmdp, SMBCO_WRITENX_FID);
	fop = cifs_findFileOpenByFid(opp->siup, fid);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_writeNX: Bad Fid %d", fid));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}

	offset = GETLE4(icmdp, SMBCO_WRITENX_OFFSET_LOW);
	if (GCAR(icmdp, SMBO_WCNT) == 14)
	{
		// Larege offset (64 bit) write
		uint64	offsetHigh = GETLE4(icmdp, SMBCO_WRITENX_OFFSET_HIGH);

		if (offsetHigh != 0)
		{
			if (largeWriteSupport == TRUE)
			{
				// The CIFS server support large write
				offset |= (offsetHigh << 32);
			}
			else
			{
				CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_writeNX: "
							"Server does not support CAP_LARGE_WRITEX"));
			}
		}
	}

	dataLen = GETLE2(icmdp, SMBCO_WRITENX_DATA_LEN_LOW);
	if (largeWriteSupport == TRUE)
	{
		dataLen |= (GETLE4(icmdp, SMBCO_WRITENX_DATA_LEN_HIGH) << 16);
	}

	writeMode = GETLE2(icmdp, SMBCO_WRITENX_WRITE_MODE);
	//remaining = GETLE2(icmdp, SMBCO_WRITENX_RAMAIN);
	dataOffset = GETLE2(icmdp, SMBCO_WRITENX_DATA_OFFSET);
	//byteCnt = GETLE2(icmdp, SMBCO_WRITENX_BYTE_CNT);


	// Validate dataOffset and dataLen
	if ((dataOffset > opp->reqMsg.smbLen) ||
		((dataOffset + dataLen) > opp->reqMsg.smbLen))
	{
		// Offset is beyond the end of the request message
		// Note: DOS error code is CIFSERR_badmem. What NTSTATUS should be?
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_UNSUCCESSFUL);
	}

	// Pointer to data to be written
	bp = opp->reqMsg.smbp + dataOffset;

	if (dataLen == 0)
	{
		// Nothing to write to file
		writtenDataLen = 0;
		rc = CIFSTATUS_OK;
	}
	else
	rc = cifs_writeFile(opp, fop, bp, offset, dataLen, 
						writeMode, &writtenDataLen);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//-------------------
	// Build the response
	//-------------------
	cifs_initNXCmdHeader(ocmdp);
	cifs_setCmdWordByteCnt(&opp->respMsg, 6, 0, TRUE);	
	PUTLE2(ocmdp, SMBCO_WP2, (writtenDataLen & 0xFF));// Number of bytes written
	PUTLE2(ocmdp, SMBCO_WP3, -1);				// Remaining (reserved)
	PUTLE4(ocmdp, SMBCO_WP4, 0);				// Reserved
	PUTLE2(ocmdp, SMBCO_WP6, 0);

	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
	{
		NFX_ASSERT(FALSE);				//TODO - handle batched request
		return cifs_processNextCmd(opp, ocmdp);
	}
	else
		return (CIFS_SENDRESPNOW);
} /* cifsm_writeNX() */

/*-----------------------------------------------------------------
 * Name			: cifsm_write
 *
 * Description	: Write a file (0x0B)
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/15/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_write(	cifs_client		*clp,
					cifs_session	*ssp,
					cifs_shareInUse	*siup,
					cifs_operation	*opp,
					uchar8			*icmdp,
					uchar8			*ocmdp)
{
	int32				rc;
	ushort16			fid;
	ushort16			bytesToWrite, bytesLeft, byteCnt;
	uint32				bytesWritten;
	uint32				offset;
	uchar8				*bp;			// Pointer to start of bytes to write
	cifs_fileOpen		*fop;

	// Check if share is an IPC
	if (siup->flags & CIFS_SIU_ISIPC)
	{
		//TODO
		CIFS_PRINT(CIFS_DBGLV_WARN, ("cifsm_writeNX: IPC write is not implemented!\n"));
		return (CIFS_SENDRESPNOW);
	}

	//--------------------------------------
	// Decode word fields in request message
	//--------------------------------------
	
	fid = GETLE2(icmdp, SMBCO_WP0);
	fop = cifs_findFileOpenByFid(opp->siup, fid);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_write: Bad Fid %d", fid));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}

	bytesToWrite = GETLE2(icmdp, SMBCO_WP1);
	offset = GETLE2(icmdp, SMBCO_WP2);
	//bytesLeft = GETLE2(icmdp, SMBCO_WP4);
	//byteCnt = GETLE2(icmdp, SMBCO_WP5);

	bp = SMBCP_BPAR(icmdp); 
	bp += 3;						// Pad is 3

	if (bytesToWrite == 0)
	{
		// File should be truncated or extended to the size indicated
		// by offset. If extended, the bytes between the old and new EOF
		// will be 0.

		//TODO: At present, we don't put 0 to file if extend the file

		fssf_attr_t			fssfAttr;
		eee_desc_t			*eDescp;
		fssf_setattrRsp_t	*setAttrResp;

		//--------------------------------------
		// Send request to FSSF to set file size 
		//--------------------------------------

		cifs_initFssfAttr(&fssfAttr);
		fssfAttr.bitmap = FSSF_ATTR_SIZE;
		fssfAttr.fsAttr.size = offset;
		rc = cifs_fssfSetattrByIdReq(opp, fop->stateid, &fssfAttr);
		cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

		//----------------------------
		// Wait for response from FSSF  
		//----------------------------
		rc = cifs_waitFssfResp(FSSF_SETATTR_RSP, (void**)&setAttrResp, &eDescp);
		cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
		cifsm_testFsStatus_ifErrReturn(opp,
										setAttrResp->status,
										eDescp);
		eee_freePacket(eDescp);
		bytesWritten = 0;
	}
	else
	{
		// Validate bytesToWrite
		if ((SMB_DIFF(bp, opp->reqMsg.smbp) + bytesToWrite) > opp->reqMsg.smbLen)
		{
			// Offset is beyond the end of the request message
			// Note: DOS error code is CIFSERR_badmem. What NTSTATUS should be?
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_UNSUCCESSFUL);
		}
		rc = cifs_writeFile(opp, fop, bp, offset,
							bytesToWrite, 0, &bytesWritten);
		cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	}

	//-------------------
	// Build the response
	//-------------------
	cifs_setCmdWordByteCnt(&opp->respMsg, 1, 0, TRUE);
	PUTLE2(ocmdp, SMBCO_WP0, (bytesWritten & 0xFF));

	return (CIFS_SENDRESPNOW);
} /* cifsm_write() */


