/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-rap.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 *
 * Name			: cifs-rap.c
 *
 * Description	: Contains functions related to Remote Administration
 *				  Protocol (RAP), which is an RPC like mechanism used 
 *				  by CIFS clients to submit requests to CIFS servers. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/03/2002
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-transport.h"
#include "cifs-encdec.h"
#include "cifs-err.h"
#include "cifs-msg.h"
#include "cifs-ipc.h"
#include "cifs-rap.h"


struct cifs_rap_cmd
{
	short16		cmdType;
	int32		(*processRapCmd)(cifs_operation		*opP,
								 uchar8				*parP,
								 uchar8				*datP,
								 ushort16			pbsCnt,
								 ushort16			dbsCnt,
								 char				**oparP,
								 uint32				*oparLen,
								 char				**odatP,
								 uint32				*odatLen);
	uint32		flags;
}  cifs_rap_cmds[4] = {
	{CIFS_RAP_CMD_NETSHAREENUM, cifs_rapNetShareEnum, 0},
	{CIFS_RAP_CMD_NETSHAREGETINFO, cifs_rapNetShareGetInfo, 0},
	{CIFS_RAP_CMD_NETSERVERGETINFO, cifs_rapNetServerGetInfo, 0},
	{-1, NULL, 0}
};

/*-----------------------------------------------------------------
 * Name			: cifs_rapCheckServerInfoLevel
 *
 * Description	: Check if level of detail about a server matches
 *				  the data descriptor string.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/04/2002
 * 
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_rapCheckServerInfoLevel(ushort16 level, char *datDescP)
{
	switch(level)
	{
		case 0:
			if (strcmp(datDescP, "B16") != 0)
				return (NFX_ERR);
			break;

		case 1:
			if (strcmp(datDescP, "B16BBDz") != 0)
				return (NFX_ERR);
			break;

		default:
				return (NFX_ERR);
			break;
	}
	return (NFX_OK);
} /* cifs_rapCheckServerInfoLevel() */

/*-----------------------------------------------------------------
 * Name			: cifs_rapCheckShareInfoLevel
 *
 * Description	: Check if level of detail about a share matches
 *				  the data descriptor string.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/04/2002
 * 
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_rapCheckShareInfoLevel(ushort16 level, char *datDescP)
{
	switch(level)
	{
		case 0:
			if (strcmp(datDescP, "B13") != 0)
				return (NFX_ERR);
			break;

		case 1:
			if (strcmp(datDescP, "B13BWz") != 0)
				return (NFX_ERR);
			break;

		case 2:
			if (strcmp(datDescP, "B13BWzWWWzB9B") != 0)
				return (NFX_ERR);
			break;

		default:
				return (NFX_ERR);
			break;
	}
	return (NFX_OK);
} /* cifs_rapCheckShareInfoLevel() */

/*-----------------------------------------------------------------
 * Name			: cifs_rapGetShareInfoSize
 *
 * Description	: Get share info size in number of bytes
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/07/2002
 * 
 *-----------------------------------------------------------------
 */
int32			// # of bytes of share info. 0 if failed
cifs_rapGetShareInfoSize(cifs_share	*shareP,
						 ushort16	level,		// Level of detail
						 uint32		*fixedLen,
						 uint32		*strLen)
{
    CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_rapGetShareInfoSize"));

	switch (level)
	{
		case 0:
			(*fixedLen) = 13;
			break;
		case 1:
			(*fixedLen) = 20;
			break;
		case 2:
			(*fixedLen) = 40;
			break;
		default:
			return (0);
	}

	if (level > 0)
	{
		// Get length of comment string (in ASCII)
		(*strLen) = unistrlen(shareP->cfg.comment);
		(*strLen)++;
	}

	if (level > 1)
	{
		// Get length of share path string
		if (unistrlen(shareP->cfg.pathName))
			(*strLen) += unistrlen(shareP->cfg.pathName) + 1;
	}

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_rapGetShareInfoSize"));

	return ((*fixedLen) + (*strLen));
} /* cifs_rapGetShareInfoSize() */

/*-----------------------------------------------------------------
 * Name			: cifs_rapFillShareInfo
 *
 * Description	: Prepare share info data of one share for response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/07/2002
 * 
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_rapFillShareInfo(cifs_share	*shareP,
					  cifs_server	*serverP,
					  ushort16		level,
					  char			**baseP,
					  char			**fixedBufP,
					  uint32		*totalFixedLen,
					  char			**strBufP,
					  uint32		*totalStrLen)
{
	char		*fbufP, *sbufP;
	uint32		fixedLen, strLen;
	char		name[CIFS_S_SHARENAME + 1];
	char		pathName[CIFS_S_PATHSTR + 1];
	char		comment[CIFS_S_SHARECOMMENT+1];
	ushort16	type;

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_rapFillShareInfo"));

	fbufP = *fixedBufP;
	sbufP = *strBufP;

	switch (level)
	{
		case 0:
			fixedLen = 13;
			break;
		case 1:
			fixedLen = 20;
			break;
		case 2:
			fixedLen = 40;
			break;
		default:
			return (NFX_ERR);
	}

	memset(name, 0, CIFS_S_SHARENAME + 1);
	memset(pathName, 0, CIFS_S_PATHSTR + 1);
	memset(comment, 0, CIFS_S_SHARECOMMENT+1);

	strUni2Dos(name, shareP->cfg.name, CIFS_S_SHARENAME);
	strUni2Dos(pathName, shareP->cfg.pathName, CIFS_S_PATHSTR);
	strUni2Dos(comment, shareP->cfg.comment, CIFS_S_SHARECOMMENT);

	if (*totalFixedLen < fixedLen)
		return (NFX_ERR);

	// Fill share name
	strcpy(fbufP, name);

	if (level > 0)
	{
		// Fill pad
		*(fbufP + 13) = 0;

		// Fill share type
		if (shareP->cfg.flags & CIFS_SHCFG_IPC)
			type = CIFS_SHARE_TYPE_IPC;
		else
			type = CIFS_SHARE_TYPE_DISK;

		PUTLE2(fbufP, 14, type);
		
		// Fill pointer to comment
		PUTLE4(fbufP, 16, SMB_DIFF(sbufP, *baseP));

		// Fill comment
		strcpy(sbufP, comment);
		if (strlen(comment))
		{
			strLen = strlen(sbufP) + 1;
			sbufP += strLen;
		}
	}

	if (level > 1)
	{
		// Fill permissions
		PUTLE2(fbufP, 20, (CIFS_RAP_SHARE_ACCESS_READ | \
							CIFS_RAP_SHARE_ACCESS_WRITE | \
							CIFS_RAP_SHARE_ACCESS_CREATE));

		// Fill max uses
		PUTLE2(fbufP, 22, -1);		// No limit on max uses

		// Fill current number of uses
		PUTLE2(fbufP, 24, shareP->useCnt);

		// Fill local path name
		if (shareP->cfg.flags & CIFS_SHCFG_ISADMIN)
		{
			PUTLE4(fbufP, 26, 0);	// For IPC$ & ADMIN$, the pointer is NULL
		}
		else
		{
			PUTLE4(fbufP, 26, SMB_DIFF(sbufP, *baseP));
			strcpy(sbufP, pathName);
			strLen += strlen(sbufP) + 1;
		}

		// Fill password and pad field
		memset(fbufP + 30, 0, 8 + 2);		// Password is NULL
					// Note: Samba's implementation on password did not follow 
					// 		 the spec draft-leach-cifs-rap-spec-00.txt
	}

	(*fixedBufP) += fixedLen;
	(*totalFixedLen) -= fixedLen;
	(*strBufP) += strLen;
	(*totalStrLen) -= strLen;

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_rapFillShareInfo"));

	return (NFX_OK);
} /* cifs_rapFillShareInfo() */

/*-----------------------------------------------------------------
 * Name			: cifs_rapNetShareEnum
 *
 * Description	: Process MSRAP NetShareEnum API. In response,
 * 				  it provides list of shares on a CIFS server.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/04/2002
 * 
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR 
cifs_rapNetShareEnum(cifs_operation	*opP,
					 uchar8			*parP,	// Ptr to parameters in MSRAP msg
					 uchar8			*datP,	// Ptr to data in MSRAP msg
					 ushort16		pbsCnt,	// Count of parameter bytes
					 ushort16		dbsCnt,	// Count of data bytes
					 char			**oparP,	// Ptr to params in response
					 uint32			*oparLen,	// # of param bytes in resp
					 char			**odatP,	// Ptr to data in response
					 uint32			*odatLen)	// # of bytes length in resp
{
	char		*parDescP;		// Ptr to parameter descriptor
	char		*datDescP;		// Ptr to data descriptor
	ushort16	level;			// Level of detail
	ushort16	rcvBufSize;		// Size of receive buffer
	uint32		offset;

	ushort16	retStatus = CIFSERR_success;		// Return status
	cifs_server	*serverP = NULL;
	cifs_share	*shareP = NULL;
	ushort16	totalShareCnt = 0, shareCnt = 0, count = 0;
	uint32		fixedLen = 0, strLen = 0, dataLen = 0;
	uint32		totalFixedLen = 0, totalStrLen = 0, totalDataLen = 0;
	char		*fixedBufP, *strBufP;
	int32		rc;

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_rapNetShareEnum"));

	// Extract descriptors and parameters
	offset = 2;
	parDescP = parP + offset;
	datDescP = parDescP + strlen(parDescP) + 1;
	offset += strlen(parDescP) + strlen(datDescP) + 2;

	level = GETLE2(parP, offset);
	rcvBufSize = GETLE2(parP, offset + 2);

	// Validate descriptors
	if (strcmp(parDescP, "WrLeh") != 0)
		return (NFX_ERR);

	if (cifs_rapCheckShareInfoLevel(level, datDescP) != NFX_OK)
	{
	    CIFS_PRINT(CIFS_DBGLV_INFO, ("cifs_rapNetShareEnum: "
					"Share info level %d does not match descriptor or "
					"is not supported", level));
		return (NFX_ERR);
	}

	serverP = opP->clp->serverp;
	totalShareCnt = serverP->shareCnt;

	// Calculate size of buffer to be used
	DLL_ITERATE(shareP, &serverP->shares, entry)
	{
		// TODO - Check user's access right at share level?

		dataLen = cifs_rapGetShareInfoSize(	shareP,
											level,
											&fixedLen,
											&strLen);
		if ((totalDataLen + dataLen) > rcvBufSize)
		{
			// TODO - need more than one message for response, because receive
			//        buffer is smaller than the size of total data to send?
		    CIFS_PRINT(CIFS_DBGLV_INFO, ("cifs_rapNetShareEnum: "
						"Total size of share info exceeds the "
						"client's receive buffer size"));
			retStatus = CIFSERR_moredata;
			break;
		}
		totalFixedLen += fixedLen;
		totalStrLen += strLen;
		totalDataLen += dataLen;
		shareCnt++;
	}

	if (totalDataLen == 0)
		return (NFX_ERR);

	if (totalDataLen > CIFS_RAP_S_BUF)
	{
		// TODO -
		// realloc(*odatP, totalDataLen);
		// memset(*odatP, 0, totalDataLen);
	}

	//-------------------
	// Build the response
	//-------------------

	count = shareCnt;
	fixedBufP = *odatP;					// Fixed data portion goes here
	strBufP = *odatP + totalFixedLen;	// Auxiliary data portion goes here

	// Fill shares info (including IPC$, ADMIN$, and hidden shares)
	DLL_ITERATE(shareP, &serverP->shares, entry)
	{
		// TODO - Check user's access right at share level?

		rc = cifs_rapFillShareInfo(	shareP,
									serverP,
									level,
									odatP,
									&fixedBufP,
									&totalFixedLen,
									&strBufP,
									&totalStrLen);
		if (rc == NFX_ERR)
			break;
		count--;
	}

	// Fill parameters
	PUTLE2(*oparP, 0, retStatus);			// Return status
	PUTLE2(*oparP, 2, 0);					// Converter word = 0
	PUTLE2(*oparP, 4, shareCnt - count);	// # of entries returned
											// Normally, 'count' should be 0
	PUTLE2(*oparP, 6, totalShareCnt);		// Total # of available entries

	*oparLen = 8;
	*odatLen = totalDataLen;

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_rapNetShareEnum"));

	return (NFX_OK);
} /* cifs_rapNetShareEnum() */

/*-----------------------------------------------------------------
 * Name			: cifs_rapNetShareGetInfo
 *
 * Description	: Process MSRAP NetShareGetInfo API. In response,
 * 				  it provides information about a particular share
 * 				  on a CIFS server.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/04/2002
 * 
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR 
cifs_rapNetShareGetInfo(cifs_operation	*opP,
						uchar8			*parP,	// Ptr to params in MSRAP msg
						uchar8			*datP,	// Ptr to data in MSRAP msg
						ushort16		pbsCnt,	// Count of parameter bytes
						ushort16		dbsCnt,	// Count of data bytes
						char			**oparP,	// Ptr to params in response
						uint32			*oparLen,	// # of param bytes in resp
						char			**odatP,	// Ptr to data in response
						uint32			*odatLen)	// # of bytes length in resp
{
	char		*parDescP;		// Ptr to parameter descriptor
	char		*datDescP;		// Ptr to data descriptor
	char		*reqShareName;	// Name of share in request
	ushort16	level;			// Level of detail
	ushort16	rcvBufSize;		// Size of receive buffer
	uint32		offset;

	ushort16	retStatus = CIFSERR_success;		// Return status
	cifs_server	*serverP = NULL;
	cifs_share	*shareP = NULL;
	char		shareName[CIFS_S_SHARENAME+1];
	uint32		fixedLen = 0, strLen = 0, dataLen = 0;
	char		*fixedBufP, *strBufP;
	int32		rc = NFX_OK;

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_rapNetShareGetInfo"));

	// Extract descriptors and parameters
	offset = 2;
	parDescP = parP + offset;
	datDescP = parDescP + strlen(parDescP) + 1;
	offset += strlen(parDescP) + strlen(datDescP) + 2;

	reqShareName = parP + offset;
	offset += strlen(reqShareName) + 1;

	level = GETLE2(parP, offset);
	rcvBufSize = GETLE2(parP, offset + 2);

	// Validate descriptors
	if (strcmp(parDescP, "zWrLh") != 0)
		return (NFX_ERR);

	if (cifs_rapCheckShareInfoLevel(level, datDescP) != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_INFO, ("cifs_rapNetShareGetInfo: "
					"Share info level %d does not match descriptor or "
					"is not supported", level));
		return (NFX_ERR);
	}

	// Get the share control block
	serverP = opP->clp->serverp;
	DLL_ITERATE(shareP, &serverP->shares, entry)
	{
		memset(shareName, 0, CIFS_S_SHARENAME + 1);
		strUni2Dos(shareName, shareP->cfg.name, CIFS_S_SHARENAME);
		if (strcmp(reqShareName, shareName) == 0)
			break;
	}

	if (shareP == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_rapNetShareGetInfo: "
					"Couldn't find share %s", reqShareName));
		return (NFX_ERR);
	}
	else
	{
		// TODO - Check user's access right at share level
		if (retStatus == CIFSERR_noaccess ||
			retStatus == CIFSERR_nonetaccess)
		{
			rc = NFX_ERR;
		}
	}

	if (rc != NFX_ERR)
	{
		// Calculate size of buffer to be used
		dataLen = cifs_rapGetShareInfoSize(	shareP,
											level,
											&fixedLen,
											&strLen);

		//-------------------
		// Build the response
		//-------------------

		// Fill data section
		fixedBufP = *odatP;
		strBufP = *odatP + fixedLen;
		rc = cifs_rapFillShareInfo(	shareP,
									serverP,
									level,
									odatP,
									&fixedBufP,
									&fixedLen,
									&strBufP,
									&strLen);

		if (dataLen > rcvBufSize)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_rapNetShareGetInfo: "
						"Total size of server info exceeds the client's "
						"receive buffer size"));
			retStatus = CIFSERR_moredata;
		}
	}

	// Fill parameters
	PUTLE2(*oparP, 0, retStatus);		// Return status
	PUTLE2(*oparP, 2, 0);				// Converter word = 0
	PUTLE2(*oparP, 4, dataLen);			// Total number of available bytes

	*oparLen = 6;
	*odatLen = MIN(dataLen, rcvBufSize);

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_rapNetShareGetInfo"));
	return (NFX_OK);
} /* cifs_rapNetShareGetInfo() */

/*-----------------------------------------------------------------
 * Name			: cifs_rapNetServerGetInfo
 *
 * Description	: Process MSRAP NetServerGetInfo API. In response,
 * 				  it provides information about the specified CIFS server.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/04/2002
 * 
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR 
cifs_rapNetServerGetInfo(cifs_operation	*opP,
						uchar8			*parP,	// Ptr to params in MSRAP msg
						uchar8			*datP,	// Ptr to data in MSRAP msg
						ushort16		pbsCnt,	// Count of parameter bytes
						ushort16		dbsCnt,	// Count of data bytes
						char			**oparP,	// Ptr to params in response
						uint32			*oparLen,	// # of param bytes in resp
						char			**odatP,	// Ptr to data in response
						uint32			*odatLen)	// # of bytes length in resp
{
	char		*parDescP;		// Ptr to parameter descriptor
	char		*datDescP;		// Ptr to data descriptor
	ushort16	level;			// Level of detail
	ushort16	rcvBufSize;		// Size of receive buffer
	uint32		offset;

	ushort16	retStatus = CIFSERR_success;		// Return status
	cifs_server	*serverP = NULL;
	uint32		fixedLen = 0, strLen = 0, dataLen = 0;
	char		comment[CIFS_S_SERVERCOMMENT+1];

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_rapNetServerGetInfo"));

	// Extract descriptors and parameters
	offset = 2;
	parDescP = parP + offset;
	datDescP = parDescP + strlen(parDescP) + 1;
	offset += strlen(parDescP) + strlen(datDescP) + 2;

	level = GETLE2(parP, offset);
	rcvBufSize = GETLE2(parP, offset + 2);

	// Validate descriptors
	if (strcmp(parDescP, "WrLh") != 0)
		return (NFX_ERR);

	if (cifs_rapCheckServerInfoLevel(level, datDescP) != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_INFO, ("cifs_rapNetServerGetInfo: "
					"Server info level %d does not match descriptor or "
					"is not supported", level));
		return (NFX_ERR);
	}

	serverP = opP->clp->serverp;

	//-------------------
	// Build the response
	//-------------------

	// Fill in server name (Question: Need to convert to uppercase or not?)
	strcpy(*odatP, serverP->cfg.name);
	fixedLen = 16;

	// Fill other info about the server
	if (level > 0)
	{
		PCAR(*odatP, fixedLen, CIFS_VERSION_MAJOR);
		PCAR(*odatP, fixedLen + 1, CIFS_VERSION_MINOR);
		PUTLE4(*odatP, fixedLen + 2, CIFS_SERVER_TYPE);
		fixedLen += 6;	

		// Fill pointer to comment
		PUTLE4(*odatP, fixedLen, fixedLen + 4);
		fixedLen += 4;

		// Fill comment
		strUni2Dos(comment, serverP->cfg.comment, CIFS_S_SERVERCOMMENT);
		strcpy(*odatP + fixedLen, comment);
		strLen = strlen(comment) + 1;
	}

	dataLen = fixedLen + strLen;
	if (dataLen > rcvBufSize)
	{
		CIFS_PRINT(CIFS_DBGLV_INFO, ("cifs_rapNetServerGetInfo: "
					"Total size of server info exceeds the "
					"client's receive buffer size"));
		retStatus = CIFSERR_moredata;
	}
	
	// Fill parameters
	PUTLE2(*oparP, 0, retStatus);		// Return status
	PUTLE2(*oparP, 2, 0);				// Converter word = 0
	PUTLE2(*oparP, 4, dataLen);			// Total number of available bytes

	*oparLen = 6;
	*odatLen = MIN(dataLen, rcvBufSize);

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_rapNetServerGetInfo"));
	return (NFX_OK);
} /* cifs_rapNetServerGetInfo() */

/*-----------------------------------------------------------------
 * Name			: cifs_msrap
 *
 * Description	: Microsoft RAP operations (server API)
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/03/2002
 * 
 *-----------------------------------------------------------------
 */
int32 			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifs_msrap(	cifs_operation	*opp,
			uchar8			*parp,		// Ptr to parameters in MSRAP msg
			uchar8			*datp,		// Ptr to data in MSRAP msg
			ushort16		pbsCnt, 	// Count of parameter bytes
			ushort16		dbsCnt)		// Count of data bytes
{
	ushort16	rapCmdType;
	int32		result, i;
	
	uint32		oparLen;
	char		*oparP;			// Parameter bytes in response message
	uint32		odatLen;
	char		*odatP;			// Data bytes in response message

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_msrap"));

	// Get RAP command type
	rapCmdType = GETLE2(parp, 0);
	
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_msrap: cmd %d", rapCmdType));

	// Look for RAP command
	for (i = 0; cifs_rap_cmds[i].cmdType != -1; i++)
	{
		if (cifs_rap_cmds[i].cmdType == rapCmdType)
			break;
	}

	if (cifs_rap_cmds[i].cmdType == -1)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_msrap: "
					"Unsupported MSRAP API (%d)", rapCmdType));
		result = NFX_ERR;
	}
	else
	{
		CIFS_RAM_ALLOC_ZERO(oparP, char, CIFS_RAP_S_BUF);
		CIFS_RAM_ALLOC_ZERO(odatP, char, CIFS_RAP_S_BUF);

		if (!oparP || !odatP)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_msrap: Insufficient memory"));
			result = NFX_ERR;
		}
		else
			result = (*cifs_rap_cmds[i].processRapCmd)(opp,
														parp,
														datp,
														pbsCnt,
														dbsCnt,
														&oparP,
														&oparLen,
														&odatP,
														&odatLen);
	}

	if (result == NFX_ERR)
	{
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_nosupport, 0);
	}
	else
	{
		cifs_replyTransac(opp, oparP, oparLen, odatP, odatLen, FALSE);
	}

	CIFS_RAM_DEALLOC(oparP);
	CIFS_RAM_DEALLOC(odatP);

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_msrap"));

	return (CIFS_SENDRESPNOW);
} /* cifs_msrap() */

