/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-sec.c,v 1.3 2002/12/10 18:22:20 laurentb Exp $
 *
 * Name			: cifs-sec.c
 *
 * Description	: Contain CIFS security functions.
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/24/2002
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-err.h"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)

sec_accessGenericMap cifs_fileAccessGenericMap = {
	SEC_GENR_FILE_READ,
	SEC_GENR_FILE_WRITE,
	SEC_GENR_FILE_EXECUTE, 
	SEC_GENR_FILE_ALL
};


//Well knowns SIDs
sec_cifsSid cifsec_nullSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_NULL_SID_AUTHORITY, { CIFSEC_NULL_RID } };

sec_cifsSid cifsec_worldSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_WORLD_SID_AUTHORITY, { CIFSEC_WORLD_RID } };

sec_cifsSid cifsec_creatorOwnerSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_CREATOR_SID_AUTHORITY, 
	{ CIFSEC_CREATOR_OWNER_RID } };

sec_cifsSid cifsec_creatorGroupSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_CREATOR_SID_AUTHORITY, 
	{ CIFSEC_CREATOR_GROUP_RID } };

sec_cifsSid cifsec_creatorOwnerServerSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_CREATOR_SID_AUTHORITY, 
	{ CIFSEC_CREATOR_OWNER_SERVER_RID } };

sec_cifsSid cifsec_creatorGroupServerSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_CREATOR_SID_AUTHORITY, 
	{ CIFSEC_CREATOR_GROUP_SERVER_RID } };
	
sec_cifsSid cifsec_networkSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_NT_AUTHORITY, { CIFSEC_NETWORK_RID } };

sec_cifsSid cifsec_anonymousSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_NT_AUTHORITY, { CIFSEC_ANONYMOUS_LOGON_RID } };

sec_cifsSid cifsec_AuthenticatedUserSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_NT_AUTHORITY, 
	{ CIFSEC_AUTHENTICATED_USER_RID } };

sec_cifsSid cifsec_builtInDomainSid = 
{ CIFSEC_SID_REVISION, 1, CIFSEC_NT_AUTHORITY, { CIFSEC_BUILTIN_DOMAIN_RID } };

sec_cifsSid cifsec_builtInAdminSid = 
{ CIFSEC_SID_REVISION, 2, CIFSEC_NT_AUTHORITY, 
	{ CIFSEC_BUILTIN_DOMAIN_RID, CIFSEC_DOM_USER_RID_ADMIN } };

sec_cifsSid cifsec_builtInGuestSid = 
{ CIFSEC_SID_REVISION, 2, CIFSEC_NT_AUTHORITY, 
	{ CIFSEC_BUILTIN_DOMAIN_RID, CIFSEC_DOM_USER_RID_GUEST } };

//Anonymous token
sec_cifsAccessToken cifsec_anonymousToken = { 
//User access token entry
{ { CIFSEC_SID_REVISION, 1, CIFSEC_NT_AUTHORITY, {CIFSEC_ANONYMOUS_LOGON_RID}},
	0 },
//Nb of groups	
2,
//Groups Access token entries
{
{ { CIFSEC_SID_REVISION, 1, CIFSEC_WORLD_SID_AUTHORITY, { CIFSEC_WORLD_RID} }, 
	CIFSEC_GROUP_ENABLED|CIFSEC_GROUP_OWNER },
{ { CIFSEC_SID_REVISION, 1, CIFSEC_NT_AUTHORITY, { CIFSEC_NETWORK_RID } },
	CIFSEC_GROUP_ENABLED } }
};
	


//****************************************************************************
//********************************* tools ************************************
//****************************************************************************


static char sidStr[512];	

/*-----------------------------------------------------------------
 * Name			: cifs_secSidSprintf
 *
 * Description	: sprintf of a security identifier
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/06/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_secSidSprintf(char *str, sec_cifsSid *sidp)
{
	char tmpStr[50];
	uint32 i;
	uint64 auth;

	auth =	(sidp->authority.val[5])
		  +	(sidp->authority.val[4] << 8)
		  +	(sidp->authority.val[3] << 16)
		  +	(sidp->authority.val[2] << 24);
		  /*TODO Add those when compiling on real system 
		  +	(sidp->authority.val[1] << 32)
		  +	(sidp->authority.val[0] << 48);*/

	sprintf(str, "S-%u-0x%llx", sidp->revision, auth);
	for (i=0; i<sidp->nbSubAuthy; i++)
	{
		sprintf(tmpStr, "-0x%x", sidp->subAuthy[i]);
		strcat(str, tmpStr);
	}
}

/*-----------------------------------------------------------------
 * Name			: cifs_secSidPrintf
 *
 * Description	: print a security identifier
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/06/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_secSidPrintf(sec_cifsSid *sidp)
{
	cifs_secSidSprintf(sidStr, sidp);
	printf("%s\n", sidStr);
}

	
//****************************************************************************
//************* allocation/duplication/deallocation functions ****************
//****************************************************************************

/*-----------------------------------------------------------------
 * Name			: cifs_secFreeSecDesc
 *
 * Description	: free Security Descriptor components if needed
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/25/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_secFreeSecDesc(sec_cifsSd *sdp, boolean dynSd)
{
	if (sdp->sAcl.acep != NULL)
		CIFS_RAM_DEALLOC(sdp->sAcl.acep);
	
	if (sdp->dAcl.acep != NULL)
		CIFS_RAM_DEALLOC(sdp->dAcl.acep);

	if (dynSd)
		CIFS_RAM_DEALLOC(sdp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secAcesClone
 *
 * Description	: clone the ACEs of an ACL 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
sec_cifsAce *	//The cloned ACEs
cifs_secAcesClone(sec_cifsAcl *aclp)
{
	sec_cifsAce *cloneAcep;
	uint32 size = sizeof(sec_cifsAce) * aclp->nbAces;

	CIFS_RAM_ALLOC(cloneAcep, sec_cifsAce, size);
	if (cloneAcep == NULL)
		return NULL;

	//Copy all the ACEs of the source ACL
	memcpy(cloneAcep, aclp->acep, size);	
	
	return (cloneAcep);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secSecDescCopy
 *
 * Description	: copy a security descriptor
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
int32    //NFX_OK or NFX_ERR
cifs_secSecDescCopy(sec_cifsSd *dsdp, sec_cifsSd *ssdp)
{
	//Copy all fields
	memcpy(dsdp, ssdp, sizeof(sec_cifsSd));
	
	//Clone the ACLs
	if ((ssdp->control & CIFSEC_SD_SACL_PRESENT) && ssdp->sAcl.nbAces != 0)
	{
		dsdp->sAcl.acep = cifs_secAcesClone(&ssdp->sAcl);
		if (dsdp->sAcl.acep == NULL)
			return (NFX_ERR);
	}
	if ((ssdp->control & CIFSEC_SD_DACL_PRESENT) && ssdp->dAcl.nbAces != 0)
	{
		dsdp->dAcl.acep = cifs_secAcesClone(&ssdp->dAcl);
		if (dsdp->dAcl.acep == NULL)
			return (NFX_ERR);
	}
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secSecDescClone
 *
 * Description	: clone a security descriptor
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
sec_cifsSd *	//The cloned security descriptor
cifs_secSecDescClone(sec_cifsSd *sdp)
{
	sec_cifsSd *cloneSdp;

	CIFS_RAM_ALLOC(cloneSdp, sec_cifsSd, sizeof(sec_cifsSd));
	if (cloneSdp == NULL)
		return NULL;

	if (cifs_secSecDescCopy(cloneSdp, sdp) != NFX_OK)
	{
		cifs_secFreeSecDesc(cloneSdp, TRUE);
		return (NULL);
	}
	
	return (cloneSdp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secSidCpy
 *
 * Description	: copy a security identifier
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/06/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_secSidCpy(sec_cifsSid *dsidp, sec_cifsSid *ssidp)
{
	uint32 i;
	
	if (dsidp == NULL || ssidp == NULL)
		return;

	dsidp->revision		= ssidp->revision;
	dsidp->nbSubAuthy	= ssidp->nbSubAuthy;
	memcpy(dsidp->authority.val, ssidp->authority.val, sizeof(sec_cifsSidAuthy));
	
	if (dsidp->nbSubAuthy > CIFSEC_SID_MAX_SUB_AUTHORITIES)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_secSidCpy: "
					"too many sub-authorities"));
		dsidp->nbSubAuthy = CIFSEC_SID_MAX_SUB_AUTHORITIES;
	}
	for (i = 0; i < dsidp->nbSubAuthy; i++)
		dsidp->subAuthy[i] = ssidp->subAuthy[i];
	
}


/*-----------------------------------------------------------------
 * Name			: cifs_secMergeSidAndRid
 *
 * Description	: Add an Rid at the end of the subauthorities list of
 *				  an SID
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/06/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR 
cifs_secMergeSidAndRid(sec_cifsSid *dsidp, sec_cifsSid *ssidp, uint32 rid)
{
	if (dsidp != ssidp)
		cifs_secSidCpy(dsidp, ssidp);
	
	if (dsidp->nbSubAuthy >= CIFSEC_SID_MAX_SUB_AUTHORITIES)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_secMergeSidAndRid: "
					"too many sub-authorities"));
		return (NFX_ERR);
	}

	dsidp->subAuthy[dsidp->nbSubAuthy++] = rid;

	return (NFX_OK);
}



/*-----------------------------------------------------------------
 * Name			: cifs_secFindRidInCreds
 *
 * Description	: Find a specific rid in the user credentials
 * 				  (Check if enabled)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/06/2002
 *
 *-----------------------------------------------------------------
 */
boolean	//TRUE if found 
cifs_secFindRidInCreds(cifsauthCred_t *credsp, uint32 rid)
{
	uint32 i;

	for (i=0; i<credsp->num_gids; i++)
	{
		if (credsp->gids[i].g_rid == rid)
		{
			if (credsp->gids[i].attr & CIFSEC_GROUP_ENABLED)
				return (TRUE);
			else
				return (FALSE);
		}
	}

	return (FALSE);
}


		
//****************************************************************************
//************************ comparison functions ******************************
//****************************************************************************


/*-----------------------------------------------------------------
 * Name			: cifs_secSidCmp
 *
 * Description	: compare 2 security identifiers
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
int32 	//0 if equal
cifs_secSidCmp(sec_cifsSid *sidp1, sec_cifsSid *sidp2)
{
	uint32 i;
	
	if (sidp1 == NULL && sidp2 == NULL)
		return (0);
	
	if (sidp1 == NULL || sidp2 == NULL)
		return (1);
	
	//Check revision
	if (sidp1->revision != sidp2->revision)
		return (2);

	//Check number of sub authorities
	if (sidp1->nbSubAuthy != sidp2->nbSubAuthy)
        return (3);
	
	//Check authority
	if (memcmp(sidp1->authority.val, sidp2->authority.val,	
				sizeof(sec_cifsSidAuthy)) != 0)
		return (4);
	
	//Check sub authorities
	for (i = 0; i < sidp1->nbSubAuthy; i++)
	{
		if (sidp1->subAuthy[i] != sidp2->subAuthy[i])
			return (5 + i);
	}
	
	return (0);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secAceCmp
 *
 * Description	: compare 2 ACEs (audit/access control entries)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
int32 	//0 if equal
cifs_secAceCmp(sec_cifsAce *acep1, sec_cifsAce *acep2)
{
	
	if (acep1 == NULL && acep2 == NULL)
		return (0);
	
	if (acep1 == NULL || acep2 == NULL)
		return (1);
	
	//Check header
	if (acep1->type != acep2->type)
		return (2);
	if (acep1->flags != acep2->flags)
        return (3);
	
	//Check access mask
	if (acep1->accessMask != acep2->accessMask)
		return (4);
	
	//Check SID 
	if (cifs_secSidCmp(&acep1->sid, &acep2->sid) != 0)
		return (5);
	
	return (0);
}



/*-----------------------------------------------------------------
 * Name			: cifs_secAclCmp
 *
 * Description	: compare 2 ACLs (audit/access control lists)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
int32 	//0 if equal
cifs_secAclCmp(sec_cifsAcl *aclp1, sec_cifsAcl *aclp2)
{
	uint32 i, j;
	boolean aceChecked;
	
	if (aclp1 == NULL && aclp2 == NULL)
		return (0);
	
	if (aclp1 == NULL || aclp2 == NULL)
		return (1);
	
	//Check revision
	if (aclp1->revision != aclp2->revision)
		return (2);

	//Check number of ACEs
	if (aclp1->nbAces != aclp2->nbAces)
        return (3);
	
	//Check ACEs
	for (i = 0; i < aclp1->nbAces; i++)
	{
		aceChecked = FALSE;
		for (j = 0; j < aclp1->nbAces; j++)
		{
			if (cifs_secAceCmp(aclp1->acep + i, aclp2->acep + j) != 0)
			{
				aceChecked = TRUE;
				break;
			}
		}
		if (aceChecked == FALSE) 
			return (4 + i);
	}
	
	return (0);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secSecDescCmp
 *
 * Description	: compare 2 security descriptors
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/25/2002
 *
 *-----------------------------------------------------------------
 */
int32 	//0 if equal
cifs_secSecDescCmp(sec_cifsSd *sdp1, sec_cifsSd *sdp2)
{
	
	if (sdp1 == NULL && sdp2 == NULL)
		return (0);
	
	if (sdp1 == NULL || sdp2 == NULL)
		return (1);
	
	//Check header
	if (sdp1->revision != sdp2->revision)
		return (2);
	if (sdp1->control != sdp2->control)
        return (3);
	
	//Check owner
	if (cifs_secSidCmp(&sdp1->ownerSid, &sdp2->ownerSid) != 0)
		return (4);
	
	//Check primary group 
	if (cifs_secSidCmp(&sdp1->groupSid, &sdp2->groupSid) != 0)
		return (5);
	
	//Check system ACL
	if (cifs_secAclCmp(&sdp1->sAcl, &sdp2->sAcl) != 0)
		return (6);
	
	//Check discretionary ACL
	if (cifs_secAclCmp(&sdp1->dAcl, &sdp2->dAcl) != 0)
		return (7);
	
	return (0);
}




//****************************************************************************
//********************** access control functions ****************************
//****************************************************************************

/*-----------------------------------------------------------------
 * Name			: cifs_secMapGenericAccess
 *
 * Description	: Map generic access bits to object specific access bits
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_secMapGenericAccess(sec_accessMask *accessp, sec_accessGenericMap *mapp)
{
	if (*accessp & SEC_GENR_READ_ACCESS)
	{
		*accessp &= ~SEC_GENR_READ_ACCESS;
		*accessp |= mapp->genericRead;
	}
	if (*accessp & SEC_GENR_WRITE_ACCESS)
	{
		*accessp &= ~SEC_GENR_WRITE_ACCESS;
		*accessp |= mapp->genericWrite;
	}
	if (*accessp & SEC_GENR_EXECUTE_ACCESS)
	{
		*accessp &= ~SEC_GENR_EXECUTE_ACCESS;
		*accessp |= mapp->genericExecute;
	}
	if (*accessp & SEC_GENR_ALL_ACCESS)
	{
		*accessp &= ~SEC_GENR_ALL_ACCESS;
		*accessp |= mapp->genericAll;
	}
}



/*-----------------------------------------------------------------
 * Name			: cifs_secSDMapGenericAccessBits
 *
 * Description	: Map generic access bits to object specific access bits
 * 				  in ACEs of the discretionary access control list of
 * 				  the security descriptor
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_secSDMapGenericAccessBits(sec_cifsSd *sdp, sec_accessGenericMap *mapp)
{
	uint32 i;
	sec_cifsAce *acep;

	acep = sdp->dAcl.acep;
	if (acep == NULL)
		return;

	for (i=0; i<sdp->dAcl.nbAces; i++)
	{
		cifs_secMapGenericAccess(&acep->accessMask, mapp);
		acep++;
	}
}


/*-----------------------------------------------------------------
 * Name			: cifs_secCreateAccessToken
 *
 * Description	: create an access token from user credentials got
 * 				  from DC if present
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/04/2002
 *
 *-----------------------------------------------------------------
 */
sec_cifsAccessToken *	//Access token created
cifs_secCreateAccessToken(cifsauthCred_t *credsp)
{
	sec_cifsAccessToken *atp;
	sec_cifsSid domSid;
	uint32 nbWellKnownSids, i, rid;
		
	if (credsp == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_secCreateAccToken: "
					"use anonymous token"));
		return (&cifsec_anonymousToken);
	}
	
	//Find out how many well-known sids we are going to add
	nbWellKnownSids = 0;	//World, Network, auth users already in token struct

	//This is ugly, but that's the only way to build the token without going
	//to the DC to get all the alias (local groups) including the user or the
	//global groups the user belongs to
	//NB: This only works for well-known Alias Users and Admins
	//    We will have to understand that further
	if (cifs_secFindRidInCreds(credsp, CIFSEC_DOM_GROUP_RID_ADMINS))
		nbWellKnownSids++;

	if (cifs_secFindRidInCreds(credsp, CIFSEC_DOM_GROUP_RID_USERS))
		nbWellKnownSids++;
	
	CIFS_RAM_ALLOC(atp, sec_cifsAccessToken,
					sizeof(sec_cifsAccessToken) 
				+ (credsp->num_gids + nbWellKnownSids) * sizeof(sec_cifsSid));
	
	if (atp == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_secCreateAccToken: "
					"failed to alloc token"));
		return (NULL);
	}
	
	//Build domain SID
	//TODO check if we should use this one or our domain SID 
	//(get it from Samba on the SSC)
	domSid.revision		= credsp->sid_rev_num;
	domSid.nbSubAuthy	= credsp->num_auths;
	if (domSid.nbSubAuthy >= CIFSEC_SID_MAX_SUB_AUTHORITIES)
	{
		domSid.nbSubAuthy = CIFSEC_SID_MAX_SUB_AUTHORITIES - 1;
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_secCreateAccToken: "
					"domSid has too many subsauths"));
	}
	memcpy(domSid.authority.val, credsp->id_auth, sizeof(sec_cifsSidAuthy));
	for (i=0; i<domSid.nbSubAuthy; i++)
		domSid.subAuthy[i] = credsp->sub_auths[i];
	
	//Create User SID
	cifs_secMergeSidAndRid(&atp->user.sid, &domSid, credsp->user_rid);
	atp->user.attributes = 0;

	//Create Groups SIDs
	for (i=0; i<credsp->num_gids; i++)
	{
		rid = credsp->gids[i].g_rid;
		cifs_secMergeSidAndRid(&atp->groupSid[i].sid, &domSid, rid);
		atp->groupSid[i].attributes = credsp->gids[i].attr;
		if (rid == credsp->group_rid)
			atp->groupSid[i].attributes |= CIFSEC_GROUP_OWNER;
	}

	//Add the default SIDs
	//NB: the number of those has to be in sync with the RAM_ALLOC above !!!!!!
	cifs_secSidCpy(&atp->groupSid[i].sid, &cifsec_worldSid);
	atp->groupSid[i++].attributes = CIFSEC_GROUP_ENABLED;
	cifs_secSidCpy(&atp->groupSid[i].sid, &cifsec_networkSid);
	atp->groupSid[i++].attributes = CIFSEC_GROUP_ENABLED;
	if (0)	//TODO add test for guest user (under study)
	{
	} else {
		cifs_secSidCpy(&atp->groupSid[i].sid, &cifsec_AuthenticatedUserSid);
		atp->groupSid[i++].attributes = CIFSEC_GROUP_ENABLED;
		
		if (cifs_secFindRidInCreds(credsp, CIFSEC_DOM_GROUP_RID_ADMINS))
		{
			cifs_secMergeSidAndRid(	&atp->groupSid[i].sid, 
									&cifsec_builtInDomainSid,
									CIFSEC_DOM_ALIAS_RID_ADMINS);
			atp->groupSid[i++].attributes = CIFSEC_GROUP_ENABLED;
		}

		if (cifs_secFindRidInCreds(credsp, CIFSEC_DOM_GROUP_RID_USERS))
		{
			cifs_secMergeSidAndRid(	&atp->groupSid[i].sid, 
									&cifsec_builtInDomainSid,
									CIFSEC_DOM_ALIAS_RID_USERS);
			atp->groupSid[i++].attributes = CIFSEC_GROUP_ENABLED;
		}
	}

	//Store the number of groups
	atp->nbGroups = i;
					
	IFDBG(CIFS_DBGLV_DBG3)
	{
		printf("cifs_secCreateAccToken:\n");
		printf("User : at=0x%02x ", atp->user.attributes);
		cifs_secSidPrintf(&atp->user.sid);
		for (i=0; i<atp->nbGroups; i++)
		{
			printf("Grp%2d: at=0x%02x ", i, atp->groupSid[i].attributes);
			cifs_secSidPrintf(&atp->groupSid[i].sid);
		}
	}
	
	return (atp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secDeleteAccessToken
 *
 * Description	: delete an access token
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/12/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_secDeleteAccessToken(sec_cifsAccessToken *atp)
{
	if (atp != &cifsec_anonymousToken)
		CIFS_RAM_DEALLOC(atp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secAceAppliesToToken
 *
 * Description	: Check if the Sid of the ACE is in the Access token
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/07/2002
 *
 *-----------------------------------------------------------------
 */
static boolean
cifs_secAceAppliesToToken(sec_cifsAce *acep, sec_cifsAccessToken *atp)
{
	uint32 i;
	
	if (cifs_secSidCmp(&acep->sid, &atp->user.sid) == 0)
		return (TRUE);
	for (i=0; i<atp->nbGroups; i++)
		if (cifs_secSidCmp(&acep->sid, &atp->groupSid[i].sid))
			return (TRUE);
	return (FALSE);		
}
		

/*-----------------------------------------------------------------
 * Name			: cifs_secCheckAceAccess
 *
 * Description	: Check access to an access control entry against 
 * 				  an access token
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/07/2002
 *
 *-----------------------------------------------------------------
 */
static uint32  //CIFSTATUS
cifs_secCheckAceAccess(	sec_cifsAce *acep,
						sec_cifsAccessToken *atp, 
						sec_accessMask *accRequired)
{

	if (acep->flags & CIFSEC_ACE_FL_INHERIT_ONLY
	 || cifs_secAceAppliesToToken(acep, atp) == FALSE)
		return (CIFSTATUS_OK);

	switch (acep->type)
	{
		case CIFSEC_ACE_TYPE_ACCESS_ALLOWED:
			//Remove access granted by this ACE from the access required
			*accRequired &= acep->accessMask;
			break;
			
		case CIFSEC_ACE_TYPE_ACCESS_DENIED:
			//If any access still required is denied by this ACE, deny access
			if (*accRequired & acep->accessMask)
				return (CIFSTATUS_ACCESS_DENIED);
			break;
		default:
			return (CIFSTATUS_INVALID_PARAMETER);
	}
	
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifs_secCheckMaxAccess
 *
 * Description	: Check maximum access against an access token
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/07/2002
 *
 *-----------------------------------------------------------------
 */
static uint32  //CIFSTATUS
cifs_secCheckMaxAccess(	sec_cifsAce *acep,
						uint32 nbAces,
						sec_cifsAccessToken *atp, 
						sec_accessMask accRequired,
						sec_accessMask *accGivenp)
{
	uint32 i, accGiven, accDenied;

	accGiven = accDenied = 0;
	
	for (i=0; i<nbAces; i++)
	{
		if ((acep->flags & CIFSEC_ACE_FL_INHERIT_ONLY) == 0
		 && cifs_secAceAppliesToToken(acep, atp))
		{
			switch (acep->type)
			{
				case CIFSEC_ACE_TYPE_ACCESS_ALLOWED:
				//Give everything from the ace rights that has not been denied
				//so far
				accGiven |= (acep->accessMask & ~accDenied);
				break;
			
				case CIFSEC_ACE_TYPE_ACCESS_DENIED:
				//Deny everything from the ace rights that has not been given
				//so far
				accDenied |= (acep->accessMask & ~accGiven);
					return (CIFSTATUS_ACCESS_DENIED);
				break;
				
				default:
					*accGivenp = 0;
					return (CIFSTATUS_INVALID_PARAMETER);
			}
		}
		acep++;
	}
	
	if (accGiven == 0 || (accGiven & accRequired) != accRequired)
	{
		*accGivenp = 0;
		return (CIFSTATUS_INVALID_PARAMETER);
	} else {
		*accGivenp = accGiven;		
		return (CIFSTATUS_OK);
	}
}


/*-----------------------------------------------------------------
 * Name			: cifs_secCheckAccess
 *
 * Description	: Check access to an security object (through its security
 * 				  descriptor) against an access token
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/06/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_secCheckAccess(sec_cifsSd *sdp,
					sec_cifsAccessToken *atp, 
					sec_accessMask accRequested,
					sec_accessMask *accGivenp)
{
	sec_cifsAce *acep;
	sec_accessMask accRequiredInDAcl;
	uint32 i, status;
		

	if (sdp == NULL
	 || !(sdp->control & CIFSEC_SD_DACL_PRESENT)
	 || sdp->dAclOff == 0)
	{
		if (accRequested & SEC_MAXIMUM_ALLOWED_ACCESS)
			*accGivenp = SEC_STDR_ALL_ACCESS | SEC_SPECR_ACCESS_MASK;
		else 
			*accGivenp = accRequested;
		return (CIFSTATUS_OK);
	}

	*accGivenp = 0;
	accRequiredInDAcl = accRequested;
	status = CIFSTATUS_OK;
	acep = sdp->dAcl.acep;
		
	IFDBG(CIFS_DBGLV_DBG3)
	{
		printf("cifs_secCheckAccess - Access Token:\n");
		printf("User : at=0x%02x ", atp->user.attributes);
		cifs_secSidPrintf(&atp->user.sid);
		for (i=0; i<atp->nbGroups; i++)
		{
			printf("Grp%2d: at=0x%02x ", i, atp->groupSid[i].attributes);
			cifs_secSidPrintf(&atp->groupSid[i].sid);
		}
		printf("cifs_secCheckAccess - Security Descriptor:\n");
		if (sdp->ownerSidOff != 0)
		{
			printf("Owner        :");
			cifs_secSidPrintf(&sdp->ownerSid);
		}
		printf("discretionary ACL:\n");
		for (i=0; i<sdp->dAcl.nbAces; i++)
		{
			printf("Ty=%x Fl=%02x Ac=0x%08x ", acep[i].type, acep[i].flags,
					acep[i].accessMask);
			cifs_secSidPrintf(&acep[i].sid);
		}
	}

	//Check whether the token includes the owner of the Security Descriptor
	//If so, grant Read and Write DAC access whatever the dACL says
	if (sdp->ownerSidOff != 0) 
	{	//Owner SID is present
		if (cifs_secSidCmp(&sdp->ownerSid, &atp->user.sid) == 0)
			accRequiredInDAcl &= ~(SEC_STDR_READ_CONTROL_ACCESS
								 | SEC_STDR_WRITE_DAC_ACCESS);
		else {
			for (i=0; i<sdp->dAcl.nbAces; i++)
			{
				if (cifs_secSidCmp(&sdp->ownerSid, &acep[i].sid) == 0)
				{
					accRequiredInDAcl &= ~(SEC_STDR_READ_CONTROL_ACCESS
										 | SEC_STDR_WRITE_DAC_ACCESS);
					break;
				}
			}
		}
	}
	
	if (accRequiredInDAcl & SEC_MAXIMUM_ALLOWED_ACCESS)
	{
		accRequiredInDAcl &= ~SEC_MAXIMUM_ALLOWED_ACCESS;
		return (cifs_secCheckMaxAccess(acep, sdp->dAcl.nbAces, atp, 
										accRequiredInDAcl, accGivenp));
	}

	for (i=0; i<sdp->dAcl.nbAces; i++)
	{
		status = cifs_secCheckAceAccess(acep, atp, &accRequiredInDAcl);
		if (status != CIFSTATUS_OK)
		{	//Access has been denied
			*accGivenp = 0;
			return status;
		}
		if (accRequiredInDAcl == 0)
			break;
		acep++;
	}
	
	if (accRequiredInDAcl == 0)
	{
		*accGivenp = accRequested;
        return (CIFSTATUS_OK);
	} else {
		*accGivenp = 0;
		return (CIFSTATUS_ACCESS_DENIED);
	}
}
	

/*-----------------------------------------------------------------
 * Name			: cifs_secCheckShareAccess
 *
 * Description	: Check access requested against access granted at share
 * 				  level
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/05/2002
 *
 *-----------------------------------------------------------------
 *
uint32  //CIFSTATUS
cifs_secCheckShareAccess(cifs_shareInUse *siup, sec_accessMask accRequested)
{
	if ((siup->access & accRequested) == accRequested)
		return (CIFSTATUS_OK);
	return (CIFSTATUS_ACCESS_DENIED);
}
*/






