/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-unicode.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 * 
 * Name			: cifs-unicode.c
 *
 * Description	: Contains unicode utility functions.
 * 				  NB: Assumes that we are running on a little endian CPU
 * 				  TODO Finalize optimizations induced by this assumption
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/20/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-err.h"
#include "cifs-transport.h"

#define UNIMAP_ISDIGIT		0x01
#define UNIMAP_ISHEXADIGIT	0x02
#define UNIMAP_ISLOWER		0x04
#define UNIMAP_ISUPPER		0x08
#define UNIMAP_ISSPACE		0x10

#define UNIMAP_S_LOWERTABLE	\
		(sizeof(unimap_lowerTable)/sizeof(unimap_charProperties))
#define UNIMAP_S_UPPERTABLE	\
		(sizeof(unimap_upperTable)/sizeof(unimap_charProperties))
		
#define UNIMAP_CHARPROPERTIES(c) \
		((c < UNIMAP_S_LOWERTABLE) ? unimap_lowerTable[c].properties : \
		 ((c < 0x10000 - UNIMAP_S_UPPERTABLE) ? 0 : \
		  (unimap_upperTable[c - (0x10000 - UNIMAP_S_UPPERTABLE)].properties) ))
		
#define UNIMAP_LOWER(c) \
		((c < UNIMAP_S_LOWERTABLE) ? unimap_lowerTable[c].lowerCase : \
		 ((c < 0x10000 - UNIMAP_S_UPPERTABLE) ? c : \
		  (unimap_upperTable[c - (0x10000 - UNIMAP_S_UPPERTABLE)].lowerCase) ))
		
#define UNIMAP_UPPER(c) \
		((c < UNIMAP_S_LOWERTABLE) ? unimap_lowerTable[c].upperCase : \
		 ((c < 0x10000 - UNIMAP_S_UPPERTABLE) ? c : \
		  (unimap_upperTable[c - (0x10000 - UNIMAP_S_UPPERTABLE)].upperCase) ))
		

// Local declarations
typedef struct {
	uni16	lowerCase;
	uni16	upperCase;
	uchar8	properties;
}unimap_charProperties;

static unimap_charProperties unimap_lowerTable[] = 
{
#include "cifs-unimap-l.h"
};
static unimap_charProperties unimap_upperTable[] = 
{
#include "cifs-unimap-u.h"
};

static uni16 char_dos2uni16[0x100];
static char  char_uni162dos[0x10000];



/*-----------------------------------------------------------------
 * Name			: cifs_initDosUnicodeMapping
 *
 * Description	: Initialize Dos codepage to unicode and unicode to dos
 * 				  conversion tables
 * 				  For 1st version we only consider codepage 437 (US)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */

uni16 dos437upper_part[0x80] = {
  0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,	//80 - 87
  0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
  0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,	//90
  0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x20a7, 0x0192,
  0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,	//A0
  0x00bf, 0x2310, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
  0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,	//B0
  0x2555, 0x2563, 0x2551, 0x2557, 0x255d, 0x255c, 0x255b, 0x2510,
  0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x255e, 0x255f,	//C0
  0x255a, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256c, 0x2567,
  0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256b,	//D0
  0x256a, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
  0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x00b5, 0x03c4,	//E0
  0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
  0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,	//F0
  0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0	//F8 - FF
};

void cifs_initDosUnicodeMapping(void)
{
	int i;

	//By default convert unicode to underline character
	memset(char_uni162dos, '_', sizeof(char_uni162dos));
	
	//Initialize lower part of both tables straight
	for (i=0; i<0x80; i++) {
		char_dos2uni16[i] = i;
		char_uni162dos[i] = i;
	}
	
	//Initialize upper part of Dos to Unicode table
	memcpy(char_dos2uni16 + 0x80, dos437upper_part, sizeof(dos437upper_part));
	
	//From there finish initialize Unicode to Dos table
	for (i=0x80; i<0x100; i++) 
		char_uni162dos[char_dos2uni16[i]] = i;
}


/*-----------------------------------------------------------------
 * Name			: strDos2Uni
 *
 * Description	: Convert a string in current dos codepage to
 *				  a MS unicode string.
 *				  If zeroEnd=TRUE, adds zero character at the end of 
 *				  the destination string.
 *				  Assumes dos codepage character are 1 byte only
 *				  Will require enhancement for 2 bytes codepages
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */
uchar8*	//pointer on address after the SOURCE string ending 0, NULL if error
strDos2Uni(uni16 *dstp, const char *srcp, ushort16 maxDestLen, boolean zeroEnd)
{
	uchar8 *sp = (uchar8 *)srcp;

	if (zeroEnd) 
		maxDestLen--;
	while (*sp) {
		PUTLE2(dstp, 0, char_dos2uni16[*sp]);
		dstp++;
		sp++;	
		if (--maxDestLen == 0 && *sp)
		{
			if (zeroEnd) 
				PUTLE2(dstp, 0, 0);
			return (NULL);
		}
	}
	if (zeroEnd) {
		PUTLE2(dstp, 0, 0);
	}

	return (sp + 1);
}	// strDos2Uni()
		


/*-----------------------------------------------------------------
 * Name			: strUni2Dos
 *
 * Description	: Convert a string in unicode to a current dos codepage string.
 *				  Assumes dos codepage character are 1 byte only
 *				  Will require enhancement for 2 bytes codepages
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/20/2001
 *
 *-----------------------------------------------------------------
 */
uchar8*	//pointer on address after the DESTINATION string ending 0, NULL if err
strUni2Dos(char *dstp, uni16 *srcp, ushort16 maxDestLen)
{
	uint32 rc = CIFSTATUS_OK;

	if (!dstp)
		return (NULL);
	
	maxDestLen--;	//ending 0
	
	while (*srcp)
	{
		*dstp = char_uni162dos[GETLE2(srcp, 0)];
		dstp++;
		srcp++;
		if (--maxDestLen == 0 && (*srcp))
		{
			*dstp = 0;
			return (NULL);
		}
	}
	*dstp = 0;
	
	return (uchar8*)(dstp + 1);
}	// strUni2Dos()


/*-----------------------------------------------------------------
 * Name			: uniprintf
 *
 * Description	: print a unicode string with an optional comment in front
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/20/2001
 *
 *-----------------------------------------------------------------
 */
void uniprintf(char *comment, uni16* unis)
{
static char str[1024];
	//Just use Unicode to Dos conversion for now
	strUni2Dos(str, unis, sizeof(str));
	printf("%s%s", comment, str);
}



/*-----------------------------------------------------------------
 * Name			: unitolower
 *
 * Description	: unicode tolower
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
uni16 unitolower(uni16 c)
{
	c = GETLE2(&c, 0);				//Unicode is in little endian
	return UNIMAP_LOWER(c);
}


/*-----------------------------------------------------------------
 * Name			: unitoupper
 *
 * Description	: unicode toupper
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
uni16 unitoupper(uni16 c)
{
	c = GETLE2(&c, 0);				//Unicode is in little endian
	return UNIMAP_UPPER(c);
}


/*-----------------------------------------------------------------
 * Name			: uniislower
 *
 * Description	: unicode islower
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
boolean uniislower(uni16 c)
{
	c = GETLE2(&c, 0);				//Unicode is in little endian
	return (UNIMAP_CHARPROPERTIES(GETLE2(&c, 0)) & UNIMAP_ISLOWER);
}
		

/*-----------------------------------------------------------------
 * Name			: uniisupper
 *
 * Description	: unicode isupper
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
boolean uniisupper(uni16 c)
{
	c = GETLE2(&c, 0);				//Unicode is in little endian
	return (UNIMAP_CHARPROPERTIES(GETLE2(&c, 0)) & UNIMAP_ISUPPER);
}
		

/*-----------------------------------------------------------------
 * Name			: uniisdigit
 *
 * Description	: unicode isdigit
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
boolean uniisdigit(uni16 c)
{
	c = GETLE2(&c, 0);				//Unicode is in little endian
	return (UNIMAP_CHARPROPERTIES(c) & UNIMAP_ISDIGIT);
}
		

/*-----------------------------------------------------------------
 * Name			: uniishexadigit
 *
 * Description	: unicode ishexadigit
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
boolean uniishexadigit(uni16 c)
{
	c = GETLE2(&c, 0);				//Unicode is in little endian
	return (UNIMAP_CHARPROPERTIES(c) & UNIMAP_ISHEXADIGIT);
}
		

/*-----------------------------------------------------------------
 * Name			: uniisspace
 *
 * Description	: unicode isspace
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
boolean uniisspace(uni16 c)
{
	c = GETLE2(&c, 0);				//Unicode is in little endian
	return (UNIMAP_CHARPROPERTIES(c) & UNIMAP_ISSPACE);
}
		

/*-----------------------------------------------------------------
 * Name			: unistrcpy
 *
 * Description	: unicode strcpy
 *				  returns length in chars
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/20/2001
 *
 *-----------------------------------------------------------------
 */
uni16 *unistrcpy(uni16 *dstp, uni16 *srcp)
{
	if (!dstp)
		return (NULL);
	*dstp = 0;
	if (srcp)
	{
		while ((*dstp++ = *srcp++) != 0);
	}
	return (dstp);
}


/*-----------------------------------------------------------------
 * Name			: unistrncpy
 *
 * Description	: unicode strncpy
 *				  returns length in chars
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/04/2001
 *
 *-----------------------------------------------------------------
 */
uni16 *unistrncpy(uni16 *dstp, uni16 *srcp, ushort16 n)
{
	uni16 *dp;

	if (!dstp || !srcp)
		return (dstp);

	dp = dstp;

	for (; *srcp && n; dp++, srcp++, n--)
		*dp = *srcp;

	while (n--)
		*dp++ = '\0';

	return (dstp);
}


/*-----------------------------------------------------------------
 * Name			: unistrlen
 *
 * Description	: unicode strlen
 *				  returns length in characters
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/20/2001
 *
 *-----------------------------------------------------------------
 */
ushort16 unistrlen(uni16 *s)
{ 
	int32	n;

	if (!s)
		return (0);

	for (n = 0; *s; s++)
		n++;

	return (n);
}


/*-----------------------------------------------------------------
 * Name			: unistrsize
 *
 * Description	: size in bytes of a unicode string including ending 0
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/20/2001
 *
 *-----------------------------------------------------------------
 */
ushort16 unistrsize(uni16 *s)
{ 
	int32	n;

	if (!s)
		return (0);

	for (n = 0; *s; s++)
		n+=2;

	return (n+2);
}


/*-----------------------------------------------------------------
 * Name			: unistrchr
 *
 * Description	: return a ptr to c in p, return NULL if not found
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/28/2001
 *
 *-----------------------------------------------------------------
 */
uni16 *unistrchr(uni16 *s, uni16 c)
{ 
	if (!s)
		return (0);

	for (; *s; s++)
		if (*s == c) return (s);

	return (0);
}


/*-----------------------------------------------------------------
 * Name			: unistrrchr
 *
 * Description	: returns a pointer to the last occurrence of the
 * 				  character c in the string s
 * 				  Can give the string length back if requested
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/19/2002
 *
 *-----------------------------------------------------------------
 */
uni16 *unistrrchr(uni16 *s, uni16 c, uint32 *slen)
{ 
	int32 len;
	
	if (!s)
		return (0);

	len = unistrlen(s);
	if (slen)
		*slen = len;
	s += len - 1;
	for (; len > 0; len--)
	{
		if (*s == c) return (s);
		s--;
	}

	return (0);
}


/*-----------------------------------------------------------------
 * Name			: unistrstr
 *
 * Description	: unicode 16 strstr
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/20/2002
 *
 *-----------------------------------------------------------------
 */
uni16 *unistrstr(uni16 *s, uni16 *subs)
{ 
	uni16 *a, *b;

    b = subs;
    if (*b == 0)
        return s;
    for ( ; *s != 0; s++)
    {
        if (*s != *b)
            continue;
        a = s;
        while (1)
        {
            if (*b == 0)
                return s;
            if (*a++ != *b++)
                break;
        }
        b = subs;
    }
    return NULL;

}


/*-----------------------------------------------------------------
 * Name			: unistrcmp
 *
 * Description	: string comparison, return < 0, 0, or > 0
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/28/2001
 *
 *-----------------------------------------------------------------
 */
int32 unistrcmp(uni16 *s1, uni16 *s2)
{ 
	while (*s1 && *s2 && (*s1 == *s2))
	{
		s1++; s2++;
	}
	return (*s1 - *s2);	
}


/*-----------------------------------------------------------------
 * Name			: unistrncmp
 *
 * Description	: string comparison of at most n characters
 * 				  Return < 0, 0, or > 0
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/20/2002
 *
 *-----------------------------------------------------------------
 */
int32 unistrncmp(uni16 *s1, uni16 *s2, ushort16 n)
{ 
	while ((n) && (*s1) && (*s2) && (*s1 == *s2))
	{
		n--; s1++; s2++;
	}

	if (n == 0)					//n first characters are identical
		return (0);
	else						//strings are different or shorter than n
		return (*s1 - *s2);
}


/*-----------------------------------------------------------------
 * Name			: unistrcasecmp
 *
 * Description	: case insensitive string comparison, return < 0, 0, or > 0
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
int32 unistrcasecmp(uni16 *s1, uni16 *s2)
{ 

	while ((*s1) && (*s2) && (unitoupper(*s1) == unitoupper(*s2)))
	{
		s1++; s2++;
	}
	return (unitoupper(*s1) - unitoupper(*s2));
}


/*-----------------------------------------------------------------
 * Name			: unistrncasecmp
 *
 * Description	: case insensitive string comparison of at most n characters
 * 				  Return < 0, 0, or > 0
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
int32 unistrncasecmp(uni16 *s1, uni16 *s2, ushort16 n)
{ 
	while ((n) && (*s1) && (*s2) && (unitoupper(*s1) == unitoupper(*s2)))
	{
		n--; s1++; s2++;
	}

	if (n == 0)					//n first characters are identical
		return (0);
	else						//strings are different or shorter than n
		return (unitoupper(*s1) - unitoupper(*s2));
}


/*-----------------------------------------------------------------
 * Name			: uniCharReplace
 *
 * Description	: replace all occurence of a character in the string
 * 					by another
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/19/2002
 *
 *-----------------------------------------------------------------
 */
void uniCharReplace(uni16 *s, uni16 charToReplace, uni16 replacementChar)
{ 
	for (; *s; s++)
		if (*s == charToReplace) *s = replacementChar;
}


/*-----------------------------------------------------------------
 * Name			: strUtf82Uni
 *
 * Description	: Convert a string in UTF-8 format to a MS unicode string.
 *				  If zeroEnd=TRUE, adds zero character at the end of 
 *				  the destination string.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFSTATUS
strUtf82Uni(uni16		*destp,		// UNICODE string
			const char	*srcp,		// UTF-8 string
			uint32		srcLen,		// Number of bytes of UTF-8 string
			uint32		*dstLen,	// Number of chars of UNICODE string
			uint32		maxDstLen,	// Maximum number of bytes of UNICODE string
			boolean		zeroEnd)	// Put 0 at the end of UNICODE string
{
	uint32	len;	// # of bytes of UTF-8 string for one character
	uchar8	*sp = (uchar8 *)srcp;
	uni16	*dstp;

	dstp = destp;
	*dstLen = 0;

	while(srcLen >= 1 && (*dstLen)*2 < maxDstLen)
	{
		if ((*sp & 0x80) == 0)
		{
			len = 1;
			*dstp = *sp;
		}
		else if ((*sp & 0xF0) == 0xE0)
		{
			if (srcLen < 3)
				// Bad UTF-8 string
				return (CIFSTATUS_OBJECT_NAME_INVALID);

			len = 3;
			*dstp = ((*sp & 0x0F) << 12) |
					((*(sp + 1) & 0x3F) << 6) |
					(*(sp + 2) & 0x3F);
		}
		else if ((*sp & 0xE0) == 0xC0)
		{
			if (srcLen < 2)
				// Bad UTF-8 string 
				return (CIFSTATUS_OBJECT_NAME_INVALID);

			len = 2;
			*dstp = ((*sp & 0x1F) << 6) | (*(sp + 1) & 0x3F);
		}

		dstp += 1;
		*dstLen += 1;
		sp += len;
		srcLen -= len;
	}

	if (srcLen > 0)
	{
		// UTF-8 string is too long
		return (CIFSTATUS_NAME_TOO_LONG);
	}

	if (zeroEnd & (GETLE2(dstp - 1, 0) != 0))
	{
		if (*dstLen < maxDstLen)
			PUTLE2(dstp, 0, 0);
		else
			return (CIFSTATUS_NAME_TOO_LONG);
	}

	return (CIFSTATUS_OK);
} /* strUtf82Uni() */



/*-----------------------------------------------------------------
 * Name			: strUni2Utf8
 *
 * Description	: Convert a MS unicode string to UTF-8 format.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/28/2002
 *
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or CIFS internal error
strUni2Utf8(uchar8		*destp,		// UTF-8 string
			uni16		*srcp,		// UNICODE string
			uint32		srcLen,		// UNICODE string length (in characters)
			uint32		*dstLen,	// Number of bytes of UTF-8 string
			uint32		maxDstLen)	// Maximum number of bytes of UTF-8 string
{
	uint32	len;	// # of bytes of UTF-8 string for one character
	uni16	src;
	uchar8	*dstp;

	dstp = destp;
	*dstLen = 0;

	maxDstLen--;		//Reserve one byte for terminating '\0'
	while(srcLen >= 1)
	{
		src = GETLE2(srcp, 0);

		if (src <= 0x7F)
		{
			if (maxDstLen < *dstLen + 1)
				// UNICODE string is too long
				return (CIFSTATUS_NAME_TOO_LONG);
			len = 1;
			*dstp = src & 0x7F;
		}
		else if (src <= 0x7FF)
		{
			if (maxDstLen < *dstLen + 2)
				// UNICODE string is too long
				return (CIFSTATUS_NAME_TOO_LONG);
			len = 2;
			*(dstp + 1) = (src & 0x3F) | 0x80;
			*dstp       = ((src >> 6) & 0x1F) | 0xC0;
		}
		else if (src <= 0xFFFF)
		{
			if (maxDstLen < *dstLen + 3)
				// UNICODE string is too long
				return (CIFSTATUS_NAME_TOO_LONG);
			len = 3;
			*(dstp + 2) = (src & 0x3F) | 0x80;
			*(dstp + 1) = ((src >> 6) & 0x3F) | 0x80;
			*dstp       = ((src >> 12) & 0x0F) | 0xE0;
		}

		dstp += len;
		*dstLen += len;
		srcp += 1;
		srcLen -= 1;
	}

	//Add ending 0
	*dstp = 0;

	return (CIFSTATUS_OK);
} /* strUni2Utf8() */




