/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs.c,v 1.2 2002/10/24 21:12:32 isabellec Exp $
 *
 * Name			: cifs.c
 *
 * Description	: Contains CIFS initialization and utility functions.
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 10/31/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-cfg.h"
#include "cifs-transport.h"
#include "cifs-filetools.h"
#include "cifs-intmsg.h"

// Extern declaration
extern void cifs_msgRcv(eee_desc_t *eDescp);

// Global definitions declarations
cifs_globalCfg		*cifs_globalCfgp;
cifs_objLists		*cifs_objectsp;
uchar8				cifs_debugLevel = CIFS_DBGLV_DBG3;

uint32 cifs_elog_map[7] = {
	0,
	LOG_ERR,						// 1, CIFS_DBGLV_ERR
	LOG_WARNING,					// 2, CIFS_DBGLV_WARN
	LOG_INFO,						// 3, CIFS_DBGLV_INFO
	LOG_INFO /*LOG_DEBUG*/,			// 4, CIFS_DBGLV_DBG
	LOG_INFO /*LOG_DEBUG*/,			// 5, CIFS_DBGLV_DBG2
	LOG_INFO /*LOG_DEBUG*/			// 6, CIFS_DBGLV_DBG3
};


// Local declarations
int32 cifs_initNextShare(void);


//***************************************************************************
//************************ Miscellaneous Tools ******************************
//***************************************************************************
	
/*-----------------------------------------------------------------
 * Name			: cifs_createEncryptionKey
 *
 * Description	: create an Encryption key for Challenge/response
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/16/2001
 *-----------------------------------------------------------------
 */
void
cifs_createEncryptionKey(char *encryptKey)
{
static char testKey[] = { 0x1d, 0xed, 0xb6, 0x3f, 0x2f, 0x62, 0xf4, 0xd8 };
/*		
	random_r(&rbuf, (int32*)encryptKey);		//Fill first 4 bytes
	random_r(&rbuf, (int32*)encryptKey+4);		//Fill last 4 bytes

	*((int32*)encryptKey)    =  random();		
	*((int32*)(encryptKey+4))  =  random();		
	*((int32*)(encryptKey+4)) +=  random();//so 4 first bytes != 4 last bytes
*/
	memcpy(encryptKey, testKey, 8);	
}


/*-----------------------------------------------------------------
 * Name			: cifs_newFid
 *
 * Description	: create a new fid for a share in use
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/03/2001
 *-----------------------------------------------------------------
 */
ushort16	//new fid 
cifs_newFid(cifs_shareInUse *siup)
{
	boolean alreadyUsed;
	
	if (siup->flags & CIFS_SIU_ISIPC)
	{	//share is an IPC
		cifs_pipe *pip;
		for (;;) 
		{
			siup->fidCnt++;
			if (siup->fidCnt == 0)
				siup->fidCnt++;
			alreadyUsed = FALSE;
			
			DLL_ITERATE(pip, &siup->u.pipe.openPipes, entry)
			{
				if (pip->fid == siup->fidCnt)
				{
					alreadyUsed = TRUE;
					break;
				}
			}
			if (alreadyUsed == FALSE)
				break;
		}
	} 
	else 
	{	//share is a disk share
    	cifs_fileOpen *dkp;
		for (;;) 
		{
			siup->fidCnt++;
			if (siup->fidCnt == 0)
				siup->fidCnt++;
			alreadyUsed = FALSE;
			DLL_ITERATE(dkp, &siup->u.disk.openFiles, entry)
			{
				if (dkp->fid == siup->fidCnt)
				{
					alreadyUsed = TRUE;
					break;
				}
			}
			if (alreadyUsed == FALSE)
				break;
		}
	}
	return (siup->fidCnt);
}


/*-----------------------------------------------------------------
 * Name			: cifs_newTid
 *
 * Description	: create a new Tid for a client
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/29/2001
 *-----------------------------------------------------------------
 */
ushort16	//new TID 
cifs_newTid(cifs_client *clp)
{
	cifs_session *ssp;
	cifs_shareInUse *siup;
	boolean alreadyUsed;
	
	for (;;) 
	{
		clp->tidCnt++;
		if (clp->tidCnt == 0)
			clp->tidCnt++;
		alreadyUsed = FALSE;
		DLL_ITERATE(ssp, &clp->sessions, entry)
		{
			DLL_ITERATE(siup, &ssp->sharesInUse, entry)
			{
				if (siup->tid == clp->tidCnt)
				{
					alreadyUsed = TRUE;
					break;
				}
			}
			if (alreadyUsed == TRUE)
				break;
		}
		if (alreadyUsed == FALSE)
			break;
	}
	return (clp->tidCnt);
}


/*-----------------------------------------------------------------
 * Name			: cifs_newUid
 *
 * Description	: create a new Uid for a client
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/27/2001
 *-----------------------------------------------------------------
 */
ushort16	//new UID 
cifs_newUid(cifs_client *clp)
{
	cifs_session *ssp;
	boolean alreadyUsed;
	
	for (;;) 
	{
		clp->uidCnt++;
		if (clp->uidCnt == 0)
			clp->uidCnt++;
		alreadyUsed = FALSE;
		DLL_ITERATE(ssp, &clp->sessions, entry)
		{
			if (ssp->uid == clp->uidCnt)
			{
				alreadyUsed = TRUE;
				break;
			}
		}
		if (alreadyUsed == FALSE)
			break;
	}
	return (clp->uidCnt);
}


/*-----------------------------------------------------------------
 * Name			: cifs_newSrvrId
 *
 * Description	: create a new server Id for a server
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *-----------------------------------------------------------------
 *
uchar8			//New server Id
cifs_newSrvrId()
{
 TODO SEE IF WE FINALLY USE THIS ?
	cifs_server	*svp = cifs_globalCfgp->serversp;
	uchar8		*serverCounter;
	boolean		alreadyUsed;

	serverCounter = &cifs_globalCfgp->serverCounter;
	for(;;)
	{
		*serverCounter++;
		alreadyUsed = FALSE;
		while(svp != NULL)
		{
			if (svp->serverId == *serverCounter)
			{
				alreadyUsed = TRUE;
				break;
			}
			svp = svp->nextp;
		}
		if (alreadyUsed == FALSE)
			break;
	}
	return (*serverCounter);
}
*/


/*-----------------------------------------------------------------
 * Name			: cifs_newCliId
 *
 * Description	: create a new client Id for a client
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *-----------------------------------------------------------------
 *
ushort16		//New client Id
cifs_newCliId(cifs_server *svp)
{
		
	cifs_client	*clp = svp->clientsp;
	ushort16	*clientCounter;
	boolean		alreadyUsed;

	clientCounter = &svp->clientCounter;
	for(;;)
	{
		*clientCounter++;
		alreadyUsed = FALSE;
		while(clp != NULL)
		{
			if (clp->clientId == *clientCounter)
			{
				alreadyUsed = TRUE;
				break;
			}
			clp = clp->nextp;
		}
		if (alreadyUsed == FALSE)
			break;
	}
	return (*clientCounter);
}
*/

#ifndef FS_STANDALONE

/*-----------------------------------------------------------------
 * Name			: cifs_getVolumeName
 *
 * Description	: Extract volume name and path name from the share path.
 * 				  Example: Input share path = /Vol/Dir1/Dir2
 * 				           Output volume name = Vol, path name = Dir1/Dir2
 *
 * Created by	: Xiaobing Zhang
 *  
 * Date Created	: 03/15/2002
 *          
 *-----------------------------------------------------------------
 */         
int32			// NFX_OK or NFX_ERR
cifs_getVolumeName(cifs_share	*sharep,
					uchar8		*volName,		// ASCII string
					uchar8		*pathName,		// UTF-8 string
					uint32		*pathNameLen)
{
	uni16	*v, *p;

	v = sharep->cfg.pathName;
	while(*v == '/')
		v++;						//Point to the beginning of volume name
	if (!*v)
		return (NFX_ERR);			//No names in the input string

	p = unistrchr(v, '/');			//Point to the end of volume name
	if (p)
		*p = 0;

	// Convert volume name from UNICODE to ASCII
	strUni2Dos(volName, v, VOL_MAX_NAME);

	if (p)
	{
		// Volume name is followed by path name
		*p++ = '/';					//Store '/' back

		// Convert path name from UNICODE to UTF-8
		strUni2Utf8(pathName, p, unistrlen(p)*2, pathNameLen, FS_PATH_MAX);
		*(pathName + *pathNameLen) = 0;
	}
	else
	{
		// No path name is the input string
		*pathName = 0;
		*pathNameLen = 0;
	}

	return (NFX_OK);
}


//***************************************************************************
//************************ Initializations **********************************
//***************************************************************************


/*-----------------------------------------------------------------
 * Name			: cifs_volIdReq
 *
 * Description	: This function sends a request to the EVM on SSC
 *				  to retrieve the Volume Id for the input name.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/08/2002
 *
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_volIdReq(uchar8 *volName,			// Volmue name in ASCII
			  cifs_share *sharep)
{
	int32				rc;
	uint32				srcPort;
	uint32				destPort;
	evm_MsgGetId_t		evmGetIdReq;

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_volIdReq: Enter"));

	// Build EVM Get Id message
	memset(&evmGetIdReq, 0, sizeof(evm_MsgGetId_t));
	evmGetIdReq.hdr.msgType = EVM_VOL_GET_ID;
	evmGetIdReq.hdr.clientPrivate = sharep->threadId;
	strcpy(evmGetIdReq.name, volName);

	// Send message to EVM
	rc = cifs_msgSend(CIFS_INTMSG_TO_EVM,
					  (uchar8 *)&evmGetIdReq,
					  sizeof(evm_MsgGetId_t),
					  FALSE);

	if (rc != NFX_OK)
		return (NFX_ERR); 

	return (NFX_OK);
} /* cifs_volIdReq() */

#endif

/*-----------------------------------------------------------------
 * Name			: cifs_shareHandleReq 
 *
 * Description	: This function sends a request to FS to
 *				  retrieve the file handle of a share.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/08/2002
 *
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_shareHandleReq(cifs_share	*sharep,
					uchar8		*pathName,		// Share path name
					uint32		pathNameLen,	// Length of share path name
					evm_volId_t	*volIdp)
{
	int32				rc;
	fs_MsgGetRootReq	*fsGetRootReqp;
	uchar8				*buf;
	uint32				msgLen;

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_shareHandleReq: Enter"));

	// Allocate message buffer
	msgLen = sizeof(fs_MsgGetRootReq) + pathNameLen + 1;
	CIFS_RAM_ALLOC_ZERO(buf, uchar8, msgLen);
	//TODO: Handle error
	NFX_ASSERT(buf != NULL);

	// Build FS get root handle message 
	fsGetRootReqp = (fs_MsgGetRootReq *)buf;
	cifs_msgBuildFSHdr(&fsGetRootReqp->hdr, FS_REQ_GET_ROOT, 0, TRUE,
											sharep->threadId);
	fsGetRootReqp->volId.whole_thing = volIdp->whole_thing;
	fsGetRootReqp->pathLen = pathNameLen;
	//strcpy(fsGetRootReqp->pathName, pathName);
	memcpy(fsGetRootReqp->pathName, pathName, pathNameLen + 1);

	// Send message to FS
	rc = cifs_msgSend(CIFS_INTMSG_TO_FS,
					  (uchar8 *)fsGetRootReqp,
					  msgLen,
					  TRUE);

	if (rc != NFX_OK)
	{
		return (NFX_ERR); 
	}

	return (NFX_OK);
} /* cifs_shareHandleReq() */


#ifndef FS_STANDALONE	

/*-----------------------------------------------------------------
 * Name			: cifs_getShareHandle
 *
 * Description	: This function is the start point to get a share handle 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/08/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_getShareHandle(esm_handle_t handle, void *infop)
{
	int32			rc = NFX_OK;
	cifs_server		*serverp = (cifs_server *)infop;
	cifs_share		*sharep;
	uchar8			volName[VOL_MAX_NAME + 1];	// ASCII string
	uchar8			pathName[FS_PATH_MAX];		// UTF-8 string
	uint32			pathNameLen;				// UTF-8 pathname len in bytes
	esm_event_t		evt;
	boolean			error = FALSE, freePkt = FALSE;

	// Search for the share
	sharep = cifs_findServerShareByThread(serverp, handle);
	if (sharep == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
					"No share with thread ID %d", handle));
		error = TRUE;
		goto cifs_getShareHandle_done;
	}

	// Get volume name and share path name
	rc = cifs_getVolumeName(sharep, volName, pathName, &pathNameLen);
	if (rc != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
					"Failed to get volume name"));
		error = TRUE;
		goto cifs_getShareHandle_done;
	}
	else
		CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
					"Volume name = %s, share path name = %s",
					volName, pathName));

	//---------------------------------------
	// Send message to EVM to query volume ID
	//---------------------------------------
	rc = cifs_volIdReq(volName, sharep);
	if (rc != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
					"Failed to send message to EVM"));
		error = TRUE;
		goto cifs_getShareHandle_done;
	}

	//-------------------------------
	// Wait for the response from EVM
	//-------------------------------
	thread_yield(&evt, CIFS_INTMSG_TIMEOUT_EVM);

	cifs_treatKillEvent(&evt);

	freePkt = TRUE;
	if (evt.event == EVM_MSG_STAT)
	{
		evm_MsgStat_t	*volIdInfop;
		evm_volId_t		*volIdp;

		volIdInfop = (evm_MsgStat_t *)(evt.p2.u.p);
		if (volIdInfop->stat != NFX_OK)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
						"Failed to get volume ID from EVM: %s",
						(char*)(volIdInfop + 1)));
			error = TRUE;
		}
		else
		{
			volIdp = &volIdInfop->payload.lVolId;

			CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
						"Got volume ID (0x%llx)!", volIdp->whole_thing));

			//---------------------------------------------------------
			// Send message to FS to query share handle given volume ID
			//---------------------------------------------------------
			rc = cifs_shareHandleReq(sharep, pathName, pathNameLen, volIdp);
			if (rc != NFX_OK)
			{
				CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
							"Failed to send message to FS"));
				error = TRUE;
			}
		}
	}
	else 
	{
		goto cifs_getShareHandle_commevent;
	}

	if (error == TRUE)
		goto cifs_getShareHandle_done;

	//------------------------------
	// Wait for the response from FS 
	//------------------------------
	thread_yield(&evt, CIFS_INTMSG_TIMEOUT_FS);

	cifs_treatKillEvent(&evt);

	if (evt.event == FS_REQ_GET_ROOT)
	{
		fs_MsgGetRootResp *rootHandleInfop;

		rootHandleInfop = (fs_MsgGetRootResp *)(evt.p2.u.p);
		if (rootHandleInfop->hdr.fsStatus != FS_ERR_NONE)
		{
			CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
						"fsStatus = %d\n", rootHandleInfop->hdr.fsStatus));
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
						"Failed to get share handle from FS"));
			error = TRUE;
		}
		else
		{
			// Got handle for the share!
			sharep->handle = rootHandleInfop->dirHandle;
			CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
						"Got share handle!"));
			cifs_decodeFileHandle(&sharep->handle);
			goto cifs_getShareHandle_done;
		}
	}
	else 
	{
		goto cifs_getShareHandle_commevent;
	}

	if (error == TRUE)
		goto cifs_getShareHandle_done;


cifs_getShareHandle_commevent:

	switch(evt.event)
	{
		case THREAD_EVENT_TIMEOUT:
			freePkt = FALSE;
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: Timeout"));
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
						"unexpected event %u", evt.event));
			break;
	}

cifs_getShareHandle_done:

	// Free received message
	if (freePkt == TRUE)
		eee_freePacket((eee_desc_t*)evt.p1.u.p);

	if (error == TRUE && sharep)
	{
		// Remove the share from server
		cifs_deleteShare(serverp, sharep);
	}

	//Init the next share not initialized yet
	cifs_initNextShare();
	
} /* cifs_getShareHandle() */

#else	//FS_STANDALONE

boolean
fs_shell_get_volid(char *id_str, evm_volId_t *volId);
		
void
cifs_getShareHandle(esm_handle_t handle, void *infop)
{
	int32			rc = NFX_OK;
	cifs_server		*serverp = (cifs_server *)infop;
	cifs_share		*sharep;
	evm_volId_t		volId, *volIdp;
	uchar8			volName[] = "0";
	uchar8			pathName[FS_PATH_MAX];		// UTF-8 string
	uint32			pathNameLen;				// UTF-8 pathname len in bytes
	esm_event_t		evt;
	boolean			error = FALSE, freePkt = FALSE;

	// Search for the share
	sharep = cifs_findServerShareByThread(serverp, handle);
	if (sharep == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
					"No share with thread ID %d", handle));
		error = TRUE;
		goto cifs_getShareHandle_done;
	}

	// Get volume name and share path name
	strcpy(pathName, "SHARE");
	pathNameLen = strlen(pathName);
	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
					"share path name = %s",	pathName));

	//---------------------------------------
	// get volume ID
	//---------------------------------------
	
	//HACK TO GET THE NUMBER OF THE VOLUME: 0  1  2  3  4  
	//										/  v  o  l  1
	volName[0] = sharep->cfg.pathName[4];
	if (!fs_shell_get_volid(volName, &volId))
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
					"Failed to get VolId from [%s]", volName));
		error = TRUE;
		goto cifs_getShareHandle_done;
	}

	volIdp = &volId;

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
				"Got volume ID (0x%llx)!", volIdp->whole_thing));

	//---------------------------------------------------------
	// Send message to FS to query share handle given volume ID
	//---------------------------------------------------------
	rc = cifs_shareHandleReq(sharep, pathName, pathNameLen, volIdp);
	if (rc != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
						"Failed to send message to FS"));
		error = TRUE;
	}
	if (error == TRUE)
		goto cifs_getShareHandle_done;

	//------------------------------
	// Wait for the response from FS 
	//------------------------------
	thread_yield(&evt, CIFS_INTMSG_TIMEOUT_FS);
	
	cifs_treatKillEvent(&evt);

	if (evt.event == FS_REQ_GET_ROOT)
	{
		fs_MsgGetRootResp *rootHandleInfop;

		rootHandleInfop = (fs_MsgGetRootResp *)(evt.p2.u.p);
		if (rootHandleInfop->hdr.fsStatus != FS_ERR_NONE)
		{
			CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
						"fsStatus = %d\n", rootHandleInfop->hdr.fsStatus));
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
						"Failed to get share handle from FS"));
			error = TRUE;
		}
		else
		{
			// Got handle for the share!
			sharep->handle = rootHandleInfop->dirHandle;
			CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_getShareHandle: "
						"Got share handle!"));
			cifs_decodeFileHandle(&sharep->handle);
			goto cifs_getShareHandle_done;
		}
	}
	else 
	{
		goto cifs_getShareHandle_commevent;
	}

	if (error == TRUE)
		goto cifs_getShareHandle_done;


cifs_getShareHandle_commevent:

	switch(evt.event)
	{
		case THREAD_EVENT_TIMEOUT:
			freePkt = FALSE;
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: Timeout"));
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_getShareHandle: "
						"unexpected event %u", evt.event));
			break;
	}

cifs_getShareHandle_done:

	// Free received message
	if (freePkt == TRUE)
		eee_freePacket((eee_desc_t*)evt.p1.u.p);

	if (error == TRUE && sharep)
	{
		// Remove the share from server
		cifs_deleteShare(serverp, sharep);
	}

	//Init the next share not initialized yet
	cifs_initNextShare();
	
}

#endif	//FS_STANDALONE

/*-----------------------------------------------------------------
 * Name			: cifs_initShare
 *
 * Description	: Initialize a CIFS share
 * 
 * Created by	: Laurent Brard
 *
 * Date Created	: 13/03/2002
 *-----------------------------------------------------------------
 */
int32 	//NFX_OK or NFX_ERR
cifs_initShare(cifs_server *svp, cifs_share *shp)
{	
	int32	rc;
	uchar8	*volName;

	// Create a thread to get share handle	
	shp->threadId = thread_create(svp);
	if (shp->threadId == 0)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_initShare: "
					"Failed to create a thread"));
		return (NFX_ERR);
	}
    
	if (thread_start(&(shp->threadId), cifs_getShareHandle) != NFX_OK)
    {
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_initShare: "
					"Failed to start a thread"));
		return (NFX_ERR);
    }

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifs_initNextShare
 *
 * Description	: Initialize the next CIFS share
 * 
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/14/2002
 *-----------------------------------------------------------------
 */
int32 	//NFX_OK or NFX_ERR
cifs_initNextShare(void)
{	
	cifs_server *svp;
	cifs_share  *shp;
	
	DLL_ITERATE(svp, &cifs_globalCfgp->servers, entry)
	{
		if (!(svp->flags & CIFS_SERVER_SHARESINITIALIZED))
		{
			DLL_ITERATE(shp, &svp->shares, entry)
			{
				if ((shp->cfg.flags & CIFS_SHCFG_ISADMIN) == 0 //Not admin share
	 			 && shp->threadId == 0)						   //Not initialized
					return cifs_initShare(svp, shp);
			}
			svp->flags |= CIFS_SERVER_SHARESINITIALIZED;
			cifs_testServerInitCompletion(svp);
		}
	}
		
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifs_initServers
 *
 * Description	: Initialize CIFS servers configured
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *-----------------------------------------------------------------
 */
int32 	//NFX_OK or NFX_ERR
cifs_initServers(void)
{
	cifs_server *svp;
	cifs_share  *shp;
	uint32	count = 0;
	
	cifs_initTransportLayer();

	DLL_ITERATE(svp, &cifs_globalCfgp->servers, entry)
	{
		cifs_initServerTransportLayer(svp);

		if (count < CIFS_N_SHAREINITS)
		{
			//Initialize the shares of the server
			DLL_ITERATE(shp, &svp->shares, entry)
			{
				if ((shp->cfg.flags & CIFS_SHCFG_ISADMIN) == 0)
				{
					cifs_initShare(svp, shp);
					if (++count >= CIFS_N_SHAREINITS)
						//We reached the max nb of simultaneous share inits
						break;
				}
			}
		}
	}
	
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name			: cifs_initSrcDstPorts 
 *
 * Description	: CIFS initialize source and destination ports to
 * 				  other modules.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/08/2002
 *-----------------------------------------------------------------
 */
void
cifs_initSrcDstPorts()
{
	int32	eee_cifsauthAppId, eee_evmAppId, eee_efsAppId, eee_fssfAppId;

#ifndef FS_STANDALONE	
	// Management and control source ports 
	// used for message exchange within the system.
	cifs_globalCfgp->mgmtSrcPort = PACK_UNICAST_PORT_ADDRESS(
												nfxMySlotId,
												nfxMyCPUId,
												cifs_globalCfgp->eee_cifsAppId,
												EEE_MGMT,
												NFX_COS_HIGH,
												PORT_BEST_FIT);

	cifs_globalCfgp->ctrlSrcPort = PACK_UNICAST_PORT_ADDRESS(
												nfxMySlotId,
												nfxMyCPUId,
												cifs_globalCfgp->eee_cifsAppId,
												EEE_CONTROL,
												NFX_COS_HIGH,
												PORT_BEST_FIT);

	// Get cifsauth EEE application ID
	eee_cifsauthAppId = eee_getApplicationId("cifsauth");
	// TODO: Handle error
	NFX_ASSERT(eee_cifsauthAppId != EEE_ERR_INVALID_APP_NAME);

	// Pack destination port to cifsauth
	cifs_globalCfgp->cifsauthDstPort = PACK_UNICAST_PORT_ADDRESS(
												nfxSscSlotId,
												NFX_SSC_UNIX,
												eee_cifsauthAppId,
												EEE_MGMT,
												NFX_COS_HIGH,
												PORT_BEST_FIT);

	// Get volume manager application ID
	eee_evmAppId = eee_getApplicationId("evm"); 
	// TODO:  Handle error
	NFX_ASSERT(eee_evmAppId != EEE_ERR_INVALID_APP_NAME);

	// Pack destination port to EVM
	cifs_globalCfgp->evmDstPort = PACK_UNICAST_PORT_ADDRESS(
												nfxSscSlotId,
												NFX_SSC_UNIX,
												eee_evmAppId,
												EEE_MGMT,
												NFX_COS_HIGH,
												PORT_BEST_FIT);

	// Get file system application ID
	eee_efsAppId = eee_getApplicationId("efs");
	// TODO:  Handle error
	NFX_ASSERT(eee_efsAppId != EEE_ERR_INVALID_APP_NAME);

	// Pack destination port to FS
	cifs_globalCfgp->efsDstPort = PACK_UNICAST_PORT_ADDRESS(
												nfxMySlotId,
												NFX_NFP_FP2,	//TODO
												eee_efsAppId,
												EEE_CONTROL,
												//TODO change to EEE_MGMT
												NFX_COS_HIGH,
												PORT_BEST_FIT);

	// Get filesystem stateful application ID
	eee_fssfAppId = eee_getApplicationId("fssf"); 
	// TODO:  Handle error
	NFX_ASSERT(eee_fssfAppId != EEE_ERR_INVALID_APP_NAME);

	// Pack destination port to FSSF
	cifs_globalCfgp->fssfDstPort = PACK_UNICAST_PORT_ADDRESS(
												nfxMySlotId,
												NFX_NFP_FP2,
												eee_fssfAppId,
												EEE_CONTROL,
												//TODO change to EEE_MGMT
												NFX_COS_HIGH,
												PORT_BEST_FIT);
	
#endif	//FS_STANDALONE	

	return;
} /* cifs_initSrcDstPorts() */

/*-----------------------------------------------------------------
 * Name			: cifs_init 
 *
 * Description	: CIFS initialize function.
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 10/31/2001
 *-----------------------------------------------------------------
 */
int32 	//NFX_OK or NFX_ERR
cifs_init(void)
{
	int32	rc;

#ifndef FS_STANDALONE	
	if (!( (nfxMySlotId == 1) && (nfxMyCPUId == 0) ))
	{
		/* Skip if not TX/RX CPU
		 * TODO:  Use Global define */
		return (NFX_OK);
	}
#endif	//FS_STANDALONE

printf("cifs_init: start\n");

	CIFS_RAM_ALLOC_ZERO(cifs_globalCfgp, cifs_globalCfg, sizeof(cifs_globalCfg));
	//TODO:  Handle error
	NFX_ASSERT(cifs_globalCfgp != NULL);

	memset(cifs_globalCfgp, 0, sizeof(cifs_globalCfg));	

#ifndef FS_STANDALONE	
	// Register with EEE and app Id
	cifs_globalCfgp->eee_cifsAppId = eee_allocateApplicationId("cifs_srvr");
	//TODO:  Handle error
	NFX_ASSERT(cifs_globalCfgp->eee_cifsAppId != EEE_ERR_INVALID_APP_NAME)
	
	// Save appId 
	rc = eee_registerReceiveFunc(cifs_globalCfgp->eee_cifsAppId, cifs_msgRcv, NULL);
	// TODO:  Handle error	
	NFX_ASSERT(rc == NFX_OK)
			
#endif	//FS_STANDALONE
			
	// Initialize source and destination ports
	cifs_initSrcDstPorts();

	// Initialize CIFS objects lists
	cifs_initObjects();
	
	// Initialize characters conversion mapping
    cifs_initDosUnicodeMapping();

	// Initialize CIFS servers list
	DLL_INIT(&cifs_globalCfgp->servers);
	
	// Load CIFS configuration
	cifs_loadCfg();
	
	// Initialize CIFS servers configured
//	cifs_initServers();	//do initServers in "cifs d i"

printf("cifs_init: end\n");

	return (NFX_OK);
}





