/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: sec-cifstruct.h,v 1.3 2002/12/10 18:22:20 laurentb Exp $
 *
 * Name			: sec-cifstruct.h
 *
 * Description	: Contains CIFS security definitions.
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/21/2002
 *
 *-----------------------------------------------------------------
 */
#ifndef _SEC_CIFSTRUCT_H
#define _SEC_CIFSTRUCT_H



//Access Mask
//
//       3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
//       1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
//      +---------------+---------------+-------------------------------+
//      |G|G|G|G|Res'd|A| StandardRights|         SpecificRights        |
//      |R|W|E|A|     |S|               |                               |
//      +-+-------------+---------------+-------------------------------+
//

#define SEC_SPECR_ACCESS_MASK					(0x0000FFFFL)

//Standard access types
#define SEC_STDR_DELETE_ACCESS			(0x00010000L)
#define SEC_STDR_READ_CONTROL_ACCESS	(0x00020000L)
#define SEC_STDR_WRITE_DAC_ACCESS		(0x00040000L)
#define SEC_STDR_WRITE_OWNER_ACCESS     (0x00080000L)
#define SEC_STDR_SYNCH_ACCESS			(0x00100000L)

//Standard combinations of standard access rights
#define SEC_STDR_REQUIRED_ACCESS		(0x000F0000L)
#define SEC_STDR_ALL_ACCESS			    (0x001F0000L)
//#define SEC_STDR_MASK				    (0x00FF0000L)TODO remove if not needed
#define SEC_STDR_EXECUTE_ACCESS		    (SEC_STDR_READ_CONTROL_ACCESS)
#define SEC_STDR_READ_ACCESS			(SEC_STDR_READ_CONTROL_ACCESS)
#define SEC_STDR_WRITE_ACCESS			(SEC_STDR_READ_CONTROL_ACCESS)


//AccessSystemAcl access type
#define SEC_SYSTEM_SECURITY_ACCESS		(0x01000000L)

// MaximumAllowed access type
#define SEC_MAXIMUM_ALLOWED_ACCESS		(0x02000000L)


//Generic rights.
#define SEC_GENR_READ_ACCESS			(0x80000000L)
#define SEC_GENR_WRITE_ACCESS			(0x40000000L)
#define SEC_GENR_EXECUTE_ACCESS		    (0x20000000L)
#define SEC_GENR_ALL_ACCESS			    (0x10000000L)


//File Specific access rights
#define SEC_FILE_READ_DATA				0x0001
#define SEC_FILE_WRITE_DATA			    0x0002
#define SEC_FILE_APPEND_DATA			0x0004
#define SEC_FILE_READ_EA				0x0008
#define SEC_FILE_WRITE_EA				0x0010
#define SEC_FILE_EXECUTE				0x0020
#define SEC_FILE_DELETE_CHILD			0x0040
#define SEC_FILE_READ_ATTRIBUTES		0x0080
#define SEC_FILE_WRITE_ATTRIBUTES		0x0100

#define SEC_FILE_ALL_ACCESS			    0x01FF

//File generic to specific rights mapping
#define SEC_GENR_FILE_READ		(SEC_STDR_READ_ACCESS|SEC_STDR_SYNCH_ACCESS\
								|SEC_FILE_READ_DATA|SEC_FILE_READ_EA\
								|SEC_FILE_READ_ATTRIBUTES)
#define SEC_GENR_FILE_WRITE	    (SEC_STDR_WRITE_ACCESS|SEC_STDR_SYNCH_ACCESS\
								|SEC_FILE_WRITE_DATA|SEC_FILE_APPEND_DATA\
								|SEC_FILE_WRITE_EA|SEC_FILE_WRITE_ATTRIBUTES)
#define SEC_GENR_FILE_EXECUTE	(SEC_STDR_EXECUTE_ACCESS\
								|SEC_STDR_SYNCH_ACCESS| SEC_FILE_EXECUTE\
								|SEC_FILE_READ_ATTRIBUTES)
#define SEC_GENR_FILE_ALL		(SEC_STDR_ALL_ACCESS|SEC_FILE_ALL_ACCESS)


typedef uint32 sec_accessMask;

//Generic mapping array. This is used to denote the mapping of each 
//generic access right to a specific access mask.
typedef struct sec_accessGenericMap {
  sec_accessMask	genericRead;
  sec_accessMask	genericWrite;
  sec_accessMask	genericExecute;
  sec_accessMask	genericAll;
} sec_accessGenericMap;

extern sec_accessGenericMap cifs_fileAccessGenericMap;


 	
//Security Identifier
//
//         1   1   1   1   1   1
//         5   4   3   2   1   0   9   8   7   6   5   4   3   2   1   0
//      +---------------------------------------------------------------+
//      |      SubAuthorityCount        |Reserved1 (SBZ)|   Revision    |
//      +---------------------------------------------------------------+
//      |                   IdentifierAuthority[0]                      |
//      +---------------------------------------------------------------+
//      |                   IdentifierAuthority[1]                      |
//      +---------------------------------------------------------------+
//      |                   IdentifierAuthority[2]                      |
//      +---------------------------------------------------------------+
//      |                                                               |
//      +- -  -  -  -  -  -  -  SubAuthority[]  -  -  -  -  -  -  -  - -+
//      |                                                               |
//      +---------------------------------------------------------------+
//
//
#define CIFSEC_SID_REVISION				1

#define CIFSEC_SID_MAX_SUB_AUTHORITIES	15
#define CIFS_N_ANYSIZE_ARRAY			1


typedef struct {
  uchar8	val[6];						// Security Identifier Authority 
}sec_cifsSidAuthy;

typedef struct {						// Security Identifier structure 
  uchar8			revision;
  uchar8			nbSubAuthy;
  sec_cifsSidAuthy 	authority;
  uint32			subAuthy[CIFSEC_SID_MAX_SUB_AUTHORITIES];
}sec_cifsSid;



//Access/Audit Control Entry (ACE)
//
//       3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
//       1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
//      +---------------+-------+-------+---------------+---------------+
//      |    AceFlags   | Resd  |Inherit|    AceSize    |     AceType   |
//      +---------------+-------+-------+---------------+---------------+
//      |                              Mask                             |
//      +---------------------------------------------------------------+
//      |                                                               |
//      +                                                               +
//      |                                                               |
//      +                              Sid                              +
//      |                                                               |
//      +                                                               +
//      |                                                               |
//      +---------------------------------------------------------------+
//
//  Mask is the access mask associated with the ACE.  This is either the
//  access allowed, access denied, audit, or alarm mask.
//
//  Sid is the Sid associated with the ACE.

//ACE type
#define CIFSEC_ACE_TYPE_ACCESS_ALLOWED	0x00
#define CIFSEC_ACE_TYPE_ACCESS_DENIED	0x01
#define CIFSEC_ACE_TYPE_SYSTEM_AUDIT	0x02
#define CIFSEC_ACE_TYPE_SYSTEM_ALARM	0x03

//ACE flags
#define CIFSEC_ACE_FL_OBJECT_INHERIT	0x01
#define CIFSEC_ACE_FL_CONTAINER_INHERIT	0x02
#define CIFSEC_ACE_FL_NO_PROPAG_INHERIT	0x04
#define CIFSEC_ACE_FL_INHERIT_ONLY		0x08
#define CIFSEC_ACE_FL_INHERITED_ACE		0x10	 	
#define CIFSEC_ACE_FL_SUCCESSFUL_ACCESS	0x40
#define CIFSEC_ACE_FL_FAILED_ACCESS		0x80
#define CIFSEC_ACE_FL_VALID_INHERIT		0x1f

typedef struct sec_cifsAce
{
  uchar8			type;				//Allowed/denied/Audit/Alarm
  uchar8			flags;				//Inheritance flags and access result
  ushort16			size;				//size in bytes of the ACE (RPC E/D)
  sec_accessMask	accessMask;			//access bits defined in cifs.h
  sec_cifsSid		sid;
}sec_cifsAce;


#define NT4_ACL_REVISION				2		//Windows NT 4 ACL revision
#define W2K_ACL_REVISION				3		//W2K ACL revision
#define CIFSEC_ACL_REVISION				W2K_ACL_REVISION	
//TODO see if we want to go for W2K ACL right away

//Audit/Access Control List
typedef struct sec_cifsAcl
{
  uchar8		revision;
  uchar8		sbz1;
  ushort16		size;					//size in bytes of the ACL (RPC (E/D)
  uint32		nbAces;					//number of ACEs
  sec_cifsAce	*acep;					//ACEs list 
}sec_cifsAcl;




//Security descriptor
//
//       3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
//       1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
//      +---------------------------------------------------------------+
//      |            Control            |Reserved1 (SBZ)|   Revision    |
//      +---------------------------------------------------------------+
//      |                            Owner                              |
//      +---------------------------------------------------------------+
//      |                            Group                              |
//      +---------------------------------------------------------------+
//      |                            Sacl                               |
//      +---------------------------------------------------------------+
//      |                            Dacl                               |
//      +---------------------------------------------------------------+
//
// In general, this data structure should be treated opaquely to ensure future
// compatibility.
//

#define CIFSEC_SD_REVISION				1

#define CIFSEC_S_SD_HDR	(sizeof(uchar8)*2 + sizeof(ushort16) + sizeof(uint32)*4)

//Security Descriptor control bits
#define CIFSEC_SD_OWNER_DEFAULTED		0x0001
#define CIFSEC_SD_GROUP_DEFAULTED		0x0002
#define CIFSEC_SD_DACL_PRESENT			0x0004
#define CIFSEC_SD_DACL_DEFAULTED		0x0008
#define CIFSEC_SD_SACL_PRESENT			0x0010
#define CIFSEC_SD_SACL_DEFAULTED		0x0020
#define CIFSEC_SD_SELF_RELATIVE			0x8000
//since W2k
#define CIFSEC_SD_DACL_AUTO_INHERIT_REQ	0x0100
#define CIFSEC_SD_SACL_AUTO_INHERIT_REQ	0x0200
#define CIFSEC_SD_DACL_AUTO_INHERITED	0x0400
#define CIFSEC_SD_SACL_AUTO_INHERITED	0x0800
#define CIFSEC_SD_DACL_PROTECTED		0x1000
#define CIFSEC_SD_SACL_PROTECTED		0x2000

//  Where:
//
//  CIFSEC_OWNER_DEFAULTED - This boolean flag, when set, indicates that the
//          SID pointed to by the Owner field was provided by a
//          defaulting mechanism rather than explicitly provided by the
//          original provider of the security descriptor.  This may
//          affect the treatment of the SID with respect to inheritence
//          of an owner.
//
//  CIFSEC_GROUP_DEFAULTED - This boolean flag, when set, indicates that the
//          SID in the Group field was provided by a defaulting mechanism
//          rather than explicitly provided by the original provider of
//          the security descriptor.  This may affect the treatment of
//          the SID with respect to inheritence of a primary group.
//
//  CIFSEC_DACL_PRESENT - This boolean flag, when set, indicates that the
//          security descriptor contains a discretionary ACL.  If this
//          flag is set and the Dacl field of the SECURITY_DESCRIPTOR is
//          null, then a null ACL is explicitly being specified.
//
//  CIFSEC_DACL_DEFAULTED - This boolean flag, when set, indicates that the
//          ACL pointed to by the Dacl field was provided by a defaulting
//          mechanism rather than explicitly provided by the original
//          provider of the security descriptor.  This may affect the
//          treatment of the ACL with respect to inheritence of an ACL.
//          This flag is ignored if the DaclPresent flag is not set.
//
//  CIFSEC_SACL_PRESENT - This boolean flag, when set,  indicates that the
//          security descriptor contains a system ACL pointed to by the
//          Sacl field.  If this flag is set and the Sacl field of the
//          SECURITY_DESCRIPTOR is null, then an empty (but present)
//          ACL is being specified.
//
//  CIFSEC_SACL_DEFAULTED - This boolean flag, when set, indicates that the
//          ACL pointed to by the Sacl field was provided by a defaulting
//          mechanism rather than explicitly provided by the original
//          provider of the security descriptor.  This may affect the
//          treatment of the ACL with respect to inheritence of an ACL.
//          This flag is ignored if the SaclPresent flag is not set.
//
//  CIFSEC_SELF_RELATIVE - This boolean flag, when set, indicates that the
//          security descriptor is in self-relative form.  In this form,
//          all fields of the security descriptor are contiguous in memory
//          and all pointer fields are expressed as offsets from the
//          beginning of the security descriptor.  This form is useful
//          for treating security descriptors as opaque data structures
//          for transmission in communication protocol or for storage on
//          secondary media.
//
//
//
//Security information
#define CIFSEC_OWNER_SEC_INFORMATION	0x00000001
#define CIFSEC_GROUP_SEC_INFORMATION	0x00000002
#define CIFSEC_DACL_SEC_INFORMATION		0x00000004
#define CIFSEC_SACL_SEC_INFORMATION		0x00000008
#define CIFSEC_ALL_SEC_INFORMATION \
		(CIFSEC_OWNER_SEC_INFORMATION|CIFSEC_GROUP_SEC_INFORMATION|\
		 CIFSEC_DACL_SEC_INFORMATION|CIFSEC_SACL_SEC_INFORMATION)
		
//(we integrate absolute and self relative security descriptor definitions to
// be able to encode/decode RPC stream and use security decriptor with the same
// structure)
typedef struct sec_cifsSd
{
  uint32		size;					//size in bytes (RPC (E/D)
  uchar8		revision;				
  uchar8		sbz1;
  ushort16		control;				//control bits
  uint32		ownerSidOff;			//owner sid offest
  uint32		groupSidOff;			//group sid offset
  uint32		sAclOff;				//system ACL offset
  uint32		dAclOff;				//discretionary ACL offset
  sec_cifsSid	ownerSid;				//owner sid
  sec_cifsSid	groupSid;				//group sid
  sec_cifsAcl	sAcl;					//system ACL
  sec_cifsAcl	dAcl;					//discretionary ACL
}sec_cifsSd;


/*
typedef struct sec_cifsSdHdr
{
  uint32		size;					//size of the whole Security Desc.
  uchar8		revision;
  uchar8		sbz1;
  ushort16		control;
  uint32		ownerSidOff;			//owner sid offest
  uint32		groupSidOff;			//group sid offset
  uint32		sAclOff;				//system ACL offset
  uint32		dAclOff;				//discretionary ACL offset
}sec_cifsSdHdr;

typedef union cifsec_selfSd
{
  sec_cifsSdHdr	hdr;
  uchar8		info[1];
}cifsec_selfSd;
*/

//Security descriptor buffer for RPC encoding / decoding
typedef struct sec_cifsSdBuf
{
  uint32	maxLen;
  uint32	sdp;
  uint32	len;
  union
  {
    sec_cifsSd sd;
    sec_cifsSd *sdp;
  }sd;
}sec_cifsSdBuf;



//Group attributes in accessToken
#define CIFSEC_GROUP_MANDATORY			(0x00000001L)
#define CIFSEC_GROUP_ENABLED_BY_DEFAULT	(0x00000002L)
#define CIFSEC_GROUP_ENABLED			(0x00000004L)
#define CIFSEC_GROUP_OWNER				(0x00000008L)
#define CIFSEC_GROUP_USE_FOR_DENY_ONLY	(0x00000010L)
#define CIFSEC_GROUP_LOGON_ID			(0xC0000000L)

//Number of default well known groups
#define CIFSEC_N_DEFAULT_GRPS	3		//World, Network, authenticated users

//Sid Name Use TODO:(see if necessary)
typedef enum sec_cifsSidNameUse
{
  SidTypeUser = 1,
  SidTypeGroup,
  SidTypeDomain,
  SidTypeAlias,
  SidTypeWellKnownGroup,
  SidTypeDeletedAccount,
  SidTypeInvalid,
  SidTypeUnknown,
  SidTypeComputer
}sec_cifsSidNameUse;


//Access Token Entry
typedef struct sec_cifsAte
{
  sec_cifsSid	sid;
  uint32		attributes;
}sec_cifsAte;

typedef struct {
  sec_cifsAte		user;
  uchar8			nbGroups;
  sec_cifsAte		groupSid[CIFSEC_N_DEFAULT_GRPS];
}sec_cifsAccessToken;


//************************** Well known SIDs and RIDs  **********************


/////////////////////////////////////////////////////////////////////////////
// Universal well-known SIDs                                               //
//                                                                         //
//     Null SID                     S-1-0-0                                //
//     World                        S-1-1-0                                //
//     Local                        S-1-2-0                                //
//     Creator Owner ID             S-1-3-0                                //
//     Creator Group ID             S-1-3-1                                //
//     Creator Owner Server ID      S-1-3-2                                //
//     Creator Group Server ID      S-1-3-3                                //
//                                                                         //
//     (Non-unique IDs)             S-1-4                                  //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#define CIFSEC_NULL_SID_AUTHORITY       {{0,0,0,0,0,0}}
#define CIFSEC_WORLD_SID_AUTHORITY      {{0,0,0,0,0,1}}
#define CIFSEC_LOCAL_SID_AUTHORITY      {{0,0,0,0,0,2}}
#define CIFSEC_CREATOR_SID_AUTHORITY    {{0,0,0,0,0,3}}
#define CIFSEC_NON_UNIQUE_AUTHORITY     {{0,0,0,0,0,4}}

#define CIFSEC_NULL_RID                 (0x00000000L)
#define CIFSEC_WORLD_RID                (0x00000000L)
#define CIFSEC_LOCAL_RID                (0X00000000L)

#define CIFSEC_CREATOR_OWNER_RID        (0x00000000L)
#define CIFSEC_CREATOR_GROUP_RID        (0x00000001L)

#define CIFSEC_CREATOR_OWNER_SERVER_RID (0x00000002L)
#define CIFSEC_CREATOR_GROUP_SERVER_RID (0x00000003L)


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// NT well-known SIDs                                                      //
//                                                                         //
//     NT Authority          S-1-5                                         //
//     Dialup                S-1-5-1                                       //
//                                                                         //
//     Network               S-1-5-2                                       //
//     Batch                 S-1-5-3                                       //
//     Interactive           S-1-5-4                                       //
//     Service               S-1-5-6                                       //
//     AnonymousLogon        S-1-5-7       (aka null logon session)        //
//     Proxy                 S-1-5-8                                       //
//     ServerLogon           S-1-5-9       (aka domain controller account) //
//     Self                  S-1-5-10      (self RID)                      //
//     Authenticated User    S-1-5-11      (Authenticated user somewhere)  //
//     Restricted Code       S-1-5-12      (Running restricted code)       //
//                                                                         //
//     (Logon IDs)           S-1-5-5-X-Y                                   //
//                                                                         //
//     (NT non-unique IDs)   S-1-5-0x15-...                                //
//                                                                         //
//     (Built-in domain)     s-1-5-0x20                                    //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////


#define CIFSEC_NT_AUTHORITY           {{0,0,0,0,0,5}}   // ntifs

#define CIFSEC_DIALUP_RID             (0x00000001L)
#define CIFSEC_NETWORK_RID            (0x00000002L)
#define CIFSEC_BATCH_RID              (0x00000003L)
#define CIFSEC_INTERACTIVE_RID        (0x00000004L)
#define CIFSEC_SERVICE_RID            (0x00000006L)
#define CIFSEC_ANONYMOUS_LOGON_RID    (0x00000007L)
#define CIFSEC_PROXY_RID              (0x00000008L)
#define CIFSEC_ENTERPRISE_CONTROLLERS_RID (0x00000009L)
#define CIFSEC_SERVER_LOGON_RID       CIFSEC_ENTERPRISE_CONTROLLERS_RID
#define CIFSEC_PRINCIPAL_SELF_RID     (0x0000000AL)
#define CIFSEC_AUTHENTICATED_USER_RID (0x0000000BL)
#define CIFSEC_RESTRICTED_CODE_RID    (0x0000000CL)

#define CIFSEC_LOGON_IDS_RID          (0x00000005L)
#define CIFSEC_LOGON_IDS_RID_COUNT    (3L)

#define CIFSEC_LOCAL_SYSTEM_RID       (0x00000012L)

#define CIFSEC_NT_NON_UNIQUE_RID      (0x00000015L)

#define CIFSEC_BUILTIN_DOMAIN_RID     (0x00000020L)


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// well-known domain relative sub-authority values (RIDs)...               //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

// Well-known users ...
#define CIFSEC_DOM_USER_RID_ADMIN          (0x000001F4L)
#define CIFSEC_DOM_USER_RID_GUEST          (0x000001F5L)
#define CIFSEC_DOM_USER_RID_KRBTGT         (0x000001F6L)

// well-known groups ...
#define CIFSEC_DOM_GROUP_RID_ADMINS        (0x00000200L)
#define CIFSEC_DOM_GROUP_RID_USERS         (0x00000201L)
#define CIFSEC_DOM_GROUP_RID_GUESTS        (0x00000202L)
#define CIFSEC_DOM_GROUP_RID_COMPUTERS     (0x00000203L)
#define CIFSEC_DOM_GROUP_RID_CONTROLLERS   (0x00000204L)
#define CIFSEC_DOM_GROUP_RID_CERT_ADMINS   (0x00000205L)
#define CIFSEC_DOM_GROUP_RID_SCHEMA_ADMINS (0x00000206L)

// well-known aliases ...
#define CIFSEC_DOM_ALIAS_RID_ADMINS        (0x00000220L)
#define CIFSEC_DOM_ALIAS_RID_USERS         (0x00000221L)
#define CIFSEC_DOM_ALIAS_RID_GUESTS        (0x00000222L)
#define CIFSEC_DOM_ALIAS_RID_POWER_USERS   (0x00000223L)

#define CIFSEC_DOM_ALIAS_RID_ACCOUNT_OPS   (0x00000224L)
#define CIFSEC_DOM_ALIAS_RID_SYSTEM_OPS    (0x00000225L)
#define CIFSEC_DOM_ALIAS_RID_PRINT_OPS     (0x00000226L)
#define CIFSEC_DOM_ALIAS_RID_BACKUP_OPS    (0x00000227L)

#define CIFSEC_DOM_ALIAS_RID_REPLICATOR    (0x00000228L)


// Allocate the System Luid.  The first 1000 LUIDs are reserved.
// Use #999 here (0x3E7 = 999)
// Note LUID: Local User Id

#define CIFSEC_SYSTEM_LUID					{ 0x3E7, 0x0 }
#define CIFSEC_ANONYMOUS_LOGON_LUID			{ 0x3e6, 0x0 }


//Well-known SIDs built from above definitions
extern sec_cifsSid cifsec_nullSid;
extern sec_cifsSid cifsec_worldSid;
extern sec_cifsSid cifsec_creatorOwnerSid;
extern sec_cifsSid cifsec_creatorGroupSid;
extern sec_cifsSid cifsec_creatorOwnerServerSid;
extern sec_cifsSid cifsec_creatorGroupServerSid;
extern sec_cifsSid cifsec_networkSid;
extern sec_cifsSid cifsec_anonymousSid;
extern sec_cifsSid cifsec_AuthenticatedUserSid;
extern sec_cifsSid cifsec_builtInDomainSid;
extern sec_cifsSid cifsec_builtInAdminSid;
extern sec_cifsSid cifsec_builtInGuestSid;

			
#endif // #ifndef _SEC_CIFSTRUCT_H 

