/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 * RCS          :  $Id: datamover-func.c,v 1.23 2002/10/03 22:31:02 pankajg Exp $
 *
 * Name         : datamover-func.c
 *
 * Description  : Action functions for the data mover.
 *                For each connection, there is a dedicated instance
 *                of the data mover.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 * Re-designed  : 30-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

#include "datamover.h"

/*------------------------------------------------------------------------------
 * Global (action) functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : dm_create
 *
 * Description  : Create a new data mover instance.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_create(esm_handle_t  handle,
          int32         nextState,
          esm_event_t   thisEvent,
          void *        cb)

{
    DM_FCT_PRINT(("> dm_create()"));
    
    /* Do nothing.
     */
    
    return(nextState);  // goto IDLE
}

/*------------------------------------------------------------------------------
 * Name         : dm_accept
 *
 * Description  : A client is connected to server.
 *                Ask user application if server accepts or rejects.
 *                (server only)
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_accept(esm_handle_t handle,
          int32        nextState,
          esm_event_t  thisEvent,
          void *       cb)
    
{
    dm_cb_t *           myCb = (dm_cb_t *) cb;
    dm_acceptReq_t      req;
    int32               rc = NFX_OK;
    
    DM_FCT_PRINT(("> Server dm_accept()"));

    /* Ask user application if server should accept or reject
     * this connection.
     */
    req.hdr.msgType = DM_ACCEPT_REQ;
    req.hdr.appHdl = myCb->appHdl;
    req.hdr.dmHdl = myCb->dmHdl;
    req.rmtAddr = myCb->rmtAddr;
    req.rmtPort = myCb->rmtPort;

    rc = dm_sendMessage((void *) &req,
                        sizeof(dm_acceptReq_t),
                        myCb->appAddr,
                        dmAddr);
    NFX_ASSERT(rc == NFX_OK);
    
    return(nextState); // goto ACCEPTING
}

/*------------------------------------------------------------------------------
 * Name         : dm_go
 *
 * Description  : User application accepts client connection.
 *                (server only)
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_go(esm_handle_t handle,
      int32        nextState,
      esm_event_t  thisEvent,
      void *       cb)
    
{
    dm_cb_t *   myCb = (dm_cb_t *) cb;
    int32       rc = NFX_OK;
    
    DM_DBG_PRINT(("> dm_go()"));

    /*
     *  Store the application handle
     */
    myCb->appHdl = thisEvent.p1.u.dw;

    if (myCb->protocol == TPL_PROT_UDP)
    {
        /* Send an empty packet to server to get it connected.
         */
        pkt_desc_t *    pktDesc = NULL;

        pktDesc = TPL_ALLOC_PACKET(myCb->tplConnId, 0);
        NFX_ASSERT(pktDesc != NULL);
        
    	pktDesc->hdrSize = 10;
    	pktDesc->dataSize = 0; // should be <= 8k
        pktDesc->cod = 0;
        pktDesc->numOfCod = 0;
        pktDesc->offset = 0;
        
        
        /* Send the packet.
         */
        rc = tpl_xmtPkt(myCb->tplConnId, pktDesc);
        if (rc != NFX_OK)
        {
            DM_ERR_PRINT(("tpl_xmtPkt() failed : %d", rc));
			PKT_FREE_PACKET(pktDesc);
        }
    }
    
    return(nextState); // goto READY
}

#if defined(NFX_SMP)
extern int32
dm_bindReqAck(void* pAppHdl, uint32 rc, void *in_pBind);
#endif /* NFX_SMP */
/*------------------------------------------------------------------------------
 * Name         : dm_openCnx
 *
 * Description  : Open the data connection to the server.
 *                (client only)
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_openCnx(esm_handle_t handle,
           int32        nextState,
           esm_event_t  thisEvent,
           void *       cb)
    
{
    int32           newState = nextState;
    dm_cb_t *       myCb = (dm_cb_t *) cb;
    dm_openReq_t *  req = (dm_openReq_t *) thisEvent.p1.u.p;
	boolean         bufIsInSharedMemory = (boolean) thisEvent.p2.u.w;
    tpl_bindCb_t    bindReq;
    tpl_bindCb_t *  bindRsp = NULL;
    int32           rc = NFX_OK;
    
    DM_FCT_PRINT(("> dm_openCnx() RmtPort %d LclPort %d", ntohs(myCb->rmtPort),
						ntohs(myCb->lclPort)));
    
    /* Setup the data connection. Connect to the server.
     */
    bzero(&bindReq, sizeof(tpl_bindCb_t));
    bindReq.tplProt = req->protocol;
    bindReq.laddr = req->lclAddr;
    bindReq.raddr = req->rmtAddr;
    bindReq.lport = TPL_ANY_PORT; // htons(myCb->lclPort); //TPL_ANY_PORT !!!
    bindReq.rport = req->rmtPort;
    //bindReq.bingFlags = xxx; - TPL internal use only
    //bindReq.pBsdSock = xxx; - TPL internal use only
    bindReq.ul_addConnInd = dm_addConnInd;
    bindReq.ul_delConnInd = dm_delConnInd;
    bindReq.ul_rcvPktInd = dm_rcvPktInd;
    bindReq.ul_xmtCmpInd = dm_xmtCmpInd;
    bindReq.usrInfo = myCb;
    {
        struct in_addr  addr;
       
	   	memset((char *)&addr, 0, sizeof(struct in_addr));
        addr.s_addr = bindReq.laddr;
        addr.s_addr = bindReq.raddr;
    }
    
#if defined(NFX_SMP)
	bindReq.ul_bindReqAck = dm_bindReqAck;
	rc = tpl_bindReq(&bindReq, myCb, (TPL_BIND_CLIENT | TPL_BIND_USE_RPC));
#endif /* NFX_SMP */
	rc = tpl_bindReq(&bindReq, &bindRsp, (TPL_BIND_CLIENT | TPL_BIND_USE_RPC));


    if (rc != NFX_OK)
    {
        DM_ERR_PRINT(("tpl_bindReq() failed (%d)", rc));
        /* Send SETUP_CNX_ERR event to this instance.
         * do we need to send an event, or can I just process the error here?
         * it depends if we could receive a addConnInd with nok
         
         rc = esm_deliverEvent(myCb->dmHdl,
                               ESM_EVENT_QUEUED,
                               ESM_DM_SETUP_CNX_ERR,
                               NULL,
                               NULL);
         newState = DM_STATE_IDLE;
        */
    }
    else
    {
        myCb->tplBindCb = (void *) bindRsp;
#ifdef ISABELLE // code to delete?
        // TPL doesn't return IP @ and port here, but in addConnInd
        // well.. may be Shaun will change that... so let's wait...
        myCb->lclAddr = bindRsp->laddr;
        myCb->lclPort = bindRsp->lport;
        {
            struct in_addr  addr;
           
            addr.s_addr = bindRsp->laddr;
        }
#endif // ISABELLE

#ifdef LINUX_TEST
        {
            int idx;
            
            idx = luc_getFreeLucUserIndex();
            lucUserTab[idx].inUse = TRUE;
            lucUserTab[idx].port = myCb->lclPort;
            lucUserTab[idx].sock = dm_openSock(myCb->lclAddr, myCb->lclPort);
            myCb->lucUserIndex = idx;
            }
#endif // LINUX_TEST

        /* Wait for addConnInd from TPL...
         */
    }
    
	if (bufIsInSharedMemory)
		eee_deallocateBuffer(req);
	else
		eee_ramDealloc(req);

    return(newState);   // goto OPENING or goto IDLE
}

/*------------------------------------------------------------------------------
 * Name         : dm_cnxUp
 *
 * Description  : Data connection is connected.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_cnxUp(esm_handle_t   handle,
         int32          nextState,
         esm_event_t    thisEvent,
         void *         cb)
    
{
    dm_cb_t *       myCb = (dm_cb_t *) cb;
    int32           rc = NFX_OK;
    dm_openRsp_t    rsp;

    DM_DBG_PRINT(("> dm_cnxUp()"));

    // IS THAT RIGHT TO KEEP???
    if (myCb->protocol == TPL_PROT_UDP)
    {
        /* Send an empty packet to server to get it connected.
         */
        pkt_desc_t *    pktDesc = NULL;

        pktDesc = TPL_ALLOC_PACKET(myCb->tplConnId, 0);
        NFX_ASSERT(pktDesc != NULL);
        
    	pktDesc->hdrSize = 10;
    	pktDesc->dataSize = 0; // should be <= 8k
        pktDesc->cod = 0;
        pktDesc->numOfCod = 0;
        pktDesc->offset = 0;
        
        /* Send the packet.
         */
        rc = tpl_xmtPkt(myCb->tplConnId, pktDesc);
        if (rc != NFX_OK)
        {
            DM_ERR_PRINT(("tpl_xmtPkt() failed : %d", rc));
			PKT_FREE_PACKET(pktDesc);
        }
    }
    
    /* Notify the user application that the connection is
     * successfully connected.
     */
    rsp.hdr.msgType = DM_OPEN_RSP;
    rsp.hdr.appHdl = myCb->appHdl;
    rsp.hdr.dmHdl = myCb->dmHdl;
    rsp.result = NFX_OK;
    rsp.lclAddr = myCb->lclAddr;
    rsp.lclPort = myCb->lclPort;
    rc = dm_sendMessage((void *) &rsp,
                         sizeof(dm_openRsp_t),
                         myCb->appAddr,
                         dmAddr);
    NFX_ASSERT(rc == NFX_OK);
    
    return(nextState);
}
/*------------------------------------------------------------------------------
 * Name         : dm_send
 *
 * Description  : Send a message/data to the target.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 05-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_send(esm_handle_t    handle,
        int32           nextState,
        esm_event_t     thisEvent,
        void *          cb)

{
    dm_cb_t         *myCb = (dm_cb_t *) cb;
    int32           newState = nextState;
    char            *msg = (char *)thisEvent.p1.u.p;
	boolean         bufIsInSharedMemory = (boolean) thisEvent.p2.u.w;
    dm_sendReq_t    *req = (dm_sendReq_t *)msg;
    int32           rc = NFX_OK;
    pkt_desc_t      *pktDesc = NULL;
    char            *sendMsg;
    dm_pktHdr_t     *dmHdr;
	ushort16        dmHdrSize;
    
    DM_DBG_PRINT(("> dm_send()"));

    /* Build DM header.
     * This is needed since TPL doesn't return the size of the header,
     * but only the size of header + data.  
     */
    dmHdrSize = sizeof(dm_pktHdr_t) + req->appHdrSize;
    sendMsg = (char *)eee_ramAlloc(dmHdrSize);
    NFX_ASSERT(sendMsg != NULL);
    bzero(sendMsg, dmHdrSize);
    dmHdr = (dm_pktHdr_t *)sendMsg;
    dmHdr->appHdrSize = req->appHdrSize;
    if (req->cod != 0)
    {
        dmHdr->isCOD = TRUE;
    }
    else
    {
        dmHdr->isCOD = FALSE;
    }
    /*
     * Add applications header at the end of pktHdr
     */
    memcpy((void *)(sendMsg + sizeof(dm_pktHdr_t)),
           (const void *)(msg + sizeof(dm_sendReq_t)),
           req->appHdrSize);
    
    /* Initialize the packet descriptor and the packet header.
     * (will be deallocated by TPL when datamover returns TPL_FREE_PKT)
     */
    pktDesc = TPL_ALLOC_PACKET(myCb->tplConnId, req->dataSize);
    NFX_ASSERT(pktDesc != NULL);

	memcpy(pktDesc->pHdr, sendMsg, dmHdrSize);
    pktDesc->hdrSize = dmHdrSize;

	/*
	 * Current implementation assumes transfer of 1 COD at a time
	 * pkt datasize == cod size ==  req->dataSize
	 */
    pktDesc->codEntry[0].cod = req->cod;
    pktDesc->codEntry[0].offset = req->offset;
    pktDesc->codEntry[0].length = req->dataSize; // should be <= 8k
    pktDesc->dataSize = req->dataSize; // should be <= 8k
	if (req->cod)
		pktDesc->numOfCod = 1;
	else
		pktDesc->numOfCod = 0;

    /* Send the packet.
     */
    rc = tpl_xmtPkt(myCb->tplConnId, pktDesc);
    if (rc != NFX_OK)
    {
        DM_ERR_PRINT(("tpl_xmtPkt() failed : %d", rc));
		PKT_FREE_PACKET(pktDesc);

        /* Send SND_DATA_ERR event to this instance.
         * do we need to send an event, or can I just process the error here?
         * it depends if we could receive a xmtPktInd with nok
         
         rc = esm_deliverEvent(myCb->dmHdl,
                               ESM_EVENT_QUEUED,
                               DM_EVENT_SND_DATA_ERR,
                               NULL,
                               NULL);
         newState = ESM_DM_STATE_IDLE; ???
        */
    }
    else
    {
        /* Do nothing.
         * Wait for xmtPktInd from TPL...
         */
    }

    eee_ramDealloc(sendMsg);
	if (bufIsInSharedMemory)
		eee_deallocateBuffer(req);
	else
		eee_ramDealloc(req);

    return(newState);   // goto SENDING or goto IDLE
}

/*------------------------------------------------------------------------------
 * Name         : dm_sendOk
 *
 * Description  : Data has been sent successfully to the target. Wait...
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_sendOk(esm_handle_t    handle,
          int32           nextState,
          esm_event_t     thisEvent,
          void *          cb)

{
    dm_cb_t *       myCb = (dm_cb_t *) cb;
    dm_sendRsp_t    rsp;
    int32           rc = NFX_OK;

    DM_DBG_PRINT(("> dm_sendOk()"));

    /* One more packet has been sent.
     */
    myCb->nbPktSntOk ++;
    /* Notify the user application.
     */
    rsp.hdr.msgType = DM_SEND_RSP;
    rsp.hdr.appHdl = myCb->appHdl;
    rsp.hdr.dmHdl = myCb->dmHdl;
    rsp.result = NFX_OK;
        
    rc = dm_sendMessage((void *) &rsp,
                        sizeof(dm_sendRsp_t),
                        myCb->appAddr,
                        dmAddr);
    NFX_ASSERT(rc == NFX_OK);
        
    /* Wait for more packet to send...
     */
    
    return(nextState);  // goto READY
}

/*------------------------------------------------------------------------------
 * Name         : dm_sendNok
 *
 * Description  : Data has been sent unsuccessfully to the target. Wait...
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_sendNok(esm_handle_t  handle,
           int32         nextState,
           esm_event_t   thisEvent,
           void *        cb)

{
    dm_cb_t *       myCb = (dm_cb_t *) cb;
    dm_sendRsp_t    rsp;
    int32           rc = NFX_OK;


    DM_DBG_PRINT(("> dm_sendNok()"));
	
    /* One more packet has been sent.
     */
    myCb->nbPktSntNok ++;
    
    /* Notify the user application.
     */
    rsp.hdr.msgType = DM_SEND_RSP;
    rsp.hdr.appHdl = myCb->appHdl;
    rsp.hdr.dmHdl = myCb->dmHdl;
    rsp.result = NFX_ERR;
        
    rc = dm_sendMessage((void *) &rsp,
                        sizeof(dm_sendRsp_t),
                        myCb->appAddr,
                        dmAddr);
    NFX_ASSERT(rc == NFX_OK);
        
    /* Wait for more packet to send...
     */
    
    return(nextState);  // goto READY
}

/*------------------------------------------------------------------------------
 * Name         : dm_rcv
 *
 * Description  : Data has been received. Forward to user application.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_rcv(esm_handle_t handle,
       int32        nextState,
       esm_event_t  thisEvent,
       void *       cb)

{
    dm_cb_t *       myCb = (dm_cb_t *) cb;
    pkt_desc_t *    pktDesc = (pkt_desc_t *) thisEvent.p1.u.p;
    char            *rcvMsg = (char *)pktDesc->pHdr;
    dm_pktHdr_t *   dmHdr = (dm_pktHdr_t *) rcvMsg;
	ushort16        dmHdrSize = sizeof(dm_pktHdr_t) + dmHdr->appHdrSize;
    nfx_cod_entry_t cod;
    char            *sndMsg;
    dm_rcvInd_t *   ntfy;
    int32           rc = NFX_OK;
    
    DM_DBG_PRINT(("> dm_rcv()"));

    /* One more packet has been received.
     */
    myCb->nbPktRcv ++;

    /* TPL doesn't return a COD to get the data. Get COD ourselves!
     * TPL doesn't return in 'hdrSize' the size of the header,
     * but the size of the header + data!
     */
    if (dmHdr->isCOD == TRUE)
    {
        /* TPL doesn't return a COD to get the data.
         * Get COD ourselves.
         * Note: TPL doesn't return in 'hdrSize' the size of the header,
         * but the size of the header + data!
         */
#ifdef SBM_REV2
        luc_sndGmcData( pktDesc,
                        (char *) (rcvMsg + dmHdrSize),
                        8192, 
						0, 			/* offset */
						0, 			/* padSize */
						&cod);
#else
        luc_sndGmcData(pktDesc,
                       (char *) (rcvMsg + dmHdrSize),
                       8192, &cod);
#endif
    }
    else
        bzero(&cod, sizeof(nfx_cod_entry_t));


    /* Forward received packet to AM agent.
     */
    sndMsg = (char *)eee_ramAlloc(sizeof(dm_rcvInd_t) + dmHdr->appHdrSize);
    NFX_ASSERT(sndMsg != NULL);
    bzero(sndMsg, (sizeof(dm_rcvInd_t) + dmHdr->appHdrSize));
    ntfy = (dm_rcvInd_t *)sndMsg;
    ntfy->hdr.msgType = DM_RCV_IND;
    ntfy->hdr.appHdl = myCb->appHdl;
    ntfy->hdr.dmHdl = myCb->dmHdl;
    ntfy->cod = cod.cod;
    ntfy->offset = cod.offset; // valid???
    ntfy->dataSize = cod.length;
    ntfy->appHdrSize = dmHdr->appHdrSize;
    /*
     * Copy AM header message beyond the local DM message
     */
    memcpy((void *) (sndMsg + sizeof(dm_rcvInd_t)),
           (const void *) (rcvMsg + sizeof(dm_pktHdr_t)),
           dmHdr->appHdrSize);
    
    rc = dm_sendMessage((void *) sndMsg,
                        dmHdr->appHdrSize + sizeof(dm_rcvInd_t),
                        myCb->appAddr,
                        dmAddr);
    NFX_ASSERT(rc == NFX_OK);
    eee_ramDealloc(sndMsg);

    /* Free received packet.
     */
    rc = PKT_FREE_PACKET(pktDesc);
    NFX_ASSERT(rc == NFX_OK);
    
    return(nextState);  // stay ACTIVE
}

/*------------------------------------------------------------------------------
 * Name         : dm_closeCnx
 *
 * Description  : Close the data connection.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_closeCnx(esm_handle_t    handle,
            int32           nextState,
            esm_event_t     thisEvent,
            void *          cb)

{
    dm_cb_t *   myCb = (dm_cb_t *) cb;
    int32       newState = nextState;
    int32       rc = NFX_OK;
    esm_event_t esmEvent;

    DM_FCT_PRINT(("> dm_closeCnx() RmtPort %d LclPort %d", ntohs(myCb->rmtPort),
						ntohs(myCb->lclPort)));

    /* Close data connection.
     */
#if 0
    if (myCb->isClient == FALSE)
    {
#endif
        rc = tpl_delConnReq(myCb->tplConnId);

        if (rc != NFX_OK)
        {
            DM_ERR_PRINT(("tpl_delConnReq() failed - %d", rc));
            esmEvent.event = DM_EVENT_CLOSE_CNX_ERR;
            esmEvent.p1.u.p = NULL;
            esmEvent.p2.u.p = NULL;
			DM_DBG_PRINT(("Posting event %d to SM %lld\n", esmEvent.event, myCb->dmHdl));
            rc = esm_deliverEvent(myCb->dmHdl, &esmEvent);
            NFX_ASSERT(rc == NFX_OK);
            // newState = ???
        }
		else 
		{
    		/* 
			 * Notify the user application that the connection is closed.
     		 */
    		dm_closeRsp_t   rsp;    // notification to user application

    		rsp.hdr.msgType = DM_CLOSE_RSP;
    		rsp.hdr.appHdl = myCb->appHdl;
    		rsp.hdr.dmHdl = myCb->dmHdl;
    		rsp.result = NFX_OK;
    		rc = dm_sendMessage((void *) &rsp,
                        sizeof(dm_closeRsp_t),
                        myCb->appAddr,
                        dmAddr);
    		NFX_ASSERT(rc == NFX_OK);
    
		}

        DM_ERR_PRINT(("dm_closeCnx request for : %s", \
                       (myCb->isClient == TRUE)?"Source":"Target"));

#if 0
    }
    else
    {
        esmEvent.event = DM_EVENT_CLOSE_CNX_OK;
        esmEvent.p1.u.p = NULL;
        esmEvent.p2.u.p = NULL;
        rc = esm_deliverEvent(myCb->dmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);
        DM_ERR_PRINT(("delcnx event for : %s", \
                       (myCb->isClient == TRUE)?"Source":"Target"));
    }
#endif

    /* Wait for delConnInd (CLOSE_CNX_OK event)...
     */

    DM_ERR_PRINT(("< dm_closeCnx()"));
    return(newState);
}

/*------------------------------------------------------------------------------
 * Name         : dm_kill
 *
 * Description  : Kill this instance of the data mover.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_kill(esm_handle_t    handle,
        int32           nextState,
        esm_event_t     thisEvent,
        void *          cb)

{
    dm_cb_t *       myCb = (dm_cb_t *) cb;
    int32           rc = NFX_OK;
    dm_closeRsp_t   rsp;    // notification to user application
    dm_msgHdr_t     req;    // 'kill me' request to data mover

    DM_FCT_PRINT(("> dm_kill() RmtPort %d LclPort %d", ntohs(myCb->rmtPort),
						ntohs(myCb->lclPort)));

#ifdef ISABELLE // under construction!!!
    // not yet implemented by TPL
    if (myCb->isClient == TRUE)
    {
        /* Connection has been deleted.
         * Unbind and clean control block.
         */
#ifdef LINUX_TEST
        luc_freeLucUserIndex(myCb->lucUserIndex);
        myCb->lucUserIndex = -1;
#endif // LINUX_TEST

        rc = tpl_unbindReq((tpl_bindCb_t *) myCb->tplBindCb);
        NFX_ASSERT(rc == NFX_OK);
        
        myCb->tplBindCb = NULL;
        myCb->tplConnId = NULL;
    }
#endif // ISABELLE    
    
    /* Notify the user application that the connection is closed.
     */
    rsp.hdr.msgType = DM_CLOSE_RSP;
    rsp.hdr.appHdl = myCb->appHdl;
    rsp.hdr.dmHdl = myCb->dmHdl;
    rsp.result = NFX_OK;
    rc = dm_sendMessage((void *) &rsp,
                        sizeof(dm_closeRsp_t),
                        myCb->appAddr,
                        dmAddr);
    NFX_ASSERT(rc == NFX_OK);

    /* Since I cannot kill myself, ask the data mover daemon to kill me.
     */
    req.msgType = DM_KILL_ME;
    req.appHdl = 0;   // not used
    req.dmHdl = myCb->dmHdl;
    rc = dm_sendMessage((void *) &req,
                        sizeof(dm_msgHdr_t),
                        dmAddr,
                        dmAddr);
    NFX_ASSERT(rc == NFX_OK);
    
    return(nextState);  // goto HALTED
}

/*------------------------------------------------------------------------------
 * Name         : dm_err
 *
 * Description  : Error: wrong event received in wrong state.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_err(esm_handle_t handle,
       int32        nextState,
       esm_event_t  thisEvent,
       void *       cb)
    
{
    dm_cb_t *       myCb = (dm_cb_t *) cb;
    int32           rc = NFX_OK;
    struct in_addr  addr;
    esm_event_t     esmEvent;

    DM_ERR_PRINT(("> dm_err() RmtPort %d LclPort %d", ntohs(myCb->rmtPort),
						ntohs(myCb->lclPort)));

    /* Print enough info to be able to debug the error.
     */
    DM_ERR_PRINT(("Internal SW error : event %d received in state %d/%d!",
                 thisEvent, esm_getCurrState(handle), nextState));
    DM_ERR_PRINT(("\thandle to data mover : %d", myCb->dmHdl));
    DM_ERR_PRINT(("\thandle to user application : %d", myCb->appHdl));
    DM_ERR_PRINT(("\tEEE address of the user application : %d", myCb->appAddr));
    DM_ERR_PRINT(("\thandle for TPL : %d", myCb->tplBindCb));
    DM_ERR_PRINT(("\thandle for the connection in TPL : %d",
                  myCb->tplConnId));
    
    DM_ERR_PRINT(("\tdata mover type : %s",
                  (myCb->isClient == TRUE)? "CLIENT" : "SERVER"));
    DM_ERR_PRINT(("\tprotocol : %s",
                  (myCb->protocol == TPL_PROT_UDP)? "UDP" : "TCP"));
    addr.s_addr = myCb->lclAddr;
    DM_ERR_PRINT(("\tlocal IP @ : %s", inet_ntoa(addr)));
    DM_ERR_PRINT(("\tlocal port : %d", ntohs(myCb->lclPort)));
    addr.s_addr = myCb->rmtAddr;
    DM_ERR_PRINT(("\tremote IP @ : %s", inet_ntoa(addr)));
    DM_ERR_PRINT(("\tremote port : %d", ntohs(myCb->rmtPort)));
    
    DM_ERR_PRINT(("\tnb packet sent successfully : %d", myCb->nbPktSntOk));
    DM_ERR_PRINT(("\tnb packet sent unsuccessfully : %d", myCb->nbPktSntNok));
    DM_ERR_PRINT(("\tnb packet received : %d", myCb->nbPktRcv));
   
    /* This is an error in the SW design.
     * Kill the instance and debug it!
     */
    esmEvent.event = DM_EVENT_KILL;
    esmEvent.p1.u.p = NULL;
    esmEvent.p2.u.p = NULL;
	DM_DBG_PRINT(("Posting event %d to SM %d\n", esmEvent.event, handle));
    rc = esm_deliverEvent(handle, &esmEvent);
    NFX_ASSERT(rc == NFX_OK);

    return(nextState);  // stay in same state 
}

/*------------------------------------------------------------------------------
 * Name         : dm_bad_state
 *
 * Description  : The instance is in a bad state.
 * Action function.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_bad_state(esm_handle_t   handle,
             int32          nextState,
             esm_event_t    thisEvent,
             void *         cb)
    
{
    int32       rc = NFX_OK;
    esm_event_t esmEvent;

    DM_ERR_PRINT(("> dm_bad_state() %d", handle));
    
    /* Print error message.
     * Reset the state.
     * Kill itself.
     */
    DM_ERR_PRINT(("Internal SW error : bad state"));
    
    nextState = DM_STATE_IDLE;
    
    esmEvent.event = DM_EVENT_KILL;
    esmEvent.p1.u.p = NULL;
    esmEvent.p2.u.p = NULL;
	DM_DBG_PRINT(("Posting event %d to SM %d\n", esmEvent.event, handle));
    rc = esm_deliverEvent(handle, &esmEvent);
    NFX_ASSERT(rc == NFX_OK);
    
    return(nextState);  // goto IDLE
}

/*----------------------------------------------------------------------------*/
