/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 * RCS          :  $Id: datamover.c,v 1.16 2002/10/03 22:30:06 pankajg Exp $
 *
 * Name         : datamover.c
 *
 * Description  : Data mover daemon.
 *
 * There is only one instance of the data mover daemon .
 * It is running on TxRx on N-NIM from NFX startup till NFX shutdown.
 * It receives all the messages, callbacks for all the data mover instances.
 * It is responsible to dispatch them to the right instances.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 * Re-designed  : 30-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

#include "datamover.h"
#include "exports/malloc-api.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* My slot and CPU identifiers..
 */
extern uint32 nfxMySlotId;
extern uint32 nfxMyCPUId;

/* My own application id and EEE address port.
 */
int32   dmAppId;
uint32  dmAddr;

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : dm_init
 *
 * Description  : Initialization of the data mover daemon.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

void
dm_init(void)

{
    int32   rc;
    
    /* Get my own application id.
     */
    dmAppId = eee_allocateApplicationId("dm");
    NFX_ASSERT(dmAppId != EEE_ERR_INVALID_APP_NAME);

    /* Calculate my own EEE address port.
     */
    dmAddr = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                       nfxMyCPUId,
                                       dmAppId,
                                       0, 0, 0);
    
    /* Register to receive messages.
     */
    rc = eee_registerReceiveFunc(dmAppId,
                                 dm_rcvMsg,
                                 dm_rcvMcastMsg);
    NFX_ASSERT(rc == NFX_OK);

#ifdef LINUX_TEST
    /* Initialize the LUC's users table.
     */
    luc_initLucUserTable();
#endif // LINUX_TEST
    
    return;
}

/*------------------------------------------------------------------------------
 * Name         : dm_cleanup
 *
 * Description  : Clean up the data mover.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

void
dm_cleanup(void)

{
    /* Unregister to receive messages.
     */
    eee_unregisterReceiveFunc(dmAppId);

    /* Deallocate my application id.
     */
    //rc = eee_deallocateApplicationId(dmAppId); - not implemented!!!
    //NFX_ASSERT(rc = NFX_OK);
    
    return;
}

#if defined(NFX_SMP)
/*------------------------------------------------------------------------------
 * Name         : dm_bindReqAck
 *
 * Description  : hanle bind request ack from TPL
 *
 * Created by   : Su Wang
 *
 * Date Created : 12/04/02
 *
 *------------------------------------------------------------------------------
 */
int32
dm_bindReqAck(void* pAppHdl, uint32 rc, void *in_pBind)
{
	
	tpl_bindCb_t	*pBind = (tpl_bindCb_t*)in_pBind;
	
	if (rc != NFX_OK)
		panic("%s: tpl bind failed\n");
	
	if (pBind->bindFlags & TPL_FLAG_CLIENT)
		((dm_cb_t*)pAppHdl)->tplBindCb = pBind;
	else
		((dm_server_t*)pAppHdl)->tplBindCb = pBind;
	
	return NFX_OK;
}
#endif /* NFX_SMP */

/*------------------------------------------------------------------------------
 * Name         : dm_rcvMsg()
 *
 * Description  : Process received message.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

void
dm_rcvMsg(eee_descPtr_t  edesc)

{
    char *          msgBuf = NULL;
	uint32			freeMsgFlag = 1;
    boolean         bufIsInSharedMemory;
    dm_msgType_t *  msgType;
    int32           rc;
    esm_event_t     esmEvent;
	uint32 			srcAddr, dstAddr;
    
    DM_DBG_PRINT(("> dm_rcvMsg()"));

	srcAddr = edesc->hdr.src_port;
	dstAddr = edesc->hdr.dest_port;

	memset(&esmEvent, 0, sizeof(esm_event_t));

    /* Get the message.
     * - calculate its length
     * - if complete message is in only one buffer
     *   - keep it in the system buffer
     *   - free the EEE descriptor
     * - if message is in several buffers
     *   - copy it into a contiguous malloc buffer
     *   - free the received packet (EEE descriptor + buffers).
     */
    {
        int32   i, numBufs;
        uint32  msgLen = 0;
        
        // should be a macro for EEE_GET_NUM_BUFS_IN_EDESC
        numBufs = ((edesc->hdr.control & EDESC_BUFCNT_MASK)
                   >> EDESC_BUFCNT_SHIFT);
        for (i = 0; i < numBufs; i++)
        {
            msgLen += edesc->bd[i+1].len;   // first buffer not used
        } // for
        
        if (numBufs == 1)
        {
            bufIsInSharedMemory = TRUE;
            msgBuf = edesc->bd[1].buf;      // first buffer not used
            eee_deallocateDesc(edesc);
        }
        else
        {
            bufIsInSharedMemory = FALSE;
            msgBuf = (char *) eee_ramAlloc(msgLen);
            NFX_ASSERT(msgBuf != NULL);
            rc = eee_copyPacketToLocalBuffer(msgBuf, edesc, msgLen, 0);
            NFX_ASSERT(rc == msgLen);
            eee_freePacket(edesc);
        }
    }

    /* Process the message according to its type.
     */
    msgType = (dm_msgType_t *) msgBuf;

    switch (*msgType)
    {
        
    case DM_LISTEN_REQ: // server only
    {
        dm_listenReq_t *    req = (dm_listenReq_t *) msgBuf;
        dm_listenRsp_t      rsp;
        tpl_bindCb_t        bindReq;
        tpl_bindCb_t *      bindRsp = NULL;
        int32               rc = NFX_OK;
        dm_server_t *       newServer = NULL;

        DM_FCT_PRINT(("Server Rcvd DM_LISTEN_REQ"));

        /* Create new server and initialize it.
         */
        newServer = (dm_server_t *) eee_ramAlloc(sizeof(dm_server_t));
        NFX_ASSERT(newServer != NULL);
        bzero((void *) newServer, sizeof(dm_server_t));
        newServer->type = DM_SERVER_CB;
        newServer->appHdl = req->hdr.appHdl;
        newServer->appAddr = srcAddr; // edesc->hdr.src_port;
        newServer->tplBindCb = NULL;
        newServer->protocol = req->protocol;
        newServer->lclAddr = req->lclAddr;
        newServer->lclPort = req->lclPort;
        
        /* Start listening to a new port.
         */
        bzero(&bindReq, sizeof(tpl_bindCb_t));
        bindReq.tplProt = req->protocol;
        bindReq.laddr = req->lclAddr;
        bindReq.lport = req->lclPort;
        //bindReq.raddr = 0; - optional - use for authentication only
        //bindReq.rport = 0; - not used
        //bindReq.bingFlags = xxx; - TPL internal use only
        //bindReq.pBsdSock = xxx; - TPL internal use only
        bindReq.ul_addConnInd = dm_addConnInd;
        bindReq.ul_delConnInd = dm_delConnInd;
        bindReq.ul_rcvPktInd = dm_rcvPktInd;
        bindReq.ul_xmtCmpInd = dm_xmtCmpInd;
        bindReq.usrInfo = newServer; 
        {
            struct in_addr  addr;
            
            addr.s_addr = bindReq.laddr;
        
			DM_FCT_PRINT(("Calling tpl_bindReq"));
            DM_FCT_PRINT(("\tlocal IP @ (server): %s", inet_ntoa(addr)));
            DM_FCT_PRINT(("\tlocal port (server): %d", ntohs(bindReq.lport)));
        }
        
#if defined(NFX_SMP)
		bindReq.ul_bindReqAck = dm_bindReqAck;
		rc = tpl_bindReq(&bindReq, newServer, (TPL_BIND_SERVER | TPL_BIND_USE_RPC));
#else
        rc = tpl_bindReq(&bindReq, &bindRsp, 
							(TPL_BIND_SERVER | TPL_BIND_USE_RPC));

		NFX_ASSERT(rc == NFX_OK);

        newServer->tplBindCb = (void *) bindRsp;
#endif /* NFX_SMP */

#ifdef LINUX_TEST
        {
           int idx;
                
           idx = luc_getFreeLucUserIndex();
           lucUserTab[idx].inUse = TRUE;
           lucUserTab[idx].port = newServer->lclPort;
           lucUserTab[idx].sock = dm_openSock(newServer->lclAddr,
                                                   newServer->lclPort);
           newServer->lucUserIndex = idx;
        }
#endif // LINUX_TEST
        
        /* Notify the user application.
         */
        rsp.hdr.msgType = DM_LISTEN_RSP;
        rsp.hdr.appHdl = newServer->appHdl;
        rsp.hdr.dmHdl = 0;  // no DM instance yet!
        rsp.result = rc;
        rsp.lclAddr = req->lclAddr;
        rsp.lclPort = req->lclPort;

        rc = dm_sendMessage((void *) &rsp,
                            sizeof(dm_listenRsp_t),
                            newServer->appAddr,
                            dmAddr);
        NFX_ASSERT(rc == NFX_OK);
    
        /* Listen and wait for client to connect...
         */
            
        break;
    } // case DM_LISTEN_REQ

    case DM_ACCEPT_RSP: // server only
    {
        dm_acceptRsp_t *    rsp = (dm_acceptRsp_t *) msgBuf;
        int32               event;

        DM_FCT_PRINT(("Server Rcvd DM_ACCEPT_RSP"));
        
        if (rsp->result == NFX_OK)
        {
            event = DM_EVENT_ACCEPT_OK;
        }
        else
        {
            event = DM_EVENT_ACCEPT_NOK;
        }
        /* Forward the response to the right instance.
         */
        esmEvent.event = event;
        esmEvent.p1.u.dw = rsp->hdr.appHdl;    // Pass the app Handle.
        esmEvent.p2.u.w = 0;
		DM_DBG_PRINT(("Posting event %d to SM %lld\n", esmEvent.event, 
							rsp->hdr.dmHdl));
        rc = esm_deliverEvent(rsp->hdr.dmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);
        
        break;
    } // case DM_ACCEPT_RSP

    case DM_OPEN_REQ: // client only
    {
        dm_openReq_t *  req = (dm_openReq_t *) msgBuf;
        dm_cb_t *       newCb = NULL;

        DM_FCT_PRINT(("Client Rcvd DM_OPEN_REQ"));
        
        /* Create control block and fill it.
         */
        newCb = (dm_cb_t *) eee_ramAlloc(sizeof(dm_cb_t));
        NFX_ASSERT(newCb != NULL);
        bzero((void *) newCb, sizeof(dm_cb_t));
        newCb->type = DM_INSTANCE_CB;
        //newCb->dmHdl = 0; - will be returned by dms_sm_create()
        newCb->appHdl = req->hdr.appHdl;
        newCb->appAddr = srcAddr; // edesc->hdr.src_port;
        //newCb->tplBindCb = 0; - will be returned in addConnInd() from TPL
        //newCb->tplConnId = 0; - will be returned in addConnInd() from TPL
        newCb->isClient = TRUE;
        newCb->protocol = req->protocol;
        newCb->lclAddr = req->lclAddr;  // could be returned by TPL...
        newCb->lclPort = req->lclPort;  // will be returned by TPL
        newCb->rmtAddr = req->rmtAddr; 
        newCb->rmtPort = req->rmtPort;
        newCb->nbPktSntOk = 0;
        newCb->nbPktSntNok = 0;
        newCb->nbPktRcv = 0;            // Make it one for source.
        {
            struct in_addr  addr;
            
            addr.s_addr = newCb->lclAddr;
            DM_FCT_PRINT(("\tlocal IP @ (source) : %s", inet_ntoa(addr)));
            DM_FCT_PRINT(("\tlocal port (source) : %d",
                          ntohs(newCb->lclPort)));
            addr.s_addr = newCb->rmtAddr;
            DM_FCT_PRINT(("\tremote IP @ (target) : %s", inet_ntoa(addr)));
            DM_FCT_PRINT(("\tremote port (target) : %d",
                          ntohs(newCb->rmtPort)));
        }
        
        /* Create a new instance of the data mover
         * and store its handle in its control block.
         */
        rc = dm_sm_create((esm_handle_t *) &newCb->dmHdl,
                          (void *) newCb);
        NFX_ASSERT(rc == NFX_OK);

		DM_DBG_PRINT(("New Client Datamover SM is %lld\n", newCb->dmHdl));
        
        /* Open the data connection.
         */
        esmEvent.event = DM_EVENT_OPEN_CNX_REQ;
        esmEvent.p1.u.p = req;
        esmEvent.p2.u.w = bufIsInSharedMemory;
		freeMsgFlag = 0; // Msg passed with event. Can't free it here
		DM_DBG_PRINT(("Posting event %d to SM %lld\n", 
						esmEvent.event, newCb->dmHdl));
        rc = esm_deliverEvent(newCb->dmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);

        break;
    } // case DM_OPEN_REQ

    case DM_SEND_REQ:
    {
        dm_sendReq_t * req = (dm_sendReq_t *) msgBuf;

        /* Forward the request to the right instance.
         */
        esmEvent.event = DM_EVENT_SND_DATA_REQ;
        esmEvent.p1.u.p = req;
        esmEvent.p2.u.w = bufIsInSharedMemory;
		freeMsgFlag = 0; // Msg passed with event. Can't free it here
		DM_DBG_PRINT(("Posting event %d to SM %lld\n", 	
						esmEvent.event, req->hdr.dmHdl));
        rc = esm_deliverEvent(req->hdr.dmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);

        break;
    } // case DM_SEND_REQ

    case DM_CLOSE_REQ:
    {
        dm_closeReq_t * req = (dm_closeReq_t *) msgBuf;

        DM_FCT_PRINT(("Rcvd DM_CLOSE_REQ for SM %lld", req->hdr.dmHdl));

        /* Forward the request to the right instance.
         */
        esmEvent.event = DM_EVENT_CLOSE_CNX_REQ;
    	esmEvent.p1.u.p = NULL;
    	esmEvent.p2.u.p = NULL;
		DM_DBG_PRINT(("Posting event %d to SM %lld\n", 	
						esmEvent.event, req->hdr.dmHdl));
        rc = esm_deliverEvent(req->hdr.dmHdl, &esmEvent);
        //NFX_ASSERT(rc == NFX_OK);

        break;
    } // case DM_CLOSE_REQ

    case DM_KILL_ME:
    {
        dm_msgHdr_t *   req = (dm_msgHdr_t *) msgBuf;
        dm_cb_t *       myCb = NULL;
        
        DM_FCT_PRINT(("Rcvd DM_KILL_ME for SM %lld", req->dmHdl));
    
        /* Destroy the instance and free its control block.
         */
        rc = dm_sm_destroy(req->dmHdl);
        NFX_ASSERT(rc == NFX_OK);

        myCb = (dm_cb_t *) esm_getUserInfo(req->dmHdl);
        NFX_ASSERT(myCb != NULL);
        rc = eee_ramDealloc(myCb);
        NFX_ASSERT(rc == NFX_OK);

        break;
    } // case DM_KILL_ME
   
    default:
    {
        DM_ERR_PRINT(("Rcvd Unknown msg type (%d)", *msgType));
        DM_ERR_PRINT(("Message is discarded"));
        break;
    }
    } // switch

    /* 
	 * Free the message buffer if it is not in use
	 * as a parameter to an event being posted here
     */
	if (freeMsgFlag)
	{
    	if (bufIsInSharedMemory == TRUE)
    	{
        	eee_deallocateBuffer(msgBuf);
    	}
    	else
    	{
        	rc = eee_ramDealloc(msgBuf);
        	NFX_ASSERT(rc == NFX_OK);
    	}
	}

    return;
}

/*------------------------------------------------------------------------------
 * Name         : dm_rcvMcastMsg()
 *
 * Description  : Process received multicast message.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

void
dm_rcvMcastMsg(eee_descPtr_t  edesc)

{
    
    DM_ERR_PRINT(("No multicast messages are expected by the data mover."));
    DM_ERR_PRINT(("Message is discarded."));
    eee_freePacket(edesc);
    
}

/*------------------------------------------------------------------------------
 * Name         : dm_addConnInd
 *
 * Description  : Indication from TPL: a connection has been added.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_addConnInd(tpl_bindCb_t *    pBind,
              void *            pConnId,
              void **           ppAppHdl,
              //uint32            status, - not yet implemented in TPL
              uint32            raddr,
              ushort16          rport)

{
    int32       rc = NFX_OK;
    ushort16    type = *((ushort16*) pBind->usrInfo);
    dm_cb_t *   myCb = NULL;
    int32       event;
    esm_event_t esmEvent;
    

    if (type == DM_SERVER_CB)
    {
        /* A client is connected.
         * Create DM instance to handle it.
         * Ask user application if it accepts or rejects the connection.
         */
        dm_server_t * server = (dm_server_t *) pBind->usrInfo;

    	DM_DBG_PRINT(("> dm_addConnInd() Server Side"));
        
        myCb = (dm_cb_t *) eee_ramAlloc(sizeof(dm_cb_t));
        NFX_ASSERT(myCb != NULL);
        bzero((void *) myCb, sizeof(dm_cb_t));
        myCb->type = DM_INSTANCE_CB;
        // myCb->appHdl = server->appHdl;
        myCb->appHdl = 0;    // This is going to AM
        myCb->appAddr = server->appAddr;
        myCb->tplBindCb = server->tplBindCb;
        myCb->tplConnId = pConnId;
        myCb->isClient = FALSE;
        myCb->protocol = server->protocol;
        myCb->lclAddr = server->lclAddr;
        myCb->lclPort = server->lclPort;
        myCb->rmtAddr = raddr; 
        myCb->rmtPort = rport;
        myCb->nbPktSntOk = 0;
        myCb->nbPktSntNok = 0;
        myCb->nbPktRcv = 1;
#ifdef LINUX_TEST
        myCb->lucUserIndex = server->lucUserIndex;
#endif // LINUX_TEST
        {
            struct in_addr  addr;
            
            addr.s_addr = myCb->lclAddr;
            addr.s_addr = myCb->rmtAddr;
        }
        
        /* Create a new instance of the data mover
         * and store its handle in its control block.
         */
        rc = dm_sm_create((esm_handle_t *) &myCb->dmHdl,
                          (void *) myCb);
        NFX_ASSERT(rc == NFX_OK);

		DM_DBG_PRINT(("New Server Datamover SM is %lld\n", myCb->dmHdl));
        
        /* Open the data connection.
         */
        esmEvent.event = DM_EVENT_OPEN_CNX_IND;
        esmEvent.p1.u.p = NULL;
        esmEvent.p2.u.p = NULL;
		DM_DBG_PRINT(("Posting event %d to SM %lld\n", 	
						esmEvent.event, myCb->dmHdl));
        rc = esm_deliverEvent(myCb->dmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);

    }
    else if (type == DM_INSTANCE_CB)
    {
        myCb = (dm_cb_t *) pBind->usrInfo;
        
        NFX_ASSERT(myCb != NULL);

    	DM_DBG_PRINT(("> dm_addConnInd() Client Side"));
   
#ifndef ISABELLE
        /* If TPL sends an indication that a connection has been added,
         * it's assumed that it has been successfully added.
         */
        event = DM_EVENT_OPEN_CNX_OK;
        myCb->tplConnId = pConnId;
        myCb->lclAddr = pBind->laddr;
        myCb->lclPort = pBind->lport;
        {
            struct in_addr  addr;
            
            addr.s_addr = myCb->lclAddr;
        }
#else
        /* To use when TPL will provide a status.
         */
        if (status == NFX_OK)
        {
            event = DMS_EVENT_OPEN_CNX_OK;
            myCb->tplConnId = pConnId;
            myCb->lclAddr = pBind->laddr;
            myCb->lclPort = pBind->lport;
            {
                struct in_addr  addr;
                
                addr.s_addr = myCb->lclAddr;
            }
        }
        else
        {
            event = DMS_EVENT_OPEN_CNX_ERR;
        }
#endif // ISABELLE
      
        esmEvent.event = event;
        esmEvent.p1.u.p = NULL;
        esmEvent.p2.u.p = NULL;
		DM_DBG_PRINT(("Posting event %d to SM %lld\n", 	
						esmEvent.event, myCb->dmHdl));
        rc = esm_deliverEvent(myCb->dmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);
    }
    else
    {
        DM_ERR_PRINT(("User info retruned by TPL is incorrect!"));
        rc = NFX_ERR;
        // NFX_ASSERT();
    }
    
    /* Give back to TPL a pointer to the datamover control block.
     */
    *ppAppHdl = myCb;
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : dm_delConnInd
 *
 * Description  : Indication from TPL: a connection has been deleted.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_delConnInd(void *    pAppHdl,
              // status - not yet implemented by TPL
              void *    pConnId)

{
    int32       rc = NFX_OK;
    dm_cb_t *   myCb = (dm_cb_t *) pAppHdl;
    int32       event;
    esm_event_t esmEvent;
    
    DM_DBG_PRINT(("> dm_delConnInd()"));
    NFX_ASSERT(myCb != NULL);

#ifndef ISABELLE
    /* If TPL sends an indication that a connection has been deleted,
     * it's assumed that it has been successfully deleted.
     */
    event = DM_EVENT_CLOSE_CNX_OK;
#else
    /* To use when TPL will provide a status.
     */
    if (status = NFX_OK)
    {
        event = DM_EVENT_CLOSE_CNX_OK;
    }
    else
    {
        event = DM_EVENT_CLOSE_CNX_ERR;
    }   
#endif // ISABELLE

    esmEvent.event = event;
    esmEvent.p1.u.p = NULL;
    esmEvent.p2.u.p = NULL;
	DM_DBG_PRINT(("Posting event %d to SM %lld\n", 	
					esmEvent.event, myCb->dmHdl));
    rc = esm_deliverEvent(myCb->dmHdl, &esmEvent);
    //NFX_ASSERT(rc == NFX_OK);
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : dm_rcvPktInd
 *
 * Description  : Indication from TPL: a packet has been received.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_rcvPktInd(void *         pAppHdl,
             pkt_desc_t *   pDesc)
    
{
    int32       rc = NFX_OK;
    dm_cb_t *   myCb = (dm_cb_t *) pAppHdl;
    esm_event_t esmEvent;
    
    DM_DBG_PRINT(("> dm_rcvPktInd()"));
    NFX_ASSERT(myCb != NULL);

//#ifdef DM_UDP
    if (myCb->protocol == TPL_PROT_UDP)
    {
        // ignore 1st packet - it's only to get the addConnInd
        if (myCb->nbPktRcv == 0)
        {
            myCb->nbPktRcv ++;
#if 0
            /* Notify the user application that the connection is
             * successfully connected.
             */
            rsp.hdr.msgType = DM_OPEN_RSP;
            rsp.hdr.appHdl = myCb->appHdl;
            rsp.hdr.dmHdl = myCb->dmHdl;
            rsp.result = NFX_OK;
            rsp.lclAddr = myCb->lclAddr;
            rsp.lclPort = myCb->lclPort;
            rc = dm_sendMessage((void *) &rsp,
                                 sizeof(dm_openRsp_t),
                                 myCb->appAddr,
                                 myAddr);
            NFX_ASSERT(rc == NFX_OK);
#endif
            return(TPL_FREE_PKT);
        }
    }
//#endif // DM_UDP

    /* Forward the packet to the right instance.
     */
    esmEvent.event = DM_EVENT_RCV_DATA_IND;
    esmEvent.p1.u.p = pDesc;
    esmEvent.p2.u.p = NULL;
	DM_DBG_PRINT(("Posting event %d to SM %lld\n", 	
					esmEvent.event, myCb->dmHdl));
    rc = esm_deliverEvent(myCb->dmHdl, &esmEvent);
    NFX_ASSERT(rc == NFX_OK);
    
    return(TPL_HOLD_PKT); /* we will pass it to application to free it */
}

/*------------------------------------------------------------------------------
 * Name         : dms_xmtCmpInd
 *
 * Description  : Indication from TPL: a packet has been transmitted on source.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
dm_xmtCmpInd(void *         pAppHdl,
             // status - not yet implemented by TPL
             pkt_desc_t *   pDesc)

{
    dm_cb_t *   myCb = (dm_cb_t *) pAppHdl;
    int32       rc = NFX_OK;
    int32       event;
    esm_event_t esmEvent;
    
    DM_DBG_PRINT(("> dm_xmtCmpInd()"));
	//pDesc->laddr, pDesc->raddr, ntohs(pDesc->lport), ntohs(pDesc->rport)));

    NFX_ASSERT(myCb != NULL);

//#ifdef DM_UDP
    if (myCb->protocol == TPL_PROT_UDP)
    {
        // ignore 1 st packet - it was sent to get target connected
        if (myCb->nbPktSntOk == 0)
        {
            myCb->nbPktSntOk++;
            return(TPL_FREE_PKT);   // free packet
        }
    }
//#endif // DM_UDP
    
#ifndef ISABELLE
    /* If TPL sends an indication that a packet has been transmitted,
     * it's assumed that it has been successfully transmitted.
     */
    event = DM_EVENT_SND_DATA_OK;
#else
    /* To use when TPL will provide a status.
     */
    if (status == NFX_OK)
    {
        event = DM_EVENT_SND_DATA_OK;
    }
    else
    {
        event = DM_EVENT_SND_DATA_ERR;
    }
#endif // ISABELLE

    esmEvent.event = event;
    esmEvent.p1.u.p = NULL;
    esmEvent.p2.u.p = NULL;
	DM_DBG_PRINT(("Posting event %d to SM %lld\n", 	
					esmEvent.event, myCb->dmHdl));
    rc = esm_deliverEvent(myCb->dmHdl, &esmEvent);
    //NFX_ASSERT(rc == NFX_OK);
    
    return(TPL_FREE_PKT);   // free packet
}



/*----------------------------------------------------------------------------*/
