/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea-rx.c
 *
 * Description  :
 * Entry module of the Eagle Agent (EA).
 * It contains the main() function.
 * This module receives all the messages sent to EA.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/file.h>
#include "ea.h"
#include "spm-api.h"
#include "raid-api.h"
#include "event-api.h"
#include "exports/eee-desc.h"
#include "neteee/eee.h"
#include "exports/eee-error.h"

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static void
ea_terminate(int sig);
static void ea_stopSession(int32 pid);
static int32 ea_startSession(int32 pid);
uint32 ea_childGetLock(char *resName);
void ea_childUnLock(void);
char *ea_printMsg(uint32 msgType);
int32 ea_createNewSocket(void);

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */


/* EA's file descriptor */
static int eaDesc;
int eaAppId;

int eaChildPidTable[MAX_EA_SESSIONS];
int eaNumSession = 0;

/*
 * Lock related defines and global variables
 */

#define EA_LOCKFILE     "/var/ea-lock/"

int lockFileFd = -1;
char lockFileName[VOL_MAX_NAME+20];

/*------------------------------------------------------------------------------
 * Functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_printMsg()
 *
 * Description  : Print ea Msg info
 *
 * Created by   :  Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
char *ea_printMsg(uint32 msgType)
{
	switch (msgType)
	{
    	case EA_LOGIN_REQ:
    		return  "EA_LOGIN_REQ";
 
    	case EA_LOGIN_RSP:
    		return "EA_LOGIN_RSP";

		case EA_SPM_GET_TOTAL_POOLS_REQ:
			return "EA_SPM_GET_TOTAL_POOLS_REQ";

		case EA_SPM_GET_TOTAL_POOLS_RSP:
			return "EA_SPM_GET_TOTAL_POOLS_RSP";

		case EA_SPM_GET_POOL_LIST_REQ:
			return "EA_SPM_GET_POOL_LIST_REQ";

		case EA_SPM_GET_POOL_LIST_RSP:
			return "EA_SPM_GET_POOL_LIST_RSP";

		case EA_SPM_GET_TOTAL_LUNS_REQ:
			return "EA_SPM_GET_TOTAL_LUNS_REQ";

		case EA_SPM_GET_TOTAL_LUNS_RSP:
			return "EA_SPM_GET_TOTAL_LUNS_RSP";

		case EA_SPM_GET_LUN_LIST_REQ:
			return "EA_SPM_GET_LUN_LIST_REQ";

		case EA_SPM_GET_LUN_LIST_RSP:
			return "EA_SPM_GET_LUN_LIST_RSP";

		case EA_SPM_MARK_LUN_AS_AGILE_REQ:
			return "EA_SPM_MARK_LUN_AS_AGILE_REQ";

		case EA_SPM_MARK_LUN_AS_AGILE_RSP:
			return "EA_SPM_MARK_LUN_AS_AGILE_RSP";

		case EA_SPM_MARK_LUN_AS_OC_REQ:
			return "EA_SPM_MARK_LUN_AS_OC_REQ";

		case EA_SPM_MARK_LUN_AS_OC_RSP:
			return "EA_SPM_MARK_LUN_AS_OC_RSP";

		case EA_SPM_UNMARK_LUN_AS_AGILE_REQ:
			return "EA_SPM_UNMARK_LUN_AS_AGILE_REQ";

		case EA_SPM_UNMARK_LUN_AS_AGILE_RSP:
			return "EA_SPM_UNMARK_LUN_AS_AGILE_RSP";

		case EA_SPM_RESCAN_LUN_REQ:
			return "EA_SPM_RESCAN_LUN_REQ";

		case EA_SPM_RESCAN_LUN_RSP:
			return "EA_SPM_RESCAN_LUN_RSP";

		case EA_SPM_GET_STORAGE_POOL_LIST_REQ:
			return "EA_SPM_GET_STORAGE_POOL_LIST_REQ";

		case EA_SPM_GET_STORAGE_POOL_LIST_RSP:
			return "EA_SPM_GET_STORAGE_POOL_LIST_RSP";

		case EA_SPM_GET_STORAGE_LUN_LIST_REQ:
			return "EA_SPM_GET_STORAGE_LUN_LIST_REQ";

		case EA_SPM_GET_STORAGE_LUN_LIST_RSP:
			return"	case EA_SPM_GET_STORAGE_LUN_LIST_RSP";
 
		case EA_SPM_SRM_INIT_IND:
			return "EA_SPM_SRM_INIT_IND";

		case EA_SPM_SRM_NEW_LUN_IND:
			return "EA_SPM_SRM_NEW_LUN_IND";

		case EA_SPM_SRM_DEL_LUN_IND:
			return "EA_SPM_SRM_DEL_LUN_IND";

		case EA_SPM_SRM_LUN_STATE_CHANGE_IND:
			return "EA_SPM_SRM_LUN_STATE_CHANGE_IND";

    	case EA_FSYS_NTFY:
    		return "EA_FSYS_NTFY";

    	case EA_FSET_NTFY:
    		return "EA_FSET_NTFY";

    	case EA_QUOTA_NTFY:
    		return "EA_QUOTA_NTFY";

    	case EA_SNAP_NTFY:
    		return "EA_SNAP_NTFY";

    	case EA_OO_VOL_NTFY:
    		return "EA_OO_VOL_NTFY";

    	case EA_DEL_VOL_NTFY:
    		return "EA_DEL_VOL_NTFY";

    	case EA_CRT_VOL_NTFY:
    		return "EA_CRT_VOL_NTFY";

    	case EA_GRW_VOL_NTFY:
    		return "EA_GRW_VOL_NTFY";

    	case EA_MOD_VOL_NTFY:
    		return "EA_MOD_VOL_NTFY";

    	case EA_FULL_VOL_NTFY:
    		return "EA_FULL_VOL_NTFY";

    	case EA_FAIL_LUN_NTFY:
    		return "EA_FAIL_LUN_NTFY";

    	case EA_TAKEOVER_VOL_NTFY:
    		return "EA_TAKEOVER_VOL_NTFY";

    	case EA_ADD_FSYS_REQ:
    		return "EA_ADD_FSYS_REQ";

    	case EA_ADD_FSYS_RSP:
    		return "EA_ADD_FSYS_RSP";

    	case EA_MOD_FSYS_REQ:
    		return "EA_MOD_FSYS_REQ";

    	case EA_MOD_FSYS_RSP:
    		return "EA_MOD_FSYS_RSP";

    	case EA_MNT_FSYS_REQ:
    		return "EA_MNT_FSYS_REQ";

    	case EA_MNT_FSYS_RSP:
    		return "EA_MNT_FSYS_RSP";

    	case EA_UMNT_FSYS_REQ:
    		return "EA_UMNT_FSYS_REQ";

    	case EA_UMNT_FSYS_RSP:
    		return "EA_UMNT_FSYS_RSP";

    	case EA_DEL_FSYS_REQ:
    		return "EA_DEL_FSYS_REQ";

    	case EA_DEL_FSYS_RSP:
    		return "EA_DEL_FSYS_RSP";

    	case EA_NUM_FSYS_REQ:
    		return "EA_NUM_FSYS_REQ";

    	case EA_NUM_FSYS_RSP:
    		return "EA_NUM_FSYS_RSP";

    	case EA_GET_FSYS_REQ:
    		return "EA_GET_FSYS_REQ";

    	case EA_GET_FSYS_RSP:
    		return "EA_GET_FSYS_RSP";

    	case EA_GET_DETAIL_FSYS_REQ:
    		return "EA_GET_DETAIL_FSYS_REQ";

    	case EA_GET_DETAIL_FSYS_RSP:
    		return "EA_GET_DETAIL_FSYS_RSP";

    	case EA_USG_FSYS_REQ:
    		return "EA_USG_FSYS_REQ";

    	case EA_USG_FSYS_RSP:
    		return "EA_USG_FSYS_RSP";

    	case EA_MOVIN_FSYS_REQ:
    		return "EA_MOVIN_FSYS_REQ";

    	case EA_MOVIN_FSYS_RSP:
    		return "EA_MOVIN_FSYS_RSP";

    	case EA_MOVOUT_FSYS_REQ:
    		return "EA_MOVOUT_FSYS_REQ";

    	case EA_MOVOUT_FSYS_RSP:
    		return "EA_MOVOUT_FSYS_RSP";

    	case EA_GROW_FSYS_REQ:
    		return "EA_GROW_FSYS_REQ";

    	case EA_GROW_FSYS_RSP:
    		return "EA_GROW_FSYS_RSP";

    	case EA_GROW_FSYS_IND:
    		return "EA_GROW_FSYS_IND";

		case EA_FORCE_MOVIN_FSYS_REQ:
			return "EA_FORCE_MOVIN_FSYS_REQ";

		case EA_FORCE_MOVIN_FSYS_RSP:
			return "EA_FORCE_MOVIN_FSYS_RSP";

    	case EA_ADD_FSET_REQ:
    		return "EA_ADD_FSET_REQ";

    	case EA_ADD_FSET_RSP:
    		return "EA_ADD_FSET_RSP";

    	case EA_MOD_FSET_REQ:
    		return "EA_MOD_FSET_REQ";

    	case EA_MOD_FSET_RSP:
    		return "EA_MOD_FSET_RSP";

    	case EA_DEL_FSET_REQ:
    		return "EA_DEL_FSET_REQ";

    	case EA_DEL_FSET_RSP:
    		return "EA_DEL_FSET_RSP";

    	case EA_NUM_FSET_REQ:
    		return "EA_NUM_FSET_REQ";

    	case EA_NUM_FSET_RSP:
    		return "EA_NUM_FSET_RSP";

    	case EA_GET_FSET_REQ:
    		return "EA_GET_FSET_REQ";

    	case EA_GET_FSET_RSP:
    		return "EA_GET_FSET_RSP";

    	case EA_USG_FSET_REQ:
    		return "EA_USG_FSET_REQ";

    	case EA_USG_FSET_RSP:
    		return "EA_USG_FSET_RSP";

    	case EA_ADD_EXP_REQ:
    		return "EA_ADD_EXP_REQ";

    	case EA_ADD_EXP_RSP:
    		return "EA_ADD_EXP_RSP";

    	case EA_MOD_EXP_REQ:
    		return "EA_MOD_EXP_REQ";

    	case EA_MOD_EXP_RSP:
    		return "EA_MOD_EXP_RSP";

    	case EA_DEL_EXP_REQ:
    		return "EA_DEL_EXP_REQ";

    	case EA_DEL_EXP_RSP:
    		return "EA_DEL_EXP_RSP";

    	case EA_DEL_PATH_REQ:
    		return "EA_DEL_PATH_REQ";

    	case EA_DEL_PATH_RSP:
    		return "EA_DEL_PATH_RSP";

    	case EA_GET_EXP_REQ:
    		return "EA_GET_EXP_REQ";

    	case EA_GET_EXP_RSP:
    		return "EA_GET_EXP_RSP";

    	case EA_GET_PATH_EXP_REQ:
    		return "EA_GET_PATH_EXP_REQ";

    	case EA_GET_PATH_EXP_RSP:
    		return "EA_GET_PATH_EXP_RSP";

    	case EA_CRT_SNAP_REQ:
    		return "EA_CRT_SNAP_REQ";

    	case EA_CRT_SNAP_RSP:
    		return "EA_CRT_SNAP_RSP";

    	case EA_DEL_SNAP_REQ:
    		return "EA_DEL_SNAP_REQ";

    	case EA_DEL_SNAP_RSP:
    		return "EA_DEL_SNAP_RSP";

    	case EA_SCH_SNAP_REQ:
    		return "EA_SCH_SNAP_REQ";

    	case EA_SCH_SNAP_RSP:
    		return "EA_SCH_SNAP_RSP";

    	case EA_GET_USG_SNAP_REQ:
    		return "EA_GET_USG_SNAP_REQ";

    	case EA_GET_USG_SNAP_RSP:
    		return "EA_GET_USG_SNAP_RSP";

    	case EA_GET_SCH_SNAP_REQ:
    		return "EA_GET_SCH_SNAP_REQ";

    	case EA_GET_SCH_SNAP_RSP:
    		return "EA_GET_SCH_SNAP_RSP";

    	case EA_GET_CHASSIS_INFO_REQ:
    		return "EA_GET_CHASSIS_INFO_REQ";

    	case EA_GET_CHASSIS_INFO_RSP:
    		return "EA_GET_CHASSIS_INFO_RSP";

		case EA_GET_FC_PORT_WWN_REQ:
			return "EA_GET_FC_PORT_WWN_REQ";

		case EA_GET_FC_PORT_WWN_RSP:
			return "EA_GET_FC_PORT_WWN_RSP";

		case EA_LUN_REDISCOVER_REQ:
			return "EA_LUN_REDISCOVER_REQ";

		case EA_LUN_REDISCOVER_RSP:
			return "EA_LUN_REDISCOVER_RSP";

    	case EA_GET_PORT_TOPOLOGY_REQ:
    		return "EA_GET_PORT_TOPOLOGY_REQ";

		case EA_GET_PORT_TOPOLOGY_RSP:
			return "EA_GET_PORT_TOPOLOGY_RSP";

		case EA_RAID_MID_GET_LIST_REQ:
			return "EA_RAID_MID_GET_LIST_REQ";
 
		case EA_RAID_MID_GET_LIST_RSP:
			return "EA_RAID_MID_GET_LIST_RSP";
	
		case EA_RAID_MID_OPEN_REQ:
			return "EA_RAID_MID_OPEN_REQ";

		case EA_RAID_MID_OPEN_RSP:
			return "EA_RAID_MID_OPEN_RSP";

		case EA_RAID_MID_CLOSE_REQ:
			return "EA_RAID_MID_CLOSE_REQ";

		case EA_RAID_MID_CLOSE_RSP:
			return "EA_RAID_MID_CLOSE_RSP";

		case EA_RAID_MID_CTRL_INFO_REQ:
			return "EA_RAID_MID_CTRL_INFO_REQ";
	
		case EA_RAID_MID_CTRL_INFO_RSP:
			return "EA_RAID_MID_CTRL_INFO_RSP";
	
		case EA_RAID_MID_CTRL_GETPARAM_REQ:
			return "EA_RAID_MID_CTRL_GETPARAM_REQ";
 
		case EA_RAID_MID_CTRL_GETPARAM_RSP:
			return "EA_RAID_MID_CTRL_GETPARAM_RSP";
 
		case EA_RAID_MID_PHYDEV_LIST_REQ:
			return "EA_RAID_MID_PHYDEV_LIST_REQ";
 
		case EA_RAID_MID_PHYDEV_LIST_RSP:
			return "EA_RAID_MID_PHYDEV_LIST_RSP";
 
		case EA_RAID_MID_PHYDEV_INFO_REQ:
			return "EA_RAID_MID_PHYDEV_INFO_REQ";
 
		case EA_RAID_MID_PHYDEV_INFO_RSP:
			return "EA_RAID_MID_PHYDEV_INFO_RSP";
 
		case EA_RAID_MID_LOGDEV_LIST_REQ:
			return "EA_RAID_MID_LOGDEV_LIST_REQ";

		case EA_RAID_MID_LOGDEV_LIST_RSP:
			return "EA_RAID_MID_LOGDEV_LIST_RSP";
 
		case EA_RAID_MID_LOGDEV_INFO_REQ:
			return "EA_RAID_MID_LOGDEV_INFO_REQ";
 
		case EA_RAID_MID_LOGDEV_INFO_RSP:
			return "EA_RAID_MID_LOGDEV_INFO_RSP";
 
		case EA_RAID_MID_LUN_CREATE_REQ:
			return "EA_RAID_MID_LUN_CREATE_REQ";

		case EA_RAID_MID_LUN_CREATE_RSP:
			return "EA_RAID_MID_LUN_CREATE_RSP";
 
		case EA_RAID_MID_LUN_CREATE_SIMPLE_REQ:
			return "EA_RAID_MID_LUN_CREATE_SIMPLE_REQ";

		case EA_RAID_MID_LUN_CREATE_SIMPLE_RSP:
			return "EA_RAID_MID_LUN_CREATE_SIMPLE_RSP";

		case EA_RAID_MID_LUN_DELETE_REQ:
			return "EA_RAID_MID_LUN_DELETE_REQ";
 
		case EA_RAID_MID_LUN_DELETE_RSP:
			return "EA_RAID_MID_LUN_DELETE_RSP";

		case EA_RAID_MID_LUN_INIT_REQ:
			return "EA_RAID_MID_LUN_INIT_REQ";

		case EA_RAID_MID_LUN_INIT_RSP:
			return "EA_RAID_MID_LUN_INIT_RSP";

		case EA_RAID_MID_LUNMASK_INFO_REQ:
			return "EA_RAID_MID_LUNMASK_INFO_REQ";

		case EA_RAID_MID_LUNMASK_INFO_RSP:
			return "EA_RAID_MID_LUNMASK_INFO_RSP";

		case EA_RAID_MID_LUNMASK_ADD_REQ:
			return "EA_RAID_MID_LUNMASK_ADD_REQ";

		case EA_RAID_MID_LUNMASK_ADD_RSP:
			return "EA_RAID_MID_LUNMASK_ADD_RSP";

		case EA_RAID_MID_LUNMASK_DEL_REQ:
			return "EA_RAID_MID_LUNMASK_DEL_REQ";

		case EA_RAID_MID_LUNMASK_DEL_RSP:
			return "EA_RAID_MID_LUNMASK_DEL_RSP";

		case EA_RAID_MID_LUN_BPLIST_REQ:
			return "EA_RAID_MID_LUN_BPLIST_REQ";

		case EA_RAID_MID_LUN_BPLIST_RSP:
			return "EA_RAID_MID_LUN_BPLIST_RSP";

		case EA_RAID_MID_LUN_CANCELBP_REQ:
			return "EA_RAID_MID_LUN_CANCELBP_REQ";

		case EA_RAID_MID_LUN_CANCELBP_RSP:
			return "EA_RAID_MID_LUN_CANCELBP_RSP";

		case SPM_GET_TOTAL_POOLS_RSP:
			return "SPM_GET_TOTAL_POOLS_RSP";

		case SPM_GET_POOL_LIST_RSP:
			return "SPM_GET_POOL_LIST_RSP";

		case SPM_GET_TOTAL_LUNS_RSP:
			return "SPM_GET_TOTAL_LUNS_RSP";

		case SPM_GET_LUN_LIST_RSP:
			return "SPM_GET_LUN_LIST_RSP";

		case SPM_MARK_LUN_AS_AGILE_RSP:
			return "SPM_MARK_LUN_AS_AGILE_RSP";

		case SPM_MARK_LUN_AS_OC_RSP:
			return "SPM_MARK_LUN_AS_OC_RSP";

		case SPM_UNMARK_LUN_AS_AGILE_RSP:
			return "SPM_UNMARK_LUN_AS_AGILE_RSP";

		case SPM_RESCAN_LUN_RSP:
			return "SPM_RESCAN_LUN_RSP";

		case SPM_GET_STORAGE_POOL_LIST_RSP:
			return "SPM_GET_STORAGE_POOL_LIST_RSP";

		case SPM_GET_STORAGE_LUN_LIST_RSP:
			return "SPM_GET_STORAGE_LUN_LIST_RSP";

		case SPM_SRM_INIT_IND:
			return "SPM_SRM_INIT_IND";

		case SPM_SRM_NEW_LUN_IND:
			return "SPM_SRM_NEW_LUN_IND";

		case SPM_SRM_DEL_LUN_IND:
			return "SPM_SRM_DEL_LUN_IND";

		case SPM_SRM_LUN_STATE_CHANGE_IND:
			return "SPM_SRM_LUN_STATE_CHANGE_IND";

		default:
			return "Unknown EA msg";

	}
}

/*------------------------------------------------------------------------------
 * Name         : ea_createNewSocket()
 *
 * Description  : This is used by a EA process to create a new
 *                socket in case it expects the response to come back
 *				  asynchronously
 *
 * Created by   :  Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
int32 ea_createNewSocket()
{
	int sock;
	struct sockaddr_eee seee;

	seee.seee_family = AF_EEE;
	seee.seee_bplane = 0;
	seee.seee_slot = SSC_SLOT;
	seee.seee_cpu = NFX_SSC_UNIX;
	seee.seee_app = 0;

	if ((sock = socket(PF_EEE, SOCK_DGRAM, 0)) >= 0) 
	{
		if (bind(sock, (struct sockaddr *)&seee, sizeof(seee)) < 0) 
		{
			close(sock);
			sock = -1;
		}
	}

	return sock;
}

/*------------------------------------------------------------------------------
 * Name         : ea_startSession()
 *
 * Description  : See if we have room for another child session
 *
 * Created by   :  Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
int32 ea_startSession(int32 pid)
{
	int32 idx;
	
	for (idx = 0; idx < MAX_EA_SESSIONS; idx++)
	{
		if (eaChildPidTable[idx] == 0)
		{
			eaChildPidTable[idx] = pid;
			eaNumSession++;
			return 1;
		}
	}

	return -1;
}

/*------------------------------------------------------------------------------
 * Name         : ea_stopSession()
 *
 * Description  : Clear a child entry for the exiting child process
 *                If pid is zero means terminate and kill all child process
 *
 * Created by   :  Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
void ea_stopSession(int32 pid)
{
	int32 idx;

	for (idx = 0; idx < MAX_EA_SESSIONS; idx++)
	{
		if ((pid == 0) || (eaChildPidTable[idx] == pid))
		{
			if (eaChildPidTable[idx])
			{
				kill(eaChildPidTable[idx], SIGTERM);
				eaChildPidTable[idx] = 0;
				eaNumSession--;

				/*
			 	 * If we found the specific process we were
			 	 * looking for then we dont need to search 
			 	 * any more
			 	 */
				if (pid) 
					break;
			}
		}
	}
}

/*------------------------------------------------------------------------------
 * Name         : ea_childGetLock()
 *
 * Description  : Get a lock and start exclusive operation
 *                Return NFX_OK if lock for exclusive operation
 *                was successful else return NFX_ERR.
 *                This lock is automatically released upon the
 *                termination of this process
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
uint32 ea_childGetLock(char *resName)
{
	sprintf(lockFileName, "%s%s", EA_LOCKFILE, resName);

	lockFileFd = open (lockFileName, O_RDWR | O_CREAT);
	if (lockFileFd < 0)
	{
		EA_ERR_PRINT(("ea_childGetLock: Could not open lock file %s\n", 
						lockFileName));
		return NFX_ERR;
	}

	if (flock(lockFileFd, LOCK_EX|LOCK_NB) == 0)
	{
		EA_FCT_PRINT(("ea_childGetLock: Got lock on %s\n", lockFileName));
		return NFX_OK;
	}
	else
	{
		EA_ERR_PRINT(("ea_childGetLock: Lock failed: Volume %s is busy\n", 
									lockFileName));
		close(lockFileFd);
		lockFileFd = -1;
		return NFX_ERR;
	}
}

/*------------------------------------------------------------------------------
 * Name         : ea_childUnLock()
 *
 * Description  : Release any locks you may be holding
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
void ea_childUnLock()
{
	/*
	 * Return if you have no locks
	 */
	if (lockFileFd < 0)
		return;

	EA_DBG_PRINT(("ea_childUnLock: Removing lock for %s\n", 
						lockFileName));
	
	close(lockFileFd);
	unlink(lockFileName);
	lockFileFd = -1;
}

/*------------------------------------------------------------------------------
 * Name         : ea_childDone()
 *
 * Description  : Signal handler for child signals
 *
 * Created by   :  Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_childDone(int sig)
{
	int pid;
	int status = 0;

	while (1)
	{
		pid = waitpid(-1, &status, WNOHANG);
		
		if (pid <= 0)
		{
			break;
		}
		else
		{
			ea_stopSession(pid);
		}
	}
}

/*------------------------------------------------------------------------------
 * Name         : ea_terminate()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_terminate(int sig)
{
    EA_FCT_PRINT(("> ea_terminate()\n"));

	signal(SIGCHLD, SIG_IGN);

    /* Cleanup different EA entities.
     */
	ea_stopSession(0);

	/*
	 * Remove any lock files
	 */
	ea_childUnLock();

    /* Unregister EA.
     */
    unRegAgileApp(eaDesc);
    
    EA_FCT_PRINT(("< ea_terminate()\n"));
    exit(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : main
 *
 * Description  : EA's entry point.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int main(int argc, char **argv, char **envp) 
{
	int pid;

    EA_FCT_PRINT(("> main()\n"));

	/*
	 * Reset child process table
	 */
	memset(&eaChildPidTable[0], 0, sizeof(int)*MAX_EA_SESSIONS);
	lockFileFd = -1;

    /*
     * Register to receive messages for EA.
     * Get its application id.
     */
	eaDesc = regAgileApp("ea");
	EA_ASSERT(eaDesc >= 0);
    eaAppId = eee_getApplicationId("ea");
    
    /*
     * Sets the procedure to follow in case of request for program termination
     * or CTRL-C interrupt.
     */
    signal(SIGTERM, ea_terminate);
    signal(SIGINT, ea_terminate);
    signal(SIGCHLD, ea_childDone);

    /* Initialize different EA entities.
     */
    
	/*
	 * Tell mcp we're ready to roll.
	 */
	kill (getppid(), SIGUSR1);

    /*
	 * Receive messages and process them.
     */

	while (TRUE)
    {
        uint32	*msgType;
        ea_senderInfo_t sender;
        ea_msg_t        msg;

        /*
         * Wait for messages...
         */
		msg.len = recvAgileMsg(eaDesc,
                               msg.buf, sizeof(msg.buf),
                               NULL,
                               sender.name,
                               &sender.cpu,
                               &sender.slot,
                               &sender.plane);
        EA_ASSERT(msg.len >= 0);

        /*
         * The first 32 bits of any messages contain its type.
         */
        msgType = (uint32 *) &msg;

		if (*msgType == NFX_EVENT_POST)
		{
			nfx_eventMsg_t * evt = (nfx_eventMsg_t *) &msg;

            EA_DBG_PRINT(("EA main process rcvd NFX_EVENT_POST message"));

			if ((evt->eventType == NFX_EVENT_CPU) &&
                (evt->param.cpuEvent.slot == NFX_SLOTTYPE_NFP) &&
                (evt->param.cpuEvent.cpu == NFX_NFP_FP1) &&   
                (evt->param.cpuEvent.state == EVENT_STATE_DOWN))
			{
				/* FP core 0 (FP1) failed. Reset any EA sessiosns
			 	 * TBD make sure in case of multiple FP
			 	 * only sessions related to the affected FP
			 	 * are reset
			 	 */
				EA_ERR_PRINT(("Process-EA Msg: NFX_EVENT_CPU -> FP1 down"));
	
				signal(SIGCHLD, SIG_IGN);
				ea_stopSession(0);
    			signal(SIGCHLD, ea_childDone);
				continue;
			}
            else
            {
                /* Ignore this event.
                 */
                continue;
            }
		}
        
        /*
         * Something has been received, fork off a child process
		 * and let it handle it. First check if we have not reached
		 * max sessions already
         */
		if (eaNumSession >= MAX_EA_SESSIONS)
		{
			/*
			 * Can't handle any more requests
			 * Just ignore the requestor should time out
			 */
			EA_ERR_PRINT(("EA cannot start any more sessions Cur (%d)\n",
							eaNumSession));
			continue;
		}

		pid = fork();

		if (pid > 0)
		{
			/*
			 * Parent forked successfully
			 * Remember the child process
			 */
			ea_startSession(pid);
		}
		else if (pid == 0)
        {
			/*
			 * Only child reaches here
			 */
			close(eaDesc);

            EA_FCT_PRINT(("Process-EA Msg: %s (%d) from %s!\n",
                         ea_printMsg(*msgType), *msgType, sender.name));
            /*
             * Determine the sender from the range of the message type,
             * and process the message accordingly.
             * (see nfx-defs.h)
             */
            if ((EA_MSG_BASE <= *msgType) && (*msgType < EV_MSG_BASE))
            {
                /*
                 * The message is coming from EV.
                 */
                ea_processEvMsg(&msg, &sender);
            }
			else if ((SPM_MSG_BASE <= *msgType) && (*msgType < NFS_MSG_BASE))
			{
                /*
                 * The message is coming from SPM.
                 */
                ea_processEvMsg(&msg, &sender);
			}
            else 
            {
                EA_ERR_PRINT(("Process-EA Msg: Unexpected message %d\n",
                             *msgType));
            }

			/* 
			 * child process can exit after it's work is done
			 * If the return was due to an error condition
			 * we may have a lock pending. Make sure to remove
			 * any lock files
			 */
			ea_childUnLock();

			exit(0); 
        }
		else
		{
			EA_ERR_PRINT(("EA could not start child process\n"));
		}

    } /* main parent process to continue the while loop */

    
    EA_DBG_PRINT(("< main()\n"));
    exit(NFX_ERR);  // should never exit by this way!
}

/*----------------------------------------------------------------------------*/
