/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea.c
 *
 * Description  : Eagle Agent (EA).
 * This files contain all the operation that can be made one a storage pool,
 * a volume, a file system and a file set.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

#include "ea.h"
#include <unistd.h>
#include "exports/rpc.h"
#include "exports/mount.h"
#include "exports/sdm-api.h"
#include "exports/fs-opt-api.h"
#include "hal.h"

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

extern int myAsyncSock;

/* Wait for EVM/EFS to respond (60s).
 */
const struct timeval waitEvmTo = {60, 0};
const struct timeval waitMountdTo = {60, 0};
const struct timeval waitEfsTo = {60, 0};
const struct timeval waitSdmTo = {60, 0};
const struct timeval waitCmTo = {60, 0};

#define BYTES_IN_MB		(1024*1024)
#define BYTES_IN_GB		(1024*1024*1024)


/*------------------------------------------------------------------------------
 * Local function prototypes
 *------------------------------------------------------------------------------
 */
int32 ea_createNewSocket(void);

        
/*------------------------------------------------------------------------------
 * Helper functions
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Action on a volume.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_crtVol()
 *
 * Description  : Create volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_crtVol(char *node,
			ea_fsysCb_t * cb,
			char *        err)

{
    int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgVolCreate_t      *evmReq = (evm_MsgVolCreate_t *) reqBuf.buf;
    evm_MsgStat_t           *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
	const struct timeval 	waitTo = {360, 0};

    EA_FCT_PRINT(("> ea_crtVol()\n"));

    /* Reset memory for request
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_CREATE;
    evmReq->hdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(evmReq->config), (const void *) (cb->vol),
           sizeof(evm_VolConfig_t));    

	myAsyncSock = ea_createNewSocket();
	if (myAsyncSock < 0)
	{
		EA_ERR_PRINT(("ea_crtVol: Could not create new socket\n"));
		return NFX_ERR;
	}
	// app_name_from_socket(myAsyncSock, evmReq->eaAppName);

{
char tmpbuf[64];
	app_name_from_socket(myAsyncSock, tmpbuf);
	EA_DBG_PRINT(("ea_crtVol: created socket with appname %s\n", tmpbuf));
}

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(myAsyncSock,
							&reqBuf, sizeof(evm_MsgVolCreate_t),
							"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							(struct timeval *) &waitTo,
							&rspBuf, sizeof(rspBuf.buf),
							node);
	if (rspSize <= 0)
	{
		sprintf(err, "ea_crtVol(): Unable to send request");
		return NFX_ERR;
	}

    /* Receive response from EVM.
     * Get volume id.
     */
	if (rspSize < sizeof(evm_MsgStat_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;

    if ((evmRsp->stat == NFX_OK) || (evmRsp->stat == EVM_VOL_GROW_PENDING))
    {
        memcpy((void *) &(cb->vol->id),
               (void *) &(evmRsp->payload.lVolId),
               sizeof(evm_volId_t));
    }

	if (evmRsp->stat != NFX_OK)
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));

    EA_FCT_PRINT(("< ea_crtVol()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_delVol()
 *
 * Description  : Delete volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_delVol(char *node,
			ea_fsysCb_t * cb,
			char *        err)

{
	int					rspSize;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
    evm_MsgVolDestroy_t *evmReq = (evm_MsgVolDestroy_t *) reqBuf.buf;
    evm_MsgStat_t       *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_delVol()\n"));

    /* Reset memory for request
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_DESTROY;
    evmReq->hdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(evmReq->id),
           (void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    
    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
							&reqBuf, sizeof(evm_MsgVolDestroy_t),
							"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							(struct timeval *) &waitEvmTo,
							&rspBuf, sizeof(rspBuf.buf),
							node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_delVol(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;

    if (evmRsp->stat != NFX_OK)
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));

    EA_DBG_PRINT(("< ea_delVol()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_renameVol()
 *
 * Description  : Rename a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 24-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_renameVol(char *node,
				ea_fsysCb_t *  cb,
				char *         err)

{
	int					rspSize;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
    evm_MsgVolRename_t  *evmReq = (evm_MsgVolRename_t *) reqBuf.buf;
    evm_MsgStat_t       *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_renameVol()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_RENAME;
    evmReq->hdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(evmReq->id),
           (void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    strcpy(evmReq->newName, cb->newFsysName);

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
							&reqBuf, sizeof(evm_MsgVolRename_t),
							"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							(struct timeval *) &waitEvmTo,
							&rspBuf, sizeof(rspBuf.buf),
							node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_renameVol(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     * Copy new name in control block.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;

    if (evmRsp->stat == NFX_OK)
        strcpy(cb->vol->name, cb->newFsysName);
    else
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));
    
    EA_FCT_PRINT(("< ea_renameVol()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_modVol()
 *
 * Description  : Modify configuration of a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 24-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_modVol(char *node,
			ea_fsysCb_t * cb,
			char *        err)

{
	int					rspSize;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
    evm_MsgSetOpt_t     *evmReq = (evm_MsgSetOpt_t *) reqBuf.buf;
    evm_MsgStat_t       *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_modVol()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_SET_LV_OPTS;
    evmReq->hdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(evmReq->id),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    evmReq->maxSize = cb->newHardQuota;
    evmReq->minAutoGrow = cb->newMinAutoGrow;
	evmReq->softQuota = cb->newSoftQuota;
	evmReq->type = cb->vol->type; // don't modify the type from EA yet!
    
    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
							&reqBuf, sizeof(evm_MsgSetOpt_t),
							"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							(struct timeval *) &waitEvmTo,
							&rspBuf, sizeof(rspBuf.buf),
							node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_modVol(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     * Update control block with modified values.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;

    if (evmRsp->stat != NFX_OK)
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));
    
    EA_FCT_PRINT(("< ea_modVol()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_takeoverVol()
 *
 * Description  : takeover a volume.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 16-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_takeoverVol(char *node, ea_fsysCb_t *cb, char *err, uint32 delay)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgVolTakeover_t	*evmReq = (evm_MsgVolTakeover_t *) reqBuf.buf;
    evm_MsgStat_t			*evmRsp = (evm_MsgStat_t *) rspBuf.buf;
	struct timeval waitTakeoverTo = {120, 0};
    
    EA_FCT_PRINT(("> ea_takeoverVol()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_TAKEOVER;
    evmReq->hdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(evmReq->id),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    evmReq->nim = cb->vol->onNIM;
	evmReq->delay = delay;

	/* adjust sendAgilemsg timeout value according to delay */
	waitTakeoverTo.tv_sec = delay + 10;
	waitTakeoverTo.tv_usec = 0;

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
							&reqBuf, sizeof(evm_MsgVolTakeover_t),
							"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            &waitTakeoverTo,
							&rspBuf, sizeof(rspBuf.buf),
							node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_takeoverVol(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;
    
    if (evmRsp->stat != NFX_OK)
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));

    EA_FCT_PRINT(("< ea_takeoverVol()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_onVol()
 *
 * Description  : Put on-line a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_onVol(char *node,
			ea_fsysCb_t *  cb,
			char *         err)

{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgVolOnline_t		*evmReq = (evm_MsgVolOnline_t *) reqBuf.buf;
    evm_MsgStat_t           *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_onVol()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_ONLINE;
    evmReq->hdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(evmReq->id),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    evmReq->nim = cb->vol->onNIM;
	evmReq->checkFlag = 1;

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(evm_MsgVolOnline_t),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEvmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);

    /* Receive response from EVM.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;
   
    if ((evmRsp->stat != NFX_OK) && (evmRsp->stat != EVM_VOL_ALREADY_ONLINE))
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));

    EA_FCT_PRINT(("< ea_onVol()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_offVol()
 *
 * Description  : Put off-line a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_offVol(char *node,
			ea_fsysCb_t * cb,
			char *        err)

{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgVolOffline_t		*evmReq = (evm_MsgVolOffline_t *) reqBuf.buf;
    evm_MsgStat_t           *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_offVol()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_OFFLINE;
    evmReq->hdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(evmReq->id),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    evmReq->nim = cb->vol->onNIM;

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(evm_MsgVolOffline_t),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEvmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_offVol(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;
    
    if ((evmRsp->stat != NFX_OK) && (evmRsp->stat != EVM_VOL_ALREADY_OFFLINE))
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));

    EA_FCT_PRINT(("< ea_offVol()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_getClusterVolIdNode()
 *
 * Description  : Get volume ID and node name for a specific volume name
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 10-OCT-2002
 *
 * Parameters   :
 *   node - OUT, node name that has this volume (only for EA on PCC)
 *               null string if local node or volume not found
 *   vol  - IN,  assume only volume name vol->name is valid in input
 *          OUT, vol->id will be filled if found
 *   err  - OUT, error message if any
 *
 *------------------------------------------------------------------------------
 */

int32
ea_getClusterVolIdNode(evm_VolConfig_t *vol, char *node, char *err)
{
	// node == NULL is for local only, should skip cluster
	if (node)
	{
		*node = '\0';
		// get id from cluster... This will only work if this
		// instance of EA is running on PCC
		if (fs_opt_get_clus_vol_id(vol->name, &(vol->id)) == NFX_OK)
		{
		char tmpb[256];

			// get name of the node that has this volume (volume id)
			// if failed, force it to local
			fs_opt_get_clus_lvol_online_node(vol->id, node);
			gethostname(tmpb, sizeof(tmpb));
			// if this is "local" or my own hostname, then
			// reset the buffer to null string...
			if (!strcmp(node, "local") || !strcmp(node, tmpb))
				*node = '\0';
			return NFX_OK;
		}
	}

	/* Unable to get volume id from cluster. Either it's a non-PCC
	 * or volume name doesn't exists on cluster. But we'll try
	 * local anyway....
	 */
	return ea_getVolId(NULL, vol, err);
}


/*------------------------------------------------------------------------------
 * Name         : ea_getVolId()
 *
 * Description  : Get volume ID from an EVM on a specific node in cluster
 *                using volume name.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       NULL or null string if local
 *
 * NOTE         : This function is a dumb function. It'll only look for
 *                volume ID from volume name on the specific node. If
 *                volume name specified is not on that specific node,
 *                this function will return error.
 *
 *------------------------------------------------------------------------------
 */

int32
ea_getVolId(char *              node,
			evm_VolConfig_t *   vol,
			char *              err)

{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgGetId_t			*evmReq = (evm_MsgGetId_t *) reqBuf.buf;
    evm_MsgStat_t           *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_getVolId()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_GET_ID;
    evmReq->hdr.clientPrivate = 0;   // not used
    strcpy(evmReq->name, vol->name);

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(evm_MsgGetId_t),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEvmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_getVolId(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     * Copy the returned volume id in control block.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;

    if (evmRsp->stat == NFX_OK)
    {
        memcpy((void *) &(vol->id),
               (const void *) &(evmRsp->payload.lVolId),
               sizeof(evm_volId_t));
    }
    else if (err)
        strcpy(err, rspBuf.buf + sizeof(evm_MsgStat_t));
    
    EA_FCT_PRINT(("< ea_getVolId()\n"));
    return(evmRsp->stat);
}

/*------------------------------------------------------------------------------
 * Name         : ea_getNodeVolIdFromVolName()
 *
 * Description  : Get node name and vol ID for a specific volume name
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 10-OCT-2002
 *
 * Parameters   :
 *   volname - IN,  volume name
 *   node    - OUT, node name where the specified volume resides
 *                  if NULL then caller is not interested in this value
 *   volId   - OUT, volume ID for the specified volume
 *                  if NULL then caller is not interested in this value
 *
 *------------------------------------------------------------------------------
 */

int32
ea_getNodeVolIdFromVolName(char *volname, char *node, evm_volId_t *volId)
{
int32 rc = NFX_OK;
evm_volId_t myid;

	// get id from cluster... This will only work if this
	// instance of EA is running on PCC
	rc = fs_opt_get_clus_vol_id(volname, &myid);
	if (rc == NFX_OK)
	{
		// get name of the node that has this volume (volume id)
		if (node)
		{
			rc = fs_opt_get_clus_lvol_online_node(myid, node);
			if (!strcmp(node, "local"))
				*node = '\0';
		}
		if (volId)
			*volId = myid;
		return rc;
	}

	/* Unable to get volume id from cluster. Either it's a non-PCC
	 * or volume name doesn't exists on cluster. But we'll try
	 * local anyway....
	 */
	if (node)
		*node = '\0';

	if (volId)
		return ea_getVolId2(NULL, volname, volId);

	return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name         : ea_getVolId2()
 *
 * Description  : Another ea_getVolId() to get volume ID. The parameter
 *                is different, that's it...
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 18-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       NULL or null string if local
 *
 * NOTE         : This function is a dumb function. It'll only look for
 *                volume ID from volume name on the specific node. If
 *                volume name specified is not on that specific node,
 *                this function will return error.
 *
 *------------------------------------------------------------------------------
 */

int32
ea_getVolId2(char *node,
			char *volname,
			evm_volId_t *volId)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgGetId_t			*evmReq = (evm_MsgGetId_t *) reqBuf.buf;
    evm_MsgStat_t           *evmRsp = (evm_MsgStat_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_getVolId2()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_GET_ID;
    evmReq->hdr.clientPrivate = 0;   // not used
    strcpy(evmReq->name, volname);

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(evm_MsgGetId_t),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEvmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_getVolId(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     * Copy the returned volume id in control block.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		evmRsp->stat = NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		evmRsp->stat = NFX_ERR;

    if ((evmRsp->stat == NFX_OK) && volId)
		*volId = evmRsp->payload.lVolId;
    
    EA_FCT_PRINT(("< ea_getVolId2()\n"));
    return(evmRsp->stat);
}


/*------------------------------------------------------------------------------
 * Name         : ea_getVolCfg()
 *
 * Description  : Get complete volume configuration.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_getVolCfg(char *node, evm_VolConfig_t *vol)

{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgVolGetAttrib_t	*evmReq = (evm_MsgVolGetAttrib_t *) reqBuf.buf;
    evm_MsgVolCreate_t		*evmRsp = (evm_MsgVolCreate_t *) rspBuf.buf;
    
    EA_FCT_PRINT(("> ea_getVolCfg()\n"));

    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    evmReq->hdr.msgType = EVM_VOL_GET_ATTRIB;
    evmReq->hdr.clientPrivate = 0;   // not used
    memcpy((void *) &(evmReq->id),
           (const void *) &(vol->id),
           sizeof(evm_volId_t));

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(evm_MsgVolGetAttrib_t),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEvmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_getVolCfg(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from EVM.
     * Save new volume configuration in control block.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
    {
        printf("ea_getVolCfg(): %d < %d\n", rspSize, sizeof(evm_msgHdr_t));
		return NFX_ERR;
    }

	if (evmRsp->hdr.msgType != EVM_VOL_CONFIG)
    {
        printf("ea_getVolCfg(): %d != %d\n", evmRsp->hdr.msgType, EVM_VOL_CONFIG);
		return NFX_ERR;
    }

    memcpy((void *) vol, (const void *) &(evmRsp->config),
           sizeof(evm_VolConfig_t));

    EA_DBG_PRINT(("\tevmRsp->config.attrs.num = %d\n",
                  evmRsp->config.attrs.num));
    EA_DBG_PRINT(("\tvolume name = %s\n", vol->name));
    EA_DBG_PRINT(("\tvolume id = %lld\n", vol->id.whole_thing));
    EA_DBG_PRINT(("\ton NIM = %d\n", vol->onNIM));
    
    EA_FCT_PRINT(("< ea_getVolCfg()\n"));

    return(NFX_OK); // EVM doesn't return a status!!!
}

/*------------------------------------------------------------------------------
 * Name         : ea_getVolIdList()
 *
 * Description  : Get list of all existing volume ids.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-NOV-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       NULL or null string if local (PCC)
 *
 * !!!NOTE!!!!  : ea_getVolIdList() will allocate memory (malloc).
 *                It's caller's responsibility to release this memory
 *
 *------------------------------------------------------------------------------
 */

ea_fsysList_t *
ea_getVolIdList(char *node)

{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    evm_MsgVolGetAttrib_t	*evmReq = (evm_MsgVolGetAttrib_t *) reqBuf.buf;
    evm_MsgVolList_t		*evmRsp = (evm_MsgVolList_t *) rspBuf.buf;
    
    uint32                  i;
	ea_fsysList_t           *fsysList;

    EA_FCT_PRINT(("> ea_getVolIdList()\n"));
    
    /* Reset memory for request and response.
     */
    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    /* Fill request.  */
    evmReq->hdr.msgType = EVM_VOL_GET_ATTRIB;
    evmReq->hdr.clientPrivate = 0;   // not used
    evmReq->id.whole_thing = -1;

    /* Send request to EVM.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(evm_MsgVolGetAttrib_t),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEvmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_getVolIdList(): unable to send request\n"));
		return NULL;
	}

    /* Receive response from EVM.
     * Save new volume configuration in control block.
     */
	if (rspSize <  sizeof(evm_msgHdr_t))
		return NULL;

	if (evmRsp->hdr.msgType != EVM_VOL_LIST)
		return NULL;

    if (evmRsp->num <= 1)
    {
		fsysList = (ea_fsysList_t *) malloc(sizeof(ea_fsysList_t));
    	if (!fsysList)
			return NULL;
		memset(fsysList, 0, sizeof(ea_fsysList_t));
    }
	else
	{
    	uint32 size = sizeof(ea_fsysList_t) +
						(evmRsp->num - 1) * sizeof(ea_fsysCfg_t);

        fsysList = (ea_fsysList_t *) malloc(size);
    	if (!fsysList)
			return NULL;
		memset(fsysList, 0, size);
	}

    for (i = 0; i < evmRsp->num; i++)
    {
        memcpy((void *) &(fsysList->fsys[i].fsysId.volId),
               (const void *) &(evmRsp->ids[i]),
               sizeof(evm_volId_t));
    } // for
    fsysList->nb = i;

    EA_FCT_PRINT(("< ea_getVolIdList()\n"));
    return fsysList;
}

/*------------------------------------------------------------------------------
 * Name         : ea_getVolList()
 *
 * Description  : Get list of all existing volume with config
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-NOV-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       NULL or null string if local (PCC)
 *
 * !!!NOTE!!!!  : ea_getVolList() will call ea_getVolIdList() to get a
 *                list of vol ids. ea_getVolIdList() will allocate memory
 *                (malloc) for this list. ea_getVolList() will not free
 *                this memory if it return this list to caller of ea_getVolList().
 *                It's caller's responsibility to release this memory
 *------------------------------------------------------------------------------
 */

ea_fsysList_t *
ea_getVolList(char *node)
{
    uint32              i;
    evm_VolConfig_t *   vol = NULL;
    int32               rc = NFX_OK;
    ea_fsysList_t       *fsysList = NULL;

    EA_FCT_PRINT(("> ea_getVolList()\n"));

    /* Get list of all volumes and ids.
     */
    fsysList = ea_getVolIdList(node);
	if (!fsysList)
		return NULL;
    
    /* Get name of all volumes.
     */
    vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
	if (!vol)
	{
		free(fsysList);
		return NULL;
	}

    for (i = 0; i < fsysList->nb; i++)
    {
        bzero((void *) vol, sizeof(evm_VolConfig_t));
        memcpy((void *) &(vol->id),
               (const void *) &(fsysList->fsys[i].fsysId.volId),
               sizeof(evm_volId_t));
        rc = ea_getVolCfg(node, vol);
        if (rc == NFX_OK)
        {
            strcpy(fsysList->fsys[i].fsysId.fsysName, vol->name);
            fsysList->fsys[i].onlineStatus = vol->onLine;
            fsysList->fsys[i].maxSize = vol->maxSize;
            fsysList->fsys[i].currSize = vol->currSize;
            fsysList->fsys[i].minAutoGrow = vol->minAutoGrow;
            fsysList->fsys[i].softQuota = vol->softQuota;
            fsysList->fsys[i].type = vol->type;
            fsysList->fsys[i].raidLevel = vol->attrs.raidLevel;

            strcpy(fsysList->fsys[i].vendorInfo[0].vendor,
                    vol->attrs.vendorInfo[0].vendor);
            strcpy(fsysList->fsys[i].vendorInfo[0].model,
                    vol->attrs.vendorInfo[0].model);
            fsysList->fsys[i].mountStatus = 1;
			
			ea_getFsysInfo(node, vol->id, &fsysList->fsys[i].fsState, 
									&fsysList->fsys[i].usedSize,
									&fsysList->fsys[i].highWaterMark);
        }
        else
            break;
    } // for
    free(vol);

    EA_FCT_PRINT(("< ea_getVolList()\n"));
    return fsysList;
}

/*------------------------------------------------------------------------------
 * Name         : ea_getVolLunList()
 *
 * Description  : Get LUN list for the specified volume
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 * Parameters   : node - node name if this is for remote node
 *                       NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */
evm_lvStrg_t *ea_getVolLunList(char *node, evm_volId_t volId, uint32 *numStrg)
{
	int					rspSize;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
    evm_MsgGetStrg_t	*strgReq = (evm_MsgGetStrg_t *) reqBuf.buf;
    evm_MsgLVStrg_t		*strgRsp = (evm_MsgLVStrg_t *) rspBuf.buf;
    
    int                 i;
	int					strgSize;
	evm_lvStrg_t *		strgList;

    bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

    strgReq->hdr.msgType = EVM_GET_LV_STRG;
    strgReq->hdr.clientPrivate = getpid();
    strgReq->id = volId;

	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(evm_MsgGetStrg_t),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEvmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_getVolLunList(): unable to send request\n"));
		return NULL;
	}

	if (rspSize <  sizeof(evm_msgHdr_t))
		return NULL;

	if (strgRsp->hdr.msgType != EVM_LV_STRG_CFG)
		return NULL;

	/*
	 * Allocate the storage block for response based
	 * on num of LUN entries
	 */
	strgSize = strgRsp->nbStrg * sizeof(evm_lvStrg_t);
	strgList = (evm_lvStrg_t *) malloc(strgSize);
    NFX_ASSERT(strgList != NULL);

	bzero((void *) strgList, strgSize);

	*numStrg = strgRsp->nbStrg;
    for (i = 0; i < strgRsp->nbStrg; i++)
		memcpy(&strgList[i], &strgRsp->strg[i], sizeof(evm_lvStrg_t));

    return strgList;
}

/*------------------------------------------------------------------------------
 * Action on a file system.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_getFsysInfo()
 *
 * Description  : Get the current used size of a file system
 *
 * Created by   : Pankaj Gupta
 *
 * Parameters   : node - node name if this is for remote node
 *                       must be NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */
int32 ea_getFsysInfo (char *node,
					evm_volId_t volId, uint32 *state,
					uint64 *usedSize,
					uint32 *highWaterMark)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgViewFsReq			*fsReq = (fs_MsgViewFsReq *) reqBuf.buf;
    fs_MsgViewFsResp		*fsRsp = (fs_MsgViewFsResp *) rspBuf.buf;
    
	fs_FSInfo *fsInfo;

	*usedSize = 0;

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
	{
		*usedSize = 0;
		*state = 0;
		*highWaterMark = 0;
		return NFX_ERR;
	}

    /* Reset memory for request
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));

	fsReq->hdr.requestType = FS_REQ_VIEWFS;
	fsReq->hdr.len = sizeof(fsReq);
	fsReq->hdr.clientPrivate = getpid();
	fsReq->type = FS_VIEW_ONE_VOLUME;
	memcpy((void *) &(fsReq->volId), (const void *) &(volId),
			              sizeof(evm_volId_t));

	rspSize = ea_sendClusterAgileMsg(-1,
                                     &reqBuf, sizeof(fs_MsgViewFsReq),
                                     "efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
                                     (struct timeval *) &waitEfsTo,
                                     &rspBuf, sizeof(rspBuf.buf),
                                     node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_getFsysInfo(): unable to send request\n"));
		return NFX_ERR;
	}

	if (rspSize < sizeof(fs_MsgViewFsResp))
	{
		*usedSize = 0;
		*state = 0;
		*highWaterMark = 0;
		return NFX_ERR;
	}

	if (fsRsp->nbrFs == 0)
	{
		*usedSize = 0;
		*state = FS_VOL_STATE_OFFLINE;
		*highWaterMark = 0;
		return NFX_ERR;
	}

	fsInfo = &(fsRsp->u.fsInfo[0]);
	*usedSize = fsInfo->stat.numBlks - fsInfo->stat.numBlksAvail;
	*state = fsInfo->state;

	if (fsInfo->percentMinFree == 100)
		*highWaterMark = DISABLE_HIGHWATERMARK;
	else
		*highWaterMark = fsInfo->percentMinFree;

	return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name         : ea_crtFsys()
 *
 * Description  : Create a file system.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 17-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_crtFsys(char *node, ea_fsysCb_t * cb)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgMakeFsReq			*fsReq = (fs_MsgMakeFsReq *) reqBuf.buf;
    fs_MsgMakeFsResp		*fsRsp = (fs_MsgMakeFsResp *) rspBuf.buf;

	struct timeval varEfsTo;

    EA_FCT_PRINT(("> ea_crtFsys()\n"));

    /* Reset memory for request
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->hdr.requestType = FS_REQ_MAKEFS;
    fsReq->hdr.len = sizeof(fs_MsgMakeFsReq);
    fsReq->hdr.clientPrivate = cb->hdr->userInfo;
    /*
      do I need the following parameters???
      fs_AttrType attrType; // Bitmap of requested attribute types.
      uint32      uid; // User ID for the request
      uint32      gid; // Group ID for the request
      fs_Proto    proto; // Client protocol.
      int32       numGids; // Number of valid entries in gids[].
      int32       pad;
      uint32      gids[FS_REQ_GIDS_MAX];
    */
    memcpy((void *) &(fsReq->volId),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    fsReq->numBlks = (cb->vol->currSize*EVM_LBLK_SIZE) / EFS_BLOCK_SIZE;
    fsReq->blkSize = EFS_BLOCK_SIZE;
    fsReq->ext_log = FALSE;

    /* Send request to FS.
     */
	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

	/*
	 * Use file system timeout as 120 sec + 60 sec for each GB
	 */
	varEfsTo.tv_sec = 120 + ((60*cb->vol->currSize*EVM_LBLK_SIZE)/BYTES_IN_GB);
	varEfsTo.tv_usec = 0;

	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgMakeFsReq),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
                        &varEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_crtFsys(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgMakeFsResp))
		fsRsp->hdr.fsStatus = NFX_ERR;

	if (fsRsp->hdr.requestType != FS_REQ_MAKEFS)
		fsRsp->hdr.fsStatus = NFX_ERR;

    
    EA_FCT_PRINT(("< ea_crtFsys()\n"));
    return(fsRsp->hdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_modFsys()
 *
 * Description  : Modify a file system.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 17-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_modFsys(char *node, ea_fsysCb_t * cb)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgSetMinFreeReq		*fsReq = (fs_MsgSetMinFreeReq *) reqBuf.buf;
    fs_MsgSetMinFreeResp	*fsRsp = (fs_MsgSetMinFreeResp *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_modFsys()\n"));

    /* Reset memory for request and response.
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->hdr.requestType = FS_REQ_SET_MIN_FREE;
    fsReq->hdr.len = sizeof(fs_MsgSetMinFreeReq);
    fsReq->hdr.clientPrivate = cb->hdr->userInfo;

    memcpy((void *) &(fsReq->volId),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));

    fsReq->minFree = cb->newHighWaterMark;

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

    /* Send request to FS.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgSetMinFreeReq),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_crtFsys(): unable to send req to node %s\n", node));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgSetMinFreeResp))
		fsRsp->hdr.fsStatus = NFX_ERR;

	if (fsRsp->hdr.requestType != FS_REQ_SET_MIN_FREE)
		fsRsp->hdr.fsStatus = NFX_ERR;

    EA_FCT_PRINT(("< ea_modFsys()\n"));

    return(fsRsp->hdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_mntFsys()
 *
 * Description  : Mount a file system.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 17-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_mntFsys(char *node, ea_fsysCb_t *cb)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgMountReq          *fsReq = (fs_MsgMountReq *) reqBuf.buf;
    fs_MsgMountResp         *fsRsp = (fs_MsgMountResp *) rspBuf.buf;

	struct timeval varEfsTo;

    EA_FCT_PRINT(("> ea_mntFsys()\n"));

    /* Reset memory for request and response.
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->hdr.requestType = FS_REQ_MOUNT;
    fsReq->hdr.len = sizeof(fs_MsgMountReq) + strlen(cb->vol->name) + 1;
    fsReq->hdr.clientPrivate = cb->hdr->userInfo;
    /*
      do I need the following parameters???
      fs_AttrType attrType; // Bitmap of requested attribute types.
      uint32      uid; // User ID for the request
      uint32      gid; // Group ID for the request
      fs_Proto    proto; // Client protocol.
      int32       numGids; // Number of valid entries in gids[].
      int32       pad;
      uint32      gids[FS_REQ_GIDS_MAX];
    */
    memcpy((void *) &(fsReq->volId),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    fsReq->fsNameLen = strlen(cb->vol->name);
    strcpy(fsReq->fsName, cb->vol->name);

    fsReq->logDevType = FS_LOG_DEVTYPE_DEFAULT;
//    uint32              logSize;    // Size of external device
//                                    //   only valid for FS_LOG_DEVTYPE_VOLID
//    evm_volId_t         logVolId;   // Volume id for ext vol logging
//                                    //   only valid for FS_LOG_DEVTYPE_VOLID

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

    /* Send request to FS.
     */
	// varEfsTo.tv_sec = 1;	// testing
	varEfsTo.tv_sec = 300;	// wait for 5 minutes first
	varEfsTo.tv_usec = 0;

	rspSize = ea_sendClusterAgileMsg(-1,
					&reqBuf, sizeof(fs_MsgMountReq) + strlen(cb->vol->name),
					"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
					(struct timeval *) &varEfsTo,
					&rspBuf, sizeof(rspBuf.buf),
					node);
	if (rspSize == 0)
	{
		int32 rc;
		uint32 myFsState = 0;
		uint64 myUsedSize = 0;
		uint32 myHighWaterMark = 0;

		// it's possible that this request timed out...
		// let's check the filesystem status first to make sure...
		while (1)
		{
			rc = ea_getFsysInfo(node, cb->vol->id,
					&myFsState, &myUsedSize, &myHighWaterMark);

			if (rc != NFX_OK)
				break;
			else if (myFsState == FS_VOL_STATE_MOUNTING)
			{
				EA_DBG_PRINT(("ea_mntFsys(): still waiting for fsys mounting\n"));
				sleep(15);
				continue;
			}
			else if (myFsState == FS_VOL_STATE_ONLINE)
			{
				EA_DBG_PRINT(("ea_mntFsys(): fsys mount return %d\n", myFsState));
				fsRsp->hdr.fsStatus = NFX_OK;
				goto ea_mntFsys_done;
			}
			else
			{
				fsRsp->hdr.fsStatus = myFsState;
				goto ea_mntFsys_done;
			}
		}

		EA_DBG_PRINT(("ea_mntFsys(): unable to send req to node %s\n", node));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgMountResp))
		fsRsp->hdr.fsStatus = NFX_ERR;

	if (fsRsp->hdr.requestType != FS_REQ_MOUNT)
		fsRsp->hdr.fsStatus = NFX_ERR;

ea_mntFsys_done:
    EA_FCT_PRINT(("< ea_mntFsys()\n"));
    return(fsRsp->hdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_umntFsys()
 *
 * Description  : Unmount a file system.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_umntFsys(char *node, ea_fsysCb_t * cb)
{
	int					rspSize;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
    fs_MsgUnmountReq	*fsReq = (fs_MsgUnmountReq *) reqBuf.buf;
    fs_MsgUnmountResp	*fsRsp = (fs_MsgUnmountResp *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_umntFsys()\n"));

    /* Reset memory for request and response.
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->hdr.requestType = FS_REQ_UNMOUNT;
    fsReq->hdr.len = sizeof(fs_MsgUnmountReq);
    fsReq->hdr.clientPrivate = cb->hdr->userInfo;
    /*
      do I need the following parameters???
      fs_AttrType attrType; // Bitmap of requested attribute types.
      uint32      uid; // User ID for the request
      uint32      gid; // Group ID for the request
      fs_Proto    proto; // Client protocol.
      int32       numGids; // Number of valid entries in gids[].
      int32       pad;
      uint32      gids[FS_REQ_GIDS_MAX];
    */
    memcpy((void *) &(fsReq->volId),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));

    /* Send request to FS.
	 * If card is dead we can go ahead with umount
     */
	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_OK;

	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgUnmountReq),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_umntFsys(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgUnmountResp))
	{
		/*
		 * FS timed out. The card is not there
		 * so we can treat umount request as successful
		 */
		fsRsp->hdr.fsStatus = NFX_OK;
    	return(fsRsp->hdr.fsStatus);
	}

	if (fsRsp->hdr.requestType != FS_REQ_UNMOUNT)
		fsRsp->hdr.fsStatus = NFX_ERR;
    
    EA_FCT_PRINT(("< ea_umntFsys()\n"));
    return(fsRsp->hdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_fsysIsExported()
 *
 * Description  : Find if a file system is being exported
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 
 *
 * Parameters   :
 *   node - specify a node name for sending request to remote cluster node
 *------------------------------------------------------------------------------
 */
int32
ea_fsysIsExported(char *node, evm_volId_t volId)
{
	int					rspSize;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
    mnt_infoReq_t		*mntInfoReq = (mnt_infoReq_t *) reqBuf.buf;
    mnt_infoResp_t		*mntInfoRsp = (mnt_infoResp_t *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_umntFsys()\n"));

    /* Reset memory for request and response.
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));

    mntInfoReq->msgHdr.private = 0;
    memcpy((void *) &(mntInfoReq->volId), (void *) &(volId), 
		sizeof(evm_volId_t));
    mntInfoReq->msgHdr.msgType = MNT_IS_VOL_EXPORTED;

	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(mnt_infoReq_t),
						MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
    {
        EA_ERR_PRINT(("ea_fsysIsExported(): unable to send request!\n"));
        return (NFX_ERR);
    }

    if (mntInfoRsp->msgHdr.status)
		return 1;
	else
		return 0;
}

/*------------------------------------------------------------------------------
 * Action on a file set.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_getFsetIdList()
 *
 * Description  : Get list of all existing file sets
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 21-NOV-2001
 *
 * !!!!NOTE!!!!!
 *   ea_getFsetIdList() will allocate memory (malloc) for fsetList.
 *   It's caller's responsibility to free this memory.
 *------------------------------------------------------------------------------
 */

ea_fsetList_t *
ea_getFsetIdList(char *node)
{
    int							rspSize;
    ea_clusmsg_t                reqBuf, rspBuf, pathReqBuf, pathRspBuf;
	fs_MsgListFileSetsReq_t 	*fsReq = (fs_MsgListFileSetsReq_t *) reqBuf.buf;
	fs_MsgListFileSetsResp_t 	*fsRsp = (fs_MsgListFileSetsResp_t *) rspBuf.buf;
	fs_MsgGetFilesetPathReq_t 	*fsPathReq
						= (fs_MsgGetFilesetPathReq_t *) pathReqBuf.buf;
	fs_MsgGetFilesetPathResp_t 	*fsPathRsp
						= (fs_MsgGetFilesetPathResp_t *) pathRspBuf.buf;

	uint32						i, j;
	uint32						totEntries = 0;
	uint32						numEntries = 0;
	ea_fsysList_t 				*fsysList = NULL;
	ea_fsetList_t				*fsetList = NULL;
	const uint32				count = 10;

    EA_FCT_PRINT(("> ea_getFsetIdList()\n"));

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NULL;

    // Get list of all volumes and ids.
	fsysList = ea_getVolList(node);
	if (fsysList == NULL)
		return NULL;

    /* 
	 * Reset and initialize memory for request and response.
     */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
	fsReq->msgHdr.requestType = FS_REQ_LIST_FILESETS;
	fsReq->msgHdr.len = sizeof(fs_MsgListFileSetsReq_t);
	fsReq->numberEntries = count;

	memset(&pathReqBuf, 0, sizeof(ea_clusmsg_t));
	fsPathReq->msgHdr.requestType = FS_REQ_GET_FILESET_PATH;
	fsPathReq->msgHdr.len = sizeof (fs_MsgGetFilesetPathReq_t);

	totEntries = 0;
	for (i = 0; i < fsysList->nb; i++)
	{
		uint32 index = 0;

		fsReq->fsid = fsysList->fsys[i].fsysId.volId;
		fsPathReq->volId = fsysList->fsys[i].fsysId.volId;

		while (1)
		{
			fsReq->startIndex = index;
		
			rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgListFileSetsReq_t),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
			if (rspSize == 0)
			{
				EA_DBG_PRINT(("ea_getFsetIdList(): unable to send fs request\n"));
				goto exit_ea_getFsetIdList;
			}

			// finished for this filesystem?
			if (fsRsp->numberEntries <= 0)
				break;

			totEntries += fsRsp->numberEntries;
			if (!fsetList)
			{
				uint32 size = sizeof(ea_fsetList_t) +
						((totEntries-1) * sizeof(ea_fsetCfg_t));

				fsetList = (ea_fsetList_t *) malloc(size);
				if (!fsetList)
					return NULL;
				memset(fsetList, 0, size);
			}
			else
			{
				char *tmpp;
				uint32 addedSize, newSize;

				addedSize = fsRsp->numberEntries * sizeof(ea_fsetCfg_t);
				newSize = sizeof(ea_fsetList_t) +
								((totEntries - 1)*sizeof(ea_fsetCfg_t));
				tmpp = (char *) realloc((void *) fsetList, newSize);
				if (!tmpp)
					goto exit_ea_getFsetIdList;

				memset(tmpp + (newSize - addedSize), 0, addedSize);
				fsetList = (ea_fsetList_t *) tmpp;
			}

			for (j = 0; j < fsRsp->numberEntries; j++)
			{
				fsPathReq->snapnum = 0;
				fsPathReq->inum =  fsRsp->filesets[j].rootInode;
				fsPathReq->fsetId = fsRsp->filesets[j].id;

				rspSize = ea_sendClusterAgileMsg(-1,
							&pathReqBuf, sizeof(fs_MsgGetFilesetPathReq_t),
							"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
							(struct timeval *) &waitEfsTo,
							&pathRspBuf, sizeof(pathRspBuf.buf),
							node);
				if (rspSize == 0)
				{
					EA_DBG_PRINT(("ea_getFsetIdList(): unable to send path request\n"));
					goto exit_ea_getFsetIdList;
				}

				fsetList->nb++;
				fsetList->fset[numEntries].fsetId.fsysId = 
									fsysList->fsys[i].fsysId;
				strcpy(fsetList->fset[numEntries].fsetId.path, fsPathRsp->path);
				EA_DBG_PRINT(("pathRsp = %s copied to fsetlist %s\n", 
					fsPathRsp->path, fsetList->fset[numEntries].fsetId.path));
				numEntries++;
			}
			index += count;
		}
	}

exit_ea_getFsetIdList:
	free(fsysList);

    EA_FCT_PRINT(("< ea_getFsetIdList()\n"));
    return fsetList;
}

/*------------------------------------------------------------------------------
 * Name         : ea_fsetGetQuota()
 *
 * Description  : Get quota block for the requested file set
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 21-NOV-2001
 *
 * !!!!NOTE!!!  : ea_fsetGetQuota() will allocate memory for fileset quota
 *                data. It's caller's responsibility to free this memory
 *------------------------------------------------------------------------------
 */

fs_MsgReportQuotaResp_t *
ea_fsetGetQuota(char *node, ea_fsetId_t *fsetId)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgReportQuotaReq_t	*fsQuotaReq = (fs_MsgReportQuotaReq_t *) reqBuf.buf;
    fs_MsgReportQuotaResp_t	*fsQuotaRsp = (fs_MsgReportQuotaResp_t *) rspBuf.buf;
    fs_MsgReportQuotaResp_t	*ret;

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NULL;

	ret = malloc(sizeof(fs_MsgReportQuotaResp_t));
    if (!ret)
		return NULL;
	memset(ret, 0, sizeof(ret));

	bzero((void *) &reqBuf, sizeof(ea_clusmsg_t));

	fsQuotaReq->msgHdr.requestType = FS_REQ_REPORT_QUOTA;
	fsQuotaReq->msgHdr.len = sizeof(fs_MsgReportQuotaReq_t);
	fsQuotaReq->fsid = fsetId->fsysId.volId;
	strcpy(fsQuotaReq->path, fsetId->path);
	fsQuotaReq->flag = FS_QUOTA_AGGREGATE;
	fsQuotaReq->startIndex = 0;
	fsQuotaReq->numberEntries = 1;

	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgReportQuotaReq_t),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_fsetGetQuota(): unable to send request\n"));
		return NULL;
	}

	if ((fsQuotaRsp->msgHdr.fsStatus != NFX_OK)  ||
			(fsQuotaRsp->numberResponses != 1))
	{
		fprintf(stderr, "Cannot retrieve quota - error %d\n",
								fsQuotaRsp->msgHdr.fsStatus);
		return NULL;
	}

	memcpy(ret, fsQuotaRsp, sizeof(fs_MsgReportQuotaResp_t));
	return ret;

}

/*------------------------------------------------------------------------------
 * Name         : ea_crtFset()
 *
 * Description  : Create a file set.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       must be NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_crtFset(char *node, ea_fsetCb_t * cb)
{
	int							rspSize;
    ea_clusmsg_t				reqBuf;
    ea_clusmsg_t				rspBuf;
    fs_MsgCreateFileSetReq_t	*fsReq = (fs_MsgCreateFileSetReq_t *) reqBuf.buf;
    fs_MsgCreateFileSetResp_t	*fsRsp = (fs_MsgCreateFileSetResp_t *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_crtFset()\n"));

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

    /* Reset memory for request and response.
     */
	memset((void *) &reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->msgHdr.requestType = FS_REQ_CREATE_FILESET;
    fsReq->msgHdr.len = sizeof(fs_MsgCreateFileSetReq_t);
    fsReq->msgHdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(fsReq->fsid), (void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    strcpy(fsReq->path, cb->path);;

    /* Send request to FS.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgCreateFileSetReq_t),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_crtFset(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgCreateFileSetResp_t))
		fsRsp->msgHdr.fsStatus = NFX_ERR;

	if (fsRsp->msgHdr.requestType != FS_REQ_CREATE_FILESET)
		fsRsp->msgHdr.fsStatus = NFX_ERR;
    
    EA_FCT_PRINT(("< ea_crtFset()\n"));
    return(fsRsp->msgHdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_delFset()
 *
 * Description  : Delete a file set.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 * Parameters   : node - indicates a node name if this is for remote
 *                       must be null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_delFset(char *node, ea_fsetCb_t * cb)
{
	int							rspSize;
    ea_clusmsg_t				reqBuf;
    ea_clusmsg_t				rspBuf;
    fs_MsgRemoveFileSetReq_t	*fsReq = (fs_MsgRemoveFileSetReq_t *) reqBuf.buf;
    fs_MsgRemoveFileSetResp_t	*fsRsp = (fs_MsgRemoveFileSetResp_t *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_delFset()\n"));

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

    /* Reset memory for request and response.
     */
	memset((void *) &reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->msgHdr.requestType = FS_REQ_REMOVE_FILESET;
    fsReq->msgHdr.len = sizeof(fs_MsgRemoveFileSetResp_t);
    fsReq->msgHdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(fsReq->fsid), (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    strcpy(fsReq->path, cb->path);

    /* Send request to FS.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgRemoveFileSetReq_t),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_delFset(): unable to send request\n"));
		return NFX_ERR;
	}
    
    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgRemoveFileSetResp_t))
		fsRsp->msgHdr.fsStatus = NFX_ERR;

	if (fsRsp->msgHdr.requestType != FS_REQ_REMOVE_FILESET)
		fsRsp->msgHdr.fsStatus = NFX_ERR;
    
    EA_FCT_PRINT(("< ea_delFset()\n"));
    return(fsRsp->msgHdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_setFsetQuota()
 *
 * Description  : Set quota on a file set.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       must be NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_setFsetQuota(char *node, ea_fsetCb_t * cb, ea_quotaCfg_t *quotaCfg)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgSetQuotaReq_t		*fsReq = (fs_MsgSetQuotaReq_t *) reqBuf.buf;
    fs_MsgSetQuotaResp_t	*fsRsp = (fs_MsgSetQuotaResp_t *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_setFsetQuota()\n"));
	
	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

    /* Reset memory for request and response.
     */
	memset((void *) &reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->msgHdr.requestType = FS_REQ_SET_QUOTA;
    fsReq->msgHdr.len = sizeof(fs_MsgSetQuotaReq_t);
    fsReq->msgHdr.clientPrivate = cb->hdr->userInfo;

    memcpy((void *) &(fsReq->fsid),
           (const void *) &(cb->vol->id),
           sizeof(evm_volId_t));
    strcpy(fsReq->path, cb->path);;
	fsReq->numberLimits = 1;
    fsReq->limits[0].flag = FS_QUOTA_AGGREGATE;
	fsReq->limits[0].id = 0;             // no quota on uid/gid
	fsReq->limits[0].inodeSoftLimit = quotaCfg->inodeSoftLimit;
	fsReq->limits[0].inodeHardLimit = quotaCfg->inodeHardLimit; 
	fsReq->limits[0].inodeTimeLimit = quotaCfg->inodeTimeLimit;
    fsReq->limits[0].blockSoftLimit = quotaCfg->blockSoftLimit;
	fsReq->limits[0].blockHardLimit = quotaCfg->blockHardLimit;
	fsReq->limits[0].blockTimeLimit = quotaCfg->blockTimeLimit;
    
    /* Send request to FS.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgSetQuotaReq_t),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_setFsetQuota(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
    EA_DBG_PRINT(("fsRsp->numberResponses = %d\n", fsRsp->numberResponses));

	if (rspSize < sizeof(fs_MsgSetQuotaResp_t))
		fsRsp->msgHdr.fsStatus = NFX_ERR;

	if (fsRsp->msgHdr.requestType != FS_REQ_SET_QUOTA)
		fsRsp->msgHdr.fsStatus = NFX_ERR;

    EA_FCT_PRINT(("< ea_setFsetQuota()\n"));
    return(fsRsp->responses[0].status);
}

/*------------------------------------------------------------------------------
 * Name         : ea_onFsetQuota()
 *
 * Description  : Turn on quota on a file set.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       must be NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_onFsetQuota(char *node, ea_fsetCb_t * cb)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgQuotaOnReq_t		*fsReq = (fs_MsgQuotaOnReq_t *) reqBuf.buf;
    fs_MsgQuotaOnResp_t		*fsRsp = (fs_MsgQuotaOnResp_t *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_onFsetQuota()\n"));

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

    /* Reset memory for request and response.
     */
	memset((void *) &reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->msgHdr.requestType = FS_REQ_QUOTA_ON;
    fsReq->msgHdr.len = sizeof(fs_MsgQuotaOnReq_t);
    fsReq->msgHdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(fsReq->fsid), (const void *) &(cb->vol->id),
			           sizeof(evm_volId_t));
    strcpy(fsReq->path, cb->path);;
    
    /* Send request to FS.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgQuotaOnReq_t),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_onFsetQuota(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgQuotaOnResp_t))
		fsRsp->msgHdr.fsStatus = NFX_ERR;

	if (fsRsp->msgHdr.requestType != FS_REQ_QUOTA_ON)
		fsRsp->msgHdr.fsStatus = NFX_ERR;

    EA_FCT_PRINT(("< ea_onFsetQuota()\n"));
    return(fsRsp->msgHdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_offFsetQuota()
 *
 * Description  : Turn off quota on a file set.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       must be NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */

int32
ea_offFsetQuota(char *node, ea_fsetCb_t * cb)
{
	int						rspSize;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
    fs_MsgQuotaOffReq_t		*fsReq = (fs_MsgQuotaOffReq_t *) reqBuf.buf;
    fs_MsgQuotaOffResp_t	*fsRsp = (fs_MsgQuotaOffResp_t *) rspBuf.buf;

    EA_FCT_PRINT(("> ea_offFsetQuota()\n"));

	if ((checkClusterAgileDest(node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(node, NFX_NFP_FP2, 1) != NFX_OK))
		return NFX_ERR;

    /* Reset memory for request and response.
     */
	memset((void *) &reqBuf, 0, sizeof(ea_clusmsg_t));

    /* Fill request.
     */
    fsReq->msgHdr.requestType = FS_REQ_QUOTA_OFF;
    fsReq->msgHdr.len = sizeof(fs_MsgQuotaOffReq_t);
    fsReq->msgHdr.clientPrivate = cb->hdr->userInfo;
    memcpy((void *) &(fsReq->fsid), (const void *) &(cb->vol->id),
           			sizeof(evm_volId_t));
    strcpy(fsReq->path, cb->path);;
    
    /* Send request to FS.
     */
	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgQuotaOffReq_t),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						(struct timeval *) &waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_offFsetQuota(): unable to send request\n"));
		return NFX_ERR;
	}

    /* Receive response from FS.
     */
	if (rspSize < sizeof(fs_MsgQuotaOffResp_t))
		fsRsp->msgHdr.fsStatus = NFX_ERR;

	if (fsRsp->msgHdr.requestType != FS_REQ_QUOTA_OFF)
		fsRsp->msgHdr.fsStatus = NFX_ERR;

    EA_FCT_PRINT(("< ea_offFsetQuota()\n"));
    return(fsRsp->msgHdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : ea_getFcPortWwn()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 * Parameters   : node - node name if this is for remote node
 *                       must be NULL or null string if local
 *
 *------------------------------------------------------------------------------
 */

uint64
ea_getFcPortWwn(char *node, uint32 slot, uint32 port)
{
	int					rspSize;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
    sdm_ioctlReqMsg_t	*req = (sdm_ioctlReqMsg_t *) reqBuf.buf;
    sdm_ioctlRspMsg_t	*rsp = (sdm_ioctlRspMsg_t *) rspBuf.buf;

	hal_adapter_info_t	*info;

	/*
	 * Get PORT/LINK INFO
	 */
	memset((void *) &reqBuf, 0, sizeof(ea_clusmsg_t));

	req->hdr.mid = SDM_MID_IOCTL_REQ;
	req->code = HAL_IOCTL_COMMON_GET_ADAPTER_INFO;
	req->slot = slot;
	req->port = port;

	rspSize = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(sdm_ioctlReqMsg_t),
						"sdm", NFX_SSC_UNIX, SSC_SLOT, 0,
						(struct timeval *) &waitSdmTo,
						&rspBuf, sizeof(rspBuf.buf),
						node);
	if (rspSize == 0)
	{
		EA_DBG_PRINT(("ea_getFcPortWwn(): unable to send request\n"));
		return -1;
	}

	if ((rsp->hdr.mid != SDM_MID_IOCTL_RSP) || 
		(rsp->eCode != SDM_OK) || 
		(rsp->dataLen < sizeof(hal_adapter_info_t)))
	{
		return -1;
	}

	info = (hal_adapter_info_t *)rsp->data;

	EA_FCT_PRINT(("ea_getFcPortWwn: Slot %d port %d WWN %016llx\n", slot, port, 
				info->u.fc_data.fc_node_name.u.wwn));

	return info->u.fc_data.fc_node_name.u.wwn;
}


/*----------------------------------------------------------------------------*/
