/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : sim-ev-tx.c
 *
 * Description  :
 * Simulates interface of Eagle Agent (EA) with Eagle Vision (EV).
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "ea.h"
#include "ea-api.h"

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

/* The name of this program. */
static char * myName;

/* Wait for EA to respond (10s). */
//static struct timeval rspTo = {10, 0};

/* */
//#define RAND_MAX 100

/*------------------------------------------------------------------------------
 * Prototype of local functions
 *------------------------------------------------------------------------------
 */

static void
print_usage(void);

/*------------------------------------------------------------------------------
 * Name         : print_usage()
 *
 * Description  : Shows the correct usage of this program.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
print_usage(void)

{
	fprintf(stderr,"Usage:\n");

	fprintf(stderr," %s add fsys <name> \n", myName);
	fprintf(stderr," %s add fset <name> <path>\n", myName);
	fprintf(stderr," %s add snap <name> <path>\n", myName);
	fprintf(stderr," %s add export <name> <path> <shareIP> <hostIP>\n", 
										myName);

	fprintf(stderr," %s del fsys <name>\n", myName);
	fprintf(stderr," %s del fset <name> <path>\n", myName);
	fprintf(stderr," %s del snap <name> <path>\n", myName);
	fprintf(stderr," %s del export <name> <path>\n", myName);
	fprintf(stderr," %s del path <name> <path>\n", myName);

	fprintf(stderr," %s get fsys \n", myName);
	fprintf(stderr," %s get fset \n", myName);
	fprintf(stderr," %s get snap <name> <path>\n", myName);
	fprintf(stderr," %s get export <name>\n", myName);
	fprintf(stderr," %s get path <name> <path>\n", myName);

}

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : It all begins here.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int
main(int argc, char *argv[])

{
	char *      actionStr;
	char *      objectStr;
    ea_msg_t    msg;
    uint32      cookie;

    /* Initialize my own name.
     * Discard path if any.
     */
	myName = argv[0];
	if (rindex(myName, '/'))
    {
		strcpy(myName, rindex(myName, '/') + 1);
	}
    
    /* Make sure there's at least the action and the object strings.
     * Get the command and the object on which to execute the command.
     */
	if (argc < 3)
    {
		print_usage();
		exit(NFX_ERR);
	}
	actionStr = argv[1];
	objectStr = argv[2];
    
    /* Initialize the cookie.
     */
    cookie = random();
    printf("cookie = %d\n", cookie);
   
    /* Each "command" has it's own action function.
     */
	if (!strcasecmp(actionStr, "add"))
    {
        if (!strcasecmp(objectStr, "fsys"))
        {
            /* Create file system.
             */
            ea_addFsysReq_t * req = (ea_addFsysReq_t *) &(msg.buf);

            req->hdr.msgType = EA_ADD_FSYS_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
            req->volSize = 102396;
            req->minAutoGrow = 0;
            req->highWaterMark = 0;
			req->attr.num = 1;
            req->nim = 1;    // N-NIM in simulation

            msg.len = sizeof(ea_addFsysReq_t);
        }
        else if (!strcasecmp(objectStr, "fset"))
        {
            /* Create file set.
             */
            ea_addFsetReq_t * req = (ea_addFsetReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_ADD_FSET_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsetId.fsysId.fsysName, argv[3]);
            strcpy(req->fsetId.path, argv[4]);
            req->quotaOn = FALSE;

			req->quotaCfg.inodeSoftLimit = 100;
			req->quotaCfg.blockSoftLimit = 25;

            msg.len = sizeof(ea_addFsetReq_t);
        }
        else if (!strcasecmp(objectStr, "snap"))
        {
            /* Create snap shot.
             */
            ea_crtSnapReq_t * req = (ea_crtSnapReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_CRT_SNAP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
            strcpy(req->snapName, argv[4]);

            msg.len = sizeof(ea_crtSnapReq_t);
        }
        else if (!strcasecmp(objectStr, "export"))
        {
            /* Create export
             */
            ea_addExpReq_t * req = (ea_addExpReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_ADD_EXP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);

            strcpy(req->expCfg.path.pathName, argv[4]);

			req->expCfg.shareIP = inet_addr(argv[5]);

			req->expCfg.num = 0;

			if (argv[6] && strlen(argv[6]) > 0)
			{
				req->expCfg.num = 1;
				req->expCfg.expHostList[0].hostIP = inet_addr(argv[6]);
			}

			req->expCfg.expHostList[0].ro_rw = 2;
			req->expCfg.expHostList[0].rtsquash = 1;

            msg.len = sizeof(ea_addExpReq_t);
        }
		else
        {
			print_usage();
			exit(NFX_ERR);
		}            
	}
	else if(!strcmp(actionStr, "mod"))
	{
        if (!strcasecmp(objectStr, "export"))
        {
            /* Create export
             */
            ea_modExpReq_t * req = (ea_modExpReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_MOD_EXP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);

            strcpy(req->expCfg.path.pathName, argv[4]);

			req->expCfg.shareIP = inet_addr(argv[5]);

			req->expCfg.num = 0;

			if (argv[6] && strlen(argv[6]) > 0)
			{
				req->expCfg.num = 1;
				req->expCfg.expHostList[0].hostIP = inet_addr(argv[6]);
			}

			req->expCfg.expHostList[0].ro_rw = 1;
			req->expCfg.expHostList[0].rtsquash = 2;

            msg.len = sizeof(ea_modExpReq_t);
        }
		else
        {
			print_usage();
			exit(NFX_ERR);
		}            
	}
	else if(!strcmp(actionStr, "del"))
    {
        if (!strcasecmp(objectStr, "fsys"))
        {
            /* Delete file system.
             */
            ea_delFsysReq_t * req = (ea_delFsysReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_DEL_FSYS_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);

            msg.len = sizeof(ea_delFsysReq_t);            
        }
        else if (!strcasecmp(objectStr, "fset"))
        {
            /* Delete file set.
             */
            ea_delFsetReq_t * req = (ea_delFsetReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_DEL_FSET_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsetId.fsysId.fsysName, argv[3]);
            strcpy(req->fsetId.path, argv[4]);

            msg.len = sizeof(ea_delFsetReq_t);
        }
        else if (!strcasecmp(objectStr, "snap"))
        {
            /* Remove snap shot.
             */
            ea_delSnapReq_t * req = (ea_delSnapReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_DEL_SNAP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
            strcpy(req->snapName, argv[4]);

            msg.len = sizeof(ea_delSnapReq_t);
        }
        else if (!strcasecmp(objectStr, "export"))
        {
            /* Delete export
             */
            ea_delExpReq_t * req = (ea_delExpReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_DEL_EXP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
			strcpy(req->expCfg.path.pathName, argv[4]);

			req->expCfg.shareIP = inet_addr(argv[5]);

			if (argv[6] && strlen(argv[6]) > 0)
			{
				req->expCfg.expHostList[0].hostIP = inet_addr(argv[6]);
				req->expCfg.num = 1;
			}

            msg.len = sizeof(ea_delExpReq_t);
        }
        else if (!strcasecmp(objectStr, "path"))
		{
            /* Delete path
             */
            ea_delPathReq_t * req = (ea_delPathReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_DEL_PATH_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
			strcpy(req->pathList[0].pathName, argv[4]);
			req->num = 1;

            msg.len = sizeof(ea_delPathReq_t);
		}
		else
        {
			print_usage();
			exit(NFX_ERR);
		}
	}
	else if(!strcmp(actionStr, "get"))
    {
        if (!strcasecmp(objectStr, "fsys"))
        {
			printf("Get file system TBD\n");
        }
        else if (!strcasecmp(objectStr, "fset"))
        {
            /* Get file set.
             */
            ea_getFsetReq_t * req = (ea_getFsetReq_t *) &(msg.buf);

            req->hdr.msgType = EA_GET_FSET_REQ;
            req->hdr.userInfo = cookie;
			req->reqType = EA_REQ_HAS_FSET_INIT;
			req->num = 0;

            msg.len = sizeof(ea_getFsetReq_t);
        }
        else if (!strcasecmp(objectStr, "snap"))
        {
            /* Create snap shot.
             */
            ea_crtSnapReq_t * req = (ea_crtSnapReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_GET_USG_SNAP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
            strcpy(req->snapName, argv[4]);

            msg.len = sizeof(ea_crtSnapReq_t);
        }
        else if (!strcasecmp(objectStr, "export"))
        {
            /* get export
             */
            ea_getExpReq_t * req = (ea_getExpReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_GET_EXP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);

            msg.len = sizeof(ea_getExpReq_t);
        }
        else if (!strcasecmp(objectStr, "path"))
        {
            /* get path
             */
            ea_getPathExpReq_t * req = (ea_getPathExpReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_GET_PATH_EXP_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
			strcpy(req->path.pathName, argv[4]);

            msg.len = sizeof(ea_getPathExpReq_t);
        }
        else if (!strcasecmp(objectStr, "chassis"))
		{
            /* get path
             */
            ea_getChassisInfoReq_t * req = (ea_getChassisInfoReq_t *) &(msg.buf);
            
            req->hdr.msgType = EA_GET_CHASSIS_INFO_REQ;
            req->hdr.userInfo = cookie;

            msg.len = sizeof(ea_getChassisInfoReq_t);
		}
		else
        {
			print_usage();
			exit(NFX_ERR);
		}
	}
	else
    {
		fprintf(stderr, "Invalid command: %s\n",actionStr);
		print_usage();
		exit(NFX_ERR);
	}

    /* Send message to EA.
     */
    if (sendAgileMsg(-1,
                     &msg.buf, msg.len,
                     "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                     NULL, NULL, 0) != 0)
    {
        perror("sendAgileMsg() failed\n");
        exit(NFX_ERR);
    }
		
	exit(NFX_OK);
}

/*----------------------------------------------------------------------------*/
