/*
 *-----------------------------------------------------------------
 *
 * Name:	eee-fwd.c
 *
 * RCS:  $Id: eee-fwd.c,v 1.85 2002/12/09 21:08:01 sandrineb Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Routines that handle the forward/receive queues
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/18/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <bqueue.h>
#include <exports/eee-api.h>
#include <exports/cm-api.h>
#include <exports/malloc-api.h>
#include <exports/mem-api.h>
#include <exports/timer-api.h>
#if defined(NFP_TXRX)
#include <exports/sbm-api.h>
#endif
#include <exports/backplane-api.h>

#ifdef SSC
#include <exports/anpssc-api.h>
#endif

#include <exports/spinlock-api.h>

#ifdef LINUX_TEST
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#ifdef LINUX_POLL_NONBLOCKING_FD
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
/* rx polling indicator socket fd for this cpu */
int eee_linuxFwdPollRxFd;
/* tx polling indicator socket fd's for destinations */
int eee_linuxFwdPollTxFd[NFX_SLOT_MAX][NFX_MAX_CPUS_PER_SLOT];
/* state of tx polling indicator socket fd's for destinations */
int eee_linuxFwdPollTxFdState[NFX_SLOT_MAX][NFX_MAX_CPUS_PER_SLOT];
#endif
#endif

#if defined(NFX_SMP)
static void eee_processPacket(struct edesc *);
#endif

uint16 eee_packetId;				/* The packet sequence identifier */

uint64 ht_bytes_rx;
nfx_spinlock_t eee_fwdqueue_spin;

#if defined(NFX_SMP)

struct local_queue
{
	nfx_spinlock_t lq_spin;
	struct edesc *lq_head;
	struct edesc *lq_tail;
};

struct local_queue local_queue;


void
eee_receivePacket(struct edesc *desc)
{
	NFX_SPIN_LOCK(&local_queue.lq_spin);
	if (local_queue.lq_tail) {
		local_queue.lq_tail->hdr.next = desc;
	}
	else {
		local_queue.lq_head = desc;
	}
	local_queue.lq_tail = desc;
	NFX_SPIN_UNLOCK(&local_queue.lq_spin);
}

struct edesc *
eee_dequeue_packet(struct local_queue *queue)
{
	struct edesc *desc;
	NFX_SPIN_LOCK(&queue->lq_spin);
	desc = queue->lq_head;
	if (desc) {
		if (!desc->hdr.next) {
			queue->lq_tail = NULL;
		}
		queue->lq_head = desc->hdr.next;
		desc->hdr.next = NULL;
	}
	NFX_SPIN_UNLOCK(&queue->lq_spin);
	return desc;
}


void
eee_poll_local_queue(void *cb, uint32 tref)
{
	struct local_queue *queue = cb;
	struct edesc *desc;

#if defined(NFP_TXRX)
	/* txrx acpu does not poll local_queue */
	if (!CM_IS_TXRX_NCPU)
		return;
#endif /* NFP_TXRX */
//	printf("local queue polled\n");
	while (--tref && (desc = eee_dequeue_packet(queue))) {
		eee_processPacket(desc);
	}
}

#if defined(NFP_TXRX)
struct local_queue local_queueA; /* local queue from Txrx ACPU */

void
eee_receivePacketA(struct edesc *desc)
{
	//printf("give message to acpu, dp %x, sp %x\n",
	//	   desc->hdr.dest_port, desc->hdr.src_port);
	
	NFX_SPIN_LOCK(&local_queueA.lq_spin);
	if (local_queueA.lq_tail) {
		local_queueA.lq_tail->hdr.next = desc;
	}
	else {
		local_queueA.lq_head = desc;
	}
	local_queueA.lq_tail = desc;
	NFX_SPIN_UNLOCK(&local_queueA.lq_spin);
}

void
eee_poll_local_queueA(void *cb, uint32 tref)
{
	struct local_queue *queue = cb;
	struct edesc *desc;

	/* txrx ncpu does not poll local_queue */
	if (CM_IS_TXRX_NCPU)
		return;
	
	while (--tref && (desc = eee_dequeue_packet(queue))) {
		eee_processPacket(desc);
	}
}

/*-----------------------------------------------------------------
 * Name :	eee_txrxFilterToAcpu
 *
 * Description: create the filter that determines whether to fwd pkt
 *				to txrx Acpu from ncpu
 *
 *
 *
 * Created by: Su Wang
 *
 * Date Created:	12/17/02
 * 
 *	 
 *-----------------------------------------------------------------
 */
#define TXRX_MANAGEMENT_FILTER	0 /* manage plane filter  */
#define TXRX_FWDQ_FILTER		1 /* fwd queue filter */
#define TXRX_NET_FILTER			2 /* network traffic filter */

char	txrx_filters[3][256];		
static void
eee_txrxMapAcpu()
{
	int i;
	
	bzero(txrx_filters, sizeof(txrx_filters));
	
	/* initilize management plane filter */
	for (i=0; i < 256; i++) {
		if (i == eee_getApplicationId("cifsauth")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}
		
		if (i == eee_getApplicationId("cifs_srvr")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}

		if (i == eee_getApplicationId("dm")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}

		if (i == eee_getApplicationId("mountd")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}

		if (i == eee_getApplicationId("mnt_srvr")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}

		if (i == eee_getApplicationId("nfs_cmd")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}

		if (i == eee_getApplicationId("nfs_srvr")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}

		if (i == eee_getApplicationId("nlm_srvr")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}

		if (i == eee_getApplicationId("nsm_srvr")) {
			txrx_filters[TXRX_MANAGEMENT_FILTER][i] = 1;
			txrx_filters[TXRX_FWDQ_FILTER][i] = 1;
			continue;
		}
	}
	
	return;
}

/*-----------------------------------------------------------------
 * Name :	eee_txrxIsPktToAcpu
 *
 * Description: SMP txrx, it determines whether a recived pkt by 
 *				NCPU txrx should be forwarded to ACPU txrx.
 *				edescp is given to acpu if
 *				1. src_cpu is fp and dst_cpu is txrx 
 *				2. if src port is dp83820 and selected appId
 *
 *
 *
 * Created by: Su Wang
 *
 * Date Created:	12/17/02
 * 
 *	 
 *-----------------------------------------------------------------
 */
static int32	
eee_txrxIsPktToAcpu(eee_descPtr_t   edesc)
{
	uint32	dp,sp, appId;
	
	dp = edesc->hdr.dest_port;
	sp = edesc->hdr.src_port;


	appId = EEE_GET_APP_ID(dp);
	
	/* msg from fp to txrx */
	if ((EEE_GET_SLOT_NUM(sp) == nfxMySlotId) && 
		( EEE_GET_SLOT_NUM(dp) == nfxMySlotId)) {

		if (txrx_filters[TXRX_FWDQ_FILTER][appId])
			return 1;
	}
	else if (EEE_GET_SLOT_NUM(sp) != nfxMySlotId) {

		/* management msg */
		if (txrx_filters[TXRX_MANAGEMENT_FILTER][appId])
			return 1;
	}
	
	return 0;
}
#endif /* NFP_TXRX */
#endif /* NFX_SMP */

/*-----------------------------------------------------------------
 * Name :	eee_remainBufLen()
 *
 * Description: return the number of bytes remaining available in the buffer
 * starting from memptr.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
int
eee_remainBufLen(eee_descBuf_t *bufdesc, uchar8 *memptr)
{
	/* Check that the pointer is inside the buffer */
	NFX_ASSERT(bufdesc->buf <= memptr && memptr < bufdesc->buf + bufdesc->len);
	return bufdesc->len - (memptr - bufdesc->buf);
}


#if defined(NETEEE_FRAGMENT)

typedef struct fragment_queue
{
	eee_descPtr_t desc;			/* The packet descriptor */
	eee_descPtr_t last_desc;	/* The last descriptor in chain */
	int timeout;				/* Decrement on timeout pass, delete
                                   the queue entry when goes to zero */
	LIST_ENTRY(fragment_queue) q_list;
} fragment_queue_t;

static LIST_HEAD(, fragment_queue) fragment_queue_head;
#if defined(NFX_SMP)
static nfx_spinlock_t fragment_spin;
#endif


/*-------------------------------------------------------------------
 * Name :	matchFragmentQueue()
 *
 * Description: return true if the packet fragment matches the fragment
 * queue
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-------------------------------------------------------------------
 */
static int
matchFragmentQueue(fragment_queue_t *q, eee_descPtr_t edesc)
{
	return (q->desc->hdr.dest_port == edesc->hdr.dest_port
			&& q->desc->hdr.src_port == edesc->hdr.src_port
			&& edesc_seqId(q->desc) == edesc_seqId(edesc));
}


/*-----------------------------------------------------------------
 * Name :	findFragmentQueue()
 *
 * Description: find a fragment queue containing the packet for the fragment
 * specified by the argument edesc.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static fragment_queue_t *
findFragmentQueue(eee_descPtr_t edesc)
{
	fragment_queue_t *q;

	for (q = LIST_FIRST(&fragment_queue_head); q != NULL;
		 q = LIST_NEXT(q, q_list)) {
		if (matchFragmentQueue(q, edesc)) {
			return q;
		}
	}
	return NULL;
}


/*-----------------------------------------------------------------
 * Name :	addFragmentQueue()
 *
 * Description: add new fragment queue to the list of queues
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static void
addFragmentQueue(fragment_queue_t *q)
{
	LIST_INSERT_HEAD(&fragment_queue_head, q, q_list);
}


/*-----------------------------------------------------------------
 * Name :	deleteFragmentQueue()
 *
 * Description: delete the queue from the list of fragment queues and
 * free it.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static void
deleteFragmentQueue(fragment_queue_t *q)
{
	LIST_REMOVE(q, q_list);
	eee_ramDealloc(q);
}


/*-----------------------------------------------------------------
 * Name :	dropFragmentQueue()
 *
 * Description: delete the fragments in the queue and remove the queue
 * from the list
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static void
dropFragmentQueue(fragment_queue_t *q)
{
	eee_freePacket(q->desc);
	deleteFragmentQueue(q);
}


/*-----------------------------------------------------------------
 * Name :	fragmentTimeout()
 *
 * Description: decrement the timeout and delete timed out fragment queues
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static int32 
fragmentTimeout(
	uint32	tmr_handle,
	void	* usr_p1,
	void	* usr_p2,
	void	* usr_p3,
	void	* usr_p4
)
{
	fragment_queue_t *nq, *q;
	for (q = LIST_FIRST(&fragment_queue_head); q != NULL; q = nq) {
		nq = LIST_NEXT(q, q_list);
		if (--q->timeout == 0) {
			printf("fragmentTimeout: drop queue\n");
			dropFragmentQueue(q);
		}
	}
	return NFX_OK;
}


/*-----------------------------------------------------------------
 * Name :	initFragmentQueue()
 *
 * Description: initialize fragment queue with the start fragment
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static void
initFragmentQueue(fragment_queue_t *q, eee_descPtr_t edesc)
{
	q->desc = edesc;
	q->last_desc = edesc;
	q->timeout = 2;
}


/*-----------------------------------------------------------------
 * Name :	allocFragmentQueue()
 *
 * Description: allocate and initialize new fragment queue
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static fragment_queue_t *
allocFragmentQueue(eee_descPtr_t edesc)
{
	fragment_queue_t *q = (fragment_queue_t *)
		eee_ramAlloc(sizeof(fragment_queue_t));
	if (q) {
		initFragmentQueue(q, edesc);
	}

	return q;
}


/*-----------------------------------------------------------------
 * Name :	fragmentOffsetMatch()
 *
 * Description: Returns true if the fragment offset matches the current
 * fragment queue contents.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static int
fragmentOffsetMatch(fragment_queue_t *q, eee_descPtr_t desc)
{
	return EEE_GET_PKT_LEN(q->desc) == edesc_fragmentOffset(desc);
}


/*-----------------------------------------------------------------
 * Name :	addFragmentToQueue()
 *
 * Description: add fragment to the fragment queue
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static int
addFragmentToQueue(eee_descPtr_t desc, fragment_queue_t *q)
{
	int naddBufs = EEE_GET_NUM_BUFS(desc);
	int startBuf = EEE_GET_NUM_BUFS(q->last_desc);
	int endBuf; 
	int newBufIdx, bufIdx;

	NFX_ASSERT(fragmentOffsetMatch(q, desc));
	if (!(q->last_desc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER))
		++startBuf;
	endBuf = startBuf + naddBufs;
	newBufIdx = (desc->hdr.control
				 & EDESC_COMMAND_USE_FIRST_BUFFER) ? 0 : 1;
	for (bufIdx = startBuf; bufIdx < endBuf; bufIdx++, newBufIdx++) {
		if (bufIdx == EEE_MAX_BUFS) {
			eee_descPtr_t newDesc = eee_allocateDesc(EEE_SHARED_POOL,
													 NFX_COS_HIGH);
			if (newDesc == NULL) {
				return FALSE;
			}
			else {
				EEE_PRINTF(("addFragmentToQueue: chain new desc\n"));
				newDesc->hdr.control = 0;
				newDesc->bd[0].control = 0;
				eee_chainDesc(q->last_desc, newDesc);
				q->last_desc = newDesc;
				bufIdx = 1;
				endBuf = endBuf - EEE_MAX_BUFS + 1;
			}
		}
		EEE_SET_NUM_BUFS(q->last_desc, EEE_GET_NUM_BUFS(q->last_desc) + 1);
		q->last_desc->bd[bufIdx] = desc->bd[newBufIdx];
	}

	EEE_SET_PKT_LEN(q->desc, EEE_GET_PKT_LEN(q->desc)
					+ EEE_GET_PKT_LEN(desc));
	
	/* The new descriptor does not own any buffers now, so no need to
	   call eee_freePacket */
	eee_deallocateDesc(desc);

	return TRUE;
}


/*-----------------------------------------------------------------
 * Name :	eee_reassemble()
 *
 * Description: Attempt to reassemble the packet. Return NULL if the
 * fragment was queued for later reassembly, or the complete packet if
 * the reassembly was complete.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static eee_descPtr_t
eee_reassemble(eee_descPtr_t edesc)
{
	uint32 offset = edesc_fragmentOffset(edesc);
	fragment_queue_t *queue;
	
	NFX_ASSERT(offset != 0 || edesc_moreFragments(edesc));

	EEE_PRINTF(("eee_reassemble: off=%d more=%s\n",
			offset, edesc_moreFragments(edesc) ? "yes" : "no"));

	NFX_SPIN_LOCK(&fragment_spin);
	if (offset == 0) {
		if ((queue = findFragmentQueue(edesc)) == NULL
			&& (queue = allocFragmentQueue(edesc)) != NULL) {
			addFragmentQueue(queue);
			NFX_SPIN_UNLOCK(&fragment_spin);
			return NULL;
		}
		else {
			if (queue != NULL) printf("duplicate queue\n");
			if (queue == NULL) printf("can't allocate queue\n");
		}
	}
	else {
		if ((queue = findFragmentQueue(edesc)) != NULL) {
			if (fragmentOffsetMatch(queue, edesc)) {
				int more_fragments = edesc_moreFragments(edesc);
				if (addFragmentToQueue(edesc, queue)) {
					if (!more_fragments) {
						/* Reassembly complete, delete the queue */
						eee_descPtr_t edesc = queue->desc;
						deleteFragmentQueue(queue);
						NFX_SPIN_UNLOCK(&fragment_spin);
						return edesc;
					}
					else {
						/* more fragments follow */
						NFX_SPIN_UNLOCK(&fragment_spin);
						return NULL;
					}
				}
				else {
					printf("eee_reassemble: can't add packet to queue, "
						   "dropping\n");
				}
			}
			else {
				printf("eee_reassemble: non-matching off=%d queue-len=%d, "
					   "dropping\n", edesc_fragmentOffset(edesc),
					   EEE_GET_PKT_LEN(queue->desc));
				deleteFragmentQueue(queue);
			}
		}
	}

	NFX_SPIN_UNLOCK(&fragment_spin);
	eee_freePacket(edesc);
	return NULL;
}


/*-----------------------------------------------------------------
 * Name :	eee_aliasData()
 *
 * Description: Make descriptor specified by the argument to reference
 * the same data of len bytes as descriptor specified by the argument
 * from starting from offset in descriptor "from". The "to" descriptor
 * will become the owner of the buffers completely contained in the
 * new descriptor.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static void
eee_aliasData(eee_descPtr_t from, int offset, eee_descPtr_t to, int len)
{
	int aliasLen = 0;
	int fromBuf;
	uchar8 *fromMem;
	int toBuf = (to->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ? 0 : 1;
	int nbufs = 0;

	NFX_ASSERT(EEE_GET_PKT_LEN(from) - offset >= len);

	EEE_PRINTF(("eee_aliasData: off=%d pktlen=%d nbufs=%d len=%d\n", offset,
			   EEE_GET_PKT_LEN(from), EEE_GET_NUM_BUFS(from), len));
	eee_skipBufsFromStart(&from, offset, &fromBuf, &fromMem);
	EEE_PRINTF(("eee_aliasData: nbufs=%d frombuf=%d fromMem=%x\n",
			   EEE_GET_NUM_BUFS(from), fromBuf, fromMem));
	while (aliasLen < len) {
		int fromBufLen = eee_remainBufLen(&from->bd[fromBuf], fromMem);
		int thisAliasLen = MIN(fromBufLen, len - aliasLen);
		to->bd[toBuf].len = thisAliasLen;
		to->bd[toBuf].buf = fromMem;
		to->bd[toBuf].control = EDESC_BD_CMD_IN_USE |
			(from->bd[fromBuf].control & EDESC_BD_BUFSIZE_MASK);
		if (fromBufLen > len - aliasLen) {
			to->bd[toBuf].control |= EDESC_BD_CMD_NO_DEALLOC;
		}
		else {
			from->bd[fromBuf].control |= EDESC_BD_CMD_NO_DEALLOC;
		}
		if (fromMem != from->bd[fromBuf].buf) {
			/* If the buffer points does not point to the beginning of the
			   original buffer, the drivers can not use the area before
			   the buffer pointer to prepend headers */
			to->bd[toBuf].control |= EDESC_BD_CMD_START_IN_USE;
		}
 		toBuf++;
 		fromBuf++;
		if (edesc_firstBufferIdx(from) +
			EEE_GET_NUM_BUFS(from) == fromBuf) {
			from = eee_nextDesc(from);
			if (from != NULL)
				fromBuf = edesc_firstBufferIdx(from);
		}
	 	nbufs++;
		if (from != NULL)
			fromMem = from->bd[fromBuf].buf;
 		aliasLen += thisAliasLen;
	} 

	EEE_SET_NUM_BUFS(to, nbufs);
	EEE_SET_PKT_LEN(to, len);
}


int
eee_nextPacketId()
{
	int id;
	NFX_SPIN_LOCK(&fragment_spin);
	id = eee_packetId;
	++eee_packetId;
	if (eee_packetId > AGILEHDR_ID_MAX)
		eee_packetId = 0;
	NFX_SPIN_UNLOCK(&fragment_spin);
	return id;
}

			
/*-----------------------------------------------------------------
 * Name :	eee_convertPacket()
 *
 * Description: fragment the packet to be output with the given mtu.
 * On txrx, copy the data from the original descriptor if sending
 * through management plane. Otherwise alias the original data into
 * new descriptors to avoid copying.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static eee_descPtr_t
eee_convertPacket(eee_descPtr_t edesc, int mtu)
{
	int hdrlen = sizeof(agile_hdr_t);
	int nextlen;
	int pktlen = EEE_GET_PKT_LEN(edesc);
	eee_descPtr_t newDesc;
	int rc = NFX_OK;
	int offset;
	int moreFragments;
	eee_descPtr_t firstDesc = NULL;
	eee_descPtr_t lastDesc = NULL;
	int alias_data = 1;
	int packetId;

	EEE_PRINTF(("eee_convertPacket: pktlen=%d mtu=%d dst=%x src=%x\n",
			   pktlen, mtu, edesc->hdr.dest_port, edesc->hdr.src_port));

#if defined(NFP_TXRX)
	if (EEE_GET_DTYPE(edesc->hdr.dest_port) == EEE_MGMT) {
		eee_descPtr_t desc = edesc;
		eee_descBuf_t *bd = edesc_firstBuf(desc);
		while (bd != 0 && alias_data) {
			if ((address_t)bd->buf < 0x2000000000 ||
				(address_t)bd->buf >= 0x2000700000) {
#if defined(DEBUG_TXRX_MGMT)
				printf("Have to convert to mgmt buf: 0x%llx from %llx\n",
					   bd->buf, eee_getBufferAllocAddr(bd->buf));
#endif				
				alias_data = 0;
			}
			bd = eee_nextBuf(&desc, bd);
		}
		if (alias_data && EEE_GET_PKT_LEN(edesc) <= mtu) {
			/* The packet is OK as is */
			return edesc;
		}
	}
#endif

	offset = 0;

	packetId = eee_nextPacketId();

	while (rc == NFX_OK && offset < pktlen) {
		nextlen = pktlen - offset;
		if (nextlen + hdrlen > mtu) {
			/* Fragments will be at least 64 bytes long */
			if (nextlen + hdrlen < mtu + 63)
				nextlen = nextlen / 2;
			else
				nextlen = mtu - hdrlen;
			nextlen &= ~63;
			moreFragments = AGILEHDR_MF;
		}
		else {
			moreFragments = 0;
		}
		
		if ((newDesc = eee_allocateDesc(EEE_SHARED_POOL,
										NFX_COS_HIGH)) != NULL) {
			eee_descHdr_t *hdr = &newDesc->hdr;

			hdr->src_port = edesc->hdr.src_port;
			hdr->dest_port = edesc->hdr.dest_port;
			hdr->offset = offset | moreFragments |
				(packetId << AGILEHDR_ID_SHIFT);
			
			if (alias_data) {
				hdr->control = 0;
				eee_aliasData(edesc, offset, newDesc, nextlen);
			}
#if defined(NFP_TXRX)
			else {
				uint8 *newBuf = eee_allocateBuffer(eee.my_pool_id,
												   EEE_BUF_MGMT,
												   NFX_COS_HIGH);
				if (newBuf) {
					eee_descBuf_t *bd = &newDesc->bd[1];
					hdr->control = nextlen | (1 << EDESC_BUFCNT_SHIFT);
					bd->len = nextlen;
					bd->buf = newBuf;
					bd->control = EDESC_BD_CMD_IN_USE | EDESC_BD_LARGE;
					eee_copyPacketData(newBuf, edesc, offset, nextlen, 0);
					eee_stats->txrx_mgmt_copied += nextlen;
				}
				else {
					rc = NFX_ERR;
				}
			}
#endif		

			if (lastDesc == NULL) {
				lastDesc = newDesc;
				firstDesc = newDesc;
			}
			else {
				eee_chainDesc(lastDesc, newDesc);
				lastDesc = newDesc;
			}
		}
		else {
			rc = NFX_ERR;
		}
		offset += nextlen;
	}

	if (rc == NFX_OK) {
		if (alias_data) {
			while (edesc) {
				eee_descPtr_t next = eee_nextDesc(edesc);
				eee_unlinkNextDesc(edesc);
				eee_deallocateDesc(edesc);
				edesc = next;
			}
		}
		else {
			eee_freePacket(edesc);
		}

		return firstDesc;
	}
	else {
		if (firstDesc != NULL) {
			eee_freePacket(firstDesc);
		}
		return NULL;
	}
}

#else

static eee_descPtr_t
eee_reassemble(eee_descPtr_t desc)
{
	return desc;
}

static int32 
fragmentTimeout(
	uint32	tmr_handle,
	void	* usr_p1,
	void	* usr_p2,
	void	* usr_p3,
	void	* usr_p4
)
{
	return NFX_OK;
}

#endif


/*-----------------------------------------------------------------
 * Name :	eee_skipBufs()
 *
 * Description: Skip len bytes from the current position in the
 * descriptor.  The position is specified by bufptr and memptr, which
 * are the current buffer index and the current pointer into the
 * descriptor.  On return the locations specified by the arguments
 * descptr, bufptr and memptr will be updated to contain new
 * descriptor position.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/23/02
 *
 *----------------------------------------------------------------- */
void
eee_skipBufs(eee_descPtr_t *descptr, int len, int *bufptr, uchar8 **memptr)
{
	int skipLen = 0;
	eee_descPtr_t desc = *descptr;
	int buf = *bufptr;
	uchar8 *mem = *memptr;
	int bufRemainLen;

	while (skipLen + (bufRemainLen = eee_remainBufLen(&desc->bd[buf], mem)) <= len) {
		skipLen += bufRemainLen;
		++buf;
		if (buf == edesc_firstBufferIdx(desc) + EEE_GET_NUM_BUFS(desc)) {
			desc = eee_nextDesc(desc);
			/* Check if tried to skip too much */
			NFX_ASSERT(desc != NULL);
			buf = edesc_firstBufferIdx(desc);
		}
		mem = desc->bd[buf].buf;
	}
	*memptr = desc->bd[buf].buf + (len - skipLen);
	*bufptr = buf;
	*descptr = desc;
}


/*-----------------------------------------------------------------
 * Name :	eee_skipBufsFromStart()
 *
 * Description: Skip len bytes from the beginning of descriptor. The
 * number of buffer containing len + 1 byte will be stored into the
 * location specified by the argument buf, the pointer into the
 * buffer will be stored into the location specified by the argument
 * memptr.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
void
eee_skipBufsFromStart(eee_descPtr_t *descptr, int len, int *bufptr, uchar8 **memptr)
{
	*bufptr = edesc_firstBufferIdx(*descptr);
	*memptr = (*descptr)->bd[*bufptr].buf;
	eee_skipBufs(descptr, len, bufptr, memptr);
}


/*-----------------------------------------------------------------
 * Name:	eee_processReceiveQueue()
 *
 * Description:	polling function that reads entries from the receive queue
 *				and acts on the entries.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/18/01
 *
 *-----------------------------------------------------------------
 */
void
eee_processReceiveQueue(void *cb, uint32 tref)
{
	struct eee_rcv_queue *queue = cb;
	address_t entry;
	
	if (!NFX_SPIN_TRYLOCK(&queue->rcv_spin)) {
		return;
	}

	if ((entry = (address_t)(*queue->rcv_head)) & EEE_FWD_OWN) {
		int cnt = 0;
		int max = queue->rcv_max;
		eee_descPtr_t *rcv_end = queue->rcv_end;
		do {
			++cnt;
			switch (entry & EEE_FWD_WHAT_MASK) {
			case EEE_FWD_ACTION_FWD:
			{
				eee_descPtr_t edesc =
					(eee_descPtr_t)(entry & EEE_FWD_ADDR_MASK);
				queue->rcv_pkts_forwarded++;
#if defined(NFP_TXRX) || defined(NFP_FP)
				if (eee_isRemoteAddr((address_t)edesc)) {
					eee_descPtr_t new_edesc =
						EEE_ALLOCATE_DESC(EEE_SHARED_POOL, NFX_COS_HIGH);
					ht_memcpy_ass(new_edesc,
								  (void *)uncached_to_cached((address_t)edesc),
								  sizeof(eee_desc_t) / CACHELINE_SIZE);
					eee_deallocateRemoteDesc(edesc);
					edesc = new_edesc;
				}
#endif
				ht_bytes_rx += sizeof(edesc) + EEE_GET_PKT_LEN(edesc);
				eee_receivePacket(edesc);
				break;
			}
			case EEE_FREE_BUF:
				eee_deallocateSharedBuffer(
					(void *) (entry & EEE_FWD_ADDR_MASK),
					(entry & EEE_FWD_BUF_SIZE_MASK) >> EEE_FWD_BUF_SIZE_SHIFT);
				queue->rcv_free_buf++;
				break;
			case EEE_FREE_DESC:
				eee_deallocateSharedDesc((void *)
										 (entry & EEE_FWD_ADDR_MASK));
				queue->rcv_free_desc++;
				break;
#if defined(NFP_TXRX)				
			case EEE_FWD_ACTION_FREE_COD:
				sbm_freeCOD((nfx_cod_t)(entry & EEE_FWD_ADDR_MASK), 1);
				queue->rcv_free_pkt++;
				break;
#endif				
			default:
				printf("Invalid rcv queue entry, %lx\n", entry);
				break;
			}

			*(queue->rcv_head) = 0;
			*(queue->rcv_tail_ptr) = queue->rcv_head;
			QUEUE_PTR_INC(	queue->rcv_head,
							rcv_end,
							queue->rcv_start);
		} while (((entry = (address_t)(*queue->rcv_head)) & EEE_FWD_OWN)
			&& cnt != max);
		queue->rcv_cnt += cnt;
		if (cnt > queue->rcv_hwm)
			queue->rcv_hwm = cnt;
	}

	NFX_SPIN_UNLOCK(&queue->rcv_spin);
}


/*-----------------------------------------------------------------
 * Name:	eee_deliverPacketToApp()
 *
 * Description: This function should deliver the packet to a
 *				registered application receive function on the current CPU.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/18/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NfX_OK
							 * EEE_ERR_INVALID_APP_ID
							 */
eee_deliverPacketToApp(
	eee_descPtr_t	edesc,
	int			app_id
)
{
#ifdef SSC
	
	uchar8	app[80];
	uint32  srcPort;

	/*
	printf("eee_deliverPackToApp(%p, 0x%x)\n", edesc, app_id);
	*/

	srcPort = edesc->hdr.src_port;
/*
	We go through the ANPSSC stuff if it's destined for the Unix CPU and
	its not addressed to a e-layer application.
 */
	if ((EEE_GET_DEST_CPU(edesc->hdr.dest_port) == NFX_SSC_UNIX) &&
	    ((app_id >= EEE_MAX_APP_CLIENTS) || !eee.rcv.uni[app_id]))
	{
		uchar8	* a;
		uchar8	* bufptr;
		uint32	pkt_len;
		extern int32	ssc_anp_fd;
		a = eee_getAppNameFromId(app_id);
		if (a) {
			memcpy(app, a, 80);
		}
		else {
			sprintf (app, "Agile%d", app_id);
		}

/*
printf("SSC : app_id (%d), sending to app string \"%s\"\n", app_id, app);
*/

		pkt_len = EEE_GET_PKT_LEN(edesc);
		bufptr = (uchar8 *)malloc(pkt_len+sizeof(ans_ssc_msghdr_t));

		((ans_ssc_msghdr_t *)bufptr)->appid =
										EEE_GET_APP_ID(edesc->hdr.src_port);
		((ans_ssc_msghdr_t *)bufptr)->slot =
										EEE_GET_SLOT_NUM(edesc->hdr.src_port);
		((ans_ssc_msghdr_t *)bufptr)->cpu =
										EEE_GET_DEST_CPU(edesc->hdr.src_port);
		((ans_ssc_msghdr_t *)bufptr)->plane =
										EEE_GET_DTYPE(edesc->hdr.src_port);


		eee_copyPacketToLocalBuffer(bufptr+sizeof(ans_ssc_msghdr_t),
														edesc, pkt_len, 0);

		if (sendAgileMsg(ssc_anp_fd,
							bufptr,
							pkt_len + sizeof(ans_ssc_msghdr_t),
							app,
							0,
							0,
							0,
							srcPort,
							0,
							0) < 0)
			printf("error sending ANP SSC sendAgileMsg\n");
		else 
			printf("ANP SSC message sent succesfully\n");
		eee_freePacket(edesc);
		return NFX_OK;
	}

#endif

	if (app_id >= EEE_MAX_APP_CLIENTS) {
		ATOMIC_INC_INT(&eee_stats->invalid_app_id, 1);
//		panic("Invalid application id, %d\n", app_id);
		return EEE_ERR_INVALID_APP_ID;
	}

	if (eee.rcv.uni[app_id]) {
		ATOMIC_INC_INT(&eee_stats->app_rcv_pkts[app_id], 1);
#ifdef EEE_ATTR
		eee_attrAssignPkt(edesc, app_id);
#endif
		(*eee.rcv.uni[app_id])(edesc);
	}
	else {
		ATOMIC_INC_INT(&eee_stats->invalid_app_id, 1);

/*
		printf("ERROR: deliverPacketToApp\n");
*/
		return EEE_ERR_INVALID_APP_ID;
	}

	return NFX_OK;

}	/* eee_deliverPacketToApp() */


/*-----------------------------------------------------------------
 * Name:	eee_receivePacket()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	11/10/01
 *
 *-----------------------------------------------------------------
 */
#if !defined(NFX_SMP)
void
eee_receivePacket(
	eee_descPtr_t	edesc
)
#else
static void
eee_processPacket(
	eee_descPtr_t   edesc
	)
#endif
{
	uint32	dp;
	uint32	app_id;

	dp = edesc->hdr.dest_port;

#if defined(NFX_SMP) && defined(NFP_TXRX)
	/* acpu, deliver to application directly */
	if (!(nfxMyCPUId == NFX_NFP_TXRX1))
		goto deliver_pkt;
#endif /* NFX_SMP, NFP_TXRX */

#ifdef NFP_TXRX
		if (EEE_GET_DTYPE(dp) == EEE_MGMT) {
			if (EEE_GET_SLOT_NUM(dp) != nfxMySlotId) {
//				printf("Got mgmt packet for slot %d\n", EEE_GET_SLOT_NUM(dp));
				EEE_SET_PORT_ID(dp, PORT_BEST_FIT);
				edesc->hdr.dest_port = dp;
				eee_forwardPacket(edesc);
				return;
			}
			else if ((EEE_GET_DEST_CPU(dp) == NFX_NFP_FP1) || 
					 (EEE_GET_DEST_CPU(dp) == NFX_NFP_FP2) ||
					(EEE_GET_DEST_CPU(dp) == NFX_NFP_FP_ANY)) {
//				printf("Got mgmt packet for cpu %d\n", EEE_GET_DEST_CPU(dp));
				EEE_SET_PORT_ID(dp, PORT_BEST_FIT);
				edesc->hdr.dest_port = dp;
				eee_forwardPacket(edesc);
				return;
			}
		}
#endif

#if defined(NFX_SMP) && defined(NFP_TXRX)
		/* is this message to be processed on ncpu or give to
		 * acpu. 
		 * edescp is given to acpu if
		 * 1. src_cpu is fp and dst_cpu is txrx 
		 * 2. if src port is dp83820 and selected appId
		 */
		if (eee_txrxIsPktToAcpu(edesc)) {
			eee_receivePacketA(edesc);
			return;
		}

deliver_pkt:
#endif /* NFX_SMP, NFP_TXRX */

	if (edesc_fragmentOffset(edesc) != 0 || edesc_moreFragments(edesc)) {
		edesc = eee_reassemble(edesc);
	}
	if (edesc != NULL) {
		/* Reassembly successful, or no reassembly was necessary */
		app_id = EEE_GET_APP_ID(dp);
		if (eee_deliverPacketToApp(edesc, app_id) != NFX_OK) {
			eee_freePacket(edesc);
		}
	}

}	/* eee_receivePacket() */


/*-----------------------------------------------------------------
 * Name :	eee_findPort()
 *
 * Description: find the driver port to send a packet to the destination
 * specified by the argument dp. 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static int
eee_findPort(uint32 port, uint32 dp, eee_port_t **drv_port)
{
	int rc = NFX_OK;
	if (port == PORT_BEST_FIT) {
		*drv_port = &eee.drv.dtype_default[EEE_GET_DTYPE(dp)];
	}
	else {
		if (dp & EDESC2_MCAST_IND) {
			panic("Multicast transmit\n");
			*drv_port = &eee.drv.multi[port];
		}
		else {
			*drv_port = &eee.drv.uni[port];
		}
	}
	
	if ((*drv_port)->func == NULL) {
		printf("no transmit function for port %d, dp = 0x%x\n", port, dp);
		rc = EEE_DROP_PACKET;
	}

	return rc;
}


/*-----------------------------------------------------------------
 * Name :	eee_doTransmit()
 *
 * Description: transmit the packet and update the statistics.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
static int
eee_doTransmit(eee_port_t *drv_port, uint32 dp, eee_descPtr_t edesc)
{
	int rc;
	ATOMIC_INC_INT(&eee_stats->fwd_pkts[EEE_GET_SLOT_NUM(dp)]
				   [EEE_GET_DEST_CPU(dp)], 1);
	rc = (*(drv_port->func))(dp, edesc);
	switch (rc) {
	case EEE_FORWARDED_PACKET:
		ATOMIC_INC_INT(&drv_port->forwarded, 1);
		break;
	case EEE_DROP_PACKET:
		ATOMIC_INC_INT(&drv_port->dropped, 1);
		printf("eee_doTransmit: packet dropped\n");
		eee_freePacket(edesc);
		break;
	case EEE_TX_QUEUE_FULL:
		panic("EEE_TX_QUEUE_FULL, dp = 0x%x\n", dp);
		break;
	default:
		panic("Bad return code, 0x%x\n", rc);
		break;
	}
	return rc;
}

	
/*-----------------------------------------------------------------
 * Name:	eee_forwardPacket()
 *
 * Description: This function is used to send a packet through the system.
 *
 *				An application that needs to send a packet should use this
 *				function.
 *				A driver that receives a packet should use this function
 *				to deliver the packet to appropriate destination.
 *
 * 				This function will handle packets destined for applications
 *				on the local CPU, other CPUs on the local slot, or
 *				sending to registered transmit drivers for delivery across
 *				the backplane switch fabrics.
 *
 *
 *				Calling function should look something like this:
 *
 *					rc = forwardPacket(edesc);
 *					if (rc == EEE_DROP_PACKET) {
 *						// The packet was dropped by either the driver or
 *						// the EEE layer.  The descriptor has been freed.
 *						// The caller should make no further references to
 *						// the packet.
 *					}
 *					else if (rc == EEE_TX_QUEUE_FULL) {
 *						// The packet could not be queued due a tx queue full
 *						// condition.  The caller can take whatever action
 *						// it wants with the packet, since it is still the
 *						// owner.
 *					}
 *					else {
 *						// The packet was queued for transmit successfully.
 *						// The caller should not free the packet (or touch
 *						// the packet in any way), as it is now owned by
 *						// someone else.
 *					}
 *
 *
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/18/01
 *
 *-----------------------------------------------------------------
 */
int32							/* EEE_FORWARDED_PACKET
								 * EEE_DROP_PACKET
								 * EEE_TX_QUEUE_FULL
								 */
eee_forwardPacket(
	eee_descPtr_t	edesc
)
{
	uint32	dp;
	uint32  slot_id, cpu_id, app_id, dtype;
	uint32	port;
	int32	rc;
	eee_port_t *drv_port;
/*
	printf("eee_forwardPacket, %p\n", edesc);
*/

	dp = edesc->hdr.dest_port;
	slot_id = EEE_GET_SLOT_NUM(dp);
	cpu_id = EEE_GET_DEST_CPU(dp);
	dtype = EEE_GET_DTYPE(dp);
	port = EEE_GET_PORT_ID(dp);

	/*
	 * Is this destined for another application on same CPU?
	 */
	if ((port == PORT_BEST_FIT) || (port == PORT_LOCAL_APP)) {
#if defined(NFP_TXRX)
		if ((slot_id == nfxMySlotId) &&
					((cpu_id == NFX_NFP_TXRX1)
					 || (cpu_id == NFX_NFP_TXRX2)
					 || (cpu_id == NFX_NFP_TXRX_ANY))) {
#elif defined(NFP_FP)
		if ((slot_id == nfxMySlotId) &&
					((cpu_id == NFX_NFP_FP2) || (cpu_id == NFX_NFP_FP_ANY)
						|| (cpu_id == NFX_NFP_FP1))) {
#else
		if ((slot_id == nfxMySlotId) && (cpu_id == nfxMyCPUId)) {
#endif
			app_id = EEE_GET_APP_ID(dp);
			if (eee_deliverPacketToApp(edesc, app_id) != NFX_OK) {
				eee_freePacket(edesc);
				return EEE_DROP_PACKET;
			}
			else
				return EEE_FORWARDED_PACKET;
		}
	}

#ifdef NFP_FP
	if (dtype == EEE_MGMT) {
		//printf("MGMT packet, change port to LOCAL_QUEUE\n");
		port = NFP_PORT_LOCAL_QUEUE;
	}
#endif

	/*
	 * Can this packet be queued locally?
	 */
#if defined(NFX_MOD_NFP) || defined(NFX_MOD_SSC)
	if (port == PORT_BEST_FIT) {
		if (slot_id == nfxMySlotId)
#ifdef NFX_MOD_NFP
			port = NFP_PORT_LOCAL_QUEUE;
#else
			port = SSC_PORT_LOCAL_QUEUE;
#endif
	}
#endif

#if defined(NFP_TXRX) && defined(NFX_SMP)
	/* if managment msg on acpu, give to ncpu */
	if ((dtype == EEE_MGMT) && (nfxMyCPUId != NFX_NFP_TXRX1)) {
		eee_receivePacket(edesc);
		return EEE_FORWARDED_PACKET;
	}
#endif
	
	if ((rc = eee_findPort(port, dp, &drv_port)) == NFX_OK) {
		eee_descPtr_t packetDesc;
		if (!(edesc->hdr.control & EDESC_COMMAND_DONT_CONVERT)) {
			int size_too_big =
				EEE_GET_PKT_LEN(edesc) + sizeof(agile_hdr_t) > drv_port->mtu;
#if defined(NFP_TXRX)
			/* Need to satisfy driver requirements about buffer location */
			int need_convert = size_too_big
				|| (dtype == EEE_MGMT && port != NFP_PORT_LOCAL_QUEUE);
#else
			int need_convert = size_too_big;
#endif
			
			if (need_convert) {
				packetDesc = eee_convertPacket(edesc, drv_port->mtu);
			}
			else {
				packetDesc = edesc;
			}
		}
		else {
			packetDesc = edesc;
		}

#if defined(NFP_FP)
		if (port == NFP_PORT_LOCAL_QUEUE)
			rc = eee_doTransmit(drv_port, dp, packetDesc);
		else {
#endif			

		if (packetDesc != NULL) {
			int rc = EEE_FORWARDED_PACKET;
			while (packetDesc != NULL) {
				eee_descPtr_t nextDesc = eee_nextDesc(packetDesc);
				eee_unlinkNextDesc(packetDesc);

				if (rc == EEE_FORWARDED_PACKET) {
					rc = eee_doTransmit(drv_port, dp, packetDesc);
				}
				else {
					printf("eee_doTransmit error, free rest of chain\n");
					eee_freePacket(packetDesc);
				}
				packetDesc = nextDesc;
			}
		}
		else {
			rc = EEE_TX_QUEUE_FULL;
			printf("Can't fragment the packet, dropped\n");
		}
#if defined(NFP_FP)		
		}
#endif		
	}
	else {
		if (rc == EEE_DROP_PACKET) {
			printf("eee_forwardPacket: find port error, packet dropped\n");
			eee_freePacket(edesc);
		}
	}

//	EEE_PRINTF(("eee_forwardPacket: returning %d\n", rc));

	return rc;

}	/* eee_forwardPacket() */

#if defined(NFX_MOD_NFP)
/*-----------------------------------------------------------------
 * Name:	eee_forwardLocalPktUni()
 *
 * Description:	This function queues a packet to a local forwarding queue
 *				for delivery to a CPU on the local card that can be accessed
 *				via shared memory.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/18/01
 *
 *-----------------------------------------------------------------
 */
int32								/* EEE_FORWARDED_PACKET
									 * EEE_TX_QUEUE_FULL
									 */
eee_forwardLocalPktUni(
	uint32		dest_port,
	eee_desc_t	* edesc
)
{
	uint32	cpu_id;
	uint32	fwd_q_index;
	int32	rc;

	cpu_id = EEE_GET_DEST_CPU(dest_port);
#ifdef NFP_TXRX
	fwd_q_index = 1;
#elif defined(NFP_FP)
	fwd_q_index = 0;
#else
	fwd_q_index = GET_FWD_INDEX_FROM_CPU_ID(cpu_id);
#endif

#ifdef EEE_ATTR
	eee_attrAssignPkt(edesc, EEE_ATTR_FWD);
#endif

	NFX_SPIN_LOCK(&eee_fwdqueue_spin);
	rc = FWD_PACKET(((address_t)edesc & ~EDESC_ADDR_MASK)
					| EEE_FWD_ACTION_FWD, fwd_q_index);
	NFX_SPIN_UNLOCK(&eee_fwdqueue_spin);
	return rc;

}	/* eee_forwardLocalPktUni() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
int32
eee_forwardLocalPktMulti(
	uint32		dest_port,
	eee_desc_t	* edesc
)
{
	printf("eee_forwardLocalPktMulti!\n");

	return EEE_DROP_PACKET;
}
#endif


/*-----------------------------------------------------------------
 * Name :	eee_initFragmentTimeout()
 *
 * Description: initialize the fragments timeout facility
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/04/02
 *
 *-----------------------------------------------------------------
 */
void
eee_initFragmentTimeout()
{
	int32 th;

	th = eee_timerCreate(2 * MSEC_PER_SEC, EEE_TIMER_REPEAT);
	if (th != TIMER_ERR_NO_TIMER_AVAILABLE) {
		int rc = eee_timerStart(th, fragmentTimeout,
								NULL, NULL, NULL, NULL);
		NFX_ASSERT(rc == NFX_OK);
	}
	else {
		panic("Can not initialize fragments timeout\n");
	}
}


/*-----------------------------------------------------------------
 * Name :	eee_addHeader()
 *
 * Description: Add a header buffer containing agile header and
 * ethernet header to the descriptor for sending the packet through
 * the backplane.
 *
 * Returns: NFX_OK if succeeeds, otherwise NFX_ERR
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/10/02
 *
 *-----------------------------------------------------------------
 */
int
eee_addHeader(eee_desc_t *pktDesc, int remotePort, uchar8 *myMacAddr)
{
	int pktLen = EEE_GET_PKT_LEN(pktDesc);
	int bufCnt = EEE_GET_NUM_BUFS(pktDesc);
	int dtype = EEE_GET_DTYPE(pktDesc->hdr.dest_port);
	char *hdrBuf;
	enet_hdr_t *ethHdr;
	agile_hdr_t *eeeHdr;
	uint32 hdrLen = sizeof(enet_hdr_t) + sizeof(agile_hdr_t);
	uint32 offset;

	// Check for room in the first buffer.
	hdrBuf = pktDesc->bd[1].buf;	
#if defined(EEE_NEW_BUFSIZE)
    offset = (address_t)hdrBuf & 0x7F;
#else
	offset = (address_t)hdrBuf & 0xFF;
#endif	

	/* The descriptor passed should not use the first buffer */
	NFX_ASSERT(!(pktDesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER));

	if( offset >= hdrLen
		&& !(pktDesc->bd[1].control & EDESC_BD_CMD_START_IN_USE)) {

#if defined(EEE_VALIDATE)		
		NFX_ASSERT((address_t)eee_getBufBase(hdrBuf) <=
				   (address_t)hdrBuf - hdrLen);
#endif		
		hdrBuf -= hdrLen;

		
		ethHdr = (enet_hdr_t *)hdrBuf;
		eeeHdr = (agile_hdr_t *)(hdrBuf + sizeof(enet_hdr_t));
		bp_setDestMacAddr(pktDesc->hdr.dest_port, remotePort, ethHdr->DA);
		memcpy(ethHdr->SA, myMacAddr, 6);
		if (dtype == EEE_SAC)
			ethHdr->type = EEE_ETH_TYPE_SAC;
		else
			ethHdr->type = EEE_ETH_TYPE;

		eeeHdr->pad = 0;
		eeeHdr->dest_port = pktDesc->hdr.dest_port;
		eeeHdr->src_port = pktDesc->hdr.src_port;
#if defined(NETEEE_FRAGMENT)
		eeeHdr->offset = pktDesc->hdr.offset;
#endif
		pktDesc->bd[1].buf = hdrBuf;
		pktDesc->bd[1].len += hdrLen;
		pktLen += hdrLen;

		pktDesc->hdr.control = (pktLen & EDESC_TOTAL_LEN) |
			((bufCnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
	}
	else {
#ifdef NFP_TXRX
		hdrBuf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_MGMT, 0);
#else
		hdrBuf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_SMALL, 0);
#endif
		if (hdrBuf == NULL) {
			printf("eee_addHeader: Error cannot allocate buffer\n");
			return NFX_ERR;
		}

		ethHdr = (enet_hdr_t *)hdrBuf;
		eeeHdr = (agile_hdr_t *)(hdrBuf + sizeof(enet_hdr_t));

		bp_setDestMacAddr(pktDesc->hdr.dest_port, remotePort, ethHdr->DA);
		memcpy(ethHdr->SA, myMacAddr, 6);
		if (dtype == EEE_SAC)
			ethHdr->type = EEE_ETH_TYPE_SAC;
		else
			ethHdr->type = EEE_ETH_TYPE;

		eeeHdr->pad = 0;
		eeeHdr->dest_port = pktDesc->hdr.dest_port;
		eeeHdr->src_port = pktDesc->hdr.src_port;
#if defined(NETEEE_FRAGMENT)
		eeeHdr->offset = pktDesc->hdr.offset;
#endif

		// Add header buffer to the edesc.
		pktDesc->bd[0].buf     = hdrBuf;
		pktDesc->bd[0].len     = hdrLen;
		pktDesc->bd[0].control = EDESC_BD_CMD_IN_USE | EDESC_BD_SMALL;
		pktLen += hdrLen;
		bufCnt++;

		pktDesc->hdr.control =
			(pktLen & EDESC_TOTAL_LEN) |
			((bufCnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK) |
			EDESC_COMMAND_USE_FIRST_BUFFER;

		CACHE_FLUSH(hdrBuf, hdrBuf + hdrLen);
	}

	return NFX_OK;
}



/*-----------------------------------------------------------------
 * Name :	eee_removeHeader()
 *
 * Description: strip the mac header and agile header, fill the descriptor
 * header.
 *
 * Returns: the size of the removed header
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	04/23/02
 *
 *-----------------------------------------------------------------
 */
int
eee_removeHeader(eee_descPtr_t edesc)
{
	agile_hdr_t *eeeHdr = (agile_hdr_t *)(edesc->bd[1].buf
										  + sizeof(enet_hdr_t));
	int hdrlen = sizeof(enet_hdr_t) + sizeof(agile_hdr_t) + eeeHdr->pad;
	edesc->hdr.dest_port = eeeHdr->dest_port;
	edesc->hdr.src_port = eeeHdr->src_port;
#if defined(NETEEE_FRAGMENT)
	edesc->hdr.offset = eeeHdr->offset;
#endif		
	edesc->bd[1].buf += hdrlen;
	edesc->bd[1].len -= hdrlen;
	return hdrlen;
}


struct eee_rcv_queue *
eee_my_receive_queue(void)
{
	 int i;
	 uint32 my_fwd_q;
	 GET_MY_FWD_INDEX(my_fwd_q);
	 for (i = 0; i < EEE_NUM_FWD_QUEUES; i++) {
		 if (i == my_fwd_q)
			 continue;
		 return &eee.rcv_queue[i];
	 }
	 return NULL;
}

/*-----------------------------------------------------------------
 * Name:	eee_fwdInit()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/18/01
 *
 *-----------------------------------------------------------------
 */
void
eee_fwdInit()
{
#ifdef LINUX_TEST
	uint32	cpu;
	int32	shmid, slot;
	uchar8	* shmptr;
#ifdef LINUX_POLL_NONBLOCKING_FD
    struct sockaddr_un sockAddr;
    char pathName[40];
#endif
	extern int32	eee_linuxForwardRemotePktUni();
	extern int32	eee_linuxForwardRemotePktMulti();
	extern void		eee_linuxForwardReceive();

#if 0
	extern int32	eee_linuxSockFwdPktUni();
	extern int32	eee_linuxSockFwdPktMulti();
	extern void		eee_linuxSockFwdRx();
#endif
#endif
	NFX_SPIN_INIT(&fragment_spin, SPINLOCK_LEVEL_UNKNOWN);

#if (EEE_NUM_FWD_QUEUES > 0)
	eee_my_receive_queue()->rcv_max = 64;
	eee_my_receive_queue()->rcv_hwm = 0;
	NFX_SPIN_INIT(&eee_my_receive_queue()->rcv_spin, SPINLOCK_LEVEL_UNKNOWN);
	eee_registerPollingFunc(EEE_HIGH_PRIORITY,
							eee_processReceiveQueue,
							eee_my_receive_queue());

#if defined(NFX_MOD_NFP)
	eee_registerTransmitFunc(
							NFP_PORT_LOCAL_QUEUE,
							eee_forwardLocalPktUni,
							eee_forwardLocalPktMulti,
							EDESC_TOTAL_LEN);
#elif defined(NFX_MOD_SSC)
#if !defined(AGILE_HARDWARE)
	eee_registerTransmitFunc(
							SSC_PORT_LOCAL_QUEUE,
							eee_forwardLocalPktUni,
							eee_forwardLocalPktMulti,
							EDESC_TOTAL_LEN);
#endif	
#endif

#endif

#if defined(NFX_SMP)
	NFX_SPIN_INIT(&local_queue.lq_spin, SPINLOCK_LEVEL_UNKNOWN);
	eee_registerPollingFunc(EEE_HIGH_PRIORITY,
							eee_poll_local_queue,
							&local_queue);


#if defined(NFP_TXRX)
	NFX_SPIN_INIT(&local_queueA.lq_spin, SPINLOCK_LEVEL_UNKNOWN);
	eee_registerPollingFunc(EEE_HIGH_PRIORITY,
							eee_poll_local_queueA,
							&local_queueA);	

	eee_txrxMapAcpu();
#endif /* NFP_TXRX */	
#endif	

#ifdef LINUX_TEST
/*
 * Management Plane queues + semaphores
 */
	/*
	 * Slot 0, SSC (cpu0 = 7000, MGMT has no MP port)
	 */
	if ((linux_mp_sem[0][0] = semget(SEM_MP_SSC, 0, 0)) == -1)
		printf("error getting MP_SSC semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_SSC, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_SSC shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_SSC shmat error\n");
	}
	printf("MP_SSC : base at 0x%x\n", shmptr);
	linux_mp[0][0] = (shm_tx_t *)shmptr;

	/*
	 * Slot 1, NFP (TXRX1, TXRX2, FP1, FP2)
	 */
	if ((linux_mp_sem[1][0] = semget(SEM_MP_TXRX1, 0, 0)) == -1)
		printf("error getting MP_TXRX1 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_TXRX1, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_TXRX1 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_TXRX1 shmat error\n");
	}
	printf("MP_TXRX1 : base at 0x%x\n", shmptr);
	linux_mp[1][0] = (shm_tx_t *)shmptr;

	if ((linux_mp_sem[1][1] = semget(SEM_MP_TXRX2, 0, 0)) == -1)
		printf("error getting MP_TXRX2 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_TXRX2, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_TXRX2 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_TXRX2 shmat error\n");
	}
	printf("MP_TXRX2 : base at 0x%x\n", shmptr);
	linux_mp[1][1] = (shm_tx_t *)shmptr;

	if ((linux_mp_sem[1][2] = semget(SEM_MP_FP1, 0, 0)) == -1)
		printf("error getting MP_FP1 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_FP1, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_FP1 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_FP1 shmat error\n");
	}
	printf("MP_FP1 : base at 0x%x\n", shmptr);
	linux_mp[1][2] = (shm_tx_t *)shmptr;

	if ((linux_mp_sem[1][3] = semget(SEM_MP_FP2, 0, 0)) == -1)
		printf("error getting MP_FP2 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_FP2, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_FP2 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_FP2 shmat error\n");
	}
	printf("MP_FP2 : base at 0x%x\n", shmptr);
	linux_mp[1][3] = (shm_tx_t *)shmptr;

	/*
	 * Slot 2, FCNIM (FC1, FC2, FC3, FC4)
	 */
	if ((linux_mp_sem[2][0] = semget(SEM_MP_FC1, 0, 0)) == -1)
		printf("error getting MP_FC1 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_FC1, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_FC1 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_FC1 shmat error\n");
	}
	printf("MP_FC1 : base at 0x%x\n", shmptr);
	linux_mp[2][0] = (shm_tx_t *)shmptr;

	if ((linux_mp_sem[2][1] = semget(SEM_MP_FC2, 0, 0)) == -1)
		printf("error getting MP_FC2 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_FC2, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_FC2 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_FC2 shmat error\n");
	}
	printf("MP_FC2 : base at 0x%x\n", shmptr);
	linux_mp[2][1] = (shm_tx_t *)shmptr;

	if ((linux_mp_sem[2][2] = semget(SEM_MP_FC3, 0, 0)) == -1)
		printf("error getting MP_FC3 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_FC3, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_FC3 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_FC3 shmat error\n");
	}
	printf("MP_FC3 : base at 0x%x\n", shmptr);
	linux_mp[2][2] = (shm_tx_t *)shmptr;

	if ((linux_mp_sem[2][3] = semget(SEM_MP_FC4, 0, 0)) == -1)
		printf("error getting MP_FC4 semaphore\n");
	if ((shmid = shmget(SHM_SEG_MP_FC4, 0, SHM_R | SHM_W)) < 0) {
		printf("error in MP_FC4 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("MP_FC4 shmat error\n");
	}
	printf("MP_FC4 : base at 0x%x\n", shmptr);
	linux_mp[2][3] = (shm_tx_t *)shmptr;

/*
 * Data Plane queues + semaphores
 */

	/*
	 * Slot 1, NFP (TXRX, FP)
	 */
	if ((linux_dp_sem[1][0] = semget(SEM_DP_TXRX, 0, 0)) == -1)
		printf("error getting DP_TXRX semaphore\n");
	if ((shmid = shmget(SHM_SEG_DP_TXRX, 0, SHM_R | SHM_W)) < 0) {
		printf("error in DP_TXRX shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("DP_TXRX shmat error\n");
	}
	printf("DP_TXRX : base at 0x%x\n", shmptr);
	linux_dp[1][0] = (shm_tx_t *)shmptr;

	if ((linux_dp_sem[1][1] = semget(SEM_DP_FP, 0, 0)) == -1)
		printf("error getting DP_FP semaphore\n");
	if ((shmid = shmget(SHM_SEG_DP_FP, 0, SHM_R | SHM_W)) < 0) {
		printf("error in DP_FP shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("DP_FP shmat error\n");
	}
	printf("DP_FP : base at 0x%x\n", shmptr);
	linux_dp[1][1] = (shm_tx_t *)shmptr;

	/*
	 * Slot 2, FCNIM (FC1, FC2, FC3, FC4)
	 */
	if ((linux_dp_sem[2][0] = semget(SEM_DP_FC1, 0, 0)) == -1)
		printf("error getting DP_FC1 semaphore\n");
	if ((shmid = shmget(SHM_SEG_DP_FC1, 0, SHM_R | SHM_W)) < 0) {
		printf("error in DP_FC1 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("DP_FC1 shmat error\n");
	}
	printf("DP_FC1 : base at 0x%x\n", shmptr);
	linux_dp[2][0] = (shm_tx_t *)shmptr;

	if ((linux_dp_sem[2][1] = semget(SEM_DP_FC2, 0, 0)) == -1)
		printf("error getting DP_FC2 semaphore\n");
	if ((shmid = shmget(SHM_SEG_DP_FC2, 0, SHM_R | SHM_W)) < 0) {
		printf("error in DP_FC2 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("DP_FC2 shmat error\n");
	}
	printf("DP_FC2 : base at 0x%x\n", shmptr);
	linux_dp[2][1] = (shm_tx_t *)shmptr;

	if ((linux_dp_sem[2][2] = semget(SEM_DP_FC3, 0, 0)) == -1)
		printf("error getting DP_FC3 semaphore\n");
	if ((shmid = shmget(SHM_SEG_DP_FC3, 0, SHM_R | SHM_W)) < 0) {
		printf("error in DP_FC3 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("DP_FC3 shmat error\n");
	}
	printf("DP_FC3 : base at 0x%x\n", shmptr);
	linux_dp[2][2] = (shm_tx_t *)shmptr;

	if ((linux_dp_sem[2][3] = semget(SEM_DP_FC4, 0, 0)) == -1)
		printf("error getting DP_FC4 semaphore\n");
	if ((shmid = shmget(SHM_SEG_DP_FC4, 0, SHM_R | SHM_W)) < 0) {
		printf("error in DP_FC4 shmget, shmid = 0x%x\n", shmid);
	}
	if ((shmptr = shmat(shmid, 0, 0)) == (void *) -1) {
		printf("DP_FC4 shmat error\n");
	}
	printf("DP_FC4 : base at 0x%x\n", shmptr);
	linux_dp[2][3] = (shm_tx_t *)shmptr;

	/*
	 * Local Forwarding queues on NFP and SSC
	 */
#if defined(NFX_MOD_NFP)
	if ((linux_fwd_sem[1][0] = semget(SEM_FWD_TO_TXRX, 0, 0)) == -1)
		printf("error getting FWD_TO_TXRX semaphore\n");
	if ((linux_fwd_sem[1][1] = semget(SEM_FWD_TO_FP, 0, 0)) == -1)
		printf("error getting FWD_TO_FP semaphore\n");
#endif

#ifdef LINUX_POLL_NONBLOCKING_FD
    /* Create, bind and register our poll indication socket */
    eee_linuxFwdPollRxFd = socket(AF_UNIX, SOCK_DGRAM, 0);
    if (eee_linuxFwdPollRxFd < 0)
    {
        perror("eee_fwd socket");
        panic("eee_fwd pollFd socket create failed");
    }
	if (fcntl(eee_linuxFwdPollRxFd, F_SETFL, O_NONBLOCK))
	{
		perror("eee_fwd fcntl");
		panic("eee_fwd pollFd socket fcntl failed");
	}
    sockAddr.sun_family = AF_UNIX;
    sprintf(pathName, "/tmp/EEE_SLOT_%d_CPU_%d", nfxMySlotId, nfxMyCPUId);
    strcpy(sockAddr.sun_path, pathName);
	/* unlink from previous run */
	unlink(pathName);
    if (bind(eee_linuxFwdPollRxFd, &sockAddr, sizeof(sockAddr)) != 0)
    {
        perror("eee_fwd bind");
        panic("eee_fwd pollFd bind of %s failed", pathName);
    }
    eee_addPollIndicationFd(eee_linuxFwdPollRxFd);
    /* Mark destination sockets as uninitialized, we'll initialize
     * each as needed.
     */
    for (slot = 0; slot < NFX_SLOT_MAX; slot++)
        for (cpu = 0; cpu < NFX_MAX_CPUS_PER_SLOT; cpu++) {
            eee_linuxFwdPollTxFd[slot][cpu] = -1;
            eee_linuxFwdPollTxFdState[slot][cpu] = -1;
        }
#endif

#if NFX_SOCKET_BACKPLANE
	eee_registerPollingFunc(EEE_HIGH_PRIORITY, eee_linuxSockFwdRx, NULL);
#endif // NFX_SOCKET_BACKPLANE

	eee_registerDefaultTransmitFunc(
									EEE_USER_DATA,
									eee_linuxForwardRemotePktUni,
									EDESC_TOTAL_LEN);
	eee_registerDefaultTransmitFunc(
									EEE_CONTROL,
									eee_linuxForwardRemotePktUni,
									EDESC_TOTAL_LEN);
	eee_registerDefaultTransmitFunc(
									EEE_MGMT,
									eee_linuxForwardRemotePktUni,
									EDESC_TOTAL_LEN);
	eee_registerDefaultTransmitFunc(
									EEE_JOURNAL,
									eee_linuxForwardRemotePktUni,
									EDESC_TOTAL_LEN);
	eee_registerDefaultTransmitFunc(
									EEE_METADATA,
									eee_linuxForwardRemotePktUni,
									EDESC_TOTAL_LEN);

#if defined(NFP_TXRX)

	/* 
	 * eee_registerTransmitFunc() will be done in luc-sim.c 
	 */

#elif defined(NFP_FP)

	eee_registerTransmitFunc(
							NFP_PORT_FP_ETH0,
							eee_linuxForwardRemotePktUni,
							eee_linuxForwardRemotePktMulti,
							EDESC_TOTAL_LEN);
	eee_registerTransmitFunc(
							NFP_PORT_FP_ETH1,
							eee_linuxForwardRemotePktUni,
							eee_linuxForwardRemotePktMulti,
							EDESC_TOTAL_LEN);
	eee_registerTransmitFunc(
							NFP_PORT_FP_ETH2,
							eee_linuxForwardRemotePktUni,
							eee_linuxForwardRemotePktMulti,
							EDESC_TOTAL_LEN);

#elif defined(NFX_MOD_FCNIM)

	eee_registerTransmitFunc(
							FCNIM_PORT_FC,
							NULL,
							NULL,
							EDESC_TOTAL_LEN);
	eee_registerTransmitFunc(
							FCNIM_PORT_BP_DATA,
							eee_linuxForwardRemotePktUni,
							eee_linuxForwardRemotePktMulti,
							EDESC_TOTAL_LEN);
	eee_registerTransmitFunc(
							FCNIM_PORT_BP_MGMT,
							eee_linuxForwardRemotePktUni,
							eee_linuxForwardRemotePktMulti,
							EDESC_TOTAL_LEN);

#elif defined(NFX_MOD_SSC)

	eee_registerTransmitFunc(
							SSC_PORT_BP_MGMT,
							eee_linuxForwardRemotePktUni,
							eee_linuxForwardRemotePktMulti,
							EDESC_TOTAL_LEN);

#endif

#if defined(NFP_TXRX) || defined(NFP_FP) || defined(FCNIM)
	/* data plane */
	eee_registerPollingFunc(EEE_HIGH_PRIORITY, eee_linuxForwardReceive, (void *)0);
#endif
#if defined(NFP_TXRX) || defined(NFP_FP) || defined(FCNIM) || defined(SSC)
	/* mgmt plane */
	eee_registerPollingFunc(EEE_HIGH_PRIORITY, eee_linuxForwardReceive, (void *)1);
#endif

#endif	/* LINUX_TEST */

}	/* eee_fwdInit() */

