/*
 *-----------------------------------------------------------------
 *
 * Name:	eee-poll.c
 *
 * RCS:  $Id: eee-poll.c,v 1.21 2002/12/05 06:47:36 tonyw Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	EEE layer polling routines
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/timer-api.h>


extern void rconPoll(void);
int32	dont_poll  CPU_PRIVATE;

/*-----------------------------------------------------------------
 * Name:	eee_registerPollingFunc(priority, polling_func, cb)
 *
 * Description:	Polling functions will be called periodically.
 *				Drivers should use polling functions to process receive
 *				rings, transmit cleanup, etc.
 *				The polling function will be called with the user-supplied
 *				control block, and a timer reference value.  The timer
 *				reference will be utilized via some EEE layer macros.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/9/01
 *
 *-----------------------------------------------------------------
 */
int32						/* returns handle for polling function or
							 * EEE_ERR_NOT_INITED
							 * EEE_ERR_POLLING_TABLE_FULL
							 * EEE_ERR_BAD_PARAM
							 */
eee_registerPollingFunc(
	uint32	priority,			/* EEE_HIGH_PRIORITY, EEE_LOW_PRIORITY */
	void	(*polling_func)(		/* function to be called as part of
									 * EEE layer polling loop
									 */
				void	* cb,		/* user defined control block for
									 * polling function
									 */
				uint32	timer_ref	/* some indication of how much time
									 * this routine can use the CPU, or
									 * how many packets to process in a
									 * loop - supplied by the EEE layer
									 */
			),
	void	* cb
)
{
	uint32	i;

	if (!eee.init_done)
		return EEE_ERR_NOT_INITED;

	if (priority == EEE_HIGH_PRIORITY) {
		for (i=0; i<EEE_MAX_POLL_HIGH_ROUTINES - 1; i++) {
			if (eee.pfunc.high[i].func == NULL) {
				eee.pfunc.high[i].func = polling_func;
				eee.pfunc.high[i].cb = cb;
				return (i | EEE_HIGH_PRIORITY_IND);
			}
		}
		return EEE_ERR_POLLING_TABLE_FULL;
	}
	else if (priority == EEE_LOW_PRIORITY) {
		for (i=0; i<EEE_MAX_POLL_LOW_ROUTINES - 1; i++) {
			if (eee.pfunc.low[i].func == NULL) {
				eee.pfunc.low[i].func = polling_func;
				eee.pfunc.low[i].cb = cb;
				return (i | EEE_LOW_PRIORITY_IND);
			}
		}
		return EEE_ERR_POLLING_TABLE_FULL;
	}
	else {
		printf("Invalid EEE priority, 0x%x\n", priority);
		return EEE_ERR_BAD_PARAM;
	}

}	/* eee_registerPollingFunc() */

#if 0
/*-----------------------------------------------------------------
 * Name:	eee_unregisterPollingFunc(handle)
 *
 * Description:	Basically provided for completeness.  A polling function
 *				can be removed from polling loop.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/9/01
 *
 *-----------------------------------------------------------------
 */
int32						/* returns NFX_OK,
							 * EEE_ERR_POLLING_FUNC_INVALID
							 * EEE_ERR_BAD_PARAM
							 */
eee_unregisterPollingFunc(
	int32	handle				/* from eee_registerPollingFunc()
								 * return code
								 */
)
{
	int32	index;

	if (handle & EEE_HIGH_PRIORITY_IND) {
		index = (handle & ~EEE_HIGH_PRIORITY_IND);
		if ((index >= 0) && (index < EEE_MAX_POLL_HIGH_ROUTINES)) {
			if (eee.pfunc.high.func[index] != NULL) {
				eee.pfunc.high.func[index] = NULL;
				return NFX_OK;
			}
			else
				return EEE_ERR_POLLING_FUNC_INVALID;
		}
		else
			return EEE_ERR_BAD_PARAM;
	}
	else if ((handle >= 0) && (handle < EEE_MAX_POLL_LOW_ROUTINES)) {
		if (eee.pfunc.low.func[handle] != NULL) {
			eee.pfunc.low.func[handle] = NULL;
			return NFX_OK;
		}
		else
			return EEE_ERR_POLLING_FUNC_INVALID;
	}
	else
		return EEE_ERR_BAD_PARAM;
}	/* eee_unregisterPollingFunc() */
#endif

/*-----------------------------------------------------------------
 * Name:	eee_poll()
 *
 * Description:	Main polling loop
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */

#ifdef EEE_POLLING_BUCKET
int32 eeePollingBuckets[EEE_NBR_POLLING_BUCKETS] = { 500,1000,10000,30000,100000,100000000};
#endif

#ifdef EEE_POLLING_PERF
int64 eeePollingCounter = 0;
int64 eeePollingCounterMin = 999999999;
int64 eeePollingCounterMax = 0;
int64 eeePollingCountersHist[EEE_POLLING_HIST] = {0,0,0,0,0,0,0,0,0,0};
int32 eeePollingCounterIndex = 0;
int32 eeePollingPrintf = 0;
#endif

void
eee_poll(
	uint32	num_loops
)
{
	struct eee_poll_cb * poll_cb;
#ifdef EEE_POLLING_PERF
    eeeRTC_t duration;
	static eeeRTC_t lastPollingTime;
	static eeeRTC_t interval;
#endif            
#if defined(EEE_POLLING_BUCKET) || defined(EEE_POLLING_PERF)
	static boolean first = TRUE;
#endif
#ifdef EEE_POLLING_BUCKET
	static eeeRTC_t eeePollingBucketsRTC[EEE_NBR_POLLING_BUCKETS];
#endif

	if (dont_poll) {
#ifdef USE_WATCHDOG
		watchdog_kick();
#endif
		return;
	}

#ifdef NFX_SMP
	rconPoll();
#endif

#if defined(EEE_POLLING_BUCKET) || defined(EEE_POLLING_PERF)
	if (first) 
    {
		first = FALSE;
#ifdef EEE_POLLING_BUCKET
		{
			int32 i;
			// Setup buckets in RTC form
			for (i=0; i<EEE_NBR_POLLING_BUCKETS; i++) {
				eee_usecToRTC((uint64)eeePollingBuckets[i],
												  &(eeePollingBucketsRTC[i]));
			}
		}
#endif
#ifdef EEE_POLLING_PERF
        lastPollingTime = *(eee_rtcZero());
#endif
	}
#endif

	while (num_loops--) {
#ifdef USE_WATCHDOG
        if ((num_loops & 0x7) == 0)
            wd_kick();
#endif
		for (poll_cb = &eee.pfunc.high[0]; poll_cb->func; ++poll_cb) {
#ifdef EEE_POLLING_BUCKET
			eeeRTC_t start, delta;
			uint32 bucket = 0, loop;

			eee_rtcStart (&start);
#endif            
			(poll_cb->func)(poll_cb->cb, 2);

#ifdef EEE_POLLING_BUCKET
			eee_rtcDuration (&start, &delta);
			for (loop = 0; loop < EEE_NBR_POLLING_BUCKETS; loop++)
			{
				if (eee_rtcCompare(&delta, &(eeePollingBucketsRTC[loop])) <= 0)
				{
					bucket = loop;
					break;
				}
			}
			if (loop >= EEE_NBR_POLLING_BUCKETS)
			{
				bucket = EEE_NBR_POLLING_BUCKETS - 1;
				if (eeePollingPrintf)
				{
					eeeGMT_t startTV, endTV;
					eeeRTC_t end;

					eee_rtcToTV(&start, &startTV);
					eee_rtcAdd(&start, &delta, &end);
					eee_rtcToTV(&end, &endTV);
					printf ("POLLING TIME %d %d %d %d %lld\n", 
							startTV.sec, startTV.usec, endTV.sec, endTV.usec,
							eee_rtcToUsec(&delta));
				}
			}
			poll_cb->bucket[bucket] += 1;
#endif
		}

		for (poll_cb = &eee.pfunc.low[0]; poll_cb->func; ++poll_cb) {
#ifdef EEE_POLLING_BUCKET
			eeeRTC_t start, delta;
			uint32 bucket = 0, loop;

			eee_rtcStart (&start);
#endif
			poll_cb->func(poll_cb->cb, 2);
#ifdef EEE_POLLING_BUCKET
			eee_rtcDuration (&start, &delta);
			for (loop = 0; loop < EEE_NBR_POLLING_BUCKETS; loop++)
			{
				if (eee_rtcCompare(&delta, &(eeePollingBucketsRTC[loop])) <= 0)
				{
					bucket = loop;
					break;
				}
			}
			if (loop >= EEE_NBR_POLLING_BUCKETS)
			{
				bucket = EEE_NBR_POLLING_BUCKETS - 1;
				if (eeePollingPrintf)
				{
					eeeGMT_t startTV, endTV;
					eeeRTC_t end;

					eee_rtcToTV(&start, &startTV);
					eee_rtcAdd(&start, &delta, &end);
					eee_rtcToTV(&end, &endTV);
					printf ("POLLING TIME %d %d %d %d %lld\n", 
							startTV.sec, startTV.usec, endTV.sec, endTV.usec,
							eee_rtcToUsec(&delta));
				}
			}
			poll_cb->bucket[bucket] += 1;
#endif
		}
	}

#ifdef EEE_POLLING_PERF
	/* This code records the number of poll loops handled per sec, for the
	   last EEE_POLLING_HIST seconds. */
	eeePollingCounter++;

	if (eee_rtcCompare(&lastPollingTime, eee_rtcZero()) == 0)
	{
		eee_rtcStart(&lastPollingTime);
		eee_msToRTC(EEE_POLLING_HIST_START_AFTER_N_SEC * MSEC_PER_SEC,
					&interval);
	}
	else
	{
		eee_rtcDuration(&lastPollingTime, &duration);

		// Todo: case when we skip 1 sec

		if (eee_rtcCompare(&duration, &interval) > 0)
		{
			eeePollingCountersHist[eeePollingCounterIndex] = eeePollingCounter;
			eeePollingCounterIndex++;
			if (eeePollingCounterIndex >= EEE_POLLING_HIST)
				eeePollingCounterIndex = 0;

			if (eeePollingCounter < eeePollingCounterMin)
				eeePollingCounterMin = eeePollingCounter;
			if (eeePollingCounter > eeePollingCounterMax)
				eeePollingCounterMax = eeePollingCounter;
		}
		else
		{
			eeePollingCounterMin = eeePollingCounter;
			eeePollingCounterMax = eeePollingCounter;
			eeePollingCounter = 0;
		}

		eeePollingCounter = 0;
		eee_rtcReset(&lastPollingTime);
	}

#endif


}	/* eee_poll() */


#ifdef LINUX_TEST
#ifdef LINUX_POLL_NONBLOCKING_FD
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
fd_set eee_linuxPollFdset;
int    eee_linuxPollFdMax = -1;
/*-----------------------------------------------------------------
 * eee_addPollIndicationFd (int fd)
 *
 * In the linux EEE simulation environment, this additional
 * registration should be made by those modules that register
 * a polling function. The caller supplies a file descriptor that
 * indicates when the polling function needs to be called.  The
 * polling function is called when a read on the file descriptor
 * would not block (ie. there is data available).  This is a simple
 * mechanism that allows the EEE linux process to block in the
 * EEE kernel instead of running continuously and soaking up CPU
 * cycles mercilessly.  An example of its use would be that whenever
 * something is added to a queue that the polling function would
 * drain, the fd is written (typically the other end of a socket),
 * and the polling function reads it when removing the enqueued item.
 *
 * Registered polling functions that do not register an fd to be
 * watched will still have their registered functions called, but
 * the resulting delays may be unacceptable to the application.
 *-----------------------------------------------------------------
 */
void
eee_addPollIndicationFd(int fd)
{
    /* Add the file descriptor to the list. */
    if (eee_linuxPollFdMax == -1)
        FD_ZERO(&eee_linuxPollFdset);
    FD_SET(fd, &eee_linuxPollFdset);
    if (fd > eee_linuxPollFdMax)
        eee_linuxPollFdMax = fd;
}
#endif
#endif /* LINUX_TEST */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/25/02
 *
 *-----------------------------------------------------------------
 */
int32
eee_start_polling_ui(
	int32	argc,
	uchar8	* argv[]
)
{
	if (argc != 2)
		return -1;

	if (!strcmp(argv[1], "on"))
		dont_poll = 0;
	else if (!strcmp(argv[1], "off"))
		dont_poll = 1;
	else
		return -1;

	return 0;
}
