/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Name:  elog-api.h
 *
 * RCS:  $Id: elog-api.h,v 1.5 2002/08/20 21:03:30 maximk Exp $
 *
 * Description:  Defines the interface and API for the elog function and MACRO
 *               of the Eagle NFX.
 *
 * Created by:	Greg Seibert
 * Changes:     Jan Vanhoutte
 *              Pankaj Gupta
 *
 * Date Created: 13 Jul 2001
 *               11 sep 2001
 *
 */


/*
 * For embedded:
 * See the E_LOG macro for usage !
 * - or - make the application specific as the e.g.NVRAM_E_LOG
 *
 * On the SSC use the elog  function
 *
 */

/*
 * The generic Eagle logging function.
 * - input parameters (R: means required)
 *   - (R) class
 *   - (R) severity < DEBUG | INFO | NOTICE | WARNING | ERROR | 
 *                    CRITICAL | ALERT | EMERGENCY >
 *   - (R) application id : a number which will be converted to 
 *     a string with:  eee_getApllicationName (id) 
 *   - appl_type: an appplication specific logging type
 *     e.g.: File system: COD, IO, REQ, MEM, etc
 *     This is used as a bitmap to enable further filtering 
 *     (for lower severity levels)
 *     e.g.: enable all traces related to 
 *           (Appl == FS) & (Type == COD | REQ) 
 *     Required for DEBUG, but good to have it always)
 *   - cookie: a system wide variable (uint32) to enable correlation of events
 *   - code: an application specific variable (uint32)
 *     Cookie and code are not really required although ignoring these
 *     makes the use of this facility rather useless.
 *     If not used, set to 0.
 *   - line and file: automatically generated by the MACRO
 *   - format and arguments (between barckets) in the MACRO
 *     the e_log function takes only a string.
 *
 * - Returns NFX_OK if all was OK. NFX_ERR otherwise.
 *
 * - Add in the function
 *   - sequence number (to indicate loss, due to overload)
 *   - the time (depending of flag)
 *
 * - Filtering
 *   - Is first done in the MACRO, even before parsing any argument or 
 *     function call
 *     - test global flag if loging is enabled
 *     - test the severity level (no traces (== debug) are generated if
 *       not requested
 *     - Application specific severity checks: (STILL TO BE IMPLEMENTED)
 *       One (or more) can have debug while all others show only errors.
 *     - an application type bitmap to show only a subset of the traces.
 *
 * - global flags to change the output: <list of possible values, [default]> 
 *   - elog_show_date             <[SHOW_DATE] | SHOW_DATE_USEC | DISABLED>
 *   - elog_show_line_file        <[DISABLED] | ENABLED>
 *   - elog_emergency_reboot      <[DISABLED] | ENABLED>
 *     If enabled, a crash file will be generated and the CPU will be restarted
 *   - elog_format                <[one_line] | multiple_lines >
 *   - others ?
 *
 * Eagle logging
 *
 * Criteria
 * - performant/fast (don't use resources if not needed/enabled)
 * - simple to use
 * - find the information (by filtering)  in potential huge data flow
 *
 * The approach is based on different levels
 *
 * - e-log generic
 *   - work on each platform and simulation system
 *     - use eee_sendMessage if embedded
 *     - use printf in standalone mode
 *     - etc
 *   - possibility to redirect to file, syslog deamon, 
 *     operator interface, etc
 *
 * - generic logging / filtering
 *   - basic checks are done before calling the
 *     logging functions, to have minimal performance impact
 *     especially for lower severity level calls.
 *
 * - application specific requirements
 *   - possibility to add appliation specific filters.
 *   - add standard info which is the same within the application
 *
 *
 * Remark: special care (and some small further extensions)
 * is needed to solve the problem of passing a variable list of 
 * argument list.
 * - The macro definition has the format string and var list between 
 *   brackets.
 *
 * - Todo list
 *   - prevent calling e_log during e_log (e.g. form eee_SendMsg)
 *   - init and start from config file.
 *   - mulicast the e_log configuration to all CPU's
 *
 */

/*
 *   This is a bit more involved since it leverages the syslog facility provided
 *   by UNIX on the SSC. The general form of the log message will be:
 *
 * MON DD HH:MM:SS hostname logger: slot:cpu:appid:cookie:code:Message
 *
 * Where:
 *
 * MON      - Is the 3 character month name
 * DD       - The day of the month
 * HH:MM:SS - The time the log message was received by the SSC.
 * hostname - The name of the network node that issued this message
 * logger   - For lack of a better name, that's what I'm calling this application
 * slot     - Which slot did the message originate from
 * cpu      - Likewise, which CPU on the board in that slot
 * appid    - and which application.
 * cookie   - Can be used to correlate a group of messages
 * code     - Going to be appid specific. Useful for event-type things
 * Message  - A character string to be used as one might see fit
 *
 * The distribution of messages is dependent upon /etc/syslog.conf. We use the
 * local0 through local6 facility and all the priorities. (We don't use local7
 * because it gets used for boot up messages by the OS sometimes.)
 *
 * For instance, we might make the following entry in /etc/syslog.conf
 * 
 * *.*						/var/agile/messages
 * local5.*					gregs
 * local6.*					radeka
 * local4.*					@thunderbolt
 *
 *
 * All messages get logged to /var/agile/messages
 *
 * When the coffee pot gets empty, I'll get the message to go make more and all 
 * messages concerned with Espresso will be sent to Radek (if we are logged on).
 * All device error messages will get forwarded to Frank's machine, thunderbolt.
 * "man syslog.conf" for more info. It's really quite flexible.
 *
 */



#if !defined (_elog_api_h)
#define _elog_api_h


#if !defined (FS_STANDALONE)
#include "nfx-incl.h"
#endif

#include "elog-def.h"

#define SYSLOG_MSG_FILE  		"/var/agile/messages"
#define SYSLOG_MSG_FILE_ALL		"/var/agile/messages*"

/*------------------------------------------------------------------------------
 * Header of the message.
 *------------------------------------------------------------------------------
 */


#if !defined (SSC)

/* 
 * The generic Eagle logging macro
 * - Will first check if the logging function has to be invoked.
 *   The only reason is performance.
 * - Will add more generic info: line number, file_name. 
 * - Calls the e_log function
 *
 */

#define E_LOG(_class_, _severity_, _appl_id_, _app_type_, _cookie_, _code_,      \
              _parameters_)                                                      \
do {                                                                             \
    if ((elog_enabled) &&                                                        \
        (_severity_ <= elog_severity))                                           \
    {                                                                            \
        if ((_severity_ < LOG_DEBUG) ||                                          \
            ((elog_severity >= LOG_DEBUG) &&                                     \
             (_severity_ >= LOG_DEBUG) &&                                        \
             (_appl_id_ >= 0) &&                                                 \
             (_appl_id_ < ELOG_MAX_APP) &&                                       \
             (elog_appl_data[_appl_id_].severity >= _severity_) &&               \
             (elog_appl_data[_appl_id_].bitmap & (1 << _app_type_))))            \
       {                                                                         \
           char *the_msg ;                                                       \
           the_msg = e_log_var_pa  _parameters_ ;                                \
           e_log (_class_, _severity_, _appl_id_, _app_type_, _cookie_, _code_,  \
                  __LINE__, __FILE__, the_msg);                                  \
           eee_ramDealloc (the_msg);                                             \
       }                                                                         \
    }                                                                            \
} while (0)

/* 
 * An application specific elog macro
 * - Just given as an example
 * - Add the generic values which are always the same for this application
 * - Invokes the generic E_LOG macro
 *
 */

/*

#define NVRAM_TRACE(_app_type_ , _code_ , _parameters_)                 \
do {                                                                    \
    E_LOG(class_1, DEBUG_LOG, nvram_appl_id, _app_type_, 0, _code_,     \
          _parameters_) ;                                               \
} while (0)

*/


/*
 * external variables
 * accessible where the macro's are used.
 */

extern int elog_severity ;
extern int elog_enabled ;

typedef struct 
{
  log_severity_t severity;
  int32 bitmap;
} elog_appl_data_t;

#define ELOG_MAX_APP    64

extern elog_appl_data_t elog_appl_data [ELOG_MAX_APP] ;




/*
 *
 *	Embedded Programing Interface
 *
 */

/* 
 *   e_log function prototype
 */

int
e_log (log_class_t class,
       log_severity_t severity,
       uint32 app_id,
       uint32 app_type,
       uint32 cookie,
       uint32 code,
       int line,
       char* file,
       char *string);

#ifdef __GNUC__
char *e_log_var_pa (char *format, ...) __attribute__ ((format (printf, 1, 2)));
#else
char *e_log_var_pa (char *format, ...);
#endif

int
e_log_init(void);

/* 
 * examples
 *

Generic
=======
  Function call
  -------------

  e_log (class_1, LOG_DEBUG, nvram_appl_id, app_type, cookie, code, 
         __LINE__, __FILE__, "nvram started");
===> for var parameters: sprintf needed before calling this function

  Macro
  -----

  E_LOG (class_1, LOG_DEBUG, nvram_appl_id, app_type, cookie, code, 
         ("nvram started with id %d 0x%llx %lld  ", 1, cod, cod));
===> + check severity before function call
===> + line, file added automatically
===> + remark: brackets needed around the format string and var par.



Application specific
====================

  Macro
  -----

  NVRAM_E_LOG (LOG_DEBUG, app_type, code, 
               ("nvram started with id %d  ", 1));
===> + less parameters
===> + check severity before function call
===> + line, file added automatically
===> + remark: brackets needed around the format string and var par.



Another example to use the fs_trace
===================================

#define FS_TRACE(_app_type_ , _parameters_)                             \
  E_LOG(class_1, LOG_DEBUG, fs_appl_id, _app_type_, 0,                  \
        (int32) pthread_getspecific(fs_simKeyCpuId),                    \
        _parameters_) ;

As an example, change existing fs_trace:
    fs_trace(FS_TRACE_REQ, "vol %lld ino %lld truncating to len %lld",
        context->volId, inode->inum, newSize);
into
    FS_TRACE(FS_TRACE_REQ, ("vol %lld ino %lld truncating to len %lld",
        context->volId, inode->inum, newSize));

'Cookie' and 'code' has still to be defined to be usefull


Output sample
=============

DEBUG    Thu Sep  6 12:38:56 2001 (301 msec) 
    line: 1450, file ../sm-nvram/nvram.c 
    nbr: 4 
    nvram_read_from_file: read 66076 Bytes, 7 segments, 16 free 8 used blocks



 *
 */



#else // SSC


// We define the same macro on the SSC
// Only the global flags are not yet available 

#define E_LOG(_class_, _severity_, _appl_id_, _app_type_, _cookie_, _code_,   \
              _parameters_)                                                   \
do {                                                                          \
        char *the_msg ;                                                       \
        the_msg = e_log_var_pa  _parameters_ ;                                \
        e_log (_class_, _severity_, _appl_id_, _app_type_, _cookie_, _code_,  \
               __LINE__, __FILE__, the_msg);                                  \
        free (the_msg);                                                       \
} while (0)



char *
e_log_var_pa (char *format, ...);


int
e_log (log_class_t class,
       log_severity_t severity,
       uint32 app_id,
       uint32 app_type,
       uint32 cookie,
       uint32 code,
       int line,
       char* file,
       char *string);


#endif // SSC


/*
 * defines (also need for the configuration itf)
 */

#define ELOG_DISABLED         0
#define ELOG_ENABLED          1


// elog_show_date
//define ELOG_DISABLED         0
#define ELOG_SHOW_DATE        1
#define ELOG_SHOW_DATE_MSEC   2
#define ELOG_SHOW_DATE_USEC   3


// elog_format
#define ELOG_ONE_LINE         0 
#define ELOG_MULTIPLE_LINE    1
#define ELOG_FORMAT_MIN_INFO  2  // show only the msg, disable all the rest (one line)
#define ELOG_FORMAT_MAX_INFO  3  // show all info == enable all flags (multiple lines)  


// elog_show_line_file
//define ELOG_DISABLED         0
//define ELOG_ENABLED          1


// elog_show_version_string
//define ELOG_DISABLED         0
//define ELOG_ENABLED          1


// elog_show_sequence
//define ELOG_DISABLED         0
//define ELOG_ENABLED          1

// elog_show_local
//define ELOG_DISABLED         0
#define ELOG_LOCAL_STDOUT      1
#define ELOG_LOCAL_STDERR      2
#define ELOG_ONLY_LOCAL_STDOUT 3   // don't send to SSC
 

/*
 * The configuration interface
 */


typedef enum {
#if (FS_STANDALONE)
    ELOG_CONFIGURE_ENABLE,
#else
    ELOG_CONFIGURE_ENABLE = ELOG_MSG_BASE,
#endif
    ELOG_CONFIGURE_SEVERITY,
    ELOG_CONFIGURE_APP_DEFAULT_SEVERITY,
    ELOG_CONFIGURE_APP_DEFAULT_BITMAP,
    ELOG_CONFIGURE_APP_SEVERITY,
    ELOG_CONFIGURE_APP_SET_BITMAP,
    ELOG_CONFIGURE_APP_CLEAR_BITMAP,
    ELOG_CONFIGURE_SHOW_DATE,
    ELOG_CONFIGURE_FORMAT,
    ELOG_CONFIGURE_SHOW_LINE,
    ELOG_CONFIGURE_SHOW_SEQ,
    ELOG_CONFIGURE_SHOW_SEVERITY,
    ELOG_CONFIGURE_SHOW_APP_NAME,
    ELOG_CONFIGURE_SHOW_LOCAL,
	ELOG_CONFIGURE_INIT,
    ELOG_CONFIGURE_TEST
} elog_config_cmd_t ;
 

typedef struct 
{
    elog_config_cmd_t  command;
    int len ;
    int value_1;
    int value_2;
} e_log_configure_msg_t;

int
e_log_configure(e_log_configure_msg_t *msg, int size);

#endif  // _elog_api_h

