// vi:ts=4
/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : elog.c
 *
 * Description  : Logging manager (elog).
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *                Pankaj Gupta
 *
 * Date Created : 23-JUL-2001
 *                10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */


#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>

#include "nfx-incl.h"

#include "exports/elog-api.h"
#include "exports/anpssc-api.h"
#include "exports/opt-api.h"

/*------------------------------------------------------------------------------
 * Constants and macros.
 *------------------------------------------------------------------------------
 */

#define LOGGER_NAME ""
#define SYSLOG_CONF_TEMPLATE  	"/usr/local/agile/etc/syslog.conf"
#define SYSLOG_CONF 			"/etc/syslog.conf"
#define SYSLOG_PID 				"/var/run/syslog.pid"
#define SYSLOG_PROG 			"/usr/sbin/syslogd"

char *elogLevelStr[] = {
                        "EMERGENCY",
						"ALERT",
						"CRITICAL",
						"ERROR",
						"WARNING",
						"NOTICE",
						"INFO",
						"DEBUG"
						};

char *elogFacilityStr[] = {
                        "local0",
						"local1",
						"local2",
						"local3",
						"local4",
						"local5",
						"local6",
						"local7"
};


/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static int
elog_init(void);

static void
elog_terminate(int);

static int
elog_setConfig(char * option,
              char * value);

static int
elog_getConfig(char * option);

static int
elog_updateConfig(opt_entry_t * entry);


/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* ELOG configuration.
 */
elog_config_t        elogCfg;

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* Event manager's file descriptor.
 */
static int fd;

/* Wait for EVM to respond (10s).
 */
static struct timeval waitRegmTo = {10, 0};

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : setElogOptionForAllCpus()
 *
 * Description  : Forward elog option to all CPUs
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 11-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
setElogOptionForAllCpus(elog_config_cmd_t cmd, int value)
{
	e_log_configure_msg_t	reqMsg;
	uint32					reqLen = sizeof(e_log_configure_msg_t);

	/*
	 * Initialize and prepare the message
	 */
	bzero(&reqMsg, reqLen);
	reqMsg.command = cmd;
	reqMsg.len = sizeof(e_log_configure_msg_t);
	reqMsg.value_1 = value;
	reqMsg.value_2 = 0;

	/*
	 * FC, cpuid 3, 2, 1, 0 are slice 0 1 2 3 respectively
	 */
    sendAgileMsg(-1, &reqMsg, reqLen, "e_log", NFX_FCNIM_CPUID3, 2, 
						MGMTPLANE, 0, 0, 0);
    sendAgileMsg(-1, &reqMsg, reqLen, "e_log", NFX_FCNIM_CPUID2, 2, 
						MGMTPLANE, 0, 0, 0);
    sendAgileMsg(-1, &reqMsg, reqLen, "e_log", NFX_FCNIM_CPUID1, 2, 
						MGMTPLANE, 0, 0, 0);
    sendAgileMsg(-1, &reqMsg, reqLen, "e_log", NFX_FCNIM_CPUID0, 2, 
						MGMTPLANE, 0, 0, 0);

	/*
	 * TxRx
	 */
    sendAgileMsg(-1, &reqMsg, reqLen, "e_log", NFX_NFP_TXRX1, 1, 
						MGMTPLANE, 0, 0, 0);

	/*
	 * FP
	 */
    sendAgileMsg(-1, &reqMsg, reqLen, "e_log", NFX_NFP_FP_ANY, 1, 
						MGMTPLANE, 0, 0, 0);


	/*
	 * SSC-MGMT
	 */
    sendAgileMsg(-1, &reqMsg, reqLen, "e_log", NFX_SSC_MGMT, SSC_SLOT, 
						MGMTPLANE, 0, 0, 0);

}

/*------------------------------------------------------------------------------
 * Name         : elog_restartSyslogd()
 *
 * Description  : Restart the syslog daemon
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 13-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void 
elog_restartSyslogd(void)
{
	FILE *pidfile = fopen (SYSLOG_PID, "r");
	pid_t syslog_pid = 0;

	if (pidfile) {
		fscanf (pidfile, "%d", &syslog_pid);
		fclose (pidfile);
	}

	if (syslog_pid) {
		kill (syslog_pid, SIGHUP);
	}
}
 
/*------------------------------------------------------------------------------
 * Name         : elog_updateSyslogConf()
 *
 * Description  : Update the /etc/syslog.conf file
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 13-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
elog_updateSyslogConf(void)
{
	FILE *templ;
	FILE *conf;
	char buf[256];
	uint32	facilityIdx = 0;

	templ = fopen (SYSLOG_CONF_TEMPLATE, "r");
	conf = fopen (SYSLOG_CONF, "w");

	if (!conf) {
		fprintf (stderr, "Could not open ELOG conf file %s\n", SYSLOG_CONF);
		return;
	}

	if (!templ) {
		fprintf (stderr, "Could not open ELOG template file %s\n", 
						SYSLOG_CONF_TEMPLATE);
	}
	else
	{
		while (!feof (templ)) {
			fgets (buf, sizeof(buf), templ);
			fputs (buf, conf);
		}

		fclose (templ);
	}

	/*
	 * Add Agile specific parameters
	 * and close the actual conf file
	 */

	/*
	 * facility.priority      SYSLOG_MSG_FILE
	 */
	facilityIdx = LOG_FAC(elogCfg.facility) - LOG_FAC(LOG_LOCAL0);
	strcpy(buf, elogFacilityStr[facilityIdx]);

	if (elogCfg.hostIpa == 0)
	{
		int fud;

		strcat(buf, ".info\t\t");	// If CompactFlash, only info and above

		/*
		 * Store syslog locally
		 */
		strcat(buf, SYSLOG_MSG_FILE);

		/*
		 * Make sure the file exists so syslog will be able to use it
		 */
		fud = open (SYSLOG_MSG_FILE, O_CREAT|O_RDWR, 0644);
		close (fud);
	}
	else
	{
		struct in_addr ip_addr;

		ip_addr.s_addr = elogCfg.hostIpa;
		strcat(buf, ".*\t\t");
		strcat(buf, "@");
		strcat(buf, (char *)inet_ntoa(ip_addr));
	}

	strcat(buf, "\n");
	fputs (buf, conf);

	/*
	 * facility.priority      /dev/console
	 */

	if (elogCfg.display == TRUE)
	{
		strcpy(buf, elogFacilityStr[facilityIdx]);
		strcat(buf, ".*\t\t");
		strcat(buf, "/dev/console");
		strcat(buf, "\n");
		fputs (buf, conf);
	}

	fclose (conf);

	/*
	 * Now, tell syslogd we've changed its configuration
 	 */
	elog_restartSyslogd();
}

/*------------------------------------------------------------------------------
 * Name         : elog_init()
 *
 * Description  : Initialize the event manager.
 *
 * Created by   : Isabelle
 *
 * Date Created : 10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
elog_init(void)

{
    ELOG_FCT_PRINT(("> elog_init()\n"));
    
    /* 
	 * Sets the procedure to follow in case of request for program termination
     * or CTRL-C interrupt.
     */
    signal(SIGTERM, elog_terminate);
    signal(SIGINT, elog_terminate);

    /* 
	 * Register to receive messages.
     */
	fd = regAgileApp("elog");
	if (fd < 0)
    {
        perror("regAgileApp() for ELOG failed");
		return(NFX_ERR);
	}

    /* 
	 * Open a connection to the system logger.
     */
	openlog(LOGGER_NAME, LOG_NDELAY, LOG_LOCAL0);

    /* 
	 * Set the configuration to default.
     */
    elogCfg.enable = TRUE;
    elogCfg.display = FALSE;
    elogCfg.levelDefault = err_s;
	elogCfg.hostIpa = 0;
	elogCfg.facility = LOG_LOCAL0;
        
	/*
	 * Start it off using this default config because we don't know what's
	 * in the syslog.conf file. It won't get changed in via elog_getConfig()
	 * unless there are entries in agile.conf In that case, the above 
	 * "default" configuration won't get enforced.
	 */
	modify_readonly_file(SYSLOG_CONF, elog_updateSyslogConf, NULL);

    /* 
	 * Get the configuration of ELOG and autosupport.
     */
    elog_getConfig(OPTIONS_ELOG_ALL);

    ELOG_FCT_PRINT(("< elog_init()\n"));
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : elog_terminate()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
elog_terminate(int sig)

{
    ELOG_FCT_PRINT(("> elog_terminate()\n"));

    /* Unregister REGM.
     */
    unRegAgileApp(fd);

    /* Close the descriptor being used to write the system logger.
     */
    closelog();
    
    ELOG_FCT_PRINT(("< elog_terminateRegistryMgr()\n"));
    exit(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : elog_setConfig()
 *
 * Description  : Set the option in the registry.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 13-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
elog_setConfig(char * option,
              char * value)
    
{
    char                 req[sizeof(opt_msgHdr_t) + sizeof(opt_setOptionReq_t)];
    opt_msgHdr_t *       reqHdr;
    opt_setOptionReq_t * reqBody;
    char                 rsp[sizeof(opt_msgHdr_t) + sizeof(opt_setOptionRsp_t)];
    opt_msgHdr_t *       rspHdr;
    opt_setOptionRsp_t * rspBody;
    
    ELOG_DBG_PRINT(("> elog_setConfig()\n"));

    /*
     */
    bzero((void *) &req, sizeof(req));
    reqHdr = (opt_msgHdr_t *) &req;
    reqBody = (opt_setOptionReq_t *) (req + sizeof(opt_msgHdr_t));
    bzero((void *) &rsp, sizeof(rsp));
    rspHdr = (opt_msgHdr_t *) &rsp;
    rspBody = (opt_setOptionRsp_t *) (rsp + sizeof(opt_msgHdr_t));

    /* Fill the request.
     */
    reqHdr->msgType = OPT_SET_OPTION_REQ;
    reqHdr->userInfo = 0; // not used    
    reqBody->del = FALSE;
    strcpy(reqBody->entry.option, option);
    strcpy(reqBody->entry.value, value);
    
    /* Send the request to the registry manager located on the SSC.
     * Receive response from the registry manager.
     */
    if (sendAgileMsg(-1,
                     req, sizeof(req),
                     "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                     &waitRegmTo,
                     rsp, sizeof(rsp)) == -1)
    {
		perror("sendAgileMsg() failed");
		return(NFX_ERR);
    }
    
    /* Read the message header and check if it's expected response.
     */
    rspHdr = (opt_msgHdr_t *) rsp;
    if (rspHdr->msgType != OPT_SET_OPTION_RSP)
    {
        ELOG_ERR_PRINT(("Message received from REGM is not the expected one!\n"));
		return(NFX_ERR);
    }
    
    /* Read result.
     */
    rspBody = (opt_setOptionRsp_t *) (rsp + sizeof(opt_msgHdr_t));
    ELOG_DBG_PRINT(("result returned = %d\n", rspBody->result));
    
    ELOG_DBG_PRINT(("< elog_setConfig()\n");   )
    return(rspBody->result);
}

/*------------------------------------------------------------------------------
 * Name         : elog_getConfig()
 *
 * Description  : Get configuration from registry manager.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
elog_getConfig(char * option)
{
    opt_msgType_t   reqType = OPT_GET_OPTION_REQ;
    elog_msg_t       req;
    elog_msg_t       rsp;
    opt_msgHdr_t *  hdr;
    void *          bufp = NULL;
    int             nbEntry = 0;
    int             i;

    ELOG_FCT_PRINT(("> elog_getConfig\n"));

    /* Fill -a %sthe request.
     */
    bufp = &req.buf;
    bzero(bufp,
          (sizeof(opt_msgHdr_t) +
           sizeof(opt_getOptionReq_t)));
    
    memcpy(bufp, (void *) &reqType, sizeof(opt_msgType_t));
    bufp += sizeof(opt_msgType_t);
    // userInfo is not used
    bufp += sizeof(opt_userInfo_t);
    memcpy(bufp, (void *) option, sizeof(opt_option_t));
    bufp += sizeof(opt_option_t);

    req.len = bufp - (void *) req.buf;

    /* Send the request to the registry manager located on the SSC.
     */
    rsp.len = sendAgileMsg(-1,
                           req.buf, req.len,
                           "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &waitRegmTo,
                           rsp.buf, sizeof(rsp.buf));
    
    /* Receive response from the registry manager.
     */
    ELOG_DBG_PRINT(("nb bytes received = %d\n", rsp.len));
    if (rsp.len == -1)
    {
		perror("sendAgileMsg failed");
		return(NFX_ERR);
    }

    /* Read the message header and check if it's expected response.
     */
    bufp = &rsp.buf;
    hdr = (opt_msgHdr_t *) rsp.buf;
    bufp += sizeof(opt_msgHdr_t);
    if (hdr->msgType != OPT_GET_OPTION_RSP)
    {
        ELOG_ERR_PRINT(("Message received from REGM is not the expected one!\n"));
		return(NFX_ERR);
    }
    
    /* Read the response and print its content.
     * XXX
     */
    memcpy((void *) &nbEntry, bufp, sizeof(int));
    bufp += sizeof(int);

    for (i = 0; i < nbEntry; i++)
    {
        opt_entry_t * entry;

        entry = (opt_entry_t *) bufp;
        bufp += sizeof(opt_entry_t);

        elog_updateConfig(entry);
    } // for
    
    ELOG_FCT_PRINT(("< elog_getConfig()\n"));
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : elog_updateConfig()
 *
 * Description  :
 * Update local configuration with info received from the registry manager.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
elog_updateConfig(opt_entry_t * entry)
    
{
    ELOG_FCT_PRINT(("> elog_updateConfig\n"));

    if (!strcmp(entry->option, OPTIONS_ELOG_ENABLE))
    {
        elogCfg.enable = (!strcmp(entry->value, "on")? TRUE : FALSE);
		setElogOptionForAllCpus(ELOG_CONFIGURE_ENABLE, elogCfg.enable);
    }
    else if (!strcmp(entry->option, OPTIONS_ELOG_DISPLAY))
    {
		uint32 oldDisplay = elogCfg.display;

        elogCfg.display = (!strcmp(entry->value, "on")? TRUE : FALSE);
		setElogOptionForAllCpus(ELOG_CONFIGURE_SHOW_LOCAL, elogCfg.display);

		/*
		 * Update the /etc/syslog.conf and 
		 * restart the syslogd
		 */
		if (oldDisplay != elogCfg.display)
		{
			modify_readonly_file(SYSLOG_CONF, elog_updateSyslogConf, NULL);
		}
    }
    else if (!strcmp(entry->option, OPTIONS_ELOG_FACILITY))
	{
		uint32 oldFacility = elogCfg.facility;

		if (!strcmp(entry->value, "local0"))
		{
			elogCfg.facility = LOG_LOCAL0;
		}
		else if (!strcmp(entry->value, "local1"))
		{
			elogCfg.facility = LOG_LOCAL1;
		}
		else if (!strcmp(entry->value, "local2"))
		{
			elogCfg.facility = LOG_LOCAL2;
		}
		else if (!strcmp(entry->value, "local3"))
		{
			elogCfg.facility = LOG_LOCAL3;
		}
		else if (!strcmp(entry->value, "local4"))
		{
			elogCfg.facility = LOG_LOCAL4;
		}
		else if (!strcmp(entry->value, "local5"))
		{
			elogCfg.facility = LOG_LOCAL5;
		}
		else if (!strcmp(entry->value, "local6"))
		{
			elogCfg.facility = LOG_LOCAL6;
		}
		else if (!strcmp(entry->value, "local7"))
		{
			elogCfg.facility = LOG_LOCAL7;
		}
		else 
		{
            ELOG_ERR_PRINT(("Wrong facility (%s)!\n", entry->option));
		}

		/*
		 * Update the /etc/syslog.conf and 
		 * restart the syslogd
		 */
		if (oldFacility != elogCfg.facility)
		{
			modify_readonly_file(SYSLOG_CONF, elog_updateSyslogConf, NULL);
		}
	}
    else if (!strcmp(entry->option, OPTIONS_ELOG_HOST))
	{
		uint32 oldIpa = elogCfg.hostIpa;
		elogCfg.hostIpa = inet_addr(entry->value);

		/*
		 * Update the /etc/syslog.conf and 
		 * restart the syslogd
		 */
		if (oldIpa != elogCfg.hostIpa)
		{
			modify_readonly_file(SYSLOG_CONF, elog_updateSyslogConf, NULL);
		}
	}
    else if (!strcmp(entry->option, OPTIONS_ELOG_LEVEL))
    {
		uint32 oldLevel = elogCfg.levelDefault;

        if (!strcmp(entry->value, "debug"))
        {
            elogCfg.levelDefault = debug_s;
        }
        else if (!strcmp(entry->value, "info"))
        {
            elogCfg.levelDefault = info_s;
        }
        else if (!strcmp(entry->value, "notice"))
        {
            elogCfg.levelDefault = notice_s;
        }
        else if (!strcmp(entry->value, "warning") ||
                 !strcmp(entry->value, "warn"))
        {
            elogCfg.levelDefault = warning_s;
        }
        else if (!strcmp(entry->value, "error") ||
                 !strcmp(entry->value, "err"))
        {
            elogCfg.levelDefault = err_s;
        }
        else if (!strcmp(entry->value, "critical"))
        {
            elogCfg.levelDefault = crit_s;
        }
        else if (!strcmp(entry->value, "alert"))
        {
            elogCfg.levelDefault = alert_s;
        }
        else if (!strcmp(entry->value, "emergency") ||
                 !strcmp(entry->value, "panic"))
        {
            elogCfg.levelDefault = emerg_s;
        }
        else
        {
            ELOG_ERR_PRINT(("Wrong level (%s)!\n", entry->option));
        }   

		setElogOptionForAllCpus(ELOG_CONFIGURE_SEVERITY,
						elogCfg.levelDefault);

		/*
		 * Update the /etc/syslog.conf and 
		 * restart the syslogd
		 */
		if (oldLevel != elogCfg.levelDefault)
		{
			modify_readonly_file(SYSLOG_CONF, elog_updateSyslogConf, NULL);
		}
    }
    else
    {
        ELOG_ERR_PRINT(("Invalid ELOG option requested\n"));
    }
    
    ELOG_FCT_PRINT(("< elog_updateConfig()\n"));
    return(NFX_OK);
}



/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : main
 *
 * Description  : Where it all begins
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 23-JUL-2001
 *                10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int
main(void) 
{
	char    senderName[256];
	uint32  cpu;
	uint32  slot;
	int 	rc;
    
    /* Initialize the event manager.
     */
	if (elog_init() != NFX_OK)
    	exit(NFX_ERR);

	 /* Tell MCP we're ready to roll
	  */
	kill (getppid(), SIGUSR1);

    /* Gather messages and format syslog strings
     */
	while (TRUE)
    {
        elog_msg_t   msg;
        
        /* Wait for messages...
         */
		msg.len = recvAgileMsg(fd,
                               msg.buf, sizeof(msg.buf),
                               NULL, 
                               senderName, &cpu, &slot, NULL);
        //ELOG_DBG_PRINT(("msg.len = %d\n", msg.len));
        
        /* Something has been received, check it.
         */
		if (msg.len > 0)
        {
            elog_msgHdr_t *  hdr;

            hdr = (elog_msgHdr_t *) msg.buf;
            
            /* Process the message according to its type.
             */
            switch (hdr->msgType)
            {
            case ELOG_MSG_NOTIFY:
            {
                elog_eventNtfy_t * logmsg = (elog_eventNtfy_t *) msg.buf;

                /* 
				 * Filter the logs.
                 */
                if ((elogCfg.enable == TRUE) &&
                    (logmsg->severity <= elogCfg.levelDefault))
                {
                    /* 
					 * Send to syslog daemon.
                     */
                    syslog(elogCfg.facility|logmsg->severity,
                           "%d:%d:%s:%s: %s", slot, cpu, 
						   	eee_getAppNameFromId(logmsg->appId), 
						   	elogLevelStr[logmsg->severity],
							logmsg->amsg);
                }
                break;
            }

            case ELOG_CONFIG_NTFY:
            {
                elog_configNtfy_t * config;

                config = (elog_configNtfy_t *) msg.buf;
                ELOG_DBG_PRINT(("entry = %s\n", config->entry.option));
                ELOG_DBG_PRINT(("value = %s\n", config->entry.value));
                elog_updateConfig((opt_entry_t*) &(config->entry));
                break;
            }
            
			case ELOG_CONFIG_SET_REQ:
			{
				elog_configNtfy_t * config;
				config = (elog_configNtfy_t *) msg.buf;

				/*
				 * Update local config
				 */
				elog_updateConfig((opt_entry_t*) &(config->entry));

				/*
				 * Update config in options file with registry manager
				 */
				elog_setConfig(config->entry.option, config->entry.value);

				/*
				 * TBD: do we need to update config for all
				 * modules on embedded side too
				 */
			}

			case ELOG_CONFIG_VIEW_REQ:
			{
				/*
				 * Send back the config block
				 */
				rc = sendAgileMsg(-1, &elogCfg, sizeof(elog_config_t),
							senderName, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							NULL, NULL, 0);

				if (rc != NFX_OK)
					printf("Could not return ELOG config\n");

			}
			break;

			case ELOG_CONFIG_GET_REQ:
			{
				e_log_configure_msg_t msg;
				msg.command = ELOG_CONFIGURE_INIT;
				msg.len = sizeof(msg);
				msg.value_1 = elogCfg.enable;
				msg.value_2 = elogCfg.levelDefault;
				if (sendAgileMsg(-1, &msg, sizeof(msg),
								 senderName, cpu, slot, MGMTPLANE,
								 NULL, NULL, 0) < 0)
					perror("sendAgileMsg");
			}
			break;

            default:

            } // switch
		}
	} // while

    exit(NFX_ERR);
}

/*----------------------------------------------------------------------------*/
