/*
 *-----------------------------------------------------------------
 *
 * Name:	esm.c
 *
 * RCS:  $Id: esm.c,v 1.53 2002/12/03 21:59:53 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Eagle State Machine scheduler
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/esm-api.h>
#include <exports/queue-api.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>
#include <exports/eee-poll.h>
#include <exports/debug.h>
#include <exports/timer-api.h>

#include "esm.h"

// limit processing time during poll 
#define ESM_MAX_CPU_TIME 10000000

// #define ESM_USE_BUCKETS

#ifdef ESM_USE_BUCKETS
#define ESM_TIME_STATS_BUCKETS_MAX  6
uint64 esm_action_stats[ESM_TIME_STATS_BUCKETS_MAX];
uint64 esm_poll_stats[ESM_TIME_STATS_BUCKETS_MAX];
int32 esm_stats_usec[ESM_TIME_STATS_BUCKETS_MAX] =
    { 1000, 5000, 10000, 15000, 20000, 50000 };
char *esm_stats_usec_name[ESM_TIME_STATS_BUCKETS_MAX] =
    { "1000", "5000", "10000", "15000", "20000", "50000" };
#endif

// Batch allocate esm_instance_t's up to some max
// after which we'll stop creating new esm instances.
#define ESM_INSTANCE_MAX        (8 * 1024)
#define ESM_INSTANCE_ALLOC      256

// Batch allocate esm_event_qelem_t's up to some max
// after which we'll stop enqueueing new events.
#define ESM_EVENT_MAX           (64 * 1024)
#define ESM_EVENT_ALLOC         1024

esm_instance_t                  *esm_currInst CPU_PRIVATE;
esm_instance_t                  **esm_instance_ptrs = NULL;
int32                           esm_instance_curr_size = 0;
int32                           esm_event_curr_size = 0;
ushort16                        esm_generation = 1;
nfx_spinlock_t                  esm_spinlock;
DLL_HDR(esm_instance_t)         esm_freeInstHdr;
DLL_HDR(esm_event_qelem_t)      esm_freeEventHdr;

// TODO - use elog for tracing.
#define ESM_TRACE

#if defined(LINUX_TEST)
#define ESM_READ_POLL_IND(x)  esm_readPollInd(x)
#define ESM_WRITE_POLL_IND(x) esm_writePollInd(x)
#else
#define ESM_READ_POLL_IND(x)
#define ESM_WRITE_POLL_IND(x)
#endif

typedef struct esm_ready_queue
{
    DLL_HDR(esm_instance_t)   dllHdr;
#ifdef ESM_USE_BUCKETS
    uint64                    stats[ESM_TIME_STATS_BUCKETS_MAX];
#endif
    int32                     maxPerPoll;
    uint32                    esm_rxHWM;
} esm_ready_queue_t;

static esm_ready_queue_t esm_readyQueue[ESM_QUEUE_PRI_MAX];

#if defined(NFP_FP) // fp threads are hogs
enum
{
    ESM_QUEUE_MAX_POLL_LO = 16,
    ESM_QUEUE_MAX_POLL_HI = 32,
};
#else
enum
{
    ESM_QUEUE_MAX_POLL_LO = 16,
    ESM_QUEUE_MAX_POLL_HI = 32,
};
#endif


/*------------------------------------------------------------------
 * Create an esm handle, disallow 0.
 */
#define ESM_CREATE_HANDLE(_handlep_, _index_)           \
    esm_generation++;                                   \
    if (esm_generation == 0) esm_generation = 1;        \
	(_handlep_) = ((esm_generation) << 16) | (_index_)

void
esm_processEventQueue(
	void	* cb,
	uint32	tref
);


/*-----------------------------------------------------------------
 * Name:	esm_extendInstances()
 *
 * Description:	allocate additional instance structures
 *
 *                 !!! must be called with spinlock held !!!
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_extendInstances()
{
    esm_instance_t  **old_ptrs, **new_ptrs, **ptr, *new_instances;
    int32           i;

    //ESM_TRACE("extending esm instances from %d\n", esm_instance_curr_size);

    // All the monkey business here is intended to provide for
    // an easy way to index the batch-allocated instances so that
    // we dont need to conduct any lengthy scans in order to find an
    // instance from a handle.  Basically, we're maintaining an array of
    // pointers to instances, but the array can grow, so this is the
    // code that allocates new instances and increases and sets up 
    // the array of pointers to them.

    if (esm_instance_curr_size >= ESM_INSTANCE_MAX)
    {
        // TODO - elog
        printf("max esm instances exceeded\n");
        return;
    }
    old_ptrs = esm_instance_ptrs;

    // Allocate new instances and a new pointer array.
    new_instances = eee_ramAlloc(sizeof(esm_instance_t) * ESM_INSTANCE_ALLOC);
    new_ptrs = eee_ramAlloc(sizeof(esm_instance_t *) * (ESM_INSTANCE_ALLOC +
                                                          esm_instance_curr_size));
    if (!new_instances || !new_ptrs)
    {
        // TODO - elog
        printf("no memory for new esm instances\n");
        if (new_instances)
            eee_ramDealloc(new_instances);
        if (new_ptrs)
            eee_ramDealloc(new_ptrs);
        return;
    }
    // Copy the old pointers to the new space.
    if (esm_instance_curr_size)
    {
        memcpy(new_ptrs, old_ptrs,
                    sizeof(esm_instance_t *) * esm_instance_curr_size);
    }
    ptr = new_ptrs + esm_instance_curr_size;
    for (i = 0; i < ESM_INSTANCE_ALLOC; i++)
    {
        // Update the pointer array.
        *(ptr + i) = new_instances + i;
        // Update the new instance struct.
        (new_instances + i)->index = esm_instance_curr_size + i;
        // Add the new instance struct to the free list.
		DLL_ISOLATE_ELEM((new_instances + i), queueEntry);
        DLL_INSERT_HEAD(&esm_freeInstHdr, (new_instances + i), queueEntry);
        // Invalidate the handle in the new instance.
        (new_instances + i)->handle = 0;
    }
    esm_instance_curr_size += ESM_INSTANCE_ALLOC;
    esm_instance_ptrs = new_ptrs;

    if (old_ptrs)
    {
        eee_ramDealloc(old_ptrs);
    }

} /* esm_extendInstances */

/*-----------------------------------------------------------------
 * Name:	esm_extendEvents
 *
 * Description:	allocates a batch of esm_event_qelem_t's for the
 *              free list.
 *
 *                 !!! must be called with spinlock held !!!
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_extendEvents()
{
    esm_event_qelem_t   *qelem;
    int32               i;

    //ESM_TRACE("extending esm events from %d\n", esm_event_curr_size);

    if (esm_event_curr_size >= ESM_EVENT_MAX)
    {
        // TODO - elog
        printf("maximum outstanding esm event limit reached (%d)\n",
               ESM_EVENT_MAX);
        return;
    }

    qelem = eee_ramAlloc(sizeof(esm_event_qelem_t) * ESM_EVENT_ALLOC);
    if (qelem == NULL)
    {
        // TODO - elog
        printf("no memory for extending esm event list\n");
        return;
    }

    for (i = 0; i < ESM_EVENT_ALLOC; i++)
    {
		DLL_ISOLATE_ELEM(qelem, dllEntry);
        DLL_INSERT_TAIL(&esm_freeEventHdr, qelem, dllEntry);
        qelem += 1;
    }

    esm_event_curr_size += ESM_EVENT_ALLOC;

} /* esm_extendEvents */

/*-----------------------------------------------------------------
 * Name:	esm_initModule()
 *
 * Description:	init function for state machine engine module
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_initModule()
{
    int32   i;

    // TODO - needs to be called only by the boot cpu for smp

    NFX_SPIN_INIT(&esm_spinlock, SPINLOCK_LEVEL_ESM);

#if defined(USE_THREAD_STACKS) || defined(USE_BSS_THREAD_STACKS)
	esm_stackInit();
#endif

    for (i = 0; i < ESM_QUEUE_PRI_MAX; i++)
    {
        DLL_INIT(&esm_readyQueue[i].dllHdr);
    }
    esm_readyQueue[ESM_QUEUE_PRI_HI].maxPerPoll = ESM_QUEUE_MAX_POLL_HI;
    esm_readyQueue[ESM_QUEUE_PRI_LO].maxPerPoll = ESM_QUEUE_MAX_POLL_LO;

    DLL_INIT(&esm_freeInstHdr);
    DLL_INIT(&esm_freeEventHdr);

    NFX_SPIN_LOCK(&esm_spinlock);
    esm_extendInstances();
    esm_extendEvents();
    NFX_SPIN_UNLOCK(&esm_spinlock);

    esm_initPlatform();

    eee_registerPollingFunc(EEE_HIGH_PRIORITY, esm_processEventQueue, NULL);

} /* esm_initModule */

/*-----------------------------------------------------------------
 * Name:	esm_getInstFromHandle()
 *
 * Description:	Find the ESM instance for the given handle.
 *
 * !!! Must be called with esm_spinlock held !!!
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
esm_instance_t *
esm_getInstFromHandle(
	esm_handle_t    handle)
{
    esm_instance_t  *esm;
	int32           index;

    if (!handle)
    {
		printf("Invalid handle, 0x%llx\n", (uint64)handle);
        return NULL;
    }

	index = ESM_GET_HANDLE_INDEX(handle);

	if (index >= esm_instance_curr_size)
    {
        // TODO - elog
		printf("Invalid handle, 0x%llx\n", (uint64)handle);
		return NULL;
	}
    esm = *(esm_instance_ptrs + index);
    if (esm->index == index &&
        esm->handle == handle)
    {
        return esm;
    }

    //ESM_TRACE("stale handle %x (esm->index %d, esm->handle %x)\n",
    //    handle, esm->index, esm->handle);

    return NULL;

}	/* esm_getInstFromHandle() */

/*-----------------------------------------------------------------
 * Name:	esm_addToReadyQueue()
 *
 * Description:	Adds the esm to the tail of the ready queue if
 *              its not already there and not busy.
 *
 *                 !!! must be called with spinlock held !!!
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
static void
esm_addToReadyQueue(
    esm_instance_t  *esm,
    int32           queuePriority)
{
    if ((esm->flags & ESM_IS_BUSY) == 0 &&
        DLL_IS_ISOLATED(esm, queueEntry))
    {
        DLL_INSERT_TAIL(&esm_readyQueue[queuePriority].dllHdr, esm, queueEntry);
    }
}

/*-----------------------------------------------------------------
 * Name:	esm_runEvent()
 *
 * Description:	state machine engine
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_runEvent(
	esm_instance_t		*esm,
	esm_event_t			event)
{
	int32				newState;
	esm_actionFunc_t  	***evFunArr;
	esm_actionFunc_t  	**evFun;
	esm_actionFunc_t  	*fun;

    esm_currInst = esm;
    if (esm->flags & ESM_CAN_YIELD)
    {
        esm_runThreadEvent(esm, event);
    }
    else
    {
        /*
	     * Call the action function appropriate for this event
	     * based on the current state.
         */
        evFunArr = esm->table->eventVectors;
        evFun = evFunArr[event.event];
        fun = evFun[esm->currentState];

        newState = fun->func(	esm->handle,
						        fun->nextState,
						        event,
						        esm->userInfo);
        
        esm->currentState = newState;
    }
    esm_currInst = NULL;

}	/* esm_runEvent() */


/*-----------------------------------------------------------------
 * Name:    esm_getCurrHandle
 *
 * Description: Return the handle for the currently running state
 *              machine.
 *
 * Created by:  Bob Fozard
 *
 * Date Created:    3/20/01
 *
 *-----------------------------------------------------------------
 */
esm_handle_t
esm_getCurrHandle()
{
    if (esm_currInst)
        return esm_currInst->handle;
    else
        return NULL;

} /* esm_getCurrHandle */


/*-----------------------------------------------------------------
 * Name:	esm_getCurrState
 *
 * Description:	Return the current state for the specified state
 *              machine.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_getCurrState(
    esm_handle_t   handle)
{
    esm_instance_t  *esm;
    int32           currState = NFX_ERR;

    NFX_SPIN_LOCK(&esm_spinlock);
    esm = esm_getInstFromHandle(handle);
    if (esm)
    {
        currState = esm->currentState;
    }
    NFX_SPIN_UNLOCK(&esm_spinlock);

    return currState;

} /* esm_getCurrState */


/*-----------------------------------------------------------------
 * Name:    esm_getUserInfo
 *
 * Description: Return the the user info for the specified state machine.
 *              Returns NULL if the state machine handle is invalid.
 * 
 * Created by:  Paul Haahr
 * 
 * Date Created:    9/13/01
 *
 *-----------------------------------------------------------------
 */
void *
esm_getUserInfo(
    esm_handle_t   handle)
{
    esm_instance_t  *esm;
    void            *userInfo = NULL;

    NFX_SPIN_LOCK(&esm_spinlock);
    esm = esm_getInstFromHandle(handle);
    if (esm)
    {
        userInfo = esm->userInfo;
    }
    NFX_SPIN_UNLOCK(&esm_spinlock);

    return userInfo;

} /* esm_getUserInfo */

/*-----------------------------------------------------------------
 * Name:	esm_enqueueEvent
 *
 * Description:	Adds an event to the event queue of a state machine instance
 *
 *                 !!! must be called with spinlock held !!!
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_enqueueEvent(
	esm_instance_t 		*esm,
	esm_event_t			event,
    int32               queuePriority)
{
    esm_event_qelem_t   *qelem;

    if (esm->qcount >= esm->table->maxQueueDepth)
    {
        // TODO - elog
        printf("max queue depth (%d) exceeded for esm handle 0x%llx name %s\n",
               esm->table->maxQueueDepth, (uint64)esm->handle, esm->table->name);
        return NFX_ERR;
    }

    DLL_REMOVE_HEAD(&esm_freeEventHdr, qelem, dllEntry);
    if (qelem == NULL)
    {
        esm_extendEvents();
        DLL_REMOVE_HEAD(&esm_freeEventHdr, qelem, dllEntry);
        if (qelem == NULL)
        {
            // TODO - elog
            printf("no memory for event queue for handle 0x%llx name %s\n",
                   (uint64)esm->handle, esm->table->name);
            return NFX_ERR;
        }
    }

    qelem->event = event;
#ifdef ESM_USE_BUCKETS
    eee_rtcStart(&qelem->qTime);
#endif
    DLL_INSERT_TAIL(&esm->eventQueueHdr, qelem, dllEntry);

	esm->qcount += 1;
    if (esm->qcount == 1)
    {
        esm_addToReadyQueue(esm, queuePriority);
    }

	return NFX_OK;

}	/* esm_enqueueEvent() */

/*-----------------------------------------------------------------
 * Name:	esm_deliverEvent()
 *
 * Description:	state machine event delivery
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
static int32
esm_doDeliverEvent(
	esm_handle_t handle,
	esm_event_t	 *event,
    int32        queuePriority)
{
	esm_instance_t  	*esm;
	int32				rc = NFX_ERR;

    NFX_SPIN_LOCK(&esm_spinlock);
	esm = esm_getInstFromHandle(handle);
	if (!esm)
	{
        // TODO - elog
		printf("esm_deliverEvent: stale handle 0x%llx\n",
			   (uint64)((uintptr_t)handle));
	}
    else
    {
        if (esm->flags & ESM_DESTROY)
        {
            // esm is pending destruction, don't
            // enqueue new events.
            // TODO - elog
    		NFX_SPIN_UNLOCK(&esm_spinlock);
			return NFX_ERR;
        }

		if ((!esm->table->canYield &&
             event->event >= esm->table->numEvents)  ||
            (esm->table->canYield &&
             event->event >= ESM_THREAD_EVENT_USER_MAX))
		{
            // TODO - elog
	    	printf(
	"esm_deliverEvent: event %d exceeds maxEvent %d for this esm type\n",
				event->event, esm->table->numEvents - 1);
    		NFX_SPIN_UNLOCK(&esm_spinlock);
			return NFX_ERR;
		}

    	/* 
	     * Enqueue event.
	     */
	   	rc = esm_enqueueEvent(esm, *event, queuePriority);
	   	if (rc != NFX_OK)
	   	{
            // TODO - elog
	   		printf("esm_deliverEvent: enqueue failed (rc: %d)\n", rc);
	   	}
        else
        {
            ESM_WRITE_POLL_IND(1);
        }
    }
    NFX_SPIN_UNLOCK(&esm_spinlock);
    
	return rc;

}	/* esm_deliverEvent() */

/*-----------------------------------------------------------------
 * Name:	esm_deliverHighPriorityEvent()
 *
 * Description:	state machine event delivery for high priority
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_deliverHighPriorityEvent(
	esm_handle_t handle,
	esm_event_t	 *event)
{
    return esm_doDeliverEvent(handle, event, ESM_QUEUE_PRI_HI);
}

/*-----------------------------------------------------------------
 * Name:	esm_deliverEvent()
 *
 * Description:	state machine event delivery
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_deliverEvent(
	esm_handle_t handle,
	esm_event_t	 *event)
{
    return esm_doDeliverEvent(handle, event, ESM_QUEUE_PRI_LO);
}

/*-----------------------------------------------------------------
 * Name:	esm_dequeueEvent()
 *
 * Description:	return the next queued event to be processed for
 *              the specified esm
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
esm_event_qelem_t *
esm_dequeueEvent(
    esm_instance_t  *esm)
{
    esm_event_qelem_t     *qelem;

    DLL_REMOVE_HEAD(&esm->eventQueueHdr, qelem, dllEntry);
    esm->qcount -= 1;

    return qelem;

} /* esm_dequeueEvent */

/*-----------------------------------------------------------------
 * Name:	esm_doProcessEventQueue()
 *
 * Description:	Process the specified event queue.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_doProcessEventQueue(
	void	* cb,
	uint32	tref,
    int32   queuePriority)
{
    esm_ready_queue_t   *readyQueue = &esm_readyQueue[queuePriority];
	esm_instance_t 		*esm;
    esm_event_qelem_t   *event_qelem;
    esm_event_t         event;
    int32               numEvents;
    eeeRTC_t            time, delta, maxTime;
    boolean             timeExpired = FALSE;
#ifdef ESM_USE_BUCKETS
	int32				i;
	eeeRTC_t			timer1, timer2, elapsed;
	static boolean		first = TRUE;
	static eeeRTC_t		esm_stats_rtc[ESM_TIME_STATS_BUCKETS_MAX];

	if (first) {
		first = FALSE;
		// Get the stats boundaries in rtc form
		for (i=0; i<ESM_TIME_STATS_BUCKETS_MAX; i++) {
			eee_usecToRTC((uint64)esm_stats_usec[i], &(esm_stats_rtc[i]));
		}
	}
	eee_rtcStart(&timer1);
#endif

    maxTime.val = ESM_MAX_CPU_TIME;
    eee_rtc(&time);

    NFX_SPIN_LOCK(&esm_spinlock);
    for (numEvents = 0; numEvents < readyQueue->maxPerPoll; numEvents++)
    {
        if (DLL_IS_EMPTY(&readyQueue->dllHdr))
        {
            break;
        }

        if (readyQueue->esm_rxHWM < numEvents)
            readyQueue->esm_rxHWM = numEvents;
        
        DLL_REMOVE_HEAD(&readyQueue->dllHdr, esm, queueEntry);
        NFX_ASSERT((esm->flags & ESM_IS_BUSY) == 0);
        esm->flags |= ESM_IS_BUSY;
        
        if (esm->flags & ESM_DESTROY)
        {
            NFX_SPIN_UNLOCK(&esm_spinlock);

            //ESM_TRACE("destroying esm handle 0x%x\n", esm->handle);

            // Drain the queue by calling the trash action for the esm.
            // Its safe to do this without the spinlock, since new
            // events won't be enqueued due to ESM_DESTROY being TRUE.
            while (esm->qcount > 0)
            {
                event_qelem = esm_dequeueEvent(esm);
                printf("esm 0x%llx pending events for destroyed sm\n",
                        (uint64)esm->handle);

                // TODO - trash functions in esm->table
                // trashfun(esm->handle, event_qelem->event.event, ...);
                               
                NFX_SPIN_LOCK(&esm_spinlock);
                DLL_INSERT_TAIL(&esm_freeEventHdr, event_qelem, dllEntry);
                NFX_SPIN_UNLOCK(&esm_spinlock);
            }

            // Indicate destruction to platform specific code.
            esm_destroyed(esm);
    
            NFX_SPIN_LOCK(&esm_spinlock);
            DLL_INSERT_TAIL(&esm_freeInstHdr, esm, queueEntry);
            esm->handle = 0;
        }
        else
        {
            NFX_ASSERT(esm->qcount > 0);
    
            event_qelem = esm_dequeueEvent(esm);
            event = event_qelem->event;
#ifdef ESM_USE_BUCKETS
			timer2 = event_qelem->qTime;
#endif
            DLL_INSERT_TAIL(&esm_freeEventHdr, event_qelem, dllEntry);
            NFX_SPIN_UNLOCK(&esm_spinlock);

            //ESM_TRACE("fp %ld calling runevent handle 0x%x event %ld\n",
            //    nfxMyCPUId, esm->handle, event);

#ifdef ESM_USE_BUCKETS
			eee_rtcDuration(&timer2, &elapsed);
			eee_rtcReset(&timer2);
            for (i = 0; i < ESM_TIME_STATS_BUCKETS_MAX; i++)
            {
                if (eee_rtcCompare(&elapsed, &(esm_stats_rtc[i])) < 0)
                {
                    break;
                }
            }
            if (i >= ESM_TIME_STATS_BUCKETS_MAX)
                i = ESM_TIME_STATS_BUCKETS_MAX - 1;
            readyQueue->stats[i] += 1;
#endif

            esm_runEvent( esm,
                          event);

            eee_rtcDuration(&time, &delta);
            if (eee_rtcCompare(&delta, &maxTime) > 0)
            {
                timeExpired = TRUE;
            }

#ifdef ESM_USE_BUCKETS
			eee_rtcDuration(&timer2, &elapsed);

            for (i = 0; i < ESM_TIME_STATS_BUCKETS_MAX; i++)
            {
                if (eee_rtcCompare(&elapsed, &(esm_stats_rtc[i])) < 0)
                {
                    break;
                }
            }
            if (i >= ESM_TIME_STATS_BUCKETS_MAX)
                i = ESM_TIME_STATS_BUCKETS_MAX - 1;
            esm_action_stats[i] += 1;
#endif

            ESM_READ_POLL_IND(1);
    
            // Put it back on the ready queue if there are still 
            // events to be processed, or if a destroy was indicated.
            NFX_SPIN_LOCK(&esm_spinlock);
            if (esm->qcount > 0 ||
                (esm->flags & ESM_DESTROY))
            {
                DLL_INSERT_TAIL(&readyQueue->dllHdr, esm, queueEntry);
            }
            esm->flags &= ~ESM_IS_BUSY;

            if (timeExpired)
            {
                break;
            }
        }
    }
    NFX_SPIN_UNLOCK(&esm_spinlock);

#ifdef ESM_USE_BUCKETS
    eee_rtcDuration(&timer1, &elapsed);
    for (i = 0; i < ESM_TIME_STATS_BUCKETS_MAX; i++)
    {
		if (eee_rtcCompare(&elapsed, &(esm_stats_rtc[i])) < 0)
        {
            break;
        }
    }
    if (i >= ESM_TIME_STATS_BUCKETS_MAX)
        i = ESM_TIME_STATS_BUCKETS_MAX - 1;
    esm_poll_stats[i] += 1;
#endif

}	/* esm_processEventQueue() */

/*-----------------------------------------------------------------
 * Name:	esm_processEventQueue()
 *
 * Description:	EEE registered polling func to process
 *              queued ESM events
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_processEventQueue(
	void	* cb,
	uint32	tref)
{
    int32   i;

#if defined(NFP_TXRX) && defined(NFX_SMP)
	/* txrx ncpu does not use esm */
	if (CM_IS_TXRX_NCPU)
		return;
#endif
 
    for (i = 0; i < ESM_QUEUE_PRI_MAX; i++)
    {
        esm_doProcessEventQueue(cb, tref, i);
    }
}

/*-----------------------------------------------------------------
 * Name:	esm_destroy_esm
 *
 * Description:	destroy an instance of a state machine by esm ptr
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_destroy_esm(
    esm_instance_t  *esm)
{
    if (!(esm->flags & ESM_DESTROY))
    {
        //ESM_TRACE("fp %ld destroy esm enqueued on handle 0x%x\n",
        //    nfxMyCPUId, esm->handle);
        esm->flags |= ESM_DESTROY;
        // Put it on the ready queue so it will
        // be destroyed on an upcoming polling loop.
        NFX_SPIN_LOCK(&esm_spinlock);
        esm_addToReadyQueue(esm, ESM_QUEUE_PRI_LO);
        NFX_SPIN_UNLOCK(&esm_spinlock);
    }

} /* esm_destroy_esm */

/*-----------------------------------------------------------------
 * Name:	esm_destroy
 *
 * Description:	destroy an instance of a state machine by handle
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_destroy(
	esm_handle_t		handle)
{
    esm_instance_t  *esm;

    NFX_SPIN_LOCK(&esm_spinlock);
	esm = esm_getInstFromHandle(handle);
	if (!esm)
	{
        // TODO - elog
		printf("esm_deliverEvent: stale handle 0x%llx\n", (uint64)handle);
        NFX_SPIN_UNLOCK(&esm_spinlock);
        return NFX_ERR;
	}
    else if (!(esm->flags & ESM_DESTROY))
    {
        //ESM_TRACE("fp %ld destroy enqueued on handle 0x%x\n",
        //  nfxMyCPUId, esm->handle);

        esm->flags |= ESM_DESTROY;
        // Put it on the ready queue so it will
        // be destroyed on an upcoming polling loop.
        esm_addToReadyQueue(esm, ESM_QUEUE_PRI_LO);
    }
    NFX_SPIN_UNLOCK(&esm_spinlock);

    return NFX_OK;

}	/* esm_destroy() */

/*-----------------------------------------------------------------
 * Name:	esm_getFreeInstance()
 *
 * Description:	find the next available free instance struct
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
esm_instance_t *
esm_getFreeInstance()
{
    esm_instance_t  *esm;
    
    NFX_SPIN_LOCK(&esm_spinlock);
    DLL_REMOVE_HEAD(&esm_freeInstHdr, esm, queueEntry);
    if (esm == NULL)
    {
        esm_extendInstances();
        DLL_REMOVE_HEAD(&esm_freeInstHdr, esm, queueEntry);
    }
    NFX_SPIN_UNLOCK(&esm_spinlock);

    return esm;

} /* esm_getFreeInstance */

/*-----------------------------------------------------------------
 * Name:	esm_instantiate()
 *
 * Description:	instantiate a state machine.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_instantiate(
	esm_table_t  			*table,
	esm_actionFunc_t  		*createAction,
	esm_handle_t			*handle,			/* filled in with new handle */
	void					*userInfo)		
{
	esm_instance_t 			*esm;
    esm_event_t             createEvent;
	int32					rc, nextState;

	if (!handle)
	{
        // TODO - elog
		printf("esm_instantiate:  NULL handle ptr\n");
		return NFX_ERR;
	}

    esm = esm_getFreeInstance();
    if (!esm)
    {
        // TODO - elog
		printf("esm_instantiate: no free instances available\n");
		return NFX_ERR;
    }
	
#if defined(NFX_SPINLOCK_DEBUG)
	LIST_INIT(&esm->sm_spinlocksHeld);
#endif

    esm->table = table;
    esm->userInfo = userInfo;
    esm->qcount = 0;
    DLL_INIT(&esm->eventQueueHdr);
    DLL_ISOLATE_ELEM(esm, yieldQueueEntry);
    if (table->canYield)
    {
        esm->flags = ESM_CAN_YIELD;
        rc = esm_allocStack(esm);
        if (rc != NFX_OK)
        {
            NFX_SPIN_LOCK(&esm_spinlock);
            DLL_INSERT_TAIL(&esm_freeInstHdr, esm, queueEntry);
            NFX_SPIN_UNLOCK(&esm_spinlock);
		    // printf("esm_instantiate: no memory for thread stack\n");
		    return NFX_ERR;
        }
    }
    else
    {
        esm->flags = 0;
        esm->stackBase = NULL;
    }
    esm->flags |= ESM_IS_BUSY;
	ESM_CREATE_HANDLE(esm->handle, esm->index);

	/*
	 * Call the create action.
	 */
    createEvent.event = 0;
    createEvent.p1.u.p = NULL;
    createEvent.p2.u.p = NULL;
	nextState = createAction->func(	esm->handle,
									createAction->nextState,
									createEvent,
									userInfo);
	esm->currentState = nextState;

    //ESM_TRACE("fp %ld instantiated handle 0x%x\n",
    //    nfxMyCPUId, esm->handle);
	*handle = esm->handle;

    NFX_SPIN_LOCK(&esm_spinlock);
    if (esm->qcount)
    {
        // Events were delivered in create action.
        esm_addToReadyQueue(esm, ESM_QUEUE_PRI_LO);
    }
    esm->flags &= ~ESM_IS_BUSY;
    NFX_SPIN_UNLOCK(&esm_spinlock);

	return NFX_OK;

}	/* esm_instantiate() */

/*-----------------------------------------------------------------
 * Name:	esm_longJump()
 *
 * Description:	Return the thread to the executive.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_longJump(
	esm_handle_t	handle)
{
	esm_instance_t  	*esm;

    NFX_SPIN_LOCK(&esm_spinlock);
	esm = esm_getInstFromHandle(handle);
    NFX_SPIN_UNLOCK(&esm_spinlock);

	if (!esm)
	{
        // TODO - elog
		printf("esm_deliverEvent: stale handle 0x%llx\n", (uint64)handle);
	}
    else
    {
        if ((esm->flags & ESM_CAN_YIELD) == 0)
        {
            panic("ERR: esm_longJump() called from non-yieldable ESM %s\n",
                esm->table->name);
        }
        else if (esm != esm_currInst)
        {
            // A thread can only long jump itself, not another.
            panic(
               "ERR: esm_longJump() called for handle 0x%x by handle 0x%x\n",
                   handle, esm_currInst->handle);
        }
        else
        {
            // Mark it for destruction.
            esm_destroy_esm(esm);
            // Long jump to exec.
            esm_threadLongJump(esm);
        }
    }
    
} /* esm_longJump */

/*-----------------------------------------------------------------
 * Name:    esm_showTimeStats()
 *
 * Description: Display statistics for time spent in
 *              all action functions and the polling loop.
 *  
 *-----------------------------------------------------------------
 */ 
int
esm_showTimeStats(
    int argc,
    char **argv)
{
#ifdef ESM_USE_BUCKETS
    printf("Type             %sus     %sus    %sus    %sus    %sus    %sus\n",
        esm_stats_usec_name[0],
        esm_stats_usec_name[1],
        esm_stats_usec_name[2],
        esm_stats_usec_name[3],
        esm_stats_usec_name[4],
        esm_stats_usec_name[5]);
    printf("------------------------------------------------------------------------------\n");
    printf("%-12s %10lld %10lld %10lld %10lld %10lld %10lld\n",
        "actionFunc",
        esm_action_stats[0],
        esm_action_stats[1],
        esm_action_stats[2],
        esm_action_stats[3],
        esm_action_stats[4],
        esm_action_stats[5]);
    printf("%-12s %10lld %10lld %10lld %10lld %10lld %10lld\n",
        "pollLoop",
        esm_poll_stats[0],
        esm_poll_stats[1],
        esm_poll_stats[2],
        esm_poll_stats[3],
        esm_poll_stats[4],
        esm_poll_stats[5]);
    printf("%-12s %10lld %10lld %10lld %10lld %10lld %10lld\n",
        "queueWaitHi",
        esm_readyQueue[ESM_QUEUE_PRI_HI].stats[0],
        esm_readyQueue[ESM_QUEUE_PRI_HI].stats[1],
        esm_readyQueue[ESM_QUEUE_PRI_HI].stats[2],
        esm_readyQueue[ESM_QUEUE_PRI_HI].stats[3],
        esm_readyQueue[ESM_QUEUE_PRI_HI].stats[4],
        esm_readyQueue[ESM_QUEUE_PRI_HI].stats[5]);
    printf("%-12s %10lld %10lld %10lld %10lld %10lld %10lld\n",
        "queueWaitLo",
        esm_readyQueue[ESM_QUEUE_PRI_LO].stats[0],
        esm_readyQueue[ESM_QUEUE_PRI_LO].stats[1],
        esm_readyQueue[ESM_QUEUE_PRI_LO].stats[2],
        esm_readyQueue[ESM_QUEUE_PRI_LO].stats[3],
        esm_readyQueue[ESM_QUEUE_PRI_LO].stats[4],
        esm_readyQueue[ESM_QUEUE_PRI_LO].stats[5]);

#else
    printf("ESM_USE_BUCKETS not defined\n");
#endif
    return NFX_OK;
}

#define ESM_SET_POLL_VIEW  0
#define ESM_SET_POLL_CLEAR 1
#define ESM_SET_POLL_HI    2
#define ESM_SET_POLL_LO    3

int
esm_poll_config (int action, int value)
{
    switch (action)
    {
    case ESM_SET_POLL_CLEAR:
        esm_readyQueue[ESM_QUEUE_PRI_LO].esm_rxHWM = 0;
        esm_readyQueue[ESM_QUEUE_PRI_HI].esm_rxHWM = 0;
        break;
    case ESM_SET_POLL_HI:
        esm_readyQueue[ESM_QUEUE_PRI_HI].maxPerPoll = value;
        break;
    case ESM_SET_POLL_LO:
        esm_readyQueue[ESM_QUEUE_PRI_LO].maxPerPoll = value;
        break;
    default:
        printf("esm LO HWM = %4d MAX = %4d\n", 
               esm_readyQueue[ESM_QUEUE_PRI_LO].esm_rxHWM,
               esm_readyQueue[ESM_QUEUE_PRI_LO].maxPerPoll);
        printf("esm HI HWM = %4d MAX = %4d\n", 
               esm_readyQueue[ESM_QUEUE_PRI_HI].esm_rxHWM,
               esm_readyQueue[ESM_QUEUE_PRI_HI].maxPerPoll);
        break;
    }
	return NFX_OK;
}

