/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-io.c
 *
 * Description  : Volume Manager (EVM) - I/O management
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

#include "nfx-incl.h"
#include "exports/debug.h"
#include "exports/timer-api.h"

#include "evm-queue.h"
#include "evm-io.h"
#include "evm-hash.h"
#include "evm-lv.h"
#include "evm-msg.h"
#include "evm-debug.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

extern evm_param_t  evmParam;

/* I/O requests are queued.
 * A request being executed is stored in the pending list.
 * When this list is full, requests are stored in the waiting list.
 */
static evm_ioQueue_t    metaQ;
static evm_ioQueue_t    userQ;

/* Handle to the timer used as clock's tics.
 */
static int32    ioClock;

/* Number to the I/O request (for debugging purpose only).
 */
static uint32   ioCnt;

extern boolean evm_mwchksum_on;
extern boolean evm_mrchksum_on;

#ifdef DEBUG
extern int32 evm_traceLevel;

uint64 offsetTab[1024];
uint32 offsetIdx = 0;

/* EVM history - for debug purpose
 */
typedef struct
{
    boolean     isReq;  // TRUE if request, FALSE if response
    evm_pReq_t  type;
    union
    {
        evm_pioCb_t         pcb;        // filled if request
        scsi_readRspMsg_t   readRsp;    // filled if read response
        scsi_writeRspMsg_t  writeRsp;   // filled if write response
    } u;  
} evm_hist_t;

#define EVM_MAX_HISTORY (2*256)
evm_hist_t evmHistory[EVM_MAX_HISTORY];
uint32 evmHistIdx;

static void incrHistIdx(void);

static void incrHistIdx(void)
{
    evmHistIdx++;
    if (evmHistIdx >= EVM_MAX_HISTORY)
        evmHistIdx = 0;
}
#endif // DEBUG

#define EVM_RETRY_NOW -999

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static int32
evm_io_createPhysReq(evm_lioCb_t *  lcb,
                     evm_pioCb_t *  pcb,
                     int32          pblks,
                     uint64         lPos);

static void
evm_io_complete(evm_ioQueue_t * queue,
                evm_lioCb_t **  LCB,
                evm_pioCb_t **  PCB);

static int32
evm_io_retry(evm_ioQueue_t *    queue,
             evm_lioCb_t *      lcb,
             evm_pioCb_t *      pcb);

static int32
evm_io_sendPioReq(evm_ioQueue_t *   queue,
                  evm_lioCb_t *     lcb,
                  evm_pioCb_t *     pcb);

static void
evm_io_rspTimeStat(evm_ioQueue_t *  queue,
                   evm_timeTab_t *  timeTab,
                   evm_pioCb_t *    pcb);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_io_createPhysReq
 * 
 * Description  : Create the physical I/O request (control block) corresponding
 *                to the logical I/O request.
 *                Assumption: 1 logical I/O <=> 1 physical I/O.
 *
 * Important    : Q element shall be previously isolated.
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 * 
 * Date Created : 28-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_io_createPhysReq(evm_lioCb_t *  lcb,    // logical control block
                     evm_pioCb_t *  pcb,    // physical control block
                     int32          pblks,  // number of physical blocks
                     uint64         lPos)   // logical vol offset
    
{
    evm_trace(EVM_TRACE_FUNC, "CALL: evm_io_createPhysReq()");

    /* Initialize the physical control block.
     */
    pcb->errCode = EVM_STAT_NO_ERROR;
    if (lcb->req.attributes & SCSI_MSG_ATTR_EXPRESS)
    {
        // no retry when "express IO"
        pcb->nbRetry = evmParam.maxIoRetry;
    }
    else
    {
        pcb->nbRetry = 0;
    }

    /* Get physical mapping of logical I/O.
     * Assumption is : 1 logical I/O <=> 1 physical I/O.
     */
    pcb->dev = evm_getPhys(lPos, pblks, lcb->lv,
                           &(pcb->offset), &(pcb->size),
                           lcb->accessPrivate);
    if ((pcb->dev == NULL) || (pcb->dev->openHandle == -1))
    {
        if (pcb->dev)
        {
            evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                      "cannot find physical device descriptor "
                      "(lPos %lld pblks %d)",
                      lcb->id, lPos, pblks);
        }
        else
        {
            evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                      "no handle for physical device (lPos %lld pblks %d)",
                      lcb->id, lPos, pblks);
        }
        return(EVM_STAT_INVALID);
    }

    /* In case of a write request, get the data to write.
     */
    if (lcb->req.reqType == EVM_WRITE_REQ)
    {
        pcb->dataHandle = lcb->req.wDataHandle;
    }
    else
    {
        pcb->dataHandle.dataPtr = NULL;
        pcb->dataHandle.codHandle= 0;
    }
    
    return(EVM_STAT_NO_ERROR);
}

/*------------------------------------------------------------------------------
 * Name         : evm_io_complete
 * 
 * Description  : Upon reception of the I/O response (SCSI/SBM),
 *                complete the I/O request (respond to the user application).
 *                If the I/O request returned with a error code,
 *                don't respond yet to the user application but retry
 *                the request (could be some temporary disk error...).
 *
 * Important    : Q shall be previously locked. Will be unlocked here.
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 * 
 * Date Created : 28-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
evm_io_complete(evm_ioQueue_t * queue,
                evm_lioCb_t **  LCB,
                evm_pioCb_t **  PCB)
    
{
    int32   rc = NFX_OK;
    boolean volIsOnline = TRUE;
    evm_lioCb_t *   lcb = *LCB;
    evm_pioCb_t *   pcb = *PCB;

    NFX_ASSERT((lcb != NULL) && (pcb != NULL));
    evm_trace(EVM_TRACE_FUNC, "CALL[%d][%d, %d]: evm_io_complete()",
              lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);

    /* Check if the volume is still online.
     */
    NFX_SPIN_LOCK(&evmVolLock);
    if (evm_getVol(lcb->req.id) == NULL)
    {
        volIsOnline = FALSE;
    }
    NFX_SPIN_UNLOCK(&evmVolLock);

    /* The following is a workaround for the "bad packet length" HW problem.
     */
    if ((pcb->errCode == EVM_RETRY_NOW) &&
        (pcb->nbRetry < evmParam.maxIoRetry) &&
        volIsOnline)
    {
        evm_io_retry(queue, lcb, pcb);
        NFX_SPIN_UNLOCK(&evmQueueLock);
        return;
    }
        
    /* If the I/O response is negative, retry the request.
     * Do nothing, wait for the request to time out.
     */
    if ((pcb->errCode != EVM_STAT_NO_ERROR) &&
        (pcb->errCode != SCSI_MSG_ECODE_OK) &&
        (pcb->nbRetry < evmParam.maxIoRetry) &&
        volIsOnline)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "Negative response (%d) for I/O on LUN 0x%llx (offset %d). "
                  "Retry request. Wait for request to time out.",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  pcb->errCode, (pcb->dev == NULL)? 0 : pcb->dev->id,
                  pcb->offset);
        NFX_SPIN_UNLOCK(&evmQueueLock);
        return;
    }

    /* Remove the request from the pending queue.
     */
    rc = evm_qRemoveFromPioPend(queue, &(pcb->handle));
    NFX_SPIN_UNLOCK(&evmQueueLock);
    if (rc != NFX_OK)
    {
        if (evmErrors.nStaleRsp == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: Stale response. "
                      "No physical I/O control block found for given handle.",
                      lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);
        }
        evmErrors.nStaleRsp++;
        
        EVM_STAT_STALES;
        return;
    }

    /* Initialize the result/status of the I/O response.
     * Do some statistics.
     */
    switch (pcb->errCode)
    {
    case EVM_STAT_NO_ERROR: // SCSI_MSG_ECODE_OK
    {
        break;
    }

    case EVM_STAT_TIMEOUT:
    {
        lcb->iostat.status = pcb->errCode;
        EVM_STAT_TIMEOUTS;
        break;
    }
    
    case EVM_STAT_SYS_ERR:
    {
        lcb->iostat.status = pcb->errCode;
        EVM_STAT_ERRORS;
        break;
    }
    
    case SCSI_MSG_ECODE_NOSUCH_DEV:
    case SCSI_MSG_ECODE_DEV_CLOSED:	
    case SCSI_MSG_ECODE_NO_RESOURCE:
    case SCSI_MSG_ECODE_INVALID_MSG:
    case SCSI_MSG_ECODE_ILLEGAL_REQUEST:
    case SCSI_MSG_ECODE_NO_MEDIA:
    case SCSI_MSG_ECODE_READ_ONLY:
    case SCSI_MSG_ECODE_OO_RANGE:
    case SCSI_MSG_ECODE_NOT_SUPPORTED:
    case SCSI_MSG_ECODE_PATH_FAILED:
    case SCSI_MSG_ECODE_DEV_IO_ERROR:
    case SCSI_MSG_ECODE_READ_CHECKSUM_ERR:
    case SCSI_MSG_ECODE_MEDIUM_ERROR:
    case SCSI_MSG_ECODE_ILI:
    case SCSI_MSG_ECODE_FM_ENCOUNTERED:
    case SCSI_MSG_ECODE_END_OF_MEDIUM:
    case SCSI_MSG_ECODE_INVALID_IOCTL_CODE:
    case SCSI_MSG_ECODE_INVALID_IOCTL_PARMS:
    case SCSI_MSG_ECODE_IOCTL_FAILED:
    case SCSI_MSG_ECODE_DATA_CHECKSUM_ERR:
    case SCSI_MSG_ECODE_TIMEOUT:
    case SCSI_MSG_ECODE_INVALID_LEN:
    case SCSI_MSG_ECODE_INVALID_TXN_SEQ:
    case SCSI_MSG_ECODE_WRITE_GATE_CLOSED:
    {
        lcb->iostat.status = EVM_STAT_DEV_ERROR;
        EVM_STAT_SCSIDEVERRS;
        EVM_STAT_ERRORS;
        break;
    }

    default:
    {
        lcb->iostat.status = EVM_STAT_SYS_ERR;
        EVM_STAT_ERRORS;
        break;
    }
    } // switch

    /* If the IO failed and the volume is not already offline,
     * put the volume offline.
     */
    if ((pcb->errCode != EVM_STAT_NO_ERROR) && volIsOnline)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "I/O request failed on LUN 0x%llx (offset %d errCode %d). "
                  "Turn the volume (0x%llx) offline.",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  (pcb->dev == NULL)? 0 : pcb->dev->id,
                  pcb->offset, pcb->errCode, lcb->req.id);

        /* Turn immediately the volume offline.
         */
        evm_offlineLV(lcb->req.id, TRUE);

        /* Notify FS and EVM about it.
         */
        evm_postEvt(EVENT_VOL_IOERROR, lcb->req.id, EVENT_STATE_DOWN);
    }
    
    /* Everything is complete. Prepare user callback.
     */
    lcb->iostat.clientPrivate = lcb->req.clientPrivate;
    
    switch (lcb->type)
    {
    case m_read:
    {
        evm_io_rspTimeStat(queue, &(evm_stats.MReadTime), pcb);
        
        /* Only 1 physical request per logical request.
         */
        lcb->iostat.rDataHandle.dataPtr = pcb->dataHandle.dataPtr;
        evm_trace(EVM_TRACE_IO, "IO[%d][%d, %d]: "
                  "MREAD iostat dataPtr 0x%x",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  lcb->iostat.rDataHandle.dataPtr);
        break;
    }
    case u_read:
    {
        evm_io_rspTimeStat(queue, &(evm_stats.UReadTime), pcb);
        
        /* Only 1 physical request per logical request.
         */
        lcb->iostat.rDataHandle.codHandle = pcb->dataHandle.codHandle;
        evm_trace(EVM_TRACE_IO, "IO[%d][%d, %d]: "
                  "UREAD cod 0x%llx",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  lcb->iostat.rDataHandle.codHandle);
        break;
    }
    case m_write:
    {
        evm_io_rspTimeStat(queue, &(evm_stats.MWriteTime), pcb);
        break;
    }
    case u_write:
    {
        evm_io_rspTimeStat(queue, &(evm_stats.UWriteTime), pcb);
        break;
    }
    default:
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "Wrong request type. Should never happened at this point. "
                  "Data corrupted!",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);
        lcb->iostat.status = EVM_STAT_SYS_ERR;
    }
    } // switch
    
    /* Invoke user callback function and remove all associated control blocks.
     */
    if (lcb->userCb)
    {
        void (*userCb)(evm_IOStat_t *stat);
        evm_IOStat_t    iostat;

        userCb = lcb->userCb;
        NFX_ASSERT_STATIC(sizeof(evm_IOStat_t) == 40);
        memcpy40(&iostat, &(lcb->iostat), sizeof(evm_IOStat_t));        
        EVM_STAT_IOCOMPLETE;
        evm_trace(EVM_TRACE_FUNC, "CALL[%d][%d, %d]: "
                  "IO complete -> calling user's callback (status %d)",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  lcb->iostat.status);
        NFX_SPIN_LOCK(&evmQueueLock);
        EVM_Q_INSERT_TO_FREE(queue, &(pcb->handle), lcb, pcb);
        NFX_SPIN_UNLOCK(&evmQueueLock);
        (*userCb)(&iostat);
    }
    else
    {
        NFX_SPIN_LOCK(&evmQueueLock);
        EVM_Q_INSERT_TO_FREE(queue, &(pcb->handle), lcb, pcb);
        NFX_SPIN_UNLOCK(&evmQueueLock);
    }
    
    /* If request in waiting queue, send it.
     */
    while(TRUE)
    {
        evm_qhandle_t * handle = NULL;
        evm_lioCb_t *   nextLcb = NULL;
        evm_pioCb_t *   nextPcb = NULL;
        
        NFX_SPIN_LOCK(&evmQueueLock);
        if ((queue->nbPioWait >= 1) && (queue->nbPioPend < queue->maxPioPend))
        {
            evm_trace(EVM_TRACE_DETAIL, "DETAIL[][]: "
                      "One request has been completed, take one from waiting Q");
            rc = evm_qRemoveFromPioWait(queue, &handle);
            if ((rc == NFX_OK) && (handle != NULL))
            {
                EVM_Q_GET_LCB(queue, handle, nextLcb);
                EVM_Q_GET_PCB(queue, handle, nextPcb);
                NFX_ASSERT((nextLcb != NULL) && (nextPcb != NULL));
                evm_trace(EVM_TRACE_DETAIL, "DETAIL[%d][%d, %d]: "
                          "Sending waiting request.",
                          nextLcb->id, nextPcb->handle.parts.idx,
                          nextPcb->handle.parts.genNb);
                rc = evm_io_sendPioReq(queue, nextLcb, nextPcb);
                if (rc == NFX_OK)
                {
                    NFX_SPIN_UNLOCK(&evmQueueLock);
                    break;
                }
                else
                {
                    nextPcb->errCode = EVM_STAT_SYS_ERR;
                    nextLcb->iostat.status = EVM_STAT_SYS_ERR;
                    evm_io_complete(queue, &nextLcb, &nextPcb);
                }
            }
        }
        else
        {
            NFX_SPIN_UNLOCK(&evmQueueLock);
            break;
        }
    } // while
    
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_io_retry
 *
 * Description  : Retry a physical I/O request.
 *
 * Important    : Q shall be previously locked.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_io_retry(evm_ioQueue_t *    queue,
             evm_lioCb_t *      lcb,
             evm_pioCb_t *      pcb)

{
    int32   rc = NFX_OK;

    NFX_ASSERT((queue != NULL) && (lcb != NULL) && (pcb != NULL));
    evm_trace(EVM_TRACE_FUNC, "CALL[%d][%d, %d]: evm_io_retry()",
              lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);

    /* Check if the volume is still online.
     */
    if (lcb->lv == NULL)
    {
        if (evmErrors.nOffline == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                      "Volume 0x%llx went offline. Don't retry I/O.",
                      lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                      lcb->req.id.whole_thing);
        }
        evmErrors.nOffline++;
        
        pcb->nbRetry = evmParam.maxIoRetry;
        return(NFX_ERR);
    }

    /* Update the handle.
     */
    queue->pMem[pcb->handle.parts.idx].handle.parts.genNb++;
    pcb->handle = queue->pMem[pcb->handle.parts.idx].handle;
    
    /* Re-send physical I/O request.
     */
    pcb->nbRetry++;

    switch (lcb->type)
    {
    case m_read:
    {
        rc = evm_sendScsiMsg(lcb, pcb);
        EVM_STAT_MREADS;
        EVM_STAT_MREADS_RETRY;
        break;
    }
    case u_read:
    {
        rc = evm_sendScsiMsg(lcb, pcb);
        EVM_STAT_UREADS;
        EVM_STAT_UREADS_RETRY;
        break;
    }
    case m_write:
    {
        rc = evm_sendScsiMsg(lcb, pcb);
        EVM_STAT_MWRITES;
        EVM_STAT_MWRITES_RETRY;
        break;
    }
    case u_write:
    {
        rc = evm_sendSbmScsiWriteMsg(lcb, pcb);
        EVM_STAT_UWRITES;
        EVM_STAT_UWRITES_RETRY;
        break;
    }
    default:
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "wrong request type %d", lcb->id,
                  pcb->handle.parts.idx, pcb->handle.parts.genNb, lcb->type);
        rc = NFX_ERR;
    }
    } // switch
    
    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "Send I/O request failed (%d) for LUN 0x%llx.",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  pcb->errCode, (pcb->dev == NULL)? 0 : pcb->dev->id);
        return(NFX_ERR);
    }
            
    /* Move request to the beginning of the list of pending requests
     * (remove it and add it to the queue).
     */
    rc = evm_qRemoveFromPioPend(queue, &(pcb->handle));
    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "cannot remove physical I/O request from pending queue",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);
        return(NFX_ERR);
    }
#ifdef DEBUG
    if (queue->nbPioPend >= queue->maxPioPend)
    {
        evm_displayStats();
        evm_trace(EVM_TRACE_ERRORS,
                  "evm_io_retry() - nbPioPend %d >= maxPioPend %d",
                  queue->nbPioPend, queue->maxPioPend);
    }
#endif // DEBUG
    rc = evm_qInsertToPioPend(queue, &(pcb->handle));
    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "cannot add physical I/O request to pending queue",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);
        return(NFX_ERR);
    }
#ifdef DEBUG
    if (evm_traceLevel & EVM_TRACE_HISTORY)
    {
        evmHistory[evmHistIdx].isReq = TRUE;
        evmHistory[evmHistIdx].type = lcb->type;
        memcpy(&(evmHistory[evmHistIdx].u.pcb), pcb, sizeof(evm_pioCb_t));
        incrHistIdx();
    }
#endif // DEBUG
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_io_sendPioReq()
 * 
 * Description  : Send physical I/O request.
 *
 * Important    : Q element shall be previously locked.
 *
 * Created by   : Isabelle Cnudde
 * 
 * Date Created : 21-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_io_sendPioReq(evm_ioQueue_t *   queue,
                  evm_lioCb_t *     lcb,
                  evm_pioCb_t *     pcb)
    
{
    int32   rc = NFX_OK;

    NFX_ASSERT((lcb != NULL) && (pcb != NULL));
    evm_trace(EVM_TRACE_FUNC, "CALL[%d][%d, %d]: evm_io_sendPioReq()",
              lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);
    
    /* Check if the volume is still online.
     */
    if (lcb->lv == NULL)
    {
        if (evmErrors.nOffline == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                      "Volume 0x%llx went offline. Don't send IO.",
                      lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                      lcb->req.id.whole_thing);

        }
        evmErrors.nOffline++;
        
        pcb->nbRetry = evmParam.maxIoRetry;
        return(NFX_ERR);
    }
    
    /* Send physical I/O request.
     */
    pcb->nbRetry = 0;
    switch (lcb->type)
    {
    case m_read:
    {
        rc = evm_sendScsiMsg(lcb, pcb);
        EVM_STAT_MREADS;
        ATOMIC_INC_DWORD(&(evm_stats.nBlkMRead), pcb->size);
        break;
    }
    case u_read:
    {
        rc = evm_sendScsiMsg(lcb, pcb);
        EVM_STAT_UREADS;
        ATOMIC_INC_DWORD(&(evm_stats.nBlkURead), pcb->size);
        break;
    }
    case m_write:
    {
        rc = evm_sendScsiMsg(lcb, pcb);
        EVM_STAT_MWRITES;
        ATOMIC_INC_DWORD(&(evm_stats.nBlkMWrite), pcb->size);
        break;
    }
    case u_write:
    {
        rc = evm_sendSbmScsiWriteMsg(lcb, pcb);
        EVM_STAT_UWRITES;
        ATOMIC_INC_DWORD(&(evm_stats.nBlkUWrite), pcb->size);
        break;
    }
    default:
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "Wrong request type %d. Should never happened at this point."
                  " Data may be corrupted!", lcb->id,
                  pcb->handle.parts.idx, pcb->handle.parts.genNb, lcb->type);
        return(NFX_ERR);
    }
    } // switch

    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "send I/O request failed (%d)",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  pcb->errCode);
        return(NFX_ERR);
    }
            
    /* Add request in list of pending requests.
     */
#ifdef DEBUG // debug only!!!
    if (queue->nbPioPend >= queue->maxPioPend)
    {
        evm_displayStats();
        evm_trace(EVM_TRACE_ERRORS,
                  "evm_io_sendPioReq() - nbPioPend %d >= maxPioPend %d",
                  queue->nbPioPend, queue->maxPioPend);
    }
#endif // DEBUG
    rc = evm_qInsertToPioPend(queue, &(pcb->handle));
    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                  "cannot add physical I/O request to pending queue",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);
        return(NFX_ERR);
    }
#ifdef DEBUG
    if (evm_traceLevel & EVM_TRACE_HISTORY)
    {
        evmHistory[evmHistIdx].isReq = TRUE;
        evmHistory[evmHistIdx].type = lcb->type;
        memcpy(&(evmHistory[evmHistIdx].u.pcb), pcb, sizeof(evm_pioCb_t));
        incrHistIdx();
    }
#endif // DEBUG
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_io_rspTimeStat
 *
 * Description  : Calculate the response it tooks to answer the caller.
 *                (Retries are not included.)
 *
 * Important    : Q element shall be previously isolated.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

static void
evm_io_rspTimeStat(evm_ioQueue_t *  queue,
                   evm_timeTab_t *  timeTab,
                   evm_pioCb_t *    pcb)
    
{
    eeeRTC_t    tmpTime;
    eeeRTC_t    arrivedTime;
	eeeRTC_t	diff;
	int32		underflow;
	static boolean  first = TRUE;
	static eeeRTC_t	binRTC[6];

    if (pcb->errCode == EVM_STAT_TIMEOUT)
    {
        return; // don't count response which timed out
    }
    
    eee_rtc(&tmpTime);
    EVM_Q_GET_TIMESTAMP(queue, &(pcb->handle), arrivedTime);
	underflow = eee_rtcSub(&tmpTime, &arrivedTime, &diff);

    if (underflow < 0)
    {
		eeeGMT_t nowTV;
		eeeGMT_t arrivedTV;

		eee_rtcToTV(&tmpTime, &nowTV);
		eee_rtcToTV(&arrivedTime, &arrivedTV);
		evm_trace(EVM_TRACE_DETAIL, "DETAIL: past is more recent than "
                  "present! now (%dsec, %dusec) - arrivedTime (%dsec, %dusec) ",
				  nowTV.sec, nowTV.usec,
                  arrivedTV.sec, arrivedTV.usec);
        return;
    }
    
	if (first) {
		first = FALSE;
		// Build up the bin times in RTC form for efficiency
		eee_msToRTC(EVM_AMSEC, &(binRTC[0]));
		eee_msToRTC(EVM_BMSEC, &(binRTC[1]));
		eee_msToRTC(EVM_CMSEC, &(binRTC[2]));
		eee_msToRTC(EVM_DMSEC, &(binRTC[3]));
		eee_msToRTC(EVM_EMSEC, &(binRTC[4]));
		eee_msToRTC(EVM_FMSEC, &(binRTC[5]));
	}
	if (eee_rtcCompare(&diff, &(binRTC[0])) < 0)
        ; // do nothing
    else if (eee_rtcCompare(&diff, &(binRTC[1])) < 0)
        ATOMIC_INC_DWORD(&(timeTab->moreThanAmsec), 1);
    else if (eee_rtcCompare(&diff, &(binRTC[2])) < 0)
        ATOMIC_INC_DWORD(&(timeTab->moreThanBmsec), 1);
    else if (eee_rtcCompare(&diff, &(binRTC[3])) < 0)
        ATOMIC_INC_DWORD(&(timeTab->moreThanCmsec), 1);
    else if (eee_rtcCompare(&diff, &(binRTC[4])) < 0)
        ATOMIC_INC_DWORD(&(timeTab->moreThanDmsec), 1);
    else if (eee_rtcCompare(&diff, &(binRTC[5])) < 0)
        ATOMIC_INC_DWORD(&(timeTab->moreThanEmsec), 1);
    else
        ATOMIC_INC_DWORD(&(timeTab->moreThanFmsec), 1);

	if (eee_rtcCompare(&diff, &timeTab->maxTime) > 0)
        timeTab->maxTime = diff;

	if (eee_rtcIncr(&timeTab->sumTime, &diff) == NFX_OK)
		ATOMIC_INC_DWORD(&(timeTab->numTime), 1);

    return;
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_io_init
 *
 * Description  : Initialize the EVM I/O module.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_io_init(void)
    
{
    int32   rc = NFX_OK;

    evm_trace(EVM_TRACE_FUNC, "CALL: evm_io_init()");
    
    /* Initialize the I/O request counter.
     */
    ioCnt = 0;
    evm_resetVolErr();

    /* Create the I/O queues.
     */
    rc = evm_qCreate(&metaQ, EVM_META_DATA_REQ,
                     (EVM_MAX_META_PIO_PEND + EVM_MAX_META_PIO_WAIT),
                     EVM_MAX_META_PIO_WAIT, EVM_MAX_META_PIO_PEND);
    NFX_ASSERT(rc == NFX_OK);
    
    rc = evm_qCreate(&userQ, EVM_USER_DATA_REQ,
                     (EVM_MAX_USER_PIO_PEND + EVM_MAX_USER_PIO_WAIT),
                     EVM_MAX_USER_PIO_WAIT, EVM_MAX_USER_PIO_PEND);
    NFX_ASSERT(rc == NFX_OK);

    /* Create and start a timer to check pending I/O requests.
     */
    ioClock = eee_timerCreate(EVM_IO_TICK, EEE_TIMER_REPEAT);
    NFX_ASSERT(ioClock != TIMER_ERR_NO_TIMER_AVAILABLE);
  
    rc = eee_timerStart(ioClock, evm_io_clockTic, NULL, NULL, NULL, NULL);
    NFX_ASSERT(rc == NFX_OK);

#ifdef DEBUG
    bzero((void *)evmHistory, EVM_MAX_HISTORY * sizeof(evm_hist_t));
    evmHistIdx = 0;
#endif // DEBUG
    return(rc);
}  

/*------------------------------------------------------------------------------
 * Name         : evm_io_terminate
 *
 * Description  : Terminate EVM I/O module.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_io_terminate(void)
    
{
    evm_trace(EVM_TRACE_FUNC, "CALL: evm_io_terminate()");

    /* Stop and destroy timer.
     */
    eee_timerStop(ioClock);
    eee_timerDestroy(ioClock);
    
    /* Destroy the I/O queues.
     */
    evm_qDestroy(&metaQ);
    evm_qDestroy(&userQ);

    return(NFX_OK);
}  

/*------------------------------------------------------------------------------
 * Name         : evm_io_req
 *
 * Description  : Client's interface for logical volume I/O request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_io_req(evm_IOReq_t *    req,
           void (*cmplt)    (evm_IOStat_t * stat),
           boolean          accessPrivate)
    
{
    evm_ioQueue_t *     queue = NULL;
    evm_qhandle_t *     handle = NULL;
    evm_lioCb_t *       lcb = NULL;
    evm_pioCb_t *       pcb = NULL;
    int32               rc = NFX_OK;
    int32               status = EVM_STAT_NO_ERROR;
    uint32              ioNb;

    NFX_ASSERT(req != NULL);
    evm_trace(EVM_TRACE_FUNC, "CALL: evm_io_req"
              "(type %d dataType %d VolId %lld offset %lld size %d)",
              req->reqType, req->dataType, req->id.whole_thing,
              req->offset, req->reqSize);

    /* Assign an incremental number to the I/O request.
     * (for debugging purpose only)
     */
    ATOMIC_INC_INT(&ioCnt, 1);
    ioNb = ioCnt;
    evm_trace(EVM_TRACE_DETAIL, "DETAIL[][]: I/O request #%d", ioNb);
    EVM_STAT_IOREQ;

    /* Select to which IO queue this request belongs.
     */
    if (req->dataType == EVM_META_DATA_REQ)
    {
        queue = &metaQ;
    }
    else if (req->dataType == EVM_USER_DATA_REQ)
    {
        queue = &userQ;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][]: %d is not a correct data type.",
                  ioNb, req->dataType);
        goto evm_IOReq_err;
    }
    
    /* Check size of the I/O request.
     * Shall be modulo 1k for scsi/fc to take it.
     */
    if (req->reqSize % 2)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                  "invalid size %d", ioNb, req->reqSize);
        status = EVM_STAT_INVALID;
        goto evm_IOReq_err;
    }

    /* Check if we can process this request.
     * Keep some free spots in the queue for high priority IOs
     * (currently their are only the volume tstamps). 
     */
    NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(queue->nbFree >= 0);
    if (queue->nbFree == 0)
    {
        /* Too many requests pending and waiting.
         * Refuse I/O request.
         */
        if (evmErrors.nQueueFull == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[%d][]: EVM server %s queue full. "
                      "Reject I/O request for volume 0x%llx.", ioNb,
                      ((req->dataType == EVM_META_DATA_REQ)? "meta":"user"),
                      req->id);
        }
        evmErrors.nQueueFull++;
        
        status = EVM_STAT_REJECT;
        EVM_STAT_IO_REJECTED;
        goto evm_IOReq_err;
    }

    /* Get free element in queue for I/O request.
     * Get the logical and physical control blocks.
     */
    rc = evm_qRemoveFromFree(queue, &handle);
    NFX_SPIN_UNLOCK(&evmQueueLock);
    if ((handle == NULL) || (rc != NFX_OK))
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                  "cannot get free spot in queue", lcb->id);
        evm_displayStats();
        status = EVM_STAT_SYS_ERR;
        goto evm_IOReq_err;
    }
    EVM_Q_GET_LCB(queue, handle, lcb);
    EVM_Q_GET_PCB(queue, handle, pcb);
    NFX_ASSERT((lcb != NULL) && (pcb != NULL));
    
    /* Save logical I/O request in control block,
     * as well as the callback function.
     */
    memcpy(&lcb->req, req, sizeof(evm_IOReq_t));
    lcb->id = ioNb;
    lcb->userCb = cmplt;
    lcb->accessPrivate = accessPrivate;
    lcb->iostat.rDataHandle.dataPtr = NULL;
    lcb->iostat.rDataHandle.codHandle = NULL;
    lcb->iostat.status = EVM_STAT_NO_ERROR;
    lcb->iostat.devErrorCode = 0;
    lcb->iostat.clientPrivate = 0;
    
    /* Check and initialize the type of the request.
     */
    if ((req->reqType == EVM_READ_REQ) &&
        (req->dataType == EVM_META_DATA_REQ))
    {
        lcb->type = m_read;
    }
    else if ((req->reqType == EVM_READ_REQ) &&
             (req->dataType == EVM_USER_DATA_REQ))
    {
        lcb->type = u_read;
    }
    else if ((req->reqType == EVM_WRITE_REQ) &&
             (req->dataType == EVM_META_DATA_REQ))
    {
        lcb->type = m_write;
    }
    else if ((req->reqType == EVM_WRITE_REQ) &&
             (req->dataType == EVM_USER_DATA_REQ))
    {
        lcb->type = u_write;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                  "wrong request (%d) and/or data type (%d)",
                  lcb->id, req->reqType, req->dataType);
        status = EVM_STAT_INVALID;
        goto evm_IOReq_err;
    }
    
    /* Get volume where to perform the I/O.
     */
    NFX_SPIN_LOCK(&evmVolLock);
    lcb->lv = evm_getVol(lcb->req.id);
    NFX_SPIN_UNLOCK(&evmVolLock);
    if (lcb->lv == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                  "cannot find volume 0x%llx", lcb->id,
                  lcb->req.id.whole_thing);
        status = EVM_STAT_INVALID;
        goto evm_IOReq_err;
    }
    evm_trace(EVM_TRACE_DETAIL, "DETAIL[%d][]: "
              "found VolId %lld VolSize %lld",
              lcb->id, lcb->lv->id.whole_thing, lcb->lv->size);

    /* Check if I/O is in volume boundary.
     */
    if (((accessPrivate == FALSE) &&
         (lcb->req.offset + lcb->req.reqSize) > lcb->lv->size) ||
        ((accessPrivate == TRUE) &&
         (lcb->req.offset + lcb->req.reqSize) > (lcb->lv->size + N_BLKS_CFG)))
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                  "try to access blocks outside volume: "
                  "offset (%lld) + size (%d) > volSize (%lld)",
                  lcb->id, lcb->req.offset,
                  lcb->req.reqSize, lcb->lv->size);
        status = EVM_STAT_INVALID;
        goto evm_IOReq_err;
    }
     
    /* Initialize the physical control block.
     */
    status = evm_io_createPhysReq(lcb, pcb,
                                  lcb->req.reqSize,
                                  lcb->req.offset);
    pcb->handle = *handle;
        
    /* Calculate checksum right now.
     */
    if (lcb->type == m_write)
    {
        if (evm_mwchksum_on)
        {
            pcb->checksum = uint32_checksum(pcb->dataHandle.dataPtr,
                                            pcb->size * EVM_LBLK_SIZE, 0);
        }
        else
        {
            pcb->checksum = 0;
        }
    }
    else if (lcb->type == u_write)
    {
        pcb->checksum = 0;
    }
    
    /* Store the request in either the waiting queue or the pending queue.
     */
    NFX_SPIN_LOCK(&evmQueueLock);
    if (queue->nbPioPend < queue->maxPioPend)
    {
        /* Send the physical request.
         */
        rc = evm_io_sendPioReq(queue, lcb, pcb);
        NFX_SPIN_UNLOCK(&evmQueueLock);
        if (rc != NFX_OK)
        {
            evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                      "cannot add physical I/O request to pending queue",
                      lcb->id);
            status = EVM_STAT_SYS_ERR;
            goto evm_IOReq_err;
        }
    }
    else
    {
        /* Add request in list of waiting request.
         */
        rc = evm_qInsertToPioWait(queue, handle);
        NFX_SPIN_UNLOCK(&evmQueueLock);
        if (rc != NFX_OK)
        {
            evm_trace(EVM_TRACE_ERRORS, "[%d][]: "
                      "cannot add physical I/O request to waiting queue",
                      lcb->id);
            status = EVM_STAT_SYS_ERR;
            goto evm_IOReq_err;
        }            
    }

    return(NFX_OK);

evm_IOReq_err:

    /* Increment number of logical I/O requests completed
     * and number of error responses returned.
     */
    EVM_STAT_IOCOMPLETE;
    EVM_STAT_ERRORS;
    
    /* Give back the element in the queue.
     */
    NFX_SPIN_LOCK(&evmQueueLock);
    EVM_Q_INSERT_TO_FREE(queue, handle, lcb, pcb);
    NFX_SPIN_UNLOCK(&evmQueueLock);
    
    return(status);
}

/*------------------------------------------------------------------------------
 * Name         : evm_io_clockTic
 * 
 * Description  : Callback function when timer expires.
 *                Go through the pending queue and check which request(s)
 *                timed out.
 *
 * Created by   : Isabelle Cnudde
 * 
 * Date Created : 02-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_io_clockTic(uint32  tmr_handle,
                void *  usr_param1,
                void *  usr_param2,
                void *  usr_param3,
                void *  usr_param4)
    
{
    evm_ioQueue_t * queue = NULL;
    int32           rc = NFX_OK;
    evm_qelt_t *    pCurr = NULL;
    evm_qelt_t *    pNext = NULL;
    evm_qhandle_t   handle;
    uint64          nbIo = 0;

    NFX_SPIN_LOCK(&evmQueueLock);

    for (queue = &metaQ; ; queue = &userQ)
    {        
        EVM_Q_GO_THRU_PIOPEND(pCurr, pNext, queue, handle)
        {
            eeeRTC_t    timestamp;
            eeeRTC_t    delta;
            eeeRTC_t    now;

            NFX_ASSERT(handle.whole != -1);
            eee_rtc(&now);
            EVM_Q_GET_TIMESTAMP(queue, &handle, timestamp);
			(void)eee_rtcSub(&now, &timestamp, &delta);
            if (eee_rtcCompare(&delta, &evmParam.ioTimeout) >= 0)
            {
                // retry
                evm_lioCb_t * lcb = NULL;
                evm_pioCb_t * pcb = NULL;

                EVM_Q_GET_LCB(queue, &handle, lcb);
                EVM_Q_GET_PCB(queue, &handle, pcb);
                NFX_ASSERT((lcb != NULL) && (pcb != NULL));
                if (pcb->nbRetry < evmParam.maxIoRetry)
                {
                    if (lcb->lv != NULL)
                    {
                        if (lcb->lv->nRetries == 0)
                        {
                            evmErrors.isError = TRUE;
                            evm_trace(EVM_TRACE_ERRORS, "[%d][%d, %d]: "
                                      "Retry I/O for volume 0x%llx LUN 0x%llx.",
                                      lcb->id, pcb->handle.parts.idx,
                                      pcb->handle.parts.genNb, 
                                      lcb->req.id.whole_thing,
                                      (pcb->dev == NULL)? 0 : pcb->dev->id);
                        }
                        lcb->lv->nRetries++;
                    }
                    
                    rc = evm_io_retry(queue, lcb, pcb);
                    if (rc != NFX_OK)
                    {
                        /* The retry failed. Do not need to retry further.
                         */
                        pcb->nbRetry = evmParam.maxIoRetry;
                        pcb->errCode = EVM_STAT_SYS_ERR;
                        evm_io_complete(queue, &lcb, &pcb);
                        NFX_SPIN_LOCK(&evmQueueLock);
                    }
                    else
                    {
                        nbIo++;
                    }
                }
                else
                {
                    /* Don't overwrite the previous error code if any.
                     */
                    if ((pcb->errCode == EVM_STAT_NO_ERROR) ||
                        (pcb->errCode == SCSI_MSG_ECODE_OK))
                    {
                        pcb->errCode = EVM_STAT_TIMEOUT;

                    }
                    evm_io_complete(queue, &lcb, &pcb);
                    NFX_SPIN_LOCK(&evmQueueLock);
                }
            }
            else
            {
                // no old request in the pending queue
                break;
            }

            /* To avoid sending too many messages in one loop,
             * the number of retries are limited.
             * (Otherwise the transmit queue would be overflowed
             * and/or EVM server will monopolize the CPU for too long.)
             */
            if (nbIo >= evmParam.maxIo)
            {
                queue = &userQ; // to exit the 2nd loop
                break; // to exit the 1st loop
            }
        } // EVM_Q_GO_THRU_PIOPEND

        if (queue == &userQ)
            break;
    } // for

    /* Print all errors of last 5s, then reset all errors counters.
     */
    if (evmErrors.isError == TRUE)
    {
        evm_printVolErr();
        evm_resetVolErr();
    }

    NFX_SPIN_UNLOCK(&evmQueueLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * SCSI responses
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_io_scsiReadRsp
 * 
 * Description  : called by EVM SRVR rcv msg to process a 
 *                SCSI_MID_READ_RSP message.  The message is contained
 *                in a system buffer, and we are responsible for freeing
 *                it here or passing it to our client for its disposition.
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 * 
 * Date Created : 28-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
evm_io_scsiReadRsp(char *   msgp,
                   uint32   msgLen,
                   uint32   srcPort)
    
{
    evm_ioQueue_t *     queue = NULL;
    scsi_readRspMsg_t * rp = (scsi_readRspMsg_t *) msgp;
    evm_qhandle_t       handle = (evm_qhandle_t) rp->hdr.cHandle;
    evm_lioCb_t *       lcb = NULL;
    evm_pioCb_t *       pcb = NULL;

#ifdef EVM_SRVR_TEST_TIMEOUT
    EVM_DEALLOC_BUF(msgp);
    return;
#endif

    EVM_STAT_IOEVENTS;
    
    /* Select to which IO queue this request belongs.
     */
    if (handle.parts.dataType == EVM_META_DATA_REQ)
    {
        queue = &metaQ;
    }
    else
    {
        queue = &userQ;
    }
    
    /* Find the physical I/O control block.
     */
    NFX_SPIN_LOCK(&evmQueueLock);
    EVM_Q_GET_PCB(queue, &handle, pcb);
    if (pcb == NULL)
    {
        if (evmErrors.nStaleRsp == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: Stale response. "
                      "No physical I/O control block found for given handle.",
                      handle.parts.idx, handle.parts.genNb);
        }
        evmErrors.nStaleRsp++;
        
        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        goto evm_io_scsiReadRsp_exit;
    }

    /* Check if response received corresponds to request sent.
     */
    if ((pcb->offset != rp->blkOffset) ||
        (pcb->size != rp->blkCnt))
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "read expected offset/size/len %d/%d/%d got %d/%d/%d\n",
                  pcb->offset, pcb->size, (pcb->size * EVM_LBLK_SIZE),
                  rp->blkOffset, rp->blkCnt, rp->dataLen);
        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        goto evm_io_scsiReadRsp_exit;
    }

    /* Get return code.
     */
    pcb->errCode = rp->eCode;
    
    /* Notify user of physical IO complete,
     */
    EVM_Q_GET_LCB(queue, &handle, lcb);
    if (lcb == NULL)
    {
        if (evmErrors.nStaleRsp == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: Stale response. "
                      "No logical I/O control block found for given handle.",
                      handle.parts.idx, handle.parts.genNb);
        }
        evmErrors.nStaleRsp++;
        
        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        goto evm_io_scsiReadRsp_exit;
    }

    if (lcb->type == m_read)
    {
        /* Meta data follows the SCSI read response message.
         */
        pcb->dataHandle.dataPtr = msgp + sizeof(scsi_readRspMsg_t);
        evm_trace(EVM_TRACE_IO, "IO[%d][%d, %d]: "
                  "SCSI_MID_READ_RSP eCode %d dataPtr 0x%x",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  pcb->errCode, pcb->dataHandle.dataPtr);
        
        /* Verify if all data expected have been received.
         */
        if ((pcb->errCode == SCSI_MSG_ECODE_OK) &&
            ((rp->dataLen != (pcb->size * EVM_LBLK_SIZE)) ||
             (msgLen != (sizeof(scsi_readRspMsg_t) + rp->dataLen))))
        {
            /* The following is a workaround for the
             * "bad packet length" HW problem.
             */
            pcb->errCode = EVM_RETRY_NOW;
            ATOMIC_INC_DWORD(&(evm_stats.nBadPacketLength), 1);
            evm_io_complete(queue, &lcb, &pcb);
            goto evm_io_scsiReadRsp_exit;
        }
        /*
          {
          pcb->errCode = EVM_STAT_MSG_ERR;
          
          evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: "
          "rp->dataLen = %d, pcb->size * EVM_LBLK_SIZE = %d\n"
          "msgLen = %d, "
          "sizeof(scsi_readRspMsg_t) + rp->dataLen = %d\n"
          "msgPtr = %p, dataPtr = %p",
          handle.parts.idx, handle.parts.genNb,
          rp->dataLen, pcb->size * EVM_LBLK_SIZE,
          msgLen, sizeof(scsi_readRspMsg_t) + rp->dataLen,
          msgp, pcb->dataHandle.dataPtr);
          
          if (evmDoAssert) NFX_ASSERT(FALSE);
          }
        */
#ifdef DEBUG
        /* Verify checksum if set by SCSI.
         */
        if (evm_mrchksum_on)
        {
            if ((pcb->errCode == SCSI_MSG_ECODE_OK) && (rp->checkSum != 0))
            {
                uint32  myCheckSum =
                    uint32_checksum(pcb->dataHandle.dataPtr,
                                    (pcb->size * EVM_LBLK_SIZE), NULL);
            
                if (myCheckSum != rp->checkSum)
                {
                    pcb->errCode = EVM_STAT_MSG_ERR;
                
                    evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: "
                              "myCheckSum = 0x%x, rp->checkSum = 0x%x\n"
                              "msgPtr = %p, dataPtr = %p",
                              handle.parts.idx, handle.parts.genNb,
                              myCheckSum, rp->checkSum,
                              msgp, pcb->dataHandle.dataPtr);
                
                    if (evmDoAssert) NFX_ASSERT(FALSE);
                }
            }
        }
#endif
    }
    else if (lcb->type == u_read)
    {
        /* The following is a workaround for the
         * "bad packet length" HW problem.
         */
        if ((pcb->errCode == SCSI_MSG_ECODE_OK) &&
            ((rp->dataLen != (pcb->size * EVM_LBLK_SIZE))))
        {
            pcb->errCode = EVM_RETRY_NOW;
            ATOMIC_INC_DWORD(&(evm_stats.nBadPacketLength), 1);
            evm_io_complete(queue, &lcb, &pcb);
            goto evm_io_scsiReadRsp_exit;
        }
        
        /* User data, return the COD.
         */
        pcb->dataHandle.codHandle = rp->codHandle;
        evm_trace(EVM_TRACE_IO, "IO[%d][%d, %d]: "
                  "SCSI_MID_READ_RSP eCode %d codHandle 0x%llx",
                  lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb,
                  pcb->errCode, pcb->dataHandle.codHandle);
    }
    
    evm_trace(EVM_TRACE_DETAIL, "DETAIL[%d][%d, %d]: READ SCSI rsp",
              lcb->id, pcb->handle.parts.idx, pcb->handle.parts.genNb);
    
    /* Notify user of physical IO complete
     */
    if (lcb->type == u_read)
    {
        EVM_STAT_SCSIUREADRSPS;
#ifdef DEBUG
        ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(srcPort)]. \
                           userReadRsp), 1);
#endif // DEBUG
    }
    else // m_read
    {
        EVM_STAT_SCSIMREADRSPS;
#ifdef DEBUG
        ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(srcPort)]. \
                           metaReadRsp), 1);
#endif // DEBUG
    }
#ifdef DEBUG
    if (evm_traceLevel & EVM_TRACE_HISTORY)
    {
        evmHistory[evmHistIdx].isReq = FALSE;
        evmHistory[evmHistIdx].type = lcb->type;
        memcpy(&(evmHistory[evmHistIdx].u.readRsp), rp,
               sizeof(scsi_readRspMsg_t));
        incrHistIdx();
    }
#endif // DEBUG
#ifdef DEBUG
   	lcb->iostat.reserved[0] = rp->queueDepth;
    ATOMIC_INC_DWORD(&(evm_stats.sumQueueDepth), rp->queueDepth);
    ATOMIC_INC_DWORD(&(evm_stats.numQueueDepth), 1);
    evm_stats.maxQueueDepth = MAX(evm_stats.maxQueueDepth, rp->queueDepth);
#endif // DEBUG
        
    /* For meta reads, the caller's data will be in msgp so don't deallocate !!
     */
    if (lcb->req.dataType == EVM_META_DATA_REQ)
    {
        evm_io_complete(queue, &lcb, &pcb);
        return;
    }
    else
    {
        evm_io_complete(queue, &lcb, &pcb);
    }
    
evm_io_scsiReadRsp_exit:

    EVM_DEALLOC_BUF(msgp);
    return;

}
 
/*------------------------------------------------------------------------------
 * Name         : evm_io_scsiWriteRsp
 * 
 * Description  : called by EVM SRVR rcv msg to process a 
 *                SCSI_MID_WRITE_RSP message
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 * 
 * Date Created : 28-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
evm_io_scsiWriteRsp(scsi_writeRspMsg_t *    wp,
                    uint32                  srcPort)    
{
    evm_ioQueue_t * queue = NULL;
    evm_qhandle_t   handle;
    evm_pioCb_t *   pcb = NULL;
    evm_lioCb_t *   lcb = NULL;

#ifdef EVM_SRVR_TEST_TIMEOUT
    return;
#endif

    EVM_STAT_IOEVENTS;
    handle = (evm_qhandle_t) wp->hdr.cHandle;
    
    /* Select to which IO queue this request belongs.
     */
    if (handle.parts.dataType == EVM_META_DATA_REQ)
    {
        queue = &metaQ;
    }
    else
    {
        queue = &userQ;
    }
    
    /* Find the physical I/O control block.
     * Remove it from the pending queue.
     */
    NFX_SPIN_LOCK(&evmQueueLock);
    EVM_Q_GET_PCB(queue, &handle, pcb);
    if (pcb == NULL)
    {
        if (evmErrors.nStaleRsp == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: Stale response. "
                      "No physical I/O control block found for given handle.",
                      handle.parts.idx, handle.parts.genNb);
        }
        evmErrors.nStaleRsp++;
        
        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        return;
    }

    /* Check if response received corresponds to request sent.
     */
    if ((wp->eCode == SCSI_MSG_ECODE_OK) &&
        ((pcb->offset != wp->blkOffset) || (pcb->size != wp->blkCnt)))
    {
        evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: "
                  "write expected offset/size %d/%d got %d/%d\n",
                  handle.parts.idx, handle.parts.genNb,
                  pcb->offset, pcb->size, wp->blkOffset, wp->blkCnt);
        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        return;
    }

    /* Get return code.
     */
    pcb->errCode = wp->eCode;

    /* Notify user of physical IO complete,
     */
    EVM_Q_GET_LCB(queue, &handle, lcb);
    if (lcb == NULL)
    {
        if (evmErrors.nStaleRsp == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: Stale response. "
                      "No logical I/O control block found for given handle.",
                      handle.parts.idx, handle.parts.genNb);
        }
        evmErrors.nStaleRsp++;

        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        return;
    }
    
    if (lcb->type == m_write)
    {
        EVM_STAT_SCSIMWRITERSPS;
#ifdef DEBUG
        ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(srcPort)]. \
                           metaWriteRsp), 1);
#endif // DEBUG
    }
    else // u_write
    {
        EVM_STAT_SCSIUWRITERSPS;
#ifdef DEBUG
        ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(srcPort)]. \
                           userWriteRsp), 1);
#endif // DEBUG
    }
        
#ifdef DEBUG
    if (evm_traceLevel & EVM_TRACE_HISTORY)
    {
        evmHistory[evmHistIdx].isReq = FALSE;
        evmHistory[evmHistIdx].type = lcb->type;
        memcpy(&(evmHistory[evmHistIdx].u.writeRsp), wp,
               sizeof(scsi_writeRspMsg_t));
        incrHistIdx();
    }
#endif // DEBUG
#ifdef DEBUG
   	lcb->iostat.reserved[0] = wp->queueDepth;
    //printf("wp->queueDepth = %d\n", wp->queueDepth);
    ATOMIC_INC_DWORD(&(evm_stats.sumQueueDepth), wp->queueDepth);
    ATOMIC_INC_DWORD(&(evm_stats.numQueueDepth), 1);
    evm_stats.maxQueueDepth = MAX(evm_stats.maxQueueDepth, wp->queueDepth);
#endif // DEBUG
    evm_io_complete(queue, &lcb, &pcb);
          
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_io_scsiErrRsp
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

void
evm_io_scsiErrRsp(scsi_errRspMsg_t *    rsp)

{
    evm_ioQueue_t * queue = NULL;
    evm_qhandle_t   handle;
    evm_lioCb_t *   lcb = NULL;
    evm_pioCb_t *   pcb = NULL;

    /* Get handle from response.
     */
    handle = (evm_qhandle_t) rsp->hdr.cHandle;

    /* Select to which IO queue this request belongs.
     */
    if (handle.parts.dataType == EVM_META_DATA_REQ)
    {
        queue = &metaQ;
    }
    else
    {
        queue = &userQ;
    }
    
    /* Find the physical I/O control block.
     */
    NFX_SPIN_LOCK(&evmQueueLock);
    EVM_Q_GET_PCB(queue, &handle, pcb);
    if (pcb == NULL)
    {
        if (evmErrors.nStaleRsp == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: Stale response. "
                      "No physical I/O control block found for given handle.",
                      handle.parts.idx, handle.parts.genNb);
        }
        evmErrors.nStaleRsp++;
        
        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        return;
    }

    /* Find the logical I/O control block.
     */
    EVM_Q_GET_LCB(queue, &handle, lcb);
    if (lcb == NULL)
    {
        if (evmErrors.nStaleRsp == 0)
        {
            evmErrors.isError = TRUE;
            evm_trace(EVM_TRACE_ERRORS, "[][%d, %d]: Stale response. "
                      "No logical I/O control block found for given handle.",
                      handle.parts.idx, handle.parts.genNb);
        }
        evmErrors.nStaleRsp++;
        
        EVM_STAT_STALES;
        NFX_SPIN_UNLOCK(&evmQueueLock);
        return;
    }
    
    pcb->errCode = rsp->eCode;
    evm_io_complete(queue, &lcb, &pcb);
    
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_io_markOutstandingIO
 *
 * Description  : A volume has been put offline,
 *                mark all outstanding IOs corresponding to this volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 15-AUG-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_io_markOutstandingIO(evm_volId_t * volId)
    
{
    evm_ioQueue_t * queue = NULL;
    evm_qelt_t *    pCurr = NULL;
    evm_qelt_t *    pNext = NULL;
    evm_qhandle_t   handle;

    NFX_SPIN_LOCK(&evmQueueLock);
    
    for (queue = &metaQ; ; queue = &userQ)
    {        
        EVM_Q_GO_THRU_PIOPEND(pCurr, pNext, queue, handle)
        {
            evm_lioCb_t * lcb = NULL;
            evm_pioCb_t * pcb = NULL;
            
            NFX_ASSERT(handle.whole != -1);
            EVM_Q_GET_LCB(queue, &handle, lcb);
            EVM_Q_GET_PCB(queue, &handle, pcb);
            NFX_ASSERT((lcb != NULL) && (pcb != NULL));
            
            if (lcb->req.id.whole_thing == volId->whole_thing)
            {
                lcb->lv = NULL;
                pcb->dev = NULL;
            }
        } // EVM_Q_GO_THRU_PIOPEND
        
        EVM_Q_GO_THRU_PIOWAIT(pCurr, pNext, queue, handle)
        {
            evm_lioCb_t * lcb = NULL;
            evm_pioCb_t * pcb = NULL;
            
            NFX_ASSERT(handle.whole != -1);
            EVM_Q_GET_LCB(queue, &handle, lcb);
            EVM_Q_GET_PCB(queue, &handle, pcb);
            NFX_ASSERT((lcb != NULL) && (pcb != NULL));
            
            if (lcb->req.id.whole_thing == volId->whole_thing)
            {
                lcb->lv = NULL;
                pcb->dev = NULL;
            }
        } // EVM_Q_GO_THRU_PIOWAIT
        
        if (queue == &userQ)
            break;
    } // for

    NFX_SPIN_UNLOCK(&evmQueueLock);
    return;
}

/*----------------------------------------------------------------------------*/
