/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-queue.c
 *
 * Description  : Volume Manager (EVM) - Queue management
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

#include "exports/timer-api.h"

#include "nfx-incl.h"
#include "exports/queue-api.h"  // DLL_xxx macros
#include "exports/malloc-api.h"
#include "evm-srvr-includes.h"
#include "evm-srvr-structs.h"
#include "evm-srvr-defines.h"
#include "evm-debug.h"
#include "evm-queue.h"

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

/* Spinlock for the queues.
 */
nfx_spinlock_t evmQueueLock;

/*------------------------------------------------------------------------------
 * Name         : evm_qCreate
 *
 * Description  : Create and initialize the I/O queues.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qCreate(evm_ioQueue_t * qPtr,
            uint32          dataType,
            uint32          qSize,
            uint32          maxPioWait,
            uint32          maxPioPend)
    
{
    uint32  i;
    int32   rc = NFX_OK;

    NFX_ASSERT(qPtr != NULL);
    NFX_ASSERT((maxPioWait + maxPioPend) == qSize);
    
    NFX_SPIN_INIT(&evmQueueLock, SPINLOCK_LEVEL_UNKNOWN);
    
    qPtr->dataType = dataType;
    
    qPtr->size = qSize;
    EVM_RAM_ALLOC_ZERO(qPtr->pMem, evm_qelt_t, (qSize * sizeof(evm_qelt_t)));
    if (qPtr->pMem == NULL)
    {
        return(NFX_ERR);
    }

    qPtr->nbFree = qSize;
    if (dataType == EVM_META_DATA_REQ)
    {
        evm_stats.qCnt.metaFree = qSize;
    }
    else
    {
        evm_stats.qCnt.userFree = qSize;
    }
    DLL_INIT(&(qPtr->pFree));
    for (i = 0; i < qSize; i++) // Build the list of free elements.
    {
        qPtr->pMem[i].handle.parts.idx = i;
        qPtr->pMem[i].handle.parts.genNb = 0;
        qPtr->pMem[i].handle.parts.dataType = dataType;
        DLL_INSERT_TAIL(&(qPtr->pFree), &(qPtr->pMem[i]), pEntry);
    } // for

    qPtr->maxPioWait = maxPioWait;
    qPtr->nbPioWait = 0;
    DLL_INIT(&(qPtr->pPioWait));

    qPtr->maxPioPend = maxPioPend;
    qPtr->nbPioPend = 0;
    DLL_INIT(&(qPtr->pPioPend));
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qDestroy
 *
 * Description  : Destroy the I/O queues.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qDestroy(evm_ioQueue_t *    qPtr)
    
{
    int32 rc = NFX_OK;

    NFX_ASSERT(qPtr != NULL);
    NFX_ASSERT(qPtr->pMem != NULL);
    
    /* Deallocate the memory.
     */
    qPtr->size = 0;
    EVM_RAM_DEALLOC(qPtr->pMem);
    qPtr->nbFree = 0;
    DLL_INIT(&(qPtr->pFree));
    qPtr->nbPioWait = 0;
    DLL_INIT(&(qPtr->pPioWait));
    qPtr->nbPioPend = 0;
    DLL_INIT(&(qPtr->pPioPend));
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qDisplay
 *
 * Description  : Display the content of the queues.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qDisplay(evm_ioQueue_t *    qPtr)
    
{
    evm_qelt_t *    pCurr;
	eeeGMT_t tmpTime;
    uint32          i;

    //NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(qPtr != NULL);
    NFX_ASSERT(qPtr->pMem != NULL);
    
    /* Display queue info.
     */
    printf("Queue:\n");
    printf("\tsize: %d\n", qPtr->size);
    printf("\tpMem: %p\n", qPtr->pMem);
    printf("\tnbFree: %d\n", qPtr->nbFree);
    printf("\tnbPioWait: %d\n", qPtr->nbPioWait);
    printf("\tnbPioPend: %d\n", qPtr->nbPioPend);
    
    /* Go through all elements and display each of them.
     */
    i = 1;
    DLL_ITERATE(pCurr, &(qPtr->pPioWait), pEntry)
    {
        printf("\tElement #%d in physical I/O waiting list\n", i);
        printf("\t\tlcb: %p\n", &(pCurr->lcb));
        printf("\t\tpcb: %p\n", &(pCurr->pcb));
		eee_rtcToTV(&pCurr->timestamp, &tmpTime);
        printf("\t\ttimestamp: %d sec %d usec\n",
               tmpTime.sec, tmpTime.usec);
        printf("\t\thandle: %d/%d\n",
               pCurr->handle.parts.idx, pCurr->handle.parts.genNb);
        i++;
    } // DLL_ITERATE
    printf("\n");
    
    i = 1;
    DLL_ITERATE(pCurr, &(qPtr->pPioPend), pEntry)
    {
        printf("\tElement #%d in physical I/O pending list\n", i);
        printf("\t\tlcb: %p\n", &(pCurr->lcb));
        printf("\t\tpcb: %p\n", &(pCurr->pcb));
		eee_rtcToTV(&pCurr->timestamp, &tmpTime);
        printf("\t\ttimestamp: %d sec %d usec\n",
               tmpTime.sec, tmpTime.usec);
        printf("\t\thandle: %d, %d\n",
               pCurr->handle.parts.idx, pCurr->handle.parts.genNb);
        i++;
    } // DLL_ITERATE
    
    //NFX_SPIN_UNLOCK(&evmQueueLock);
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qInsertToFree
 *
 * Description  : Add an element into the list of free elements.
 *                Free control block if not yet done.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qInsertToFree(evm_ioQueue_t *   qPtr,
                  evm_qhandle_t *   handle)
    
{
    int32   rc;
    int16   index = handle->parts.idx;

    //NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(qPtr != NULL);

    if ((0 <= index) && (index < qPtr->size))
    {
        qPtr->pMem[index].lcb.id = 0;
        qPtr->pMem[index].pcb.handle.whole = 0;        
        qPtr->pMem[index].timestamp = *(eee_rtcZero());
        qPtr->pMem[index].handle.parts.genNb++;
        DLL_INSERT_TAIL(&(qPtr->pFree), &(qPtr->pMem[index]), pEntry);
        qPtr->nbFree++;
        rc = NFX_OK;
        EVM_STAT_Q_INSERT_FREE;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "EVM Error: invalid handle (%d,%d)\n",
                  handle->parts.idx, handle->parts.genNb);
        rc = NFX_ERR;
    }
    
    //NFX_SPIN_UNLOCK(&evmQueueLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qRemoveFromFree
 *
 * Description  : Remove an element from the list of free elements.
 *                (<=> Get a free element.)
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qRemoveFromFree(evm_ioQueue_t *     qPtr,
                    evm_qhandle_t **    handle)
    
{
    evm_qelt_t *    pNewElt = NULL;
    int32           rc;

    //NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(qPtr != NULL);

    DLL_REMOVE_HEAD(&(qPtr->pFree), pNewElt, pEntry);
    if (pNewElt != NULL)
    {   
        *handle = &(pNewElt->handle);
        qPtr->nbFree--;
        rc = NFX_OK;
        EVM_STAT_Q_REMOVE_FREE;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "No more free element - queue full\n");
        *handle = NULL;
        rc = NFX_ERR;
    }
    
    //NFX_SPIN_UNLOCK(&evmQueueLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qInsertToPioWait
 *
 * Description  : Add an element into the list of physical I/O waiting requests.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qInsertToPioWait(evm_ioQueue_t *    qPtr,
                     evm_qhandle_t *    handle)
    
{
    int32   rc;
    int16   index = handle->parts.idx;

    //NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(qPtr != NULL);

    if ((0 <= index) && (index < qPtr->size))
    {
        eee_rtc(&(qPtr->pMem[index].timestamp));
        DLL_INSERT_HEAD(&(qPtr->pPioWait), &(qPtr->pMem[index]), pEntry);
        qPtr->nbPioWait++;
        rc = NFX_OK;
        EVM_STAT_Q_INSERT_PIOWAIT;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "EVM Error: invalid handle (%d,%d)\n",
                  handle->parts.idx, handle->parts.genNb);
        rc = NFX_ERR;
    }
    
    //NFX_SPIN_UNLOCK(&evmQueueLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qRemoveFromPioWait
 *
 * Description  : Remove an element from the list of physical I/O
 *                waiting requests.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qRemoveFromPioWait(evm_ioQueue_t *  qPtr,
                       evm_qhandle_t ** handle)
    
{
    evm_qelt_t *    pElt = NULL;
    int32           rc;

    //NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(qPtr != NULL);

    DLL_REMOVE_TAIL(&(qPtr->pPioWait), pElt, pEntry);
    if (pElt != NULL)
    {   
        *handle = &(pElt->handle);
        qPtr->nbPioWait--;
        rc = NFX_OK;
        EVM_STAT_Q_REMOVE_PIOWAIT;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "No more waiting element - queue empty\n");
        *handle = NULL;
        rc = NFX_ERR;
    }
    
    //NFX_SPIN_UNLOCK(&evmQueueLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qInsertToPioPend
 *
 * Description  : Add an element into the list of physical I/O pending requests.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 * Note         : insert element at the tail and not at the head in order to
 *                use DLL_ITERATE_NEXT()
 *------------------------------------------------------------------------------
 */

int32
evm_qInsertToPioPend(evm_ioQueue_t *    qPtr,
                     evm_qhandle_t *    handle)
    
{
    int32   rc;
    int16   index = handle->parts.idx;

    //NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(qPtr != NULL);

    if ((0 <= index) && (index < qPtr->size))
    {
        eee_rtc(&(qPtr->pMem[index].timestamp));
        DLL_INSERT_TAIL(&(qPtr->pPioPend), &(qPtr->pMem[index]), pEntry);
        qPtr->nbPioPend++;
        rc = NFX_OK;
        EVM_STAT_Q_INSERT_PIOPEND;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "EVM Error: invalid handle (%d,%d)\n",
                  handle->parts.idx, handle->parts.genNb);
        rc = NFX_ERR;
    }
    
    //NFX_SPIN_UNLOCK(&evmQueueLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_qRemoveFromPioPend
 *
 * Description  : Remove an element from the list of physical I/O
 *                pending requests.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_qRemoveFromPioPend(evm_ioQueue_t *  qPtr,
                       evm_qhandle_t *  handle)
    
{
    int32   rc;
    int16   index = handle->parts.idx;

    //NFX_SPIN_LOCK(&evmQueueLock);
    NFX_ASSERT(qPtr != NULL);

    if ((0 <= index) && (index < qPtr->size))
    {
        if (!DLL_IS_ISOLATED(&(qPtr->pMem[index]), pEntry))
        {
            DLL_REMOVE(&(qPtr->pPioPend), &(qPtr->pMem[index]), pEntry);
            qPtr->nbPioPend--;
            EVM_STAT_Q_REMOVE_PIOPEND;
        }
        rc = NFX_OK;
    }
    else
    {
        evm_trace(EVM_TRACE_ERRORS, "Invalid handle (%d,%d).\n",
                  handle->parts.idx, handle->parts.genNb);
        rc = NFX_ERR;
    }
    
    //NFX_SPIN_UNLOCK(&evmQueueLock);
    return(rc);
}

/*----------------------------------------------------------------------------*/
