/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-takeover.c,v 1.22 2002/10/30 21:21:04 janv Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-takeover.c
 *
 * Description  : Volume Manager (EVM) - volume takeover management
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

#ifdef LINUX_TEST
#include <sys/time.h>
#endif
#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // alarm()
#include <signal.h>     // signal()

#include "exports/anpssc-api.h"
#include "exports/timer-api.h"
#include "exports/eee-desc.h"
#include "exports/scsi-api.h"
#include "exports/sdm-api.h"
#include "evm-api.h"
#include "evm.h"
#include "evm-takeover.h"

/*------------------------------------------------------------------------------
 * Type definition.
 *------------------------------------------------------------------------------
 */

typedef struct _evm_tk_volume_t evm_tk_volume_t;
struct _evm_tk_volume_t
{
    evm_volId_t         id;
    boolean             checkOnly;
    uint32              delay;
    uint64              clientPrivate;
    void                (*callback) (evm_volId_t   volId,
                                     uint32        status,
                                     uint64        clientPrivate);
    uint32              startTime;
    scsi_id_t           firstLunId;
    evm_StrgAttr_t *    sAttr;
    int32               nbReadMgrArea;
    int32               lastMgrTstamp;
    int32               nbNoMgrAreaChange;
    evm_tk_volume_t *   next;
};

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* Pointer to the beginning of the list of volumes to takeover.
 */
static evm_tk_volume_t *    listTkVol;

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static int32
evm_tk_addVol(evm_tk_volume_t * vol);

static void
evm_tk_delVol(evm_tk_volume_t **    vol);

static void
evm_tk_takeoverDone(evm_tk_volume_t **  vol,
                    int32               rc);

static int32
evm_tk_writeTkoArea(evm_tk_volume_t * vol);

static int32
evm_tk_readMgrArea(evm_tk_volume_t ** vol);

static int32
evm_tk_readTkoArea(evm_tk_volume_t * vol);

/*------------------------------------------------------------------------------
 * Name         : evm_tk_addVol
 *
 * Description  : Add volume to the list.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_tk_addVol(evm_tk_volume_t * vol)

{
    evm_tk_volume_t *   currVol = listTkVol;
    
    /* Make sure this volume is not already in the list.
     */
    while (currVol != NULL)
    {
        if (currVol->id.whole_thing == vol->id.whole_thing)
            return(NFX_ERR);
        currVol = currVol->next;
    } // while
    
    /* Add volume to the list.
     */
    if (listTkVol != NULL)
    {
        vol->next = listTkVol;
    }
    listTkVol = vol;

    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_tk_delVol
 *
 * Description  : Remove volume from the list.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

static void
evm_tk_delVol(evm_tk_volume_t **    vol)

{
    evm_tk_volume_t *   pCurr = listTkVol;
    evm_tk_volume_t *   pPrev = pCurr;
    
    while (pCurr != NULL)
    {
        if (pCurr->id.whole_thing == (*vol)->id.whole_thing)
        {
            if (pCurr == listTkVol)
            {
                listTkVol = pCurr->next;
            }
            else
            {
                pPrev->next = pCurr->next;
            }
            EVM_FREE(pCurr->sAttr);
            EVM_FREE(pCurr);
            EVM_DBG_PRINT(("*vol == NULL"));
            *vol = NULL;
            break;
        }
        pPrev = pCurr;
        pCurr = pCurr->next;
    } // while
}

/*------------------------------------------------------------------------------
 * Name         : evm_tk_takeoverDone
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

static void
evm_tk_takeoverDone(evm_tk_volume_t **  vol,
                    int32               rc)

{
    EVM_FCT_PRINT(("evm_tk_takeoverDone(0x%llx)", (*vol)->id));
    
    /* Return response to caller.
     */
    if ((*vol)->callback != NULL)
    {
        (*vol)->callback((*vol)->id, rc, (*vol)->clientPrivate);
    }

    /* Remove element from the list.
     */
    evm_tk_delVol(vol);

    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_tk_writeTkoArea
 *
 * Description  : Write the takeover private area on the LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_tk_writeTkoArea(evm_tk_volume_t * vol)

{
    int32           rc = NFX_OK;
    evm_tkoArea_t   data;

    EVM_FCT_PRINT(("evm_tk_writeTkoArea(0x%llx)", vol->id.whole_thing));

    data.magic = NFX_STRG_MAGIC;
    memcpy(&data.tkoId, &myNfx, sizeof(evm_nfxId_t));
    EVM_GET_TIME(data.timestamp);
    rc = evm_writeTkoArea(vol->sAttr, &data);
        
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_tk_readMgrArea
 *
 * Description  : Read the manager private area on the LUN
 *                and check if somebody is still managing the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_tk_readMgrArea(evm_tk_volume_t **   vol)

{
    int32                   rc = NFX_OK;
    sdm_readDataRspMsg_t *  readRsp = NULL;
    evm_mgrArea_t *         data = NULL;

    NFX_ASSERT((*vol != NULL) && ((*vol)->sAttr != NULL));

    /* Read the manager area.
     */
    rc = evm_readMgrArea((*vol)->sAttr, &readRsp);
    if ((rc != NFX_OK) || (readRsp == NULL))
    {
        goto evm_tk_readMgrArea_exit;
    }
    (*vol)->nbReadMgrArea++;
    data = (evm_mgrArea_t *) readRsp->data;

#ifdef ISABELLE // debug only!!!
    printf("READ MANAGER AREA:\n");
    printf("\tVOL magic = %x\n", data->magic);
    printf("\tVOL manager id = %d\n", data->mgrId);
    printf("\tVOL timestamp = %d\n", data->timestamp);
    printf("(*vol)->lastMgrTstamp = %d", (*vol)->lastMgrTstamp);
#endif //ISABELLE
    
    /* Check if the timestamp has not been modified since last 2 reads.
     * If it hasn't, the volume is not anymore being managed by a node.
     * We can take it over now.
     */
    if ((*vol)->nbReadMgrArea == 1)
    {
        (*vol)->lastMgrTstamp = data->timestamp;
        if (data->timestamp == 0)
        {
            evm_tk_takeoverDone(vol, rc);
        }
    }
    else if ((*vol)->nbReadMgrArea == 2)
    {
        (*vol)->lastMgrTstamp = data->timestamp;
        (*vol)->nbNoMgrAreaChange = 0;        
    }
    else
    {
        if (data->timestamp != (*vol)->lastMgrTstamp)
        {
            (*vol)->lastMgrTstamp = data->timestamp;
            (*vol)->nbNoMgrAreaChange = 0;
        }
        else
        {
            (*vol)->nbNoMgrAreaChange++;
            if ((*vol)->nbNoMgrAreaChange == 2)
            {
                evm_tk_takeoverDone(vol, rc);
            }
        }
    }

  evm_tk_readMgrArea_exit:
    EVM_FREE(readRsp);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_tk_readTkoArea
 *
 * Description  : Read the takeover private area on the LUN
 *                and check if somebody else wants to take over the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_tk_readTkoArea(evm_tk_volume_t * vol)

{
    int32                   rc = NFX_OK;
    sdm_readDataRspMsg_t *  readRsp = NULL;
    evm_tkoArea_t *         data = NULL;

    NFX_ASSERT((vol != NULL) && (vol->sAttr != NULL));
    
    rc = evm_readTkoArea(vol->sAttr, &readRsp);
    if ((rc != NFX_OK) || (readRsp == NULL))
    {
        goto evm_tk_readTkoArea_exit;
    }
    data = (evm_tkoArea_t *) readRsp->data;

#ifdef ISABELLE // debug only!!!
    printf("READ TAKEOVER AREA:\n");
    printf("\tTKO magic = %x\n", data->magic);
    printf("\tTKO manager id = %d\n", data->tkoId);
    printf("\tTKO timestamp = %d\n", data->timestamp);
#endif //ISABELLE    

    if (!memcmp(&(data->tkoId), &myNfx, sizeof(evm_nfxId_t)))
    {
        // so far, I am the only one trying to takeover this volume
        // write the takeover area
        rc = evm_tk_writeTkoArea(vol);
    }
    else
    {
        /* We were not able to take over the volume during assigned
         * time frame. 
         * Send negative response to caller.
         */
        EVM_ERR_PRINT(("Node %s/%s/%d is also trying to takeover volume 0x%llx!\n",
                       data->tkoId.cluster, data->tkoId.node,
                       data->tkoId.chassisId, vol->id.whole_thing));
        if (vol->callback != NULL)
        {
            vol->callback(vol->id, NFX_ERR, vol->clientPrivate);
        }
        
        /* Remove volume from the list.
         */
        evm_tk_delVol(&vol);
    }

  evm_tk_readTkoArea_exit:
    EVM_FREE(readRsp);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Glocal functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_tk_init
 *
 * Description  : Initialize 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 23-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_tk_init(void)

{
    int32   rc = NFX_OK;

    EVM_FCT_PRINT(("evm_tk_init()"));

    /* Initialize the list of volumes to takeover.
     */
    listTkVol = NULL;

    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_tk_clockTick
 *
 * Description  : Callback function when timer expires.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 23-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_tk_clockTick(void)

{
    int32               rc = NFX_OK;
    evm_tk_volume_t *   currVol = listTkVol;
    evm_tk_volume_t *   nextVol = NULL;
    uint32              now;

    EVM_GET_TIME(now);
    
    while (currVol != NULL)
    {
        nextVol = currVol->next;
        
        EVM_DBG_PRINT(("\tevm_tk_clockTick() - volume = 0x%llx",
                       currVol->id.whole_thing));
        if ((currVol->delay != 0) &&
            (now - currVol->startTime >= currVol->delay))
        {
            /* We were not able to take over the volume during assigned
             * time frame. 
             * Send negative response to caller.
             */
            EVM_ERR_PRINT(("Give up trying to takeover 0x%llx.",
                           currVol->id.whole_thing));
            evm_tk_takeoverDone(&currVol, NFX_ERR);
        }
        else
        {
            rc = evm_tk_readMgrArea(&currVol);
            if (rc != NFX_OK)
            {
                evm_tk_takeoverDone(&currVol, rc);
            }
            
            if ((currVol != NULL) && (currVol->checkOnly == FALSE))
            {
                rc = evm_tk_readTkoArea(currVol);
                if (rc != NFX_OK)
                {
                    evm_tk_takeoverDone(&currVol, rc);
                }
             }
        }
        currVol = nextVol;
    } // while
     
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_tk_takeoverVol
 *
 * Description  : Takeover a volume.
 *
 * check if volume exists (in the view of this NFX)
 * check if volume is not already on-line in this NFX
 * if checks ok then:
 *   start monitoring manager timestamp
 *   start writing the takeover timestamp
 *
 * when delta(owner timestamp) == 0
 *   stop timer
 *   notify EA that volume takeover can proceed
 *
 * when timer expires
 *   notify EA that volume cannot be taken over
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_takeoverVol(evm_MsgVolTakeover_t *  req,
                void (*callback) (evm_volId_t   volId,
                                  uint32        status,
                                  uint64        clientPrivate),
                boolean     checkOnly,
                char *      err)

{
    int32               rc = NFX_OK;
    evm_tk_volume_t *   vol = NULL;
    evm_VolConfig_t     cfg;
    evm_volStrg_t *     strg = NULL;

    NFX_ASSERT(checkOnly == FALSE); // not use for simple online anymore
    
    /* Check if volume exists (in the view of this node).
     */
    rc = get_lv_cfg_byid(req->id, &cfg, err);
    if (rc != NFX_OK)
    {
        EVM_ERR_PRINT(("Volume 0x%llx not seen by this node.",
                       req->id.whole_thing));
        goto evm_takeoverVol_exit;
    }

    /* Check if volume is not already on-line in this node.
     */
    if (cfg.onLine == TRUE)
    {
        sprintf(err, "Volume 0x%llx already managed by this node.",
                req->id.whole_thing);
        EVM_ERR_PRINT(("%s", err));
        rc = EVM_VOL_ALREADY_ONLINE;
        goto evm_takeoverVol_exit;
    }

    /* Get the LUNs composing the volume.
     */
    rc = get_lv_strg(req->id, &strg, err);
    NFX_ASSERT((rc == NFX_OK) && (strg != NULL));
    NFX_ASSERT(strg->num >= 1);
    
    /* Check the volume consistency.
     */
    {
        int             i;
        evm_StrgAttr_t *attr = NULL;
        evm_strgCfg_t   label;
        uint64          totalSize = 0;
        
        for (i = 0; i < strg->num; i++)
        {
            if (strg->au[i].strgId == 0)
            {
                sprintf(err, "Volume 0x%llx is missing a LUN. "
                        "Cannot put online a broken volume.",
                        req->id.whole_thing);
                EVM_ERR_PRINT(("%s", err));
                rc = NFX_ERR;
                goto evm_takeoverVol_exit;
            }
            rc = get_pstrg_cfg(strg->au[i].strgId, &attr, err);
            NFX_ASSERT((rc == NFX_OK) && (attr != NULL));
            rc = evm_readLunLabel(strg->au[i].strgId, attr, &label);
            NFX_ASSERT(rc == NFX_OK);
            if (strcmp(label.cluster, myCluster) ||
                (label.id.whole_thing != req->id.whole_thing))
            {
                EVM_ERR_PRINT(("Volume 0x%llx - LUN #%d 0x%llx label\n"
                               "- label.cluster: %s, myCluster: %s\n"
                               "- label.id: 0x%llx, req->id: 0x%llx",
                               req->id.whole_thing, i+1, strg->au[i].strgId,
                               label.cluster, myCluster,
                               label.id.whole_thing, req->id.whole_thing));
                sprintf(err, "Volume 0x%llx is not consistent (LUN 0x%llx).",
                        req->id.whole_thing, strg->au[i].strgId);
                EVM_ERR_PRINT(("%s", err));
                rc = NFX_ERR;
                goto evm_takeoverVol_exit;
            }
            if (i == 0)
                totalSize = label.currSize;
            totalSize -= attr->blkCnt;
            EVM_FREE(attr);
        } // for
        if (totalSize != 0)
        {
            EVM_ERR_PRINT(("Volume 0x%llx is not complete.\n"
                           "- totalSize: %lld, currSize: %lld",
                           req->id.whole_thing, totalSize, label.currSize));
            sprintf(err, "Volume 0x%llx is not complete.", req->id.whole_thing);
            EVM_ERR_PRINT(("%s", err));
            rc = NFX_ERR;
            goto evm_takeoverVol_exit;
        }
    }
    
    /* Allocate volume and initialize it.
     * Get first LUN and its attributes.
     */
    EVM_ALLOC_ZERO(vol, evm_tk_volume_t, sizeof(evm_tk_volume_t));
    vol->id = req->id;
    vol->checkOnly = checkOnly;
    vol->delay = req->delay;
    vol->clientPrivate = req->hdr.clientPrivate;
    vol->callback = (void *) callback;
    vol->nbReadMgrArea = 0;
    vol->lastMgrTstamp = -1;
    EVM_GET_TIME(vol->startTime);
    vol->firstLunId = strg->au[0].strgId;
    rc = get_pstrg_cfg(vol->firstLunId, &(vol->sAttr), err);
    NFX_ASSERT((rc == NFX_OK) && (vol->sAttr->openHandle >= 0));

    /* Add volume to list of volumes to takeover.
     */
    rc = evm_tk_addVol(vol);
    if (rc != NFX_OK)
    {
        goto evm_takeoverVol_exit;
    }

#ifdef ISABELLE // debug only!!!
    printf("checkOnly = %s\n", (vol->checkOnly == FALSE)? "FALSE":"TRUE");
    printf("clientPrivate = 0x%llx\n", (uint64)(address_t) vol->clientPrivate);
    printf("get_lv_strg()\n");
    printf("volId = %llx\n", req->id.whole_thing);
    printf("rc/err = %d/%s\n", rc, err);
    printf("strg->num = %d\n", strg->num);
    printf("strg->au[0].strgId = %llx\n", strg->au[0].strgId);
#endif // ISABELLE
    
    /* Start monitoring manager timestamp.
     */
    rc = evm_tk_readMgrArea(&vol);
    //NFX_ASSERT(vol != NULL);
    if (rc != NFX_OK)
    {
        goto evm_takeoverVol_exit;
    }
    
    /* Start writing the takeover timestamp.
     */
    if ((vol != NULL) && (vol->checkOnly == FALSE))
    {
        rc = evm_tk_writeTkoArea(vol);
        if (rc != NFX_OK)
        {
            goto evm_takeoverVol_exit;
        }
    }

  evm_takeoverVol_exit:

    if ((rc != NFX_OK) && (vol != NULL))
    {
        evm_tk_delVol(&vol);
    }
    
    EVM_FREE(strg);
    return(rc);
}

/*----------------------------------------------------------------------------*/
