/*------------------------------------------------------------------------------
 *
 * Name: evm_lvdb.c
 *
 * RCS:  $Id: evm_lvdb.c,v 1.40 2002/10/01 16:57:37 isabellec Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: This file contains the methods used to manipulate
 *              EVM's Volume Configuration database. There is a single
 *              instance of these on each SSC.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 2 Jul 2001
 *
 *------------------------------------------------------------------------------
 */
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <db.h>
#include <string.h>
#include "exports/elog-api.h"

#include "evm.h"
#include "evm_loc_defs.h"

/*
	lvdb uses the logical volume id (evm_volId_t) as the key and 
	evm_VolConfig_t as the data.

	nameidx has name string as key and evm_volId_t as data
 */
static DB * lvdb = NULL;    // Logical volume config database
static DB * nameidx = NULL;	// Name string index into above
static DB * lvstrg = NULL;	// id keyed evm_volStrg_t records

/*-----------------------------------------------------------------
 * Name: init_lvdb
 *
 * Description: Open/Initialize the logical volume configuration databases
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 2 May 2001
 *
 *-----------------------------------------------------------------
 */
int init_lvdb(void)
{
    
/*
	Create the local database EVM will use to look up logical
	volume info. This exists only for the life of this run. Thus
	the O_TRUNC. We also create a "name index".

	The configurationd database will be keyed on lvol id. The name index
	will be keyed on the logical volume string name and contain the
	lvol id as its data.
 */
	lvdb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
	nameidx = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_HASH, NULL);
	lvstrg = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
	if (!lvdb || !nameidx || !lvstrg)
    {
		EVM_ERR_PRINT(("EVM failed opening local databases %s",
                       strerror(errno)));
		EVM_ERR_PRINT(("evm_cfgd exits."));
		exit(1);
	}
    
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: close_lvdb
 *
 * Description: 
 *
 * Created by:	
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */

int
close_lvdb(void)

{
    int32 rc = NFX_OK;

    if (lvdb != NULL)
        rc = lvdb->close(lvdb);
    if (nameidx != NULL)
        rc = nameidx->close(nameidx);
    if (lvstrg != NULL)
        rc = lvstrg->close(lvstrg);

    return(rc);
}

/******************************************************************/
// Logical volume specific stuff

/*-----------------------------------------------------------------
 * Name: init_lv_cfg
 *
 * Description: Read the volume configuration and send the message
 *              to evm_lvmgr to assemble its logical volumes and
 *				then to ask the file system to mount them.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 2 May 2001
 *
 *-----------------------------------------------------------------
 */
int init_lv_cfg(void)
{
/*
	Read sequentially through the internal database and send the
	appropriate messages to evm_cntrl (NIMS), and evm_lvmgr.
 */
	return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name         : upd_lv_cfg
 *
 * Description  : Update logical volume config database with this
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 02-MAY-2001
 *
 *------------------------------------------------------------------------------
 */

int
upd_lv_cfg(evm_VolConfig_t *    cfg,
           int                  upd_od,
           char *               err)
    
{
    int             rc = NFX_OK;
	DBT             key;
	DBT             dbrec;
	DBT             idxkey;
	DBT             idxrec;
	int             stat;
    evm_volStrg_t * lvstrg = NULL;

    /* Make sure it's already in there (an update will replace existing)
     */
	key.data = &cfg->id;
	key.size = sizeof(cfg->id);

	stat = lvdb->get(lvdb, &key, &dbrec, 0);
	if (stat)
    {
        /* Key not there or error
         */
		if (stat < 0)
        {
			if (err)
            {
				sprintf(err, "upd: LVol id %llx: get error %s", 
                        cfg->id.whole_thing, strerror(errno));
				EVM_DBG_PRINT(("%s", err));
			}
			rc = NFX_ERR;
            goto upd_lv_cfg_exit;
		}
		else
        {
			if (err)
            {
				sprintf(err, "upd: LVol id %llx not found",
                        cfg->id.whole_thing);
				EVM_DBG_PRINT(("%s", err));
			}
			rc = NFX_ERR;
            goto upd_lv_cfg_exit;
		}
	}
    
    /* Replace the database entry and name index record
     * Errors with the name index DB are not really fatal but we should make
     * some noise.
     */
	else    // The key was found
    {
        /* Use the old name string for deleting the name index because that's
         * what could be changing after all and using the new one would
         * be wrong.
         */
		char *  namstr = ((evm_VolConfig_t *)(dbrec.data))->name;
        
		idxkey.data = namstr;
		idxkey.size = VOL_MAX_NAME;

		stat = nameidx->del(nameidx, &idxkey, 0);
		if (stat < 0)
        {
			if (err)
            {
				sprintf(err, "upd: idx err name %s: del ind %s\n", 
                        namstr, strerror(errno));
				EVM_DBG_PRINT(("%s", err));
			}
		}
		else if (stat)
        {
			if (err)
            {
				sprintf(err, "name idx not found %s",namstr);
				EVM_DBG_PRINT(("%s", err));
			}
		}
        
        /* Use the new name for the index
         */
		idxkey.data = cfg->name;
		idxkey.size = VOL_MAX_NAME;
		idxrec.data = &cfg->id;
		idxrec.size = sizeof(cfg->id);
		stat = nameidx->put (nameidx, &idxkey, &idxrec, 0);
		if (stat < 0)
        {
			if (err)
            {
				sprintf(err, "idx err name %s: put idx %s", 
                        namstr, strerror(errno));
				EVM_DBG_PRINT(("%s", err));
			}
		}

		dbrec.data = cfg;	// Update record with this data
		dbrec.size = sizeof(*cfg);

		stat = lvdb->put (lvdb, &key, &dbrec, 0);
		if (stat < 0)
        {
			if (err)
            {
				sprintf(err, "LVol id %llx: put err %s",
                        cfg->id.whole_thing, strerror(errno));
				EVM_DBG_PRINT(("%s", err));
			}
			rc = NFX_ERR;
            goto upd_lv_cfg_exit;
		}
        
        /* Update on disk vol config data
         */
        if (upd_od &&
            ((get_lv_strg(cfg->id, &lvstrg, err) == NFX_ERR) ||
            (upd_od_vol_cfg(cfg, lvstrg) == NFX_ERR)))
        {
            if (err)
            {
                sprintf(err, "LVol id %llx: cannot update ondisk config",
                        cfg->id.whole_thing);
            }
            EVM_DBG_PRINT(("%s", err));
            rc = NFX_ERR;
            goto upd_lv_cfg_exit;
        }

        /* Report the event that this LV's config has somehow changed
         */
        evm_sendEvt(EVENT_VOL_MODIFY, cfg->id, EVENT_STATE_NOCHANGE);
	}

upd_lv_cfg_exit:
    EVM_FREE(lvstrg);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : add_lv_cfg
 *
 * Description  : Add a logical volume to the database. Checks to make
 *                sure it doesn't exist either by name string nor LVolID.
 *
 * Note         : The volume name might be changed by this function
 *                in the case of duplicate name.
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 14-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

int
add_lv_cfg(evm_VolConfig_t *    cfg,
           boolean              allowDuplicate,
           char *               err)

{
	DBT idx;
	DBT key;
	DBT dbrec;
	int stat;
	evm_volStrg_t strgAlloc;
    
    /* We'll use the feature of the database ops to make sure we're
     * not adding a record with a duplicate id.
     * First we see if the name string exists.
     */
    idx.data = cfg->name;
	idx.size = VOL_MAX_NAME;

	stat = nameidx->get (nameidx, &idx, &dbrec, 0);
	if (stat < 0)
    {
		if (err)
        {
			sprintf(err, "add_lv_cfg: name get error %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (!stat)
    {
        if (allowDuplicate == TRUE)
        {
            char    newVolName[VOL_MAX_NAME];
            int     nb = 1;
            int     nbMax = 9999;
            
            EVM_ERR_PRINT(("EVM: duplicate name %s", cfg->name));
            
            for (nb = 1; nb < nbMax; nb++)
            {
                bzero(newVolName, VOL_MAX_NAME);
                sprintf(newVolName, "%s_duplicate#%d", cfg->name, nb);
                idx.data = newVolName;
                idx.size = VOL_MAX_NAME;
                stat = nameidx->get(nameidx, &idx, &dbrec, 0);
                if (stat < 0)
                {
                    if (err)
                    {
                        sprintf(err, "add_lv_cfg: name get error %s",
                                strerror(errno));
                        EVM_DBG_PRINT(("%s", err));
                    }
                    return NFX_ERR;
                }
                else if (!stat)
                {
                    continue;
                }
                
                EVM_ERR_PRINT(("EVM: renamed %s to %s", cfg->name, newVolName));
                strcpy(cfg->name, newVolName);
                break;
            } // for
            
            if (nb == nbMax)
            {
                EVM_ERR_PRINT(("EVM: cannot rename %s", cfg->name));
                return NFX_ERR;
            }
        }
        else // (allowDuplicate == FALSE)
        {
            sprintf(err, "EVM: duplicate name %s", cfg->name);
            EVM_ERR_PRINT(("%s", err)); 
            return(NFX_ERR);
        }
    }
    
    /* Name string is unique so try to add the database record and let
     * the DB ops ensure the id is unique.
     */
	key.data = &cfg->id;
	key.size = sizeof(cfg->id);
	dbrec.data = cfg;
	dbrec.size = sizeof(*cfg);

	stat = lvdb->put (lvdb, &key, &dbrec, R_NOOVERWRITE);
	if (stat < 0) {
		if (err) {
			sprintf (err, "add_lv_cfg: rec put error %s", strerror(errno));
			EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (stat) {
		if (err) {
			sprintf (err, "add_lv_cfg: dupicate id %llx", cfg->id.whole_thing);
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
    
    /* Database rec added OK so add the name string index now.
     */
	dbrec.data =  &cfg->id;
	dbrec.size = sizeof(cfg->id);
	stat = nameidx->put (nameidx, &idx, &dbrec, R_NOOVERWRITE);
    
    /* These tests are redundant with the get above, but if not done 
     * the put will surely fail at a very inopportune time.
     */
	if (stat < 0) {
		if (err) {
			sprintf (err, "add_lv_cfg: name put error %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (stat) {
		if (err) {
			sprintf (err, "add_lv_cfg: dupicate name %s",cfg->name);
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
    
    /* Finally, add the empty storage allocation configuration
     */
	memset (&strgAlloc, 0, sizeof(strgAlloc));
	dbrec.data = &strgAlloc;
	dbrec.size = sizeof(strgAlloc);

	if (lvstrg->put (lvstrg, &key, &dbrec, 0)) {
		if (err) {
			sprintf (err, "Error creating LV storage alloc");
		}
	}
    
    /* Report the event that this LV's config has somehow changed
     */
    //evm_sendEvt(EVENT_VOL_CREATE, cfg->id, EVENT_STATE_DOWN);

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: get_lv_cfg_byid
 *
 * Description: Get a logical volume config record from the local DB
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 14 June 2001
 *
 *-----------------------------------------------------------------
 */
int
get_lv_cfg_byid(evm_volId_t         id,
                evm_VolConfig_t *   cfg,
                char *              err)
{
	int             stat;
	DBT             key;
	DBT             data;
    evm_VolConfig_t config;
    
    /* Get the config from the LV config DB
     */
	key.data = &id;
	key.size = sizeof(id);

	stat = lvdb->get (lvdb, &key, &data, 0);
	if (stat <0)
    {
		if (err)
        {
			sprintf(err, "get_lv_cfg_byid: db get error %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (stat)
    {
		return NFX_ERR;
	}

	//*cfg = *(evm_VolConfig_t *)(data.data);
    memcpy((void *) &config, (const void *) data.data, data.size);
    *cfg = config;

	return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : del_lv_cfg
 *
 * Description  : Delete a logical volume config record from the local DB
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 14-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

int
del_lv_cfg(evm_volId_t  id,
           char *       err)

{
	int             stat;
	DBT             key;
	DBT             dbdata;
	DBT             idxkey;
    evm_VolConfig_t cfg;
    
    /* Get the LV Config for that id
     */
	key.data = &id;
	key.size = sizeof(id);

	stat = lvdb->get (lvdb, &key, &dbdata, 0);
	if (stat < 0)
    {
		if (err)
        {
			sprintf (err, "del: id get err %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (stat)
    {
		if (err)
        {
			sprintf (err, "del: id %llx not found", id.whole_thing);
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}

    memcpy((void *) &cfg,
           dbdata.data,
           sizeof(evm_VolConfig_t));

    /* Cannot destroy an online LV.
     */
	if (cfg.onLine == TRUE)
    {
		if (err)
        {
			sprintf (err, "Cannot delete online vol");
		}
		return NFX_ERR;
	}
    
    /* Remove the name index entry first
     */
	idxkey.data = cfg.name;
	idxkey.size = VOL_MAX_NAME;

	stat = nameidx->del (nameidx, &idxkey, 0);
	if (stat < 0)
    {
		if (err)
        {
			sprintf (err, "del: name del err %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (stat)
    {
		if (err)
        {
			sprintf (err, "del: name del nf %s", (char *)idxkey.data);
            EVM_DBG_PRINT(("%s", err));
		}
	}
    
    /* Then the LV record
     */
	stat = lvdb->del (lvdb, &key, 0);
	if (stat < 0)
    {
		if (err)
        {
			sprintf (err, "del: lv del err %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (stat)
    {
		if (err)
        {
			sprintf (err, "del: failed id %llx", id.whole_thing);
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
    
    /* Finally, remove the storage record (the caller should have
     * released this storage already).
     */
	stat = lvstrg->del (lvstrg, &key, 0);	// LV's gone so what to do on fail

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: get_lv_ids
 *
 * Description: Dump all LV ids in config database (up to max)
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 14 June 2001
 *
 *-----------------------------------------------------------------
 */
int get_lv_ids(int max, evm_volId_t *ids, char *err)
{
	int num = 0;
	int stat;
	DBT key;
	DBT data;
/*
	Sequentially go through the LV config db recording the LV ids
	therein
 */
	stat = lvdb->seq (lvdb, &key, &data, R_FIRST);

	while (!stat && (num < max)) {
//		ids[num] = ((evm_VolConfig_t *)(data.data))->id;
		memcpy (&ids[num], &(((evm_VolConfig_t *)(data.data))->id), sizeof(evm_volId_t));
		num++;

		stat = lvdb->seq (lvdb, &key, &data, R_NEXT);
	}
/*
	If there was an error or we reached max before coming to the end,
	it is a problem.
 */
	if (stat < 0) {
		if (err) {
			sprintf (err, "Error reading LV database %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
	}
	else if (!stat) {
		if (err) {
			sprintf (err, "Did not get all LV ids");
            EVM_DBG_PRINT(("%s", err));
		}
	}

	return num;
}

/*------------------------------------------------------------------------------
 * Name         : evm_getVolIdsAndNames
 *
 * Description  : Return the list of all volumes ids and names.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int
evm_getVolIdsAndNames(int           max,
                      evm_volId_t * ids,
                      char          names[][VOL_MAX_NAME],
                      char *        err)
{
	int     num = 0;
	int     stat;
	DBT     key;
	DBT     data;
    char    names2[MAX_IDS][VOL_MAX_NAME];
    
    /* Sequentially go through the LV config db recording the LV ids therein
     */
	stat = lvdb->seq (lvdb, &key, &data, R_FIRST);

	while (!stat && (num < max))
    {
		memcpy((void *) &ids[num],
               (const void *) &(((evm_VolConfig_t *)(data.data))->id),
               sizeof(evm_volId_t));
        strcpy(names2[num], ((evm_VolConfig_t *)(data.data))->name);
#ifdef ISABELLE // debug only
        printf("*** %s / %s\n",
               ((evm_VolConfig_t *)(data.data))->name, names2[num]);
#endif // ISABELLE
		num++;

		stat = lvdb->seq (lvdb, &key, &data, R_NEXT);
	}
    
    /*  there was an error or we reached max before coming to the end,
     * it is a problem.
     */
	if (stat < 0)
    {
		if (err)
        {
			sprintf (err, "Error reading LV database %s", strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
	}
	else if (!stat)
    {
		if (err)
        {
			sprintf (err, "Did not get all LV ids");
            EVM_DBG_PRINT(("%s", err));
		}
	}

    memcpy((void *) names, (const void *) names2,
           MAX_IDS * VOL_MAX_NAME * sizeof(char));
    
	return(num);
}

/*------------------------------------------------------------------------------
 * Name         : get_lv_id_byname
 *
 * Description  : Get the LV id given its name
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 20-JUN-2001
 *
 *------------------------------------------------------------------------------
 */
int
get_lv_id_byname(char *         name,
                 evm_volId_t *  id,
                 char *         err)
{
	DBT key;
	DBT data;
	int stat;
    
    /* Get the volume id from the name index.
     */
	key.data = name;
	key.size = VOL_MAX_NAME;

	stat = nameidx->get (nameidx, &key, &data, 0);
	if (stat < 0)
    {
		if (err)
        {
			sprintf (err, "get_lv_id_byname: idx get error %s", 
			         strerror(errno));
            EVM_DBG_PRINT(("%s", err));
		}
		return NFX_ERR;
	}
	else if (stat)
    {
		if (err)
        {
			sprintf (err, "Volume named '%s' not found.", name);
		}
		return NFX_ERR;
	}
    
	memcpy (id, data.data, sizeof(evm_volId_t));
    
	return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name         : get_lv_strg
 *
 * Description  : Retrieve a logical volume's physical storage config
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 03-JUL-2001
 *
 *------------------------------------------------------------------------------
 */

int
get_lv_strg(evm_volId_t         id,
            evm_volStrg_t **    strg,
            char *              err)
    
{
	DBT idkey;
	DBT rec;

	idkey.data = &id;
	idkey.size = sizeof(id);

	if (lvstrg->get(lvstrg, &idkey, &rec, 0))
    {
        sprintf(err, "Error getting LV's storage cfg");
		EVM_DBG_PRINT(("%s", err));
		return(NFX_ERR);
	}

    //*strg = (evm_volStrg_t *) malloc(rec.size);
    EVM_ALLOC(*strg, evm_volStrg_t, rec.size);
    memcpy((void *) *strg, (const void *) rec.data, rec.size);
    
	return(NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: upd_lv_strg
 *
 * Description: Update a logical volume's physical storage config
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 3 Jul 2001
 *
 *-----------------------------------------------------------------
 */
int upd_lv_strg(evm_volId_t id, evm_volStrg_t *strg, char *err)
{
	DBT idkey;
	DBT rec;

	idkey.data = &id;
	idkey.size = sizeof(id);

	rec.data = strg;
	rec.size = VOLSTRG_SIZE(strg);

	if (lvstrg->put (lvstrg, &idkey, &rec, 0)) {
		sprintf (err, "Error updating LV's storage cfg");
		return NFX_ERR;
	}

/*
    No need to update the lv storage config on disk
    upd_lv_cfg() will and evm usually updates both DBs
 */

	return NFX_OK;
}

/*----------------------------------------------------------------------------*/
