/*------------------------------------------------------------------------------
 * RCS:  $Id: evm_reg.c,v 1.16 2002/11/04 22:39:08 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name: evm_reg.c
 *
 * Description: This file contains the functions to manipulate
 *              the EVM parameters in the registry
 *
 * Created by:  Lan Chin
 *
 * Date Created: 05 Nov 2001
 *
 *------------------------------------------------------------------------------
 */
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <db.h>
#include <string.h>

#include "exports/anpssc-api.h"
#include "exports/opt-api.h"
#include "evm-api.h"
#include "evm_loc_defs.h"
#include "evm.h"

/*------------------------------------------------------------------------------
 * Name         : evm_regGetCluster
 *
 * Description  : Get the cluster name from the registry
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_regGetCluster(void)
{
    char ** clusterName = NULL;

    clusterName = get_cfg_val("cluster.name");
    if ((clusterName == NULL) || (clusterName[0] == NULL))
    {
        /* If not found, set if with the incoming value
         */
        EVM_DBG_PRINT(("cluster name not found, using hostname\n"));
        strcpy(myCluster, myNode);
    }
    else        
    {
        strcpy(myCluster, clusterName[1]);
        EVM_DBG_PRINT(("cluster name is %s\n", myCluster));
    }
    free_cfg_val(clusterName);
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_getNewVolId
 *
 * Description  : Generate a new volume id and make sure it's unique.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-JUL-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_getNewVolId(evm_volId_t *   newVolId,
                char *          err)
    
{
    char ** clusterEntries = NULL;
    char ** slaveEntries = NULL;
    char ** reservedEntries = NULL;
    int     i;

    /* Generate a new volume id.
     */
    EVM_GET_NEW_VOLID((*newVolId));
    
    /* Get all volume'ids we know about (stored in configuration file).
     */
    clusterEntries = get_cfg_val("cluster.lv.*");
    slaveEntries = get_cfg_val("slave.*");    
    reservedEntries = get_cfg_val("evm.reserved.volid.*");    
#ifdef ISABELLE // debug only
    {
        int n = 0;
        evm_volId_t id;
        
        printf("\n");
        for (i = 0; clusterEntries[i] != NULL; i = i+2)
        {
            sscanf((rindex(clusterEntries[i], '.') + 1), "%llx",
                   &(id.whole_thing));
            printf("\tclusterEntries[%d] = 0x%llx\n",
                   i, id.whole_thing);
            n++;
        } // for    
        printf("\n");
        for (i = 0; slaveEntries[i] != NULL; i = i+2)
        {
            sscanf((rindex(slaveEntries[i], '.') + 1), "%llx",
                   &(id.whole_thing));
            printf("\tslaveEntries[%d] = 0x%llx\n",
                   i, id.whole_thing);
            n++;
        } // for    
        printf("\n");
        for (i = 0; reservedEntries[i] != NULL; i = i+2)
        {
            sscanf((rindex(reservedEntries[i], '.') + 1), "%llx",
                   &(id.whole_thing));
            printf("\treservedEntries[%d] = 0x%llx\n", i, id.whole_thing);
            n++;
        } // for    
        printf("\n");
        printf("n = %d\n", n);
    }
#endif // ISABELLE

    /* Make sure it's unique (ie. not already assigned within the cluster).
     * Add 1 till volume id is unique.
     */
    while (TRUE)
    {
        boolean     volidIsUnique = TRUE;
        evm_volId_t id;
        
        if (clusterEntries != NULL)
        {
            for (i = 0; clusterEntries[i] != NULL; i = i+2)
            {   
                sscanf((rindex(clusterEntries[i], '.') + 1),
                       "%llx", &(id.whole_thing));
                if (id.whole_thing == newVolId->whole_thing)
                {
                    volidIsUnique = FALSE;
                    break;
                }
            } // for   
        } 

        if ((volidIsUnique) && (slaveEntries != NULL))
        {
            for (i = 0; slaveEntries[i] != NULL; i = i+2)
            {
                sscanf((rindex(slaveEntries[i], '.') + 1),
                       "%llx", &(id.whole_thing));
                if (id.whole_thing == newVolId->whole_thing)
                {
                    volidIsUnique = FALSE;
                    break;
                }
            } // for
        }

        if ((volidIsUnique) && (reservedEntries != NULL))
        {
            for (i = 0; reservedEntries[i] != NULL; i = i+2)
            {
                sscanf((rindex(reservedEntries[i], '.') + 1),
                       "%llx", &(id.whole_thing));
                if (id.whole_thing == newVolId->whole_thing)
                {
                    volidIsUnique = FALSE;
                    break;
                }
            } // for
        }
#ifdef ISABELLE // debug only
        printf("newVolId 0x%llx %s unique\n", newVolId->whole_thing,
               volidIsUnique? "is":"is NOT");
#endif // ISABELLE
        if (volidIsUnique)
        {
            break;
        }
        else
        {
            newVolId->whole_thing++;
        }
    } // while

#ifdef ISABELLE // debug only
    printf("New volume id:\n"
           "\twhole_thing %lld\n"
           "\tclusterUnique %d\n"
           "\tworldUnique %d\n",
           newVolId->whole_thing,
           newVolId->details.u.parts.clusterUnique,
           newVolId->details.worldUnique);
#endif // ISABELLE

    if (clusterEntries != NULL) free_cfg_val(clusterEntries);
    if (slaveEntries != NULL) free_cfg_val(slaveEntries);
    if (reservedEntries != NULL) free_cfg_val(reservedEntries);

    /* Reserve it by storing the volume id in the configuration file.
     */
    {
        char * newReserved[3];

        asprintf(&newReserved[0],
                 "evm.reserved.volid.%llx", newVolId->whole_thing);
        asprintf(&newReserved[1], "null");
        newReserved[2] = NULL;
        put_cfg_val(newReserved);
        free(newReserved[0]);
        free(newReserved[1]);
    }
    
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_isVolNameUnique
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 13-AUG-2002
 *
 *------------------------------------------------------------------------------
 */

boolean
evm_isVolNameUnique(char *  volName)
    
{
    boolean nameIsUnique = TRUE;
    char ** nameEntries = NULL;
    int     i;

    /* Get all volumes' names we know about (stored in configuration file).
     */
    nameEntries = get_cfg_val("lvname.*");

    /* Go through all the names and check if the given name is unique.
     */
    if (nameEntries != NULL)
    {
        for (i = 0; nameEntries[i] != NULL; i = i+2)
        {   
            if (!strcmp((rindex(nameEntries[i], '.') + 1), volName))
            {
                nameIsUnique = FALSE;
                break;
            }
        } // for
        
        free_cfg_val(nameEntries);
    } 

    return(nameIsUnique);
}

/*----------------------------------------------------------------------------*/
