/*
 *-----------------------------------------------------------------
 *
 * Name:	file.c
 *
 * RCS:  $Id: file.c,v 1.24 2002/11/20 23:48:58 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	UNIX-like file interface to eagle filesystem.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	11/07/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>
#include <exports/eee-api.h>
#include <exports/cod-api.h>
#include <exports/esm-api.h>
#include <exports/sbm-api.h>
#include <exports/evm-api.h>
#include <exports/fs-api.h>
#include "file-api.h"
#if defined(SSC)
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <exports/anpssc-api.h>
#define FILE_SSC_SLOT SSC_SLOT
#else
#define FILE_SSC_SLOT nfxSscSlotId
#endif

#if !defined(MIN)
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) < (b)) ? (b) : (a))
#endif

#define FILE_TRACE 

/**
 * Table of open files.
 */
typedef struct file_TableEntry file_TableEntry;
struct file_TableEntry
{
    fs_Handle       handle;
    fs_Handle       parentHandle;   // TODO - need this?
    int64           offset;
    boolean         inUse;
    int32           slot;
};
#define FILE_ENTRY_MAX  64
static file_TableEntry files[FILE_ENTRY_MAX];

/**
 * Messages defined by the file library.
 */
enum
{
    FILE_MSG_SBM_RESP       = FILE_MSG_BASE,
    FILE_MSG_PROXY_WRITE,
    FILE_MSG_MAX

};

/**
 * Header for file library messages.
 */
typedef struct file_MsgHdr
{
    uint32          msgType;
    esm_handle_t	esmHandle;	
    uint64          offset;
    fs_Handle       fileHandle;

} file_MsgHdr;

NFX_ASSERT_STATIC(sizeof(file_MsgHdr) <= FILE_HEADER_MAX);


#if defined(SSC)
#define eee_ramAlloc    malloc
#define eee_ramDealloc  free
#define panic
#else
nfx_spinlock_t  file_spin;
#define FILE_RESPONSE_EVENT     0


int32 file_codCreates = 0;
int32 file_writes = 0;
int32 file_reads = 0;
int32 file_commits = 0;
int32 file_lookups = 0;
int32 file_getAttrs = 0;
int32 file_mkdirs = 0;


#ifdef NFP_TXRX
/**
 * Build and send an fs write request per the specified
 * arguments.
 */
static int32
file_proxySendWriteReq(
    uint32          srcPort,
    esm_handle_t    esmHandle,
    fs_Handle       fileHandle,
    uint64          offset,
    uint32          length,
    nfx_cod_t       cod)
{
    fs_MsgWriteReq  *fsReq;
    eee_desc_t      *fsedesc;
    uint32          destPort;
    int32           rc;
    uchar8          *buf;

    destPort = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                         NFX_NFP_FP_ANY,
                                         eee_getApplicationId("efs"),
                                         EEE_MGMT,
                                         NFX_COS_HIGH,
                                         PORT_BEST_FIT);

    fsedesc = eee_allocateDesc(EEE_FP_POOL, NFX_COS_HIGH);
    if (fsedesc == NULL)
    {
         printf("file:  doProxyWrite failed to alloc fsedesc\n");
		sbm_freeCOD(cod, 1);
        return NFX_ERR;
    }
    buf = eee_allocateBuffer(EEE_FP_POOL, EEE_BUF_LARGE, NFX_COS_HIGH);
    if (buf == NULL)
    {
        printf("file:  doProxyWrite failed to alloc buffer\n");
		sbm_freeCOD(cod, 1);
        eee_deallocateDesc(fsedesc);
        return NFX_ERR;
    }

    fsReq = (void *) buf;
    fsReq->hdr.requestType = FS_REQ_WRITE;
    fsReq->hdr.attrType = 0;
    fsReq->hdr.proto = FS_PROTO_NFS_V3;
    fsReq->hdr.clientPrivate = esmHandle;
    fsReq->hdr.uid = 0;
    fsReq->hdr.numGids = 0;
    fsReq->hdr.clientPrivate = esmHandle;
    fsReq->fileHandle = fileHandle;
    fsReq->fileOffset = offset;
    fsReq->writeLength = length;
    fsReq->stableHow = FS_WRITE_STABLE_UNSTABLE;
    fsReq->cod[0].cod = cod;
    fsReq->cod[0].offset = 0;
    fsReq->cod[0].length = length;
    fsReq->nbrUsedCod = 1;
    fsedesc->hdr.next = NULL;
    fsedesc->hdr.control = 0;
#if defined(NETEEE_FRAGMENT)
    fsedesc->hdr.offset = 0;
#endif           

    /*
     * Tell the fs to respond to the originator
     * on the MGMT plane, that way it can get
     * to any originator (SSC, TXRX, FC).
     */
    srcPort &= ~EDESC2_DATA_TYPE_MASK;
    srcPort |= EEE_MGMT << EDESC2_DATA_TYPE_SHIFT;
    fsedesc->hdr.src_port = srcPort;

    fsedesc->hdr.dest_port = destPort;
    fsedesc->bd[1].buf = buf;
    fsedesc->bd[1].len = sizeof(fs_MsgWriteReq);
    fsedesc->bd[1].control = EDESC_BD_LARGE | EDESC_BD_CMD_IN_USE;
    EEE_SET_NUM_BUFS(fsedesc, 1);
    EEE_SET_PKT_LEN(fsedesc, sizeof(fs_MsgWriteReq));

    rc = eee_forwardPacket(fsedesc);
    if (rc != NFX_OK)
    {
        printf("file:  doProxyWrite failed to forward fs message\n");
		sbm_freeCOD(cod, 1);
        eee_freePacket(fsedesc);
    }

    return rc;

} /* file_proxySendWriteReq */

/**
 * Handle the proxy write request.  This is used with
 * REV_2 of the SBM.  It runs on the TXRX proc and receives
 * write requests from other cpu's in the system.  Here we
 * store the data as a cod and send the write request to the
 * fp.  The fp will respond to the originator of the request.
 */
static void
file_doProxyWrite(
    eee_desc_t  *edesc)
{
#ifdef SBM_REV_2
    eee_descBuf_t   *bd;
    file_MsgHdr     *fileHdr;
    nfx_cod_t       cod;
    uint32          srcPort = edesc->hdr.src_port;
    int32           rc;

    // Move the first buffer pointer past
    // the fileMsgHdr.
    bd = edesc_firstBuf(edesc);
    fileHdr = (file_MsgHdr *)bd->buf;
    bd->len -= sizeof(file_MsgHdr);
    bd->buf += sizeof(file_MsgHdr);

    // Decrement length.
    EEE_SET_PKT_LEN(edesc, EEE_GET_PKT_LEN(edesc) - sizeof(file_MsgHdr));

    if (!sbm_verifyPktInTXRXPool(edesc))
    {
        eee_descPtr_t   tmp;
        uint32          numBufs = EEE_GET_NUM_BUFS(edesc);

        if (sbm_rx_copy_to_txrx(edesc, &tmp, numBufs) != NFX_OK)
        {
            printf("file_doProxy: Error from sbm_rx_copy_to_txrx\n");
            eee_freePacket(edesc);
            return;
        }
        edesc = tmp;
    }

    // Store it in the sac.  Sac now owns the descriptor.
    cod = sbm_getCODFromDesc(edesc, nfxMySlotId, nfxMyCPUId);
    sbm_sendCODToSac(cod);

    // Send the write request to the fp, indicating
    // the originator of the write as the source.
    rc = file_proxySendWriteReq(srcPort,
                                fileHdr->esmHandle,
                                fileHdr->fileHandle,
                                fileHdr->offset,
                                EEE_GET_PKT_LEN(edesc),
                                cod);
    if (rc != NFX_OK)
    {
        printf("file: proxy write failed (%d)\n", rc);
    }

#endif // SBM_REV_2

} /* file_doProxyWrite */

#endif // NFP_TXRX


/**
 * Receive a message from EEE.
 */
void
file_receiveEEEMessage(
    eee_desc_t   *edesc)
{
    eee_descBuf_t   *bd;
    fs_MsgRespHdr   *fsRespHdr;
    evm_msgHdr_t    *evmRespHdr;
    file_MsgHdr     *fileMsgHdr;
    union {
        file_MsgHdr     file;
        fs_MsgRespHdr   fs;
        evm_msgHdr_t    evm;
    }               flatHdr;
    esm_handle_t    esmHandle = 0;
    esm_event_t     esmEvent;
    uint32          msgType;
    int32           rc, msgLen;
    void            *msgHdr;

    msgLen = EEE_GET_PKT_LEN(edesc);
    //FILE_TRACE("file: received %d byte eee message from %x to %x\n", msgLen,
    //    edesc->hdr.src_port, edesc->hdr.dest_port);

    //eee_descDisplay(edesc);
    //bmc12500Eth_printBuf(edesc->bd[1].buf, edesc->bd[1].len, 1);

    bd = edesc_firstBuf(edesc);
    if (EEE_GET_NUM_BUFS(edesc) > 1 &&
        bd->len < sizeof flatHdr)
    {
        // Copy the header to a flat buffer.
        rc = eee_copyPacketToLocalBuffer(&flatHdr,
                                         edesc,
                                         sizeof flatHdr,
                                         0);
        msgType = *((uint32 *) &flatHdr);
        msgHdr = &flatHdr;
    }
    else
    {
        msgType = *((uint32 *)bd->buf);
        msgHdr = bd->buf;
    }

    if (msgType >= FS_MSG_BASE && msgType < FS_REQ_MAX)
    {
        if (msgLen < sizeof(fs_MsgRespHdr))
        {
            printf("file: short message, dropped\n");
            eee_freePacket(edesc);
            return;
        }
        fsRespHdr = (fs_MsgRespHdr *) msgHdr;
        esmHandle = fsRespHdr->clientPrivate;
        //FILE_TRACE("file: response message from fs, esmHandle %x\n",
        //            esmHandle);
    }
    else if (msgType >= EVM_MSG_BASE && msgType < EVM_DUMMY)
    {
        if (msgLen < sizeof(evm_msgHdr_t))
        {
            printf("file: short message, dropped\n");
            eee_freePacket(edesc);
            return;
        }
        evmRespHdr = (evm_msgHdr_t *) msgHdr;
        esmHandle = evmRespHdr->clientPrivate;
        //FILE_TRACE("file: response message from evm, esmHandle %x\n",
        //            esmHandle);
    }
    else if (msgType >= FILE_MSG_BASE && msgType < FILE_MSG_MAX)
    {
        if (msgLen < sizeof(file_MsgHdr))
        {
            printf("file: short message, dropped\n");
            eee_freePacket(edesc);
            return;
        }
        fileMsgHdr = (file_MsgHdr *) msgHdr;
        esmHandle = fileMsgHdr->esmHandle;

        if (msgType == FILE_MSG_PROXY_WRITE)
        {
            //FILE_TRACE("file: proxy write req from esmHandle %x\n",
            //           esmHandle);
            esmHandle = 0;
        }
        else if (msgType == FILE_MSG_SBM_RESP)
        {
            //FILE_TRACE("file: response message from sbm, esmHandle %x\n",
            //           esmHandle);
        }
        else
        {
            printf("file: unknown message type, dropped (%d)\n",
                   msgType);
            eee_freePacket(edesc);
            return;
        }
    }
    else
    {
        printf("file: unknown msgType in response message %x\n", msgType);
    }

    if (esmHandle)
    {
        uchar8  *msg;

        // Copy the message to a contiguous malloc buffer
        // and free the received packet.
        msg = eee_ramAlloc(msgLen);
        if (!msg)
        {
            printf("file: no memory\n");
            eee_freePacket(edesc);
            return;
        }
    
        rc = eee_copyPacketToLocalBuffer(msg,
                                         edesc,
                                         msgLen,
                                         0);
        eee_freePacket(edesc);
        if (rc != msgLen)
        {
            printf("file: failed to eee_copyPacketToLocalBuffer\n");
            return;
        }

        // Deliver the response to the waiting thread.
        esmEvent.event = FILE_RESPONSE_EVENT;
        esmEvent.p1.u.p = msg;
        esmEvent.p2.u.w = msgLen;
        rc = esm_deliverEvent(esmHandle, &esmEvent);
        if (rc != NFX_OK)
        {
            printf("file: failed to esm_deliverEvent for response message\n");
        }
    }
    else if (msgType == FILE_MSG_PROXY_WRITE)
    {
#ifdef NFP_TXRX
        if (msgLen > sizeof(file_MsgHdr))
        {
            file_doProxyWrite(edesc);
        }
        else
        {
            printf("file: short packet, no data in write\n");
            eee_freePacket(edesc);
        }
#else
        printf("file: proxy write rcvd on wrong proc, ignored\n");
        eee_freePacket(edesc);
#endif
    }
    else
    {
        printf("file: bad esm handle in response message\n");
        eee_freePacket(edesc);
    }
}

/**
 * Initialize this module, when running under EEE.
 */
void
file_initModule()
{
    uint32  appId;
    int32   rc;

    // Obtain our application id.
    appId = eee_allocateApplicationId("file");
    if (appId == NFX_ERR)
    {
        panic("failed to allocate file application id");
    }

    // Register receive message function with eee.
    rc = eee_registerReceiveFunc(   appId,
                                    file_receiveEEEMessage,
                                    file_receiveEEEMessage);
    if (rc != NFX_OK)
    {
        panic("failed to register file receive message function");
    }

    NFX_SPIN_INIT(&file_spin, SPINLOCK_LEVEL_FILE);

} /* file_initModule */
#endif // SSC

/**
 * Send a message to the specified destination, optionally waiting
 * for and delivering a response.
 */
static int32
file_sendMsg(
    void            *msg,
    int32           len,
    void            *respBuf,
    int32           respBufLen,
    int32           slot,
    int32           cpu,
    char            *app,
    int32           dataType)
{
    esm_event_t     event;
    uint32          srcPort, dstPort;
    int32           rc, respSize;
    char            *respPtr;

#if defined(SSC)
    struct timeval      timeout = {20, 0};

    //FILE_TRACE("file: sending %d byte eee msg to slot %d cpu %d app %s\n",
    //    len, slot, cpu, app);

    rc = sendAgileMsg(-1,                 // system assigned tmp socket
                      msg,
                      len,
                      app,
                      cpu,
                      slot,
                      MGMTPLANE,
                      &timeout,
                      respBuf,
                      respBufLen);
    if (rc < 0)
    {
        //FILE_TRACE("file: sendAgileMsg returned error %d\n", rc);
        return NFX_ERR;
    }
    else if (respBuf && rc == 0)
    {
        //FILE_TRACE("file: sendAgileMsg returned 0 length response\n");
    }
#else

    //FILE_TRACE("file: sending %d byte eee msg to slot %d cpu %d app %s\n",
    //    len, slot, cpu, app);

    srcPort = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                        nfxMyCPUId,
                                        eee_getApplicationId("file"),
                                        dataType,
                                        NFX_COS_HIGH,
                                        PORT_BEST_FIT);
    dstPort = PACK_UNICAST_PORT_ADDRESS(slot,
                                        cpu,
                                        eee_getApplicationId(app),
                                        dataType,
                                        NFX_COS_HIGH,
                                        PORT_BEST_FIT);
    rc = eee_sendMessage(msg,
                         len,
                         dstPort,
                         srcPort);
    if (rc != NFX_OK)
    {
        //FILE_TRACE("file: eee_sendMessage returned %d\n", rc);
        return NFX_ERR;
    }
    else
    {
        //FILE_TRACE("file: sent from 0x%x to 0x%x\n", srcPort, dstPort);
    }

    if (respBuf)
    {
        //FILE_TRACE("file: waiting for response message\n");
        esm_yield(&event, 20);
        //FILE_TRACE("file: response message received\n");
        if (event.event != FILE_RESPONSE_EVENT)
        {
            //FILE_TRACE("file: response wait resumed on wrong event %d\n",
            //           event.event);
            return NFX_ERR;
        }

        respPtr = (char *) event.p1.u.p;
        respSize = (int32) event.p2.u.w;
        if (respPtr == NULL ||
            respSize <= 0)
        {
            //FILE_TRACE("file: response wait resumed on NULL/0 (%x/%d) response\n", 
            //           respPtr, respSize);
            return NFX_ERR;
        }

        memcpy(respBuf, respPtr, MIN(respSize, respBufLen));
        eee_ramDealloc(respPtr);
    }
#endif

    return NFX_OK;

} /* file_sendMsg */

/**
 * Determine the slot a volume is instantiated on.
 */
static int32
file_getSlotForVolume(
    evm_volId_t     volId,
    int32           *slot)
{
    evm_MsgVolGetAttrib_t   req;
    evm_MsgVolCreate_t      resp;
    evm_MsgStat_t           *evmStat;
    evm_msgHdr_t            *evmHdr;
    int32                   rc;

    memset(&req, 0, sizeof req);
    req.hdr.msgType = EVM_VOL_GET_ATTRIB;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    req.id = volId;

    rc = file_sendMsg(&req,
                      sizeof req,
                      &resp,
                      sizeof resp,
                      FILE_SSC_SLOT,
                      NFX_SSC_UNIX,
                      "evm",
                      EEE_MGMT);
    if (rc != NFX_OK)
    {
        //printf("file: Failed to identify slot volume is hosted on.\n");
        return NFX_ERR;
    }
    evmHdr = (evm_msgHdr_t *) &resp;
    if (evmHdr->msgType == EVM_MSG_STAT)
    {
        evmStat = (evm_MsgStat_t *) &resp;
        //printf("file: EVM returned error status (%d)\n", evmStat->stat);
        return NFX_ERR;
    }    
    *slot = resp.config.onNIM;

    return NFX_OK;

} /* file_getSlotForVolume */

/**
 * Decompose "volume:/dir1/dir2/file" into a evm_volId_t and pathname.
 */
static int32
file_decomposePathName(
    char            *name,
    int32           *slot,
    evm_volId_t     *volId,
    char            **pathName)
{
    evm_MsgGetId_t      req;
    evm_MsgStat_t       resp;
    int32               rc;
    char                *sepPtr;

    sepPtr = strchr(name, ':');
    if (sepPtr == NULL)
    {
        //FILE_TRACE("file: no %c separater found in name\n", ':');
        return NFX_ERR;
    }

    // Terminate volume name.
    *sepPtr = '\0';

    // Set pathName ptr.
    *pathName = sepPtr + 1;

    // Do some sanity checking on the resulting strings.
    if (strlen(name) <= 0 ||
        strlen(*pathName) <= 0 ||
        strlen(*pathName) > FS_PATH_MAX)
    {
        // TODO - other checks???
        *sepPtr = ':';
        //FILE_TRACE("file: name %s fails sanity checks", name);
        return NFX_ERR;
    }

    // Send a message to EVM to obtain the volume Id.
    memset(&req, 0, sizeof req);
    req.hdr.msgType = EVM_VOL_GET_ID;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    strncpy(req.name, name, VOL_MAX_NAME);
    
    //FILE_TRACE("file: looking up volumeId for %s\n", req.name);

    // Replace the ':' char.
    *sepPtr = ':';

    rc = file_sendMsg(&req,
                      sizeof req,
                      &resp,
                      sizeof resp,
                      FILE_SSC_SLOT,
                      NFX_SSC_UNIX,
                      "evm",
                      EEE_MGMT);
    if (rc != NFX_OK)
    {
        //printf("file: failed to obtain volume ID\n");
        return NFX_ERR;
    }
    if (resp.stat != NFX_OK)
    {
        //printf("file: EVM returned error status (%d)\n", resp.stat);
        return NFX_ERR;
    }

    *volId = resp.payload.lVolId;

    rc = file_getSlotForVolume(*volId, slot);

    return rc;

} /* file_decomposePathName */

/**
 * Obtain the root fs_Handle for a volume.
 */
static int32
file_getRoot(
    int32           slot,
    evm_volId_t     volId,
    fs_Handle       *handle)
{
    fs_MsgGetRootReq    req;
    fs_MsgGetRootResp   resp;
    int32               rc;

    memset(&req, 0, sizeof req);
    req.hdr.requestType = FS_REQ_GET_ROOT;
    req.hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    req.volId = volId;
    
    rc = file_sendMsg(&req,
                      sizeof req,
                      &resp,
                      sizeof resp,
                      slot,
                      NFX_NFP_FP2,      // TODO - either fp will do
                      "efs",
                      EEE_MGMT);
    if (rc != NFX_OK)
        return rc;
    if (resp.hdr.fsStatus != NFX_OK)
    {
        //printf("file: fs getroot returned %d\n", resp.hdr.fsStatus);
        return NFX_ERR;
    }

    *handle = resp.dirHandle;

    return NFX_OK;

} /* file_getRoot */

/**
 * Create the file "name" in parentHandle directory
 * and return a handle to it.
 */
static int32
file_createFile(
    char        *name,
    uint32      mode,
    int32       slot,
    fs_Handle   parentHandle,
    fs_Handle   *fileHandle)
{
    fs_MsgCreateReq     *createReq;
    fs_MsgCreateResp    createResp;
    int32               rc, reqSize;
    char                reqBuf[sizeof(fs_MsgCreateReq) + FS_NAME_MAX + 1];

    reqSize = strlen(name) + 1 + sizeof *createReq;
    createReq = (fs_MsgCreateReq *) reqBuf;
    memset(createReq, 0, sizeof *createReq);
    createReq->hdr.requestType = FS_REQ_CREATE;
    createReq->hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    createReq->hdr.clientPrivate = esm_getCurrHandle();
#endif
    createReq->parentHandle = parentHandle;
    createReq->u.attr.mode = mode;
    createReq->nameLen = strlen(name);
    strncpy(createReq->fileName, name, FS_NAME_MAX);

    rc = file_sendMsg(createReq,
                      reqSize,
                      &createResp,
                      sizeof createResp,
                      slot,
                      NFX_NFP_FP2,      // TODO -either fp will do
                      "efs",
                      EEE_MGMT);

    if (rc != NFX_OK)
        return rc;

    if (createResp.hdr.fsStatus != NFX_OK)
    {
        //printf("file: fs create returned %d\n", createResp.hdr.fsStatus);
        return NFX_ERR;
    }

    *fileHandle = createResp.fileHandle;

    return NFX_OK;

} /* file_createFile */

/**
 * Create the directory "name" in parentHandle directory
 * and return a handle to it.
 */
static int32
file_createDir(
    char        *name,
    uint32      mode,
    int32       slot,
    fs_Handle   parentHandle)
{
    fs_MsgMkdirReq      *mkdirReq;
    fs_MsgMkdirResp     mkdirResp;
    int32               rc, reqSize;
    char                reqBuf[sizeof(fs_MsgMkdirReq) + FS_NAME_MAX + 1];

    reqSize = strlen(name) + 1 + sizeof *mkdirReq;
    mkdirReq = (fs_MsgMkdirReq *) reqBuf;
    memset(mkdirReq, 0, sizeof *mkdirReq);
    mkdirReq->hdr.requestType = FS_REQ_MKDIR;
    mkdirReq->hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    mkdirReq->hdr.clientPrivate = esm_getCurrHandle();
#endif
    mkdirReq->parentHandle = parentHandle;
    mkdirReq->dirAttr.mode = mode;
    mkdirReq->nameLen = strlen(name);
    strncpy(mkdirReq->dirName, name, FS_NAME_MAX);

    rc = file_sendMsg(mkdirReq,
                      reqSize,
                      &mkdirResp,
                      sizeof mkdirResp,
                      slot,
                      NFX_NFP_FP2,      // TODO -either fp will do
                      "efs",
                      EEE_MGMT);

    if (rc != NFX_OK)
        return rc;

    if (mkdirResp.hdr.fsStatus != NFX_OK)
    {
        //printf("file: fs mkdir returned %d\n", mkdirResp.hdr.fsStatus);
        return NFX_ERR;
    }

    return NFX_OK;

} /* file_createDir */

/**
 * Lookup the file "name" in parentHandle dir and
 * return a handle to it.
 */
static int32
file_lookupHandle(
    char        *name,
    int32       slot,
    fs_Handle   parentHandle,
    fs_Handle   *fileHandle)
{
    fs_MsgLookupReq     *lookupReq;
    fs_MsgLookupResp    lookupResp;
    int32               rc, reqSize;
    char                reqBuf[sizeof(fs_MsgLookupReq) + FS_NAME_MAX + 1];

    //FILE_TRACE("file_lookupHandle: lookup of %s\n", name);

    reqSize = strlen(name) + 1 + sizeof *lookupReq;
    lookupReq = (fs_MsgLookupReq *) reqBuf;
    memset(lookupReq, 0, sizeof *lookupReq);
    lookupReq->hdr.requestType = FS_REQ_LOOKUP;
    lookupReq->hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    lookupReq->hdr.clientPrivate = esm_getCurrHandle();
#endif
    lookupReq->parentHandle = parentHandle;
    lookupReq->nameLen = strlen(name);
    strncpy(lookupReq->fileName, name, FS_NAME_MAX);
    
    rc = file_sendMsg(lookupReq,
                      reqSize,
                      &lookupResp,
                      sizeof lookupResp,
                      slot,
                      NFX_NFP_FP2,      // TODO -either fp will do
                      "efs",
                      EEE_MGMT);
    if (rc != NFX_OK)
        return rc;

    if (lookupResp.hdr.fsStatus != NFX_OK)
    {
        //printf("file: fs lookup returned %d\n", lookupResp.hdr.fsStatus);
        return NFX_ERR;
    }

    *fileHandle = lookupResp.fileHandle;

    return NFX_OK;

} /* file_lookupHandle */

/**
 * Iterate over pathName, return with fileName and parentHandle
 * updated.
 */
static int32
file_nameIterate(
    int32       slot,
    evm_volId_t volId,
    char        *pathName,
    char        **fileName,
    fs_Handle   *parentHandle)
{
    int32   rc;
    char    *sepPtr, *pathNext;

    rc = file_getRoot(slot, volId, parentHandle);
    if (rc != NFX_OK)
        return NFX_ERR;

    // Eat leading '/' if they exist.
    while (*pathName && *pathName == '/')
        pathName += 1;

    *fileName = pathNext = pathName;
    if (strlen(pathNext) == 0)
    {
        return NFX_ERR;
    }

    sepPtr = strchr(pathNext, '/');
    while(sepPtr != NULL)
    {
        *fileName = sepPtr + 1;
        if (strlen(*fileName) == 0)
            break;

        *sepPtr = '\0';
        if (strlen(pathNext) == 0)
        {
            *sepPtr = '/';
            break;
        }
        rc = file_lookupHandle(pathNext, slot, *parentHandle, parentHandle);
        *sepPtr = '/';
        if (rc != NFX_OK)
            return NFX_ERR;

        pathNext = *fileName;
        sepPtr = strchr(pathNext, '/');
    }

    if (*fileName)
        return NFX_OK;
    else
        return NFX_ERR;

} /* file_nameIterate */

/**
 * Free a file table entry.
 */
static int32
file_freeTableEntry(
    int32       index)
{
    NFX_SPIN_LOCK(&file_spin);
    if (index >= 0 && index < FILE_ENTRY_MAX)
        files[index].inUse = FALSE;
    NFX_SPIN_UNLOCK(&file_spin);

    return NFX_OK;

} /* file_freeTableEntry */

/**
 * Allocate a file table entry.
 */
static int32
file_allocTableEntry(
    int32       *index)
{
    int32   i, rc = NFX_ERR;

    NFX_SPIN_LOCK(&file_spin);
    for (i = 0; i < FILE_ENTRY_MAX; i++)
    {
        if (!files[i].inUse)
            break;
    }
    if (i < FILE_ENTRY_MAX)
    {
        files[i].inUse = TRUE;
        files[i].offset = 0;

        *index = i;
        rc = NFX_OK;
    }
    else
    {
        printf("file: max open files reached\n");
    }
    NFX_SPIN_UNLOCK(&file_spin);

    return rc;

} /* file_allocTableEntry */

/**
 * Obtain a cod handle to tag some write data with.
 */
static uint32
file_getNextCodHandle(
    uint64  *cod)
{
#if defined(SSC)
    uint64  nextCod;
    int32   rc, fd;

    fd = open("/tmp/.codNext", O_RDWR|O_CREAT|O_SYNC, 0666);
    if (fd < 0)
        return NFX_ERR;
    rc = flock(fd, LOCK_EX);
    if (rc < 0) {
        perror("nextCod flock");
        goto err;
    }
    rc = read(fd, &nextCod, sizeof nextCod);
    if (rc == 0) {
        // Empty file, init it.
        nextCod = 0;
    } else if (rc < 0) {
        perror("nextCod read");
        goto err;
    }
    nextCod += 1;
    rc = lseek(fd, 0, SEEK_SET);
    if (rc != 0) {
        perror("nextCod seek");
        goto err;
    }
    rc = write(fd, &nextCod, sizeof nextCod);
    if (rc != sizeof nextCod) {
        perror("nextCod write");
        goto err;
    }   
    rc = flock(fd, LOCK_UN);
    if (rc < 0) {
        perror("nextCod unflock");
        goto err;
    }
    close(fd);

    *cod = ((uint64) FILE_SSC_SLOT << COD_SLOT_SHIFT |
            (uint64) NFX_SSC_UNIX << COD_CPU_SHIFT |
            (uint64) 0 << COD_CHASSIS_SHIFT |  // TODO - chassis of 0 a problem?
            nextCod);
    return NFX_OK;
 err:
    flock(fd, LOCK_UN);
    close(fd);
    printf("file: error getting next cod handle\n");
    return NFX_ERR;
#else
    *cod = COD_GET_NEXT();
    return NFX_OK;
#endif

} /* file_getNextCodHandle */

/**
 * create a file, returns fd on success, < 0 on error.
 */
int32
file_create(
    char        *name,      // eg. "volumeName:/path1/.../pathN/fileName"
    uint32      mode)
{
    fs_Handle   parentHandle, handle;
    evm_volId_t volId;
    int32       rc, fd, slot;
    char        *pathName, *fileName;

    //FILE_TRACE("file: +++ create %s %o +++\n", name, mode);

    rc = file_allocTableEntry(&fd);
    if (rc != NFX_OK)
        return rc;

    rc = file_decomposePathName(name, &slot, &volId, &pathName);
    if (rc != NFX_OK)
        return rc;

    rc = file_nameIterate(slot, volId, pathName, &fileName, &parentHandle);
    if (rc != NFX_OK)
        return rc;
    
    rc = file_createFile(fileName, mode, slot, parentHandle, &handle);
    if (rc != NFX_OK)
    {
        file_freeTableEntry(fd);
        return rc;
    }

    files[fd].slot = slot;
    files[fd].parentHandle = parentHandle;
    files[fd].handle = handle;

    //FILE_TRACE("file: --- created %s ---\n", name);

    return fd;

} /* file_create */

/**
 * create a directory, returns 0 on success, < 0 on error.
 */
int32
file_mkdir(
    char        *name,      // eg. "volumeName:/path1/.../pathN/dirName"
    uint32      mode)
{
    fs_Handle   parentHandle;
    evm_volId_t volId;
    int32       rc, slot;
    char        *pathName, *fileName;

    //FILE_TRACE("file: +++ mkdir %s %o +++\n", name, mode);

    rc = file_decomposePathName(name, &slot, &volId, &pathName);
    if (rc != NFX_OK)
        return rc;

    rc = file_nameIterate(slot, volId, pathName, &fileName, &parentHandle);
    if (rc != NFX_OK)
        return rc;
    
    rc = file_createDir(fileName, mode, slot, parentHandle);
    if (rc != NFX_OK)
        return rc;

    //FILE_TRACE("file: --- made dir %s ---\n", name);

    return NFX_OK;

} /* file_mkdir */


/**
 * Obtain the attributes for a file.
 */
static int32
file_getAttr(
    int32           slot,
    fs_Handle       handle,
    fs_Attr         *attrs)
{
    fs_MsgGetAttrReq    req;
    fs_MsgGetAttrResp   resp;
    int32               rc;

    memset(&req, 0, sizeof req);
    req.hdr.requestType = FS_REQ_GET_ATTR;
    req.hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    req.fileHandle = handle;

    rc = file_sendMsg(&req,
                      sizeof req,
                      &resp,
                      sizeof resp,
                      slot,
                      NFX_NFP_FP2,      // TODO - either fp will do
                      "efs",
                      EEE_MGMT);
    if (rc != NFX_OK)
        return rc;
    if (resp.hdr.fsStatus != NFX_OK)
    {
        //printf("file: fs getAttr returned %d\n", resp.hdr.fsStatus);
        return NFX_ERR;
    }

    *attrs = resp.attr;

    return NFX_OK;

} /* file_getAttr */

/**
 * open an existing file for reading/writing, returns fd on success,
 * < 0 on error.
 */
int32
file_open(
    char        *name)      // eg. "volumeName:/path1/.../pathN/fileName"
{
    fs_Handle   parentHandle, handle;
    evm_volId_t volId;
    int32       rc, fd, slot;
    char        *pathName, *fileName;

    //FILE_TRACE("file: +++ open %s +++\n", name);

    rc = file_allocTableEntry(&fd);
    if (rc != NFX_OK)
        return rc;

    rc = file_decomposePathName(name, &slot, &volId, &pathName);
    if (rc != NFX_OK)
        return rc;

    rc = file_nameIterate(slot, volId, pathName, &fileName, &parentHandle);
    if (rc != NFX_OK)
        return rc;
    
    rc = file_lookupHandle(fileName, slot, parentHandle, &handle);
    if (rc != NFX_OK)
    {
        file_freeTableEntry(fd);
        return rc;
    }

    files[fd].slot = slot;
    files[fd].parentHandle = parentHandle;
    files[fd].handle = handle;

    //FILE_TRACE("file: --- opened %s ---\n", name);

    return fd;

} /* file_open */

/**
 * close an open file.  Send a commit request to the filesystem
 * to stabilize all dirty file data to disk.
 */
int32
file_close(
    int32   fd)
{
    fs_MsgCommitReq     req;
    fs_MsgCommitResp    resp;
    int32               rc;

    //FILE_TRACE("file: close fd %d\n", fd);

    if (fd < 0 || fd >= FILE_ENTRY_MAX)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }
    if (!files[fd].inUse)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }

    req.hdr.requestType = FS_REQ_COMMIT;
    req.hdr.proto = FS_PROTO_NFS_V3;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    req.fsHandle = files[fd].handle;
    req.offset = 0;
    req.count = 0;

    rc = file_sendMsg(&req,
                      sizeof req,
                      &resp,
                      sizeof resp,
                      files[fd].slot,
                      NFX_NFP_FP2,
                      "efs",
                      EEE_MGMT);
    if (rc != NFX_OK)
    {
        printf("file: close failed to send FS commit msg (%d)\n", rc);
        file_freeTableEntry(fd);
        return rc;
    }

    if (resp.hdr.fsStatus != NFX_OK)
    {
        printf("file: FS commit returned %d\n", resp.hdr.fsStatus);
        file_freeTableEntry(fd);
        return NFX_ERR;
    }

    return file_freeTableEntry(fd);
}


/**
 * Write to a file.  Returns num bytes written or < 0 on error.
 * Use a preallocated buffer to send messages.
 */

int32
file_write_use_prealloc(
    int32   fd,
    char    *buf,
    int32   length,
    char    *msgbuf)
{
    fs_MsgWriteReq      req;
    fs_MsgWriteResp     resp;
    file_MsgHdr         *fileHdr;
    gmc_app_hdr_t       *gmcreq;
    int32               thisLength, bdSize;
    int32               rc, numDone = 0;
    nfx_cod_t           cod;
    eee_desc_t          *edesc;
    eee_descBuf_t       *ebuf;
    uchar8              *dataptr;

    //FILE_TRACE("file: +++ write fd %d buf %p len %d +++\n", fd, buf, length);

    if (fd < 0 || fd >= FILE_ENTRY_MAX)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }
    if (!files[fd].inUse)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }

    req.hdr.requestType = FS_REQ_WRITE;
    req.hdr.proto = FS_PROTO_NFS_V3;
    req.hdr.attrType = 0;
    req.hdr.uid = 0;
    req.hdr.numGids = 0;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    req.fileHandle = files[fd].handle;

#ifdef SBM_REV_2
#ifdef NFP_TXRX
    // We're on the TXRX, no need to involve the
    // filelib proxy.

    while (numDone < length)
    {
        thisLength = MIN(EEE_BUFSIZE_LARGE, length - numDone);

        // Store the data in the sac.
        edesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
        if (edesc == NULL)
        {
            printf("file: file_write failed to allocate cod edesc\n");
            return NFX_ERR;
        }
        if (thisLength <= EEE_BUFSIZE_SMALL)
        {
            bdSize = EDESC_BD_SMALL;
            ebuf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_SMALL,
                                      NFX_COS_HIGH);
        }
        else
        {
            bdSize = EDESC_BD_LARGE;
            ebuf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_LARGE,
                                      NFX_COS_HIGH);
        }
        if (ebuf == NULL)
        {
            printf("file: file_write failed to allocate cod buf\n");
            eee_deallocateDesc(edesc);
            return NFX_ERR;
        }
        memcpy(ebuf, buf + numDone, thisLength);
        edesc->hdr.next = NULL;
        edesc->hdr.control = 0;
#if defined(NETEEE_FRAGMENT)
        edesc->hdr.offset = 0;
#endif           
        edesc->hdr.src_port = 0;
        edesc->hdr.dest_port = 0;
        edesc->bd[1].buf = (uchar8 *)ebuf;
        edesc->bd[1].len = thisLength;
        edesc->bd[1].control = bdSize | EDESC_BD_CMD_IN_USE;
        EEE_SET_NUM_BUFS(edesc, 1);
        EEE_SET_PKT_LEN(edesc, thisLength);

        cod = sbm_getCODFromDesc(edesc, nfxMySlotId, nfxMyCPUId);
        sbm_sendCODToSac(cod);

        // Send the write req to the filesystem, and wait here.
        req.fileOffset = files[fd].offset;
        req.writeLength = thisLength;
        req.stableHow = FS_WRITE_STABLE_UNSTABLE;
        req.cod[0].cod = cod;
        req.cod[0].offset = 0;
        req.cod[0].length = thisLength;
        req.nbrUsedCod = 1;

        rc = file_sendMsg(&req,
                          sizeof req,
                          &resp,
                          sizeof resp,
                          files[fd].slot,
                          NFX_NFP_FP_ANY,
                          "efs",
                          EEE_MGMT);
        if (rc != NFX_OK)
        {
            printf("file: write failed to send FS msg (%d)\n", rc);
			sbm_freeCOD(cod, 1);
            break;
        }

        if (resp.hdr.fsStatus != NFX_OK)
        {
            printf("file: FS write returned %d\n", resp.hdr.fsStatus);
            break;
        }
        numDone += thisLength;
        files[fd].offset += thisLength;
    }

#else // NFP_TXRX

    // Not on the TXRX, must send data to filelib
    // proxy running there.
    fileHdr = (void *) msgbuf;
    while (numDone < length)
    {
        thisLength = MIN(FILE_BLOCK_SIZE, length - numDone);

        fileHdr->msgType = FILE_MSG_PROXY_WRITE;
#if !defined(SSC)
        fileHdr->esmHandle = esm_getCurrHandle();
#endif
        fileHdr->offset = files[fd].offset;
        fileHdr->fileHandle = files[fd].handle;

        dataptr = msgbuf + sizeof(file_MsgHdr);
        memcpy(dataptr, buf + numDone, thisLength);

#ifdef FCNIM
        rc = file_sendMsg(fileHdr,
                          sizeof(file_MsgHdr) + thisLength,
                          &resp,
                          sizeof resp,
                          files[fd].slot,
                          NFX_NFP_TXRX_ANY,
                          "file",
                          EEE_MGMT);
#else
        rc = file_sendMsg(fileHdr,
                          sizeof(file_MsgHdr) + thisLength,
                          &resp,
                          sizeof resp,
                          files[fd].slot,
                          NFX_NFP_TXRX_ANY,
                          "file",
                          EEE_USER_DATA);
#endif
        if (rc != NFX_OK)
        {
            printf("file: write failed to send FS msg (%d)\n", rc);
            break;
        }

        if (resp.hdr.fsStatus != NFX_OK)
        {
            printf("file: FS write returned %d\n", resp.hdr.fsStatus);
            break;
        }

        numDone += thisLength;
        files[fd].offset += thisLength;
    }

#endif // NFP_TXRX
#else // SBM_REV_2

    gmcreq = (gmc_app_hdr_t *) msgbuf;
    memset(gmcreq, 0, sizeof *gmcreq);
    memset(&req, 0, sizeof req);
    req.hdr.requestType = FS_REQ_WRITE;
    req.hdr.proto = FS_PROTO_NFS_V3;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    req.fileHandle = files[fd].handle;
    while (numDone < length)
    {
        req.fileOffset = files[fd].offset;
        req.writeLength = MIN(FILE_BLOCK_SIZE, length - numDone);
        req.stableHow = FS_WRITE_STABLE_UNSTABLE;
        rc = file_getNextCodHandle(&cod);
        if (rc != NFX_OK)
            break;
        req.cod[0].cod = cod;
        req.cod[0].offset = 0;
        req.cod[0].length = req.writeLength;
        req.nbrUsedCod = 1;

        gmcreq->control = eee_getApplicationId("efs"); // Notify filesystem
        gmcreq->id.cod = cod;
        memcpy(gmcreq + 1, buf + numDone, req.writeLength);

#ifdef FCNIM
        rc = file_sendMsg(gmcreq,
                          sizeof *gmcreq + req.writeLength,
                          NULL,
                          0,
                          files[fd].slot,
                          NFX_NFP_FP_ANY,
                          "sac",
                          EEE_SAC);
#else
        rc = file_sendMsg(gmcreq,
                          sizeof *gmcreq + req.writeLength,
                          NULL,
                          0,
                          files[fd].slot,
                          NFX_NFP_TXRX_ANY,
                          "sac",
                          EEE_USER_DATA);
#endif
        if (rc != NFX_OK)
        {
            printf("file: write failed to send SBM msg (%d)\n", rc);
            break;
        }
        rc = file_sendMsg(&req,
                          sizeof req,
                          &resp,
                          sizeof resp,
                          files[fd].slot,
                          NFX_NFP_FP2,
                          "efs",
                          EEE_MGMT);
        if (rc != NFX_OK)
        {
            printf("file: write failed to send FS msg (%d)\n", rc);
            break;
        }

        if (resp.hdr.fsStatus != NFX_OK)
        {
            printf("file: FS write returned %d\n", resp.hdr.fsStatus);
            break;
        }
        numDone += req.writeLength;
        files[fd].offset += req.writeLength;
    }

#endif // SBM_REV_2

    //FILE_TRACE("file: --- wrote fd %d len %d ---\n", fd, numDone);

    return numDone;
}


/**
 * Write to a file.  Returns num bytes written or < 0 on error.
 * Writes are done with UNSTABLE (NFSv3) semantics.  A commit
 * request will be sent to stabilize the file to disk upon
 * close().
 */
int32
file_write(
    int32   fd,
    char    *buf,
    int32   length)
{
	char *msgbuf = eee_ramAlloc(sizeof(file_WriteHeader) + length);
	if (msgbuf == NULL) {
		//FILE_TRACE("file_write: not enough memory for msgbuf\n");
		return NFX_ERR;
	}
	else {
		int32 rc = file_write_use_prealloc(fd, buf, length, msgbuf);
		eee_ramDealloc(msgbuf);
		return rc;
	}
} /* file_write */


/**
 * read from a file.  Returns num bytes read, or < 0 on error.
 */
int32
file_read(
    int32   fd,
    char    *buf,
    int32   length)
{
    fs_MsgReadReq           req;
    fs_MsgReadResp          resp;
    fs_MsgTransmitComplete  txcompl;
    struct { // this is ugly...provide enough space for this combo, but copy
             // file_MsgHdr to sbm.msg...yuk
        sbm_sendControlAndSACData_t   sbm;
        file_MsgHdr                   fileMsgSpace;
    } sbmMsg;
    file_MsgHdr         fileMsg;
    int32               rc, numDone = 0;
    char                *rxTmpBuf;

    //FILE_TRACE("file: +++ read fd %d buf %p len %d +++\n", fd, buf, length);

    if (fd < 0 || fd >= FILE_ENTRY_MAX)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }
    if (!files[fd].inUse)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }

    rxTmpBuf = eee_ramAlloc(MIN(FILE_BLOCK_SIZE, length) + sizeof fileMsg);
    if (!rxTmpBuf)
    {
        //FILE_TRACE("file: no memory\n");
        return NFX_ERR;
    }

    memset(&req, 0, sizeof req);
    req.hdr.requestType = FS_REQ_READ;
    req.hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    req.hdr.clientPrivate = esm_getCurrHandle();
#endif
    req.fileHandle = files[fd].handle;
    memset(&txcompl, 0, sizeof txcompl);
    txcompl.hdr.requestType = FS_REQ_TX_COMPL;
    txcompl.hdr.proto = FS_PROTO_NFS_V2;
    memset(&sbmMsg, 0, sizeof sbmMsg);
    sbmMsg.sbm.msg_len = sizeof fileMsg;
    sbmMsg.sbm.cmd = CMD_TO_MCPU_SEND_CONTROL_AND_DATA_TO_SRC;
    fileMsg.msgType = FILE_MSG_SBM_RESP;
#if !defined(SSC)
    fileMsg.esmHandle = esm_getCurrHandle();
#endif
    memcpy(&sbmMsg.sbm.msg, &fileMsg, sizeof fileMsg);

    while (numDone < length)
    {
        req.fileOffset = files[fd].offset;
        req.readLength = MIN(FILE_BLOCK_SIZE, length - numDone);
    
        // Send read req to fs.
        rc = file_sendMsg(&req,
                          sizeof req,
                          &resp,
                          sizeof resp,
                          files[fd].slot,
                          NFX_NFP_FP2,
                          "efs",
                          EEE_MGMT);
        if (rc != NFX_OK)
            break;
        if (resp.hdr.fsStatus != NFX_OK)
        {
            printf("file: fs read returned %d\n", resp.hdr.fsStatus);
            break;
        }

        if (resp.bytesRead > 0)
        {
            // Retrieve cod from gmc.
            sbmMsg.sbm.COD = resp.cod[0].cod;

            if ((resp.nbrUsedCod != 1) ||
                (resp.bytesRead != resp.cod[0].length))
            { 
                printf("file: fs read problem (nbrUsedCod %d, length %d %d)\n",
                       resp.nbrUsedCod, resp.bytesRead, resp.cod[0].length);
                break;
            }
            
            // sbmMsg.sbm.length = resp.bytesRead;
            sbmMsg.sbm.length = resp.cod[0].length;
            sbmMsg.sbm.offset = resp.cod[0].offset;
            rc = file_sendMsg(&sbmMsg,
                              sizeof sbmMsg,
                              rxTmpBuf,
                              resp.bytesRead + sizeof fileMsg,
                              files[fd].slot,
                              NFX_NFP_TXRX_ANY,
                              "sbm",
                              EEE_MGMT);
            if (rc != NFX_OK)
                break;
    
            memcpy(buf + numDone, rxTmpBuf + sizeof fileMsg, resp.bytesRead);
    
            numDone += resp.bytesRead;
            files[fd].offset += resp.bytesRead;
    
            // Send TX_COMPL to fs.
            txcompl.fileHandle = resp.fileHandle;   
            txcompl.fileOffset = resp.fileOffset;   
            txcompl.bytesRead = resp.bytesRead;
            txcompl.fsPrivate = resp.fsPrivate;
            rc = file_sendMsg(&txcompl,
                              sizeof txcompl,
                              NULL,
                              0,
                              files[fd].slot,
                              NFX_NFP_FP2,  // TODO - either fp will do
                              "efs",
                              EEE_MGMT);
            if (rc != NFX_OK)
                break;

            if (resp.bytesRead < req.readLength)
            {
                // eof implied
                break;
            }
        }
        else
        {
            // eof implied
            break;
        }
    }

    eee_ramDealloc(rxTmpBuf);

    //FILE_TRACE("file: --- read fd %d len %d ---\n", fd, numDone);

    return numDone;

} /* file_read */

/**
 * Unlink (remove) a file.
 */
int32
file_unlink(
    char    *name)
{
    fs_MsgRemoveReq     *removeReq;
    fs_MsgRemoveResp    removeResp;
    fs_Handle           parentHandle;
    evm_volId_t         volId;
    int32               rc, reqSize, slot;
    char                *pathName, *fileName;
    char                reqBuf[sizeof(fs_MsgRemoveReq) + FS_NAME_MAX + 1];

    //FILE_TRACE("file: +++ remove %s +++\n", name);

    rc = file_decomposePathName(name, &slot, &volId, &pathName);
    if (rc != NFX_OK)
        return rc;

    rc = file_nameIterate(slot, volId, pathName, &fileName, &parentHandle);
    if (rc != NFX_OK)
        return rc;

    reqSize = strlen(fileName) + 1 + sizeof *removeReq;
    removeReq = (fs_MsgRemoveReq *) reqBuf;
    memset(removeReq, 0, sizeof *removeReq);
    removeReq->hdr.requestType = FS_REQ_REMOVE;
    removeReq->hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    removeReq->hdr.clientPrivate = esm_getCurrHandle();
#endif
    removeReq->parentHandle = parentHandle;
    removeReq->nameLen = strlen(fileName);
    strncpy(removeReq->name, fileName, FS_NAME_MAX);
    
    rc = file_sendMsg(removeReq,
                      reqSize,
                      &removeResp,
                      sizeof removeResp,
                      slot,
                      NFX_NFP_FP2,      // TODO -either fp will do
                      "efs",
                      EEE_MGMT);
    if (rc != NFX_OK)
        return rc;

    if (removeResp.hdr.fsStatus != NFX_OK)
    {
        //printf("file: fs remove returned %d\n", removeResp.hdr.fsStatus);
        return NFX_ERR;
    }

    //FILE_TRACE("file: --- unlink'd %s ---\n", name);

    return NFX_OK;

} /* file_unlink */


/**
 * Rename a file
 */
int32
file_rename(char *nameFrom, char *nameTo)
{
	int reqSize;
	fs_MsgRenameReq *renameReq;
	fs_MsgRenameResp renameResp;
	fs_Handle parentHandleFrom;
	fs_Handle parentHandleTo;
	evm_volId_t volIdFrom;
	char *pathNameFrom;
	char *fileNameFrom;
	int32 slotFrom;
	evm_volId_t volIdTo;
	char *pathNameTo;
	char *fileNameTo;
	int32 slotTo;
	char reqBuf[sizeof(fs_MsgRenameReq) + FS_NAME_MAX + FS_NAME_MAX + 2];
	int32 rc;

	//FILE_TRACE("file: +++ rename %s %s +++\n", nameFrom, nameTo);

	rc = file_decomposePathName(nameFrom, &slotFrom, &volIdFrom, &pathNameFrom);
	if (rc != NFX_OK)
		return rc;

	rc = file_decomposePathName(nameTo, &slotTo, &volIdTo, &pathNameTo);
	if (rc != NFX_OK)
		return rc;

	rc = file_nameIterate(slotFrom, volIdFrom, pathNameFrom, &fileNameFrom,
						  &parentHandleFrom);
	if (rc != NFX_OK)
		return rc;

	rc = file_nameIterate(slotTo, volIdTo, pathNameTo, &fileNameTo,
						  &parentHandleTo);

	if (rc != NFX_OK)
		return rc;

	if (strlen(fileNameFrom) >= FS_NAME_MAX || strlen(fileNameTo) >= FS_NAME_MAX)
		return NFX_ERR;

	reqSize = strlen(fileNameFrom) + strlen(fileNameTo) + 2 + sizeof(fs_MsgRenameReq);
	renameReq = (fs_MsgRenameReq *)reqBuf;
	memset(renameReq, 0, reqSize);
#if !defined(SSC)
    renameReq->hdr.clientPrivate = esm_getCurrHandle();
#endif
	renameReq->hdr.requestType = FS_REQ_RENAME;
	renameReq->hdr.proto = FS_PROTO_NFS_V2;
	renameReq->fromDirHandle = parentHandleFrom;
	renameReq->toDirHandle = parentHandleTo;
	renameReq->fromNameLen = strlen(fileNameFrom);
	renameReq->toNameLen = strlen(fileNameTo);
	strncpy(renameReq->fromName, fileNameFrom, FS_NAME_MAX);
	strncpy(renameReq->fromName + renameReq->fromNameLen + 1,
			fileNameTo, FS_NAME_MAX);

	rc = file_sendMsg(renameReq, reqSize, &renameResp, sizeof(renameResp),
					  slotTo, NFX_NFP_FP2, "efs", EEE_MGMT);
	if (rc != NFX_OK)
		return rc;

	if (renameResp.hdr.fsStatus != NFX_OK) {
		//printf("file: fs rename returned %d\n", renameResp.hdr.fsStatus);
		return NFX_ERR;
	}

	//FILE_TRACE("file: --- renamed %s %s ---\n", nameFrom, nameTo);

	return NFX_OK;
}


/**
 * Remove a directory.
 */
int32
file_rmdir(
    char    *name)
{
    fs_MsgRmdirReq      *rmdirReq;
    fs_MsgRmdirResp     rmdirResp;
    fs_Handle           parentHandle;
    evm_volId_t         volId;
    int32               rc, reqSize, slot;
    char                *pathName, *fileName;
    char                reqBuf[sizeof(fs_MsgRemoveReq) + FS_NAME_MAX + 1];

    //FILE_TRACE("file: +++ rmdir %s +++\n", name);

    rc = file_decomposePathName(name, &slot, &volId, &pathName);
    if (rc != NFX_OK)
        return rc;

    rc = file_nameIterate(slot, volId, pathName, &fileName, &parentHandle);
    if (rc != NFX_OK)
        return rc;

    reqSize = strlen(fileName) + 1 + sizeof *rmdirReq;
    rmdirReq = (fs_MsgRmdirReq *) reqBuf;
    memset(rmdirReq, 0, sizeof *rmdirReq);
    rmdirReq->hdr.requestType = FS_REQ_RMDIR;
    rmdirReq->hdr.proto = FS_PROTO_NFS_V2;
#if !defined(SSC)
    rmdirReq->hdr.clientPrivate = esm_getCurrHandle();
#endif
    rmdirReq->parentHandle = parentHandle;
    rmdirReq->nameLen = strlen(fileName);
    strncpy(rmdirReq->name, fileName, FS_NAME_MAX);
    
    rc = file_sendMsg(rmdirReq,
                      reqSize,
                      &rmdirResp,
                      sizeof rmdirResp,
                      slot,
                      NFX_NFP_FP2,      // TODO -either fp will do
                      "efs",
                      EEE_MGMT);
    if (rc != NFX_OK)
        return rc;

    if (rmdirResp.hdr.fsStatus != NFX_OK)
    {
        //printf("file: fs rmdir returned %d\n", rmdirResp.hdr.fsStatus);
        return NFX_ERR;
    }

    //FILE_TRACE("file: --- rmdir'd %s ---\n", name);

    return NFX_OK;

} /* file_rmdir */


/**
 * Seek on a file.
 */
int32
file_lseek(
    int32   fd,
    int32   whence,
    int64   offset)
{
    int32   rc;

    //FILE_TRACE("file: +++ lseek fd %d whence %d offset %lld +++\n",
    //            fd, whence, offset);

    if (fd < 0 || fd >= FILE_ENTRY_MAX)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }
    if (!files[fd].inUse)
    {
        //FILE_TRACE("file: bad descriptor\n");
        return NFX_ERR;
    }

    switch (whence)
    {
    case FILE_SEEK_SET:
        files[fd].offset = offset;
        break;
    case FILE_SEEK_CUR:
        files[fd].offset += offset;
        break;
    case FILE_SEEK_END: {
        /* Get current size, adjust offset */
        fs_Attr attrs;

        rc = file_getAttr(files[fd].slot, files[fd].handle, &attrs);
        if (rc != NFX_OK)
        {
            //FILE_TRACE("file: lseek failed to get attrs (%d)\n", rc);
            return NFX_ERR;
        }
            
        files[fd].offset = attrs.size + offset;
        break;
    }
    default:
        //FILE_TRACE("file: lseek bad whence (%d)\n", whence);
        return NFX_ERR;
    }

    //FILE_TRACE("file: lseek new offset %lld\n", files[fd].offset);

    return files[fd].offset;

} /* file_lseek */


///
/// TODO - result codes!!!
///
