/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * $Id: bt.h,v 1.16 2002/08/27 18:44:59 lanc Exp $
 */

typedef struct fs_Bt_split fs_Bt_split;
typedef struct fs_Bt_insert fs_Bt_insert;
typedef struct fs_Bt_remove fs_Bt_remove;
typedef struct fs_Bufpair fs_Bufpair;
typedef struct fs_Bt_eek fs_Bt_eek;
typedef struct fs_Bt_eeksearch fs_Bt_eeksearch;
typedef struct fs_Bt_walk fs_Bt_walk;

struct fs_Bt_insert {
	enum {
		FS_BT_INSERT_OK,	/* insert successful */
		FS_BT_NEEDS_SPLIT,	/* node must split */
	} status;

	uint64 key;			/* after a split, key and blknum for new block */
	uint64 blknum;			/* which got promoted */

	boolean got_new_leaf;		/* TRUE when the insert causes a new leaf to get allocated */
};

struct fs_Bt_remove {
	enum {
		FS_BT_REMOVE_OK,
		FS_BT_REMOVE_OK_EMPTY,
		FS_BT_REMOVE_NEEDS_DISPOSAL
	} status;

	/*
	 * Number of blocks freed in this remove operation.
	 *
	 * For directories, this counts only level 0 directory blocks.
	 * For extent maps, this counts only the total length of the
	 * extents freed, not the blocks of the btree index itself.
	 */
	uint64 nblocks;
};

/*
 * After a btree split, new {key,buffer} tuple returned for the records copied
 * into a new child on the left hand side.
 */
struct fs_Bt_split {
	uint64 key;		/* after a split, key which points to left child */
	fs_Buf *buf;		/* after a split, new left child buf */
};

/*
 * Two blknums, returned after a B-tree cleave operation on a node.
 */
struct fs_Bufpair {
	uint64 left, right;
};

struct fs_Bt_eek {
	uint8 *bitmap;
	int linkcount;
	fs_Eekmark marktype;

    /*
     * Set to false if no corrections should be done (eg during walk)
     */
    boolean fixIt;

	/*
	 * Keep track of the blknum at each level last visited;
	 * used to avoid double-counting B-tree inner nodes.
	 */
	uint64 last_blk[FS_INOROOT_MAXLEVEL];

    /*
     * For dir eek scans, >1 hash can be found for collisions,
     * so a successful scan is for a hash,dir_index pair match
     */
    int32 dir_index;
};

struct fs_Bt_eeksearch {
	uint64 key;
	enum {
		FS_BT_EEK_OK,
		FS_BT_EEK_DISPOSE,
        FS_BT_EEK_FAILED,
        FS_BT_EEK_COW,
	} status;
    /*
     * if status is FS_BT_EEK_COW return cow block so parent can modify key
     */
    uint64 cowBlock;
    /*
     * Set to false if no corrections should be done (eg during walk)
     * note: normally this is used to return status/result but in some
     *       instances it is used to check if corrections should be
     *       made (eg query_dispose routines)
     */
    boolean fixIt;
    /*
     * For dir scan searchs, return TRUE if (hash,dir_index) pair was found
     */
    boolean found;
};

struct fs_Bt_walk {
	boolean (*worker)(fs_Context *context, fs_Inode *inode, fs_WalkResult *result, void *arg0);
	void *arg;

	/*
	 * Keep track of the blknum at each level last visited;
	 * used to avoid double-counting B-tree inner nodes.
	 */
	uint64 last_blk[FS_INOROOT_MAXLEVEL];
};

/*
 * Btree access "macros"
 */
static inline uint64 *
get_key(fs_Btree *t, int i)
{
	return &t->keys[t->index[i]];
}

static inline uint64 *
get_left(fs_Btree *t, int i)
{
	return &t->left[t->index[i]];
}


extern fs_Buf *get_buf(uint64 blknum);
extern fs_Buf *alloc_buf(fs_Inode *inode);
extern void fs_put_buf(fs_Buf *);
extern fs_Buf *f_get_buf(fs_Inode *, uint64 offset);
extern fs_Inode *alloc_inode(void);
extern fs_Inode *get_inode(uint64 inum);

/*
 * Generic TOUCH macros -- XXX: should generalize outside of btree
 * library.
 */
#define TOUCH_ELEM(buf, p) fs_bt_touch(context, buf, p, sizeof *(p))
#define TOUCH(buf, p, len) fs_bt_touch(context, buf, p, len)

static inline void
fs_bt_touch(fs_Context *context, fs_Buf *buf, void *p0, int len)
{
	char *p = p0;
	/* range check with unsigned compare! */
	assert((unsigned) (p - buf->u.data) < PAGE_SIZE);
	assert((unsigned) (p + len - buf->u.data) <= PAGE_SIZE);
	FS_BUF_WRITE(buf, p - buf->u.data, len);
}

void fs_emap_walk_extent(
	fs_Context *context,
	fs_Inode *inode,
	fs_Extent e,
    uint64 offset,
	boolean (*worker)(fs_Context *context, fs_Inode *inode, fs_WalkResult *result, void *arg0),
	void *arg0);
