/*
 * Name:  fs-cod-rev2.c
 *
 * RCS:  $Id: fs-cod-rev2.c,v 1.12 2002/12/05 22:54:54 maximk Exp $
 *
 * Description:  Functions for communicating cod manipulations
 *               with revision 2 of the SAC Buffer Manager (SBM).
 * 
 * Date Created:  8/19/2002
 */

#if !defined(SSC)

#include "fs-includes.h"

#if defined(FS_STANDALONE)
#include "mcpu-sim.h"
#endif


/**
 * A special pre-initialized cod filled with '0's.
 *
 * This cod should never be freed.
 */
nfx_cod_t fs_nullCOD = -1;
uint32    fs_codCreates = 0;
uint32    fs_codFrees = 0;
uint32    fs_codMerges = 0;
uint32    fs_codSplits = 0;
uint32    fs_codRxNfs = 0;

union bd_control_len
{
	struct {
		ushort16 control;
		ushort16 len;
		uint32 res_;
	} u;
	uint64 control_len;
};
  


static inline void
fill_short_packet_descriptor(volatile eee_desc_t *edesc,
							 uint32 msglen,
							 uint32 dest_port,
							 uint32 src_port)
{
  union bd_control_len cl;

  edesc->hdr.control = (1 << EDESC_BUFCNT_SHIFT)
	  | EDESC_COMMAND_USE_FIRST_BUFFER
	  | msglen;
  edesc->hdr.offset = 0;
  edesc->hdr.dest_port = dest_port;
  edesc->hdr.src_port = src_port;

  cl.u.control = EDESC_BD_CMD_IN_USE | EDESC_BD_CMD_NO_DEALLOC;
  cl.u.len = msglen;
  *((volatile uint64 *)edesc + 4) = cl.control_len;

  edesc->bd[0].buf = (uchar8 *)&edesc->bd[1];
}


/**
 * Split a cod into two cods at the specified offset.
 */
int32
fs_cod_split (
    nfx_cod_t   src_COD,
    uint32      src_offset,
    nfx_cod_t   dst_COD)
{
#if defined(FS_STANDALONE)

  FS_TRACE(NULL, FS_TRACE_COD,
           ("fs_cod_split, src cod %#llx, dst cod %#llx", src_COD, dst_COD));

  fs_codSplits += 1;

  return (mcpu_sim_split (src_COD, src_offset, dst_COD));

#else

  volatile sbm_splitCOD_t  *split;
  int32            rc;
  volatile eee_desc_t *edesc;
 
  FS_TRACE(NULL, FS_TRACE_COD, ("fs_cod_split, src cod %#llx, dst cod %#llx", 
           src_COD, dst_COD));

  edesc = eee_allocateDesc(EEE_REMOTE_POOL, NFX_COS_HIGH);
  if (edesc == NULL)
  {
      fs_error(NULL, FS_EEE_ERR, "no desc avail for split");
  }

  fill_short_packet_descriptor(edesc, sizeof(sbm_splitCOD_t),
							   fs_mcpuDestPortId,
							   fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));

  split = (sbm_splitCOD_t *)&edesc->bd[1];
  split->cmd = CMD_TO_MCPU_MANIPULATE_SPLIT;
  split->src_offset = src_offset;
  split->src_COD = src_COD;
  split->dest_COD = dst_COD;

  rc = eee_forwardLocalPktUni(fs_mcpuDestPortId, (struct edesc *)edesc);
  if (rc != NFX_OK)
  {
      fs_error(NULL, FS_EEE_ERR, "cod split msg");
  }

  ATOMIC_INC_INT(&fs_codSplits, 1);

  return rc;
#endif
}


/**
 * Merge the contents of src_COD into dst_COD.  After the
 * merge, src_COD is freed by the sbm.
 */
int32
fs_cod_merge (
    nfx_cod_t   src_COD,
    uint32      src_len,
    uint32      src_offset,
    nfx_cod_t   dst_COD,
    uint32      dst_offset)
{
  /*
   * NOTE:  AFTER THE MERGE, THE SOURCE COD IS INVALID!!!!!!
   * The src_COD is freed  !!!
   */

#if defined(FS_STANDALONE)

  FS_TRACE(NULL, FS_TRACE_COD,
	("fs_cod_merge, src %#llx len %d off %d dst %#llx off %d",
		src_COD, src_len, src_offset, dst_COD, dst_offset));

  fs_codMerges += 1;

  return (mcpu_sim_merge (src_COD, src_len, src_offset, dst_COD, dst_offset));

#else

  volatile sbm_mergeCOD_t  *merge;
  int32            rc;
  volatile eee_desc_t *edesc;

#if 0
  FS_TRACE(NULL, FS_TRACE_COD,
	("fs_cod_merge, src %#llx len %d off %d dst %#llx off %d",
		src_COD, src_len, src_offset, dst_COD, dst_offset));
#endif

  if (src_len == 0)
  {
     fs_cod_free_cod(src_COD);
     return NFX_OK;
  }

  edesc = eee_allocateDesc(EEE_REMOTE_POOL, NFX_COS_HIGH);
  if (edesc == NULL)
  {
      fs_error(NULL, FS_EEE_ERR, "no desc avail for merge");
  }

  fill_short_packet_descriptor(edesc, sizeof(sbm_mergeCOD_t),
							   fs_mcpuDestPortId,
							   fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));

  merge = (sbm_mergeCOD_t *)&edesc->bd[1];
  
  merge->cmd = CMD_TO_MCPU_MANIPULATE_MERGE;
  merge->src_len = src_len;
  merge->src_COD = src_COD;
  merge->src_offset = src_offset;
  merge->dest_offset = dst_offset;
  merge->dest_COD = dst_COD;

  rc = eee_forwardLocalPktUni(fs_mcpuDestPortId, (struct edesc *)edesc);
  if (rc != NFX_OK)
  {
      fs_error(NULL, FS_EEE_ERR, "cod merge msg");
  }

  ATOMIC_INC_INT(&fs_codMerges, 1);

  return rc;
#endif
}


/**
 * Fill the specified cod with 0's for the specified
 * offset and length.  Updates the COD pointer with a
 * new cod handle, the previous one is freed by the sbm.
 */
int32
fs_cod_fill_null (
    nfx_cod_t   *COD,
    uint32      len,
    uint32      offset)
{

#if defined(FS_STANDALONE)

  FS_TRACE(NULL, FS_TRACE_COD,
           ("fs_cod_fill_null, cod %#llx, len %d, offset %d",
           *COD, len, offset));

  return (mcpu_sim_cod_null (*COD, len, offset));
  
#else
  // Create a zero-fill cod, merge the caller's cod into it.
  nfx_cod_t     zeroCod;
  int32         rc;

  zeroCod = fs_cod_get_new_cod();

  FS_TRACE(NULL, FS_TRACE_COD,
           ("fs_cod_fill_null, cod %#llx, len %d, offset %d new cod %#llx", 
           *COD, len, offset, zeroCod));

  rc = fs_cod_create_null(zeroCod, FS_BLOCK_SIZE, 0);
  rc |= fs_cod_merge (*COD, offset, 0, zeroCod, 0);

  *COD = zeroCod;

  return rc;
#endif
}


/**
 * Create a zero-filled cod with the specified handle
 * and length.
 */
int32
fs_cod_create_null (
    nfx_cod_t   COD,
    uint32      len,
    uint32      offset)
{

#if defined(FS_STANDALONE)

  FS_TRACE(NULL, FS_TRACE_COD,
           ("fs_cod_create_null, cod %#llx, len %d, offset %d", 
           COD, len, offset));

  fs_codCreates += 1;

  return (mcpu_sim_cod_null (COD, len, offset));
  
#else
  int32            rc;
  volatile sbm_createCOD_t *nullmsg;
  volatile eee_desc_t *edesc;

  FS_TRACE(NULL, FS_TRACE_COD,
           ("fs_cod_create_null, cod %#llx, len %d, offset %d", 
           COD, len, offset));

  edesc = eee_allocateDesc(EEE_REMOTE_POOL, NFX_COS_HIGH);
  if (edesc == NULL)
  {
      fs_error(NULL, FS_EEE_ERR, "no desc avail for nullmsg");
  }

  fill_short_packet_descriptor(edesc, sizeof(sbm_createCOD_t),
							   fs_mcpuDestPortId,
							   fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));
  
  nullmsg = (sbm_createCOD_t *)&edesc->bd[1];
  nullmsg->cmd = CMD_TO_MCPU_MANIPULATE_CREATE;
  nullmsg->len = len;
  nullmsg->COD = COD;
  nullmsg->fill = 0;

  rc = eee_forwardLocalPktUni(fs_mcpuDestPortId, (struct edesc *)edesc);
  if (rc != NFX_OK)
  {
    fs_error(NULL, FS_EEE_ERR, "cod nullmsg msg");
  }

  ATOMIC_INC_INT(&fs_codCreates, 1);

  return rc;
#endif
}


/**
 * Create a new COD handle.
 */
nfx_cod_t fs_cod_get_new_cod()
{
#if defined(FS_STANDALONE)
    nfx_cod_t COD;

	COD = fs_sim_get_new_cod (3) ; 
    // use id == 3, just a value for the sim, must be diff for all NFS instances
    // use fs_sim_get_new_cod, to ensure unique COD after reading logfile

    FS_TRACE(NULL, FS_TRACE_COD, ("fs_cod_get_new_cod, cod %#llx", 
             COD));

    return COD;
#else
    eee_desc_t  *desc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);

    if (desc == NULL)
    {
        fs_error(NULL, FS_EEE_ERR, "no desc avail for cod");
    }

    return sbm_getCODFromDesc(desc, nfxMySlotId, nfxMyCPUId);
#endif
}


/**
 * Free the specified cod.
 */
void
fs_cod_free_cod(
    nfx_cod_t COD)
{
#if defined(FS_STANDALONE)
  int32 rc ;

  FS_TRACE(NULL, FS_TRACE_COD, ("fs_cod_free_cod, cod %#llx", 
           COD));
  
  rc = mcpu_sim_sbm_freeCOD(COD) ;
  
  if (NFX_OK != rc)
    {
        fs_error(NULL, FS_INTERNAL_ERR,
                "fs_cod_free_cod failed %#llx\n", COD);
    }
  
  fs_codFrees += 1;

#else

  FS_TRACE(NULL, FS_TRACE_COD, ("fs_cod_free_cod, cod %#llx", COD));

  NFX_SPIN_LOCK(&eee_fwdqueue_spin);
  FWD_PACKET(COD | EEE_FWD_ACTION_FREE_COD, EEE_REMOTE_QUEUE_INDEX);
  NFX_SPIN_UNLOCK(&eee_fwdqueue_spin);

  ATOMIC_INC_INT(&fs_codFrees, 1);

#endif

}


/**
 * Initialize this module.
 */
void
fs_cod_init()
{
}

#endif // SSC
