/*
 * Name:  fs-err.c
 *
 * RCS:  $Id: fs-err.c,v 1.57 2002/10/24 21:12:53 isabellec Exp $
 *
 * Description:  Code for logging/printing errors in the filesystem module.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"

int32   fs_errorCounts[FS_ERROR_TYPE_MAX];

typedef struct fs_ErrorTable
{
    char    *message;

} fs_ErrorTable;

fs_ErrorTable fs_errorTable[FS_ERROR_TYPE_MAX] =
{
    { "esm event delivery"},      // FS_DELIVER_EVT_ERR
    { "unknown message"},         // FS_UNK_MSG_ERR
    { "simulator bad args"},      // FS_SIM_ARGS_ERR
    { "simulator file open"},     // FS_SIM_OPEN_ERR
    { "simulator lseek"},         // FS_SIM_LSEEK_ERR
    { "simulator i/o"},           // FS_SIM_IO_ERR
    { "eee error"},               // FS_EEE_ERR
    { "internal error"},      	// FS_INTERNAL_ERR
    { "ram alloc"},               // FS_RAM_ALLOC_ERR
    { "event delivery failed"},   // FS_EVENT_DELIVERY_FAIL
    { "esm create failed"}, 		// FS_ESM_CREATE_ERR
    { "esm destroy failed"},		// FS_ESM_DESTROY_ERR
    { "volume read"},             // FS_READ_ERR
    { "volume write"},            // FS_WRITE_ERR
    { "log entry error"},         // FS_LOG_ENTRY_ERR
    { "incorrect log device"},    // FS_LOG_BAD_DEV_ERR
    { "log write error"},         // FS_LOG_WRITE_ERR 
    { "log read error"},          // FS_LOG_READ_ERR  
    { "log id not found"},        // FS_LOG_ID_NOT_FOUND
    { "log dep not found"},       // FS_LOG_DEP_NOT_FOUND
    { "log device full"},         // FS_LOG_FULL_ERR
    { "volume not mounted"},      // FS_VOLUME_NOT_MOUNTED
    { "volume already mounted"},  // FS_ALREADY_MOUNTED_ERR
    { "bad super magic"},         // FS_BAD_SUPER_MAGIC
    { "bad super version"},       // FS_BAD_SUPER_VERSION
    { "mkfs incomplete"},         // FS_MKFS_INCOMPLETE
    { "bad version in log"},      // FS_LOG_BAD_VERSION
    { "bad magic in log"},       	// FS_LOG_BAD_MAGIC
    { "volume exception"},     	// FS_VOLUME_EXCEPTION
    { "event timeout"},           // FS_EVENT_TIMEOUT_ERR
    { "event volume broken"},         // FS_EVENT_VOL_BROKEN
    { "event volume IO error"},       // FS_EVENT_VOL_IOERROR
    { "event volume mgt conflict"},   // FS_EVENT_VOL_CONFLICT
    { "event volume takeover"},       // FS_EVENT_VOL_TAKEOVER
    { "event volume corrupted"}       // FS_EVENT_VOL_CORRUPT
};

#define CALLER_STRING_MAX   128

boolean fsDoAssert = FALSE;

int32 fs_changeDoAssert(int32 argc, char8 *argv[])
{
    if (argc != 2)
        goto fs_changeDoAssert_printUsage;
    
    if (!strcmp(argv[1], "on"))
    {
        fsDoAssert = TRUE;
        printf("FS assert turned ON.\n");
    }
    else if (!strcmp(argv[1], "off"))
    {
        fsDoAssert = FALSE;
        printf("FS assert turned OFF.\n");
    }
    else
    {
        printf("FS assert is turned %s.\n", (fsDoAssert)? "ON":"OFF");
    }
    return(NFX_OK);
    
  fs_changeDoAssert_printUsage:
    printf("USAGE: fsassert {show|on|off}\n");
    return(NFX_ERR);    
}


/**
 * Thread function to handle turning out the lights at
 * the end of a volume exception.
 */
static void
fs_volumeExceptionLightsOut(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context  *context = arg;
    int i;

    FS_ASSERT(context->isThread);

    // Flush any IO's sitting on the retry queues.
    fs_rdwrDestroyRetryQueue(context);

    fs_volWaitPendingIO(context, TRUE);

    fs_bufStopWritebackDaemonWait(context, TRUE);
    fs_logStopDaemon(context, TRUE);

    fs_volWaitPendingIO(context, TRUE);

    fs_inodeCacheDestroyVolume(context, context->fileSystem);
    fs_bufCacheDestroyVolume(context, context->fileSystem);
    fs_logDestroyVolume(context, context->fileSystem);

    fs_doUnmount(context, context->fileSystem, TRUE);
    
#if !defined(FS_STANDALONE)
    /* Send an event that the filesystem aborted.
     */
    fs_postEvt(context, EVENT_VOL_FSYSABORT, EVENT_STATE_DOWN);
#endif
    
    fs_ramDealloc(FS_RAM_CONTEXT, context);
}

 
/**
 * Handle an exception condition for the specified
 * volume.  Basically, clean up and forcibly unmount.
 */
void
fs_volumeException(
    fs_Context      *context,
    fs_FileSystem   *fs,
    int32           errorType)
{
    esm_handle_t    smHandle = 0;
    boolean         isThread = FALSE;
    boolean         isFirst = FALSE;

    if (context != NULL)
    {
        smHandle = context->smHandle;
        isThread = context->isThread;
    }

    //printf("fs_volumeException: smHandle 0x%x fs %p\n", 
    //           smHandle, fs);

    if (fs == NULL)
    {
        //printf("fs_volumeException NULL fs\n");

        // Assumed to be the mount thread here. (What
        // else would not have a filesystem yet?
        FS_ASSERT(context->reqHdr->requestType == FS_REQ_MOUNT);
        FS_ASSERT(isThread);

        (fs_reqTypeTbl[context->reqHdr->requestType].respFunc)
                                                (context, NFX_ERR);
        (fs_reqTypeTbl[context->reqHdr->requestType].clnupFunc)
                                                (context, FALSE);
        thread_exit();

        /* NOTREACHED */
    }

    if (fs->lightsOut)
    {
        // If anything happens to get this far during
        // lights out processing (the buf alloc thread
        // does, since its never on the active thread
        // list and doesn't get killed earlier), terminate
        // it.

        if (isThread)
        {
            thread_exit();
        }
        else
        {
            return;
        }
    }

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);

    if (fs->state != FS_VOL_STATE_FAILING)
    {
        fs_warn(context, errorType,
            "filesystem error on volId 0x%llx (%s), aborting volume\n",
            fs->volId.whole_thing, fs->super->mountName);

        fs->state = FS_VOL_STATE_FAILING;
        isFirst = TRUE;
    }
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

#if defined(FS_DEBUG_DUMP_LOCK_OWNERS)

    // IF debug print as much as possible and jump into debugger

    if (isFirst)
    {
        fs_Context      *threadContext;

        NFX_SPIN_LOCK(&fs->spin);
        DLL_ITERATE(threadContext, &fs->activeThreadHdr, fsEntry)
        {
            fs_lockPrintOwned (threadContext);
        }
        NFX_SPIN_UNLOCK(&fs->spin);
		printf("MetaWriteback esm 0x%llx\n", 
			   (uint64)((uintptr_t)fs->bufList[FS_BUFLIST_META].wbHandle));
		printf("UserWriteback esm 0x%llx\n", 
			   (uint64)((uintptr_t)fs->bufList[FS_BUFLIST_USER].wbHandle));
    }	
			
#endif

    if (fsDoAssert)
    {
		switch(errorType)
        {
        case FS_READ_ERR:
        case FS_WRITE_ERR:
        case FS_EVENT_VOL_BROKEN:
        case FS_EVENT_VOL_IOERROR:
        case FS_EVENT_VOL_CONFLICT:
        case FS_EVENT_VOL_TAKEOVER:
        case FS_EVENT_VOL_CORRUPT:
        // continue and abort the filesystem
            break;
                
        default:
            // jump to debugger
#ifdef FS_DEBUG
            FS_ASSERT(FALSE);
#else
            fs_abort();
#endif
        } // switch
    }
        
    if (isFirst)
    {
        fs_Context      *threadContext;
        esm_event_t     event;
        int32           rc;

        // Remove this context from the active thread list and
        // signal remaining contexts for termination.
        NFX_SPIN_LOCK(&fs->spin);
        if (context != NULL &&
            !DLL_IS_ISOLATED(context, fsEntry))
        {
            DLL_REMOVE(&context->fileSystem->activeThreadHdr,
                       context, fsEntry);
        }
        DLL_ITERATE(threadContext, &fs->activeThreadHdr, fsEntry)
        {
            event.event = FS_THREAD_EVENT_EXCEPTION;
            rc = thread_event(threadContext->smHandle, &event);
            if (rc != NFX_OK)
            {
                fs_error(NULL, FS_EVENT_DELIVERY_FAIL,
                    "thread exception event");
            }
        }
        NFX_SPIN_UNLOCK(&fs->spin);
    }

    if (context != NULL &&
        context->reqHdr != NULL)
    {
        int32   errorType = NFX_ERR;

        // Send an error response to this context's requestor.
        if (context->reqHdr->requestType < FS_REQ_NFS_MAX)
            errorType = FS_ERR_IO;
        (fs_reqTypeTbl[context->reqHdr->requestType].respFunc)
                                                (context, errorType);
        (fs_reqTypeTbl[context->reqHdr->requestType].clnupFunc)
                                                (context, FALSE);
    }
    else if (isThread)
    {
        // Its a system related thread, terminate in the
        // usual fashion.
        fs_threadTerminate(context);
    }
    else
    {
        // Its a pure state machine or our ioDone
        // routine, not a thread.
    }

    // context ptr no longer valid, don't try to use
    // it from here on.
    context = NULL;

    //printf("ioPendingCount %d active threads %s\n", fs->ioPendingCount,
    //    DLL_IS_EMPTY(&fs->activeThreadHdr) ? "NO" : "YES");

    if (DLL_IS_EMPTY(&fs->activeThreadHdr))
    {
        // I'm the last one standing, I'll turn
        // out the lights.

        // Create a tmp context for the remainder.
        fs_Context *tmpContext = fs_ramAlloc(FS_RAM_CONTEXT,
                                             sizeof(fs_Context));
        bzero(tmpContext, sizeof *tmpContext);
#ifdef FS_DEBUG
        DLL_INIT(&tmpContext->ownerList);
#endif
        tmpContext->fileSystem = fs;
        tmpContext->volId = fs->volId;
        tmpContext->isThread = TRUE;

        fs->lightsOut = TRUE;

        if (isThread)
        {
            fs_volumeExceptionLightsOut(0, tmpContext);
        }
        else
        {
            esm_handle_t    handle;
            int32           rc = NFX_OK;
            // Non-thread can't call lightsOut, so
            // fork one here to handle it.
            handle = thread_create(tmpContext);
            if (handle != 0)
                rc = thread_start(&handle, fs_volumeExceptionLightsOut);
            else
                rc = NFX_ERR;
            if (rc != NFX_OK)
                fs_error(tmpContext, FS_ESM_CREATE_ERR, "exception lights out");
        }
    }

    if (isThread)
    {
        // Return to the executive.
        thread_exit();
    }

} /* fs_volumeException */


/**
 * Print a pretty error message.
 */
static void
fs_printErrorMessage(
    fs_Context      *context,
    int32           elogSeverity,
    int32           errorType,
    char            *format,
    va_list         ap)
{
    evm_volId_t volId;
    char        callerString[CALLER_STRING_MAX];
    
    if (context != NULL)
    {
        volId = context->volId;
    }

    if (format != NULL && format[0] != '\0')
    {
        vsnprintf(callerString, CALLER_STRING_MAX, format, ap);
    }
    else
    {
        callerString[0] = '\0';
    }

    switch (elogSeverity)
    {
    case LOG_CRIT:
        if (context != NULL)
        {
            E_LOG (class_1, LOG_CRIT, fs_applicationId, 0, 0, 0, 
                   ("FATAL FILESYSTEM ERROR on volId 0x%llx: %s, %s\n",
                    volId.whole_thing,
                    fs_errorTable[errorType].message,
                    callerString));
        }
        else
        {
            E_LOG (class_1, LOG_CRIT, fs_applicationId, 0, 0, 0, 
                   ("FATAL FILESYSTEM ERROR: %s, %s\n",
                    fs_errorTable[errorType].message,
                    callerString));
        }
        break;
    default:
        if (context != NULL)
        {
            E_LOG (class_1, LOG_WARNING, fs_applicationId, 0, 0, 0, 
                   ("FILESYSTEM WARNING on volId 0x%llx: %s, %s\n",
                    volId.whole_thing,
                    fs_errorTable[errorType].message,
                    callerString));
        }
        else
        {
            E_LOG (class_1, LOG_WARNING, fs_applicationId, 0, 0, 0, 
                   ("FILESYSTEM WARNING: %s, %s\n",
                    fs_errorTable[errorType].message,
                    callerString));
        }
        break;
    }

    if (context != NULL)
    {
#if defined(FS_DEBUG)
        fs_contextDump(context);
#endif
    }
}


/**
 * Print an error message and then abort the system.
 */
EXPORTED noreturn
fs_error(
    fs_Context      *context,
    int32           errorType,
    char            *format,
    ...
)
{
    va_list ap;
    va_start(ap, format);
    fs_printErrorMessage(context, LOG_CRIT, errorType, format, ap);
    va_end(ap);

    fs_abort();
}


/**
 * Print a panic message and then abort the system.
 */

noreturn fs_panic(char *format, ...)
{
    va_list ap;
    char string[CALLER_STRING_MAX];

    va_start(ap, format);
    vsnprintf(string, CALLER_STRING_MAX - 1, format, ap);
    va_end(ap);

    E_LOG (class_1, LOG_ERR, fs_applicationId, 0, 0, 0, 
           ("%s", string));

    fs_abort();
}


/**
 * Abort the system, after printing a backtrace.
 */
noreturn fs_abort()
{
#if defined(FS_STANDALONE)
    btrace_print();
    abort();
#else
    FS_ELOG(NULL, LOG_ERR, 0, ("fs_abort"));
    panic("fs_abort");
    // Just in case somebody decided to make
    // panic simply get nervous, force an
    // exception here.
    *((volatile uint32 *)0) = 0;
#endif
}


/**
 * Print an error message, don't abort the system.
 */
EXPORTED void
fs_warn(
    fs_Context      *context,
    int32           errorType,
    char            *format,
    ...
)
{
    va_list ap;
    va_start(ap, format);
    fs_printErrorMessage(context, LOG_WARNING, errorType, format, ap);
    va_end(ap);
} /* fs_warn */

#endif // SSC
