/*
 * Name:  fs-igrow.c
 *
 * RCS:  $Id: fs-igrow.c,v 1.21 2002/12/10 19:14:34 bobonfeline Exp $
 *
 * Description:  Inode file grow routines.
 *
 * Created By:  Byron Rakitzis
 *
 * Date Created:  11/28/2001
 */

#if !defined(SSC)

#include "fs-includes.h"

void
fs_growInodeFile(
    fs_Context *context,
    fs_Parameters *oldParams,
    fs_Parameters *params)
{
    fs_Inode *inosum = I_GET_EXCL(0, FS_FULL_INODE_BLOCK_INODE);
    fs_Inode *dirsum = I_GET_EXCL(0, FS_DIRECTORY_IN_INODE_BLOCK_INODE);
    uint64 old_summary_blocks = oldParams->inodeSummaryCount;
    uint64 newSize = params->inodeSummaryCount * FS_BLOCK_SIZE;
    uint64 oldInodeBlocks = oldParams->maxInodeBlocks;
    uint64 i;

    // clear the sentinel bits past the old end of the file system

    fs_bitmapClearBitsAfter(context, inosum, oldInodeBlocks);
    fs_bitmapClearBitsAfter(context, dirsum, oldInodeBlocks);

    // adjust the file sizes

    inosum->dinode.size = newSize;
    fs_inodeWrite(context, inosum, INO_WRITE_LOG);
    dirsum->dinode.size = newSize;
    fs_inodeWrite(context, dirsum, INO_WRITE_LOG);

    // grow inode summary files explicitly

    for (i = old_summary_blocks; i < params->inodeSummaryCount; i++)
    {
        fs_Buf *buf;

        buf = O_CREATE(inosum, i);
        bzero(buf->u.data, params->blockSize);
        FS_BUF_ZERO_ALL(buf);

        buf = O_CREATE(dirsum, i);
        bzero(buf->u.data, params->blockSize);
        FS_BUF_ZERO_ALL(buf);
    }

    // mark past the end of the file system in the inode summaries

    fs_bitmapSetBitsAfter(context, inosum, params->maxInodeBlocks);
    fs_bitmapSetBitsAfter(context, dirsum, params->maxInodeBlocks);

    fs_reallocShadow(context, context->fileSystem);
}


void
fs_igrowThread(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context      *context = arg;
    fs_FileSystem   *fs = context->fileSystem;
    fs_Parameters   *params = &fs->params;
    fs_SuperBlock   *super = fs->super;
    boolean         didGrow = FALSE;

    // PR #1789: acquire global access lock
    fs_lockShared(context, &context->fileSystem->accessLock, NULL);
    fs_lockExclusive(context, &fs->modifyLock, NULL);

    /*
     * Grow the inode file until the ratio of blocks to inodes is 1.
     */
    while(super->blockInodeRatio > 1 &&
          !fs_haveMininodesAvail(fs))
    {
    	fs_Parameters oldParams = *params;

    	//
    	// Divide the inode ratio by 2, compute new size params.
    	//
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        super->blockInodeRatio /= 2;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        *params = fs_computeSizeParameters(super);

        //fs_printParameters(params);

        if (context->logSummary == NULL) {
            fs_logBeginTransaction(context);
        }

        fs_growInodeFile(context, &oldParams, params);
        didGrow = TRUE;

        fs_writeSuperblocksTrans(context, super);

        fs_logCommit(context, FS_LOG_WRITE_COMMIT_STABLE, 0);
    }

    NFX_SPIN_LOCK(&fs->spin);
    assert(fs->igrowThreadStarted);
    fs->igrowThreadStarted = FALSE;
    NFX_SPIN_UNLOCK(&fs->spin);

    // release global locks
    fs_unlockExclusive(context, &fs->modifyLock);
    fs_unlockShared(context, &fs->accessLock);

    if (didGrow && (!context->eek.inUse))
    {
        // yield until log is completed
        // must be out of the accessLock, to avoid deadlock 
        context->fileSystem->logDevTypeTable->commitTrans(context);
    }
    
    fs_threadTerminate(context);
}

#endif // SSC
