/*
 * Name		: fs-increstore.c
 * RCS		: $Id: fs-increstore.c,v 1.100 2002/12/05 21:08:57 bob Exp $
 * Description	: Incremental Restore
 * Created by	: Rahul Khona
 * Date Created	: 7/11/2001
 */

#if !defined(SSC)

#include "fs-includes.h"
#include "fs-restore.h"
#include "fs-ndmp-api.h"
#include <exports/ndmp-fs-api.h>

#define D   printf

#define testSpecialDirectories(dirp)    \
    {                               \
        if (strcmp(dirp->name, ".") == 0                                \
         || strcmp(dirp->name, "..") == 0                               \
         || strcmp(dirp->name, ".snapshot") == 0)                       \
        {                                                               \
            FS_READDIR_ADV_ENTRY_PTR(dirp);                             \
            continue;                                                   \
        }                                                               \
    }

PRIVATE fs_Status
fs_restoreExtractObject(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_removeTempLink(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    char                *path,
    fs_Inode            **inode
);
PRIVATE fs_Status
fs_createTempLink(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    char                *path,
    fs_Inode            **inode
);
PRIVATE fs_Status
fs_restoreExtractOneName(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_restoreExtractNames(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_restoreCreateDir(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_restoreCreateFset(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_restoreCreateFile(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_restoreCreateSymLink(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_restoreCreateNode(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext,
    uint8                   type
);
PRIVATE fs_Status
fs_restoreRename(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
);
PRIVATE fs_Status
fs_restoreLink(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
);
PRIVATE fs_Status
fs_restoreUnlink(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
PRIVATE fs_Status
fs_restoreCompleteTempDir(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    char                *cp,
    fs_Inumber          rootInum
);
PRIVATE fs_Status
fs_restoreCompleteTemp(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext
);
PRIVATE fs_Status
fs_ensureParentExists(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    int                 temp
);
PRIVATE fs_Status
fs_createTempDir(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    fs_Inode          **parent,
    char                *p,
    fs_Inode          **inode
);

PRIVATE fs_Status
fs_pathFound(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext,
    char                *path
);
PRIVATE fs_Status
fs_drainIncrObject(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
);
PRIVATE fs_Status
fs_createTempRoot(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
);
PRIVATE fs_Status
fs_removeTempRoot(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
);

/*
 * Create a temporary directory inode where intermediate
 * files during restore would be stored
 */

PRIVATE fs_Status
fs_createTempRoot(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
)
{
    fs_Status           status = FS_ERR_NONE;
    fs_Inode            *tempInode, *tempDirInode;
    int                 rc;
    char                name[64];
    fs_Time             now = fs_getTimeOfDay();
	uint64				unstableId;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    fs_growIfNeeded(context);
    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("IN create temp root"));
    fs_logBeginTransaction(context);
    tempDirInode = I_GET_EXCL(0, FS_TEMPDIR_INODE);
    tempInode = fs_allocateDirectoryInode(context, tempDirInode, 0777,
                                    0, 0, 0);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
    ("TempRootInode is %lld\n 0x%llx", tempInode->inum, tempInode->inum));
    restoreContext->tempInodeNum = tempInode->inum;
    sprintf(name, "%lld", tempInode->inum);
    fs_dir_addent(context, tempDirInode, name, strlen(name), tempInode->inum);
    tempDirInode->dinode.atime = now;
    tempDirInode->dinode.ctime = now;
    tempDirInode->dinode.mtime = now;
    tempDirInode->dinode.version++;
    fs_inodeWrite(context, tempDirInode, INO_WRITE_LOG);
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = tempInode->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif
    if (rc != NFX_OK) {
        restoreContext->tempInodeNum = 0;
        fs_inodeReleaseExclusive(context, tempInode);
        fs_inodeReleaseExclusive(context, tempDirInode);
        status = FS_ERR_IO;
    }
    return status;
}
/*
 * Remove a subtree from the filesystem
 */

PRIVATE void
fs_removeSubTreeObject( 
    fs_Context              *context,
    fs_Inumber              inum,
    char                    *name,
    fs_Inumber              parentInum)
{
    fs_Inode    *inode, *parentInode;
    boolean     dontCare = FALSE;
	uint64      unstableId;
    int         rc;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);

    fs_logBeginTransaction(context);

    parentInode = I_GET_EXCL(0, parentInum);
    inode       =  I_GET_EXCL(0, inum);
/*    
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
("IN remove sub tree - name %s curInum %lld link cnt %d node link %d", name,
  parentInode, curInode->dinode.linkCnt, inode->dinode.linkCnt));
*/  
    if (!FS_I_IS_DIR(inode->dinode.attrs)) {
        fs_removeNonDirNode(context, parentInode, inode, name,
                strlen(name), NULL, NULL, NULL, &dontCare);
    } else {
        fs_removeDirNode(context, parentInode, inode, name, 
                strlen(name), NULL, NULL, NULL);
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = inode->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif
    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
    }
}


typedef struct fs_removeSubTreeContext_t    {
    DLL_ENTRY(struct fs_removeSubTreeContext_t)  list;
    fs_Inode        *inode;
    char            *name;
    fs_Inumber      parentInum;
            
    fs_Inumber      inum;
    uint32          numEntries;
    int             eof;
    int             i;
    uint64          cookie;
    fs_DirEntry     *dirp;
    char            buf[FS_MAXPATHLEN];        
}fs_removeSubTreeContext;

PRIVATE fs_Status
fs_removeSubTree(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext,
    fs_Inode                *inode,
    char                    *name,
    fs_Inumber              parentInum
)
{
    DLL_HDR(fs_removeSubTreeContext) rmSubTreeContexts;
    fs_removeSubTreeContext *ctx, *ctx2;
    boolean             goingDown;
    fs_Inumber          curInum;
/*
elog_enabled = TRUE;
elog_severity = LOG_DEBUG;
elog_appl_data[fs_applicationId].severity = LOG_DEBUG;
elog_appl_data[fs_applicationId].bitmap |= (1 << FS_TRACE_DUMP_RESTORE);
*/
    //Initialize the contexts list
    DLL_INIT(&rmSubTreeContexts);

    //Create initial context
    ctx = fs_ramAlloc(FS_RAM_RESTORE_DIRENTS, sizeof(fs_removeSubTreeContext));
    ctx->inode          = inode;
    ctx->name           = name;
    ctx->parentInum     = parentInum;
    ctx->inum           = inode->inum;
    ctx->eof            = 0;
    ctx->cookie         = 0;
          
    do  {
        goingDown = FALSE;
            
        /*
         * If have been using the cpu since a long time, give others a chance
         */

        fs_yieldPeriodic(context);

        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("IN remove sub tree for %s", ctx->name));

        if (!FS_I_IS_DIR(ctx->inode->dinode.attrs)) {
            goto regular_object;
        }

        while (!ctx->eof) {
            ctx->dirp = (fs_DirEntry*)ctx->buf;
            fs_dir_readdir(context, ctx->inode, ctx->dirp, FS_MAXPATHLEN + 1,
                            ctx->cookie, FALSE, &ctx->numEntries, &ctx->eof);
            fs_inodeReleaseExclusive(context, ctx->inode);
            for (ctx->i = 0; ctx->i < ctx->numEntries; ctx->i++) {
                ctx->cookie = ctx->dirp->cookie;
                curInum = ctx->dirp->inode;
                
                testSpecialDirectories(ctx->dirp);
                
                if (!FS_I_IS_DIR(ctx->inode->dinode.attrs)) {
                    fs_removeSubTreeObject(context, curInum, ctx->dirp->name,
                                            ctx->inum);
                } else {
                    ctx2 = fs_ramAlloc(FS_RAM_RESTORE_DIRENTS, 
                                        sizeof(fs_removeSubTreeContext));
                    ctx2->inode         = I_GET_EXCL(0, curInum);
                    ctx2->name          = ctx->dirp->name;
                    ctx2->parentInum    = ctx->inum;
                    ctx2->inum          = curInum;
                    ctx2->eof           = 0;
                    ctx2->cookie        = 0;
                    DLL_ISOLATE_ELEM(ctx, list);
                    DLL_INSERT_TAIL(&rmSubTreeContexts, ctx, list);
                    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("going down from %s to %s", ctx->name, ctx2->name));

                    ctx = ctx2;
                    //We are now in the context of the new dir
                    goingDown = TRUE;
                    break;
                }

goUp:                
                FS_READDIR_ADV_ENTRY_PTR(ctx->dirp);
            }
            if (goingDown) break;
            ctx->inode = I_GET_EXCL(0, ctx->inum);
        }
        if (goingDown) continue;
        
regular_object:
        fs_inodeReleaseExclusive(context, ctx->inode);

        fs_removeSubTreeObject(context, ctx->inum, ctx->name, ctx->parentInum);

        //Retrieve upper context
        ctx2 = ctx;        
        DLL_REMOVE_TAIL(&rmSubTreeContexts, ctx, list);
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("going up from %s back to %s", ctx2->name,
                  (ctx!=NULL && ctx!=DLL_NIL) ? ctx->name : "[END]"));
        //Free previous context
        fs_ramDealloc(FS_RAM_RESTORE_DIRENTS, ctx2);        
        
        if (ctx == NULL || ctx == DLL_NIL)  {
            break;
        }            
                
        goto goUp;

    } while (1);     
                
    return FS_ERR_NONE;
}

/*
 * Clear the temporary tree that has been created during incremental
 * restore phase
 */

PRIVATE fs_Status
fs_removeTempRoot(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
)
{
    fs_Status           status = FS_ERR_NONE;
    fs_Inode            *tempInode;
    int                 rc;
    fs_DirEntry         *dirents, *dirp;
    int                 eof = 0;
    uint32              numEntries;
    uint64              cookie = 0;
    uint32              i;
    fs_Inode            *curInode, *tempDirInode;
    fs_Inumber          curInum;
    char                name[64];
	uint64				unstableId;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("IN remove temp root"));
    sprintf(name, "%lld", restoreContext->tempInodeNum);
    
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("in restoreComplete Temp"));
    dirents = fs_ramAlloc(FS_RAM_RESTORE_DIRENTS, FS_MAXPATHLEN);

    tempDirInode = I_GET_EXCL(0, FS_TEMPDIR_INODE);
    tempInode = I_GET_EXCL(0, restoreContext->tempInodeNum);
    while (!eof) {
        dirp = dirents;
        fs_dir_readdir(context, tempInode, dirp, FS_MAXPATHLEN + 1,
                        cookie, FALSE, &numEntries, &eof);
        fs_inodeReleaseExclusive(context, tempInode);
        for (i = 0; i < numEntries; i++) {
            cookie = dirp->cookie;
            curInum = dirp->inode;
                
            testSpecialDirectories(dirp);
                
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
              ("calling remove sub tree for %s", dirp->name));
            curInode = I_GET_EXCL(0, curInum);
            fs_removeSubTree(context, restoreContext, curInode,
                             dirp->name, restoreContext->tempInodeNum);
            FS_READDIR_ADV_ENTRY_PTR(dirp);
        }
        tempInode =  I_GET_EXCL(0, restoreContext->tempInodeNum);
    }
    fs_ramDealloc(FS_RAM_RESTORE_DIRENTS, dirents);
    fs_logBeginTransaction(context);
    fs_truncateFile(context, tempInode, 0, FALSE, NULL);
    fs_deallocateInode(context, tempInode);
    rc = fs_dir_removeent(  context,
                            tempDirInode,
                            name,
                            strlen(name));
    tempDirInode->dinode.ctime = fs_getTimeOfDay();
    tempDirInode->dinode.mtime = tempDirInode->dinode.ctime;
    tempDirInode->dinode.version++;
    fs_inodeWrite(context, tempDirInode, INO_WRITE_LOG);
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = tempInode->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
    }
    restoreContext->tempInodeNum = 0;
    return status;
}
/*
 * Given a tape record, extract names from it into restoreContext
 */

PRIVATE fs_Status
fs_restoreExtractOneName(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_DumpOperation_t      *operation;
    uint16                  nameLen;
    char                    *name, *cp;
    int                     dataLen;
    int                     copyLen;
    fs_Status               status = FS_ERR_NONE;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("IN extract one name"));
    operation = (fs_DumpOperation_t *)&restoreContext->record.u.operation;
    if (operation->header.flags & FS_DR_TEMP_NAME) {
        restoreContext->isTargetTemp = 1;
    }
    nameLen = operation->len - 1;
    dataLen = FS_DR_TBSIZE - sizeof (fs_DumpOperation_t);
    name = restoreContext->record.u.dummy + sizeof (fs_DumpOperation_t);
    
    restoreContext->targetBufLen = nameLen;
    if (nameLen <= dataLen) {
        strcpy(restoreContext->targetBuf, name);
        FS_ASSERT(strlen(restoreContext->targetBuf) == 
                  restoreContext->targetBufLen);
        return status;
    }

    strncpy(restoreContext->targetBuf, name, dataLen);
    cp = restoreContext->targetBuf + dataLen;
    nameLen -= dataLen;
    while (nameLen > 0) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
              ("Reading additional block for names in extract one name"));
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            return status;
        }
        dataLen = FS_DR_TBSIZE;
        copyLen = MIN(dataLen, nameLen);
        strncpy(cp, restoreContext->record.u.dummy, copyLen);
        cp += copyLen;
        nameLen -= copyLen;
    }

    FS_ASSERT(strlen(restoreContext->targetBuf)==restoreContext->targetBufLen);
    return status;

}
PRIVATE fs_Status
fs_restoreExtractNames(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_DumpOperation_t      *operation;
    uint16                  nameLen;
    char                    *name, *cp;
    int                     dataLen;
    fs_Status               status = FS_ERR_NONE;
    char                    *zeroPtr;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("IN extract names"));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("In restoreExtractNames"));
    operation = (fs_DumpOperation_t *)&restoreContext->record.u.operation;
    if (operation->header.flags & FS_DR_TEMP_NAME) {
        FS_TRACE(context,
                 FS_TRACE_DUMP_RESTORE, ("In restoreExtractNames - temp target"));
        restoreContext->isTargetTemp = 1;
    }
    if (operation->header.flags & FS_DR_LINK_TEMP) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In restoreExtractNames - temp link"));
        restoreContext->isSourceTemp = 1;
    }
    nameLen = operation->len;
    dataLen = FS_DR_TBSIZE - sizeof (fs_DumpOperation_t);
    name = restoreContext->record.u.dummy + sizeof (fs_DumpOperation_t);

    if (nameLen <= dataLen) {

        /*
         * Both source and target are contained within this buffer
         */

        strcpy(restoreContext->sourceBuf, name);
        restoreContext->sourceBufLen = strlen(name);
        strcpy(restoreContext->targetBuf, 
               name + restoreContext->sourceBufLen + 1);
        restoreContext->targetBufLen = nameLen - restoreContext->sourceBufLen
                                               - 2;
        return status;
    }

    /*
     * Read in one name at a time
     */

    restoreContext->sourceBufLen = 0;
    restoreContext->targetBufLen = 0;
    cp = restoreContext->sourceBuf;
    zeroPtr = memchr(name, '\0', dataLen);
get_source:
    if (zeroPtr != NULL) {
        strcpy(cp, name);
        restoreContext->sourceBufLen += strlen(name);
        name += restoreContext->sourceBufLen + 1;
        
        /*
         * Find out start of the target name
         */

        zeroPtr++;

        /*
         * If have exhausted the current tape record, read in the new one
         * and point to the start of the new record for target name
         */

        if (zeroPtr >= &restoreContext->record.u.dummy[FS_DR_TBSIZE]) {
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
          ("Reading additional block for names in extract name - source"));
            status = fs_readTape(context, restoreContext);
            if (status != FS_ERR_NONE) {
                return status;
            }
            dataLen = FS_DR_TBSIZE;
            name = restoreContext->record.u.dummy;
            zeroPtr = memchr(name, '\0', dataLen);
        } else {
                dataLen -= (restoreContext->sourceBufLen + 1);
        }

        /*
         * Prepare to extract target
         */

        cp = restoreContext->targetBuf;
        goto get_target;
    }

    if (zeroPtr == NULL) {
        memcpy(cp, name, dataLen);
        restoreContext->sourceBufLen += dataLen;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
        ("Reading additional block for names in extract name - target"));
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            return status;
        }
        cp += dataLen;
        dataLen = FS_DR_TBSIZE;
        name = restoreContext->record.u.dummy;
        zeroPtr = memchr(name, '\0', dataLen);
        goto get_source;
    }

    
get_target:
    if (zeroPtr != NULL) {
        restoreContext->targetBufLen = nameLen - restoreContext->sourceBufLen
                                               - 2;
        if (dataLen >= restoreContext->targetBufLen) {
                strcpy(cp, name);
                return status;
        } else {
                zeroPtr = NULL;
        }
    }

    if (zeroPtr == NULL) {
        strncpy(cp, name, dataLen);
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            return status;
        }
        cp += dataLen;
        dataLen = FS_DR_TBSIZE;
        name = restoreContext->record.u.dummy;
        zeroPtr = memchr(name, '\0', dataLen);
        goto get_target;
    }

    /* NEVEREACHED */
    FS_ASSERT(0);
    return status; 
}

/*
 * Create a Directory. If the target path is temp, we need
 * to create it in the temp dir
 */

PRIVATE fs_Status
fs_restoreCreateDir(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_Inumber              rootInum;
    fs_Inode                *component, *parentDir;
    fs_path_t               *path;
    char                    *p, *c;
    char                    *pathName = restoreContext->targetBuf;
    uint16                  pathNameLen = restoreContext->targetBufLen;
    int                     temp = restoreContext->isTargetTemp;
    int32                   rc;
    fs_Status               status = FS_ERR_NONE;
    char                    *newPath;
	uint64  				unstableId;
    fs_Time                 now;
    fs_SecurityDescNfs      sdNew;
    
    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
            ("in Create Dir for %s", pathName));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In restoreCreateDir - path %s temp %d", pathName, temp));
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif

    if (fs_shouldIgnore(context, restoreContext, pathName, pathNameLen, 
                        &newPath)) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("Draining %s for create dir", pathName));
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        return fs_drainIncrObject(context, restoreContext);
    }

    if (temp) {

        /*
         * We need to create the object in the temp directory
         */
        rootInum = restoreContext->tempInodeNum;
    }

    /*
     * If we are to create an object in temporary directories,
     * if the intermediate ancestors are missing create them. For
     * in place creation fail if intermediate parents are missing
     */

    if (newPath) {
        strcpy(restoreContext->targetBuf, newPath);
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }


    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {

        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name\n");
            return FS_ERR_NOENT;
        }

        component = fs_lookup(context, context->fileSystem, 0,
                              parentDir, p);
        if (component == NULL) {
            c =  fs_getNextComponent(path);

            if (c != NULL) {
                if (temp) {
                    status = fs_createTempDir(context, restoreContext,
                                              &parentDir, p, &component);
                    if (status != FS_ERR_NONE) {
                        fs_inodeReleaseExclusive(context, parentDir);
                        fs_destroyPath(path);
                        return status;
                    }
                    p = c;
                    fs_inodeReleaseExclusive(context, parentDir);
                    parentDir = component;
                    component = NULL;
                    continue;
                } else {
                    fs_inodeReleaseExclusive(context, parentDir);

                    fs_destroyPath(path);
                    fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NOENT, pathName,
                                          "Invalid Path name\n");
                    return FS_ERR_NOENT;
                }
            }
            break;
        }
        parentDir = component;
        component = NULL;
        p = fs_getNextComponent(path);
    }

    FS_ASSERT(component == NULL);

    if (p == NULL) {

        /*
         * We found the last component. Ensure that is a directory
         */

        component = parentDir;
        if (FS_I_IS_DIR(component->dinode.attrs)) {
            parentDir = NULL;
            fs_logBeginTransaction(context);
            fs_destroyPath(path);
            goto set_attr;
        } else {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Different type of object exists\n");
            return FS_ERR_EXIST;
        }
    }

    /*
     * Now create the entry in the target Directory
     */

    fs_logBeginTransaction(context);
    component = fs_allocateDirectoryInode(context, parentDir, 0777,
                                    restoreContext->attr.uid,
                                    restoreContext->attr.gid,
                                    parentDir->dinode.fsetNum);
    if (component == NULL) {
        /*
         * Send a log message
         */
        fs_destroyPath(path);
        fs_ndmpSendRestoreLog(context, restoreContext,
                               NDMP_LOG_MSG_TYP_ERR, FS_ERR_NOENT,
                               FS_ERR_NOENT, pathName, "Could not create directory\n");

        fs_inodeReleaseExclusive(context, parentDir);
        return FS_ERR_IO;
    }

    rc = fs_dir_addent(context, parentDir, p, strlen(p), component->inum);
    fs_destroyPath(path);
    if (rc != NFX_OK) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    status = fs_addLink(context, 0, component, parentDir->inum);
    if (status != FS_ERR_NONE) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }

set_attr:
    now = fs_getTimeOfDay();
    component->dinode.mtime = restoreContext->attr.mtime;
    component->dinode.ctime = now;
    component->dinode.version++;
    component->dinode.atime = restoreContext->attr.atime;
    component->dinode.attrs |= (restoreContext->attr.mode & FS_IATR_MASK);

    sdNew = component->sd->u.nfs;
    sdNew.mode = (restoreContext->attr.mode & FS_IMODE_MASK);
    fs_securityDescSet(context, component, &sdNew.hdr);
    // BOB - Rahul, please visit this
    component->dinode.userAttr = restoreContext->attr.simpleAttrs;
    //component->dinode.simpleAttrs =  (component->dinode.simpleAttrs & 
                                  //(FS_INODE_SATTR_NO_SNAP|
                                    //FS_INODE_SATTR_DIR_SPLIT)) |
                                    //restoreContext->attr.simpleAttrs;
    if (restoreContext->extract) {
        component->dinode.ctime = now;
        component->dinode.mtime = component->dinode.atime =
            component->dinode.ctime;
    }
    fs_inodeWrite(context, component, INO_WRITE_LOG);
    if (parentDir) {
        parentDir->dinode.linkCnt += 1;
        if (restoreContext->extract) {
            parentDir->dinode.ctime = now;
            parentDir->dinode.mtime = parentDir->dinode.atime = 
                parentDir->dinode.ctime;
        parentDir->dinode.version++;
        } else {
            parentDir->dinode.ctime = now;
            parentDir->dinode.version++;
        }
        fs_inodeWrite(context, parentDir, INO_WRITE_LOG);
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
    }
    return status;
}

/*
 * Create a Directory. If the target path is temp, we need
 * to create it in the temp dir
 */

PRIVATE fs_Status
fs_restoreCreateFset(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_Inumber              rootInum;
    fs_Inode                *component, *parentDir, *fsetInfoInode;
    fs_FileSetIncoreRecord_t    *fset, *parentFset;
    fs_path_t               *path;
    char                    *p, *c;
    char                    *pathName = restoreContext->targetBuf;
    uint16                  pathNameLen = restoreContext->targetBufLen;
    int                     temp = restoreContext->isTargetTemp;
    int32                   rc;
    fs_Status               status = FS_ERR_NONE;
    char                    *newPath;
    fs_FileSystem           *core = context->fileSystem;
	uint64  				unstableId;
    fs_Time                 now;
    fs_SecurityDescNfs      sdNew;

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("in Create fset for %s", pathName));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In restoreCreateDir - path %s temp %d", pathName, temp));
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif

    if (fs_shouldIgnore(context, restoreContext, pathName, pathNameLen, 
                        &newPath)) {
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("Draining %s for create fset", pathName));
        return fs_drainIncrObject(context, restoreContext);
    }

    // See if we can allocate a new fileset
    NFX_SPIN_LOCK(&core->filesetListLock);
    if (core->filesetsCount > FS_FILESET_MAX_ID)  {
        NFX_SPIN_UNLOCK(&core->filesetListLock);
        if (newPath) {
             fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        return FS_ERR_FBIG;
    } else {
        //reserve it so that no other can be created if we reach the limit
        core->filesetsCount++;
    }

    
    if (temp) {

        /*
         * We need to create the object in the temp directory
         */
        rootInum = restoreContext->tempInodeNum;
    }

    /*
     * If we are to create an object in temporary directories,
     * if the intermediate ancestors are missing create them. For
     * in place creation fail if intermediate parents are missing
     */

    if (newPath) {
        strcpy(restoreContext->targetBuf, newPath);
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }


    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {

        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            NFX_SPIN_LOCK(&core->filesetListLock);
            FS_ASSERT(core->filesetsCount != 1)
            core->filesetsCount--;
            NFX_SPIN_UNLOCK(&core->filesetListLock);
            
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name");
            
            return FS_ERR_NOENT;
        }

        component = fs_lookup(context, core, 0, parentDir, p);
        if (component == NULL) {
            c =  fs_getNextComponent(path);

            if (c != NULL) {
                if (temp) {
                    status = fs_createTempDir(context, restoreContext,
                                              &parentDir, p, &component);
                    if (status != FS_ERR_NONE) {
                        NFX_SPIN_LOCK(&core->filesetListLock);
                        FS_ASSERT(core->filesetsCount != 1)
                        core->filesetsCount--;
                        NFX_SPIN_UNLOCK(&core->filesetListLock);
            
                        fs_inodeReleaseExclusive(context, parentDir);
                        fs_destroyPath(path);
                        return status;
                    }
                    p = c;
                    fs_inodeReleaseExclusive(context, parentDir);
                    parentDir = component;
                    component = NULL;
                    continue;
                } else {
                    NFX_SPIN_LOCK(&core->filesetListLock);
                    FS_ASSERT(core->filesetsCount != 1)
                    core->filesetsCount--;
                    NFX_SPIN_UNLOCK(&core->filesetListLock);
            
                    fs_inodeReleaseExclusive(context, parentDir);

                    fs_destroyPath(path);
                    fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NOENT, pathName,
                                          "Invalid Path name\n");
                    return FS_ERR_NOENT;
                }
            }
            break;
        }
        parentDir = component;
        component = NULL;
        p = fs_getNextComponent(path);
    }

    FS_ASSERT(component == NULL);

    if (p == NULL) {

        /*
         * We found the last component. Ensure that is a directory
         */

        component = parentDir;
        if (FS_I_IS_DIR(component->dinode.attrs)) {
            parentDir = NULL;
            fs_logBeginTransaction(context);
            fs_destroyPath(path);
            goto set_attr;
        } else {
            NFX_SPIN_LOCK(&core->filesetListLock);
            FS_ASSERT(core->filesetsCount != 1)
            core->filesetsCount--;
            NFX_SPIN_UNLOCK(&core->filesetListLock);
            
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Different type of object exists\n");
            return FS_ERR_EXIST;
        }
    }

    /*
     * Now create the entry in the target Directory
     */

    fs_logBeginTransaction(context);
    component = fs_allocateDirectoryInode(context, parentDir, 0777,
                                    restoreContext->attr.uid,
                                    restoreContext->attr.gid,
                                    parentDir->dinode.fsetNum);
    if (component == NULL) {
        /*
         * Send a log message
         */
        NFX_SPIN_LOCK(&core->filesetListLock);
        FS_ASSERT(core->filesetsCount != 1)
        core->filesetsCount--;
        NFX_SPIN_UNLOCK(&core->filesetListLock);
            
        fs_destroyPath(path);
        fs_ndmpSendRestoreLog(context, restoreContext,
                               NDMP_LOG_MSG_TYP_ERR, FS_ERR_NOENT,
                               FS_ERR_NOENT, pathName, "Could not create directory\n");

        fs_inodeReleaseExclusive(context, parentDir);
        return FS_ERR_IO;
    }

    rc = fs_dir_addent(context, parentDir, p, strlen(p), component->inum);
    fs_destroyPath(path);
    if (rc != NFX_OK) {
        NFX_SPIN_LOCK(&core->filesetListLock);
        FS_ASSERT(core->filesetsCount != 1)
        core->filesetsCount--;
        NFX_SPIN_UNLOCK(&core->filesetListLock);
            
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    status = fs_addLink(context, 0, component, parentDir->inum);
    if (status != FS_ERR_NONE) {
        NFX_SPIN_LOCK(&core->filesetListLock);
        FS_ASSERT(core->filesetsCount != 1)
        core->filesetsCount--;
        NFX_SPIN_UNLOCK(&core->filesetListLock);
            
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
set_attr:
    now = fs_getTimeOfDay();
    component->dinode.mtime = restoreContext->attr.mtime;
    component->dinode.ctime = now;
    component->dinode.version++;
    component->dinode.atime = restoreContext->attr.atime;
    component->dinode.attrs |= (restoreContext->attr.mode & FS_IATR_MASK);

    sdNew = component->sd->u.nfs;
    sdNew.mode = (restoreContext->attr.mode & FS_IMODE_MASK);
    fs_securityDescSet(context, component, &sdNew.hdr);
    // BOB - Rahul, please visit this
    component->dinode.userAttr = restoreContext->attr.simpleAttrs;
    //component->dinode.simpleAttrs =  (component->dinode.simpleAttrs & 
                                  //(FS_INODE_SATTR_NO_SNAP|
                                    //FS_INODE_SATTR_DIR_SPLIT)) |
                                    //restoreContext->attr.simpleAttrs;

    if (restoreContext->extract) {
        component->dinode.ctime = now;
        component->dinode.mtime = component->dinode.atime =
            component->dinode.ctime;
        component->dinode.version++;
    }
    fs_inodeWrite(context, component, INO_WRITE_LOG);

    fsetInfoInode = I_GET_EXCL(0, FS_FILESET_INFO_INODE);
    parentFset = parentDir->fset = fs_getFset(context, core, parentDir);
    fs_lockExclusive(context, &parentFset->lock, NULL);
    fset = fs_makeFileSet(context, core, parentFset, component, fsetInfoInode);
    fs_unlockExclusive(context, &parentFset->lock);
    if (!fs_inodeWasModified(context, fsetInfoInode)) {
        fs_inodeReleaseExclusive(context, fsetInfoInode);
    }

    if (parentDir) {
        parentDir->dinode.linkCnt += 1;
        if (restoreContext->extract) {
            parentDir->dinode.ctime = now;
            parentDir->dinode.atime =
                parentDir->dinode.mtime = parentDir->dinode.ctime;
            parentDir->dinode.version++;
        } else {
            parentDir->dinode.ctime = now;
            parentDir->dinode.version++;
        }
        fs_inodeWrite(context, parentDir, INO_WRITE_LOG);
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
    }
    return status;
}

/*
 * Create a file 
 */

PRIVATE fs_Status
fs_restoreCreateFile(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_Inumber              rootInum, inum;
    fs_Inode                *component, *parentDir;
    fs_path_t               *path;
    char                    *p, *c;
    char                    *pathName = restoreContext->targetBuf;
    uint16                  pathNameLen = restoreContext->targetBufLen;
    int                     temp = restoreContext->isTargetTemp;
    int32                   rc;
    fs_DumpDataExtension_t  *dataExtension;
    uint32                  numExtents, i, j;
    fs_DumpExtent_t         *drExtent;
    char                    *buf;
    fs_Status               status = FS_ERR_NONE;
    uint64                  headerId = restoreContext->currentId;
    char                    *newPath;
    uint32                  genNum;
	uint64  				unstableId;
    fs_Time                 now;
    fs_SecurityDescNfs      sdNew;

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("in Create file for %s is temp %d", pathName, temp));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In restore create file - temp %d", temp));
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif
    if (fs_shouldIgnore(context, restoreContext, pathName, pathNameLen, 
                        &newPath)) {
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Draining %s for create file", pathName));
        return fs_drainIncrObject(context, restoreContext);
    }

    if (temp) {

        /*
         * We need to create the object in the temp directory
         */
        rootInum = restoreContext->tempInodeNum;
    }

    /*
     * If we are to create an object in temporary directories,
     * if the intermediate ancestors are missing create them. For
     * in place creation fail if intermediate parents are missing
     */

    if (newPath) {
        strcpy(restoreContext->targetBuf, newPath);
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }

    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {

        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name\n");
            return FS_ERR_NOENT;
        }
        component = fs_lookup(context, context->fileSystem, 0,
                              parentDir, p);
        if (component == NULL) {
            c =  fs_getNextComponent(path);

            if (c != NULL) {
                if (temp) {
                    status = fs_createTempDir(context, restoreContext,
                                              &parentDir, p, &component);
                    if (status != FS_ERR_NONE) {
                        fs_inodeReleaseExclusive(context, parentDir);
                        fs_destroyPath(path);
                        return status;
                    }
                    p = c;
                    fs_inodeReleaseExclusive(context, parentDir);
                    parentDir = component;
                    component = NULL;
                    continue;
                } else {
                    fs_inodeReleaseExclusive(context, parentDir);
                    fs_destroyPath(path);
                    fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NONE, pathName,
                                          "Invalid Path name\n");
                    return FS_ERR_NOENT;
                }
            }
            break;
        }
        parentDir = component;
        component = NULL;
        p = fs_getNextComponent(path);
    }

    FS_ASSERT(component == NULL);

    now = fs_getTimeOfDay();
    
    if (p == NULL) {
        /*
         * Objectr with the same name exists. Ensure that it is a file
         */

        component = parentDir;
        if (!FS_I_IS_REG(component->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Different type of object exists\n");
            return FS_ERR_EXIST;
        }
        parentDir = NULL;
        fs_logBeginTransaction(context);
        fs_truncateFile(context, component, 0, FALSE, NULL);
        fs_destroyPath(path);
        goto set_attr;
    }

    /*
     * Now create the entry in the target Directory
     */

    fs_logBeginTransaction(context);
    //component = fs_allocateFileInode(context, parentDir, 0777,
    component = fs_allocateFileInode(context, parentDir, 0,
                                    restoreContext->attr.uid,
                                    restoreContext->attr.gid,
                                    parentDir->dinode.fsetNum);
    if (component == NULL) {
        /*
         * Send a log message
         */
        fs_destroyPath(path);
        fs_ndmpSendRestoreLog(context, restoreContext,
                               NDMP_LOG_MSG_TYP_ERR, FS_ERR_NOENT,
                               FS_ERR_NOENT, pathName, "Could not create File\n");

        fs_inodeReleaseExclusive(context, parentDir);
        return FS_ERR_IO;
    }

    inum = component->inum;
    genNum = component->dinode.genNum;

    rc = fs_dir_addent(context, parentDir, p, strlen(p), component->inum);
    if (rc != NFX_OK) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        fs_destroyPath(path);
        return FS_ERR_IO;
    }

    /*
     * Now that we have added an entry to the directory, get rid of
     * the path
     */

    fs_destroyPath(path);

    status = fs_addLink(context, 0, component, parentDir->inum);
    if (status != FS_ERR_NONE) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    if (restoreContext->extract) {
        parentDir->dinode.ctime = now;
        parentDir->dinode.atime = parentDir->dinode.mtime =
            parentDir->dinode.ctime;
        parentDir->dinode.version++;
    } else {
        parentDir->dinode.ctime = now;
        parentDir->dinode.version++;
    }
    fs_inodeWrite(context, parentDir, INO_WRITE_LOG);

set_attr:
    inum = component->inum;
    component->dinode.mtime = restoreContext->attr.mtime;
    component->dinode.ctime = now;
    component->dinode.version++;
    component->dinode.atime = restoreContext->attr.atime;
    component->dinode.size = restoreContext->attr.sizeDev.size;
    component->dinode.attrs |= (restoreContext->attr.mode & FS_IATR_MASK);

    sdNew = component->sd->u.nfs;
    sdNew.mode = (restoreContext->attr.mode & FS_IMODE_MASK);
    fs_securityDescSet(context, component, &sdNew.hdr);
    // BOB - Rahul, please visit this
    component->dinode.userAttr = restoreContext->attr.simpleAttrs;
    //component->dinode.simpleAttrs =  (component->dinode.simpleAttrs & 
    //                              (FS_INODE_SATTR_NO_SNAP|
    //                                FS_INODE_SATTR_DIR_SPLIT)) |
    //                                restoreContext->attr.simpleAttrs;

    if (restoreContext->extract) {
        component->dinode.ctime = now;
        component->dinode.mtime = component->dinode.atime =
            component->dinode.ctime;
        component->dinode.version++;
    }
    fs_inodeWrite(context, component, INO_WRITE_LOG);
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
    }
//    component = I_GET_EXCL(0, inum);

    /*
     * Now restore the file
     */

    /*
     * For incremental restores, we don't store any extents information
     * in the main operation header, but in a subsequent Data header
     */

    status = fs_readTape(context, restoreContext);
    if (status != FS_ERR_NONE) {
        /*
         * Log an error message
         */

//        fs_inodeReleaseExclusive(context, component);
        return status;
    }

    component = I_GET_EXCL(0, inum);
    buf = fs_ramAlloc(FS_RAM_RESTORE_DATA, FS_BLOCK_SIZE);

    dataExtension =
            (fs_DumpDataExtension_t *)&restoreContext->record.u.extension;

    if (dataExtension->header.op != FS_DR_DATA ||
        dataExtension->header.headerId != headerId) {
        /*
         * Push back the read header
         */

        fs_restorePushBackRead(context, restoreContext);
        fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
        fs_inodeReleaseExclusive(context, component);
        return status;
    }
    numExtents = dataExtension->header.dataExtents;
    drExtent = dataExtension->extents;

    drExtent = (fs_DumpExtent_t *)(restoreContext->record.u.dummy +
                                   sizeof (fs_DumpOperation_t));

write_data:
    for (i = 0; i < numExtents; i++) {

        /*
         * Process each extent
         */

        j = 0;
        while (j < drExtent->len) {

            /*
             * Drop the inode lock while read from the tape
             */

            fs_inodeReleaseExclusive(context, component);
            status = fs_readDataTape(context, restoreContext, buf);
            if (status != FS_ERR_NONE) {
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
//                component = I_GET_EXCL(0, inum);
                return status;
            }
            component = I_GET_EXCL(0, inum);

            /*
             * Restore only if file was not removed when we dropped the lock
             */

            if (component->dinode.genNum == genNum) {
                status = fs_restoreWriteFile(context, restoreContext,
                                             &component, buf,
                                        drExtent->offset + (j << FS_DR_TBSHIFT),
                                         FS_DR_TBSIZE);
            }
            if (status != FS_ERR_NONE) {
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
                fs_inodeReleaseExclusive(context, component);
                return status;
            }
            j++ ;
        }
        drExtent++;
    }
        

    /*
     * Read the tape till, we see DumpExtension records for this headerId
     */

    fs_inodeReleaseExclusive(context, component);
    fs_readTape(context, restoreContext);
    component = I_GET_EXCL(0, inum);
    dataExtension = &restoreContext->record.u.extension;
    if (dataExtension->header.headerId != headerId) {
        fs_restorePushBackRead(context, restoreContext);
        fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
        fs_inodeReleaseExclusive(context, component);
        return status;
    }
    
    numExtents = dataExtension->header.dataExtents;
    drExtent = dataExtension->extents;
    goto write_data;

    /* NEVER REACHED */
}

/*
 * Restore a symlink
 */

PRIVATE fs_Status
fs_restoreCreateSymLink(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_Inumber              rootInum, inum;
    fs_Inode                *component, *parentDir;
    fs_path_t               *path;
    char                    *p, *c;
    char                    *pathName = restoreContext->targetBuf;
    uint16                  pathNameLen = restoreContext->targetBufLen;
    int                     temp = restoreContext->isTargetTemp;
    int32                   rc;
    uint32                  readLen;
    char                    buf [FS_MAXPATHLEN + 1], *cp;
    fs_Status               status = FS_ERR_NONE;
    char                    *newPath;
    uint32                  genNum;
	uint64  				unstableId;
    fs_Time                 now;
    fs_SecurityDescNfs      sdNew;

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("in Create symlink for %s\n", pathName));
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif
    if (fs_shouldIgnore(context, restoreContext, pathName, pathNameLen, 
                        &newPath)) {
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Draining %s for create symlink\n", pathName));
        return fs_drainIncrObject(context, restoreContext);
    }

    if (temp) {

        /*
         * We need to create the object in the temp directory
         */
        rootInum = restoreContext->tempInodeNum;
    }

    if (newPath) {
        strcpy(restoreContext->targetBuf, newPath);
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }
    /*
     * If we are to create an object in temporary directories,
     * if the intermediate ancestors are missing create them. For
     * in place creation fail if intermediate parents are missing
     */

    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {

        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name\n ");
            return FS_ERR_NOENT;
        }
        component = fs_lookup(context, context->fileSystem, 0,
                              parentDir, p);
        if (component == NULL) {
            c =  fs_getNextComponent(path);

            if (c != NULL) {
                if (temp) {
                    status = fs_createTempDir(context, restoreContext,
                                              &parentDir, p, &component);
                    if (status != FS_ERR_NONE) {
                        fs_inodeReleaseExclusive(context, parentDir);
                        fs_destroyPath(path);
                        return status;
                    }
                    p = c;
                    fs_inodeReleaseExclusive(context, parentDir);
                    parentDir = component;
                    component = NULL;
                    continue;
                } else {
                    fs_inodeReleaseExclusive(context, parentDir);

                    fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NOENT, NULL,
                                          "Invalid Path name\n");
                    fs_destroyPath(path);
                    return FS_ERR_NOENT;
                }
            }
            break;
        }
        parentDir = component;
        component = NULL;
        p = fs_getNextComponent(path);
    }

    FS_ASSERT(component == NULL);
    
    now = fs_getTimeOfDay();
    
    if (p == NULL) {
        /*
         * Object with same name exists. Ensure that it is a symlink
         */

        component = parentDir;
        if (!FS_I_IS_SYML(component->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Different type of object exists\n");
            return FS_ERR_EXIST;
        }
        parentDir = NULL;
        fs_logBeginTransaction(context);
        if (component->dinode.numblocks > 0) {

            /*
             * Special case truncation of symlinks contained within inode
             * as emap_shrink does not handle it correctly !!!
             */

            fs_truncateFile(context, component, 0, FALSE, NULL);
        } else {
            component->dinode.size = 0;
        }
        fs_destroyPath(path);
        goto set_attr;
    }

    /*
     * Now create the entry in the target Directory
     */

    fs_logBeginTransaction(context);
    //component = fs_allocateFileInode(context, parentDir, 0777 |
    component = fs_allocateFileInode(context, parentDir,
                                     FS_IFMT_SYML,
                                    restoreContext->attr.uid,
                                    restoreContext->attr.gid,
                                    parentDir->dinode.fsetNum);
    if (component == NULL) {
        /*
         * Send a log message
         */
        fs_destroyPath(path);
        fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NOENT, pathName,
                                          "Could not create file\n");

        fs_inodeReleaseExclusive(context, parentDir);
        return FS_ERR_IO;
    }

    inum = component->inum;

    rc = fs_dir_addent(context, parentDir, p, strlen(p), component->inum);
    fs_destroyPath(path);
    if (rc != NFX_OK) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    status = fs_addLink(context, 0, component, parentDir->inum);
    if (status != FS_ERR_NONE) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    if (restoreContext->extract) {
        parentDir->dinode.atime =
        parentDir->dinode.mtime =
        parentDir->dinode.ctime = now;
        parentDir->dinode.version++;
    } else {
        parentDir->dinode.ctime = now;
        parentDir->dinode.version++;
    }
    fs_inodeWrite(context, parentDir, INO_WRITE_LOG);

set_attr:
    component->dinode.mtime = restoreContext->attr.mtime;
    component->dinode.ctime = now;
    component->dinode.version++;
    component->dinode.atime = restoreContext->attr.atime;
    component->dinode.size = restoreContext->attr.sizeDev.size;
    component->dinode.attrs |= (restoreContext->attr.mode & FS_IATR_MASK);

    sdNew = component->sd->u.nfs;
    sdNew.mode = (restoreContext->attr.mode & FS_IMODE_MASK);
    fs_securityDescSet(context, component, &sdNew.hdr);
    // BOB - Rahul, please visit this
    component->dinode.userAttr = restoreContext->attr.simpleAttrs;
    //component->dinode.simpleAttrs =  (component->dinode.simpleAttrs & 
                                  //(FS_INODE_SATTR_NO_SNAP|
                                    //FS_INODE_SATTR_DIR_SPLIT)) |
                                    //restoreContext->attr.simpleAttrs;

    if (restoreContext->extract) {
        component->dinode.ctime = now;
        component->dinode.version++;
        component->dinode.mtime = component->dinode.atime =
            component->dinode.ctime;
    }

    genNum = component->dinode.genNum;
    inum = component->inum;

    /*
     * Drop the lock on the component when we block on reading the tape.
     * Commit the transaction and create a new one as we cannot have
     * locks dropped across a transaction
     */

    fs_inodeWrite(context, component, INO_WRITE_LOG);
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_yieldPeriodic(context);
    
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
        return status;
    }
    

    /*
     * Now read in the data from the tape. We always dump a DR_DATA header
     * for symlink, drain it
     */

    status = fs_readTape(context, restoreContext);
    if (status != FS_ERR_NONE) {
        if (parentDir) {
            fs_inodeReleaseExclusive(context, parentDir);
        }
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }

    /*
     * Now read in the extent map.
     */

    status = fs_readTape(context, restoreContext);
    if (status != FS_ERR_NONE) {
        if (parentDir) {
            fs_inodeReleaseExclusive(context, parentDir);
        }
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }

    /*
     * Now get the actual link content, we know that we
     * always write 1 block worth of extents only
     */

    cp = buf;
more_symlink:
    FS_ASSERT((cp - buf) <= FS_PATH_MAX);
    status = fs_readDataTape(context, restoreContext, cp);
    if (status != FS_ERR_NONE) {
        return FS_ERR_IO;
    }
    cp += FS_DR_TBSIZE;
    if (((cp - buf) + 1) < component->dinode.size) {
        goto more_symlink;
    }
    readLen = FS_DR_TBSIZE;

    /*
     * Get the inode again. If the generation cound does not match
     * ingore it
     */

    component = I_GET_EXCL(0, inum);
    if (component->dinode.genNum != genNum) {
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_NONE;
    } 
    fs_logBeginTransaction(context);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Readin length is %d buf is %s", strlen(buf), buf));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Attr length is %lld ", component->dinode.size));
    if (component->dinode.size <= FS_INODE_INLINE_LEN) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Coping to inline syml"));
        memcpy(component->dinode.u.data, buf, component->dinode.size);
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Symlink is %s",
                                component->dinode.u.data));
    } else {
        fs_Buf  *bp = NULL;
        fs_emaproot_init(context, component);
        bp = O_CREATE(component, 0);
		FS_ASSERT(bp != NULL);
		FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("bp block is %lld", bp->block.block));
        memcpy(bp->u.data, buf, component->dinode.size);
        FS_BUF_ZERO_ALL(bp);
        FS_BUF_WRITE(bp, 0, component->dinode.size);
    }

    fs_inodeWrite(context, component, INO_WRITE_LOG);
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
    }
    return status;
}

PRIVATE fs_Status
fs_restoreRename(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
)
{

    fs_Status   status = FS_ERR_NONE;
    fs_Inode    *sourceParent, *targetParent, *source;
    fs_Inumber  sourceParentInum, targetParentInum, sourceInum, targetInum;
    char        *sourceName, *targetName, *newPath = NULL;
    uint16      sourceNameLen, targetNameLen;
    int32       rc;
	uint64				unstableId;
    fs_Time     now;
    fs_SecurityDescNfs sdNew;

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

#if 1
    if ((restoreContext->extract &&
         !fs_isRestored(context, restoreContext, restoreContext->targetBuf)) ||
	    fs_shouldIgnore(context, restoreContext, restoreContext->targetBuf, 
                        restoreContext->targetBufLen, &newPath)) {
#else
    if (fs_shouldIgnore(context, restoreContext,
                        restoreContext->targetBuf, &newPath)) {
#endif
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Draining %s for create unlink",
                   restoreContext->targetBuf));
	    return fs_drainIncrObject(context, restoreContext);
    }
    /*
     * First ensure that the path to target exists
     */

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("in Create rename "));
    status = fs_ensureParentExists(context, restoreContext,
                                    restoreContext->isTargetTemp);
    if (status != FS_ERR_NONE) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->targetBuf,
                              "Path to target does not exist");
        // Instead of returning failure, let restore continue and try
        // to restore whatever is possible
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        return FS_ERR_NONE;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Source buf is %s", restoreContext->sourceBuf));
    status = fs_decomposePath(context, restoreContext,
                              restoreContext->sourceBuf,
                              &sourceParentInum, &sourceInum, &sourceName,
                              &sourceNameLen, restoreContext->isSourceTemp);

    /*
     * If we did not find path to the source, let restore carry on
     * with other files
     */

    if (status == FS_ERR_NOENT) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                          "cannot rename as path to source does not exist\n");
        
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        return FS_ERR_NONE;
    } else if (status != FS_ERR_NONE){
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                          "cannot rename as path to source does not exist\n");
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        return status;
    } else if (sourceInum == 0) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                          "cannot rename as path to source does not exist\n");
        
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        return FS_ERR_NONE;
    }
    FS_ASSERT(sourceInum != 0);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("target buf is %s", restoreContext->targetBuf));
    status = fs_decomposePath(context, restoreContext,
                              restoreContext->targetBuf,
                              &targetParentInum, &targetInum, &targetName,
                              &targetNameLen, restoreContext->isTargetTemp);
    if (status == FS_ERR_NOENT) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                      "cannot rename as path to target parent does not exist\n");
        
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        return FS_ERR_NONE;
    } else if (status != FS_ERR_NONE){
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->targetBuf,
                          "cannot rename as path to target does not exist\n");
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        return status;
    }

    if (targetInum != 0) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->targetBuf,
                          "cannot rename as path to target object exists\n");
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        return status;
    }
    FS_ASSERT(targetParentInum != 0);
    /*
     * Initialize a transaction
     */

    fs_logBeginTransaction(context);

    /*
     * Get the inodes
     */

    if (sourceParentInum == targetParentInum) {
        FS_TRACE(context,
                 FS_TRACE_DUMP_RESTORE,
                 ("Source and Target parents are same"));
        sourceParent = targetParent = I_GET_EXCL(0, sourceParentInum);
    } else {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Source and Target parents are different"));
        if (sourceParentInum < targetParentInum) {
            sourceParent = I_GET_EXCL(0, sourceParentInum);
            targetParent = I_GET_EXCL(0, targetParentInum);
        } else {
            sourceParent = I_GET_EXCL(0, sourceParentInum);
            targetParent = I_GET_EXCL(0, targetParentInum);
        }
    }

    source = I_GET_EXCL(0, sourceInum);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("targetName is %s", targetName));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("sourceinum is %lld", source->inum));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("targetparent is %lld", targetParent->inum));
    rc = fs_dir_addent(context, targetParent, targetName, targetNameLen,
                       source->inum);
    if (rc != NFX_OK) {
        status = FS_ERR_IO;       
        /*
         * Send a log message
         */
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              FS_ERR_IO, NULL,
                              "I/O error occured while writing dir\n");
        fs_inodeReleaseExclusive(context, source);
        fs_inodeReleaseExclusive(context, sourceParent);
        if (sourceParentInum != targetParentInum) {
            fs_inodeReleaseExclusive(context, targetParent);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("could not add entry to target"));
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        return status;
    }
    if (sourceParentInum != targetParentInum) {
        if (FS_I_IS_DIR(source->dinode.attrs)) {
            fs_dirroot_set_parent(context, source, targetParentInum);
        }
        status = fs_addLink(context, 0, source, targetParent->inum);
        if (status != FS_ERR_NONE) {
            fs_inodeReleaseExclusive(context, source);
            fs_inodeReleaseExclusive(context, sourceParent);
            if (sourceParentInum != targetParentInum) {
                fs_inodeReleaseExclusive(context, targetParent);
            }
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("could not add link to target"));
            if (sourceName) {
                fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
            }
            if (targetName) {
                fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
            }
            return status;
        }
    }

    /*
     * Remove the entry from parent
     */

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("sourceName is %s", sourceName));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("sourceParentInum is %lld", sourceParent->inum));
    (void)fs_dir_removeent(context, sourceParent, sourceName, sourceNameLen);
    if (sourceParentInum != targetParentInum) {
        status = fs_removeLink(context, 0, source, sourceParent->inum);
        if (status != FS_ERR_NONE) {
            fs_inodeReleaseExclusive(context, source);
            fs_inodeReleaseExclusive(context, sourceParent);
            if (sourceParentInum != targetParentInum) {
                fs_inodeReleaseExclusive(context, targetParent);
            }
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("could not remove link from source"));
            if (newPath) {
                fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
            }
            if (sourceName) {
                fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
            }
            if (targetName) {
                fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
            }
            return status;
        }
    }

    /*
     * Update inode attributes
     */

    now = fs_getTimeOfDay();
    source->dinode.ctime = now;
    source->dinode.version++;
    source->dinode.mtime = restoreContext->attr.mtime;
    source->dinode.atime = restoreContext->attr.atime;
    source->dinode.attrs |= (restoreContext->attr.mode & FS_IATR_MASK);

    sdNew = source->sd->u.nfs;
    sdNew.uid = restoreContext->attr.uid;
    sdNew.gid = restoreContext->attr.gid;
    sdNew.mode = (restoreContext->attr.mode & FS_IMODE_MASK);
    fs_securityDescSet(context, source, &sdNew.hdr);
    // BOB - Rahul, please visit this
    source->dinode.userAttr = restoreContext->attr.simpleAttrs;
    //source->dinode.simpleAttrs =  (source->dinode.simpleAttrs & 
    //                              (FS_INODE_SATTR_NO_SNAP|
    //                                FS_INODE_SATTR_DIR_SPLIT)) |
    //                                restoreContext->attr.simpleAttrs;

    fs_inodeWrite(context, source, INO_WRITE_LOG);

    sourceParent->dinode.ctime = now;
    sourceParent->dinode.version++;
    sourceParent->dinode.mtime = sourceParent->dinode.ctime;
    fs_inodeWrite(context, sourceParent, INO_WRITE_LOG);

    /* RAHUL 
     * Handle link cnt update for parents if this is a directory
     */

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("source mode is %o", source->sd->u.nfs.mode));
    if (FS_I_IS_DIR(source->dinode.attrs)) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("Updating link cnt source parent %lld target parent %lld",
                sourceParent->inum, targetParent->inum));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("Updating link cnt source parent link %d target parent link %d",
                sourceParent->dinode.linkCnt, targetParent->dinode.linkCnt));
        sourceParent->dinode.linkCnt--;
        targetParent->dinode.linkCnt++;
    }

    if (sourceParentInum != targetParentInum) {
        targetParent->dinode.ctime = now;
        targetParent->dinode.version++;
        targetParent->dinode.mtime = targetParent->dinode.ctime;
        fs_inodeWrite(context, targetParent, INO_WRITE_LOG);
    }

    if (newPath) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }
    if (sourceName) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
    }
    if (targetName) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = source->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != FS_ERR_NONE) {
        status = FS_ERR_IO;
        /*
         * Log the message
         */
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("restore rename transaction failed"));
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Rename completed with status %d", status));
    return status;
}

/*
 * Create a link
 */

PRIVATE fs_Status
fs_restoreLink(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
)
{

    fs_Status   status = FS_ERR_NONE;
    fs_Inode    *sourceParent, *targetParent, *source;
    fs_Inumber  sourceParentInum, targetParentInum, sourceInum, targetInum;
    char        *sourceName, *targetName;
    uint16      sourceNameLen, targetNameLen;
    int32       rc;
    char                    *newPath;
	uint64				unstableId;

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("in Create link"));
    /*
     * First ensure that the path to target exists
     */

    if (fs_shouldIgnore(context, restoreContext, restoreContext->targetBuf,
                        restoreContext->targetBufLen, &newPath)) {
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Draining %s for create lin", restoreContext->targetBuf));
        return fs_drainIncrObject(context, restoreContext);
    }

    if (newPath) {
        strcpy(restoreContext->targetBuf, newPath);
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }
    status = fs_ensureParentExists(context, restoreContext,
                                    restoreContext->isTargetTemp);
    if (status != FS_ERR_NONE) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->targetBuf,
                          "cannot link as path to target does not exist\n");
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("in restore link: ensure parent exists returned %d",
                    status));
        // Instead of returning failure, let restore continue and try
        // to restore whatever is possible
        return FS_ERR_NONE;
    }

    status = fs_decomposePath(context, restoreContext,
                              restoreContext->sourceBuf,
                              &sourceParentInum, &sourceInum, &sourceName,
                              &sourceNameLen, restoreContext->isSourceTemp);
    if (status == FS_ERR_NOENT) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                          "cannot link as path to source does not exist\n");
        
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
            ("in restore link: decomose path for source buf %s returned %d",
            restoreContext->sourceBuf, status));
        return FS_ERR_NONE;
    } else if (status != FS_ERR_NONE){
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                          "cannot link as path to source does not exist\n");
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
            ("in restore link: decomose path for source buf %s returned %d",
            restoreContext->sourceBuf, status));
        }
        return status;
    } else if (sourceInum == 0) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                          "cannot link as path to source does not exist\n");
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
            ("in restore link: decomose path for source buf %s returned %d",
            restoreContext->sourceBuf, status));
        }
        return status;
    
    }
    FS_ASSERT(sourceInum != 0);
    status = fs_decomposePath(context, restoreContext,
                              restoreContext->targetBuf,
                              &targetParentInum, &targetInum, &targetName,
                              &targetNameLen, restoreContext->isTargetTemp);

    if (status == FS_ERR_NOENT) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->sourceBuf,
                      "cannot link as path to target parent does not exist\n");
        
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
        ("in restore link: decomose path for target buf %s returned %d",
            restoreContext->targetBuf, status));
        return FS_ERR_NONE;
    } else if (status != FS_ERR_NONE){
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->targetBuf,
                          "cannot link as path to target does not exist\n");
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
        ("in restore link: decomose path for target buf %s returned %d",
            restoreContext->targetBuf, status));
        return status;
    }

    if (targetInum != 0) {
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              status, restoreContext->targetBuf,
                          "cannot link as target exists\n");
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
        ("in restore link: decomose path for target buf %s returned %d",
            restoreContext->targetBuf, status));
        return status;
    }
    FS_ASSERT(targetParentInum != 0);
    /*
     * Initialize a transaction
     */

    fs_logBeginTransaction(context);

    /*
     * Get the inodes
     */

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
        ("in restore link source buf %s target buf %s", 
          restoreContext->sourceBuf, restoreContext->targetBuf));
    if (sourceParentInum == targetParentInum) {
        sourceParent = targetParent = I_GET_EXCL(0, sourceParentInum);
    } else {
        if (sourceParentInum < targetParentInum) {
            sourceParent = I_GET_EXCL(0, sourceParentInum);
            targetParent = I_GET_EXCL(0, targetParentInum);
        } else {
            sourceParent = I_GET_EXCL(0, sourceParentInum);
            targetParent = I_GET_EXCL(0, targetParentInum);
        }
    }

    source = I_GET_EXCL(0, sourceInum);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("in restore link add link addin to inode %lld", source->inum));
    rc = fs_dir_addent(context, targetParent, targetName, targetNameLen,
                       source->inum);
    if (rc != NFX_OK) {
        status = FS_ERR_IO;       
        /*
         * Send a log message
         */
        fs_ndmpSendRestoreLog(context, restoreContext,
                              status, NDMP_LOG_MSG_TYP_ERR,
                              FS_ERR_IO, NULL,
                              "I/O error occured while writing dir\n");
        fs_inodeReleaseExclusive(context, source);
        fs_inodeReleaseExclusive(context, sourceParent);
        if (sourceParentInum != targetParentInum) {
            fs_inodeReleaseExclusive(context, targetParent);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("in restore link add ent failed for %s", targetName));
        return status;
    }
    status = fs_addLink(context, 0, source, targetParent->inum);
    if (status != FS_ERR_NONE) {
        fs_inodeReleaseExclusive(context, source);
        fs_inodeReleaseExclusive(context, sourceParent);
        if (sourceParentInum != targetParentInum) {
            fs_inodeReleaseExclusive(context, targetParent);
        }
        if (sourceName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
        }
        if (targetName) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("in restore link add link failed for %s", targetName));
        return status;
    }
    if (sourceParentInum != targetParentInum) {

        /*
         * At this point, we don't need the source parent
         */

        fs_inodeReleaseExclusive(context, sourceParent);
    }
    if (restoreContext->extract) {
        targetParent->dinode.atime =
        targetParent->dinode.mtime =
        targetParent->dinode.ctime = fs_getTimeOfDay();
        targetParent->dinode.version++;
    } else {
        targetParent->dinode.ctime = fs_getTimeOfDay();
        targetParent->dinode.version++;
    }
    fs_inodeWrite(context, targetParent, INO_WRITE_LOG);

    /*
     * Increase the link count of the source inode
     */

    source->dinode.linkCnt += 1;
    fs_inodeWrite(context, source, INO_WRITE_LOG);
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = source->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != FS_ERR_NONE) {
        status = FS_ERR_IO;
        /*
         * Log the message
         */
    }
    if (sourceName) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, sourceName);
    }
    if (targetName) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, targetName);
    }
    return status;
}

/*
 * Perform an unlink operation
 */

PRIVATE fs_Status
fs_restoreUnlink(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_Inumber              rootInum;
    fs_Inode                *component = NULL, *parentDir = NULL,
                            *fsetInfoInode;
    fs_path_t               *path;
    char                    *p, *c, *newPath = NULL;
    char                    *pathName = restoreContext->targetBuf;
    uint32                  pathNameLen = restoreContext->targetBufLen;
    int                     temp = restoreContext->isTargetTemp;
    int32                   rc;
    fs_Status               status = FS_ERR_NONE;
    fs_FileSystem           *core = context->fileSystem;
	uint64				unstableId;

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("in Create unlink for %s", pathName));
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif
    if (temp) {

        /*
         * We need to create the object in the temp directory
         */
        rootInum = restoreContext->tempInodeNum;
    }
#if 1
    if ((restoreContext->extract &&
         !fs_isRestored(context, restoreContext, restoreContext->targetBuf)) ||
	     fs_shouldIgnore(context, restoreContext, pathName, pathNameLen,
                         &newPath))
    {
#else
    if (fs_shouldIgnore(context, restoreContext, pathName, &newPath)) {
#endif
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Draining %s for create unlink", pathName));
	    return fs_drainIncrObject(context, restoreContext);
    }

    /*
     * If we are to create an object in temporary directories,
     * if the intermediate ancestors are missing create them. For
     * in place creation fail if intermediate parents are missing
     */

    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {
        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            return FS_ERR_NONE;
        }

        component = fs_lookupKeepParentLocked(context, core, 0, parentDir, p);
        if (component == NULL) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            if (newPath) {
                fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
            }
            return FS_ERR_NONE;
        }
        c = fs_getNextComponent(path);
        if (c == NULL) {
            break;
        }
        p = c;
        fs_inodeReleaseExclusive(context, parentDir);
        parentDir = component;
        component = NULL;
    }

    
    FS_ASSERT(parentDir != NULL && component != NULL);

    /*
     * Now remove the entry in the target Directory
     */

    /*
     * If the object is directory ensure that is is empty
     */

    if (FS_I_IS_DIR(component->dinode.attrs)) {
        if (!fs_dir_is_empty(context, component)) {
            fs_destroyPath(path);   // CHECKIN
            fs_inodeReleaseExclusive(context, component);
            fs_inodeReleaseExclusive(context, parentDir);
            return FS_ERR_NOTEMPTY;
        }
    }

    fs_logBeginTransaction(context);
    component->fset = fs_getFset(context, core, component);
    if (component->inum == component->fset->cachedRecord.rootInode) {

        /*
         * This is a fileset root. Remove it
         */

        fsetInfoInode = I_GET_EXCL(0, FS_FILESET_INFO_INODE);
        fs_removeFileSet(context, core, component->fset, component, 
                            fsetInfoInode);

        if (!fs_inodeWasModified(context, fsetInfoInode)) {
            fs_inodeReleaseExclusive(context, fsetInfoInode);
        }

        //Decrement the number of filesets
        NFX_SPIN_LOCK(&core->filesetListLock);
        FS_ASSERT(core->filesetsCount != 1)
        core->filesetsCount--;
        NFX_SPIN_UNLOCK(&core->filesetListLock);
    }

    /*
     * For Non Directory inodes, truncate the file
     */

    if (!FS_I_IS_DIR(component->dinode.attrs)) {
        if (component->dinode.linkCnt ==1 && 
            (!FS_I_IS_SYML(component->dinode.attrs) ||
            (component->dinode.numblocks > 0))) {
            fs_truncateFile(context, component, 0, FALSE, NULL);
        }
    }
    rc = fs_dir_removeent(context, parentDir, p, strlen(p));
    status = fs_removeLink(context, 0, component, parentDir->inum);
    fs_destroyPath(path);
    if (status != FS_ERR_NONE) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        return FS_ERR_IO;
    }

    if (FS_I_IS_DIR(component->dinode.attrs)) {
        parentDir->dinode.linkCnt -= 1;
    }
    component->dinode.linkCnt--;
    fs_inodeWrite(context, parentDir, INO_WRITE_LOG);
    fs_inodeWrite(context, component, INO_WRITE_LOG);
    if (component->dinode.linkCnt == 0 ||
         FS_I_IS_DIR(component->dinode.attrs)) {

        /*
         * We don't support hardlinks to directories, so this will
         * work as unlink for dir would not be called otherwise
         */

        fs_truncateFile(context, component, 0, FALSE, NULL);
        fs_deallocateInode(context, component);
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
    }
    if (newPath) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }
    return status;
}

/*
 * Read the tape and performing various operations
 */

EXPORTED fs_Status
fs_incrRestore(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{

    fs_Status               status;
    fs_DumpGenericHeader_t  *genHeader;
    uint16                  op;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
#ifdef FS_DEBUG_1
    fs_unlockShared(context, &context->fileSystem->accessLock);
#endif
    /*
     * We have already read the first record to determine that
     * we have to perform incremental restore
     */

    fs_createTempRoot(context, restoreContext);

    while (1) {

        restoreContext->isSourceTemp = restoreContext->isTargetTemp = 0;
        restoreContext->sourceBuf[0] = restoreContext->targetBuf[0] = '\0';
        status = fs_readTape(context, restoreContext);
        genHeader = (fs_DumpGenericHeader_t *)&restoreContext->record;
        restoreContext->attr = genHeader->attr;
        restoreContext->currentId = genHeader->headerId;

        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("operation is %d for id %lld",
                    genHeader->op, genHeader->headerId));
        switch (genHeader->op) {

        case FS_DR_CREATE_DIR:

            /*
             * Get the name of the directory to be created and also
             * determine if we need to create it in temp dir or not
             */

            status = fs_restoreExtractOneName(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreCreateDir(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_NAME_ONLY:

            /*
             * Object did not have a namespace operation, but its
             * contents were changed, so we need to restore the
             * contents again. The inode of the object is still of
             * same type
             */

            status = fs_restoreExtractOneName(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreExtractObject(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_CREATE_FILE:
            status = fs_restoreExtractOneName(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreCreateFile(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_CREATE_SYML:
            status = fs_restoreExtractOneName(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreCreateSymLink(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_CREATE_CHAR:
        case FS_DR_CREATE_BLOCK:
        case FS_DR_CREATE_FIFO:
            op = genHeader->op;
            status = fs_restoreExtractOneName(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreCreateNode(context, restoreContext, op);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_CREATE_FSET:
            /*
             * Get the name of the directory to be created and also
             * determine if we need to create it in temp dir or not
             */

            status = fs_restoreExtractOneName(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreCreateFset(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_CREATE_LINK:
            status = fs_restoreExtractNames(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreLink(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_RENAME:
            status = fs_restoreExtractNames(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreRename(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_REMOVE:
        case FS_DR_REMOVE_DIR:
            status = fs_restoreExtractOneName(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
            status = fs_restoreUnlink(context, restoreContext);
            if (status != FS_ERR_NONE) {
                goto done;
            }
#ifdef FS_DEBUG_1
            fs_lockTerminateContext(context);
#endif
            break;
        case FS_DR_TAPE_END:
            goto done;
        case FS_DR_SET_FQUOTA:
        case FS_DR_SET_UQUOTA:
        case FS_DR_SET_GQUOTA:
        default:
            break;
        }
    }
done:
#ifdef FS_DEBUG_1
    fs_lockTerminateContext(context);
#endif
    if (status == FS_ERR_NONE) {

        /*
         * Rename objects from temp dir to their final destintions
         */
        if (restoreContext->tempInodeNum != 0) {
            status = fs_restoreCompleteTemp(context, restoreContext);
        }
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("calling remove Temp root"));
#ifdef FS_DEBUG_1
    fs_lockTerminateContext(context);
#endif
    fs_removeTempRoot(context, restoreContext);
#ifdef FS_DEBUG_1
    fs_lockShared(context, &context->fileSystem->accessLock, NULL);
#endif
    return status;
}
typedef struct fs_restoreCompleteTempDirContext_t   {
    DLL_ENTRY(struct fs_restoreCompleteTempDirContext_t)  list;
    char                *cp;
    fs_Inumber          rootInum;
            
    int                 eof;
    uint32              numEntries;
    uint32              i;
    uint64              cookie;
    fs_Inumber          curInum;
    fs_DirEntry         *dirp;
    char                buf[FS_MAXPATHLEN];
}fs_restoreCompleteTempDirContext;

PRIVATE fs_Status
fs_restoreCompleteTempDir(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext,
    char                *cp,
    fs_Inumber          rootInum
)
{
    DLL_HDR(fs_restoreCompleteTempDirContext) restoreComplTempDirCtxts;
    fs_restoreCompleteTempDirContext *ctx, *ctx2;
    boolean             goingDown;
    fs_Inode            *tempRoot;
    fs_Inode            *curInode;
    fs_Status           status = FS_ERR_NONE;
/*    
elog_enabled = TRUE;
elog_severity = LOG_DEBUG;
elog_appl_data[fs_applicationId].severity = LOG_DEBUG;
elog_appl_data[fs_applicationId].bitmap |= (1 << FS_TRACE_DUMP_RESTORE);
*/    
    //Initialize the contexts list
    DLL_INIT(&restoreComplTempDirCtxts);

    //Create initial context
    ctx = fs_ramAlloc(FS_RAM_RESTORE_DIRENTS, 
                      sizeof(fs_restoreCompleteTempDirContext));
    ctx->cp         = cp;
    ctx->rootInum   = rootInum;
    ctx->eof        = 0;
    ctx->cookie     = 0;
    
    do  {
        goingDown = FALSE;
            
        /*
         * If have been using the cpu since a long time, give others a chance
         */

        fs_yieldPeriodic(context);
        
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In restoreCompleteTempDir for %lld", ctx->rootInum));
#ifndef FS_NO_LOCK_TIMEOUT
        tempRoot =  I_GET_SHARED(0, ctx->rootInum);
#else
        tempRoot =  I_GET_EXCL(0, ctx->rootInum);
#endif
        *ctx->cp = '/';
        ctx->cp++;
        *ctx->cp = '\0';
        while (!ctx->eof) {
            ctx->dirp = (fs_DirEntry*)ctx->buf;
            fs_dir_readdir(context, tempRoot, ctx->dirp, FS_MAXPATHLEN + 1,
                            ctx->cookie, FALSE, &ctx->numEntries, &ctx->eof);
#ifndef FS_NO_LOCK_TIMEOUT
            fs_inodeRelease(context, tempRoot);
#else
            fs_inodeReleaseExclusive(context, tempRoot);
#endif
            for (ctx->i = 0; ctx->i < ctx->numEntries; ctx->i++) {
                ctx->cookie = ctx->dirp->cookie;
                ctx->curInum = ctx->dirp->inode;
                
                testSpecialDirectories(ctx->dirp);
                
                strncpy(ctx->cp, ctx->dirp->name, ctx->dirp->nameLen);
                status = fs_pathFound(context, restoreContext,
                                      restoreContext->currentName);
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                        ("Path found for %s returned %d curInum %lld",
                         restoreContext->currentName, status, ctx->curInum));
                if (status != FS_ERR_NONE && status != FS_ERR_NOENT) {
                    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                            ("status returned by pathFound %d", status));
                    /*
                     * Log the message with ndmp and return failure
                     */
                    fs_ndmpSendRestoreLog(context, restoreContext,
                            status, NDMP_LOG_MSG_TYP_ERR, FS_ERR_NOENT, NULL, 
                            "Error completing moves from temp dir\n");
                    goto goUp;
                }
        
                if (status == FS_ERR_NONE) {
                    ctx->cp[ctx->dirp->nameLen] = '\0';
#ifndef FS_NO_LOCK_TIMEOUT
                    curInode = I_GET_SHARED(0, ctx->curInum);
#else
                    curInode = I_GET_EXCL(0, ctx->curInum);
#endif
                    if (FS_I_IS_DIR(curInode->dinode.attrs)) {
#ifndef FS_NO_LOCK_TIMEOUT
                        fs_inodeRelease(context, curInode);
#else
                        fs_inodeReleaseExclusive(context, curInode);
#endif
                       
                        ctx2 = fs_ramAlloc(FS_RAM_RESTORE_DIRENTS, 
                                      sizeof(fs_restoreCompleteTempDirContext));
                        ctx2->cp         = ctx->cp + ctx->dirp->nameLen - 1;
                        ctx2->rootInum   = ctx->curInum;
                        ctx2->eof        = 0;
                        ctx2->cookie     = 0;
                        DLL_ISOLATE_ELEM(ctx, list);
                        DLL_INSERT_TAIL(&restoreComplTempDirCtxts, ctx, list);
                        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                                ("going down from %s to %s", ctx->cp, 
                                 ctx->dirp->name));
                        ctx = ctx2;
                        //We are now in the context of the new dir
                        goingDown = TRUE;
                        break;

goUp:                   //Adjusting cp
                        ctx->cp--;
                        *ctx->cp = '\0'; 

                        //Retrieve upper context
                        ctx2 = ctx;
                        DLL_REMOVE_TAIL(&restoreComplTempDirCtxts, ctx, list);
                        if (ctx != NULL && ctx != DLL_NIL)  {
                            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                                    ("going up from %s back to %s", ctx->dirp->name,
                                     ctx->cp));
                        } else {
                            FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                                    ("going up (end)")); 
                        }

                        //Free previous context
                        fs_ramDealloc(FS_RAM_DUMP_DIRENT, ctx2);

                        if (ctx == NULL || ctx == DLL_NIL)  {
                             //We are done, return here
                             return status;
                        }
                        
                        if (status != FS_ERR_NONE) {
                            goto goUp;
                        }
                    } else {
#ifndef FS_NO_LOCK_TIMEOUT
                        fs_inodeRelease(context, curInode);
#else
                        fs_inodeReleaseExclusive(context, curInode);
#endif
                        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                                ("releasing as non dir for %s", 
                                 ctx->dirp->name));
                    }
                    FS_READDIR_ADV_ENTRY_PTR(ctx->dirp);
                    continue;
                }

                /*
                 * Need to move this object
                 */

                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("Removing object in CompleteTempDir 1 for %s",
                            restoreContext->currentName));
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                            ("Dirp at this point is %s", ctx->dirp->name));

                curInode = I_GET_EXCL(0, ctx->curInum);
                fs_createTempLink(context, restoreContext,
                                  restoreContext->currentName, &curInode);
//            fs_inodeReleaseExclusive(context, curInode);
                fs_removeTempLink(context, restoreContext,
                                  restoreContext->currentName, &curInode);
                fs_inodeReleaseExclusive(context, curInode);
                FS_READDIR_ADV_ENTRY_PTR(ctx->dirp);
                
            }
            if (goingDown) break;
#ifndef FS_NO_LOCK_TIMEOUT
            tempRoot =  I_GET_SHARED(0, ctx->rootInum);
#else
            tempRoot =  I_GET_EXCL(0, ctx->rootInum);
#endif
        }
        if (goingDown) continue;

#ifndef FS_NO_LOCK_TIMEOUT
        fs_inodeRelease(context, tempRoot);
#else
        fs_inodeReleaseExclusive(context, tempRoot);
#endif
        status = FS_ERR_NONE;
        goto goUp;
        
    } while (1);

}

/*
 * Complete the rename operations that were temporarily performed to
 * a temp dir. Now move the objects to their right place.
 *
 * Walk the temp tree in preorder fashion. If a corresponding node
 * does not exist in the target tree, move the subtree.
 */
PRIVATE fs_Status
fs_restoreCompleteTemp(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext
)
{
    fs_Inode            *tempRoot;
    fs_Inumber          curInum;
    fs_Inode            *curInode;
    fs_DirEntry         *dirents, *dirp;
    int                 eof = 0;
    uint32              numEntries;
    uint64              cookie = 0;
    uint32              i;
    char                *cp;
    fs_Status           status = FS_ERR_NONE;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("In restore complete Temp"));
    

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In restoreComplete Temp"));
    dirents = fs_ramAlloc(FS_RAM_RESTORE_DIRENTS, FS_MAXPATHLEN);
#ifndef FS_NO_LOCK_TIMEOUT
    tempRoot =  I_GET_SHARED(0, restoreContext->tempInodeNum);
#else
    tempRoot =  I_GET_EXCL(0, restoreContext->tempInodeNum);
#endif
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Temp inum in complete temp %lld", tempRoot->inum));
    cp = restoreContext->currentName;
    *cp = '/';
    cp++;
    *cp = '\0';
    while (!eof) {
        dirp = dirents;
        fs_dir_readdir(context, tempRoot, dirp, FS_MAXPATHLEN + 1,
                        cookie, FALSE, &numEntries, &eof);
#ifndef FS_NO_LOCK_TIMEOUT
        fs_inodeRelease(context, tempRoot);
#else
        fs_inodeReleaseExclusive(context, tempRoot);
#endif
        //dirp = dirents;

        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In restoreComplete Temp numEntries %d", numEntries));
        for (i = 0; i < numEntries; i++) {
            cookie = dirp->cookie;
            curInum = dirp->inode;
                
            testSpecialDirectories(dirp);
                
            strncpy(cp, dirp->name, dirp->nameLen);
            if (fs_pathFound(context, restoreContext,
                             restoreContext->currentName) == FS_ERR_NONE) {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                    ("Path found in completTemp for %s returned err none",
                    restoreContext->currentName));
                cp[dirp->nameLen] = '\0';
                curInode = I_GET_EXCL(0, curInum);
                if (FS_I_IS_DIR(curInode->dinode.attrs)) {
                    fs_inodeReleaseExclusive(context,curInode);
                    status =fs_restoreCompleteTempDir(context, restoreContext,
                                             cp + dirp->nameLen - 1, curInum);
                    if (status != FS_ERR_NONE) {
                        fs_ramDealloc(FS_RAM_RESTORE_DIRENTS, dirents);
                        return status;
                    }
                } else {

                        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                            ("object exists %s", dirp->name));
                    /*
                     * Object exists. Remove Tree will free it
                     */

                    fs_inodeReleaseExclusive(context, curInode);
                }
                FS_READDIR_ADV_ENTRY_PTR(dirp);
                continue;
            }

            FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
             ("Path found in complete Temp 2 for %s did not return success",
                restoreContext->currentName));
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("Dirp at this point is %s", dirp->name));
            /*
             * Need to move this object
             */

            curInode = I_GET_EXCL(0, curInum);
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                ("Removing object in CompleteTemp 2 %s",
                 restoreContext->currentName));
            fs_createTempLink(context, restoreContext,
                              restoreContext->currentName, &curInode);
            fs_removeTempLink(context, restoreContext,
                              restoreContext->currentName, &curInode);
            fs_inodeReleaseExclusive(context, curInode);
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Continuing at end"));
            FS_READDIR_ADV_ENTRY_PTR(dirp);
            
        }
#ifndef FS_NO_LOCK_TIMEOUT
        tempRoot =  I_GET_SHARED(0, restoreContext->tempInodeNum);
#else
        tempRoot =  I_GET_EXCL(0, restoreContext->tempInodeNum);
#endif
    }
#ifndef FS_NO_LOCK_TIMEOUT
    fs_inodeRelease(context, tempRoot);
#else
    fs_inodeReleaseExclusive(context, tempRoot);
#endif
    return status;
    
}

/*
 * Utility routines
 */

/*
 * Given a pathname, findout the parent of the last component, inode
 * for the last component and name of the last component
 */

EXPORTED fs_Status
fs_decomposePath(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext,
    char                *pathName,
    fs_Inumber          *parent,
    fs_Inumber          *inum,
    char                **name,
    uint16              *nameLen,
    int                 temp
)
{
    fs_path_t           *path;
    char                *c, *p;
    fs_Inumber          rootInum = FS_ROOT_INODE;
    fs_Inode            *parentDir = NULL;
    fs_Inode            *component = NULL;
    fs_Inumber          parentInum;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif
    

    *name = NULL;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("Decompose %s tmp %d", pathName, temp));
    path = fs_makePath(pathName);
    if (temp) {
        rootInum = restoreContext->tempInodeNum;
    }
    parentDir = I_GET_EXCL(0, rootInum);
    parentInum = parentDir->inum;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("root inum %lld parent inum %lld snapnum %d\n", rootInum,
               parentInum, parentDir->snapnum));

    p = fs_getNextComponent(path);
    if (p == NULL) {
        fs_destroyPath(path);
        *parent = parentInum;
        if (inum) {
            *inum = 0;
        }
        *name = NULL;
        fs_inodeReleaseExclusive(context, parentDir);
        return FS_ERR_NOENT;
    }
    while (p != NULL) {
        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name\n ");
			FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("Returning no ent for %s as parent is not a dir",
                      pathName));
            return FS_ERR_NOENT;
        }

        parentInum = parentDir->inum;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("decomposePath:parent Inum %lld", parentInum));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("p is %s, len %d", p, strlen(p)));
        component = fs_lookup(context, context->fileSystem, 0,
                              parentDir, p);
        if (component == NULL) {
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("decomposePath:lookup failed"));
            c =  fs_getNextComponent(path);

            if (c != NULL) {
                fs_inodeReleaseExclusive(context, parentDir);

                fs_destroyPath(path);
                fs_ndmpSendRestoreLog(context, restoreContext,
                                      FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                      FS_ERR_NOENT, pathName,
                                      "Invalid Path name\n ");
				FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("Returning no ent for %s - %s", pathName, c));
                return FS_ERR_NOENT;
            }
            fs_inodeReleaseExclusive(context, parentDir);
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("p = %p, c = %p", p, c));
            break;
        }
        c = fs_getNextComponent(path);
        if (c == NULL) {
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("c is null"));
            break;
        }
        p = c;
        parentDir = component;
    }

    if (!p) {
        *name = NULL;
    } else {
        *nameLen = strlen(p);
        *name = fs_ramAlloc(FS_RAM_RESTORE_NAME, *nameLen + 1);
        strcpy(*name, p);
    }
    fs_destroyPath(path);
    *parent = parentInum;
    if (inum) {
        if (component) {
            *inum = component->inum;
        } else {
            *inum = 0;
        }
    }

    if (component) {
        fs_inodeReleaseExclusive(context, component);
    }
    return FS_ERR_NONE;
}


/*
 * Create a directory in parent with read/write/execute permission for
 * everybody
 */

PRIVATE fs_Status
fs_createTempDir(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    fs_Inode          **parent,
    char                *p,
    fs_Inode          **inode
)
{
    int32               rc;
    fs_Inumber          parentInum = (*parent)->inum;
    fs_Inumber          inum;
	uint64				unstableId;

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
      ("IN createTempDir for %s parent %lld\n", p, (*parent)->inum));

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }
    fs_logBeginTransaction(context);

    *inode = fs_allocateDirectoryInode(context, *parent, 0777, 0, 0,
                                 (*parent)->dinode.fsetNum);
    inum = (*inode)->inum;
    rc = fs_dir_addent(context, *parent, p, strlen(p), (*inode)->inum);
    (void)fs_addLink(context, 0, (*inode), (*parent)->inum);
    (*parent)->dinode.linkCnt += 1;
    (*parent)->dinode.ctime = fs_getTimeOfDay();
    (*parent)->dinode.mtime = (*parent)->dinode.ctime;
    (*parent)->dinode.version++;
    fs_inodeWrite(context, (*parent), INO_WRITE_LOG);

    fs_inodeWrite(context, *inode, INO_WRITE_LOG);

#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = (*inode)->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    *parent = I_GET_EXCL(0, parentInum);
    *inode = I_GET_EXCL(0, inum);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
      ("IN createTempDir for %s inode link cnt %d for %lld\n",
       p, (*inode)->dinode.linkCnt, inum));
    return FS_ERR_NONE;
}


/*
 * Given a pathname, ensure that all the ancestors in the path exist.
 * If they don't then create them. This are used only on tmp dir, create
 * the objects with read/write/execute permissions for everybody
 */

PRIVATE fs_Status
fs_ensureParentExists(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    int                 temp
)
{
    fs_path_t           *path;
    char                *pathName = restoreContext->targetBuf;
    char                *last = strrchr(pathName, '/');
    char                *p, *c;
    fs_Inode            *parentDir = NULL, *component = NULL;
    fs_Inumber          rootInum = FS_ROOT_INODE;
    fs_Status           status = FS_ERR_NONE;   

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif
    
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("IN ensure path exists for %s temp %d", pathName, temp));
    if (last == NULL) {
        return FS_ERR_NOENT;
    }
    if (last == pathName) {
        return FS_ERR_NONE;
    }
    *last = '\0';
    if (temp) {
        rootInum = restoreContext->tempInodeNum;
    }
    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {
        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name\n ");
            return FS_ERR_NOENT;
        }

        component = fs_lookup(context, context->fileSystem, 0,
                              parentDir, p);
        if (component == NULL) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("IN ensure path exists for component is null for %s", p));
            c =  fs_getNextComponent(path);

            if (c != NULL || c == NULL) {
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
      ("IN ensure path exists for component is null for %s c is %s", p, c));
                if (temp) {
                    status = fs_createTempDir(context, restoreContext,
                                              &parentDir, p, &component);
                    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
      ("IN ensure path createTempDir returned %d", status));
                } else {
                    fs_destroyPath(path);
                    fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NOENT, NULL,
                                          "Invalid Path name\n ");
                    return FS_ERR_NOENT;
                }
                if (status != FS_ERR_NONE) {
                    *last =  '/';
                    fs_inodeReleaseExclusive(context, parentDir);
                    fs_destroyPath(path);
                    return status;
                }
                p = c;
                fs_inodeReleaseExclusive(context, parentDir);
                parentDir = component;
                component = NULL;
                continue;
            }

            /*
             * Immediate parent does not exist and we cannot create it
             * return failure
             */
            fs_destroyPath(path);
            *last =  '/';
            fs_inodeReleaseExclusive(context, parentDir);
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
      ("IN ensure path exists for component is null for %s c is NULL", p));
            return FS_ERR_NOENT;
        }
        parentDir = component;
        component = NULL;
        p = fs_getNextComponent(path);
    }

    fs_destroyPath(path);
    *last =  '/';
    fs_inodeReleaseExclusive(context, parentDir);
    return FS_ERR_NONE;
}

PRIVATE fs_Status
fs_createTempLink(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    char                *path,
    fs_Inode            **inode
)
{
    fs_Status           status;
    fs_Inumber          parent;
    fs_Inode            *parentInode;
    char                *name;
    uint16              nameLen;
    fs_Inumber          curInum = (*inode)->inum;
    int                 rc;
	uint64				unstableId;

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("IN create temp link %s ", path));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("In createTempLink"));
    status = fs_decomposePath(context, restoreContext, path, &parent, NULL,
                              &name, &nameLen, 0);
    if (status != FS_ERR_NONE) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("decompose path returned %d", status));
        if (name) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, name);
        }
        return status;
    }

    FS_ASSERT(name != NULL);
    parentInode = I_GET_EXCL(0, parent);
    if (fs_dir_lookupent(context, parentInode, name, nameLen) != 0) {
        fs_inodeReleaseExclusive(context, parentInode);

        fs_ramDealloc(FS_RAM_RESTORE_NAME, name);
        return status;
    }
    fs_logBeginTransaction(context);
    fs_dir_addent(context, parentInode, name, nameLen, curInum);

    // RAHUL -- Update link count appropriately
    if (FS_I_IS_DIR((*inode)->dinode.attrs)) {
        parentInode->dinode.linkCnt++;
    }
    fs_inodeWrite(context, parentInode, INO_WRITE_LOG);
    status = fs_addLink(context, 0, *inode, parent);
    if (!fs_inodeWasModified(context, (*inode))) {
        fs_inodeReleaseExclusive(context, (*inode));
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = (*inode)->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    *inode = I_GET_EXCL(0, curInum);
    if (name) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, name);
    }
    return status;
}

PRIVATE fs_Status
fs_removeTempLink(
    fs_Context          *context,
    fs_RestoreContext_t  *restoreContext,
    char                *path,
    fs_Inode            **inode
)
{
    fs_Status           status;
    fs_Inumber          parent;
    fs_Inode            *parentInode;
    char                *name;
    uint16              nameLen;
    fs_Inumber          curInum = (*inode)->inum;
    int                 rc;
	uint64				unstableId;

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("In Remove Temp Link for %s", path));

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("In removeTempLink"));
    status = fs_decomposePath(context, restoreContext, path, &parent, NULL,
                              &name, &nameLen, 1);
    if (status != FS_ERR_NONE) {
        if (name) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, name);
        }
        return status;
    }

    fs_logBeginTransaction(context);
    parentInode = I_GET_EXCL(0, parent);
    fs_dir_removeent(context, parentInode, name, nameLen);
    
    // RAHUL -- Update link count appropriately
    if (FS_I_IS_DIR((*inode)->dinode.attrs)) {
        parentInode->dinode.linkCnt--;
    }
    fs_inodeWrite(context, parentInode, INO_WRITE_LOG);
    status = fs_removeLink(context, 0, *inode, parent);
    if (!fs_inodeWasModified(context, (*inode))) {
        fs_inodeReleaseExclusive(context, (*inode));
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = (*inode)->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    *inode = I_GET_EXCL(0, curInum);
    if (name) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, name);
    }
    return status;
}
PRIVATE fs_Status
fs_restoreExtractObject(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
)
{
    fs_Inumber              rootInum, inum;
    fs_Inode                *component, *parentDir;
    fs_path_t               *path;
    char                    *p, *c;
    char                    *pathName = restoreContext->targetBuf;
    uint16                  pathNameLen = restoreContext->targetBufLen;
    int                     temp = restoreContext->isTargetTemp;
    int32                   rc;
    fs_DumpDataExtension_t  *dataExtension;
    uint32                  numExtents, i, j;
    fs_DumpExtent_t         *drExtent;
    char                    *buf;
    uint64                  headerId = restoreContext->currentId;
    fs_Status               status = FS_ERR_NONE;
    char                    *newPath;
    int                     extractExtents = 0;
	uint64					unstableId;
    fs_Time                 now;
    fs_SecurityDescNfs      sdNew;

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("In restoreExtractObject"));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("extracting %s for extract object for id %lld",
                      pathName, headerId));
    if (fs_shouldIgnore(context, restoreContext, pathName, pathNameLen,
                        &newPath)) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("In restoreExtractObject draining"));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("Draining %s for extract object", pathName));
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        return fs_drainIncrObject(context, restoreContext);
    }
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif

    if (temp) {

        /*
         * We need to create the object in the temp directory
         */
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("extracting %s in tmp", pathName));
        rootInum = restoreContext->tempInodeNum;
    }

    if (newPath) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("extracting %s to newpath %s", pathName, newPath));
        strcpy(restoreContext->targetBuf, newPath);
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }
    /*
     * If we are to create an object in temporary directories,
     * if the intermediate ancestors are missing create them. For
     * in place creation fail if intermediate parents are missing
     */

    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("In restoreExtractObject p %s", p));
        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name\n ");
            return FS_ERR_NOENT;
        }

        component = fs_lookup(context, context->fileSystem, 0,
                              parentDir, p);
        if (component == NULL) {
            c =  fs_getNextComponent(path);

            if (c != NULL) {
                if (temp) {
                    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                      ("In restoreExtractObject calling createTempDir %s %s",
                        p, c));
                    status = fs_createTempDir(context, restoreContext,
                                              &parentDir, p, &component);
                    if (status != FS_ERR_NONE) {
                        fs_inodeReleaseExclusive(context, parentDir);
                        fs_destroyPath(path);
                        return status;
                    }
                    p = c;
                    fs_inodeReleaseExclusive(context, parentDir);
                    parentDir = component;
                    component = NULL;
                    continue;
                } else {
                    fs_inodeReleaseExclusive(context, parentDir);
                    fs_destroyPath(path);
                    fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NONE, pathName,
                                          "Invalid Path name\n");
                    return FS_ERR_NOENT;
                }
            }
            break;
        }
        parentDir = component;
        component = NULL;
        p = fs_getNextComponent(path);
    }

    now = fs_getTimeOfDay();
    
    if (p == NULL) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In restoreExtractObject p is null"));
        FS_ASSERT(component == NULL);
        /*
         * Objectr with the same name exists. Ensure that it is of
         * same type as the object attributes
         */

        component = parentDir;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
         ("object mode %o attr mode %o\n", component->sd->u.nfs.mode,
            restoreContext->attr.mode));
        if ((component->dinode.attrs & FS_IFMT_MASK) !=
            (restoreContext->attr.mode & FS_IFMT_MASK)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Different type of object exists\n");
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("Object exists in restoreExtractObject so returning"));
            return FS_ERR_EXIST;
        }
        fs_logBeginTransaction(context);
        parentDir = NULL;
        if (!FS_I_IS_DIR(component->dinode.attrs)) {
			if (!FS_I_IS_SYML(component->dinode.attrs) ||
				component->dinode.numblocks > 0) {
					fs_truncateFile(context, component, 0, FALSE, NULL);
			} else {
				FS_ASSERT(FS_I_IS_SYML(component->dinode.attrs));
				component->dinode.size = 0;
			}
        }
        fs_destroyPath(path);
        goto set_attr;
    }
    fs_logBeginTransaction(context);
    if (FS_I_IS_DIR(restoreContext->attr.mode))  {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("allocating dir inode for %s", pathName));
        component = fs_allocateDirectoryInode(context, parentDir, 0, //0777,
                                        restoreContext->attr.uid,
                                        restoreContext->attr.gid,
                                        parentDir->dinode.fsetNum);
    } else {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("allocating inode for %s", pathName));
        //component = fs_allocateFileInode(context, parentDir, 0777,
        component = fs_allocateFileInode(context, parentDir,
                                        restoreContext->attr.mode &
                                            FS_IFMT_MASK,
                                        restoreContext->attr.uid,
                                        restoreContext->attr.gid,
                                        parentDir->dinode.fsetNum);
    }
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("allocating dir entry for %s", pathName));
    rc = fs_dir_addent(context, parentDir, p, strlen(p), component->inum);
    fs_destroyPath(path);
    if (rc != NFX_OK) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("allocating link entry for %s", pathName));
    status = fs_addLink(context, 0, component, parentDir->inum);
    if (FS_I_IS_DIR(component->dinode.attrs)) {
            parentDir->dinode.linkCnt++;
    }
    if (status != FS_ERR_NONE) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    if (restoreContext->extract) {
        parentDir->dinode.atime =
        parentDir->dinode.mtime =
        parentDir->dinode.ctime = now;
        parentDir->dinode.version++;
    } else {
        parentDir->dinode.ctime = now;
        parentDir->dinode.version++;
    }
    fs_inodeWrite(context, parentDir, INO_WRITE_LOG);

set_attr:
    inum = component->inum;
    if (!FS_I_IS_DIR(component->dinode.attrs)) {
			if (!FS_I_IS_SYML(component->dinode.attrs) ||
				component->dinode.numblocks > 0) {
					fs_truncateFile(context, component, 0, FALSE, NULL);
			} else {
				FS_ASSERT(FS_I_IS_SYML(component->dinode.attrs));
				component->dinode.size = 0;
			}
    }
    component->dinode.mtime = restoreContext->attr.mtime;
    component->dinode.ctime = now;
    component->dinode.version++;
    component->dinode.atime = restoreContext->attr.atime;
    if (!FS_I_IS_DIR(component->dinode.attrs)) {
        component->dinode.size = restoreContext->attr.sizeDev.size;
    }
    component->dinode.attrs |= (restoreContext->attr.mode & FS_IATR_MASK);

    sdNew = component->sd->u.nfs;
    sdNew.mode = (restoreContext->attr.mode & FS_IMODE_MASK);
    fs_securityDescSet(context, component, &sdNew.hdr);
    // BOB - Rahul, please visit this
    component->dinode.userAttr = restoreContext->attr.simpleAttrs;
    //component->dinode.simpleAttrs =  (component->dinode.simpleAttrs & 
                                  //(FS_INODE_SATTR_NO_SNAP|
                                    //FS_INODE_SATTR_DIR_SPLIT)) |
                                    //restoreContext->attr.simpleAttrs;

    if (FS_I_IS_BLK(component->dinode.attrs) ||
        FS_I_IS_CHR(component->dinode.attrs)) {
        component->dinode.size = 0;
        component->dinode.u.devNum.major =
                restoreContext->attr.sizeDev.devNum.major;
        component->dinode.u.devNum.minor =
                restoreContext->attr.sizeDev.devNum.minor;
    }
    if (restoreContext->extract) {
        component->dinode.ctime = now;
        component->dinode.mtime = component->dinode.atime =
            component->dinode.ctime;
        component->dinode.version++;
    }
    
    fs_inodeWrite(context, component, INO_WRITE_LOG);

    if (FS_I_IS_REG(component->dinode.attrs)) {
        extractExtents = 1;
    }
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != NFX_OK) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("logCommit Failed\n"));
        status = FS_ERR_IO;
    }

    if (!extractExtents) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("returing as extractExtents is not set for %s", pathName));
        return status;
    }
    component = I_GET_EXCL(0, inum);

    /*
     * Now restore the object
     */

    headerId =
        ((fs_DumpGenericHeader_t *)&restoreContext->record)->headerId;
    /*
     * For incremental restores, we don't store any extents information
     * in the main operation header, but in a subsequent Data header
     */

    status = fs_readTape(context, restoreContext);
    if (status != FS_ERR_NONE) {
        /*
         * Log an error message
         */
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("Returning as read tape failed 1"));
        fs_inodeReleaseExclusive(context, component);

        return status;
    }

    buf = fs_ramAlloc(FS_RAM_RESTORE_DATA, FS_BLOCK_SIZE);

write_data:
    dataExtension =
            (fs_DumpDataExtension_t *)&restoreContext->record.u.extension;

    if (dataExtension->header.op != FS_DR_DATA ||
        dataExtension->header.headerId != headerId + 1) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
        ("returing as op is %d and id is %lld where as expected id is %lld",
            dataExtension->header.op, dataExtension->header.headerId,
            headerId + 1));
        /*
         * Push back the read header
         */

        fs_restorePushBackRead(context, restoreContext);
        fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);

        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("extract Object:Pushed back data"));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                  ("extract Object:op = %d", dataExtension->header.op));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("extract Object:id = %lld, expected id = %lld",
                 dataExtension->header.headerId, headerId));
        fs_inodeReleaseExclusive(context, component);
        return status;
    }

    /* SO FAR NOT USED...
     * If this object was a symlink, we have 1 following block containing
     * link data
     */

    if (FS_I_IS_SYML(component->dinode.attrs)) {
        fs_logBeginTransaction(context);
        if (component->dinode.size < FS_INODE_INLINE_LEN) {
            memcpy(component->dinode.u.data, &restoreContext->record.u.dummy,
                    component->dinode.size);                   
        } else {
            fs_Buf  *bp = NULL;
            fs_emaproot_init(context, component);
            bp = O_CREATE(component, 0);
            memcpy(bp->u.data, &restoreContext->record.u.dummy,
                   component->dinode.size);
            FS_BUF_ZERO_ALL(bp);
            FS_BUF_WRITE(bp, 0, component->dinode.size);
        }

        fs_inodeWrite(context, component, INO_WRITE_LOG);
#if 0
        rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);

    if (rc != NFX_OK) {
            status = FS_ERR_IO;
        }
        fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
        return status;
    }

    numExtents = dataExtension->header.dataExtents;
    drExtent = dataExtension->extents;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("restoreExtractObject numExt %d\n", numExtents));

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("restoring %d extents in restoreExtractObject", numExtents));
    for (i = 0; i < numExtents; i++) {

        /*
         * Process each extent
         */

        j = 0;
        while (j < drExtent->len) {
            fs_inodeReleaseExclusive(context, component);
            status = fs_readDataTape(context, restoreContext, buf);
            if (status != FS_ERR_NONE) {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                        ("Returning as read tape failed 2"));
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
                return status;
            }
            component = I_GET_EXCL(0, inum);

            /*
             * If error occurs inode is released
             */

#ifdef FS_DEBUG
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("restoreExtractObject: cksum 0x%x\n",
                            fs_checksum(buf, FS_DR_TBSIZE, ~0)));
#endif
            status = fs_restoreWriteFile(context, restoreContext, &component,
                                         buf,
                                        drExtent->offset + (j << FS_DR_TBSHIFT),
                                         FS_DR_TBSIZE);
            if (status != FS_ERR_NONE) {
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
                fs_inodeReleaseExclusive(context, component);
                return status;
            }
            j++ ;
        }
        drExtent++;
    }
        

    /*
     * Read the tape till, we see DumpExtension records for this headerId
     */

    fs_inodeReleaseExclusive(context, component);
    status = fs_readTape(context, restoreContext);
    component = I_GET_EXCL(0, inum);
    if (status != FS_ERR_NONE){
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                        ("Returning as read tape failed 3"));
            fs_inodeReleaseExclusive(context, component);
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
    }
    goto write_data;

    /* NEVER REACHED */
}

PRIVATE fs_Status
fs_drainIncrObject(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext
)
{
    fs_Status           status = FS_ERR_NONE;
    fs_DumpGenericHeader_t  *header;
    fs_DumpOperation_t      *operation;
    fs_DumpDataExtension_t  *extension;
    uint64              i;
    fs_DumpExtent_t     *drExtent;
    uint32              numExtents;
    uint64              blocks;
    int                 nHeaderBlocks = 1;
    uint64              id;
    char                *buf;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
            ("IN drain incr object "));
    buf = fs_ramAlloc(FS_RAM_RESTORE_DATA, FS_DR_TBSIZE);
    header = (fs_DumpGenericHeader_t *)&restoreContext->record;
    id = header->headerId;
more:

    header = (fs_DumpGenericHeader_t *)&restoreContext->record;
    if (header->headerId != id) {
        if (header->op == FS_DR_DATA && (header->headerId == id + 1)) {
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Found DR_DATA with headerId %lld", header->headerId));
        } else {
            status = fs_restorePushBackRead(context, restoreContext);
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
        }
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Draining op %d", header->op));
    switch (header->op) {

    case FS_DR_CREATE_SOCK:
    case FS_DR_CREATE_BLOCK:
    case FS_DR_CREATE_FIFO:
    case FS_DR_CREATE_FSET:
    case FS_DR_CREATE_DIR:
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Draining fset, dir or special objects"));
        operation = (fs_DumpOperation_t *)header;
        nHeaderBlocks = (operation->len + sizeof (fs_DumpOperation_t) +
                         FS_DR_TBSIZE - 1) >> FS_DR_TBSHIFT;
        if (nHeaderBlocks > 1) {
            while (nHeaderBlocks > 1) {
                status = fs_readTape(context, restoreContext);
                if (status != FS_ERR_NONE) {
                    return status;
                }
                nHeaderBlocks--;
            }
            
        }

        // Read the next block
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
        }
        goto more;
    case FS_DR_CREATE_SYML:
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Draining symlink"));
        operation = (fs_DumpOperation_t *)header;
        nHeaderBlocks = (operation->len + sizeof (fs_DumpOperation_t) +
                         FS_DR_TBSIZE - 1) >> FS_DR_TBSHIFT;
        if (nHeaderBlocks > 1) {
            while (nHeaderBlocks > 1) {
                status = fs_readTape(context, restoreContext);
                if (status != FS_ERR_NONE) {
                    return status;
                }
                nHeaderBlocks--;
            }
            
        } else  if (header->dataExtents == 0) {
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return FS_ERR_NONE;
        }
        // Read the next block
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
        }
        goto more;
    case FS_DR_NAME_ONLY:
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Draining name"));
        operation = (fs_DumpOperation_t *)header;
        nHeaderBlocks = (operation->len + sizeof (fs_DumpOperation_t) +
                         FS_DR_TBSIZE - 1) >> FS_DR_TBSHIFT;
        if (nHeaderBlocks > 1) {
            while (nHeaderBlocks > 1) {
                status = fs_readTape(context, restoreContext);
                if (status != FS_ERR_NONE) {
                    return status;
                }
                nHeaderBlocks--;
            }
            
        }
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
        }
        goto more;
    case FS_DR_CREATE_FILE:
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Draining create file"));
        operation = (fs_DumpOperation_t *)header;
        nHeaderBlocks = (operation->len + sizeof (fs_DumpOperation_t) +
                         FS_DR_TBSIZE - 1) >> FS_DR_TBSHIFT;
        if (nHeaderBlocks > 1) {
            while (nHeaderBlocks > 1) {
                status = fs_readTape(context, restoreContext);
                if (status != FS_ERR_NONE) {
                    return status;
                }
                nHeaderBlocks--;
            }
            goto more;
        }
        if (header->dataExtents == 0) {
            status = fs_readTape(context, restoreContext);
            if (status != FS_ERR_NONE) {
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
                return status;
            }
            goto more;
        }
        numExtents = header->dataExtents;
        drExtent = (fs_DumpExtent_t *)(restoreContext->record.u.dummy +
                                       sizeof (operation));
        blocks = 0;
        for (i = 0; i < numExtents; i++) {
            //blocks += (drExtent->len >> FS_DR_TBSHIFT);
            blocks += drExtent->len;
            drExtent++;
        }
        for (i = 0; i < blocks; i++) {
            status = fs_readDataTape(context, restoreContext, buf);
            if (status != FS_ERR_NONE) {
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
                return status;
            }
        }
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
        }
        goto more;
    case FS_DR_DATA:
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Draining data"));
        extension = (fs_DumpDataExtension_t *)header;
        if (header->dataExtents == 0) {

            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Zero extents for DR_DATA"));
            status = fs_readTape(context, restoreContext);
            if (status != FS_ERR_NONE) {
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
                return status;
            }
            goto more;
        }
        numExtents = header->dataExtents;
        drExtent =  extension->extents;
        blocks = 0;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("draining %d extents for DR_DATA", numExtents));
        for (i = 0; i < numExtents; i++) {
            //blocks += (drExtent->len >> FS_DR_TBSHIFT);
            blocks += drExtent->len;
            drExtent++;
        }
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("draining %lld blocks for DR_DATA", blocks));
        for (i = 0; i < blocks; i++) {
            status = fs_readDataTape(context, restoreContext, buf);
            if (status != FS_ERR_NONE) {
                fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
                return status;
            }
        
        }
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
        }
        goto more;

    case FS_DR_REMOVE:
    case FS_DR_REMOVE_DIR:
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Draining remove"));
        fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
        return status;
    case FS_DR_RENAME:
    case FS_DR_CREATE_LINK:
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Draining rename and link"));
        operation = (fs_DumpOperation_t *)header;
        nHeaderBlocks = (operation->len + sizeof (fs_DumpOperation_t) +
                         FS_DR_TBSIZE - 1) >> FS_DR_TBSHIFT;
        if (nHeaderBlocks > 1) {
            while (nHeaderBlocks > 1) {
                status = fs_readTape(context, restoreContext);
                if (status != FS_ERR_NONE) {
                    return status;
                }
                nHeaderBlocks--;
            }
            
        }
        // Read the next block
        status = fs_readTape(context, restoreContext);
        if (status != FS_ERR_NONE) {
            fs_ramDealloc(FS_RAM_RESTORE_DATA, buf);
            return status;
        }
        goto more;

    default:
        FS_ASSERT(0);
        goto more;
    }
}

PRIVATE fs_Status
fs_restoreCreateNode(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext,
    uint8                   type
)
{
    fs_Inumber              rootInum, inum;
    fs_Inode                *component, *parentDir;
    fs_path_t               *path;
    char                    *p, *c;
    char                    *pathName = restoreContext->targetBuf;
    uint16                  pathNameLen = restoreContext->targetBufLen;
    int                     temp = restoreContext->isTargetTemp;
    int32                   rc;
    fs_Status               status = FS_ERR_NONE;
    uint16                  mode;
    char                    *newPath;
	uint64				unstableId;
    fs_SecurityDescNfs  sdNew;

    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("IN restore create node for %s ", pathName));
    fs_logBeginTransaction(context);
#if 0
    rootInum = FS_ROOT_INODE;
#else
    rootInum = restoreContext->cwdInum;
#endif
    if (fs_shouldIgnore(context, restoreContext, pathName, pathNameLen,
                        &newPath)) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Draining %s for create node", pathName));
        if (newPath) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
        }
        return fs_drainIncrObject(context, restoreContext);
    }

    if (temp) {

        /*
         * We need to create the object in the temp directory
         */
        rootInum = restoreContext->tempInodeNum;
    }

    if (newPath) {
        strcpy(restoreContext->targetBuf, newPath);
        fs_ramDealloc(FS_RAM_RESTORE_NAME, newPath);
    }
    switch (type) {

        case FS_DR_CREATE_BLOCK:
            mode = FS_IFMT_BLK;
            break;
        case FS_DR_CREATE_CHAR:
            mode = FS_IFMT_CHR;
            break;
        case FS_DR_CREATE_SOCK:
            mode = FS_IFMT_SOCK;
            break;
        case FS_DR_CREATE_FIFO:
            mode = FS_IFMT_FIFO;
            break;
        default: 
            FS_ASSERT(0);
            mode = FS_IFMT_BLK;
            break;
    }

    /*
     * If we are to create an object in temporary directories,
     * if the intermediate ancestors are missing create them. For
     * in place creation fail if intermediate parents are missing
     */

    path = fs_makePath(pathName);
    parentDir = I_GET_EXCL(0, rootInum);

    p = fs_getNextComponent(path);
    while (p != NULL) {
        if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Invalid Path name\n ");
            return FS_ERR_NOENT;
        }

        component = fs_lookup(context, context->fileSystem, 0,
                              parentDir, p);
        if (component == NULL) {
            c =  fs_getNextComponent(path);

            if (c != NULL) {
                if (temp) {
                    status = fs_createTempDir(context, restoreContext,
                                              &parentDir, p, &component);
                    if (status != FS_ERR_NONE) {
                        fs_inodeReleaseExclusive(context, parentDir);
                        fs_destroyPath(path);
                        return status;
                    }
                    p = c;
                    fs_inodeReleaseExclusive(context, parentDir);
                    parentDir = component;
                    component = NULL;
                    continue;
                } else {
                    fs_inodeReleaseExclusive(context, parentDir);
                    fs_destroyPath(path);
                    fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NOENT, pathName,
                                          "Invalid Path name\n ");
                    return FS_ERR_NOENT;
                }
            }
            break;
        }
        parentDir = component;
        component = NULL;
        p = fs_getNextComponent(path);
    }

    FS_ASSERT(component == NULL);

    if (p == NULL) {

        /*
         * A node with this name already exists. Ensure that it is of the
         * type we are restoring
         */

        component = parentDir;
        if ((component->dinode.attrs & FS_IFMT_MASK) != mode) {
            fs_inodeReleaseExclusive(context, parentDir);

            fs_destroyPath(path);
            fs_ndmpSendRestoreLog(context, restoreContext,
                                  FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                  FS_ERR_NOENT, pathName,
                                  "Different type of object exists\n");
            return FS_ERR_EXIST;
        }
        parentDir = NULL;
        fs_destroyPath(path);
        goto set_attr;
    }

    /*
     * Now create the entry in the target Directory
     */

    component = fs_allocateFileInode(context, parentDir, mode, // | 0777,
                                    restoreContext->attr.uid,
                                    restoreContext->attr.gid,
                                    parentDir->dinode.fsetNum);
    if (component == NULL) {
        /*
         * Send a log message
         */
        fs_destroyPath(path);
        fs_ndmpSendRestoreLog(context, restoreContext,
                                          FS_ERR_NOENT, NDMP_LOG_MSG_TYP_ERR,
                                          FS_ERR_NOENT, NULL,
                                          "Could not create file\n ");

        fs_inodeReleaseExclusive(context, parentDir);
        return FS_ERR_IO;
    }

    inum = component->inum;

    rc = fs_dir_addent(context, parentDir, p, strlen(p), component->inum);
    fs_destroyPath(path);
    if (rc != NFX_OK) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    status = fs_addLink(context, 0, component, parentDir->inum);
    if (status != FS_ERR_NONE) {
        fs_inodeReleaseExclusive(context, parentDir);
        fs_inodeReleaseExclusive(context, component);
        return FS_ERR_IO;
    }
    fs_inodeWrite(context, parentDir, INO_WRITE_LOG);
set_attr:
    component->dinode.mtime = restoreContext->attr.mtime;
    component->dinode.ctime = fs_getTimeOfDay();
    component->dinode.version++;
    component->dinode.atime = restoreContext->attr.atime;
    component->dinode.size = 0;
    component->dinode.attrs |= (restoreContext->attr.mode & FS_IATR_MASK);

    sdNew = component->sd->u.nfs;
    sdNew.mode = (restoreContext->attr.mode & FS_IMODE_MASK);
    fs_securityDescSet(context, component, &sdNew.hdr);
    // BOB - Rahul, please visit this
    component->dinode.userAttr = restoreContext->attr.simpleAttrs;
    //component->dinode.simpleAttrs =  (component->dinode.simpleAttrs & 
                                  //(FS_INODE_SATTR_NO_SNAP|
                                    //FS_INODE_SATTR_DIR_SPLIT)) |
                                    //restoreContext->attr.simpleAttrs;

    if (FS_I_IS_BLK(component->dinode.attrs) ||
        FS_I_IS_CHR(component->dinode.attrs)) {
        component->dinode.u.devNum.major =
                restoreContext->attr.sizeDev.devNum.major;
        component->dinode.u.devNum.minor =
                restoreContext->attr.sizeDev.devNum.minor;
    }

    fs_inodeWrite(context, component, INO_WRITE_LOG);
#if 0
    rc = fs_logCommitEntry(context);
#else
	unstableId = component->inum;
	vl_addUnstableId(context, &unstableId);
	fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);
	rc = NFX_OK;
    fs_makeSureLogIsNotFull(context);
#endif

    fs_restoreSleepIfNeeded(context);
    
    if (rc != NFX_OK) {
        status = FS_ERR_IO;
    }
    return status;
}

PRIVATE fs_Status
fs_pathFound(
    fs_Context          *context,
    fs_RestoreContext_t *restoreContext,
    char                *path
)
{
    fs_Status           status;
    fs_Inumber          parent, last;
    char                *name;
    uint16              nameLen;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("IN path found for %s ", path));
    status = fs_decomposePath(context, restoreContext,
                              path, &parent, &last, &name, &nameLen, 0);
    if (status != FS_ERR_NONE) {
        if (name) {
            fs_ramDealloc(FS_RAM_RESTORE_NAME, name);
        }
        return status;
    }
    if (name) {
        fs_ramDealloc(FS_RAM_RESTORE_NAME, name);
    }

    if (parent != 0 && last != 0) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Found %s", path));
        return FS_ERR_NONE;
    } else {
        return FS_ERR_NOENT;
    }
}

EXPORTED fs_Status
fs_cleanRestoreResidue(
    fs_Context          *context
)
{
    fs_RestoreContext_t *restoreContext;
    fs_Status           status = FS_ERR_NONE;
    fs_DirEntry         *dirents, *dirp;
    int                 eof = 0;
    uint32              numEntries;
    uint64              cookie = 0;
    uint32              i;
    fs_Inode            *tempDirInode;
    fs_Inumber          curInum;
    int                 change = 0;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    fs_growIfNeeded(context);
    if (!fs_haveMinfreeAvail(context->fileSystem)) {
        return FS_ERR_NOSPC;
    }

    restoreContext = fs_ramAlloc(FS_RAM_DUMP_CONTEXT,
                          sizeof (fs_RestoreContext_t) +
                          FS_DR_TBSIZE);
     bzero(restoreContext, sizeof (fs_RestoreContext_t));

    dirents = fs_ramAlloc(FS_RAM_RESTORE_DIRENTS, FS_MAXPATHLEN);

retry:
    tempDirInode = I_GET_EXCL(0, FS_TEMPDIR_INODE);
    change = 0;
    while (!eof) {
        dirp = dirents;
        fs_dir_readdir(context, tempDirInode, dirp, FS_MAXPATHLEN + 1,
                        cookie, FALSE, &numEntries, &eof);
        fs_inodeReleaseExclusive(context, tempDirInode);
        for (i = 0; i < numEntries; i++) {
            cookie = dirp->cookie;
            curInum = dirp->inode;
            if (strcmp(dirp->name, ".") == 0) {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("found ."));
                FS_READDIR_ADV_ENTRY_PTR(dirp);
                continue;
            }
            if (strcmp(dirp->name, "..") == 0) {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("found .."));
                FS_READDIR_ADV_ENTRY_PTR(dirp);
                continue;
            }
            if (strcmp(dirp->name, ".snapshot") == 0) {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("found snapshot"));
                FS_READDIR_ADV_ENTRY_PTR(dirp);
                continue;
            }
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
              ("calling remove sub tree for %s", dirp->name));
            restoreContext->tempInodeNum = dirp->inode;
            change = 1;
            fs_removeTempRoot(context, restoreContext);
            FS_READDIR_ADV_ENTRY_PTR(dirp);
        }
        if (change) {
            cookie = 0 ;
            eof = 0 ;
            goto retry;
        }
        tempDirInode = I_GET_EXCL(0, FS_TEMPDIR_INODE);
    }
    fs_inodeReleaseExclusive(context, tempDirInode);
    if (change) {
        cookie = 0 ;
        eof = 0 ;
        goto retry;
    }
    fs_ramDealloc(FS_RAM_RESTORE_DIRENTS, dirents);
    fs_ramDealloc(FS_RAM_DUMP_CONTEXT, restoreContext);
    return status;
}

#endif  //SSC
