/*
 * Name		: fs-link.c
 * RCS		: $ID$
 * Description	: Change log implementation 
 * Created by	: Rahul Khona
 * Date Created	: 8/10/2001
 */

#if !defined(SSC)

#include "fs-includes.h"

PRIVATE fs_Status fs_newLink(fs_Context *, uint64, fs_Inumber, fs_Inumber);
PRIVATE fs_Status fs_remLink(fs_Context *, uint64, fs_Inumber, fs_Inumber);
PRIVATE fs_Status fs_findInternalParent(fs_Context *, uint64, fs_Inumber,
                                        fs_Inumber *, uint64 *);
/*
 * We maintain a link file which has entries for all the hardlinks
 * in a given filesystem. Each entry is of the form 
 * inode, parent inode, numlinks
 * where inode is the inode in question, parent inode is the inode
 * number which contains 1 or more entries for the inode and numlinks
 * is the count of number of links to the inode from a given parent
 *
 * Since hardlinks are not very common now a days and there would be very
 * few of them per filesystem, we will keep this a linear file. If it
 * becomes too much of a problem, we will make it a btree based index
 */

/*
  When we modify a buffer (FS_BUF_WRITE) we hold the buf Excl,
  If we would fs_inodeReleaseExclusive(context, linkTable);
    Another transaction, can get the inode, and try to get the same buffer,
    The first transaction can have a need to find/update another link,
    But he may not get the FS_LINK_INODE. 
    This breaks the lock order rule (first inode, then buffers)
*/

PRIVATE fs_Status
fs_newLink(
    fs_Context          *context,
    uint64              snapnum,
    fs_Inumber          inum,
    fs_Inumber          parent
)
{
    fs_Inode            *linkTable;
    fs_Buf              *buf;
    uint64              offset;
    fs_LinkEntry_t      *linkEntry;
    int                 nEntries;
    int                 i;
    uint64              freeOffset;

    FS_TRACE(context, FS_TRACE_REQ, ("In newLink"));
    linkTable = I_GET_EXCL(snapnum, FS_LINK_INODE);
    FS_ASSERT(linkTable != NULL);

    freeOffset = linkTable->dinode.size;
    offset = 0;

    while (offset < linkTable->dinode.size) 
    {
        buf = O_GET_EXCL(linkTable, (uint64)(offset >> (int)FS_BLOCK_SHIFT));

        FS_ASSERT(buf != NULL);

        nEntries = FS_BLOCK_SIZE / sizeof (fs_LinkEntry_t);
        linkEntry = (fs_LinkEntry_t *)buf->u.data;
        i = 0;
        while (i < nEntries) 
        {
            if (linkEntry->inum == inum) 
            {
                if (linkEntry->parent == parent) 
                {
                    linkEntry->numLinks++;
                    FS_BUF_WRITE(buf, 
                                 (int32)((char*)(&linkEntry->numLinks) - (char*)(buf->u.data)),
                                 sizeof (linkEntry->numLinks));

                    // Dirty the link inode to avoid deadlocks
                    // see comment at top of file
                    linkTable->dinode.mtime = fs_getTimeOfDay();
                    fs_inodeWrite(context, linkTable, INO_WRITE_LOG);
                    return FS_ERR_NONE;
                }
            }

            if ((offset < freeOffset) && (linkEntry->numLinks == 0))
            {
                freeOffset = offset;
            }
            linkEntry++;
            i++;
        }

        fs_bufReleaseExclusive(context, buf);
        offset += FS_BLOCK_SIZE;
    }

    /*
     * We did not find an entry. Add it to the file. We have already 
     * looked for a free entry. Use it
     */

    buf = O_GET_EXCL(linkTable, (uint64)(freeOffset >> (int)FS_BLOCK_SHIFT));
    if (buf == NULL) 
    {
        FS_ASSERT(freeOffset == linkTable->dinode.size);
        /*
         * Create a new buffer
         */
        buf = O_CREATE(linkTable, (uint64)(freeOffset >> (int)FS_BLOCK_SHIFT));

        FS_ASSERT(buf != NULL);
        bzero(buf->u.data, FS_BLOCK_SIZE);
        FS_BUF_ZERO_ALL(buf);
        linkTable->dinode.size += FS_BLOCK_SIZE;
        fs_inodeWrite(context, linkTable, INO_WRITE_LOG);
    }
    else
    {
        // Dirty the link inode to avoid deadlocks
        // see comment at top of file
        linkTable->dinode.mtime = fs_getTimeOfDay();
        fs_inodeWrite(context, linkTable, INO_WRITE_LOG);
    }

    nEntries = FS_BLOCK_SIZE / sizeof (fs_LinkEntry_t);
    linkEntry = (fs_LinkEntry_t *)buf->u.data;
    i = 0;
    while (i < nEntries) 
    {
        if (linkEntry->numLinks == 0) 
        {
            linkEntry->inum = inum;
            linkEntry->parent = parent;
            linkEntry->numLinks++; 
            FS_BUF_WRITE(buf, 
                         (int32)((char*)(linkEntry) - (char*)(buf->u.data)),
                         sizeof (fs_LinkEntry_t));
            return FS_ERR_NONE;
        }
        linkEntry++;
        i++;
    }

    FS_ASSERT(0);
    return FS_ERR_NOENT;
}

/*
 * Remove a link to the inode from the links table
 */

PRIVATE fs_Status
fs_remLink(
    fs_Context              *context,
    uint64                  snapnum,
    fs_Inumber              inum,
    fs_Inumber              parentInum
)
{
    fs_Inode            *linkTable;
    fs_Buf              *buf;
    uint64              offset;
    fs_LinkEntry_t      *linkEntry;
    int                 nEntries;
    int                 i;

    FS_TRACE(context, FS_TRACE_REQ, ("In remLink"));
    linkTable = I_GET_EXCL(snapnum, FS_LINK_INODE);
    FS_ASSERT(linkTable != NULL);

    //offset = linkTable->dinode.size;
    offset = 0;
    while (offset < linkTable->dinode.size) 
    {
        buf = O_GET_EXCL(linkTable, (uint64)(offset >> (int)FS_BLOCK_SHIFT));
        if (buf == NULL) 
        {
            fs_inodeReleaseExclusive(context, linkTable);
            return FS_ERR_NOENT;
        }

        FS_ASSERT(buf != NULL);

        nEntries = FS_BLOCK_SIZE / sizeof (fs_LinkEntry_t);
        linkEntry = (fs_LinkEntry_t *)buf->u.data;
        i = 0;
        while (i < nEntries) 
        {
            if (linkEntry->inum == inum) 
            {
                if (linkEntry->parent == parentInum) 
                {
                    FS_ASSERT(linkEntry->numLinks > 0);
                    linkEntry->numLinks--;
                    if (linkEntry->numLinks == 0)
                    {
                         linkEntry->parent = linkEntry->inum = 0;
                    }
                    FS_BUF_WRITE(buf, 
                                 (int32)((char*)(linkEntry) - (char*)(buf->u.data)),
                                 sizeof (fs_LinkEntry_t));
					
                    // Dirty the link inode to avoid deadlocks
                    // see comment at top of file
                    linkTable->dinode.mtime = fs_getTimeOfDay();
                    fs_inodeWrite(context, linkTable, INO_WRITE_LOG);
                    return FS_ERR_NONE;
                }
            }

            linkEntry++;
            i++;
        }

        fs_bufReleaseExclusive(context, buf);
        offset += FS_BLOCK_SIZE;
    }

    fs_inodeReleaseExclusive(context, linkTable);
    return FS_ERR_NOENT;
}

/*
 * Given an inun and a cookie, find the next parent for the inode
 */

PRIVATE fs_Status
fs_findInternalParent(
    fs_Context              *context,
    uint64                  snapnum,
    fs_Inumber              inum,
    fs_Inumber              *parent,
    uint64                  *cookie
)
{
    fs_Inode            *linkTable;
    fs_Buf              *buf;
    uint64              offset = 0;
    fs_LinkEntry_t      *linkEntry;
    int                 nEntries;
    int                 i = 0;
    uint32              boff;

    linkTable = I_GET_SHARED(snapnum, FS_LINK_INODE);
    FS_ASSERT(linkTable != NULL);

    offset = *cookie;
    boff = offset & (FS_BLOCK_SIZE - 1);
    while (offset < linkTable->dinode.size) 
    {
        buf = O_GET_EXCL(linkTable, (uint64)(offset >> (int)FS_BLOCK_SHIFT));
        if (buf == NULL) 
        {
            fs_inodeRelease(context, linkTable);
            return FS_ERR_NOENT;
        }

        FS_ASSERT(buf != NULL);

        nEntries = (FS_BLOCK_SIZE - boff) / sizeof (fs_LinkEntry_t);
        linkEntry = (fs_LinkEntry_t *)(buf->u.data + boff);
        i = 0;
        while (i < nEntries) 
        {
            if (linkEntry->inum == inum) 
            {
                *cookie = offset + boff + sizeof (fs_LinkEntry_t);
                *parent = linkEntry->parent;
                fs_bufReleaseExclusive(context, buf);
                fs_inodeRelease(context, linkTable);
                return FS_ERR_NONE;
            }
            boff += sizeof (fs_LinkEntry_t);
            linkEntry++;
            i++;
        }

        fs_bufReleaseExclusive(context, buf);
        offset += FS_BLOCK_SIZE;
        offset &= ~(FS_BLOCK_SIZE - 1);
        boff = 0;
    }

    fs_inodeRelease(context, linkTable);
    return FS_ERR_NOENT;
}

EXPORTED fs_Status
fs_addLink(
    fs_Context          *context,
    uint64              snapnum,
    fs_Inode            *inode,
    fs_Inumber          parent
)
{
    FS_TRACE(context, FS_TRACE_REQ, ("In addLink"));
    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));
    FS_ASSERT(context->logSummary != NULL);

    if (FS_I_IS_DIR(inode->dinode.attrs))
    {
        // Don't record this for dirs, they
        // never have hard links, and the parent
        // inum is available in the btree.
        return FS_ERR_NONE;
    }
    if (inode->dinode.piu.parentInum == 0) {
        // First time here, just record the parent inum.
        inode->dinode.piu.parentInum = parent;
        fs_inodeWrite(context, inode, INO_WRITE_LOG);
        return FS_ERR_NONE;
    }
    return fs_newLink(context, snapnum, inode->inum, parent);
}

EXPORTED fs_Status
fs_removeLink(
    fs_Context          *context,
    uint64              snapnum,
    fs_Inode            *inode,
    fs_Inumber          parent
)
{
    fs_Status           status;
    
    FS_TRACE(context, FS_TRACE_REQ, ("In removeLink"));
    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));
    FS_ASSERT(context->logSummary != NULL);
    if (FS_I_IS_DIR(inode->dinode.attrs))
    {
        // Don't record this for dirs, they
        // never have hard links, and the parent
        // inum is available in the btree.
        return FS_ERR_NONE;
    }

    if (inode->dinode.piu.parentInum == parent) {
        inode->dinode.piu.parentInum = 0;
        fs_inodeWrite(context, inode, INO_WRITE_LOG);
        return FS_ERR_NONE;
    }

    status = fs_remLink(context, snapnum, inode->inum, parent);

#if 0
    if (status == FS_ERR_NOENT && inode->dinode.piu.parentInum == parent) {
        inode->dinode.piu.parentInum = 0;
        fs_inodeWrite(context, inode, INO_WRITE_LOG);
        return FS_ERR_NONE;
    }
#endif

    return status;
}

EXPORTED fs_Status
fs_findParent(
    fs_Context          *context,
    uint64              snapnum,
    fs_Inumber          inum,
    fs_Inumber          *parent,
    uint64              *cookie
)
{
    fs_Inode            *inode;
    uint64              mycookie = 0, *tcookie = cookie;

    FS_TRACE(context, FS_TRACE_REQ, ("In findParent"));
//    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));
    if (inum == FS_ROOT_INODE) {
        *parent = 0;
        return FS_ERR_NONE;
    }
    inode = I_GET_SHARED(snapnum, inum);
    if (FS_I_IS_DIR(inode->dinode.attrs))
    {
        *parent = fs_dirroot_get_parent(context, inode); 
        fs_inodeRelease(context, inode);
        if (cookie) {
            *cookie = 0;
        }
        return FS_ERR_NONE;
    }
    if (inode->dinode.piu.parentInum != 0) {
        *parent = inode->dinode.piu.parentInum;
        fs_inodeRelease(context, inode);
        if (cookie) {
            *cookie = 0;
        }
        return FS_ERR_NONE;
    }

    fs_inodeRelease(context, inode);

    if (tcookie == NULL) {
        tcookie = &mycookie;
    }
    return fs_findInternalParent(context, snapnum, inum, parent, tcookie);
}

EXPORTED void
fs_eek_link_inode(
    fs_Context  *context,
    uint32      snapnum,
    boolean     fixIt
)
{
    eeeRTC_t            beginTime, delta;
    fs_Inode            *linkTable, *inode;
    fs_Buf              *buf;
    uint64              offset;
    fs_LinkEntry_t      *linkEntry;
    int                 nEntries;
    int                 i, linkcount;
    uint64              freeOffset;
    boolean             writeBuf, wasModified = FALSE;

    linkTable = I_GET_EXCL(snapnum, FS_LINK_INODE);
    FS_ASSERT(linkTable != NULL);
    freeOffset = linkTable->dinode.size;
    offset = 0;

    eee_rtc(&beginTime);

    while (offset < linkTable->dinode.size)
    {
        buf = O_GET_EXCL(linkTable, (uint64)(offset >> (int)FS_BLOCK_SHIFT));
        FS_ASSERT(buf != NULL);

        nEntries = FS_BLOCK_SIZE / sizeof (fs_LinkEntry_t);
        linkEntry = (fs_LinkEntry_t *)buf->u.data;
        i = 0;
        while (i < nEntries)
        {
            eee_rtcDuration(&beginTime, &delta);
            if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
            {
                fs_sleep(context, 0);
                eee_rtc(&beginTime);
            }

            writeBuf = FALSE;

            // do some basic checks of inums first
            if (linkEntry->inum > context->fileSystem->params.maxInodeNumber)
            {
                fs_eek_print_status(context, FALSE,
                        "INVALID inum entry I#%lld", linkEntry->inum);
                linkcount = 0;
                writeBuf = fixIt;
                context->eek.errcount++;
            }
            else if (linkEntry->parent > context->fileSystem->params.maxInodeNumber)
            {
                fs_eek_print_status(context, FALSE,
                        "INVALID parent inum entry I#%lld", linkEntry->parent);
                linkcount = 0;
                writeBuf = fixIt;
                context->eek.errcount++;
            }
            else if ((linkEntry->inum != 0) && (linkEntry->parent != 0))
            {
                // check if numLinks match what we found by scanning all directories
                linkcount = fs_eek_find_parent_linkcount(context,
                                    linkEntry->parent, linkEntry->inum);
                // adjust for the one stored in piu.parentInum
                inode = I_GET_SHARED(snapnum, linkEntry->inum);
                if (linkcount && inode->dinode.piu.parentInum &&
                    (inode->dinode.piu.parentInum == linkEntry->parent)) {
                    linkcount--;
                }
                fs_inodeRelease(context, inode);
                if (linkcount && linkcount != linkEntry->numLinks)
                {
                    fs_eek_print_status(context, FALSE, 
                        "MISMATCHED num links for parent I#%lld ino I#%lld, EXPECTED %d, GOT %d",
                        linkEntry->parent, linkEntry->inum,
                        linkcount, linkEntry->numLinks);
                    writeBuf = fixIt;
                    context->eek.errcount++;
                }
            }

            if (writeBuf)
            {
                if (fs_eek_query(context, "Fix?"))
                {
                    linkEntry->numLinks = linkcount;
                    if (linkcount == 0)
                    {
                        linkEntry->inum = linkEntry->parent = 0;
                    }
                    FS_BUF_WRITE(buf,
                                (int32)((char*)(&linkEntry->numLinks) - (char*)(buf->u.data)),
                                sizeof (linkEntry->numLinks));
                    wasModified = TRUE;
                }
            }

            if ((offset < freeOffset) && (linkEntry->numLinks == 0))
            {
                freeOffset = offset;
            }
            linkEntry++;
            i++;
        }
        offset += FS_BLOCK_SIZE;
        if (!fs_bufWasModified(context, buf))
            fs_bufReleaseExclusive(context, buf);
    }
    if (wasModified)
    {
        linkTable->dinode.mtime = fs_getTimeOfDay();
        fs_inodeWrite(context, linkTable, INO_WRITE_LOG);
    }
    else
    {
        fs_inodeReleaseExclusive(context, linkTable);
    }

    // check for missing entries
    fs_eek_scavenge_parent_linkcounts(context, snapnum, fixIt);
}

#endif  //SSC
