// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-log.c
 *
 * RCS:  $Id: fs-log.c,v 1.243 2002/12/10 19:34:28 bob Exp $
 *
 * Description:  Filesystem transaction logging.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"
#include "fs-log-private.h"

/**
 * Names for debugging.
 */
static const char *logRecTypeNames[] = {
    "?",
    "meta1",
    "meta2",
    "user",
    "alloc",
    "zero",
    "commit",
    "remove",
    "bitmap",
    "ftAlloc",
    "ftUsed",
    "ftDealloc",
    "ftFreed",
    "done",
    "inode",
    "inodeFlush"
};


/**
 * Log device specific functions are collected into this
 * table.
 */
fs_LogDevTypeTable fs_logDevTypeTable[FS_LOG_DEVTYPE_MAX] = {
    // NULL log dev
    { 
      0,
      fs_logNullInitDevice,
      fs_logNullCommitWrite,
      fs_logNopCommitTrans,
      fs_logNopSyncInodeBuf,
      fs_logNopSyncInodeLog,
      fs_logNopWriteDevHdr,
      fs_logNopRecordCksum,
      fs_logNullWrite,
      fs_logNullReadInit,
      fs_logNullRead,
      fs_logNullReadDone,
      fs_logNopVerifyCksum,
      fs_logNopIsEmpty,
    },
    // NVRAM log dev
    { 
      FS_LOG_ID_BYTES,
      fs_logNvramInitDevice,
      fs_logNvramCommitWrite,
      fs_logNopCommitTrans,
      fs_logNopSyncInodeBuf,
      fs_logNopSyncInodeLog,
      fs_logNvramWriteDevHdr,
      fs_logNvramRecordCksum,
      fs_logNvramWrite,
      fs_logNvramReadInit,
      fs_logNvramRead,
      fs_logNvramReadDone,
      fs_logNvramVerifyCksum,
      fs_logNvramIsEmpty,
    },
    // Volume same as filesystem
    { 
      0,
      fs_logVolInitDevice,
      fs_logVolCommitWrite,
      fs_logVolCommitTrans,
      fs_logVolSyncInodeBuf,
      fs_logVolSyncInodeLog,
      fs_logVolWriteDevHdr,
      fs_logNopRecordCksum,
      fs_logVolWrite,
      fs_logVolReadInit,
      fs_logVolRead,
      fs_logVolReadDone,
      fs_logNopVerifyCksum,
      fs_logVolIsEmpty,
    },
    // Volume separate from filesystem
    { 
      0,
      fs_logVolInitDevice,
      fs_logVolCommitWrite,
      fs_logVolCommitTrans,
      fs_logVolSyncInodeBuf,
      fs_logVolSyncInodeLog,
      fs_logVolWriteDevHdr,
      fs_logNopRecordCksum,
      fs_logVolWrite,
      fs_logVolReadInit,
      fs_logVolRead,
      fs_logVolReadDone,
      fs_logNopVerifyCksum,
      fs_logVolIsEmpty,
    },
};
    

/**
 * Find/allocate and initialize a log device.
 */
int32
fs_logInitDevice(
    fs_Context      *context)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;

    return devTypeTable->initDevice(context);
}


/**
 * Allocate a log transaction id.
 */
static fs_LogId
fs_logAllocId(
    fs_Context      *context)
{
    fs_FileSystem   *fileSystem = context->fileSystem;
    uint32          id;

    NFX_SPIN_LOCK(&fileSystem->logRoot->spin);
    fileSystem->logRoot->nextId += 1;
    if (fileSystem->logRoot->nextId >= FS_LOG_ID_MAX)
    {
        fileSystem->logRoot->nextId = FS_LOG_ID_MIN;
    }
    id = fileSystem->logRoot->nextId;
    NFX_SPIN_UNLOCK(&fileSystem->logRoot->spin);

    return id;

} /* fs_logAllocId */


/**
 * Init the LockedObjects struct.
 */
static inline void
fs_initLockedObjects(
    fs_LogLockedObjects  *objs)
{
    int32   i;

    objs->next = NULL;
    objs->last = NULL;

    for (i = 0; i < FS_LOG_OBJ_ARR_MAX; i++)
    {
        objs->obj[i] = NULL;
    }
}

/**
 * Create an in-core log summary record for the specified context.
 */
static fs_LogSummary *
fs_makeLogSummary(
    fs_Context *context)
{
    fs_LogSummary *summary;

    summary = fs_ramAlloc(FS_RAM_LOG_SUMM, sizeof(fs_LogSummary));
    summary->id = fs_logAllocId(context);
    summary->nextRecNum  = 0;
    summary->bytesUsed  = 0;
    summary->blksUsed = 0;
    summary->inodesUsed = 0;
    fs_initLockedObjects(&summary->bufArr);
    summary->bufCount = 0;
    fs_initLockedObjects(&summary->inodeArr);
    summary->inodeCount = 0;
    fs_initLockedObjects(&summary->btreeArr);
    summary->btreeCount = 0;
    DLL_INIT(&summary->undoHdr);
    summary->undoCount = 0;
    summary->outerSummary = NULL;
    DLL_ISOLATE_ELEM(summary, hashEntry);

    return summary;
}


/**
 * Start a transaction log in the specified context, which must not have
 * a current log summary.
 */
void 
fs_logBeginTransaction(
    fs_Context *context)
{
    assert(context->logSummary == NULL);
    context->logSummary = fs_makeLogSummary(context);
}


/**
 * Start a nested transaction log in the specified context.
 *
 * For now, assumes that the context has an existing log summary; that is,
 * a transaction is already in progress.  We could also make this work to
 * handle either case by just removing an assertion, but it seems to be
 * worth checking.
 * In case of adding a "transaction done" entry, no log sum was in progress
 */
void 
fs_logNestTransaction(
    fs_Context *context)
{
    fs_LogSummary *summary = fs_makeLogSummary(context);
    // assert(context->logSummary != NULL);
    summary->outerSummary = context->logSummary;
    context->logSummary = summary;
}


/**
 * Indicate whether or not log size is above threshold
 * and writeback engine should be active.
 */
boolean
fs_logAboveWritebackThreshold(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    fs_LogFsRoot    *logRoot = fileSystem->logRoot;

    if (logRoot->currLen >
         (FS_LOG_LOOP_BYTES(fileSystem->logSize) *
                fs_incoreTune->logPctUseLoWtrMk / 100 ))
    {
        return TRUE;
    }

    return FALSE;
}


/**
 * Return the log segment number that the current head
 * of the log resides in.
 */
int32
fs_logGetHeadSegnum(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    return fileSystem->logRoot->head / fileSystem->logRoot->segmentSize;
}


/**
 * Return the size of the specified log record type.
 */
uint32
fs_logRecSize(
    fs_Context *context,
    int recType)
{
    uint32 size;
    switch (recType)
    {
    case FS_LOG_REC_COMMIT:
        size = 0;
        break;
    case FS_LOG_REC_META_ONE:
        size = sizeof (fs_LogRecMetaOne);
        break;
    case FS_LOG_REC_META_TWO:
        size = sizeof (fs_LogRecMetaTwo);
        break;
    case FS_LOG_REC_USER:
        size = sizeof (fs_LogRecUser);
        break;
    case FS_LOG_REC_ALLOC:
        size = sizeof (fs_LogRecAlloc);
        break;
#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
        size = sizeof (fs_LogRecMetaZero);
        break;
    case FS_LOG_REC_USER_ZERO:
        size = sizeof (fs_LogRecUserZero);
        break;
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
        size = sizeof (fs_LogRecZero);
        break;
#endif // NEW_ZF
    case FS_LOG_REC_BITMAP:
        size = sizeof (fs_LogRecBitmap);
        break;
    case FS_LOG_REC_INODE:
        size = sizeof (fs_LogRecInode);
        break;
    case FS_LOG_REC_INODE_FLUSH:
        size = sizeof (fs_LogRecInodeFlush);
        break;
    case FS_LOG_REC_BUF_FLUSH:
        size = sizeof (fs_LogRecBufFlush);
        break;
    case FS_LOG_REC_MARK_DONE:
        size = sizeof (fs_LogRecMarkDone);
        break;
    default:
        LOG_TRACE(("bad log record type %d", recType));
        return 0;
    }

    size += sizeof (fs_LogRecHdr);
    return size;

} /* fs_logRecSize */

/**
 * This assertion checks that the size computed above is correct.  It
 * would probably be simpler to write "size += offsetof(fs_LogRecord, u)"
 * but that's rather cryptic.
 */
NFX_ASSERT_STATIC(sizeof (fs_LogRecHdr) == offsetof(fs_LogRecord, u));


/**
 * Called to get size of a fs_LogRecord
 */
int32
fs_logGetRecLen(
    fs_Context      *context,
    fs_LogRecord    *record)
{
    uint32  recLen = 0;

    switch (record->hdr.recType)
    {
        // fixed length fields
        case FS_LOG_REC_USER:
        case FS_LOG_REC_ALLOC:
#ifdef NEW_ZF
        case FS_LOG_REC_META_ZERO:
        case FS_LOG_REC_USER_ZERO:
#else // NEW_ZF
        case FS_LOG_REC_ZERO_FILL:
#endif // NEW_ZF
        case FS_LOG_REC_BITMAP:
        case FS_LOG_REC_MARK_DONE:
        case FS_LOG_REC_INODE:
        case FS_LOG_REC_INODE_FLUSH:
        case FS_LOG_REC_BUF_FLUSH:
            recLen = fs_logRecSize(context, record->hdr.recType);
            break;
        case FS_LOG_REC_COMMIT:
            recLen = sizeof (fs_LogRecHdr);
            break;

        // variable length fields

        case FS_LOG_REC_META_ONE:
            recLen = (alignto(record->u.metaOne.length, FS_LOG_ALIGN_64)
                        + sizeof (fs_LogRecMetaOne)
                        + sizeof (fs_LogRecHdr));
            break;
        case FS_LOG_REC_META_TWO:
            recLen = (alignto(record->u.metaTwo.length, FS_LOG_ALIGN_64)
                        + sizeof (fs_LogRecMetaTwo)
                        + sizeof (fs_LogRecHdr));
            break;

        default:
            break;
    }

    return recLen;

} /* fs_logGetRecLen */

/**
 * Perform truncation.  SegNum is at the head and its
 * count has reached 0.
 */
static void
fs_logTruncateSegment(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    int32           segNum)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;
    fs_LogSegment       *seg = &logRoot->logSeg[segNum];
    int32               segsTrunc = 0;

    //printf("truncate called with segNum %d\n", segNum);

    FS_ASSERT(logRoot->logSeg[segNum].head == logRoot->head);
    FS_ASSERT(seg->active);

    NFX_SPIN_LOCK(&logRoot->spin);
    while (!(logRoot->tail / logRoot->segmentSize == segNum) &&
           seg->depCount == 0 &&
           seg->undoCount == 0 && 
           seg->active)
    {
        FS_ASSERT(seg->active);
        seg->head = logRoot->segmentSize * segNum;
	//printf("set1 seg %p segnum %d head to %d\n",
        //       seg, segNum, seg->head);
        FS_ASSERT(seg->head != 0 || segNum == 0);
        seg->active = FALSE;
        segsTrunc += 1;
	//printf("truncate deactivated segNum %d\n", segNum);

        segNum = (segNum + 1) % FS_LOG_MAX_SEGMENTS;
        seg = &logRoot->logSeg[segNum];
    }

    LOG_TRACE(("stop truncate seg %d dep %d undo %d active %d seghead %d",
	    segNum, seg->depCount, seg->undoCount, seg->active, seg->head));

#ifdef FS_DEBUG
    if (logRoot->tail == (logRoot->segmentSize * segNum))
    {
        // If the tail resides at the beginning of
        // the segment we're stopping at, that segment
        // should not be active, since we guarantee at
        // least a full segment between head and tail
        // (should be caught in fs_writeRecord).
        FS_ASSERT(!seg->active);
    }
    else
    {
        // The log is not empty, so the segment we stopped
        // at must be active.
        FS_ASSERT(seg->active);
    }
#endif

    FS_ASSERT(seg->head != 0 || segNum == 0);

    if (logRoot->tail / logRoot->segmentSize == segNum &&
        seg->depCount == 0 &&
        seg->undoCount == 0)
    {
        logRoot->head = logRoot->tail;
        seg->head = logRoot->tail;
	//printf("set2 seg %p segnum %d head to %d\n",
        //       seg, segNum, seg->head);
        FS_ASSERT(seg->head != 0 || segNum == 0);
    }
    else
    {
        logRoot->head = seg->head;
    }

    if (logRoot->head > logRoot->tail)
    {
        logRoot->currLen = FS_LOG_LOOP_BYTES(context->fileSystem->logSize) -
                                    (logRoot->head - logRoot->tail);
    }
    else 
    {
        logRoot->currLen = logRoot->tail - logRoot->head;
    }
    NFX_SPIN_UNLOCK(&logRoot->spin);

    LOG_TRACE(("truncate %d segments currLen %d head %d tail %d",
			segsTrunc, logRoot->currLen, logRoot->head, logRoot->tail));

    if (segsTrunc > logRoot->logStats.nbrTruncateMax)
    {
        logRoot->logStats.nbrTruncateMax = segsTrunc;
    }

    if ( ! logRoot->inReplay)
    {
        devTypeTable->writeDevHdr(context, FS_LOG_WRITE_HEAD, logRoot->head);
    }
}


/**
 * Decrement the counter for the specified log
 * segment.
 */
static void
fs_logSegmentDecr(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    int32           segNum)
{
    FS_ASSERT(logRoot->logSeg[segNum].depCount > 0);
    FS_ASSERT(logRoot->logSeg[segNum].active);

    //printf("decr dep segNum %d was %d\n", segNum,
    //    logRoot->logSeg[segNum].depCount);

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logSeg[segNum].depCount -= 1;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    if (logRoot->logSeg[segNum].depCount == 0 &&
        logRoot->logSeg[segNum].undoCount == 0)
    {
        if (logRoot->logSeg[segNum].head == logRoot->head)
        {
            fs_logTruncateSegment(context, logRoot, segNum);
        }
    }
}


/**
 * Remove the specified dependencies from the
 * log.
 */
void
fs_logRemoveDeps(
    fs_Context          *context,
    fs_LogSegmentDeps   *logSegDep,
    nfx_spinlock_t      *spinp)
{
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;
    int32           numBytes = FS_LOG_MAX_SEGMENTS / BITS_PER_BYTE;
    int32           i, j, b;
    int32           numDone = 0;

    LOG_TRACE(("removedeps cnt %d from %p",
               logSegDep->count, logSegDep));

    for (i = 0; i < numBytes; i++)
    {
        if (!logSegDep->u.bytes[i])
            continue;

        for (j = 0, b = 1; j < BITS_PER_BYTE; j++, b <<= 1)
        {
            if (logSegDep->u.bytes[i] & b)
            {
                fs_logSegmentDecr(context, logRoot, (i * BITS_PER_BYTE) + j);
                if (spinp)
                    NFX_SPIN_LOCK(spinp);
                logSegDep->u.bytes[i] &= ~b;
                if (spinp)
                    NFX_SPIN_UNLOCK(spinp);
                numDone += 1;
            }
        }
    }

    FS_ASSERT(logSegDep->count == numDone);
    if (spinp)
        NFX_SPIN_LOCK(spinp);
    logSegDep->count = 0;
    if (spinp)
        NFX_SPIN_UNLOCK(spinp);
}



/**
 * Set the segment dependency bit in the specified logSegDep
 * if not already set.  Returns TRUE in didSet if the bit was
 * set and TRUE in wasFirst if it was the first bit set for
 * the object.
 */
static void
fs_logSetSegmentBit(
    fs_Context          *context,
    fs_LogSegmentDeps   *logSegDep,
    int32               segNum,
    boolean             *didSet,
    boolean             *wasFirst)
{
    int32   bytenum, bitnum, bitmask;

    bytenum = segNum / BITS_PER_BYTE;
    bitnum = segNum % BITS_PER_BYTE;
    bitmask = 1 << bitnum;

    *wasFirst = FALSE;
    *didSet = FALSE;

    if (!(logSegDep->u.bytes[bytenum] & bitmask))
    {
        logSegDep->u.bytes[bytenum] |= bitmask;
        logSegDep->count += 1;
        if (logSegDep->count == 1)
        {
            *wasFirst = TRUE;
        }
        *didSet = TRUE;
    }
}


/**
 * Transfer setment dependencies from srcSegDep
 * to dstSegDep.
 */
void
fs_logDoDepTransfer(
    fs_Context          *context,
    fs_LogSegmentDeps   *dst,
    fs_LogSegmentDeps   *src)
{
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;
    int32           numBytes = FS_LOG_MAX_SEGMENTS / BITS_PER_BYTE;
    int32           i, j, b;
    int32           numWasSet = 0;

    // Compare the bits - if both are set, we must
    // decrement the logRoot segment count by one.
    // Otherwise, simply OR the src into the dst,
    // incrementing dst->count if only src is set.

    for (i = 0; i < numBytes; i++)
    {
        if (!src->u.bytes[i])
            continue;

        for (j = 0, b = 1; j < BITS_PER_BYTE; j++, b <<= 1)
        {
            if (src->u.bytes[i] & b)
            {
                if (dst->u.bytes[i] & b)
                {
                    //printf("decr seg %d, set in both during xfer\n",
                    //       (i * BITS_PER_BYTE) + j);
                    fs_logSegmentDecr(context, logRoot,
                                      (i * BITS_PER_BYTE) + j);
                }
                else
                {
                    dst->count += 1;
                    dst->u.bytes[i] |= b;
                }
                src->u.bytes[i] &= ~b;
                numWasSet += 1;
            }
        }
    }

    FS_ASSERT(src->count == numWasSet);
    src->count = 0;
}


/**
 * Return TRUE if seg1 is older than seg2.  If its
 * equal or newer, return FALSE.
 */
boolean
fs_logSegIsOlder(
    fs_LogFsRoot    *logRoot,
    int16           seg1,
    int16           seg2) 
{
    int32 tailSeg, headSeg;

    if (seg1 == -1 || seg2 == -1)
    {
        // Don't perform compare unless both
        // segments are valid.
        return FALSE;
    }

    tailSeg = logRoot->tail / logRoot->segmentSize;
    headSeg = logRoot->head / logRoot->segmentSize;

    if (logRoot->head < logRoot->tail)
    {
        FS_ASSERT(seg1 <= tailSeg &&
                  seg1 >= headSeg);
        FS_ASSERT(seg2 <= tailSeg &&
                  seg2 >= headSeg);
        // Non wrap-around case.
        return (seg1 < seg2);
    }
    else
    {
        FS_ASSERT(seg1 <= tailSeg ||
                  seg1 >= headSeg);
        FS_ASSERT(seg2 <= tailSeg ||
                  seg2 >= headSeg);
        // Wrap-around case.
        if (seg1 >= headSeg)
        {
            if (seg2 >= headSeg)
            {
                return (seg1 < seg2);
            }
            else
            {
                return TRUE;
            }
        }
        else
        {
            if (seg2 >= headSeg)
            {
                return FALSE;
            }
            else
            {
                return (seg1 < seg2);
            }
        }
    }
}


/**
 * Transfer the specified log dependencies to the
 * specified object of the specified type.
 */
void
fs_logTransferDeps(
    fs_Context          *context,
    fs_LogSegmentDeps   *srcSegDep,
    int16               srcHeadSeg,
    void                *objectPtr,
    int32               objectType)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    fs_LogSegmentDeps   *dstSegDep;
    boolean             doHeadSegUpdate = FALSE;

    LOG_TRACE(("transferring %d deps from %p",
                   srcSegDep->count, srcSegDep));

    if (srcSegDep->count == 0)
        return;

    switch(objectType)
    {
    case FS_LOG_DEP_BUF: {
        fs_Buf *buf = objectPtr;
        dstSegDep = &buf->logSegDep;
        LOG_TRACE(("  transfer deps to buf inum %lld page %lld",
                    buf->inum, buf->page));       
        NFX_SPIN_LOCK(&buf->spin);

        if (dstSegDep->count == 0 ||
            fs_logSegIsOlder(logRoot, srcHeadSeg, buf->headLogSeg))
        {
            doHeadSegUpdate = TRUE;
        }
        fs_logDoDepTransfer(context, dstSegDep, srcSegDep);

        NFX_SPIN_UNLOCK(&buf->spin);

        if (doHeadSegUpdate)
        {
            fs_bufSetDirtySegment(context, buf, srcHeadSeg, -1);
        }

        break;
    }
    default:
        fs_error(context,  FS_INTERNAL_ERR,
            "log dep transfer unsupported on object type");
        break;
    }

} /* fs_logTransferDeps */


/**
 * Set the segment dependency bit in the specified object if
 * not already set.  Returns TRUE if the bit was set during
 * this call, FALSE otherwise.
 */
static boolean
fs_logSetSegmentDepInObject(
    fs_Context          *context,
    fs_LogObjectSumm    *objSumm,
    int32               segNum)
{
    fs_LogSegmentDeps   *logSegDep;
    nfx_spinlock_t      *spin = NULL;
    char                desc[FS_LOG_DEP_DESCRIPTION];
    boolean             wasFirst, didSet;
    void                (*setDirtySegFunc)(fs_Context *, void *, int32, int32);
    int16               headLogSeg;

    switch(objSumm->objectType)
    {
    case FS_LOG_DEP_BUF: {
        fs_Buf *buf = objSumm->objectPtr;
        logSegDep = &buf->logSegDep;
        spin = &buf->spin;
        setDirtySegFunc = fs_bufSetDirtySegment;
        headLogSeg = buf->headLogSeg;
        break;
    }
        
    case FS_LOG_DEP_BTREE:
    case FS_LOG_DEP_INODE: {
        fs_Inode *inode = objSumm->objectPtr;
        logSegDep = &inode->logSegDep;
        setDirtySegFunc = fs_inodeSetDirtySegment;
        headLogSeg = inode->headLogSeg;
        break;
    }
        
#ifdef NEW_ZF
    case FS_LOG_DEP_UZERO:
    {
        fs_zfPage * page = objSumm->objectPtr;
        logSegDep = &page->logSegDep;
        //setDirtySegFunc = NULL;
        //headLogSeg = -1;
        break;
    }
#endif // NEW_ZF
        
    default:
        fs_error(context,  FS_INTERNAL_ERR, "log object type unknown: %d",
                 objSumm->objectType);
        UNREACHED();
    } // switch

    if (spin != NULL)
        NFX_SPIN_LOCK(spin);

    fs_logSetSegmentBit(context, logSegDep, segNum, &didSet, &wasFirst);

    if (spin != NULL)
        NFX_SPIN_UNLOCK(spin);

#ifdef NEW_ZF
    if ((objSumm->objectType != FS_LOG_DEP_UZERO) &&
        (wasFirst ||
         fs_logSegIsOlder(context->fileSystem->logRoot, segNum, headLogSeg)))
#else // NEW_ZF
    if (wasFirst ||
        fs_logSegIsOlder(context->fileSystem->logRoot, segNum, headLogSeg))
#endif // NEW_ZF
    {
        setDirtySegFunc(context, objSumm->objectPtr, segNum, objSumm->offset);

        LOG_TRACE(("setSegBit first,  handle %llx %s count %d",
                   (uint64)((uintptr_t)context->smHandle), 
                   fs_logDepDescribe(context, objSumm->objectPtr,
                                     objSumm->objectType, desc, FS_LOG_DEP_DESCRIPTION-1),
                   logSegDep->count));
    }

    return didSet;

} /* fs_logSetSegmentDepInObject */


/**
 * Set the specified segment as active and update
 * its head pointer.
 *
 * !!! Must be called with logRoot->spin held !!!
 */
static void
fs_logActivateSegment(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          offset,
    uint32          endOffset,
    int32           segStart,
    int32           segThis,
    int32           segEnd)
{
    if (!logRoot->logSeg[segThis].active)
    {
        //printf("activating seg %d/%d/%d offset %d endOffset %d\n",
        //       segStart, segThis, segEnd, offset, endOffset);

        logRoot->logSeg[segThis].active = TRUE;

        if (segThis == segStart)
        {
            logRoot->logSeg[segThis].head = offset;
        }
        else
        {
            logRoot->logSeg[segThis].head = endOffset;
        }
    }
    else if (segThis == segStart &&
             offset < logRoot->logSeg[segThis].head) 
    {
        //printf("update head from offset %d/%d/%d offset %d "
        //       "endOffset %d curr head %d\n",
        //    segStart, segThis, segEnd, offset, endOffset,
        //    logRoot->logSeg[segThis].head );

        logRoot->logSeg[segThis].head = offset;
    }

    FS_ASSERT(logRoot->logSeg[segThis].head != 0 ||
              segThis == 0);
}


/**
 * Increment the undoCount in the indicated segment (segThis).
 */
static void
fs_logIncrUndoCount(
    fs_Context          *context,
    fs_LogFsRoot        *logRoot,
    fs_LogObjectSumm    *objSumm,
    int32               segStart,
    int32               segThis,
    int32               segEnd,
    uint32              endOffset)
{
    //printf("incr undo cnt seg %d was %d\n", segThis,
    //  logRoot->logSeg[segThis].undoCount);

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logSeg[segThis].undoCount += 1;

    fs_logActivateSegment(context, logRoot, objSumm->offset, endOffset,
                          segStart, segThis, segEnd);
    NFX_SPIN_UNLOCK(&logRoot->spin);
}


/**
 * Decrement the undoCount in the indicated segment (segThis).
 */
static void
fs_logDecrUndoCount(
    fs_Context          *context,
    fs_LogFsRoot        *logRoot,
    fs_LogObjectSumm    *objSumm,
    int32               segStart,
    int32               segThis,
    int32               segEnd,
    uint32              endOffset)
{
    //printf("decr undo cnt seg %d was %d\n", segThis,
    //       logRoot->logSeg[segThis].undoCount);

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logSeg[segThis].undoCount -= 1;
    NFX_SPIN_UNLOCK(&logRoot->spin);
}


/**
 * Record a log segment dependency per the specified
 * object summary in the indicated segment (segThis).
 */
static void
fs_logSetObjectDependency(
    fs_Context          *context,
    fs_LogFsRoot        *logRoot,
    fs_LogObjectSumm    *objSumm,
    int32               segStart,
    int32               segThis,
    int32               segEnd,
    uint32              endOffset)
{
    boolean didSet;

    didSet = fs_logSetSegmentDepInObject(context, objSumm, segThis);
    if (didSet)
    {
        //printf("incr dep segNum %d was %d\n", segThis,
        //    logRoot->logSeg[segThis].depCount);

        NFX_SPIN_LOCK(&logRoot->spin);
        logRoot->logSeg[segThis].depCount += 1;

        fs_logActivateSegment(context, logRoot, objSumm->offset, endOffset,
                              segStart, segThis, segEnd);
        NFX_SPIN_UNLOCK(&logRoot->spin);
    }
    else
    {
        FS_ASSERT(logRoot->logSeg[segThis].active);
    }
}


/**
 * Perform the specified operation on the range of segments
 * that are covered by the specified objectSummary.
 */
static void
fs_logDoSegmentRange(
    fs_Context          *context,
    fs_LogFsRoot        *logRoot,
    fs_LogObjectSumm    *objSumm,
    void                (*func)(
                            fs_Context          *context,
                            fs_LogFsRoot        *logRoot,
                            fs_LogObjectSumm    *objSumm,
                            int32               segStart,
                            int32               segThis,
                            int32               segEnd,
                            uint32              endOffset))
{
    uint32   segStart = objSumm->offset / logRoot->segmentSize;
    uint32   endOffset;
    uint32   segEnd;
    int32    segRange, i;
    
    endOffset = objSumm->offset;
    FS_LOG_INCR_PTR(context->fileSystem->logSize,
                    endOffset,
                    objSumm->length);
    if (endOffset == 0)
    {
        segEnd = FS_LOG_MAX_SEGMENTS - 1;
    }
    else
    {
        segEnd = (endOffset - 1) / logRoot->segmentSize;
    }

    FS_ASSERT(segStart >= 0 && segStart < FS_LOG_MAX_SEGMENTS);
    FS_ASSERT(segEnd >= 0 && segEnd < FS_LOG_MAX_SEGMENTS);

    if (segStart < segEnd)
        segRange = segEnd - segStart;
    else if (segEnd < segStart)
        segRange = FS_LOG_MAX_SEGMENTS - (segStart - segEnd);
    else
        segRange = 0;

    //printf("doSegRange offset %d endOffset %d seg %d/x/%d %d\n",
    //       objSumm->offset, endOffset, segStart, segEnd, segRange);

    for (i = 0; i <= segRange; i++)
    {
        int32 segThis = (segStart + i) % FS_LOG_MAX_SEGMENTS;
        
        func(context, logRoot, objSumm, segStart, segThis, segEnd, endOffset);
    }
}


/**
 * Returns TRUE if the specified logSegmentDeps
 * bitmap has segment set for the current
 * head of the log.
 */
boolean
fs_logAreDepsAtHeadSegment(
    fs_Context          *context,
    fs_FileSystem       *fileSystem,
    fs_LogSegmentDeps   *segDeps)
{
    fs_LogFsRoot    *logRoot = fileSystem->logRoot;
    uint32          headSeg = logRoot->head / logRoot->segmentSize;
    int32           bytenum, bitnum, bitmask;

    bytenum = headSeg / BITS_PER_BYTE;
    bitnum = headSeg % BITS_PER_BYTE;
    bitmask = 1 << bitnum;

    return (segDeps->u.bytes[bytenum] & bitmask);
}


/**
 * Add an undo record summary to the
 * log summary.
 */
static void
fs_logAddUndoSummary(
    fs_Context          *context,
    fs_LogSummary       *logSummary,
    fs_LogRecord        *logRec,
    fs_LogObjectSumm    *objSumm)
{
    fs_Buf              *buf = objSumm->objectPtr;
    fs_LogUndoSummary   *undo;

    FS_ASSERT(objSumm->objectType = FS_LOG_DEP_BUF);

    undo = fs_ramAlloc(FS_RAM_LOG_UNDO_SUMM, sizeof *undo);
    if (logRec->hdr.recType == FS_LOG_REC_BITMAP)
    {
        undo->type = FS_LOG_UNDO_TYPE_BITMAP;
        memcpy(&undo->u, &logRec->u, sizeof undo->u);
    }
    else
    {
        undo->type = FS_LOG_UNDO_TYPE_NOP;
    }
    undo->objSumm = *objSumm;
    undo->logId = logSummary->id;
    DLL_ISOLATE_ELEM(undo, logEntry);
    DLL_ISOLATE_ELEM(undo, bufEntry);
    DLL_INSERT_TAIL(&logSummary->undoHdr, undo, logEntry);
    logSummary->undoCount++;
    NFX_SPIN_LOCK(&buf->spin);
    DLL_INSERT_TAIL(&buf->undoHdr, undo, bufEntry);
    NFX_SPIN_UNLOCK(&buf->spin);

} /* fs_logAddUndoSummary */


/**
 * Record the specified dependency type in the specified
 * object for the specified log record.
 */
void
fs_logAddDep(
    fs_Context          *context,
    fs_LogRecHdr        *recHdr,
    fs_LogObjectSumm    *objSumm,
    boolean             addUndo)
{
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;

    //printf("logAddDep obj %p offset %d undo %d\n",
    //       objSumm->objectPtr, objSumm->offset, addUndo);

    if (addUndo)
    {
        fs_logAddUndoSummary(context, context->logSummary,
                             (fs_LogRecord *) recHdr, objSumm);
        fs_logDoSegmentRange(context, logRoot, objSumm,
                             fs_logIncrUndoCount);
    }
    else
    {
        fs_logDoSegmentRange(context, logRoot, objSumm,
                             fs_logSetObjectDependency);
    }

} /* fs_logAddDep */


/**
 * Activate the log segment if its not already activated.
 */
static void
fs_logMarkSegmentActive(
    fs_Context          *context,
    fs_LogFsRoot        *logRoot,
    fs_LogObjectSumm    *objSumm,
    int32               segStart,
    int32               segThis,
    int32               segEnd,
    uint32              endOffset)
{
    NFX_SPIN_LOCK(&logRoot->spin);
    fs_logActivateSegment(context, logRoot, objSumm->offset, endOffset,
                          segStart, segThis, segEnd);
    NFX_SPIN_UNLOCK(&logRoot->spin);
}


/**
 * Activate the affected log segments if they are
 * not already activated.
 */
void
fs_logMarkActive(
    fs_Context          *context,
    fs_LogObjectSumm    *objSumm)
{
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;

    fs_logDoSegmentRange(context, logRoot, objSumm,
                         fs_logMarkSegmentActive);

    FS_ASSERT(objSumm->objectType == FS_LOG_DEP_NONE);
    if (logRoot->head == objSumm->offset)
    {
        // We just wrote a non-dependent record to
        // an empty log, truncate it now.
        fs_logTruncateSegment(context, logRoot,
                              objSumm->offset / logRoot->segmentSize);
    }
}


/**
 * Add an object lock to a log summary.  This provides
 * for unlocking the objects during commit.
 */
static void
fs_logAddObjectLock(
    fs_Context          *context,
    fs_LogSummary       *logSummary,
    void                *object,
    fs_LogLockedObjects *objects,
    int                 *objcount,
    fs_LogDepType       objectType)
{
    fs_LogLockedObjects *currObjects;

    FS_ASSERT(logSummary != NULL);

#ifdef FS_DEBUG_OBJECT_LOCK
{
    int32               i;
    // Check for dup first, we use these to perform unlocking
    // during commit, and we wouldn't want to unlock the same
    // one twice!
    //
    currObjects = objects;
    for (i = 0; i < *objcount; i++)
    {
        FS_ASSERT(objects != NULL);

	// When the counter wraps, move onto the next chain
	// of objects.
        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            currObjects = currObjects->next;
        }

        if (object == currObjects->obj[i % FS_LOG_OBJ_ARR_MAX])
        {           
            // Already recorded this object, not allowed.
            FS_ELOG(context, LOG_ERROR, FS_TRACE_LOG,
                    ("dup object id %d object %p objectType %d", 
                     logSummary->id, object, objectType));
            FS_ASSERT(FALSE);
            return;
        }
    }
    FS_ASSERT(currObjects->next == NULL);
}
#endif

    // We should be using the last array in the list.
    if (objects->last == NULL)
    {
        currObjects = objects;
    }
    else
    {
        currObjects = objects->last;
    }

    // Check for space.
    if (*objcount > 0 && (*objcount % FS_LOG_OBJ_ARR_MAX) == 0)
    {
        // Need to allocate additional space.
        FS_ASSERT(currObjects->next == NULL);
        currObjects->next = fs_ramAlloc(FS_RAM_LOG_BUF_ARR, sizeof *objects);
        bzero(currObjects->next, sizeof *objects);
        currObjects = currObjects->next;
        objects->last = currObjects;
    }   

    currObjects->obj[*objcount % FS_LOG_OBJ_ARR_MAX] = object;
    ++*objcount;

} /* fs_logAddObjectLock */


/**
 * Add an inode lock to the log summary for
 * the transaction.  This provides for unlocking the inode
 * during transaction commit.
 */
void
fs_logAddInodeLock(
    fs_Context      *context,
    fs_LogSummary   *logSummary,
    fs_Inode        *inode)
{
    LOG_TRACE(("id %d addinum %lld",
        logSummary->id, inode->inum));

    fs_logAddObjectLock(context, logSummary, inode, &logSummary->inodeArr,
                        &logSummary->inodeCount, FS_LOG_DEP_INODE);
    fs_lockSetDirty(context, &inode->lock);

} /* fs_logAddInodeLock */


/**
 * Add a buf lock to a log summary.  This provides
 * for unlocking the bufs during commit.
 */
void
fs_logAddBufLock(
    fs_Context      *context,
    fs_LogSummary   *logSummary,
    fs_Buf          *buf)
{
    LOG_TRACE(("id %d addbuf inum %lld page %lld blk %lld flag %#x",
                logSummary->id, buf->inum, buf->page,
                fs_blockNumber(buf->block), buf->flags));

    fs_logAddObjectLock(context, logSummary, buf, &logSummary->bufArr,
                        &logSummary->bufCount, FS_LOG_DEP_BUF);
    fs_lockSetDirty(context, &buf->lock);

} /* fs_logAddBufLock */


/**
 * Add a btree lock to a log summary.  This provides
 * for unlocking the btrees during commit.
 */
void
fs_logAddBtreeLock(
    fs_Context      *context,
    fs_LogSummary   *logSummary,
    fs_Inode        *inode)
{
    LOG_TRACE(("id %d addbtree %lld",
        logSummary->id, inode->inum));
        
    fs_logAddObjectLock(context, logSummary, inode, &logSummary->btreeArr,
                        &logSummary->btreeCount, FS_LOG_DEP_BTREE);
    fs_lockSetDirty(context, &inode->btreeLock);

} /* fs_logAddBtreeLock */

/**
 * Describe the object containing a dependency set.
 */
char *fs_logDepDescribe(
    fs_Context  *context,
    void        *objectPtr,
    int32       objectType,
    char        *string,
    int32       remain)
{
    char *p = string;
    int32      this;

    if (objectPtr == NULL)
        return "";

    switch (objectType) {

        case FS_LOG_DEP_BUF: {
            fs_Buf *buf = objectPtr;
            this = snprintf(p, remain, "buffer");
            p += this;
            remain -= this;
            if (fs_blockIsValid(buf->block))
                this = snprintf(p, remain, " block %lld", fs_blockNumber(buf->block));
            if (buf->inum != 0)
                this = snprintf(p, remain, " inode %lld page %lld", buf->inum, buf->page);
            p += this;
            remain -= this;
            break;
        }

        case FS_LOG_DEP_BTREE:
        case FS_LOG_DEP_INODE: {
            fs_Inode *inode = objectPtr;
            this = snprintf(p, remain, "inode %lld", inode->inum);
            p += this;
            remain -= this;
            break;
        }

        case FS_LOG_DEP_NONE:
            this = snprintf(p, remain, "none");
            p += this;
            remain -= this;
            break;

        default:
            fs_error(context, FS_INTERNAL_ERR, "unexpexted log dep type: %d",
                     objectType);
            UNREACHED();
    }

    return string;

} /* fs_logDepDescribe */


/**
 * Dump the data portion of a log record for debugging.
 */
static uint32
fs_logDumpData(
    void    *datap,
    int32   length,
    char    *data,
    int32   remain)
{
    int32   i;
    uchar8  *charp;
    char    *p = data;
    int32   this;

#define FS_LOG_MAX_LOG_DUMP_DATA 64

    if (length > FS_LOG_MAX_LOG_DUMP_DATA)
        length = FS_LOG_MAX_LOG_DUMP_DATA;

    charp = datap;
    this = snprintf(p, remain, "%08d: %02x", 0, *(charp + 0));
    p += this;
    remain -= this;
    for (i = 1; i < length; i += 1)
    {
        if (i % 16 == 0)
        {
            this = snprintf(p, remain, "\n%08d: ", i);
            p += this;
            remain -= this;
        }
        else if (i % 4 == 0)
        {
            this = snprintf(p, remain, " ");
            p += this;
            remain -= this;
        }

        this = snprintf(p, remain, "%02x", *(charp + i));
        p += this;
        remain -= this;
    }
    if (length >= FS_LOG_MAX_LOG_DUMP_DATA)
        this = snprintf(p, remain, " ...\n");
    else
        this = snprintf(p, remain, "\n");
    p += this;
    remain -= this;
    
    return (p - data);

} /* fs_logDumpData */


/**
 * Dump the contents of a log record for debugging.
 * Return the new offset or -1 if FS_LOG_ID_EOL
 * Called with offset 0, will give the first log record.
 */
static uint32
fs_logDumpOneRecord(
    char            *str,
    int32           remain,
    fs_Context      *context,
    fs_LogRecord    *record,
    int32           *recordLen)
{
    uint32              recLen;
    char                *s = str;
    int32               this;

    this = snprintf(s, remain, " hdr.id %d\n", record->hdr.id);
    s += this;
    remain -= this;
    this = snprintf(s, remain, " hdr.type %d\n", record->hdr.recType);
    s += this;
    remain -= this;
    this = snprintf(s, remain, " hdr.flags 0x%x\n", record->hdr.flags);
    s += this;
    remain -= this;

    switch(record->hdr.recType)
    {
    case FS_LOG_REC_META_ONE:
        this = snprintf(s, remain, "META_ONE type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "blkId inum %lld page %lld block %lld\n",
               record->u.metaOne.blkId.inum, record->u.metaOne.blkId.page,
               fs_blockNumber(record->u.metaOne.blkId.block));
        s += this;
        remain -= this;
        this = snprintf(s, remain, "offset %d\n", record->u.metaOne.offset);
        s += this;
        remain -= this;
        this = snprintf(s, remain, "length %d\n", record->u.metaOne.length);
        s += this;
        remain -= this;
        this = fs_logDumpData(record->u.metaOne.data, record->u.metaOne.length, s, remain);
        s += this;
        remain -= this;
        recLen = alignto(record->u.metaOne.length, FS_LOG_ALIGN_64)
            + sizeof(fs_LogRecMetaOne);
        break;
    case FS_LOG_REC_META_TWO:
        this = snprintf(s, remain, "META_TWO type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "srcBlkNum %lld\n", record->u.metaTwo.srcBlkNum);
        s += this;
        remain -= this;
        this = snprintf(s, remain, "dstBlkId inum %lld page %lld block %lld\n",
               record->u.metaTwo.dstBlkId.inum, record->u.metaTwo.dstBlkId.page,
               fs_blockNumber(record->u.metaTwo.dstBlkId.block));
        s += this;
        remain -= this;
        this = snprintf(s, remain, "offset %d\n", record->u.metaTwo.offset);
        s += this;
        remain -= this;
        this = snprintf(s, remain, "length %d\n", record->u.metaTwo.length);
        s += this;
        remain -= this;
        this = fs_logDumpData(record->u.metaTwo.data, record->u.metaTwo.length, s, remain);
        s += this;
        remain -= this;
        recLen = alignto(record->u.metaTwo.length, FS_LOG_ALIGN_64)
            + sizeof(fs_LogRecMetaOne);
        break;
    case FS_LOG_REC_USER:
        this = snprintf(s, remain, "USER type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "blkId inum %lld page %lld block %lld\n",
               record->u.user.blkId.inum, record->u.user.blkId.page,
               fs_blockNumber(record->u.user.blkId.block));
        s += this;
        remain -= this;
        this = snprintf(s, remain, "codHandle 0x%llx\n", record->u.user.codHandle);
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecUser);
        break;
    case FS_LOG_REC_ALLOC:
        this = snprintf(s, remain, "ALLOC type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "blkId inum %lld page %lld block %lld\n",
               record->u.alloc.blkId.inum, record->u.alloc.blkId.page,
               fs_blockNumber(record->u.alloc.blkId.block));
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecAlloc);
        break;
#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
        this = snprintf(s, remain, "META_ZERO type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "blkId inum %lld page %lld block %lld\n",
                     record->u.metaZero.blkId.inum,
                     record->u.metaZero.blkId.page,
                     fs_blockNumber(record->u.metaZero.blkId.block));
        s += this;
        remain -= this;
        this = snprintf(s, remain, "offset %d\n", record->u.metaZero.offset);
        s += this;
        remain -= this;
        this = snprintf(s, remain, "length %d\n", record->u.metaZero.length);
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecMetaZero); 
       break;
    case FS_LOG_REC_USER_ZERO:
        this = snprintf(s, remain, "USER_ZERO type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "blkId inum %lld page %lld block %lld\n",
                     record->u.userZero.blkId.inum,
                     record->u.userZero.blkId.page,
                     fs_blockNumber(record->u.userZero.blkId.block));
        s += this;
        remain -= this;
        this = snprintf(s, remain, "offset in 1st page %d\n", record->u.userZero.offset);
        s += this;
        remain -= this;
        this = snprintf(s, remain, "nb page %d\n", record->u.userZero.nbPage);
        s += this;
        remain -= this;
        this = snprintf(s, remain, "action: %s\n", (record->u.userZero.zero?
                                         "zero fill" : "don't zero fill"));
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecUserZero); 
       break;
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
        this = snprintf(s, remain, "ZERO_FILL type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "blkId inum %lld page %lld block %lld\n",
               record->u.zero.blkId.inum, record->u.zero.blkId.page,
               fs_blockNumber(record->u.zero.blkId.block));
        s += this;
        remain -= this;
        this = snprintf(s, remain, "offset %d\n", record->u.zero.offset);
        s += this;
        remain -= this;
        this = snprintf(s, remain, "length %d\n", record->u.zero.length);
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecZero); 
       break;
#endif // NEW_ZF
    case FS_LOG_REC_COMMIT:
        this = snprintf(s, remain, "COMMIT type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "numRecs %d\n", record->hdr.numRecs);
        s += this;
        remain -= this;
        recLen = 0;
        break;
    case FS_LOG_REC_BITMAP:
        this = snprintf(s, remain, "BITMAP type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "blkId inum %lld page %lld block %lld\n",
               record->u.bitmap.blkId.inum, record->u.bitmap.blkId.page,
               fs_blockNumber(record->u.bitmap.blkId.block));
        s += this;
        remain -= this;
        this = snprintf(s, remain, "offset %d, length %d, %s\n", 
                        record->u.bitmap.offset, record->u.bitmap.length,
                        ((record->u.bitmap.op == FS_LOG_REC_BITMAP_SET) ? "SET" : "RESET"));
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecBitmap);
        break;
    case FS_LOG_REC_INODE:
        this = snprintf(s, remain, "INODE type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "inum %lld\n",
                     record->u.inode.inum);
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecInode);
        break;
    case FS_LOG_REC_INODE_FLUSH:
        this = snprintf(s, remain, "INODE_FLUSH type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "inum %lld\n",
                     record->u.inodeFlush.inum);
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecInodeFlush);
        break;
    case FS_LOG_REC_BUF_FLUSH:
        this = snprintf(s, remain, "BUF_FLUSH type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "list %d inum %lld page %lld blknum %lld\n",
                     record->u.bufFlush.bufList,
                     record->u.bufFlush.inum,
                     record->u.bufFlush.startPage,
                     fs_blockNumber(record->u.bufFlush.block));
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecBufFlush);
        break;
    case FS_LOG_REC_MARK_DONE:
        this = snprintf(s, remain, "MARK DONE type\n");
        s += this;
        remain -= this;
        this = snprintf(s, remain, "Mark Transaction %d done\n", record->u.markDone.logId);
        s += this;
        remain -= this;
        recLen = sizeof(fs_LogRecMarkDone);
        break;
    default:
        this = snprintf(s, remain, "UNKNOWN type\n");
        s += this;
        remain -= this;
        recLen = 0;
        break;
    }

    this = snprintf(s, remain, "\n");
    s += this;
    remain -= this;

    recLen += sizeof(fs_LogRecHdr);

    *recordLen = recLen;
    
    return (s - str);
}

static uint32
fs_logDumpLogRecord(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          *offset,
    char            *data,
    int32           remain,
    int32           criteria,
    uint64          value,
    boolean         *match)
{
    fs_LogRecord        *record;
    uint32              recLen;
    uint32              volSize;
    char                *s = data;
    int32               this;

    *match = FALSE;
    if (criteria == 0)
    {
        *match = TRUE;
    }

    if (*offset == -1)
    {
        s += snprintf(s, remain, "EOL\n");
        return s - data;
    }
    
    if ( ! *offset)
    {
        *offset = logRoot->head;
    }
    
    switch (logRoot->logDevType) 
    {
    case FS_LOG_DEVTYPE_NVRAM:
        if ((logRoot->logDev.nvHandle == 0) || (logRoot->logDev.nvHandle == NVRAM_INVALID_HANDLE))
        {
            s += snprintf(s, remain, "No nvram found for VolId %lld\n", context->volId.whole_thing);
            return s - data;
        }

        record = fs_ramAlloc(FS_RAM_LOG_READ_BUF, NVRAM_RW_MAX);
        fs_logReadNvram (logRoot, (uchar8 *)record, *offset, NVRAM_RW_MAX);
        break;
    case FS_LOG_DEVTYPE_SELF:
    case FS_LOG_DEVTYPE_VOLID:
        // No need to read data
        volSize = sizeof(fs_LogRecord);

        if (logRoot->tail >= *offset)
            volSize = MIN(volSize, logRoot->tail - *offset);
        else
            volSize = MIN(volSize, (logRoot->tail + context->fileSystem->logSize - *offset));
        
        record = (fs_LogRecord*)fs_logReadVol (context, logRoot, *offset, &volSize);
        break;
    case FS_LOG_DEVTYPE_NULL:
        s += snprintf(s, remain, "No LogRecord FS_LOG_DEVTYPE_NULL\n");
        return s - data;
    default:
        fs_error(context, FS_LOG_BAD_DEV_ERR, "Not supported device type");
        UNREACHED();
    }


    this = snprintf(s, remain, "Log: currLen  %d head %d tail %d\n", 
                   logRoot->currLen, logRoot->head, logRoot->tail);
    s += this;
    remain -= this;

    switch (logRoot->logDevType) 
    {
    case FS_LOG_DEVTYPE_NVRAM:
        if (record->hdr.id == FS_LOG_ID_EOL)
        {
            s += snprintf(s, remain, "LogRecord at offset %d : EOL \n", *offset);
            *offset = -1;
            fs_ramDealloc(FS_RAM_LOG_READ_BUF, record);
            return s - data;
        }
        break;
    case FS_LOG_DEVTYPE_SELF:
    case FS_LOG_DEVTYPE_VOLID:
        if ((record == NULL) || (volSize == 0))
        {
            s += snprintf(s, remain, "LogRecord at offset %d : EOL \n", *offset);
            *offset = -1;
            return s - data;
        }
        break;
    }

    this = snprintf(s, remain, "LogRecord at offset %d\n", *offset);
    s += this;
    remain -= this;

    this = fs_logDumpOneRecord(s, remain, context, record, &recLen);
    s += this;
    remain -= this;
    
    switch (logRoot->logDevType) 
    {
    case FS_LOG_DEVTYPE_NVRAM:
        fs_ramDealloc(FS_RAM_LOG_READ_BUF, record);
        break;
    case FS_LOG_DEVTYPE_SELF:
    case FS_LOG_DEVTYPE_VOLID:
        if (recLen > volSize)
        {
            volSize = recLen;
            record = (fs_LogRecord*)fs_logReadVol (context, logRoot, *offset, &volSize);
            if (recLen > volSize)
            {
                this = snprintf(s, remain, "LogRecord too short %d > %d \n", recLen, volSize);
                s += this;
                remain -= this;
            }
        }
        break;
    }

    switch (criteria)
    {
    case 1: // transId
        if (record->hdr.id == value)
            *match = TRUE;
        break;
    case 2: // blocknbr
        switch(record->hdr.recType)
        {
        case FS_LOG_REC_META_ONE:
            if (fs_blockNumber(record->u.metaOne.blkId.block) == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_META_TWO:
            if (fs_blockNumber(record->u.metaTwo.dstBlkId.block) == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_USER:
            if (fs_blockNumber(record->u.user.blkId.block) == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_ALLOC:
            if (fs_blockNumber(record->u.alloc.blkId.block) == value)
                *match = TRUE;
            break;
#ifdef NEW_ZF
        case FS_LOG_REC_META_ZERO:
            if (fs_blockNumber(record->u.metaZero.blkId.block) == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_USER_ZERO:
            if (fs_blockNumber(record->u.userZero.blkId.block) == value)
                *match = TRUE;
            break;
#else // NEW_ZF
        case FS_LOG_REC_ZERO_FILL:
            if (fs_blockNumber(record->u.zero.blkId.block) == value)
                *match = TRUE;
            break;
#endif // NEW_ZF
        case FS_LOG_REC_BITMAP:
            if (fs_blockNumber(record->u.bitmap.blkId.block) == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_BUF_FLUSH:
            if (fs_blockNumber(record->u.bufFlush.block) == value)
                *match = TRUE;
        }
        break;
    case 3: // inum
        switch(record->hdr.recType)
        {
        case FS_LOG_REC_META_ONE:
            if (record->u.metaOne.blkId.inum == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_META_TWO:
            if (record->u.metaTwo.dstBlkId.inum == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_USER:
            if (record->u.user.blkId.inum == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_ALLOC:
            if (record->u.alloc.blkId.inum == value)
                *match = TRUE;
            break;
#ifdef NEW_ZF
        case FS_LOG_REC_META_ZERO:
            if (record->u.metaZero.blkId.inum == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_USER_ZERO:
            if (record->u.userZero.blkId.inum == value)
                *match = TRUE;
            break;
#else // NEW_ZF
        case FS_LOG_REC_ZERO_FILL:
            if (record->u.zero.blkId.inum == value)
                *match = TRUE;
            break;
#endif // NEW_ZF
        case FS_LOG_REC_BITMAP:
            if (record->u.bitmap.blkId.inum == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_INODE:
            if (record->u.inode.inum == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_INODE_FLUSH:
            if (record->u.inodeFlush.inum == value)
                *match = TRUE;
            break;
        case FS_LOG_REC_BUF_FLUSH:
            if (record->u.bufFlush.inum == value)
                *match = TRUE;
        }
        break;
    case 99: // find valid record
        if ((record->hdr.id >= FS_LOG_ID_MIN) &&
            (record->hdr.id < FS_LOG_ID_MAX) &&
            (record->hdr.recType >= FS_LOG_REC_META_ONE) &&
            (record->hdr.recType < FS_LOG_REC_TYPE_MAX) &&
            (recLen <= (FS_BLOCK_SIZE + sizeof(fs_LogRecord))))
        {
            *match = TRUE;
        }
        else
        {
            recLen = 8;
        }
        
        break;
    }

        
    if (recLen > FS_BLOCK_SIZE + sizeof(fs_LogRecord))
    {
        this = snprintf(s, remain, "Insane recLen %d, terminating scan\n", recLen);
        s += this;
        remain -= this;
        *offset = -1;
    }
    else
    {
        *offset += recLen ;
    }
    
    return s - data;

} /* fs_logDumpLogRecord */


/**
 * Transition the undo record to a segment dependency.
 */
void
fs_logConvertUndoToPerm(
    fs_Context          *context,
    fs_LogUndoSummary   *undo,
    fs_LogSegmentDeps   *logSegDep)
{
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;

    //printf("convert undo offset %d\n", undo->objSumm.offset);

    fs_logDoSegmentRange(context, logRoot, &undo->objSumm,
                         fs_logSetObjectDependency);

    fs_logDoSegmentRange(context, logRoot, &undo->objSumm,
                         fs_logDecrUndoCount);
}


/**
 * Perform undo operations listed in listBuf
 * on data pointed to by buf.
 *
 * Must be called with buf->spin held.
 */
void
fs_logUndoBuf(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_Buf      *listBuf)
{
    fs_LogUndoSummary   *undo;
    int32               i;

    LOG_TRACE(("undo buf inum %lld page %lld blk %lld flag %#x",
               buf->inum, buf->page, fs_blockNumber(buf->block), buf->flags));

    FS_ASSERT (buf->flags & FS_BUF_TYPE_META);

    DLL_ITERATE(undo, &listBuf->undoHdr, bufEntry)
    {
        switch(undo->type)
        {
        case FS_LOG_UNDO_TYPE_BITMAP:
            for (i = 0; i < undo->u.bitmap.length; i++)
            {
                if (undo->u.bitmap.op == FS_LOG_REC_BITMAP_SET)
                {
                    buf->u.bytes[undo->u.bitmap.offset + i] &= 
                                                    ~undo->u.bitmap.bits[i];
                }
                else
                {
                    buf->u.bytes[undo->u.bitmap.offset + i] |= 
                                                    undo->u.bitmap.bits[i];
                }
            }
            break;
        case FS_LOG_UNDO_TYPE_NOP:
            break;
        default:
            FS_ASSERT(FALSE);
            break;
        }
    }

} /* fs_logUndoBuf */


/**
 * Write a record to the log device.  Returns the offset and length
 * of the written record in objSumm.
 */
static void
fs_logWriteRecord(
    fs_Context          *context,
    fs_LogRecord        *record,    // Ptr to record to add.
    uchar8              *data,      // Ptr to new data for meta mods only.
    int32               dataLen,    // Len of new data for meta mods only.
    fs_LogStableWrite   stable,
    uint64              inum,
    fs_LogObjectSumm    *objSumm)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;
    fs_LogSummary       *logSummary = context->logSummary;
    fs_LogRecHdr        *recHdr = &record->hdr;
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    uchar8              *datap[LOG_IOVEC_MAX];
    int32               lenp[LOG_IOVEC_MAX];
    uint32              offset = 0, length = 0;
    uint32              i;
    uint64              pad;


    NFX_SPIN_LOCK(&logRoot->spin);
    context->logSummary->nextRecNum++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    datap[LOG_IOVEC_HDR] = (uchar8 *) recHdr;
    lenp[LOG_IOVEC_HDR] = fs_logRecSize(context, recHdr->recType);
    datap[LOG_IOVEC_DATA] = data;
    lenp[LOG_IOVEC_DATA] = dataLen;
    datap[LOG_IOVEC_PAD] = (uchar8 *) &pad;
    lenp[LOG_IOVEC_PAD] = FS_LOG_ALIGN_64 - (dataLen & 0x7);
    lenp[LOG_IOVEC_PAD] = lenp[LOG_IOVEC_PAD] == 8 ? 0 : lenp[LOG_IOVEC_PAD];
    lenp[LOG_IOVEC_EOL] = 0;

    for (i = 0; i < LOG_IOVEC_MAX; i++)
	length += lenp[i];

#ifdef ISABELLE // test only
    if ((devTypeTable->initDevice != fs_logNullInitDevice) && (length >= 8192))
    {
        printf("length %d\n", length);
        //FS_ASSERT(length < 8192);
    }
#endif // ISABELLE

    if (logRoot->inReplay)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("writerecord in state %d: rec id %d type %d len %d",
                 context->fileSystem->state, record->hdr.id,
                 record->hdr.recType, length));
	FS_ASSERT(FALSE);
    }
    
    LOG_TRACE(("id %d write %s hdrlen %d datalen %d padlen %d",
         logSummary->id,
         logRecTypeNames[recHdr->recType],
		 lenp[LOG_IOVEC_HDR],
		 lenp[LOG_IOVEC_DATA],
         lenp[LOG_IOVEC_PAD]));

    FS_ASSERT(lenp[LOG_IOVEC_HDR] != 0);
    FS_ASSERT(length < FS_LOG_RECORD_LEN_MAX);

    LOG_TRACE(("  incr %d currLen %d head %d tail %d",
               length, logRoot->currLen, logRoot->head, logRoot->tail));


    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->currLen += length;
    logRoot->logStats.logTotalBytes += length;

    // Confirm space on the log device.  Ensure at least a full
    // segment is always free.
    if (length + logRoot->segmentSize > FS_LOG_FREE_BYTES(logRoot))
    {
#ifdef FS_DEBUG_DUMP_FULL_LOG
        char *data;
        uint32 offset;
        boolean match;
        int32 this, remain = FS_LOGDATA_MAX_SIZE - 1;

        data = fs_ramAlloc(FS_RAM_CORE_MISC, FS_LOGDATA_MAX_SIZE);
        this = snprintf(data, remain, "Record at head of log:\n");
        remain -= this;
        offset = logRoot->head;
        fs_logDumpLogRecord(context, logRoot, &offset, &data[this], remain, 0, 0, &match);
        printf ("%s\n", data);
        fs_ramDealloc(FS_RAM_CORE_MISC, data);
#endif
        NFX_SPIN_UNLOCK(&logRoot->spin);

        fs_warn(context, FS_LOG_FULL_ERR, "fs_logWriteRecord");
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }
    logSummary->bytesUsed += length; // just a tuning aid
    NFX_SPIN_UNLOCK(&logRoot->spin);

    // Wakeup writeback daemons if approaching log dirty
    // threshold.
    if (logRoot->currLen > 
            (FS_LOG_LOOP_BYTES(context->fileSystem->logSize) *
                 fs_incoreTune->logPctUseHiWtrMk / 100))
    {
        if (fs_inodeBlockingLogTruncation(context, context->fileSystem))
        {
            // Wakeup the inode cleaner.  It will signal the
            // meta writeback daemon when its finished.
            fs_inodeSignalCleaner(context, FS_INODE_CLEANER_LOG);
        }
        else
        {
            // Wakeup the meta buffer writeback daemon.
            fs_bufSignalWriteback(context, FS_BUFLIST_META, FS_WB_WAKEUP_LOG);

            // And wakeup the user buffer writeback daemon.
            // to clean the zero fill
            fs_bufSignalWriteback(context, FS_BUFLIST_USER, FS_WB_WAKEUP_LOG);
        }
    }

    // Write to the specified log device.
    (devTypeTable->logWrite)(context,
                             &offset,
                             datap,
                             lenp,
                             stable,
                             inum);

#if defined(FS_DEBUG_CHECK_CRASH)
    if (fs_debugCrashEnabled)
    {
        fs_debugCheckCrash(context);
    }
#endif

    if (objSumm != NULL)
    {
        objSumm->offset = offset;
        objSumm->length = length;
    }

} /* fs_logWriteRecord */


/**
 * Append a log record describing a simple block modification.
 */
int32
fs_logAddDataRecord(
    fs_Context      *context,
    fs_Buf          *buf,       // Buf that was modified.
    int32           offset,     // Byte offset in buf->data where modification
                                // begins (n/a for user (cod) bufs).
    int32           length,     // Length in bytes of modification (n/a
                                // for user (cod) bufs).
    uint64          srcBlkNum,  // Source/old blk number.
    uint32          flags)
{
    fs_LogFsRoot        *logRoot;
    fs_LogRecord        record;
    fs_LogObjectSumm    objSumm;
    uchar8              *datap;
    boolean             addUndo = (flags & FS_LOG_ADD_FLAG_UNDO) ?
                                        TRUE : FALSE;

    FS_ASSERT (context->logSummary);

    if (!context->fileSystem->persistentUserBufCache &&
        (buf->flags & FS_BUF_TYPE_USER))
    {
        LOG_TRACE(
            ("id %d add user rec ignore inum %lld page %lld "
             "blk %lld flags %#x flags %#x",
               context->logSummary->id,
               buf->inum, buf->page, fs_blockNumber(buf->block),
                buf->flags, flags));

        return NFX_OK;
    }

    FS_ASSERT (fs_blockIsValid(buf->block));

    logRoot = (context->fileSystem)->logRoot;

    datap = NULL;
    record.hdr.flags = 0;
    record.hdr.id = context->logSummary->id;

    if (flags & FS_LOG_ADD_FLAG_ZERO)
    {
        LOG_TRACE(
            ("id %d add zero fill inum %lld page %lld blk %lld flags %#x "
             "off %d len %d newblk %lld flags %#x",
               context->logSummary->id,
               buf->inum, buf->page, fs_blockNumber(buf->block), buf->flags,
               offset, length, srcBlkNum, flags));

#ifdef NEW_ZF
        FS_ASSERT(buf->flags & FS_BUF_TYPE_META);

        record.hdr.recType = FS_LOG_REC_META_ZERO;
        record.u.metaZero.offset = offset;
        record.u.metaZero.length = length;
        length = 0;  // No '0' data actually recorded in log.
        record.u.metaZero.blkId.block = buf->block;
        record.u.metaZero.blkId.inum = buf->inum;
        record.u.metaZero.blkId.page = buf->page;
        NFX_SPIN_LOCK(&logRoot->spin);
        logRoot->logStats.nbrZeroMeta++;
        NFX_SPIN_UNLOCK(&logRoot->spin);
#else // NEW_ZF
        record.hdr.recType = FS_LOG_REC_ZERO_FILL;
        record.u.zero.offset = offset;
        record.u.zero.length = length;
        record.u.zero.isMeta = ((buf->flags & FS_BUF_TYPE_META) != 0);
        length = 0;  // No '0' data actually recorded in log.
        record.u.zero.blkId.block = buf->block;
        record.u.zero.blkId.inum = buf->inum;
        record.u.zero.blkId.page = buf->page;
        NFX_SPIN_LOCK(&logRoot->spin);
        if (record.u.zero.isMeta)
            logRoot->logStats.nbrZeroMeta++;
        else
            logRoot->logStats.nbrZeroUser++;
        NFX_SPIN_UNLOCK(&logRoot->spin);
#endif // NEW_ZF
    }
    else if (buf->flags & FS_BUF_TYPE_USER)
    {
	// This is left around for possible future
	// support of an NVRAM log device.  Not currently
	// supported, and very probably broken in various
	// ways. -- bob.
	FS_ASSERT(FALSE);

        LOG_TRACE(("id %d add user rec inum %lld page %lld blk %lld "
                   "flags %#x off %d len %d newblk %lld flags %#x",
                   context->logSummary->id,
                   buf->inum, buf->page, fs_blockNumber(buf->block),
                   buf->flags, offset, length, srcBlkNum, flags));

        // Assumption here is that a codHandle represents
        // a single blkSize block at this time.
        // TODO - validate this assumption!!!
        offset = 0;
        length = 0;
        record.hdr.recType = FS_LOG_REC_USER;
        record.u.user.codHandle = buf->u.user.codHandle;
        record.u.user.codLength = buf->u.user.codLength;
        record.u.user.blkId.block = buf->block;
        record.u.user.blkId.inum = buf->inum;
        record.u.user.blkId.page = buf->page;
        NFX_SPIN_LOCK(&logRoot->spin);
        logRoot->logStats.nbrUser++;
        NFX_SPIN_UNLOCK(&logRoot->spin);
    }
    else
    {
        if (buf->flags & FS_BUF_COPIED)
        {
            LOG_TRACE(("id %d add meta 2 rec inum %lld page %lld blk %lld "
                       "flag %#x off %d len %d newblk %lld flags %#x",
                       context->logSummary->id,
                       buf->inum, buf->page, fs_blockNumber(buf->block),
                       buf->flags, offset, length, srcBlkNum, flags));

            // Buffer is COW, so we're recording changes to it
            // for its new copy.  (Actually the length of the
            // changes can be 0, eg. just a straight copy)
            record.hdr.recType = FS_LOG_REC_META_TWO;
            record.u.metaTwo.srcBlkNum = srcBlkNum;
            record.u.metaTwo.dstBlkId.block = buf->block;
            record.u.metaTwo.dstBlkId.inum = buf->inum;
            record.u.metaTwo.dstBlkId.page = buf->page;
            record.u.metaTwo.offset = offset;
            record.u.metaTwo.length = length;
            NFX_SPIN_LOCK(&logRoot->spin);
            logRoot->logStats.nbrMetaTwo++;
            NFX_SPIN_UNLOCK(&logRoot->spin);
        }
        else
        {
            LOG_TRACE(("id %d add meta 1 rec inum %lld page %lld blk %lld "
                       "flag %#x off %d len %d newblk %lld flags %#x",
                       context->logSummary->id, buf->inum, buf->page,
                       fs_blockNumber(buf->block), buf->flags,
                       offset, length, srcBlkNum, flags));

            // Simple, non-COW metadata buffer modification.
            record.hdr.recType = FS_LOG_REC_META_ONE;
            record.u.metaOne.blkId.block = buf->block;
            record.u.metaOne.blkId.inum = buf->inum;
            record.u.metaOne.blkId.page = buf->page;
            record.u.metaOne.offset = offset;
            record.u.metaOne.length = length;
            NFX_SPIN_LOCK(&logRoot->spin);
            logRoot->logStats.nbrMetaOne++;
            NFX_SPIN_UNLOCK(&logRoot->spin);
        }
        if (length)
        {
            datap = buf->u.data + offset;
        }
    }

    objSumm.objectPtr = buf;
    objSumm.objectType = FS_LOG_DEP_BUF;

    // Write the log record to the log device.
    fs_logWriteRecord(context,
                      &record,
                      datap,
                      length,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    // Add in-core summary for the record.
    fs_logAddDep(context,
                 &record.hdr,
                 &objSumm,
                 addUndo);

    return NFX_OK;

} /* fs_logAddDataRecord */


/**
 * Append a log record describing a block
 * allocation.
 */
int32
fs_logAddAllocRecord(
    fs_Context      *context,
    fs_Buf          *buf,       // Buf that has been allocated for.
    boolean         isCow,      // Is this a COW alloc?
    fs_Block        oldBlock)   // If COW, original src or old block
{
    fs_LogFsRoot        *logRoot;
    fs_LogRecord        record;
    fs_LogObjectSumm    objSumm;
    fs_LogRecAlloc      *alloc = &record.u.alloc;

    FS_ASSERT(buf->inum);
    FS_ASSERT(fs_blockIsValid(buf->block));
    
    if (!context->fileSystem->persistentUserBufCache &&
        (buf->flags & FS_BUF_TYPE_USER))
    {
        LOG_TRACE(
            ("id %d add alloc user rec ignore inum %lld page %lld "
             "blk %lld flags %#x",
               context->logSummary->id, buf->inum, buf->page,
               fs_blockNumber(buf->block), buf->flags));

        return NFX_OK;
    }

    LOG_TRACE(("id %d addallocrec inum %lld page %lld blk %lld flag %#x "
               "isCow %d old %lld",
               context->logSummary->id,
               buf->inum, buf->page, fs_blockNumber(buf->block), buf->flags,
               isCow, fs_blockNumber(oldBlock)));

    record.hdr.id = context->logSummary->id;
    record.hdr.recType = FS_LOG_REC_ALLOC;
    record.hdr.flags = 0;

    alloc->blkId.inum = buf->inum;
    alloc->blkId.page = buf->page;
    alloc->blkId.block = buf->block;
    alloc->isMeta = (buf->flags & FS_BUF_TYPE_META) != 0;
    alloc->isCow = isCow;
    alloc->oldBlkNum = fs_blockNumber(oldBlock);

    logRoot = (context->fileSystem)->logRoot;

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrAlloc++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    objSumm.objectPtr = buf;
    objSumm.objectType = FS_LOG_DEP_BUF;

    // Write the log record to the log device.
    fs_logWriteRecord(context,
                      &record,
                      NULL,
                      0,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    // Add in-core summary for the record.
    fs_logAddDep(context,
                 &record.hdr,
                 &objSumm,
                 FALSE);

    return NFX_OK;

} /* fs_logAddAllocRecord */


/**
 * Append a log record describing a bitmap change.
 */
int32
fs_logAddBitmapRecord(
    fs_Context      *context,
    fs_Buf          *buf,
    int32           offset,
    int32           length,
    uchar8          *bits,
    int32           op,
    fs_BitmapOperation *debugOp)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    fs_LogRecord        record;
    fs_LogObjectSumm    objSumm;

    FS_ASSERT(buf->inum);
    FS_ASSERT(fs_blockIsValid(buf->block));
    FS_ASSERT(length <= FS_LOG_BITMAP_MAX_BYTES);
    FS_ASSERT(buf->flags & FS_BUF_TYPE_META);

    LOG_TRACE(
        ("id %d addbitmap inum %lld page %lld blk %lld flag %#x "
         "offset %d length %d op %s",
             context->logSummary->id,
             buf->inum, buf->page, fs_blockNumber(buf->block), buf->flags,
             offset,
             length,
             op == FS_LOG_REC_BITMAP_SET ? "set" : "reset"));

    record.hdr.id = context->logSummary->id;
    record.hdr.recType = FS_LOG_REC_BITMAP;
    record.hdr.flags = 0;

    record.u.bitmap.blkId.inum = buf->inum;
    record.u.bitmap.blkId.page = buf->page;
    record.u.bitmap.blkId.block = buf->block;
    record.u.bitmap.offset = offset;
    record.u.bitmap.length = length;
    memcpy(&record.u.bitmap.bits, bits, length);
    record.u.bitmap.op = op;
    if (debugOp != NULL)
    {
	if (buf->inum == FS_REF_COUNT_INODE ||
            buf->inum == FS_FREE_BUDDY_TREE_INODE)
	{
		fs_Block startBlock = fs_extentStartBlock(debugOp->u.extent);
		FS_ASSERT(fs_blockNumber(startBlock) >= 0);
		FS_ASSERT(fs_blockNumber(startBlock) + 
			fs_extentLength(debugOp->u.extent) <= FS_MAX_BLOCKS);
	}

	if (buf->inum == FS_INODE_INODE)
	{
		FS_ASSERT(debugOp->u.inum < buf->fileSystem->super->numBlks);
	}

	record.u.bitmap.debugOp.u = debugOp->u;
	record.u.bitmap.debugOp.op = debugOp->op;
    }
    else
    {
	bzero(&record.u.bitmap.debugOp, sizeof(record.u.bitmap.debugOp));
    }

#ifdef FS_DEBUG
{
    int32   i;

    for (i = 0; i < length; i++)
    {
        if (op == FS_LOG_REC_BITMAP_SET)
        {
            FS_ASSERT((buf->u.bytes[offset + i] & record.u.bitmap.bits[i]) ==
                                record.u.bitmap.bits[i]);
        }
        else
        {
            FS_ASSERT((buf->u.bytes[offset + i] & record.u.bitmap.bits[i]) ==
                                0);
        }
    }
}
#endif

    logRoot = (context->fileSystem)->logRoot;

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrBitmap++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    objSumm.objectPtr = buf;
    objSumm.objectType = FS_LOG_DEP_BUF;

    // Write the log record to the log device.
    fs_logWriteRecord(context,
                      &record,
                      NULL,
                      0,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    // Add in-core summary for the record.
    fs_logAddDep(context,
                 &record.hdr,
                 &objSumm,
                 TRUE);

    return NFX_OK;

} /* fs_logAddBitmapRecord */


/**
 * Append a log record for a modified inode.
 */
void
fs_logAddInodeRecord(
    fs_Context  *context,
    fs_Inode    *inode)
{
    fs_LogFsRoot *logRoot = context->fileSystem->logRoot;
    fs_LogRecord record;
    fs_LogObjectSumm objSumm;

    LOG_TRACE(("id %d addInodeRec %lld",
               context->logSummary->id,
               inode->inum));

    record.hdr.id = context->logSummary->id;
    record.hdr.recType = FS_LOG_REC_INODE;
    record.hdr.flags = 0;

    record.u.inode.inum = inode->inum;
    record.u.inode.snapnum = inode->snapnum;
    record.u.inode.dinode = inode->dinode;

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrInode++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    objSumm.objectPtr = inode;
    objSumm.objectType = FS_LOG_DEP_INODE;

    // Write the log record to the log device.
    fs_logWriteRecord(context,
                      &record,
                      NULL,
                      0,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    // Add in-core summary for the record.
    fs_logAddDep(context,
                 &record.hdr,
                 &objSumm,
                 FALSE);

} /* logAddInodeRecord */


/**
 * Create and commit a zero-fill transaction for the
 * specified buf.  This is only called if the log device
 * is on a disk volume.
 */
void
fs_logZeroFillAndCommit(
    fs_Context      *context,
    fs_Buf          *buf)
{
    fs_LogRecord            rec;
    fs_LogFsRoot            *logRoot = context->fileSystem->logRoot;
    fs_LogSummary           *nestedSummary;
    fs_LogObjectSumm        objSumm;

    FS_ASSERT(fs_blockIsValid(buf->block));
#ifdef NEW_ZF
    FS_ASSERT(buf->flags & FS_BUF_TYPE_META);
#endif // NEW_ZF

    fs_logNestTransaction (context);
    nestedSummary = context->logSummary;

    buf->zeroFillLogId = nestedSummary->id;

    LOG_TRACE(
        ("zero-fill id %d buf inum %lld page %lld blk %lld flags %#x", 
        nestedSummary->id, buf->inum, buf->page,
        fs_blockNumber(buf->block), buf->flags));

    rec.hdr.id = nestedSummary->id;
    rec.hdr.flags = 0;
#ifdef NEW_ZF
    rec.hdr.recType = FS_LOG_REC_META_ZERO;
    rec.u.metaZero.offset = 0;
    rec.u.metaZero.blkId.block = buf->block;
    rec.u.metaZero.blkId.inum = buf->inum;
    rec.u.metaZero.blkId.page = buf->page;
    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrZeroMeta++;
    rec.u.metaZero.length = FS_BLOCK_SIZE;
#else // NEW_ZF
    rec.hdr.recType = FS_LOG_REC_ZERO_FILL;
    rec.u.zero.offset = 0;
    rec.u.zero.isMeta = ((buf->flags & FS_BUF_TYPE_META) != 0);
    rec.u.zero.blkId.block = buf->block;
    rec.u.zero.blkId.inum = buf->inum;
    rec.u.zero.blkId.page = buf->page;
    NFX_SPIN_LOCK(&logRoot->spin);
    if (rec.u.zero.isMeta)
    {
        logRoot->logStats.nbrZeroMeta++;
        rec.u.zero.length = FS_BLOCK_SIZE;
    }
    else
    {
        logRoot->logStats.nbrZeroUser++;
        rec.u.zero.length = buf->u.user.codLength;
    }
#endif // NEW_ZF
    NFX_SPIN_UNLOCK(&logRoot->spin);

    objSumm.objectPtr = buf;
    objSumm.objectType = FS_LOG_DEP_BUF;

    // Write the zero-fill log record to the log device.
    fs_logWriteRecord(context,
                      &rec,
                      NULL,
                      0,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    // Add in-core summary for the record.
    fs_logAddDep(context,
                 &rec.hdr,
                 &objSumm,
                 FALSE);

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrCommit++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    vl_addUnstableId(context, &buf->inum);

    rec.hdr.id = nestedSummary->id;
    rec.hdr.recType = FS_LOG_REC_COMMIT;
    rec.hdr.flags = 0;
    rec.hdr.numRecs = nestedSummary->nextRecNum;

    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;

    // Write the commit record to the log device.
    fs_logWriteRecord(context,
                      &rec,
                      NULL,
                      0,
                      FS_LOG_WRITE_COMMIT_UNSTABLE,
                      buf->inum,
                      &objSumm);

    fs_logMarkActive(context, &objSumm);

    // Pop the transaction.
    context->logSummary = nestedSummary->outerSummary;
    fs_ramDealloc(FS_RAM_LOG_SUMM, nestedSummary);

} // fs_logZeroFillAndCommit


/**
 * Create a nested transaction that completes the specified
 * transaction on the log (ie. prevents replay of the
 * specified transaction).
 * 
 * This is only called if the log device is on a disk
 * volume.
 */
void
fs_logComplete(
    fs_Context      *context,
    fs_LogId        transaction,
    uint64          inum)
{
    fs_LogRecord            rec;
    fs_LogFsRoot            *logRoot;
    fs_LogSummary           *nestedSummary;
    fs_LogObjectSumm        objSumm;

    fs_logNestTransaction(context);

    nestedSummary = context->logSummary;
    logRoot = context->fileSystem->logRoot;

    LOG_TRACE(("fs_logComplete trans %d nest id %d", 
            transaction, nestedSummary->id));

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrMarkDone++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    rec.hdr.id = nestedSummary->id;
    rec.hdr.flags = 0;
    rec.hdr.recType = FS_LOG_REC_MARK_DONE;
    rec.u.markDone.logId = transaction;

    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;

    // Write the log record to the log device.
    fs_logWriteRecord(context,
                      &rec,
                      NULL,
                      0,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    fs_logMarkActive(context, &objSumm);

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrCommit++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    vl_addUnstableId(context, &inum);

    rec.hdr.id = nestedSummary->id;
    rec.hdr.recType = FS_LOG_REC_COMMIT;
    rec.hdr.flags = 0;
    rec.hdr.numRecs = nestedSummary->nextRecNum;

    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;

    // Write the commit record to the log device.
    fs_logWriteRecord(context,
                      &rec,
                      NULL,
                      0,
                      FS_LOG_WRITE_COMMIT_UNSTABLE,
                      inum,
                      &objSumm);

    fs_logMarkActive(context, &objSumm);

    // Pop the transaction.
    context->logSummary = nestedSummary->outerSummary;

    fs_ramDealloc(FS_RAM_LOG_SUMM, nestedSummary);

} /* fs_logComplete */


/**
 * Log that an inode was removed.
 */
void
fs_logAddInodeFlushRecord(
    fs_Context  *context,
    fs_Inode    *inode)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    fs_LogRecord        record;
    fs_LogObjectSumm    objSumm;

    LOG_TRACE(("id %d addInodeFlushRec %lld",
                context->logSummary->id,
                inode->inum));

    record.hdr.id = context->logSummary->id;
    record.hdr.recType = FS_LOG_REC_INODE_FLUSH;
    record.hdr.flags = 0;

    record.u.inodeFlush.inum = inode->inum;

    if (FS_LOG_IS_EMPTY(logRoot))
    {
        // Don't add it, since this would be the head
        // entry in the log and would stall the log.
        // (Since its the head entry it isn't necessary
        // anyway).
        return;
    }

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrInodeFlush++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;

    // Write the record to the log device.
    fs_logWriteRecord(context,
                      &record,
                      NULL,
                      0,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    fs_logMarkActive(context, &objSumm);

} /* logAddInodeFlushRecord */


/**
 * Log that one or more bufs were removed.
 */
void
fs_logAddBufFlushRecord(
    fs_Context  *context,
    int32       bufListType,
    uint64      inum,
    uint64      startPage,
    fs_Block    block)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    fs_LogRecord        record;
    fs_LogObjectSumm    objSumm;

    if (!context->fileSystem->persistentUserBufCache &&
        bufListType == FS_BUFLIST_USER)
    {
        LOG_TRACE(
            ("id %d flush user buf rec ignore inum %lld page %lld blk %lld",
               context->logSummary->id,
               inum, startPage, fs_blockNumber(block)));
        return;
    }

    FS_ASSERT(fs_blockIsValid(block));

    LOG_TRACE(("id %d addBufFlushRec type %d inum %lld page %lld bknum %lld",
                context->logSummary->id,
                bufListType,
                inum,
                startPage,
                fs_blockNumber(block)));

    if (FS_LOG_IS_EMPTY(logRoot))
    {
        // Don't add it, since this would be the head
        // entry in the log and would stall the log.
        // (Since its the head entry it isn't necessary
        // anyway).
        return;
    }

    record.hdr.id = context->logSummary->id;
    record.hdr.recType = FS_LOG_REC_BUF_FLUSH;
    record.hdr.flags = 0;

    record.u.bufFlush.bufList = bufListType;
    record.u.bufFlush.inum = inum;
    record.u.bufFlush.startPage = startPage;
    record.u.bufFlush.block = block;

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrBufFlush++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;

    // Write the record to the log device.
    fs_logWriteRecord(context,
                      &record,
                      NULL,
                      0,
                      FS_LOG_WRITE_OTHER,
                      0,
                      &objSumm);

    fs_logMarkActive(context, &objSumm);

} /* logAddBufFlushRecord */

#ifdef NEW_ZF
/**
 * Add a "user zero" log record.
 */
void
fs_logAddUserZeroRecord(fs_Context *    context,
                        fs_Inode *      inode,
                        uint64          startPage,
                        uint32          offsetInFirstPage,
                        uint32          nbPage,
                        boolean         doZero,
                        boolean         doCommit,
                        fs_Buf **       bufList)
{
    fs_LogFsRoot *      logRoot = context->fileSystem->logRoot;
    fs_LogRecord        record;
    fs_LogObjectSumm    objSumm;

    //printf("fs_logAddUserZeroRecord(inum %lld startPage %lld offset %d "
    //"nbPage %d doZero %s)\n", inode->inum, startPage,
    //offsetInFirstPage, nbPage, doZero? "TRUE":"FALSE");

    /* Initialize the user (don't) zero fill record.
     */
    record.hdr.id = context->logSummary->id;
    record.hdr.recType = FS_LOG_REC_USER_ZERO;
    record.hdr.flags = 0;
    
    record.u.userZero.blkId.block = FS_INVALID_BLOCK;
    record.u.userZero.blkId.inum = inode->inum;
    record.u.userZero.blkId.page = startPage;
    record.u.userZero.offset = offsetInFirstPage;
    record.u.userZero.nbPage = nbPage;
    record.u.userZero.zero = doZero;

    NFX_SPIN_LOCK(&logRoot->spin);
    logRoot->logStats.nbrZeroUser++;
    NFX_SPIN_UNLOCK(&logRoot->spin);

    /* Write the record to the log
     * and get objSumm.offset and .length initialized.
     */
    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;
    fs_logWriteRecord(context, &record, NULL, 0,
                      FS_LOG_WRITE_OTHER, 0, &objSumm);

    /* Add "zero fill" dependencies to the new buffers.
     * Note: no dependency for the "don't zero fill" record.
     */
    if (doZero == TRUE)
    {
        uint32 i;

        FS_ASSERT(bufList != NULL);
        objSumm.objectType = FS_LOG_DEP_BUF;

        if (bufList[0]->page != startPage)
        {
            nbPage--;
        }
        
        for (i = 0; i < nbPage; i++)
        {
            objSumm.objectPtr = bufList[i];
            fs_logAddDep(context, &record.hdr, &objSumm, FALSE);
        } // for
    }
    else
    {
        NFX_SPIN_LOCK(&(context->fileSystem->zfSpin));
        fs_zfDelPage(context, inode->inum, startPage, TRUE);
        NFX_SPIN_UNLOCK(&(context->fileSystem->zfSpin));
        
        fs_logMarkActive(context, &objSumm);
    }

    /* Commit the record if requested.
     */
    if (doCommit == TRUE)
    {
        vl_addUnstableId(context, &(inode->inum));
        
        record.hdr.id = context->logSummary->id;
        record.hdr.recType = FS_LOG_REC_COMMIT;
        record.hdr.flags = 0;
        
        objSumm.objectPtr = NULL;
        objSumm.objectType = FS_LOG_DEP_NONE;
        
        // Write the commit record to the log device.
        fs_logWriteRecord(context, &record, NULL, 0,
                          FS_LOG_WRITE_COMMIT_UNSTABLE, inode->inum, &objSumm);
        
        NFX_SPIN_LOCK(&logRoot->spin);
        logRoot->logStats.nbrCommit++;
        NFX_SPIN_UNLOCK(&logRoot->spin);

        fs_logMarkActive(context, &objSumm);
    }

    return;
}
#endif // NEW_ZF

/**
 * Copy modified inodes to the log at the end of a transaction.
 */
static void
fs_logCopyInodes(
    fs_Context      *context,
    fs_LogSummary   *logSummary)
{
    fs_LogLockedObjects     *objects;
    fs_Inode                *inode;
    int32                   i;

    // Copy inodes from the inode array.
    objects = &logSummary->inodeArr;
    for (i = 0; i < logSummary->inodeCount; i++)
    {
        FS_ASSERT(objects != NULL);

        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            objects = objects->next;
        }

        inode = objects->obj[i % FS_LOG_OBJ_ARR_MAX];

        if (inode == NULL)
        {
            // Inode was pending deallocation, its already
            // been handled.
            continue;
        }

        LOG_TRACE(("  copying inode inum %lld %p deps %d",
                 inode->inum, inode, inode->logSegDep.count));

        fs_inodeUpdateLog(context, inode);
    }

    // Copy inodes from the inode btree array.
    objects = &logSummary->btreeArr;
    for (i = 0; i < logSummary->btreeCount; i++)
    {
        FS_ASSERT(objects != NULL);

        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            objects = objects->next;
        }

        inode = objects->obj[i % FS_LOG_OBJ_ARR_MAX];

        if (inode == NULL)
        {
            // Inode was pending deallocation, its already
            // been handled.
            continue;
        }

        LOG_TRACE(("  copying inode btree inum %lld %p deps %d",
                 inode->inum, inode, inode->logSegDep.count));

        fs_inodeUpdateLog(context, inode);
    }

} /* fs_logCopyInodes */


/**
 * Unlock inodes at the end of a transaction.
 */
void
fs_logCommitInodes(
    fs_Context      *context,
    fs_LogSummary   *logSummary)
{
    fs_LogLockedObjects     *objects;
    int32                   i;

    // Unlock inodes and release space held by the
    // inode array.
    objects = &logSummary->inodeArr;
    for (i = 0; i < logSummary->inodeCount; i++)
    {
        fs_Inode    *inode;

        FS_ASSERT(objects != NULL);

        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            fs_LogLockedObjects *prev = objects;
            objects = objects->next;
            prev->next = NULL;
            if (prev != &logSummary->inodeArr)
            {
                fs_ramDealloc(FS_RAM_LOG_BUF_ARR, prev);
            }
        }

        inode = objects->obj[i % FS_LOG_OBJ_ARR_MAX];

        if (inode == NULL)
        {
            // Inode was pending deallocation, its already
            // been handled.
            continue;
        }

        LOG_TRACE(("  commiting inode inum %lld %p deps %d",
                 inode->inum, inode, inode->logSegDep.count));

        fs_inodeCommit(context, inode);
    }
    if (objects != &logSummary->inodeArr)
    {
        fs_ramDealloc(FS_RAM_LOG_BUF_ARR, objects);
    }

} /* fs_logCommitInodes */


/**
 * Comparison function for ordering inodes by their inode number.
 */
static int
fs_logInumCompare(
    const void *i0,
    const void *i1)
{
    const fs_Inode    *inode0 = *(fs_Inode **) i0;
    const fs_Inode    *inode1 = *(fs_Inode **) i1;

    if (inode0->inum < inode1->inum)
        return -1;
    if (inode0->inum > inode1->inum)
        return 1;
    
    FS_ASSERT(FALSE);              // array should not have duplicates
    UNREACHED();
}


/**
 * Complete the deallocation of inodes at the end of a transaction.
 */
static void
fs_logCommitInodeDeallocate(
    fs_Context      *context)
{
    fs_LogLockedObjects     *objects;
    fs_Inode                *sortedInodes[FS_LOG_MAX_INODE_DEALLOC];
    int32                   i, j, numToDo = 0;

    objects = &context->logSummary->inodeArr;
    for (i = 0; i < context->logSummary->inodeCount; i++)
    {
        fs_Inode    *inode;

        FS_ASSERT(objects != NULL);

        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            objects = objects->next;
        }

        inode = objects->obj[i % FS_LOG_OBJ_ARR_MAX];

        if (!(inode->flags & INO_FLAGS_PENDING_DEALLOC))
        {
            continue;
        }

        // Remove it so we wont copy it to the log
        // later.
        objects->obj[i % FS_LOG_OBJ_ARR_MAX] = NULL;

        FS_ASSERT(numToDo < FS_LOG_MAX_INODE_DEALLOC);

        sortedInodes[numToDo++] = inode;
    }

    if (numToDo > 0)
    {
        // Clear them from the btree array if we're holding them.
        objects = &context->logSummary->btreeArr;
        for (i = 0; i < context->logSummary->btreeCount; i++)
        {
            fs_Inode    *inode;
    
            FS_ASSERT(objects != NULL);
    
            if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
            {
                objects = objects->next;
            }
    
            inode = objects->obj[i % FS_LOG_OBJ_ARR_MAX];
    
            for (j = 0; j < numToDo; j++)
            {
                if (inode == sortedInodes[j])
                {
                    objects->obj[i % FS_LOG_OBJ_ARR_MAX] = NULL;
                    fs_inodeBtreeCommit(context, inode);
                    break;
                }
            }
        }

        if (numToDo > 1)
        {
            // Do them in order so lock order on the inode
            // file is maintained.
            qsort(sortedInodes,
                  numToDo,
                  sizeof sortedInodes[0],
                  fs_logInumCompare);
        }

        for (i = 0; i < numToDo; i++)
        {
            fs_commitInodeDeallocate(context, sortedInodes[i]);
        }
    }

} /* fs_logCommitInodeDeallocate */


/**
 * Unlock inode btrees at the end of a transaction.
 */
static void
fs_logCommitBtrees(
    fs_Context      *context,
    fs_LogSummary   *logSummary)
{
    fs_LogLockedObjects          *objects;
    int32                   i;

    // Unlock inode btrees and release space held by the
    // inode array.
    objects = &logSummary->btreeArr;
    for (i = 0; i < logSummary->btreeCount; i++)
    {
        fs_Inode    *inode;

        FS_ASSERT(objects != NULL);

        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            fs_LogLockedObjects *prev = objects;
            objects = objects->next;
            prev->next = NULL;
            if (prev != &logSummary->btreeArr)
            {
                fs_ramDealloc(FS_RAM_LOG_BUF_ARR, prev);
            }
        }

        inode = objects->obj[i % FS_LOG_OBJ_ARR_MAX];

        if (inode == NULL)
        {
            // Inode was pending deallocation, its already
            // been handled.
            continue;
        }

        LOG_TRACE(("  commiting btree inum %lld %p deps %d",
                 inode->inum, inode, inode->logSegDep.count));

        fs_inodeBtreeCommit(context, inode);
    }
    if (objects != &logSummary->btreeArr)
    {
        fs_ramDealloc(FS_RAM_LOG_BUF_ARR, objects);
    }

} /* fs_logCommitBtrees */


/**
 * Commit the summmary at the end of a transaction.
 *
 */
int32
fs_logCommitSummary(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    fs_LogSummary   *logSummary)
{
    fs_LogLockedObjects *objects;
    int32               i;

    LOG_TRACE(("commit summary %d recs %d", logSummary->id,
               logSummary->nextRecNum));

    // Unlock buffers and release space held by the buf array.
    objects = &logSummary->bufArr;
    for (i = 0; i < logSummary->bufCount; i++)
    {
        fs_Buf  *buf;

        FS_ASSERT(objects != NULL);

        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            fs_LogLockedObjects *prev = objects;
            objects = objects->next;
            prev->next = NULL;
            if (prev != &logSummary->bufArr)
            {
                fs_ramDealloc(FS_RAM_LOG_BUF_ARR, prev);
            }
        }

        buf = objects->obj[i % FS_LOG_OBJ_ARR_MAX];

        if (buf == NULL)
        {
            continue;
        }
   
        LOG_TRACE(
                ("commiting buf inum %lld page %lld blk %lld flag %#x buf %p",
                  buf->inum, buf->page, fs_blockNumber(buf->block),
                  buf->flags, buf));
        fs_bufCommit(context, buf);
    }
    if (objects != &logSummary->bufArr)
    {
        fs_ramDealloc(FS_RAM_LOG_BUF_ARR, objects);
    }

    FS_ASSERT(DLL_IS_EMPTY(&logSummary->undoHdr));

    context->logSummary = NULL;
    fs_ramDealloc(FS_RAM_LOG_SUMM, logSummary);

    return NFX_OK;
}


/**
 * Remove the specified buf from the log summary.
 * Used when we give a buffer to the WB during a transaction.
 * The WB becomes the owner and will release the buffer when ioDone.
 * The logcommit has nothing to do for this buffer anymore.
 *
 * Remark:  The bufArr can have holes as a result of this (NULL entries !!!)
 */
boolean
fs_logRemoveBufLogSummary(
    fs_Context      *context,
    fs_Buf          *buf)
{
    fs_LogSummary   *logSummary;
    fs_LogLockedObjects    *objects;
    int32           i;

    logSummary = context->logSummary;
    if (logSummary == NULL)
    {
        return FALSE;
    }

    // Return TRUE if buf exists in this log summary.
    objects = &logSummary->bufArr;
    for (i = 0; i < logSummary->bufCount; i++)
    {
        FS_ASSERT(objects != NULL);

        if (i > 0 && (i % FS_LOG_OBJ_ARR_MAX) == 0)
        {
            objects = objects->next;
        }

        if (objects->obj[i % FS_LOG_OBJ_ARR_MAX] == buf)
        {
            objects->obj[i % FS_LOG_OBJ_ARR_MAX] = NULL;
            fs_lockUnsetDirty(context, &buf->lock);
            return TRUE;
        }
    }

    return FALSE;

} /* fs_logRemoveBufLogSummary */


/**
 * fs_logCommitEntry, fs_logCommitWrite, fs_logCommitEntryDone
 *
 * Called to commit a log entry to the log device at
 * the end of a transaction.
 *
 * fs_logCommitEntry: is always stable  (FILE_SYNC)
 * fs_logCommitWrite: used in the nfs write thread
 *    stableHow indicates request, and can be overwritten
 * fs_logCommitEntryDone: if previous commit was unstable
 *    the user can send the commit req
 *
 */
int32
fs_logCommit(
    fs_Context           *context,
    fs_LogStableWrite    stable,
    uint64               inum)
{
    fs_LogRecord         rec;
    fs_LogFsRoot         *logRoot = context->fileSystem->logRoot;
    fs_LogSummary        *logSummary = context->logSummary;
    fs_LogSummary        *outerSummary;
    fs_LogObjectSumm     objSumm;

    LOG_TRACE(("  id %d hdl %llx %s req commit %d recs %d bytes of log",
               logSummary->id, (uint64)((uintptr_t)context->smHandle),
               (context->reqHdr != NULL) ?
               fs_reqTypeTbl[context->reqHdr->requestType].name : "daemon",
               logSummary->nextRecNum,
               logSummary->bytesUsed));

    //printf("id %d hdl %#x %s req commit %d bytes of log\n",
    //  logSummary->id, smHandle,
    //  (context->reqHdr != NULL) ?
    //  fs_reqTypeTbl[context->reqHdr->requestType].name : "daemon",
    //  logSummary->bytesUsed);

    // Commit any pending inode deallocations.
    fs_logCommitInodeDeallocate(context);

    // Commit any pending space deallocations.
    fs_commitDeallocate(context);

    // Copy any modified inodes to the log.
    fs_logCopyInodes(context, logSummary);

    if (!context->eek.inUse)
    {
        NFX_SPIN_LOCK(&logRoot->spin);
        logRoot->logStats.nbrCommit++;
        NFX_SPIN_UNLOCK(&logRoot->spin);

        // Write the log record to the log device.
        rec.hdr.id = logSummary->id;
        rec.hdr.recType = FS_LOG_REC_COMMIT;
        rec.hdr.flags = 0;
	rec.hdr.numRecs = logSummary->nextRecNum;

        objSumm.objectPtr = NULL;
        objSumm.objectType = FS_LOG_DEP_NONE;

#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
        if (context->flags & FS_CONTEXT_TRANSACTION_COUNTED)
        {
            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->nbrOngoingTrans--;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);
            context->flags &= ~FS_CONTEXT_TRANSACTION_COUNTED;
        }
#endif           

        // Write the log record to the log device.
        fs_logWriteRecord(context,
                          &rec,
                          NULL,
                          0,
                          stable,
                          inum,
                          &objSumm);

        fs_logMarkActive(context, &objSumm);
    }

    fs_logCommitBtrees(context, logSummary);
    fs_logCommitInodes(context, logSummary);

    // fs_logCommitSummary will free logSummary,
    // so don't reference it after calling.
    outerSummary = logSummary->outerSummary;
    logSummary->outerSummary = NULL;
    fs_logCommitSummary(context, logRoot, logSummary);

    // Pop the transaction.
    context->logSummary = outerSummary;

    return NFX_OK;

} /* fs_logCommit */


/**
 * Commit the transaction to the log.
 */
int32
fs_logCommitEntry(
    fs_Context        *context)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;
    int32               rc = NFX_OK;

    LOG_TRACE(("id %d commit", context->logSummary->id));

    // Write commit record, unlock bound objects, store
    // incore log summary (TODO - this may change if fs_logCommit
    // is exploded).  This does not block.
    fs_logCommit(context, FS_LOG_WRITE_COMMIT_STABLE, 0);

    if (!context->eek.inUse)
    {
        rc = devTypeTable->commitTrans(context);
    }
    
    return rc;
}


/**
 * Commit outstanding log and user bufs for the
 * specified inode number.  Blocks until data is
 * persistent on disk if stableHow is FILE_SYNC.
 */
int32
fs_logCommitWrite(
    fs_Context         *context,
    fs_Inode           *inode,
    uint64             *inum,
    fs_WriteStableHow  *stableHow)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;
    int32               rc;

    LOG_TRACE(("id %d commit write", context->logSummary->id));

    rc = devTypeTable->commitWrite(context, inode, stableHow);

    fs_bufStartIoThreshold(context, inode, FS_BUFLIST_USER);

    return (rc);
}


/**
 * Commit outstanding log and user bufs for the
 * specified inode.  Blocks until data is persistent
 * on disk.
 */
int32
fs_logSyncInodeBuf(
    fs_Context         *context,
    fs_Inode           *inode)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;

    LOG_TRACE(("sync'ing inode %lld", inode->inum));

    return devTypeTable->syncInodeBuf(context, inode);
}

/**
 * Commit outstanding log and user bufs for the
 * specified inode.  Blocks until data is persistent
 * on disk.
 */
void
fs_logSyncInodeLog(
    fs_Context         *context,
    uint64             inum)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;

    LOG_TRACE(("sync'ing inum %lld", inum));

    devTypeTable->syncInodeLog(context, inum);
}


/**
 * If the specified buf has not been allocated for, perform
 * allocation and if stableHow is UNSTABLE, record a zero-fill
 * log entry for it.
 */
void
fs_logWriteBuf(
    fs_Context        *context,
    fs_Inode          *inode,
    fs_Buf            *buf,
    fs_WriteStableHow stableHow)
{
    FS_ASSERT(buf->flags & FS_BUF_TYPE_USER);

    LOG_TRACE(("id %d write user buf inum %lld page %lld blk %lld flags %#x", 
               context->logSummary->id, buf->inum, buf->page,
               fs_blockNumber(buf->block), buf->flags));

    switch (context->fileSystem->logRoot->logDevType) 
    {
    case FS_LOG_DEVTYPE_NVRAM:
        break;
    case FS_LOG_DEVTYPE_SELF:
    case FS_LOG_DEVTYPE_VOLID:
        vl_addUserBuf (context, inode, buf, stableHow);
        break;
    case FS_LOG_DEVTYPE_NULL:
        break;
    default:
        fs_error(context, FS_INTERNAL_ERR, "Unknown logDevType");
        break;
    }
}


/**
 * Initialize the log segment size and each segment's
 * head pointer.
 */
void
fs_logInitLogSegments(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          logSize)
{
    int32   i;

    bzero(logRoot->logSeg, sizeof(logRoot->logSeg));

    logRoot->segmentSize = logSize / FS_LOG_MAX_SEGMENTS;
    for (i = 0; i < FS_LOG_MAX_SEGMENTS; i++)
    {
        logRoot->logSeg[i].head = logRoot->segmentSize * i;
    }
}


/**
 * Destroy the specified fs_LogLockedObjects.
 */
static void
fs_logDestroyLockedObjects(
    fs_Context          *context,
    fs_LogLockedObjects *objects)
{
    fs_LogLockedObjects *objectsNext;

    while (objects != NULL)
    {
        objectsNext = objects->next;
        fs_ramDealloc(FS_RAM_LOG_BUF_ARR, objects);
        objects = objectsNext;
    }
}


/**
 * Destroy the specifed logSummary, freeing it and memory
 * for all structures chained to it.
 */
void
fs_logDestroySummary(
    fs_Context      *context,
    fs_LogSummary   *logSummary)
{
    fs_LogUndoSummary   *undo, *undoNext;

    if (logSummary->outerSummary)   
    {
        fs_logDestroySummary(context, logSummary->outerSummary);
        logSummary->outerSummary = NULL;
    }

    // Free any pending bufArr elements.
    fs_logDestroyLockedObjects(context, logSummary->bufArr.next);

    // Free any pending inodeArr elements.
    fs_logDestroyLockedObjects(context, logSummary->inodeArr.next);

    // Free any pending btreeArr elements.
    fs_logDestroyLockedObjects(context, logSummary->btreeArr.next);

    // Free any pending undo entries.
    DLL_ITERATE_NEXT(undo, undoNext, &logSummary->undoHdr, logEntry)
    {
        logSummary->undoCount--;
        fs_ramDealloc(FS_RAM_LOG_UNDO_SUMM, undo);
    }

    // Finally, dealloc the summary.
    fs_ramDealloc(FS_RAM_LOG_SUMM, logSummary);
}


/**
 * Destroy the in-core log summary information for the
 * specified volume.
 *
 * This is called when a volume is being forcibly
 * unmounted.
 */
void
fs_logDestroyVolume(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    //fs_LogFsRoot    *logRoot = fileSystem->logRoot;

} /* fs_logDestroyVolume */


/**
 * Dump log data
 */
int32
fs_logMakeDumpLogRecord(
    fs_Context  *context,
    char        *data,
    int32       remain)
{
    fs_MsgLogdataReq *req = (fs_MsgLogdataReq*) context->reqHdr;
    uint32            offset = req->offset;
    uint32            nbrEntries = req->nbrEntries;
    char              localStr[FS_LOGDATA_MAX_SIZE];
    int32             this = 0;
    char             *p = data;
    boolean           match;
    fs_LogFsRoot     *logRoot;

    if (context == NULL || context->fileSystem == NULL || data == NULL)
        return 0;

    logRoot = context->fileSystem->logRoot;
    if (logRoot == NULL)
        return 0;
    
    if (nbrEntries == 0)
    {
        if (offset == 3)
        {
#ifdef FS_DEBUG_CANDO_STATS
            extern int32 fs_bufWritebackCandoDeallocMax;
            extern int32 fs_ramCandoMaxDealloc;
#endif
#ifdef FS_DEBUG_INODE_CLEANER_TIMING
            extern int32 fs_inodeCleanerMaxPerLoop;
            extern eeeRTC_t fs_inodeCleanerMaxRuntime;
            extern int32 fs_inodeCleanerWakeupLog;
            extern int32 fs_inodeCleanerWakeupAll;
            extern int32 fs_inodeCleanerWakeupCache;
            extern int32 fs_inodeCleanerWakeupSegdep;
            extern int32 fs_inodeCleanerWakeupStart;
            extern int32 fs_inodeCleanerWakeupRestart;
            extern int32 fs_inodeCleanerWakeupRefDecr;
            extern int32 fs_inodeCleanerWakeupOther;
#endif
#ifdef FS_DEBUG_WRITEBACK_TIMING
            extern uint32 fs_userBufWbMaxPerLoop;
            extern uint32 fs_metaBufWbMaxPerLoop;
            extern eeeRTC_t fs_userBufMaxRuntime;
            extern eeeRTC_t fs_metaBufMaxRuntime;
#endif
            extern uint64 fs_stat_write_stable;
            extern uint64 fs_stat_write_un_stable;

            fs_vl_showStats(context);
#ifdef FS_DEBUG_CANDO_STATS
            printf("nbrTruncateMax: %d\n",
                   logRoot->logStats.nbrTruncateMax);
            printf("fs_ramCandoMaxDealloc: %d\n", fs_ramCandoMaxDealloc);
            printf("fs_bufWritebackCandoDeallocMax: %d\n",
                   fs_bufWritebackCandoDeallocMax);
#endif
#ifdef FS_DEBUG_INODE_CLEANER_TIMING
            printf("fs_inodeCleanerMaxPerLoop: %d (reset to 0)\n",
                   fs_inodeCleanerMaxPerLoop);
            printf("inodeCleanerMaxRuntime: %lld msec (wakeup cnt %d %d %d %d %d %d %d %d)\n",
                   eee_rtcToMS(&fs_inodeCleanerMaxRuntime),
                   fs_inodeCleanerWakeupLog,
                   fs_inodeCleanerWakeupAll,
                   fs_inodeCleanerWakeupCache,
                   fs_inodeCleanerWakeupSegdep,
                   fs_inodeCleanerWakeupStart,
                   fs_inodeCleanerWakeupRestart,
                   fs_inodeCleanerWakeupRefDecr,
                   fs_inodeCleanerWakeupOther);
            fs_inodeCleanerMaxPerLoop = 0;
			fs_inodeCleanerMaxRuntime = *(eee_rtcZero());
#endif
#ifdef FS_DEBUG_WRITEBACK_TIMING
            printf("bufWritebackMaxPerLoop: meta %d user %d (reset to 0)\n", 
                   fs_metaBufWbMaxPerLoop, fs_userBufWbMaxPerLoop);
            printf("bufWritebackMaxRuntime: meta %lld user %lld msec\n", 
                   eee_rtcToMS(&fs_metaBufMaxRuntime),
				   eee_rtcToMS(&fs_userBufMaxRuntime));
            fs_userBufWbMaxPerLoop = 0;
            fs_metaBufWbMaxPerLoop = 0;
            fs_userBufMaxRuntime = *(eee_rtcZero());
            fs_metaBufMaxRuntime = *(eee_rtcZero());
#endif

            printf("fs_stat_write_stable      %10lld - fs_stat_write_un_stable %10lld \n",
                   fs_stat_write_stable, fs_stat_write_un_stable); 
            
        }
        if (offset == 4)
        {
            fs_incorePrintResSettings(context);
            fs_incorePrintCurrResources(context);
            fs_incorePrintStats(context);
        }
        if (offset == 5)
        {
            fs_logPrintStats(&logRoot->logStats, logRoot->currLen);
        }
    }

    if (nbrEntries && (offset & 0x7))
    {
        this = snprintf(p, remain, "offset not 8-byte aligned, rejecting\n");
        nbrEntries = 0;
    }
    if (nbrEntries && (offset > context->fileSystem->logSize))
    {
        this = snprintf(p, remain, "offset %d is to big (max %d)\n", offset, context->fileSystem->logSize);
        nbrEntries = 0;
    }
    p += this;
    remain -= this;

    while ((offset >= 0) && 
           (nbrEntries-- > 0)) 
    {
        this = fs_logDumpLogRecord(context, logRoot, &offset, 
                                   localStr, FS_LOGDATA_MAX_SIZE - 1,
                                   req->criteria, req->value, &match);
        if (match)
        {
            if (this < remain)
            {
                memcpy(p, localStr, this) ;
                p += this;
                remain -= this;
            } 
            else 
            {
                break;
            }
            
            printf ("%s", localStr) ;
        }
    }
    
    return (p - data);
}


/*-----------------------------------------------------------------------------
 * Device type table functions for the NULL device.
 */
int32
fs_logNullInitDevice(
    fs_Context  *context)
{
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;

    logRoot->headPtr = 0;
    logRoot->head = 0;
    logRoot->tail = 0;
    logRoot->currLen = 0;
    logRoot->nextId = FS_LOG_ID_MIN;

    return NFX_OK;
}

int32
fs_logNopCommitTrans(
    fs_Context *context)
{
    return NFX_OK;
}

int32
fs_logNopSyncInodeLog(
    fs_Context  *context,
    uint64      inum)
{
    return NFX_OK;
}

int32
fs_logNopSyncInodeBuf(
    fs_Context  *context,
    fs_Inode    *inode)
{
    return NFX_OK;
}

int32
fs_logNullCommitWrite(
    fs_Context          *context,
    fs_Inode            *inode,
    fs_WriteStableHow   *stableHow)
{
    int32   rc;

    // Perform the necessary steps to complete the
    // transaction, but the commit record doesn't
    // actually get written (TODO - this should change
    // if/when we break up fs_logCommit)
    rc = fs_logCommit(context, FS_LOG_WRITE_OTHER, 0);

    // We don't modify stableHow here, use of this
    // device type is for debug only.

    return rc;
}

int32
fs_logNopWriteDevHdr(
    fs_Context  *context,
    int32       operation,
    uint32      count)
{
    return NFX_OK;
}

int32
fs_logNopRecordCksum(
    fs_Context      *context,
    fs_LogRecHdr    *recordHdr,
    int32           recordHdrLen,
    void            *data,
    int32           dataLen)
{
    return NFX_OK;
}

void
fs_logNullWrite(
    fs_Context          *context,
    uint32              *offset,
    uchar8              *datap[LOG_IOVEC_MAX],
    uint32              lenp[LOG_IOVEC_MAX],
    fs_LogStableWrite   stable,
    uint64              inum)
{
    fs_LogUndoSummary   *undo, *undoNext;
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    int32               length = 0, i;

    *offset = logRoot->tail;

    // Increment the tail pointer so that each record
    // in-core will have a distinct offset and thus be
    // retrievable when the dependency is cleared.
    // This must match the record length in order for
    // proper truncation to occur.
    for (i = 0; i < LOG_IOVEC_MAX; i++)
	length += lenp[i];

    NFX_SPIN_LOCK(&logRoot->spin);
    FS_LOG_INCR_PTR(context->fileSystem->logSize,
                    logRoot->tail,
                    length);
    NFX_SPIN_UNLOCK(&logRoot->spin);

    if (stable == FS_LOG_WRITE_COMMIT_STABLE ||
        stable == FS_LOG_WRITE_COMMIT_UNSTABLE)
    {
        DLL_ITERATE_NEXT(undo, undoNext, &context->logSummary->undoHdr,
                         logEntry)
        {
            fs_Buf  *buf = undo->objSumm.objectPtr;

            FS_ASSERT(undo->objSumm.objectType == FS_LOG_DEP_BUF);
            FS_ASSERT(buf->flags & FS_BUF_TYPE_META);
            DLL_REMOVE(&context->logSummary->undoHdr, undo, logEntry);

            NFX_SPIN_LOCK(&buf->spin);
            DLL_REMOVE(&buf->undoHdr, undo, bufEntry);
            DLL_INSERT_TAIL(&buf->candoHdr, undo, bufEntry);
            NFX_SPIN_UNLOCK(&buf->spin);
        }
    }
}

int32
fs_logNullReadInit(
    fs_Context  *context,
    fs_LogInfo  *info)
{
    // TBD: only for action FS_LOG_ACTION_REPLAY ??
    info->head = 0;
    info->tail = 0;

    return NFX_OK;
}

int32
fs_logNullRead(
    fs_Context  *context,
    uchar8      **datap,
    uint32      offset,
    uint32      *bytesRead,
    uint32      tail)
{
    return NFX_OK;
}

void
fs_logNullReadDone(
    fs_Context  *context,
    uchar8      *datap)
{
    return;
}

int32
fs_logNopVerifyCksum(
    fs_Context  *context,
    uint32      *data,
    uint32      length)
{
    return NFX_OK;
}

int32
fs_logNopIsEmpty(
    fs_Context  *context)
{
    return TRUE;
}


/*
 * END OF Device type table functions for the NULL device.
 *---------------------------------------------------------------------------*/


/**
 * Stop the logger daemon for the specified
 * filesystem.
 * 
 * TODO - convert this to use the log-device-type table.
 */
void
fs_logStopDaemon(
    fs_Context  *context,
    boolean     isForce)
{
    fs_FileSystem   *fs = context->fileSystem;
    int32           rc;

    switch (fs->logRoot->logDevType) 
    {
    case FS_LOG_DEVTYPE_NVRAM:
        rc = nvram_deallocate(NVRAM_CLIENT_EFS,
                              fs->logRoot->logDev.nvHandle);
        if (rc != NFX_OK)
        {
            fs_warn(context, FS_INTERNAL_ERR,
                "nvram_dealloc logStopDaemon");
        }
        break;
    case FS_LOG_DEVTYPE_SELF:
    case FS_LOG_DEVTYPE_VOLID:
        vl_stopIODaemon (context, isForce);
        break;
    case FS_LOG_DEVTYPE_NULL:
        break;
    default:
        fs_warn(context, FS_LOG_BAD_DEV_ERR, "logStopDaemon");
        break;
    }

} /* fs_logStopDaemon */


/*---------------------------------------------------------------------------
 * Debugging stuff for forcing crashes.
 */
#ifdef FS_DEBUG_CHECK_CRASH

void
fs_logCrash(
    fs_Context  *context,
    int32       crashPhase)
{
    printf("\nlog test: crash phase %d, logRoot stats:\n", crashPhase);
    fs_logPrintStats(&context->fileSystem->logRoot->logStats,
                     context->fileSystem->logRoot->currLen);
    fs_debugCrash();
}


extern void
fs_logTestForceCrash(
    fs_Context  *context,
    int32       crashPhase,
    int32       n)
{
    int32   i;
    boolean isDir;
    uint16  perm = 0644;
    uint16  fperm = perm | FS_IFMT_REG, dperm = perm | FS_IFMT_DIR;
    fs_Time  now = {0,0};
    uint32  uid = 0, gid = 0, fileset = 0;
    char    name[256];

    uint64      inums[FS_LOG_MAX_TESTS];
    fs_Buf      *buf;
    fs_Extent   extents[FS_LOG_MAX_TESTS];
    fs_Inode    *root, *inodes[FS_LOG_MAX_TESTS];

    root = I_GET_EXCL(0, FS_ROOT_INODE);
    assert(root);

    if (!n) n = FS_LOG_DEFAULT_TESTS;
    if (n > FS_LOG_MAX_TESTS) n = FS_LOG_MAX_TESTS;

    // create a transaction context

    fs_logBeginTransaction(context);

    for (i = 0; i < n; i++)
    {
        now = fs_getTimeOfDay();

        // step 1: allocate inode -> FS_LOG_REC_BITMAP/set

        isDir = (i % 4) == 0;
        if (isDir)
        {
            inodes[i] = fs_allocateDirectoryInode(context, root, dperm, uid, gid, fileset);
        }
        else
        {
            inodes[i] = fs_allocateFileInode(context, root, fperm, uid, gid, fileset);
        }
        assert(inodes[i]);
        inums[i] = inodes[i]->inum;

        // allocate extent so we can deallocate them in step 4 below
        // to add dependencies before deallocating the inodes

        if (!isDir)
        {
            extents[i] = fs_allocateExtent(context, inodes[i], 0, FS_LOG_NPAGE, FALSE);
            assert(fs_extentLength(extents[i]));
            fs_emap_add(context, inodes[i], 0, extents[i]);
        }

        // step 2: write inode -> FS_LOG_REC_INODE (fs_logCommit() time)

        inodes[i]->dinode.size = isDir ? 0 : fs_extentLength(extents[i]);
        inodes[i]->dinode.attrs = isDir ? FS_IFMT_DIR : FS_IFMT_REG;
        inodes[i]->dinode.linkCnt = isDir ? 2 : 1;
        inodes[i]->dinode.ctime = now;
        inodes[i]->dinode.mtime = now;
        inodes[i]->dinode.atime = now;

        fs_inodeWrite(context, inodes[i], INO_WRITE_LOG);

        // add dir entry to root

        sprintf(name, "replay%d", i+now.secs);
        fs_dir_addent(context, root, name, strlen(name), inodes[i]->inum);
        root->dinode.atime = now;
        root->dinode.mtime = now;
        if (isDir) root->dinode.linkCnt += 1;
        fs_inodeWrite(context, root, INO_WRITE_LOG);

        // step 3: write bufs
        //         write zero first page -> FS_LOG_REC_ZERO_FILL

        if (!isDir)
        {
            buf = O_GET_EXCL(inodes[i], 0);
            assert(buf);
            FS_BUF_ZERO_ALL(buf);
        }

        //printf("log test[%d]: inum %lld %s attrs %o size %lld\n",
        //        i, inums[i], (isDir ? "is dir" : "reg"),
        //        inodes[i]->dinode.attrs, inodes[i]->dinode.size);
    }

    // check for crash phase 1: before commiting transaction
    // no commit record: verify all records added are discarded during replay

    if (crashPhase == FS_LOG_TEST_PHASE1)
    {
        fs_logCrash(context, crashPhase);
    }

    // commit and release the transaction context

    fs_logCommitEntry(context);

    // TBD -> META_TWO

    fs_snapCreate(context, "s1");

    fs_logBeginTransaction(context);
    for (i = 0; i < n; i++)
    {
        fs_Inode    *inode;
        inode = I_GET_EXCL(0, inums[i]);
        assert(inode);
        isDir = FS_I_IS_DIR(inode->dinode.attrs);
        if (!isDir)
        {
            buf = O_GET_EXCL(inode, 0);
            assert(buf);
            FS_BUF_ZERO_ALL(buf);
        }
    }
    fs_logCommitEntry(context);

    // check for crash phase 2: after commiting transaction
    // commit record added: verify all records added are replayed

    if (crashPhase == FS_LOG_TEST_PHASE2)
    {
        fs_logCrash(context, crashPhase);
    }

    // create a transaction context again

    fs_logBeginTransaction(context);

    root = I_GET_EXCL(0, FS_ROOT_INODE);
    assert(root);
    for (i = 0; i < n; i++)
    {
        // step 4: deallocate inode -> FS_LOG_REC_BITMAP/clear,
        //         FS_LOG_REC_INODE_FLUSH
        //         need to add dependencies !!
        //         FS_ASSERT(logSummary->recsPend > 0) in
        //         fs_logAddDep

        I_GET_EXCL(0, inodes[i]->inum);
        isDir = FS_I_IS_DIR(inodes[i]->dinode.attrs);
#if 0
        if (!isDir)
        {
            fs_deallocateExtent(context, extents[i]);
            fs_emap_remove(context, inodes[i], 0);
        }
#endif
        if (isDir)
        {
            sprintf(name, "replay%d", i+now.secs);
            fs_dir_removeent(context, root, name, strlen(name));
            root->dinode.linkCnt -= 1;
            fs_inodeWrite(context, root, INO_WRITE_LOG);
            fs_deallocateInode(context, inodes[i]);
            //printf("test replay: deallocate dir inode %lld\n",
            //               inodes[i]->inum);
        }
        else
        {
            fs_inodeReleaseExclusive(context, inodes[i]);
        }
    }

    // commit and release the transaction context

    fs_logCommitEntry(context);
}
/*
 * END OF Debugging stuff for forcing crashes.
 *---------------------------------------------------------------------------*/
#endif


/**
 * Print the in-core summary information for the
 * specified logSummary.
 */
void
fs_logDumpIncoreData(
    fs_Context      *context,
    fs_LogSummary   *logSummary)
{
    printf("logSummary at %p id %d nextRecNum %d\n",
           logSummary, logSummary->id, logSummary->nextRecNum);

    printf(" bufCnt %d inoCnt %d btCnt %d undoCnt %d\n",
           logSummary->bufCount, logSummary->inodeCount,
           logSummary->btreeCount, logSummary->undoCount);

} /* fs_logDumpIncoreData */

#endif // SSC
