/*
 * Name:  fs-log.h
 *
 * RCS:  $Id: fs-log.h,v 1.115 2002/12/10 19:34:28 bob Exp $
 *
 * Description:  API data structures for transaction logging.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_LOG_H)
#define FS_LOG_H

/*-----------------------------------------------------------------------------
 * Filesystem transaction log API.  Various modules in the filesystem
 * use this API to record and commit log records, and to remove log
 * dependencies.
 *
 * For transaction logging to work on this system:
 *
 *  1 - an exclusive lock must be obtained on each metadata block
 *      to be modified (inodes are an exception, only the inode is
 *      locked exclusively, the block itself is shared locked).
 *  2 - each metadata block modification must be logged (fs_logAddDataRecord).
 *  3 - the exclusive lock cannot be released until after the log entry
 *      is commited to the log device (fs_logCommitEntry, which takes
 *      care of releasing the locks)
 *  4 - as modified blocks are written to stable storage, the dependency
 *      must be removed from the log entry (fs_logRemoveDepend)
 *
 * Definitions:
 *
 *    log entry - Composed of one or more data records and a commit record.
 *                During log redo, the absense of a commit record for a
 *                log entry causes all of its data records to be ignored.
 *                A committed log entry describes a complete filesystem
 *                transaction.
 *
 *      record  - A component of a log entry.  It indicates which log
 *                entry it is a part of, and describes a single contiguous
 *                modification to a filesystem object (block).  The records
 *                that make up a log entry are not necessarily stored
 *                contiguously in the log.
 *
 * A log is composed of a list of entry records that runs circularly on
 * the log device (typically a segment of NVRAM).  There is a pointer
 * maintained outside of the log (but on the same device) that points to
 * the head (oldest record). The end of the log is known by the last field
 * in a record being a well-known value, which is over-written by each
 * subsequently appended record.  If a log record's consistency check
 * fails (a checksum over the record), the log itself is not considered
 * invalid, but rather it is determined that the end of the log has
 * been reached (with the previous entry being the last valid record).
 * This case could occur if a system failure occurs while writing a
 * log record.  An empty log is identified when the head pointer points
 * to an id matching the well-known "last" value (FS_LOG_ID_LAST).
 *
 * For each data record in a committed log entry, a dependency is tracked
 * in-core.  For every record added, the block that is modified must
 * include a reference to the log transaction id.  As each writeback
 * completes for the modified block(s), the I/O completion routine calls
 * into this module to remove the dependency for the record. As a dependency
 * is met, the record can be removed from the log. Once all of the
 * outstanding dependencies are met the commit record can be removed from
 * the log. Removal of a log entry from the log device is accomplished by
 * moving the head pointer past those records that have been obsoleted.
 *
 * If space in the log reaches a low mark this module will schedule
 * dirty block writebacks.
 */

#define FS_LOG_ID_BYTES  sizeof(fs_LogId)

#define FS_LOG_MIN_NBR_BLOCKS        ((1024 * 1024) / FS_BLOCK_SIZE)
#define FS_LOG_MAX_NBR_BLOCKS        ((64 * 1024 * 1024) / FS_BLOCK_SIZE)
// #define FS_LOG_MAX_NBR_BLOCKS        ((16 * 1024 * 1024) / FS_BLOCK_SIZE)

enum {
	LOG_IOVEC_HDR = 0,
    LOG_IOVEC_DATA,
	LOG_IOVEC_PAD,
	LOG_IOVEC_EOL,
	LOG_IOVEC_MAX
};

#define FS_LOG_WRITE_HEAD            1
#define FS_LOG_WRITE_INCR_COMMITS    2
#define FS_LOG_WRITE_DECR_COMMITS    3

/*
 * Maximum number of inode deallocations per single transaction
 */
#define FS_LOG_MAX_INODE_DEALLOC     4        // arbitrary small value

/*
 * The enum to describe hoe the log is written.
 * Currently, this is only usefull for Vol log, 
 *   nvram is stable by implementation.
 * All log records are of the mode: FS_LOG_WRITE_OTHER
 * All commits are written in STABLE mode, only the NFS write req
 * in NFSv3 with user flag set to FS_WRITE_STABLE_UNSTABLE
 * indicates the log write mode as FS_LOG_WRITE_COMMIT_UNSTABLE.
 *
 * If log write mode is STABLE: the commit will yield until the
 *   log record is confirmed to be written out to disk.
 * In unstable mode: no yield is needed.
 */ 
typedef enum fs_LogStableWrite
{
    FS_LOG_WRITE_COMMIT_STABLE      = 1,
    FS_LOG_WRITE_COMMIT_UNSTABLE    = 2,
    FS_LOG_WRITE_OTHER              = 3
} fs_LogStableWrite;


/*-----------------------------------------------------------------------------
 * A log entry consists of one or more data records and
 * a commit record.  Each record minimally consists of an
 * fs_LogRecHdr.
 *
 */
typedef struct fs_LogRecHdr
{
    fs_LogId    id;                     // Transaction id this record
                                        // belongs to.
#define FS_LOG_ID_MIN      1
#define FS_LOG_ID_MAX      0xfeedb00f   // Highest valid transaction id
#define FS_LOG_ID_EOL      0xfeedbeef   // Indicates end of log.
#define FS_LOG_ID_INVALID  0xfeedbeff   // Indicates that record does not belong to 
                                        // to a transcation (trash records during replay)

    uchar8      recType;                // Type of record:
#define FS_LOG_REC_META_ONE     1       // Contains metadata only, refers to
                                        // a single block as src and sink.
#define FS_LOG_REC_META_TWO     2       // Contains metadata only, refers to
                                        // one block as src and another as
                                        // sink.  This type of entry is
                                        // only used when the source block
                                        // is COW (which is cleared as a
                                        // result of the modification being
                                        // recorded, thus only ONE of these
                                        // entries for a given block should
                                        // ever be present in the log).
#define FS_LOG_REC_USER         3       // Contains user data only, refers to
                                        // a COD handle as src and a single
                                        // block as sink.
#define FS_LOG_REC_ALLOC        4       // Provides block allocation info for
                                        // blocks that had previously been
                                        // logged with ino,offset (logical)
                                        // location info only.
#ifdef NEW_ZF
#define FS_LOG_REC_META_ZERO    5       // Zero fill a meta disk block.
#define FS_LOG_REC_USER_ZERO    6       // Zero fill a user disk block.
#define FS_LOG_REC_COMMIT       7       // Commit record.
#define FS_LOG_REC_USER_RM      8       // Indicates that the block with user
                                        // has been removed (truncated)
#define FS_LOG_REC_BITMAP       9       // Intent log record - bitmap.  Can
                                        // be undone.

#define FS_LOG_REC_MARK_DONE    10       // Mark a previous transaction as done.
                                        // This transaction must not be replayed !!
#define FS_LOG_REC_INODE        11      // Modified an inode.
#define FS_LOG_REC_INODE_FLUSH  12      // Removed an inode, provides for flushing
                                        // it from the cache during replay.
#define FS_LOG_REC_BUF_FLUSH    13      // Removed one or more bufs, provides for
                                        // flushing them from the cache during
                                        // replay.
#define FS_LOG_REC_TYPE_MAX     14
#else // NEW_ZF
#define FS_LOG_REC_ZERO_FILL    5       // Zero fill a disk block.
#define FS_LOG_REC_COMMIT       6       // Commit record.
#define FS_LOG_REC_USER_RM      7       // Indicates that the block with user
                                        // has been removed (truncated)
#define FS_LOG_REC_BITMAP       8       // Intent log record - bitmap.  Can
                                        // be undone.

#define FS_LOG_REC_MARK_DONE    9       // Mark a previous transaction as done.
                               
         // This transaction must not be replayed !!
#define FS_LOG_REC_INODE        10      // Modified an inode.
#define FS_LOG_REC_INODE_FLUSH  11      // Removed an inode, provides for flushing
                                        // it from the cache during replay.
#define FS_LOG_REC_BUF_FLUSH    12      // Removed one or more bufs, provides for
                                        // flushing them from the cache during
                                        // replay.
#define FS_LOG_REC_TYPE_MAX     13
#endif // NEW_ZF
    uchar8      flags;
#define FS_LOG_BLK_LOGICAL      0x01    // Blk number not yet allocated for
                                        // location of result block (dest
                                        // referred to by ino,offset).
#define FS_LOG_REPLAY_COMMITTED 0x02    // Replay entry committed, awaiting
                                        // blk allocation.
    ushort16    numRecs;                // Number of records for this
                                        // transaction, valid in commit record
                                        // only.
} fs_LogRecHdr;

/*-----------------------------------------------------------------------------
 * The physical location of a block is not always known at the time
 * the log entry is recorded (allows for more intelligent space allocation
 * policy).  In such cases, the inode number and offset are recorded now,
 * (and recHdr.flags has FS_LOG_BLK_LOGICAL set) and a later logged
 * transaction will record the space allocation.  During log replay, in
 * the absense of the later space allocation transaction, space will be
 * allocated as part of the replay (and we ensure that free space never
 * runs low enough to cause the late allocation to fail).
 */
struct fs_LogBlkId
{
    uint64          inum;       // Inode number (can be FS_INVALID_INODE)
    uint64          page;       // Byte offset in file 
    fs_Block        block;      // storage block in filesystem (can be FS_INVALID_BLOCK)
};

/*-----------------------------------------------------------------------------
 * Log entry data records are of twelve different types.  A record describes
 * a modification for a SINGLE filesystem block ONLY!
 */
struct fs_LogRecMetaOne
{
    fs_LogBlkId     blkId;              // Block being modified.
    uint32          offset;             // Byte offset on the block where
                                        // the new data starts.
    uint32          length;             // Length in bytes of data[0].
    uchar8          data[0];            // The new data.

};

struct fs_LogRecMetaTwo
{
    uint64      srcBlkNum;              // Block number of source
                                        // block.
    fs_LogBlkId dstBlkId;               // Block location of dest
                                        // block.
    uint32      offset;                 // Byte offset where new
                                        // data starts.
    uint32      length;
    uchar8      data[0];                // The new data.
};

struct fs_LogRecUser
{
    fs_LogBlkId         blkId;          // Block being modified.
    nfx_cod_t           codHandle;      // Handle on new user data. 
    int32               codLength;
    int32               pad;
};

struct fs_LogRecUserRm
{
    fs_LogBlkId         blkId;          // Block being modified.
    nfx_cod_t           codHandle;      // Handle on new user data. 
};

struct fs_LogRecAlloc
{
    fs_LogBlkId         blkId;          // Inum,offset pair containing
                                        // block being allocated for.
    boolean             isMeta;         // Is this a metadata buffer?
    boolean             isCow;          // Is this a COW alloc?
    uint64              oldBlkNum;      // Block number of old for COW alloc
};

#ifdef NEW_ZF
struct fs_LogRecMetaZero
{
    fs_LogBlkId     blkId;      // Block being zeroed
    uint32          offset;     // Where to start zero-fill.
    uint32          length;     // Length of zero-fill.
};

struct fs_LogRecUserZero
{
    fs_LogBlkId     blkId;      // Inum + start page
    uint32          offset;     // Offset in first page
    uint32          nbPage;     // Number of pages (0 = all pages of the file)
    boolean         zero;       // TRUE: zero fill, FALSE: don't zero fill
	uint32			pad;
};
#else // NEW_ZF
struct fs_LogRecZero
{
    fs_LogBlkId         blkId;          // Block being zeroed
    uint32              offset;         // Where to start zero-fill.
    uint32              length;         // Length of zero-fill.
    boolean             isMeta;         // Is this a metadata buffer?
	uint32				pad;
};
#endif // NEW_ZF

struct fs_LogRecBitmap
{
	fs_BitmapOperation	debugOp;
    fs_LogBlkId         blkId;          // Block containing bitmap.
    uint32              offset;         // Byte offset of bitmap in block.
#define FS_LOG_BITMAP_MAX_BYTES 32
    uchar8              bits[FS_LOG_BITMAP_MAX_BYTES];// Bits to be set or reset.
    ushort16            length;         // Num of bytes in bits[]
#define FS_LOG_REC_BITMAP_SET   1
#define FS_LOG_REC_BITMAP_RESET 2
    ushort16            op;             // SET or RESET
};

/**
 * Modified an inode.
 */
struct fs_LogRecInode
{
    uint64      inum;
	int32		snapnum;
	int32		pad;
    fs_Dinode   dinode;
};

/**
 * Removed an inode.
 */
struct fs_LogRecInodeFlush
{
    uint64      inum;
};

/**
 * Removed one or more bufs.
 */
struct fs_LogRecBufFlush
{
    int32       bufList;
	int32		pad;
    fs_Block    block;      // If single meta block
    uint64      inum;       // inum, startPage if
    uint64      startPage;  // one or more file page
};

/**
 * Mark a transaction completed (not-replayable)
 */
struct fs_LogRecMarkDone
{
	fs_LogId	logId;
	uint32		pad;
};

struct fs_LogRecord
{
    fs_LogRecHdr        hdr;

    union
    {
        fs_LogRecMetaOne        metaOne;
        fs_LogRecMetaTwo        metaTwo;
        fs_LogRecUser           user;
        fs_LogRecAlloc          alloc;
        fs_LogRecUserRm         userRm;
        fs_LogRecBitmap         bitmap;
        fs_LogRecInode          inode;
        fs_LogRecInodeFlush     inodeFlush;
        fs_LogRecBufFlush       bufFlush;
        fs_LogRecMarkDone		markDone;
#ifdef NEW_ZF
        fs_LogRecUserZero       userZero;
        fs_LogRecMetaZero       metaZero;
#else // NEW_ZF
        fs_LogRecZero           zero;        
#endif // NEW_ZF
    } u;
};

#define FS_LOG_ALIGN_64 sizeof(uint64)

/**
 * Maximal log record size - enough to contain a complete filesystem
 * block, the record header, and any additional padding.
 */
#define FS_LOG_RECORD_LEN_MAX (FS_MAX_BLOCK_SIZE + 		\
                               sizeof(fs_LogRecord) + 	\
                               FS_LOG_ALIGN_64)

/**
 * The type of an object which contains log dependencies.
 */
typedef enum {
    FS_LOG_DEP_NONE,
    FS_LOG_DEP_BUF,
    FS_LOG_DEP_INODE,
    FS_LOG_DEP_BTREE,
#ifdef NEW_ZF
    FS_LOG_DEP_UZERO,
#endif // NEW_ZF
} fs_LogDepType;


/**
 * Structure to contain the object and log location the object's
 * data was written to.
 */
struct fs_LogObjectSumm
{
    uint32          offset;         // Byte offset on log device where record
                                    // starts.
    int32           length;         // Bytes consumed by record on device,
                                    // including any padding applied.
    void            *objectPtr;     // Object this record depends on.
    fs_LogDepType   objectType;     // Type of object for objectPtr.
};


/**
 * Holds the list of inodes modified for a given transaction.
 * They're held locked until the transaction commits. An array
 * and next pointer to facilitate expanding the number of inodes
 * for a particular log entry.
 */
struct fs_LogLockedObjects
{
#define FS_LOG_OBJ_ARR_MAX    8
    void            	 *obj[FS_LOG_OBJ_ARR_MAX];
    fs_LogLockedObjects  *next;
    fs_LogLockedObjects  *last; // Valid only at head of chain.
};

/**
 * Holds in-core data that provides for undo'ing a change.  This
 * is used during buf writeback of bufs containing uncommitted
 * changes.  Only certain log record types are undo'able.
 */
struct fs_LogUndoSummary
{
    fs_LogObjectSumm        objSumm;
    int32                   type;
    uint32                  logId;
    union
    {
        fs_LogRecBitmap     bitmap;
        // Add others as needed
    } u;
    DLL_ENTRY(fs_LogUndoSummary)
                            logEntry;
    DLL_ENTRY(fs_LogUndoSummary)
                            bufEntry;
};

enum {
    FS_LOG_UNDO_TYPE_BITMAP = 1,
    FS_LOG_UNDO_TYPE_NOP    = 2,
};

/**
 * Used for fs_LogDevTypeTable's logReadInit() routine
 * Called to get log device info for logRead()
 */
typedef struct
{
    uint32  logSize;    // how many bytes in the log
    uint32  nbrCommits; // how many commits
    uint32  head;       // start of log data
    uint32  tail;       // end of log data
                        //   NVRAM: can set to NULL if unknown
    uint32  readChunkSize;  // number of bytes to read at a time
                            //   NVRAM_RW_MAX for nvram
                            //   VOL_MIN_TO_READ for vol log
    uint32  maxSize;    // maximum bytes for a read
} fs_LogInfo;


/**
 * Number of characters needed to describe a dependency.
 */
enum { FS_LOG_DEP_DESCRIPTION = 100 };


/**
 * Describe the object containing a dependency set.
 */
extern char *fs_logDepDescribe(
    fs_Context *context,
    void *objectPtr,
    int32 objectType,
    char        *string,
    int32       remain);


/**
 * Macros to manipulate log size head/tail pointers.
 */
#define FS_LOG_DEV_RESERVED_BYTES                             \
    ((logRoot->logDevType == FS_LOG_DEVTYPE_NVRAM) ?          \
        FS_LOG_RESERVED_BYTES : 0 )

#define FS_LOG_LOOP_BYTES(_logSize_)                          \
    ((_logSize_) - FS_LOG_DEV_RESERVED_BYTES)

#define FS_LOG_INCR_PTR(_logSize_,_ptr_,_num_)                \
    (_ptr_) += (_num_);                                       \
    if ((_ptr_) >= (_logSize_))                               \
    {                                                         \
        (_ptr_) = (_ptr_) - FS_LOG_LOOP_BYTES(_logSize_);     \
    }

#define FS_LOG_FREE_BYTES(_logRoot_)                          \
    (FS_LOG_LOOP_BYTES(context->fileSystem->logSize) -        \
        (_logRoot_)->currLen)

#define FS_LOG_DECR_PTR(_logSize_,_ptr_,_num_)                \
    if ((_ptr_) - FS_LOG_DEV_RESERVED_BYTES < (_num_))        \
    {                                                         \
        (_ptr_) = (_logSize_) - ((_num_) -                    \
             (((_ptr_)) - FS_LOG_DEV_RESERVED_BYTES));        \
    }                                                         \
    else                                                      \
    {                                                         \
        (_ptr_) -= (_num_);                                   \
    }

#define FS_LOG_IS_EMPTY(_logRoot_)                            \
    ((_logRoot_)->currLen == 0 ||                             \
     (_logRoot_)->currLen == FS_LOG_ID_BYTES)

#define LOG_TRACE(args)   FS_TRACE(context, FS_TRACE_LOG, args)


/**
 * In-core summary of log entry.  This is used to track
 * outstanding dependencies on committed log entries.  Once all
 * dependencies are met, the log entry and this summary are
 * obsoleted.
 */
struct fs_LogSummary
{
    fs_LogId                    id;         // Transaction id of this log entry.
    uint32                      nextRecNum; // Next record number for this id.
    uint32                      bytesUsed;  // Total bytes log space consumed.

    int64                       blksUsed;   // blocks allocated/freed in this thread
    int64                       inodesUsed; // inodes allocated/freed in this thread

    fs_LogLockedObjects         bufArr;  	// Set of locked bufs for the log
                                            // entry.
    int                         bufCount;	// count of bufArr elements
    fs_LogLockedObjects         inodeArr;	// Set of locked inodes for the log
                                            // entry.
    int                         inodeCount;	// count of inodeArr elements
    fs_LogLockedObjects         btreeArr;	// Set of locked btrees for the log
                                            // entry.
    int                         btreeCount;	// count of btreeArr elements
    DLL_HDR(fs_LogUndoSummary)  undoHdr;    // List of undo'able records.
    int                         undoCount;	// count of undo elements
    fs_LogSummary               *outerSummary;// For a nested transaction, the
                                              // next transaction's log summary.
    DLL_ENTRY(fs_LogSummary)    hashEntry;  // During replay, entry for list of
};


/**
 * Header for a log, located at offset 0 on each volume's log device.
 */
struct fs_LogDevHdr
{
    uint32      head;            // Current byte offset of head of log.
    uint32      nbrCommits;      // Gives the indication how complete the
                                 // replay was (in case of no EOL or CRC error)
    uint32      cksum;
    uint32      magic;
#define FS_LOG_MAGIC        0xd00fbead
    uint32      version;
#define FS_LOG_VERSION_ONE  0x00000001
    uint32      timeStamp;
    int32       logSize;         // Size in bytes of log device.
    int32       spare[1];
};
#define FS_LOG_RESERVED_BYTES   (2 * sizeof(fs_LogDevHdr))


/**
 * This maintains/tracks the statistics during a log replay/test 
 */
typedef struct fs_LogStats
{
    uint32              nbrMetaOne;
    uint32              nbrMetaTwo;

    uint32              nbrUser;
    uint32              nbrAlloc;

    uint32              nbrZeroUser;
    uint32              nbrZeroMeta;

    uint32              nbrUserRm;
    uint32              nbrBitmap;

    uint32              nbrInode;
    uint32              nbrInodeFlush;

    uint32              nbrBufFlush;
    uint32              nbrMarkDone;

    uint32              nbrCommit;
    uint32              nbrEOL;

    uint32              nbrEntries;
    uint32              nbrMaxTransaction;

    uint32              nbrCurrTransaction;
    uint32              MaxRecLen;

    uint32              nbrTruncateRecord;
    uint32              nbrTruncateCall;

    uint32              nbrTruncateMax;
    uint32              nbrExtraBitmapsCreate;

    uint32              nbrExtraBitmapsAccess;
    uint32              ioTotalLogWrites;

    uint32              ioPartialLogWrites;
    uint32              postponedPartialLogWrites;
    uint32              forceWrite;
    uint32              forceWriteDone;

#ifdef LOG_TUNE_STATISTICS
    uint32              nbrFixedInodeDep;
    uint32              nbrVariableInodeDep;
    uint32              maxInodeDep;
    uint32              minInodeDep;

    uint32              nbrFixedBufDep;
    uint32              nbrVariableBufDep;
    uint32              maxBufDep;
    uint32              minBufDep;

    uint32              nbrFixedioDoneUsed;
    uint32              nbrVariableioDoneUsed;
    uint32              ioDoneUsed;
    uint32              bufUsed;
    uint32              ioDoneMax;
    uint32              ioDoneMin;
#endif

    uint64              ioPartialBytes;

    uint64              ioTotalBytes;

    uint64              logTotalBytes;

} fs_LogStats;


/**
 * Structure for tracking state of an individual segment of
 * the log.
 */
typedef struct fs_LogSegment
{
    uint32      depCount;       // Count of outstanding dependencies.
    uint32      undoCount;      // Count of outstanding undo records.
    uint32      head;           // Valid head pointer when the previous
                                // segment is truncated.
    ushort16    active;         // If segment is currently active in the log.
    ushort16    segNum;         // Segment number this pertains to.
} fs_LogSegment;


/**
 * Per filesystem log root in memory.  This maintains/tracks the 
 * utilization for a single log.
 */
struct fs_LogFsRoot
{
    nfx_spinlock_t      spin;
    uint32              head;           // Address on the log device of the
                                        // head of the log.  This is a shadow
                                        // of **headPtr so we don't have to
                                        // read it from the device.
    uint32              *headPtr;       // Address of head pointer on the log
                                        // device.  The value at this address
                                        // always points to the oldest record
                                        // in the log.
    uint32              currLen;        // Space currently consumed by log
                                        // records.
    uint32              tail;           // Address on the log device where
                                        // the next record should be written.
    uint32              segmentSize;    // logSize / FS_LOG_MAX_SEGMENTS
    fs_LogSegment       logSeg[FS_LOG_MAX_SEGMENTS];  // Segment structs for
                                        // the log.
    uint32              nextId;         // Next id to use for log entries.
    int32               logDevType;     // Type of device being logged to.
    fs_LogReplay        *replay;
    union
    {
        nvram_handle_t  nvHandle;       // Handle if LOG_DEVTYPE_NVRAM
        evm_volId_t     volId;          // Id if LOG_DEVTYPE_VOLID
    } logDev;
    void                *volBufRoot;    // data strutures for volume logging
    fs_LogStats         logStats;       // log statistics
	boolean				inReplay;		// Replay currently in progress?
    fs_LogInfo          logInfo;
    boolean             skipReplay;
};

/**
 * Holds a list of COD's for a given transaction.
 * A list for used COD's and for removed COD's is needed.
 * An array and next pointer to facilitate expanding the number 
 * of COD's for a particular log entry.
 */
typedef struct fs_LogCodArr fs_LogCodArr;
struct fs_LogCodArr
{
#define FS_LOG_COD_ARR_MAX    8
    nfx_cod_t            cod[FS_LOG_COD_ARR_MAX];
    fs_LogCodArr         *next;
};


/**
 * Structures for maintaining in-progress replay state.
 */
#define FS_LOG_EXTRA_BITMAP      13      // (1 << size) number of bits
                                         // 13 gives 8192 bits = 1K bytes
                                         // !! warning must be at least 3 (8 bits, 1 byte)

typedef struct fs_LogReplayExtraBitmap fs_LogReplayExtraBitmap;

struct fs_LogReplayExtraBitmap 
{
    DLL_ENTRY(fs_LogReplayExtraBitmap)   entry;
    uint32                               bitmapId;
    uchar8                               bits[(1 << FS_LOG_EXTRA_BITMAP) / BITS_PER_BYTE];      // bitmap    
} ;

struct fs_LogReplayBitmap
{
    uchar8  *bits;      // bitmap
    int32   numBits;    // size of bitmap in bits
    uint32  firstId;    // transaction id at index 0
	uint32  maxForwIndex;
	uint32  maxRevIndex;
    DLL_HDR(fs_LogReplayExtraBitmap) extra;
};

struct fs_LogReplay
{
#define FS_LOG_SUMM_HASH_MAX    1024
#define FS_LOG_SUMMARY_HASH(_id_)                             \
    ((_id_) % FS_LOG_SUMM_HASH_MAX)

    DLL_HDR(fs_LogSummary)  summHdr[FS_LOG_SUMM_HASH_MAX];
#ifndef NEW_ZF
    DLL_HDR(fs_Buf)         zeroFillHdr;
#endif // NEW_ZF
    fs_LogReplayBitmap      bitmap;
    uint32                  firstDepOffset;
};

#ifdef NEW_ZF

#define FS_ZF_INUM_BUCKETS  FS_BUFLIST_HASH_MAX // use the same hashing
#define FS_ZF_PAGE_BUCKETS  FS_BUFLIST_HASH_MAX // use the same hashing

/* List of user data pages to zero per inodes per filesystem.
 */
//typedef struct fs_zfPage_t fs_zfPage;
struct fs_zfPage_t
{
    fs_LogSegmentDeps       logSegDep;  // Log dependencies
    evm_volId_t             volId;      // to hash only
    uint64                  inum;       // to hash only
    uint64                  page;
    uint32                  offset;
    DLL_ENTRY(fs_zfPage)    entry;      
};

//typedef struct fs_zfInode_t fs_zfInode;
struct fs_zfInode_t
{
    evm_volId_t             volId;      // to hash only
    uint64                  inum;
    DLL_HDR(fs_zfPage)      pageList;
    DLL_ENTRY(fs_zfInode)   entry;
};
#endif // NEW_ZF

/**
 * Log scan actions.
 */
typedef enum {
    /** scan the log for transactions which are done */
    FS_LOG_ACTION_DONE = 1,
    /** replay the log into cache */
    FS_LOG_ACTION_REPLAY,
} fs_LogAction;


/**
 * Defines for conducting log unit tests.
 */
#define FS_LOG_MAX_TESTS       1024
#define FS_LOG_DEFAULT_TESTS   100
#define FS_LOG_NPAGE           10


/**
 * Find/allocate and initialize a log device.
 */
extern int32
fs_logInitDevice(
    fs_Context      *context);

/**
 * Append a log record describing a block modification.
 */
extern int32
fs_logAddDataRecord(
    fs_Context      *context,
    fs_Buf          *buf,       // Buf that was modified.
    int32           offset,     // Byte offset in buf->data where modification
                                // begins (0 for user (cod) bufs).
    int32           length,     // Length in bytes of modification (0 for
                                // user (cod) bufs).
    uint64          srcBlkNum,  // Source/old blk number.  Used to identify
                                // source or "before COW" blknum.
                                // When adding a record for a COW block,
                                // buf must refer to new/copied blknum and have
                                // FS_BUF_COPIED set when calling this function!
    uint32          flags
#define FS_LOG_ADD_FLAG_ZERO    0x0001 // Zero-fill the buffer.
#define FS_LOG_ADD_FLAG_UNDO    0x0002 // Add an undo record to the buf.
);


/**
 * Append a log record describing a block
 * allocation.
 */
extern int32
fs_logAddAllocRecord(
    fs_Context      *context,
    fs_Buf          *buf,       // Buf that is being allocated for.
    boolean         isCow,      // Is this a COW alloc?
    fs_Block        oldBlock    // For COW allocs
);


/**
 * Append a log record describing a bitmap
 * modification.  
 */
extern int32
fs_logAddBitmapRecord(
    fs_Context      *context,
    fs_Buf          *buf,
    int32           offset,
    int32           length,
    uchar8          *bits,
    int32           op,
	fs_BitmapOperation *debugOp
);


/**
 * Log a modified inode.
 */
extern void fs_logAddInodeRecord(
    fs_Context  *context,
    fs_Inode    *inode);


/**
 * Log that an inode was removed
 */
extern void fs_logAddInodeFlushRecord(
    fs_Context  *context,
    fs_Inode    *inode);


/**
 * Log that one or more bufs were removed.
 */
void
fs_logAddBufFlushRecord(
    fs_Context  *context,
    int32       bufListType,
    uint64      inum,
    uint64      startPage,
    fs_Block    block);


/**
 * Add an exclusively held buf to a log summary.
 * This provides for unlocking the bufs during commit.
 * During replay, the logSummary is related to the transaction, 
 * not to the context.
 */
extern void
fs_logAddBufLock(
    fs_Context      *context,
    fs_LogSummary   *logSummary,
    fs_Buf          *buf
);


/**
 * Add an inode lock to the log summary for
 * the transaction.  This provides for unlocking the lock
 * during transaction commit.
 */
extern void
fs_logAddInodeLock(
    fs_Context      *context,
    fs_LogSummary   *logSummary,
    fs_Inode        *inode 
);

/**
 * Add an inode btree lock to the log summary for
 * the transaction.  This provides for unlocking the lock
 * during transaction commit.
 */
extern void
fs_logAddBtreeLock(
    fs_Context      *context,
    fs_LogSummary   *logSummary,
    fs_Inode        *inode 
);


/**
 * Create and commit a zero-fill transaction for the
 * specified buf.
 */
extern void
fs_logZeroFillAndCommit(
    fs_Context      *context,
    fs_Buf          *buf
);


/**
 * Commit a log entry to the log device at
 *              the end of a transaction.
 *
 * fs_logCommitEntry: is always stable  (FILE_SYNC)
 * fs_logCommitWrite: used in the nfs write thread
 *    stableHow indicates request, and can be overwritten
 * fs_logCommitEntryDone: if previous commit was unstable
 *    the user can send the commit req
 */
extern int32
fs_logCommitEntry(
    fs_Context        *context
);


/**
 * Commit outstanding log and user bufs for the
 * specified inode number.  Blocks until data is
 * persistent on disk if stableHow is FILE_SYNC.
 */
extern int32
fs_logCommitWrite(
    fs_Context        *context,
    fs_Inode          *inode,
    uint64            *inum,
    fs_WriteStableHow *stableHow);


/**
 * fs_logCommitEntry, fs_logCommitWrite, fs_logCommitEntryDone
 *
 * Called to commit a log entry to the log device at
 * the end of a transaction.
 *
 * fs_logCommitEntry: is always stable  (FILE_SYNC)
 * fs_logCommitWrite: used in the nfs write thread
 *    stableHow indicates request, and can be overwritten
 * fs_logCommitEntryDone: if previous commit was unstable
 *    the user can send the commit req
 *
 */
extern int32
fs_logCommit(
    fs_Context        *context,
    fs_LogStableWrite stable,
    uint64            inum);


/**
 * Commit outstanding log and user bufs for the
 * specified inode.  Blocks until data is persistent
 * on disk.
 */
extern int32
fs_logSyncInodeBuf(
    fs_Context        *context,
    fs_Inode          *inode);

extern void
fs_logSyncInodeLog(
    fs_Context        *context,
    uint64            inum);


/**
 * If the specified buf has not been allocated for, perform
 * allocation and if stableHow is UNSTABLE, record a zero-fill
 * log entry for it.
 */
extern void
fs_logWriteBuf(
    fs_Context        *context,
    fs_Inode          *inode,
    fs_Buf            *buf,
    fs_WriteStableHow stableHow);

/**
 * Create a nested transaction that completes the specified
 * transaction.
 * 
 * Caller must hold the logroot spinlock.
 */
extern void
fs_logComplete(
    fs_Context      *context,
    fs_LogId        transaction,
    uint64          inum);

/**
 * Transition the undo record to a segment dependency.
 */
extern void
fs_logConvertUndoToPerm(
    fs_Context          *context,
    fs_LogUndoSummary   *undo,
    fs_LogSegmentDeps   *logSegDep);

/**
 * Remove the specified dependencies from the
 * log.
 */
void
fs_logRemoveDeps(
    fs_Context          *context,
    fs_LogSegmentDeps   *logSegDep,
    nfx_spinlock_t      *spinp);

void
fs_logDoDepTransfer(
    fs_Context          *context,
    fs_LogSegmentDeps   *dst,
    fs_LogSegmentDeps   *src);
    
/**
 * Remove the specified buf from the log summary.
 * Used when we give a buffer to the WB during a transaction.
 * The WB becomes the owner and will release the buffer when ioDone.
 * The logcommit has nothing to do for this buffer anymore.
 *
 * Remark:  The bufArr can have holes as a result of this (NULL entries !!!)
 */
boolean
fs_logRemoveBufLogSummary(
    fs_Context      *context,
    fs_Buf          *buf
);


/**
 * Start a transaction log in the specified context, which must not have
 * a current log summary.
 */
extern void fs_logBeginTransaction(
    fs_Context *context);


/**
 * Start a nested transaction log in the specified context.
 */
extern void fs_logNestTransaction(
    fs_Context *context);


/**
 * Dump log data
 */
extern int32
fs_logMakeDumpLogRecord(
    fs_Context *context,
    char       *data,       // text string to write log records
    int32      remain       // remain size in text string
);


/**
 * Perform undo operations indicated in listBuf
 * on buf.
 */
extern void
fs_logUndoBuf(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_Buf      *listBuf
);


/**
 * Dump log stats
 */
extern void
fs_logPrintStats(
    fs_LogStats *stats,
    uint32 logLen);


/**
 * fs_logReplay
 *
 * Description: Called to replay the complete log when mounting a 
 *              dirty volume.
 *              This will put all the committed records found in the
 *              log-device into the cache.
 *
 */
extern int32
fs_logReplay(
    fs_Context      *context,
    boolean         skipReplay,
    fs_LogFsRoot    *logRoot);


/**
 * Transfer the specified log dependencies to the
 * specified object of the specified type.
 */
extern void
fs_logTransferDeps(
    fs_Context          *context,
    fs_LogSegmentDeps   *deps,
    int16               srcHeadSeg,
    void                *objectPtr,
    int32               objectType);


/**
 * Initialize the log segment size and each segment's
 * head pointer. 
 */
extern void
fs_logInitLogSegments(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          logSize);


/**
 * Returns TRUE if the specified logSegmentDeps
 * bitmap has segment set for the current
 * head of the log.
 */
extern boolean
fs_logAreDepsAtHeadSegment(
    fs_Context          *context,
    fs_FileSystem       *fileSystem,
    fs_LogSegmentDeps   *segDeps);


/**
 * Destroy the in-core log summary information for the
 * specified volume.
 *
 * This is called when a volume is being forcibly
 * unmounted.
 */
extern void
fs_logDestroyVolume(
    fs_Context      *context,
    fs_FileSystem   *fileSystem);


/**
 * Destroy the specifed logSummary, freeing it and memory
 * for all structures chained to it.
 */
extern void
fs_logDestroySummary(
    fs_Context      *context,
    fs_LogSummary   *logSummary);


/**
 * Indicate whether or not log size is above threshold
 * and writeback engine should be active.
 */
extern boolean
fs_logAboveWritebackThreshold(
    fs_Context      *context,
    fs_FileSystem   *fileSystem);


/**
 * Return the log segment number that the current head
 * of the log resides in.
 */
int32
fs_logGetHeadSegnum(
    fs_Context      *context,
    fs_FileSystem   *fileSystem);


/**
 * Log device type specific function table.
 */
typedef struct fs_LogDevTypeTable
{
    int32   logRecIdTrailerLen;
    int32   (* initDevice)(
                fs_Context          *context);
    int32   (* commitWrite)(
                fs_Context          *context,
                fs_Inode            *inode,
                fs_WriteStableHow   *stableHow);
    int32   (* commitTrans)(
                fs_Context          *context);
    int32   (* syncInodeBuf)(
                fs_Context          *context,
                fs_Inode            *inode);
    int32   (* syncInodeLog)(
                fs_Context          *context,
                uint64              inum);
    int32   (* writeDevHdr)(
                fs_Context          *context,
                int32               operation,
                uint32              count);
    int32   (* recordCksum)(
                fs_Context          *context,
                fs_LogRecHdr        *recordHdr,
                int32               recordHdrLen,
                void                *data,
                int32               dataLen);
    void    (* logWrite)(
                fs_Context          *context,
                uint32              *offset,
                uchar8              *datap[LOG_IOVEC_MAX],
                uint32              lenp[LOG_IOVEC_MAX],
                fs_LogStableWrite   stable,
                uint64              inum);
    int32   (* logReadInit)(
                fs_Context          *context,
                fs_LogInfo          *info);
    int32   (* logRead)(
                fs_Context          *context,
                uchar8              **datap,
                uint32              offset,
                uint32              *bytesRead,
                uint32              tail);
    void    (* logReadDone)(
                fs_Context          *context,
                uchar8              *datap);
    int32   (* verifyCksum)(
                fs_Context          *context,
                uint32              *data,
                uint32              length);
    int32   (* logIsEmpty)(
                fs_Context          *context);
} fs_LogDevTypeTable;
extern fs_LogDevTypeTable fs_logDevTypeTable[FS_LOG_DEVTYPE_MAX];

extern int32
fs_logNullInitDevice(
    fs_Context  *context);

extern int32
fs_logNopCommitTrans(
    fs_Context *context);

extern int32
fs_logNullCommitWrite(
    fs_Context          *context,
    fs_Inode            *inode,
    fs_WriteStableHow   *stableHow);

extern int32
fs_logNopSyncInodeLog(
    fs_Context  *context,
    uint64      inum);

extern int32
fs_logNopSyncInodeBuf(
    fs_Context  *context,
    fs_Inode    *inode);

extern int32
fs_logNopWriteDevHdr(
    fs_Context  *context,
    int32       operation,
    uint32      count);

extern int32
fs_logNopRecordCksum(
    fs_Context      *context,
    fs_LogRecHdr    *recordHdr,
    int32           recordHdrLen,
    void            *data,
    int32           dataLen);

extern void
fs_logNullWrite(
    fs_Context          *context,
    uint32              *offset,
    uchar8              *datap[LOG_IOVEC_MAX],
    uint32              lenp[LOG_IOVEC_MAX],
    fs_LogStableWrite   stable,
    uint64              inum);

extern void
fs_logTestForceCrash(
    fs_Context  *context,
    int32       crashPhase,
    int32       n);

extern int32
fs_logTestReplay(
    fs_Context  *context,
    int32       crashPhase,
    int32       n);

extern int32
fs_logNullReadInit(
    fs_Context  *context,
    fs_LogInfo  *info);

extern int32
fs_logNullRead(
    fs_Context  *context,
    uchar8      **datap,
    uint32      offset,
    uint32      *bytesRead,
    uint32      tail);

extern void
fs_logNullReadDone(
    fs_Context  *context,
    uchar8      *datap);

extern int32
fs_logNopVerifyCksum(
    fs_Context  *context,
    uint32      *data,
    uint32      length);

extern int32
fs_logNopIsEmpty(
    fs_Context  *context);

extern void
fs_logDumpIncoreData(
	fs_Context		*context,
	fs_LogSummary	*logSummary);

extern void
fs_logStopDaemon(
    fs_Context  *context,
    boolean     isForce);

/**
 * Return TRUE if seg1 is older than seg2.  If its
 * equal or newer, return FALSE.
 */
extern boolean
fs_logSegIsOlder(
    fs_LogFsRoot    *logRoot,
    int16           seg1,
    int16           seg2);

extern void
fs_logReplayMarkBufDirty(
	fs_Context	*context,
	fs_BufList	*bufList, 
	fs_Buf		*buf);

extern void
fs_logReplayFindBuf(
    fs_Context      *context,
    fs_LogReplay    *replay,
    uint64          inum, 
    uint64          page,
    fs_Block        block,
    boolean         isMeta,
    boolean         doRead,
    fs_Buf          **bufp);

#ifdef NEW_ZF
extern void
fs_zfInit(void);

extern void
fs_logAddUserZeroRecord(fs_Context *    context,
                        fs_Inode *      inode,
                        uint64          startPage,
                        uint32          offsetInFirstPage,
                        uint32          nbPage,
                        boolean         doZero,
                        boolean         doCommit,
                        fs_Buf **       bufList);

extern void
fs_zfZeroPage(fs_Context *  context,
              fs_Inode *    inode,
              fs_zfPage *   zfPage);

extern void
fs_userZeroThread(esm_handle_t  handle,
                  void *        arg);


extern fs_zfInode *
fs_zfGetInode(fs_Context *  context,
              uint64        inum);

extern fs_zfPage *
fs_zfGetPage(fs_Context *   context,
             uint64         inum,
             uint64         page);

extern fs_zfInode *
fs_zfAddInode(fs_Context *  context,
              uint64        inum);

extern void
fs_zfAddPage(fs_Context *       context,
             uint64             inum,
             uint64             startPage,
             uint32             offsetInFirstPage,
             uint32             nbPage,
             fs_LogRecord *     record,
             fs_LogObjectSumm * obj);

extern void
fs_zfDelInode(fs_Context *  context,
              uint64        inum);

extern void
fs_zfDelPage(fs_Context *   context,
             uint64         inum,
             uint64         startPage,
             boolean        tillEnd);

void
fs_zfDisplayList(fs_Context *   context);

uint32
fs_zfInumHash(const void * key);

int32
fs_zfInumCompare(const void * obj1,
                 const void * obj2);

void
fs_zfInumRelease(void * key,
                 void * obj);

uint32
fs_zfPageHash(const void * key);

int32
fs_zfPageCompare(const void * obj1,
                 const void * obj2);

void
fs_zfPageRelease(void * key,
                 void * obj);

#endif // NEW_ZF

#endif /* FS_LOG_H */
