/*
 * Name:  fs-proto.h
 *
 * RCS:  $Id: fs-proto.h,v 1.135 2002/12/10 22:25:39 bobonfeline Exp $
 *
 * Description:  Filesystem internal function prototypes.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_PROTO_H)
#define FS_PROTO_H

#if defined(FS_STANDALONE)
// linux build
#include <string.h>

#else
#if defined SSC
// BSD build
#include <string.h>

#else
// embedded (FC, TXRX, FP)
#include <exports/libc.h>
#endif
#endif

 
/**
 * Return the time in seconds since the Unix epoch
 */
/*-----------------------------------------------------------------------------
 * fs_getTimeOfDaySecs
 *
 * Description: Returns current time since January 1, 1970 in seconds.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
static inline uint32
fs_getTimeOfDaySecs()
{
    eeeGMT_t now;
    
    eee_rtcGMT (&now);
    return now.sec;
} /* fs_getTimeOfDaySecs */


/**
 * Set the time in seconds and useconds since the Unix epoch
 */
static inline fs_Time
fs_getTimeOfDay(void)
{
    fs_Time time;
#ifdef FS_STANDALONE
    struct timeval  now;

    gettimeofday(&now, NULL);
    time.secs = now.tv_sec;
    time.nsecs = now.tv_usec * 1000;
#else // FS_STANDALONE
    eeeGMT_t    now;

    eee_rtcGMT(&now);
    time.secs = now.sec;
    time.nsecs = now.usec * 1000;
#endif // FS_STANDALONE
    return time;
} /* fs_getTimeOfDay */


/**
 * Print a debug message, if the _app_type_ trace is enabled
 */
#ifdef FS_DEBUG
#define FS_TRACE(_context_, _app_type_, _parameters_)                         \
do {                                                                          \
    if ((elog_enabled) &&                                                     \
        (elog_severity >= LOG_DEBUG))                                         \
    {                                                                         \
       if ((elog_appl_data[fs_applicationId].severity >= LOG_DEBUG) &&        \
           (elog_appl_data[fs_applicationId].bitmap & (1 << _app_type_)))     \
       {                                                                      \
          char *the_msg ;                                                     \
          the_msg = e_log_var_pa  _parameters_ ;                              \
          fs_contextLog(_context_, LOG_DEBUG, _app_type_,                     \
                        __LINE__, __FILE__, the_msg);                         \
          eee_ramDealloc(the_msg);                                            \
       }                                                                      \
    }                                                                         \
} while (0)
#else
#define FS_TRACE(_context_, _app_type_, _parameters_) 
#endif

/**
 * Print a syslog message, if the logging severity and trace level are
 * sufficient.
 */
#define FS_ELOG(_context_, _severity_ , _app_type_, _parameters_)           \
do {                                                                          \
    if ((elog_enabled) &&                                                     \
        (_severity_ <= elog_severity))                                        \
    {                                                                         \
        if ((_severity_ < LOG_DEBUG) ||                                       \
            ((elog_severity >= LOG_DEBUG) &&                                  \
             (_severity_ >= LOG_DEBUG) &&                                     \
             (elog_appl_data[fs_applicationId].severity >= _severity_) &&     \
             (elog_appl_data[fs_applicationId].bitmap & (1 << _app_type_))))  \
       {                                                                      \
          char *the_msg ;                                                     \
          the_msg = e_log_var_pa  _parameters_ ;                              \
          fs_contextLog(_context_, _severity_, _app_type_,                    \
                        __LINE__, __FILE__, the_msg);                         \
          eee_ramDealloc (the_msg);                                           \
       }                                                                      \
    }                                                                         \
} while (0)

/**
 * Use elog to send a trace message, if the right flags are set.
 * Used by the FS_TRACE_MACRO
 */
void fs_trace_elog(
    fs_Context *context,
    int32      level,
    int        line, 
    char*      file,
    char       *traceString);

/**
 * Ram alloc/dealloc wrappers for helping to
 * find what's being used.
 *	// Do not add to this list without updating
 *	// fs_ramTypeName[]!!!
 */
typedef enum fs_RamAllocType
{
    FS_RAM_REQ_MSG,
    FS_RAM_RESP_MSG,
    FS_RAM_NFS_RESP_ARG,
    FS_RAM_COD_DATA,
    FS_RAM_CONTEXT,
    FS_RAM_BUF_PTR,
    FS_RAM_BUF_SHADOW_DATA,
    FS_RAM_BUF_PAGE_HDR,
    FS_RAM_INODE,
    FS_RAM_LOG_SUMM,
    FS_RAM_LOG_DEP,
    FS_RAM_LOG_SUMM_ARR,
    FS_RAM_LOG_BUF_ARR,
    FS_RAM_LOG_INO_ARR,
    FS_RAM_LOG_REC,
    FS_RAM_RDWR,
    FS_RAM_LOCK_WAIT,
    FS_RAM_BUF_DATA,
    FS_RAM_READ_DATA,
    FS_RAM_WRITE_DATA,
    FS_RAM_MKFS_DATA,
    FS_RAM_SIM_MSG_QENTRY,
    FS_RAM_SIM_MISC,
    FS_RAM_LOG_READ_BUF,
    FS_RAM_LOG_WRITE_BUF,
    FS_RAM_CORE_MISC,
    FS_RAM_FS_COD,
    FS_RAM_MCPU,
    FS_RAM_SUPER,
    FS_RAM_ARENA,
	FS_RAM_QUOTA,
	FS_RAM_FILESET,
	FS_RAM_FSET_CONTEXT,
	FS_RAM_QUOTA_CONTEXT,
	FS_RAM_QUOTA_CONTEXT_LIST,
	FS_RAM_MAKEPATH,
	FS_RAM_EXTERNAL,
	FS_RAM_SKIPHEAP,
    FS_RAM_DUMP_DIRENT,
    FS_RAM_DUMP_RENAMELINK,
    FS_RAM_DUMP_NAME,
    FS_RAM_DUMP_NAMEENTRY,
    FS_RAM_DUMP_CONTEXT,
    FS_RAM_RESTORE_ATTR,
    FS_RAM_RESTORE_DATA,
    FS_RAM_RESTORE_NAME,
    FS_RAM_RESTORE_DIRENTS,
    FS_RAM_RESTORE_CONTEXT,
    FS_RAM_BUF_CLONE,
    FS_RAM_BUF_CLONE_DATA,
    FS_RAM_AM_SOURCE_SESSION,
    FS_RAM_AM_TARGET_SESSION,
    FS_RAM_SNAP_NAME,
    FS_RAM_AM_LAST_SNAP,
    FS_RAM_EXTENT_VECTOR,
    FS_RAM_REGION_SUMMARY,
    FS_RAM_LOG_BUF,
    FS_RAM_GET_FILESET_PATH,
    FS_RAM_LOG_UNDO_SUMM,
    FS_RAM_LOG_BUF_DEP,
    FS_RAM_DIR_LEAF,
    FS_RAM_RESTORE_IODONE,
    FS_RAM_WAITER,
    FS_RAM_PENDING_REQUESTS,
    FS_RAM_BENCHMARK_DESC,
    FS_RAM_CTX_SHARED_LOCKS,
    FS_RAM_RESTORE_IODONE_Q,
    FS_RAM_FORK_CONTEXT,
    FS_RAM_FULL_FH_CONTEXT,
    FS_RAM_FILE_HISTORY_QUEUE,
    FS_RAM_READ_AHEAD,
    FS_RAM_SDESC,
    FS_RAM_DUMP_IOBUF,
    FS_RAM_COMPACT_LEAF,
    FS_RAM_READDIR_ENTRIES,
    FS_RAM_READDIR_CACHEENT,
    FS_RAM_DNLC_ENTRY,
    FS_RAM_ZF,  // zero fill
    FS_RAM_VLC, // volume log cache
    FS_RAM_DIRHASH_TP,
    FS_RAM_DIRHASH_TABLE,
    FS_RAM_DUMP_SESSION,
	// Do not add to this list without updating
	// fs_ramTypeName[]!!!
    FS_RAM_TYPE_MAX
} fs_RamAllocType;

#define FS_RAM_STATS_ALLOC    0
#define FS_RAM_STATS_DEALLOC  1
#define FS_RAM_STATS_HOW_MAX  2
extern uint32 fs_ramStats[FS_RAM_STATS_HOW_MAX][FS_RAM_TYPE_MAX];
extern nfx_spinlock_t fs_ramStatsSpinId;


#ifndef SBM_REV_2
EXPORTED int32
fs_sacStoreInstantiate(
    sbm_ack_msg_t    *msg,
    int32           bufType,
    uint32          src
);
#endif

#if ! defined(SSC)

/*
 * dump mem usage
 */
extern void
fs_dumpAlloc(int32 what, int32 numBytes);

#ifdef FS_USE_MALLOC_SLAB
extern struct mm_cache *fs_compact_leaf_cache;
extern struct mm_cache *fs_bufptr_cache;
extern struct mm_cache *fs_bufpagehdr_cache;
extern struct mm_cache *fs_context_cache;
extern struct mm_cache *fs_inode_cache;
extern struct mm_cache *fs_logsummary_cache;
extern struct mm_cache *fs_vl_unstableFH_cache;
extern struct mm_cache *fs_log_buf_cache;
extern struct mm_cache *fs_log_undo_summ_cache;
extern struct mm_cache *fs_waiter_cache;
#endif

extern uint32 fs_ramStats[FS_RAM_STATS_HOW_MAX][FS_RAM_TYPE_MAX];
extern nfx_spinlock_t fs_ramStatsSpinId;

/**
 * Dynamically allocate memory.
 *
 * The what parameter indicates the purpose of the allocated
 * memory and is used for profiling and statistics gathering.
 *
 * define as inline, so that switch is resolved at compile time
 */
static inline void *
fs_doRamAlloc(int32 what, int32 numBytes, boolean allowFail)
{
    char *p = NULL;
#if !defined(FS_STANDALONE)
    void *ra = (void *) (address_t) get_ra();
#endif

#if defined(FS_STANDALONE)
    p = (char*)fs_simAlloc(numBytes);
#else
    switch (what)
    {
#ifdef FS_USE_MALLOC_SLAB
    case FS_RAM_COMPACT_LEAF:
        assert(numBytes == fs_compact_leaf_cache->c_objsize);
        p = cache_alloc_ra(fs_compact_leaf_cache, ra);
        break;
    case FS_RAM_BUF_PTR:
        assert(numBytes == fs_bufptr_cache->c_objsize);
        p = cache_alloc_ra(fs_bufptr_cache, ra);
        break;
    case FS_RAM_BUF_PAGE_HDR:
        assert(numBytes == fs_bufpagehdr_cache->c_objsize);
        p = cache_alloc_ra(fs_bufpagehdr_cache, ra);
        break;
    case FS_RAM_LOG_SUMM:
        assert(numBytes == fs_logsummary_cache->c_objsize);
        p = cache_alloc_ra(fs_logsummary_cache, ra);
        break;
    case FS_RAM_INODE:
        assert(numBytes == fs_inode_cache->c_objsize);
        p = cache_alloc_ra(fs_inode_cache, ra);
        break;
    case FS_RAM_CONTEXT:
        // ctor overhead
#ifdef MM_VERIFY_ALLOCS
        assert(numBytes == fs_context_cache->c_objsize - sizeof(void*) * 2);
#else
        assert(numBytes == fs_context_cache->c_objsize - sizeof(void*));
#endif
        p = cache_alloc_ra(fs_context_cache, ra);
        break;
    case FS_RAM_LOG_WRITE_BUF:
        assert(numBytes == fs_vl_unstableFH_cache->c_objsize);
        p = cache_alloc_ra(fs_vl_unstableFH_cache, ra);
        break;
    case FS_RAM_LOG_BUF:
        assert(numBytes == fs_log_buf_cache->c_objsize);
        p = cache_alloc_ra(fs_log_buf_cache, ra);
        break;
    case FS_RAM_LOG_UNDO_SUMM:
        assert(numBytes == fs_log_undo_summ_cache->c_objsize);
        p = cache_alloc_ra(fs_log_undo_summ_cache, ra);
        break;
    case FS_RAM_WAITER:
        assert(numBytes == fs_waiter_cache->c_objsize);
        p = cache_alloc_ra(fs_waiter_cache, ra);
        break;
#endif

    case FS_RAM_BUF_DATA:
    case FS_RAM_BUF_CLONE_DATA:
    case FS_RAM_READ_DATA:
    case FS_RAM_WRITE_DATA:
    case FS_RAM_SUPER:
        FS_ASSERT(numBytes <= FS_BLOCK_SIZE);

        p = (char*)eee_allocateBuffer(EEE_SHARED_POOL,
                                      EEE_BUF_ULTRA, NFX_COS_HIGH);
        // We have an ultra buffer, preserve the unused leader for
        // packet headers.
        if (p != NULL)
        {
#if defined(EEE_NEW_BUFSIZE)
            p += FS_ULTRA_BUF_OFFSET;
#else			
            p += EEE_BUFSIZE_ULTRA - FS_BLOCK_SIZE;
#endif			
        }
        break;
        
    default:
        p = eee_ramAllocRA(numBytes, ra);
        break;
    }
    
#endif

    if (p == NULL && !allowFail)
	{
        fs_dumpAlloc(what, numBytes);
    }
    else
    {
        NFX_SPIN_LOCK(&fs_ramStatsSpinId);
        fs_ramStats[FS_RAM_STATS_ALLOC][what] += 1;
        NFX_SPIN_UNLOCK(&fs_ramStatsSpinId);
    }
        
    return (void*)p;
}

/**
 * Dynamically allocate memory.  Will not return if memory is not
 * available.  The what parameter indicates the purpose of the allocated
 * memory and is used for profiling and statistics gathering.
 */
static inline void *
fs_ramAlloc(int32 what, int32 numBytes)
{
    return fs_doRamAlloc(what, numBytes, FALSE);
}


/**
 * Dynamically allocate memory.  WILL return if memory is not
 * available.  The what parameter indicates the purpose of the allocated
 * memory and is used for profiling and statistics gathering.
 */
static inline void *
fs_ramAllocFail(int32 what, int32 numBytes)
{
    return fs_doRamAlloc(what, numBytes, TRUE);
}


/**
 * Free dynamically allocated memory.  The what parameter indicates the
 * purpose of the allocated memory and is used for profiling.
 */
static inline void
fs_ramDealloc(int32 what, void *ptr)
{
#if !defined(FS_STANDALONE)
    void *ra = (void *) (address_t) get_ra();
#endif

    NFX_SPIN_LOCK(&fs_ramStatsSpinId);
    fs_ramStats[FS_RAM_STATS_DEALLOC][what] += 1;
    NFX_SPIN_UNLOCK(&fs_ramStatsSpinId);

#if defined(FS_STANDALONE)
    fs_simDealloc(ptr);
#else
    switch (what)
    {
#ifdef FS_USE_MALLOC_SLAB
    case FS_RAM_COMPACT_LEAF:
        cache_free_ra(fs_compact_leaf_cache, ptr, ra);
        break;
    case FS_RAM_BUF_PTR:
        cache_free_ra(fs_bufptr_cache, ptr, ra);
        break;
    case FS_RAM_BUF_PAGE_HDR:
        cache_free_ra(fs_bufpagehdr_cache, ptr, ra);
        break;
    case FS_RAM_LOG_SUMM:
        cache_free_ra(fs_logsummary_cache, ptr, ra);
        break;
    case FS_RAM_INODE:
        cache_free_ra(fs_inode_cache, ptr, ra);
        break;
    case FS_RAM_CONTEXT:
        cache_free_ra(fs_context_cache, ptr, ra);
        break;
    case FS_RAM_LOG_WRITE_BUF:
        cache_free_ra(fs_vl_unstableFH_cache, ptr, ra);
        break;
    case FS_RAM_LOG_BUF:
        cache_free_ra(fs_log_buf_cache, ptr, ra);
        break;
    case FS_RAM_LOG_UNDO_SUMM:
        cache_free_ra(fs_log_undo_summ_cache, ptr, ra);
        break;
    case FS_RAM_WAITER:
        cache_free_ra(fs_waiter_cache, ptr, ra);
        break;
#endif

    case FS_RAM_BUF_DATA:
    case FS_RAM_BUF_CLONE_DATA:
    case FS_RAM_READ_DATA:
    case FS_RAM_WRITE_DATA:
    case FS_RAM_SUPER:
        eee_deallocateBufferRA(ptr, (address_t) ra);
        break;
        
    default:
        eee_ramDeallocRA(ptr, ra);
        break;
    }
#endif
}

#endif // SSC

/**
 * Handle an exception condition for the specified
 * volume.
 */
EXPORTED void
fs_volumeException(
    fs_Context      *context,
    fs_FileSystem   *fs,
    int32           errorType);

/**
 *  fs error codes to error strings conversion
 */   
EXPORTED char *fs_errCodeToString(fs_Status status);
        
/*
 * Quota related
 */

EXPORTED void
fs_holdQuota(
	fs_QuotaIncoreRecord_t	*quota);

EXPORTED void
fs_releQuota(
	fs_QuotaIncoreRecord_t	*quota);

EXPORTED fs_QuotaIncoreRecord_t *
fs_getUGQuota(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*fset,	// Fileset on which we are working
	int							type,	// whether user or group
	uint32						id,		// uid or gid
    boolean                     failIfDoesNotExist
);

EXPORTED fs_QuotaDiskRecord_t *
fs_getUGSnappedQuota(
	fs_Context					*context,
	fs_FileSystem				*core,
    uint64                      snapnum,
	fs_FileSetDiskRecord_t	    *fset,	// Fileset on which we are working
	int							type,	// whether user or group
	uint32						id		// uid or gid
);
EXPORTED uint64
fs_getQuotaReservation(
	fs_Context	*context,	
	fs_FileSystem	*core,		
	fs_Inode	*inode,	// Inode which has blocks being assigned
						// or directory entry being added
	uint64		count,	// count of resources required
	int			flag,	// count inodes or blocks
	uint32		uid,	// uid from the creditials
	uint32		gid,	// primary gid from the credentials
    int         alwaysSucced,   // succeed reservation inspite of excess
	fs_QuotaContextList_t	**list	// return the contextlist
);

EXPORTED void
fs_setUGQuota(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*fset,	// File set we are operating on
	fs_QuotaBlk_t				*qblk	// quota we want to set
);

EXPORTED fs_Status
fs_setFileSetQuota(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*fset,	// fileset we are working on
	fs_QuotaBlk_t				*qblk	// new quota information
);

EXPORTED void
fs_freeQuotaResources(
	fs_Context	*context,	
	fs_FileSystem	*core,		
	fs_Inode	*inode,
	uint64		count,	// count of resources required
	int			flag,	// count inodes or blocks
	uint32		uid,
	uint32		gid,
	fs_QuotaContextList_t	**list	// return the contextlist
);

EXPORTED void
fs_freeQuotaContextList(
	fs_Context		*context,
	fs_FileSystem		*core,	// fileSystem
	fs_QuotaContextList_t	*q	// free this context list
);

EXPORTED void
fs_trimQuotaReservation(
	fs_Context				*context,
	fs_FileSystem				*core,
	fs_QuotaContextList_t	*contextList,
	uint64					amount);

EXPORTED void
fs_commitContextList(
	fs_Context				*context,
	fs_FileSystem				*core,
	fs_QuotaContextList_t	*contextList,
	boolean					otherCommitFollow);	//set if fs_commitContextList is
												//to be called again in the thread

EXPORTED void
fs_QuotaAdjustmentNoCheck(
	fs_Context	*context,	
	fs_FileSystem	*core,		
	fs_Inode	*inode,		// Inode changed
	fs_Inode	*newInode,	// New Inode created, if count inodes
	int64		count,	// resources (if + : required, if - : to be released)
	int			flag,	// count inodes or blocks
	uint32		uid,
	uint32		gid,
	fs_QuotaContextList_t	**list	// return the contextlist
);

/*
 * Fileset related
 */
EXPORTED void
fs_holdFileSet(
	fs_Context			*context,	// context
	fs_FileSystem			*core,		// core
	fs_FileSetIncoreRecord_t	*fset	// fileset to be freed
);


EXPORTED void
fs_releFileSet(
	fs_Context			*context,	// context
	fs_FileSystem			*core,		// core
	fs_FileSetIncoreRecord_t	*fset	// fileset to be freed
);

EXPORTED int
fs_isFileSetHeld(
	fs_Context			*context,	// context
	fs_FileSystem			*core,		// core
	fs_FileSetIncoreRecord_t	*fset	// fileset to be freed
);

EXPORTED fs_FileSetIncoreRecord_t *
fs_getFset(
	fs_Context	*context,
	fs_FileSystem	*core,
	fs_Inode	*inode	// Get fileset belonging to this inode
);

EXPORTED fs_FileSetDiskRecord_t *
fs_getSnappedFset(
	fs_Context	*context,
	fs_FileSystem	*core,
	fs_Inode	*inode	// Get fileset belonging to this inode
);

EXPORTED void
fs_buildFileSetList(
	fs_Context	    *context,
	fs_FileSystem	*core
);

EXPORTED void
fs_cleanFileSetNQuotaLists(
	fs_Context      *context,
	fs_FileSystem	*core
);

EXPORTED fs_FileSetIncoreRecord_t *
fs_makeDefaultFileSet(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*parent,	// parent of the fileset
	fs_Inode					*fsetRoot,	// Fileset for this inode
	fs_Inode					*fsetInfoInode	// fileset info file inode
);
EXPORTED fs_FileSetIncoreRecord_t *
fs_makeFileSet(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*parent,	// parent of the fileset
	fs_Inode					*fsetRoot,	// Fileset for this inode
	fs_Inode					*fsetInfoInode	// fileset info file inode
);

EXPORTED void
fs_removeFileSet(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*fset,	// fileset being removed
	fs_Inode					*root,	// root of the fileset
	fs_Inode					*fsetInfoInode	// fileset info file
);

EXPORTED fs_Status
fs_findFset(
	fs_Context					*context,
	fs_FileSystem					*core,
	char						*pt,
	fs_Inode					**root,
	fs_FileSetIncoreRecord_t	**rfset);

EXPORTED int
fs_fsetInfoHasFreeFsets(
	fs_Context		*context,
	fs_FileSystem		*core,
	fs_Inode		*fsetInfoInod);
EXPORTED int32
fs_listFileSet(
	fs_Context					*context,	
	fs_FileSystem					*core,
	fs_MsgListFileSetsReq_t		*req,
	fs_MsgListFileSetsResp_t		*resp
);

/*
 * Turn quota on a given fileset
 */

EXPORTED void
fs_quotaOn(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*fset
);

EXPORTED void
fs_quotaOff(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*fset
);

EXPORTED int32
fs_reportFileSetQuota(
	fs_Context					*context,	
	fs_FileSystem					*core,
	fs_MsgReportQuotaReq_t		*req,
	fs_MsgReportQuotaResp_t		*resp
);

EXPORTED int32
fs_reportUGQuota(
	fs_Context					*context,	
	fs_FileSystem					*core,
	fs_MsgReportQuotaReq_t		*req,
	fs_MsgReportQuotaResp_t		*resp
);

/*
 * File related. Probably duplicates
 */


EXPORTED fs_Inode *
fs_create(
	fs_Context	  *context,
	fs_FileSystem *fs,
	uint64		  snapnum,	// snapshot number
	fs_Inode	  *parent,	// parent inode
	char		  *name,	// name of the file/dir to be created
	int			   mode		// mode bits for the new file
);

EXPORTED fs_Inode *
fs_lookup(
	fs_Context	  *context,
	fs_FileSystem *fs,
	uint64		  snapnum,	// snapshot number
	fs_Inode	  *parent,	// directory in which search is made
	char		  *name		// name of the file/dir being lookedup
);

EXPORTED fs_Inode *
fs_lookupKeepParentLocked(
	fs_Context	  *context,
	fs_FileSystem *fs,
	uint64		  snapnum,	// snapshot number
	fs_Inode	  *parent,	// directory in which search is made
	char		  *name		// name of the file/dir being lookedup
);
EXPORTED fs_Inode *
fs_lookupShared(
	fs_Context	  *context,
	fs_FileSystem *fs,
	uint64		  snapnum,	// snapshot number
	fs_Inode	  *parent,	// directory in which search is made
	char		  *name		// name of the file/dir being lookedup
);

/*
 * Path related
 */

EXPORTED fs_path_t *
fs_makePath(char *data);

EXPORTED void
fs_destroyPath(fs_path_t *path);

EXPORTED char *
fs_getNextComponent(fs_path_t *path);

/**
 * Return a random 64-bit integer, using the state in the specified
 * context, between 0 and the specified upper limit.
 */
extern uint64 fs_random64(
    fs_Context *context,
    uint64     max);


/**
 * Quiesce the file system from modification.
 */
extern void fs_freeze(
    fs_Context *context);


/**
 * Allow modification.
 */
extern void fs_thaw(
    fs_Context *context);


/**
 * Quiesce the file system from access.
 */
extern void fs_freezeAll(
    fs_Context *context);


/**
 * Allow access.
 */
extern void fs_thawAll(
    fs_Context *context);


/**
 * Return true if the name is "." or ".."
 */
extern boolean fs_nameIsDotOrDotDot(
    char *);


/**
 * Return NFX_OK is the name passed in is valid for
 * creation/removal, else return the error appropriate
 * for the violation.
 */
extern int32
fs_nameIsValid(
    char    *name,
    int32   length);


/*
 * NDMP related protos that are called only by fs
 */
extern int32 fs_dumpSetupThread( fs_Context      *context);
extern int32 fs_dumpNamesThread( fs_Context      *context);
extern int32 fs_dumpStartThread( fs_Context          *context);
extern int32 fs_dumpIODoneThread( fs_Context          *context);
extern int32 fs_dumpAbortThread( fs_Context          *context);
extern fs_Status fs_ndmpWriteTape( fs_Context *context,
                                 fs_DumpContext_t *dumpContext,
                                 char *buf);
extern int32 fs_ndmpSendDumpLog(
    fs_Context              *context,
    fs_DumpContext_t        *dumpContext,
    fs_Status               status,
    int32                   type,
    fs_Status               fileStatus,
    char                    *fileName,
    char                    *message
);
#if 0
extern int32
fs_ndmpSendDumpFileHistory(
    fs_Context              *context,
    fs_DumpContext_t        *dumpContext,
    char                    *name,
    fs_Inumber              parent,
    fs_Inode                *inode
);
extern int32
fs_ndmpSendDumpFileHistoryParentsAlso(
    fs_Context              *context,
    fs_DumpContext_t        *dumpContext,
    char                    *name,
    fs_Inode                *inode
);
extern int32
fs_ndmpSendDumpFileHistoryParent(
    fs_Context              *context,
    fs_DumpContext_t        *dumpContext,
    char                    *name
);
#endif
extern int32
fs_ndmpSetDumpDate(
    fs_Context              *context,
    fs_DumpContext_t        *dumpContext,
    char                    *snapname,
    uint32                  date
);
extern int32
fs_ndmpSendDoneOrAbort(
    fs_Context              *context,
    fs_DumpContext_t        *dumpContext,
    fs_Status               status
);
extern int32
fs_restoreSetupThread(
    fs_Context      *context
);
extern int32
fs_restoreNamesThread(
    fs_Context      *context
);
extern int32
fs_restoreStartThread(
    fs_Context          *context
);
extern int32
fs_restoreIODoneThread(
    fs_Context          *context
);
extern int32
fs_restoreAbortThread(
    fs_Context          *context
);
extern int32
fs_ndmpReadTape(
    fs_Context              *context,
    fs_RestoreContext_t     *restoreContext
);
extern int
fs_ndmpFetchTapeData(fs_Context	*context, fs_RestoreContext_t *restoreContext);

extern int32
fs_ndmpSendRestoreLog(
    fs_Context              *context,
    fs_RestoreContext_t        *restoreContext,
    fs_Status               status,
    int32                   type,
    fs_Status               fileStatus,
    char                    *fileName,
    char                    *message
);
extern int32
fs_ndmpSendRestoreDoneOrAbort(
    fs_Context              *context,
    fs_RestoreContext_t        *restoreContext,
    fs_Status               status
);
int32
fs_getHandleFromPath(
    fs_Context  *context,
    char        *pathName,
    fs_Handle   *handle
);
boolean
fs_permissionOkay(
    fs_Context  *context,
    fs_Inode    *inode,             /* file or dir for operation */
    int32       type
);
extern void
fs_initAllocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem
);
extern void
fs_reallocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem
);
extern void
fs_freeAllocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem
);
extern void
fs_initShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem
);

extern void
fs_removeNonDirNode(
    fs_Context  			*context,
    fs_Inode    			*parentInode,
    fs_Inode    			*inode,
    char        			*name,
    int32       			nameLen,
	fs_QuotaContextList_t   **rlist1,
	fs_QuotaContextList_t   **rlist2,
	fs_QuotaContextList_t   **rlist3,
    boolean     *wasDestroyed
);


/**
 * Initialize memory allocator.
 */
extern void fs_memInit(void);


/**
 * Search for memory leaks.
 */
extern void fs_memCheckLeaks(void);

extern int fs_sendNdmpResponse(fs_Context *, fs_ReqType, uint32, uint64,
                               fs_Status);
/**
 * Remove a directory node from parent dir
 */
extern void
fs_removeDirNode(
    fs_Context  			*context,
    fs_Inode    			*parentInode,
    fs_Inode    			*inode,
    char        			*name,
    int32       			nameLen,
	fs_QuotaContextList_t   **rlist1,
	fs_QuotaContextList_t   **rlist2,
	fs_QuotaContextList_t   **rlist3);


/**
 * Grow the volume if required
 * Returns TRUE when needed
 */
extern boolean fs_growIfNeeded(fs_Context *c);

extern void fs_destroyRestoreContextOnUmount(fs_Context  *context);
extern void fs_destroyDumpContextOnUmount(fs_Context  *context);

extern void fs_yieldTemp(fs_Context	*context);
extern void fs_yieldPeriodic(fs_Context	*context);
extern inline void fs_makeSureLogIsNotFull(fs_Context *context);
extern inline void fs_restoreSleepIfNeeded(fs_Context *context);

extern int32 fs_setMinFreeThread(fs_Context      *context);
extern int32 fs_getMinFreeThread(fs_Context      *context);
extern void fs_readAhead(fs_Context *context, fs_Inode *inode,
                         uint64 reqOffset, uint32 reqLen, int meta);
extern boolean fs_waitLogEmpty(fs_Context *context);
extern boolean fs_isBackupRestoreInProgress(fs_Context *context);
extern void fs_ndmpDumpDirHistory(fs_Context *context,
                                  fs_DumpContext_t *dumpContext,
                                  fs_Inumber parent, fs_Inumber inum, 
                                  char *name, int nameLen);
extern void fs_ndmpDumpNodeHistory(fs_Context *context,
                       fs_DumpContext_t *dumpContext, fs_Inode   *inode);
extern int fs_ndmpFlushLastBuffer(fs_Context *context,
                                   fs_DumpContext_t    *dumpContext);
extern void fs_ndmpWaitForDumpBuffers(fs_Context *context,
                                      fs_DumpContext_t *dumpContext);
extern void fs_ndmpDumpNodeHistoryNodeBuf(fs_Context *context,
                       fs_DumpContext_t *dumpContext,
                       fs_NdmpHistoryNode_t   *node);
extern void fs_ndmpFlushNodeHistory(fs_Context *context,
                                    fs_DumpContext_t *dumpContext);
extern int fs_ndmpFlushLastBuffer(fs_Context *context,
                                   fs_DumpContext_t    *dumpContext);
extern void fs_ndmpFlushDirHistory(fs_Context *context,
                                    fs_DumpContext_t *dumpContext);

/**
 * Initialize the dnlc.
 */
extern void
fs_dnlcInit();

/**
 * Lookup the specified entry in the dnlc by name filling
 * in parent and child attributes and child inode genNum
 * if found.   Returns whether the entry exists in the
 * directory, doesn't exist in the directory, or the dnlc
 * doesn't know and a full probe of the directory is
 * required.
 */
enum {
    FS_DNLC_UNKNOWN,
    FS_DNLC_EXISTS,
    FS_DNLC_NOT_EXISTS
};
extern int32
fs_dnlcLookup(
    fs_FileSystem   *fileSystem,
    fs_Proto        protocol,
    uint64          parentInumSnap,
    char            *name,
    int32           len,
    fs_Attr         *attrp,
    uint32          *genNum
);

/**
 * Flush the dnlc for the specified filesystem.
 */
extern void
fs_dnlcDestroyVolume(
    fs_FileSystem *fileSystem
);

/**
 * Insert the specified entry into the dnlc if it
 * doesn't already exist in the cache.  Returns TRUE
 * if the insertion takes place or its already in the
 * cache, FALSE if it cannot be inserted.
 */
extern boolean
fs_dnlcInsert(
    fs_FileSystem *fileSystem,
    uint64        parentInumSnap,
    char          *name,
    int32         len,
    uint64        inum
);

/**
 * Negate the specified entry in the dnlc.
 * Returns TRUE if the entry was negated (or a
 * negative entry inserted), FALSE if it couldn't
 * be.
 */
boolean
fs_dnlcNegate(
    fs_FileSystem   *fileSystem,
    uint64          parentInumSnap,
    char            *name,
    int32           len
);

/**
 * Remove the specified entry from the dnlc.
 * Returns TRUE if the entry was removed or
 * FALSE if it wasn't found.
 */
boolean
fs_dnlcRemove(
    fs_FileSystem   *fileSystem,
    uint64          parentInumSnap,
    char            *name,
    int32           len
);


/**
 * Flush the dnlc for the specified filesystem,snapnum.
 */
extern void
fs_dnlcDestroySnapnum(
    fs_FileSystem *fileSystem,
    int32         snapnum
);


/**
 * Copy the attributes from inode into attr.
 */
typedef enum fs_SnapType
{
    FS_NO_SNAPSHOT_TYPE         = 0x00,
    FS_SNAPSHOT_DIR_TYPE        = 0x01,
    FS_SNAPSHOT_BOUNDARY_TYPE   = 0x02
} fs_SnapType;

extern void
fs_doGetAttr(
    evm_volId_t     volId,
    fs_Attr         *attr,
    fs_Inode        *inode,
    fs_SnapType     snapType,   // FS_SNAPSHOT_DIR_TYPE or FS_SNAPSHOT_DIR_ENTRY_TYPE
    uint64          snapInfo,   // snapnumInum if FS_SNAPSHOT_DIR_ENTRY_TYPE
                                // parentInum if FS_SNAPSHOT_DIR_TYPE
    int32           protocol,
    fs_AttrType     attrType);

#ifdef FS_HIST_WAIT
extern void fs_backupInit();
extern void fs_receiveFileHistWindowUpdate(struct eee_window_update *);
#endif
extern void fs_restoreInvalidateTxrxCache(fs_Context *context);
#endif /* FS_PROTO_H */
