/*
 * Name		: fs-quota.h
 * RCS		: $ID$
 * Description	: Quota implementation data structures
 * Created by	: Rahul Khona
 * Date Created	: 8/10/2001
 */

#ifndef _FS_QUOTA_H
#define _FS_QUOTA_H

//TODO UPDATE COMMENT WITH NEW QUOTA SCHEME
/*
 * Incore quota record. One quota record per user id or group id that
 * is in use would be maintained incore. This can be freely recycled
 * if they are on the freelist.
 *
 * Each quota structure hangs on 2 lists a freelist and a hash list.
 * These lists are maintained in a global cache which spans across file systems
 * for better resource utilization. 
 *
 * User quota and group quota structures would be hanging off hash
 * queues, but would be hanging off a common freelist for more efficiency.
 * Whenever an operation is being made for a user / group and quotas are
 * turned on, this structure would be brought incore for the entities, if
 * they are not already in core. They would be taken off the freelist when
 * in use. Once there use is complete, they would be put back on the freelist
 * while still being retained on the hash list. Whenever a structure is
 * reused, it will move to a different hash list. This structures would
 * be periodically freeded if they are above a certain threshold.
 *
 * Spin lock would protect incoreRecords which are used for tracking
 * incore reservations. Once we are ready to commit the transaction
 * we would lock the structure exclusively, apply the reservation for the
 * particular request which holds it exclusively to the disk records and
 * keep the structure locked till it is logged. The reason is that quota
 * structures can be point of contention, if several allocation requests
 * come in for the same user. By separating incore reservation of quotas
 * from actual update of quotas to disk, we don't have to keep the
 * quota structure for the user locked for the whole duration and
 * thus reduce contention.
 *
 * In addition disk quota have to be written in log order, so that
 * replay of transaction log will not reset the quota and consumption
 * information incorrectly. The problem would be best explained by
 * an example. If 2 requests came in for user A simultaneously to allocate
 * 5 blocks to file 1 and 10 blocks to file 2. If the consumption at the
 * before these requests was say Y. If the ondisk updates were in the
 * order of Y + 5, followed by (Y + 5) + 10 and if the log was written
 * in (Y + 5) + 10 and Y + 5, then replay could show the consumption
 * as Y + 5 where as it should have been Y + 15. Remeber that on disk
 * we record absolute consumption value and not the deltas that need
 * to be applied as we log the disk blocks and not the operations.
 * 
 * lock structure would be used to i/o lock the diskrecord and the
 * buf structures during transaction logging.
 *
 * count is protected by the list spin lock. It indicates how many
 * contexts have this structure in use. Routines that make the structure
 * in use should increment the count holding the freelist lock and which
 * release the structure should decrement the count. If the count drops to
 * 0, the structure should be put back on freelist and if the count changes
 * from 0 to 1, it should be taken off the freelist.
 */



/*
 * Structure of quota and consumption records for user or group.
 * This would be the structure of an individual record in user
 * or group quota file. Each record would be 128 bytes. To ensure
 * that quota records will never span 2 i/o buffers
 * 
 * The user quota file and the group quota file are linear files. The first
 * record is the default user/group quota. 
 * 
 * NB: The disk quota record might go beyond 128 bytes if the SID is too long
 *     (never seen so far but still possible). In that case the flag 
 *     FS_UG_QREC_HAS_CONTIN (this record has a continuation) would be set and
 *     the next record would be the continuation (even if not in the same fs
 *     block)
 */
                                                
typedef struct fs_QuotaCfgDiskRecord {
	fs_quota_state  		state;
	uint32					flags;
    uint32                  lastChanged;    //Time when limits changed
	fs_QuotaRecord_t		defaultQuota;   //Default limits
}fs_QuotaCfgDiskRecord_t;	

typedef struct fs_QuotaDiskRecord {
	uint32					flags;
#define FS_UG_QNOT_BLK_TIMELIMIT	0x80000000	//user/group quota notification
#define FS_UG_QNOT_BLK_HARDLIMIT	0x40000000	//flags
#define FS_UG_QNOT_INO_TIMELIMIT	0x20000000
#define FS_UG_QNOT_INO_HARDLIMIT	0x10000000
#define FS_UG_DEFAULT_QUOTALIMITS  	0x01000000  //U/G uses default quota limits
#define FS_UG_QREC_HAS_CONTIN      	0x00000001  //this record has a continuation
//FIRST GUESS. IMPLEMENTATION WILL DETERMINE IF THE FOLLOWING ARE NECESSARY    
#define FS_UG_QREC_IS_CONTIN       	0x00000002  //this record is a continuation
#define FS_UG_QREC_FREE           	0x00000004  //This record is free

    uint32                  lastChanged;    //Time when limits changed
	fs_QuotaRecord_t		quota;			//limits
	fs_ConsumptionRecord_t	consumption;	//usage
    uint32                  unixId;         //uid/gid associated with this rec 
    uint16                  idFlags;        //identity flags
#define FS_UG_QTID_UNIX             0x0001  //record defined by unix identity
#define FS_UG_QTID_WIN              0x0002  //record defined by windows identity
#define FS_UG_QTID_MASK         (FS_UG_QTID_UNIX|FS_UG_QTID_WIN)
#define FS_UG_QTIDMAP_KNOWN         0x0004  //identity mappping is known
#define FS_UG_QTIDMAP_VALID         0x0008  //1:mapping is valid, 0:no mapping
#define FS_UG_QTIDMAP_MASK      (FS_UG_QTIDMAP_KNOWN|FS_UG_QTIDMAP_VALID)
    
    uchar8                  pad;
    uchar8                  sidLen;
    sec_cifsSid             sid;
} fs_QuotaDiskRecord_t;	

typedef union fs_QuotaDiskRecordPadded {
    fs_QuotaCfgDiskRecord_t cfg;
      //Because quota "disk" record can be bigger than 128, we can no longer
      //use this structure to access it. 
      //We have to cast pointer on fs_QuotaDiskRecordPadded to pointer on 
      //fs_QuotaDiskRecord_t.
//    fs_QuotaDiskRecord_t    qrec;
    char8                   bytes[128];    
}fs_QuotaDiskRecordPadded;

NFX_ASSERT_STATIC(FS_BLOCK_SIZE % sizeof(fs_QuotaDiskRecordPadded) == 0);
#define FS_QUOTA_RECS_PER_PAGE \
        (FS_BLOCK_SIZE / sizeof(fs_QuotaDiskRecordPadded))

#define FS_QUOTA_REC_SHIFT  7
NFX_ASSERT_STATIC((1<<FS_QUOTA_REC_SHIFT) == sizeof(fs_QuotaDiskRecordPadded));
        

#define	FS_QUOTA_INCORE_NEEDS_INIT	1	// indicate that incore inode
										// needs initialization
#define FS_QUOTA_INCORE_USER        2   // user quota
#define FS_QUOTA_INCORE_GROUP       4   // group quota

typedef struct fs_QuotaIncoreRecord fs_QuotaIncoreRecord_t;
struct fs_QuotaIncoreRecord {
    evm_volId_t volId;
	uint32		count;		    // shared holders of this lock
	uint32		id;			    // user or group id
	uint32		fileSetId;	    // fileset id
	uint8		flag;		    // indicates whether cachedRecords is valid
	fs_Lock		lock;		    // lock the structure for i/o and logging
	nfx_spinlock_t	spinlock;	// protect reservations, incore flags
								// and other things
	DLL_ENTRY(fs_QuotaIncoreRecord_t)	freeList;
	DLL_ENTRY(fs_QuotaIncoreRecord_t)	hashList;
	fs_QuotaDiskRecord_t	cachedRecords;	// disk cache
};

/*
 * Quota reservation and accounting context structure. Find a suitable
 * home for this structure.
 *
 * It is basically a list of all the quota structures and fileset
 * structures against which reservation is to be performed for a file
 * operation.
 */

typedef struct fs_QuotaContext fs_QuotaContext_t;
struct fs_QuotaContext {
	DLL_ENTRY(fs_QuotaContext_t)	list;
	fs_QuotaIncoreRecord_t		*quota;
	fs_FileSetIncoreRecord_t	*fset;
    uint64                      count;  //How many blocks/inodes accounted for
	int						    notify;
	int							flagsWr;//flags need to be written to disk
};

typedef struct fs_FsetContext fs_FsetContext_t;
struct fs_FsetContext {
	DLL_ENTRY(fs_FsetContext_t)	list;
	fs_FileSetIncoreRecord_t		*fset;
    uint64                          count;//How many blocks/inodes accounted for
	int								notify;
};

#define	FS_COUNT_INODES		0x00000001
#define	FS_COUNT_BLOCKS		0x00000002
#define	FS_FREE_RESOURCES	0x80000000

typedef struct fs_QuotaContextList {
	uint64		availCount;	// blocks/inodes to account for
	uint32		uid;	// id for whom we are counting
	uint32		gid;	// id for whom we are counting
	int			flag;	// indicate inodes or blocks
    fs_Inode    *newInode;  // newly allocated inode

	/*
	 * Objects against which quota is to be accounted for
	 */

	DLL_HDR(fs_QuotaContext_t)	uQuotaList; // list of quotas
	DLL_HDR(fs_QuotaContext_t)	gQuotaList; // list of quotas
	DLL_HDR(fs_FsetContext_t)	fsetList; // list of fsets
} fs_QuotaContextList_t;

#define	FS_QUOTA_ON_EVENT	1

#if defined(FS_STANDALONE) 
#define	FS_QUOT_EXCEPTION_MSG_RECV	FS_SIM_MSG_ADDR_SSC
#else	/*FS_STANDALONE*/
#define	FS_QUOT_EXCEPTION_MSG_RECV	// TBD
#endif	/*FS_STANDALONE*/


void fs_quotaCacheInit(void);

#endif	/*_FS_QUOTA_H*/
