/*
 * Name:  fs-rdwr.c
 *
 * RCS:  $Id: fs-rdwr.c,v 1.68 2002/12/10 18:51:39 bobonfeline Exp $
 *
 * Description:  Code for doing reads/writes on a volume.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */


#if !defined(SSC)

#include "fs-includes.h"

/**
 * Stack of fs_RdwrReq structs.  We manage memory for
 * these locally to ensure that we know how to clean them
 * up during exception conditions.
 */
nfx_spinlock_t   fs_rdwrSpin;
fs_RdwrReq  *fs_rdwrStart;
fs_RdwrReq  *fs_rdwrEnd;
fs_RdwrReq  **fs_rdwrTop;
int32       fs_rdwrNumFree;

nfx_spinlock_t   fs_ioSpin;
int32            fs_ioPending = 0;
int32            fs_ioWritesPending = 0;
int32            fs_ioMaxWritesPending = 0;
int32            fs_ioReadsPending = 0;
int32            fs_ioMaxReadsPending = 0;
uint32           fs_ioTotalMetaWrites = 0;
uint32           fs_ioTotalMetaReads = 0;
uint32           fs_ioTotalMetaReadAheads = 0;
uint32           fs_ioTotalUserWrites = 0;
uint32           fs_ioTotalUserReads = 0;
uint32           fs_ioTotalUserReadAheads = 0;
uint32           fs_ioTotalLogWrites = 0;
uint32           fs_ioTotalLogReads = 0;



/**
 * Queues of IO requests to retry.
 */
fs_Io2retryQueue fs_io2retryQ[EVM_DATA_TYPE_MAX];


/**
 * Forward delcarations.
 */
static int32
fs_rdwrRetryTimeout(
    uint32  timer,
    void    *param1,
    void    *param2,
    void    *param3,
    void    *param4);


#if !defined(FS_STANDALONE)
/**
 * Signal the IO retry daemon that an IO was just
 * enqueued to the retry queue.
 *
 * Must be called with the rdwr spinlock held!
 */
static void
fs_rdwrSignalRetryDaemon(int32 dataType)
{
    if (fs_io2retryQ[dataType].timerSignal == TRUE)
        return;
    fs_io2retryQ[dataType].timerSignal = TRUE;

    if (eee_timerStart(fs_io2retryQ[dataType].timerHandle,
                       fs_rdwrRetryTimeout,
                       (void *) (address_t) dataType, NULL, NULL, NULL) != NFX_OK)
    {
        fs_error(NULL, FS_INTERNAL_ERR, "start rdwr retry timer");
    }
}
#else // !FS_STANDALONE

static void
fs_rdwrSignalRetryDaemon(int32 dataType)
{
}

#endif // !FS_STANDALONE


/**
 * Send the specified IO request to EVM.  If EVM is full and
 * this is not a retry, enqueue it for retry.
 */
static int32
fs_rdwrSendEvmReq(
    fs_RdwrReq  *rdwr,
    boolean     isRetry)
{
    fs_FileSystem   *fileSystem = rdwr->fileSystem;
    int32           rc, reqType = rdwr->evmReq.reqType;

    rc = evm_IOReq(&rdwr->evmReq, fs_ioDone);
    if (rc != NFX_OK)
    {
        if (rc == EVM_STAT_REJECT)
        {
            //printf("EVM srvr rejected IO req %p, type %d\n",
            //        rdwr, rdwr->evmReq.dataType);

            if (!isRetry)
            {
                // EVM queue is full, need to retry this later.
                //printf("enqueue for retry\n");

                NFX_SPIN_LOCK(&fs_rdwrSpin);
                DLL_INSERT_TAIL(&fs_io2retryQ[rdwr->evmReq.dataType].dllHdr,
                                rdwr, retryEntry);
                fs_rdwrSignalRetryDaemon(rdwr->evmReq.dataType);
                NFX_SPIN_UNLOCK(&fs_rdwrSpin);
            }
        }
        // else caller handles exception case
    }
    else
    {
        if (reqType == EVM_WRITE_REQ)
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioWritesPending += 1;
            fs_ioPending += 1;
            if (fs_ioWritesPending > fs_ioMaxWritesPending)
                fs_ioMaxWritesPending = fs_ioWritesPending;
            NFX_SPIN_UNLOCK(&fs_ioSpin);
    
            if (fileSystem != NULL)
            {
                NFX_SPIN_LOCK(&fileSystem->spin);
                fileSystem->ioWritesPending += 1;
                fileSystem->ioPendingCount += 1;
                if (fileSystem->ioWritesPending > fileSystem->ioMaxWritesPending)
                {
                    fileSystem->ioMaxWritesPending = fileSystem->ioWritesPending;
                }
                NFX_SPIN_UNLOCK(&fileSystem->spin);
            }
        }
        else
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioReadsPending += 1;
            fs_ioPending += 1;
            if (fs_ioReadsPending > fs_ioMaxReadsPending)
                fs_ioMaxReadsPending = fs_ioReadsPending;
            NFX_SPIN_UNLOCK(&fs_ioSpin);
    
            if (fileSystem != NULL)
            {
                NFX_SPIN_LOCK(&fileSystem->spin);
                fileSystem->ioReadsPending += 1;
                fileSystem->ioPendingCount += 1;
                if (fileSystem->ioReadsPending > fileSystem->ioMaxReadsPending)
                {
                    fileSystem->ioMaxReadsPending = fileSystem->ioReadsPending;
                }
                NFX_SPIN_UNLOCK(&fileSystem->spin);
            }
        }
    }

    return rc;
}


/**
 * Initiate IO retries from the specified retry queue.
 */
static void
fs_rdwrUnloadRetryQueue(int32 type)
{
    fs_RdwrReq  *rdwr;
    int32       rc, numDone = 0;

#define FS_RETRY_MAX_PER_EVENT  128

    NFX_SPIN_LOCK(&fs_rdwrSpin);

    while (!DLL_IS_EMPTY(&fs_io2retryQ[type].dllHdr))
    {
        DLL_REMOVE_HEAD(&fs_io2retryQ[type].dllHdr, rdwr, retryEntry);

        NFX_SPIN_UNLOCK(&fs_rdwrSpin);

        //printf("doing io retry for req %p type %d\n", rdwr, type);

        if (rdwr->fileSystem &&
            rdwr->fileSystem->state == FS_VOL_STATE_FAILING)
        {
            fs_ioDoneError(rdwr);
            NFX_SPIN_LOCK(&fs_rdwrSpin);
            continue;
        }

        rc = fs_rdwrSendEvmReq(rdwr, TRUE);
        if (rc != NFX_OK)
        {
            if (rc == EVM_STAT_REJECT)
            {
                // Put it back on the head to preserve IO order.
                NFX_SPIN_LOCK(&fs_rdwrSpin);
                DLL_INSERT_HEAD(&fs_io2retryQ[type].dllHdr, rdwr, retryEntry);
                break;
            }
            else
            {
                int32 errorType = (rdwr->evmReq.reqType == EVM_READ_REQ) ?
                    FS_READ_ERR : FS_WRITE_ERR;
                fs_rdwrDealloc(rdwr);
                fs_volumeException(NULL, rdwr->fileSystem, errorType);
                NFX_SPIN_LOCK(&fs_rdwrSpin);
                continue;
            }
        }

        NFX_SPIN_LOCK(&fs_rdwrSpin);

        if (++numDone >= FS_RETRY_MAX_PER_EVENT)
            break;
    }

    NFX_SPIN_UNLOCK(&fs_rdwrSpin);
}


#if !defined(FS_STANDALONE)
/**
 * Timer function for rdwr retry queues.  This is activated whenever
 * we have IO sitting on retry queues.
 */
static int32
fs_rdwrRetryTimeout(
    uint32  timer,
    void    *param1,
    void    *param2,
    void    *param3,
    void    *param4)
{
    int32   dataType = (int32) (address_t) param1;

    fs_rdwrUnloadRetryQueue(dataType);

    NFX_SPIN_LOCK(&fs_rdwrSpin);
    if (DLL_IS_EMPTY(&fs_io2retryQ[dataType].dllHdr))
    {
        fs_io2retryQ[dataType].timerSignal = FALSE;
        NFX_SPIN_UNLOCK(&fs_rdwrSpin);
        return NFX_OK;
    }
    NFX_SPIN_UNLOCK(&fs_rdwrSpin);

    if (eee_timerStart(timer,
                       fs_rdwrRetryTimeout,
                       (void *) (address_t) dataType, NULL, NULL, NULL) != NFX_OK)
    {
        fs_error(NULL, FS_INTERNAL_ERR, "start rdwr retry timer");
    }

    FS_ASSERT(fs_io2retryQ[dataType].timerSignal == TRUE);

    return NFX_OK;
}
#endif // FS_STANDALONE


/*-----------------------------------------------------------------------------
 * fs_volReadWrite
 *
 * Description: Function to initiate read or write to volume.  Caller
 *              provides fs_rdwrReq_t, which will be returned as
 *              event data (param1) to the waiting state machine.  See
 *              fs_ioDone().
 *
 *              If caller is using FS_RDWR_SYNC, the function
 *              will eee_yield() the CPU and return when the next
 *              event is delivered to the yielding state machine.
 *              Normally, this will be the I/O completion event and
 *              p1 will be the original fs_RdwrReq with the status
 *              field updated. If a different event triggers our
 *              awakening, panic.
 */
void
fs_volReadWrite(
    fs_Context      *context,
    fs_RdwrReq      *req)       // Request structure must persist until
                                // I/O completion is signalled.
{
	esm_event_t		event;
    int32           rc, reqType = req->evmReq.reqType;
    int32           dataType = req->evmReq.dataType;
    int32           doneEvent = req->doneEvent;
    boolean         isSync = (req->flags & FS_RDWR_FLAGS_SYNC);

    /*
    printf("fs_volReadWrite: %s handle %d\n",
        req->evmReq.reqType == EVM_READ_REQ ? "READ" : "WRITE",
        context->smHandle);

    printf("volReadWrite: offset %lld num %lld\n",
                req->evmReq.offset, req->evmReq.reqSize);
    */

#if defined(AGILE_HARDWARE)
#ifdef FS_RDWR_TIMESTAMP
    eee_rtc (&req->timestamp_start); 
#endif
#endif

    if (context->fileSystem &&
        context->fileSystem->state == FS_VOL_STATE_FAILING)
    {
        fs_rdwrDealloc(req);
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
        return;
    }

    // If there are already requests on the waiting queue,
    // don't start this request ahead of them.
    NFX_SPIN_LOCK(&fs_rdwrSpin);
    if (!DLL_IS_EMPTY(&fs_io2retryQ[dataType].dllHdr))
    {
        DLL_INSERT_TAIL(&fs_io2retryQ[dataType].dllHdr, req, retryEntry);
        NFX_SPIN_UNLOCK(&fs_rdwrSpin);
    }
    else
    {
        NFX_SPIN_UNLOCK(&fs_rdwrSpin);

        rc = fs_rdwrSendEvmReq(req, FALSE);
        if (rc != NFX_OK && rc != EVM_STAT_REJECT)
        {
            int32 errorType = (reqType == EVM_WRITE_REQ) ? FS_WRITE_ERR : FS_READ_ERR;
            fs_rdwrDealloc(req);
            fs_volumeException(context, context->fileSystem, errorType);
            return;
        }
    }

    if (isSync)
    {
        if (reqType == EVM_WRITE_REQ)
        {
            if (dataType == EVM_META_DATA_REQ)
            {
                fs_yield(context, &event, FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_MW);
            }
            else
            {
                fs_yield(context, &event, FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_UW);
            }
        }
        else
        {
            if (dataType == EVM_META_DATA_REQ)
            {
                fs_yield(context, &event, FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_MR);
            }
            else
            {
                fs_yield(context, &event, FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_UR);
            }
        }
        
        if (event.event != doneEvent ||
            event.p1.u.p != req)
        {
            // Some other event woke us up.
			fs_error(context, FS_INTERNAL_ERR,
				"fs_volRW: wrong event: got %d/%p/%p, exp %d/%p\n",
  				event.event, event.p1.u.p, event.p2.u.p, doneEvent, req);
        }
    }

} /* fs_volReadWrite */


/**
 * Wait for all pending IO requests to complete.
 */
void
fs_volWaitPendingIO(
    fs_Context  *context,
    boolean     isForce)
{
    fs_FileSystem   *fs = context->fileSystem;
    int32           numWaits = 0;

    FS_ASSERT(context->isThread);

// about 60 seconds
#define MAX_WAITS   50
#define WAIT_MSECS  50
    // Now wait until all threads are terminated and
    // all IO is completed.
    NFX_SPIN_LOCK(&fs->spin);
    while (numWaits++ < MAX_WAITS)
    {
        if (fs->ioPendingCount > 0)
        {
            if (numWaits > MAX_WAITS / 2)
            {
                FS_ELOG(context, LOG_DEBUG, FS_TRACE_SPACE,
                        ("unmount %s, still waiting for %d pending IO's\n",
                        fs->super->mountName, fs->ioPendingCount));
            }
        }
        else
        {
            break;
        }

        NFX_SPIN_UNLOCK(&fs->spin);
        if (isForce)
        {
            // We're in exception condition already,
            // don't call fs_sleep, since it will
            // call the exception handler.
            thread_sleep(NULL, numWaits * WAIT_MSECS);
        }
        else
        {
            fs_sleep(context, numWaits * WAIT_MSECS);
        }
        NFX_SPIN_LOCK(&fs->spin);
    }
    NFX_SPIN_UNLOCK(&fs->spin);

    if (numWaits >= MAX_WAITS)
    {
        if (isForce)
        {
            fs_error(context, FS_INTERNAL_ERR, "volWaitPendingIo failed");
        }
        else
        {
            fs_volumeException(context, context->fileSystem,
                               FS_VOLUME_EXCEPTION);
        }
    }
#undef MAX_WAITS
#undef WAIT_MSECS

}


/**
 * Flush the IO's on the retry queue for the
 * specified filesystem.
 */
void
fs_rdwrDestroyRetryQueue(
    fs_Context  *context)
{
    fs_RdwrReq  *rdwr, *rdwrNext;
    DLL_HDR(fs_RdwrReq) destroyHdr;
    int32 dataType;

    DLL_INIT(&destroyHdr);

    NFX_SPIN_LOCK(&fs_rdwrSpin);
    for (dataType = 0; dataType < EVM_DATA_TYPE_MAX; dataType++)
    {
        DLL_ITERATE_NEXT(rdwr, rdwrNext, &fs_io2retryQ[dataType].dllHdr, retryEntry)
        {
            if (rdwr->fileSystem && rdwr->fileSystem == context->fileSystem)
            {
                DLL_REMOVE(&fs_io2retryQ[dataType].dllHdr, rdwr, retryEntry);
                DLL_INSERT_HEAD(&destroyHdr, rdwr, retryEntry);
            }
        }
    }
    NFX_SPIN_UNLOCK(&fs_rdwrSpin);

    DLL_ITERATE_NEXT(rdwr, rdwrNext, &destroyHdr, retryEntry)
    {
        DLL_REMOVE(&destroyHdr, rdwr, retryEntry);

        NFX_SPIN_LOCK(&rdwr->fileSystem->spin);
        rdwr->fileSystem->ioPendingCount -= 1;
        NFX_SPIN_UNLOCK(&rdwr->fileSystem->spin);
        fs_ioDoneError(rdwr);
    }
}


/**
 * Initialize common portions of the given request block with the
 * specified context, offset, and size parameters.
 */
static void fs_volInitRequest(
    fs_Context      *context,
    fs_RdwrReq      *req,
    esm_handle_t    handle,
    evm_volOffset_t offset,
    evm_volSize_t   size)
{
    FS_ASSERT(req >= fs_rdwrStart &&
              req < fs_rdwrEnd);

    req->smHandle              = handle;
    req->fileSystem            = context->fileSystem;
    req->evmReq.id             = context->volId;
    req->evmReq.reqSize        = size / EVM_LBLK_SIZE;
    req->evmReq.offset         = offset / EVM_LBLK_SIZE;
    req->evmReq.clientPrivate  = (address_t) req;
    req->evmReq.attributes     = 0;
}


/**
 * Convenience function for the typical meta data read request case.
 */
void fs_volRead(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size)
{
    FS_TRACE(context, FS_TRACE_IO,
             ("vol %llx meta read blk %lld len %lld req %p flags %#x",
              context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             req,
             req->flags));

    fs_volInitRequest(context, req, context->smHandle, offset, size);
    req->doneEvent = FS_THREAD_EVENT_IO_DONE;
    req->flags = FS_RDWR_FLAGS_SYNC;
    req->evmReq.reqType = EVM_READ_REQ;
    req->evmReq.dataType = EVM_META_DATA_REQ;

    fs_volReadWrite(context, req);
}


/**
 * Convenience function for the typical meta data block read request case.
 */
void fs_volReadBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Block        block)
{
    fs_volRead(context, req,
               fs_blockNumber(block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE);
}


/**
 * Convenience function for the typical user data read request case.
 */
void fs_volReadUser(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size)
{
    FS_TRACE(context, FS_TRACE_IO,
             ("vol %llx user read blk %lld len %lld req %p flags %#x",
              context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             req,
             req->flags));

    fs_volInitRequest(context, req, context->smHandle, offset, size);
    req->doneEvent = FS_THREAD_EVENT_IO_DONE;
    req->flags = FS_RDWR_FLAGS_SYNC;
    req->evmReq.reqType = EVM_READ_REQ;
    req->evmReq.dataType = EVM_USER_DATA_REQ;

    fs_volReadWrite(context, req);
}


/**
 * Convenience function for the typical user data block read request case.
 */
void fs_volReadUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Block        block)
{
    fs_volReadUser(context, req,
                   fs_blockNumber(block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE);
}


/**
 * Convenience function for an async meta data read request.
 */
void fs_volReadAsync(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    int32           evt,
    uint64          attributes,
    void            *priv)
{
    FS_TRACE(context, FS_TRACE_IO,
     ("vol %llx meta read async blk %lld len %lld req %p flags %#x",
             context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             req,
             req->flags));

    fs_volInitRequest(context, req, context->smHandle, offset, size);
    req->doneEvent = evt;
    req->private = priv;
    req->flags = 0;
    req->evmReq.reqType = EVM_READ_REQ;
    req->evmReq.dataType = EVM_META_DATA_REQ;
    req->evmReq.codOffset = 0;
    req->evmReq.attributes = attributes;

    fs_volReadWrite(context, req);
}


/**
 * Convenience function for an async meta data block read request.
 */
void fs_volReadAsyncBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Block        block,
    int32           evt,
    void            *priv)
{
    fs_volReadAsync(context, req,
                    fs_blockNumber(block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE,
                    evt, 0, priv);
}


/**
 * Convenience function for the typical meta data write request case.
 */
void fs_volWrite(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    void            *ptr)
{
    FS_TRACE(context, FS_TRACE_IO,
           ("vol %llx meta write blk %lld len %lld datap %p req %p flags %#x",
            context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             ptr,
             req,
             req->flags));

    fs_volInitRequest(context, req, context->smHandle, offset, size);
    req->doneEvent = FS_THREAD_EVENT_IO_DONE;
    req->flags = FS_RDWR_FLAGS_SYNC;
    req->evmReq.reqType = EVM_WRITE_REQ;
    req->evmReq.dataType = EVM_META_DATA_REQ;
    req->evmReq.wDataHandle.dataPtr = ptr;

    fs_volReadWrite(context, req);
}


/**
 * Convenience function for the typical meta data block write request case.
 */
void fs_volWriteBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Block        block,
    void            *ptr)
{
    fs_volWrite(context, req,
                fs_blockNumber(block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE,
                ptr);
}


/**
 * Convenience function for the typical user data write request case.
 */
void fs_volWriteUser(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    nfx_cod_t       cod)
{
    FS_TRACE(context, FS_TRACE_IO,
          ("vol %llx user write blk %lld len %lld cod %#llx req %p flags %#x",
           context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             cod,
             req,
             req->flags));

    fs_volInitRequest(context, req, context->smHandle, offset, size);
    req->doneEvent = FS_THREAD_EVENT_IO_DONE;
    req->flags = FS_RDWR_FLAGS_SYNC;
    req->evmReq.reqType = EVM_WRITE_REQ;
    req->evmReq.dataType = EVM_USER_DATA_REQ;
    req->evmReq.wDataHandle.codHandle = cod;
    req->evmReq.codOffset = 0;

    fs_volReadWrite(context, req);
}


/**
 * Convenience function for the typical user data block write request case.
 */
void fs_volWriteUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Block        block,
    nfx_cod_t       cod)
{
    fs_volWriteUser(context, req,
                    fs_blockNumber(block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE,
                    cod);
}


/**
 * Convenience function for an async meta data write request.
 */
void fs_volWriteAsync(
    fs_Context      *context,
    fs_RdwrReq      *req,
    esm_handle_t    handle,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    void            *ptr,
    int32           evt,
    uint64          attributes,
    void            *priv)
{
    FS_TRACE(context, FS_TRACE_IO,
     ("vol %llx meta write async blk %lld len %lld datap %p req %p flags %#x",
             context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             ptr,
             req,
             req->flags));

    fs_volInitRequest(context, req, handle, offset, size);
    req->doneEvent = evt;
    req->private = priv;
    req->flags = 0;
    req->evmReq.reqType = EVM_WRITE_REQ;
    req->evmReq.dataType = EVM_META_DATA_REQ;
    req->evmReq.wDataHandle.dataPtr = ptr;
    req->evmReq.codOffset = 0;
    req->evmReq.attributes = attributes;

    fs_volReadWrite(context, req);
}


/**
 * Convenience function for an async meta data block write request.
 */
void fs_volWriteAsyncBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    esm_handle_t    handle,
    fs_Block        block,
    void            *ptr,
    int32           evt,
    void            *priv)
{
#ifdef FS_DEBUG
    { // try to catch the guy writing superblock without calculating the CRC
        int i, n;
        
        for (i = 0, n = fs_superBlockCount(); i < n; i++)
        {
            if (fs_blockNumber(block) == fs_blockNumber(fs_superBlockBlock(i)))
            {
                fs_SuperBlock * super = (fs_SuperBlock*)ptr;
                uint32          savedCrc = super->crc32;

                // printf("writing superblock (i=%d) "
                //       "(n=%d) crc: 0x%x\n", i, n, savedCrc);
                super->crc32 = 0;
                super->crc32 = fs_checksum(super, sizeof(fs_SuperBlock), ~0);
                FS_ASSERT(savedCrc == super->crc32);            
            }
        } // for
    }
#endif // DEBUG    
  
    fs_volWriteAsync(context, req, handle,
                     fs_blockNumber(block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE,
                     ptr, evt, 0, priv);
}


/**
 * Convenience function for an async user data write request.
 */
void fs_volWriteAsyncUser(
    fs_Context      *context,
    fs_RdwrReq      *req,
    esm_handle_t    handle,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    nfx_cod_t       cod,
    int32           evt,
    void            *priv)
{
    FS_TRACE(context, FS_TRACE_IO,
    ("vol %llx user write async blk %lld len %lld cod %#llx req %p flags %#x",
             context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             cod,
             req,
             req->flags));

    FS_ASSERT(size <= FS_BLOCK_SIZE);

    fs_volInitRequest(context, req, handle, offset, size);
    req->doneEvent = evt;
    req->private = priv;
    req->flags = 0;
    req->evmReq.reqType = EVM_WRITE_REQ;
    req->evmReq.dataType = EVM_USER_DATA_REQ;
    req->evmReq.wDataHandle.codHandle = cod;
    req->evmReq.codOffset = 0;

    fs_volReadWrite(context, req);
}


/**
 * Convenience function for an async user data block write request.
 */
void fs_volWriteAsyncUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    esm_handle_t    handle,
    fs_Block        block,
    nfx_cod_t       cod,
    int32           len,
    int32           evt,
    void            *priv)
{
    fs_volWriteAsyncUser(context, req, handle,
                         fs_blockNumber(block) * FS_BLOCK_SIZE, len,
                         cod, evt, priv);
}


/**
 * Clean up rdwr from error return.
 */
void
fs_ioDoneError(
    fs_RdwrReq  *rdwr)
{
    if (rdwr->evmReq.reqType == EVM_READ_REQ)
    {
        if (rdwr->evmReq.dataType == EVM_USER_DATA_REQ)
        {
            // Free the just read cod.
            if (rdwr->evmStat.rDataHandle.codHandle)
            {
                fs_cod_free_cod(rdwr->evmStat.rDataHandle.codHandle);
                rdwr->evmStat.rDataHandle.codHandle = 0;
            }
        }
        else if (rdwr->evmStat.rDataHandle.dataPtr)
        {
            // Free the just read metadata.
            fs_ramDealloc(FS_RAM_READ_DATA, rdwr->evmStat.rDataHandle.dataPtr);
            rdwr->evmStat.rDataHandle.dataPtr = NULL;
        }
    }
    else
    {
        // Chance for memory leak here, but most
        // writes are from buffer cache or vol logging
        // daemon, so we don't want to free datap here.
        //
        // Currently, only mkfs/mount/unmount operations
        // are a potential leak.
        //
        // TODO - this good enuf?
    }

    fs_rdwrDealloc(rdwr);
}


/**
 * Initialize fs_RdwrReq stack.
 */
void
fs_rdwrInit()
{
    fs_RdwrReq  *rdwr;
    int32       numToAlloc = FS_RDWR_MAX + 2048;  // Add a few for safety
    int32       i;

#define FS_RDWR_RETRY_TIMEOUT 50    // msecs

    fs_rdwrStart = fs_ramAlloc(FS_RAM_RDWR, sizeof(fs_RdwrReq) * numToAlloc);
    fs_rdwrEnd = fs_rdwrStart + numToAlloc;
    fs_rdwrTop = fs_ramAlloc(FS_RAM_RDWR, sizeof(fs_RdwrReq *) * numToAlloc);

    for (i = 0; i < numToAlloc; i++)
    {
        rdwr = fs_rdwrStart + i;
        rdwr->flags = FS_RDWR_FLAGS_FREE;
        DLL_ISOLATE_ELEM(rdwr, dllEntry);
        DLL_ISOLATE_ELEM(rdwr, retryEntry);
        *(fs_rdwrTop + i) = rdwr;
    }
    fs_rdwrNumFree = numToAlloc;

    /* Initialize the queues of IOs to retry.
     */
    for (i = 0; i < EVM_DATA_TYPE_MAX; i++)
    {
        DLL_INIT(&(fs_io2retryQ[i].dllHdr));

#if !defined(FS_STANDALONE)
        fs_io2retryQ[i].timerHandle = eee_timerCreate(FS_RDWR_RETRY_TIMEOUT,
                                                      EEE_TIMER_ONCE);
        if (fs_io2retryQ[i].timerHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
        {
            fs_error(NULL, FS_EEE_ERR, "create rdwr retry timer");
        }
        fs_io2retryQ[i].timerSignal = FALSE;
#endif
    }

    NFX_SPIN_INIT(&fs_rdwrSpin,SPINLOCK_LEVEL_UNKNOWN);

    //printf("rdwr init start 0x%x end 0x%x top 0x%x num %d\n",
    //  fs_rdwrStart, fs_rdwrEnd, fs_rdwrTop, fs_rdwrNumFree);
}


/**
 * Allocate an fs_RdwrReq.
 */
fs_RdwrReq *
fs_rdwrAlloc()
{
    fs_RdwrReq  *rdwr;

    NFX_SPIN_LOCK(&fs_rdwrSpin);
    if (--fs_rdwrNumFree <= 0)
    {
        fs_error(NULL, FS_INTERNAL_ERR, "no free rdwr structs");
    }
    rdwr = *fs_rdwrTop++;
    FS_ASSERT(rdwr->flags & FS_RDWR_FLAGS_FREE);
    rdwr->flags = 0;
    rdwr->fileSystem = NULL;
    NFX_SPIN_UNLOCK(&fs_rdwrSpin);

    //printf("rdwr alloc %p\n", rdwr);

    FS_ASSERT(rdwr >= fs_rdwrStart &&
              rdwr < fs_rdwrEnd);

    return rdwr;
}


/**
 * Deallocate an fs_RdwrReq.
 */
void
fs_rdwrDealloc(
    fs_RdwrReq  *rdwr)
{
    //printf("rdwr dealloc %p\n", rdwr);
    FS_ASSERT(rdwr >= fs_rdwrStart &&
              rdwr < fs_rdwrEnd);

    FS_ASSERT(!(rdwr->flags & FS_RDWR_FLAGS_FREE));

    FS_ASSERT(DLL_IS_ISOLATED(rdwr, dllEntry));
    FS_ASSERT(DLL_IS_ISOLATED(rdwr, retryEntry));

    NFX_SPIN_LOCK(&fs_rdwrSpin);
    rdwr->flags = FS_RDWR_FLAGS_FREE;
    fs_rdwrTop--;
    *fs_rdwrTop = rdwr;
    fs_rdwrNumFree++;
    NFX_SPIN_UNLOCK(&fs_rdwrSpin);
}


//   W R I T E   Q U E U E S


/**
 * Initialize a write queue for a given number of outstanding requests.
 *
 * Once a request from a write queue is inuse, it may not be
 * waited for individually.  The only available operation is to wait
 * for the entire queue to be written.
 */
void
fs_initWriteQueue(
    fs_Context      *context,
    fs_WriteQueue   *queue,
    int             size)
{
    FS_ASSERT(size > 0);

    queue->context = context;
    queue->max     = size;
    queue->ioPending = 0;
    queue->enqueued = 0;
    DLL_INIT(&queue->dllHdr);
}


/**
 * Wait for a scheduled IO request to be completed.
 */
static fs_RdwrReq *
fs_writeQueueWait(
    fs_WriteQueue   *queue)
{
    esm_event_t     event;
    fs_RdwrReq      *rdwr;
        
    FS_ASSERT(queue->ioPending > 0);

    fs_yield(queue->context, &event, FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_MW);

    queue->ioPending -= 1;
        
    if (event.event != FS_THREAD_EVENT_IO_DONE)
    {
        fs_error(queue->context, FS_INTERNAL_ERR,
                 "fs_writeQueueWait: unexpected event: %d", event.event);
    }

    rdwr = event.p1.u.p;
    return rdwr;
}


/**
 * Allocate a request buffer from the specified write queue, waiting
 * for one to become free if none are currently free.
 *
 * The fields of the request object are in an unknown state when it
 * is handed to the user.
 */
static fs_RdwrReq *
fs_writeQueueAcquire(
    fs_WriteQueue   *queue)
{
    if (queue->enqueued < queue->max)
    {
        return fs_rdwrAlloc();
    }
    
    fs_writeQueueSync(queue);
    FS_ASSERT(queue->enqueued < queue->max);
    return fs_rdwrAlloc();
}


/**
 * Wait for all outstanding writes in the specified queue to complete.
 */
void
fs_writeQueueSync(
    fs_WriteQueue   *queue)
{
    fs_RdwrReq  *req, *reqNext;

    FS_TRACE(queue->context, FS_TRACE_IO,
             ("writeQueueSync: %d enqueued %d ioPending",
                queue->enqueued, queue->ioPending));

    // Schedule the requests for IO.
    DLL_ITERATE_NEXT(req, reqNext, &queue->dllHdr, dllEntry)
    {
        DLL_REMOVE(&queue->dllHdr, req, dllEntry);
        queue->enqueued -= 1;
        queue->ioPending += 1;

        fs_volWriteAsync(queue->context,
                         req,
                         queue->context->smHandle,
                         req->evmReq.offset,
                         req->evmReq.reqSize,
                         req->evmReq.wDataHandle.dataPtr,
                         FS_THREAD_EVENT_IO_DONE,
                         0,
                         req);
    }

    // Wait for all to complete.
    while (queue->ioPending > 0)
    {
        fs_RdwrReq *rdwr = fs_writeQueueWait(queue);

        NFX_SPIN_LOCK(&queue->context->fileSystem->spin);
        queue->context->fileSystem->ioPendingCount -= 1;
        NFX_SPIN_UNLOCK(&queue->context->fileSystem->spin);

        if (rdwr->evmStat.status != NFX_OK)
        {
            fs_rdwrDealloc(rdwr);
            printf("fs_writeQueueSync io error returned, aborting\n");
            fs_volumeException(queue->context, queue->context->fileSystem,
                               FS_WRITE_ERR);
            return;
        }

        fs_rdwrDealloc(rdwr);
    }
}


/**
 * Queue writing meta data.
 */
void
fs_queueWrite(
    fs_WriteQueue   *queue,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    void            *data)
{
    fs_RdwrReq *req = fs_writeQueueAcquire(queue);

    // Borrow these fields, they'll be properly
    // initialized when we schedule the IO.
    req->evmReq.reqSize = size;
    req->evmReq.offset = offset;
    req->evmReq.wDataHandle.dataPtr = data;

    queue->enqueued += 1;
    DLL_INSERT_TAIL(&queue->dllHdr, req, dllEntry);
}


/**
 * Queue writing a meta data block.
 */
void
fs_queueWriteBlock(
    fs_WriteQueue   *queue,
    fs_Block        block,
    int32           blockSize,
    void            *data)
{
    fs_queueWrite(queue,
                  fs_blockNumber(block) * blockSize, blockSize, data);
}

/**
 * Convenience function for the user data read ahead request case.
 */
void fs_volReadAheadUser(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    esm_handle_t    raDoneHandle)
{
    FS_TRACE(context, FS_TRACE_IO,
             ("vol %llx user read blk %lld len %lld req %p flags %#x",
              context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             req,
             req->flags));

    fs_volInitRequest(context, req, context->smHandle, offset, size);
    req->smHandle = raDoneHandle;
    req->doneEvent = FS_WB_EVENT_RA_DONE;
    req->flags = 0;
    req->evmReq.reqType = EVM_READ_REQ;
    req->evmReq.dataType = EVM_USER_DATA_REQ;

    fs_volReadWrite(context, req);
}

/**
 * Convenience function for the typical meta data read ahead request case.
 */
void fs_volReadAhead(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    esm_handle_t    raDoneHandle)
{
    FS_TRACE(context, FS_TRACE_IO,
             ("vol %llx meta read blk %lld len %lld req %p flags %#x",
              context->volId.whole_thing,
             offset / FS_BLOCK_SIZE,
             size,
             req,
             req->flags));

    fs_volInitRequest(context, req, context->smHandle, offset, size);
    req->doneEvent = FS_WB_EVENT_RA_DONE;
    req->smHandle = raDoneHandle;
    req->flags = 0;
    req->evmReq.reqType = EVM_READ_REQ;
    req->evmReq.dataType = EVM_META_DATA_REQ;

    fs_volReadWrite(context, req);
}

/**
 * Convenience function for the typical user data block read ahead request case.
 */
void fs_volReadAheadUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Buf          *buf,
    esm_handle_t    raDoneHandle)
{
    req->private = (void *)buf;
    fs_volReadAheadUser(context, req,
                   fs_blockNumber(buf->block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE,
                    raDoneHandle);
}

/**
 * Convenience function for the typical user data block read request ahead case.
 */
void fs_volReadAheadBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Buf          *buf,
    esm_handle_t    raDoneHandle)
{
    req->private = (void *)buf;
    fs_volReadAhead(context, req,
                   fs_blockNumber(buf->block) * FS_BLOCK_SIZE, FS_BLOCK_SIZE,
                   raDoneHandle);
}
#endif // SSC
