// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/**
 * Test code for skip heaps.
 *
 * Paul Haahr <haahr@agilestorage.com>
 */

#include "fs-includes.h"


#define PERMUTATIONS 6


/** Largest "meaningful" difference between integers */

enum { BIG = 30000 }; 


//  U T I L I T I E S


#define NEW(type)       ((type *) zalloc(sizeof (type)))
#define P(p)            printf("%s = %p\n", #p, p)

#define streq(s, t)     (strcmp(s, t) == 0)

#ifdef  EFENCE
extern void *ef_malloc(size_t size);
extern void ef_free(void *ptr);
#define malloc  ef_malloc
#define free    ef_free
#endif


/**
 * Allocate zero-initialized memory.
 */
void *zalloc(int size)
{
    void *p = malloc(size);
    assert(p != NULL);
    bzero(p, size);
    return p;
}


/**
 * Return true if the specified integer is odd.
 */
boolean isOdd(int n)
{
    return (n & 1) != 0;
}



//   S I M U L A T E D   A G I L E   E N V I R O N M E N T


/**
 * Globals we wish we didn't refer to.
 */
int32 fs_applicationId = 0;
uint32 nfxMySlotId = 0;


/**
 * If checkLogging is turned on, all buffer writes are checked to ensure
 * that the buffer hasn't changed.
 */
boolean checkLogging = TRUE;


/**
 * Print a panic message and then abort the system.
 */
noreturn fs_panic(char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    vprintf(format, ap);
    va_end(ap);

    fs_abort();
}


/**
 * Called to indicate that a buffer has been modified.
 */
void fs_bufWrite(
    fs_Context  *context,
    fs_Buf      *buf,
    int32       offset,
    int32       length,
    uint32      flags,
    void        *bitmap)
{
    if (checkLogging) {
        int i;
        uint8 *old = buf->shadow;
        uint8 *new = buf->u.data;
        uint32 size = buf->page;
        boolean clean = TRUE;

        for (i = 0; i < offset; i++) {
            if (old[i] != new[i]) {
                fprintf(stderr, "byte %d changed: old %d new %d\n",
                        i, old[i], new[i]);
                clean = FALSE;
            }
        }
        for (i = offset; i < offset + length; i++) {
            old[i] = new[i];
        }
        for (i = offset + length; i < size; i++) {
            if (old[i] != new[i]) {
                fprintf(stderr, "byte %d changed: old %d new %d\n",
                        i, old[i], new[i]);
                clean = FALSE;
            }
        }
        if (!clean) {
            fprintf(stderr, "... was writing %d+%d\n", offset, length);
            fs_abort();
        }
    }
}


/**
 * Abort the system, after printing a backtrace.
 */
noreturn fs_abort()
{
    btrace_print();
    abort();
}


/**
 * Shadow the agile allocator.
 */
void *eee_ramAlloc(uint32 size)
{
    void *p = malloc(size);
    assert(p != NULL);
    return p;
}


/**
 * Shadow the agile allocator.
 */
void *fs_ramAlloc(int32 reason, int32 size)
{
    void *p = malloc(size);
    assert(p != NULL);
    return p;
}



//  S E T U P


/**
 * Template for ints.
 */
fs_SkipHeapTemplate *intTemplate;


/**
 * Comparison function for ints.
 */
int intcmp(void *payload1, void *payload2)
{
    int i1 = *(int *) payload1;
    int i2 = *(int *) payload2;
    if (i1 < i2) {
        unsigned delta = i2 - i1;
        return -MIN(delta, BIG);
    }
    if (i1 > i2) {
        unsigned delta = i1 - i2;
        return MIN(delta, BIG);
    }
    return 0;
}


/**
 * Template for strings of length 32 or under.
 */
fs_SkipHeapTemplate *stringTemplate;


/**
 * Initialize the templates we're using.
 */
void initializeTemplates(int size)
{
    intTemplate
        = fs_makeSkipHeapTemplate(size, sizeof (int), 0, intcmp);

    stringTemplate
        = fs_makeSkipHeapTemplate(
            size, 32, 0, (int (*)(void *, void *)) strcmp);
}


/**
 * Make a skip heap using the specified template, in a new buffer.
 */
fs_SkipHeap *makeHeap(
    fs_SkipHeapTemplate *template)
{
    uint32 size = fs_skipHeapTemplateBlockSize(template);
    fs_Buf *buf = NEW(fs_Buf);
    fs_SkipHeap *heap = NEW(fs_SkipHeap);

    buf->page = size;
    buf->u.data = eee_ramAlloc(size);
    buf->shadow = eee_ramAlloc(size);
    memset(buf->u.data, 'z', size);
    memcpy(buf->shadow, buf->u.data, size);
    *heap = fs_makeSkipHeap(NULL, template, buf);
    fs_skipHeapInit(heap);
    return heap;
}


/**
 * Print an integer skip heap.
 */
void printIntHeap(fs_SkipHeap *heap)
{
    int *p;
    for (p = fs_skipHeapFirst(heap);
         p != NULL;
         p = fs_skipHeapNext(heap, p))
    {
        printf("%d\n", *p);
    }
}



//  T E S T S


/**
 * Test simple get and put behavior on int heaps.
 */
void testIntSimple(void)
{
    int i;
    int *p;
    int *prev;
    
    fs_SkipHeap *heap = makeHeap(intTemplate);

    assert(fs_skipHeapIsEmpty(heap));
    assert(!fs_skipHeapIsFull(heap));
    assert(fs_skipHeapCount(heap) == 0);
    assert(fs_skipHeapRemaining(heap) == fs_skipHeapCapacity(heap));
    assert(fs_skipHeapCapacity(heap) > 0);
    assert(fs_skipHeapFirst(heap) == NULL);

    // fill the heap
    
    for (i = 0;; i++) {
        int n = isOdd(i) ? -i : i;
        assert(fs_skipHeapCount(heap) == i);
        if (!fs_skipHeapAdd(heap, &n))
            break;
        fs_skipHeapCheck(heap);
        assert(!fs_skipHeapIsEmpty(heap));
        assert(fs_skipHeapFirst(heap) != NULL);
    }
    fs_skipHeapCheck(heap);
    assert(fs_skipHeapIsFull(heap));
    assert(fs_skipHeapCount(heap) == fs_skipHeapCapacity(heap));
    assert(fs_skipHeapRemaining(heap) == 0);
    assert(fs_skipHeapFirst(heap) != NULL);

    // iterate over the heap

    for (prev = NULL, p = fs_skipHeapFirst(heap), i = 0;
         p != NULL;
         prev = p, p = fs_skipHeapNext(heap, p), i++)
    {
        if (prev != NULL) {
            assert(*prev < *p);
        }
    }
    assert(i == fs_skipHeapCapacity(heap));
}


/**
 * Run a test, with elements permuted in the specified order.
 */
void runPermutedTest(int max, int permuted[])
{
    int i;
    int *p;
    char *s;

    fs_SkipHeap *iheap = makeHeap(intTemplate);
    fs_SkipHeap *sheap = makeHeap(stringTemplate);

    char *words[] = {
        "alpha", "bravo", "charlie", "delta", "echo", "foxtrot", "golf",
        "hotel", "india", "juliet", "kilo", "lima", "mike", "november",
        "oscar", "papa", "quebec", "romeo", "sierra", "tango", "uniform",
        "victor", "whiskey", "x-ray", "yankee", "zulu", NULL
    };

#if 0
    printf("%d: ", max);
    for (i = 0; i < max; i++) {
        printf(" %d", i);
    }
    printf(" -");
    for (i = 0; i < max; i++) {
        printf(" %d", permuted[i]);
    }
    printf("\n");
#endif

    assert(fs_skipHeapIsEmpty(iheap));
    assert(fs_skipHeapCount(iheap) == 0);
    for (i = 0; i < max; i++) {
        assert(fs_skipHeapCount(iheap) == i);
        fs_skipHeapAdd(iheap, &permuted[i]);
        assert(!fs_skipHeapIsEmpty(iheap));
    }
    assert(fs_skipHeapCount(iheap) == max);

    for (i = 0, p = fs_skipHeapFirst(iheap);
         i < max;
         i++, p = fs_skipHeapNext(iheap, p))
    {
        assert(*p == i);
    }

    assert(fs_skipHeapIsEmpty(sheap));
    assert(fs_skipHeapCount(sheap) == 0);
    for (i = 0; i < max; i++) {
        assert(fs_skipHeapCount(sheap) == i);
        fs_skipHeapAdd(sheap, words[permuted[i]]);
        assert(!fs_skipHeapIsEmpty(sheap));
    }
    assert(fs_skipHeapCount(sheap) == max);

    for (i = 0, s = fs_skipHeapFirst(sheap);
         i < max;
         i++, s = fs_skipHeapNext(sheap, s))
    {
        assert(streq(s, words[i]));
    }

    free(iheap->buffer->u.data);
    free(iheap->buffer->shadow);
    free(iheap->buffer);
    free(iheap);

    free(sheap->buffer->u.data);
    free(sheap->buffer->shadow);
    free(sheap->buffer);
    free(sheap);
}


void testPermutations(int pivot, int max, int elements[])
{
    if (pivot < max) {
        int permuted[PERMUTATIONS];
        int next = pivot + 1;
        int i;
        for (i = 0; i < max; i++)
            permuted[i] = elements[i];
        testPermutations(next, max, permuted);
        for (i = next; i < max; i++) {
            permuted[pivot] = elements[i];
            permuted[i] = elements[i - 1];
            testPermutations(next, max, permuted);
        }
    }
    else {
        runPermutedTest(max, elements);
    }
}


/**
 * Test permuted insertions.
 */
void testPermute(void)
{
    int i;
    int elements[PERMUTATIONS];
    for (i = 0; i < PERMUTATIONS; i++) {
        elements[i] = i;
    }
    for (i = 0; i <= PERMUTATIONS; i++) {
        boolean logging = checkLogging;
        checkLogging = (i < 3);
        testPermutations(0, i, elements);
        checkLogging = logging;
    }
}


/**
 * Test searching for the nearest value.
 */
void testNearest(void)
{
    int i;
    int n;
    int *p;
    fs_SkipHeap *heap = makeHeap(intTemplate);

    n = 42;
    p = fs_skipHeapFindNearest(heap, &n);
    assert(p == NULL);

    i = 13;
    fs_skipHeapAdd(heap, &i);
    p = fs_skipHeapFindNearest(heap, &n);
    assert(*p == i);
    fs_skipHeapRemove(heap, p);
    assert(fs_skipHeapIsEmpty(heap));

    i = 144;
    fs_skipHeapAdd(heap, &i);
    p = fs_skipHeapFindNearest(heap, &n);
    assert(*p == i);

    i = 100;
    fs_skipHeapAdd(heap, &i);
    p = fs_skipHeapFindNearest(heap, &n);
    assert(*p == 100);

    n = 142;
    p = fs_skipHeapFindNearest(heap, &n);
    assert(*p == 144);

    n = 30000;
    p = fs_skipHeapFindNearest(heap, &n);
    assert(*p == 144);

    n = 122;
    p = fs_skipHeapFindNearest(heap, &n);
    assert(*p == 100 || *p == 144);

    n = 100;
    fs_skipHeapRemove(heap, &n);
    n = 144;
    fs_skipHeapRemove(heap, &n);

    for (i = 1; i > 0; i *= 2)
        fs_skipHeapAdd(heap, &i);

    for (i = 4; i < BIG && i > 0; i *= 2) {
        int j = i / 2;
        n = i - (i / 4);
        p = fs_skipHeapFindNearest(heap, &n);
        assert(*p == i || *p == j);
        n += 1;
        p = fs_skipHeapFindNearest(heap, &n);
        assert(*p == i);
        n -= 2;
        p = fs_skipHeapFindNearest(heap, &n);
        assert(*p == j);
    }
}


/**
 * Public entry point for testing.
 */
int main(int argc, char **argv)
{
    int i;
    int sizes[] = { 1024, 4096, 8192, 65536, 999, 0 };
    
    for (i = 0; sizes[i] != 0; i++) {
        int size = sizes[i];
        checkLogging = (size < 5000);
        initializeTemplates(size);
        testIntSimple();
        testPermute();
        testNearest();
    }

    printf("%s passed\n", argv[0]);
    return 0;
}
