/*------------------------------------------------------------------------------
 * RCS:  $Id: fssf-fs.c,v 1.1 2002/03/27 00:09:48 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fssf-fs.c
 *
 * Description  : File System State-Full (FSSF)
 *                Interface with FS
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "fssf.h"
#include "fssf-hash.h"

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_mcl
 *
 * Description  : Multi Component Lookup
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

fssf_status_t
fssf_mcl(fssf_cb_t *    cb,         // input
         fs_Handle *    objHandle,  // output
         fs_Attr *      objAttr)    // output

{
    fssf_openReq_t *    openReq = (fssf_openReq_t *) (cb->msgBuf);
    fssf_status_t       status = NFX_OK;
    uint32              reqSize;
    fs_MsgLookupReq *   req = NULL;
    esm_event_t         evt;
    
    // TODO!!! need MCL support from EFS
    // right now I am using LOOKUP so pathName is not used
    
    /* Initialize the request.
     */
    reqSize = sizeof(fs_MsgLookupReq) + strlen(openReq->fileName);
    req = (fs_MsgLookupReq *) fssf_ramAllocZero(reqSize);
    req->hdr.requestType = FS_REQ_LOOKUP;
    req->hdr.len = reqSize;
	req->hdr.attrType = FS_POST_ATTRS;
	req->hdr.uid = 0;   // root
	req->hdr.gid = 0;   // root
    req->hdr.proto = openReq->hdr.protocol;
    req->hdr.numGids =0;
    req->hdr.clientPrivate = cb->threadHandle;
    
    memcpy((void *) &(req->parentHandle),
           (const void *) &(openReq->parentHandle),
           sizeof(fs_Handle));
    req->nameLen = strlen(openReq->fileName);
    strcpy(req->fileName, openReq->fileName);
    
    /* Send the request to FS.
     */
    FSSF_STAT_FSLOOKUPREQ;
    fssf_sendMsg(fssfPort, fsPort, (void *) req, reqSize);
    thread_yield(&evt, FSSF_YIELD_TIMEOUT);

    /* A response has been received.
     */
    if (evt.event == FS_REQ_LOOKUP)
    {
        fs_MsgLookupResp *  rsp = (fs_MsgLookupResp *) evt.p1.u.p;
        
        status = rsp->hdr.fsStatus;
        memcpy((void *) objHandle, (const void *) &(rsp->fileHandle),
               sizeof(fs_Handle));
        if (status == NFX_OK)
        {
            NFX_ASSERT(rsp->numAttrs == 1);
            NFX_ASSERT(rsp->attr[0].attrType == FS_POST_ATTRS);
            memcpy((void *) objAttr, (const void *) &(rsp->attr[0]),
                   sizeof(fs_Attr));
        }
    }
    else
    {
        status = NFX_ERR;
    }
    
    return(status);
}

/*------------------------------------------------------------------------------
 * Name         : fssf_createFile
 *
 * Description  : Create a file
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

fssf_status_t
fssf_createFile(fssf_cb_t * cb,         // input
                fs_Handle * dirHandle,  // input
                fs_Handle * fileHandle, // output
                fs_Attr *   fileAttr)   // output

{
    fssf_openReq_t *    openReq = (fssf_openReq_t *) (cb->msgBuf);
    fssf_status_t       status = NFX_OK;
    uint32              reqSize;
    fs_MsgCreateReq *   req = NULL;
    esm_event_t         evt;
    
    /* Initialize the request.
     */
    reqSize = sizeof(fs_MsgCreateReq) + strlen(openReq->fileName);
    req = (fs_MsgCreateReq *) fssf_ramAllocZero(reqSize);
    req->hdr.requestType = FS_REQ_CREATE;
    req->hdr.len = reqSize;
	req->hdr.attrType = FS_POST_ATTRS;
	req->hdr.uid = 0;   // root
	req->hdr.gid = 0;   // root
    req->hdr.proto = openReq->hdr.protocol;
    req->hdr.numGids =0;
    req->hdr.clientPrivate = cb->threadHandle;
    memcpy((void *) &(req->parentHandle), (const void *) dirHandle,
           sizeof(fs_Handle));
    memcpy((void *) &(req->u), (const void *) &(openReq->u), sizeof(req->u));
    req->flags = openReq->createHow;
    req->nameLen = strlen(openReq->fileName);
    strcpy(req->fileName, openReq->fileName);

    /* Send the request to FS.
     */
    FSSF_STAT_FSCREATEREQ;
    fssf_sendMsg(fssfPort, fsPort, (void *) req, reqSize);
    thread_yield(&evt, FSSF_YIELD_TIMEOUT);

    /* A response has been received.
     */
    if (evt.event == FS_REQ_CREATE)
    {
        fs_MsgCreateResp *  rsp = (fs_MsgCreateResp *) evt.p1.u.p;
        
        status = rsp->hdr.fsStatus;
        memcpy((void *) fileHandle, (const void *) &(rsp->fileHandle),
               sizeof(fs_Handle));
        if (status == NFX_OK)
        {
            NFX_ASSERT(rsp->numAttrs == 1);
            NFX_ASSERT(rsp->attr[0].attrType == FS_POST_ATTRS);
            memcpy((void *) fileAttr, (const void *) &(rsp->attr[0]),
                   sizeof(fs_Attr));
        }
    }
    else
    {
        status = NFX_ERR;
    }
    
    return(status);
}

/*----------------------------------------------------------------------------*/
