/*------------------------------------------------------------------------------
 * RCS:  $Id: fssf-owner.c,v 1.1 2002/03/27 00:09:48 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fssf-owner.c
 *
 * Description  : File System State-Full (FSSF)
 *                Operations on the ownerTab.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "fssf.h"
#include "fssf-hash.h"

/*------------------------------------------------------------------------------
 * Constants.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static fssf_status_t
fssf_delOwner(fssf_owner_t **   owner);

/*------------------------------------------------------------------------------
 * Local funstions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_delOwner
 *
 * Description  : Remove an owner in ownerTab.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static fssf_status_t
fssf_delOwner(fssf_owner_t **   owner)  // input/ouput

{
    fssf_status_t   status = NFX_OK;

    FSSF_DBG_PRINT(("FSSF: fssf_delOwner(0x%llx)", (*owner)->owner));

#ifdef ISABELLE // test only
    {
        fssf_owner_t * ow = NULL;
        
        fssf_getOwner((*owner)->owner, &ow);
        if (ow == NULL)
        {
            printf("owner(0x%llx) not in ownerTab.\n", (*owner)->owner);
        }
        else
        {
            printf("owner(0x%llx) found!\n", (*owner)->owner);
        }    
    }
#endif // ISABELLE
    
    if (fssf_chash_remove(ownerTab, (void *) &((*owner)->owner)) < 0)
    {
        FSSF_ERR_PRINT(("FSSF ERROR: fssf_chash_remove() failed"));
        status = NFX_ERR;
    }
    
    return(status);
}

/*------------------------------------------------------------------------------
 * Global funstions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_initOwnerTab
 *
 * Description  : Initialize the client with open file(s) table.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_initOwnerTab(int initSize)

{
    NFX_SPIN_LOCK(&fssf_tab_lock);
    ownerTab = fssf_chash_create(FSSF_HASH_BUCKETS_OWNER,
                            fssf_owner_hashFunc,
                            fssf_owner_hashCompare,
                            fssf_owner_hashDestroy);
    NFX_SPIN_UNLOCK(&fssf_tab_lock);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_owner_hashFunc
 *
 * Description  : Hash function for the owners table.
 *                Use the client id.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

uint32
fssf_owner_hashFunc(const void * key)

{
    return(*((fssf_clientid_t *) key) % FSSF_HASH_BUCKETS_OWNER);
}

/*------------------------------------------------------------------------------
 * Name         : fssf_owner_hashCompare
 *
 * Description  : Compare function for the owners table.
 *                The 2 client ids should be identical.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32
fssf_owner_hashCompare(const void * key1,
                       const void * key2)
    
{
    if (*((fssf_clientid_t *) key1) == *((fssf_clientid_t *) key2))
   {
        return(0);  // it matches
    }
    else
    {
        return(1);
    }
}

/*------------------------------------------------------------------------------
 * Name         : fssf_owner_hashDestroy
 *
 * Description  : Destroy function for the owners table.
 *                Free the "owner" element.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_owner_hashDestroy(void *   key,
                       void *   obj)
    
{
    fssf_ramDealloc(key);
    fssf_ramDealloc(obj);
}

/*------------------------------------------------------------------------------
 * Name         : fssf_getOwner
 *
 * Description  : Return a pointer to the matching entry in ownerTab.
 *                NULL if not found.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_getOwner(fssf_clientid_t   id,     // input
              fssf_owner_t **   owner)  // ouput

{
    *owner = (fssf_owner_t *) fssf_chash_find(ownerTab, (void *) &id);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_addNewOwner
 *
 * Description  : Add a new owner in ownerTab.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

fssf_status_t
fssf_addNewOwner(fssf_clientid_t    id,         // input
                 fssf_owner_t **    newOwner)   // ouput

{
    fssf_status_t       status = NFX_OK;
    fssf_clientid_t *   newId = NULL;
    
    FSSF_DBG_PRINT(("FSSF: fssf_addNewOwner(0x%llx)", id));
    
    newId = (fssf_clientid_t *) fssf_ramAllocZero(sizeof(fssf_clientid_t));
    *newId = id;
    
    *newOwner = (fssf_owner_t *) fssf_ramAllocZero(sizeof(fssf_owner_t));
    (*newOwner)->owner = id;
    DLL_INIT(&((*newOwner)->stateList));

    if (fssf_chash_insert(ownerTab, (void *) newId, (void *) *newOwner) < 0)
    {
        FSSF_ERR_PRINT(("FSSF ERROR: fssf_chash_insert() failed" ));
        fssf_ramDealloc(*newOwner);
        status = NFX_ERR;
    }
    
    return(status);
}

/*------------------------------------------------------------------------------
 * Name         : fssf_addStateInOwner
 *
 * Description  : Add a new state for a given owner.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_addStateInOwner(fssf_state_t * state,  // input
                     fssf_owner_t * owner)  // input / output
    
{
    NFX_ASSERT((state != NULL) && (owner != NULL));
    DLL_INSERT_TAIL(&(owner->stateList), state, statePerOwner);    
    state->owner = owner;
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_delStateFromOwner
 *
 * Description  : Remove a state for a given owner.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

fssf_status_t
fssf_delStateFromOwner(fssf_state_t *   state)  // input / output
    
{
    fssf_status_t   status = NFX_OK;
    
    NFX_ASSERT((state != NULL) && (state->owner != NULL));
    
    DLL_REMOVE(&(state->owner->stateList), state, statePerOwner);
    if (DLL_IS_EMPTY(&(state->owner->stateList)))
    {
        FSSF_DBG_PRINT(("FSSF: remove owner (0x%llx)", state->owner->owner));
        status = fssf_delOwner(&(state->owner));
    }
    state->owner = NULL;
    
    return(status);
}

/*----------------------------------------------------------------------------*/
