/*
 *-----------------------------------------------------------------
 *
 * Name: mem-api.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2004, OnStor Inc.,
 *
 * Description: Contains utility routines to convert memory address
 *              from cache to uncache, from virtual to physical etc.
 *              Ported from sm-gt64240/mem-api.c.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 05/14/04
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/hwMap.h>
#include <exports/mem-api.h>

#ifndef PROM
/* for proms the memorysize variables is declared in main.c */
uint64	memorysize = 0;
#endif
uint32	memorysize_MB = 0;
uint32	pcisize = 0;

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 *-----------------------------------------------------------------
 */
void
mem_map_init()
{

#ifdef FC_USE_NEW_MEM_MAP
	extern void mem_map_tlb_init();
	extern void mem_map_gt_init();

	mem_map_tlb_init();

	mem_map_gt_init();
#endif

}

/*-----------------------------------------------------------------
 * Name:	is_valid_addr(adr)
 *
 * Description: returns 0 or 1
 *
 *-----------------------------------------------------------------
 */
int32
is_valid_addr(
	address_t	adr
)
{
	if ((adr >= K0BASE) && (adr < K2BASE)) 
		return 1;

	/* R9K local memory - cached */
	if ((adr >= LOCAL_MEM_BASE) && 
		(adr < (LOCAL_MEM_BASE+LOCAL_MEM_SIZE)))
		return 1;

	/* GT memory - cached */
	if ((adr >= SW_RAM_BASE_CACHED) && 
		(adr < ((address_t)SW_RAM_BASE_CACHED + SW_RAM_SIZE-1)))
		return 1;

	/* GT memory - uncached */
	if ((adr >= (address_t)SW_RAM_BASE_UNCACHED) && 
		(adr < ((address_t)SW_RAM_BASE_UNCACHED + SW_RAM_SIZE-1)))
		return 1;
	
	return 0;
}

/*-----------------------------------------------------------------
 * Name: virt_to_phys
 *
 * Description: converts virtual address to physical
 *
 *-----------------------------------------------------------------
 */
address_t
virt_to_phys(
	address_t	virt
)
{
	address_t	phys;

	/* 
	 * Bobcat memory map:
	 * virtual             physical       size
	 * 0x80000000          0x00000000     256MB    Cached
	 * 0xA0000000          0x00000000     256MB    Uncached
	 * 0x10000000          0x20000000     256MB    Cached
	 * 0x40000000          0x40000000     256MB    Cached
	 * 0x50000000          0x50000000     256MB    Cached
	 * 0x60000000          0x60000000     256MB    Cached
	 * 0x70000000          0x70000000     256MB    Cached
	 * 0xC0000000          0x40000000     256MB    Uncached
	 * 0xD0000000          0x50000000     256MB    Uncached
	 * 0xE0000000          0x60000000     256MB    Uncached
	 * 0xF0000000          0x70000000     256MB    Uncached
	 */
	if ((virt >= K0BASE) && (virt < K2BASE))
		phys = KVA_TO_PA(virt);
	else if ((virt >= LOCAL_MEM_BASE) && 
			 (virt < (LOCAL_MEM_BASE+LOCAL_MEM_SIZE)))
		phys = virt - LOCAL_MEM_BASE + HW_LOCAL_MEM_BASE;
	else if ((virt >= SW_RAM_BASE_CACHED) && 
			 (virt < ((address_t)SW_RAM_BASE_CACHED+SW_RAM_SIZE-1)))
		phys = virt;
	else if ((virt >= (address_t)SW_RAM_BASE_UNCACHED) && 
			 (virt < ((address_t)SW_RAM_BASE_UNCACHED+SW_RAM_SIZE-1)))
		phys = virt - SW_RAM_BASE_UNCACHED + HW_GT_MEM_0_BASE;
	else {
		address_t   ra;

		ra = (address_t)get_ra();
		panic("Invalid virt, %p, ra = %p\n", virt, ra);
		phys = 0;
	}

/*
	printf("%s: virt = %x phys=%x\n", 
			__FUNCTION__, (uint32)virt, (uint32)phys);
*/
	return phys;
}

/*-----------------------------------------------------------------
 * Name: phys_to_virt_cached
 *
 * Description: Converts physical address to virtual cached address
 *
 *-----------------------------------------------------------------
 */
address_t
phys_to_virt_cached(
	address_t	phys
)
{
	address_t	virt;

	if ((phys >= 0x0) && (phys < LOCAL_MEM_SIZE))
		virt = PHYS_TO_K0(phys);
	else if ((phys >= HW_LOCAL_MEM_BASE) && 
			 (phys < ((address_t)HW_LOCAL_MEM_BASE + LOCAL_MEM_SIZE)))
		virt = phys - HW_LOCAL_MEM_BASE + LOCAL_MEM_BASE;
	else if ((phys >= HW_GT_MEM_0_BASE) && 
			 (phys < ((address_t)HW_GT_MEM_0_BASE + SW_RAM_SIZE-1)))
		virt = phys - HW_GT_MEM_0_BASE + SW_RAM_BASE_CACHED;
	else {
		address_t   ra;

		ra = (address_t)get_ra();
		panic("Invalid virt, %p, ra = %p\n", phys, ra);
		virt = 0;
	}

/*
	printf("%s: virt = %x phys=%x\n", 
			__FUNCTION__, (uint32)virt, (uint32)phys);
*/
	return virt;
}

/*-----------------------------------------------------------------
 * Name: phys_to_virt_uncached
 *
 * Description: Converts physical address to virtual uncached address
 *
 *-----------------------------------------------------------------
 */
address_t
phys_to_virt_uncached(
	address_t	phys
)
{
	address_t	virt;

	if ((phys >= 0x0) && (phys < LOCAL_MEM_SIZE))
		virt = PHYS_TO_K1(phys);
	else if ((phys >= HW_GT_MEM_0_BASE) && 
			 (phys < ((address_t)HW_GT_MEM_0_BASE + SW_RAM_SIZE-1)))
		virt = phys - HW_GT_MEM_0_BASE + SW_RAM_BASE_UNCACHED;
	else {
		address_t   ra;

		ra = (address_t)get_ra();
		panic("Invalid virt, %p, ra = %p\n", phys, ra);
		virt = 0;
	}

/*
	printf("%s: virt = %x phys=%x\n", 
			__FUNCTION__, (uint32)virt, (uint32)phys);
*/
	return virt;
}

/*-----------------------------------------------------------------
 * Name: cached_to_uncached
 *
 * Description: Converts cached address to uncached address
 *
 *-----------------------------------------------------------------
 */
address_t
cached_to_uncached(
	address_t	cached
)
{
	address_t	ra;

	if ((cached >= K0BASE) && (cached < K1BASE))
		return K0_TO_K1(cached);

	if ((cached >= SW_RAM_BASE_CACHED) &&
		(cached < ((address_t)SW_RAM_BASE_CACHED + SW_RAM_SIZE -1)))
		return (cached - SW_RAM_BASE_CACHED + SW_RAM_BASE_UNCACHED);

	ra = (address_t)get_ra();
	panic("Invalid cached addr, %p, ra = %p\n", cached, ra);

	return 0;
}

/*-----------------------------------------------------------------
 * Name: uncached_to_cached
 *
 * Description: Converts uncached address to cached address
 *
 *-----------------------------------------------------------------
 */
address_t
uncached_to_cached(
	address_t	uncached
)
{
	address_t	ra;

	if ((uncached >= K1BASE) && (uncached < K2BASE))
		return K1_TO_K0(uncached);

	if ((uncached >= (address_t)SW_RAM_BASE_UNCACHED) &&
		(uncached < ((address_t)SW_RAM_BASE_UNCACHED + SW_RAM_SIZE-1)))
		return (uncached - SW_RAM_BASE_UNCACHED + SW_RAM_BASE_CACHED);

	ra = (address_t)get_ra();
	panic("Invalid uncached addr, %p, ra = %p\n", uncached, ra);

	return 0;
}

