/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ipmd-vif.c
 *
 * RCS:  $Id: ipmd-vif.c
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: virtual interface functions
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/7/2002
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <neteee/eee.h>
#include <neteee/eee_if.h>
#include <sys/types.h>
#include <signal.h>

#include "nfx-error.h"
#include "nfx-types.h"
#include "nfx-defs.h"
#include "nfx-ports.h"
#include "exports/anpssc-api.h"
#include "exports/ipm-api.h"
#include "exports/evm-api.h"
#include "exports/ncm-api.h"
#include "exports/event-api.h"
#include "exports/opt-api.h"
#include "exports/event-api.h"
#include "../sm-ipm/ipm.h"
#include "ipmd-kern.h"
#include "ipmd-vif.h"

#if defined(LINK_AGG)

// #define NO_TXRX

#define IPMD_VIF_PRINTF(a)	//printf a

#define IPMD_VIF_EMPTY(_vif) (*((_vif)->vif_name) == '\0')
#define IPMD_VIF_GETLINK(_vif, _link) (&((_vif)->vif_links[_link][0]))

#define IPMD_VIF_TOTXRX()														\
{																				\
	if (txrx_reply == 1) {														\
		IPMD_VIF_PRINTF(("send msg to slot %d, cpu %d, size %d\n",				\
				ipm_vifp->vif_slot, ipmd_txrx_cpu, ipm_hdrp->ipmh_size));		\
		if (ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipm_vifp->vif_slot,		\
						 ipmd_txrx_cpu, "ipm",									\
						 ipm_hdrp, IPM_MSG_MAX_SIZE, 0, -1) <= 0) {				\
			ipm_hdrp->ipmh_flags = IPM_MSG_CMD_TIMEOUT;							\
			ipm_stats.err_fwd ++;												\
			rc = NFX_ERR;														\
		}																		\
		else if (!(ipm_hdrp->ipmh_flags & IPM_MSG_CMD_OK))	{					\
			ipm_stats.adm_fwd ++;												\
			rc = NFX_ERR;														\
		}																		\
	}																			\
	else {																		\
		ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_NREPLY;							    \
		ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipm_vifp->vif_slot,			\
					 ipmd_txrx_cpu, "ipm", NULL, 0, 0, -1);						\
		ipm_stats.adm_fwd ++;													\
	}																			\
}																				


extern uint64			resp_buf[];
extern int				ipmd_txrx_slot;
extern int				ipmd_txrx_cpu;
extern int				ipmd_ssc_slot;
extern int				ipmd_ssc_cpu;
extern ipm_stats_t		ipm_stats;

/* keep a record of vif in the system */
static ipm_vif_t ipmd_ssc_vif[IPM_MAX_VIF];

#define		PREF_STR		"pref."
static void ipmd_AddVif(void	*msg, char	*app, uint32 flags);
static void ipmd_PrefLinkVif(void	*msg, char	*app, uint32 flags);

static int32	ipmd_parseVifName(char	*vifname, int *slot, int *link)
{
	char	*tmpChar;
	
	if (strncmp(vifname, IPM_VIF_STR, strlen(IPM_VIF_STR)))
		return NFX_ERR;
	
	/* get slot num */
	tmpChar = ((char*)vifname) + strlen(IPM_VIF_STR);
	if  ((*tmpChar < '0') || (*tmpChar > '9')) 
		return NFX_ERR;
	*slot = atoi(tmpChar);
	
	/* bypass dot */
	tmpChar ++;
	if (*tmpChar != '.')
		return NFX_ERR;
	
	/* get Link */
	tmpChar ++;
	*link = atoi(tmpChar);
	
	return NFX_OK;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_SaveVif
 *
 * Description: save virtual info to agile.conf. 
 *				this function is called whenever make a
 *				configuration change.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/12/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_SaveVif(ipm_vif_t	*myVifp)
{
	char	option[OPT_MAX_OPTION_LEN];
	char	value[MAX_VALUE_LEN];
	char	*tmpLink;
	int		i;
	
	if (!myVifp) {
		IPMD_VIF_PRINTF(("%s: no incoming vif\n", __FUNCTION__));
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}

	/* save vif */
	sprintf(option, "%s.%s", OPTIONS_IPM_VIF_CONF, myVifp->vif_name);
	value[0] = '\0';
	for (i = 0; i < IPM_MAX_VIF; i++) {
		tmpLink = IPMD_VIF_GETLINK(myVifp, i);
		if (*tmpLink == '\0')
			continue;
		
		strcat(value, tmpLink);
		strcat(value, ",");
	}
	
	/* save vif pref */
	if ((myVifp->vif_flags & IPM_VIF_MODE_FO))	{
		strcat(value, PREF_STR);
		strcat(value, myVifp->vif_prefLink);
	}

	/* send to agile.conf */
	if (setRegistryOption(option, value) != NFX_OK) 
		IPM_LOGERR(IPM_ERR_VIF);

	return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_UnsaveVif
 *
 * Description: remove virtual info to agile.conf. 
 *				this function is called whenever vif
 *				is removed
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/12/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_UnsaveVif(ipm_vif_t	*myVifp)
{
	char	option[OPT_MAX_OPTION_LEN];
	char		**vifPp;
	
	if (!myVifp) {
		IPMD_VIF_PRINTF(("%s: no incoming vif\n", __FUNCTION__));
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}
	
	sprintf(option, "%s.%s", OPTIONS_IPM_VIF_CONF, myVifp->vif_name);
	if ((vifPp = get_cfg_val(option)) == NULL) {
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}

	/* remove vif */
	del_cfg_obj(vifPp);
	free_cfg_val(vifPp);
	
	return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_RestoreVif
 *
 * Description: restore all vif configurations.
 *				this function is called in two places:
 *				1) when mcp first started.
 *				2) when txrx restarted
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/12/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_RestoreVif()
{
	char		*vifKey;
	char		**vifNameList;
	char		**vifPp;
	char		*strp;
	ipm_mvif_t	*ipm_vifp;
	ipm_mhdr_t	*ipm_hdrp;
	int			msgSize;
	int			j, i;
	
	
	vifKey = OPTIONS_IPM_VIF_CONF;
	vifKey = "ipm.vif.*";
	
	if ((vifNameList = get_cfg_val(vifKey)) == NULL)		
		return;
	
	msgSize = sizeof(ipm_mvif_t) + sizeof(ipm_mhdr_t);
	
	ipm_hdrp = (ipm_mhdr_t*)malloc(msgSize);
	ipm_hdrp->ipmh_flags = 0;
	ipm_hdrp->ipmh_size = msgSize;

	ipm_vifp = (ipm_mvif_t*)(ipm_hdrp + 1);
	
	for (vifPp = vifNameList; *vifPp; vifPp = vifPp+2) {
		bzero(ipm_vifp, sizeof(ipm_mvif_t));
		i = j = 0;

	    ipm_hdrp->ipmh_mtype = IPM_MSG_VIF_CREAT;      
		
		/* tmp is option */
		strp = *vifPp;
		
		/* get vif name */
		strcpy(ipm_vifp->vif_name, (strp + strlen(OPTIONS_IPM_VIF_CONF) + 1));
		
		ipm_vifp->vif_slot = ipmd_txrx_slot;
		
		/* tmp is value,
		 * "fp1.0,fp1.1,pref.fp1.0 
		 */
		strp = *(vifPp+1);
		
		/* copy all links */
		while (*strp) {
			/* get prefered link */
			if (strncmp(strp, PREF_STR, strlen(PREF_STR)) == 0) {
				strp += strlen(PREF_STR);
				strcpy(ipm_vifp->vif_prefLink, strp);
				ipm_vifp->vif_flags = IPM_VIF_MODE_FO;
				break;
			}
			
			/* end of a link,  */
			if (*strp == ',') {
				j ++;			/* advance to next link */
				i = 0;
				ipm_vifp->vif_linkCnt ++;

				
			}
			else {				/* cpy by each character */
				ipm_vifp->vif_links[j][i] = *strp;
				i ++;
			}
			strp ++;
		}

		if (!(ipm_vifp->vif_flags & IPM_VIF_MODE_FO))
			ipm_vifp->vif_flags = IPM_VIF_MODE_LA;
		
		/* print out msg */
		IPMD_VIF_PRINTF(("RESTORE VIF****\n"));
		
		//ipmd_PrintVif(ipm_vifp, ipm_hdrp->ipmh_mtype);
		
		/* send to kernel and txrx */
		ipmd_AddVif(ipm_hdrp, NULL, IPMD_VIFCONF_RESTORE);

		if ((ipm_vifp->vif_flags & IPM_VIF_MODE_FO)) {
			ipm_hdrp->ipmh_mtype = IPM_MSG_VIF_PREF;
			ipmd_PrefLinkVif(ipm_hdrp, NULL, IPMD_VIFCONF_RESTORE);
		}
	}
	
	free_cfg_val(vifNameList);
	free(ipm_hdrp);
	
	return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_VifSendCcm
 *
 * Description:  send virtual intf configuration to ccm.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/12/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_VifSendCcm(ipm_mhdr_t	*ipm_hdrp) 
{
	int32				mtype = ipm_hdrp->ipmh_mtype;
	int32				flags = ipm_hdrp->ipmh_flags;
	ipm_mvif_t			*ipm_vifp;
	nfx_eventMsg_t		evt_msg;

	/* if configuration not ok, don't send */
	if (!(flags & IPM_MSG_CMD_OK))
		return;
	
	/* post the event to eventd */
	memset (&evt_msg, 0, sizeof(evt_msg));
	evt_msg.msgType = NFX_EVENT_POST;
	gethostname (evt_msg.nodeName, sizeof(evt_msg.nodeName));

	switch (mtype) {
	case IPM_MSG_VIP_ADD:
	case IPM_MSG_VIP_DEL:
		/* get old msg ptr */
		ipm_vifp = (ipm_mvif_t*)(ipm_hdrp + 1);

		evt_msg.eventType = NFX_EVENT_IP_IF;

		/* fill in msgs */
		sprintf(evt_msg.param.ipIfEvent.portName, "%s:%s",
				evt_msg.nodeName, ipm_vifp->vif_name);
		IPMD_VIF_PRINTF(("port name is %s\n", 
						 evt_msg.param.ipIfEvent.portName));
		
		evt_msg.param.ipIfEvent.ipa = ipm_vifp->vif_ip;
		evt_msg.param.ipIfEvent.mask = ipm_vifp->vif_mask;
		if (mtype ==  IPM_MSG_VIP_DEL)
			evt_msg.param.ipIfEvent.state = EVENT_STATE_REMOVE;
		else
			evt_msg.param.ipIfEvent.state = EVENT_STATE_UP;
		break;
		
	default:
		return;
	}

	sendAgileMsg (-1, &evt_msg, sizeof(evt_msg),
	              "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
	
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_UpdateActLink
 *
 * Description: update active link, triggered by event sent from
 *				txrx. This function will be called
 *				on when a) failover occurs. b) perfered link
 *				become operational. c) ipmd_prefLinkVif has been
 *				called, and the new prefLink is not operational.
 *
 *  In vif msg, vif->prefLink:	old active link.
 *				vif->prefAct:	new active link.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_UpdateActLink(void	*msg)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1);
	char		*vifName = ipm_vifp->vif_name;
	ipm_vif_t	*myVifp;

	if ((myVifp = ipmd_FindVifByName(vifName)) == NULL) {
		IPMD_VIF_PRINTF(("bad vif name %s\n", vifName));
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}

	/* sanity check: the current actLink should be old 
	 * actLink in msg.
	 */
	if (strcmp(ipm_vifp->vif_prefLink, myVifp->vif_actLink)) {
		IPMD_VIF_PRINTF(("bad actLink msgLink %s,  locLink %s\n", 
						 ipm_vifp->vif_prefLink,myVifp->vif_actLink));
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}

	/* sanity check: new link exist */
	if (ipmd_FindVifByLink(ipm_vifp->vif_actLink) != myVifp) {
		IPMD_VIF_PRINTF(("new act link %s not in vif %s\n", 
						 ipm_vifp->vif_actLink, myVifp->vif_name));
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}
	
	/* switch IP configuration */
	ipmd_SwitchLinkIP(myVifp->vif_actLink, ipm_vifp->vif_actLink);
	
	/* update to new  actLink */
	memcpy(myVifp->vif_actLink, ipm_vifp->vif_actLink, IPM_STR_SIZE);
	
	return;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_SwitchLinkIP
 *
 * Description: switch ip addresses from one link to the other.

 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
void
ipmd_SwitchLinkIP(char *srcLink,  char *dstLink)
{
	ipm_ip_t	*ipm_ipp;
	ipm_if_t	ipm_if;
	int			need=0, count=0, i;
	ipm_rt_t	*ipm_rtp;
	ipm_arp_t	*ipm_arpp;
	

	IPMD_VIF_PRINTF(("%s: switch link from %s to %s\n",
					 __FUNCTION__, srcLink, dstLink));
	
	/* dskLink exist */
	if (ipmd_GetKernIfByName(dstLink, &ipm_if, sizeof(ipm_if), &count, &need)
		!= IPM_GET_OK || count == 0) {
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}
	
	/* get ip associated with interface */
	ipm_ipp = (ipm_ip_t*)resp_buf;
	need=0, count=0;
	if (ipmd_GetKernIpByName(srcLink, ipm_ipp,IPM_SSC_BUFSIZE*8, &count, &need)
		!= IPM_GET_OK) {
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}
	
	/* IP switch interfaces */
	for (i = 0; i < count; i++) {
		ipmd_DelKernIp(srcLink, ipm_ipp->ip_addr);
		ipmd_AddKernIp(dstLink, ipm_ipp->ip_addr, ipm_ipp->ip_mask);
		ipm_ipp++;
		
	}

	/* need swich rt as well ???? */
	ipm_rtp = (ipm_rt_t*)resp_buf;
	need=0, count=0;
	if (ipmd_GetKernRtTbl(ipm_rtp, IPM_SSC_BUFSIZE*8, &count, &need)
		!= IPM_GET_OK) {
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}
	for (i = 0; i < count; i++) {
		if (strcmp(ipm_rtp->rt_ifname, srcLink) == 0) {
			IPMD_VIF_PRINTF(("%s: switch route %x\n",
							 __FUNCTION__, ipm_rtp->rt_addr)); 
			ipmd_DelKernRt(ipm_rtp->rt_addr, ipm_rtp->rt_mask, ipm_rtp->rt_nhop);
			ipmd_AddKernRt(ipm_rtp->rt_addr, ipm_rtp->rt_mask, ipm_rtp->rt_nhop);
		}
		ipm_rtp++;
	}
	
	/* remove all arp on srcLInk */
	ipm_arpp = (ipm_arp_t*)resp_buf;
	need=0, count=0;
	if (ipmd_GetKernArpTbl(ipm_arpp, IPM_SSC_BUFSIZE*8, &count, &need)
		!= IPM_GET_OK) {
		IPM_LOGERR(IPM_ERR_VIF);
		return;
	}
	for (i = 0; i < count; i++) {
		if (strcmp(ipm_arpp->arp_ifname, srcLink) == 0) {
			IPMD_VIF_PRINTF(("%s: switch arp %x\n",
							 __FUNCTION__, ipm_arpp->arp_addr)); 
			ipmd_DelKernArp(ipm_arpp->arp_addr);
		}
		ipm_arpp ++;
	}
			
	IPMD_VIF_PRINTF(("link switch completed\n"));
	return;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetIfFromTxrx
 *
 * Description: get link from txrx and copy to given buffer
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_CMDERR_TIMEOUT */
ipmd_GetIfFromTxrx(char	*linkName, ipm_if_t	*ipm_ifp)
{
	ipm_mhdr_t	*ipm_mhdrp = 0;
	ipm_mintf_t	*ipm_mintfp;
	boardType_t btype;
	char		nodeName[48];
	int			slot, link, cpu;
	int32		rc = IPM_GET_OK;
			
	ipm_mhdrp = malloc(sizeof(ipm_mhdr_t) + sizeof(ipm_mintf_t));
	if (!ipm_mhdrp) {
		IPM_LOGERR(IPM_ERR_BUF);
		rc =  IPM_GET_ERR;
		goto done;
	}
			
	/* fill in msg hdr */
	ipm_mhdrp->ipmh_mtype =  IPM_MSG_INTF_REQ;
	ipm_mhdrp->ipmh_size = sizeof(ipm_mhdr_t) + sizeof(ipm_mintf_t);
	ipm_mhdrp->ipmh_flags = 0;
			
	/* fill in interface request info */
	if (getPortInfo (linkName, nodeName, &btype, &slot, &link)) {
		IPM_LOGERR(IPM_ERR_BUF);
		rc =  IPM_GET_ERR;
		goto done;
	}
	cpu = GET_CPU_FROM_ULINK(link);
	link = GET_LINK_FROM_ULINK(link);
			
	ipm_mintfp = (ipm_mintf_t*)(ipm_mhdrp+1);
	ipm_mintfp->if_slot = slot;
	ipm_mintfp->if_link  = link;
	ipm_mintfp->if_cpu  =  cpu;
	ipm_mintfp->if_dtype = (ushort16)btype;
			
	/* send to txrx for reply */
	if (ipmd_SendMsg(ipm_mhdrp, ipm_mhdrp->ipmh_size, slot, cpu, "ipm", 
					 ipm_mhdrp, ipm_mhdrp->ipmh_size, 0, -1) <= 0) {
		/* txrx did not response */
		rc = IPM_CMDERR_TIMEOUT;
		goto done;
	}
			
	if (!(ipm_mhdrp->ipmh_flags & IPM_MSG_CMD_OK)) {
		rc =  IPM_GET_ERR;
		goto done;
	}

	/* we have a possitive reply, parsing */
	strcpy(ipm_ifp->if_name, linkName);
	memcpy(ipm_ifp->if_mac, ipm_mintfp->if_mac, 6);
	ipm_ifp->if_flags = ipm_mintfp->if_flags;
	
	ipm_ifp->if_auxFlags = 0;
	if (ipm_mintfp->if_flags & IFF_RUNNING)
		ipm_ifp->if_auxFlags |= IPM_AUXIF_RUNNING;
	
	if (ipm_mintfp->if_flags & IFF_LINK0)
		ipm_ifp->if_auxFlags |= IPM_AUXIF_AUTO;
	
	ipm_ifp->if_mtusize = ipm_mintfp->if_mtusize;
 
  done:
	if (ipm_mhdrp)
		free (ipm_mhdrp);

	return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_CmpVif
 *
 * Description: find vif on ssc.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32 
ipmd_CmpVif(ipm_vif_t	*vifp1,	ipm_vif_t	*vifp2)
{
	return NFX_OK;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_FindVifByName
 *
 * Description: find vif on ssc.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
ipm_vif_t*
ipmd_FindVifByName(char	*vifName) 
{
	int i;
	 
	for ( i = 0 ; i<IPM_MAX_VIF; i++) {
		if (strcmp(ipmd_ssc_vif[i].vif_name, vifName) == 0)
			return &ipmd_ssc_vif[i];
	}

	return NULL;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_FindVifByLink
 *
 * Description: find vif by given link
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
ipm_vif_t*
ipmd_FindVifByLink(char	*linkName) 
{
	int i, j;
	 
	for ( i = 0 ; i<IPM_MAX_VIF; i++) {
		if (!IPMD_VIF_EMPTY(&ipmd_ssc_vif[i])) {
			for (j=0; j < IPM_MAXLINK_VIF; j++) {
				char	*myLinkName;
				
				myLinkName = IPMD_VIF_GETLINK(&ipmd_ssc_vif[i], j);
				if (strcmp(myLinkName, linkName) == 0)
					return &ipmd_ssc_vif[i];
				
			}
		}
	}

	return NULL;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_VifHasIP
 *
 * Description: does interface has IP,
 *				
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
char*							/* linkName, NULL */
ipmd_VifHasIP(uint32 ip, ipm_vif_t	*vifp)
{
	char		*tmpLink, *myLink;
	ipm_vif_t	*tmpVifp, *myVifp = vifp;
	ipm_ip_t	*ipm_ipp = (ipm_ip_t*)resp_buf;
	int			need=0, count=0;
	int			i;
	

	/* does IP address exit */
	if (ipmd_GetKernIpByIp(ip, 1, 
						   ipm_ipp, sizeof(*ipm_ipp), 
						   &count, &need) != IPM_GET_OK) 
		return NULL;
	
	/* get virtual that has this IP */
	tmpLink = ipm_ipp->ip_ifname;
	tmpVifp = ipmd_FindVifByLink(tmpLink);
		

	if (tmpVifp != myVifp) 
		return NULL;
	else {
		for (i=0; i<IPM_MAXLINK_VIF; i++) {
			myLink = IPMD_VIF_GETLINK(myVifp, i);
			if (strcmp(tmpLink, myLink) == 0)
				return myLink;
		}
	}
	
	return NULL;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_AddLinkToVif
 *
 * Description: add specified link to vif
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK, NFX_ERR, IPM_GET_NO_BUF */
ipmd_AddLinkToVif(char	*linkName, ipm_vif_t *vifp, uint32 flags) 
{
  ipm_if_t	*ipm_ifp = (ipm_if_t*)resp_buf;
  ipm_ip_t	*ipm_ipp = (ipm_ip_t*)resp_buf;
  int		count=0, need=0, i;
  char		*myLink = NULL, *tmpLink;
  ipm_vif_t	*tmpVifp;

  /* is link name ok */
  if (strncmp(linkName, "sc", 2) == 0)
	  return NFX_ERR;

  IPMD_VIF_PRINTF(("add link %s, flag is %x\n", linkName, flags));
  
  if ((ipmd_GetKernIfByName(linkName, ipm_ifp, IPM_SSC_BUFSIZE*8, &count, &need)
       == IPM_GET_ERR || count == 0)) 
	  return NFX_ERR;

  /* does this link used by other virtual interface */
  tmpVifp = ipmd_FindVifByLink(linkName);
  if (tmpVifp) {
	  if (tmpVifp == vifp)		/* link belongs to virtual interface already */
		  return NFX_OK;
	  else						/* link belongs to another virtual interface */
		  return NFX_ERR;
  }
  
  /* does this link has IP assigned to it */
  need = 0, count = 0;
  ipmd_GetKernIpByName(linkName, ipm_ipp, IPM_SSC_BUFSIZE*8, &count, &need);
  if (count) {
	  /* if flag is IPMD_VIFCONF_RESTORE
	   * function is called during initialization
	   * remove all ip address associated with this link
	   */
	  if ((flags & IPMD_VIFCONF_RESTORE)) {
		  int	i;
		  for (i=0; i < count; i++) {
			  IPMD_VIF_PRINTF(("%s: ip %x removed from link %s\n",
							   __FUNCTION__, ipm_ipp->ip_addr, ipm_ipp->ip_ifname));
			  ipmd_DelKernIp(ipm_ipp->ip_ifname, ipm_ipp->ip_addr);
			  ipm_ipp ++;
		  }
	  }
	  else
		  return NFX_ERR;
  }
  
  
  /* is the link duplicate */
  for (i=0; i < IPM_MAXLINK_VIF; i++) {
	  tmpLink = IPMD_VIF_GETLINK(vifp, i);
	  if (strcmp(tmpLink, linkName) == 0)
      return NFX_OK;
  }
  
  if (vifp->vif_linkCnt >= IPM_MAXLINK_VIF)
	  return IPM_GET_NO_BUF;

  /* find a slot */
  for (i=0; i < IPM_MAXLINK_VIF; i++) {
	  if (*IPMD_VIF_GETLINK(vifp, i) == '\0') {
		  myLink = IPMD_VIF_GETLINK(vifp, i);
		  break;
	  }
  }

  /* something bad happend,vif_links and vif_linkCnt don't agree */
  if (!myLink) {
    IPM_LOGERR(IPM_ERR_BUF);
    return IPM_GET_NO_BUF;
  }
    
  memcpy(myLink, linkName, IPM_STR_SIZE);
  vifp->vif_linkCnt++;


#if 0  
  /* first link by default will be prefered link */
  if ((vifp->vif_flags & IPM_VIF_MODE_FO) && (vifp->vif_linkCnt == 1)) {
	  strcpy(vifp->vif_prefLink, myLink);
	  strcpy(vifp->vif_actLink, myLink);
  }
#endif /* 0 */

  return NFX_OK;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_RmLinkFromVif
 *
 * Description: remove specified link from vif
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_RmLinkFromVif(char	*linkName, ipm_vif_t *vifp)
{
	int		i;
	char	*myLink = 0;
	
	for (i=0; i<IPM_MAXLINK_VIF; i++) {
		myLink = IPMD_VIF_GETLINK(vifp, i);
		if (strcmp(myLink, linkName) == 0) {
			bzero(myLink, IPM_STR_SIZE);
			vifp->vif_linkCnt--;
			return NFX_OK;
		}
	}
	
	return NFX_ERR;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_AddVif
 *
 * Description: create a VIF
 * flags set:	IPM_MSG_CMD_NOVIF	-- should not happen
 *				IPM_MSG_CMD_DUPVIF	-- vif exist
 *				IPM_MSG_CMD_OK		-- OK
 *				IPM_MSG_CMD_TIMEOUT -- txrx timeout
 *				IPM_MSG_CMD_MANYVIF -- too many vifs on system
 *
 *				
 *	
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_AddVif(void	*msg, char	*app, uint32 flags) 
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1);
	char		*vifName = ipm_vifp->vif_name;
	ipm_vif_t	*myVifp = NULL;
	int			count = ipm_vifp->vif_linkCnt;
	int			i;
	int32		rc = NFX_OK;
	int			txrx_reply = 1;
	int			link = -1, slot = -1;

	/* vif name is legal and no dup */
	if ((ipmd_parseVifName(vifName, &slot, &link) != NFX_OK) ||
		(slot != ipm_vifp->vif_slot) || 
		(link > IPM_LINK_MAX) || (link < 0)) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		IPM_LOGERR(IPM_ERR_VIF);
		rc = NFX_ERR;
		goto reply;
	}

	if (ipmd_FindVifByName(vifName)) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_DUPVIF;
		rc = NFX_ERR;
		goto reply;
	}

	if (ipm_vifp->vif_slot != ipmd_txrx_slot) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		rc = NFX_ERR;
		goto reply;
	}

	/* find empty slot to add new virtual interface */
	for (i = 0; i < IPM_MAX_VIF; i++) {
		if (IPMD_VIF_EMPTY(&ipmd_ssc_vif[i])) {
		  myVifp = & ipmd_ssc_vif[i];
		  break;
		}
	}

	/* is there space for new vif */
	if (!myVifp) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_MANYVIF;
		rc = NFX_ERR;
		goto reply;
	}

	/* add link to virtual interface structure,
	 * any failure to add link will cause entire
	 * operation to fail.
	 */
	/* if vif is added from configure restoration,
	 * it is possible that link already has ip
	 * assigned
	 */
	for (i = 0; i< count; i++) {
		rc = ipmd_AddLinkToVif(IPMD_VIF_GETLINK(ipm_vifp, i), myVifp, flags);
		if (rc == NFX_ERR) {
			ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOINTF;
			goto reply;
		}
		else if (rc == IPM_GET_NO_BUF) {
			IPM_LOGERR(IPM_ERR_VIF);
			goto reply;
		}
	}
	
	/* copy link and flags, vif created on ssc */
	memcpy(myVifp->vif_name, vifName, IPM_STR_SIZE);
	myVifp->vif_flags = ipm_vifp->vif_flags;
	myVifp->vif_slot = ipm_vifp->vif_slot;

	/* set active link */
	if ((myVifp->vif_flags & IPM_VIF_MODE_LA))
		memcpy(myVifp->vif_actLink, myVifp->vif_links[0], IPM_STR_SIZE);
	
#if defined(NO_TXRX)
	ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 

reply:
	if ((flags & IPMD_VIFCONF_RESTORE))
		return;

	if (rc != NFX_OK) {
		/* an error has occured, remove virtual interface */
		if (myVifp)
			bzero(myVifp, sizeof(*myVifp));
	}
	else {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_OK;
		/* save configuration */
		ipmd_SaveVif(myVifp);
	}

	ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipmd_ssc_slot, ipmd_ssc_cpu, app, 
				 NULL, 0, 0, -1);

	return;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_DelVif
 *
 * Description: delete a VIF
 *
 * flags set:	IPM_MSG_CMD_NOVIF	-- vif does exist
 *				IPM_MSG_CMD_OK		-- OK
 *				IPM_MSG_CMD_HASIP_VIF -- cannot del vif because 
 *										 ip assigned to it.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_DelVif(void	*msg, char	*app, uint32 flags)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1), *myVifp;
	char		*vifName = ipm_vifp->vif_name;
	int			txrx_reply = 0;
	int32		rc = NFX_OK;

	/* virtual interface exist */
	if ((myVifp = ipmd_FindVifByName(vifName)) == NULL) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		rc = NFX_ERR;
		goto to_txrx;
	}

	/* make sure there is no IP associated with it */
	if (myVifp->vif_ipCnt) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_HASIP_VIF;
		rc = NFX_ERR;
		goto to_txrx;
	}
		
		
	/* clear virtual interface */
	if (myVifp) {
		/* save configuration,
		 *  must be done here because vif record is erased
		 */
		ipmd_UnsaveVif(myVifp);
		bzero(myVifp, sizeof(*myVifp));
	}

to_txrx:
#if defined(NO_TXRX)
	//ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	//rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 	
	
	/* reply back to sender */
	if (rc  == NFX_OK) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_OK;
	}
	
	ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipmd_ssc_slot, ipmd_ssc_cpu, app, 
				 NULL, 0, 0, -1);

	return;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_PrefLinkVif
 *
 * Description: define prefered link in a virtual interface.
 *
 *  flags set:	IPM_MSG_CMD_NOINTF	-- link does not exist
 *				IPM_MSG_CMD_OK		-- OK
 *				IPM_MSG_CMD_TIMEOUT -- txrx timeout
 *				IPM_MSG_CMD_BADMODE_VIF -- incorrect mode
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_PrefLinkVif(void	*msg, char *app, uint32 flags)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1);
	char		*vifName = ipm_vifp->vif_name;
	char		*prefLinkName = ipm_vifp->vif_prefLink;
	ipm_vif_t	*myVifp;
	int32		rc = NFX_OK;
	int			txrx_reply = 1;
	
	/* make sure link and virtual interface matches */
	myVifp = ipmd_FindVifByLink(prefLinkName);
	if (!myVifp || strcmp(myVifp->vif_name, vifName)) {
		rc = NFX_ERR;
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOINTF;
		goto reply;
	}

	if (!(myVifp->vif_flags & IPM_VIF_MODE_FO)) {
		rc = NFX_ERR;
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_BADMODE_VIF;
		goto reply;
	}
		
	/* send to txrx */
#if defined(NO_TXRX)
	ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 

	if (rc != NFX_OK)
		goto reply;

	/* taken care of vif initialization */
	if (myVifp->vif_prefLink[0] == '\0') {
		memcpy(myVifp->vif_prefLink, prefLinkName, IPM_STR_SIZE);
		memcpy(myVifp->vif_actLink, prefLinkName, IPM_STR_SIZE);
		goto reply;
	}
	
	/* set new prefered link */
	if (strcmp(myVifp->vif_prefLink, prefLinkName)) {
		/* reset actLink and prefLink to same link, 
		 * if txrx found pref is not active, it will
		 * send an event to update actLink
		 */
		memcpy(myVifp->vif_prefLink, prefLinkName, IPM_STR_SIZE);		
	}
	
	/* switch IP address from old prefLink to new prefLink */
	if (strcmp(myVifp->vif_actLink, prefLinkName)) {
		ipmd_SwitchLinkIP(myVifp->vif_actLink,  prefLinkName);
		memcpy(myVifp->vif_actLink, prefLinkName, IPM_STR_SIZE);
		
	}

reply:	
	if ((flags & IPMD_VIFCONF_RESTORE))
		return;

	if (rc == NFX_OK) {
		ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
		/* save configuration */
		ipmd_SaveVif(myVifp);
	}
	
	ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipmd_ssc_slot, ipmd_ssc_cpu, app, 
				 NULL, 0, 0, -1);	

	IPMD_VIF_PRINTF(("%s: set preference finished\n", __FUNCTION__));
	
	return;
	
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_AddVifLink
 *
 * Description: add link to vif.
 *
 *  flags set:	IPM_MSG_CMD_NOINTF	-- link does not exist
 *				IPM_MSG_CMD_NOVIF	-- vif does not exist
 *				IPM_MSG_CMD_OK		-- OK
 *				IPM_MSG_CMD_TIMEOUT -- txrx timeout
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_AddVifLink(void	*msg, char *app, uint32 flags)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1);
	int			count = ipm_vifp->vif_linkCnt;
	int32		rc = NFX_OK;
	ipm_vif_t	*myVifp;
	int			txrx_reply = 1, i;
	
	/* find virtual interface */
	myVifp = ipmd_FindVifByName(ipm_vifp->vif_name);
	if (!myVifp) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		rc = NFX_ERR;
		goto reply;
	}
	
	/* chck max num does exit */
	/* TODO chck for max link num */

	/* add link to virtual interface */
	for (i=0; i < count; i++) {
		rc = ipmd_AddLinkToVif(IPMD_VIF_GETLINK(ipm_vifp, i), myVifp, 0);
		if (rc != NFX_OK) 
			break;
	}

	/* if rc != NFX_OK, need to clear link already added */
	if (rc != NFX_OK) {
		for (; (i-1) >=0; i--)
			ipmd_RmLinkFromVif(IPMD_VIF_GETLINK(ipm_vifp, i), myVifp);
		
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOINTF;
		goto reply;
	}

	/* send to txrx */
#if defined(NO_TXRX)
	ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 
	
	/* txrx don't like it, remove all added interface */
	if (rc != NFX_OK) {
		for (i=0; i < count; i++)
			ipmd_RmLinkFromVif(IPMD_VIF_GETLINK(ipm_vifp, i), myVifp);
	}
	
reply:
	if (rc == NFX_OK) {
		ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
		/* save configuration */
		ipmd_SaveVif(myVifp);
	}
	
	ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipmd_ssc_slot, ipmd_ssc_cpu, app, 
				 NULL, 0, 0, -1);

	return;
}



/*-----------------------------------------------------------------
 *
 * Name: ipmd_RmVifLink
 *
 * Description: remove  links from vif.
 *
 *  flags set:	IPM_MSG_CMD_NOINTF	-- link does not exist
 *				IPM_MSG_CMD_NOVIF	-- vif does not exist
 *				IPM_MSG_CMD_OK		-- OK
 *				IPM_MSG_CMD_ACTLINK_VIF
 *									-- can not remove active
 *									-- link.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_RmVifLink(void	*msg, char *app, uint32 flags)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1);
	int			count = ipm_vifp->vif_linkCnt;
	int			rc = NFX_OK, i;
	ipm_vif_t	*myVifp;
	int			txrx_reply = 0;
	char		*tmpLink = NULL;
	
	
	/* find virtual interface */
	myVifp = ipmd_FindVifByName(ipm_vifp->vif_name);

	if (!myVifp) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		rc = NFX_ERR;
		goto to_txrx;
	}

	/* remove links from virtual interface */
	for (i=0; i < count; i++) {
		tmpLink = IPMD_VIF_GETLINK(ipm_vifp, i);
		
		if (strcmp(myVifp->vif_actLink, tmpLink) == 0) {
			/* link cannot be activelink */
			rc = NFX_ERR;
			ipm_hdrp->ipmh_flags = IPM_MSG_CMD_ACTLINK_VIF;
			break;
		}
		else if (strcmp(myVifp->vif_prefLink, tmpLink) == 0) {
			/* prefLink removed, assign new prefLink = actLink */
			memcpy(myVifp->vif_prefLink, myVifp->vif_actLink,\
				   IPM_STR_SIZE);
		}
		
		rc = ipmd_RmLinkFromVif(IPMD_VIF_GETLINK(ipm_vifp, i), myVifp);
		if (rc != NFX_OK) {
			ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOINTF;
			break;
		}
	}

	/* if rc != NFX_OK, need to add link already removed */
	if (rc != NFX_OK) {
		for (; (i-1) >=0; i--)
			ipmd_AddLinkToVif(IPMD_VIF_GETLINK(ipm_vifp, i), myVifp, 0);
	}

to_txrx:		
	/* send to txrx */
#if defined(NO_TXRX)
	//ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	//rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 
	
	/* reply */
	if (rc == NFX_OK) {
		/* save configuration */
		ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
		ipmd_SaveVif(myVifp);
	}
	
	ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipmd_ssc_slot, ipmd_ssc_cpu, app, 
				 NULL, 0, 0, -1);

	return;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_AddVifIP
 *
 * Description: assign IP to virtual link
 *
 *  flags set:	IPM_MSG_CMD_NOVIF	-- vif does not exist
 *				IPM_MSG_CMD_NOLINK_VIF	-- vif contains no link
 *				IPM_MSG_CMD_OK		-- OK
 *				IPM_MSG_CMD_DUPIP	-- ip already used	
 *				IPM_MSG_CMD_TIMEOUT -- txrx doesnot respond
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_AddVifIP(void	*msg, char *app, uint32 flags)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1);
	ipm_vif_t	*myVifp;
	ipm_ip_t	ipm_ip;
	ipm_if_t	*ipm_ifp = (ipm_if_t*)resp_buf;
	int32		rc = NFX_OK;
	char		*myLink = NULL;
	int			count=0, need=0;
	int			txrx_reply = 1;

	/* virtual interface exit and contains some link */
	myVifp = ipmd_FindVifByName(ipm_vifp->vif_name);
	if (!myVifp) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		rc = NFX_ERR;
		goto reply;
	}
	if (!myVifp->vif_linkCnt) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOLINK_VIF;
		rc = NFX_ERR;
		goto reply;
	}
	
	/* does IP address already been used */
	if (ipmd_GetKernIpByIp(ipm_vifp->vif_ip, 1, 
						   &ipm_ip, sizeof(ipm_ip), 
						   &count, &need) == IPM_GET_OK) {

		/* if IP already assigned to another vif, and link
		 * return error.
		 */
		if (ipmd_VifHasIP(ipm_vifp->vif_ip, myVifp) == NULL) {
			ipm_hdrp->ipmh_flags = IPM_MSG_CMD_DUPIP;
			rc = NFX_ERR;
		}
		else {
			ipm_hdrp->ipmh_flags = IPM_MSG_CMD_OK;
			rc = NFX_OK;
		}

		/* reply immediately NFX_ERR or NFX_OK
		 * IP already exist
		 */
		goto reply;
	}
	
	/* add IP address to bsd kern,
	 * the link maynot be active, that is OK
	 * we will reply on event handler to choose an
	 * active link.
	 */
	myLink = myVifp->vif_actLink;
	
	/* make sure the link is a valid link */
	if ((ipmd_GetKernIfByName(myLink, ipm_ifp, IPM_SSC_BUFSIZE*8, &count, &need)
		 == IPM_GET_ERR || count == 0)) {
		/* this is err condition */
		IPM_LOGERR(IPM_ERR_VIF);
		rc = NFX_ERR;
		goto reply;
	}

	/* send to txrx */
#if defined(NO_TXRX)
	ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 
	if (rc != NFX_OK) 
		goto reply;
	
	/* add  now */
	ipmd_AddKernIp(myLink, ipm_vifp->vif_ip, ipm_vifp->vif_mask);
	myVifp->vif_ipCnt++;
	
reply:
	if (rc == NFX_OK) {
		ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
		/* send configuration to ccm */
		ipmd_VifSendCcm(ipm_hdrp);
	}

	ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipmd_ssc_slot, ipmd_ssc_cpu, app, 
				 NULL, 0, 0, -1);
	
	return;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_DelVifIP
 *
 * Description: delete IP to virtual link
 *
 *  flags set:	IPM_MSG_CMD_NOIP	-- ip does not exist
 *				IPM_MSG_CMD_NOVIF	-- vif does not exist
 *				IPM_MSG_CMD_OK		-- OK
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_DelVifIP(void	*msg, char *app, uint32 flags)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_hdrp+1);
	ipm_vif_t	*myVifp;

	int32		rc = NFX_OK;
	char		*myLink = NULL;
	int			txrx_reply = 0;

	/* virtual interface exsit */
	myVifp = ipmd_FindVifByName(ipm_vifp->vif_name);
	if (!myVifp) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		rc = NFX_ERR;
		goto to_txrx;
	}

	/* ip belongs to this virtual interface */
	if ((myLink =ipmd_VifHasIP(ipm_vifp->vif_ip, myVifp)) == NULL) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOIP;
		rc = NFX_ERR;
		goto to_txrx;
	}

	/* fail over, all ip should be active link;
	 * if not, something is wrong. 
	 */
	if ((myVifp->vif_flags & IPM_VIF_MODE_FO) &&
		(strcmp(myLink, myVifp->vif_actLink))) {
		IPM_LOGERR(IPM_ERR_VIF);
		goto to_txrx;
	}
	
	/* remove ip from bsd kern */
	ipmd_DelKernIp(myLink, ipm_vifp->vif_ip);
	myVifp->vif_ipCnt--;
	
to_txrx:
	/* send to txrx */
#if defined(NO_TXRX)
	//ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	//rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 

	if (rc == NFX_OK) {
		ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
		/* send configuration to ccm */
		ipmd_VifSendCcm(ipm_hdrp);

		/* replace IP route, if there is one available */
		ipmd_ReplaceIp(ipm_vifp->vif_ip);
	}
	
	ipmd_SendMsg(ipm_hdrp, ipm_hdrp->ipmh_size, ipmd_ssc_slot, ipmd_ssc_cpu, app, 
				 NULL, 0, 0, -1);

	return;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetVif
 *
 * Description: get virtual interface info
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_GetVif(void	*msg, char *app, uint32 flags)
{
	ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
	ipm_mstat_req_t *ipm_stat_reqp = (ipm_mstat_req_t*)(ipm_hdrp + 1);
	ipm_mvif_t	*ipm_vifp = (ipm_mvif_t*)(ipm_stat_reqp+1);
	char		*vifName = ipm_stat_reqp->req.ifname;
	ipm_vif_t	*myVifp;
	int			msgSize = sizeof(ipm_mhdr_t) + sizeof(ipm_mstat_req_t);
	ipm_mvif_t	*respVifp;
	ipm_if_t	*respIfp;
	int			i;
	char		*tmpLink;
	int32		rc = IPM_GET_OK;
	int32		count = 0;


	/* return summary of all vifs 
	 * if no vif name specified
	 */
	if (vifName[0] == '\0') {
		ipm_vifp = (ipm_vif_t*)(((char*)resp_buf) + sizeof(ipm_mhdr_t) + 
			sizeof(ipm_mstat_req_t));
		for (i = 0; i < IPM_MAX_VIF; i++) {
			myVifp = &ipmd_ssc_vif[i];
			if (myVifp->vif_name[0] == '\0')
				continue;
			
			memcpy(ipm_vifp, myVifp, sizeof(ipm_vif_t));
			msgSize += sizeof(ipm_vif_t);
			count ++;
			ipm_vifp ++;
		}

		rc = NFX_OK;
		goto reply;
	}
	
	myVifp = ipmd_FindVifByName(vifName);
	if (!myVifp) {
		ipm_hdrp->ipmh_flags = IPM_MSG_CMD_NOVIF;
		rc = IPM_GET_ERR;
		goto reply;
	}

#if 0							/* not working, and not needed */
	/* for test only, get vif info from txrx  */
#if defined(NO_TXRX)
	ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	rc = NFX_OK;
#else
	IPMD_VIF_TOTXRX();
#endif 
	if (rc != IPM_GET_OK) {
		rc = IPM_CMDERR_TIMEOUT;
		goto reply;
	}
	ipmd_CmpVif(ipm_vifp, myVifp);
#endif /* 0 */

	/* calculate the response msg size */
	memcpy((char*)resp_buf, (char*)msg, msgSize);
	
	/* cpy vif to response buffer */
	respVifp = (ipm_mvif_t*)(((char*)resp_buf) + msgSize);
	memcpy((char*)respVifp, myVifp, sizeof(ipm_mvif_t));
	msgSize += sizeof(ipm_mvif_t);
	respIfp = (ipm_if_t*)(respVifp + 1);
	
	/* getting each link info from txrx */
	for (i=0; i < IPM_MAXLINK_VIF; i++) {
		tmpLink = IPMD_VIF_GETLINK(myVifp, i);
		
		/* bypass empty slot */
		if (*tmpLink == '\0')
			continue;
		
		rc = ipmd_GetIfFromTxrx(tmpLink, respIfp);
		if (rc != IPM_GET_OK) {
			msgSize =  sizeof(ipm_mhdr_t) + sizeof(ipm_mstat_req_t);
			goto reply;
		}
		
		respIfp++;
		msgSize += sizeof(ipm_if_t);
	}
	
	/* successfully get single vif */
	count = 1;

reply:
	/* update msgsize */
	
	/* copy the hdr from  request buffer */
	memcpy((char*)resp_buf, ipm_hdrp, sizeof(ipm_mhdr_t));
	
	ipm_hdrp = (ipm_mhdr_t*)resp_buf;
	ipm_hdrp->ipmh_size = msgSize;
 
	/* update flags */
	if (rc == IPM_GET_OK) {
		ipm_hdrp->ipmh_flags |= IPM_MSG_CMD_OK;
	}
	
	/* update num. of record copied */
	ipm_stat_reqp = (ipm_mstat_req_t*)(ipm_hdrp + 1);
	ipm_stat_reqp->count = count;	
	
	ipmd_SendMsg((void*)ipm_hdrp, msgSize, ipmd_ssc_slot, 
				 ipmd_ssc_cpu, app, NULL, 0, 0, -1);

	return;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_ProcessVif
 *
 * Description: process virtual interface requests from applications
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
void ipmd_ProcessVif(int32 mtype, void* msg, char *app)
{
	IPMD_VIF_PRINTF(("%s: SSC vif  received\n", __FUNCTION__));
	//ipmd_PrintVif((ipm_vif_t*)(((ipm_mhdr_t*)msg) + 1), mtype);

	ipm_stats.adm_rcv++;

	switch (mtype) {
	case IPM_MSG_VIP_ADD:
		ipmd_AddVifIP(msg, app, 0);
		break;
			
	case IPM_MSG_VIP_DEL:
		ipmd_DelVifIP(msg, app, 0);
		break;
		
	case IPM_MSG_VIF_ADDLNK:
		ipmd_AddVifLink(msg, app, 0);
		break;

	case IPM_MSG_VIF_RMLNK:
		ipmd_RmVifLink(msg, app, 0);
		break;
		
	case IPM_MSG_VIF_CREAT:
		ipmd_AddVif(msg, app, 0);
		break;
		
	case IPM_MSG_VIF_DEL:
		ipmd_DelVif(msg, app, 0);
		break;
		
	case IPM_MSG_VIF_PREF:
		ipmd_PrefLinkVif(msg, app, 0);
		break;
		
	case IPM_MSG_VIFTBL_REQ:
		ipmd_GetVif(msg, app, 0);
		break;
		
	default:
	}
	
	return;
}

/*----------------------------------------------------------------
 *
 * Name: ipmd_ProcessVifTxrx
 *
 * Description: process event from txrx 
 *
 * Created by: Su Wang
 *
 * Date Created: 10/14/02
 *
 *----------------------------------------------------------------
 */
void ipmd_ProcessVifTxrx(int32 mtype, void	*msg)
{

	IPMD_VIF_PRINTF(("%s: TXRX vif  received\n", __FUNCTION__));
	//ipmd_PrintVif((ipm_vif_t*)(((ipm_mhdr_t*)msg) + 1), mtype);
	
	switch(mtype) {
	case IPM_MSG_VIF_EVNT:
		ipmd_UpdateActLink(msg);
	}
	
	return;
}

 
/*-----------------------------------------------------------------
 *
 * Name: ipmd_InitVif
 *
 * Description: 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
void ipmd_InitVif()
{

	bzero(ipmd_ssc_vif, sizeof(ipmd_ssc_vif));
	
	/* restore any saved configuration */
	ipmd_RestoreVif();
	return;
}

#endif /* LINK_AGG */
