/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_ct.c
 *
 * RCS:		$Id: ispfc_ct.c,v 1.12 2002/10/22 01:30:38 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver CT (common transport) code
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/5/01
 *
 *------------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_iocb.h"
#include "ispfc_mbox.h"
#include "ispfc_link.h"
#include "ispfc_ct.h"

#if defined(ISPFC_SIM)
extern int ispfc_fcadmin_disk_count;
extern int ispfc_fcadmin_tape_count;
#endif

void ispfc_dev_ms_login_done_callout(ispfc_cb_t *, void *, int32);
void ispfc_ct_common_done(ct_desc_t *);

/* 
****************************************************************************
* External interface
****************************************************************************
*/
int32 
ispfc_ct_send(ct_desc_t *ctd)
{
	ispfc_cb_t *	cb = hal_dt_entry_get()->hal_handle;
	VERIFY(cb);

	/* 
	 * Cleanse return status 
	 */
	ctd->status = 0;

	if (cb->interface_type == HAL_IF_FC_L_PORT ||
	    cb->interface_type == HAL_IF_FC_N_PORT ||
	    LINK_STATE(cb) == ISPFC_CLS_STATE_OFFLINE) {
		ctd->status = CT_STATUS_ILLEGAL_OP;
		(*ctd->callback)(ctd);
		return(NFX_ERR);
	}
	if (ctd->server_addr != CT_ADDRESS_NS &&
	    ctd->server_addr != CT_ADDRESS_MS) {
		ctd->status = CT_STATUS_INVALID_PARM;
		(*ctd->callback)(ctd);
		return(NFX_ERR);
	}

	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;

	/* 
	 * Stick descriptor on waiting queue; it will be
	 * picked up and dispatched by the CT handler.  
	 */
	ctd->internal = 0;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32
ispfc_ct_dispatch_desc(ispfc_cb_t *cb, ct_desc_t *ctd)
{
	eee_desc_t *	ed;
	eee_descBuf_t *	bd;
	int		len = 0;
	int		buf_count = 0;

	if ((ed = ctd->e_desc) == NULL) {
		ctd->status = CT_STATUS_INVALID_PARM;
		(*ctd->callback)(ctd);
		return(NFX_OK);
	}

	bd = &ed->bd[1];

	/*
	 * Verify the CT request buffer
	 */
	if (bd->control & EDESC_BD_CMD_IN_USE && bd->buf) {
		len += bd->len;
		++buf_count;
	}
	else {
		ctd->status = CT_STATUS_INVALID_PARM;
		(*ctd->callback)(ctd);
		return(NFX_OK);
	}
	
	++bd;

	/*
	 * If the CT response buffer is not valid, allocate one
	 */
	if (!(bd->control & EDESC_BD_CMD_IN_USE) || bd->buf == NULL) {	
		bd->buf = eee_allocateBuffer(EEE_SHARED_POOL, EDESC_BD_ULTRA, 0);
		bd->len = EEE_BUFSIZE_ULTRA;
		bd->control = EDESC_BD_CMD_IN_USE | EDESC_BD_ULTRA;
		len += bd->len;
		++buf_count;
		ed->hdr.control = ((buf_count << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK) | 
				  (len & EDESC_TOTAL_LEN);
	}

#if defined(ISPFC_SIM)
	{

#define SIM_SWITCH_COUNT	4
#define SIM_SWITCH_PORT_COUNT	16
#define SIM_SWITCH_WWN_BASE	0x0000112233000000LL

		struct {
			wwn_t	wwn;
			char	sym_name[32];
			int	len;
		} sim_switches[SIM_SWITCH_COUNT] = {
			{ 0x2000000000000000LL | SIM_SWITCH_WWN_BASE | 1, "blue",   4 },
			{ 0x2000000000000000LL | SIM_SWITCH_WWN_BASE | 2, "red",    3 },
			{ 0x2000000000000000LL | SIM_SWITCH_WWN_BASE | 3, "green",  5 },
			{ 0x2000000000000000LL | SIM_SWITCH_WWN_BASE | 4, "yellow", 6 }
		};

	
		ispfc_ct_req_t *	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
		ispfc_ct_resp_t *	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
		ushort16		cmd = req->header.command_code;
		int			i;
		wwn_t			switch_name;
		wwn_t			port_name;
		uint32			domain, area, port;
		ispfc_ct_resp_GID_FT_t *p;
		uint32			port_id;
		
		ctd->status = CT_STATUS_OK;
		resp->header.response_code = SNS_CT_FS_ACC;

		inplace_ntohs(&cmd);
		switch(cmd) {
		case CT_NS_CMD_GID_FT:	
			if (ispfc_fcadmin_disk_count == 0) {
				resp->header.response_code = SNS_CT_FS_RJT;
				resp->header.reason_code = SNS_CT_RSN_UNABLE_TO_PERFORM;
			}
			else {

				if (ispfc_fcadmin_disk_count > ISPFC_SIM_MAX_DISKS)
					ispfc_fcadmin_disk_count = ISPFC_SIM_MAX_DISKS;

				p = &resp->u.GID_FT.list[0];

				for (i = 0; i < ispfc_fcadmin_disk_count + ispfc_fcadmin_tape_count; ++i, ++p) {

					if (i < ispfc_fcadmin_disk_count) {
						/* 
						 * Spread disks over first two
						 * switches, over the first 8
						 * ports. The remaining 8 ports
						 * are reserved for simulated
						 * E-ports.
						 */
						domain = i / (ISPFC_SIM_MAX_DISKS / 2);
						port = i % (ISPFC_SIM_MAX_DISKS / 2);
						area = port % (SIM_SWITCH_PORT_COUNT/2);
						port = ispfc_loopid_to_alpa(port / (SIM_SWITCH_PORT_COUNT/2));
						p->portid = (domain << 16) | (area << 8) | port;
					}
					else {
						/* 
						 * Spread disks over second two
						 * switches, over the first 8
						 * ports. The remaining 8 ports
						 * are reserved for simulated
						 * E-ports.
						 */
						domain = i / (ISPFC_SIM_MAX_DISKS / 2) + 2;
						port = i % (ISPFC_SIM_MAX_DISKS / 2);
                                                area = port % (SIM_SWITCH_PORT_COUNT/2);
                                                port = ispfc_loopid_to_alpa(port / (SIM_SWITCH_PORT_COUNT/2));
						p->portid = (domain << 16) | (area << 8) | port;
					}
#ifdef LITTLE_ENDIAN 
					SWAP3((byte *)p+1);
#endif
					if (i == ispfc_fcadmin_disk_count + ispfc_fcadmin_tape_count - 1)
						p->control_byte = 0x80;
				}
			}
			break;

		case CT_NS_CMD_GFPN_ID:
			port_id = 0;
			memcpy(&port_id, &req->u.GFPN_ID, sizeof(uint32));
			inplace_ntohl(&port_id);

			domain = (port_id & 0xFF0000) >> 16;
			area = (port_id & 0x00FF00) >> 8;
			WWN_SET(&switch_name, &sim_switches[domain].wwn);
			port_name.u.wwn = switch_name.u.wwn | 0x1000000000000000LL | ((uint64)area << 48);
			WWN_SET(resp->u.GFPN_ID.node_name, &port_name);
			break;

		case CT_NS_CMD_GPN_ID:	
		case CT_NS_CMD_GNN_ID:
		case CT_NS_CMD_RFT_ID:
		case CT_NS_CMD_RSPN_ID:
			break;

		case CT_MS_CMD_GIEL:
			resp->u.GIEL.ie_count = 4;
			inplace_htonl(&resp->u.GIEL.ie_count);
			WWN_SET(resp->u.GIEL.ie[0].ie_name, &sim_switches[0].wwn);
			WWN_SET(resp->u.GIEL.ie[1].ie_name, &sim_switches[1].wwn);
			WWN_SET(resp->u.GIEL.ie[2].ie_name, &sim_switches[2].wwn);
			WWN_SET(resp->u.GIEL.ie[3].ie_name, &sim_switches[3].wwn);
			break;

		case CT_MS_CMD_GIELN:
			for (i = 0; i < SIM_SWITCH_COUNT; ++i) {
				if (WWN_EQUAL(req->u.GIELN.ie_name, &sim_switches[i].wwn)) {
					resp->u.GIELN.ie_logical_name.len = sim_switches[i].len;
					memcpy(resp->u.GIELN.ie_logical_name.name, sim_switches[i].sym_name, sim_switches[i].len);
					break;
				}
			}
			VERIFY(i < SIM_SWITCH_COUNT);
			break;

		case CT_MS_CMD_GFN:
			WWN_SET(resp->u.GFN.fabric_name, &sim_switches[0].wwn);
			break;

		case CT_MS_CMD_GPL:
			WWN_SET(&switch_name, req->u.GPL.ie_name);
			resp->u.GPL.port_count = SIM_SWITCH_PORT_COUNT;
			inplace_htonl(&resp->u.GPL.port_count);
			for (i = 0; i < SIM_SWITCH_PORT_COUNT; ++i) {
				port_name.u.wwn = switch_name.u.wwn | 0x1000000000000000LL | ((uint64)i << 48);
				WWN_SET(resp->u.GPL.port[i].port_name, &port_name);
				if (i < 8)
					resp->u.GPL.port[i].port_type = (i % 2) ? PORT_TYPE_F : PORT_TYPE_FL;
				else
					resp->u.GPL.port[i].port_type = PORT_TYPE_E;
			}
			break;

		case CT_MS_CMD_GPPN:
			WWN_SET(&port_name, req->u.GPPN.port_name);
			resp->u.GPPN.port_number = (port_name.u.wwn >> 48) & 0xFF;
			inplace_htonl(&resp->u.GPPN.port_number);
			break;

		case CT_MS_CMD_GAPNL:
			break;
		default: 
			VERIFY(0);
		}
		inplace_ntohs(&resp->header.response_code);
	}
#endif /* defined(ISPFC_SIM) */

#if !defined(ISPFC_SIM)
	if (ispfc_send_ct_desc(cb, ctd) != NFX_OK)
		return(NFX_ERR);
#endif

	ispfc_queue_desc_on_ct_busy_queue(cb, ctd);

	return(NFX_OK);
}

void
ispfc_dispose_of_ct_desc(ispfc_cb_t *cb, ct_desc_t *ctd, int retry)
{
	if (!retry || ctd->retry_count == 0) {
		if (ctd->priv_callback)
			(*ctd->priv_callback)(ctd);
		else
			(*ctd->callback)(ctd);
	}
	else {
		/* 
		 * Cleanse return status 
		 */
		ctd->status = 0;

		/* Stick descriptor back on waiting queue */
		ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);
	}
}

int32 
ispfc_ct_desc_start_handler(ispfc_cb_t *cb, uint32 timer_ref)
{
	ct_desc_t *	ctd;
	int32		rc;

	/* If link down, don't bother */
	if (LINK_STATE(cb) != ISPFC_CLS_STATE_UP && 
	    LINK_STATE(cb) != ISPFC_CLS_STATE_REESTABLISHING)
		return(NFX_OK);

	/* If link not attached to fabric, don't bother */
	if (cb->interface_type == HAL_IF_FC_L_PORT || cb->interface_type == HAL_IF_FC_N_PORT)
		return(NFX_OK);

	/* NS should never be unmapped */
	VERIFY(NS_STATE(cb) == ISPFC_CLS_NS_STATE_MAPPED);

	switch (MS_STATE(cb)) {
	case ISPFC_CLS_MS_STATE_NOT_MAPPED:
		/* If mailbox busy, bail */
		if (ispfc_mbox_is_busy(cb))
			break;

		/* 
		 * Fire off a fabric login mailbox command to map the
		 * MS. Turn on optional bit 1 to disable attempt to do
		 * a process login; Qlogic Sanbox2 switches don't
		 * understand a process login on the management server
		 * port. 
		 */
		cb->mbox_in_valid_count = 4;
		cb->mbox_in[0] = ISPFC_LOGIN_FABRIC_PORT;
		cb->mbox_in[1] = (ISPFC_MANAGEMENT_SERVER_PORTID << 8) | 0x0002;
		cb->mbox_in[2] = (CT_ADDRESS_MS & ISPFC_DEV_PHYS_DEV_ID_MASK_HI) >> 16;
		cb->mbox_in[3] = CT_ADDRESS_MS & ISPFC_DEV_PHYS_DEV_ID_MASK_LO;
		SET_MS_STATE(cb, ISPFC_CLS_MS_STATE_MAP_PENDING);
		ispfc_mbox_start(cb, ispfc_dev_ms_login_done_callout, NULL);
		break;
		
	case ISPFC_CLS_MS_STATE_MAP_PENDING:
		/* Waiting for mailbox command -- nothing to do */
		break;

	case ISPFC_CLS_MS_STATE_MAPPED:
		/* 
		 * If anything in the waiting queues, start it.
		 */
		while ((ctd = ispfc_dequeue_desc_off_ct_waiting_queue(cb))) {
			rc = ispfc_ct_dispatch_desc(cb, ctd);
			/* If dispatch failed, stick it back on the waiting list and bail */
			if (rc != NFX_OK) {
				ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);
				break;
			}
		}
		break;

	default:
		VERIFY(0);
	}
	return(NFX_OK);
}

void 
ispfc_dev_ms_login_done_callout(ispfc_cb_t *cb, void *cookie, int32 status)
{
	VERIFY(MS_STATE(cb) == ISPFC_CLS_MS_STATE_MAP_PENDING);

	if (status != MBOX_CMD_STATUS_SUCCESS) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		SET_MS_STATE(cb, ISPFC_CLS_MS_STATE_MAPPED);
	}
}

/* 
****************************************************************************
* Internal Name (directory) server interfaces
****************************************************************************
*/
int32 
ispfc_ns_GID_FT(ispfc_cb_t *cb, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_NS;
	req->header.fs_subtype = FS_SUB_TYPE_NAME_SERVICE;
	req->header.command_code = CT_NS_CMD_GID_FT;
	req->header.maximum_resp_size = sizeof(resp->u.GID_FT)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
        req->u.GID_FT.fc_4_type_code = FC_4_TYPE_FCP;
	
	ctd->server_addr = CT_ADDRESS_NS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ns_GNN_ID(ispfc_cb_t *cb, uint32 port_id, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_NS;
	req->header.fs_subtype = FS_SUB_TYPE_NAME_SERVICE;
	req->header.command_code = CT_NS_CMD_GNN_ID;
	req->header.maximum_resp_size = sizeof(resp->u.GNN_ID)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);

        req->u.GNN_ID.portid[0] = (port_id & 0xFF0000) >> 16;
        req->u.GNN_ID.portid[1] = (port_id & 0x00FF00) >> 8;
        req->u.GNN_ID.portid[2] = (port_id & 0x0000FF);

	ctd->server_addr = CT_ADDRESS_NS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ns_GPN_ID(ispfc_cb_t *cb, uint32 port_id, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_NS;
	req->header.fs_subtype = FS_SUB_TYPE_NAME_SERVICE;
	req->header.command_code = CT_NS_CMD_GPN_ID;
	req->header.maximum_resp_size = sizeof(resp->u.GPN_ID)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	
        req->u.GPN_ID.portid[0] = (port_id & 0xFF0000) >> 16;
        req->u.GPN_ID.portid[1] = (port_id & 0x00FF00) >> 8;
        req->u.GPN_ID.portid[2] = (port_id & 0x0000FF);

	ctd->server_addr = CT_ADDRESS_NS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ns_GFPN_ID(ispfc_cb_t *cb, uint32 port_id, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_NS;
	req->header.fs_subtype = FS_SUB_TYPE_NAME_SERVICE;
	req->header.command_code = CT_NS_CMD_GFPN_ID;
	req->header.maximum_resp_size = sizeof(resp->u.GFPN_ID)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	
        req->u.GFPN_ID.portid[0] = (port_id & 0xFF0000) >> 16;
        req->u.GFPN_ID.portid[1] = (port_id & 0x00FF00) >> 8;
        req->u.GFPN_ID.portid[2] = (port_id & 0x0000FF);

	ctd->server_addr = CT_ADDRESS_NS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ns_RFT_ID(ispfc_cb_t *cb, uint32 port_id, byte fc4_type, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;
	int			byte_index;
	int			bit_index;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_NS;
	req->header.fs_subtype = FS_SUB_TYPE_NAME_SERVICE;
	req->header.command_code = CT_NS_CMD_RFT_ID;
	req->header.maximum_resp_size = 0;
	inplace_htons(&req->header.command_code);

        req->u.RFT_ID.portid[0] = (port_id & 0xFF0000) >> 16;
        req->u.RFT_ID.portid[1] = (port_id & 0x00FF00) >> 8;
        req->u.RFT_ID.portid[2] = (port_id & 0x0000FF);

#ifdef LITTLE_ENDIAN	
	byte_index = 3 - ((fc4_type/8) % 4) + (fc4_type/32) * 4;
	bit_index = fc4_type%8;
#else
	byte_index = ((fc4_type/8) % 4) + (fc4_type/32) * 4;
	bit_index = fc4_type%8;
#endif
	req->u.RFT_ID.fc_4_types[byte_index] |= 1 << bit_index;

	ctd->server_addr = CT_ADDRESS_NS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ns_RSPN_ID(ispfc_cb_t *cb, uint32 port_id, byte str_len, char *str, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_NS;
	req->header.fs_subtype = FS_SUB_TYPE_NAME_SERVICE;
	req->header.command_code = CT_NS_CMD_RSPN_ID;
	req->header.maximum_resp_size = 0;
	inplace_htons(&req->header.command_code);

        req->u.RSPN_ID.portid[0] = (port_id & 0xFF0000) >> 16;
        req->u.RSPN_ID.portid[1] = (port_id & 0x00FF00) >> 8;
        req->u.RSPN_ID.portid[2] = (port_id & 0x0000FF);

	VERIFY(str_len <= sizeof(req->u.RSPN_ID.string));
	req->u.RSPN_ID.str_length = str_len;
	memcpy(req->u.RSPN_ID.string, str, str_len);

	ctd->server_addr = CT_ADDRESS_NS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

/* 
****************************************************************************
* Internal Management server interfaces
****************************************************************************
*/
int32 
ispfc_ms_GIEL(ispfc_cb_t *cb, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GIEL;
	req->header.maximum_resp_size = sizeof(resp->u.GIEL)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	
	ctd->server_addr = CT_ADDRESS_MS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ms_GFN(ispfc_cb_t *cb, wwn_t *ie_name, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GFN;
	req->header.maximum_resp_size = sizeof(resp->u.GFN)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	WWN_SET(&req->u.GFN.ie_name, ie_name);

	ctd->server_addr = CT_ADDRESS_MS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32
ispfc_ms_GIELN(ispfc_cb_t *cb, wwn_t *ie_name, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GIELN;
	req->header.maximum_resp_size = sizeof(resp->u.GIELN)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	WWN_SET(&req->u.GIELN.ie_name, ie_name);

	ctd->server_addr = CT_ADDRESS_MS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ms_GPL(ispfc_cb_t *cb, wwn_t *ie_name, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GPL;
	req->header.maximum_resp_size = sizeof(resp->u.GPL)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	WWN_SET(&req->u.GPL.ie_name, ie_name);

	ctd->server_addr = CT_ADDRESS_MS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32
ispfc_ms_GPPN(ispfc_cb_t *cb, wwn_t *port_name, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GPPN;
	req->header.maximum_resp_size = sizeof(resp->u.GPPN)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	WWN_SET(&req->u.GPPN.port_name, port_name);

	ctd->server_addr = CT_ADDRESS_MS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

int32 
ispfc_ms_GAPNL(ispfc_cb_t *cb, wwn_t *port_name, ct_desc_t *ctd)
{
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	resp;

	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
        bzero(resp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GAPNL;
	req->header.maximum_resp_size = sizeof(resp->u.GAPNL)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);
	WWN_SET(&req->u.GAPNL.port_name, port_name);

	ctd->server_addr = CT_ADDRESS_MS;
	VERIFY(ctd->callback);

	ctd->priv_callback = ispfc_ct_common_done;
	ctd->state = CT_DESC_STATE_WAITING;
	eee_rtc(&ctd->first_issued_timestamp);
	ctd->timeout_count = 10;
	ctd->retry_count = 3;
	ctd->internal = 1;
	ispfc_queue_desc_on_ct_waiting_queue(cb, ctd);

	return(NFX_OK);
}

/*
****************************************************************************
* Common completion routine for internal interfaces
****************************************************************************
*/
void 
ispfc_ct_common_done(ct_desc_t *ctd)
{
	ispfc_cb_t *		cb = hal_dt_entry_get()->hal_handle;
	ispfc_ct_req_t *	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	ispfc_ct_resp_t *	resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	char *			str = NULL;
	ispfc_ct_resp_GID_FT_t*	p;
	int			i;

	VERIFY(cb);

	inplace_ntohs(&req->header.command_code);

	switch(req->header.command_code) {
	case CT_NS_CMD_GPN_ID:	       str = "Name server query GPN_ID"; break;
	case CT_NS_CMD_GNN_ID:	       str = "Name server query GNN_ID"; break;
	case CT_NS_CMD_GFPN_ID:	       str = "Name server query GFPN_ID"; break;
	case CT_NS_CMD_GID_FT:	       str = "Name server query GID_FT"; break;
	case CT_NS_CMD_RFT_ID:	       str = "Name server registration RFT_ID"; break;
	case CT_NS_CMD_RSPN_ID:	       str = "Name server registration RSPN_ID"; break;
	case CT_MS_CMD_GIEL:	       str = "Management server query GIEL"; break;
	case CT_MS_CMD_GFN:	       str = "Management server query GFN"; break;
	case CT_MS_CMD_GIELN:	       str = "Management server query GIELN"; break;
	case CT_MS_CMD_GPL:	       str = "Management server query GPL"; break;
	case CT_MS_CMD_GPPN:	       str = "Management server query GPPN"; break;
	case CT_MS_CMD_GAPNL:	       str = "Management server query GAPNL"; break;
	default: 
		VERIFY(0);
	}

	if (ctd->status != CT_STATUS_OK) {
		FC_DBG(0, ("%s: %s transport failure: status = 0x%x\n", 
			   cb->debug_name, str, ctd->status));
		(*ctd->callback)(ctd);
		return;
	}

	inplace_ntohs(&resp->header.response_code);
	ctd->status |= (resp->header.response_code << CT_STATUS_RESP_SHIFT) & CT_STATUS_RESP_MASK;
	ctd->status |= CT_STATUS_RESP_VALID;
	
	if (resp->header.response_code != SNS_CT_FS_ACC) {
		FC_DBG(0, ("%s: %s logical failure: response_code 0x%x, reason_code 0x%x\n",
			   cb->debug_name, str, resp->header.response_code, resp->header.reason_code));
		(*ctd->callback)(ctd);
		return;
	}
	else {
		FC_DBG(1, ("%s: %s succeeded: response_code 0x%x, reason_code 0x%x\n",
			   cb->debug_name, str, resp->header.response_code, resp->header.reason_code));
	}

	/* Swap appropriate fields */
	switch(req->header.command_code) {
	case CT_NS_CMD_GPN_ID:
	case CT_NS_CMD_GNN_ID:
	case CT_NS_CMD_GFPN_ID:
		break;

	case CT_NS_CMD_GID_FT:
#ifdef LITTLE_ENDIAN
		for (i = 0, p = resp->u.GID_FT.list; ; ++i, ++p) {
			SWAP3((byte *)p+1);
			if (p->control_byte & 0x80)
				break;
		}
#endif
		break;

	case CT_NS_CMD_RFT_ID:
	case CT_NS_CMD_RSPN_ID:
		break;

	case CT_MS_CMD_GIEL:
		inplace_ntohl(&resp->u.GIEL.ie_count);
		VERIFY(resp->u.GIEL.ie_count <= MAX_IES);
		break;

	case CT_MS_CMD_GFN:
	case CT_MS_CMD_GIELN:
		break;

	case CT_MS_CMD_GPL:
		inplace_ntohl(&resp->u.GPL.port_count);
		VERIFY(resp->u.GPL.port_count <= MAX_PORTS);
		break;

	case CT_MS_CMD_GPPN:
		inplace_ntohl(&resp->u.GPPN.port_number);
		break;

	case CT_MS_CMD_GAPNL:
		inplace_ntohl(&resp->u.GAPNL.port_count);
		VERIFY(resp->u.GAPNL.port_count <= MAX_ATTACHED_PORTS);
		break;

	default: 
		VERIFY(0);
	}
	
	(*ctd->callback)(ctd);
}

