/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_ioctl.c
 *
 * RCS:		$Id: ispfc_ioctl.c,v 1.31 2002/11/14 00:52:20 nitinb Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver IOCTL code
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/5/01
 *
 *------------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_iocb.h"
#include "ispfc_ioctl.h"
#include "ispfc_mbox.h"
#include "ispfc_ct.h"
#include "ispfc_device.h"
#include "ispfc_td.h"
#include "ispfc.h"
#include <exports/env-api.h>

extern void ispfc_send_fc_state_msg(int);

typedef struct {
	void *			cookie;
	void (*callback)(void *, uint32);
} ispfc_ioctl_sm_t;

typedef struct {
	uint32			type; /* 0 = BYPASS, 1 = UNBYPASS */
	ispfc_lfa_cb_t		lfa_req;
	ispfc_lfa_resp_t	lfa_resp;
	wwn_t			channel_wwn;	
	void *			cookie;
	void (*callback)(void *, uint32);
} ispfc_lpc_sm_t;

#define ISPFC_LPC_BYPASS	0
#define ISPFC_LPC_UNBYPASS	1

typedef struct {
	ispfc_lfa_cb_t		lfa_req;
	ispfc_lfa_resp_t	lfa_resp;
	wwn_t			channel_wwn;
	void *			cookie;
	void (*callback)(void *, uint32);
} ispfc_lip_sm_t;

typedef struct {
	ispfc_lfa_cb_t		lfa_req;
	ispfc_lfa_resp_t	lfa_resp;
	wwn_t			channel_wwn;
	hal_device_map_t *	caller_map;
	ispfc_position_map_t 	loop_map;
	void (*callback)(wwn_t, hal_device_map_t *, int, void *);
	void *			cookie;
} ispfc_get_position_map_sm_t;
	
void ispfc_ioctl_done(ispfc_cb_t *, void *cookie, int32);
void ispfc_bypass_device(ispfc_cb_t *, wwn_t, void (*)(void *, uint32), void *);
void ispfc_unbypass_device(ispfc_cb_t *, wwn_t, void (*)(void *, uint32), void *);
void ispfc_unbypass_all(ispfc_cb_t *, wwn_t, void (*)(void *, uint32), void *);
void ispfc_lpc_done(ispfc_cb_t *, void *, int32);
void ispfc_lip(ispfc_cb_t *, wwn_t, uint32, uint32, void (*)(void *, uint32), void *);
void ispfc_lip_done(ispfc_cb_t *, void *, int32);
void ispfc_get_position_map(ispfc_cb_t *, wwn_t, hal_device_map_t *, 
                            void (*)(wwn_t, hal_device_map_t *, int, void *), void *);
void ispfc_get_position_map_done(ispfc_cb_t *, void *, int32);
void scsi_getstats (hal_fc_link_stats_t *stats, uint32 isClear);


#ifdef SCSI_TIMESTAMP
extern scsi_timeTab_t  totReadDataRspBuckets;
extern scsi_timeTab_t  totWriteRspBuckets;

extern scsi_timeTab_t  fcReadDataRspBuckets;
extern scsi_timeTab_t  fcWriteRspBuckets;
#endif


int32
ispfc_ioctl(uint32 code, ...)
{
	hal_dispatch_entry_t *	hal_dt = hal_dt_entry_get();
	ispfc_cb_t *		cb;
	va_list			ap;

	cb = hal_dt->hal_handle;
	VERIFY(cb);

	switch (code) {
	case HAL_IOCTL_COMMON_GET_ADAPTER_INFO:
		{
			hal_adapter_info_t	*p;
			
			va_start(ap, code);
			p = va_arg(ap, hal_adapter_info_t *);
			va_end(ap);

#if defined(ISPFC_SIM)
			sprintf(p->name, "Agile Storage Sim");
			sprintf(p->firmware_rev, "%d.%d.%d", 1, 0, 42);
			sprintf(p->hardware_rev, "%d", 42);
#else
			sprintf(p->name, "QLogic 2300");
			sprintf(p->firmware_rev, "%d.%d.%d",
				cb->firmware_version[0],
				cb->firmware_version[1],
				cb->firmware_version[2]);
			sprintf(p->hardware_rev, "%d",
				cb->chip_rev);
#endif
			p->interface_type = cb->interface_type;
			p->max_tags_per_device = ISPFC_MAX_SCSI_TAGS;
			WWN_SET(&p->u.fc_data.fc_node_name, &cb->adapter_wwn);
			inplace_ntohll(&p->u.fc_data.fc_node_name);
			p->u.fc_data.link_speed = cb->link_speed;
			p->u.fc_data.fc_port_id = cb->host_port_id;
			p->u.fc_data.fc_state_change_generation_num = cb->reestablish_count;
			p->u.fc_data.fc_packet_size = cb->fw_init_cb->max_frame_len;
			p->u.fc_data.sram_parity = cb->sram_parity_available;
		}
		break;

	case HAL_IOCTL_COMMON_RESET_ADAPTER:
		if (!(cb->link_state & ISPFC_CLF_FC_OFFLINE_LATCH)) {
			FC_DBG(0, ("%s: Resetting link\n", cb->debug_name));
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		}
		break;

	case HAL_IOCTL_COMMON_DISABLE_ADAPTER:
		FC_DBG(0, ("%s: Offlining link\n", cb->debug_name));
		E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
		       ("ispfc: Disabling fibrechannel link"));
		ispfc_send_fc_state_msg(ISPFC_CLS_STATE_DOWN);
		if (IS_LINK_UP(cb)) {
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_OFFLINE_REQUESTED);
		}
		else {
			cb->link_state |= ISPFC_CLF_FC_ASYNC_OFFLINE_REQUEST;
		}
		cb->link_state |= ISPFC_CLF_FC_OFFLINE_LATCH;
		ispfc_set_autolinkenable_off();
		break;

	case HAL_IOCTL_COMMON_ENABLE_ADAPTER:
		if (IS_LINK_OFFLINE(cb)) {
			FC_DBG(0, ("%s: Onlining link\n", cb->debug_name));
			E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
			       ("ispfc: Enabling fibrechannel link"));
			ispfc_send_fc_state_msg(ISPFC_CLS_STATE_UP);
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
			cb->link_state &= ~ISPFC_CLF_FC_OFFLINE_LATCH;
		}
		ispfc_set_autolinkenable_on();
		break;

	case HAL_IOCTL_COMMON_RESET_DEVICE:
		{
			uint32		hal_handle;
			ispfc_dev_t *	dev;
			
                        va_start(ap, code);
                        hal_handle = va_arg(ap, uint32);
                        va_end(ap);
			
			dev = ispfc_hash_table_obj_find(cb->dev_handle_hash, hal_handle);
			if (dev) {
				FC_DBG(0, ("%s: Resetting phys_dev_id 0x%x, handle %d\n",
					   cb->debug_name, dev->phys_dev_id, dev->dev_handle));
				SET_DEV_STATE(dev, ISPFC_DS_STATE_RESET_REQUESTED);
			}
			else 
				return(NFX_ERR);
		}
		break;

 	case HAL_IOCTL_COMMON_GET_DEVICE_MAP:
		{
			wwn_t			channel_wwn;
			hal_device_map_t *	map; 
			void (*callback)(wwn_t, hal_device_map_t *, int, void *);
			void *			cookie;

			va_start(ap, code);
			channel_wwn = va_arg(ap, wwn_t);
			map = va_arg(ap, void *);
			callback = va_arg(ap, void *);
			cookie = va_arg(ap, void *);
			va_end(ap);

			if (cb->interface_type == HAL_IF_FC_L_PORT)
				ispfc_get_position_map(cb, channel_wwn, map, callback, cookie);
			else
			if (cb->interface_type == HAL_IF_FC_N_PORT) {
				/* Create a fake map with just two ports on it */
				map->hostid = cb->host_port_id;
				map->port_count = 2;
				map->port_map[0] = cb->host_port_id;
				map->port_map[1] = (cb->host_port_id == 0xef) ? 0xe8 : 0xef;
				(*callback)(channel_wwn, map, NFX_OK, cookie);
			}
			else
				return(NFX_ERR);
			break;
		}
		break;

	case HAL_IOCTL_COMMON_RESCAN:
		FC_DBG(0, ("%s: Rescanning link\n", cb->debug_name));
		E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
		       ("ispfc: Rescanning fibrechannel link"));

		/* Mark all devices for re-registration */
		{
			ispfc_dev_t *	dev;

			FOR_ALL_DEVS({
				dev->dev_state |= ISPFC_DF_REGISTRATION_REQUESTED;
			});
		}

		if (IS_LINK_UP(cb)) {
			if (cb->interface_type == HAL_IF_FC_L_PORT) {
				SET_LINK_STATE(cb, ISPFC_CLS_STATE_LIP_OCCURRED);
			}
			else
			if (cb->interface_type == HAL_IF_FC_FL_PORT || cb->interface_type == HAL_IF_FC_F_PORT) {
				SET_LINK_STATE(cb, ISPFC_CLS_STATE_NSCN_OCCURRED);
			}
			else {
				SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
			}
		}
		break;

	case HAL_IOCTL_COMMON_KILL_WRITES:
		{
			uint32		hal_handle;
			uint32		lun;
			ispfc_dev_t *	dev;
			scsi_desc_t *	sd;

                        va_start(ap, code);
                        hal_handle = va_arg(ap, uint32);
			lun = va_arg(ap, uint32);
                        va_end(ap);

			dev = ispfc_hash_table_obj_find(cb->dev_handle_hash, hal_handle);
			if (dev) {

				FC_DBG(0, ("%s: Killing all writes for phys_dev_id 0x%x, lun %d, handle %d\n",
					   cb->debug_name, dev->phys_dev_id, lun, dev->dev_handle));
				
				/*  
				 * Grope thru the busy and waiting
				 * queue and "kill" all queued
				 * writes. We do this by setting the
				 * max time to live to 0, thus when
				 * this command is to be scheduled
				 * again it will be ABORTED.  
				 */
				if ((sd = dev->scsi_desc_busy)) {
					do {
						if ((sd->control & SD_OPCODE_MASK) == SD_OPCODE_WRITE && SCSI_PATH_GET_LUN(sd->path) == lun)
							sd->maxTimeToLive = 0;
						++sd->retryCnt;
					} while ((sd = sd->next) != dev->scsi_desc_busy);
				}
				if ((sd = dev->scsi_desc_waiting)) {
					do {
						if ((sd->control & SD_OPCODE_MASK) == SD_OPCODE_WRITE && SCSI_PATH_GET_LUN(sd->path) == lun)
							sd->maxTimeToLive = 0;
						++sd->retryCnt;
					} while ((sd = sd->next) != dev->scsi_desc_waiting);
				}
				/*
				 * Finally, fire off an abort
				 * taskset. This will abort all
				 * outstanding I/Os, but all other
				 * than writes will be retried.
				 */
				ispfc_dev_abort_taskset(cb, dev, lun);
				return(NFX_OK);
			}
			else 
				return(NFX_ERR);
			

		}
		break;

	case HAL_IOCTL_FC_GET_ADAPTER_LINK_STATS:
		{
			hal_fc_link_stats_t *	stats;
			void (*callback)(void *, uint32);
			void *			cookie;
			ispfc_ioctl_sm_t *	sm = NULL;

			va_start(ap, code);
			stats = va_arg(ap, void *);
			callback = va_arg(ap, void *);
			cookie = va_arg(ap, void *);
			va_end(ap);

			stats->link_failure = -1;
			stats->loss_of_sync = -1;
			stats->loss_of_signal = -1;
			stats->prim_proto_error = -1;
			stats->invalid_xmit_words = -1;
			stats->invalid_crc = -1;
			stats->frames_out = cb->adap_frames_out;
			stats->frames_in = cb->adap_frames_in;
			stats->transport_errors = cb->adap_transport_error_cnt;
			stats->underrun_cnt = cb->adap_underrun_cnt;
			stats->read_6_cmd_cnt = cb->adap_read_6_cmd_cnt;
			stats->read_10_cmd_cnt = cb->adap_read_10_cmd_cnt;
			stats->write_6_cmd_cnt = cb->adap_write_6_cmd_cnt;
			stats->write_10_cmd_cnt = cb->adap_write_10_cmd_cnt;
			stats->other_cmd_cnt = cb->adap_other_cmd_cnt;
			stats->count_per_usec = count_per_usec;

			scsi_getstats(stats, FALSE);

			/* If mailbox busy, or link not UP bail */
			if (ispfc_mbox_is_busy(cb) || !IS_LINK_UP(cb)) {
				(*callback)(cookie, NFX_OK);
				return(NFX_OK);
			}				
			
			CACHE_FLUSH(stats, stats + 1);

			sm = eee_ramAlloc(sizeof(ispfc_ioctl_sm_t));
			VERIFY(sm);
			sm->cookie = cookie;
			sm->callback = callback;

			/* Fire off a GET LINK STATUS mailbox command */
			cb->mbox_in_valid_count = 8;
			cb->mbox_in[0] = ISPFC_GET_LINK_STATUS;
			cb->mbox_in[1] = cb->host_loop_id << 8;
			cb->mbox_in[2] = VA_TO_IO32(&stats->link_failure) >> 16;
			cb->mbox_in[3] = VA_TO_IO32(&stats->link_failure) & 0xFFFF;
			cb->mbox_in[4] = 0;
			cb->mbox_in[5] = 0;
			cb->mbox_in[6] = 0;
			cb->mbox_in[7] = 0;
			ispfc_mbox_start(cb, ispfc_ioctl_done, sm);
		}
		break;

	case HAL_IOCTL_FC_CLR_ADAPTER_LINK_STATS:
		{
			cb->adap_frames_out = 0;
			cb->adap_frames_in  = 0;
			cb->adap_underrun_cnt = 0;
			cb->adap_transport_error_cnt = 0;
			cb->adap_read_6_cmd_cnt = 0;
			cb->adap_read_10_cmd_cnt = 0;
			cb->adap_write_6_cmd_cnt = 0;
			cb->adap_write_10_cmd_cnt = 0;
			cb->adap_other_cmd_cnt = 0;
		}
		break;

	case HAL_IOCTL_FC_LIP:
		{
			wwn_t			channel_wwn;
			uint32			lip_byte3;
			uint32			lip_byte4;
			void (*callback)(void *, uint32);
			void *			cookie;

			va_start(ap, code);
			channel_wwn = va_arg(ap, wwn_t);
			lip_byte3 = va_arg(ap, uint32);
			lip_byte4 = va_arg(ap, uint32);
			callback = va_arg(ap, void *);
			cookie = va_arg(ap, void *);
                        va_end(ap);

			inplace_htonll(&channel_wwn);

			ispfc_lip(cb, channel_wwn, lip_byte3, lip_byte4, callback, cookie);
		}
		break;

	case HAL_IOCTL_FC_LOOP_PORT_BYPASS:
	case HAL_IOCTL_FC_LOOP_PORT_ENABLE:
		{
			wwn_t			dev_portname;
			void (*callback)(void *, uint32);
			void *			cookie;

			va_start(ap, code);
			dev_portname = va_arg(ap, wwn_t);
			callback = va_arg(ap, void *);
			cookie = va_arg(ap, void *);
                        va_end(ap);

			inplace_htonll(&dev_portname);

			if (code == HAL_IOCTL_FC_LOOP_PORT_BYPASS)
				ispfc_bypass_device(cb, dev_portname, callback, cookie);
			else
				ispfc_unbypass_device(cb, dev_portname, callback, cookie);
		}
		break;

	case HAL_IOCTL_FC_LOOP_PORT_ENABLE_ALL:
		{
			wwn_t			channel_wwn;
			void (*callback)(void *, uint32);
			void *			cookie;

			va_start(ap, code);
			channel_wwn = va_arg(ap, wwn_t);
			callback = va_arg(ap, void *);
			cookie = va_arg(ap, void *);
                        va_end(ap);

			inplace_htonll(&channel_wwn);

			ispfc_unbypass_all(cb, channel_wwn, callback, cookie);
		}

	case HAL_IOCTL_FC_CONNECTION_STATE:
		{
			uint32      *p;

                        va_start(ap, code);
                        p = va_arg(ap, uint32 *);
                        va_end(ap);

			if (IS_LINK_UP(cb))
				*p = HAL_IOCTL_FC_LINK_STATE_LINK_UP;
			else
			if (IS_LINK_OFFLINE(cb))
				*p = HAL_IOCTL_FC_LINK_STATE_LINK_OFFLINE;
			else
				*p = HAL_IOCTL_FC_LINK_STATE_LINK_DOWN;
		}
		break;

	case HAL_IOCTL_FC_TOPOLOGY:
		{
			void *		topology;
			uint32 *	topology_len;
			
			va_start(ap, code);
			topology = va_arg(ap, void *);
			topology_len = va_arg(ap, uint32 *);
			va_end(ap);

			fctd_format_topology(topology, topology_len);
		}
		break;

	default:
		return(NFX_ERR);
	}
	return(NFX_OK);
}

void 
ispfc_ioctl_done(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_ioctl_sm_t *	sm = cookie;

	FC_DBG(2, ("%s: ispfc_ioctl_done: status = 0x%x\n", 
		   cb->debug_name, status));
	
	if (status == MBOX_CMD_STATUS_TIMEOUT)
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	
	(*sm->callback)(sm->cookie, status == MBOX_CMD_STATUS_SUCCESS ? NFX_OK : NFX_ERR);
	eee_ramDealloc((void *)sm);
	return;
}

void
ispfc_bypass_device(ispfc_cb_t *cb, wwn_t dev_portname, void (*callback)(void *, uint32), void *cookie)
{
	ispfc_lpc_sm_t *	sm;
	ispfc_dev_t *		dev;
	channel_name_t		ch;

	FC_DBG(0, ("%s: ispfc_bypass_device: dev_portname 0x%llx\n", cb->debug_name, dev_portname.u.wwn));
	
	sm = eee_ramAlloc(sizeof(ispfc_lpc_sm_t));
	VERIFY(sm);

	sm->cookie = cookie;
	sm->callback = callback;
	sm->type = ISPFC_LPC_BYPASS;

	dev = ispfc_device_lookup_by_portname(cb, dev_portname);
	if (dev == NULL)
		goto error;

	WWN_SET(&sm->channel_wwn, &dev->channel_wwn);

	/* If mailbox busy, bail */
	if (ispfc_mbox_is_busy(cb))
		goto error;

	if (cb->interface_type == HAL_IF_FC_L_PORT) {
		VERIFY(WWN_EQUAL(&sm->channel_wwn, &cb->adapter_wwn) || 
		       WWN_EQUAL(&sm->channel_wwn, &null_wwn));
		/* Fire off a LOOP PORT BYPASS mailbox command */
		cb->mbox_in_valid_count = 2;
		cb->mbox_in[0] = ISPFC_LOOP_PORT_BYPASS;
		cb->mbox_in[1] = ispfc_alpa_to_loopid(ISPFC_DEV_PHYS_DEV_TO_ALPA(dev->phys_dev_id)) << 8;
		ispfc_mbox_start(cb, ispfc_lpc_done, sm);
	}
	else {
		if (WWN_EQUAL(&sm->channel_wwn, &cb->adapter_wwn) ||
		    WWN_EQUAL(&sm->channel_wwn, &null_wwn)) {
			goto error;
		}
		WWN_SET(&ch.channel_wwn, &sm->channel_wwn);
		if (ispfc_channel_lookup(&ch) != NFX_OK ||
		    ch.channel_interface_type != HAL_IF_FC_FL_PORT) {
			goto error;
		}
		bzero(&sm->lfa_req, sizeof(ispfc_lfa_cb_t));
		sm->lfa_req.lfa_response_buf_len = 4;
		sm->lfa_req.lfa_response_buf_lo32 = VA_TO_IO32(&sm->lfa_resp);
		sm->lfa_req.lfa_command_len = 20;
		sm->lfa_req.lfa_fl_port = ch.u.fc_f_or_fc_fl.channel_fl_port;
		sm->lfa_req.lfa_command_code = LFA_COMMAND_CODE_LPC;
		sm->lfa_req.u.lpc.lfa_lpc_port_control_function = LFA_LPC_PORT_CONTROL_FUNCTION_NORMAL;

		/* 
		 * Construct LPB bit map
		 */
		{
			int	short_offset, bit_offset;
			uint32	loop_id = ispfc_alpa_to_loopid(ISPFC_DEV_PHYS_DEV_TO_ALPA(dev->phys_dev_id));

			short_offset = 7 - (loop_id / 16);
			bit_offset = loop_id % 16;
			sm->lfa_req.u.lpc.lfa_lpc_lpb_alpa_bitmap[short_offset] |= (1 << bit_offset);
#if 0		/* Seems like Qlogic byte swaps LFA autonomously */
			inplace_htons(&sm->lfa_req.u.lpc.lfa_lpc_lpb_alpa_bitmap[short_offset]);
#endif
		}

#if 0		/* Seems like Qlogic byte swaps LFA autonomously */
		inplace_ntohs(&sm->lfa_req.lfa_response_buf_len);
		inplace_ntohl(&sm->lfa_req.lfa_response_buf_lo32);
		inplace_ntohs(&sm->lfa_req.lfa_command_len);
		inplace_ntohl(&sm->lfa_req.lfa_fl_port);
		inplace_ntohs(&sm->lfa_req.lfa_command_code);
		inplace_ntohs(&sm->lfa_req.u.lpc.lfa_lpc_port_control_function);
#endif

		CACHE_FLUSH(sm, sm + 1);

		/* Fire off a SEND LFA mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_SEND_LFA;
		cb->mbox_in[1] = sizeof(ispfc_lfa_cb_t)/sizeof(ushort16);
		cb->mbox_in[2] = VA_TO_IO32(&sm->lfa_req) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(&sm->lfa_req) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		ispfc_mbox_start(cb, ispfc_lpc_done, sm);
	}

	return;

 error:
	(*sm->callback)(sm->cookie, NFX_ERR);
	eee_ramDealloc((void *)sm);
}

void
ispfc_unbypass_device(ispfc_cb_t *cb, wwn_t dev_portname, void (*callback)(void *, uint32), void *cookie)
{
	ispfc_lpc_sm_t *	sm;
	ispfc_dev_t *		dev;
	channel_name_t		ch;

	FC_DBG(0, ("%s: ispfc_unbypass_device: dev_portname 0x%llx\n", cb->debug_name, dev_portname.u.wwn));
	
	sm = eee_ramAlloc(sizeof(ispfc_lpc_sm_t));
	VERIFY(sm);

	sm->cookie = cookie;
	sm->callback = callback;
	sm->type = ISPFC_LPC_UNBYPASS;

	dev = ispfc_device_lookup_by_portname(cb, dev_portname);
	if (dev == NULL)
		goto error;

	WWN_SET(&dev->channel_wwn, &sm->channel_wwn);

	/* If mailbox busy, bail */
	if (ispfc_mbox_is_busy(cb))
		goto error;

	if (cb->interface_type == HAL_IF_FC_L_PORT) {
		VERIFY(WWN_EQUAL(&sm->channel_wwn, &cb->adapter_wwn) || 
		       WWN_EQUAL(&sm->channel_wwn, &null_wwn));
		/* Fire off a LOOP PORT UNBYPASS mailbox command */
		cb->mbox_in_valid_count = 2;
		cb->mbox_in[0] = ISPFC_LOOP_PORT_ENABLE;
		cb->mbox_in[1] = ispfc_alpa_to_loopid(ISPFC_DEV_PHYS_DEV_TO_ALPA(dev->phys_dev_id)) << 8;
		ispfc_mbox_start(cb, ispfc_lpc_done, sm);
	}
	else {
		if (WWN_EQUAL(&sm->channel_wwn, &cb->adapter_wwn) ||
		    WWN_EQUAL(&sm->channel_wwn, &null_wwn)) {
			goto error;
		}
		WWN_SET(&ch.channel_wwn, &sm->channel_wwn);
		if (ispfc_channel_lookup(&ch) != NFX_OK ||
		    ch.channel_interface_type != HAL_IF_FC_FL_PORT) {
			goto error;
		}
		bzero(&sm->lfa_req, sizeof(ispfc_lfa_cb_t));
		sm->lfa_req.lfa_response_buf_len = 4;
		sm->lfa_req.lfa_response_buf_lo32 = VA_TO_IO32(&sm->lfa_resp);
		sm->lfa_req.lfa_command_len = 20;
		sm->lfa_req.lfa_fl_port = ch.u.fc_f_or_fc_fl.channel_fl_port;
		sm->lfa_req.lfa_command_code = LFA_COMMAND_CODE_LPC;
		sm->lfa_req.u.lpc.lfa_lpc_port_control_function = LFA_LPC_PORT_CONTROL_FUNCTION_NORMAL;

		/* 
		 * Construct LPE bit map
		 */
		{
			int	short_offset, bit_offset;
			uint32	loop_id = ispfc_alpa_to_loopid(ISPFC_DEV_PHYS_DEV_TO_ALPA(dev->phys_dev_id));

#if 0
			if (loop_id == 0xFF) {
				sm->lfa_req.u.lpc.lfa_lpc_lpe_alpa_bitmap[0] |= 0x8000;
				inplace_htons(&sm->lfa_req.u.lpc.lfa_lpc_lpe_alpa_bitmap[0]);
			}
#endif

			short_offset = 7 - (loop_id / 16);
			bit_offset = loop_id % 16;
			sm->lfa_req.u.lpc.lfa_lpc_lpe_alpa_bitmap[short_offset] |= (1 << bit_offset);
#if 0		/* Seems like Qlogic byte swaps LFA autonomously */
			inplace_htons(&sm->lfa_req.u.lpc.lfa_lpc_lpe_alpa_bitmap[short_offset]);
#endif
		}

#if 0		/* Seems like Qlogic byte swaps LFA autonomously */
		inplace_ntohs(&sm->lfa_req.lfa_response_buf_len);
		inplace_ntohl(&sm->lfa_req.lfa_response_buf_lo32);
		inplace_ntohs(&sm->lfa_req.lfa_command_len);
		inplace_ntohl(&sm->lfa_req.lfa_fl_port);
		inplace_ntohs(&sm->lfa_req.lfa_command_code);
		inplace_ntohs(&sm->lfa_req.u.lpc.lfa_lpc_port_control_function);
#endif

		CACHE_FLUSH(sm, sm + 1);

		/* Fire off a SEND LFA mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_SEND_LFA;
		cb->mbox_in[1] = sizeof(ispfc_lfa_cb_t)/sizeof(ushort16);
		cb->mbox_in[2] = VA_TO_IO32(&sm->lfa_req) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(&sm->lfa_req) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		ispfc_mbox_start(cb, ispfc_lpc_done, sm);
	}

	return;

 error:
	(*sm->callback)(sm->cookie, NFX_ERR);
	eee_ramDealloc((void *)sm);
}

void
ispfc_unbypass_all(ispfc_cb_t *cb, wwn_t channel_wwn, void (*callback)(void *, uint32), void *cookie)
{
	ispfc_lpc_sm_t *	sm;
	channel_name_t		ch;

	FC_DBG(0, ("%s: ispfc_unbypass_all: channel_wwn 0x%llx\n", cb->debug_name, channel_wwn.u.wwn));
	
	sm = eee_ramAlloc(sizeof(ispfc_lpc_sm_t));
	VERIFY(sm);

	sm->cookie = cookie;
	sm->callback = callback;
	sm->type = ISPFC_LPC_UNBYPASS;

	WWN_SET(&sm->channel_wwn, &channel_wwn);

	/* If mailbox busy, bail */
	if (ispfc_mbox_is_busy(cb))
		goto error;

	if (cb->interface_type == HAL_IF_FC_L_PORT) {
		VERIFY(WWN_EQUAL(&sm->channel_wwn, &cb->adapter_wwn) || 
		       WWN_EQUAL(&sm->channel_wwn, &null_wwn));
		/* Fire off a LOOP PORT UNBYPASS mailbox command */
		cb->mbox_in_valid_count = 2;
		cb->mbox_in[0] = ISPFC_LOOP_PORT_ENABLE;
		cb->mbox_in[1] = 0xFF << 8;
		ispfc_mbox_start(cb, ispfc_lpc_done, sm);
	}
	else {
		if (WWN_EQUAL(&sm->channel_wwn, &cb->adapter_wwn) ||
		    WWN_EQUAL(&sm->channel_wwn, &null_wwn)) {
			goto error;
		}
		WWN_SET(&ch.channel_wwn, &sm->channel_wwn);
		if (ispfc_channel_lookup(&ch) != NFX_OK ||
		    ch.channel_interface_type != HAL_IF_FC_FL_PORT) {
			goto error;
		}
		bzero(&sm->lfa_req, sizeof(ispfc_lfa_cb_t));
		sm->lfa_req.lfa_response_buf_len = 4;
		sm->lfa_req.lfa_response_buf_lo32 = VA_TO_IO32(&sm->lfa_resp);
		sm->lfa_req.lfa_command_len = 20;
		sm->lfa_req.lfa_fl_port = ch.u.fc_f_or_fc_fl.channel_fl_port;
		sm->lfa_req.lfa_command_code = LFA_COMMAND_CODE_LPC;
		sm->lfa_req.u.lpc.lfa_lpc_port_control_function = LFA_LPC_PORT_CONTROL_FUNCTION_NORMAL;

		/* 
		 * Construct LPE bit map
		 */
		sm->lfa_req.u.lpc.lfa_lpc_lpe_alpa_bitmap[0] |= 0x8000;
#if 0		/* Seems like Qlogic byte swaps LFA autonomously */
		inplace_htons(&sm->lfa_req.u.lpc.lfa_lpc_lpe_alpa_bitmap[0]);
#endif

#if 0		/* Seems like Qlogic byte swaps LFA autonomously */
		inplace_ntohs(&sm->lfa_req.lfa_response_buf_len);
		inplace_ntohl(&sm->lfa_req.lfa_response_buf_lo32);
		inplace_ntohs(&sm->lfa_req.lfa_command_len);
		inplace_ntohl(&sm->lfa_req.lfa_fl_port);
		inplace_ntohs(&sm->lfa_req.lfa_command_code);
		inplace_ntohs(&sm->lfa_req.u.lpc.lfa_lpc_port_control_function);
#endif

		CACHE_FLUSH(sm, sm + 1);

		/* Fire off a SEND LFA mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_SEND_LFA;
		cb->mbox_in[1] = sizeof(ispfc_lfa_cb_t)/sizeof(ushort16);
		cb->mbox_in[2] = VA_TO_IO32(&sm->lfa_req) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(&sm->lfa_req) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		ispfc_mbox_start(cb, ispfc_lpc_done, sm);
	}

	return;

 error:
	(*sm->callback)(sm->cookie, NFX_ERR);
	eee_ramDealloc((void *)sm);
}

void
ispfc_lpc_done(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_lpc_sm_t *	sm = cookie;
	channel_name_t		ch;

	FC_DBG(0, ("%s: ispfc_lpc_done: status = 0x%x\n", cb->debug_name, status));
	
	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else
	if (status == MBOX_CMD_STATUS_SUCCESS) {
		if (cb->interface_type == HAL_IF_FC_FL_PORT ||
		    cb->interface_type == HAL_IF_FC_F_PORT) {
			WWN_SET(&ch.channel_wwn, &sm->channel_wwn);
			if (ispfc_channel_lookup(&ch) != NFX_OK) {
				goto error;
			}

			inplace_ntohl(&sm->lfa_resp.lfa_response_code);
			if (sm->lfa_resp.lfa_response_code != LFA_RESPONSE_CODE_LS_ACC) {
				FC_DBG(0, ("%s: LPC on remote loop %s failed: response code 0x%x\n", 
					   cb->debug_name,
					   ch.channel_name,
					   sm->lfa_resp.lfa_response_code));
				goto error;
			}
			if (sm->lfa_resp.u.lpc.lfa_lpc_status != LFA_LINIT_STATUS_SUCCESS) {
				FC_DBG(0, ("%s: Failure of LPC on remote loop %s. (status 0x%x)\n",
					   cb->debug_name,
					   ch.channel_name,
					   sm->lfa_resp.u.lpc.lfa_lpc_status));
				goto error;
			}
		}
	}

	(*sm->callback)(sm->cookie, status == MBOX_CMD_STATUS_SUCCESS ? NFX_OK : NFX_ERR);
	eee_ramDealloc((void *)sm);
	return;

 error:
	(*sm->callback)(sm->cookie, NFX_ERR);
	eee_ramDealloc((void *)sm);
	return;

}

void
ispfc_lip(ispfc_cb_t *cb, wwn_t channel_wwn, uint32 lip_byte3, uint32 lip_byte4, void (*callback)(void *, uint32), void *cookie)
{
	ispfc_lip_sm_t *	sm;
	channel_name_t		ch;

	FC_DBG(0, ("%s: ispfc_lip: channel_wwn 0x%llx, byte 3 0x%x, byte 4 0x%x\n", 
		   cb->debug_name, channel_wwn.u.wwn, lip_byte3, lip_byte4));

	sm = eee_ramAlloc(sizeof(ispfc_lip_sm_t));
	VERIFY(sm);

	sm->cookie = cookie;
	sm->callback = callback;

	WWN_SET(&sm->channel_wwn, &channel_wwn);

	/* If mailbox busy, bail */
	if (ispfc_mbox_is_busy(cb))
		goto error;

	if (cb->interface_type == HAL_IF_FC_L_PORT) {
		VERIFY(WWN_EQUAL(&channel_wwn, &cb->adapter_wwn) ||
                       WWN_EQUAL(&channel_wwn, &null_wwn));

		if (LINK_STATE(cb) != ISPFC_CLS_STATE_LIP_GENERATED) {
			/* Fire off a INITIATE LIP mailbox command */
			cb->mbox_in_valid_count = 1;
			cb->mbox_in[0] = ISPFC_INITIATE_LIP;
			ispfc_mbox_start(cb, ispfc_lip_done, sm);

			if (LINK_STATE(cb) == ISPFC_CLS_STATE_UP)
				SET_LINK_STATE(cb, ISPFC_CLS_STATE_LIP_GENERATED);
		}
	}
	else {
		if (WWN_EQUAL(&channel_wwn, &cb->adapter_wwn) ||
                    WWN_EQUAL(&channel_wwn, &null_wwn)) {
                        goto error;
                }
                WWN_SET(&ch.channel_wwn, &channel_wwn);
                if (ispfc_channel_lookup(&ch) != NFX_OK ||
                    ch.channel_interface_type != HAL_IF_FC_FL_PORT) {
                        goto error;
                }
		bzero(&sm->lfa_req, sizeof(ispfc_lfa_cb_t));
		sm->lfa_req.lfa_response_buf_len = 4;
		sm->lfa_req.lfa_response_buf_lo32 = VA_TO_IO32(&sm->lfa_resp);
		sm->lfa_req.lfa_command_len = 4;
		sm->lfa_req.lfa_fl_port = ch.u.fc_f_or_fc_fl.channel_fl_port;
		sm->lfa_req.lfa_command_code = LFA_COMMAND_CODE_LINIT;
		sm->lfa_req.u.linit.lfa_linit_init_function = LFA_LINIT_INIT_FUNCTION_NORMAL;
		sm->lfa_req.u.linit.lfa_linit_lip_byte_3 = lip_byte3;
		sm->lfa_req.u.linit.lfa_linit_lip_byte_4 = lip_byte4;

#if 0		/* Seems like Qlogic byte swaps LFA autonomously */
		inplace_ntohs(&sm->lfa_req.lfa_response_buf_len);
		inplace_ntohl(&sm->lfa_req.lfa_response_buf_lo32);
		inplace_ntohs(&sm->lfa_req.lfa_command_len);
		inplace_ntohl(&sm->lfa_req.lfa_fl_port);
		inplace_ntohs(&sm->lfa_req.lfa_command_code);
		inplace_ntohs(&sm->lfa_req.u.linit.lfa_linit_init_function);
#endif

		CACHE_FLUSH(sm, sm + 1);

		/* Fire off a SEND LFA mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_SEND_LFA;
		cb->mbox_in[1] = sizeof(ispfc_lfa_cb_t)/sizeof(ushort16);
		cb->mbox_in[2] = VA_TO_IO32(&sm->lfa_req) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(&sm->lfa_req) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		ispfc_mbox_start(cb, ispfc_lip_done, sm);

		if (LINK_STATE(cb) == ISPFC_CLS_STATE_UP)
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_NSCN_GENERATED);
	}

	return;

 error:
	(*sm->callback)(sm->cookie, NFX_ERR);
	eee_ramDealloc((void *)sm);
}

void
ispfc_lip_done(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_lip_sm_t *	sm = cookie;
	channel_name_t		ch;

	FC_DBG(0, ("%s: ispfc_lip_done: status = 0x%x\n", 
		   cb->debug_name, status));
	
	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else
	if (status == MBOX_CMD_STATUS_SUCCESS) {
		if (cb->interface_type == HAL_IF_FC_FL_PORT ||
		    cb->interface_type == HAL_IF_FC_F_PORT) {
			WWN_SET(&ch.channel_wwn, &sm->channel_wwn);
			if (ispfc_channel_lookup(&ch) != NFX_OK) {
				goto error;
			}
		    
			inplace_ntohl(&sm->lfa_resp.lfa_response_code);
			if (sm->lfa_resp.lfa_response_code != LFA_RESPONSE_CODE_LS_ACC) {
				FC_DBG(0, ("%s: LINIT LIP(%x,%x) on remote loop %s failed: response code 0x%x\n", 
					   cb->debug_name,
					   sm->lfa_req.u.linit.lfa_linit_lip_byte_3,
					   sm->lfa_req.u.linit.lfa_linit_lip_byte_4,
					   ch.channel_name,
					   sm->lfa_resp.lfa_response_code));
				goto error;
			}
			if (sm->lfa_resp.u.linit.lfa_linit_status != LFA_LINIT_STATUS_SUCCESS) {
				FC_DBG(0, ("%s: Failure of LINIT LIP(%x,%x) on remote loop %s. (status 0x%x)\n", 
					   cb->debug_name,
					   sm->lfa_req.u.linit.lfa_linit_lip_byte_3,
					   sm->lfa_req.u.linit.lfa_linit_lip_byte_4,
					   ch.channel_name,
					   sm->lfa_resp.u.linit.lfa_linit_status));
				goto error;
			}
		}
	}

	(*sm->callback)(sm->cookie, status == MBOX_CMD_STATUS_SUCCESS ? NFX_OK : NFX_ERR);
	eee_ramDealloc((void *)sm);
	return;

 error:
	(*sm->callback)(sm->cookie, NFX_ERR);
	eee_ramDealloc((void *)sm);
	return;
}

void
ispfc_get_position_map(ispfc_cb_t *cb, 
		       wwn_t channel_wwn, 
		       hal_device_map_t *caller_map,
		       void (*callback)(wwn_t, hal_device_map_t *, int, void *),
		       void *cookie)
{
	ispfc_get_position_map_sm_t *	sm;
	channel_name_t			ch;

	sm = eee_ramAlloc(sizeof(ispfc_get_position_map_sm_t));
	VERIFY(sm);
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	sm = (ispfc_get_position_map_sm_t *)cached_to_uncached((address_t)sm);
#else
	(ps_uint_t)sm |= 0x20000000L; /* Turn on uncached access */
#endif
#endif
	WWN_SET(&sm->channel_wwn, &channel_wwn);
	sm->caller_map = caller_map;
	sm->callback = callback;
	sm->cookie = cookie;

	/* If mailbox busy, bail */
	if (ispfc_mbox_is_busy(cb))
		goto error;

	if (cb->interface_type == HAL_IF_FC_L_PORT) {
		VERIFY(WWN_EQUAL(&channel_wwn, &cb->adapter_wwn) || 
		       WWN_EQUAL(&channel_wwn, &null_wwn));
		
		/* Fire off a GET LOOP MAP mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_GET_FCAL_POSITION_MAP;
		cb->mbox_in[1] = 0;
		cb->mbox_in[2] = VA_TO_IO32(&sm->loop_map) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(&sm->loop_map) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		ispfc_mbox_start(cb, ispfc_get_position_map_done, sm);
	}
	else
	if (cb->interface_type == HAL_IF_FC_FL_PORT ||
	    cb->interface_type == HAL_IF_FC_F_PORT) {
		if (WWN_EQUAL(&channel_wwn, &cb->adapter_wwn) ||
		    WWN_EQUAL(&channel_wwn, &null_wwn)) {
			goto error;
		}
		WWN_SET(&ch.channel_wwn, &channel_wwn);
		if (ispfc_channel_lookup(&ch) != NFX_OK) {
			goto error;
		}

		bzero(&sm->lfa_req, sizeof(ispfc_lfa_cb_t));
		sm->lfa_req.lfa_response_buf_len = 4;
		sm->lfa_req.lfa_response_buf_lo32 = VA_TO_IO32(&sm->lfa_resp);
		sm->lfa_req.lfa_command_len = 2;
		sm->lfa_req.lfa_fl_port = ch.u.fc_f_or_fc_fl.channel_fl_port;
		sm->lfa_req.lfa_command_code = LFA_COMMAND_CODE_LSTS;

		inplace_ntohs(&sm->lfa_req.lfa_response_buf_len);
		inplace_ntohl(&sm->lfa_req.lfa_response_buf_lo32);
		inplace_ntohs(&sm->lfa_req.lfa_command_len);
		inplace_ntohl(&sm->lfa_req.lfa_fl_port);
		inplace_ntohs(&sm->lfa_req.lfa_command_code);

		/* Fire off a SEND LFA mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_SEND_LFA;
		cb->mbox_in[1] = sizeof(ispfc_lfa_cb_t)/sizeof(ushort16);
		cb->mbox_in[2] = VA_TO_IO32(&sm->lfa_req) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(&sm->lfa_req) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		ispfc_mbox_start(cb, ispfc_get_position_map_done, sm);
	}
	else
		VERIFY(0);

	return;

 error:
	(*sm->callback)(channel_wwn, caller_map, NFX_ERR, cookie);
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	sm = (ispfc_get_position_map_sm_t *)uncached_to_cached((address_t)sm);
#else
	(ps_uint_t)sm &= ~0x20000000L; /* Turn off uncached access */
#endif
#endif
	eee_ramDealloc((void *)sm);
}

void
ispfc_get_position_map_done(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_get_position_map_sm_t *	sm = cookie;
	channel_name_t			ch;
	int				i;

	FC_DBG(1, ("%s: ispfc_get_position_map_done: status = 0x%x\n", 
		   cb->debug_name, status));
	
	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else
	if (status == MBOX_CMD_STATUS_SUCCESS) {
		if (cb->interface_type == HAL_IF_FC_FL_PORT ||
		    cb->interface_type == HAL_IF_FC_F_PORT) {
			WWN_SET(&ch.channel_wwn, &sm->channel_wwn);
			if (ispfc_channel_lookup(&ch) != NFX_OK) {
				goto error;
			}
		    
			inplace_ntohl(&sm->lfa_resp.lfa_response_code);
			if (sm->lfa_resp.lfa_response_code != LFA_RESPONSE_CODE_LS_ACC) {
				FC_DBG(0, ("%s: LSTS on remote loop %s failed: response code 0x%x\n", 
					   cb->debug_name,
					   ch.channel_name,
					   sm->lfa_resp.lfa_response_code));
				goto error;
			}
			if (sm->lfa_resp.u.lsts.lfa_lsts_loop_state != LFA_LSTS_LOOP_STATE_ONLINE) {
				FC_DBG(0, ("%s: Remote loop %s not online: state 0x%x\n", 
					   cb->debug_name,
					   ch.channel_name,
					   sm->lfa_resp.u.lsts.lfa_lsts_loop_state));
				goto error;
			}
			memcpy(&sm->loop_map, 
			       &sm->lfa_resp.u.lsts.lfa_lsts_position_map, 
			       sizeof(sm->lfa_resp.u.lsts.lfa_lsts_position_map));
		}
		if (cb->interface_type == HAL_IF_FC_F_PORT)
			sm->caller_map->hostid = ispfc_alpa_to_loopid(0);
		else 
			sm->caller_map->hostid = cb->host_loop_id;
		sm->caller_map->port_count = sm->loop_map.port_count;
		for (i = 0; i < sm->loop_map.port_count; ++i)
			sm->caller_map->port_map[i] = ispfc_alpa_to_loopid(sm->loop_map.port_map[i]);
	}

	(*sm->callback)(sm->channel_wwn, sm->caller_map, status == MBOX_CMD_STATUS_SUCCESS ? NFX_OK : NFX_ERR, sm->cookie);
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	sm = (ispfc_get_position_map_sm_t *)uncached_to_cached((address_t)sm);
#else
	(ps_uint_t)sm &= ~0x20000000L; /* Turn off uncached access */
#endif
#endif
	eee_ramDealloc((void *)sm);
	return;

 error:
	(*sm->callback)(sm->channel_wwn, sm->caller_map, NFX_ERR, sm->cookie);
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	sm = (ispfc_get_position_map_sm_t *)uncached_to_cached((address_t)sm);
#else
	(ps_uint_t)sm &= ~0x20000000L; /* Turn off uncached access */
#endif
#endif
	eee_ramDealloc((void *)sm);
	return;
}

