/*
 *-----------------------------------------------------------------------------
 *
 * Name:	ispfc_link.c
 *
 * RCS:		$Id: ispfc_link.c,v 1.44 2002/11/14 00:52:20 nitinb Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver connection, link and adapter 
 *              initialization state machine.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 5/31/01
 *
 *-----------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_mbox.h"
#include "ispfc_link.h"
#include "ispfc_ct.h"
#include "ispfc_device.h"
#include "ispfc_aen_handler.h"
#include "ispfc_scsi.h"
#include "ispfc_iocb.h"
#include "../sm-eee/eee-test.h"
#include <exports/event-api.h>
#include <exports/anpssc-api.h>
#include <exports/eee-api.h>
#include <exports/cm-api.h>

void ispfc_init_adap(ispfc_cb_t *, void (*)(ispfc_cb_t *, void *, int32), void *);
void ispfc_init_adap_1(ispfc_cb_t *, void * cookie, int32);
void ispfc_init_adap_link_up_timeout(uint32, ispfc_cb_t *, ispfc_adap_init_state_t *);

void ispfc_cleanup_link_state(ispfc_cb_t *);

void ispfc_init_conn(ispfc_cb_t *, void (*)(ispfc_cb_t *, void *, int32), void *);
void ispfc_init_conn_1(ispfc_cb_t *, void *, int32);
void ispfc_init_conn_fw_ready_delay(uint32, ispfc_cb_t *, ispfc_conn_init_state_t *);

void ispfc_init_link(ispfc_cb_t *, void (*)(ispfc_cb_t *, void *, int32), void *);
void ispfc_init_link_1(ispfc_cb_t *, void *, int32);
void ispfc_init_link_2(ispfc_cb_t *, void *, int32);

void ispfc_resetting_done_callout(ispfc_cb_t *, void *, int32);
void ispfc_reestablishing_done_callout(ispfc_cb_t *, void *, int32);

#define ISPFC_INIT_ADAP_RETRY	1
#define ISPFC_INIT_CONN_RETRY	2

void ispfc_led_control(ispfc_cb_t *, int);

#define ISPFC_LED_CONTROL_INIT	     0
#define ISPFC_LED_CONTROL_FLT_ON     1
#define ISPFC_LED_CONTROL_FLT_TOGGLE 2
#define ISPFC_LED_CONTROL_ACT_ON     3
#define ISPFC_LED_CONTROL_ACT_TOGGLE 4

#define DIAG_TRANSFER_LENGTH (65536 * 4)

/* SYSDVT */
extern sysdvt_qlogic_t  *qdvt;

/*******************************************************************************
 * ispfc_init_adap -
 *
 *   Primary entry point into adapter init state machine. State machine
 *   is initialized and the first state is entered.  
 ******************************************************************************/
void
ispfc_init_adap(ispfc_cb_t *cb, void (*adap_init_callback)(ispfc_cb_t *, void *, int32), void *adap_init_cookie)
{
	ispfc_adap_init_state_t	* st = cb->adap_init_state;

	VERIFY(st);

	FC_DBG(1, ("%s: Attempting %s initialization.\n", 
		   cb->debug_name,
		   ((cb->fw_init_cb->additional_fw_options & IFW_AFO_CONN_TYPE_MASK) == IFW_AFO_LOOP_ONLY) ? 
		   "Loop" : 
		   "Point-to-Point"));
	
	st->adap_init_retry = ISPFC_INIT_ADAP_RETRY;
	st->adap_init_state = 0;
	st->adap_init_callback = adap_init_callback;
	st->adap_init_cookie = adap_init_cookie;

	ispfc_init_adap_1(cb, st, MBOX_CMD_STATUS_SUCCESS);
}

/*******************************************************************************
 * ispfc_init_adap_1 -
 * 
 *  Runs the adapter init state machine.
 *   Retry up to 3 times:
 *	   State 0: 
 *		 o Soft reset adapter
 *		 o Enable SRAM parity, part 1
 *		 o mailbox(INIT_SRAM)
 *		 --> state 1
 *	   State 1:
 *		 FAILURE --> state 0 (retry)
 *		 o Enable SRAM parity, part 2
 *		 o mailbox(NOP)
 *		 --> state 2
 *	   State 2:
 *		 FAILURE --> state 0 (retry)
 *		 o mailbox(LOAD_RISC)
 *		 --> state 3
 *	   State 3:
 *		 FAILURE --> state 0 (retry)
 *		 o mailbox(VERIFY_CKSUM)
 *		 --> state 4
 *	   State 4:
 *		 FAILURE --> state 0 (retry)
 *		 o mailbox(EXEC_FIRM)
 *		 --> state 5
 *         State 5:
 *		 FAILURE --> state 0 (retry)
 *		 o mailbox(SET_FIRMWARE_OPTIONS)
 *		 --> state 6
 *	   State 6:
 *		 FAILURE --> state 0 (retry)
 *		 o mailbox(INITIALIZE_FIRM)
 *		 --> state 7
 *         State 7:
 *		 FAILURE --> state 0 (retry)
 *		 o wait for link up or timeout
 *		 --> state 8
 *	   State 8:
 *		 if (TIMEOUT)
 *		    o mailbox(LIP_RESET)
 *		    --> state 0 (retry)
 *		 FAILURE --> state 0 (retry)
 *		 SUCCESS --> return(SUCCESS)
 *
 *         State 9:
 *               DIAG_LOOP_BACK
 *		 
 *   return(FAILURE)
 ******************************************************************************/
void
ispfc_init_adap_1(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_adap_init_state_t *st = cookie;
	int			 rc;

 next_state:

	FC_DBG(2, ("%s: ispfc_init_adap_1: state = %d, status = 0x%x\n", 
		   cb->debug_name, st->adap_init_state, status));

	switch(st->adap_init_state) {
	case 0:
		/* Out of retries? Return failure */
		if (st->adap_init_retry == 0) {
			/* Failure */
			(*st->adap_init_callback)(cb, st->adap_init_cookie, ADAP_INIT_STATUS_FAILED);
			return;
		}
		/* Reset adapter */
		if (ispfc_soft_reset_adapter(cb) != NFX_OK) {
			/* Failure */
			(*st->adap_init_callback)(cb, st->adap_init_cookie, ADAP_INIT_STATUS_FAILED);
			return;
		}
		/* If SRAM parity available, (re)initialize SRAM */
		if (cb->sram_parity_available) {
			/* Pause RISC */
			ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_PAUSE);
			/* Enable parity checking */
			ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_WRT_PAR_ENABLE | 0x0001);
			/* Release RISC */
			ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_RELEASE);
			st->adap_init_generic_counter = 0x800;
			/* Fire off the first in a sequence of ISPFC_INIT_SRAM mailbox commands */
			cb->mbox_in_valid_count = 9;
			cb->mbox_in[0] = ISPFC_INIT_SRAM_EXT;
			cb->mbox_in[1] = (st->adap_init_generic_counter & 0xFFFF);
			cb->mbox_in[2] = 0x0800;
			cb->mbox_in[3] = 0xBEEF;
			cb->mbox_in[4] = 0;
			cb->mbox_in[5] = 0;
			cb->mbox_in[6] = 0;
			cb->mbox_in[7] = 0;
			cb->mbox_in[8] = st->adap_init_generic_counter >> 16;
			st->adap_init_generic_counter += 0x0800;
			st->adap_init_state = 1;
			ispfc_mbox_start(cb, ispfc_init_adap_1, st);
			return;
		}
		else {
			st->adap_init_state = 1;
			status = MBOX_CMD_STATUS_SUCCESS;
			goto next_state;
		}
		break;

	case 1:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		/* If SRAM parity available, continue (re)initialize SRAM */
                if (cb->sram_parity_available) {
			if (st->adap_init_generic_counter < 0x80000) { /* 512K 16-bit words */
				/* Fire off the next in a sequence of ISPFC_INIT_SRAM mailbox commands */
				cb->mbox_in_valid_count = 9;
				cb->mbox_in[0] = ISPFC_INIT_SRAM_EXT;
				cb->mbox_in[1] = (st->adap_init_generic_counter & 0xFFFF);
				cb->mbox_in[2] = 0x0800;
				cb->mbox_in[3] = 0xBEEF;
				cb->mbox_in[4] = 0;
				cb->mbox_in[5] = 0;
				cb->mbox_in[6] = 0;
				cb->mbox_in[7] = 0;
				cb->mbox_in[8] = st->adap_init_generic_counter >> 16;
				st->adap_init_generic_counter += 0x0800;
				st->adap_init_state = 1;
				ispfc_mbox_start(cb, ispfc_init_adap_1, st);
				return;
			}

			/* Reset RISC and wait for ready */
			if (ispfc_soft_reset_adapter(cb) != NFX_OK) {
				/* Failure */
				(*st->adap_init_callback)(cb, st->adap_init_cookie, ADAP_INIT_STATUS_FAILED);
				return;
			}
			/* Reenable SRAM parity checking */
			ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_PAUSE);
			ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_WRT_PAR_ENABLE | 0x0001);
			ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_RELEASE);
		}
		
		/* Fire off a ISPFC_NOP mailbox command */
		cb->mbox_in_valid_count = 1;
		cb->mbox_in[0] = ISPFC_NOP;
		st->adap_init_state = 2;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;
		
	case 2:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		if ((cb->link_state & ISPFC_CLS_MODE_MASK) == ISPFC_CLS_MODE_DIAG_SRAM) {
			st->adap_init_state = 21;
			goto next_state;
		}

		{
			int		i;
			ushort16 *	p = cb->risc_code;
			ushort16	checksum = 0;
			
			for (i = 0; i < cb->risc_code_length; ++i)
				checksum += *p++;
			FC_DBG(2, ("%s: ispfc_init_adap_1: Firmware address %p, size 0x%x, checksum 0x%x\n", 
				   cb->debug_name, cb->risc_code, cb->risc_code_length * 2, checksum));
		}

		CACHE_FLUSH((address_t)cb->risc_code, (address_t)cb->risc_code + cb->risc_code_length * 2);

		/* Fire off ISPFC_LOAD_RISC mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_LOAD_RISC;
		cb->mbox_in[1] = cb->risc_code_addr;
		cb->mbox_in[2] = VA_TO_IO32(cb->risc_code) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(cb->risc_code) & 0xFFFF;
		cb->mbox_in[4] = cb->risc_code_length;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		st->adap_init_state = 3;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 3:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}
		
		cb->risc_code_copy = eee_ramAlloc(cb->risc_code_length * 2 + CACHELINE_SIZE);
		cb->risc_code_copy_aligned = HAL_ALIGN_TO(cb->risc_code_copy, CACHELINE_SIZE);
		CACHE_INVALIDATE_STRUCT((address_t)cb->risc_code_copy_aligned, cb->risc_code_length * 2);

		/* Fire off ISPFC_DUMP_RISC mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_DUMP_RISC;
		cb->mbox_in[1] = cb->risc_code_addr;
		cb->mbox_in[2] = VA_TO_IO32(cb->risc_code_copy_aligned) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(cb->risc_code_copy_aligned) & 0xFFFF;
		cb->mbox_in[4] = cb->risc_code_length;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		st->adap_init_state = 4;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;
		
	case 4:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			eee_ramDealloc(cb->risc_code_copy);
			cb->risc_code_copy = cb->risc_code_copy_aligned = NULL;
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		{
			int		i;
			ushort16 *	p = cb->risc_code_copy_aligned;
			ushort16	checksum = 0;
			
			for (i = 0; i < cb->risc_code_length; ++i)
				checksum += *p++;
			FC_DBG(2, ("%s: ispfc_init_adap_1: Copy firmware address %p, size 0x%x, checksum 0x%x\n", 
				   cb->debug_name, cb->risc_code_copy_aligned, cb->risc_code_length * 2, checksum));
		}
		
		/* Compare buffers */
		if (memcmp((char *)cb->risc_code, (char *)cb->risc_code_copy_aligned, cb->risc_code_length * 2) != 0) {
			FC_DBG(0, ("%s: ispfc_init_adap_1: Firmware miscompare.\n", cb->debug_name));
			eee_ramDealloc(cb->risc_code_copy);
			cb->risc_code_copy = cb->risc_code_copy_aligned = NULL;
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}
		eee_ramDealloc(cb->risc_code_copy);
		cb->risc_code_copy = cb->risc_code_copy_aligned = NULL;
		st->adap_init_state = 5;
		goto next_state;

	case 5:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		/* Fire off ISPFC_VERIFY_CKSUM mailbox command */
		cb->mbox_in_valid_count = 2;
		cb->mbox_in[0] = ISPFC_VERIFY_CKSUM;
		cb->mbox_in[1] = cb->risc_code_addr;
		st->adap_init_state = 6;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 6:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		/* Fire off ISPFC_EXEC_FIRM mailbox command */
		cb->mbox_in_valid_count = 2;
		cb->mbox_in[0] = ISPFC_EXEC_FIRM;
		cb->mbox_in[1] = cb->risc_code_addr;
		st->adap_init_state = 7;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 7:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		/* 
		 * Fire off SET_ADDITION_FW_OPTIONS to enable
		 * reception of LIP(F8) and disable auto-switching
		 * from LIP(FF) to LIP(F7) 
		 */
		cb->mbox_in_valid_count = 4;
		cb->mbox_in[0] = ISPFC_SET_FW_OPTIONS;
		cb->mbox_in[1] = FO_NO_GPIO_CONTROL | FO_LIP_F8_AEN | FO_NO_LIP_F7_SWITCHING | FO_LIP_ERROR_AEN;
		cb->mbox_in[2] = 0;
		cb->mbox_in[3] = 0;
		st->adap_init_state = 8;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 8:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		ispfc_led_control(cb, ISPFC_LED_CONTROL_INIT);

		CACHE_FLUSH_STRUCT((address_t)cb->fw_init_cb, sizeof(ispfc_init_fw_cb_t));

		/* Fire off ISPFC_INITIALIZE_FIRM mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_INITIALIZE_FW;
		cb->mbox_in[1] = 0;
		cb->mbox_in[2] = VA_TO_IO32(cb->fw_init_cb) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(cb->fw_init_cb) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		st->adap_init_state = 9;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 9:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		cb->fw_init_cb_copy = eee_ramAlloc(sizeof(ispfc_init_fw_cb_t) + CACHELINE_SIZE);
		cb->fw_init_cb_copy_aligned = HAL_ALIGN_TO(cb->fw_init_cb_copy, CACHELINE_SIZE);
		CACHE_INVALIDATE_STRUCT((address_t)cb->fw_init_cb_copy_aligned, sizeof(ispfc_init_fw_cb_t));

		/* Fire off ISPFC_GET_ICB mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_GET_ICB;
		cb->mbox_in[1] = 0;
		cb->mbox_in[2] = VA_TO_IO32(cb->fw_init_cb_copy_aligned) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(cb->fw_init_cb_copy_aligned) & 0xFFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		st->adap_init_state = 10;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 10:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			eee_ramDealloc(cb->fw_init_cb_copy);
			cb->fw_init_cb_copy = cb->fw_init_cb_copy_aligned = NULL;
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		/* Compare buffers */
		if (memcmp((char *)cb->fw_init_cb, (char *)cb->fw_init_cb_copy_aligned, sizeof(ispfc_init_fw_cb_t)) != 0) {
			FC_DBG(0, ("%s: ispfc_init_adap_1: Init Block miscompare.\n", cb->debug_name));
			eee_ramDealloc(cb->fw_init_cb_copy);
			cb->fw_init_cb_copy = cb->fw_init_cb_copy_aligned = NULL;
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}
		eee_ramDealloc(cb->fw_init_cb_copy);
		cb->fw_init_cb_copy = cb->fw_init_cb_copy_aligned = NULL;
		st->adap_init_state = 11;
		goto next_state;

	case 11:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			--st->adap_init_retry;
			st->adap_init_state = 0;
			goto next_state;
		}

		if ((cb->link_state & ISPFC_CLS_MODE_MASK) == ISPFC_CLS_MODE_DIAG_LOOPBACK_INTERNAL) {
			st->adap_diag_count = 0;
			st->adap_diag_transfer_length = DIAG_TRANSFER_LENGTH;
			st->adap_diag_buf1 = eee_ramAlloc(st->adap_diag_transfer_length + 64);
			st->adap_diag_buf2 = eee_ramAlloc(st->adap_diag_transfer_length + 64);
			st->adap_diag_receive_buffer = HAL_ALIGN_TO(st->adap_diag_buf1, 64);
			st->adap_diag_transmit_buffer = HAL_ALIGN_TO(st->adap_diag_buf2, 64);

			printf("Starting post-SERDES internal loop-back test: Data size %d ....\n",
			       st->adap_diag_transfer_length);

			memset(st->adap_diag_transmit_buffer, st->adap_diag_count, st->adap_diag_transfer_length);

			CACHE_FLUSH_STRUCT((address_t)st->adap_diag_transmit_buffer, st->adap_diag_transfer_length);
			CACHE_INVALIDATE_STRUCT((address_t)st->adap_diag_receive_buffer, st->adap_diag_transfer_length);

			/* Fire off ISPFC_DIAG_LOOP_BACK mailbox command */
			cb->mbox_in_valid_count = 20;
			cb->mbox_in[0] = ISPFC_DIAG_LOOP_BACK;
			cb->mbox_in[1] = 0x00B1;
			cb->mbox_in[2] = 0;
			cb->mbox_in[3] = 0;
			cb->mbox_in[10] = st->adap_diag_transfer_length & 0xFFFF;
			cb->mbox_in[11] = (st->adap_diag_transfer_length >> 16) & 0xFFFF;
			cb->mbox_in[12] = 0;
			cb->mbox_in[13] = 0;
			cb->mbox_in[14] = VA_TO_IO32(st->adap_diag_transmit_buffer) & 0xFFFF;
			cb->mbox_in[15] = (VA_TO_IO32(st->adap_diag_transmit_buffer) >> 16) & 0xFFFF;
			cb->mbox_in[16] = VA_TO_IO32(st->adap_diag_receive_buffer) & 0xFFFF;
			cb->mbox_in[17] = (VA_TO_IO32(st->adap_diag_receive_buffer) >> 16) & 0xFFFF;
			cb->mbox_in[18] = 1;
			cb->mbox_in[19] = 0;
			st->adap_init_state = 20;
			ispfc_mbox_start(cb, ispfc_init_adap_1, st);
			qdvt->qlogic_sent++;
			qdvt->qlogic_bytes_sent+=(st->adap_diag_transfer_length);
			return;
		}

		/* Wait for link up or timeout - register AEN handlers and start a timeout timer */
		ispfc_register_aen_handler(cb, ISPFC_CS_LIP_OCCURRED, (void *)ispfc_init_adap_1, st, ISPFC_CS_LIP_OCCURRED);
		ispfc_register_aen_handler(cb, ISPFC_CS_LIP_RESET, (void *)ispfc_init_adap_1, st, ISPFC_CS_LIP_RESET);
		ispfc_register_aen_handler(cb, ISPFC_CS_LIP_F8_RECVD, (void *)ispfc_init_adap_1, st, ISPFC_CS_LIP_F8_RECVD);
		ispfc_register_aen_handler(cb, ISPFC_CS_PNT_TO_PNT_CONNECTION, (void *)ispfc_init_adap_1, st, ISPFC_CS_PNT_TO_PNT_CONNECTION);
		ispfc_register_aen_handler(cb, ISPFC_CS_LOOP_UP, (void *)ispfc_init_adap_1, st, ISPFC_CS_LOOP_UP);

		st->adap_init_timer_handle = eee_timerCreate(cb->lip_timeout, EEE_TIMER_ONCE);
		VERIFY(st->adap_init_timer_handle >= 0);
		eee_timerStart(st->adap_init_timer_handle, (void *)ispfc_init_adap_link_up_timeout, cb, st, NULL, NULL);
		st->adap_init_last_link_event = 0;
		st->adap_init_state = 12;
                return;
		
	case 12:
#if defined(ISPFC_SIM)
#ifdef ISPFC_FABRIC_SIM
		st->adap_init_last_link_event = ISPFC_CS_PNT_TO_PNT_CONNECTION;
#endif
		status = ISPFC_CS_LOOP_UP;
#endif /* defined(ISPFC_SIM) */

		if (status == NFX_ERR) {
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LIP_OCCURRED);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LIP_RESET);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LIP_F8_RECVD);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_PNT_TO_PNT_CONNECTION);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LOOP_UP);

			/* Fire off LIP-RESET mailbox command */
			cb->mbox_in_valid_count = 3;
			cb->mbox_in[0] = ISPFC_LIP_RESET;
			cb->mbox_in[1] = 0xFFFF;
			cb->mbox_in[2] = 0;
			st->adap_init_state = 0;
			--st->adap_init_retry;
			ispfc_mbox_start(cb, ispfc_init_adap_1, st);
			return;
		}
		else 
		if (status != ISPFC_CS_LOOP_UP) {
			st->adap_init_last_link_event = status;
			return;
		}
		else {
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LIP_OCCURRED);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LIP_RESET);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LIP_F8_RECVD);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_PNT_TO_PNT_CONNECTION);
			ispfc_deregister_aen_handler(cb, ISPFC_CS_LOOP_UP);

#if !defined(ISPFC_SIM)	/* timer stopped/destroyed by ispfc_init_adap_link_up_timeout */
			rc = eee_timerStop(st->adap_init_timer_handle);
			VERIFY(rc == NFX_OK);
			rc = eee_timerDestroy(st->adap_init_timer_handle);
			VERIFY(rc == NFX_OK);
			st->adap_init_timer_handle = 0;
#endif /* !defined(ISPFC_SIM) */
			if (st->adap_init_last_link_event == ISPFC_CS_PNT_TO_PNT_CONNECTION) {
                                FC_DBG(2, ("%s: Point-to-Point connection established.\n", cb->debug_name));
			}
			else {
                                FC_DBG(2, ("%s: Loop connection established.\n", cb->debug_name));
			}

			if ((cb->link_state & ISPFC_CLS_MODE_MASK) == ISPFC_CLS_MODE_DIAG_LOOPBACK_EXTERNAL) {
				st->adap_diag_count = 0;
				st->adap_diag_transfer_length = DIAG_TRANSFER_LENGTH;
				st->adap_diag_buf1 = eee_ramAlloc(st->adap_diag_transfer_length + 64);
				st->adap_diag_buf2 = eee_ramAlloc(st->adap_diag_transfer_length + 64);
				st->adap_diag_receive_buffer = HAL_ALIGN_TO(st->adap_diag_buf1, 64);
				st->adap_diag_transmit_buffer = HAL_ALIGN_TO(st->adap_diag_buf2, 64);

				printf("Starting post-SERDES external loop-back test: Data size %d ....\n",
				       st->adap_diag_transfer_length);

				memset(st->adap_diag_transmit_buffer, st->adap_diag_count, st->adap_diag_transfer_length);

				CACHE_FLUSH_STRUCT((address_t)st->adap_diag_transmit_buffer, st->adap_diag_transfer_length);
				CACHE_INVALIDATE_STRUCT((address_t)st->adap_diag_receive_buffer, st->adap_diag_transfer_length);

				/* Fire off ISPFC_DIAG_LOOP_BACK mailbox command */
				cb->mbox_in_valid_count = 20;
				cb->mbox_in[0] = ISPFC_DIAG_LOOP_BACK;
				cb->mbox_in[1] = 0x00B2; 
				cb->mbox_in[2] = 0;
				cb->mbox_in[3] = 0;
				cb->mbox_in[10] = st->adap_diag_transfer_length & 0xFFFF;
				cb->mbox_in[11] = (st->adap_diag_transfer_length >> 16) & 0xFFFF;
				cb->mbox_in[12] = 0;
				cb->mbox_in[13] = 0;
				cb->mbox_in[14] = VA_TO_IO32(st->adap_diag_transmit_buffer) & 0xFFFF;
				cb->mbox_in[15] = (VA_TO_IO32(st->adap_diag_transmit_buffer) >> 16) & 0xFFFF;
				cb->mbox_in[16] = VA_TO_IO32(st->adap_diag_receive_buffer) & 0xFFFF;
				cb->mbox_in[17] = (VA_TO_IO32(st->adap_diag_receive_buffer) >> 16) & 0xFFFF;
				cb->mbox_in[18] = 1;
				cb->mbox_in[19] = 0;
				st->adap_init_state = 20;
				ispfc_mbox_start(cb, ispfc_init_adap_1, st);
				qdvt->qlogic_sent++;
				qdvt->qlogic_bytes_sent+=(st->adap_diag_transfer_length);
				return;
			}
			else {
				/* Successfully initialized link */
				(*st->adap_init_callback)(cb, st->adap_init_cookie, ADAP_INIT_STATUS_SUCCESS);
				return;
			}
		}

	case 20:
		st->adap_diag_count++;

		qdvt->qlogic_rcvd ++;
		qdvt->qlogic_bytes_rcvd +=(st->adap_diag_transfer_length);
		if ((st->adap_diag_count % 100) == 0)
			printf("Iteration %d, "
			       "CRC %d, "
			       "Disparity %d, "
			       "Frame Length %d, "
			       "Miscompares %d\n",
			       st->adap_diag_count,
			       st->adap_diag_crc_count,
			       st->adap_diag_disparity_count,
			       st->adap_diag_frame_len_count,
			       st->adap_diag_miscompares);
				
		if (status == MBOX_CMD_STATUS_TIMEOUT) {
			printf("Resetting adapter ...\n");
			eee_ramDealloc(st->adap_diag_buf1);
			eee_ramDealloc(st->adap_diag_buf2);
			st->adap_diag_buf1 = st->adap_diag_receive_buffer = NULL;
			st->adap_diag_buf2 = st->adap_diag_transmit_buffer = NULL;
			st->adap_init_state = 0;
			goto next_state;
		}
		else
		if (status == MBOX_CMD_STATUS_FAILED) {
			if (cb->mbox_out[0] == 0x400C) {
				st->adap_diag_crc_count += cb->mbox_out[1];
				st->adap_diag_disparity_count += cb->mbox_out[2];
				st->adap_diag_frame_len_count += cb->mbox_out[3];
			}
		}
		else
		if (memcmp(st->adap_diag_transmit_buffer, 
			   st->adap_diag_receive_buffer, 
			   st->adap_diag_transfer_length)) {
			printf("Miscompare\n");
			++st->adap_diag_miscompares;
			qdvt->qlogic_dataErr++;
		}

		if ((cb->link_state & ISPFC_CLS_MODE_MASK) != ISPFC_CLS_MODE_DIAG_LOOPBACK_INTERNAL &&
		    (cb->link_state & ISPFC_CLS_MODE_MASK) != ISPFC_CLS_MODE_DIAG_LOOPBACK_EXTERNAL) {
			printf("Terminating loopback test\n");
			eee_ramDealloc(st->adap_diag_buf1);
			eee_ramDealloc(st->adap_diag_buf2);
			st->adap_diag_buf1 = st->adap_diag_receive_buffer = NULL;
			st->adap_diag_buf2 = st->adap_diag_transmit_buffer = NULL;
			st->adap_init_state = 0;
			qdvt->qlogic_crcErr = st->adap_diag_crc_count;
			qdvt->qlogic_dispCnt = st->adap_diag_disparity_count;
			qdvt->qlogic_FrameLenCnt = st->adap_diag_frame_len_count;
			goto next_state;
		}

		memset(st->adap_diag_transmit_buffer, st->adap_diag_count, st->adap_diag_transfer_length);

		CACHE_FLUSH_STRUCT((address_t)st->adap_diag_transmit_buffer, st->adap_diag_transfer_length);
		CACHE_INVALIDATE_STRUCT((address_t)st->adap_diag_receive_buffer, st->adap_diag_transfer_length);

		/* Fire off ISPFC_DIAG_LOOP_BACK mailbox command */
		cb->mbox_in_valid_count = 20;
		cb->mbox_in[0] = ISPFC_DIAG_LOOP_BACK;
		if ((cb->link_state & ISPFC_CLS_MODE_MASK) == ISPFC_CLS_MODE_DIAG_LOOPBACK_EXTERNAL)
			cb->mbox_in[1] = 0x00B2;
		else
			cb->mbox_in[1] = 0x00B1;
		cb->mbox_in[2] = 0;
		cb->mbox_in[3] = 0;
		cb->mbox_in[10] = st->adap_diag_transfer_length & 0xFFFF;
		cb->mbox_in[11] = (st->adap_diag_transfer_length >> 16) & 0xFFFF;
		cb->mbox_in[12] = 0;
		cb->mbox_in[13] = 0;
		cb->mbox_in[14] = VA_TO_IO32(st->adap_diag_transmit_buffer) & 0xFFFF;
		cb->mbox_in[15] = (VA_TO_IO32(st->adap_diag_transmit_buffer) >> 16) & 0xFFFF;
		cb->mbox_in[16] = VA_TO_IO32(st->adap_diag_receive_buffer) & 0xFFFF;
		cb->mbox_in[17] = (VA_TO_IO32(st->adap_diag_receive_buffer) >> 16) & 0xFFFF;
		cb->mbox_in[18] = 1;
		cb->mbox_in[19] = 0;
		st->adap_init_state = 20;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		qdvt->qlogic_sent++;
		qdvt->qlogic_bytes_sent+=(st->adap_diag_transfer_length);
		return;

	case 21:
		st->adap_diag_count = 0;
		st->adap_diag_transfer_length = 0x800 * 2;
		st->adap_diag_buf1 = eee_ramAlloc(st->adap_diag_transfer_length + 64);
		st->adap_diag_buf2 = eee_ramAlloc(st->adap_diag_transfer_length + 64);
		st->adap_diag_receive_buffer = HAL_ALIGN_TO(st->adap_diag_buf1, 64);
		st->adap_diag_transmit_buffer = HAL_ALIGN_TO(st->adap_diag_buf2, 64);

		printf("Starting SRAM read/write test: Transfer size %d ....\n",
		       st->adap_diag_transfer_length);

		memset(st->adap_diag_transmit_buffer, st->adap_diag_count, st->adap_diag_transfer_length);

		CACHE_FLUSH_STRUCT((address_t)st->adap_diag_transmit_buffer, st->adap_diag_transfer_length);

		st->adap_init_generic_counter = 0x800;

		/* Fire off the first in a sequence of ISPFC_LOAD_SRAM_EXT mailbox commands */
		cb->mbox_in_valid_count = 9;
		cb->mbox_in[0] = ISPFC_LOAD_SRAM_EXT;
		cb->mbox_in[1] = (st->adap_init_generic_counter & 0xFFFF);
		cb->mbox_in[2] = (VA_TO_IO32(st->adap_diag_transmit_buffer) >> 16) & 0xFFFF;
		cb->mbox_in[3] = VA_TO_IO32(st->adap_diag_transmit_buffer) & 0xFFFF;
		cb->mbox_in[4] = st->adap_diag_transfer_length / 2;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		cb->mbox_in[8] = st->adap_init_generic_counter >> 16;
		st->adap_init_state = 22;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 22:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			eee_ramDealloc(st->adap_diag_buf1);
			eee_ramDealloc(st->adap_diag_buf2);
			st->adap_diag_buf1 = st->adap_diag_receive_buffer = NULL;
			st->adap_diag_buf2 = st->adap_diag_transmit_buffer = NULL;
			ispfc_trigger_pci_analyzer();
			ispfc_dump_registers(cb);
			st->adap_init_state = 0;
			goto next_state;
		}

		if ((cb->link_state & ISPFC_CLS_MODE_MASK) != ISPFC_CLS_MODE_DIAG_SRAM) {
			printf("Terminating SRAM read/write test\n");
			eee_ramDealloc(st->adap_diag_buf1);
			eee_ramDealloc(st->adap_diag_buf2);
			st->adap_diag_buf1 = st->adap_diag_receive_buffer = NULL;
			st->adap_diag_buf2 = st->adap_diag_transmit_buffer = NULL;
			st->adap_init_state = 0;
			goto next_state;
		}

		CACHE_INVALIDATE_STRUCT((address_t)st->adap_diag_receive_buffer, st->adap_diag_transfer_length);

		/* Fire off  ISPFC_DUMP_SRAM_EXT mailbox commands */
		cb->mbox_in_valid_count = 9;
		cb->mbox_in[0] = ISPFC_DUMP_SRAM_EXT;
		cb->mbox_in[1] = (st->adap_init_generic_counter & 0xFFFF);
		cb->mbox_in[2] = (VA_TO_IO32(st->adap_diag_receive_buffer) >> 16) & 0xFFFF;
		cb->mbox_in[3] = VA_TO_IO32(st->adap_diag_receive_buffer) & 0xFFFF;
		cb->mbox_in[4] = st->adap_diag_transfer_length / 2;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		cb->mbox_in[8] = st->adap_init_generic_counter >> 16;
		st->adap_init_state = 23;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;

	case 23:
		if (status != MBOX_CMD_STATUS_SUCCESS) {
			eee_ramDealloc(st->adap_diag_buf1);
			eee_ramDealloc(st->adap_diag_buf2);
			st->adap_diag_buf1 = st->adap_diag_receive_buffer = NULL;
			st->adap_diag_buf2 = st->adap_diag_transmit_buffer = NULL;
			ispfc_trigger_pci_analyzer();
			ispfc_dump_registers(cb);
			st->adap_init_state = 0;
			goto next_state;
		}

		/* Verify receive buffer */
		if (memcmp(st->adap_diag_receive_buffer, st->adap_diag_transmit_buffer, st->adap_diag_transfer_length) != 0) {
			printf("SRAM miscompare at offset 0x%x.\n", st->adap_init_generic_counter);
			eee_ramDealloc(st->adap_diag_buf1);
			eee_ramDealloc(st->adap_diag_buf2);
			st->adap_diag_buf1 = st->adap_diag_receive_buffer = NULL;
			st->adap_diag_buf2 = st->adap_diag_transmit_buffer = NULL;
			ispfc_trigger_pci_analyzer();
			st->adap_init_state = 0;
			goto next_state;
		}

		CACHE_FLUSH_STRUCT((address_t)st->adap_diag_transmit_buffer, st->adap_diag_transfer_length);

		st->adap_init_generic_counter += 0x0800;
		if (st->adap_init_generic_counter < 0x80000) { /* 512K 16-bit words */
			/* Fire off the next in a sequence of ISPFC_LOAD_SRAM_EXT mailbox commands */
			cb->mbox_in_valid_count = 9;
			cb->mbox_in[0] = ISPFC_LOAD_SRAM_EXT;
			cb->mbox_in[1] = (st->adap_init_generic_counter & 0xFFFF);
			cb->mbox_in[2] = (VA_TO_IO32(st->adap_diag_transmit_buffer) >> 16) & 0xFFFF;
			cb->mbox_in[3] = VA_TO_IO32(st->adap_diag_transmit_buffer) & 0xFFFF;
			cb->mbox_in[4] = st->adap_diag_transfer_length / 2;
			cb->mbox_in[5] = 0;
			cb->mbox_in[6] = 0;
			cb->mbox_in[7] = 0;
			cb->mbox_in[8] = st->adap_init_generic_counter >> 16;
			st->adap_init_state = 22;
			ispfc_mbox_start(cb, ispfc_init_adap_1, st);
			return;
		}

		/* Done with one iteration */
		printf("SRAM TEST iteration %d%c", st->adap_diag_count++, '\015');

		memset(st->adap_diag_transmit_buffer, st->adap_diag_count, st->adap_diag_transfer_length);

		CACHE_FLUSH_STRUCT((address_t)st->adap_diag_transmit_buffer, st->adap_diag_transfer_length);

		st->adap_init_generic_counter = 0x800;

		/* Fire off the first in a sequence of ISPFC_LOAD_SRAM_EXT mailbox commands */
		cb->mbox_in_valid_count = 9;
		cb->mbox_in[0] = ISPFC_LOAD_SRAM_EXT;
		cb->mbox_in[1] = (st->adap_init_generic_counter & 0xFFFF);
		cb->mbox_in[2] = (VA_TO_IO32(st->adap_diag_transmit_buffer) >> 16) & 0xFFFF;
		cb->mbox_in[3] = VA_TO_IO32(st->adap_diag_transmit_buffer) & 0xFFFF;
		cb->mbox_in[4] = st->adap_diag_transfer_length / 2;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		cb->mbox_in[8] = st->adap_init_generic_counter >> 16;
		st->adap_init_state = 22;
		ispfc_mbox_start(cb, ispfc_init_adap_1, st);
		return;
	}
}

void 
ispfc_init_adap_link_up_timeout(uint32 timer_handle, ispfc_cb_t *cb, ispfc_adap_init_state_t *st)
{ 
	int32 rc;

	/* Destroy timer */
	VERIFY(timer_handle == st->adap_init_timer_handle);
	rc = eee_timerDestroy(st->adap_init_timer_handle);
	VERIFY(rc == NFX_OK);

	/* Call back into state machine */
	st->adap_init_timer_handle = 0;
	ispfc_init_adap_1(cb, st, NFX_ERR);
}

void 
ispfc_cleanup_link_state(ispfc_cb_t *cb)
{
	ispfc_dev_t *		dev;
	scsi_desc_t *		sd;
	scsi_task_mgmt_t *	tmo;
	ct_desc_t *		ctd;
	int			rc;

	/* Kill any in progress mailbox command */
	if (ispfc_mbox_is_busy(cb))
		ispfc_mbox_abort(cb);

	/* We might have been waiting for link up; cleanup .... */
	{
		ispfc_adap_init_state_t	* st = cb->adap_init_state;
		if (st->adap_init_timer_handle) {
			rc = eee_timerStop(st->adap_init_timer_handle);
			VERIFY(rc == NFX_OK);
			rc = eee_timerDestroy(st->adap_init_timer_handle);
			VERIFY(rc == NFX_OK);
			st->adap_init_timer_handle = 0;
			ispfc_reset_aen_handlers(cb);
		}
		if(st->adap_diag_buf1)
			eee_ramDealloc(st->adap_diag_buf1);
		if(st->adap_diag_buf2)
			eee_ramDealloc(st->adap_diag_buf2);
		st->adap_diag_buf1 = st->adap_diag_receive_buffer = NULL;
		st->adap_diag_buf2 = st->adap_diag_transmit_buffer = NULL;
	}

	/* Move all device states to "resetting" */
	FOR_ALL_DEVS({
		SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_REQUESTED);
	});

	/* Flush all device scsi descriptor and TMO busy queues */
	FOR_ALL_DEVS({
		while ((sd = dev->scsi_desc_busy) != NULL) {
			int retry;
			
			retry = 0;

			switch(sd->state) {
			case SCSI_DESC_STATE_TIMED_OUT:
				/* Command timed out; retry up to retryCnt limit */
				SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_EXCESSIVE_TIMEOUTS);
				++sd->current_to_retry_count;
				--sd->retryCnt;
				++retry;
				break;

			case SCSI_DESC_STATE_BUSY:
				/* Command is a victim; retry but don't penalize retry counter */
				++retry;
				break;
				
			case SCSI_DESC_STATE_ABORTED:
				/* Command was aborted by the caller; return it. */
				SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_ABORTED);
				break;
				
			default:
				VERIFY(0);
			}
			ispfc_dequeue_desc_off_dev_busy_queue(cb, dev, sd);
			ispfc_dispose_of_scsi_desc(cb, dev, sd, retry);
		} /* while (sd) */

		/*
		 * Any pending TMO was already "killed" when
		 * we aborted the pending mailbox command.
		 * All that remains is to clean up.
		 */
		if ((tmo = dev->scsi_task_mgmt_busy)) {
			tmo->status = NFX_ERR;
			(*tmo->callback)(tmo);
			dev->scsi_task_mgmt_busy = NULL;
		}
	});


	/* Flush busy CT queue */
	while ((ctd = cb->ct_desc_busy) != NULL) {
		ispfc_dequeue_desc_off_ct_busy_queue(cb, ctd);
		ispfc_dispose_of_ct_desc(cb, ctd, 1);
	} /* while (ctd) */

	/*
	 * Just "drop" all queued "internal" CT
	 * descriptors. Since "internal" CT descriptors are
	 * part of statically allocated structures, which are
	 * about to get zeroed, leaving these descriptors on
	 * the queue cause problems. This is needed for direct 
	 * REESTABLISHING-->RESET_REQUESTED link state transitions.
	 * Yeah, this is a bit of an ugly hack .....
	 */
	ispfc_purge_internal_desc_off_ct_waiting_queue(cb);
}

int
ispfc_soft_reset_adapter(ispfc_cb_t *cb)
{
#if !defined(ISPFC_SIM)
	int i;

	/* Hit the soft reset */
	ISPFC_REG_WRITE(cb, ISPFC_REG_CONSTAT, CONSTAT_SOFT_RESET);
	eee_delay(100);

	/* Clear the mailbox semaphore */
	ISPFC_REG_WRITE(cb, ISPFC_REG_SEMAPHORE, 0);
	eee_delay(100);

	ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_RELEASE);
        for (i = 10000; ; ) {
		eee_delay(100);
                if (!(ISPFC_REG_READ(cb, ISPFC_REG_HCCR) & (HCCR_PAUSE | HCCR_RESET)))
			break;
                if (i-- == 0) {
                        FC_DBG(0, ("%s: Restart failed.\n", cb->debug_name));
                        return(NFX_ERR);
                }
        }
	
	/* Poll looking for zero status */
        while (ISPFC_REG_READ(cb, ISPFC_REG_MBOX0))
		;

	ispfc_led_control(cb, ISPFC_LED_CONTROL_INIT);

        cb->risc_rom_rev = ISPFC_REG_READ(cb, ISPFC_REG_MBOX7);

#endif /* !defined(ISPFC_SIM) */

	cb->request_in = 0;
	cb->request_out = 0;
	cb->response_in = 0;
	cb->prev_response_in = 0;
	cb->response_out = 0;
	ispfc_iocb_handle_init(cb);

        return(NFX_OK);
}

/*******************************************************************************
 * ispfc_init_conn -
 *
 *   Primary entry point into connection init state machine. State machine
 *   is initialized and the first state is entered.  
 ******************************************************************************/
void
ispfc_init_conn(ispfc_cb_t *cb, void (*conn_init_callback)(ispfc_cb_t *, void *, int32), void *conn_init_cookie)
{
	ispfc_conn_init_state_t	*st = cb->conn_init_state;

	VERIFY(st);

	st->conn_init_retry = ISPFC_INIT_CONN_RETRY;
	st->conn_init_state = 0;
	st->conn_init_callback = conn_init_callback;
	st->conn_init_cookie = conn_init_cookie;

	ispfc_init_conn_1(cb, st, NFX_OK);
}

/*******************************************************************************
 * ispfc_init_conn_1 -
 * 
 *  Runs the connection init state machine.
 *	   State 0: 
 *		 o mailbox(GET_FW_STATE)
 *		 --> state 1
 *	   State 1:
 *		 TIMEOUT --> return(FAILURE)
 *		 FAILURE --> state 0 (retry)
 *		 READY   --> state 2
 *		 o delay 1 second
 *		 --> state 0 (retry)
 *	   State 2:
 *		 o mailbox(GET_PORT_ID)
 *		 --> state 3
 *	   State 3:
 *		 TIMEOUT --> return(FAILURE)
 *		 FAILURE --> state 0 (retry)
 *		 if (!fabric)
 *		    return(SUCCESS)
 *		 o mailbox(GET_PORT_NAME)
 *		 --> state 4
 *	   State 4:
 *		 TIMEOUT --> return(FAILURE)
 *		 FAILURE --> state 0 (retry)
 *		 o mailbox(GET_RETRY_COUNT)
 *		 --> state 5
 *	   State 5:
 *		 TIMEOUT --> return(FAILURE)
 *		 FAILURE --> state 0 (retry)
 *		 o mailbox(GET_DATA_RATE)
 *		 --> state 6
 *	   State 6:
 *		 TIMEOUT --> return(FAILURE)
 *		 FAILURE --> state 0 (retry)
 *		 SUCCESS --> return(SUCCESS)
 ******************************************************************************/
void
ispfc_init_conn_1(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_conn_init_state_t *st = cookie;
	channel_name_t		 ch;
	int			 rc;

 next_state:

	FC_DBG(2, ("%s: ispfc_init_conn_1: state = %d, status = 0x%x\n", 
		   cb->debug_name, st->conn_init_state, status));

	switch(st->conn_init_state) {
	case 0:
		/* Out of retries? Return failure */
		if (st->conn_init_retry == 0) {
			/* Failure */
			(*st->conn_init_callback)(cb, st->conn_init_cookie, CONN_INIT_STATUS_FAILED);
			return;
		}

		/* Fire off ISPFC_GET_FW_STATE mailbox command */ 
		cb->mbox_in_valid_count = 1;
		cb->mbox_out_valid_count = 2;
		cb->mbox_in[0] = ISPFC_GET_FW_STATE;
		st->conn_init_state = 1;
		ispfc_mbox_start(cb, ispfc_init_conn_1, st);
		return;

	case 1:
		if (status == MBOX_CMD_STATUS_TIMEOUT) {
			(*st->conn_init_callback)(cb, st->conn_init_cookie, CONN_INIT_STATUS_FAILED);
			return;
		}
		else
		if (status == MBOX_CMD_STATUS_FAILED) {
			--st->conn_init_retry;
			st->conn_init_state = 0;
			goto next_state;
		}
		else {
#if defined(ISPFC_SIM)
		  cb->mbox_out[1] = ISPFC_STATE_READY;
#endif
			/* If not ready, delay for 1 second and try again */
			if (cb->mbox_out[1] == ISPFC_STATE_READY) {
				st->conn_init_state = 2;
				goto next_state;
			}
			st->conn_init_timer_handle = eee_timerCreate(1000, EEE_TIMER_ONCE);
			VERIFY(st->conn_init_timer_handle >= 0);
			eee_timerStart(st->conn_init_timer_handle, (void *)ispfc_init_conn_fw_ready_delay, cb, st, NULL, NULL);
			--st->conn_init_retry;
			st->conn_init_state = 0;
			return;
		}
		break;

	case 2:
		/* Fire off ISPFC_GET_PORT_ID mailbox command */ 
		cb->mbox_in_valid_count = 1;
		cb->mbox_out_valid_count = 7;
		cb->mbox_in[0] = ISPFC_GET_PORT_ID;
		st->conn_init_state = 3;
		ispfc_mbox_start(cb, ispfc_init_conn_1, st);
		return;
		
	case 3:
		if (status == MBOX_CMD_STATUS_TIMEOUT) {
			(*st->conn_init_callback)(cb, st->conn_init_cookie, CONN_INIT_STATUS_FAILED);
			return;
		}
		else
		if (status == MBOX_CMD_STATUS_FAILED) {
			--st->conn_init_retry;
			st->conn_init_state = 0;
			goto next_state;
		}

#if defined(ISPFC_SIM)
#ifdef ISPFC_FABRIC_SIM
		cb->mbox_out[6] = 3;
#endif
#endif /* defined(ISPFC_SIM) */

		/* Loop ID */
		cb->host_loop_id = cb->mbox_out[1];

		/* FC port ID */
		cb->host_port_id = 
			((uint32)cb->mbox_out[3] << 16) | cb->mbox_out[2];

		/* Connection type */
		switch(cb->mbox_out[6]) {
                case 0:                         /* Loop (no fabric) */
                        cb->interface_type = HAL_IF_FC_L_PORT;
			FC_DBG(0, ("%s: Private loop connection established: host_loop_id %d\n", 
				   cb->debug_name, cb->host_loop_id));
                        break;
                case 1:                         /* FL_port in loop */
                        cb->interface_type = HAL_IF_FC_FL_PORT;
			FC_DBG(0, ("%s: Public loop connection established: host_loop_id %d, host_port_id 0x%x\n", 
				   cb->debug_name, cb->host_loop_id, cb->host_port_id));
                        break;
                case 2:                         /* N_port (point-to-point) */
                        cb->interface_type = HAL_IF_FC_N_PORT;
			FC_DBG(0, ("%s: Private point-to-point connection established: host_port_id 0x%x\n", 
				   cb->debug_name, cb->host_port_id));
                        break;
                case 3:                         /* F_port (point-to-point) */
                        cb->interface_type = HAL_IF_FC_F_PORT;
			FC_DBG(0, ("%s: Public point-to-point connection established: host_port_id 0x%x\n", 
				   cb->debug_name, cb->host_port_id));
                        break;
                default:
			VERIFY(0);
                }

		/* Add a channel entry for the host adapter -- don't know peer WWN yet. */
		bzero(&ch, sizeof(channel_name_t));
		ch.channel_wwn.u.wwn = cb->adapter_wwn.u.wwn;
		rc = ispfc_channel_lookup(&ch);
		if (rc != NFX_OK ||
		    ch.channel_interface_type != cb->interface_type) {
			if (rc == NFX_OK) {
				rc = ispfc_channel_remove(&ch);
				VERIFY(rc == NFX_OK);
			}
			bzero(&ch, sizeof(channel_name_t));
			sprintf(ch.channel_name, "%d.%d:",
				hal_dt_entry_get()->hal_bus_id, 
				hal_dt_entry_get()->hal_dev_id);
			ch.channel_wwn.u.wwn = cb->adapter_wwn.u.wwn;
			ch.channel_interface_type = cb->interface_type;

			if (cb->interface_type == HAL_IF_FC_F_PORT ||
			    cb->interface_type == HAL_IF_FC_FL_PORT) {
				ch.u.fc_n_or_fc_nl.channel_ha_bus_id = hal_dt_entry_get()->hal_bus_id;
				ch.u.fc_n_or_fc_nl.channel_ha_dev_id = hal_dt_entry_get()->hal_dev_id;
			}
			else {
				ch.u.fc_l.channel_ha_bus_id = hal_dt_entry_get()->hal_bus_id;
				ch.u.fc_l.channel_ha_dev_id = hal_dt_entry_get()->hal_dev_id;
			}
			rc = ispfc_channel_add(&ch);
			VERIFY(rc == NFX_OK);
		}

		/* 
		 * If fabric attached, verify successfull FLOGI --
		 * fire off a GET_PORT_NAME mailbox command, to get
		 * the WWN of the FL-port.
		 */
		if (cb->interface_type == HAL_IF_FC_FL_PORT ||
		    cb->interface_type == HAL_IF_FC_F_PORT) {
			cb->mbox_in_valid_count = 2;
			cb->mbox_in[0] = ISPFC_GET_PORT_NAME;
			cb->mbox_in[1] = ISPFC_FABRIC_PORT_PORTID << 8;
			st->conn_init_state = 4;
			ispfc_mbox_start(cb, ispfc_init_conn_1, st);
			return;
		}
		else {
			/* Fire off GET_DATA_RATE mailbox command to get link speed */
			cb->mbox_in_valid_count = 2;
			cb->mbox_out_valid_count = 3;
			cb->mbox_in[0] = ISPFC_DATA_RATE;
			cb->mbox_in[1] = 0;
			st->conn_init_state = 6;
			ispfc_mbox_start(cb, ispfc_init_conn_1, st);
			return;
		}
        
	case 4:
		if (status == MBOX_CMD_STATUS_TIMEOUT) {
                        (*st->conn_init_callback)(cb, st->conn_init_cookie, CONN_INIT_STATUS_FAILED);
                        return;
                }
                else
		if (status == MBOX_CMD_STATUS_FAILED) {
			--st->conn_init_retry;
			st->conn_init_state = 0;
			goto next_state;
		}

		/* Fire off GET_RETRY_COUNT mailbox command to get switch's RA_TOV */
		cb->mbox_in_valid_count = 1;
		cb->mbox_out_valid_count = 4;
		cb->mbox_in[0] = ISPFC_GET_RETRY_COUNT;
		st->conn_init_state = 5;
		ispfc_mbox_start(cb, ispfc_init_conn_1, st);
		return;

	case 5:
		if (status == MBOX_CMD_STATUS_TIMEOUT) {
                        (*st->conn_init_callback)(cb, st->conn_init_cookie, CONN_INIT_STATUS_FAILED);
                        return;
                }
                else
		if (status == MBOX_CMD_STATUS_FAILED) {
			--st->conn_init_retry;
			st->conn_init_state = 0;
			goto next_state;
		}

		cb->mbox_timeout = cb->mbox_out[3] * 100;
		cb->switch_ra_tov = cb->mbox_out[3] * 100 / 2;

		/* Fire off GET_DATA_RATE mailbox command to get link speed */
		cb->mbox_in_valid_count = 2;
		cb->mbox_out_valid_count = 3;
		cb->mbox_in[0] = ISPFC_DATA_RATE;
		cb->mbox_in[1] = 0;
		st->conn_init_state = 6;
		ispfc_mbox_start(cb, ispfc_init_conn_1, st);
		return;

	case 6:
		if (status == MBOX_CMD_STATUS_TIMEOUT) {
                        (*st->conn_init_callback)(cb, st->conn_init_cookie, CONN_INIT_STATUS_FAILED);
                        return;
                }
                else
		if (status == MBOX_CMD_STATUS_FAILED) {
			--st->conn_init_retry;
			st->conn_init_state = 0;
			goto next_state;
		}

		cb->link_speed = cb->mbox_out[1] == 0 ? 1 : 2;

		(*st->conn_init_callback)(cb, st->conn_init_cookie, CONN_INIT_STATUS_SUCCESS);
		return;
	}
}

void 
ispfc_init_conn_fw_ready_delay(uint32 timer_handle, ispfc_cb_t *cb, ispfc_conn_init_state_t *st)
{
	int32 rc;

	/* Destroy timer */
	VERIFY(timer_handle == st->conn_init_timer_handle);
	rc = eee_timerDestroy(st->conn_init_timer_handle);
	VERIFY(rc == NFX_OK);

	/* Call back into state machine */
	ispfc_init_conn_1(cb, st, NFX_OK);
}

/*******************************************************************************
 * ispfc_init_link -
 *
 *   Primary entry point into link init state machine. State machine
 *   is initialized and the first state is entered.  
 ******************************************************************************/
void
ispfc_init_link(ispfc_cb_t *cb, void (*link_init_callback)(ispfc_cb_t *, void *, int32), void *link_init_cookie)
{
	ispfc_link_init_state_t	*st = cb->link_init_state;

	VERIFY(st);

	/* If running diags, force into loop mode */
	if ((cb->link_state & ISPFC_CLS_MODE_MASK) != ISPFC_CLS_MODE_RUNTIME) 
		cb->link_state &= ~ISPFC_CLF_POINT_TO_POINT_OK;
	
	if (cb->link_state & ISPFC_CLF_POINT_TO_POINT_OK) 
		st->link_init_state = 5;
	else 
		st->link_init_state = 0;

	st->link_init_callback = link_init_callback;
	st->link_init_cookie = link_init_cookie;

	ispfc_init_link_1(cb, st, NFX_OK);
}

/*******************************************************************************
 * ispfc_init_link_1 -
 * 
 *  Runs the link init state machine.
 *	   if ("previously point-to-point")
 *		 --> state 5
 *	   else
 *		 --> state 0
 *
 *	   State 0: // Loop not attempted
 *		 o Attempt link init in loop mode
 *		 --> state 1
 *	   State 1: 
 *		 FAILURE --> return(FAILURE)
 *		 SUCCESS (no FL-port) --> return(SUCCESS_L-port)
 *		 o Attempt link init in p-to-p mode
 *		 --> state 2
 *	   State 2:
 *		 FAILURE --> state 3
 *		 SUCCESS --> return(SUCCESS_N-port)
 *	   State 3: // loop OK; p-to-p not successfull
 *		 o Attempt link init in loop mode
 *		 --> state 4
 *	   State 4:
 *		 FAILURE --> return(FAILURE)
 *		 SUCCESS --> return(SUCCESS_NL-port)
 * 
 *	   State 5: // Point-to-point successfull
 *		 o Attempt link init in p-to-p mode
 *		 --> state 6
 *	   State 6:
 *		 SUCCESS --> return(SUCCESS_N-port)
 *		 FAILURE --> state 0
 *
 * Note that each "step" requires an adapter init and a connection init,
 * hence the two micro-states, ispfc_init_link_1 and ispfc_init_link_2.
 ******************************************************************************/
void
ispfc_init_link_1(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_link_init_state_t *st = cookie;

 next_state:

	FC_DBG(2, ("%s: ispfc_init_link_1: state = %d, status = 0x%x\n", 
		   cb->debug_name, st->link_init_state, status));

	switch(st->link_init_state) {
	case 0:
		/* Fire off an adapter initialization in loop-only mode */
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		cb->fw_init_cb->additional_fw_options &= ~IFW_AFO_CONN_TYPE_MASK;
		cb->fw_init_cb->additional_fw_options |= IFW_AFO_LOOP_ONLY;
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		st->link_init_state = 1;
		ispfc_init_adap(cb, ispfc_init_link_2, st);
		return;

	case 1:
		if (status != CONN_INIT_STATUS_SUCCESS) {
			(*st->link_init_callback)(cb, st->link_init_cookie, LINK_INIT_STATUS_FAILED);
			return;
		}
		
		if (cb->interface_type == HAL_IF_FC_L_PORT) {
			(*st->link_init_callback)(cb, st->link_init_cookie, LINK_INIT_STATUS_SUCCESS);
			return;
		}
		
		/* Fire off an adapter initialization in point-to-point-only mode */
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		cb->fw_init_cb->additional_fw_options &= ~IFW_AFO_CONN_TYPE_MASK;
		cb->fw_init_cb->additional_fw_options |= IFW_AFO_POINT_TO_POINT_ONLY;
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		st->link_init_state = 2;
		ispfc_init_adap(cb, ispfc_init_link_2, st);
		return;

	case 2:
		if (status != CONN_INIT_STATUS_SUCCESS) {
			cb->link_state &= ~ISPFC_CLF_POINT_TO_POINT_OK;
			st->link_init_state = 3;
			goto next_state;
		}
		cb->link_state |= ISPFC_CLF_POINT_TO_POINT_OK;
		(*st->link_init_callback)(cb, st->link_init_cookie, LINK_INIT_STATUS_SUCCESS);
		return;

	case 3:
		/* Fire off an adapter initialization in loop-only mode */
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		cb->fw_init_cb->additional_fw_options &= ~IFW_AFO_CONN_TYPE_MASK;
		cb->fw_init_cb->additional_fw_options |= IFW_AFO_LOOP_ONLY;
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		st->link_init_state = 4;
		ispfc_init_adap(cb, ispfc_init_link_2, st);
		return;

	case 4:
		if (status != CONN_INIT_STATUS_SUCCESS)
			(*st->link_init_callback)(cb, st->link_init_cookie, LINK_INIT_STATUS_FAILED);
		else
			(*st->link_init_callback)(cb, st->link_init_cookie, LINK_INIT_STATUS_SUCCESS);
		return;
		
	case 5:
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		/* Fire off an adapter initialization in point-to-point-only mode */
		cb->fw_init_cb->additional_fw_options &= ~IFW_AFO_CONN_TYPE_MASK;
		cb->fw_init_cb->additional_fw_options |= IFW_AFO_POINT_TO_POINT_ONLY;
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(&cb->fw_init_cb->additional_fw_options);
#endif
		st->link_init_state = 6;
		ispfc_init_adap(cb, ispfc_init_link_2, st);
		return;

	case 6:
		if (status == CONN_INIT_STATUS_SUCCESS) {
			cb->link_state |= ISPFC_CLF_POINT_TO_POINT_OK;
			(*st->link_init_callback)(cb, st->link_init_cookie, LINK_INIT_STATUS_SUCCESS);
			return; 
		}
		else {
#if 1
			cb->link_state &= ~ISPFC_CLF_POINT_TO_POINT_OK;
			st->link_init_state = 0;
			goto next_state;
#else
			cb->link_state &= ~ISPFC_CLF_POINT_TO_POINT_OK;
			(*st->link_init_callback)(cb, st->link_init_cookie, LINK_INIT_STATUS_FAILED);
#endif
		}
	}
}

void
ispfc_init_link_2(ispfc_cb_t *cb, void *cookie, int32 status)
{
	if (status != ADAP_INIT_STATUS_SUCCESS) {
		ispfc_init_link_1(cb, cookie, CONN_INIT_STATUS_FAILED);
		return;
	}
	ispfc_init_conn(cb, ispfc_init_link_1, cookie);
}

#ifdef DEBUG
void	
ispfc_check_state_machine(ispfc_cb_t *cb, int new_state)
{
	char *state_string;

	VERIFY((new_state & ~ISPFC_CLS_STATE_MASK) == 0);
	switch(new_state) {
	case ISPFC_CLS_STATE_RESETTING:		state_string = "ISPFC_CLS_STATE_RESETTING"; break;
	case ISPFC_CLS_STATE_REESTABLISHING:	state_string = "ISPFC_CLS_STATE_REESTABLISHING"; break;
	case ISPFC_CLS_STATE_RESET_REQUESTED:	state_string = "ISPFC_CLS_STATE_RESET_REQUESTED"; break;
	case ISPFC_CLS_STATE_DOWN:		state_string = "ISPFC_CLS_STATE_DOWN"; break;
	case ISPFC_CLS_STATE_OFFLINE_REQUESTED: state_string = "ISPFC_CLS_STATE_OFFLINE_REQUESTED"; break;
	case ISPFC_CLS_STATE_OFFLINE_PENDING:	state_string = "ISPFC_CLS_STATE_OFFLINE_PENDING"; break;
	case ISPFC_CLS_STATE_OFFLINE:		state_string = "ISPFC_CLS_STATE_OFFLINE"; break;
	case ISPFC_CLS_STATE_LIP_OCCURRED:	state_string = "ISPFC_CLS_STATE_LIP_OCCURRED"; break;
	case ISPFC_CLS_STATE_LIP_GENERATED:	state_string = "ISPFC_CLS_STATE_LIP_GENERATED"; break;
	case ISPFC_CLS_STATE_NSCN_OCCURRED:	state_string = "ISPFC_CLS_STATE_NSCN_OCCURRED"; break;
	case ISPFC_CLS_STATE_NSCN_GENERATED:	state_string = "ISPFC_CLS_STATE_NSCN_GENERATED"; break;
	case ISPFC_CLS_STATE_UP:		state_string = "ISPFC_CLS_STATE_UP"; break;
	default:				VERIFY(0);
	}

	FC_DBG(1, ("%s: New state %s (0x%08x-->0x%08x)\n",
		   cb->debug_name, state_string, cb->link_state, new_state));

	/* Sanity check link state machine */
	switch(LINK_STATE(cb)) { 
	case ISPFC_CLS_STATE_UP:
		VERIFY(new_state == ISPFC_CLS_STATE_LIP_OCCURRED ||
		       new_state == ISPFC_CLS_STATE_NSCN_OCCURRED ||
		       new_state == ISPFC_CLS_STATE_DOWN ||
		       new_state == ISPFC_CLS_STATE_RESET_REQUESTED ||
		       new_state == ISPFC_CLS_STATE_LIP_GENERATED ||
		       new_state == ISPFC_CLS_STATE_NSCN_GENERATED ||
		       new_state == ISPFC_CLS_STATE_OFFLINE_REQUESTED);
		break;
	case ISPFC_CLS_STATE_DOWN:
		VERIFY(new_state == ISPFC_CLS_STATE_UP ||
		       new_state == ISPFC_CLS_STATE_RESET_REQUESTED ||
		       new_state == ISPFC_CLS_STATE_OFFLINE_REQUESTED);		       
		break;
	case ISPFC_CLS_STATE_LIP_GENERATED:
		VERIFY(new_state == ISPFC_CLS_STATE_LIP_OCCURRED ||
		       new_state == ISPFC_CLS_STATE_RESET_REQUESTED);
		break;
	case ISPFC_CLS_STATE_NSCN_GENERATED:
		VERIFY(new_state == ISPFC_CLS_STATE_NSCN_OCCURRED ||
		       new_state == ISPFC_CLS_STATE_RESET_REQUESTED);
		break;
	case ISPFC_CLS_STATE_LIP_OCCURRED:
		VERIFY(new_state == ISPFC_CLS_STATE_LIP_OCCURRED ||
		       new_state == ISPFC_CLS_STATE_REESTABLISHING ||
		       new_state == ISPFC_CLS_STATE_RESET_REQUESTED);
		break;
	case ISPFC_CLS_STATE_NSCN_OCCURRED:
		VERIFY(new_state == ISPFC_CLS_STATE_NSCN_OCCURRED ||
		       new_state == ISPFC_CLS_STATE_REESTABLISHING ||
		       new_state == ISPFC_CLS_STATE_RESET_REQUESTED);
		break;
	case ISPFC_CLS_STATE_REESTABLISHING:
		VERIFY(new_state == ISPFC_CLS_STATE_RESET_REQUESTED ||
		       new_state == ISPFC_CLS_STATE_UP ||
		       new_state == ISPFC_CLS_STATE_LIP_OCCURRED ||
		       new_state == ISPFC_CLS_STATE_NSCN_OCCURRED);
		break;
	case ISPFC_CLS_STATE_RESET_REQUESTED:
		VERIFY(new_state == ISPFC_CLS_STATE_RESETTING ||
		       new_state == ISPFC_CLS_STATE_OFFLINE_REQUESTED);
		break;
	case ISPFC_CLS_STATE_RESETTING:
		VERIFY(new_state == ISPFC_CLS_STATE_REESTABLISHING ||
		       new_state == ISPFC_CLS_STATE_RESET_REQUESTED ||
		       new_state == ISPFC_CLS_STATE_OFFLINE_REQUESTED);
		break;
	case ISPFC_CLS_STATE_OFFLINE:
		VERIFY(new_state == ISPFC_CLS_STATE_RESET_REQUESTED);
		break;
	case ISPFC_CLS_STATE_OFFLINE_REQUESTED:
		VERIFY(new_state == ISPFC_CLS_STATE_OFFLINE_PENDING ||
		       new_state == ISPFC_CLS_STATE_OFFLINE);
		break;
	case ISPFC_CLS_STATE_OFFLINE_PENDING:
		VERIFY(new_state == ISPFC_CLS_STATE_OFFLINE);
		break;
	default:
		VERIFY(0);
	}
}
#endif /* DEBUG */


/*******************************************************************************
 * ispfc_link_management_handler - 
 *
 *   Routine periodically called from the EEE poll routine, to handle link
 *   state and manage recovery.
 ******************************************************************************/
int32
ispfc_link_management_handler(ispfc_cb_t *cb, uint32 timer_ref)
{
	ispfc_dev_t *		dev;
	static eeeRTC_t		timer;
	static boolean		first = TRUE;
	eeeRTC_t			duration;
	static eeeRTC_t		interval;

	if (!first) {
		eee_rtcDuration(&timer, &duration);
	}
	else {
		first = FALSE;
		/* Set out interval to about 1/8 of a second. */
		eee_usecToRTC(0x20000LL, &interval);
		/* Force duration to be larger than interval the first time. */
		eee_usecToRTC(0x40000LL, &duration);
		eee_rtcStart(&timer);
	}
	if (eee_rtcCompare(&duration, &interval) > 0) {
		eee_rtcReset(&timer);
		if (LINK_STATE(cb) == ISPFC_CLS_STATE_UP && cb->scsi_desc_busy_count)
			ispfc_led_control(cb, ISPFC_LED_CONTROL_ACT_TOGGLE); 
		else
		if (LINK_STATE(cb) == ISPFC_CLS_STATE_UP && cb->scsi_desc_busy_count == 0)
			ispfc_led_control(cb, ISPFC_LED_CONTROL_ACT_ON);
		else
		if (LINK_STATE(cb) == ISPFC_CLS_STATE_OFFLINE && (cb->link_state & ISPFC_CLF_FC_OFFLINE_LATCH))
			ispfc_led_control(cb, ISPFC_LED_CONTROL_FLT_ON);
		else
			ispfc_led_control(cb, ISPFC_LED_CONTROL_FLT_TOGGLE);
	}

	switch (LINK_STATE(cb)) {
	case ISPFC_CLS_STATE_OFFLINE:
		if (!(cb->link_state & ISPFC_CLF_FC_OFFLINE_LATCH)) {
			/* Offline is transient state -- back to attempting to bring up the link */
			cb->link_state &= ~ISPFC_CLF_LINK_UP_LATCH;
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		}
		break;

	case ISPFC_CLS_STATE_UP:
		/* Up is stable state -- nothing to do */
		break;

	case ISPFC_CLS_STATE_RESETTING:
		/*
		 * Waiting for link init done event -- nothing to do 
		 */
		break;

	case ISPFC_CLS_STATE_REESTABLISHING:
		/*
		 * Waiting for reestablishing done event -- nothing to do 
		 */
		break;

	case ISPFC_CLS_STATE_RESET_REQUESTED:
		/* 
		 * Link reset requested -- start the link init state
		 * machine and wait for completion.  
		 */
		E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
		       ("ispfc: Resetting Fibrechannel link"));
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESETTING);

		ispfc_cleanup_link_state(cb);

		/*
		 * Start the link initialization state machine
		 */
		cb->link_state |= ISPFC_CLF_POINT_TO_POINT_OK;
		ispfc_init_link(cb, ispfc_resetting_done_callout, NULL);
		break;

	case ISPFC_CLS_STATE_DOWN:
#if 0
		/*
		 * Waiting for link up event. If link down more that
		 * LINK_DOWN_RECOVERY_SECS without coming up by
		 * itself, poke it. Otherwise nothing to do.  
		 */

		if (SECS_SINCE_LINK_DOWN(cb) > LINK_DOWN_RECOVERY_SECS) {
			/* Link down too long -- poke it */
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		}
#else
		if (cb->link_state & ISPFC_CLF_LINK_UP_LATCH &&
		    SECS_SINCE_LINK_DOWN(cb) > MAX_LINK_DOWN_SECS_WITHOUT_RESET) {
			/* Link down too long -- offline the link */
			FC_DBG(0, ("%s: Fibrechannel link failed to come up in %d seconds. Invalidating devices.\n",
				   cb->debug_name, MAX_LINK_DOWN_SECS_WITHOUT_RESET));
			E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
			       ("ispfc: Fibrechannel link failed to come up in %d seconds. Invalidating devices.",
				MAX_LINK_DOWN_SECS_WITHOUT_RESET));
			ispfc_send_fc_state_msg(ISPFC_CLS_STATE_DOWN);
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_OFFLINE_REQUESTED);
			return(NFX_OK);
		}
#endif
		break;

	case ISPFC_CLS_STATE_OFFLINE_REQUESTED:
		/*
		 * An link offline request has been made -- "clean up"
		 * link state and soft reset adapter.
		 */
		ispfc_cleanup_link_state(cb);
		ispfc_soft_reset_adapter(cb);

		/*
		 * Reinitialize fabric mapping map and unmap devices.
		 */
		if (cb->interface_type == HAL_IF_FC_FL_PORT ||
		    cb->interface_type == HAL_IF_FC_F_PORT) {
			ispfc_fabric_map_init(cb);
			FOR_ALL_DEVS({
				dev->dev_state &= ~ISPFC_DF_MAPPING_ID_ALLOCATED;
			});
		}

		/* 
		 * Put all device state machines into invalidation
		 * state.  
		 */
		FOR_ALL_DEVS({
			dev->dev_state |= ISPFC_DF_INVALIDATION_REQUESTED;
		});

		/* 
		 * Clear point-to-point OK flag so next time we'll
		 * attempt loop init also. 
		 */
		cb->link_state &= ~ISPFC_CLF_POINT_TO_POINT_OK;

		/*
		 * Clear ASYNC offline request 
		 */
		cb->link_state &= ~ISPFC_CLF_FC_ASYNC_OFFLINE_REQUEST;

		SET_LINK_STATE(cb, ISPFC_CLS_STATE_OFFLINE_PENDING);
		break;

	case ISPFC_CLS_STATE_OFFLINE_PENDING:
		/*
		 * Waiting for all devices to become invalidated -- poll device state.
		 */
		if (cb->dev_count == 0) {
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_OFFLINE);
			if (cb->link_state & ISPFC_CLF_FC_OFFLINE_LATCH)
				E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
				       ("ispfc: Fibrechannel link now disabled"));
		}
		break;

	case ISPFC_CLS_STATE_LIP_OCCURRED:
		/* 
		 * A LIP event occurred - awaiting debounce timer to
		 * expire -- nothing much else to do.  
		 */
		if (SECS_SINCE_LIP_OCCURRED(cb) > LINK_SETTLE_DEBOUNCE_SECS) {
			/*
			 * Start the "reestablishing connection state
			 * machine to [re]discover devices.  
			 */
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_REESTABLISHING);
			ispfc_reestablish_connection(cb, ispfc_reestablishing_done_callout, NULL);
		}
		break;

	case ISPFC_CLS_STATE_NSCN_OCCURRED:
		/* 
		 * A NSCN event occurred - awaiting debounce timer to
		 * expire -- nothing much else to do.  
		 */
		if (SECS_SINCE_NSCN_OCCURRED(cb) > LINK_SETTLE_DEBOUNCE_SECS) {
			/*
			 * Start the "reestablishing connection state
			 * machine to [re]discover devices.  
			 */
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_REESTABLISHING);
			ispfc_reestablish_connection(cb, ispfc_reestablishing_done_callout, NULL);
		}
		break;
		
	case ISPFC_CLS_STATE_LIP_GENERATED:
		/*
		 * Awaiting a LIP_OCCURRED event -- if it doesn't
		 * happen within LINK_SETTLE_DEBOUNCE_SECS, we need to
		 * reset the link 
		 */
		if (SECS_SINCE_LIP_GENERATED(cb) > LINK_SETTLE_DEBOUNCE_SECS) {
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		}
		break;

	case ISPFC_CLS_STATE_NSCN_GENERATED:
		/*
		 * Awaiting a NSCN_OCCURRED event -- if it doesn't
		 * happen within LINK_SETTLE_DEBOUNCE_SECS, we need to
		 * reset the link 
		 */
		if (SECS_SINCE_NSCN_GENERATED(cb) > LINK_SETTLE_DEBOUNCE_SECS) {
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		}
		break;

	default:
		VERIFY(0);
	}

	return(NFX_OK);
}

void 
ispfc_resetting_done_callout(ispfc_cb_t *cb, void *cookie, int32 status)
{
	/*
	 * Check that link hasn't been down for longer that MAX_LINK_DOWN_SECS
	 */
	if (status != LINK_INIT_STATUS_SUCCESS &&
	    cb->link_state & ISPFC_CLF_LINK_UP_LATCH &&
	    SECS_SINCE_LINK_DOWN(cb) > MAX_LINK_DOWN_SECS_WITH_RESET) {
		/* Link down too long -- offline the link */
		FC_DBG(0, ("%s: Fibrechannel link failed to come up in %d seconds. Invalidating devices.\n",
			   cb->debug_name, MAX_LINK_DOWN_SECS_WITH_RESET));
		E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
		       ("ispfc: Fibrechannel link failed to come up in %d seconds. Invalidating devices.",
			MAX_LINK_DOWN_SECS_WITH_RESET));
		ispfc_send_fc_state_msg(ISPFC_CLS_STATE_DOWN);
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_OFFLINE_REQUESTED);
		return;
	}

	/* 
	 * Check if ASYNC offline request was made while resetting.
	 */
	if (cb->link_state & ISPFC_CLF_FC_ASYNC_OFFLINE_REQUEST) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_OFFLINE_REQUESTED);
		return;
	}

	VERIFY(LINK_STATE(cb) == ISPFC_CLS_STATE_RESETTING);

	if (status != LINK_INIT_STATUS_SUCCESS) {
		/*
		 * Is this the first time we've tried to initialize link?
		 * If so, offline the link 
		 */
		if (!(cb->link_state & ISPFC_CLF_LINK_UP_LATCH)) {
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_OFFLINE_REQUESTED);
			return;
		}

		/* 
		 * Link Initialization failed -- request another.
		 */
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		ispfc_dev_t *dev;

		/*
		 * Link initialization was successfull -- cleanse internal state.
		 */

		/*
		 * Reinitialize fabric mapping map and unmap devices.
		 */
		if (cb->interface_type == HAL_IF_FC_FL_PORT ||
		    cb->interface_type == HAL_IF_FC_F_PORT) {
			ispfc_fabric_map_init(cb);
			FOR_ALL_DEVS({
				dev->dev_state &= ~ISPFC_DF_MAPPING_ID_ALLOCATED;
			});
		}

		/*
		 * Unmap MS/NS; NS is always mapped, MS is dynamically managed.
		 */
		SET_NS_STATE(cb, ISPFC_CLS_NS_STATE_MAPPED);
		SET_MS_STATE(cb, ISPFC_CLS_MS_STATE_NOT_MAPPED);

		/*
		 * Start the "reestablishing connection state machine
		 * to [re]discover devices.  
		 */
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_REESTABLISHING);
		ispfc_reestablish_connection(cb, ispfc_reestablishing_done_callout, NULL);
	}
}

void 
ispfc_reestablishing_done_callout(ispfc_cb_t *cb, void *cookie, int32 status)
{
	VERIFY(LINK_STATE(cb) == ISPFC_CLS_STATE_REESTABLISHING);

	if (status != REESTABLISH_STATUS_SUCCESS) {
		/* 
		 * Device [re]discovery failed -- request another link reset.
		 */
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		/*
		 * Device [re]discovery was successfull -- link is now
		 * up.  All SCSI descriptors queued in waiting queues
		 * will be started by device management handler.
		 */
		E_LOG (class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0, 
		       ("ispfc: Fibrechannel link now enabled and up"));
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_UP);
	}
}
 
int
ispfc_fabric_map_init(ispfc_cb_t *cb)
{
	int i;

        ASSERT(cb->interface_type == HAL_IF_FC_FL_PORT ||
               cb->interface_type == HAL_IF_FC_F_PORT);
	
        for (i = 0; i < LOOPID_SPACE_SIZE; ++i)
                cb->mapping_id_map[i] = FABRIC_MAP_UNMAPPED;

        cb->map_cache_hits = cb->map_cache_misses = 0;

        /*
         * 0x7E        reserved for F_port
         * 0x7F        reserved for Fabric Controller
         * 0x80        reserved for Name Server
	 * 0x81        reserved (by driver) for Management Server
         * 0xFF        reserved for IP Broadcast
         */
        cb->mapping_id_map[0x7E] = FABRIC_MAP_MAPPED_PERM;
        cb->mapping_id_map[0x7F] = FABRIC_MAP_MAPPED_PERM;
        cb->mapping_id_map[0x80] = FABRIC_MAP_MAPPED_PERM;
        cb->mapping_id_map[0x81] = FABRIC_MAP_MAPPED_PERM;
        cb->mapping_id_map[0xFF] = FABRIC_MAP_MAPPED_PERM;

        if (cb->interface_type == HAL_IF_FC_FL_PORT) {
                /*
                 * 0x00 - 0x7D reserved for loop devices
                 */
                for (i = 0; i <= 0x7D; ++i)
                        cb->mapping_id_map[i] = FABRIC_MAP_UNUSABLE;
                cb->avail_mapping_id_count = cb->max_avail_mapping_id_count = 126;
                cb->next_avail_mapping_id = 0x81;
        }
        else {
                cb->avail_mapping_id_count = cb->max_avail_mapping_id_count = 251;
                cb->next_avail_mapping_id = 0;
        }

        return(NFX_OK);
}

int
ispfc_fabric_map_allocate_id(ispfc_cb_t *cb)
{
	int mapping_id = NFX_ERR;
	int i;

	if (cb->avail_mapping_id_count) {
		for (i = 0; i < LOOPID_SPACE_SIZE; ++i) {
                        if (cb->next_avail_mapping_id >= LOOPID_SPACE_SIZE)
                                cb->next_avail_mapping_id = 0;
                        if (cb->mapping_id_map[cb->next_avail_mapping_id] == FABRIC_MAP_UNMAPPED) {
                                mapping_id = cb->next_avail_mapping_id++;
                                cb->mapping_id_map[mapping_id] = FABRIC_MAP_MAPPED_TEMP;
                                --cb->avail_mapping_id_count;
                                break;
                        }
			else
				++cb->next_avail_mapping_id;
                }
        }

	return(mapping_id);
}

int
ispfc_fabric_map_deallocate_id(ispfc_cb_t *cb, int mapping_id)
{
	if (mapping_id < 0 ||
	    cb->mapping_id_map[mapping_id] != FABRIC_MAP_MAPPED_TEMP)
		return(NFX_ERR);
	cb->mapping_id_map[mapping_id] = FABRIC_MAP_UNMAPPED;
        ++cb->avail_mapping_id_count;
	return(NFX_OK);
}

int 
ispfc_fabric_map_is_allocated_id(ispfc_cb_t *cb, int mapping_id)
{
	return(cb->mapping_id_map[mapping_id] != FABRIC_MAP_UNMAPPED);
}

ispfc_dev_t *
ispfc_fabric_map_find_lru(ispfc_cb_t *cb)
{
	ispfc_dev_t *	dev;
	ispfc_dev_t *	lru_dev = NULL;

	FOR_ALL_DEVS({
		if (dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED) {
			if (lru_dev == NULL ||
			    dev->last_accessed_timestamp < lru_dev->last_accessed_timestamp)
				lru_dev = dev;
		}
	});
	return(lru_dev);
}

#define NO_BITS			0xFFFFFFFFFFFFFFFFLL
#define BITS_PER_WORD		(sizeof(cb->handle_map[0]) * 8)
#define WORDS_IN_MAP		(sizeof(cb->handle_map)/sizeof(cb->handle_map[0]))

int
ispfc_device_handle_init(ispfc_cb_t *cb)
{
	bzero(cb->handle_map, sizeof(cb->handle_map));
	cb->avail_handle_count = WORDS_IN_MAP * BITS_PER_WORD;
	cb->next_avail_handle = 0;

	return(NFX_OK);
}

int32
ispfc_device_handle_allocate(ispfc_cb_t *cb)
{
	int32	handle;
	int	i, word, bit;
	uint64	mask;

	if (cb->avail_handle_count) {
		/* 
		 * Find a word with available bits, starting at cb->next_avail_handle
		 */
		handle = cb->next_avail_handle;
		word = handle / BITS_PER_WORD;
		for (i = 0;
		     i < WORDS_IN_MAP; 
		     ++i, word = ((word == WORDS_IN_MAP-1) ? 0 : word+1)) {

			if (cb->handle_map[word] == NO_BITS)
				continue;

			/*
			 * Slide the mask until an available bit is found
			 */

			bit = handle % BITS_PER_WORD;
			mask = 1LL << bit;

			for (; bit < BITS_PER_WORD; ++bit, mask <<= 1, ++handle) {
				if (!(cb->handle_map[word] & mask)) {
					cb->handle_map[word] |= mask;
					--cb->avail_handle_count;
					cb->next_avail_handle = 
						(handle == WORDS_IN_MAP * BITS_PER_WORD - 1) ? 0 : handle+1;
					return(handle);
				}
			}
		}
	}

	return(NFX_ERR);
}

int
ispfc_device_handle_deallocate(ispfc_cb_t *cb, int32 handle)
{
	int	word, bit;
	uint64	mask;

	VERIFY(cb->avail_handle_count > 0);
	
	word = handle / BITS_PER_WORD;
	bit = handle % BITS_PER_WORD;
	mask = 1LL << bit;
	
	if (!(cb->handle_map[word] & mask))
		return(NFX_ERR);
	else {
		cb->handle_map[word] &= ~mask;
		++cb->avail_handle_count;
		return(NFX_OK);
	}
}

/*******************************************************************************
 * ispfc_queue_desc_on_ct_busy_queue -
 *
 *   Insert descriptor into BUSY queue, sorted by timeout value. The
 *   device global timeout timer is adjusted to the maximum time out value
 *   of all descriptors currently pending/busy.
 ******************************************************************************/
void
ispfc_queue_desc_on_ct_busy_queue(ispfc_cb_t *cb, ct_desc_t *ctd)
{
        ct_desc_t *	p;

	FC_DBG(20, ("%s: queueing CT desc %p on B q\n", 
		    cb->debug_name, ctd));

	VERIFY(ctd->state == CT_DESC_STATE_WAITING);
        ctd->state = CT_DESC_STATE_BUSY;
        cb->ct_desc_busy_count++;

        /*
         * Insert into list sorted by timeout value.
         */
        if ((p = cb->ct_desc_busy)) {
                while (p->next != cb->ct_desc_busy &&
                       ctd->timeout_count < p->timeout_count)
                        p = p->next;
                if (p->timeout_count > ctd->timeout_count) {
                        /* Insert after p */
                        p->next->prev = ctd;
                        ctd->next = p->next;
                        ctd->prev = p;
                        p->next = ctd;
                }
                else {
                        /* Insert before p */
                        p->prev->next = ctd;
                        ctd->prev = p->prev;
                        ctd->next = p;
                        p->prev = ctd;
                        if (p == cb->ct_desc_busy)
                                cb->ct_desc_busy = ctd;
                }
        }
        else {
                cb->ct_desc_busy = ctd;
                ctd->next = ctd->prev = ctd;
        }

        /*
         * Reset the global timer value to the max -- we don't know in
         * which order commands get executed on drives.
         */
        cb->ct_global_timeout_timer = cb->ct_desc_busy->timeout_count + DEVICE_QUIESCE_TIMER;
}

/*******************************************************************************
 * ispfc_queue_desc_on_ct_waiting_queue -
 *
 *   Insert descriptor into WAITING queue, sorted by arrival time; i.e. 
 *   new arrivals are inserted at the end of the queue.
 ******************************************************************************/
void
ispfc_queue_desc_on_ct_waiting_queue(ispfc_cb_t *cb, ct_desc_t *ctd)
{
	FC_DBG(20, ("%s: queueing CT desc %p on W q\n", 
		    cb->debug_name, ctd));

	VERIFY(ctd->state == CT_DESC_STATE_WAITING);

	if (cb->ct_desc_waiting == NULL) {
		ctd->next = ctd->prev = ctd;
		cb->ct_desc_waiting = ctd;
	}
	else {
		ctd->prev = cb->ct_desc_waiting->prev;
		ctd->next = cb->ct_desc_waiting;
		cb->ct_desc_waiting->prev->next = ctd;
		cb->ct_desc_waiting->prev = ctd;
	}

}

/*******************************************************************************
 * ispfc_dequeue_desc_off_ct_busy_queue -
 *
 *   Remove descriptor from BUSY queue. The
 *   device global timeout timer is adjusted to the maximum time out value
 *   of all remaining descriptors currently pending/busy.
 ******************************************************************************/
void
ispfc_dequeue_desc_off_ct_busy_queue(ispfc_cb_t *cb, ct_desc_t *ctd)
{
        ct_desc_t *	head;

	FC_DBG(20, ("%s: dequeueing CT desc %p off B q\n", 
		    cb->debug_name, ctd));

        VERIFY(ctd->state == CT_DESC_STATE_BUSY ||
               ctd->state == CT_DESC_STATE_TIMED_OUT);
        VERIFY(cb->ct_desc_busy_count);
	
        ctd->state = CT_DESC_STATE_WAITING;
        cb->ct_desc_busy_count--;

        head = cb->ct_desc_busy;
        VERIFY(head && ctd->next);
        if (head == ctd) {
                if (head->next == head) {
                        VERIFY(head->prev == head);
                        cb->ct_desc_busy = NULL;
                }
                else {
                        cb->ct_desc_busy = head->next;
                }
        }
        /* dequeue it from the list */
        ctd->prev->next = ctd->next;
        ctd->next->prev = ctd->prev;

        ctd->next = ctd->prev = NULL;

        /*
         * Recompute next global timeout value. Just use the current
         * timeout value of the command at the head of the list. This
         * *should* be the one with the largest timer value.
         */
        if (cb->ct_desc_busy)
                cb->ct_global_timeout_timer = cb->ct_desc_busy->timeout_count + DEVICE_QUIESCE_TIMER;
        else
                cb->ct_global_timeout_timer = 0;
}

/*******************************************************************************
 * ispfc_dequeue_desc_off_ct_waiting_queue -
 *
 *   Remove descriptor from WAITING queue, from the head; i.e. 
 *   in FIFO order.
 ******************************************************************************/
ct_desc_t *
ispfc_dequeue_desc_off_ct_waiting_queue(ispfc_cb_t *cb)
{
	ct_desc_t *	ctd;

	if ((ctd = cb->ct_desc_waiting)) {
		ctd->next->prev = ctd->prev;
		ctd->prev->next = ctd->next;
		cb->ct_desc_waiting = ctd->next;
		if (ctd->next == ctd)
			cb->ct_desc_waiting = NULL;
		ctd->next = ctd->prev = NULL;
#if 0
		VERIFY(ctd->state == CT_DESC_STATE_WAITING);
#else
		/* XXX debug code */
		if (ctd->state != CT_DESC_STATE_WAITING) {
			printf("Assertion: ispfc_dequeue_desc_off_ct_waiting_queue: ct %p, state %d, internal %d\n", 
			       ctd, ctd->state, ctd->internal);
			VERIFY(ctd->state == CT_DESC_STATE_WAITING);
		}
#endif			
		FC_DBG(20, ("%s: dequeueing CT desc %p off W q\n", 
			    cb->debug_name, ctd));
	}

	return(ctd);
}

/*******************************************************************************
 * ispfc_purge_internal_desc_off_ct_waiting_queue -
 *
 *   "Toss" all CT descriptors from WAITING queue, that have the internal
 *   field set.
 ******************************************************************************/
void
ispfc_purge_internal_desc_off_ct_waiting_queue(ispfc_cb_t *cb)
{
	ct_desc_t *	ctd;
	ct_desc_t *	next;

 again:
	if ((ctd = cb->ct_desc_waiting)) {
		do {
			next = ctd->next;

			if (ctd->internal) {
				ctd->next->prev = ctd->prev;
				ctd->prev->next = ctd->next;
				if (ctd->next == ctd)
					cb->ct_desc_waiting = NULL;
				goto again;
			}

			ctd = next;
		} while (ctd != cb->ct_desc_waiting);
	}
}

void 
ispfc_led_control(ispfc_cb_t *cb, int code)
{
#ifdef AGILE_HARDWARE
	static ushort16 gpiod = 0;

	/* 
	 * GPIO bits:
	 *  bit 7 : Output: 0 = link fail LED indicator
	 *  bit 6 : Output: 0 = link powered ACT LED indicator
	 *  bit 5 : Output: 1 = SFP TX disable
	 *  bit 4 : Input:  0 = module inserted, 1 = module removed
	 *  bit 3 : Input:  1 = loss of signal
	 *  bit 2 : Input:  1 = TX fault (may not be supported on all modules)
	 *  bit 1 : Reserved
	 *  bit 0 : Reserved
	 */

#if 0
10 green
00 red
01 red
11 off
#endif

	switch(code) {
	case ISPFC_LED_CONTROL_INIT:
		ISPFC_REG_WRITE(cb, ISPFC_REG_GPIOE, 0xE0);
		gpiod = 0xC0;
		ISPFC_REG_WRITE(cb, ISPFC_REG_GPIOD, gpiod);
		break;

	case ISPFC_LED_CONTROL_FLT_ON:
		gpiod = 0x40;
		ISPFC_REG_WRITE(cb, ISPFC_REG_GPIOD, gpiod);
		break;

	case ISPFC_LED_CONTROL_FLT_TOGGLE:
		gpiod |= 0x40;
		gpiod ^= 0x80;
		ISPFC_REG_WRITE(cb, ISPFC_REG_GPIOD, gpiod);
		break;

	case ISPFC_LED_CONTROL_ACT_ON:
		gpiod = 0x80;
		ISPFC_REG_WRITE(cb, ISPFC_REG_GPIOD, gpiod);
		break;

	case ISPFC_LED_CONTROL_ACT_TOGGLE:
		gpiod |= 0x80;
		gpiod ^= 0x40;
		ISPFC_REG_WRITE(cb, ISPFC_REG_GPIOD, gpiod);
		break;

	default:
		FC_DBG(2, ("%s: ispfc_led_control: invalid code %d\n",
			   cb->debug_name, code));
	}

#endif /* AGILE_HARDWARE */
}

void
ispfc_send_fc_state_msg(int state)
{
	nfx_eventMsg_t	eventMsg;
	uint32			dest;
	uint32			source;

	memset((void *)&eventMsg, 0, sizeof(nfx_eventMsg_t));
	eventMsg.msgType = NFX_EVENT_POST;
	eventMsg.needAck = 0;
	eventMsg.eventScope = NFX_EVENT_GLOBAL;
	eventMsg.eventType = NFX_EVENT_PORT;
	if(state == ISPFC_CLS_STATE_DOWN) {
		eventMsg.status = NFX_EVENT_STATUS_ERR;
		eventMsg.param.portEvent.state = EVENT_STATE_DOWN;
	}
	else {
		eventMsg.status = NFX_EVENT_STATUS_OK;
		eventMsg.param.portEvent.state = EVENT_STATE_UP;
	}

	dest = PACK_UNICAST_PORT_ADDRESS(NFX_SLOTTYPE_SSC, NFX_SSC_UNIX,
						eee_getApplicationId("eventd"),
						EEE_MGMT, NFX_COS_HIGH, 0);

	source = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId, 
						eee_getApplicationId("scsi"),
						EEE_MGMT, NFX_COS_HIGH, 0);

	eventMsg.param.portEvent.portAddr = source;

	eee_sendMessage((void *)&eventMsg, sizeof(eventMsg), dest, source);
}
