/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_nvram.c
 *
 * RCS:		$Id: ispfc_nvram.c,v 1.17 2002/10/25 22:24:36 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver NVRAM code.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/05/01
 *
 *------------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_nvram.h"
#include <exports/cm-api.h>
#include <exports/seep-api.h>

#define NVRAM_DELAY()	ispfc_nv_delay() /* 10 uSec delay */

ushort16 ispfc_nvram_request(ispfc_cb_t *, uint32);
void	 ispfc_nv_write(ispfc_cb_t *, ushort16);
void	 ispfc_nv_delay(void);

/*******************************************************************************
 * ispfc_nvram_config - Initialize interesting values from NVRAM
 *
 * Input:
 *      cb                = adapter control block pointer.
 *
 * Output:
 *      initialization control block in response_ring
 *      host adapters parameters in host adapter block
 *
 * Returns:
 *      NFX_OK if NVRAM valid, else NFX_ERR
 ******************************************************************************/
int
ispfc_nvram_config(ispfc_cb_t *cb)
{
	int			status = NFX_OK;
	ushort16		cnt;
	nvram2300_t *		nv = (nvram2300_t *)cb->request;
	ushort16 *		wptr = (ushort16 *)nv;
	byte			chksum = 0;
	wwn_t *			name;
	wwn_t			dummy_wwn;

	bzero(nv, sizeof(nvram2300_t));

	/* Read the whole NVRAM */
	for (cnt = 0; cnt < NV_BYTE_COUNT/sizeof(ushort16); ++cnt) {
		*wptr = ispfc_get_nvram_word(cb, cnt);
#if defined(BIG_ENDIAN_EVAL_BOARD)
		SWAP2(wptr);
#endif
		chksum += (byte)*wptr;
		chksum += (byte)(*wptr >> 8);
		wptr++;
	}

#if defined(ISPFC_SIM)
	chksum = 1;
#endif
	
	/* If bad NVRAM data, set defaults parameters. */
	if (chksum || 
	    nv->id[0] != 'I' || nv->id[1] != 'S' || nv->id[2] != 'P' || nv->id[3] != ' ' || 
	    nv->nvram_version < NV_MINIMUM_VERSION) {

		FC_DBG(1, ("ispfc_nvram_config: NVRAM checksum invalid -- attempting retrieval from SEEP.\n"));

		/* Set defaults */
#if defined(AGILE_HARDWARE)
		nv->adapter_p.RISC_ram_parity = 1;
#else
		nv->adapter_p.RISC_ram_parity = 0;
#endif

		name = &dummy_wwn;

#if defined(AGILE_HARDWARE)
		WWN_SET(&null_wwn, name);
		memcpy(&(name->u.name.ieee_assigned[0]), (char *)seepGetMyExtMacAddr(nfxMyCPUId), 6);
		if (name->u.name.ieee_assigned[0] == 0 &&
		    name->u.name.ieee_assigned[1] == 0 &&
		    name->u.name.ieee_assigned[2] == 0) 
#endif
		{
			FC_DBG(0, ("ispfc_nvram_config: Cannot determine WWN from SEEP; setting default.\n"));
			name->u.name.nport_identifier = 0x5001;
			name->u.name.ieee_assigned[0] = 0x0a;
			name->u.name.ieee_assigned[1] = 0x98;
			name->u.name.ieee_assigned[2] = 0x00;
			name->u.name.company_assigned[0] = (nfxMyCPUId & 0x00FF0000) >> 16;
			name->u.name.company_assigned[1] = (nfxMyCPUId & 0x0000FF00) >> 8;
			name->u.name.company_assigned[2] = (nfxMyCPUId & 0x000000FF) >> 0;
		}
		name->u.name.nport_identifier = 0x5001;
		inplace_htons(&name->u.name.nport_identifier);
		WWN_SET(nv->port_name, name);
		WWN_SET(nv->node_name, name);
		status = NFX_ERR;
	} /* if (chksum) */

	FC_DBG(0, ("ispfc_nvram_config: NVRAM dump: \n"));
	FC_DBG(0, ("     port_name ........ %s\n", wwn_to_str((wwn_t *)nv->port_name, NULL)));
	FC_DBG(0, ("     node_name ........ %s\n", wwn_to_str((wwn_t *)nv->node_name, NULL)));
	FC_DBG(0, ("     RISC_ram_parity .. %d\n", nv->adapter_p.RISC_ram_parity));

	/* Set FW Init CB fields */
	WWN_SET(cb->fw_init_cb->portname, nv->port_name);
	WWN_SET(cb->fw_init_cb->nodename, nv->node_name);

	/* Set operational flags */
	cb->sram_parity_available = nv->adapter_p.RISC_ram_parity;

	return(status); 
}

/*******************************************************************************
 * ispfc_get_nvram_word - Read NVRAM data word
 *      Calculates word position in NVRAM and calls request routine to
 *      get the word from NVRAM.
 *
 * Input:
 *      cb      = adapter control block pointer.
 *      address = NVRAM word address.
 *
 * Returns:
 *      data word.
 ******************************************************************************/
ushort16
ispfc_get_nvram_word(ispfc_cb_t *cb, ushort16 address)
{
	uint32		nv_cmd;
	ushort16	data;

	nv_cmd = address << 16;
	nv_cmd |= NV_READ_OP;

	data = ispfc_nvram_request(cb, nv_cmd);
	return(data);
}

/*******************************************************************************
 * ispfc_nvram_request - NVRAM request
 *      Sends read command to NVRAM and gets data from NVRAM.
 *
 * Input:
 *      cb     = adapter control block pointer.
 *      nv_cmd = Bit 26     = start bit
 *               Bit 25, 24 = opcode
 *               Bit 23-16  = address
 *               Bit 15-0   = write data
 *
 * Returns:
 *      data word.
******************************************************************************/
ushort16
ispfc_nvram_request(ispfc_cb_t *cb, uint32 nv_cmd) 
{
	byte		cnt;
	ushort16	data = 0;
	ushort16	reg_data;

	/* Send command to NVRAM. */
	nv_cmd <<= 5;
	for (cnt = 0; cnt < 11; ++cnt) {
                if (nv_cmd & BIT_31)
			ispfc_nv_write(cb, NVRAM_DATA_OUT);
                else
			ispfc_nv_write(cb, 0);
                nv_cmd <<= 1;
	}

	/* Read data from NVRAM. */
	for (cnt = 0; cnt < 16; ++cnt) {
		ISPFC_REG_WRITE(cb, ISPFC_REG_NVRAM, NVRAM_SELECT | NVRAM_CLOCK);
                NVRAM_DELAY();
                data <<= 1;
                reg_data = ISPFC_REG_READ(cb, ISPFC_REG_NVRAM);
                if (reg_data & NVRAM_DATA_IN)
			data |= BIT_0;
                ISPFC_REG_WRITE(cb, ISPFC_REG_NVRAM, NVRAM_SELECT);
                NVRAM_DELAY();
	}

	/* Deselect chip. */
	ISPFC_REG_WRITE(cb, ISPFC_REG_NVRAM, NVRAM_DESELECT);
	NVRAM_DELAY();

	return(data);
}

void
ispfc_nv_write(ispfc_cb_t *cb, ushort16 data) 
{
	ISPFC_REG_WRITE(cb, ISPFC_REG_NVRAM, data | NVRAM_SELECT);
	NVRAM_DELAY();
	ISPFC_REG_WRITE(cb, ISPFC_REG_NVRAM, data | NVRAM_SELECT | NVRAM_CLOCK);
	NVRAM_DELAY();
	ISPFC_REG_WRITE(cb, ISPFC_REG_NVRAM, data | NVRAM_SELECT);
	NVRAM_DELAY();
}

void
ispfc_nv_delay(void) 
{
	eee_delay(NV_DELAY_COUNT);
}

