/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_private.h
 *
 * RCS:		$Id: ispfc_private.h,v 1.50 2002/10/22 01:30:38 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver private definitions
 *
 * Created by:	Radek Aster
 *
 * Date Created: 5/11/01
 *
 *------------------------------------------------------------------------------
 */

/*
 * Two types of "simulators" are allowed: 
 *
 * (a) simulation of disks in LINUX files. Is only available when
 *     LINUX_TEST is defined. Set ISPFC_SIM to ISPFC_SIM_LINUX_FILE.
 * 
 * (b) simulation of disks in RAM, i.e. RAM disks. Is available when
 *     either LINUX_TEST or EVAL_BOARD are defined.  Set ISPFC_SIM to
 *     ISPFC_SIM_RAM_DISK. 
 */
#define ISPFC_SIM_LINUX_FILE 1
#define ISPFC_SIM_RAM_DISK   2
#ifdef LINUX_TEST
#define ISPFC_SIM ISPFC_SIM_LINUX_FILE
#endif

#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
/* #define ISPFC_SIM ISPFC_SIM_RAM_DISK */
#endif

#ifdef ISPFC_SIM
#define ISPFC_SIM_MAX_DISKS			20
#define ISPFC_SIM_MAX_TAPES			5
#define ISPFC_SIM_DISK_FMT			"%s/sim_disk_%d_%016llx"
#define ISPFC_SIM_DISK_SECTOR_SIZE		520
#define ISPFC_SIM_DISK_CAPACITY			102400
#define ISPFC_SIM_TAPE_DATA_FMT			"%s/sim_tape_data_%016llx.%d"
#define ISPFC_SIM_TAPE_DIR_FMT			"%s/sim_tape_dir_%016llx.%d"
#define ISPFC_SIM_TAPE_DEFAULT_BLOCKSIZE	1024
#define ISPFC_SIM_TAPE_MAX_BYTES		(100LL * 1024LL * 1024LL) /* 100 MB */
#endif

#if defined(ISPFC_SIM)
/* 
 * ISPFC_FABRIC_SIM -- simulate switch attached disks, otherwise loop
 * attached disks are simulated.  
 */
#define ISPFC_FABRIC_SIM 
#endif

#ifndef __ISPFC_PRIVATE_H__
#define __ISPFC_PRIVATE_H__

#include <nfx-incl.h>
#include <exports/eee-api.h>
#include "../sm-eee/eee.h"		/* XXX */
#include <exports/malloc-api.h>
#include <exports/timer-api.h>
#include <exports/scsi-api.h>
#include <exports/elog-api.h>
#include <exports/hal.h>
#include <exports/wwn.h>
#include <exports/debug.h>
#include <exports/mem-api.h>
#include "ispfc_hash.h"
#include "ispfc_channels.h"
#include "ispfc_debug.h"

#if defined(LINUX_TEST)
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif /* defined(LINUX_TEST) */

#if 1					/* XXX */
#include "ispfc_extern.h"
#endif

#define MSEC_PER_SEC 1000

#define SCSI_DEVICE_NOTIFY scsi_llDeviceNotify
#define SCSI_RECEIVE	   scsi_llReceive


/*
 * Operational constants
 */
#define MAX_LINK_DOWN_SECS_WITH_RESET		15
#define MAX_LINK_DOWN_SECS_WITHOUT_RESET	10
#define LINK_DOWN_RECOVERY_SECS			5	
#define LINK_SETTLE_DEBOUNCE_SECS		3
#define DEVICE_DEFERRED_LOGIN_RETRY		5
#define LOOPID_SPACE_SIZE			256
#define HANDLE_SPACE_SIZE			65536

/* Default timer values */
#if defined(ISPFC_SIM)
#define LOOP_UP_TIMEOUT				(1 * MSEC_PER_SEC) /* 1 second */
#else
#define LOOP_UP_TIMEOUT				(5 * MSEC_PER_SEC) /* 5 second */
#endif
#define TIMEOUT_HANDLER_PERIOD			(3 * MSEC_PER_SEC) /* 3 seconds */
#define ISPFC_MBOX_TIMEOUT_DEFAULT		(5 * MSEC_PER_SEC) /* 5 seconds */
#define ISPFC_MBOX_TIMEOUT_MAX_PREFERRED	(9 * MSEC_PER_SEC) /* 9 seconds */
#define ISPFC_RA_TOV_DEFAULT			(4 * MSEC_PER_SEC) /* 4 seconds */
#define ISPFC_TD_POLL_PERIOD			(5 * MSEC_PER_SEC) /* 5 seconds */
#define DEVICE_DEFERRED_LOGIN_TIMER		(1 * MSEC_PER_SEC) /* 1 seconds */
#define DEVICE_QUIESCE_TIMER			(3)		   /* 3 seconds */
#define DEVICE_QUEUE_FLUSH_TIMER		(2 * MSEC_PER_SEC) /* 2 seconds */

#define ISPFC_RESPONSE_THROTTLE			16		   /* Response limit per pass */

/* PCI config register definitions */
#define ISPFC_REG_PCI_VENDOR_ID           0x00
#define ISPFC_REG_PCI_DEVICE_ID           0x02
#define ISPFC_REG_PCI_COMMAND             0x04
#define ISPFC_REG_PCI_STATUS              0x06
#define ISPFC_REG_PCI_REVISION            0x08
#define ISPFC_REG_PCI_CLASS               0x09
#define ISPFC_REG_PCI_CACHE_LINE          0x0c
#define ISPFC_REG_PCI_LATENCY             0x0d
#define ISPFC_REG_PCI_HEADER_TYPE         0x0e
#define ISPFC_REG_PCI_IO_BASE             0x10
#define ISPFC_REG_PCI_MEM_BASE            0x14
#define ISPFC_REG_PCI_ROM_BASE            0x30
#define ISPFC_REG_PCI_MAX_LATENCY         0x3f

/* ISPFC_REG_PCI_COMMAND bit definitions */
#define CMD_FAST_BACK_ENABLE		  0x0200
#define CMD_SYS_ERR_ENABLE		  0x0100
#define CMD_ADDR_STEP			  0x0080
#define CMD_PARITY_RESPONSE		  0x0040
#define CMD_VGA_SNOOP			  0x0020
#define CMD_MEM_WRITE_INVALID		  0x0010
#define CMD_SPECIAL_CYCLES		  0x0008
#define CMD_BUS_MASTER			  0x0004
#define CMD_MEM_ADDR_ENABLE		  0x0002
#define CMD_IO_ADDR_ENABLE		  0x0001

#define ISPFC_LATENCY                     0x40
#define ISPFC_CACHE_LINE_SIZE             8               /*  4 byte quantities */
#define ISPFC_MAX_FC_FRAME_SIZE		  2048

typedef struct {
        byte		version;		/* Version */
        byte		reserved1;			
        ushort16	fw_options;		/* Firmware options */
        ushort16	max_frame_len;		/* Maximum frame length */
	ushort16	max_IOCB_resources;	/* Max. IOCB resource allocation */
	ushort16	execution_throttle;	/* Execution throttle */
        byte		retry_count;		/* Retry count */
        byte		retry_delay;		/* Retry delay */
        byte		portname[8];		/* Port name (WWN) */
        ushort16	hard_address;		/* Hard/previous address */
        byte		inquiry_data;		/* Inquiry data */
	byte		login_time_out;		/* Login timeout */
	byte		nodename[8];		/* Node name (WWN) */
        ushort16	request_out_pointer;	/* Request queue out-pointer */
        ushort16	response_in_pointer;	/* Response queue in-pointer */
        ushort16	request_queue_len;	/* Request queue size in entries */
        ushort16	response_queue_len;	/* Response queue size in entries */
        uint32		request_queue_addr[2];	/* 64-bit physical address of beginning of request queue */
        uint32		response_queue_addr[2];	/* 64-bit physical address of beginning of response queue */
	ushort16	lun_enables;		/* LUN enables */
	byte		command_rsrc_cnt;	/* Command resource count */
	byte		imm_notify_rsrc_cnt;	/* Immediate notify resource count */
        ushort16	timeout;		/* Timeout */
	ushort16	reserved2;
        ushort16	additional_fw_options;	/* Addition firmware options */
        ushort16	reserved3;
        ushort16	special_options;	/* Special options */
	ushort16	reserved4[13];
} ispfc_init_fw_cb_t;

/* Firmware options definitions */
#define IFW_FO_EXTENDED_CONTROL_BLOCK	   0x8000
#define IFW_FO_NAME_OPTION		   0x4000
#define IFW_FO_FORCE_PLOGI_ON_LIP	   0x2000
#define IFW_FO_ASSERT_LIPA		   0x0800
#define IFW_FO_DESCENDING_LOOP_ID_SEARCH   0x0400
#define IFW_FO_DISABLE_INIT_LIP		   0x0200
#define IFW_FO_PORT_DB_UPDATE_AE	   0x0100
#define IFW_FO_ENABLE_TARGET_DEVICE_TYPE   0x0080
#define IFW_FO_ENABLE_ADISC		   0x0040
#define IFW_FO_DISABLE_INIT		   0x0020
#define IFW_FO_ENABLE_TARGET		   0x0010
#define IFW_FO_FAST_POSTING		   0x0008
#define IFW_FO_FULL_DUPLEX		   0x0004
#define IFW_FO_ENABLE_FAIRNESS		   0x0002
#define IFW_FO_ASSERT_LIHA		   0x0001

/* Additional firmware options definitions */
#define IFW_AFO_CONN_TYPE_MASK		   0x0070
#define IFW_AFO_LOOP_ONLY		   0x0000
#define IFW_AFO_POINT_TO_POINT_ONLY	   0x0010
#define IFW_AFO_LOOP_PREFERRED		   0x0020
#define IFW_AFO_POINT_TO_POINT_PREFERRED   0x0030

/* Special options definitions */
#define IFW_SO_DATA_RATE_MASK		   0xC000
#define IFW_SO_1GIGABIT			   0x0000
#define IFW_SO_2GIGABIT			   0x4000
#define IFW_SO_AUTO_NEGOTIATE_SPEED	   0x8000
#define IFW_SO_50_OHM			   0x0000
#define IFW_SO_75_OHM			   0x2000

#define ISPFC_MAX_CT_REQUESTS		   10
#define ISPFC_MAX_SCSI_TAGS		   5
#define ISPFC_MAX_LUNS_PER_TARGET	   10
#define ISPFC_MAX_TARGETS		   128

#define ISPFC_IOCBS_PER_SCSI_REQ	   10
#define ISPFC_IOCBS_PER_CT_REQ		   1

#define ISPFC_MAX_OUTSTANDING_REQUESTS \
	(ISPFC_MAX_CT_REQUESTS + \
	 (ISPFC_MAX_SCSI_TAGS * ISPFC_MAX_LUNS_PER_TARGET * ISPFC_MAX_TARGETS))
#define ISPFC_REQUEST_QUEUE_SIZE \
	((ISPFC_MAX_CT_REQUESTS * ISPFC_IOCBS_PER_CT_REQ) + \
	 (ISPFC_MAX_TARGETS * ISPFC_MAX_LUNS_PER_TARGET * (1 + ISPFC_MAX_SCSI_TAGS * ISPFC_IOCBS_PER_SCSI_REQ)))
#define ISPFC_RESPONSE_QUEUE_SIZE \
	(ISPFC_MAX_OUTSTANDING_REQUESTS)

#define AEN_HANDLERS			   50

#define INC_Q_PTR(_ptr_, _s_) \
	{ \
		ASSERT((_ptr_) < (_s_)); \
		(_ptr_) = (((_ptr_) == (_s_) - 1) ? 0 : (_ptr_) + 1); \
	}

typedef struct {
	byte  bytes[64];
} ispfc_command_entry_t;

typedef struct {
	byte  bytes[64];
} ispfc_status_entry_t;

typedef struct {
	/* Request and Response rings */
   	ispfc_command_entry_t * request; /* request[ISPFC_REQUEST_QUEUE_SIZE]; */
   	ispfc_status_entry_t *  response; /* response[ISPFC_RESPONSE_QUEUE_SIZE]; */
	ushort16		request_in;
	ushort16		request_out;	
	ushort16		response_in;           
	ushort16		prev_response_in;           
	ushort16		response_out;           

	ispfc_init_fw_cb_t *	fw_init_cb;
	void *			fw_init_cb_copy;
	ispfc_init_fw_cb_t *	fw_init_cb_copy_aligned;

	/* IOCB management state */
	struct {
		void *		cookie;
		uint32		alloced; /* 0, ISPFC_ENTRY_TYPE_COMMAND2, ISPFC_ENTRY_TYPE_CTIOCB */
	} iocb_handles[ISPFC_MAX_OUTSTANDING_REQUESTS];
	uint32			next_iocb_handle;
	int32			available_iocb_handles;

	/* Base address of chip registers in memory space */
	void *			reg_mem_base;

	/* ISP 2300 RISC code */
	byte *			firmware_version;
	ushort16		risc_code_addr;
	ushort16		risc_code_length;
	ushort16 *		risc_code;
	void *			risc_code_copy;
	ushort16 *		risc_code_copy_aligned;

	/* WWNs */
	wwn_t			adapter_wwn;
	wwn_t			fabric_wwn;

	/* For registration with fabric */
	byte			symbolic_name[255];
	byte			fc4_type;

	/* Chip attributes */
	byte			chip_rev;
	byte			risc_rom_rev;
	byte			sram_parity_available:1,
				link_speed:2;

	/* Interface attributes and status */
	byte			interface_type;
	uint32			host_loop_id;
	uint32			host_port_id;

	uint32			mbox_timeout;		/* mailbox timeout in mSecs */
	uint32			switch_ra_tov;		/* RA_TOV in mSecs */
	uint32			lip_timeout;		/* LIP timeout in mSecs */

	/* Adapter name */
	char 			debug_name[32];

	/* Mailbox state */
	void *			mbox_sm;
	ushort16		mbox_in[32];		/* IN to the adapter */
	ushort16		mbox_out[32];		/* OUT from the adapter */
	byte			mbox_in_valid_count;
	byte			mbox_out_valid_count;
	   
	/* Thread state machines */
	int32			ispfc_timeout_handler_timer_handle;

	/* Device loopid mapping state */
	byte			mapping_id_map[LOOPID_SPACE_SIZE];	/* 256 available mapping IDs */
	byte			avail_mapping_id_count;
	byte			next_avail_mapping_id;
	byte			max_avail_mapping_id_count;
	uint32			map_cache_misses;
	uint32			map_cache_hits;
	uint32			dev_count_at_last_rebalance;

	/* Device handle mapping state */
	uint64			handle_map[HANDLE_SPACE_SIZE/64];	/* 65536 available handles */
	uint32			avail_handle_count; 
	uint32			next_avail_handle;

	/* Busy requests */
	ushort16		scsi_desc_busy_count;
	ushort16		ct_desc_busy_count;

	/* Attached Device state */
	void *			dev_physid_hash;
	void *			dev_handle_hash;
	void *			dev_list;
	uint32			dev_list_refcount;
	uint32			dev_count;
	uint32			dev_generation;

	/* CT state and requests */
	ct_desc_t *		ct_desc_busy;
	ct_desc_t *		ct_desc_waiting;
	byte                    ct_global_timeout_timer;

	/* Link State */
	uint32			link_state;
	eeeRTC_t		link_down_timestamp;
	eeeRTC_t		lip_generated_timestamp;
	eeeRTC_t		nscn_generated_timestamp;
	eeeRTC_t		lip_occurred_timestamp;
	eeeRTC_t		nscn_occurred_timestamp;
	uint32			reestablish_count;

	void *			adap_init_state; /* Adapter Init state machine */
	void *			conn_init_state; /* Connection Init state machine */
	void *			link_init_state; /* Link Init state machine */
	void *			dev_disc_state;  /* Device Discovery state machine */

	/* Registered AEN handlers */
	void *			aen_handler[AEN_HANDLERS];

	/* Stats/Counters */
	uint64			adap_frames_out;
	uint64			adap_frames_in;
	uint32			adap_underrun_cnt;
	uint32			adap_transport_error_cnt;
	uint32			adap_read_6_cmd_cnt;
	uint32			adap_read_10_cmd_cnt;
	uint32			adap_write_6_cmd_cnt;
	uint32			adap_write_10_cmd_cnt;
	uint32			adap_other_cmd_cnt;

#ifdef DEBUG
#define PIO_TRACE_MAX 512
	ushort16		pio_trace_i;
	struct {
		uint32		trace_val;
		uint32		trace_count;
		eeeRTC_t	trace_rtc;
	}			pio_trace[PIO_TRACE_MAX];
#endif /* DEBUG */
} ispfc_cb_t; 

/* 
 * Link state
 */
#define ISPFC_CLS_STATE_MASK			0x000000FF
#define ISPFC_CLS_STATE_RESETTING		0
#define ISPFC_CLS_STATE_REESTABLISHING		1
#define ISPFC_CLS_STATE_RESET_REQUESTED		2
#define ISPFC_CLS_STATE_DOWN			3
#define ISPFC_CLS_STATE_OFFLINE_REQUESTED	4
#define ISPFC_CLS_STATE_OFFLINE_PENDING 	5
#define ISPFC_CLS_STATE_OFFLINE			6
#define ISPFC_CLS_STATE_LIP_OCCURRED		7	/* loop only */
#define ISPFC_CLS_STATE_LIP_GENERATED		8	/* loop only */
#define ISPFC_CLS_STATE_NSCN_OCCURRED		9	/* fabric only */
#define ISPFC_CLS_STATE_NSCN_GENERATED		10	/* fabric only */
#define ISPFC_CLS_STATE_UP			11

#define ISPFC_CLS_MS_STATE_MASK			0x00000300
#define ISPFC_CLS_MS_STATE_NOT_MAPPED		0x00000000
#define ISPFC_CLS_MS_STATE_MAP_PENDING		0x00000100
#define ISPFC_CLS_MS_STATE_MAPPED		0x00000200

#define ISPFC_CLS_NS_STATE_MASK			0x00000C00
#define ISPFC_CLS_NS_STATE_NOT_MAPPED		0x00000000
#define ISPFC_CLS_NS_STATE_MAP_PENDING		0x00000400
#define ISPFC_CLS_NS_STATE_MAPPED		0x00000800

#define ISPFC_CLF_STATE_MASK			0x0FFF0000
#define ISPFC_CLF_POINT_TO_POINT_OK		0x00010000
#define ISPFC_CLF_LINK_UP_LATCH			0x00020000
#define ISPFC_CLF_NESTED_LIP			0x00040000
#define ISPFC_CLF_NESTED_NSCN			0x00040000
#define ISPFC_CLF_FC_GS3_NOT_SUPPORTED		0x00080000
#define ISPFC_CLF_FC_OFFLINE_LATCH		0x00100000
#define ISPFC_CLF_FC_ASYNC_OFFLINE_REQUEST	0x00200000

#define ISPFC_CLS_MODE_MASK			0xF0000000
#define ISPFC_CLS_MODE_RUNTIME			0x00000000
#define ISPFC_CLS_MODE_DIAG_SRAM		0x10000000
#define ISPFC_CLS_MODE_DIAG_LOOPBACK_INTERNAL	0x20000000
#define ISPFC_CLS_MODE_DIAG_LOOPBACK_EXTERNAL	0x30000000

#ifdef DEBUG
void    ispfc_check_state_machine(ispfc_cb_t *, int);
#define CHECK_LINK_STATE(_cb_, _s_) \
        ispfc_check_state_machine(_cb_, _s_)
#else
#define CHECK_LINK_STATE(_cb_, _s_)
#endif /* DEBUG */
#define IS_LINK_DOWN(_cb_) \
        (((_cb_)->link_state & ISPFC_CLS_STATE_MASK) == ISPFC_CLS_STATE_DOWN)
#define IS_LINK_UP(_cb_) \
        (((_cb_)->link_state & ISPFC_CLS_STATE_MASK) == ISPFC_CLS_STATE_UP)
#define IS_LINK_OFFLINE(_cb_) \
        (((_cb_)->link_state & ISPFC_CLS_STATE_MASK) == ISPFC_CLS_STATE_OFFLINE)
#define LINK_STATE(_cb_) \
        ((_cb_)->link_state & ISPFC_CLS_STATE_MASK)
#define SET_LINK_STATE(_cb_, _s_) \
        { \
		CHECK_LINK_STATE(_cb_, _s_); \
                if (IS_LINK_UP(_cb_)) \
			eee_rtc(&(_cb_)->link_down_timestamp); \
                if ((_s_) == ISPFC_CLS_STATE_LIP_GENERATED) \
			eee_rtc(&(_cb_)->lip_generated_timestamp); \
		else \
                if ((_s_) == ISPFC_CLS_STATE_NSCN_GENERATED) \
			eee_rtc(&(_cb_)->nscn_generated_timestamp); \
                else \
                if ((_s_) == ISPFC_CLS_STATE_LIP_OCCURRED) \
			eee_rtc(&(_cb_)->lip_occurred_timestamp); \
		else \
		if ((_s_) == ISPFC_CLS_STATE_NSCN_OCCURRED) \
			eee_rtc(&(_cb_)->nscn_occurred_timestamp); \
                else \
                if ((_s_) == ISPFC_CLS_STATE_UP) \
                        (_cb_)->link_state |= ISPFC_CLF_LINK_UP_LATCH; \
                (_cb_)->link_state = ((_cb_)->link_state & ~ISPFC_CLS_STATE_MASK) | (_s_); \
        }
#define SET_NS_STATE(_cb_, _s_) \
	(_cb_)->link_state = ((_cb_)->link_state & ~ISPFC_CLS_NS_STATE_MASK) | (_s_)
#define SET_MS_STATE(_cb_, _s_) \
	(_cb_)->link_state = ((_cb_)->link_state & ~ISPFC_CLS_MS_STATE_MASK) | (_s_)
#define NS_STATE(_cb_) \
	((_cb_)->link_state & ISPFC_CLS_NS_STATE_MASK)
#define MS_STATE(_cb_) \
	((_cb_)->link_state & ISPFC_CLS_MS_STATE_MASK)

#define SECS_SINCE_LINK_DOWN(_cb_) \
	(eee_rtcDurationSec(&(_cb_)->link_down_timestamp))
#define SECS_SINCE_LIP_GENERATED(_cb_) \
	(eee_rtcDurationSec(&(_cb_)->lip_generated_timestamp))
#define SECS_SINCE_NSCN_GENERATED(_cb_) \
	(eee_rtcDurationSec(&(_cb_)->nscn_generated_timestamp))
#define SECS_SINCE_LIP_OCCURRED(_cb_) \
	(eee_rtcDurationSec(&(_cb_)->lip_occurred_timestamp))
#define SECS_SINCE_NSCN_OCCURRED(_cb_) \
	(eee_rtcDurationSec(&(_cb_)->nscn_occurred_timestamp))

#if defined(ISPFC_SIM)
typedef int tape_dir_ent_t;
#endif

typedef struct ispfc_dev {
	uint32			phys_dev_id;
	int32			dev_handle;
	wwn_t			dev_nodename;
	wwn_t			dev_portname;
	wwn_t			dev_old_portname;
	wwn_t			channel_wwn;
	uint32			dev_generation;
	byte			mapping_id;
	uint64			last_accessed_timestamp;
	uint64			last_cached_in_timestamp;

	/* SCSI descriptor queues */
	scsi_desc_t *		scsi_desc_busy;
	scsi_desc_t *		scsi_desc_waiting;
	ushort16		dev_cmd_busy_count;

	/* Task management request queues */
	scsi_task_mgmt_t *	scsi_task_mgmt_busy;
	scsi_task_mgmt_t *	scsi_task_mgmt_waiting;

	uint32			dev_state;

	uint32                  dev_global_timeout_timer;
        byte                    dev_tmo_timer;
        byte                    dev_reset_timer;
        int32			paused_timer_handle;
        int32                   deferred_login_timer_handle;
	int32			queue_flush_timer_handle;
	byte			device_login_level;
	void *			port_database;
	void *			lfa_req;
	void *			lfa_resp;
	
	void *			cb;


#if defined(ISPFC_SIM) 
	
	char			scsi_vendor_id[8+1];
	char			scsi_product_id[16+1];
	char			scsi_revision[4+1];
	char			scsi_serialno[8+1];

	int			sim_dev_type; /* 0 == DISK, 1 == TAPE */

	/* For disks only */
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
	char			filename[128];
	int			fd;
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
	char *			ram_disk_buf;
#endif
        int			capacity;                /* In sectors */
	int			sector_size;

	/* For tapes only */
	uint64			max_bytes_to_eod;

	char			current_data_filename[128];
	int			current_data_fd;
	char			current_dir_filename[128];
	int			current_dir_fd;

	int			current_fileno;
	int			current_blockno;
	int			fixed_blocksize;
	int			compression_enabled;
	int			write_protected;
	uint8			bytes_at_eod;

	tape_dir_ent_t *	current_dir;
	int			current_dir_size;
	int			current_dir_alloced_size;
	int			current_dir_dirty;

#endif /* defined(ISPFC_SIM) */

	struct ispfc_dev *	next;
	struct ispfc_dev *	prev;
} ispfc_dev_t;

/*
 * phys_dev_id definitions
 */
#define ISPFC_DEV_PHYS_DEV_ID_MASK		0x00FFFFFF
#define ISPFC_DEV_PHYS_DEV_ID_MASK_HI		0x00FF0000
#define ISPFC_DEV_PHYS_DEV_ID_MASK_LO		0x0000FFFF
#define ISPFC_DEV_PHYS_DEV_TO_FLA(_d_)		(uint32)(((_d_) & 0x00FFFF00L))
#define ISPFC_DEV_PHYS_DEV_TO_ALPA(_d_)		(uint32)(((_d_) & 0x000000FFL))

/*
 * Device state
 */
#define ISPFC_DS_STATE_MASK			0x000000FF
#define ISPFC_DS_STATE_LOGIN_PENDING		0
#define ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING	1
#define ISPFC_DS_STATE_LOGIN_REQUESTED		2
#define ISPFC_DS_STATE_DEFERRED_LOGIN		3
#define ISPFC_DS_STATE_LOGOUT_PENDING		4
#define ISPFC_DS_STATE_LOGGED_OUT		5
#define ISPFC_DS_STATE_INVALIDATE_PENDING	6
#define ISPFC_DS_STATE_UP			7
#define ISPFC_DS_STATE_PAUSED		        8
#define ISPFC_DS_STATE_QUIESCED			9
#define ISPFC_DS_STATE_TMO_PENDING		10
#define ISPFC_DS_STATE_TMO_REQUESTED		11
#define ISPFC_DS_STATE_RESET_PENDING		12
#define ISPFC_DS_STATE_RESET_REQUESTED		13
#define ISPFC_DS_STATE_QUEUE_FLUSH_PENDING	14
#define ISPFC_DS_STATE_GPDB_PENDING		15
#define ISPFC_DS_STATE_ID_VERIFICATION_PENDING	16
#define ISPFC_DS_STATE_INVALID			17

#define ISPFC_DS_STATE_SUBSTATE_MASK		0x0000FF00
#define ISPFC_DS_STATE_SUBSTATE_SHIFT		16

#define ISPFC_DF_STATE_ACTION_FLAGS_MASK	0x00FF0000
#define ISPFC_DF_INVALIDATION_REQUESTED		0x00010000
#define ISPFC_DF_ID_VERIFICATION_REQUESTED	0x00020000
#define ISPFC_DF_LOGOUT_REQUESTED		0x00040000
#define ISPFC_DF_REINIT_REQUESTED		0x00080000
#define ISPFC_DF_REGISTRATION_REQUESTED		0x00100000

#define ISPFC_DF_STATE_PASSIVE_FLAGS_MASK	0xFF000000
#define ISPFC_DF_REGISTERED			0x01000000
#define ISPFC_DF_MAPPING_ID_ALLOCATED		0x02000000
#if defined(ISPFC_SIM)
#define ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE	0x04000000
#endif

#ifdef DEBUG
void    ispfc_dev_check_state_machine(ispfc_dev_t *, int);
#define CHECK_DEV_STATE(_dev_, _s_) \
        ispfc_dev_check_state_machine(_dev_, _s_)
#else
#define CHECK_DEV_STATE(_dev_, _s_)
#endif /* DEBUG */

#define DEV_STATE(_dev_) \
	((_dev_)->dev_state & ISPFC_DS_STATE_MASK)
#define SET_DEV_STATE(_dev_, _s_) \
	{ \
		CHECK_DEV_STATE(_dev_, _s_); \
		if (_s_ == ISPFC_DS_STATE_LOGIN_REQUESTED) { \
			int32 _rc_; \
			if (DEV_STATE(_dev_) == ISPFC_DS_STATE_DEFERRED_LOGIN && (_dev_)->deferred_login_timer_handle) { \
				_rc_ = eee_timerStop((_dev_)->deferred_login_timer_handle); \
				VERIFY(_rc_ == NFX_OK); \
				_rc_ = eee_timerDestroy((_dev_)->deferred_login_timer_handle); \
				VERIFY(_rc_ == NFX_OK); \
				(_dev_)->deferred_login_timer_handle = 0; \
			} \
			else \
			if (DEV_STATE(_dev_) == ISPFC_DS_STATE_PAUSED && (_dev_)->paused_timer_handle) { \
				_rc_ = eee_timerStop((_dev_)->paused_timer_handle); \
				VERIFY(_rc_ == NFX_OK); \
				_rc_ = eee_timerDestroy((_dev_)->paused_timer_handle); \
				VERIFY(_rc_ == NFX_OK); \
				(_dev_)->paused_timer_handle = 0; \
			} \
			else \
			if (DEV_STATE(_dev_) == ISPFC_DS_STATE_QUEUE_FLUSH_PENDING && (_dev_)->queue_flush_timer_handle) { \
				_rc_ = eee_timerStop((_dev_)->queue_flush_timer_handle); \
				VERIFY(_rc_ == NFX_OK); \
				_rc_ = eee_timerDestroy((_dev_)->queue_flush_timer_handle); \
				VERIFY(_rc_ == NFX_OK); \
				(_dev_)->queue_flush_timer_handle = 0; \
			} \
		} \
		(_dev_)->dev_state = ((_dev_)->dev_state & ~ISPFC_DS_STATE_MASK) | (_s_); \
	}

#define DEV_SUBSTATE(_dev_) \
	(((_dev_)->dev_state & ISPFC_DS_STATE_SUBSTATE_MASK) >> ISPFC_DS_STATE_SUBSTATE_SHIFT)
#define SET_DEV_SUBSTATE(_dev_, _ss_) \
	{ \
		(_dev_)->dev_state = ((_dev_)->dev_state & ~ISPFC_DS_STATE_SUBSTATE_MASK) | \
				(((_ss_) << ISPFC_DS_STATE_SUBSTATE_SHIFT) & ISPFC_DS_STATE_SUBSTATE_MASK); \
	}

/* 
 * Fabric map Mapping ID values
 */
#define FABRIC_MAP_UNMAPPED     0x00
#define FABRIC_MAP_MAPPED_TEMP  0x01
#define FABRIC_MAP_MAPPED_PERM  0xFE
#define FABRIC_MAP_UNUSABLE     0xFF

/* 
 * SCSI descriptor states
 */
#define SCSI_DESC_STATE_WAITING		0
#define SCSI_DESC_STATE_BUSY		1
#define SCSI_DESC_STATE_TIMED_OUT	2
#define SCSI_DESC_STATE_ABORTED		3

/*******************************************************************************
 * 
 * ISP 23xx Register definitions
 *
 * Address	Register
 * -------	--------------------------------------------------
 * 0x0000	ISPFC_REG_FLASH_ADDR   - Flash BIOS Address Register
 * 0x0002	ISPFC_REG_FLASH_DATA   - Flash BIOS Data Register
 * 0x0006	ISPFC_REG_CONSTAT      - ISP Control/Status register
 * 0x0008	ISPFC_REG_INTERRUPT    - ISP to PCI Interrupt Control Register
 * 0x000a	ISPFC_REG_INTSTAT      - ISP to PCI Interrupt Status Register
 * 0x000c	ISPFC_REG_SEMAPHORE    - ISP Semaphore Register
 * 0x000e	ISPFC_REG_NVRAM	       - ISP NVRAM Interface Register
 * 0x0010	ISPFC_REG_REQUEST_IN   - Request Queue In-Pointer Register
 * 0x0012	ISPFC_REG_REQUEST_OUT  - Request Queue Out-Pointer Register
 * 0x0014	ISPFC_REG_RESPONSE_IN  - Response Queue In-Pointer Register
 * 0x0016	ISPFC_REG_RESPONSE_OUT - Response Queue Out-Pointer Register
 * 0x0018	ISPFC_REG_R_TO_H_LOW   - RISC to Host Status Low Register
 * 0x001a	ISPFC_REG_R_TO_H_HI    - RISC to Host Status High Register
 * 0x00c0	ISPFC_REG_HCCR	       - Host Command and Control Register
 * 0x00cc	ISPFC_REG_GPIOD	       - GPIO Data Register
 * 0x00ce	ISPFC_REG_GPIOE	       - GPIO Enable Register
 * 0x0040	ISPFC_REG_MBOX0	       - Mailbox register 0
 * 0x0042	ISPFC_REG_MBOX1	       - Mailbox register 1
 * ...
 * ...
 * 0x007e	ISPFC_REG_MBOX31       - Mailbox register 31
 ******************************************************************************/

#define ISPFC_REG_FLASH_ADDR	0x00
#define ISPFC_REG_FLASH_DATA	0x02
#define ISPFC_REG_CONSTAT	0x06
#define ISPFC_REG_INTERRUPT	0x08
#define ISPFC_REG_INTSTAT	0x0a
#define ISPFC_REG_SEMAPHORE	0x0c
#define ISPFC_REG_NVRAM		0x0e

#define ISPFC_REG_REQUEST_IN	0x10
#define ISPFC_REG_REQUEST_OUT	0x12
#define ISPFC_REG_RESPONSE_IN	0x14
#define ISPFC_REG_RESPONSE_OUT	0x16
#define ISPFC_REG_R_TO_H_LOW	0x18
#define ISPFC_REG_R_TO_H_HI	0x1a
#define ISPFC_REG_R_TO_H	0x18	/* Hi and Low - Can be access as 32-bit quantity */

#define ISPFC_REG_HCCR		0xc0
#define ISPFC_REG_GPIOD		0xcc
#define ISPFC_REG_GPIOE		0xce

/* MAILBOX registers */
#define	ISPFC_REG_MBOX_BASE	0x40
#define ISPFC_REG_MBOX(_ii_)	(ISPFC_REG_MBOX_BASE + 2 * (_ii_))
#define ISPFC_REG_MBOX0		(ISPFC_REG_MBOX_BASE + 0x00)
#define ISPFC_REG_MBOX1		(ISPFC_REG_MBOX_BASE + 0x02)
#define ISPFC_REG_MBOX2		(ISPFC_REG_MBOX_BASE + 0x04)
#define ISPFC_REG_MBOX3		(ISPFC_REG_MBOX_BASE + 0x06)
#define ISPFC_REG_MBOX4		(ISPFC_REG_MBOX_BASE + 0x08)
#define ISPFC_REG_MBOX5		(ISPFC_REG_MBOX_BASE + 0x0a)
#define ISPFC_REG_MBOX6		(ISPFC_REG_MBOX_BASE + 0x0c)
#define ISPFC_REG_MBOX7		(ISPFC_REG_MBOX_BASE + 0x0e)
#define ISPFC_REG_MBOX8		(ISPFC_REG_MBOX_BASE + 0x10)
#define ISPFC_REG_MBOX9		(ISPFC_REG_MBOX_BASE + 0x12)
#define ISPFC_REG_MBOX10	(ISPFC_REG_MBOX_BASE + 0x14)
#define ISPFC_REG_MBOX11	(ISPFC_REG_MBOX_BASE + 0x16)
#define ISPFC_REG_MBOX12	(ISPFC_REG_MBOX_BASE + 0x18)
#define ISPFC_REG_MBOX13	(ISPFC_REG_MBOX_BASE + 0x1a)
#define ISPFC_REG_MBOX14	(ISPFC_REG_MBOX_BASE + 0x1c)
#define ISPFC_REG_MBOX15	(ISPFC_REG_MBOX_BASE + 0x1e)
#define ISPFC_REG_MBOX16	(ISPFC_REG_MBOX_BASE + 0x20)
#define ISPFC_REG_MBOX17	(ISPFC_REG_MBOX_BASE + 0x22)
#define ISPFC_REG_MBOX18	(ISPFC_REG_MBOX_BASE + 0x24)
#define ISPFC_REG_MBOX19	(ISPFC_REG_MBOX_BASE + 0x26)
#define ISPFC_REG_MBOX20	(ISPFC_REG_MBOX_BASE + 0x28)
#define ISPFC_REG_MBOX21	(ISPFC_REG_MBOX_BASE + 0x2a)
#define ISPFC_REG_MBOX22	(ISPFC_REG_MBOX_BASE + 0x2c)
#define ISPFC_REG_MBOX23	(ISPFC_REG_MBOX_BASE + 0x2e)
#define ISPFC_REG_MBOX24	(ISPFC_REG_MBOX_BASE + 0x30)
#define ISPFC_REG_MBOX25	(ISPFC_REG_MBOX_BASE + 0x32)
#define ISPFC_REG_MBOX26	(ISPFC_REG_MBOX_BASE + 0x34)
#define ISPFC_REG_MBOX27	(ISPFC_REG_MBOX_BASE + 0x36)
#define ISPFC_REG_MBOX28	(ISPFC_REG_MBOX_BASE + 0x38)
#define ISPFC_REG_MBOX29	(ISPFC_REG_MBOX_BASE + 0x3a)
#define ISPFC_REG_MBOX30	(ISPFC_REG_MBOX_BASE + 0x3c)
#define ISPFC_REG_MBOX31	(ISPFC_REG_MBOX_BASE + 0x3e)

/* CONSTAT register definitions */
#define CONSTAT_SOFT_RESET	0x0001	/* ISP23xx Soft Reset */
#define CONSTAT_FLASH_RW   	0x0002	/* Flash BIOS read/write enable */
#define CONSTAT_64BIT		0x0004	/* PCI 64-bit Bus Slot */
#define CONSTAT_FLASH_UPPER_SEL	0x0008	/* Flash Upper 64K bank select */
#define CONSTAT_MS_FB		0x0010	/* Frame buffer registers */
#define CONSTAT_MS_FPM0		0x0020	/* Fibre Protocol Module 0 registers */
#define CONSTAT_MS_FPM1		0x0030	/* Fibre Protocol Module 1 registers */

/* INTERRUPT register definitions */
#define INTERRUPT_XMT_DMA	0x0001	/* Enable transmit DMA interrupts */
#define INTERRUPT_RCV_DMA	0x0002	/* Enable receive DMA interrupts */
#define INTERRUPT_CMD_DMA	0x0004	/* Enable command DMA interrupts */
#define INTERRUPT_RISC		0x0008	/* Enable RISC processor interrupts */ 
#define INTERRUPT_FB		0x0010	/* Enable frame buffer interrupts */
#define INTERRUPT_FPM		0x0020	/* Enable FPM interrupts */
#define INTERRUPT_2300		0x8000	/* Enables interrupt types */

#define ENABLE_ISP_INTERRUPTS	(INTERRUPT_2300 | INTERRUPT_RISC)
#define DISABLE_ISP_INTERRUPTS	(0)

/* INTSTAT register definitions */
#define INTSTAT_XMT_DMA		0x0001	/* Transmit DMA interrupt request */
#define INTSTAT_RCV_DMA		0x0002	/* Receive DMA interrupt request */
#define INTSTAT_CMD_DMA		0x0004	/* Command DMA interrupt request */
#define INTSTAT_RISC		0x0008	/* RISC processor interrupt request */
#define INTSTAT_FB		0x0010	/* FB interrupt request */
#define INTSTAT_FPM		0x0020	/* FPM interrupt request */
#define INTSTAT_CODE_MASK	0x0F00	/* Interrupt status code mask */
#define INTSTAT_ALL		0x8000	/* Interrupt types */

/* SEMAPHORE register definitions */
#define SEMAPHORE_LOCK		0x0001	/* 1 sets lock, 0 clears it */
#define SEMAPHORE_STATUS	0x0002	/* Current semaphore status */

/* NVRAM interface register definitions */
#define NVRAM_CLOCK		0x0001	/* NVRAM Clock */
#define NVRAM_SELECT		0x0002	/* NVRAM Chip Select */
#define NVRAM_DATA_OUT		0x0004	/* NVRAM Data Out */
#define NVRAM_DATA_IN		0x0008	/* NVRAM Data In */
#define NVRAM_DESELECT		0x0000

/* R_TO_H_LOW register definitions */
#define R_TO_H_LOW_RISC_INT	0x8000	/* RISC to Host Interrupt Request */
#define R_TO_H_LOW_PAUSED	0x0100	/* RISC Paused */
#define R_TO_H_LOW_STATUS_MASK	0x00FF	/* Interrupt status mask */

#define R_TO_H_LOW_STATUS_ROM_SUCCESSFULL	0x01
#define R_TO_H_LOW_STATUS_ROM_FAILED		0x02
#define R_TO_H_LOW_STATUS_MBOX_SUCCESSFULL	0x10
#define R_TO_H_LOW_STATUS_MBOX_FAILED		0x11
#define R_TO_H_LOW_STATUS_ASYNC_EVENT		0x12
#define R_TO_H_LOW_STATUS_RESPONSE_UPDATED	0x13

/* HCCR register definitions */
#define HCCR_PAUSE		0x0020	/* RISC processor paused (R/O) */
#define HCCR_RESET		0x0040	/* RISC processor reset (R/O) */
#define HCCR_INTR		0x0080	/* Host Interrupt pending (R/O) */
#define HCCR_FB_PE		0x0100	/* Frame Buffer Parity Error (R/O) */
#define HCCR_FB_PE_MASK		0x0200	/* Frame Buffer Parity Error Mask (R/O) */
#define HCCR_SRAM_PARITY_ENABLE	0x0400	/* SRAM Parity Enabled (R/O) */	
#define HCCR_SRAM_PARITY_ERROR	0x0800	/* SRAM Parity Error (R/O) */	
#define HCCR_COMMAND_MASK	0xF000	/* HCCR Command Field */

/* HCCR commands */
#define HCCR_CMD_NOP		0x0000	/* HOP */
#define HCCR_CMD_RESET		0x1000	/* Reset RISC */
#define HCCR_CMD_PAUSE		0x2000	/* Pause RISC */
#define HCCR_CMD_RELEASE	0x3000	/* Release RISC */
#define HCCR_CMD_MASK_FB_PARITY	0x4000	/* Mask Frame Buffer Parity */
#define HCCR_CMD_SET_HOST_INTR	0x5000	/* Set Host Interrupt (to RISC) */
#define HCCR_CMD_CLR_HOST_INTR	0x6000	/* Clear Host Interrupt (to RISC) */
#define HCCR_CMD_CLR_RISC_INTR	0x7000	/* Clear RISC Interrupt (to Host) */
#define HCCR_CMD_WRT_PAR_ENABLE 0xa000  /* Enable/Disable SRAM parity if bit-0 1/0 */
#define HCCR_CMD_FRC_PARITY_ERR 0xe000	/* Force SRAM parity error */

/*
 * Pre-allocated loop IDs -- See Qlogic document titled "ISP2100 Firmware
 * Interface Specification Addendum: Fabric Support", dated 07/22/98
 */
#define ISPFC_FABRIC_PORT_PORTID	0x7E    /* Internal mapping of FC portid 0xfffffe */
#define ISPFC_FABRIC_CONTROLLER_PORTID	0x7F    /* Internal mapping of FC portid 0xfffffd */
#define ISPFC_DIRECTORY_SERVER_PORTID	0x80    /* Internal mapping of FC portid 0xfffffc */

/*
 * Additionally, the driver manages the Management server mapping.
 */
#define ISPFC_MANAGEMENT_SERVER_PORTID	0x81	/* Driver managed mapping of FC portid 0xfffffa */

#include "ispfc_misc.h"

/* Register access macros */
#ifndef DEBUG
#define ISPFC_REG_READ(_cb_, _offset_) \
		(*(volatile ushort16 *)((_cb_)->reg_mem_base + (_offset_)))
#define ISPFC_REG_WRITE(_cb_, _offset_, _value_) \
		(*(volatile ushort16 *)((_cb_)->reg_mem_base + (_offset_)) = (_value_))
#else
#define ISPFC_REG_READ(_cb_, _offset_) \
		_ispfc_reg_read((_cb_), (_offset_))
#define ISPFC_REG_WRITE(_cb_, _offset_, _value_) \
		_ispfc_reg_write((_cb_), (_offset_), (_value_))
#endif /* DEBUG */

#ifdef EVAL_BOARD
#if 0	/* For Galileo LITTLE_ENDIAN eval board; #if 1 for BIG_ENDIAN board */
#undef LITTLE_ENDIAN
#define BIG_ENDIAN_EVAL_BOARD
#else
#define LITTLE_ENDIAN 1	
#define LITTLE_ENDIAN_EVAL_BOARD
#endif
#endif

#ifdef AGILE_HARDWARE
#define LITTLE_ENDIAN 1	
#endif

/*
 * Big-endian to Little-endian in-place swap routines
 */
#ifdef LITTLE_ENDIAN
#define inplace_ntohll(_p_)	SWAP8(_p_)
#define inplace_ntohl(_p_)	SWAP4(_p_)
#define inplace_ntohs(_p_)	SWAP2(_p_)
#define inplace_htonll(_p_)	SWAP8(_p_)
#define inplace_htonl(_p_)	SWAP4(_p_)
#define inplace_htons(_p_)	SWAP2(_p_)
#else
#define inplace_ntohll(_p_)
#define inplace_ntohl(_p_)
#define inplace_ntohs(_p_)
#define inplace_htonll(_p_) 
#define inplace_htonl(_p_)
#define inplace_htons(_p_)
#endif /* LITTLE_ENDIAN */

#define SWAP2(_p_) \
	{ \
		  *((ushort16 *)(_p_)) = (((*((ushort16 *)(_p_)) >>  8) & 0x00FF) | \
				          ((*((ushort16 *)(_p_)) <<  8) & 0xFF00)); \
	}
#define SWAP3(_p_) \
	{ \
	  byte _t_ = *(byte *)(_p_); \
	  *(byte *)(_p_) = *((byte *)(_p_)+2); \
	  *((byte *)(_p_)+2) = _t_; \
	}
#define SWAP4(_p_) \
	{ \
		  *((uint32 *)(_p_)) = (((*((uint32 *)(_p_)) >> 24) & 0x000000FFL) | \
				        ((*((uint32 *)(_p_)) >>  8) & 0x0000FF00L) | \
				        ((*((uint32 *)(_p_)) <<  8) & 0x00FF0000L) | \
				        ((*((uint32 *)(_p_)) << 24) & 0xFF000000L)); \
	}
#define SWAP8(_p_) \
	{ \
		  *((uint64 *)(_p_)) = (((*((uint64 *)(_p_)) >> 56) & 0x00000000000000FFLL) | \
				        ((*((uint64 *)(_p_)) >> 40) & 0x000000000000FF00LL) | \
				        ((*((uint64 *)(_p_)) >> 24) & 0x0000000000FF0000LL) | \
				        ((*((uint64 *)(_p_)) >>  8) & 0x00000000FF000000LL) | \
				        ((*((uint64 *)(_p_)) <<  8) & 0x000000FF00000000LL) | \
				        ((*((uint64 *)(_p_)) << 24) & 0x0000FF0000000000LL) | \
				        ((*((uint64 *)(_p_)) << 40) & 0x00FF000000000000LL) | \
				        ((*((uint64 *)(_p_)) << 56) & 0xFF00000000000000LL)); \
	}

#endif /* __ISPFC_PRIVATE_H__ */
