/*
 *-----------------------------------------------------------------
 *
 * Name:	transport-drv.c
 *
 * RCS:  $Id: transport-drv.c,v 1.16 2002/10/08 03:16:18 johnh Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	For communication between simulated cards in the Linux Test
 *				environment, we'll define a driver that uses shared memory
 *				and semaphores to pass messages.  The driver should be
 *				registered as the transport for intercard messages when
 *				LINUX_TEST is defined.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/28/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>

#ifdef LINUX_TEST
#ifndef SHMEM
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <sys/un.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <errno.h>

#include <exports/cm-api.h>
#include <exports/eee-api.h>
#include <exports/spinlock-api.h>
#include "../sm-eee/eee.h"
#include "../sm-eee/eee-platform.h"

int32 linux_mp_sem[NFX_NUM_SLOTS_PER_CHASSIS][NFX_MAX_CPUS_PER_SLOT] = {{-1}};
int32 linux_dp_sem[NFX_NUM_SLOTS_PER_CHASSIS][NFX_MAX_CPUS_PER_SLOT] = {{-1}};
int32 linux_fwd_sem[NFX_NUM_SLOTS_PER_CHASSIS][NFX_MAX_CPUS_PER_SLOT] = {{-1}};
shm_tx_t *	linux_mp[NFX_NUM_SLOTS_PER_CHASSIS][NFX_MAX_CPUS_PER_SLOT];
shm_tx_t *	linux_dp[NFX_NUM_SLOTS_PER_CHASSIS][NFX_MAX_CPUS_PER_SLOT];


#define SHM_MSG_OWN		0x80000000
#define SHM_MSG_DONE	0x40000000

/*-----------------------------------------------------------------
 * Name:	eee_linuxForwardReceive()
 *
 * Description:	Polling function to receive packets from shared memory
 *				queue between Linux applications.  This replaces the
 *				backplane drivers for the Linux simulation environment.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	7/1/01
 *
 *-----------------------------------------------------------------
 */
void
eee_linuxForwardReceive(
	void	* cb,
	uint32	tref
)
{
	uint32			slot, cpu;
	int32			sid;
	struct sembuf	sem;
	shm_tx_t		* cptr;
	uchar8			* tx, * rx, * start, * end;
	uint32			avail;
	uint32			d;
	uint32			pkt_len;
	uint32			dp, sp, attr;
	uint32			i, j;
	uint32			bytes_removed;
	eee_descPtr_t	edesc;
	uchar8			* buf[EEE_MAX_BUFS];
	uint32			size[EEE_MAX_BUFS];
	uint32			len[EEE_MAX_BUFS];
	uint32			bytes_remain;
	eee_descBuf_t	* bd;
	uint32			num_bytes;
	uchar8			* rx_base;
	enet_hdr_t		* enet_hdr;
	agile_hdr_t		* agile_hdr;
    int             data;
    extern int      eee_linuxFwdPollRxFd;
	uint32			tot_bytes_removed = 0;
	uint32			cnt = 0;

	slot = nfxMySlotId;
	cpu = nfxMyCPUId;

	if ((uint32)cb == 0) {			/* data plane */
#if defined(NFP_TXRX)
		sid = linux_dp_sem[slot][0];
		cptr = linux_dp[slot][0];
#elif defined(NFP_FP)
		sid = linux_dp_sem[slot][1];
		cptr = linux_dp[slot][1];
#else
		sid = linux_dp_sem[slot][cpu];
		cptr = linux_dp[slot][cpu];
#endif
	}
	else if ((uint32)cb == 1) {		/* mgmt plane */
		sid = linux_mp_sem[slot][cpu];
		cptr = linux_mp[slot][cpu];
	}
	else
		panic("Invalid cb = %d\n", cb);

	/*
	 * Get Semaphore for shared memory area into which we will write
	 * the data to send to the remote card.
	 */
	SPINLOCK_GET(sid);

	if (cptr->tx_offset == 0)
		goto rx_release_sem;

	tx = (uchar8 *)cptr + cptr->tx_offset;
	rx = (uchar8 *)cptr + cptr->rx_offset;
	start = (uchar8 *)cptr + cptr->start_offset;
	end = (uchar8 *)cptr + cptr->end_offset;
	avail = cptr->tx_bytes_avail;

	/*
	 * Read data from queue, and formulate EEE packet
	 */
	while ((d = *(uint32 *)rx) & SHM_MSG_OWN) {
		if (cnt++ >= 20)
			break;
/*
		printf("RX GOT SEMAPHORE\n");
		printf("shared memory control structure at 0x%x\n", cptr);
		printf("tx = 0x%x, rx = 0x%x, start = 0x%x, end = 0x%x, avail = 0x%x\n",
											tx, rx, start, end, avail);
*/

		rx_base = rx;

		bytes_removed = 0;
		pkt_len = d & 0x00FFFFFF;
#ifdef FIX_HDR_STUFF
		enet_hdr = (enet_hdr_t *)(rx + sizeof(uint32));
		agile_hdr = (agile_hdr_t *)(rx + sizeof(uint32) + sizeof(enet_hdr_t));
		rx = rx + sizeof(uint32) + sizeof(enet_hdr_t) + sizeof(agile_hdr_t);
		if (rx == end)
			rx = start;
		dp = agile_hdr->dest_port;
		sp = agile_hdr->src_port;
#else
		dp = *++(uint32 *)rx;
		sp = *++(uint32 *)rx;
/*
		attr = *++(uint32 *)rx;
*/
		rx = (uchar8 *)((uint32 *)rx+1);
#endif

		bytes_remain = pkt_len;

		edesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
		if (!edesc)
			goto rx_release_sem;

		for (i=0; i<EEE_MAX_BUFS; i++) {

			if (bytes_remain == 0)
				break;

			if (bytes_remain <= eee.buf_size[EEE_BUF_SMALL] - eee.buf_offset) {
				if (buf[i] = eee_allocateBuffer(EEE_SHARED_POOL,
												EEE_BUF_SMALL,
												NFX_COS_HIGH)) {
					size[i] = EEE_BUF_SMALL;
					len[i] = bytes_remain;
					bytes_remain = 0;
					continue;
				}
			}
			if (bytes_remain <= eee.buf_size[EEE_BUF_LARGE] - eee.buf_offset) {
				if (buf[i] = eee_allocateBuffer(EEE_SHARED_POOL,
												EEE_BUF_LARGE,
												NFX_COS_HIGH)) {
					size[i] = EEE_BUF_LARGE;
					len[i] = bytes_remain;
					bytes_remain = 0;
					continue;
				}
			}

			if (buf[i] = eee_allocateBuffer(EEE_SHARED_POOL,
											EEE_BUF_LARGE,
											NFX_COS_HIGH)) {
				size[i] = EEE_BUF_LARGE;
				len[i] = eee.buf_size[EEE_BUF_LARGE] - eee.buf_offset;
				bytes_remain -= (eee.buf_size[EEE_BUF_LARGE] - eee.buf_offset);
			}
			else if (buf[i] = eee_allocateBuffer(EEE_SHARED_POOL,
											EEE_BUF_SMALL,
											NFX_COS_HIGH)) {
				size[i] = EEE_BUF_SMALL;
				len[i] = eee.buf_size[EEE_BUF_SMALL] - eee.buf_offset;
				bytes_remain -= (eee.buf_size[EEE_BUF_SMALL] - eee.buf_offset);
			}
			else {
				printf("Cannot allocate buffers\n");
				eee_deallocateDesc(edesc);
				for (j=0; j<i; j++)
					eee_deallocateBuffer(buf[j]);
				goto rx_release_sem;
			}
		}

		if (bytes_remain) {
			eee_deallocateDesc(edesc);
			for (j=0; j<EEE_MAX_BUFS; j++)
				eee_deallocateBuffer(buf[j]);
			goto rx_release_sem;
		}

		if (i == EEE_MAX_BUFS) {
			edesc->hdr.control =	EDESC_COMMAND_USE_FIRST_BUFFER |
									(i << EDESC_BUFCNT_SHIFT) |
									pkt_len;
			bd = &(edesc->bd[0]);
		}
		else {
			edesc->hdr.control =	(i << EDESC_BUFCNT_SHIFT) |
									pkt_len;
			bd = &(edesc->bd[1]);
		}
		edesc->hdr.dest_port = dp;
		edesc->hdr.src_port = sp;
/*
		edesc->hdr.attr = attr;
*/

		for (j=0; j<i; bd++, j++) {
			bd->control = EDESC_BD_CMD_IN_USE | size[j];
			bd->len = len[j];
			bd->buf = buf[j];
			for (num_bytes = 0; num_bytes < len[j]; num_bytes++) {
				bd->buf[num_bytes] = *rx;
				if (++rx == end)
					rx = start;
			}
		}
		rx += 31;
		(uint32)rx &= ~31;
		if (rx == end)
			rx = start;

#ifdef FIX_HDR_STUFF
		bytes_removed = pkt_len + sizeof(uint32) +
									sizeof(enet_hdr_t) +
									sizeof(agile_hdr_t);
#else
		bytes_removed = pkt_len + (sizeof(uint32) * 3);	/* own, dp, sp,/attr/ */
#endif
		bytes_removed += 31;
		bytes_removed &= ~31;

		avail += bytes_removed;

		for (i=0; i<bytes_removed; i++) {
			*rx_base = 0;
			if (++rx_base == end)
				rx_base = start;
		}

		SPINLOCK_RELEASE(sid);

		eee_receivePacket(edesc);

		SPINLOCK_GET(sid);

		cptr->tx_bytes_avail += bytes_removed;

#ifdef LINUX_POLL_NONBLOCKING_FD
        // Drain a byte from our poll indicator socket.
        // Ignore the return value, there may be nothing
        // to read and its set to NONBLOCK.
        read(eee_linuxFwdPollRxFd, &data, 1);
#endif
	}

	cptr->rx_offset = rx - (uchar8 *)cptr;

/*
	tx = (uchar8 *)cptr + cptr->tx_offset;
	rx = (uchar8 *)cptr + cptr->rx_offset;
	start = (uchar8 *)cptr + cptr->start_offset;
	end = (uchar8 *)cptr + cptr->end_offset;
	avail = cptr->tx_bytes_avail;
	printf("tx = 0x%x, rx = 0x%x, start = 0x%x, end = 0x%x, avail = 0x%x\n",
											tx, rx, start, end, avail);

	printf("RELEASED SEMAPHORE\n");
*/

rx_release_sem:
	SPINLOCK_RELEASE(sid);

}	/* eee_liuxForwardReceive() */

/*-----------------------------------------------------------------
 * Name:	eee_linuxForwardRemotePktUni()
 *
 * Description:	transmit function for Linux simulation environment for
 *				packets that will ultimately be transmitted over backplane
 *				switch fabric.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/28/01
 *
 *-----------------------------------------------------------------
 */
int32
eee_linuxForwardRemotePktUni(
	uint32		dest_port,
	eee_desc_t	* edesc
)
{
	uint32			slot, cpu, dtype;
	uint32			size;
	int32			sid;
	struct sembuf	sem;
	uint32			i, j;
	shm_tx_t		* cptr;
	uchar8			* tx, * rx, * start, * end;
	uint32			* tx_int;
	uint32			avail;
	uint32			pkt_len, tot_len, buf_len, num_bufs;
	eee_descBuf_t	* bd;
	int32			rc = EEE_FORWARDED_PACKET;
	enet_hdr_t		* enet_hdr;
	agile_hdr_t		* agile_hdr;
    int             data;
    extern int      eee_linuxFwdPollTxFd[NFX_SLOT_MAX][NFX_MAX_CPUS_PER_SLOT];
    extern int      eee_linuxFwdPollTxFdState[NFX_SLOT_MAX]
												[NFX_MAX_CPUS_PER_SLOT];

/*
	printf("Remote Uni, edesc = 0x%x\n", edesc);
*/

	slot = EEE_GET_SLOT_NUM(dest_port);
	cpu = EEE_GET_DEST_CPU(dest_port);
	dtype = EEE_GET_DTYPE(dest_port);

	if ((slot == NFX_SLOTTYPE_SSC) && (cpu == NFX_SSC_MGMT))
		printf("WARNING: Cannot send traffic to SSC_MGMT cpu, please fix\n");

	if ((nfxMySlotId == NFX_SLOTTYPE_SSC) && (nfxMyCPUId == NFX_SSC_MGMT))
		printf("WARNING: Cannot send traffic from SSC_MGMT cpu, please fix\n");

	if ((dtype == EEE_USER_DATA) || (dtype == EEE_CONTROL)) {
		if (nfxMySlotId == NFX_SLOTTYPE_SSC) {
			printf("WARNING: Cannot send data plane pkt from SSC, please fix\n");
			dtype = EEE_MGMT;
		}
		if (slot == NFX_SLOTTYPE_SSC) {
			printf("WARNING: Cannot send data plane pkt to SSC, please fix\n");
			dtype = EEE_MGMT;
		}
	}

	if ((dtype == EEE_USER_DATA) || (dtype == EEE_CONTROL)) {
		if (slot == NFX_SLOTTYPE_SSC)
			panic("Invalid data type to slot 0\n");
		else if (slot == NFX_SLOTTYPE_NFP) {
			if ((cpu == NFX_NFP_TXRX_ANY) ||
				(cpu == NFX_NFP_TXRX1) ||
				(cpu == NFX_NFP_TXRX2)) {
				sid = linux_dp_sem[1][0];
				cptr = linux_dp[1][0];
			}
			else if ((cpu == NFX_NFP_FP_ANY) ||
					(cpu == NFX_NFP_FP1) ||
					(cpu == NFX_NFP_FP2)) {
					sid = linux_dp_sem[1][1];
					cptr = linux_dp[1][1];
			}
			else
				panic("Invalid CPU identifier, %d, to slot 1\n", cpu);
		}
		else {
			sid = linux_dp_sem[slot][cpu];
			cptr = linux_dp[slot][cpu];
		}
		size = SHM_SEG_DP_SIZE;
	}
	else if (dtype == EEE_MGMT) {
		if (slot == NFX_SLOTTYPE_NFP) {
			if (cpu >= NFX_NFP_NUM_CPUS)
				panic("Invalid MGMT CPU identifier, %d, to slot 1\n", cpu);
		}
		sid = linux_mp_sem[slot][cpu];
		cptr = linux_mp[slot][cpu];
		size = SHM_SEG_MP_SIZE;
		if (EEE_GET_DTYPE(edesc->hdr.src_port) != EEE_MGMT)
			printf("WARNING: Sending pkt on MP, but src_port dtype != MP, please fix\n");
	}
	else
		panic("Invalid data type to eee_linuxForwardRemotePktUni, dtype = %d\n",
																		dtype);

	if (sid < 0)
		panic("Invalid sid < 0\n");

	/*
	 * Get Semaphore for shared memory area into which we will write
	 * the data to send to the remote card.
	 */
	SPINLOCK_GET(sid);

	/*
	 * Initialize control structure
 	 */
	if (cptr->tx_offset == 0) {
		cptr->start_offset = 0x20;
		cptr->tx_offset = cptr->start_offset;
		cptr->rx_offset = cptr->start_offset;
		cptr->end_offset = size;
		cptr->tx_bytes_avail = size - 0x20;
	}

/*
	printf("shared memory control structure at 0x%x\n", cptr);
*/
	tx = (uchar8 *)cptr + cptr->tx_offset;
	rx = (uchar8 *)cptr + cptr->rx_offset;
	start = (uchar8 *)cptr + cptr->start_offset;
	end = (uchar8 *)cptr + cptr->end_offset;
	avail = cptr->tx_bytes_avail;
/*
	printf("tx = 0x%x, rx = 0x%x, start = 0x%x, end = 0x%x, avail = 0x%x\n",
											tx, rx, start, end, avail);
*/

	/*
	 * Validate enough space for new message
	 */
	pkt_len = edesc->hdr.control & EDESC_TOTAL_LEN;
#ifdef FIX_HDR_STUFF
	tot_len = pkt_len + sizeof(uint32) +
						sizeof(enet_hdr_t) +
						sizeof(agile_hdr_t);
#else
	tot_len = pkt_len + (sizeof(uint32) * 3);	/* own, dp, sp, /attr/ */
#endif
	tot_len += 31;
	tot_len &= ~31;		/* round to next 32 byte boundary */
/*
	printf("check space, len = 0x%x, bytes_avail = 0x%x\n", tot_len, avail);
*/
	if (tot_len > avail) {
printf("returning EEE_TX_QUEUE_FULL, tot_len = %d, avail = %d\n", tot_len, avail);
		rc = EEE_TX_QUEUE_FULL;
		goto tx_release_sem;
	}

	/*
	 * Write data from descriptor into queue
	 */
/*
	printf("add bytes to queue, tx at 0x%x, rx at 0x%x\n", tx, rx);
*/

	tx_int = (uint32 *)tx;
#ifdef FIX_HDR_STUFF
	*tx_int++ = SHM_MSG_OWN | pkt_len;
	tx = (uchar8 *)tx_int;
	enet_hdr = (enet_hdr_t *)tx;
	agile_hdr = (agile_hdr_t *)(tx + sizeof(enet_hdr_t));
	tx = tx + sizeof(enet_hdr_t) + sizeof(agile_hdr_t);
	if (sizeof(uint32) + sizeof(enet_hdr_t) + sizeof(agile_hdr_t) > 32)
		panic("header does not fit in 32 bytes\n");
	if (tx == end)
		tx = start;
	agile_hdr->dest_port = edesc->hdr.dest_port;
	agile_hdr->src_port = edesc->hdr.src_port;
#else
	*tx_int++ = SHM_MSG_OWN | pkt_len;
	*tx_int++ = edesc->hdr.dest_port;
	*tx_int++ = edesc->hdr.src_port;
/*
	*tx_int++ = edesc->hdr.attr;
*/
	tx = (uchar8 *)tx_int;
#endif

	bd = &(edesc->bd[0]);
	if (!(edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER))
		bd++;

	num_bufs = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
	for (i=0; i<num_bufs; i++, bd++) {
		buf_len = bd->len;
		for (j=0; j<buf_len; j++) {
			*tx = bd->buf[j];
			if (++tx == end)
				tx = start;
		}
	}

	cptr->tx_bytes_avail -= tot_len;

	(uint32)tx += 31;
	(uint32)tx &= ~31;
	if (tx == end)
		tx = start;

	cptr->tx_offset = tx - (uchar8 *)cptr;

	tx = (uchar8 *)cptr + cptr->tx_offset;
	rx = (uchar8 *)cptr + cptr->rx_offset;
	start = (uchar8 *)cptr + cptr->start_offset;
	end = (uchar8 *)cptr + cptr->end_offset;
	avail = cptr->tx_bytes_avail;
/*
	printf("tx = 0x%x, rx = 0x%x, start = 0x%x, end = 0x%x, avail = 0x%x\n",
											tx, rx, start, end, avail);
*/
#ifdef LINUX_POLL_NONBLOCKING_FD
    if (eee_linuxFwdPollTxFd[slot][cpu] == -1)
    {
        // Create socket for dest slot/cpu.
        eee_linuxFwdPollTxFd[slot][cpu] = socket(AF_UNIX, SOCK_DGRAM, 0);
        if (eee_linuxFwdPollTxFd[slot][cpu] < 0)
        {
            perror("shm transport socket");
            panic("failed to create pollInd socket");
        }
        if (fcntl(eee_linuxFwdPollTxFd[slot][cpu], F_SETFL, O_NONBLOCK))
        {
            perror("shm transport fcntl");
            panic("socket fcntl failed");
        }
    }
	if (eee_linuxFwdPollTxFdState[slot][cpu] == -1)
	{
        struct sockaddr_un      sockAddr;
        char                    pathName[40];

		// Try to connect to destination.  If this fails,
		// don't panic, we'll just retry next time.  It could
		// be that the destination isn't up yet.
        sockAddr.sun_family = AF_UNIX;
        sprintf(pathName, "/tmp/EEE_SLOT_%d_CPU_%d", slot, cpu);
        strcpy(sockAddr.sun_path, pathName);
        if (connect(eee_linuxFwdPollTxFd[slot][cpu],
						(struct sockaddr *) &sockAddr, sizeof(sockAddr)) < 0)
        {
			/* hush this
            perror("linuxpoll connect");
            printf(
			"failed to connect pollInd socket slot %d cpu %d will try later",
				slot, cpu);
			*/
        }
		else
		{
			eee_linuxFwdPollTxFdState[slot][cpu] = 1;
		}
	}
	else
	{
        // Disregard return value, it could be WOULD_BLOCK.
    	write(eee_linuxFwdPollTxFd[slot][cpu], &data, 1);
	}
#endif

tx_release_sem:
	/*
	 * Release Semaphore
	 */
	SPINLOCK_RELEASE(sid);

	if (rc == EEE_FORWARDED_PACKET)
		eee_freePacket(edesc);

	return rc;

}	/* eee_linuxForwardRemotePktUni() */

/*-----------------------------------------------------------------
 * Name:	eee_linuxForwardRemotePktMulti()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/28/01
 *
 *-----------------------------------------------------------------
 */
int32
eee_linuxForwardRemotePktMulti(
	uint32		dest_port,
	eee_desc_t	* edesc
)
{
	printf("Remote Multi, edesc = 0x%x\n", edesc);


}	/* eee_linuxForwardRemotePktMulti() */

#endif	/* !SHMEM */
#endif	/* LINUX_TEST */

