/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:         lock-hash.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description:  Byte range lock clients and owners hashing functions.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */

#include <exports/vs-hash-api.h>
#include "lock-incls.h"

/*-----------------------------------------------------------------
 * Name:         lock_cifsLockOwnerHashFunc
 *
 * Description:
 *
 *     Generic hashing function for hash table of CIFS lock owner.
 *     Called to return hash key.
 *
 * Arguments:
 *
 *     keyp - Pointer to CIFS owner hash key, which is combination of
 *            Tid, Fid and Pid.
 *
 * Return Value:
 *
 *     Hash value.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */
uint32
lock_cifsLockOwnerHashFunc(void *keyp)
{
	uint32					hashval;
	lock_htCifsOwnerKey_t	*cifsOwnerKeyp;

	cifsOwnerKeyp = (lock_htCifsOwnerKey_t *)keyp;
	hashval = (cifsOwnerKeyp->tid +
			   cifsOwnerKeyp->fid +
			   cifsOwnerKeyp->pid) % LOCK_HASH_CIFS_OWNER_BUCKETS;

	return (hashval);
} /* lock_cifsLockOwnerHashFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_cifsLockOwnerHashCompare
 *
 * Description:
 *
 *     Comparison function for hash table of CIFS lock owners.
 *     Called to check for a CIFS lock owner match.
 *
 * Arguments:
 *
 *     htOwnerIdx - Index to a CIFS lock owner, which is stored in the
 *                  hash table.
 *
 *     searchOwnerIdx - Index to a CIFS lock owner, which contains
 *                      CIFS owner key to compare.
 *
 * Return Value:
 *
 *     0 - Match.
 *
 *     1 - No match.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_cifsLockOwnerHashCompare(vs_mem_index_t	htOwnerIdx,
							  vs_mem_index_t	searchOwnerIdx)
{
	lock_cifsLockOwner_t	*htOwnerp;
	lock_cifsLockOwner_t	*searchOwnerp;

	htOwnerp = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(htOwnerIdx);
	if (htOwnerp == NULL)
		return (1);

	searchOwnerp = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(searchOwnerIdx);
	if (searchOwnerp == NULL)
		return (1);

	if ((htOwnerp->tid == searchOwnerp->tid) &&
		(htOwnerp->fid == searchOwnerp->fid) &&
		(htOwnerp->pid == searchOwnerp->pid))
		return (0);
	else
		return (1);
} /* lock_cifsLockOwnerHashCompare() */

/*-----------------------------------------------------------------
 * Name:         lock_cifsLockOwnerHashRelease
 *
 * Description:
 *
 *     Release function for hash table of CIFS lock owners.
 *     Called to destroy the entire hash table.
 *
 * Arguments:
 *
 *     htOwnerIdx - Index to a CIFS lock owner stored in the hash table.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_cifsLockOwnerHashRelease(vs_mem_index_t htOwnerIdx)
{
	/* Nothing to do */
	return;
} /* lock_cifsLockOwnerHashRelease() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmLockClientHashFunc
 *
 * Description:
 *
 *     Generic hashing function for hash table of NLM lock client.
 *     Called to return hash key.
 *
 * Arguments:
 *
 *     keyp - Pointer to NLM client hash key, which is the client
 *            IP address.
 *
 * Return Value:
 *
 *     Hash value.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */
uint32
lock_nlmLockClientHashFunc(void *keyp)
{
	uint32					hashval;
	lock_htNlmClientKey_t	*nlmClientIpAddrp;
	
	nlmClientIpAddrp = (lock_htNlmClientKey_t *)keyp;
	hashval = ((uint32)nlmClientIpAddrp->clientIpAddr) %
		   									LOCK_HASH_NLM_CLIENT_BUCKETS;
	/*TODO: Change the hash value calculation after ip_index_t is defined*/

	return (hashval);
} /* lock_nlmLockClientHashFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmLockClientHashCompare
 *
 * Description:
 *
 *     Comparison function for hash table of NLM lock clients.
 *     Called to check for an NLM lock client match.
 *
 * Arguments:
 *
 *     htClientIdx - Index to an NLM lock client, which is stored in the
 *                   hash table.
 *
 *     searchClientIdx - Index to an NLM lock client, whick contains
 *                       NLM client IP address to compare.
 *
 * Return Value:
 *
 *     0 - Match.
 *
 *     1 - No match.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_nlmLockClientHashCompare(vs_mem_index_t	htClientIdx,
							  vs_mem_index_t	searchClientIdx)
{
	lock_nlmLockClient_t		*htClientp;
	lock_nlmLockClient_t		*searchClientp;

	htClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(htClientIdx);
	if (htClientp == NULL)
		return (1);

	searchClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(searchClientIdx);
	if (searchClientp == NULL)
		return (1);

	if (htClientp->clientIpAddr == searchClientp->clientIpAddr)
		return (0);
	else
		return (1);
} /* lock_nlmLockClientHashCompare() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmLockClientHashRelease
 *
 * Description:
 *
 *     Release function for hash table of NLM lock clients.
 *     Called to destroy the entire hash table.
 *
 * Arguments:
 *
 *     htClientIdx - Index to an NLM lock client stored in the hash table.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_nlmLockClientHashRelease(vs_mem_index_t htClientIdx)
{
	/* Nothing to do */
	return;
} /* lock_nlmLockClientHashRelease() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmLockOwnerHashFunc
 *
 * Description:
 *
 *     Generic hashing function for hash table of NLM lock owner.
 *     Called to return hash key.
 *
 * Arguments:
 *
 *     keyp - Pointer to NLM owner hash key, which is combination of
 *            NLM client IP address and owner Pid.
 *
 *
 * Return Value:
 *
 *     Hash value.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
uint32
lock_nlmLockOwnerHashFunc(void *keyp)
{
	uint32					hashval;
	lock_htNlmOwnerKey_t	*nlmOwnerKeyp;

	nlmOwnerKeyp = (lock_htNlmOwnerKey_t *)keyp;
	hashval = (((uint32)nlmOwnerKeyp->clientIpAddr) + nlmOwnerKeyp->pid) %
												LOCK_HASH_NLM_OWNER_BUCKETS;
	/*TODO: Change the hash value calculation after ip_index_t is defined*/

	return (hashval);
} /* lock_nlmLockOwnerHashFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmLockOwnerHashCompare
 *
 * Description:
 *
 *     Comparison function for hash table of NLM lock owners.
 *     Called to check for an NLM lock owner match.
 *
 * Arguments:
 *
 *     htOwnerIdx - Index to an NLM lock owner, which is stored in the
 *                  hash table.
 *
 *     searchOwnerIdx - Index to an NLM lock owner, from whick we can get
 *                      NLM client IP address and owner Pid to compare.
 *
 * Return Value:
 *
 *     0 - Match.
 *
 *     1 - No match.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_nlmLockOwnerHashCompare(vs_mem_index_t	htOwnerIdx,
							 vs_mem_index_t	searchOwnerIdx)
{
	lock_nlmLockOwner_t		*htOwnerp;
	lock_nlmLockOwner_t		*searchOwnerp;
	lock_nlmLockClient_t	*htClientp;
	lock_nlmLockClient_t	*searchClientp;

	htOwnerp = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(htOwnerIdx);
	if (htOwnerp == NULL)
		return (1);

	htClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(htOwnerp->lockClient);
	if (htClientp == NULL)
		return (1);

	searchOwnerp = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(searchOwnerIdx);
	if (searchOwnerp == NULL)
		return (1);

	searchClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(searchOwnerp->lockClient);
	if (searchClientp == NULL)
		return (1);

	if ((htClientp->clientIpAddr == searchClientp->clientIpAddr) &&
		(htOwnerp->pid == searchOwnerp->pid))
		return (0);
	else
		return (1);
} /* lock_nlmLockOwnerHashCompare() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmLockOwnerHashRelease
 *
 * Description:
 *
 *     Release function for hash table of NLM lock owners.
 *     Called to destroy the entire hash table.
 *
 * Arguments:
 *
 *     htClientIdx - Index to an NLM lock owner stored in the hash table.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/14/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_nlmLockOwnerHashRelease(vs_mem_index_t htOwnerIdx)
{
	/* Nothing to do */
	return;
} /* lock_nlmLockOwnerHashRelease() */

/*-----------------------------------------------------------------
 * Name:         lock_htLookupCifsLockOwner
 *
 * Description:
 *
 *     Search for a CIFS lock owner in the hash table.
 *
 * Arguments:
 *
 *     tid - Tid of CIFS lock owner.
 *
 *     fid - Fid of CIFS lock owner.
 *
 *     pid - Pid of CIFS lock owner.
 *
 *     searchOwnerIdx - Index to a CIFS lock owner, which contains
 *                      CIFS owner key to compare.
 *
 *     foundOwnerIdxp - This is output. Index to the found CIFS lock owner.
 *
 * Return Value:
 *
 *     Pointer to the CIFS lock owner structure - If found.
 *
 *     NULL - If not found.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
lock_cifsLockOwner_t *
lock_htLookupCifsLockOwner(uint16			tid,
						   uint16			fid,
						   uint16			pid,
						   vs_mem_index_t	searchOwnerIdx,
						   vs_mem_index_t	*foundOwnerIdxp)
{
	lock_htCifsOwnerKey_t	key;
	lock_cifsLockOwner_t	*ownerp = NULL;

	key.tid = tid;
	key.fid = fid;
	key.pid = pid;

	*foundOwnerIdxp = vs_hash_lookup(lock_globalCfgp->cifsOwnerHash,
									 (void *)&key,
									 searchOwnerIdx);
	if (!VS_MEM_IDX_IS_NULL(*foundOwnerIdxp))
	{
		/* Found the CIFS lock owner */
		ownerp = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(*foundOwnerIdxp);
	}

	return (ownerp);
} /* lock_htLookupCifsLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_htInsertCifsLockOwner
 *
 * Description:
 *
 *     Insert a CIFS lock owner in the hash table.
 *
 * Arguments:
 *
 *     tid - Tid of CIFS lock owner.
 *
 *     fid - Fid of CIFS lock owner.
 *
 *     pid - Pid of CIFS lock owner.
 *
 *     ownerIdx - Index to the CIFS lock owner.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
inline int32
lock_htInsertCifsLockOwner(uint16			tid,
						   uint16			fid,
						   uint16			pid,
						   vs_mem_index_t	ownerIdx)
{
	int32					rc;
	lock_htCifsOwnerKey_t	key;

	key.tid = tid;
	key.fid = fid;
	key.pid = pid;

	rc =  vs_hash_insert(lock_globalCfgp->cifsOwnerHash,
						 (void *)&key,
						 ownerIdx);

	return (rc);
} /* lock_htInsertCifsLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_htDeleteCifsLockOwner
 *
 * Description:
 *
 *     Delete a CIFS lock owner from the hash table.
 *
 * Arguments:
 *
 *     tid - Tid of CIFS lock owner.
 *
 *     fid - Fid of CIFS lock owner.
 *
 *     pid - Pid of CIFS lock owner.
 *
 *     ownerIdx - Index to the CIFS lock owner.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_htDeleteCifsLockOwner(uint16			tid,
						   uint16			fid,
						   uint16			pid,
						   vs_mem_index_t	ownerIdx)
{
	lock_htCifsOwnerKey_t	key;
	vs_mem_index_t			delOwnerIdx;

	key.tid = tid;
	key.fid = fid;
	key.pid = pid;

	delOwnerIdx = vs_hash_delete(lock_globalCfgp->cifsOwnerHash,
				   				 (void *)&key,
								 ownerIdx);

	if (VS_MEM_IDX_IS_NULL(delOwnerIdx))
		return (NFX_ERR);
	else
		return (NFX_OK);
} /* lock_htDeleteCifsLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_htLookupNlmLockClient
 *
 * Description:
 *
 *     Search for an NLM client in the hash table.
 *
 * Arguments:
 *
 *     clientIpAddr - IP address of the NLM client.
 *
 *     searchClientIdx - Index to an NLM client, which contains
 *                       NLM client IP address to compare.
 *
 *     foundClientIdxp - This is output. Index to the found NLM client.
 *
 * Return Value:
 *
 *     Pointer to the NLM client structure - If found.
 *
 *     NULL - If not found.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
lock_nlmLockClient_t *
lock_htLookupNlmLockClient(ip_index_t		clientIpAddr,
						   vs_mem_index_t	searchClientIdx,
						   vs_mem_index_t	*foundClientIdxp)
{
	lock_htNlmClientKey_t	key;
	lock_nlmLockClient_t	*clientp = NULL;

	/*TODO: Change this after ip_index_t is defined*/
	key.clientIpAddr = clientIpAddr;

	*foundClientIdxp = vs_hash_lookup(lock_globalCfgp->nlmClientHash,
									  &key,
									  searchClientIdx);
	if (!VS_MEM_IDX_IS_NULL(*foundClientIdxp))
	{
		/* Found the NLM client */
		clientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(*foundClientIdxp);
	}

	return (clientp);
} /* lock_htLookupNlmLockClient() */

/*-----------------------------------------------------------------
 * Name:         lock_htInsertNlmLockClient
 *
 * Description:
 *
 *     Insert an NLM client in the hash table.
 *
 * Arguments:
 *
 *     clientIpAddr - IP address of the NLM client.
 *
 *     clientIdx - Index to the NLM client.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
inline int32
lock_htInsertNlmLockClient(ip_index_t		clientIpAddr,
						   vs_mem_index_t	clientIdx)
{
	int32					rc;
	lock_htNlmClientKey_t	key;

	/*TODO: Change this after ip_index_t is defined*/
	key.clientIpAddr = clientIpAddr;

	rc =  vs_hash_insert(lock_globalCfgp->nlmClientHash, &key, clientIdx);
	return (rc);
} /* lock_htInsertNlmLockClient() */

/*-----------------------------------------------------------------
 * Name:         lock_htDeleteNlmLockClient
 *
 * Description:
 *
 *     Delete an NLM client in the hash table.
 *
 * Arguments:
 *
 *     clientIpAddr - IP address of the NLM client.
 *
 *     clientIdx - Index to the NLM client.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_htDeleteNlmLockClient(ip_index_t		clientIpAddr,
						   vs_mem_index_t	clientIdx)
{
	lock_htNlmClientKey_t	key;
	vs_mem_index_t			delClientIdx;

	/*TODO: Change this after ip_index_t is defined*/
	key.clientIpAddr = clientIpAddr;

	delClientIdx = vs_hash_delete(lock_globalCfgp->nlmClientHash,
								  &key,
								  clientIdx);

	if (VS_MEM_IDX_IS_NULL(delClientIdx))
		return (NFX_ERR);
	else
		return (NFX_OK);
} /* lock_htDeleteNlmLockClient() */

/*-----------------------------------------------------------------
 * Name:         lock_htLookupNlmLockOwner
 *
 * Description:
 *
 *     Search for an NLM owner in the hash table.
 *
 * Arguments:
 *
 *     clientIpAddr - IP address of the NLM client.
 *
 *     pid - Pid of the NLM owner.
 *
 *     searchOwnerIdx - Index to an NLM lock owner, from whick we can get
 *                      NLM client IP address and owner Pid to compare.
 *
 *     foundOwnerIdxp - This is output. Index to the found NLM owner.
 *
 * Return Value:
 *
 *     Pointer to the NLM owner structure - If found.
 *
 *     NULL - If not found.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
lock_nlmLockOwner_t *
lock_htLookupNlmLockOwner(ip_index_t		clientIpAddr,
						  int32				pid,
						  vs_mem_index_t	searchOwnerIdx,
						  vs_mem_index_t	*foundOwnerIdxp)
{
	lock_htNlmOwnerKey_t	key;
	lock_nlmLockOwner_t		*ownerp = NULL;

	/*TODO: Change this after ip_index_t is defined*/
	key.clientIpAddr = clientIpAddr;
	key.pid = pid;

	*foundOwnerIdxp = vs_hash_lookup(lock_globalCfgp->nlmOwnerHash,
									 &key,
									 searchOwnerIdx);
	if (!VS_MEM_IDX_IS_NULL(*foundOwnerIdxp))
	{
		/* Found the NLM owner */
		ownerp = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(*foundOwnerIdxp);
	}

	return (ownerp);
} /* lock_htLookupNlmLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_htInsertNlmLockOwner
 *
 * Description:
 *
 *     Insert an NLM owner in the hash table.
 *
 * Arguments:
 *
 *     clientIpAddr - IP address of the NLM client.
 *
 *     pid - Pid of the NLM owner.
 *
 *
 *     ownerIdx - Index to the NLM owner.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
inline int32
lock_htInsertNlmLockOwner(ip_index_t		clientIpAddr,
						  int32				pid,
						  vs_mem_index_t	ownerIdx)
{
	int32					rc;
	lock_htNlmOwnerKey_t	key;

	/*TODO: Change this after ip_index_t is defined*/
	key.clientIpAddr = clientIpAddr;
	key.pid = pid;

	rc =  vs_hash_insert(lock_globalCfgp->nlmOwnerHash, &key, ownerIdx);
	return (rc);
} /* lock_htInsertNlmLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_htDeleteNlmLockOwner
 *
 * Description:
 *
 *     Delete an NLM lock owner from the hash table.
 *
 * Arguments:
 *
 *     clientIdx - Index to the NLM lock client.
 *
 *     pid - Pid of NLM lock owner.
 *
 *     ownerIdx - Index to the NLM lock owner.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/15/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_htDeleteNlmLockOwner(vs_mem_index_t	clientIdx,
						  uint16			pid,
						  vs_mem_index_t	ownerIdx)
{
	lock_nlmLockClient_t	*clientp;
	lock_htNlmOwnerKey_t	key;
	vs_mem_index_t			delOwnerIdx;

	clientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(clientIdx);
	if (clientp == NULL)
		return (NFX_ERR);

	/*TODO: Change this after ip_index_t is defined*/
	key.clientIpAddr = clientp->clientIpAddr;
	key.pid = pid;

	delOwnerIdx = vs_hash_delete(lock_globalCfgp->nlmOwnerHash,
				   				 &key,
								 ownerIdx);

	if (VS_MEM_IDX_IS_NULL(delOwnerIdx))
		return (NFX_ERR);
	else
		return (NFX_OK);
} /* lock_htDeleteNlmLockOwner() */


