/*
 *-----------------------------------------------------------------
 *
 * Name:	malloc.c
 *
 * RCS:  $Id: malloc.c,v 1.7 2002/11/22 22:03:40 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	internal functions for embedded memory allocator
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/malloc-api.h>
#include "malloc.h"

#undef DEBUG
#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x) 
#endif

/*
 * stack structures
 */
uchar8	* mm_stack_buf_base[MM_NUM_STACK_SIZES];
uchar8	* mm_stack_buf_end[MM_NUM_STACK_SIZES];
uint32		mm_stack_size_ratio[MM_NUM_STACK_SIZES];
uint32		mm_stack_num_bufs[MM_NUM_STACK_SIZES];
uint32		mm_stack_buf_size[MM_NUM_STACK_SIZES];
uint32		mm_stack_bdesc_size[MM_NUM_STACK_SIZES];
mm_stack_t	mm_stack[MM_NUM_STACK_SIZES];

uint32		mm_num_bufs;		/* number of buffers in pool */
uint32		mm_max_contig;
uint32		mm_num_free;
mm_stats_t	mm_stats;

mm_free_node_t	mm_free_list;
mm_alloc_node_t	mm_alloc_list;
mm_free_node_stack_t	mm_free_node_stack;
mm_alloc_node_stack_t	mm_alloc_node_stack;
mm_bdesc_t		* mm_bdesc_list;
address_t		mm_bdesc_base;
address_t		mm_bdesc_end;

mm_free_node_t	* mm_free_stack_base;
mm_alloc_node_t	* mm_alloc_stack_base;
mm_free_node_t	* mm_free_node_list;
mm_alloc_node_t	* mm_alloc_node_list;

uint32			* mm_validate;

#ifdef MEM_HISTOGRAM
int32 mem_histogram [3][HISTOGRAM_MAX] ;
#endif
int32 mem_histogram_cont [3] ;

#ifdef MEM_HISTOGRAM_BUCKET
int32 memHistogramBucketsSize[MEM_HIST_NBR_BUCKETS] = { 10,100,1000,10000,100000,100000000};
int64 memHistogramBuckets [MEM_HIST_BUCKET_TYPES][MEM_HIST_NBR_BUCKETS];
#endif


#define CONTIG_ALLOC_INDICATOR ((void *)1)

mm_base_t mmBase ;

struct mm_cache **size_to_cache;
struct mm_cache *slab_cache;
struct mm_cache *cache_cache;
void *mm_pool_start;
void *mm_pool_end;
void *mm_pool_begin;
struct kmem_usage *mm_usage;
int mm_usage_num;
unsigned long mm_pool_inuse;
nfx_spinlock_t mm_pool_spin;

CIRCLEQ_HEAD(, mm_cache) mm_sys_caches;

#if defined(MEM_HISTOGRAM)
#define UPDATE_STAT_INC(stat) ++(stat)
#define UPDATE_MAXSTAT_INC(stat, maxstat) \
	{ if (++(stat) > maxstat) maxstat = stat; }
#define UPDATE_STAT_DEC(stat) --(stat)
#define UPDATE_STAT_MINUS(stat, dec) ((stat) -= dec)
#define UPDATE_STAT_PLUS(stat, plus) ((stat) += plus)
#define UPDATE_MAXSTAT_PLUS(stat, plus, max) \
	{ if (((stat) += (plus) > maxstat) maxstat = stat; }
#define UPDATE_MAXSTAT(stat, max) \
 	{ if ((stat) > (maxstat)) maxstat = stat; }
#else
#define UPDATE_STAT_INC(stat)
#define UPDATE_MAXSTAT_INC(stat, maxstat)
#define UPDATE_STAT_DEC(stat)
#define UPDATE_STAT_MINUS(stat, dec)
#define UPDATE_STAT_PLUS(stat, plus)
#define UPDATE_MAXSTAT_PLUS(stat, plus, max)
#define UPDATE_MAXSTAT(stat, maxstat) 
#endif

#if defined(MM_VALIDATE)
#define MM_VERIFY(x) VERIFY(x)
#else
#define MM_VERIFY(x)
#endif

#define MM_COPYLEN 20 

struct mm_pool_entry
{
	LIST_ENTRY(mm_pool_entry) pe_list_entry;
};

LIST_HEAD(, mm_pool_entry) mm_pool;

#define ALIGN(addr, align) (((address_t)addr + align - 1) & ~(align - 1))


static inline struct kmem_usage *
addr_to_ku(void *addr)
{
	unsigned long idx = (((address_t)addr - (address_t)mm_pool_begin)
						 >> SLAB_IDX_SHIFT);
	MM_VERIFY(addr >= mm_pool_start && addr < mm_pool_end);
	return mm_usage + idx;
}


static inline struct kmem_usage *
mm_pool_entry_end_ku(struct kmem_usage *ku)
{
	return ku + (unsigned long)ku->ku_slab / SLAB_MIN_SIZE - 1;
}


void
mm_pool_init(void)
{
	/* Put the whole pool on the free list as one object */
	struct mm_pool_entry *pe = (struct mm_pool_entry *)mm_pool_start;
	LIST_INIT(&mm_pool);
	NFX_SPIN_INIT(&mm_pool_spin, SPINLOCK_LEVEL_UNKNOWN);
	struct kmem_usage *start = addr_to_ku(mm_pool_start);
	struct kmem_usage *end = addr_to_ku(mm_pool_end - 1);
	start->ku_slab = (void *)((address_t)mm_pool_end
							  - (address_t)mm_pool_start);
	end->ku_slab = (void *)((address_t)mm_pool_end
							  - (address_t)mm_pool_start);
	LIST_INSERT_HEAD(&mm_pool, pe, pe_list_entry);
}


void *
mm_pool_alloc(unsigned long size, unsigned long align)
{
	struct mm_pool_entry *pe;
    struct mm_pool_entry *best_pe = NULL;
    unsigned long best_diff = 0xFFFFFFFF;

	debug(("mm_pool_alloc: sz %ld align %ld\n", size, align));

	ASSERT(!(align & (SLAB_MIN_SIZE - 1)));
	ASSERT(!(size & (SLAB_MIN_SIZE - 1)));

 retry:
	NFX_SPIN_LOCK(&mm_pool_spin);
	LIST_FOREACH(pe, &mm_pool, pe_list_entry) {
		char *aligned_mem = (char *)ALIGN(pe, align);
		struct kmem_usage *ku = addr_to_ku(pe);
		unsigned long entry_size = (unsigned long)ku->ku_slab;
        unsigned long diff = (aligned_mem - (char *)pe) +
            (((char *)pe + entry_size) - (aligned_mem + size));
		MM_VERIFY(ku->ku_cache == NULL);
        
		if (aligned_mem + size <= (char *)pe + entry_size) {
            if (diff < best_diff) {
                best_pe = pe;
                best_diff = diff;
            }
        }
        if (best_diff == 0
            || (LIST_NEXT(pe, pe_list_entry) == NULL && best_pe != NULL)) {
            char *aligned_mem = (char *)ALIGN(best_pe, align);
            struct kmem_usage *ku = addr_to_ku(best_pe);
            unsigned long entry_size = (unsigned long)ku->ku_slab;
            
			unsigned long begin_size =
				(address_t)aligned_mem - (address_t)best_pe;
			unsigned long rest_size = entry_size - begin_size - size;
			
			debug(("mm_pool_alloc: found entry at %p sz %ld new alloc %p\n",
				   best_pe, entry_size, aligned_mem));
			
			if (rest_size) {
				struct mm_pool_entry *rest_entry =
					(struct mm_pool_entry *)(aligned_mem + size);
				struct kmem_usage *rest_ku = addr_to_ku(rest_entry);
				struct kmem_usage *rest_ku_end =
					addr_to_ku((char *)rest_entry + rest_size - 1);

				debug(("mm_pool_alloc: split tail sz %ld\n", rest_size));
				
                ASSERT(rest_ku->ku_cache == NULL);
                ASSERT(rest_ku_end->ku_cache == NULL);

				rest_ku->ku_slab = (void *)rest_size;
				rest_ku_end->ku_slab = (void *)rest_size;
				LIST_INSERT_HEAD(&mm_pool, rest_entry, pe_list_entry);
			}
			if (begin_size) {
				debug(("mm_pool_alloc: split beginning sz %ld\n", begin_size));
                ASSERT(ku->ku_cache == NULL);
				ku->ku_slab = (void*)begin_size;
				if (begin_size > SLAB_MIN_SIZE) {
					ASSERT(mm_pool_entry_end_ku(ku)->ku_cache == NULL);
					mm_pool_entry_end_ku(ku)->ku_slab = (void *)begin_size;
				}
			}
			else {
				LIST_REMOVE(best_pe, pe_list_entry);
			}

			/* Save the sizes of allocations and indicate obj was
			   allocated for pool list */
			struct kmem_usage *ku_start = addr_to_ku(aligned_mem);
			struct kmem_usage *ku_end = addr_to_ku(aligned_mem + size - 1);
			ku_start->ku_slab = (void *)size;
			ku_start->ku_cache = CONTIG_ALLOC_INDICATOR;
			if (ku_end != ku_start) {
				ku_end->ku_slab = (void *)size;
				ku_end->ku_cache = CONTIG_ALLOC_INDICATOR;
			}
			mm_pool_inuse += size;
			NFX_SPIN_UNLOCK(&mm_pool_spin);
			return aligned_mem;
		}
	}
	NFX_SPIN_UNLOCK(&mm_pool_spin);
    if (!cleanup_caches())
        return NULL;
    else
        goto retry;
}


void
mm_pool_free(void *obj)
{
	struct kmem_usage *ku = addr_to_ku(obj);
	MM_VERIFY(((address_t)obj & (SLAB_MIN_SIZE - 1)) == 0);
	MM_VERIFY(ku->ku_cache == CONTIG_ALLOC_INDICATOR);
	MM_VERIFY(ku->ku_slab != NULL);
#if defined(MM_VALIDATE)
	struct kmem_usage *ku_end = addr_to_ku((char *)obj +
										   (unsigned long)ku->ku_slab - 1);
#endif	
	MM_VERIFY(ku->ku_slab == ku_end->ku_slab);
	MM_VERIFY(ku_end->ku_cache == CONTIG_ALLOC_INDICATOR);
	mm_pool_free_size(obj, (unsigned long)ku->ku_slab);
}

void
mm_pool_free_size(void *obj, unsigned long size)
{
	struct kmem_usage *this_start = addr_to_ku(obj);
	struct kmem_usage *this_end = addr_to_ku((char *)obj + size - 1);

	debug(("mm_pool_free_size: %p sz %ld\n", obj, size));

	NFX_SPIN_LOCK(&mm_pool_spin);
	
	mm_pool_inuse -= size;

	/* Indicate obj goes to free list */
	this_end->ku_cache = NULL;
	this_start->ku_cache = NULL;

	MM_VERIFY(this_end->ku_slab == this_start->ku_slab);

	/* check if the next obj is free */
	struct kmem_usage *ku_next = addr_to_ku((char *)obj + size);
	char *next_addr = (char *)obj + size;
	if (ku_next->ku_cache == NULL) {
		unsigned long next_size =
			(unsigned long)ku_next->ku_slab;
		struct kmem_usage *next_end =
			addr_to_ku(next_addr + next_size - 1);

		debug(("mm_pool_free_size: coalesce with next\n"));
		
        ASSERT(next_end->ku_cache == NULL);
        
		/* Update block size and indicate it is on free list */
		next_end->ku_slab = (void *)(size + next_size);
		if (next_end != ku_next) {
			ku_next->ku_slab = NULL;
		}
		
		LIST_REMOVE((struct mm_pool_entry *)next_addr, pe_list_entry);

		this_end->ku_slab = NULL;

		this_start->ku_cache = NULL;
		this_start->ku_slab = (void *)(size + next_size);

		/* Update total free size */
		this_end = next_end;
		size += next_size;
	}

	/* Check previous */
	struct kmem_usage *prev_end = this_start - 1;
	if (obj > mm_pool_start && prev_end->ku_cache == NULL) {
		/* Previous object on free list */
		unsigned long prev_size = (unsigned long)prev_end->ku_slab;
		struct kmem_usage *prev_start =
			addr_to_ku((char *)obj - prev_size);
			
		debug(("mm_pool_free_size: coalesce with prev\n"));
		
		MM_VERIFY((unsigned long)prev_start->ku_slab == prev_size);
		ASSERT(prev_start->ku_cache == NULL);
        
		prev_start->ku_slab = (void *)(prev_size + size);
		this_end->ku_slab = (void *)(prev_size + size);
        this_start->ku_slab = NULL;

		if (prev_end != prev_start) {
			prev_end->ku_slab = NULL;
			prev_end->ku_cache = NULL;
		}
	}
	else {
		/* Previous is busy, add this obj to free list */
		this_start->ku_slab = (void *)size;
		this_end->ku_slab = (void *)size;
		LIST_INSERT_HEAD(&mm_pool,
						 (struct mm_pool_entry *)obj, pe_list_entry);
	}
	NFX_SPIN_UNLOCK(&mm_pool_spin);
}


void *
mm_pool_alloc_raw(unsigned size, unsigned align)
{
	debug(("mm_pool_alloc_raw: start=%p end=%p sz=%d align=%d from %p\n",
		   mm_pool_start, mm_pool_end, size, align, (void *)get_ra()));
	
	if (mm_pool_start + size + align <= mm_pool_end) {
		void *ptr;
		mm_pool_start += align - 1;
		mm_pool_start -= (address_t)mm_pool_start % align;
		debug(("mm_pool_alloc_raw: after align: start=%p\n", mm_pool_start));
		ptr = mm_pool_start;
		mm_pool_start += size;
		return ptr;
	}
	else {
		return NULL;
	}
}


#if defined(MM_VERIFY_ALLOCS)
static void
mark_free_object(struct mm_cache *cache, void *obj)
{
	int i;
    if (!cache->c_ctor) {
        /* Copy some known text into the object to detect duplicate frees */
        for (i = 0; i < MIN(cache->c_offset / sizeof(unsigned int),
                            MM_COPYLEN); ++i) {
            *((unsigned int *)obj + i) = 0xdeadbeef;
        }
    }
}


static void
mark_allocated_object(struct mm_cache *cache, void *obj)
{
    if (!cache->c_ctor) {
        int i;
        for (i = 0; i < MIN(cache->c_objsize / sizeof(unsigned int), MM_COPYLEN);
             ++i) {
            *((unsigned int *)obj + i) = 0xbaddcafe;
        }
    }
    
}


static void
verify_free_obj(struct mm_cache *cache, void *obj)
{
	int i;
    if (!cache->c_ctor) {
        for (i = 0; i < MIN(cache->c_offset / sizeof(unsigned int),
                            MM_COPYLEN); ++i) {
            if (*((unsigned int *)obj + i) != 0xdeadbeef) {
                printf("buffer %p freed by %p modified on free list at %d: ", obj,
                       ((struct mm_bufctl *)((char *)obj +
                                             cache->c_offset))->bc_addr, i);
                for (; i < MIN(cache->c_objsize / sizeof(unsigned int),
                               MM_COPYLEN); ++i) {
                    printf("%x ", *((unsigned int *)obj + i));
                }
                printf("\n");
                panic("\n");
            }
        }
    }
}

#else
void mark_free_object(struct mm_cache *cache, void *obj);
void mark_allocated_object(struct mm_cache *cache, void *obj);
void verify_free_obj(struct mm_cache *cache, void *obj);
#define mark_free_object(c, o)
#define mark_allocated_object(c, o)
#define verify_free_obj(c, o)
#endif


#if defined(MM_EXPENSIVE_VALIDATE)
static void
cache_verify_slab_list(struct mm_cache *cache)
{
	int done_fulls = 0;
	int done_partials = 0;
	struct mm_slab *slab;
	TAILQ_FOREACH(slab, &cache->c_slabs, sl_entry) {
		if (slab->sl_inuse == cache->c_num) {
			if (done_fulls)
				panic("full slab in the middle of list\n");
			if (SLIST_FIRST(&slab->sl_buffers)) {
				panic("list not empty for full slab\n");
			}
		}
		else {
			if (!done_fulls) {
				if (cache->c_current != slab)
					panic("current does not point to first partial slab\n");
			}
			done_fulls = 1;
			if (slab->sl_inuse) {
				if (done_partials)
					panic("partial slab at tail\n");
				if (!SLIST_FIRST(&slab->sl_buffers))
					panic("list is empty for partial slab\n");
			}
			else {
				done_partials = 1;
				if (!SLIST_FIRST(&slab->sl_buffers))
					panic("list is empty for all free slab\n");
			}
		}
	}
}

static void
cache_verify_free_list(struct mm_cache *cache, struct mm_slab *slab)
{
	struct mm_bufctl *ctl;
	SLIST_FOREACH(ctl, &slab->sl_buffers, bc_next) {
		verify_free_obj(cache, (char *)ctl - cache->c_offset);
	}
}

#else
void cache_verify_slab_list(struct mm_cache *cache);
#define cache_verify_slab_list(cache)
void cache_verify_free_list(struct mm_cache *cache, struct mm_slab *slab);
#define cache_verify_free_list(c, s)
#endif


void
cache_init(struct mm_cache *cache, int flags,
		   int size, int align, int slabsize,
		   void (*ctor)(void *, uint32),
		   void (*dtor)(void *))
{
	TAILQ_INIT(&cache->c_slabs);
	cache->c_objsize = ALIGN(size, sizeof(unsigned long));
	if (ctor) {
		if (!(flags & CACHE_INMEM_SLAB))
			panic("not supported\n");
		/* Reserve the linkage space for contstructed objects */
		cache->c_objsize += sizeof(struct mm_bufctl);
	}
	cache->c_align = align;
	if (flags & CACHE_INMEM_SLAB) {
		cache->c_num = (slabsize - sizeof(struct mm_slab)) /
			ALIGN(cache->c_objsize, cache->c_align);
	}
	else {
		cache->c_num = slabsize /
			ALIGN(cache->c_objsize, cache->c_align);
	}
	NFX_SPIN_INIT(&cache->c_spin, SPINLOCK_LEVEL_UNKNOWN);
	cache->c_current = 0;
	cache->c_slabsize = slabsize;
	cache->c_offset = cache->c_objsize - sizeof(struct mm_bufctl);
	cache->c_ctor = ctor;
	cache->c_dtor = dtor;
	cache->c_calls = 0;
	cache->c_maxinuse = 0;
	cache->c_inuse = 0;
	cache->c_numslabs = 0;
	cache->c_maxnumslabs = 0;
	cache->c_flags = flags;
}


static void
cache_shrink(struct mm_cache *cache)
{
	struct mm_slab *slab = TAILQ_LAST(&cache->c_slabs, mm_slab_queue);
	TAILQ_REMOVE(&cache->c_slabs, slab, sl_entry);
	char *mem = slab->sl_mem;
	struct kmem_usage *ku = addr_to_ku(mem);
	int i;
	if (cache->c_current == slab)
		cache->c_current = NULL;
	UPDATE_STAT_DEC(cache->c_numslabs);
	if (cache->c_dtor) {
		int i;
		char *dtor_mem = mem;
		for (i = 0; i < cache->c_num; ++i) {
			cache->c_dtor(dtor_mem);
			dtor_mem += cache->c_objsize;
			dtor_mem = (char *)ALIGN(dtor_mem, cache->c_align);
		}
	}
	for (i = 0; i < cache->c_slabsize / SLAB_MIN_SIZE; ++i, ++ku) {
		ku->ku_cache = NULL;
		ku->ku_slab = NULL;
	}
	mm_pool_free_size(slab->sl_mem, cache->c_slabsize);
	if (!(cache->c_flags & CACHE_INMEM_SLAB))
		cache_free(slab_cache, slab);
}

#define MAX_CLEANUP (10 * 1024 * 1024)

int
cleanup_caches(void)
{
	struct mm_cache *cache;
	int success = 0;
	unsigned long total_free = 0;
	CIRCLEQ_FOREACH(cache, &mm_sys_caches, c_entry) {
		NFX_SPIN_LOCK(&cache->c_spin);
		struct mm_slab *slab = TAILQ_LAST(&cache->c_slabs, mm_slab_queue);
		while (slab && !slab->sl_inuse) {
			cache_shrink(cache);
			total_free += cache->c_slabsize;
			if (total_free > MAX_CLEANUP) {
				NFX_SPIN_UNLOCK(&cache->c_spin);
				goto done;
			}
			success = 1;
			slab = TAILQ_LAST(&cache->c_slabs, mm_slab_queue);
		}
		NFX_SPIN_UNLOCK(&cache->c_spin);
	}
 done:
	debug(("cleanup_caches: freed %ld bytes\n", total_free));
	return success;
}


static void
cache_grow(struct mm_cache *cache)
{
	int size = cache->c_slabsize;
	void *mem; 
	struct mm_slab *slab;
	void (*ctor)(void *, uint32);
	
	debug(("cache_grow: sz=%d num=%d\n", cache->c_objsize, cache->c_num));

	NFX_SPIN_UNLOCK(&cache->c_spin);
	mem = mm_pool_alloc(size, size);
		
	if (mem) {
		char *obj = mem;
		struct kmem_usage *ku = addr_to_ku(mem);
		int i;

		if (cache->c_flags & CACHE_INMEM_SLAB) {
			slab = (struct mm_slab *)((char *)mem + size -
									  sizeof(struct mm_slab));
		}
		else {
			slab = cache_alloc(slab_cache);
		}

		debug(("cache_grow: alloced memory at %p slab at %p\n", mem, slab));

		if (slab) {
			for (i = 0; i < size / SLAB_MIN_SIZE; ++i, ++ku) {
				ku->ku_cache = cache;
				ku->ku_slab = slab;
			}

			size -= sizeof(struct mm_slab);
			slab->sl_mem = mem;
			SLIST_INIT(&slab->sl_buffers);
			slab->sl_inuse = 0;
			ctor = cache->c_ctor;
			for (i = 0; i < cache->c_num; ++i) {
				SLIST_INSERT_HEAD(
					&slab->sl_buffers,
					(struct mm_bufctl *)((char *)obj + cache->c_offset),
					bc_next);
				mark_free_object(cache, obj);
				if (ctor) {
					ctor(obj, cache->c_objsize - sizeof(struct mm_bufctl));
				}
				obj += cache->c_objsize;
				obj = (void *)ALIGN(obj, cache->c_align);
			}

			debug(("cache_grow: insert slab to queue\n"));
			
			NFX_SPIN_LOCK(&cache->c_spin);
			cache_verify_free_list(cache, slab);
			TAILQ_INSERT_TAIL(&cache->c_slabs, slab, sl_entry);
			if (cache->c_current == NULL) {
				cache->c_current = slab;
				ASSERT(SLIST_FIRST(&(cache->c_current->sl_buffers)));
			}
			cache_verify_slab_list(cache);
			UPDATE_MAXSTAT_INC(cache->c_numslabs, cache->c_maxnumslabs);
			return;
		}
	}
	NFX_SPIN_LOCK(&cache->c_spin);
}


int32
eee_ramDeallocRA(void *obj, void * ra)
{
	struct kmem_usage *ku = addr_to_ku(obj);

//printf("eee_ramDeallocRA: ra %p addr %p \n", ra, obj);

	if (ku->ku_cache != (struct mm_cache *)CONTIG_ALLOC_INDICATOR) {
		cache_free_ra(addr_to_ku(obj)->ku_cache, obj, ra);
	}
	else {
		mm_pool_free(obj);
	}
	return NFX_OK;
}


#if defined(MM_VERIFY_ALLOCS)

void
cache_free(struct mm_cache *cache, void *obj)
{
	cache_free_ra(cache, obj, (void*)get_ra());
}

void
cache_free_ra(struct mm_cache *cache, void *obj, void *ra)
#else
void
cache_free(struct mm_cache *cache, void *obj)
#endif
{
	struct mm_slab *slab;
	struct mm_bufctl *ctl = (struct mm_bufctl *)
		((char *)obj + cache->c_offset);
	int i __attribute__((unused));
	
	if (cache->c_flags & CACHE_INMEM_SLAB) {
		slab = (struct mm_slab *)
			(((address_t)obj & ~((address_t)cache->c_slabsize - 1)) +
			 (address_t)cache->c_slabsize - sizeof(struct mm_slab));
	}
	else {
		slab = addr_to_ku(obj)->ku_slab;
	}

	/* Check for unaligned free */
	MM_VERIFY(((address_t)obj - (address_t)slab->sl_mem) %
			  ALIGN(cache->c_objsize, cache->c_align) == 0);

	NFX_SPIN_LOCK(&cache->c_spin);
				  
#if defined(MM_VERIFY_ALLOCS)
	if (*((unsigned int *)obj) == 0xdeadbeef) {
		/* Looks like a free object, verify if it is on free list */
		struct mm_bufctl *bc;
		SLIST_FOREACH(bc, &slab->sl_buffers, bc_next) {
			if ((void *)bc == ctl) {
				panic("duplicate free from %p, previous from %p\n",
					  ra, bc->bc_addr);
			}
		}
	}
	ctl->bc_addr = ra;
#endif

	mark_free_object(cache, obj);

	UPDATE_STAT_DEC(cache->c_inuse);
	if (SLIST_FIRST(&slab->sl_buffers)) {
		/* Check if objects were really allocated from this slab */
		MM_VERIFY(slab->sl_inuse > 0);
		SLIST_INSERT_HEAD(&slab->sl_buffers, ctl, bc_next);
		--slab->sl_inuse;
		if (slab->sl_inuse == 0) {
			if (TAILQ_NEXT(slab, sl_entry) &&
				TAILQ_NEXT(slab, sl_entry)->sl_inuse) {
				/* Move to tail of the list */
				if (cache->c_current == slab) {
					cache->c_current = TAILQ_NEXT(slab, sl_entry);
				}
				TAILQ_REMOVE(&cache->c_slabs, slab, sl_entry);
				TAILQ_INSERT_TAIL(&cache->c_slabs, slab, sl_entry);
				cache_verify_slab_list(cache);
			}
		}
	}
	else {
		/* Slab was full before free */
		MM_VERIFY(slab->sl_inuse == cache->c_num);
		--slab->sl_inuse;
		SLIST_INSERT_HEAD(&slab->sl_buffers, ctl, bc_next);
		if (TAILQ_NEXT(slab, sl_entry) &&
			TAILQ_NEXT(slab, sl_entry)->sl_inuse == cache->c_num) {
			TAILQ_REMOVE(&cache->c_slabs, slab, sl_entry);
			if (cache->c_current)
				TAILQ_INSERT_BEFORE(cache->c_current, slab, sl_entry);
			else
				TAILQ_INSERT_TAIL(&cache->c_slabs, slab, sl_entry);
		}
		cache->c_current = slab;
		cache_verify_slab_list(cache);
	}

	NFX_SPIN_UNLOCK(&cache->c_spin);
}


#if defined(MM_VERIFY_ALLOCS)

void *
cache_alloc(struct mm_cache *cache)
{
	return cache_alloc_ra(cache, (void *)get_ra());
}

void *
cache_alloc_ra(struct mm_cache *cache, void *ra)
#else
void *
cache_alloc(struct mm_cache *cache)
#endif
{
	void *obj;
	struct mm_slab *cur;
	struct mm_bufctl *bc;
	int i __attribute__((unused));

	NFX_SPIN_LOCK(&cache->c_spin);
	cur = cache->c_current;
	/* Check current */
	if (cur == NULL) {
		cache_grow(cache);
		cur = cache->c_current;
		if (cur == NULL) {
			obj = NULL;
			goto done;
		}
	}

	if ((bc = SLIST_FIRST(&cur->sl_buffers))) {
		obj = (char *)bc - cache->c_offset;
#if defined(MM_VERIFY_ALLOCS)
		verify_free_obj(cache, obj);
		((struct mm_bufctl *)obj)->bc_addr = ra;
#endif
		
		UPDATE_STAT_INC(cache->c_calls);
		UPDATE_MAXSTAT_INC(cache->c_inuse, cache->c_maxinuse);

		SLIST_REMOVE_HEAD(&cur->sl_buffers, bc_next);
		cur->sl_inuse++;

		mark_allocated_object(cache, obj);
		
		if (SLIST_FIRST(&cur->sl_buffers)) {
			MM_VERIFY(cur->sl_inuse < cache->c_num);
		}
		else {
			MM_VERIFY(cur->sl_inuse == cache->c_num);
			cache->c_current = TAILQ_NEXT(cur, sl_entry);
			cache_verify_slab_list(cache);
		}
	}
	else {
		obj = NULL;
	}
 done:
	NFX_SPIN_UNLOCK(&cache->c_spin);
	return obj;
}

struct mm_cache *
cache_create(uint32 flags, uint32 objsize, uint32 align,
			 uint32 slabsize, void (*ctor)(void *, uint32),
			 void (*dtor)(void *))
{
	struct mm_cache *cache = cache_alloc(cache_cache);
	if (cache) {
		cache_init(cache, flags, objsize, align,
				   slabsize, ctor, dtor);
		CIRCLEQ_INSERT_TAIL(&mm_sys_caches, cache, c_entry);
	}
	return cache;
}


/*-----------------------------------------------------------------
 * Name:	eee_ramAllocRA()
 *
 * Description:	Allocate a flat memory area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocRA(
	uint32	num_bytes,			/* size of buffer requested */
	void 	* ra
)
{
    void *ptr;

    NFX_ASSERT(num_bytes != 0);
    
	if (num_bytes < SLAB_ALLOC_MAX_SIZE) {
		ptr = cache_alloc_ra(size_to_cache[(num_bytes - 1) / 64], ra);
//printf("eee_ramAllocRA cache: ra %p addr %p len %d\n", ra, ptr, num_bytes);

	}
	else {
		ptr =  mm_pool_alloc((num_bytes + SLAB_MIN_SIZE - 1)
							 & ~(SLAB_MIN_SIZE - 1),
							 SLAB_MIN_SIZE);
//printf("eee_ramAllocRA pool : ra %p addr %p len %d (%d)\n", ra, ptr, num_bytes,
//    (num_bytes + SLAB_MIN_SIZE - 1) & ~(SLAB_MIN_SIZE - 1));

	}

    return ptr;

}	/* eee_ramAlloc() */


/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	2/21/02
 *
 *-----------------------------------------------------------------
 */
void *
eee_ramAlloc(
	uint32	num_bytes
)
{
	void	* ra;

	ra = (void *)get_ra();

	return eee_ramAllocRA(num_bytes, ra);
}


void
cache_show()
{
#if defined(MEM_HISTOGRAM)
	struct mm_cache *cache;
	unsigned long total_alloced = 0;
	unsigned long total_inuse = 0;
	unsigned long free_slabs;
	struct mm_slab *slab;

	printf("cache usage\n");
	printf("%10s %10s %10s %10s %10s %10s %10s\n",
		   "objsize", "numobjects", "numslabs", "freeslabs",
		   "totalused", "maxinuse", "calls");
	CIRCLEQ_FOREACH(cache, &mm_sys_caches, c_entry) {
		total_inuse += cache->c_inuse * cache->c_objsize;
		total_alloced += cache->c_numslabs * cache->c_slabsize;
		free_slabs = 0;
		TAILQ_FOREACH(slab, &cache->c_slabs, sl_entry) {
			if (slab->sl_inuse == 0) {
				++free_slabs;
			}
		}
		if (cache->c_numslabs) {
			printf("%10d %10ld %10ld %10ld %10ld %10ld %10ld\n",
				   cache->c_objsize,
				   cache->c_inuse,
				   cache->c_numslabs,
				   free_slabs,
				   cache->c_numslabs * cache->c_slabsize,
				   cache->c_maxinuse,
				   cache->c_calls);
		}
	}
	printf("total alloced: %ld total inuse: %ld fragmentation: %ld%%\n",
		   total_alloced, total_inuse,
		   ((total_alloced - total_inuse) * (unsigned long)100)
		   / total_alloced);
	printf("\n");
#endif
}


void
pool_show()
{
#if defined(MEM_HISTOGRAM)
	unsigned long max_size = 0;
	unsigned long total_size = 0;
	unsigned long cnt = 0;    
	struct mm_pool_entry *pe;

	printf("pool usage: size %ld allocated %ld\n",
		   mm_pool_end - mm_pool_start, mm_pool_inuse);

	NFX_SPIN_LOCK(&mm_pool_spin);
	LIST_FOREACH(pe, &mm_pool, pe_list_entry) {
		struct kmem_usage *ku = addr_to_ku(pe);
		MM_VERIFY(ku->ku_cache == NULL);
		MM_VERIFY(mm_pool_entry_end_ku(ku)->ku_cache == NULL);
		MM_VERIFY(mm_pool_entry_end_ku(ku)->ku_slab == ku->ku_slab);

        if ((unsigned long)ku->ku_slab > max_size)
            max_size = (unsigned long)ku->ku_slab;
        total_size += (unsigned long)ku->ku_slab;
        cnt++;
        
		if (cnt < 5)
            printf("[%p %ld]->", pe, (unsigned long)ku->ku_slab);
	}
	NFX_SPIN_UNLOCK(&mm_pool_spin);
	printf("\n");

    printf ("Total free size      %ld\n", total_size);
    printf ("Max free entry size  %ld\n", max_size);
    printf ("Nbr of entries       %ld\n", cnt);
    
#endif	
}


/*-----------------------------------------------------------------
 * Name:	mm_isMallocBuffer()
 *
 * Description:	Checks whether or not the given address is from a flexible or stack malloc pool
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isMallocBuffer(
	void	* addr
)
{
	return mm_pool_begin <= addr && addr < mm_pool_end;
}	/* mm_isMallocBuffer() */

/*-----------------------------------------------------------------
 * Name:	mm_isSizeOkay()
 *
 * Description:	Checks malloc buffer to see if the len exceeds buffer length.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isSizeOkay(
	void	* addr,
	uint32	len,
	uint32	* buf_size
)
{
	struct mm_cache *cache = addr_to_ku(addr)->ku_cache;
	if (cache && cache != CONTIG_ALLOC_INDICATOR) {
		if (cache->c_objsize >= len) {
			return TRUE;
		}
		else {
			*buf_size = cache->c_objsize;
			return FALSE;
		}
	}
	else {
		return FALSE;
	}
}
	/* mm_isSizeOkay() */

struct cache_sizes {
    int objsize;
    int size;
    int flags;
};


#if defined(NFP_FP)
struct cache_sizes init_cache_sizes[] = {
    { 64, 8 * 1024, CACHE_INMEM_SLAB },
    { 128, 8 * 1024, CACHE_INMEM_SLAB },
    { 320, 8 * 1024, CACHE_INMEM_SLAB },
    { 1024, 8 * 1024, 0 },
    { 7232, 64 * 1024, CACHE_INMEM_SLAB },
    { 8384, 64 * 1024, CACHE_INMEM_SLAB },
    { 32 * 1024, 64 * 1024, 0 },
    { 0, 0, 0 }
};
#endif

/*-----------------------------------------------------------------
 * Name:	mallocInit()
 *
 * Description:	Initialize memory manager structures for 'Malloc' and 'Free'
 *				functions.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void
mallocInit(
	uchar8	* base,
	uint32	num_bytes
)
{
    malloc_size_t   flexSize;
    malloc_size_t   stackSize;

/*
	printf("mallocInit, base = %p, size = 0x%x bytes\n", base, num_bytes);
*/
	mm_stats.total_flex_size = 0;
	mm_stats.total_stack_size = 0;
	mm_stats.total_cont_size = 0;
    mm_stats.no_stack = 0;
	mm_stats.contHistogram[MM_HIST_MAX] = 0;
	mm_stats.contHistogram[MM_HIST_CUR] = 0;
	mm_stats.contHistogram[MM_HIST_TOT] = 0;

#ifdef MEM_HISTOGRAM
	bzero(mem_histogram, sizeof(mem_histogram));
#endif
	bzero(mem_histogram_cont, sizeof(mem_histogram_cont));

#ifdef MEM_HISTOGRAM_BUCKET
    bzero(memHistogramBuckets, sizeof(memHistogramBuckets));
#endif



	flexSize = (num_bytes / 100 * MM_FLEX_MEM_PCT) & ~0x3F;

	// remark stack size will be bigger, see further for updates
	stackSize = (num_bytes / 100 * MM_STACK_MEM_PCT) & ~0x3F;

    // assert ((MM_FLEX_MEM_PCT + MM_STACK_MEM_PCT) < 100);

    // general layout
    mmBase.flex = base;
    mmBase.flexSize = flexSize;
    mmBase.stack = base + flexSize;
    mmBase.stackSize = stackSize;

	CIRCLEQ_INIT(&mm_sys_caches);

	mm_pool_begin = mmBase.flex;
	mm_pool_start = mmBase.flex;
	mm_pool_end = (void *)(((address_t)mmBase.flex + flexSize) & ~(4096 - 1));
	mm_usage_num = ((address_t)mm_pool_end - (address_t)mm_pool_begin)
		/ SLAB_MIN_SIZE;
	mm_usage = mm_pool_alloc_raw(mm_usage_num * sizeof(struct kmem_usage),
								 sizeof(struct kmem_usage));
	if (mm_usage == NULL)
		panic("can not alloc usage structure\n");
	debug(("allocated mm_usage at %p\n", mm_usage));
	bzero(mm_usage, mm_usage_num * sizeof(struct kmem_usage));
	
	size_to_cache = mm_pool_alloc_raw(
		(64 * 1024) / 64 * sizeof(struct mm_cache *), sizeof(unsigned long));
	if (size_to_cache == NULL)
		panic("can not allocate size_to_cache structure\n");
	debug(("allocated size_to_cache structure at %p\n", size_to_cache));
		  
	slab_cache = mm_pool_alloc_raw(sizeof(struct mm_cache),
								   sizeof(unsigned long));
	if (slab_cache == NULL)
		panic("can not allocate slab_cache\n");
	debug(("allocated slab_cache at %p\n", slab_cache));
	cache_init(slab_cache, CACHE_INMEM_SLAB,
			   sizeof(struct mm_slab), sizeof(unsigned long),
			   64 * 1024, NULL, NULL);

	cache_cache = mm_pool_alloc_raw(sizeof(struct mm_cache),
									sizeof(unsigned long));
	if (cache_cache == NULL)
		panic("can not allocate cache_cache\n");
	debug(("allocated cache cache at %p\n", cache_cache));
	cache_init(cache_cache, CACHE_INMEM_SLAB,
			   sizeof(struct mm_cache), sizeof(unsigned long),
			   64 * 1024, NULL, NULL);

	mm_pool_alloc_raw(0, SLAB_MIN_SIZE);

	mm_pool_init();

	NFX_SPIN_LOCK(&slab_cache->c_spin);
	cache_grow(slab_cache);
	NFX_SPIN_UNLOCK(&slab_cache->c_spin);
	if (slab_cache->c_current == NULL)
		panic("can not allocate first slab for slab mgmt cache\n");
	debug(("after cache_grow: slab_cache->c_current=%p\n",
		   slab_cache->c_current));
	
	NFX_SPIN_LOCK(&cache_cache->c_spin);
	cache_grow(cache_cache);
	NFX_SPIN_UNLOCK(&cache_cache->c_spin);
	if (cache_cache->c_current == NULL)
		panic("can not allocate first slab for cache_cache\n");
	debug(("after cache_grow: cache_cache->c_current=%p\n",
		   cache_cache->c_current));

#if defined(NFP_FP)
    struct cache_sizes *cs = &init_cache_sizes[0];
    int size = 0;
    for (; cs->size != 0; ++cs) {
        struct mm_cache *cache = cache_create(
            cs->flags,
            cs->objsize,
            64,
            cs->size,
            NULL, NULL);
        if (cache == NULL)
            panic("can not allocate initial cache\n");
        for (; size < cs->objsize; size += 64)
            size_to_cache[size / 64] = cache;
    }
#else    
	/* Allocate initial caches */
	int i;
	for (i = 0; i < 1024; ++i) {
		struct mm_cache *cache = cache_create(
			CACHE_INMEM_SLAB, 
			(i + 1) * 64, 64, 64 * 1024,
			NULL, NULL);
		if (cache == NULL)
			panic("can not allocate initial cache\n");
		size_to_cache[i] = cache;
	}
#endif

	CIRCLEQ_INSERT_TAIL(&mm_sys_caches, cache_cache, c_entry);
	CIRCLEQ_INSERT_TAIL(&mm_sys_caches, slab_cache, c_entry);
}	/* mallocInit() */


/*-----------------------------------------------------------------
 *
 * Description: Reports the total/used memory sizes
 *
 *-----------------------------------------------------------------
 */
malloc_size_t
eee_ramTotalHeapSize()
{
	return mm_pool_end - mm_pool_start;
}

malloc_size_t
eee_ramTotalHeapAvail()
{
	return (mm_pool_end - mm_pool_start - mm_pool_inuse);
}

malloc_size_t
eee_ramTotalFlexSize()
{
    return (mm_pool_end - mm_pool_start);
}

malloc_size_t
eee_ramTotalFlexAvail()
{
    return (mm_pool_end - mm_pool_start - mm_pool_inuse);

}

malloc_size_t
eee_ramTotalFlexUsed()
{
	return mm_pool_inuse;
}

