/*
 *-----------------------------------------------------------------
 *
 * Name:	malloc.h
 *
 * RCS:  $Id: malloc.h,v 1.5 2002/11/22 22:03:40 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Local definitions for memory allocation module
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _MALLOC_H
#define _MALLOC_H

// CONT:    contigious mem bocks of 512 bytes
// STACK:   fixed size buffers (e.g. 64/256/1536/8704 size buffers, each a fixed max nbr)
// FLEX:    flexible size and nbr of buffers.

#include "malloc-api.h"

#define SLAB_MAX_SIZE (256 * 1024)
#define SLAB_MIN_SIZE 4096
#define SLAB_IDX_SHIFT 12

#define MM_NUM_BUFS		mm_num_bufs

#if defined(NFP_FP)
#define SLAB_ALLOC_MAX_SIZE (32 * 1024)
#else
#define SLAB_ALLOC_MAX_SIZE (64 * 1024)
#endif

extern uint32		mm_num_bufs;
extern uint32		mm_max_contig;
extern uint32		mm_num_free;

extern uint32		* mm_validate;

extern void malloc_showStats(void);
extern void malloc_showFreeList(void);
extern void malloc_showAllocList(void);
extern void malloc_showFreeNodeStack(void);
extern void malloc_showAllocNodeStack(void);
extern void malloc_showAllocRanges(void);
extern void malloc_clearHistogram(void);
extern void malloc_showHistogram(int mode, int output, int type, int nbr);
extern int malloc_test(int argc, char **argv);
extern void malloc_showOwner(void);



/*------------------------------------------------------------------
 *
 * Buffer definition
 *
 * The buffer pointer returned to the user, is really a pointer within
 * a buffer descriptor.  There is a head and a tail structure for
 * debugging.
 *
 *------------------------------------------------------------------*/
#define MM_HEAD_SIG		0x12344321
#define MM_TAIL_SIG		0xFFEDDEFF
#define ALLOC_BUF_END	0xA7A7A7A7

typedef struct mm_base {
	void	           * flex;
    uint64             flexSize;
    nfx_spinlock_t     flexSpin;
	void	           * stack;
    uint64             stackSize;
    nfx_spinlock_t     stackSpin;
	void	           * cont;
    nfx_spinlock_t     contSpin;
} mm_base_t;

extern mm_base_t mmBase ;

typedef struct mm_flex_head mm_flex_head;

struct mm_flex_head {
	LIST_ENTRY(mm_flex_head)  entry;
	SLIST_ENTRY(mm_flex_head) pool;  // the pool list
	unsigned long reserved_1;
	uint32	size;               // actual allocated size
	uint32	length;             // buffer length, where to find the tail
	void	* malloc_ra;
	void	* free_ra;
#ifdef MM_USE_FLEX_TAIL
	uint32	reserved;
	uint32	sig;
#else
	void	* prev_free_ra;
#endif
};


#ifdef MM_USE_FLEX_TAIL
typedef struct mm_flex_tail {
	uint32	sig;
	uint32	res;
	void	* prev_malloc_ra;
	void	* prev_free_ra;
	uint32	res_1[2];            // align to 32 bytes
	uint32	res_2[8];            // align to 64 bytes
                                 // Todo: use this for debugging info
                                 // - more prev ra
                                 // - timestamp, when malloc, free
                                 // - other ?

} mm_flex_tail;
#endif

typedef struct mm_head {
	uint32	sig;
	uint32	length;
	void	* malloc_ra;
	void	* free_ra;
#if __long64
	uint32	res[2];
#endif
} mm_head_t;

typedef struct mm_tail {
	uint32	res;				/* could be overwritten by ALLOC_BUF_END, */
								/* so don't put anything useful here! */
	void	* prev_malloc_ra;
	void	* prev_free_ra;
	uint32	sig;
} mm_tail_t;

#define MM_BUFSIZE		512

typedef struct mm_bdesc {
	mm_head_t	head;
	uchar8		buf[MM_BUFSIZE];
	mm_tail_t	tail;
} mm_bdesc_t;

extern address_t	mm_bdesc_base;
extern address_t	mm_bdesc_end;


typedef struct mm_flex {
    LIST_HEAD(,mm_flex_head) freeList;
    LIST_HEAD(,mm_flex_head) useList;
	uint32	                 free;         // nbr of current free buffers
    uint32                   unavailable;
    uint64                   histogram[3];
#ifdef __long64
	uint64					 res_1;
	uint64					 res_2;
#else
	uint64					 res_1;
	uint64					 res_2;
	uint64					 res_3;
#endif	
} mm_flex_t;

typedef struct mm_flex_desc {
    void                     *start;       // Begin of flex buffer pool descriptors (fixed after init)
    void                     *bufStart;    // Begin of buffers in the Flex pool (fixed after init)
    void                     *poolStart;   // Begin of Pool (will move to poolend, as mem is used)
    void                     *poolEnd;     // End of memory pool (fixed after init)
    uint64                   poolAvail;    // Nbr of bytes available in pool (will decr to 0)
    uint64                   nbrFlexLists; // nbr of diff flexible buffers (fixed at init)
	SLIST_HEAD(, mm_flex_head) poolList;
} mm_flex_desc_t;


struct mm_bufctl
{
	SLIST_ENTRY(mm_bufctl) bc_next;
#ifdef MM_VERIFY_ALLOCS
	void *bc_addr;
#endif	
};


struct kmem_usage
{
	struct mm_cache *ku_cache;
	struct mm_slab *ku_slab;
};

extern struct mm_cache **size_to_cache;
extern struct kmem_usage *mm_usage;

/*------------------------------------------------------------------
 *
 * Free Buffer Stack definitions
 *
 *------------------------------------------------------------------*/

#define MM_NUM_STACK_SIZES		4
#define MM_FLEX_MAX             (64*1024)    // max malloc size in bytes, provided for flex buffers
#define MM_FLEX_MODULO_SHIFT    6            // shift for the modulo operation
                                             // !!! must be modulo: MALLOC_ALIGNED_SIZE !!!!


#ifdef NFP_FP

#define MM_FLEX_MEM_PCT         100      // % of mem used for flexible mem allocation

#define MM_STACK_MEM_PCT		0       // % of memory used for fixed size buf (stacks)

#define MM_STACK_SIZE_1 		(128-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_2 		(512-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_3 		(8192-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_4 		(0x4000-sizeof(mm_head_t)-sizeof(mm_tail_t))

#define MM_STACK_NUM_1_BUFS	    16		/* these are numbers relative to */
#define MM_STACK_NUM_2_BUFS	    32		/* each other */
#define MM_STACK_NUM_3_BUFS	    16
#define MM_STACK_NUM_4_BUFS	    1

#else  // NFP_TXRX, FCNIM 
#ifdef NFP_TXRX

#define MM_FLEX_MEM_PCT         100       // % of mem used for flexible mem allocation

#define MM_STACK_MEM_PCT		0		/* 50 % memory for stacks (fixed size buf) */

#define MM_STACK_SIZE_1 		(128-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_2 		(512-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_3 		(2048-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_4 		(0x4000-sizeof(mm_head_t)-sizeof(mm_tail_t))

#define MM_STACK_NUM_1_BUFS	16		/* these are numbers relative to */
#define MM_STACK_NUM_2_BUFS	16		/* each other */
#define MM_STACK_NUM_3_BUFS	16
#define MM_STACK_NUM_4_BUFS	8

#else // FCNIM and all others ;-)

#define MM_FLEX_MEM_PCT         100       // % of mem used for flexible mem allocation

#define MM_STACK_MEM_PCT		0	/* 50 % memory for stacks (fixed size buf) */

#define MM_STACK_SIZE_1 		(128-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_2 		(512-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_3 		(1024-sizeof(mm_head_t)-sizeof(mm_tail_t))
#define MM_STACK_SIZE_4 		(2048-sizeof(mm_head_t)-sizeof(mm_tail_t))

#define MM_STACK_NUM_1_BUFS	16		/* these are numbers relative to */
#define MM_STACK_NUM_2_BUFS	16		/* each other */
#define MM_STACK_NUM_3_BUFS	16
#define MM_STACK_NUM_4_BUFS	16

#endif
#endif

/* If these structures aren't powers of 2 in length (see MM_STACK_SIZE_*),
 * the compiler generates a divide instruction in MM_STACK_FREE_BUF.
 */
typedef struct mm_stack_1 {
	mm_head_t	head;
	uchar8		buf[MM_STACK_SIZE_1];
	mm_tail_t	tail;
} mm_stack_bdesc_1_t;

typedef struct mm_stack_2 {
	mm_head_t	head;
	uchar8		buf[MM_STACK_SIZE_2];
	mm_tail_t	tail;
} mm_stack_bdesc_2_t;

typedef struct mm_stack_3 {
	mm_head_t	head;
	uchar8		buf[MM_STACK_SIZE_3];
	mm_tail_t	tail;
} mm_stack_bdesc_3_t;

typedef struct mm_stack_4 {
	mm_head_t	head;
	uchar8		buf[MM_STACK_SIZE_4];
	mm_tail_t	tail;
} mm_stack_bdesc_4_t;

typedef struct buffer_stack {
	mm_bdesc_t	** top;
	uint32		avail;			/* how many are available in the free stack */
	uint32		unavail;		/* how many times have there not been enough */
								/* buffers in the malloc stack */
	uint32		allocated;		/* how many buffers have been allocated */
								/* from the free stack */
	uint32		freed;			/* how many buffers have been freed back to */
								/* the stack */
} mm_stack_t;

extern uchar8		* mm_stack_buf_base[];
extern uchar8		* mm_stack_buf_end[];
extern uint32		mm_stack_size_ratio[];
extern uint32		mm_stack_num_bufs[];
extern uint32		mm_stack_bdesc_size[];
extern uint32		mm_stack_buf_size[];
extern mm_stack_t	mm_stack[];

/*-----------------------------------------------------------------
 * Name:	MM_STACK_FREE_BUF
 *
 * Description:	To free a buffer back on the malloc stacks.
 *				Before calling this macro, we should have determined to which
 *				size stack the buffer belongs.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/14/01
 *
 *-----------------------------------------------------------------
 */
#if 1
#define MM_STACK_FREE_BUF(_buf, _index, _ra) {								\
	mm_bdesc_t	* _bdesc;													\
																			\
	if (_index == 0)														\
		_bdesc = (mm_bdesc_t *)(mm_stack_buf_base[_index] +					\
					(((_buf - mm_stack_buf_base[_index]) /					\
				sizeof(mm_stack_bdesc_1_t)) * sizeof(mm_stack_bdesc_1_t)));	\
	else if (_index == 1)													\
		_bdesc = (mm_bdesc_t *)(mm_stack_buf_base[_index] +					\
					(((_buf - mm_stack_buf_base[_index]) /					\
				sizeof(mm_stack_bdesc_2_t)) * sizeof(mm_stack_bdesc_2_t)));	\
	else if (_index == 2)													\
		_bdesc = (mm_bdesc_t *)(mm_stack_buf_base[_index] +					\
					(((_buf - mm_stack_buf_base[_index]) /					\
				sizeof(mm_stack_bdesc_3_t)) * sizeof(mm_stack_bdesc_3_t)));	\
	else if (_index == 3)													\
		_bdesc = (mm_bdesc_t *)(mm_stack_buf_base[_index] +					\
					(((_buf - mm_stack_buf_base[_index]) /					\
				sizeof(mm_stack_bdesc_4_t)) * sizeof(mm_stack_bdesc_4_t)));	\
	else																	\
		panic("Invalid index, %d, in MM_STACK_FREE_BUF\n", _index);			\
	if (((mm_head_t *)_bdesc)->free_ra) panic("dup memfree");               \
	((mm_head_t *)_bdesc)->length = 0;										\
	((mm_head_t *)_bdesc)->free_ra = _ra;									\
	mm_stack[_index].avail++;												\
	mm_stack[_index].freed++;												\
    mm_stats.total_stack_used -= mm_stack_buf_size[_index];					\
	*--mm_stack[_index].top = _bdesc;										\
}
#else
#define MM_STACK_FREE_BUF(_buf, _index, _ra) {								\
	mm_bdesc_t	* _bdesc;													\
																			\
	_bdesc = (mm_bdesc_t *)(mm_stack_buf_base[_index] +						\
					(((_buf - mm_stack_buf_base[_index]) /					\
									mm_stack_bdesc_size[_index]) *			\
											mm_stack_bdesc_size[_index]));	\
	if (((mm_head_t *)_bdesc)->free_ra) panic("dup memfree");               \
	((mm_head_t *)_bdesc)->length = 0;										\
	((mm_head_t *)_bdesc)->free_ra = _ra;									\
	mm_stack[_index].avail++;												\
	mm_stack[_index].freed++;												\
    mm_stats.total_stack_used -= mm_stack_buf_size[_index];					\
	*--mm_stack[_index].top = _bdesc;										\
}
#endif


#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
/*-----------------------------------------------------------------
 * Name:	MM_STACK_ALLOC_BUF
 *
 * Description:	To try and alloc a buffer from the malloc stacks.
 *				This will return 0 if a buffer could not be allocated from
 *				the stacks.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/14/01
 *
 *-----------------------------------------------------------------
 */
#define MM_STACK_ALLOC_BUF(_size, _ra) ({							\
	mm_bdesc_t	* _bdesc = 0;										\
	mm_head_t	* _head;											\
	mm_tail_t	* _tail;											\
	uchar8		* _r = 0;											\
	uint32		_i = MM_NUM_STACK_SIZES;							\
																	\
	if (_size <= MM_STACK_SIZE_1) {									\
		if (mm_stack[0].avail)										\
			_i = 0;													\
		else {														\
			mm_stack[0].unavail++;									\
			if (mm_stack[1].avail)									\
				_i = 1;												\
			else													\
				mm_stack[1].unavail++;								\
		}															\
	}																\
	else if (_size <= MM_STACK_SIZE_2) {							\
		if (mm_stack[1].avail)										\
			_i = 1;													\
		else {														\
			mm_stack[1].unavail++;									\
			if (mm_stack[2].avail)									\
				_i = 2;												\
			else													\
				mm_stack[2].unavail++;								\
		}															\
	}																\
	else if (_size <= MM_STACK_SIZE_3) {							\
		if (mm_stack[2].avail)										\
			_i = 2;													\
		else {														\
			mm_stack[2].unavail++;									\
			if (mm_stack[3].avail)									\
				_i = 3;												\
			else													\
				mm_stack[3].unavail++;								\
		}															\
	}																\
	else if (_size <= MM_STACK_SIZE_4) {							\
		if (mm_stack[3].avail)										\
			_i = 3;													\
		else														\
			mm_stack[3].unavail++;									\
	}																\
																	\
	if (_i < MM_NUM_STACK_SIZES) {									\
		_bdesc = (mm_bdesc_t *)*mm_stack[_i].top++;					\
		mm_stack[_i].avail--;										\
		mm_stack[_i].allocated++;									\
    	mm_stats.total_stack_used += mm_stack_buf_size[_i];		    \
		_head = (mm_head_t *)_bdesc;								\
		_tail = (mm_tail_t *)(_bdesc->buf + mm_stack_buf_size[_i]);	\
		_tail->prev_malloc_ra = _head->malloc_ra;					\
		_tail->prev_free_ra = _head->free_ra;						\
		_head->length = _size;										\
		_head->malloc_ra = _ra;										\
		_head->free_ra = 0;											\
		CACHEOP(Hit_Writeback_Inv_S, _head);						\
		CACHEOP(Hit_Writeback_Inv_S, _tail);						\
		_r = _bdesc->buf;											\
	}																\
	_r;																\
})

#else

/*-----------------------------------------------------------------
 * Name:	MM_STACK_ALLOC_BUF
 *
 * Description:	To try and alloc a buffer from the malloc stacks.
 *				This will return 0 if a buffer could not be allocated from
 *				the stacks.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/14/01
 *
 *-----------------------------------------------------------------
 */
#define MM_STACK_ALLOC_BUF(_size, _ra) ({							\
	mm_bdesc_t	* _bdesc = 0;										\
	mm_head_t	* _head;											\
	mm_tail_t	* _tail;											\
	uchar8		* _r = 0;											\
	uint32		_i = MM_NUM_STACK_SIZES;							\
																	\
	if (_size <= MM_STACK_SIZE_1) {									\
		if (mm_stack[0].avail)										\
			_i = 0;													\
		else {														\
			mm_stack[0].unavail++;									\
			if (mm_stack[1].avail)									\
				_i = 1;												\
			else													\
				mm_stack[1].unavail++;								\
		}															\
	}																\
	else if (_size <= MM_STACK_SIZE_2) {							\
		if (mm_stack[1].avail)										\
			_i = 1;													\
		else {														\
			mm_stack[1].unavail++;									\
			if (mm_stack[2].avail)									\
				_i = 2;												\
			else													\
				mm_stack[2].unavail++;								\
		}															\
	}																\
	else if (_size <= MM_STACK_SIZE_3) {							\
		if (mm_stack[2].avail)										\
			_i = 2;													\
		else {														\
			mm_stack[2].unavail++;									\
			if (mm_stack[3].avail)									\
				_i = 3;												\
			else													\
				mm_stack[3].unavail++;								\
		}															\
	}																\
	else if (_size <= MM_STACK_SIZE_4) {							\
		if (mm_stack[3].avail)										\
			_i = 3;													\
		else														\
			mm_stack[3].unavail++;									\
	}																\
																	\
	if (_i < MM_NUM_STACK_SIZES) {									\
		_bdesc = (mm_bdesc_t *)*mm_stack[_i].top++;					\
		mm_stack[_i].avail--;										\
		mm_stack[_i].allocated++;									\
    	mm_stats.total_stack_used += mm_stack_buf_size[_i];		    \
		_head = (mm_head_t *)_bdesc;								\
		_tail = (mm_tail_t *)(_bdesc->buf + mm_stack_buf_size[_i]);	\
		_tail->prev_malloc_ra = _head->malloc_ra;					\
		_tail->prev_free_ra = _head->free_ra;						\
		_head->length = _size;										\
		_head->malloc_ra = _ra;										\
		_head->free_ra = 0;											\
		_r = _bdesc->buf;											\
	}																\
	_r;																\
})
#endif	/* EVAL_BOARD || AGILE_HARDWARE */

/*------------------------------------------------------------------
 *
 * List structure definitions
 *
 *------------------------------------------------------------------*/

typedef struct mm_free_node {
	uint32				size;
	mm_bdesc_t			* addr;
	struct mm_free_node	* Snext;		/* for size-based list */
	struct mm_free_node	* Sprev;
	struct mm_free_node	* Anext;		/* for address-based list */
	struct mm_free_node	* Aprev;
} mm_free_node_t;

typedef struct mm_alloc_node {
	uint32					size;
	mm_bdesc_t				* addr;
	struct mm_alloc_node	* next;
} mm_alloc_node_t;

typedef struct free_stack {
	uint32			avail;
	mm_free_node_t	** top;
} mm_free_node_stack_t;

typedef struct alloc_stack {
	uint32			avail;
	mm_alloc_node_t	** top;
} mm_alloc_node_stack_t;

extern mm_free_node_t			mm_free_list;
extern mm_alloc_node_t			mm_alloc_list;
extern mm_free_node_stack_t		mm_free_node_stack;
extern mm_alloc_node_stack_t	mm_alloc_node_stack;
extern mm_bdesc_t	* mm_bdesc_list;

extern mm_free_node_t	* mm_free_stack_base;
extern mm_alloc_node_t	* mm_alloc_stack_base;
extern mm_free_node_t	* mm_free_node_list;
extern mm_alloc_node_t	* mm_alloc_node_list;

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	Scans free_size_list and finds best fit for the requested
 *				number of blocks.  Returns the node pointer or NULL.
 *
 *				Really, NULL should never return, since if there is
 *				not a contiguous buffer large enough, we should have
 *				checked that before calling this macro.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_FIND_FREE_SPACE(_nb)			\
({										\
	mm_free_node_t	* _ptr, * _r;		\
										\
	_r = NULL;							\
	_ptr = mm_free_list.Snext;			\
										\
	while (_ptr != &(mm_free_list)) {	\
		if (_ptr->size >= (_nb)) {		\
			_r = _ptr;					\
			break;						\
		}								\
		else							\
			_ptr = _ptr->Snext;			\
	}									\
	_r;									\
})

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	Retrieves a free_node_t node from the free pool.  Returns
 *				pointer to the node.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_GET_FREE_NODE()					\
({											\
	mm_free_node_t	* _r;					\
											\
	if (mm_free_node_stack.avail) {			\
		_r = *mm_free_node_stack.top++;		\
		mm_free_node_stack.avail--;			\
	}										\
	else									\
		panic("Error in GET_FREE_NODE\n");	\
											\
	_r;										\
})

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	Releases free_node_t node back into free pool
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_FREE_FREE_NODE(_np)							\
{														\
	*--mm_free_node_stack.top = (mm_free_node_t *)_np;	\
	mm_free_node_stack.avail++;							\
}

/*-----------------------------------------------------------------
 * Name:
 *

 * Description:	Retrives an alloc_node_t node from the free pool.  Returns
 *				pointer to the node.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_GET_ALLOC_NODE()					\
({											\
	mm_alloc_node_t	* _r;					\
											\
	if (mm_alloc_node_stack.avail) {		\
		_r = *mm_alloc_node_stack.top++;	\
		mm_alloc_node_stack.avail--;		\
	}										\
	else									\
		panic("Error in GET_ALLOC_NODE\n");	\
											\
	_r;										\
})

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	Releases alloc_node_t node back into free pool
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_FREE_ALLOC_NODE(_anp)							\
{															\
	*--mm_alloc_node_stack.top = (mm_alloc_node_t *)_anp;	\
	mm_alloc_node_stack.avail++;							\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	Create an alloc node and insert it into alloc_list.
 *				The node should be filled in with the real buffer base,
 *				and the number of buffers included in this area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_INSERT_IN_ALLOC_LIST(_anp)				\
{													\
	mm_alloc_node_t	* _ptr;							\
													\
	_ptr = &(mm_alloc_list);						\
													\
	while (_ptr->next) {							\
		if (_ptr->next->addr < _anp->addr)			\
			_ptr = _ptr->next;						\
		else										\
			break;									\
	}												\
													\
	_anp->next = _ptr->next;						\
	_ptr->next = _anp;								\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_FIND_BUFFER_IN_ALLOC_LIST(_addr, _bdesc) {			\
	mm_alloc_node_t	* _ptr;										\
																\
	_ptr = mm_alloc_list.next;									\
	_bdesc = 0;													\
																\
	while (_ptr) {												\
		if (((address_t)_addr >= (address_t)_ptr->addr->buf) &&				\
			((address_t)_addr <												\
				(address_t)_ptr->addr->buf + _ptr->addr->head.length)) {	\
			_bdesc = _ptr->addr;								\
			break;												\
		}														\
		else													\
			_ptr=_ptr->next;									\
	}															\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	Searches the alloc'd list for the buffer which contains
 *				the address being freed.  Removes the node from the 
 *				alloc list, and returns the real buffer base, and number
 *				of buffers that had been allocated for this area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_DELETE_BUFFER_FROM_ALLOC_LIST(_bufptr, _bufbase, _nb)	\
{																	\
	mm_alloc_node_t	* _ptr, * _t1, * _t2;							\
																	\
	_t1 = &(mm_alloc_list);											\
	_t2 = _t1->next;												\
																	\
	if (_t2->addr > (mm_bdesc_t *)_bufptr)							\
		panic("Error in DELETE_BUFFER\n");							\
																	\
	_ptr = _t2->next;												\
	if (!_ptr)														\
		goto mm_delete_found;										\
																	\
	while (_ptr) {													\
		if (_ptr->addr > (mm_bdesc_t *)_bufptr)						\
			goto mm_delete_found;									\
		else {														\
			_t1 = _t2;												\
			_t2 = _ptr;												\
			_ptr = _ptr->next;										\
		}															\
	}																\
																	\
	if (!_ptr) {													\
		if (_t2->addr <= (mm_bdesc_t *)_bufptr)						\
			goto mm_delete_found;									\
		else														\
			panic("Error in DELETE BUFFER\n");						\
	}																\
																	\
mm_delete_found:													\
	_bufbase = _t2->addr;											\
	_nb = _t2->size;												\
	_t1->next = _ptr;												\
	MM_FREE_ALLOC_NODE(_t2);										\
}
	
/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	Removes node from free list, and frees node back onto
 *				free node stack.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_REMOVE_FREE_NODE(_np)					\
{													\
	_np->Sprev->Snext = _np->Snext;					\
	_np->Snext->Sprev = _np->Sprev;					\
													\
	_np->Aprev->Anext = _np->Anext;					\
	_np->Anext->Aprev = _np->Aprev;					\
													\
	MM_FREE_FREE_NODE(_np);							\
													\
	MM_UPDATE_MAX_CONTIG();							\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_INSERT_IN_FREE_LIST(_addr, _size)					\
{																\
	mm_free_node_t	* _np, * _tmp;								\
																\
	_np = MM_GET_FREE_NODE();									\
																\
	_np->addr = _addr;											\
	_np->size = _size;											\
																\
	MM_INSERT_IN_FREE_ADDR_LIST(_np);							\
																\
	if (_np->Anext != &(mm_free_list)) {						\
		if (_np->Anext->addr == _addr + _size) {				\
			_np->size += _np->Anext->size;						\
			_tmp = _np->Anext;									\
			MM_REMOVE_FREE_NODE(_tmp);							\
		}														\
	}															\
																\
	if (_np->Aprev != &(mm_free_list)) {						\
		if (_np->Aprev->addr + _np->Aprev->size == _addr) {		\
			_np->addr = _np->Aprev->addr;						\
			_np->size += _np->Aprev->size;						\
			_tmp = _np->Aprev;									\
			MM_REMOVE_FREE_NODE(_tmp);							\
		}														\
	}															\
																\
	MM_INSERT_IN_FREE_SIZE_LIST(_np);							\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_INSERT_IN_FREE_SIZE_LIST(_np)								\
{																		\
	mm_free_node_t	* _ptr;												\
																		\
	_ptr = &(mm_free_list);												\
																		\
	while (1) {															\
		if ((_ptr->Snext == &(mm_free_list)) ||							\
								(_ptr->Snext->size > _np->size)) {		\
			_np->Snext = _ptr->Snext;									\
			_np->Sprev = _ptr;											\
			_ptr->Snext->Sprev = _np;									\
			_ptr->Snext = _np;											\
			break;														\
		}																\
		else															\
			_ptr = _ptr->Snext;											\
	}																	\
																		\
	MM_UPDATE_MAX_CONTIG();												\
}

#define MM_INSERT_IN_FREE_ADDR_LIST(_np)								\
{																		\
	mm_free_node_t	* _ptr;												\
																		\
	_ptr = &(mm_free_list);												\
																		\
	while (1) {															\
		if ((_ptr->Anext == &(mm_free_list)) ||							\
								(_ptr->Anext->addr > _np->addr)) {		\
			_np->Anext = _ptr->Anext;									\
			_np->Aprev = _ptr;											\
			_ptr->Anext->Aprev = _np;									\
			_ptr->Anext = _np;											\
			break;														\
		}																\
		else															\
			_ptr = _ptr->Anext;											\
	}																	\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	_np has new contents.  This macro should remove it from
 *				its current location and insert it at the proper location
 *				based on size.  The addr should still be in the same order
 *				as before, so no change is due to the address order.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_CHANGE_FREE_SIZE_NODE(_np)					\
{														\
	if (_np->size >= _np->Sprev->size) {				\
		/*												\
		 * still in order, do nothing					\
		 */												\
		MM_UPDATE_MAX_CONTIG();							\
	}													\
	else {												\
		/*												\
		 * Remove from list								\
		 */												\
		_np->Sprev->Snext = _np->Snext;					\
		_np->Snext->Sprev = _np->Sprev;					\
		/*												\
		 * Find new location, and insert it				\
		 */												\
		MM_INSERT_IN_FREE_SIZE_LIST(_np);				\
	}													\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#define MM_UPDATE_MAX_CONTIG()							\
{														\
	mm_max_contig = mm_free_list.Sprev->size;			\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#ifdef MM_VALIDATE

#define MM_VALIDATE_ALLOC(_bdesc, _nb)						\
{															\
	uint32	_bdesc_num;										\
	uint32	_word_num, _bit_pos;							\
	uint32	_size;											\
															\
	_bdesc_num = (uint32)((address_t)_bdesc -				\
			(address_t)mm_bdesc_list)/sizeof(mm_bdesc_t);	\
															\
	for (_size = 0; _size < _nb; _size++) {					\
		_word_num = (_bdesc_num+_size) / 32;				\
		_bit_pos = (_bdesc_num+_size) % 32;					\
															\
		if (mm_validate[_word_num] & (1 << _bit_pos))		\
			panic("bdesc already allocated\n");				\
		else												\
			mm_validate[_word_num] |= (1 << _bit_pos);		\
	}														\
}

#else

#define MM_VALIDATE_ALLOC(_bdesc, _nb)

#endif

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#ifdef MM_VALIDATE

#define MM_VALIDATE_FREE(_bdesc, _nb)						\
{															\
	uint32	_bdesc_num;										\
	uint32	_word_num, _bit_pos;							\
	uint32	_size;											\
															\
	_bdesc_num = (uint32)((address_t)_bdesc -				\
			(address_t)mm_bdesc_list)/sizeof(mm_bdesc_t);	\
															\
	for (_size = 0; _size < _nb; _size++) {					\
		_word_num = (_bdesc_num+_size) / 32;				\
		_bit_pos = (_bdesc_num+_size) % 32;					\
															\
		if (mm_validate[_word_num] & (1 << _bit_pos))		\
			mm_validate[_word_num] &= ~(1 << _bit_pos);		\
		else												\
			panic("bdesc already freed\n");					\
	}														\
}

#else

#define MM_VALIDATE_FREE(_bdesc, _nb)

#endif

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:	For debugging and recordkeeping, increment counters.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#ifdef MM_VALIDATE

#define MM_INC_BYTE_RANGE(_n)											\
{																		\
	uint32	_i, _j, _inc;												\
																		\
	_i = 0;																\
	_j = 0;																\
																		\
	for (_inc=0x20 ; _j < 0x200; _j += _inc, _i++) {	/* 0 - 512 */		\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	for (_inc=0x40 ; _j < 0x400; _j += _inc, _i++) {	/* 512 - 1K */		\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	for (_inc=0x80 ; _j < 0x1000; _j += _inc, _i++) {	/* 1K - 4K */		\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	for (_inc=0x100 ; _j < 0x2000; _j += _inc, _i++) {	/* 4K - 8K */		\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	for (_inc=0x100 ; _j < 0x4000; _j += _inc, _i++) {	/* 8K - 16K */		\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	for (_inc=0x400 ; _j < 0x8000; _j += _inc, _i++) {	/* 16K - 32K */		\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	for (_inc=0x800 ; _j < 0x10000; _j += _inc, _i++) {	/* 32K - 64K */		\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	for (_inc=0x4000 ; _j < 0x80000; _j += _inc, _i++) {/* 64K - 512K */	\
		if (_n < _j+_inc) {												\
			mm_stats.malloc_size[_i]++;									\
			goto _done_with_inc;										\
		}																\
	}																	\
	mm_stats.malloc_size[_i]++;						/* >= 512K */		\
																		\
_done_with_inc:															\
	if (_i > MM_NUM_BYTE_RANGES)										\
		panic("Invalid byte range\n");									\
}

#else

#define MM_INC_BYTE_RANGE(_n)

#endif


/*******************************************************************
 * Defines for enabling the histogram of the memory usage
 *******************************************************************/

#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef NFP_FP
#define MEM_HISTOGRAM 1
//#define MEM_HISTOGRAM_BUCKET 1
#endif
#ifdef NFP_TXRX
#define MEM_HISTOGRAM 1
//#define MEM_HISTOGRAM_BUCKET 1
#endif
#ifdef FCNIM
#define MEM_HISTOGRAM 1
//#define MEM_HISTOGRAM_BUCKET 1
#endif
#endif


#ifdef MEM_HISTOGRAM_BUCKET

#define MEM_HIST_NBR_BUCKETS    6
#define MEM_HIST_BUCKET_TYPE_FLEX_ALLOC       0
#define MEM_HIST_BUCKET_TYPE_STACK_ALLOC      1
#define MEM_HIST_BUCKET_TYPE_CONT_ALLOC       2
#define MEM_HIST_BUCKET_TYPE_FLEX_DEALLOC     3
#define MEM_HIST_BUCKET_TYPE_STACK_DEALLOC    4
#define MEM_HIST_BUCKET_TYPE_CONT_DEALLOC     5
#define MEM_HIST_BUCKET_TYPES                 6

extern int32 memHistogramBucketsSize[MEM_HIST_NBR_BUCKETS];
extern int64 memHistogramBuckets [MEM_HIST_BUCKET_TYPES][MEM_HIST_NBR_BUCKETS];

#endif // MEM_HISTOGRAM_BUCKET


#define MM_HIST_MAX   0
#define MM_HIST_CUR   1
#define MM_HIST_TOT   2
#define MM_HIST_SORT  MM_HIST_MAX

#ifdef MEM_HISTOGRAM

#define HISTOGRAM_MAX      10000
#define HISTOGRAM_MODULO   4

extern int32 mem_histogram [3][HISTOGRAM_MAX] ;

#endif // MEM_HISTOGRAM

extern int32 mem_histogram_cont [3] ;

extern void *mm_pool_alloc(unsigned long size, unsigned long align);
extern void mm_pool_free(void *obj);
extern void mm_pool_free_size(void *obj, unsigned long size);
extern void free(void *obj);
extern void cache_free_ra(struct mm_cache *cache, void *obj, void *ra);
extern void *mm_alloc_block(unsigned int size);
extern void mm_free_block(void *mem, unsigned int size);

extern void cache_show(void);
extern void pool_show(void);

#endif	/* _MALLOC_H */


