/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ndmp-auth-cmd.c
 *
 * Description  : Command line interface for NDMP authentication.
 *
 * This interface allows to manage NDMP clients and own NDMP accounts.
 * The information is stored persistently in the registry file.
 * Passwords are encrypted with a 2-ways algorithm.
 * Example:
 *      options.ndmp.auth.server.gilbert=56ab6f
 *      options.ndmp.auth.client.peter=4g2d30
 *      options.ndmp.auth.client.isabelle=23eda67
 *
 * Note/TODO    :
 * - How many NDMP servers do we have? Current assumption is as many as we want!
 * - Should we have a limitation in the number of NDMP clients?
 *   If yes, how many?
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "ndmp-auth.h"
#include "ndmp-auth-api.h"
#include <ctype.h>

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* The name of this program.
 */
static char * myName;

/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */

static void
ndmp_printUsage(void);

static int
ndmp_sendSetReq(opt_setOptionReq_t *    setReq);

static int
ndmp_addClient(char * id,
               char * passwd);

static int
ndmp_delClient(char * id);

static int
ndmp_modClient(char * id,
               char * passwd);

static int
ndmp_addServer(char * id,
               char * passwd);

static int
ndmp_delServer(char * id);

static int
ndmp_modServer(char * id,
               char * passwd);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ndmp_printUsage()
 *
 * Description  : Print the correct usage of this program.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ndmp_printUsage(void)

{
	fprintf(stderr,"Usage:\n");
    
	fprintf(stderr," ndmp auth client add USERNAME\n");
	fprintf(stderr," ndmp auth client del USERNAME\n");
	fprintf(stderr," ndmp auth client mod USERNAME\n");
	// fprintf(stderr,"\n");
    fprintf(stderr," ndmp auth server add USERNAME\n");
	fprintf(stderr," ndmp auth server del USERNAME\n");
	fprintf(stderr," ndmp auth server mod USERNAME\n");
}

/*------------------------------------------------------------------------------
 * Name         : ndmp_sendSetReq()
 *
 * Description  :
 * Send a 'set option' request to the registry manager.
 * Return the response from the registry manager.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ndmp_sendSetReq(opt_setOptionReq_t *    setReq)
    
{
    opt_msgType_t   reqType = OPT_SET_OPTION_REQ;
    ndmp_msg_t      req;
    ndmp_msg_t      rsp;
    opt_msgHdr_t *  hdr;
    void *          bufp = NULL;
    int             result;

    NDMP_FCT_PRINT(("> ndmp_sendSetReq()\n"));

    /* Fill the request.
     */
    bufp = &req.buf;
    bzero(bufp,
          (sizeof(opt_msgHdr_t) +
           sizeof(opt_setOptionReq_t)));
    
    memcpy(bufp, (void *) &reqType, sizeof(opt_msgType_t));
    bufp += sizeof(opt_msgType_t);
    // userInfo is not used
    bufp += sizeof(opt_userInfo_t);
    memcpy(bufp, (void *) setReq, sizeof(opt_setOptionReq_t));
    bufp += sizeof(opt_setOptionReq_t);

    req.len = bufp - (void *) req.buf;

    /* Send the request to the registry manager located on the SSC.
     */
    rsp.len = sendAgileMsg(-1,
                           req.buf, req.len,
                           "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &waitRegmTo,
                           rsp.buf, sizeof(rsp.buf));
    
    /* Receive response from the registry manager.
     */
    NDMP_DBG_PRINT(("nb bytes received = %d\n", rsp.len));
    if (rsp.len == -1)
    {
		perror("sendAgileMsg failed");
		return(NFX_ERR);
    }

    /* Read the message header and check if it's expected response.
     */
    bufp = &rsp.buf;
    hdr = (opt_msgHdr_t *) rsp.buf;
    bufp += sizeof(opt_msgHdr_t);
    if (hdr->msgType != OPT_SET_OPTION_RSP)
    {
        NDMP_ERR_PRINT(("Message received from REGM is not the expected one!\n"));
		return(NFX_ERR);
    }
    
    /* Read result.
     */
    memcpy((void *) &result, bufp, sizeof(int));
    bufp += sizeof(int);
    
    NDMP_FCT_PRINT(("< ndmp_sendSetReq()\n"));
    return(result);
}

/*------------------------------------------------------------------------------
 * Name         : ndmp_addClient()
 *
 * Description  : Add a NDMP client with its (encoded) password.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ndmp_addClient(char * id,
               char * passwd)

{
    int                 rc;
    opt_setOptionReq_t  req;
    char                encodedPasswd[CRYPT_MAX_PASSWD_LEN];

    /* Encode password.
     */
    rc = crypt_encode(passwd, encodedPasswd);

    if (rc == NFX_OK)
    {
        /* Fill request.
         */
        req.del = FALSE;
        strcpy(req.entry.option, "options.ndmp.auth.client.");
        strcat(req.entry.option, id);
        strcpy(req.entry.value, encodedPasswd);
        
        /* Send the request.
         */
        rc = ndmp_sendSetReq(&req);
    }
    
    NDMP_FCT_PRINT(("< ndmp_addClient()\n"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : ndmp_delClient()
 *
 * Description  : Delete a NDMP client.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ndmp_delClient(char * id)

{
    int                 rc;
    opt_setOptionReq_t  req;

    /* Fill request.
     */
    req.del = TRUE;
    strcpy(req.entry.option, "options.ndmp.auth.client.");
    strcat(req.entry.option, id);
    strcpy(req.entry.value, "");    // not used

    /* Send the request.
     */
    rc = ndmp_sendSetReq(&req);

    NDMP_FCT_PRINT(("< ndmp_addClient()\n"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : ndmp_modClient()
 *
 * Description  : Modify the password of a NDMP client.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ndmp_modClient(char * id,
               char * passwd)

{
    return(ndmp_addClient(id, passwd));
}

/*------------------------------------------------------------------------------
 * Name         : ndmp_addServer()
 *
 * Description  : Add own NDMP account and (encoded) password.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ndmp_addServer(char * id,
               char * passwd)

{
    int                 rc;
    opt_setOptionReq_t  req;
    char                encodedPasswd[CRYPT_MAX_PASSWD_LEN];

    /* Encode password.
     */
    rc = crypt_encode(passwd, encodedPasswd);

    if (rc == NFX_OK)
    {
        /* Fill request.
         */
        req.del = FALSE;
        strcpy(req.entry.option, "options.ndmp.auth.server.");
        strcat(req.entry.option, id);
        strcpy(req.entry.value, encodedPasswd);
        
        /* Send the request.
         */
        rc = ndmp_sendSetReq(&req);
    }
    
    NDMP_FCT_PRINT(("< ndmp_addServer()\n"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : ndmp_delServer()
 *
 * Description  : Delete own NDMP account.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ndmp_delServer(char * id)

{
    int                 rc;
    opt_setOptionReq_t  req;

    /* Fill request.
     */
    req.del = TRUE;
    strcpy(req.entry.option, "options.ndmp.auth.server.");
    strcat(req.entry.option, id);
    strcpy(req.entry.value, "");    // not used

    /* Send the request.
     */
    rc = ndmp_sendSetReq(&req);

    NDMP_FCT_PRINT(("< ndmp_addServer()\n"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : ndmp_modServer()
 *
 * Description  : Modify the password of own NDMP account.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ndmp_modServer(char * id,
               char * passwd)

{
    return(ndmp_addServer(id, passwd));
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ndmp_authenticate()
 *
 * Description  : It all begins here.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int
main(int argc, char *argv[])

{
    int     rc, i;
	char *  actionStr;
	char *  idStr;

    /* Initialize my own name.
     * Discard path if any.
     */
	myName = argv[0];
	if (rindex(myName, '/'))
    {
		strcpy(myName, rindex(myName, '/') + 1);
	}
    
    /* Make sure there's at least the action and the identifier strings.
     */
	if (argc < 3)
    {
		ndmp_printUsage();
		exit(NFX_ERR);
	}
	actionStr = argv[1];

    if (!strcmp(argv[2], "server"))
    {
        if (argc != 4)
        {
            ndmp_printUsage();
            exit(NFX_ERR);
        }
        idStr = argv[3];
    }
    else
    {
        if (argc != 3)
        {
            ndmp_printUsage();
            exit(NFX_ERR);
        }
        idStr = argv[2];
    }

    if (strlen(idStr) > 64)
    {
        fprintf(stderr, "User name too long. (Max. 64 characters)\n");
        ndmp_printUsage();
        exit(NFX_ERR);
    }
    else
    {
        i = 0;
        while (i < strlen(idStr))
        {
            if (!isalpha(idStr[i]) && !isdigit(idStr[i]) &&
                 idStr[i] != '-' && idStr[i] != '_')
            {
                fprintf(stderr, "Only [0-9], [a-z], [A-Z], [-](dash) and [_](underscore) characters allowed in USERNAME\n");
                ndmp_printUsage();
                exit(NFX_ERR);
            }
            i++;
        }
    }
    
    /* Each "command" has it's own action function.
     */
	if (!strcasecmp(actionStr, "add"))
    {
        // add account into file
        char    password[CRYPT_MAX_PASSWD_LEN];

        rc = crypt_getPasswd(password);
        if (rc == NFX_OK)
        {
            if (!strcmp(argv[2], "server"))
            {
                rc = ndmp_addServer(idStr, password);
            }
            else
            {
                rc = ndmp_addClient(idStr, password);
            }
        }
	}
	else if(!strcmp(actionStr, "del"))
    {
        // delete account from file
        if (!strcmp(argv[2], "server"))
        {
            rc = ndmp_delServer(idStr);
        }
        else
        {
            rc = ndmp_delClient(idStr);
        }
	}
	else if(!strcmp(actionStr, "mod"))
    {
        // modify password of the account into file
        char    password[CRYPT_MAX_PASSWD_LEN];

        rc = crypt_getPasswd(password);
        if (rc == NFX_OK)
        {
            if (!strcmp(argv[2], "server"))
            {
                rc = ndmp_modServer(idStr, password);
            }
            else
            {
                rc = ndmp_modClient(idStr, password);
            }
        }
	}
	else
    {
		fprintf(stderr, "Invalid command: %s\n", actionStr);
		ndmp_printUsage();
		rc = NFX_ERR;
	}
    
	exit(rc);
}

/*----------------------------------------------------------------------------*/
