/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ndmp-auth.c
 *
 * Description  : NDMP authentication library.
 * Contain all routines for NDMP authentication.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "ndmp-auth.h"
#include "ndmp-auth-api.h"

/*------------------------------------------------------------------------------
 * Constants and macros.
 *------------------------------------------------------------------------------
 */

/* Length of the NDMP entry in the registry file.
 */
#define NDMP_AUTH_ENTRY_LEN (OPT_MAX_OPTION_LEN + MAX_VALUE_LEN)

/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */

static int                          // NFX_OK, NFX_ERR  - OUTPUT
ndmp_getPassword(char * id,         // identifier       - INPUT
                 char * password);  // password         - OUTPUT
    
/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ndmp_getPassword()
 *
 * Description  : Get the password of a given NDMP client from the registry.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */


static int                          // NFX_OK, NFX_ERR  - OUTPUT
ndmp_getPassword(char * id,         // identifier       - INPUT
                 char * password)   // password         - OUTPUT

{
    int             rc = NFX_OK;
    char            encodedPasswd[CRYPT_MAX_PASSWD_LEN];
    boolean         found = FALSE;
    opt_msgType_t   reqType = OPT_GET_OPTION_REQ;
    ndmp_msg_t      req;
    ndmp_msg_t      rsp;
    opt_msgHdr_t *  hdr;
    void *          bufp = NULL;
    opt_option_t    option;
    int             nbEntry;
    int             i;

    NDMP_FCT_PRINT(("> ndmp_getPassword()\n"));

    /* Fill the request.
     */
    bufp = &req.buf;
    bzero(bufp,
          (sizeof(opt_msgHdr_t) +
           sizeof(opt_getOptionReq_t)));
    
    memcpy(bufp, (void *) &reqType, sizeof(opt_msgType_t));
    bufp += sizeof(opt_msgType_t);
    // userInfo is not used
    bufp += sizeof(opt_userInfo_t);
    strcpy(option, "options.ndmp.auth.client.");
    strcat(option, id);
    memcpy(bufp, (void *) option, sizeof(opt_option_t));
    bufp += sizeof(opt_option_t);

    req.len = bufp - (void *) req.buf;

    /* Send the request to the registry manager located on the SSC.
     */
    rsp.len = sendAgileMsg(-1,
                           req.buf, req.len,
                           "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &waitRegmTo,
                           rsp.buf, sizeof(rsp.buf));
    
    /* Receive response from the registry manager.
     */
    NDMP_DBG_PRINT(("nb bytes received = %d\n", rsp.len));
    if (rsp.len == -1)
    {
		perror("sendAgileMsg failed");
		return(NFX_ERR);
    }

    /* Read the message header and check if it's expected response.
     */
    bufp = &rsp.buf;
    hdr = (opt_msgHdr_t *) rsp.buf;
    bufp += sizeof(opt_msgHdr_t);
    if (hdr->msgType != OPT_GET_OPTION_RSP)
    {
        NDMP_ERR_PRINT("Message received from REGM is not the expected one!\n");
		return(NFX_ERR);
    }
    
    /* Read the response and print its content.
     */
    memcpy((void *) &nbEntry, bufp, sizeof(int));
    bufp += sizeof(int);

    for (i = 0; i < nbEntry; i++)
    {
        opt_entry_t * entry;
        char          currId[NDMP_AUTH_ENTRY_LEN];

        entry = (opt_entry_t *) bufp;
        bufp += sizeof(opt_entry_t);
        NDMP_DBG_PRINT(("%-45.45s\t%s\n", entry->option, entry->value));

        strcpy(currId, rindex((char *) entry, '.') + 1);
        if (!strcmp(currId, id))
        {
            found = TRUE;
            strcpy(encodedPasswd, entry->value);
            NDMP_DBG_PRINT(("Used?...\n"));
            if (index(encodedPasswd, '\n'))
            {
                NDMP_DBG_PRINT(("...YES!\n"));
                *index(encodedPasswd, '\n') = 0;
            }
            break;
        }
    } // for

    if (found == TRUE)
    {
        /* Decode password.
         */
        rc = crypt_decode(encodedPasswd, password);
    }
    else
    {
        NDMP_ERR_PRINT(("User id %s doesn't exist!\n", id));
        rc = NFX_ERR;
    }
    
    NDMP_FCT_PRINT(("< ndmp_getPassword()\n"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ndmp_authenticate()
 *
 * Description  : Authenticate an NDMP client.
 *
 * - find user in file
 * - get corresponding password
 * - if no encryption
 *   - compare passwords
 * - if MD5 encryption,
 *   - calculate digest with password and challenge,
 *   - compare digests.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */


boolean                                 // TRUE if authentication is successful
ndmp_authenticate(int       type,       // NDMP_AUTH_TEXT, NDMP_AUTH_MD5
                  char *    id,         // identifier
                  byte *    secret,     // password or digest (if MD5)
                  byte *    challenge)  // challenge string (if MD5)

{
    boolean authentic = FALSE;
    char    password[CRYPT_MAX_PASSWD_LEN];
    
    NDMP_FCT_PRINT(("> ndmp_authenticate()\n"));

    /* Find user in file and get corresponding password.
     */
    if (ndmp_getPassword(id, password) == NFX_OK)
    {
        NDMP_DBG_PRINT(("Password is %s\n", password));

        switch (type)
        {
        case NDMP_AUTH_TEXT:
        {
            /* No encryption, compare passwords.
             */
            if (!strcmp(secret, password))
            {
                authentic = TRUE;
            }
            break;
        }
        case NDMP_AUTH_MD5:
        {
            /* MD5 encryption,
             * calculate digest with password and challenge,
             * compare digests.
             *
             * Reference: draft-skardal-ndmpv4-02.txt
             */
            int         rc;
            char        msg[128];
            crypt_md_t  md;
            
            bzero((void *) msg, sizeof(msg));
            memcpy((void *) &msg[0], password, strlen(password));
            memcpy((void *) &msg[128 - strlen(password)],
                   password, strlen(password));
            memcpy((void *) &msg[128 - strlen(password) - 64],
                   challenge, 64);
            
            rc = crypt_digest(MD5, msg, sizeof(msg), &md);
            if (rc == NFX_OK)
            {
                int j;
                
                authentic = TRUE;

                for (j = 0; j < md.len; j++)
                {
                    NDMP_DBG_PRINT(("%d: %02x == %02x\n",
                                       j, secret[j], md.value[j]));
                    if (secret[j] != md.value[j])
                    {
                        authentic = FALSE;
                        break;
                    }
                } //for
            }
            break;
        }
        default:
        {
            NDMP_DBG_PRINT(("Wrong type %d\n", type));
        }
        } // switch
    }
    
    NDMP_FCT_PRINT(("< ndmp_authenticate()\n"));
    return(authentic);
}

/*----------------------------------------------------------------------------*/
