/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-nim-util.c
 *
 * RCS:  $Id: ndmp-nim-util.c,v 1.35 2002/11/28 01:46:24 anhv Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:		Contain utility funcs used by NDMP server
 *					on NNIM.
 *
 * Created by: Anh Vu
 *
 * Date Created:        9/06/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>
#include <exports/scsi-api.h>
#include <exports/sdm-agent-api.h>
#include "ndmp.h"
#include "ndmp-debug.h"
#include <exports/scsiio-api.h>
#include "ndmp-mover-sm.h"
#include "ndmp-data-sm.h"
#include "ndmp-data-v2-sm.h"

extern ndmp_NNIMCtrl_t ndmp_NNIMCtrl;
extern ndmpNNIMMsgEnt_t ndmpNNIMMsgTbl[];
extern int ndmp_fp_msgStats[];

/*-----------------------------------------------------------------
 * Name:  ndmpGetNNIMMsgEnt
 *
 * Description: Find message entry in our NNIM Message Entry table.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        11/7/01
 *
 *-----------------------------------------------------------------
 */
ndmpNNIMMsgEnt_t *
ndmpGetNNIMMsgEnt(uint32 msgid)
{
    ndmpNNIMMsgEnt_t *entry;
 
    for (entry = ndmpNNIMMsgTbl; entry->msgid != 0; entry++)
    {
        if (entry->msgid == msgid)
            return(entry);
    }
    return(0);
} /* ndmpNNIMGetMsgEnt */       
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_sendCmdRsp
 *
 * Description:     Sends an NFX NDMP Internal response back to the NDMP server
 *					on SSC.  Used when no additional data need to be sent except 
 *					for an error code. 
 *                  
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *
 *------------------------------------------------------------------------------
 */
int
ndmp_sendCmdRsp( ndmp_NFXReq_t *req, int32 dst_port, uint32 error, int32 userInfo )
{
    ndmp_NFXRsp_t rsp;
	ndmpNNIMMsgEnt_t    *msgent;
	int rc;

    /*
     * Create reject message and send it.
     */
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendCmdRsp: Begin\n"));

	if (req == NULL)
	{
    	NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendCmdRsp: Req == NULL, return.\n"));
    	E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_sendCmdRsp: Req == NULL, return.\n"));
    	return (NFX_OK);
	}

    memset(&rsp, 0, sizeof(rsp));

    rsp.userInfo     = userInfo;
    rsp.content[0]     = (char) error;
    rsp.hdr.clientHdl = req->hdr.clientHdl;
    rsp.hdr.sessId  = req->hdr.sessId;
    rsp.hdr.mid     = req->hdr.mid;
    rsp.ndmp_mid     = req->ndmp_mid;
	memcpy(&rsp.cl_hdr, &req->cl_hdr, sizeof(ndmp_Hdr));

    rc = eee_sendMessage((uchar8 *)&rsp, sizeof(rsp), dst_port, ndmp_NNIMCtrl.src_port);

	if (rc != NFX_OK)
	{
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendCmdRsp: eee_sendMessage failed for "
						"msgId %d! sessId %d\n", req->hdr.mid, req->hdr.sessId));
        E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_sendCmdRsp: eee_sendMessage failed for "
						 "msgId %d! sessId %d\n", req->hdr.mid, req->hdr.sessId));
		return NFX_ERR;
    } 

	if ((msgent = ndmpGetNNIMMsgEnt(req->hdr.mid)) == 0)
	{
		NDMP_DBG_PRINTF (D_INFO, ("ndmp_NNIMReceiveMsg:   unrecognized message sent %d, "
						"sessId %d.\n", req->hdr.mid, req->hdr.sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_NNIMReceiveMsg:   unrecognized message sent %d, "
						 "sessId %d.\n", req->hdr.mid, req->hdr.sessId));
    } else {
#ifdef NFX_SMP 
		ATOMIC_INC_INT(&ndmp_fp_msgStats[msgent->stat_idx_snt], 1);
#else
		ndmp_fp_msgStats[msgent->stat_idx_snt]++;
#endif
	}
    return (NFX_OK);
} /* ndmp_sendCmdRsp */
/*-----------------------------------------------------------------
 * Name: ndmp_save_env
 *
 * Description:     
 *
 * Created by:  Anh Vu
 *
 * Date Created: 11/12/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_save_env(char8 *env_data, int32 num_env, Data_t *data)
{
	int32	i, value_len, name_len;

	if (data->env_lst)
    {
        // Deallocate old list if exists, then create new one /w new info
        NDMP_FP_FREE(data->env_lst, "env_lst");
        data->envLen = 0;
	}
    NDMP_FP_ALLOC(data->env_lst, ndmp_pval, (num_env * sizeof(ndmp_pval)), "env_lst");
    if (data->env_lst == NULL)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_save_env:  Error w/ mallocing env_lst\n"));
        E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_save_env:  Error w/ mallocing env_lst\n"));
        return NFX_ERR;
    }
    /* remember to dealloc when session is removed
	 * env_data has data stored as follows, repeats of the following 4:
	 * 	- len of 'name' env. var (4 bytes)
	 *  - the name string itself (length of string = len above)
	 *  - len of 'value'
	 *  - the value string itself (length of string = len above)
	 */
	data->envLen = num_env;	
	for (i = 0; i < data->envLen; i++)
	{
		/* Save environment variable NAME */
		name_len = *((int32 *) env_data);
        NDMP_FP_ALLOC(data->env_lst[i].name, char, (name_len + 1), "env_name");
    	if (data->env_lst[i].name == NULL)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_save_env:  Error w/ mallocing name.\n"));
        	E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
							("ndmp_save_env:  Error w/ mallocing name.\n"));
			ndmp_free_env(data);
        	return NFX_ERR;
    	}
		env_data += 4;		// skip name_len
		memcpy(data->env_lst[i].name, env_data, name_len);
		data->env_lst[i].name[name_len] = '\0';
		NDMP_INT_ALIGN(name_len);
		env_data += name_len;	// skip name string, proceed to value_len

		/* Save environment variable VALUE */
		value_len = *((int32 *) env_data);
        NDMP_FP_ALLOC(data->env_lst[i].value, char, (value_len + 1), "env_val");
    	if (data->env_lst[i].value == NULL)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_save_env:  Error w/ mallocing value.\n"));
        	E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
							("ndmp_save_env:  Error w/ mallocing value.\n"));
			ndmp_free_env(data);
        	return NFX_ERR;
    	}
		env_data += 4;		// skip value_len
		memcpy(data->env_lst[i].value, env_data, value_len);
		data->env_lst[i].value[value_len] = '\0';
		NDMP_INT_ALIGN(value_len);
		env_data += value_len;	// skip value string, proceed to next envir pair
		
	}
	return NFX_OK;	
}
/*-----------------------------------------------------------------
 * Name: ndmp_free_env
 *
 * Description:	Free the list of environment variables we keep 
 *	in Data Server.
 *
 * Created by:  Anh Vu
 *
 * Date Created: 11/12/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_free_env(Data_t *data)
{
    int32   i;

    if (data->env_lst)
    {
		for (i = 0; i < data->envLen; i++)
    	{	
            NDMP_FP_FREE(data->env_lst[i].name, "env_name");
            NDMP_FP_FREE(data->env_lst[i].value, "env_value");
		}
        NDMP_FP_FREE(data->env_lst, "env_lst");
	}
	data->env_lst = NULL;
    data->envLen = 0;
	return NFX_OK;
} /* ndmp_free_env */
        
/*-----------------------------------------------------------------
 * Name: ndmp_save_flist
 *
 * Description:   Retain the restored file list in our Data_t
 *	structure.
 *
 * Created by:  Anh Vu
 *
 * Date Created: 11/12/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_save_flist(char8 *flist_data, int32 num_fil, Data_t *data)
{
	int32	i, new_FileLen, old_FileLen;

    /* If there are any old file list, free them */
    ndmp_free_flist(data);

    NDMP_FP_ALLOC(data->fil_lst, ndmp_name, (num_fil * sizeof(ndmp_name)), "fil_lst");
    if (data->fil_lst == NULL)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_save_flist:  Error w/ mallocing\n"));
        E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_save_flist:  Error w/ mallocing\n"));
        return NFX_ERR;
    }
    // remember to dealloc when session is removed
	data->flist_len = num_fil;	
	for (i = 0; i < data->flist_len; i++)
	{
		/* Save orig_path name  */
		old_FileLen = *((int32 *) flist_data);
		//NDMP_DBG_PRINTF(D_DBG, ("ndmp_save_flist:  old_FileLen = %d\n", old_FileLen));

        NDMP_FP_ALLOC(data->fil_lst[i].orig_path, char, (old_FileLen + 1), "orig_path");
		flist_data += 4;		// skip old file name_len
		memcpy(data->fil_lst[i].orig_path, flist_data, old_FileLen);
		data->fil_lst[i].orig_path[old_FileLen] = '\0';		//null terminated it
		NDMP_INT_ALIGN(old_FileLen);
		//NDMP_DBG_PRINTF(D_DBG, ("ndmp_save_flist:  Padded old_FileLen = %d\n", old_FileLen));
		flist_data += old_FileLen;	// skip name string, proceed to value_len

		/* Save dest_path name  */
		new_FileLen = *((int32 *) flist_data);
		//NDMP_DBG_PRINTF(D_DBG, ("ndmp_save_flist:  new_FileLen = %d\n", new_FileLen));

        NDMP_FP_ALLOC(data->fil_lst[i].dest_dir, char, (new_FileLen + 1), "dest_dir");
		flist_data += 4;		// skip new file name length
		memcpy(data->fil_lst[i].dest_dir, flist_data, new_FileLen);
		data->fil_lst[i].dest_dir[new_FileLen] = '\0';
		NDMP_INT_ALIGN(new_FileLen);
		//NDMP_DBG_PRINTF(D_DBG, ("ndmp_save_flist:  Padded new_FileLen = %d\n", new_FileLen));
		flist_data += new_FileLen;	// skip value string, proceed to next envir pair
		
		// We don't do Direct Access so we don't care about new_name, 
		// other_name, node & fh_info 
	}
	return NFX_OK;	
}
/*-----------------------------------------------------------------
 * Name: ndmp_free_flist
 *
 * Description:	Free the list of restored files we keep
 *  in Data Server
 *
 * Created by:  Anh Vu
 *
 * Date Created: 11/12/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_free_flist(Data_t *data)
{
    int32   i;

    if (data->fil_lst)
    {
        for (i = 0; i < data->flist_len; i++)
        {
            NDMP_FP_FREE(data->fil_lst[i].orig_path, "fl.orig_path");
            NDMP_FP_FREE(data->fil_lst[i].dest_dir, "fl.dest_dir");
        }
        NDMP_FP_FREE(data->fil_lst, "fil_lst");
    }
    data->fil_lst = NULL;
    data->flist_len = 0;
    return NFX_OK;

} /* ndmp_free_flist */

#ifdef NFX_SMP
/*
 *------------------------------------------------------------------------------
 * Function Name:   print_sessid_tbl
 *------------------------------------------------------------------------------
 */
void
print_sessid_tbl()
{
	int i;
	sess_node_t *np;

	printf("idx   SessId(d)    SessId(x)   Servp\n"
			"---   ---------    ---------   -----\n");
	for (i = 0; i<NDMP_MAX_SESS; i++)
	{
		np = &ndmp_NNIMCtrl.sessLst[i];
		if (np->sessid != -1)
		{
			printf("%3d   %10d   0x%x   %p\n", 
					i, np->sessid, np->sessid, np->servp);
		}
	}
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_add_sessid_entry
 *
 * Description:     Add sessId into our sessid_table, at the idx slot.
 *
 * Created by:      Anh Vu 
 *
 * Data Created:    5/16/02
 *
 * Returns:         NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
sess_node_t *
ndmp_add_sessid_entry(int32 idx, int32 sessId)
{

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_add_sessid_entry, idx %d\n", idx));

	NFX_SPIN_LOCK(&ndmp_NNIMCtrl.spin);
	ndmp_NNIMCtrl.sessLst[idx].sessid = sessId;
	NFX_SPIN_UNLOCK(&ndmp_NNIMCtrl.spin);

	if (ndmp_dbg_level)
		print_sessid_tbl();

	return (&ndmp_NNIMCtrl.sessLst[idx]);
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_del_sessid_entry
 *
 * Description:     Remove sessid entry from sessid_table.
 *
 * Created by:      Anh Vu 
 *
 * Data Created:    5/16/02
 *
 * Returns:         NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int
cleanup_msg_q(nfx_spinlock_t *msgq_lock, msg_q_t *rcv_q)
{
	uint64 seq_id;
	msgCntxt_t *msgI;
	int32		rc;
	q_headNodePtr_t *msg_q = &rcv_q->msg_q;


	//printf("cleanup_msg_q - begin, msg_q ptr %p\n", rcv_q->msg_q);
	NFX_SPIN_LOCK(msgq_lock);
	if (*msg_q == NULL)
	{
		//printf("Queue's empty\n");
	} else {
		/* Clean out the SSC/FP msg rcvd Q */
		if (q_resetReadPointer(*msg_q) != NFX_OK)
		{
			NFX_SPIN_UNLOCK(msgq_lock);
			return NFX_ERR;
		}
		while ((rc = q_deleteQueueItem(*msg_q, &seq_id, &msgI)) != Q_ERR_EMPTY_QUEUE)
		{
			//printf("freeing msgI %p, seqId %lld \n", msgI, seq_id);
			NDMP_FP_FREE(msgI, "msgI del_sessid_entry");
		}
		q_destroyQueue(*msg_q);
		*msg_q = NULL;	
	}
	rcv_q->nxt_seq_expected = 0;
	rcv_q->msgq_cnt = 0;
	NFX_SPIN_UNLOCK(msgq_lock);

	return NFX_OK;
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_del_sessid_entry
 *
 * Description:     Remove sessid entry from sessid_table.
 *
 * Created by:      Anh Vu 
 *
 * Data Created:    5/16/02
 *
 * Returns:         NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int
ndmp_del_sessid_entry(int32 idx, int32 sessId)
{
	sess_node_t *np = &ndmp_NNIMCtrl.sessLst[idx];

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_del_sessid_entry:  sessId %d\n", sessId));
	if (np->sessid == -1)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_del_sessid_entry:  Trying to del sessid entry %d that does not exist, idx %d\n", sessId, idx));
		E_LOG(class_1, warning_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_del_sessid_entry:  Trying to del sessid entry %d that does not exist, idx %d\n", 
						 sessId, idx));
		return NFX_OK;
	}
	if (np->sessid == sessId)
	{
		np->sessid = -1;	
		np->servp = NULL;	

	} else {
		NDMP_DBG_PRINTF(D_INFO, ("Warning:  No such sessId %d "
			"at indicated index %d.\n", sessId, idx));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("Warning:  No such sessId %d at indicated index %d.\n", sessId, idx));
	}	

	if (ndmp_dbg_level)
		print_sessid_tbl();
	return NFX_OK;
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_find_sessid_entry
 *
 * Description:     Find sessid entry from sessid_table.
 *
 * Created by:      Anh Vu 
 *
 * Data Created:    5/16/02
 *
 * Returns:         -1 or idx
 *
 *------------------------------------------------------------------------------
 */
sess_node_t *
ndmp_find_sessid_entry(int32 sessId, int *idx)
{
	int i;
	//sess_node_t *np;

	for (i=0; i < NDMP_MAX_SESS; i++)
	{
//#ifdef DBG_SESS_ID
		//printf("ndmp_find_sessid_entry:  " "idx %d, sessid %d\n", i, ndmp_NNIMCtrl.sessLst[i].sessid); 
//#endif
		if (ndmp_NNIMCtrl.sessLst[i].sessid == sessId)
		{
//#ifdef DBG_SESS_ID
			//printf("ndmp_find_sessid_entry: Found it, idx = %d, sessId = %d, servp %p.\n", 
					//i, sessId, ndmp_NNIMCtrl.sessLst[i].servp);
//#endif
			*idx = i;
			return (&ndmp_NNIMCtrl.sessLst[i]);
		}	
	}
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_find_sessid_entry:  sessId %d not found\n", sessId)); 
	*idx = -1;
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name: ndmp_seqId_cmp
 *
 * Description:	Add request to pending queue for IO rsp.
 *
 * Created by:  Anh Vu 
 *
 * Date Created: 01/03/03
 *
 *-----------------------------------------------------------------
 */
int32 ndmp_seqId_cmp(const void *xptr, const void *yptr)
{
	uint64 x, y;

	x = *(uint64 *)xptr;
	y = *(uint64 *)yptr;

	return x - y;
}
/*-----------------------------------------------------------------
 * Name: ndmp_send_evnt
 *
 * Description:	
 *
 * Created by:  Anh Vu 
 *
 * Date Created: 01/03/03
 *
 *-----------------------------------------------------------------
 */
int32 ndmp_send_evnt(sess_node_t *node)
{
	esm_event_t event;
	int32	rc;

	event.event = NDMP_THREAD_EVENT_GOT_MSG;
	event.p1.u.p = node;  //to verify i'm the sender
	//printf("ndmp_send_evnt:  EVENT_GOT_MSG, sessid %d, ndmp_NNIMCtrl.msgThreadHdl %d\n", 
	//					node->sessid, ndmp_NNIMCtrl.msgThreadHdl);
	rc = thread_event(ndmp_NNIMCtrl.msgThreadHdl, &event);
	//printf("ndmp_send_evnt:  rc = %d\n", rc);
	return (rc);
}

/*-----------------------------------------------------------------
 * Name: ndmp_InsertMsgInOrder
 *
 * Description:	Add request to pending queue for IO rsp, whether
 * 				sessid .
 *
 * Created by:  Anh Vu 
 *
 * Date Created: 01/03/03
 *
 *-----------------------------------------------------------------
 */
int32 ndmp_InsertMsgInOrder(uint64 *seq_id, 
						msgCntxt_t **msgI,  
						sess_node_t *node, 
						int32 cardtyp)
{
	
	int32 rc;
	q_headNodePtr_t *msg_q;

	msg_q = (cardtyp == SSC_CARD)? &(node->ssc_rcv_q.msg_q): &(node->fp_rcv_q.msg_q);

	//printf("ndmp_InsertMsgInOrder: sessid %d, seq_id %lld, msgI %p, cardtyp %s\n", 
	//				(*msgI)->sessid, *seq_id, *msgI, (cardtyp == SSC_CARD)? "SSC":"FP");
	if (*msg_q == NULL)
	{
		*msg_q  = q_createQueue(sizeof(long), sizeof(long),
						ndmp_seqId_cmp, ndmp_seqId_cmp);
		//printf("%s msg queue ptr %p\n", (cardtyp == SSC_CARD)? "SSC":"FP", *msg_q);

	} 
	/* For SSC we need to insert these in order w/ the
	* sequence id in the msg.
	*/
	if (cardtyp == SSC_CARD)
	{
		rc = q_insertItemInOrder(*msg_q, (void *)seq_id, (void *)msgI,  
					Q_FIND_ONE | Q_INSERT_ORDER_ASCENDING);
	} else {
		/* For FP, we just q them as them arrive since they
		* are sent by FS which is on the same FP board
		* which means the eee_send is executed in FS context,
		* and msgs will arrive in order.  Get it?
		*/
		rc = q_enQueue(*msg_q, (void *)msgI, (void *)msgI);
	}
	if (rc == NFX_OK)
	{
		if (ndmp_send_evnt(node) == NFX_OK) 
		{
			return((*msg_q)->size);
		} else {
			printf("ndmp send GOT_MSG event failed for sessid %d\n", (*msgI)->sessid);
			NFX_ASSERT(1);
		}
	}
	return 0;
}
#else
/*
 *------------------------------------------------------------------------------
 * Function Name:   print_sessid_tbl
 *------------------------------------------------------------------------------
 */
void
print_sessid_tbl()
{
	int i;
	sess_node_t *np;

	printf("idx   SessId(d)    SessId(x)   Servp\n"
			"---   ---------    ---------   -----\n");
	for (i = 0; i<NDMP_MAX_SESS; i++)
	{
		for (np = ndmp_NNIMCtrl.sessLst[i]; np != NULL; np = np->nxt)
		{
			printf("%3d   %10d   0x%x   %p\n", 
			i, np->sessid, np->sessid, np->servp);
		}
	}
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_add_sessid_entry
 *
 * Description:     Add sessId into our sessid_table, at the idx slot.
 *
 * Created by:      Anh Vu 
 *
 * Data Created:    5/16/02
 *
 * Returns:         NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
sess_node_t *
ndmp_add_sessid_entry(int32 idx, int32 sessId)
{

	sess_node_t *node;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_add_sessid_entry, idx %d, ndmp_NNIMCtrl.sessLst[idx] = %p\n",
		idx, ndmp_NNIMCtrl.sessLst[idx]));

	NDMP_FP_ALLOC(node, sess_node_t, sizeof(sess_node_t), "node_t");
	if (!node)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_add_sessid_entry : Error eee_ramAlloc sess_node.\n"));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_add_sessid_entry : Error eee_ramAlloc sess_node.\n"));
		return NULL;
	}
	node->sessid = sessId;
	node->nxt = ndmp_NNIMCtrl.sessLst[idx];
	ndmp_NNIMCtrl.sessLst[idx] = node;

	if (ndmp_dbg_level)
		print_sessid_tbl();

	return node;
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_del_sessid_entry
 *
 * Description:     Remove sessid entry from sessid_table.
 *
 * Created by:      Anh Vu 
 *
 * Data Created:    5/16/02
 *
 * Returns:         NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int
ndmp_del_sessid_entry(int32 idx, int32 sessId)
{
	int32 found = FALSE;
	sess_node_t *curp, *prev;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_del_sessid_entry:  sessId %d\n", sessId));
	if (ndmp_NNIMCtrl.sessLst[idx] == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_del_sessid_entry:  Trying to del sessid entry %d that does not exist, idx %d\n", sessId, idx));
		E_LOG(class_1, warning_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_del_sessid_entry:  Trying to del sessid entry %d that does not exist, idx %d\n", 
						 sessId, idx));
		return NFX_OK;
	}
	for (prev = NULL, curp = ndmp_NNIMCtrl.sessLst[idx]; curp != NULL; curp = curp->nxt)
	{
#ifdef DBG_SESS_ID
		printf("idx = %d, prev %p, curp %p, prev->sessid %d, curp->sessid %d\n", 
					idx, prev, curp, prev? prev->sessid:0, curp? curp->sessid:0);
#endif				
		if (curp->sessid == sessId)
		{		
			if (prev==NULL)
			{
				ndmp_NNIMCtrl.sessLst[idx] = curp->nxt;
			} else {
				prev->nxt = curp->nxt;
			}
			NDMP_FP_FREE (curp, "node_t");
			found = TRUE;
			break;
		} else {
			prev = curp;
		}
	}
	if (!found)
	{	
		NDMP_DBG_PRINTF(D_INFO, ("Warning:  No such sessId %d "
			"at indicated index %d.\n", sessId, idx));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("Warning:  No such sessId %d at indicated index %d.\n", sessId, idx));
	}	

	if (ndmp_dbg_level)
		print_sessid_tbl();
	return NFX_OK;
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_find_sessid_entry
 *
 * Description:     Find sessid entry from sessid_table.
 *
 * Created by:      Anh Vu 
 *
 * Data Created:    5/16/02
 *
 * Returns:         -1 or idx
 *
 *------------------------------------------------------------------------------
 */
sess_node_t *
ndmp_find_sessid_entry(int32 sessId, int *idx)
{
	int i;
	sess_node_t *np;

	for (i=0; i < NDMP_MAX_SESS; i++)
	{
		for (np = ndmp_NNIMCtrl.sessLst[i]; np != NULL; np=np->nxt)
		{
#ifdef DBG_SESS_ID
			printf("ndmp_find_sessid_entry:  " "np->sessid %d, idx %d\n", np->sessid, i); 
#endif
			if (np->sessid == sessId)
			{
#ifdef DBG_SESS_ID
				printf("ndmp_find_sessid_entry: Found it, idx = %d, node = %p, sessId = %d, servp %p.\n", 
					i, np, np->sessid, np->servp);
#endif
				*idx = i;
				return np;
			}
		}	
	}
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_find_sessid_entry:  sessId %d not found\n", sessId)); 
	*idx = -1;
	return (NULL);
}
#endif
/*-----------------------------------------------------------------
 * Name: ndmp_addReqToQueue
 *
 * Description:	Add request to pending queue for IO rsp.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 09/16/02
 *
 *-----------------------------------------------------------------
 */
void ndmp_addReqToQueue(Mover_t *mover, uint64 ioCookie, char8 *buf)
{
    ndmp_ioReqNode_t *new;

    //printf("ndmp_addReqToQueue:  Add ioCookie %lld to q\n", ioCookie);
    NDMP_FP_ALLOC(new, ndmp_ioReqNode_t, sizeof(ndmp_ioReqNode_t), "ioReqNode");
    new->ioCookie = ioCookie;
    new->buf = buf;
    new->next = NULL;
    if (mover->ioPendQ.head == NULL)
        mover->ioPendQ.head = new;
    else
        mover->ioPendQ.last->next = new;
    mover->ioPendQ.last = new;
    mover->ioPendQ.ioReqCnt++;

    return;
} /* ndmp_addReqToQueue */

/*-----------------------------------------------------------------
 * Name: ndmp_getNextFromQ
 *
 * Description:	Get next request from pending queue for IO rsp.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 09/16/02
 *
 *-----------------------------------------------------------------
 */
ndmp_ioReqNode_t *ndmp_getNextFromQ(Mover_t *mover)
{
    if (mover->ioPendQ.curr == mover->ioPendQ.last)
        return NULL;
    else if (mover->ioPendQ.curr == NULL)
        mover->ioPendQ.curr = mover->ioPendQ.head;
    else
        mover->ioPendQ.curr = mover->ioPendQ.curr->next;
    return (mover->ioPendQ.curr);
} /* ndmp_getNextFromQ */

/*-----------------------------------------------------------------
 * Name: ndmp_delReqFromQueue
 *
 * Description:	Delete first request from pending queue for IO rsp.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 09/16/02
 *
 *-----------------------------------------------------------------
 */
int ndmp_delReqFromQueue(Mover_t *mover)
{
    ndmp_ioReqNode_t *tmp;

    if (mover->ioPendQ.head == NULL)
        return NFX_ERR;

    tmp = mover->ioPendQ.head;

    if (mover->ioPendQ.head == mover->ioPendQ.last)
    {
        mover->ioPendQ.head = NULL;
        mover->ioPendQ.last = NULL;
        mover->ioPendQ.curr = NULL;
    }
    else 
	{
        mover->ioPendQ.head = mover->ioPendQ.head->next;
	}	
    NDMP_FP_FREE(tmp, "ioReqNode");
    mover->ioPendQ.ioReqCnt--;

    return NFX_OK;
} /* ndmp_delReqFromQueue */

/*-----------------------------------------------------------------
 * Name: 		ndmp_scsiio_write_done
 *
 * Description:	To handle fragmentation to 8k bufs for xmission.
 *
 * Created by:  Radek Aster
 *
 * Date Created: 
 *
 *-----------------------------------------------------------------
 */
void 
ndmp_scsiio_write_done(int scsi_status, void *buf, uint32 datalen, void *cookie, void *ioCookie)
{
	int32               sessId = (address_t)cookie & 0xFFFFFFFF;
	scsi_writeRspMsg_t  *write_rsp;
    NdmpNIMService_t    *servp;
	esm_handle_t		hdl;
    esm_event_t			esmEvent;
	int32				rc, idx;
	sess_node_t			*node = NULL;

    NDMP_FP_ALLOC(write_rsp, scsi_writeRspMsg_t, sizeof(scsi_writeRspMsg_t), "scsiio_write_rsp");
	NFX_ASSERT(write_rsp);

	write_rsp->eCode = scsi_status;

	node = ndmp_find_sessid_entry(sessId, &idx);

	if ( (idx == -1) || (!node) || (node && (node->servp == NULL)) )
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_scsiio_write_done:  Invalid session %d, ignore.\n", sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_scsiio_write_done:  Invalid session %d, ignore.\n", sessId));
        NDMP_FP_FREE(write_rsp, "write response");
		return;
	}

	servp = node->servp;
	hdl = servp->Mover.my_smHdl;
    bzero(&esmEvent, sizeof(esm_event_t));
    esmEvent.event = NDMP_MOVER_EVENT_WRITE_DONE;
    esmEvent.p1.u.p = (void *)write_rsp;
    esmEvent.p2.u.dw = (int64) ioCookie;
    rc = esm_deliverEvent(hdl, &esmEvent);
    if (rc != NFX_OK)
    {
         NDMP_DBG_PRINTF(D_INFO, ("ndmp_scsiio_write_done:  SessId %d -- ERROR, cannot post event %d to SM hdl %d",
								  sessId, esmEvent.event, hdl));
         E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						 ("ndmp_scsiio_write_done:  SessId %d -- ERROR, cannot post event %d to SM hdl %d",
						  sessId, esmEvent.event, hdl));
        NDMP_FP_FREE(write_rsp, "write response");
    }

    return;
}

/*-----------------------------------------------------------------
 * Name: 		ndmp_scsiio_read_done
 *
 * Description:	To handle fragmentation to 8k bufs on reception
 *
 * Created by:  Radek Aster
 *
 * Date Created: 
 *
 *-----------------------------------------------------------------
 */
void 
ndmp_scsiio_read_done(int scsi_status, void *buf, uint32 datalen, void *cookie, void *ioCookie)
{
	int32               sessId = (address_t)cookie & 0xFFFFFFFF;
    scsi_readRspMsg_t   *read_rsp;      //EEE_BUF_RST
	ndmp_NFXScsiRdRsp   *ndmp_read_rsp = NULL;
    NdmpNIMService_t    *servp;
	esm_handle_t		hdl;
    esm_event_t			esmEvent;
	int32				rc, idx;
	sess_node_t			*node = NULL;
    eee_desc_t          *eDesc = (eee_desc_t *) buf;
    eee_descBuf_t       *bd;

    //read_rsp = (scsi_readRspMsg_t *)(buf - sizeof(scsi_readRspMsg_t));

    /* There will always be an eee desc no matter what the scsi_status is.
     * just the datalen will be 0 in an error case.
     */
    bd = &eDesc->bd[1];
    read_rsp = (scsi_readRspMsg_t *)(bd->buf - sizeof(scsi_readRspMsg_t)); 
              /* get to the scsi hdr that is before the data portion pointed to by bd->buf */

    NDMP_DBG_PRINTF(D_INFO, ("ebuf          : %p, ioCookie %llu\n" 
    	    "Message ID   : %d\n" "Txn sequence : %d\n" "Data Length  : %d\n"
    	    "Cod Handle   : %llx\n" "read_rsp     : %p\n",
		    buf, (uint64) ioCookie,	read_rsp->hdr.mid, read_rsp->transSeqId, 
		    read_rsp->dataLen, read_rsp->codHandle, read_rsp));

    NDMP_FP_ALLOC(ndmp_read_rsp, ndmp_NFXScsiRdRsp, sizeof(ndmp_NFXScsiRdRsp), "ndmp_NFXScsiRdRsp");
    if (ndmp_read_rsp == NULL)
    {
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_scsiio_read_done:  Cannot malloc ndmp_read_rsp"
            ", sessid %d, return.\n", sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_scsiio_read_done:  Cannot malloc ndmp_read_rsp, session %d, return.\n", sessId));
        free_eee_buf_chain(buf);
        goto scsiio_rddone_snd_event; //try to reply best we could
    }
	ndmp_read_rsp->eCode = scsi_status;
	ndmp_read_rsp->dataLen = datalen;
//EEE_BUF_RST
	ndmp_read_rsp->eBufptr = (eee_desc_t *)buf;



scsiio_rddone_snd_event:

	node = ndmp_find_sessid_entry(sessId, &idx);

	if ( (idx == -1) || (!node) || (node && (node->servp == NULL)) )
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_scsiio_read_done:  Invalid session %d, ignore.\n", sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_scsiio_read_done:  Invalid session %d, ignore.\n", sessId));
        NDMP_FP_FREE(ndmp_read_rsp, "ndmp_read_rsp ndmp_scsiio_read_done");
        free_eee_buf_chain(buf);
		return;
	}

	servp = node->servp;
	hdl = servp->Mover.my_smHdl;
    bzero(&esmEvent, sizeof(esm_event_t));
    esmEvent.event = NDMP_MOVER_EVENT_READ_DONE;
    esmEvent.p1.u.p = (void *)ndmp_read_rsp;
    esmEvent.p2.u.dw = (int64) ioCookie;
    rc = esm_deliverEvent(hdl, &esmEvent);
    if (rc != NFX_OK)
    {
         NDMP_DBG_PRINTF(D_INFO, ("ndmp_scsiio_read_done:  SessId %d -- ERROR, cannot post event %d to SM hdl %d",
								  sessId, esmEvent.event, hdl));
         E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						 ("ndmp_scsiio_read_done:  SessId %d -- ERROR, cannot post event %d to SM hdl %d",
						  sessId, esmEvent.event, hdl));
        NDMP_FP_FREE(ndmp_read_rsp, "ndmp_read_rsp ndmp_scsiio_read_done");
        free_eee_buf_chain(buf);
    }

    return;
}

/*-----------------------------------------------------------------
 * Name: 		ndmp_sendSCSIRdWrCmd
 *
 * Description:	Rewrite to send/receive data from SCSIIO
 *
 * Created by:  Radek Aster
 *
 * Date Created: 
 *
 *-----------------------------------------------------------------
 */
int
ndmp_sendSCSIRdWrCmd(int type, uint64 ioCookie, NdmpNIMService_t *servp, char8 *buf)
{
    int                         rc = NFX_OK;
	Mover_t						*mover = &servp->Mover;
	char						*read_buf = NULL;
	eeeRTC_t					proctime;

	switch (type)
	{
		case SCSI_MID_WRITE_REQ:
	
			if (buf == NULL)
			{
				NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendSCSIRdWrCmd:  NULL buf for Write.\n"));
				E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
								("ndmp_sendSCSIRdWrCmd:  NULL buf for Write.\n"));
				return NFX_ERR;
			}

			NDMP_DBG_PRINTF(D_INFO, ("*** ndmp_sendSCSIRdWrCmd:  WRITE req, tape hdl 0x%x," 
									"ioCookie %lld, buf 0x%p.\n", 
					mover->tape_hdl, ioCookie, buf));

			/* Inititate SCSI write operation */
			rc = scsiio_write(mover->tape_hdl, buf, mover->recSiz, ndmp_scsiio_write_done,
							  (void *)(uint64)servp->sessId, (void *)ioCookie, 0); //don't free memory for me, Rahul will free himself.
			if (rc != NFX_OK) 
			{
				NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendSCSIRdWrCmd:  scsiio_write failed.\n"));
				E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
								("ndmp_sendSCSIRdWrCmd:  scsiio_write failed.\n"));
				return NFX_ERR;
			}
#ifdef NFX_SMP 
			ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_WR_DATA_REQ_SNT], 1);
#else
			ndmp_fp_msgStats[SCSI_WR_DATA_REQ_SNT]++;
#endif
			break; 

		case SCSI_MID_READ_REQ:
			eee_rtc(&proctime);
			NDMP_DBG_PRINTF(D_INFO, ("*** ndmp_sendSCSIRdWrCmd:  READ req, tape hdl 0x%x, ioCookie %lld, REQ time %lld.\n", 
					mover->tape_hdl, ioCookie, proctime.val));

			/* Inititate SCSI read operation */

			/* Allocate memory here w/o using NDMP_FP_ALLOC macro because
			 * we don't want to account for the stats for this malloc.  Reason
			 * being that FP will release it and we cannot account for that.
			 * So it may throw us off during debugging.  6/26/02
			 */
#if 0 //EEE_BUF_RST
			read_buf = eee_ramAllocZero(mover->recSiz + sizeof(scsi_readRspMsg_t));
			if (read_buf == NULL)
			{
				NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendSCSIRdWrCmd:  Read buffer malloc error\n"));
				E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
								("ndmp_sendSCSIRdWrCmd:  Read buffer malloc error\n"));
				return NFX_ERR;
			}
			read_buf += sizeof(scsi_readRspMsg_t);
#endif
			rc = scsiio_read(mover->tape_hdl, read_buf, mover->recSiz, 
							 ndmp_scsiio_read_done,	
							 (void *)(uint64)servp->sessId, (void *)ioCookie);
			if (rc != NFX_OK) 
			{
				NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendSCSIRdWrCmd:  scsiio_read failed.\n"));
				E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
								("ndmp_sendSCSIRdWrCmd:  scsiio_read failed.\n"));
				return NFX_ERR;
			}
#ifdef NFX_SMP 
			ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT], 1);
#else
			ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT]++;
#endif
			break; 

		default:
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendSCSIRdWrCmd:  Wrong send type request %d\n", type));
			E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
							("ndmp_sendSCSIRdWrCmd:  Wrong send type request %d\n", type));
			rc = NFX_ERR;
	}

	return rc;

} /* ndmp_sendSCSIRdWrCmd */ 
/*
 *------------------------------------------------------------------------------
 * Function Name:   decode_data_event
 * Description:		
 * Created by:      Anh Vu
 * Data Created:    10/10/02
 * Returns:
 *------------------------------------------------------------------------------
 */
char *
decode_data_event (int event)
{
	switch (event)
	{
		case NDMP_DATA_EVENT_LISTEN: 	return "listen";
		case NDMP_DATA_EVENT_CONNECT: 	return "connect";
		case NDMP_DATA_EVENT_CONNECTED: return "connected";
		case NDMP_DATA_EVENT_START_OPS: return "start_ops";
		case NDMP_DATA_EVENT_FS_OPS_RSP: return "fs_ops_rsp";
		case NDMP_DATA_EVENT_IO_REQ: 	return "io_req";
		case NDMP_DATA_EVENT_IO_DONE: 	return "io_done";
		case NDMP_DATA_EVENT_GET_ENV:  	return "get_env";
		case NDMP_DATA_EVENT_GET_STATE: return "get_state";
		case NDMP_DATA_EVENT_DATA_READY:return "data_ready";
		case NDMP_DATA_EVENT_SND_INFO:  return "snd_info";
		case NDMP_DATA_EVENT_ABORT:		return "abort";
		case NDMP_DATA_EVENT_STOP:		return "stop";
		case NDMP_DATA_EVENT_INTERNAL_ERR: return "start_ops"; 
		case NDMP_DATA_EVENT_FS_ABORT_RSP:  return "fs_abort"; 
		case NDMP_DATA_EVENT_CONN_ERR: 	return "conn_err";
		default:						return 	"unknown";
	}
}	

/*
 *------------------------------------------------------------------------------
 * Function Name:   decode_data_state
 * Description:		
 * Created by:      Anh Vu
 * Data Created:    10/10/02
 * Returns:
 *------------------------------------------------------------------------------
 */
char *
decode_data_state (int state)
{
	switch (state)
	{
		case NDMP_DATA_STATE_IDLE:		return "idle";
		case NDMP_DATA_STATE_LISTEN:	return "listen";
		case NDMP_DATA_STATE_CONNECTED:	return "connected";
		case NDMP_DATA_STATE_ACTIVE:	return "active";
		case NDMP_DATA_STATE_HALTED:	return "halted";
   		default:						return "unknown";
	}
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   decode_mover_state
 * Description:		
 * Created by:      Anh Vu
 * Data Created:    10/10/02
 * Returns:
 *------------------------------------------------------------------------------
 */
char *
decode_mover_state (int state)
{
	switch (state)
	{
		case NDMP_MOVER_STATE_IDLE:		return "idle";
		case NDMP_MOVER_STATE_LISTEN:	return "listen";
		case NDMP_MOVER_STATE_PAUSED:	return "paused";
		case NDMP_MOVER_STATE_ACTIVE:	return "active";
		case NDMP_MOVER_STATE_HALTED:	return "halted";
   		default:						return "unknown";
	}
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   decode_scsi_ecode
 * Description:		
 * Created by:      Anh Vu
 * Data Created:    10/10/02
 * Returns:
 *------------------------------------------------------------------------------
 */
char *
decode_scsi_ecode (int ecode)
{
	switch (ecode)
	{
		case SCSI_MSG_ECODE_OK:  return "Ok";
		case SCSI_MSG_ECODE_ILI:  return "ILI";
		case SCSI_MSG_ECODE_MEDIUM_ERROR:  return "Medium Error";
		case SCSI_MSG_ECODE_FM_ENCOUNTERED:  return "FM Encountered";
		case SCSI_MSG_ECODE_END_OF_MEDIUM:  return "EOM";
		case SCSI_MSG_ECODE_PATH_FAILED:  return "Device path failed";
		case SCSI_MSG_ECODE_NOSUCH_DEV: return "No such device";
		case SCSI_MSG_ECODE_DEV_IO_ERROR: return "Device I/O error";
		case SCSI_MSG_ECODE_DEV_CLOSED: return "Device Closed";
		case SCSI_MSG_ECODE_NO_RESOURCE: return "Device resource not found";
		case SCSI_MSG_ECODE_TIMEOUT: return "SCSI msg timeout";
		case SCSI_MSG_ECODE_INVALID_IOCTL_CODE: return "SCSI Invalid IOCTL";
		case SCSI_MSG_ECODE_INVALID_IOCTL_PARMS: return "SCSI Invalid IOCTL parameters";
		case SCSI_MSG_ECODE_IOCTL_FAILED: return "SCSI IOCTL Failed";
		case SCSI_MSG_ECODE_DATA_CHECKSUM_ERR: return "SCSI data checksum error";
		case SCSI_MSG_ECODE_INVALID_LEN: return "SCSI invalid len";
		case SCSI_MSG_ECODE_INVALID_TXN_SEQ: return "SCSI invalid transmit sequence";
		case SCSI_MSG_ECODE_INVALID_MSG: return "SCSI invalid sequence msg received";
   		default:	return "Unknown device error";
	}
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   free_eee_buf_chain
 * Description:     
 * Created by:      Anh Vu
 * Data Created:    11/19/02
 * Returns:
 *------------------------------------------------------------------------------
 */
void
free_eee_buf_chain(eee_desc_t *ebuf)
{
    if (ebuf)
        eee_freePacket(ebuf);
    return;
}

