/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-ssc-util.c
 *
 * RCS:  $Id: ndmp-ssc-util.c,v 1.39 2002/12/05 19:46:09 anhv Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:		Contain utilities funcs used by NDMP servers on SSC
 *
 * Created by: Anh Vu
 *
 * Date Created:        9/06/01
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/anpssc-api.h>
#include <exports/elog-api.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/tape-api.h>
#include <exports/opt-api.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"
#include <exports/event-api.h>

char    					tbuf[256];
extern ndmpMsgEnt_t 		ndmpMsgTbl[];
extern ndmpAgileMsgEnt_t 	ndmpAgileMsgTbl[];
extern ndmpParentUIMsgEnt_t ndmpParentUIMsgTbl[];
extern sessId_entry_t  		sessId_Tbl[NDMP_MAX_SESS];
extern int					nxt_sessId;

static struct timeval waitRegm = {10, 0};

// void ndmp_sendSnapRemoveIndication(const int64 volId, const char * const snapName);

static int32 ndmp_regGetOption(char *option, int32 optionLen, char *rsp_buf);
void ndmp_getCfgVal(char *opt_str, char *def_str, int *setVar, 
				int min, int max);

#define FILE_MODE 0644	/* RW for owner, R for group and others */

#define NDMP_MAP_FILE   1
#define NDMP_UNMAP_FILE 2

#define DUMPFILE_FMT "fsname: 0x%llx, ddate: %d, client-ip: 0x%x, dump_level: %d, current_bkup_ind: %d  snapshot: %s\n"

/*-----------------------------------------------------------------
 * Name:  ndmpGetMsgEnt
 *
 * Description: Find message entry in our Message Entry table.
 *				This table contains msgs sent by the NDMP client.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
ndmpMsgEnt_t *
ndmpGetMsgEnt(ndmpMsgEnt_t *entry, ndmp_Msg msgid)
{
    for ( ; entry->msgid != 0; entry++)
    {
        if (entry->msgid == msgid)
            return(entry);
    }
    return(0);
} /* ndmpGetMsgEnt */

/*-----------------------------------------------------------------
 * Name:  ndmpGetAgileMsgEnt
 *
 * Description: Find message entry in our Agile Message Entry table.
 *				This table contains msgs sent by UI process to
 *				the children NDMP or the
 *				NDMP on FP board.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
ndmpAgileMsgEnt_t *
ndmpGetAgileMsgEnt(int msgid)
{
    ndmpAgileMsgEnt_t *entry;

    for (entry = ndmpAgileMsgTbl; entry->msgid != 0; entry++)
    {
        if (entry->msgid == msgid)
            return(entry);
    }
    return(0);
} /* ndmpGetAgileMsgEnt */

/*-----------------------------------------------------------------
 * Name:  ndmpGetParentUIMsgEnt
 *
 * Description: Find message entry in our UI msg table
 *				received by parent NDMP server from ndmp UI
 *				process.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        2/24/02
 *
 *-----------------------------------------------------------------
 */
ndmpParentUIMsgEnt_t *
ndmpGetParentUIMsgEnt(int msgid)
{
    ndmpParentUIMsgEnt_t *entry;

    for (entry = ndmpParentUIMsgTbl; entry->msgid != 0; entry++)
    {
        if (entry->msgid == msgid)
            return(entry);
    }
    return(0);
} /* ndmpGetParentUIMsgEnt */

/*-----------------------------------------------------------------
 * Name:  ndmp_InitSessIdTbl
 *
 * Description:  Get an unused session Id.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/16/02
 *
 *-----------------------------------------------------------------
 */
void
ndmp_InitSessIdTbl()
{
    int32 i;
	struct timeval tt;

    for (i = 0; i < NDMP_MAX_SESS; i++)
    {
        sessId_Tbl[i].in_use = FALSE;
        sessId_Tbl[i].sessId = -1;
        sessId_Tbl[i].pid = -1;
    }
	gettimeofday(&tt, NULL);
	nxt_sessId = tt.tv_sec;
	
	return;
} /* ndmpInitSessIdTbl */

/*-----------------------------------------------------------------
 * Name:  ndmpGetFreeSessId
 *
 * Description:  Get an unused session Id.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/16/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmpGetFreeSessId(int *sessId)
{
    int32 i;

	NDMP_DBG_PRINTF(D_INFO, ("ndmpGetFreeSessId:\n"));

   	for (i = 0; i < NDMP_MAX_SESS; i++)
   	{
        if (sessId_Tbl[i].in_use == FALSE)
		{
			sessId_Tbl[i].in_use = TRUE;
			*sessId = nxt_sessId;
			NDMP_DBG_PRINTF(D_INFO, ("ndmpGetFreeSessId:  Return Id %d, "
					"sessId = %d\n", i, nxt_sessId));
			if (nxt_sessId == MAX_SESSID)
				nxt_sessId = 0;
			else
				nxt_sessId++;
            return(i);
		}
		NDMP_DBG_PRINTF(D_INFO, ("ndmpGetFreeSessId:  [%d], in_use = %s, "
			"sessId %d\n", i, (sessId_Tbl[i].in_use == TRUE)? "yes":"no", 
			sessId_Tbl[i].sessId ));
	}	
	NDMP_DBG_PRINTF(D_INFO, ("ndmpGetFreeSessId:  SessId_tbl are full.\n"));
	return(-1);
	

} /* ndmpGetFreeSessId */

/*-----------------------------------------------------------------
 * Name:  ndmpSetSessId
 *
 * Description:  Set an session Id entry w/ process id.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/16/02
 *
 *-----------------------------------------------------------------
 */
void
ndmpSetSessId(int idx, int32 sessId, int32 pid, struct sockaddr_in  *from)
{

	if (sessId_Tbl[idx].in_use == TRUE)
	{
		sessId_Tbl[idx].pid = pid;
		sessId_Tbl[idx].sessId = sessId;
		memcpy(&sessId_Tbl[idx].from, from, sizeof(struct sockaddr_in));
	} else {
		NDMP_DBG_PRINTF(D_INFO, ("ndmpSetSessId:  Error, sessId "
				"entry %d is not in use\n", idx));
    }
    return;

} /* ndmpSetSessId */

void
ndmpMarkFreeSessId(pid_t pid)
{
    int32 i; 

	if (pid == -1)
		return;

	NDMP_DBG_PRINTF(D_INFO, ("ndmpMarkFreeSessId:  Freeing process %d\n", pid));

	/* Find the entry that is used by specified client */
	for (i = 0; i < NDMP_MAX_SESS; i++)
	{
		if (sessId_Tbl[i].pid == pid)
		{ 
			sessId_Tbl[i].sessDel = 1;
			break;
		}
	}

    return;
}

/*-----------------------------------------------------------------
 * Name:  ndmpFreeSessId
 *
 * Description:  Mark a session Id entry as available.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/16/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmpFreeSessId(pid_t pid, int32 idx)
{
    int32 i, find_by_pid = TRUE; 
	int32 Found = FALSE;

	if (pid == -1)
		find_by_pid = FALSE;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmpFreeSessId:  Freeing process %d\n", pid));

	if (find_by_pid)
	{
		/* Find the entry that is used by specified client */
		for (i = 0; i < NDMP_MAX_SESS; i++)
		{
			if (sessId_Tbl[i].pid == pid)
			{ 
				Found = TRUE;
				break;
			}
		}
	} else if (idx != -1) {
		i = idx;
		Found = TRUE;
	}

	if (Found)
	{
		bzero(&sessId_Tbl[i], sizeof(sessId_entry_t));
		sessId_Tbl[i].pid = -1;
		sessId_Tbl[i].sessId = -1;

		NDMP_DBG_PRINTF(D_INFO, ("ndmpFreeSessId:  Entry %d with pid %d released.\n", i, pid));

	} else {

		NDMP_DBG_PRINTF(D_INFO, ("ndmpFreeSessId:  No entry released.  "
				"Entry with pid %d not found\n", pid));
	}
    return(NFX_OK);

} /* ndmpFreeSessId */

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_readDumpRec
 *
 * Description: Read an entry from the memory mapped pointer to the
 *              dumpdate file.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

static void ndmp_readDumpRec(const char *const ptr, dmpdate_t * const * const drec)
{
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_readDumpRec: Begin\n"));
    sscanf(ptr, DUMPFILE_FMT,
            &((*drec)->fs_name.whole_thing), &((*drec)->date), &((*drec)->client_ip),
            &((*drec)->dump_level), &((*drec)->curr_bkup_ind), ((*drec)->snapshot));

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_readDumpRec: End\n"));
    return;
}

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_writeDumpRec
 *
 * Description: Write an entry to the file pointer.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

static void ndmp_writeDumpRec(FILE *fp, const dmpdate_t *const drec)
{
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_writeDumpRec: Begin\n"));
    fprintf(fp, DUMPFILE_FMT, drec->fs_name.whole_thing, drec->date, drec->client_ip,
                              drec->dump_level, drec->curr_bkup_ind, drec->snapshot);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_writeDumpRec: End\n"));
    return;
}

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_accessDumpDateFile
 *
 * Description: Open the dumpdate file and map it to the pointer.
 *              The fd is static so that the next call to the same
 *              function would close the file and unmap the memory
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

static size_t ndmp_accessDumpDateFile(char **ptr, const char *const fileName,
                                      const int32 mode, const int32 map_mode,
                                      const int32 map)
{
    static int  fd, flen = 0;
    struct stat status;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_accessDumpDateFile: Begin\n"));
    if (map == NDMP_MAP_FILE)
    {
        /*
         * Map the file name to pointer
         */
        *ptr = NULL;
        if ((fd = open(fileName, mode, 0)) == -1 ) {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_accessDumpDateFile: Error open %s\n",
                                      fileName));
            return 0;
        }
        fstat(fd, &status);
        flen = status.st_size;
        if (flen == 0)
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_accessDumpDateFile: File empty\n"));
            close(fd);
            return 0;
        }

        *ptr = mmap(0, flen, PROT_READ, map_mode, fd, 0);
        if (*ptr == MAP_FAILED)
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_accessDumpDateFile: Map %s failed\n",
                                      fileName));
            close(fd);
            return 0;
        }
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_accessDumpDateFile: %s mapped\n", fileName));
    }
    else
    {
        /*
         * Unmap the file name to pointer
         */
        munmap(*ptr, flen);
        close(fd);
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_accessDumpDateFile: Unmapped\n"));
        *ptr = NULL;
        fd = 0;
        flen = 0;
    }

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_accessDumpDateFile: End\n"));
    return flen;
}

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_getNextDumpRecord
 *
 * Description: Get the next record from the memory mapped pointer.
 *              Set the pointers to point to the next record for
 *              the getNext call.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

static int32 ndmp_getNextDumpRecord(char **start_ptr, char **last_ptr,
                              dmpdate_t * const * const drec, const int32 fromLastRec)
{
    char *the_ptr;
    int32 found = 0;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_getNextDumpRecord: Begin\n"));
    while (*last_ptr >= *start_ptr && !found)
    {
        if (fromLastRec)
        {
            if (*last_ptr > *start_ptr)
            {
                (*last_ptr)--;
                if ((**last_ptr) != '\n')
                    continue;
            }

            found = 1;
            the_ptr = (*last_ptr == *start_ptr)? *last_ptr : (*last_ptr)+1;
            ndmp_readDumpRec(the_ptr, drec);
        }
        else
        {
            if (*start_ptr == *last_ptr)
                return 0;
            the_ptr = *start_ptr;
            ndmp_readDumpRec(the_ptr, drec);
            found = 1;
            while ((**start_ptr) != '\n') (*start_ptr)++;
            *start_ptr = (*start_ptr == *last_ptr)?*start_ptr
                                                  :(*start_ptr)+1; // Next one
        }
    }

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_getNextDumpRecord: End\n"));
    return found;
}

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_sendSnapRemoveIndication
 *
 * Description: Send snap remove indication to the file system.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

void ndmp_sendSnapRemoveIndication(const int64 volId, const char * const snapName)
{
    fs_MsgSnapRemoveReq *req;
    int32  len;

    NDMP_DBG_PRINTF(D_INFO,("ndmp_sendSnapRemoveIndication : %s\n", snapName));
    len = sizeof(fs_MsgSnapRemoveReq) + strlen(snapName)+1;

    NDMP_SSC_ALLOC(req, fs_MsgSnapRemoveReq, len);
    req->hdr.requestType = FS_REQ_SNAP_REMOVE;
    req->hdr.len = len;
    // req->volId = (evm_volId_t)data->fs_volId;
    req->volId.whole_thing = volId;
    req->nameLen = strlen(snapName);
    strcpy(req->snapDirName, snapName);

    sendAgileMsg(-1, req, len, "efs", NFX_NFP_FP2, 1, MGMTPLANE, NULL, NULL, 0);
    E_LOG (class_1, info_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
              ("ndmp_sendSnapRemoveIndication: remove snapshot %s", snapName));

    NDMP_SSC_FREE(req);
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendSnapRemoveIndication: End\n"));
    return;

}

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_deleteDumpRecord
 *
 * Description: Delete one record from the dumpdate file.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

void ndmp_deleteDumpRecord(const char *const fileName, const char *const snapname)
{
    FILE *fp;
    int  flen;
    char *ptr, *last_ptr, *mmf;
    char bkupLink[1024];
    dmpdate_t *d;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_deleteDumpRecord: Begin\n"));
    bzero(bkupLink, 1024);
    strcpy(bkupLink, fileName);
    strcat(bkupLink, "_tmp_link");

    flen = ndmp_accessDumpDateFile(&mmf, fileName, O_RDONLY,
                                                  MAP_PRIVATE, NDMP_MAP_FILE);
    if (flen == 0 || !mmf)
    {
        NDMP_DBG_PRINTF(D_INFO,("ndmp_deleteDumpRecord: Unable to open file %s\n", fileName));
        return;
    }

    fp = fopen(bkupLink, "w");
    ptr = mmf;
    last_ptr = mmf + flen - 1;
    NDMP_SSC_ALLOC(d, dmpdate_t, sizeof(dmpdate_t));
    while (ndmp_getNextDumpRecord(&ptr, &last_ptr, &d, 0))
    {
        if (strcasecmp(d->snapshot, snapname))
            ndmp_writeDumpRec(fp, d);
    }
    E_LOG (class_1, info_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
              ("ndmp_deleteDumpRecord: delete dumprec %s", snapname));

    ndmp_accessDumpDateFile(&mmf, NULL, 0, 0, NDMP_UNMAP_FILE);
    rename(bkupLink, fileName);
    fclose(fp);
    NDMP_SSC_FREE(d);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_deleteDumpRecord: End\n"));
    return;
}

/*-----------------------------------------------------------------
 * Name:  ndmp_getDumpDate
 *
 * Description:     Get dump date from our dumpdates file
 *                  and return the date record to caller.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/29/01
 *
 *-----------------------------------------------------------------
 */
int32 
ndmp_getDumpDate(dmpdate_t *const *const ddrec, const int32 dump_level,
                 const int64 volid, const char *const volname,
                 const uint32 client_ip, char *const err)
{
    int         flen = 0, found = 0;
	char		*ptr, *last_ptr, dumpDateLink[1024];

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_getDumpDate: Begin\n"));
    if (!volname)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_getDumpDate:  Volname not specified\n"));
        return NFX_ERR;
    }

    /*
     * Read file through the link
     */
    bzero(dumpDateLink, 1024);
    sprintf(dumpDateLink, "%s%lld.dmp", dumpDateDir, volid);
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_getDumpDate:  openfile %s\n", dumpDateLink));

    flen = ndmp_accessDumpDateFile(&ptr, dumpDateLink, O_RDONLY,
                                                      MAP_SHARED, NDMP_MAP_FILE);
    if (flen == 0 || !ptr)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_getDumpDate: %s does not exist\n",
                                  dumpDateLink));
        sprintf (err, "User err: No prior full-backup. Continuing with backup");
        return NFX_OK;
    }

	last_ptr = ptr + flen - 1;
	if (*last_ptr != '\n')
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_getDumpDate: EOF not expected\n"));
        sprintf (err, "Server err: [NL]. Continuing with backup");
        ndmp_accessDumpDateFile(&ptr, NULL, 0, 0, NDMP_UNMAP_FILE);
		return NFX_OK;
	}
    while (*last_ptr == '\n')
        last_ptr--;
    if (last_ptr == ptr)
	{
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_getDumpDate: File's empty\n"));
        sprintf (err, "Server err: [M-NL]. Continuing with backup");
        ndmp_accessDumpDateFile(&ptr, NULL, 0, 0, NDMP_UNMAP_FILE);
        return NFX_OK;
    }
    last_ptr++; // Point to the first new-line of a dump record.


    while (ndmp_getNextDumpRecord(&ptr, &last_ptr, ddrec, 1))
    {
		/*
         * We know that dump date records are written chronologically
         * based on time bkup was done, so we find a record that matches
         * the volId and w/ a lower dump level than the one passed in args.
		 */
        if ((*ddrec)->curr_bkup_ind)
            continue;

        if ((*ddrec)->fs_name.whole_thing != volid)
        {
            if (last_ptr == ptr)
            {
                memset((*ddrec), 0, sizeof(dmpdate_t));
                break;
            }
            else
                continue;
        }

        if ((*ddrec)->client_ip != client_ip)
        {
            if (last_ptr == ptr)
            {
                memset((*ddrec), 0, sizeof(dmpdate_t));
                break;
            }
            else
                continue;
        }

        if ((*ddrec)->dump_level >= dump_level)
        {
            if (last_ptr == ptr)
            {
                memset((*ddrec), 0, sizeof(dmpdate_t));
                break;
            }
            else
                continue;
        }

        found = 1;
        NDMP_DBG_PRINTF(D_INFO,("ndmp_getDumpDate\n"));
        NDMP_DBG_PRINTF(D_INFO,("----------------\n"));
        NDMP_DBG_PRINTF(D_INFO,("fsname: 0x%llx\n",(*ddrec)->fs_name.whole_thing));
        NDMP_DBG_PRINTF(D_INFO,("date  : %d\n",(*ddrec)->date));
        NDMP_DBG_PRINTF(D_INFO,("client-ip: 0x%x\n",(*ddrec)->client_ip));
        NDMP_DBG_PRINTF(D_INFO,("dump_level: %d\n",(*ddrec)->dump_level));
        NDMP_DBG_PRINTF(D_INFO,("curr_bkup_ind: %d\n",(*ddrec)->curr_bkup_ind));
        NDMP_DBG_PRINTF(D_INFO,("snapshot: %s\n",(*ddrec)->snapshot));
        NDMP_DBG_PRINTF(D_INFO,("----------------\n"));
        break;
    }

    if (!found)
        sprintf (err, "User err: No prior backup less than %d level. Continuing with backup", dump_level);
   
    ndmp_accessDumpDateFile(&ptr, NULL, 0, 0, NDMP_UNMAP_FILE);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_getDumpDate: End\n"));
    return NFX_OK;
} /* ndmp_getDumpDate */

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_manageSnapShots
 *
 * Description: When we have decided to commit a snap record to the
 *              file, we should clean the dumpdate entries and delete
 *              the redundant snapshots.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

void ndmp_manageSnapShots(const char *const fileName, const int64 volId,
                          const dmpdate_t *const ddrec, const int32 tmp_all)
{
    FILE *fp;
    int  flen;
    char *ptr, *last_ptr, *mmf;
    char bkupLink[1024];
    dmpdate_t *d;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_manageSnapShots: Begin\n"));
    if (ddrec)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ddrec->dump_level: %d\n", ddrec->dump_level));
        NDMP_DBG_PRINTF(D_INFO, ("ddrec->client_ip: 0x%x\n", ddrec->client_ip));
        NDMP_DBG_PRINTF(D_INFO, ("ddrec->snapshot: %s\n", ddrec->snapshot));
    }
    bzero(bkupLink, 1024);
    strcpy(bkupLink, fileName);
    strcat(bkupLink, "_tmp_link");

    flen = ndmp_accessDumpDateFile(&mmf, fileName, O_RDONLY,
                                                  MAP_PRIVATE, NDMP_MAP_FILE);
    if (flen == 0 || !mmf)
    {
        NDMP_DBG_PRINTF(D_INFO,("ndmp_manageSnapShots: Unable to open file %s\n", fileName));
        return;
    }

    fp = fopen(bkupLink, "w");
    ptr = mmf;
    last_ptr = mmf + flen - 1;
    NDMP_SSC_ALLOC(d, dmpdate_t, sizeof(dmpdate_t));
    bzero(d, sizeof(dmpdate_t));
    while (ndmp_getNextDumpRecord(&ptr, &last_ptr, &d, 0))
    {
        if (tmp_all)
        {
            if (!d->curr_bkup_ind)
                ndmp_writeDumpRec(fp, d);
            else if (strlen(d->snapshot) > 0)
                /* Send file system indication that snapshot is not needed */
                ndmp_sendSnapRemoveIndication(volId, d->snapshot);
        }
        else
        {
            if (d->curr_bkup_ind)
                ndmp_writeDumpRec(fp, d);
            else if (d->dump_level < ddrec->dump_level)
                ndmp_writeDumpRec(fp, d);
            else if (d->client_ip != ddrec->client_ip)
                ndmp_writeDumpRec(fp, d);
            else if (strlen(d->snapshot) > 0)
                /* Send file system indication that snapshot is not needed */
                ndmp_sendSnapRemoveIndication(volId, d->snapshot);
        }
        bzero(d, sizeof(dmpdate_t));
    }

    ndmp_accessDumpDateFile(&mmf, NULL, 0, 0, NDMP_UNMAP_FILE);
    rename(bkupLink, fileName);
    fclose(fp);
    NDMP_SSC_FREE(d);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_manageSnapShots: End\n"));
    return;
}

/*
 *-----------------------------------------------------------------
 * Name:  ndmp_cleanDumpDate
 *
 * Description: When we receive volume events for a volume we will
 *              check against the dumpdate file and clean the
 *              entries or remove the file.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/9/02
 *
 *-----------------------------------------------------------------
 */

void ndmp_cleanDumpDate(char *const msg)
{
    nfx_eventMsg_t *evt = (nfx_eventMsg_t *)msg;
    char dumpDateFile[1024], dumpDateLink[1024], newdumpDateFile[1024];
    int32 delFile = 0, flen;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_cleanDumpDate:  Begin\n"));

    /* Don't send back the response */
    bzero(dumpDateFile, 1024);
    bzero(dumpDateLink, 1024);
    if (evt->eventType == NFX_EVENT_MKFS)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_cleanDumpDate: Mkfs event for %lld\n", evt->param.mkfsEvent.volId.whole_thing));
        sprintf(dumpDateLink, "%s%lld.dmp", dumpDateDir, evt->param.mkfsEvent.volId.whole_thing);
        delFile = 1;
    }
    else if (evt->eventType == NFX_EVENT_VOLUME &&
             evt->param.volEvent.event == EVENT_VOL_DELETE)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_cleanDumpDate: Volume delete event for %lld\n", evt->param.volEvent.volId.whole_thing));
        sprintf(dumpDateLink, "%s%lld.dmp", dumpDateDir, evt->param.volEvent.volId.whole_thing);
        delFile = 1;
    }
    else if (evt->eventType == NFX_EVENT_VOLUME &&
             evt->param.volEvent.event == EVENT_VOL_ONLINE)
    {
        bzero(newdumpDateFile, 1024);
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_cleanDumpDate: Volume online event for %lld\n", evt->param.volEvent.volId.whole_thing));
        sprintf(dumpDateLink, "%s%lld.dmp", dumpDateDir, evt->param.volEvent.volId.whole_thing);

        if (!((flen = readlink(dumpDateLink, dumpDateFile, 1024)) == -1))
        {
            ndmp_manageSnapShots(dumpDateFile, evt->param.volEvent.volId.whole_thing, NULL, 1);
        }
        
    }
    else
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_cleanDumpDate: Event unsupported by NDMP %d\n", evt->eventType));
        return;
    }

    if (delFile)
    {
        if (readlink(dumpDateLink, dumpDateFile, 1024) != -1)
        {
            /* Delete the file */
            NDMP_DBG_PRINTF(D_INFO,("ndmp_cleanDumpDate: file %s\n", dumpDateFile));
            unlink(dumpDateFile);
            unlink(dumpDateLink);
        }
    }

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_cleanDumpDate: End\n"));
    return;
}

/*-----------------------------------------------------------------
 * Name:  ndmp_setDumpDate
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/29/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_setDumpDate(ndmpSess_t *const sess, char8 *const rcvbuf, ndmp_src_t *const src)
{
    FILE         *fp;
    ndmp_NFXRsp_t *ndmp_intRsp = (ndmp_NFXRsp_t *) rcvbuf;
    dmpdate_t   *ddrec= (dmpdate_t *)ndmp_intRsp->content;
    char        dumpDateFile[1024], dumpDateLink[1024], newdumpDateFile[1024];
    int         flen;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_setDumpDate: Begin\n"));
    if (!ddrec->fs_name.whole_thing)
    {
        NDMP_DBG_PRINTF(D_INFO,("ndmp_setDumpDate: Empty file system entry\n"));
        return NFX_OK;
    }
    if (strlen(ddrec->snapshot) <= 0)
    {
        NDMP_DBG_PRINTF(D_INFO,("ndmp_setDumpDate: Empty snapshot name\n"));
        return NFX_OK;
    }
    if (sess->partial_bkup == TRUE)
    {
        NDMP_DBG_PRINTF(D_INFO,("ndmp_setDumpDate: Partial back. Entry not created\n"));
        if (ddrec->curr_bkup_ind == 0)
            ndmp_sendSnapRemoveIndication(ddrec->fs_name.whole_thing, ddrec->snapshot);
        return NFX_OK;
    }

    bzero(dumpDateFile, 1024);
    bzero(dumpDateLink, 1024);
    sprintf(dumpDateLink, "%s%lld.dmp", dumpDateDir, sess->Serv.Data.fs_volId);
    if ((flen = readlink(dumpDateLink, dumpDateFile, 1024)) == -1)
    {
        /* Link does not exist. We need to create it */
        sprintf(dumpDateFile, "%s%s_%lld.dmp", dumpDateDir, sess->Serv.Data.fs_volname, sess->Serv.Data.fs_volId);
        symlink(dumpDateFile, dumpDateLink);
    }
    else
    {
        bzero(newdumpDateFile, 1024);
        sprintf(newdumpDateFile, "%s%s_%lld.dmp", dumpDateDir, sess->Serv.Data.fs_volname, sess->Serv.Data.fs_volId);
        if (strcasecmp(newdumpDateFile, dumpDateFile))
        {
            /* Vol Name has changed. So change the link name for this file */
            rename(dumpDateFile, newdumpDateFile);
            unlink(dumpDateLink);
            symlink(newdumpDateFile, dumpDateLink);
            bzero(dumpDateFile, 1024);
            strcpy(dumpDateFile, newdumpDateFile);
        }
    }
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_setDumpDate : Create/Append file %s\n", dumpDateFile));

    /*
     * Clean dumpdate file and send file system indication to remove snapshot
     */
    ddrec->client_ip = sess->host_ipaddr.sin_addr.s_addr;
    if (ddrec->curr_bkup_ind == 0)
    {
        E_LOG (class_1, info_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
              ("ndmp_setDumpDate: commit dumprec %s", ddrec->snapshot));
        ndmp_deleteDumpRecord(dumpDateFile, ddrec->snapshot);
        ndmp_manageSnapShots(dumpDateFile, sess->Serv.Data.fs_volId, ddrec, 0);
    }
    else
        E_LOG (class_1, info_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
              ("ndmp_setDumpDate: Temp dumprec %s", ddrec->snapshot));

    if( (fp = fopen(dumpDateFile, "a")) == NULL )
    {
       	/* If not created yet, we create it */
         NDMP_DBG_PRINTF(D_INFO, ("Error: Unable to open dumpdate file: %s\n", dumpDateFile));
         return(NFX_ERR);
    }
    NDMP_DBG_PRINTF(D_INFO, ("Creating new Dumpdate file......\n"));

    ndmp_writeDumpRec(fp, ddrec);

    /* write(fd, ddrec, sizeof(dmpdate_t)); */
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_setDumpDate\n")); 
	NDMP_DBG_PRINTF(D_INFO, ("----------------\n")); 
	NDMP_DBG_PRINTF(D_INFO, ("fsname: 0x%llx\n", ddrec->fs_name.whole_thing));
	NDMP_DBG_PRINTF(D_INFO, ("date  : %d\n", ddrec->date));
	NDMP_DBG_PRINTF(D_INFO, ("Client-ip: 0x%x\n", ddrec->client_ip));
	NDMP_DBG_PRINTF(D_INFO, ("dump_level: %d\n", ddrec->dump_level));
	NDMP_DBG_PRINTF(D_INFO, ("snapshot: %s\n", ddrec->snapshot));
    fclose(fp);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_setDumpDate: End\n"));
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_fwdMsgtoNNIM
 *
 * Description: Forward MOVER & DATA server control message to our Mover 
 *				or Data state machines on the N-NIM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_fwdMsgtoNNIM(ndmpSess_t *sess, uint32 ndmp_nfx_mid, int32 userInfo)
{
    ndmp_NFXReq_t    *ndmp_req;   /* relay client cmd to Mover on NNIM */
	int32			 msglen, rc = NFX_OK;
	ndmpMsgEnt_t 	 *msgent = (ndmpMsgEnt_t *) sess->cl_msgent;
	

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_fwdMsgtoNNIM.\n"));

    /* Forward client cmd request to Mover on N_NIM */

	msglen =  sizeof(ndmp_NFXReq_t) + NDMP_MAX_REQ_SIZ -1;
    NDMP_SSC_ALLOC(ndmp_req, ndmp_NFXReq_t, msglen);
    if (ndmp_req == NULL)
	{
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_fwdMsgtoNNIM:  Agile message send to CLOSE tape failed\n"));
		return NFX_ERR;	
	}
    memset((void*)ndmp_req, 0, msglen);
    ndmp_req->hdr.mid        = ndmp_nfx_mid;
    ndmp_req->hdr.sessId     = sess->sessId;
#ifdef NFX_SMP
	ndmp_req->hdr.seqId     = sess->xmt2fp_seqId;
	sess->xmt2fp_seqId ++;
	ndmp_req->hdr.idx        = sess->sess_idx;
#endif

	memcpy(&ndmp_req->cl_hdr, &sess->msgRcvd.hdr, sizeof(ndmp_Hdr));
    ndmp_req->ndmp_mid       = sess->msgRcvd.hdr.msgId;
    ndmp_req->userInfo       = userInfo;
    ndmp_req->len1       	 = NDMP_MAX_REQ_SIZ;
    memcpy(ndmp_req->content, sess->msgRcvd.content, NDMP_MAX_REQ_SIZ);

    if (sendAgileMsg (sess->Agilesock, ndmp_req, msglen,
                              "ndmp",       /* NDMP on the N-NIM */
                               NFX_NFP_FP2, 1, MGMTPLANE, /* cpu, slot, plane */
                               NULL, NULL, 0) < 0)  
	{
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_fwdMsgtoNNIM:  Agile message send to FP failed\n"));
		rc = NFX_ERR;
    }
	if (msgent->stat_idx_snt != -1)
		sess->ndmp_ssc_chld_msgStats[msgent->stat_idx_snt]++;
	if (ndmp_req)
        NDMP_SSC_FREE(ndmp_req);
	return rc;

} /* ndmp_fwdMsgtoNNIM */

/*-----------------------------------------------------------------
 * Name:    getChallengStr
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
void getChallengStr(char8 *dst)
{
    int randnum;

    srand(10);
    randnum = rand();
    NDMP_DBG_PRINTF(D_INFO, ("getChallengStr:  %d", randnum));

    memset(dst, 0, NDMP_CHALLENGE_LEN);
    sprintf (dst, "%d", randnum);

} /* getChallengStr */

/*-----------------------------------------------------------------
 * Name:   ndmp_rspClient 
 *
 * Description:	 Send a response to NDMP client.  Return NFX_ERR if
 *				can't and NFX_OK if can.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_rspClient(ndmpSess_t *sess, void *cl_rsp)
{

	if (ndmpXmtMsg(sess, NDMP_MSG_RSP, sess->msgRcvd.hdr.msgId, 
			sess->msgRcvd.hdr.xmtSeq, NDMP_NO_ERR, (void *)cl_rsp) < 0)
	{
		E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
				("ndmp_rspClient: error sending %d rsp.\n", 
				sess->msgRcvd.hdr.msgId));
    	NDMP_DBG_PRINTF(D_INFO, ("ndmp_rspClient: error sending %d rsp.\n", 
				sess->msgRcvd.hdr.msgId));
		return NFX_ERR;
    }
	return NFX_OK;

} /* ndmp_rspClient */

/*-----------------------------------------------------------------
 * Name:    ndmp_chkDevOpen
 *
 * Description: Check if a tape or a media changer device is already
 * 				opened.  If not, send client an error msg.  User is
 *				to set error code in cl_rsp of correct 
 *				respond msg type based on msg request type
 *				prior to this call.
 *				
 *				dev_typ = TAPE_DEV or SCSI_DEV
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_chkDevOpen(ndmpSess_t *sess, void *cl_rsp, int32 dev_typ)
{

	uint32 handle;

	handle = (dev_typ == TAPE_DEV)? sess->Serv.Tape.hdl: sess->Serv.SCSI.hdl;

    if (!handle)
    {
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
						("ndmp_chkDevOpen: %s device is not open.\n", 
						(dev_typ == TAPE_DEV)? "TAPE": "SCSI"));
    	NDMP_DBG_PRINTF(D_INFO, ("ndmp_chkDevOpen: %s device is not open.\n",
                        (dev_typ == TAPE_DEV)? "TAPE": "SCSI"));
		ndmp_rspClient(sess, cl_rsp);
		return NFX_ERR;
    }
    return NFX_OK;

} /* ndmp_chkDevOpen */

/*-----------------------------------------------------------------
 * Name: sndMsgtoTapeDrv
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32 sndMsgtoTapeDrv(ndmpSess_t *sess, void *Req, int32 req_len, void *cl_rsp)
{

	ndmpMsgEnt_t *msgent = (ndmpMsgEnt_t *) sess->cl_msgent;

    /* Query tape driver */
    if (sendAgileMsg(sess->Agilesock, Req, req_len,
                        "tape-driver", 0, 0, 0,  /* to which applic */
                        NULL, NULL, 0) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("Agile message send failed\n"));
        if (ndmpXmtMsg(sess, NDMP_MSG_RSP, sess->msgRcvd.hdr.msgId, 
				sess->msgRcvd.hdr.xmtSeq, NDMP_NO_ERR, (void *)cl_rsp) < 0)
        {
            E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
            		("ndMsgtoTapeDrv:  Error sending rsp to client, msgid %d.\n", sess->msgRcvd.hdr.msgId));
		
        }
        return NFX_ERR;
    }
	if (msgent->stat_idx_snt != -1)
		sess->ndmp_ssc_chld_msgStats[msgent->stat_idx_snt]++;
    return NFX_OK;

}  /* sndMsgtoTapeDrv */

/*-----------------------------------------------------------------
 * Name: free_env_on_SSC
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        2/2/02
 *
 *-----------------------------------------------------------------
 */
int32
free_env_on_SSC(Data_SSC_t *data)
{
    int32   i;

    if (data->env_lst)
    {
        for (i = 0; i < data->srvr_env_len; i++)
        {
            if (data->env_lst[i].name)
                NDMP_SSC_FREE(data->env_lst[i].name);
            if (data->env_lst[i].value)
                NDMP_SSC_FREE(data->env_lst[i].value);
        }
        NDMP_SSC_FREE(data->env_lst);
    }
    data->env_lst = NULL;
    data->env_len = 0;

    return NFX_OK;
} /* free_env_on_SSC */

/*-----------------------------------------------------------------
 * Name: decode_tape_ecode
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/2/02
 *
 *-----------------------------------------------------------------
 */
char *
decode_tape_ecode(int32 ecode)
{
	switch (ecode)
	{
		case TAPE_ECODE_SUCCESS: return "Success";
		case TAPE_ECODE_DEVICE_BUSY: return "Device Busy";
		case TAPE_ECODE_NO_DEVICE: return "No Device";
		case TAPE_ECODE_NO_TAPE_LOADED: return "No Tape Loaded";
		case TAPE_ECODE_ILI: return "Incorrect-length-indicator (ILI)";
		case TAPE_ECODE_EOM_ENCOUNTERED: return "EOM";
		case TAPE_ECODE_FM_ENCOUNTERED: return "File Mark Encountered";
		case TAPE_ECODE_DATA_SIZE_MISMATCH: return "Data size must be multiple of fixed block size";
		case TAPE_ECODE_MEDIUM_ERR: return "Media Err";
		case TAPE_ECODE_EOD_ENCOUNTERED: return "End of Data";
		case TAPE_ECODE_IO_ERROR: return "I/O Error";
		case TAPE_ECODE_DEVICE_NOT_OPEN: return "Device Not Open";
		default: return "Invalide Tape Ecode";
	}
}
/*-----------------------------------------------------------------
 * Name: decode_tape_mtio
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/2/02
 *
 *-----------------------------------------------------------------
 */
char *
decode_tape_mtio(int32 tape_op)
{
	switch (tape_op)
	{
		case NDMP_MTIO_FSF : return "Fwdspace-Filemark";      // fwd space filemarks
		case NDMP_MTIO_BSF : return "Bkspace-Filemark";       // bwd space filemarks
		case NDMP_MTIO_FSR : return "Fwdspace-Rec";           // fwd space tape records
		case NDMP_MTIO_BSR : return "Bkspace-Rec";            // bwd space tape records
		case NDMP_MTIO_REW : return "Rewind";                 // rewind
		case NDMP_MTIO_EOF : return "Write-Filemarks";        // write end-of-file marks
		case NDMP_MTIO_OFF : return "Unload";                 // unload, eject tape
		default:   return	"Undefined Ops";
	}
}

/*-----------------------------------------------------------------
 * Name: ndmp_regGetOption
 *
 * Description: Get registry option for NDMP variables.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 12/03/02
 *
 *-----------------------------------------------------------------
 */
static int32 ndmp_regGetOption(char *option, int32 optionLen, char *rsp_buf)
{
    opt_msgHdr_t        *hdr;
    opt_getOptionReq_t  *req;
    opt_getOptionRsp_t  *rsp;
    char                msg[1024];
    int32               len, rc;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_regGetOption : Get option %s\n", option));

    memset(msg, 0, sizeof(char)*1024);
    hdr = (opt_msgHdr_t *) msg;
    hdr->msgType = OPT_GET_OPTION_REQ;
    hdr->userInfo = 0;
    req = (opt_getOptionReq_t *) (msg + sizeof(opt_msgHdr_t));
    memcpy((void *) &req->option, (void *) option, optionLen);
    len = sizeof(opt_msgHdr_t) + sizeof(opt_getOptionReq_t) + 1;
    rc = sendAgileMsg(-1, msg, len, "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &waitRegm, rsp_buf, sizeof(char)*1024);
    if (rc <= 0)
        return -1;

    hdr = (opt_msgHdr_t *) rsp_buf;
    if (hdr->msgType != OPT_GET_OPTION_RSP)
        return -1;

    rsp = (opt_getOptionRsp_t *) (rsp_buf + sizeof(opt_msgHdr_t));
    if (rsp->nbEntry == 0)
        NDMP_DBG_PRINTF(D_INFO, ("No registry entry found\n"));

    return rc;

} /* ndmp_regGetOption */

/*-----------------------------------------------------------------
 * Name: ndmp_regSetOption
 *
 * Description: Set registry option for NDMP variables.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 12/03/02
 *
 *-----------------------------------------------------------------
 */
void ndmp_regSetOption(char *option, char *value)
{
    opt_msgHdr_t        *hdr;
    opt_setOptionReq_t  *req;
    opt_setOptionRsp_t  *rsp;
    char                msg[1024], rsp_buf[1024];
    int32               len, rc;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_regSetOption : Set option %s\n", option));

    memset(msg, 0, sizeof(char)*1024);
    memset(rsp_buf, 0, sizeof(char)*1024);
    hdr = (opt_msgHdr_t *) msg;
    hdr->msgType = OPT_SET_OPTION_REQ;
    hdr->userInfo = 0;
    req = (opt_setOptionReq_t *) (msg + sizeof(opt_msgHdr_t));
    strcpy(req->entry.option, option);
    strcpy(req->entry.value, value);
    len = sizeof(opt_msgHdr_t) + sizeof(opt_setOptionReq_t) + 1;
    rc = sendAgileMsg(-1, msg, len, "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &waitRegm, rsp_buf, sizeof(char)*1024);

    if (rc <= 0)
        return;

    hdr = (opt_msgHdr_t *) rsp_buf;
    if (hdr->msgType != OPT_SET_OPTION_RSP)
        return;

    rsp = (opt_setOptionRsp_t *) (rsp_buf + sizeof(opt_msgHdr_t));
    if (rsp->result != NFX_OK)
        NDMP_DBG_PRINTF(D_INFO, ("Registry entry error.\n"));

    return;

} /* ndmp_regSetOption */

/*-----------------------------------------------------------------
 * Name: ndmp_getCfgVal
 *
 * Description: Make the below code (taken from ndmp_readRegistryEntry()
 * 				into a func so multiple	variables can be added easily. 
 * 				Mainly, read the option string from the registry and 
 * 				assign it to setVar.  If there's error, set the same
 * 				option string in the registry w/ the def_str.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  12/04/02
 *
 * Input:  opt_str - option string to get in Registry
 *		   def_str - default string you want to set to in registry
 * 		   setVar  - variable that you want to set the value 
 *				you just got from the registry.
 *		   min  - min value opt_str can be, used for validation
 *		   max  - max value, ditto above 
 *-----------------------------------------------------------------
 */
void ndmp_getCfgVal(char *opt_str, char *def_str, int *setVar, int min, int max)
{
    int32               len, n, val;
    char                msg[1024];
    opt_getOptionRsp_t  *rsp;

    n = strlen(opt_str);
    len = ndmp_regGetOption(opt_str, n, msg);
    if (len > 0)
    {
        rsp = (opt_getOptionRsp_t *)(msg + sizeof(opt_msgHdr_t));
        if (rsp->nbEntry)
        {
			val = atoi(rsp->entry[0].value);
			if ((val >= min) && (val <= max))
				*setVar = val;
			else {
				/* Just in case someone mess around w/ config file
				* manually.  If value read is outside range, we'll
				* reset it to default.
				*/
				printf("Config for %s = %d which is outside of range %d-%d."
							" Resetting to default: %s\n",
							opt_str, val, min, max, def_str);
            	ndmp_regSetOption(opt_str, def_str);
			}
        }
        else
            ndmp_regSetOption(opt_str, def_str);
    }
    else
        ndmp_regSetOption(opt_str, def_str);
}

/*-----------------------------------------------------------------
 * Name: ndmp_readRegistryEntry
 *
 * Description:
 *
 * Created by:  Nitin Jain
 *
 * Date Created:  12/02/02
 *
 *-----------------------------------------------------------------
 */
void ndmp_readRegistryEntry()
{

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_readRegistryEntry : Begin\n"));

	/* Obtain stored config for 'feature enabled/disabled', default
	 * is ON. 
	 */
	ndmp_getCfgVal(OPTIONS_NDMP_ENBL_MODE, "1", &ndmp_enabled, 0, 1);

	/* Obtain stored config for 'highest NDMP version supported'
	 * Default is ver 3.
	 */
	ndmp_getCfgVal(OPTIONS_NDMP_HI_VER, "3", &ndmp_hi_ver, 2, 3);

	/* Obtain stored config for 'restore mode' */
	ndmp_getCfgVal(OPTIONS_NDMP_RESTORE_MODE, "0", 
					&ndmp_restore_mode, 0, 1);

	/* Obtain stored config for 'trace file mode', i.e. capture
	 * traces between server and client or not.  Default is 
	 * ON.
	 */
	ndmp_getCfgVal(OPTIONS_NDMP_TRACE_MODE, "1", &ndmp_trace_msg, 0, 1);

	/* Obtain stored config for 'trace file duration', i.e. the captured
	 * traces between server and client will be retained for how many
	 * days.   Default is 1 day.
	 */
	ndmp_getCfgVal(OPTIONS_NDMP_TRACE_DURATION, "1", 
					&ndmp_trace_files, 1, 10);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_readRegistryEntry : End\n"));

    return;
}
