/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb-ns.h,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb-ns.h
 *
 * Description	: NetBIOS name service definitions.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */
#ifndef _NB_NS_H
#define _NB_NS_H

/*-------------------------
 * Name service definitions
 *-------------------------
 */
#define NB_NS_NAMETYPE_WS			0x00
#define NB_NS_NAMETYPE_SERVER		0x20
#define NB_NS_NAMETYPE_DOMAINCTRLR	0x1C
#define NB_NS_NAMETYPE_MASTBROWSER	0x1D

#define NB_NS_TCP_PORT				137
#define NB_NS_UDP_PORT				137
#define NB_NS_TTL					3000000000U	// milliseconds. Over 3 days.
#define NB_NS_INFINITE_TTL			0
#define NB_NS_CONFLICT_TIMER		1000		// 1 second

/* Opcode definitions */
#define NB_NS_OPCODE_QUERY			0x0
#define NB_NS_OPCODE_REG			0x05
#define NB_NS_OPCODE_RELEASE		0x06
#define NB_NS_OPCODE_WACK			0x07
#define NB_NS_OPCODE_REFRESH_8		0x08
#define NB_NS_OPCODE_REFRESH_9		0x09	// Alternative refresh opcode
#define NB_NS_OPCODE_MULTIHOMED_REG	0x0F	// Multi-homed name registration

/* Question type definitions */
#define NB_NS_QTYPE_NB				0x0020	
								// NetBIOS general Name Service Resource Record
#define NB_NS_QTYPE_NBSTAT			0x0021	
								// NetBIOS NODE STATUS Resource Record

/* Question class definitions */
#define NB_NS_QCLASS_IN				0x0001	// Internet class

/* Resource record type definitions */
#define NB_NS_RRTYPE_A				0x0001	// IP address Resource Record
#define NB_NS_RRTYPE_NS				0x0002	// Name Server Resource Record
#define NB_NS_RRTYPE_NULL			0x000A	// NULL Resource Record
#define NB_NS_RRTYPE_NB				0x0020
								// NetBIOS general Name Service Resource Record
#define NB_NS_RRTYPE_NBSTAT			0x0021
								// NetBIOS NODE STATUS Resource Record

/* Resource record class definitions */
#define NB_NS_RRCLASS_IN			0x0001	// Internet class

/* NB_FLAGS field of the Resource Record RDATA filed for RR_TYPE of "NB" */
#define NB_NS_NBFLAG_BNODE			0x00	// Broadcast node type
#define NB_NS_NBFLAG_PNODE			0x20	// Point-to-point node type
#define NB_NS_NBFLAG_MNODE			0x40	// Mixed bcast and p-p node type
#define NB_NS_NBFLAG_HNODE			0x60	// Microsoft 'hybrid' node type
#define NB_NS_NBFLAG_GROUP			0x80	// RR_NAME is a GROUP NetBIOS name
#define NB_NS_NBFLAG_UNIQUE			0x00	// RR_NAME is a UNIQUE NetBIOS name
#define NB_NS_NBFLAG_NODE_MASK		0x60	// node type mask
#define NB_NS_NBFLAG_MASK			0xE0	// NB_FLAGS mask

/* Error codes for NetBIOS name service requests */
#define NB_NS_FMT_ERR	0x1		// Packet format error
#define NB_NS_SRV_ERR	0x2		// Internal server error
#define NB_NS_NAM_ERR	0x3		// Name does not exist
#define NB_NS_IMP_ERR	0x4		// Request not implemented
#define NB_NS_RFS_ERR	0x5		// Request refused
#define NB_NS_ACT_ERR	0x6		// Active error - name owned by another host
#define NB_NS_CFT_ERR	0x7		// Name in conflict error

/*----------------------------------------
 * Name service packet related structures
 *----------------------------------------
 */
 
/* NM flags - flags for operation */
typedef struct nb_nsNMFlags
{
	boolean		bcast;			// Broadcast
	boolean 	recAval;		// Recursion available
	boolean		recDesired;		// Recursion desired
	boolean		trunc;			// Truncation
	boolean		auth;			// Authoritative answer
} nb_nsNMFlags_t;

/* Name service packet header */
typedef struct nb_nsHeader
{
	uint16			nameTransID;	// Name service transaction ID
	uint16			opcode;			// Packet type code
	boolean			rsp;			// TURE if pkt is response, FALSE if request
	nb_nsNMFlags_t	nmFlags;		// Flags for operation
	uint16			rcode;			// Result codes of request
	uint16			qdCount;		// Number of question entries
	uint16			anCount;		// Number of answer resource records
	uint16			nsCount;		// Number of authority resource records
	uint16			arCount;		// Number of additional resource records
} nb_nsHeader_t;

/* Name service packet question entry */
typedef struct nb_nsQuestion
{
	nb_name_t		qName;			// Question name
	uint16			qType;			// Question type
	uint16			qClass;			// Question class
} nb_nsQuestion_t;

/* Name service packet resource record */
typedef struct nb_nsResource
{
	nb_name_t	rrName;			// Name corresponding to the resource record
	uint16		rrType;			// Resource record type code
	uint16		rrClass;		// Resource record class code
	uint64		ttl;			// TTL of the resource record's name
	uint16		rdLength;		// Number of bytes in RDATA field
	char		rdata[NB_S_MAX_IP_DATAGRAM]; 
								// Resource info for the NetBIOS name
} nb_nsResource_t;

/* Name service packet */
typedef struct nb_nsPacket
{
	nb_nsHeader_t	header;
	nb_nsQuestion_t	question;
	nb_nsResource_t	*answer;
	nb_nsResource_t	*authority;
	nb_nsResource_t	*additional;
} nb_nsPacket_t;


typedef struct nb_nsRspParam	// Parameters of functions used to generate
								// a response message.
{
	nb_nsPacket_t	*nsInReqPktP;	// Incoming request packet
	uint16			rcode;			// Result code
	uint16			reqType;		// Request type
	uint16			opcode;			// Opcode
	uint64			ttl;			// Time to live
	uint8			*rdata;			// rdata
	uint16			dataLen;
} nb_nsRspParam_t;

/*--------------------
 * Function prototypes
 *--------------------
 */
nb_nsPacket_t *nb_nsCreateReqPkt(nb_name_t *name, boolean bcast);

void nb_nsDeletePkt(nb_nsPacket_t *nsPktP);

void nb_nsSetNBFlags(uchar8 *rdataP, uchar8 gFlag, uchar8 nodeType);

uint8 nb_nsGetNBFlags(uchar8 *rdata);

boolean nb_nsIsGroup(uchar8 *rdata);

int32 nb_nsCreateAddRsrc(nb_nsPacket_t *nsPktP, uint32 ipAddr, boolean group);

void nb_nsClearRD(nb_nsPacket_t *nsPktP, pkt_desc_t *nsXmtPktP);

nb_nsPacket_t *nb_nsCreateNameRegPkt(nb_srvCB_t *nbSrvCbP,
									 nb_name_t *reqName,
									 boolean bcast,
									 boolean group);

nb_nsPacket_t *nb_nsCreateNameQueryPkt(nb_srvCB_t *nbSrvCbP,
									   nb_name_t *reqName,
									   boolean bcast);

nb_nsPacket_t *nb_nsCreateNameReleasePkt(nb_srvCB_t *nbSrvCbP,
										 nb_name_t *reqName,
										 boolean bcast,
										 boolean group);

nb_nsPacket_t *nb_nsCreateNameRefreshPkt(nb_srvCB_t *nbSrvCbP,
										 nb_name_t *reqName,
										 boolean bcast,
										 boolean group);

int32 nb_nsSendRequest(nb_reqCB_t *nbReqCbP,
					   nb_name_t *reqName,
					   uint32 raddr,
					   uint16 rport,
					   uint16 reqType,
					   boolean group);

nb_reqCB_t *nb_nsRequest(nb_srvCB_t *nbSrvCbP,
						 nb_name_t *reqName,
						 uint16 reqType,
						 boolean group);

int32 nb_nsLookupNameInCache(nb_srvCB_t *nbSrvCbP,
							 nb_name_t *nbName,
							 nb_queryCB_t **nbQueryCbP);

int32 nb_nsProcWackRsp(nb_nsPacket_t *nsInRspPktP, nb_reqCB_t *nsReqCbP);

int32 nb_nsProcNameRegRsp(nb_nsPacket_t *nsInRspPktP,
						  nb_reqCB_t *nsReqCbP);

int32 nb_nsProcNameQueryRsp(nb_incomingMsgCB_t *inMsgCbP,
							nb_nsPacket_t *nsInRspPktP,
							nb_reqCB_t *nsReqCbP);

int32 nb_nsProcNameReleaseRsp(nb_nsPacket_t *nsInRspPktP,
							  nb_reqCB_t *nsReqCbP);

void nb_nsProcResponse(nb_incomingMsgCB_t *inMsgCbP,
					   nb_nsPacket_t *nsInRspPktP);

void nb_nsCreateRspPkt(nb_nsRspParam_t *rspParamP,
					   nb_nsPacket_t *nsOutRspPktP);

int32 nb_nsResponse(nb_incomingMsgCB_t *inMsgCbP,
					nb_nsRspParam_t *rspParamP);

void nb_nsProcNameRegReq(nb_incomingMsgCB_t *inMsgCbP,
						 nb_nsPacket_t *nsInReqPktP);

void nb_nsProcNameRefreshReq(nb_incomingMsgCB_t *inMsgCbP,
							 nb_nsPacket_t *nsInReqPktP);

void nb_nsProcNameQueryReq(nb_incomingMsgCB_t *inMsgCbP,
						   nb_nsPacket_t *nsInReqPktP);

void nb_nsProcNodeStatusReq(nb_incomingMsgCB_t *inMsgCbP,
							nb_nsPacket_t *nsInReqPktP);

void nb_nsProcNameReleaseReq(nb_incomingMsgCB_t *inMsgCbP,
							 nb_nsPacket_t *nsInReqPktP);

void nb_nsProcRequest(nb_incomingMsgCB_t *inMsgCbP,
					  nb_nsPacket_t *nsInReqPktP);

int32 nb_nsRcvMsg(nb_incomingMsgCB_t *inMsgCbP,
				  pkt_desc_t *rcvPktDescP);

void nb_nsProcNameRegResult(nb_reqCB_t *nbReqCbP,
							nb_nsPacket_t *rspPktP,
			   				int32 result);

void nb_nsProcNameQueryResult(nb_reqCB_t *nbReqCbP,
							  nb_nsPacket_t *rspPktP,
							  int32 result);

void nb_nsProcNameReleaseResult(nb_reqCB_t *nbReqCbP,
								int32 result);

int32 nb_nsReqTimeout(int32 timerHdl, void* param1,
					  void* param2, void* param3, void* param4);

int32 nb_nsConflTimeout(int32 timerHdl, void* param1,
						void* param2, void* param3, void* param4);

int32 nb_nsNameQueryReq(nb_srvCB_t *nbSrvCbP,
						nb_name_t name,
						boolean group,
						void *userInfo,
						void (*nb_nsNameQueryInd) (uint16 ipAddrCnt,
													uint32 *ipAddrs,
													void *userInfo,
													int32 status));

uint32 nb_nsStartNameListTimer(nb_srvCB_t *nbSrvCbP);

int32 nb_nsNameListTimeout(int32 timerHdl, void* param1,
						   void* param2, void* param3, void* param4);
#endif /* _NB_NS_H */
