/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfp-cmd.c
 *
 * RCS:  $Id: nfp-cmd.c,v 1.8 2002/11/06 03:35:43 joseph Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: 
 *
 * Created by: Joseph Vu 
 *
 * Date Created:        9/18/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/nfs-api.h>

int32 nfp_cmdShow ( int32 argc, char *argv[] );
int32 nfp_cmdTrace ( int32 argc, char *argv[] );
int32 nfp_displayStat(void);

typedef struct
{
	int32   (*cmdFunc)( int32 argc, char *argv[] );
	char    *cmd;
	char    *helpTxt;
} nfp_cmdTbl_t;

nfp_cmdTbl_t    nfp_cmdTable[ ] =
{
	{ nfp_cmdShow,  "show",     "nfp show statistics" },
	{ nfp_cmdTrace, "trace",    "nfp trace on|off" }
};

#define NFP_CMD_MAX sizeof(nfp_cmdTable)/sizeof(nfp_cmdTbl_t)

/*-----------------------------------------------------------------
 * Name: nfp_cmdHelpAll
 *
 * Description: Display help-texts for all NFP commands.
 *
 * Created by : Joseph Vu
 *
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfp_cmdHelpAll (void)
{
	int32   i;

	printf("Usage:");
	for (i=0; i < NFP_CMD_MAX; i++)
	{
		printf("\t%s\n", nfp_cmdTable[i].helpTxt);
	}
	return (0);
}

/*-----------------------------------------------------------------
 * Name: nfp_cmdHelpCmd
 *
 * Description: Display help-texts for an individual NFP command.
 *
 * Created by : Joseph Vu
 *
 * Date Created: 10/5/2001
 *-----------------------------------------------------------------
 */
int32 nfp_cmdHelpCmd (char *pCmd)
{
	int32   i;

	for (i=0; i < NFP_CMD_MAX; i++)
	{
		if (strequ(pCmd, nfp_cmdTable[i].cmd))
		{
			printf("Usage:");
			printf("\t%s\n", nfp_cmdTable[i].helpTxt);
			return 0;
		}
	}
	return (nfp_cmdHelpAll( ));

}

/*-----------------------------------------------------------------
 * Name: nfp_cmdHelp
 *
 * Description: Display NFP help-texts.
 *
 *              nfp help [show | trace]
 *
 * Created by : Joseph Vu
 *
 * Date Created: 10/05/2001
 *-----------------------------------------------------------------
 */
int32 nfp_cmdHelp ( int32 argc, char *argv[] )
{
	//int32 i;

	if (argc < 3)
		return (nfp_cmdHelpAll());
	else
		return (nfp_cmdHelpCmd( argv[2] ));
}

/*-----------------------------------------------------------------
 * Name: nfp_cmdShow
 *
 * Description: Dispatch for nfp show commands.
 *
 *              nfp show [stat]
 *
 * Created by : Joseph Vu
 *
 * Date Created: 10/05/2001
 *-----------------------------------------------------------------
 */
int32 nfp_cmdShow ( int32 argc, char *argv[] )
{
	if (argc < 3)
	{
		nfp_cmdHelpCmd("show");
		return (0);
	}

	if ( !(strncmp(argv[2], "stat", 4)))
	{
		return (nfp_displayStat());
	}
	else
		return (nfp_cmdHelpCmd("show"));
}

/*-----------------------------------------------------------------
 * Name: nfp_cmdTrace
 *
 * Description: Enables or disables NFP tracing.
 *
 *              nfp trace <on|off>
 *
 * Created by : Joseph Vu
 *
 * Date Created: 10/05/2001
 *-----------------------------------------------------------------
 */
int32 nfp_cmdTrace ( int32 argc, char *argv[] )
{
	if (argc < 3)
	{
		nfp_cmdHelpCmd("trace");
		return 0;
	}
	if (strequ(argv[2], "on"))
	{
		//trace_enable = 1;
		printf("NFP trace has been turned on.\n");
	}
	else if (strequ(argv[2], "off"))
	{
        //trace_enable = 0;
		printf("NFS trace has been turned off\n");
	}
	else
		nfp_cmdHelpCmd("trace");

	return 0;
}
/*-----------------------------------------------------------------
 * Name: nfp_cmd
 *
 * Description: Main dispatch for NFP commands.  Called from
 *              command table.
 *
 *              nfp show statistics
 *              nfp trace on|off
 * Created by : Joseph Vu
 *
 * Date Created: 10/5/2001
 *-----------------------------------------------------------------
 */
int32 nfp_cmd ( int32 argc, char *argv[] )
{
	int32	i;

	if (argc < 2)
	{
		return (nfp_cmdHelp(argc, argv));
	}

	for (i=0; i < NFP_CMD_MAX; i++)
	{
		if (strequ(argv[1], nfp_cmdTable[i].cmd))
		{
			return ((*nfp_cmdTable[i].cmdFunc)(argc, argv));
		}
	}
	return (nfp_cmdHelp(argc, argv));
}

/*-----------------------------------------------------------------
 * Name: nfp_displayStat 
 *
 * Description: 
 *              show NFP statistics
 * Created by : Joseph Vu
 *
 * Date Created: 10/5/2001
 *-----------------------------------------------------------------
 */
int32 nfp_displayStat(void)
{
//	uint32                  i;
//	nfp_fileCBPtr_t         fileCBp;
	nfp_stat_t              *statp;
	statp = &nfp_globalCfgp->debugStat;

	printf("NFP Debug Statistic Counters\n");
	printf("----------------------------\n");
	printf("No. of Avail Client Control Blocks :           %d\n", 
					nfp_globalCfgp->clientCbCfgList.freeEntriesCnt);
	printf("No. of Alloc Client Control Blocks :           %d\n", statp->numOfAllocClientCB);
	printf("No. of Free Client Control Blocks :            %d\n", statp->numOfFreeClientCB);
	printf("Client Control Blocks LWM:                     %d\n", statp->clientCbLwm);
	printf("No. of Avail File System Control Block:        %d\n",
					nfp_globalCfgp->fileSysCbCfgList.freeEntriesCnt);
	printf("No. of Alloc File System Control Block:        %d\n", statp->numOfAllocFileSysCB);
	printf("No. of Free File System Control Block:         %d\n", statp->numOfFreeFileSysCB);
	printf("File System Control Block LWM:                 %d\n", statp->fileSysCbLwm);
	printf("No. of Avail File Control Blocks:              %d\n", 
					nfp_globalCfgp->fileCbCfgList.freeEntriesCnt);
	printf("No. of Alloc File Control Blocks:              %d\n", statp->numOfAllocFileCB);
	printf("No. of Free File Control Blocks:               %d\n", statp->numOfFreeFileCB);
	printf("File Control Blocks LWM:                       %d\n", statp->fileCbLwm);
	printf("No. of Avail Request Control Blocks:           %d\n", 
					nfp_globalCfgp->reqCbCfgList.freeEntriesCnt);
	printf("No. of Alloc Request Control Blocks:           %d\n", statp->numOfAllocReqCB);
	printf("No. of Free Request Control Blocks:            %d\n", statp->numOfFreeReqCB);
	printf("Request Control Blocks LWM:                    %d\n", statp->reqCbLwm);
	printf("No. of Avail Reply Control Blocks:             %d\n", 
					nfp_globalCfgp->replyCbCfgList.freeEntriesCnt);
	printf("No. of Alloc Reply Control Blocks:             %d\n", statp->numOfAllocReplyCB);
	printf("No. of Free Reply Control Blocks:              %d\n", statp->numOfFreeReplyCB);
	printf("Reply Control Blocks LWM:                      %d\n", statp->replyCbLwm);
	printf("No. of Avail Lock Control Blocks:              %d\n", 
					nfp_globalCfgp->lockCbCfgList.freeEntriesCnt);
	printf("No. of Alloc Lock Control Blocks:              %d\n", statp->numOfAllocLockCB);
	printf("No. of Free Lock Control Blocks:               %d\n", statp->numOfFreeLockCB);
	printf("Lock Control Blocks LWM:                       %d\n", statp->lockCbLwm);
	printf("No. of Avail Generic Nodes:                    %d\n", 
					nfp_globalCfgp->nodeCbCfgList.freeEntriesCnt);
	printf("No. of Alloc Generic Nodes:                    %d\n", statp->numOfAllocNode);
	printf("No. of Alloc Generic Nodes:                    %d\n", statp->numOfFreeNode);
	printf("No. of Client Control Blocks Err:              %d\n", statp->numOfBadClientCB);
	printf("No. of File System Control Blocks Err:         %d\n", statp->numOfBadFileSysCB);
	printf("No. of File Control Blocks Err:                %d\n", statp->numOfBadFileCB);
	printf("No. of Request Control Blocks Err:             %d\n", statp->numOfBadReqCB);
	printf("No. of Reply Control Blocks Err:               %d\n", statp->numOfBadReplyCB);
	printf("No. of Lock Control Blocks Err:                %d\n", statp->numOfBadLockCB);
	printf("No. of Nodes Control Blocks Err:               %d\n", statp->numOfBadNode);

	return (NFX_OK);
}

