/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfs-cache.c
 *
 * RCS:  $Id: nfs-cache.c,v 1.2 2002/11/16 01:29:12 shaun Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: Contain NFS attribute cache  related routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 10/10/02
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX
    
#include <nfx-incl.h>
#include <exports/evm-api.h>
#include <exports/fs-api.h>
#include <exports/nfs-api.h>
#include <exports/queue-api.h>
#include <exports/malloc-api.h>
#include <exports/timer-api.h>
#include "nfs-cache.h"


#define	NFS_CACHE_AGE_TIME			1000	/* 1 sec aging timer period */
#define NFS_CACHE_MAX_AGE			60		/* age out after 60s inactive */
#define NFS_CACHE_MAX_LOOP_CNT		100		/* 100 entry per timer loop */

#define NFS_CACHE_MAX_SYMLNK_ENTRY	(512)	/* max. # of symlnk cache */
#define NFS_CACHE_MAX_SYMLNK_HASH	(2 * NFS_CACHE_MAX_SYMLNK_ENTRY)

#define NFS_CACHE_MAX_FSINFO_ENTRY	(512)	/* max. # of fsinfo cache */
#define NFS_CACHE_MAX_FSINFO_HASH	(2 * NFS_CACHE_MAX_FSINFO_ENTRY)

#define NFS_CACHE_FLUSH_TO_PERIOD	10		/* 10 ms flush timer */
#define NFS_CACHE_F_INUSE			0x8000

#define NFS_CACHE_STATE_ENABLED		1
#define NFS_CACHE_STATE_DISABLED	0

/*-------------------------------
 * Attribute Cache data structure
 *-------------------------------
 */
typedef struct _nfs_attrCache_t		nfs_attrCache_t;
struct _nfs_attrCache_t
{
    fs_Handle                   	fhdl;		/* file handle (hash key) */
    fs_Attr                     	fattr;		/* file attribute */
    int32                       	hashIdx;	/* hash table index */
    ushort16                       	flags;		/* flags */
    ushort16                       	hitCnt;		/* # of hit */
    ushort16                       	ageCnt;		/* aging count, del when 0 */
    ushort16                       	nameLen;	/* */
    DLL_ENTRY(nfs_attrCache_t)  	cacheList;	/* link to cache list */
    DLL_ENTRY(nfs_attrCache_t)  	hashList;	/* link to hash list */
};

/*-------------------------------
 * Symlink Cache data structure
 *-------------------------------
 */
typedef struct _nfs_symlnkCache_t		nfs_symlnkCache_t;
struct _nfs_symlnkCache_t
{
    fs_Handle                   	fhdl;		/* file handle (hash key) */
    fs_Attr                     	fattr;		/* file attribute */
    int32                       	hashIdx;	/* hash table index */
    ushort16                       	flags;		/* flags */
    ushort16                       	hitCnt;		/* # of hit */
    ushort16                       	ageCnt;		/* aging count, del when 0 */
    ushort16                       	nameLen;	/* len of symlink name */
	char							fname[32];
    DLL_ENTRY(nfs_symlnkCache_t)  	cacheList;	/* link to cache list */
    DLL_ENTRY(nfs_symlnkCache_t)  	hashList;	/* link to free list */
};

/*-------------------------------
 * FS Info Cache data structure
 *-------------------------------
 */
typedef struct _nfs_fsinfoCache_t		nfs_fsinfoCache_t;
struct _nfs_fsinfoCache_t
{
    fs_Handle                   	fhdl;		/* file handle (hash key) */
    fs_Attr                     	fattr;		/* file attribute */
    fs_Fsinfo                     	fsinfo;		/* file attribute */
    int32                       	hashIdx;	/* hash table index */
    ushort16                       	flags;		/* flags */
    ushort16                       	ageCnt;		/* aging count, del when 0 */
    DLL_ENTRY(nfs_fsinfoCache_t)  	cacheList;	/* link to cache list */
    DLL_ENTRY(nfs_fsinfoCache_t)  	hashList;	/* link to free list */
};


/*-------------------------------
 * Attribute Cache List to 
 * maintain free, used & hash lists.
 *-------------------------------
 */
typedef struct  _nfs_attrCacheList_t
{
    DLL_HDR(nfs_attrCache_t)    list;
} nfs_attrCacheList_t;

/*-------------------------------
 * Symbolic Link Cache List to 
 * maintain free, used & hash lists.
 *-------------------------------
 */
typedef struct  _nfs_symlnkCacheList_t
{
    DLL_HDR(nfs_symlnkCache_t)    list;
} nfs_symlnkCacheList_t;

/*-------------------------------
 * FS Info Cache List to 
 * maintain free, used & hash lists.
 *-------------------------------
 */
typedef struct  _nfs_fsinfoCacheList_t
{
    DLL_HDR(nfs_fsinfoCache_t)    list;
} nfs_fsinfoCacheList_t;


/*-------------------------------
 * NFS Cache statistics.
 *-------------------------------
 */
typedef struct
{
	uint32		findCnt;			/* # of attr cache lookup */
	uint32		addCnt;				/* # of attr cache add */
	uint32		delCnt;				/* # of attr cache delete */
	uint32		hitCnt;				/* # of attr cache hit */
	uint32		updateCnt;			/* # of cache update */
	uint32		rsrcErr;			/* # of no resource err */
	uint32		hashHWM;			/* high water mark for hash */
	uint32		hashIdx;			/* hash idx assoc. with HWM */
} nfs_cacheStats_t;

/*-------------------------------
 * NFS Cache config.
 *-------------------------------
 */
typedef struct
{
	int32					state;				/* state of cache */
	int32					attrCacheCount;		/* max # attr cache entry*/
	int32					attrHashCount;		/* max # attr hash entry */
	int32					attrCacheSize;		/* attr cache table size */
	int32					attrHashSize;		/* hash table size */

	int32					symlnkCacheSize;	/* symlnk cache table size */
	int32					symlnkHashSize;		/* hash table size */

	int32					fsinfoCacheSize;	/* fsinfo cache table size */
	int32					fsinfoHashSize;		/* hash table size */

	int32					ageTmrHdl;			/* cache aging timer handle */

	nfs_attrCache_t			*pCacheSaved;		/* start of next timer loop */
	nfs_attrCache_t			*pAttrCacheTbl;		/* attr. cache table */
	nfs_attrCacheList_t		*pAttrHashTbl;		/* attr. cache hash table */
	nfs_attrCacheList_t		attrCacheList;		/* DLL of attr cache */

	nfs_symlnkCache_t		*pSymlnkCacheTbl;	/* symlnk cache table */
	nfs_symlnkCacheList_t	*pSymlnkHashTbl;	/* symlnk cache hash table */
	nfs_symlnkCacheList_t	symlnkCacheList;	/* DLL of symlnk cache */

	nfs_fsinfoCache_t		*pFsinfoCacheTbl;	/* symlnk cache table */
	nfs_fsinfoCacheList_t	*pFsinfoHashTbl;	/* symlnk cache hash table */
	nfs_fsinfoCacheList_t	fsinfoCacheList;	/* DLL of symlnk cache */

	nfs_cacheStats_t		attrStats;			/* attr cache statistics */
	nfs_cacheStats_t		symlnkStats;		/* symlnk cache statistics */
	nfs_cacheStats_t		fsinfoStats;		/* fsinfo cache statistics */

	uint32					getAttrCnt;			/* # of get attr cache reply */
	uint32					accessCnt;			/* # of access cache reply */
	uint32					readlnkCnt;			/* # of readlnk cache reply */
	uint32					fsinfoCnt;			/* # of fsinfo cache reply */

	int32					flushTmrHdl;		/* Flush Timer handle */
	uint32					flushAttrIdx;		/* attr hash idx to flush */
	uint32					flushSymlnkIdx;		/* sym link hash idx to flush*/
	uint32					flushFsinfoIdx;		/* fsinfo hash idx to flush*/

}	nfs_cacheCfg_t;

nfs_cacheCfg_t	*pCacheCfg;

extern	uint64 get_memorysize();

/*-----------------------------------------------------------------
 * Name:        nfs_cacheShowStats
 *
 * Description: Shows cache related stats & config.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void
nfs_cacheShowStats( void )
{
	nfs_cacheStats_t	*pStats;

	printf("NFS Attr. Cache Statistics:\n");
	printf("------------------------------------------\n");
	printf("\tAttr Cache   Size: %10d  Count: %10d\n", 
			pCacheCfg->attrCacheSize, pCacheCfg->attrCacheCount);
	printf("\tAttr Hash    Size: %10d  Count: %10d\n", 
			pCacheCfg->attrHashSize, pCacheCfg->attrHashCount);

	pStats = &pCacheCfg->attrStats;

	printf("\tAttr     Find: %10d\n", pStats->findCnt);
	printf("\tAttr      Add: %10d\n", pStats->addCnt);
	printf("\tAttr      Del: %10d\n", pStats->delCnt);
	printf("\tAttr      Hit: %10d\n", pStats->hitCnt);
	printf("\tAttr   Update: %10d\n", pStats->updateCnt);
	printf("\tAttr Rsrc Err: %10d\n", pStats->rsrcErr);
	printf("\tAttr Hash HWM: %10d\n", pStats->hashHWM);
	printf("\tAttr Hash Idx: %10d\n", pStats->hashIdx);

	pStats = &pCacheCfg->symlnkStats;

	printf("\tSymlnk     Find: %10d\n", pStats->findCnt);
	printf("\tSymlnk      Add: %10d\n", pStats->addCnt);
	printf("\tSymlnk      Del: %10d\n", pStats->delCnt);
	printf("\tSymlnk      Hit: %10d\n", pStats->hitCnt);
	printf("\tSymlnk   Update: %10d\n", pStats->updateCnt);
	printf("\tSymlnk Rsrc Err: %10d\n", pStats->rsrcErr);
	printf("\tSymlnk Hash HWM: %10d\n", pStats->hashHWM);
	printf("\tSymlnk Hash Idx: %10d\n", pStats->hashIdx);

	pStats = &pCacheCfg->fsinfoStats;

	printf("\tFsinfo     Find: %10d\n", pStats->findCnt);
	printf("\tFsinfo      Add: %10d\n", pStats->addCnt);
	printf("\tFsinfo      Del: %10d\n", pStats->delCnt);
	printf("\tFsinfo      Hit: %10d\n", pStats->hitCnt);
	printf("\tFsinfo   Update: %10d\n", pStats->updateCnt);
	printf("\tFsinfo Rsrc Err: %10d\n", pStats->rsrcErr);
	printf("\tFsinfo Hash HWM: %10d\n", pStats->hashHWM);
	printf("\tFsinfo Hash Idx: %10d\n", pStats->hashIdx);

	printf("\tGetAttr  Rsp: %10d\n", pCacheCfg->getAttrCnt);
	printf("\tAccess   Rsp: %10d\n", pCacheCfg->accessCnt);
	printf("\tReadln   Rsp: %10d\n", pCacheCfg->readlnkCnt);
	printf("\tFsinfo   Rsp: %10d\n", pCacheCfg->fsinfoCnt);
//	printf("\tAgeing   HWM: %10d\n", pStats->ageHWM);

//	nfs_cacheFlush( 0 );
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheInit
 *
 * Description: Allocate & initialize nfs cache data structures.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void 
nfs_cacheInit( void )
{
	int32					i;
	uint64					memSize;
	nfs_attrCache_t			*pAttrCache;
	nfs_attrCacheList_t		*pAttrHash;
	nfs_symlnkCache_t		*pSymlnkCache;
	nfs_symlnkCacheList_t	*pSymlnkHash;
	nfs_fsinfoCache_t		*pFsinfoCache;
	nfs_fsinfoCacheList_t	*pFsinfoHash;

    /*-------------------------------
     * Allocate cache cfg.
     *-------------------------------
     */
	pCacheCfg = eee_ramAlloc(sizeof(nfs_cacheCfg_t));
	if (pCacheCfg == NULL)
		panic("%s: pCacheCfg alloc failed!\n", __FUNCTION__);

    /*-------------------------------
     * Allocate attr. cache entries.
	 * 1M entries if 4G mem. else 
	 * 1K entries.
     *-------------------------------
     */
	memSize = 4;
	memSize *= 1024;
	memSize *= 1024;
	memSize *= 1024;
	pCacheCfg->attrCacheCount = (get_memorysize() >= memSize) ?
								(1024*1024) :
								(1*1024);
		
	pCacheCfg->attrCacheSize = sizeof(nfs_attrCache_t) * 
						   		pCacheCfg->attrCacheCount;
	pCacheCfg->pAttrCacheTbl = eee_ramAlloc(pCacheCfg->attrCacheSize);
	if (pCacheCfg->pAttrCacheTbl == NULL)
		panic("%s: AttrCacheTbl alloc failed!\n", __FUNCTION__);

    /*-------------------------------
     * Allocate attr. hash table.
	 * 2 x # of attr cache entries.
     *-------------------------------
     */
	pCacheCfg->attrHashCount= (pCacheCfg->attrCacheCount * 2);
	pCacheCfg->attrHashSize = sizeof(nfs_attrCacheList_t) *
						  	  pCacheCfg->attrHashCount;
	pCacheCfg->pAttrHashTbl = eee_ramAlloc(pCacheCfg->attrHashSize);
	if (pCacheCfg->pAttrHashTbl == NULL)
		panic("%s: AttrHashTbl alloc failed!\n", __FUNCTION__);

    /*-------------------------------
     * Allocate symlnk cache entries.
     *-------------------------------
     */
	pCacheCfg->symlnkCacheSize = sizeof(nfs_symlnkCache_t) * 
						   		NFS_CACHE_MAX_SYMLNK_ENTRY;
	pCacheCfg->pSymlnkCacheTbl = eee_ramAlloc(pCacheCfg->symlnkCacheSize);
	if (pCacheCfg->pSymlnkCacheTbl == NULL)
		panic("%s: SymlnkCacheTbl alloc failed!\n", __FUNCTION__);

    /*-------------------------------
     * Allocate symlnk hash table.
     *-------------------------------
     */
	pCacheCfg->symlnkHashSize = sizeof(nfs_symlnkCacheList_t) *
						  		NFS_CACHE_MAX_SYMLNK_HASH;
	pCacheCfg->pSymlnkHashTbl = eee_ramAlloc(pCacheCfg->symlnkHashSize);
	if (pCacheCfg->pSymlnkHashTbl == NULL)
		panic("%s: SymlnkHashTbl alloc failed!\n", __FUNCTION__);

    /*-------------------------------
     * Allocate fsinfo cache entries.
     *-------------------------------
     */
	pCacheCfg->fsinfoCacheSize = sizeof(nfs_fsinfoCache_t) * 
						   		NFS_CACHE_MAX_FSINFO_ENTRY;
	pCacheCfg->pFsinfoCacheTbl = eee_ramAlloc(pCacheCfg->fsinfoCacheSize);
	if (pCacheCfg->pFsinfoCacheTbl == NULL)
		panic("%s: FsinfoCacheTbl alloc failed!\n", __FUNCTION__);

    /*-------------------------------
     * Allocate fsinfo hash table.
     *-------------------------------
     */
	pCacheCfg->fsinfoHashSize = sizeof(nfs_fsinfoCacheList_t) *
						  		NFS_CACHE_MAX_FSINFO_HASH;
	pCacheCfg->pFsinfoHashTbl = eee_ramAlloc(pCacheCfg->fsinfoHashSize);
	if (pCacheCfg->pFsinfoHashTbl == NULL)
		panic("%s: FsinfoHashTbl alloc failed!\n", __FUNCTION__);

    /*-------------------------------
     * Init attr. & symlnk cache lists.
     *-------------------------------
     */
	DLL_INIT(&pCacheCfg->attrCacheList.list);
	DLL_INIT(&pCacheCfg->symlnkCacheList.list);
	DLL_INIT(&pCacheCfg->fsinfoCacheList.list);

    /*-------------------------------
     * Insert attr. cache entries 
	 * into cache list.
     *-------------------------------
     */
	pAttrCache = pCacheCfg->pAttrCacheTbl;
	for (i=0; i < pCacheCfg->attrCacheCount; i++)
	{
		memset(pAttrCache, 0, sizeof(nfs_attrCache_t));

		DLL_INSERT_TAIL(&pCacheCfg->attrCacheList.list, pAttrCache, cacheList);

		pAttrCache++;
	}

    /*-------------------------------
     * Init attr. hash table.
     *-------------------------------
     */
	pAttrHash = pCacheCfg->pAttrHashTbl;
	for (i=0; i < pCacheCfg->attrHashCount; i++)
	{
		DLL_INIT(&pAttrHash->list);
		pAttrHash++;
	}

    /*-------------------------------
     * Insert symlnk. cache entries 
	 * into cache list.
     *-------------------------------
     */
	pSymlnkCache = pCacheCfg->pSymlnkCacheTbl;
	for (i=0; i < NFS_CACHE_MAX_SYMLNK_ENTRY; i++)
	{
		memset(pSymlnkCache, 0, sizeof(nfs_symlnkCache_t));

		DLL_INSERT_TAIL(&pCacheCfg->symlnkCacheList.list, 
						pSymlnkCache, 
						cacheList);

		pSymlnkCache++;
	}

    /*-------------------------------
     * Init symlnk. hash table.
     *-------------------------------
     */
	pSymlnkHash = pCacheCfg->pSymlnkHashTbl;
	for (i=0; i < NFS_CACHE_MAX_SYMLNK_HASH; i++)
	{
		DLL_INIT(&pSymlnkHash->list);
		pSymlnkHash++;
	}

    /*-------------------------------
     * Insert fsinfo. cache entries 
	 * into cache list.
     *-------------------------------
     */
	pFsinfoCache = pCacheCfg->pFsinfoCacheTbl;
	for (i=0; i < NFS_CACHE_MAX_FSINFO_ENTRY; i++)
	{
		memset(pFsinfoCache, 0, sizeof(nfs_fsinfoCache_t));

		DLL_INSERT_TAIL(&pCacheCfg->fsinfoCacheList.list, 
						pFsinfoCache, 
						cacheList);

		pFsinfoCache++;
	}

    /*-------------------------------
     * Init fsinfo. hash table.
     *-------------------------------
     */
	pFsinfoHash = pCacheCfg->pFsinfoHashTbl;
	for (i=0; i < NFS_CACHE_MAX_FSINFO_HASH; i++)
	{
		DLL_INIT(&pFsinfoHash->list);
		pFsinfoHash++;
	}

    /*-------------------------------
     * Create Timer for Flushing Cache
	 * entries. Used when we receive
	 * FS_REQ_CACHE_INV msg from FP.
     *-------------------------------
     */
    pCacheCfg->flushTmrHdl = eee_timerCreate( NFS_CACHE_FLUSH_TO_PERIOD,
                                              EEE_TIMER_REPEAT);
	if (pCacheCfg->flushTmrHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
		panic("%s: Flush timer create failed!\n", __FUNCTION__);

#ifdef NOTUSED
    /*-------------------------------
     * Create & start aging timer
     *-------------------------------
     */
	pCacheCfg->ageTmrHdl = eee_timerCreate(NFS_CACHE_AGE_TIME,
											EEE_TIMER_REPEAT);
	if (pCacheCfg->ageTmrHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
		panic("%s: Age timer create failed!\n", __FUNCTION__);

	eee_timerStart( pCacheCfg->ageTmrHdl, nfs_cacheAgeTmrHandler, 0, 0, 0, 0);
#endif

    pCacheCfg->state = NFS_CACHE_STATE_ENABLED;
}


/*-----------------------------------------------------------------
 * Name:        nfs_cacheUpdateAttr
 *
 * Description: Update attribute cache.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void 
nfs_cacheUpdateAttr( fs_Handle *fhdlp, fs_Attr *attrp)
{
	nfs_attrCache_t			*pCache;
	int32					idx;

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return;

	pCache = nfs_cacheFindAttr(fhdlp, &idx);

	if (pCache)
	{
		pCacheCfg->attrStats.updateCnt++;

		pCache->fattr = *attrp;
	}
	else
	{
		nfs_cacheAddAttr(fhdlp, attrp, idx);
	}
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheFindAttr
 *
 * Description: Find attribute cache on a given file handle.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void	*
nfs_cacheFindAttr(fs_Handle *fhdlp, int32 *idx)
{
	uint32					cnt;
	uint32					hashIdx;
	nfs_attrCache_t			*pCache;
	nfs_attrCacheList_t		*pHash;
	nfs_cacheStats_t		*pStats;

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return NULL;

	pStats = &pCacheCfg->attrStats;

	pStats->findCnt++;

    /*-------------------------------
     * Make hashIdx using volId & inum
	 * from file handle.
     *-------------------------------
     */
	hashIdx  = (uint32)(fhdlp->volId.whole_thing >> 32);
	hashIdx += (uint32)(fhdlp->volId.whole_thing & 0xFFFFFFFF);
	hashIdx += (uint32)(fhdlp->inum >> 32);
	hashIdx += (uint32)(fhdlp->inum & 0xFFFFFFFF);

	hashIdx &= (pCacheCfg->attrHashCount-1);

    /*-------------------------------
     * Save hash index.
     *-------------------------------
     */
	*idx = hashIdx;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: volId=%llx inum=%llx hashIdx=%d\n", 
			__FUNCTION__, 
			fhdlp->volId.whole_thing,
			fhdlp->inum,
			hashIdx);
#endif

    /*-------------------------------
     * Get hash entry from idx.
     *-------------------------------
     */
	pHash = (pCacheCfg->pAttrHashTbl+hashIdx);
	
    /*-------------------------------
     * Check if hash list if empty.
     *-------------------------------
     */
	if (DLL_IS_EMPTY(&pHash->list))
		return NULL;

    /*-------------------------------
     * Go thru hash list to find 
	 * matching file handle.
     *-------------------------------
     */
	cnt = 0;
	DLL_ITERATE(pCache, &pHash->list, hashList)
	{
		if (pCache == NULL)
        {
			return (NULL);
        }

	    /*-------------------------------
	     * Record high water mark for debug.
	     *-------------------------------
	     */
		if (pStats->hashHWM < ++cnt)
		{
			pStats->hashHWM = cnt;
			pStats->hashIdx = hashIdx;
		}
			
	    /*-------------------------------
	     * Match entire file handle.
	     *-------------------------------
	     */
		if ((pCache->fhdl.volId.whole_thing == fhdlp->volId.whole_thing) &&
			(pCache->fhdl.inum              == fhdlp->inum) 			&&
			(pCache->fhdl.gen               == fhdlp->gen)  			&&
			(pCache->fhdl.mkfsTime          == fhdlp->mkfsTime)  		&&
			(pCache->fhdl.mntPnt            == fhdlp->mntPnt))
		{
		    /*-------------------------------
		     * Incr. count & reset age.
		     *-------------------------------
		     */
			pStats->hitCnt++;
			pCache->hitCnt++;
			pCache->ageCnt  = NFS_CACHE_MAX_AGE;

		    /*-------------------------------
		     * Return Cache ptr.
		     *-------------------------------
		     */
			return (pCache);
		}
	}

    return (NULL);
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheAddAttr
 *
 * Description: Add attribute cache of a given file handle.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void *
nfs_cacheAddAttr( fs_Handle *fhdlp, fs_Attr *attrp, int32 hashIdx )
{
	nfs_attrCache_t			*pCache;
	nfs_attrCacheList_t     *pHash;
	nfs_cacheStats_t		*pStats;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: idx=%d\n", __FUNCTION__, hashIdx );
#endif

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return NULL;

	pStats = &pCacheCfg->attrStats;

    /*-------------------------------
	 * Remove old cache entry from 
	 * head of used list.
     *-------------------------------
     */
	DLL_REMOVE_HEAD(&pCacheCfg->attrCacheList.list, pCache, cacheList);
			
	if (pCache == NULL)
	{
		pStats->rsrcErr++;
		return NULL;
	}

	if (pCache->flags & NFS_CACHE_F_INUSE)
	{
		pHash = (pCacheCfg->pAttrHashTbl+pCache->hashIdx);

	    /*-------------------------------
		 * Remove from hash list.
	   	 *-------------------------------
	   	 */
		DLL_REMOVE(&pHash->list, pCache, hashList);
	}

    /*-------------------------------
     * Incr. add count.
     *-------------------------------
     */
	pStats->addCnt++;

    /*-------------------------------
     * Get new hash entry.
     *-------------------------------
     */
	pHash = (pCacheCfg->pAttrHashTbl+hashIdx);

    /*-------------------------------
     * Store file handle & attr in cache.
     *-------------------------------
     */
	pCache->fhdl  = *fhdlp;
	pCache->fattr = *attrp;
	pCache->hashIdx = hashIdx;
	pCache->ageCnt  = NFS_CACHE_MAX_AGE;

	pCache->flags = NFS_CACHE_F_INUSE;

    /*-------------------------------
     * Insert cache into hash lists.
     *-------------------------------
     */
	DLL_INSERT_TAIL(&pHash->list, pCache, hashList );

    /*-------------------------------
     * Insert cache into list.
	 * Used by aging timer to remove
	 * aged out cache entries.
     *-------------------------------
     */
	DLL_INSERT_TAIL(&pCacheCfg->attrCacheList.list, pCache, cacheList);

	return ((void *)pCache);
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheFindSymlnk
 *
 * Description: Find Symlnk cache on a given file handle.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void	*
nfs_cacheFindSymlnk(fs_Handle *fhdlp, int32 *idx)
{
	uint32					cnt;
	uint32					hashIdx;
	nfs_symlnkCache_t		*pCache;
	nfs_symlnkCacheList_t	*pHash;
	nfs_cacheStats_t		*pStats;

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return NULL;

	pStats = &pCacheCfg->symlnkStats;


	pStats->findCnt++;

    /*-------------------------------
     * Make hashIdx using volId & inum
	 * from file handle.
     *-------------------------------
     */
	hashIdx  = (uint32)(fhdlp->volId.whole_thing >> 32);
	hashIdx += (uint32)(fhdlp->volId.whole_thing & 0xFFFFFFFF);
	hashIdx += (uint32)(fhdlp->inum >> 32);
	hashIdx += (uint32)(fhdlp->inum & 0xFFFFFFFF);

	hashIdx &= (NFS_CACHE_MAX_SYMLNK_HASH-1);

    /*-------------------------------
     * Save hash index.
     *-------------------------------
     */
	*idx = hashIdx;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: volId=%llx inum=%llx hashIdx=%d\n", 
			__FUNCTION__, 
			fhdlp->volId.whole_thing,
			fhdlp->inum,
			hashIdx);
#endif

    /*-------------------------------
     * Get hash entry from idx.
     *-------------------------------
     */
	pHash = (pCacheCfg->pSymlnkHashTbl+hashIdx);
	
    /*-------------------------------
     * Check if hash list if empty.
     *-------------------------------
     */
	if (DLL_IS_EMPTY(&pHash->list))
		return NULL;

    /*-------------------------------
     * Go thru hash list to find 
	 * matching file handle.
     *-------------------------------
     */
	cnt = 0;
	DLL_ITERATE(pCache, &pHash->list, hashList)
	{
		if (pCache == NULL)
			return (NULL);

	    /*-------------------------------
	     * Record high water mark for debug.
	     *-------------------------------
	     */
		if (pStats->hashHWM < ++cnt)
		{
			pStats->hashHWM = cnt;
			pStats->hashIdx = hashIdx;
		}
			
	    /*-------------------------------
	     * Match entire file handle.
	     *-------------------------------
	     */
		if ((pCache->fhdl.volId.whole_thing == fhdlp->volId.whole_thing) &&
			(pCache->fhdl.inum              == fhdlp->inum) 			&&
			(pCache->fhdl.gen               == fhdlp->gen)  			&&
			(pCache->fhdl.mkfsTime          == fhdlp->mkfsTime)  		&&
			(pCache->fhdl.mntPnt            == fhdlp->mntPnt))
		{
		    /*-------------------------------
		     * Incr. count & reset age.
		     *-------------------------------
		     */
			pStats->hitCnt++;
			pCache->hitCnt++;
			pCache->ageCnt  = NFS_CACHE_MAX_AGE;

		    /*-------------------------------
		     * Return Fs Attr ptr.
		     *-------------------------------
		     */
			return (pCache);
		}
	}

	return (NULL);
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheAddSymlnk
 *
 * Description: Add symlnk cache of a given file handle.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void *
nfs_cacheAddSymlnk( fs_Handle 	*fhdlp, 
					fs_Attr 	*attrp, 
					int32 		hashIdx,
					char 		*namep,
					uint32		nameLen )
{
	nfs_symlnkCache_t		*pCache;
	nfs_symlnkCacheList_t   *pHash;
	nfs_cacheStats_t		*pStats;

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return NULL;

	pStats = &pCacheCfg->symlnkStats;


#ifdef NFS_DBG_PRINTF_ON
	printf("%s: idx=%d\n", __FUNCTION__, hashIdx );
#endif


    /*-------------------------------
	 * Remove old cache entry from 
	 * head of list.
     *-------------------------------
     */
	DLL_REMOVE_HEAD(&pCacheCfg->symlnkCacheList.list, pCache, cacheList);
			
	if (pCache == NULL)
	{
		pStats->rsrcErr++;
		return NULL;
	}

	if (pCache->flags & NFS_CACHE_F_INUSE)
	{
		pHash = (pCacheCfg->pSymlnkHashTbl+pCache->hashIdx);

	    /*-------------------------------
		 * Remove from hash list.
	     *-------------------------------
	     */
		DLL_REMOVE(&pHash->list, pCache, hashList);
	}

    /*-------------------------------
     * Incr. add count.
     *-------------------------------
     */
	pStats->addCnt++;

    /*-------------------------------
     * Get new hash entry.
     *-------------------------------
     */
	pHash = (pCacheCfg->pSymlnkHashTbl+hashIdx);

    /*-------------------------------
     * Store file handle & attr in cache.
     *-------------------------------
     */
	pCache->fhdl  = *fhdlp;
	pCache->fattr = *attrp;
	pCache->hashIdx = hashIdx;
	pCache->ageCnt  = NFS_CACHE_MAX_AGE;

	pCache->flags = NFS_CACHE_F_INUSE;

	pCache->nameLen = nameLen;

	memcpy(pCache->fname, namep, nameLen);

    /*-------------------------------
     * Insert cache into hash lists.
     *-------------------------------
     */
	DLL_INSERT_TAIL(&pHash->list, pCache, hashList );

    /*-------------------------------
     * Insert cache into tail of list.
     *-------------------------------
     */
	DLL_INSERT_TAIL(&pCacheCfg->symlnkCacheList.list, pCache, cacheList);
	
	return ((void *)pCache);
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheUpdateSymlnk
 *
 * Description: Update Symlnk cache
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void 
nfs_cacheUpdateSymlnk( void *cachep, fs_Attr *attrp)
{
	nfs_symlnkCache_t			*pCache;

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: cachep=%p\n", __FUNCTION__, cachep );
#endif
	
	pCacheCfg->symlnkStats.updateCnt++;

//	pCache = (nfs_symlnkCache_t *)cachep;
	pCache = cachep;
	pCache->fattr = *attrp;
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheFindFsinfo
 *
 * Description: Find Fsinfo cache on a given file handle.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void	*
nfs_cacheFindFsinfo(fs_Handle *fhdlp, int32 *idx)
{
	uint32					cnt;
	uint32					hashIdx;
	nfs_fsinfoCache_t		*pCache;
	nfs_fsinfoCacheList_t	*pHash;
	nfs_cacheStats_t		*pStats;

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return NULL;

	pStats = &pCacheCfg->fsinfoStats;


	pStats->findCnt++;

    /*-------------------------------
     * Make hashIdx using volId & inum
	 * from file handle.
     *-------------------------------
     */
	hashIdx  = (uint32)(fhdlp->volId.whole_thing >> 32);
	hashIdx += (uint32)(fhdlp->volId.whole_thing & 0xFFFFFFFF);
	hashIdx += (uint32)(fhdlp->inum >> 32);
	hashIdx += (uint32)(fhdlp->inum & 0xFFFFFFFF);

	hashIdx &= (NFS_CACHE_MAX_FSINFO_HASH-1);

    /*-------------------------------
     * Save hash index.
     *-------------------------------
     */
	*idx = hashIdx;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: volId=%llx inum=%llx hashIdx=%d\n", 
			__FUNCTION__, 
			fhdlp->volId.whole_thing,
			fhdlp->inum,
			hashIdx);
#endif

    /*-------------------------------
     * Get hash entry from idx.
     *-------------------------------
     */
	pHash = (pCacheCfg->pFsinfoHashTbl+hashIdx);
	
    /*-------------------------------
     * Check if hash list if empty.
     *-------------------------------
     */
	if (DLL_IS_EMPTY(&pHash->list))
		return NULL;

    /*-------------------------------
     * Go thru hash list to find 
	 * matching file handle.
     *-------------------------------
     */
	cnt = 0;
	DLL_ITERATE(pCache, &pHash->list, hashList)
	{
		if (pCache == NULL)
			return (NULL);

	    /*-------------------------------
	     * Record high water mark for debug.
	     *-------------------------------
	     */
		if (pStats->hashHWM < ++cnt)
		{
			pStats->hashHWM = cnt;
			pStats->hashIdx = hashIdx;
		}
			
	    /*-------------------------------
	     * Match entire file handle.
	     *-------------------------------
	     */
		if ((pCache->fhdl.volId.whole_thing == fhdlp->volId.whole_thing) &&
			(pCache->fhdl.inum              == fhdlp->inum) 			&&
			(pCache->fhdl.gen               == fhdlp->gen)  			&&
			(pCache->fhdl.mkfsTime          == fhdlp->mkfsTime)  		&&
			(pCache->fhdl.mntPnt            == fhdlp->mntPnt))
		{
		    /*-------------------------------
		     * Incr. count & reset age.
		     *-------------------------------
		     */
			pStats->hitCnt++;
			pCache->ageCnt  = NFS_CACHE_MAX_AGE;

		    /*-------------------------------
		     * Return Fs Attr ptr.
		     *-------------------------------
		     */
			return (pCache);
		}
	}

	return (NULL);
}

/*-----------------------------------------------------------------
 * Name:        nfs_cacheAddFsinfo
 *
 * Description: Add Fsinfo cache of a given file handle.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
void *
nfs_cacheAddFsinfo( fs_Handle 	*fhdlp, 
					fs_Attr 	*attrp, 
					fs_Fsinfo 	*infop, 
					int32 		hashIdx)
{
	nfs_fsinfoCache_t		*pCache;
	nfs_fsinfoCacheList_t   *pHash;
	nfs_cacheStats_t		*pStats;

	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return NULL;

	pStats = &pCacheCfg->fsinfoStats;


#ifdef NFS_DBG_PRINTF_ON
	printf("%s: idx=%d\n", __FUNCTION__, hashIdx );
#endif


    /*-------------------------------
	 * Remove old cache entry from 
	 * head of list.
     *-------------------------------
     */
	DLL_REMOVE_HEAD(&pCacheCfg->fsinfoCacheList.list, pCache, cacheList);
			
	if (pCache == NULL)
	{
		pStats->rsrcErr++;
		return NULL;
	}

	if (pCache->flags & NFS_CACHE_F_INUSE)
	{
		pHash = (pCacheCfg->pFsinfoHashTbl+pCache->hashIdx);

	    /*-------------------------------
		 * Remove from hash list.
	     *-------------------------------
	     */
		DLL_REMOVE(&pHash->list, pCache, hashList);
	}

    /*-------------------------------
     * Incr. add count.
     *-------------------------------
     */
	pStats->addCnt++;

    /*-------------------------------
     * Get new hash entry.
     *-------------------------------
     */
	pHash = (pCacheCfg->pFsinfoHashTbl+hashIdx);

    /*-------------------------------
     * Store file handle & attr in cache.
     *-------------------------------
     */
	pCache->fhdl  = *fhdlp;
	pCache->fattr = *attrp;
	pCache->fsinfo = *infop;
	pCache->hashIdx = hashIdx;
	pCache->ageCnt  = NFS_CACHE_MAX_AGE;

	pCache->flags = NFS_CACHE_F_INUSE;

    /*-------------------------------
     * Insert cache into hash lists.
     *-------------------------------
     */
	DLL_INSERT_TAIL(&pHash->list, pCache, hashList );

    /*-------------------------------
     * Insert cache into tail of list.
     *-------------------------------
     */
	DLL_INSERT_TAIL(&pCacheCfg->fsinfoCacheList.list, pCache, cacheList);
	
	return ((void *)pCache);
}


#ifdef NOTUSED
/*-----------------------------------------------------------------
 * Name:        nfs_cacheAgeTmrHandler
 *
 * Description: Aging timer handle to age out inactive attributes.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 10/10/2002
 *-----------------------------------------------------------------
 */
int32
nfs_cacheAgeTmrHandler( int32 	timer_handle, 
						void 	*param1, 
						void 	*param2, 
						void 	*param3, 
						void 	*param4) 
{
	int32					cnt;
	nfs_attrCache_t			*pCache;
	nfs_attrCacheList_t     *pHash;

    /*-------------------------------
     * Go thru used list to check age
	 * of inactive cache entries.
     *-------------------------------
     */
	cnt = 0;
	DLL_ITERATE(pCache, &pCacheCfg->usedCacheList.list, usedList)
	{
		if (pCache == NULL)
			return NFX_OK;

	    /*-------------------------------
	     * Start from previously left off.
	     *-------------------------------
	     */
		if ((cnt == 0) && (pCacheCfg->pCacheSaved != NULL))
		{
			pCache = pCacheCfg->pCacheSaved;
			pCacheCfg->pCacheSaved = NULL;
		}

	    /*-------------------------------
		 * Reached max loop count,
	     * save for next timer loop.
	     *-------------------------------
	     */
		if (cnt > NFS_CACHE_MAX_LOOP_CNT)
		{
			pCacheCfg->pCacheSaved = pCache;
			return NFX_OK;
		}

	    /*-------------------------------
		 * Record aging high water mark.
	     *-------------------------------
	     */
		if (pCacheCfg->stats.ageHWM < ++cnt)
			pCacheCfg->stats.ageHWM = cnt;

		pCache->ageCnt--;

#ifdef NFS_DBG_PRINTF_ON
		printf("%s: pCache=%p hits=%d age=%d idx=%d\n", 
			__FUNCTION__, 
			pCache, 
			pCache->hitCnt, 
			pCache->ageCnt,
			pCache->hashIdx);
#endif

	    /*-------------------------------
		 * Delete aged out cache 
	     *-------------------------------
	     */
		if (pCache->ageCnt <= 0)
		{
			pHash = (pCacheCfg->pAttrHashTbl+pCache->hashIdx);

		    /*-------------------------------
			 * Remove from used list.
		     *-------------------------------
		     */
			DLL_REMOVE(&pCacheCfg->usedCacheList.list, pCache, usedList);
			
		    /*-------------------------------
			 * Remove from hash list.
		     *-------------------------------
		     */
			DLL_REMOVE(&pHash->list, pCache, hashList);

		    /*-------------------------------
			 * Insert back to free list.
		     *-------------------------------
		     */
			DLL_INSERT_TAIL(&pCacheCfg->freeCacheList.list, pCache, freeList);

			pCacheCfg->stats.delCnt++;
		}
		
	}
	
	return NFX_OK;
}
#endif

int32	
nfs_cacheSendGetAttrResp(nfs_request_t *nfsReqp, void *cachep)
{
	nfs_pktCounters_t		*txPktStatp;
	nfp_clientCBPtr_t       clientCBp;
	pkt_desc_t				*pDesc;
	char                    *txMsgp;
	int32                   txMsgLen;
	int32					length;
	nfs_attrCache_t			*pCache;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s Enter\n", __FUNCTION__);
#endif

    /*-------------------------------
	 * Move recently used cache to 
	 * tail of used list.
     *-------------------------------
     */
	pCache = cachep;
	DLL_MOVE_TO_TAIL(&pCacheCfg->attrCacheList.list, pCache, cacheList);

	clientCBp = nfsReqp->clientCBp;

	pDesc = (pkt_desc_t *)nfs_tplAllocPkt(clientCBp, 
											nfsReqp,
											0 );
	if (pDesc == NULL)
	{
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		nfs_msgQueReplyMsg(clientCBp, NFS_TX_CACHE_MSG_TYPE,
							(void *)nfsReqp,0,0);
		return (NFX_ERR);
	}
	
	pCacheCfg->getAttrCnt++;

	txMsgp = (char *)pDesc->pHdr;
	txMsgLen = 0;

	/*
	 *Start with RPC reply header
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, 
								nfsReqp->rpcCallHdrp->rpc_xid,
                                RPC_REPLY_MSG_ACCEPTED, 
								RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	/*
	 *Set Status
	 */
	*(uint32 *)txMsgp = NFS_OK;
	txMsgp   += 4;
	txMsgLen += 4;
        
	txPktStatp = &nfs_globalCfgp->nfs_xmtPktCounters;

	/*      
	 *Store attributes for the object
	 */
	length = nfs_storeV3Attributes( nfsReqp->reqType, 
									NFS_ATTR_AVAIL,
									&pCache->fattr,
									txMsgp); 

	/*
	 *Update message buffer pointer & length
	 *Increment transmit statistic counter
	 */
	txMsgp += length;
	txMsgLen += length;

	txPktStatp->getAttr++;

	/*
	 *Set packet desc hdr size
	 */
	pDesc->hdrSize = txMsgLen;

	if (nfs_msgXmtPkt(pDesc, nfsReqp) != NFX_OK)
	{
		//Error: Set reqState to DONE so that
		//nfs_reqTimeoutHandler() will free nfsReqp
		nfsReqp->pktDescp = (void *)pDesc;
		nfsReqp->reqState = NFS_REQ_DONE;
	}
#ifdef NFS_DBG_PRINTF_ON
	printf("%s Exit\n", __FUNCTION__);
#endif

	return (NFX_OK);
}



void
nfs_cacheGetOwnership( nfs_request_t 	*nfsReqp, 
					   fs_Attr			*pAttr,
					   boolean  		*user,
					   boolean  		*group,
					   boolean  		*root)
{
	rpc_authSysInfo_t	*pAuthSys;
	uint32				uid;
	uint32				gid;
	int32				i;

	pAuthSys = &nfsReqp->rpcCallHdrp->authFlavor.authSys;
	uid      = pAuthSys->userId;
	gid      = pAuthSys->groupId;

	if (uid == 0)
	{
		*user  = TRUE;
		*group = TRUE;
		*root  = TRUE;
		return;
	}
	
	*user  = FALSE;
	*group = FALSE;
	*root  = FALSE;
	
	if (uid == pAttr->uid)
	{
		*user  = TRUE;
		return;
	}

	if (gid == pAttr->gid)
	{
		*group = TRUE;
		return;
	}

	for (i=0; i < pAuthSys->auxGidCount; i++)
	{
		if (pAuthSys->auxGids[i] == gid)
		{
			*group = TRUE;
			return;
		}
	}
}

boolean
nfs_cacheCheckPermission( nfs_request_t 	*nfsReqp, 
						  fs_Attr 			*pAttr,
						  int32 			type )
{
	boolean			isOwner;
	boolean			isGroupOwner;
	boolean			isSuperUser;

	nfs_cacheGetOwnership(  nfsReqp, 
							pAttr, 
							&isOwner, 
							&isGroupOwner, 
							&isSuperUser);

	switch (type)
	{
		case	FS_PERM_LOOKUP:
				if (isSuperUser)
					return TRUE;

				if (isOwner && (pAttr->mode & FS_IMODE_XUSR))
					return TRUE;

				if (isGroupOwner && (pAttr->mode & FS_IMODE_XGRP))
					return TRUE;

				if (!isOwner && !isGroupOwner && (pAttr->mode & FS_IMODE_XOTH))
					return TRUE;

				break;

		case	FS_PERM_READ:
				if (isSuperUser)
					return TRUE;

				if (isOwner && (pAttr->mode & FS_IMODE_RUSR))
					return TRUE;

				if (isGroupOwner && (pAttr->mode & FS_IMODE_RGRP))
					return TRUE;

				if (!isOwner && !isGroupOwner && (pAttr->mode & FS_IMODE_ROTH))
					return TRUE;

				break;

		case	FS_PERM_WRITE:
				if (isSuperUser) 
					return TRUE;

				if (isOwner && (pAttr->mode & FS_IMODE_WUSR)) 
					return TRUE;

				if (isGroupOwner && (pAttr->mode & FS_IMODE_WGRP))
					return TRUE;

				if (!isOwner && !isGroupOwner && pAttr->mode & FS_IMODE_WOTH)
					return TRUE;
				break;

		case	FS_PERM_DIRENT:
				if (isSuperUser) 
					return TRUE;

				if (isOwner && 
					((pAttr->mode & (FS_IMODE_WUSR|FS_IMODE_XUSR)) ==
									(FS_IMODE_WUSR|FS_IMODE_XUSR)))
					return TRUE;

				if (isGroupOwner && 
					((pAttr->mode & (FS_IMODE_WGRP|FS_IMODE_XGRP)) ==
									(FS_IMODE_WGRP|FS_IMODE_XGRP)))
					return TRUE;

				if (!isOwner && !isGroupOwner &&
					((pAttr->mode & (FS_IMODE_WOTH|FS_IMODE_XOTH)) ==
									(FS_IMODE_WOTH|FS_IMODE_XOTH)))
					return TRUE;

				break;

		case	FS_PERM_READDIR:
				if (isSuperUser) 
					return TRUE;

				if (isOwner &&
					((pAttr->mode & (FS_IMODE_RUSR|FS_IMODE_XUSR)) ==
									(FS_IMODE_RUSR|FS_IMODE_XUSR)))
					return TRUE;

				if (isGroupOwner &&
					((pAttr->mode & (FS_IMODE_RGRP|FS_IMODE_XGRP)) ==
									(FS_IMODE_RGRP|FS_IMODE_XGRP)))
					return TRUE;

				if (!isOwner && !isGroupOwner &&
					((pAttr->mode & (FS_IMODE_ROTH|FS_IMODE_XOTH)) ==
									(FS_IMODE_ROTH|FS_IMODE_XOTH)))
					return TRUE;

				break;

		case	FS_PERM_EXEC:
				if (isSuperUser) 
					return TRUE;

				if (isOwner && (pAttr->mode & FS_IMODE_XUSR))
					return TRUE;

				if (isGroupOwner && (pAttr->mode & FS_IMODE_XGRP))
					return TRUE;

				if (!isOwner && !isGroupOwner && (pAttr->mode & FS_IMODE_XOTH))
					return TRUE;

				break;

		default:
				printf("%s: unknow permission type\n", __FUNCTION__);
				break;
	}

	return FALSE;
}


uint32
nfs_cacheGetAccessBits( nfs_request_t *nfsReqp, fs_Attr *pAttr, uint32 reqBits)
{
	uint32			type;
	uint32			accessBits;

	accessBits = 0;

	if (reqBits & FS_ACCESS_READ)
	{
		type = (pAttr->type == FS_FTYPE_DIR) ? 
				FS_PERM_READDIR : 
				FS_PERM_READ;

        if (nfs_cacheCheckPermission(nfsReqp, pAttr, type))
        {
            accessBits |= FS_ACCESS_READ;
        }
	}

    if (reqBits & FS_ACCESS_LOOKUP)
    {
		if (pAttr->type == FS_FTYPE_DIR)
        {
            type = FS_PERM_LOOKUP;
        	if (nfs_cacheCheckPermission(nfsReqp, pAttr, type))
            {
                accessBits |= FS_ACCESS_LOOKUP;
            }
        }
    }

    if (reqBits & FS_ACCESS_MODIFY)
    {
		type = (pAttr->type == FS_FTYPE_DIR) ? 
            	FS_PERM_DIRENT: 
				FS_PERM_WRITE;

       	if (nfs_cacheCheckPermission(nfsReqp, pAttr, type))
        {
            accessBits |= FS_ACCESS_MODIFY; 
        }
    }

    if (reqBits & FS_ACCESS_EXTEND)
    {
		type = (pAttr->type == FS_FTYPE_DIR) ? 
            	FS_PERM_DIRENT: 
				FS_PERM_WRITE;

       	if (nfs_cacheCheckPermission(nfsReqp, pAttr, type))
        {
            accessBits |= FS_ACCESS_EXTEND;
        }
    }

    if (reqBits & FS_ACCESS_DELETE)
    {
		if (pAttr->type == FS_FTYPE_DIR)
        {
            type = FS_PERM_DIRENT;
       		if (nfs_cacheCheckPermission(nfsReqp, pAttr, type))
            {
                accessBits |= FS_ACCESS_DELETE;
            }
        }
    }

    if (reqBits & FS_ACCESS_EXECUTE)
    {
		if (pAttr->type != FS_FTYPE_DIR)
        {
            type = FS_PERM_EXEC;
       		if (nfs_cacheCheckPermission(nfsReqp, pAttr, type))
            {
                accessBits |= FS_ACCESS_EXECUTE;
            }
        }
    }

//	printf("%s: accessBits=%x\n",__FUNCTION__, accessBits);

	return accessBits;

}

int32	
nfs_cacheSendAccessResp(nfs_request_t	*nfsReqp, 
						void 			*cachep, 
						uint32			reqBits)
{
	nfs_pktCounters_t		*txPktStatp;
	nfp_clientCBPtr_t       clientCBp;
	pkt_desc_t				*pDesc;
	char                    *txMsgp;
	int32                   txMsgLen;
	int32					length;
	nfs_attrCache_t			*pCache;
	uint32					accessBits;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s Enter\n", __FUNCTION__);
#endif

    /*-------------------------------
	 * Move recently used cache to 
	 * tail of used list.
     *-------------------------------
     */
	pCache = cachep;
	DLL_MOVE_TO_TAIL(&pCacheCfg->attrCacheList.list, pCache, cacheList);

	clientCBp = nfsReqp->clientCBp;

	pDesc = (pkt_desc_t *)nfs_tplAllocPkt(clientCBp, 
											nfsReqp,
											0 );
	if (pDesc == NULL)
	{
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		nfs_msgQueReplyMsg(clientCBp, NFS_TX_CACHE_MSG_TYPE,
							(void *)nfsReqp,0,0);
		return (NFX_ERR);
	}
	
	pCacheCfg->accessCnt++;

	txMsgp = (char *)pDesc->pHdr;
	txMsgLen = 0;

	/*
	 *Start with RPC reply header
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, 
								nfsReqp->rpcCallHdrp->rpc_xid,
                                RPC_REPLY_MSG_ACCEPTED, 
								RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	/*
	 *Set Status
	 */
	*(uint32 *)txMsgp = NFS_OK;
	txMsgp   += 4;
	txMsgLen += 4;
        
	txPktStatp = &nfs_globalCfgp->nfs_xmtPktCounters;

	/*      
	 *Store attributes for the object
	 */
	length = nfs_storeV3Attributes( nfsReqp->reqType, 
									NFS_ATTR_AVAIL,
									&pCache->fattr,
									txMsgp); 

	/*
	 *Update message buffer pointer & length
	 *Increment transmit statistic counter
	 */
	txMsgp += length;
	txMsgLen += length;

	accessBits = nfs_cacheGetAccessBits(nfsReqp, &pCache->fattr, reqBits);
	*(uint32 *)txMsgp = accessBits;

	txMsgp += sizeof(uint32);
	txMsgLen += sizeof(uint32);

	txPktStatp->access++;

	/*
	 *Set packet desc hdr size
	 */
	pDesc->hdrSize = txMsgLen;

	if (nfs_msgXmtPkt(pDesc, nfsReqp) != NFX_OK)
	{
		//Error: Set reqState to DONE so that
		//nfs_reqTimeoutHandler() will free nfsReqp
		nfsReqp->pktDescp = (void *)pDesc;
		nfsReqp->reqState = NFS_REQ_DONE;
	}
#ifdef NFS_DBG_PRINTF_ON
	printf("%s Exit\n", __FUNCTION__);
#endif

	return (NFX_OK);
}

int32	
nfs_cacheSendReadLinkResp(nfs_request_t	*nfsReqp, 
						void 			*cachep )
{
	nfs_pktCounters_t		*txPktStatp;
	nfp_clientCBPtr_t       clientCBp;
	pkt_desc_t				*pDesc;
	char                    *txMsgp;
	int32                   txMsgLen;
	int32					length;
	nfs_symlnkCache_t		*pCache;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s Enter\n", __FUNCTION__);
#endif

    /*-------------------------------
	 * Move recently used cache to 
	 * tail of list.
     *-------------------------------
     */
	pCache = cachep;
	DLL_MOVE_TO_TAIL(&pCacheCfg->symlnkCacheList.list, pCache, cacheList);


	clientCBp = nfsReqp->clientCBp;

	pDesc = (pkt_desc_t *)nfs_tplAllocPkt(clientCBp, 
											nfsReqp,
											0 );
	if (pDesc == NULL)
	{
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		nfs_msgQueReplyMsg(clientCBp, NFS_TX_CACHE_MSG_TYPE,
							(void *)nfsReqp,0,0);
		return (NFX_ERR);
	}
	
	pCacheCfg->readlnkCnt++;

	txMsgp = (char *)pDesc->pHdr;
	txMsgLen = 0;

	/*
	 *Start with RPC reply header
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, 
								nfsReqp->rpcCallHdrp->rpc_xid,
                                RPC_REPLY_MSG_ACCEPTED, 
								RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	/*
	 *Set Status
	 */
	*(uint32 *)txMsgp = NFS_OK;
	txMsgp   += 4;
	txMsgLen += 4;
        
	txPktStatp = &nfs_globalCfgp->nfs_xmtPktCounters;

	/*      
	 *Store attributes for the object
	 */
	length = nfs_storeV3Attributes( nfsReqp->reqType, 
									NFS_ATTR_AVAIL,
									&pCache->fattr,
									txMsgp); 

	/*
	 *Update message buffer pointer & length
	 *Increment transmit statistic counter
	 */
	txMsgp += length;
	txMsgLen += length;

    /*
     *Get data associated with the symbolic link
     */
    length = nfs_storeString(txMsgp,
                            pCache->fname,
                            pCache->nameLen,
                            FS_PATH_MAX);
    txMsgLen += length;
	txMsgp += length;

	txPktStatp->readLink++;

	/*
	 *Set packet desc hdr size
	 */
	pDesc->hdrSize = txMsgLen;

	if (nfs_msgXmtPkt(pDesc, nfsReqp) != NFX_OK)
	{
		//Error: Set reqState to DONE so that
		//nfs_reqTimeoutHandler() will free nfsReqp
		nfsReqp->pktDescp = (void *)pDesc;
		nfsReqp->reqState = NFS_REQ_DONE;
	}
#ifdef NFS_DBG_PRINTF_ON
	printf("%s Exit\n", __FUNCTION__);
#endif

	return (NFX_OK);
}

int32	
nfs_cacheSendFsinfoResp(nfs_request_t	*nfsReqp, 
						void 			*cachep )
{
	nfs_pktCounters_t		*txPktStatp;
	nfp_clientCBPtr_t       clientCBp;
	pkt_desc_t				*pDesc;
	char                    *txMsgp;
	int32                   txMsgLen;
	int32					length;
	nfs_fsinfoCache_t		*pCache;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s Enter\n", __FUNCTION__);
#endif

    /*-------------------------------
	 * Move recently used cache to 
	 * tail of list.
     *-------------------------------
     */
	pCache = cachep;
	DLL_MOVE_TO_TAIL(&pCacheCfg->fsinfoCacheList.list, pCache, cacheList);


	clientCBp = nfsReqp->clientCBp;

	pDesc = (pkt_desc_t *)nfs_tplAllocPkt(clientCBp, 
											nfsReqp,
											0 );
	if (pDesc == NULL)
	{
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		nfs_msgQueReplyMsg(clientCBp, NFS_TX_CACHE_MSG_TYPE,
							(void *)nfsReqp,0,0);
		return (NFX_ERR);
	}
	
	pCacheCfg->fsinfoCnt++;

	txMsgp = (char *)pDesc->pHdr;
	txMsgLen = 0;

	/*
	 *Start with RPC reply header
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, 
								nfsReqp->rpcCallHdrp->rpc_xid,
                                RPC_REPLY_MSG_ACCEPTED, 
								RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	/*
	 *Set Status
	 */
	*(uint32 *)txMsgp = NFS_OK;
	txMsgp   += 4;
	txMsgLen += 4;
        
	txPktStatp = &nfs_globalCfgp->nfs_xmtPktCounters;

	/*      
	 *Store attributes for the object
	 */
	length = nfs_storeV3Attributes( nfsReqp->reqType, 
									NFS_ATTR_AVAIL,
									&pCache->fattr,
									txMsgp); 

	/*
	 *Update message buffer pointer & length
	 *Increment transmit statistic counter
	 */
	txMsgp += length;
	txMsgLen += length;

	length = nfs_storeFsinfo(txMsgp, &pCache->fsinfo);

    txMsgLen += length;
	txMsgp += length;

	txPktStatp->fsinfo++;

	/*
	 *Set packet desc hdr size
	 */
	pDesc->hdrSize = txMsgLen;

	if (nfs_msgXmtPkt(pDesc, nfsReqp) != NFX_OK)
	{
		//Error: Set reqState to DONE so that
		//nfs_reqTimeoutHandler() will free nfsReqp
		nfsReqp->pktDescp = (void *)pDesc;
		nfsReqp->reqState = NFS_REQ_DONE;
	}
#ifdef NFS_DBG_PRINTF_ON
	printf("%s Exit\n", __FUNCTION__);
#endif

	return (NFX_OK);
}


void
nfs_cacheFlushAttr( void )
{
	uint32					hashIdx;
	int32					loopCnt;
	nfs_attrCache_t			*pCache, *pCacheNext;
	nfs_attrCacheList_t		*pHash;

	hashIdx = pCacheCfg->flushAttrIdx;
	loopCnt = NFS_CACHE_MAX_LOOP_CNT;

	// printf("%s: hashIdx=%d\n", __FUNCTION__, hashIdx);

    /*-------------------------------
     * Make sure loopCnt is in range.
     *-------------------------------
     */
	if ((hashIdx + NFS_CACHE_MAX_LOOP_CNT) > pCacheCfg->attrHashCount)
		loopCnt = pCacheCfg->attrHashCount - hashIdx;

	pCacheCfg->flushAttrIdx += loopCnt;

    /*-------------------------------
     * Get attr hash entry from idx.
     *-------------------------------
     */
	pHash = (pCacheCfg->pAttrHashTbl+hashIdx);

	for (; loopCnt; pHash++, loopCnt--)
	{
		/*-------------------------------
		 * Check if hash list if empty.
		 *-------------------------------
		 */
		if (DLL_IS_EMPTY(&pHash->list))
		{
			continue;
		}

		/*-------------------------------
		 * Go thru hash list & invalidate
		 *-------------------------------
		 */
		DLL_ITERATE_NEXT(pCache, pCacheNext, &pHash->list, hashList)
		{
			if (pCache == NULL)
				break;

			/*-------------------------------
			 * Clear in-use flag.
			 *-------------------------------
			 */
			pCache->flags = 0;

			/*-------------------------------
			 * Remove from hash list.
			 *-------------------------------
			 */
			DLL_REMOVE(&pHash->list, pCache, hashList);
	        pCacheCfg->attrStats.delCnt++;
		}
	}

}

void
nfs_cacheFlushSymlnk( void )
{
	uint32					hashIdx;
	int32					loopCnt;
	nfs_symlnkCache_t		*pCache, *pCacheNext;
	nfs_symlnkCacheList_t	*pHash;

	hashIdx = pCacheCfg->flushSymlnkIdx;
	loopCnt = NFS_CACHE_MAX_LOOP_CNT;

	// printf("%s: hashIdx=%d\n", __FUNCTION__, hashIdx);

    /*-------------------------------
     * Make sure loopCnt is in range.
     *-------------------------------
     */
	if ((hashIdx + NFS_CACHE_MAX_LOOP_CNT) > NFS_CACHE_MAX_SYMLNK_HASH)
		loopCnt = NFS_CACHE_MAX_SYMLNK_HASH - hashIdx;

	pCacheCfg->flushSymlnkIdx += loopCnt;

    /*-------------------------------
     * Get attr hash entry from idx.
     *-------------------------------
     */
	pHash = (pCacheCfg->pSymlnkHashTbl+hashIdx);

	for (; loopCnt; pHash++, loopCnt--)
	{
		/*-------------------------------
		 * Check if hash list if empty.
		 *-------------------------------
		 */
		if (DLL_IS_EMPTY(&pHash->list))
		{
			continue;
		}

		/*-------------------------------
		 * Go thru hash list & invalidate
		 *-------------------------------
		 */
		DLL_ITERATE_NEXT(pCache, pCacheNext, &pHash->list, hashList)
		{
			if (pCache == NULL)
				break;

			/*-------------------------------
			 * Remove from hash list.
			 *-------------------------------
			 */
			DLL_REMOVE(&pHash->list, pCache, hashList);
	        pCacheCfg->symlnkStats.delCnt++;
		}
	}
}

void
nfs_cacheFlushFsinfo( void )
{
	uint32					hashIdx;
	int32					loopCnt;
	nfs_symlnkCache_t		*pCache, *pCacheNext;
	nfs_symlnkCacheList_t	*pHash;

	hashIdx = pCacheCfg->flushFsinfoIdx;
	loopCnt = NFS_CACHE_MAX_LOOP_CNT;

	// printf("%s: hashIdx=%d\n", __FUNCTION__, hashIdx);

    /*-------------------------------
     * Make sure loopCnt is in range.
     *-------------------------------
     */
	if ((hashIdx + NFS_CACHE_MAX_LOOP_CNT) > NFS_CACHE_MAX_FSINFO_HASH)
		loopCnt = NFS_CACHE_MAX_FSINFO_HASH - hashIdx;

	pCacheCfg->flushFsinfoIdx += loopCnt;

    /*-------------------------------
     * Get attr hash entry from idx.
     *-------------------------------
     */
	pHash = (pCacheCfg->pFsinfoHashTbl+hashIdx);

	for (; loopCnt; pHash++, loopCnt--)
	{
		/*-------------------------------
		 * Check if hash list if empty.
		 *-------------------------------
		 */
		if (DLL_IS_EMPTY(&pHash->list))
		{
			continue;
		}

		/*-------------------------------
		 * Go thru hash list & invalidate
		 *-------------------------------
		 */
		DLL_ITERATE_NEXT(pCache, pCacheNext, &pHash->list, hashList)
		{
			if (pCache == NULL)
				break;

			/*-------------------------------
			 * Remove from hash list.
			 *-------------------------------
			 */
			DLL_REMOVE(&pHash->list, pCache, hashList);
	        pCacheCfg->fsinfoStats.delCnt++;
		}
	}
}

int32
nfs_cacheFlushTimerHandler( int32 	tmrHdl, 
							void 	*param1,
							void 	*param2,
							void 	*param3,
							void 	*param4)
{

	if (pCacheCfg->flushAttrIdx < pCacheCfg->attrHashCount)
	{
	    /*-------------------------------
	     * Flush attribute cache
	     *-------------------------------
	     */
		nfs_cacheFlushAttr( );
	}
#ifdef NOTUSED
	else if (pCacheCfg->flushSymlnkIdx < NFS_CACHE_MAX_SYMLNK_HASH)
	{
	    /*-------------------------------
	     * Flush symlink cache
	     *-------------------------------
	     */
		nfs_cacheFlushSymlnk( );
	}
	else if (pCacheCfg->flushFsinfoIdx < NFS_CACHE_MAX_FSINFO_HASH)
	{
	    /*-------------------------------
	     * Flush fsinfo cache
	     *-------------------------------
	     */
		nfs_cacheFlushFsinfo( );
	}
#endif
	else
	{
	    /*-------------------------------
		 * All caches have been flushed,
	     * stop timer for flushing cache.
	     *-------------------------------
	     */
    	eee_timerStop( pCacheCfg->flushTmrHdl);
		pCacheCfg->state = NFS_CACHE_STATE_ENABLED;
	}
	
	return 0;
}

void 
nfs_cacheFlush( int32 mode )
{
	int32 rc;

    /*-------------------------------
     * Check if flush is in progress.
     *-------------------------------
     */
	if (pCacheCfg->state == NFS_CACHE_STATE_DISABLED)
		return;

    /*-------------------------------
     * Reset indices & disable cache.
     *-------------------------------
     */
	pCacheCfg->flushAttrIdx   = 0;
	pCacheCfg->flushSymlnkIdx = 0;
	pCacheCfg->flushFsinfoIdx = 0;
	pCacheCfg->state = NFS_CACHE_STATE_DISABLED;

    /*-------------------------------
     * Start timer for flushing cache.
     *-------------------------------
     */
    rc = eee_timerStart( pCacheCfg->flushTmrHdl,
                        (void *)nfs_cacheFlushTimerHandler, 0, 0, 0, 0);
    if (rc != NFX_OK)
	{
        printf("%s: Failure to start flush timer!\n", __FUNCTION__);
	}

}

#endif /* NFP_TXRX */
