/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfs.c
 *
 * RCS:  $Id: nfs.c,v 1.79 2002/10/30 00:50:06 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *  Contain NFS utility functions.
 *
 * Created by: Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/nfs-api.h>
#include "nfs-hash.h"

#define	NFS_DBG_ON	


#define		NFS_MAX_IP_ADDR_STRLEN	32
/*
#define NFS_USE_SHMEM
*/

#define NFS_CAL_PERCENT(pktCnt, totalPktCnt) ((pktCnt * 100) / totalPktCnt)
/*
 *Extern declaration
 */
extern void nfs_msgRcv( eee_desc_t *eDescp );

extern uint32 nfs_debugCmpPktInd;		//for debugging only
extern uint32 nfs_callmsgFreeAll;		//for  debugging only
extern uint32 nfs_callmsgFreeAll2;		//for debugging only
nfs_globalCfg_t		*nfs_globalCfgp;

extern int32   cntNoDescErr1;		//for dedugging only
extern int32   cntNoDescErr2;		//for debugging only

extern nfp_fileSysCBPtr_t mnt_findFileSysCB(
    nfp_clientCBPtr_t clientCBp,        /*MOUNT Client control block*/
    fs_Handle *fs_handlep);              /*filehandle*/

extern uint32  nfs_tmpDebug1;
extern uint32  nfs_tmpDebug2;

uint32	nfs_bootTimeInSecs = 0;
int32	cacheEnable = 0;            /*Default: cache disabled*/


typedef struct
{
    uint32  volId[2];
    uint32  inum[2];
    uint32  gen;
    uint32  mkfsTime;
    uint32  mntPnt[32];
} nfs_Fhandle_t;

/* 
 * Shaun 09/19/02
 *
 * Maps FS attribute bit map to NFS attr. index.
 * to improve NFS-TX performance.
 */ 
int32	nfs_attrIdxArray[ FS_POST_TPAR_ATTRS + 1];

/*-----------------------------------------------------------------
 * Name: nfs_init 
 *
 * Description:
 *	NFS initialize function.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 	/*
	 	 *NFX_OK or NFX_ERR
	 	 */
nfs_init(void)
{
	int32		rc;

#if defined(NFX_SMP) && defined(NFP_TXRX)
	if (!(nfxMySlotId == 1))
#else
	if (!( (nfxMySlotId == 1) && (nfxMyCPUId == 0) ))
#endif 
	{
		/*
		 *Skip if not TX/RX CPU
		 *TODO:  Use Global define
		 */
		return (NFX_OK);
	}

	nfs_bootTimeInSecs = nfs_utilGetTimeOfDayInSecs();

	/*-------------------------------
	 * Allocate global config block.
	 *-------------------------------
	 */
	nfs_globalCfgp =  (nfs_globalCfg_t  *)eee_ramAlloc(sizeof(nfs_globalCfg_t));
	if (nfs_globalCfgp == NULL)
		panic ("%s: Failure to alloc NFS Global Cfg memory!\n", __FUNCTION__);
	memset(nfs_globalCfgp, 0, sizeof(nfs_globalCfg_t));	

	/*-------------------------------
	 * Allocate hash table for ReqCB look up
	 *-------------------------------
	 */
	nfs_globalCfgp->pReqCbHashTbl = nfs_hashCreateReqCb( NFS_MAX_HASH_REQCB );
	if (nfs_globalCfgp->pReqCbHashTbl == NULL)
		panic ("%s: Failure to alloc ReqCB hash table!\n", __FUNCTION__);

	/*-------------------------------
	 * Allocate list to keep track of 
	 * ReqCBs in use.
	 *-------------------------------
	 */
	nfs_globalCfgp->pReqCbList = (void *)eee_ramAlloc(sizeof(nfp_list_t));
	if (nfs_globalCfgp->pReqCbList == NULL)
		panic ("%s: Failure to alloc ReqCB list!\n", __FUNCTION__);
	memset(nfs_globalCfgp->pReqCbList, 0, sizeof(nfp_list_t));	

	/*-------------------------------
	 * Allocate NFS App Id.
	 *-------------------------------
	 */
	nfs_globalCfgp->eee_nfsAppId = eee_allocateApplicationId("nfs_srvr");
	if (nfs_globalCfgp->eee_nfsAppId == EEE_ERR_INVALID_APP_NAME)
		panic("%s: Failure to alloc Application ID!\n", __FUNCTION__);

	/*-------------------------------
	 * Register nfs rcv function with EEE 
	 *-------------------------------
	 */
	if (eee_registerReceiveFunc(nfs_globalCfgp->eee_nfsAppId, 
								nfs_msgRcv, 
								NULL) != NFX_OK)
		panic("%s: eee_registerReceiveFunc failed!\n", __FUNCTION__);

	/*-------------------------------
	 * Get file system App Id
	 *-------------------------------
	 */
	nfs_globalCfgp->eee_efsAppId = eee_getApplicationId("efs");
	if (nfs_globalCfgp->eee_efsAppId == EEE_ERR_INVALID_APP_NAME)
		panic("%s: Failure to get EFS Application ID!\n", __FUNCTION__);

	/*-------------------------------
	 * Create Timer for freeing ReqCB
	 *-------------------------------
	 */
	nfs_globalCfgp->reqTimeoutHandle = eee_timerCreate( NFS_REQ_TIMEOUT_PERIOD,
														EEE_TIMER_REPEAT);
	if (nfs_globalCfgp->reqTimeoutHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
		panic ("%s: Failure to create requests timer!\n", __FUNCTION__);

	/*-------------------------------
	 * Start Timer for freeing ReqCB
	 *-------------------------------
	 */
	rc = eee_timerStart( nfs_globalCfgp->reqTimeoutHandle,
						(void *)nfs_reqTimeoutHandler, 0, 0, 0, 0);
	if (rc != NFX_OK)
		panic("%s: Failure to start requests timer!\n", __FUNCTION__);

	/*-------------------------------
	 * Create Timer for retry ReqCB
	 *-------------------------------
	 */
	nfs_globalCfgp->txReqRetryHandle = eee_timerCreate( NFS_REQ_RETRY_PERIOD,
														EEE_TIMER_REPEAT);
	if (nfs_globalCfgp->txReqRetryHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
		panic ("%s: Failure to create retry TX  timer!\n", __FUNCTION__);

	/*-------------------------------
	 * Start Timer for retry ReqCB
	 *-------------------------------
	 */
	rc = eee_timerStart(nfs_globalCfgp->txReqRetryHandle,
						(void *)nfs_msgTxTimerHandler, 0, 0, 0, 0);
	if (rc != NFX_OK)
		panic("%s: Failure to start TX replies timer!\n", __FUNCTION__);

	/*-------------------------------
	 * Allocate NFS CMD app id.
	 *-------------------------------
	 */
	nfs_globalCfgp->eee_nfsCmdAppId = eee_allocateApplicationId("nfs_cmd");
	if (nfs_globalCfgp->eee_nfsCmdAppId == EEE_ERR_INVALID_APP_NAME)
	{
		panic("%s: Failure to alloc Application NFS CMD ID!\n",__FUNCTION__);
	}

	/*-------------------------------
	 * Register nfs cmd function with EEE 
	 *-------------------------------
	 */
	if (eee_registerReceiveFunc(nfs_globalCfgp->eee_nfsCmdAppId, 
								nfs_cmdRcv, NULL) != NFX_OK)
	{
		panic("%s: Failure to register nfsCmdRcv!\n",__FUNCTION__);
	}


	//Register timer to clean up UDP connections
	nfs_globalCfgp->nfs_tplUdpActTimeOutHandle = 
			eee_timerCreate(NFS_TPL_UDP_ACT_TIMOUT_PERIOD,
							EEE_TIMER_REPEAT);
	if (nfs_globalCfgp->nfs_tplUdpActTimeOutHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		nfs_globalCfgp->nfs_debugPktCounters.nfs_tplUDPConnActTimerErr++;				
		printf("nfs_init 7: Failure to create NFS/TPL UDP Connection Activity timer!\n");
	}
	rc = eee_timerStart(nfs_globalCfgp->nfs_tplUdpActTimeOutHandle,
					(void *)nfs_utilCheckUdpConn, 0, 0, 0, 0);
	if (rc != NFX_OK)
	{
		nfs_globalCfgp->nfs_debugPktCounters.nfs_tplUDPConnActTimerErr++;				
		printf("nfs_init 8: Failure to start NFS/TPL UDP Connection Activity timer!\n");	
	}

	/*
	 * Shaun: 03/13 Bind NFS Server with TPL
	 */
    if (nfs_tplBindReq( 0 ) != NFX_OK)
    {
        panic( "nfs_init: TPL BIND failed\n");
    }

    /*
     * Shaun: 08/18/02
     *
     * Setup nfs & efs port Id to be used
     * in nfs_msgSendToFs().
     */
    nfs_globalCfgp->eee_nfsPortId = PACK_UNICAST_PORT_ADDRESS(
                                            nfxMySlotId,
                                            nfxMyCPUId,
                                            nfs_globalCfgp->eee_nfsAppId,
                                            EEE_CONTROL,
                                            NFX_COS_HIGH,
                                            0);

    nfs_globalCfgp->eee_efsPortId = PACK_UNICAST_PORT_ADDRESS(
                                            nfxMySlotId,
                                            NFX_NFP_FP2,
                                            nfs_globalCfgp->eee_efsAppId,
                                            EEE_CONTROL,
                                            NFX_COS_HIGH,
                                            0);

	nfs_globalCfgp->nfs_BootUpTimeCnt = 0;
	//Initialize NFP NFS max request to 90%
	nfs_globalCfgp->nfs_maxReqsThreshold = (NFP_MAX_REQUEST * 90) / 100;


	/* 
	 * Shaun 09/19/02
	 *
	 * Setup FS attribute bit map to NFS attr. index map.
	 */ 
	memset(nfs_attrIdxArray, 0, sizeof(nfs_attrIdxArray) );

	nfs_attrIdxArray[FS_PRE_PAR_ATTRS]   = NFS_ATTR_IDX_PRE_PARENT;
	nfs_attrIdxArray[FS_POST_PAR_ATTRS]  = NFS_ATTR_IDX_POST_PARENT;
	nfs_attrIdxArray[FS_PRE_ATTRS] 		 = NFS_ATTR_IDX_PRE;
	nfs_attrIdxArray[FS_POST_ATTRS] 	 = NFS_ATTR_IDX_POST;
	nfs_attrIdxArray[FS_PRE_FPAR_ATTRS]  = NFS_ATTR_IDX_PRE_FROM_PARENT ;
	nfs_attrIdxArray[FS_POST_FPAR_ATTRS] = NFS_ATTR_IDX_POST_FROM_PARENT;
	nfs_attrIdxArray[FS_PRE_TPAR_ATTRS]  = NFS_ATTR_IDX_PRE_TO_PARENT;
	nfs_attrIdxArray[FS_POST_TPAR_ATTRS] = NFS_ATTR_IDX_POST_TO_PARENT;

	nfs_cacheInit();

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:	nfs_extractCharVal 
 *
 * Description:
 *      Retrieves a byte value information from message buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_extractCharVal (char *nfsMsgp,  	/*source message buffer*/
				uchar8 *value				/*return value*/
)
{
	*value = *nfsMsgp;
	return (sizeof(uchar8));
}

/*-----------------------------------------------------------------
 * Name: nfs_extractShortVal	
 *
 * Description:
 *      Retrieves a short value information from message buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_extractShortVal(char *nfsMsgp,    /*source message buffer*/
				ushort16 *value				/*return value*/
)
{
	*value = ntohs(*(ushort16 *)nfsMsgp);
	return (sizeof (ushort16));
} 

/*-----------------------------------------------------------------
 * Name: nfs_extractLongVal 
 *
 * Description:
 *      Retrieves a long (32 bits) value information from message buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_extractLongVal(char *nfsMsgp,    /*source message buffer*/
				uint32 *value				/*return value*/
)
{
	*value = ntohl(*(uint32 *)nfsMsgp);
	return (sizeof(uint32));
}

/*-----------------------------------------------------------------
 * Name:	nfs_extractLongLongVal 
 *
 * Description:
 * Retrieves a long (64 bits) value information from message buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_extractLongLongVal(char *nfsMsgp,	/*source message buffer*/
				uint64 *value				/*return value*/
)
{
	int32			*srcMsg32p;
	int32			*destMsg32p;


	srcMsg32p = (int32 *)nfsMsgp;
	destMsg32p = (int32 *)value;
//	destMsg32p[0] = srcMsg32p[1];
//	destMsg32p[1] = srcMsg32p[0];
	destMsg32p[0] = ntohl(srcMsg32p[1]);
	destMsg32p[1] = ntohl(srcMsg32p[0]);

	return (sizeof(uint64));
}				

/*-----------------------------------------------------------------
 * Name: nfs_extractFhandle 
 *
 * Description:
 * Extract file handle from message buffer received from the client.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*
	 	 *NFX_ERR: Failure to extract file handle 
	 	 *length:  Filehandle length in bytes
	 	 */
nfs_extractFhandle(int32 nfs_version, 	/*NFS version*/
			char *nfsMsgp,		/*Source message buffer to extract filehandle*/
			fs_Handle *fhandlep	/*store filehandle*/
)
{
	int32	ourFhandleLen;

	ourFhandleLen = sizeof(fs_Handle);

	switch (nfs_version)
	{
		case NFS_VER2:
			memcpy(fhandlep, nfsMsgp, ourFhandleLen);		
			/* 
			 * Convert big to little endian
			 */
			xdr_swap(fhandlep, ourFhandleLen, 0);
			return (NFS_V2_FHANDLE_LEN);	
		case NFS_VER3:	
		{
			uint32		*msgp;
			uint32		fhandleLen;
			uint32		fieldLen;

			msgp = (int32 *)nfsMsgp;
			fhandleLen = ntohl(*msgp++);
			fieldLen = sizeof(uint32);
			if (fhandleLen != ourFhandleLen)
				return (NFX_ERR);

			memcpy(fhandlep, (char *)msgp, fhandleLen);

			xdr_swap(fhandlep, fhandleLen, 0);
			fieldLen += fhandleLen;
			return (fieldLen);
		}
	}

	return (NFX_ERR);
}


/*-----------------------------------------------------------------
 * Name:       nfs_extractString	
 *
 * Description:
 *	Extract string from source buffer and stores it in
 *	destination buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*length = number of bytes of the the object */
nfs_extractString (char *sourceBufp, 	/*source message buffer*/
		char *destNamep, 	/*return name string*/
		ushort16 *nameLen,	/*return name length in bytes*/
		ushort16 maxLen		/*maximum len allowed*/
)
{
	char	*currMsgp;
	uint32	*msgp;
	uint32	*destbufp;
	uint32	strLen;
	int32	i;
	int32	iteration;
	uint32	fieldLen;
	uint32	padLen;

	if ( (sourceBufp == NULL) || (destNamep == NULL) )
		return (NFX_ERR);

	msgp = (uint32 *)sourceBufp;
	currMsgp = sourceBufp;

	fieldLen = padLen = strLen = 0;
	strLen = ntohl(*msgp++);
	*nameLen = strLen;
	currMsgp += sizeof(uint32);
	fieldLen += sizeof(uint32);

	if (strLen > maxLen) 
	{
		/*
		return (NFSERR_FILENAME_TLONG);	
		*/
		return (NFX_ERR);
	}
	destbufp = (uint32 *)destNamep;

	iteration = strLen >> 2;
	/*
	 *Need account for padding bytes if string length is not divisible by 4.
	 *XDR primitive type is represented in integral number of 4 byte XDR unit.
	 */
	if (strLen & 0x03)
		iteration++;

	for (i = 0;  i < iteration; i++)
	{
		*destbufp++ = (*msgp++);
	}
	fieldLen += (iteration << 2);

	return (fieldLen);		
}

/*-----------------------------------------------------------------
 * Name: nfs_storeFhandle 
 *
 * Description:
 *      Store an input  file handle into message buffer to be transmitted 
 *	to NFS client.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*
	 	 *lenth of filehandle in bytes
	 	 */
nfs_storeFhandle(int32 		nfs_version, 	/*NFS version*/
				 fs_Handle 	*fhandlep, 		/*Source file handle*/
				 char 		*destbufp)		/*destination buffer */
{
	if (nfs_version == NFS_VER3)
			return (nfs_storeV3Fhandle( fhandlep, destbufp ));
	else
			return (nfs_storeV2Fhandle( fhandlep, destbufp ));
}

/*-----------------------------------------------------------------
 * Name: 		nfs_storeV3Fhandle 
 *
 * Description: Store NFS V3 file handle into message buffer 
 *
 * Date Created: 08/22/02
 *-----------------------------------------------------------------
 */
int32	
nfs_storeV3Fhandle( fs_Handle	*fhandlep, 		/*Source file handle*/
					char 		*destbufp)		/*destination buffer */
{
    nfs_Fhandle_t   *nfsFhp;

	/*
	 * fill the byte count field
	 */
	*(uint32 *)destbufp = sizeof(fs_Handle);

	destbufp += sizeof(uint32);

    nfsFhp = (nfs_Fhandle_t *)destbufp;

    nfsFhp->volId[1] = (uint32 )(fhandlep->volId.whole_thing >> 32);
    nfsFhp->volId[0] = (uint32 )(fhandlep->volId.whole_thing & 0xFFFFFFFF);

    nfsFhp->inum[1]  = (uint32 )(fhandlep->inum >> 32);
    nfsFhp->inum[0]  = (uint32 )(fhandlep->inum & 0xFFFFFFFF);

    nfsFhp->gen      = fhandlep->gen;
    nfsFhp->mkfsTime = fhandlep->mkfsTime;

    nfsFhp->mntPnt[1] = (uint32 )(fhandlep->mntPnt >> 32);
    nfsFhp->mntPnt[0] = (uint32 )(fhandlep->mntPnt & 0xFFFFFFFF);

	return (sizeof(fs_Handle) + sizeof(uint32));
}

/*-----------------------------------------------------------------
 * Name: 		nfs_storeV2Fhandle 
 *
 * Description: Store NFS V2 file handle into message buffer 
 *
 * Date Created: 08/22/02
 *-----------------------------------------------------------------
 */
int32	
nfs_storeV2Fhandle( fs_Handle	*fhandlep, 		/*Source file handle*/
					char 		*destbufp)		/*destination buffer */
{
	int32   ourFhandleLen;

	ourFhandleLen = sizeof(fs_Handle);
	
	memcpy (destbufp, (char *)fhandlep, ourFhandleLen);
	destbufp += ourFhandleLen;
	memset (destbufp, 0,(NFS_V2_FHANDLE_LEN - ourFhandleLen));
	return (NFS_V2_FHANDLE_LEN);

}

/*-----------------------------------------------------------------
 * Name:       	nfs_storeByteValToBuf	
 *
 * Description:
 *      Stores a 8 bits value into destination buffer. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_storeByteValToBuf(uchar8 *destBuf, uchar8  byteVal)
{
	*destBuf = byteVal;
	return (sizeof(uchar8));
}

/*-----------------------------------------------------------------
 * Name:       	nfs_storeShortValToBuf	
 *
 * Description:
 *      Stores a 16 bits value into destination buffer. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_storeShortValToBuf(char *destBuf, ushort16 shortVal)
{
	*(ushort16 *)destBuf = shortVal;
	return (sizeof(ushort16));
}

/*-----------------------------------------------------------------
 * Name:       	nfs_storeLongValToBuf	
 *
 * Description:
 *      Stores a 32 bits value into destination buffer. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_storeLongValToBuf(char *destBuf, uint32 longVal)
{
	*(uint32 *)destBuf = longVal;
	return (sizeof(uint32));
}

/*-----------------------------------------------------------------
 * Name:       	nfs_storeLongLongValToBuf	
 *
 * Description:
 *      Stores a 64 bits value into destination buffer. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_storeLongLongValToBuf(char *destBuf, uint64 longLongVal)
{
	uint32	*dst = (uint32 *)destBuf;

	dst[0] = (uint32 )(longLongVal >> 32);
	dst[1] = (uint32 )(longLongVal & 0xFFFFFFFF);

	return (sizeof(uint64));
}

/*-----------------------------------------------------------------
 * Name:       nfs_storeString
 *
 * Description:
 *      Store string into destination buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */

int32
nfs_storeString(
	char *DestBufp,     /*destination buffer to store input string*/
	char *string,       /*string*/
	uint32 strLen,      /*string length in bytes*/
	uint32 maxLen		/*maximum string length allowed*/
)
{
    char        *destp;
    char        *sourcep;
    uint32      length = 0;
    int32       i;
    int32       iteration;

    if ((DestBufp == NULL) || (string == NULL))
        return (0);
    if  ((strLen < 0) || (strLen > maxLen) )
        return (0);
   
    length = 0;
    destp = DestBufp;
	*(uint32 *)destp = strLen; 
    /*
     *String length in number of bytes
     */
	*(uint32 *)destp = strLen; 
	destp += 4;
    length += 4;

    sourcep = string;
    iteration = strLen >> 2;
    /*
     *Need account for padding bytes if string length is not divisible by 4.
     *XDR primitive type is represented in integral number of 4 byte XDR unit.
     */
    if (strLen & 0x03)
        iteration++;

    for (i = 0;  i < iteration; i++)
    {
		destp[3] = sourcep[0];
		destp[2] = sourcep[1];
		destp[1] = sourcep[2];
		destp[0] = sourcep[3];
		destp += 4;
		sourcep += 4;
    }

    length += (iteration << 2);
    return (length);

}

/*-----------------------------------------------------------------
 * Name:     nfs_storeIpAddr  
 *
 * Description:
 *      Store IP address in string format into destination buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_storeIpAddr(char *destBufp,		//dest buffer to store
				uint32 ipAddr)
{
	char		*ipStrp;
	uint32		ipAddrLen;
	char		ipAddrStr[NFS_MAX_IP_ADDR_STRLEN];
	int32		length;

	if (destBufp == NULL) 
		return (NULL);

	ipStrp = (char *)&ipAddr;
	memset(ipAddrStr, 0, sizeof(ipAddrStr));
	sprintf(ipAddrStr, "%u.%u.%u.%u",0xff&ipStrp[0], 
				0xff&ipStrp[1], 0xff&ipStrp[2], 0xff&ipStrp[3]);

	ipAddrLen = strlen(ipAddrStr);

	length = nfs_storeString(destBufp, ipAddrStr, ipAddrLen,
							NFS_MAX_IP_ADDR_STRLEN);
	return (length);
}

/*-----------------------------------------------------------------
 * Name:       	nfs_storeAuthFlavor 
 * 
 * Description:
 *   	Store authentication flavors support by our server. 
 * 
 * Created by:  Joseph Vu
 * 
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*
	 	 *length = number of bytes of the field
	 	 */
nfs_storeAuthFlavor(char *msgp)
{
	uint32		*destMsgp;

	destMsgp = (uint32 *)msgp;
	/*
	 *Store AUTH_UNIX for now
	 */
	*destMsgp++ = NFS_PRC_AUTH_SYS;
	*destMsgp++ = NFS_PRC_AUTH_SYS;			

	return (8);	
}


/*-----------------------------------------------------------------
 * Name:	nfs_updateStatCounter	
 *
 * Description:
 *     	Update statistic counters. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
void
nfs_updateXmtStatCounter(nfs_request_t *nfsReqp)
{
	nfs_pktCounters_t		*txPktStatp;

	txPktStatp = &nfs_globalCfgp->nfs_xmtPktCounters;

	switch (nfsReqp->reqType)
	{
		case NFS_MKNOD:
			txPktStatp->mknod++;	
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.mknod++;	
			break;

		case NFS_REMOVE:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.remove++;	
			txPktStatp->remove++;	
			break;

		case NFS_RMDIR:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.rmdir++;	
			txPktStatp->rmdir++;	
			break;

		case NFS_RENAME:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.rename++;	
			txPktStatp->rename++;	
			break;

		case NFS_CREATE:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.create++;	
			txPktStatp->create++;	
			break;

		case NFS_MKDIR:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.mkdir++;	
			txPktStatp->mkdir++;	
			break;

		case NFS_GETATTR:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.getAttr++;	
			txPktStatp->getAttr++;
			break;
		case NFS_SETATTR:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.setAttr++;	
			txPktStatp->setAttr++;	
			break;
		case NFS_LOOKUP:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.lookUp++;	
			txPktStatp->lookUp++;	
			break;
		case NFS_ACCESS:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.access++;	
			txPktStatp->access++;	
			break;

		case NFS_READLINK:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.readLink++;	
			txPktStatp->readLink++;	
			break;
		case NFS_READ:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.read++;	
			txPktStatp->read++;	
			break;

		case NFS_WRITE:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.write++;	
			txPktStatp->write++;	
			break;
		case NFS_SYMLINK:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.symlink++;	
			txPktStatp->symlink++;	
			break;
		case NFS_LINK:
			txPktStatp->link++;	
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.link++;	
			break;
		case NFS_READDIR:
			txPktStatp->readdir++;	
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.readdir++;	
			break;
		case NFS_STATFS:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.statfs++;	
			txPktStatp->statfs++;	

			break;
		case NFS_FSSTAT:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.fsstat++;	
			txPktStatp->fsstat++;	
			break;

		case NFS_FSINFO:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.fsinfo++;	
			txPktStatp->fsinfo++;	
			break;

		case NFS_PATHCONF:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.pathconf++;	
			txPktStatp->pathconf++;	
			break;

		case NFS_COMMIT:
			nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.commit++;	
			txPktStatp->commit++;	
			break;

	}

}

void nfs_setStat(int32 fsStatus)
{
	switch (fsStatus)
	{
		case FS_ERR_NONE:					// NFX_OK, NFS_OK, NO_ERROR
			nfs_globalCfgp->nfs_statCounters.nfs_ok++;
			break;
    	case FS_ERR_PERM:
			nfs_globalCfgp->nfs_statCounters.nfs_errPerm++;
			break;
		case FS_ERR_NOENT:
			nfs_globalCfgp->nfs_statCounters.nfs_errNoEnt++;
			break;
		case FS_ERR_IO:
			nfs_globalCfgp->nfs_statCounters.nfs_errIO++;
			break;
		case FS_ERR_NXIO:
			nfs_globalCfgp->nfs_statCounters.nfs_errNxio++;
			break;
		case FS_ERR_ACCESS:					//permission denied
			nfs_globalCfgp->nfs_statCounters.nfs_errAccess++;
			break;
		case FS_ERR_EXIST:
			nfs_globalCfgp->nfs_statCounters.nfs_errFileExist++;
			break; 
		case FS_ERR_XDEV:
		case FS_ERR_NODEV:
			nfs_globalCfgp->nfs_statCounters.nfs_errNoDev++;
			break;
    	case FS_ERR_NOTDIR:
			nfs_globalCfgp->nfs_statCounters.nfs_errNotDir++;
			break;
    	case FS_ERR_ISDIR:
			nfs_globalCfgp->nfs_statCounters.nfs_errIsDir++;
			break;
    	case FS_ERR_INVAL:
			nfs_globalCfgp->nfs_statCounters.nfs_errInVal++;
			break;
    	case FS_ERR_FBIG:
			nfs_globalCfgp->nfs_statCounters.nfs_errFileBig++;
			break;
    	case FS_ERR_NOSPC:
			nfs_globalCfgp->nfs_statCounters.nfs_errNoSpc++;
			break;
    	case FS_ERR_ROFS:
			nfs_globalCfgp->nfs_statCounters.nfs_errRofs++;
			break;
    	case FS_ERR_MLINK:
			nfs_globalCfgp->nfs_statCounters.nfs_errMlink++;
			break;
    	case FS_ERR_NAMETOOLONG:
			nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;	
			break;
    	case FS_ERR_NOTEMPTY:
			nfs_globalCfgp->nfs_statCounters.nfs_errNotEmpty++;
			break;
    	case FS_ERR_DQUOT:
			nfs_globalCfgp->nfs_statCounters.nfs_errDquota++;
			break;
    	case FS_ERR_STALE:
			nfs_globalCfgp->nfs_statCounters.nfs_errStaleFhandle++;
			break;
    	case FS_ERR_REMOTE:
			nfs_globalCfgp->nfs_statCounters.nfs_errRemote++;
			break;
    	case FS_ERR_WFLUSH:
			nfs_globalCfgp->nfs_statCounters.nfs_errFlush++;
			break;
		case FS_ERR_BADHANDLE:
			nfs_globalCfgp->nfs_statCounters.nfs_errBadHandle++;
			break;
		case FS_ERR_NOT_SYNC:
			nfs_globalCfgp->nfs_statCounters.nfs_errNotSync++;
			break;
		case FS_ERR_BAD_COOKIE:
			nfs_globalCfgp->nfs_statCounters.nfs_errBadCookie++;
			break;	
		case FS_ERR_NOTSUPP:
			nfs_globalCfgp->nfs_statCounters.nfs_errNotSupp++;
			break;
		default:
			nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
	}	//switch

}
/*-----------------------------------------------------------------
 * Name:	nfs_totalPktCnt 
 *  
 * Description:
 *		Return total packets count on the NFS server based 
 * for a NFS version.
 * 
 * Created by:  Joseph Vu
 *  
 * Date Created:        5/28/01
 *  
 *-----------------------------------------------------------------
 */ 
uint64 nfs_totalPktCnt(nfs_pktCounters_t   *pktStatp, uint32 nfs_version)
{
	int32			totalPktCnt = 0;

	if (nfs_version == NFS_VER2)
	{
		totalPktCnt = 
			pktStatp->null + pktStatp->getAttr +  pktStatp->setAttr + pktStatp->lookUp + 
			pktStatp->readLink + pktStatp->read +  pktStatp->write + pktStatp->create +
			pktStatp->mkdir + pktStatp->symlink + pktStatp->remove + pktStatp->rmdir +
			pktStatp->rename + pktStatp->link + pktStatp->readdir + pktStatp->statfs;
	}
	else
	{
		//NFS V3
		totalPktCnt = 
			pktStatp->null + pktStatp->getAttr +  pktStatp->setAttr + pktStatp->lookUp + 
			pktStatp->readLink + pktStatp->read +  pktStatp->write + pktStatp->create +
			pktStatp->mkdir + pktStatp->symlink + pktStatp->remove + pktStatp->rmdir +
			pktStatp->rename + pktStatp->link + pktStatp->readdir + 
			pktStatp->access +  pktStatp->mknod + pktStatp->readdirplus + pktStatp->fsstat +
			pktStatp->fsinfo + pktStatp->pathconf + pktStatp->commit; 

	}

	return (totalPktCnt);
}

/*-----------------------------------------------------------------
 * Name:   	nfs_displayMntInfo 
 *
 * Description:
 *  Display MOUNT client information. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
void nfs_displayMntInfo(void)
{
	nfp_nodePtr_t		nodep;
	nfp_list_t 			*nodeListp;
	nfp_fileSysCBPtr_t 	fileSysCBp;
//	nfs_clientInfo_t    *clientInfop;
	nfp_clientCBPtr_t   clientCBp;
    uint32              clientCount;
	uint32				nodeCount;
	uint32				clientIpAddr;
	uint32				maxClient;

	clientCount = 0;	
	maxClient = 32;			//Set max client to 32 for now.  
							//Changed to NFP_MAX_CLIENT later
	printf("Client IP         Mnt Pnt     VolId  Inum      Mount Pt    Last Access Time\n");
	printf("-----------------------------------------------------------------------------------\n");

	for (clientCBp= (nfp_clientCBPtr_t)nfp_globalCfgp->clientCbCfgList.listBasep;
		((clientCount < maxClient) && (clientCBp != NULL)); clientCount++, clientCBp++)
	{
		if ((nfp_validateClientCB(clientCBp) != NFX_OK) ||
			  (!(clientCBp->nfp_flag & NFP_ENTRY_INUSED)) ||
				(!clientCBp->nfp_ipAddr) )	
			continue;
		clientIpAddr = ntohl(clientCBp->nfp_ipAddr);
		/*
		printf("%d.%d.%d.%d", 
				((clientIpAddr >> 24) & 0xff), 
				((clientIpAddr >> 16) & 0xff),
				((clientIpAddr >> 8) & 0xff),
				(clientIpAddr & 0xff) );
		*/		
		//Display client mount points
		nodeListp = &clientCBp->nfp_fileSysList;
		nodeCount = nodeListp->nfp_reqCnt;
		if (nodeCount > 16)
			nodeCount = 16;			//Set to 16 for now.  
									//It should be set to NFP_MAX_FILESYS
    	for (nodep = nodeListp->nfp_headp; (nodep != NULL && nodeCount > 0);
        	nodep = nodep->nfp_nextp, nodeCount--)
    	{
        	fileSysCBp = (nfp_fileSysCBPtr_t)nodep->nfp_entryCBp;
			if ( (nfp_validateFileSysCB(fileSysCBp) != NFX_OK) ||
				(!(fileSysCBp->nfp_flag & NFP_ENTRY_INUSED )) )
				continue;
			printf("%d.%d.%d.%d", 
				((clientIpAddr >> 24) & 0xff), 
				((clientIpAddr >> 16) & 0xff),
				((clientIpAddr >> 8) & 0xff),
				(clientIpAddr & 0xff) );
			printf("%12s", fileSysCBp->nfp_name);

			printf("%10llx %10llx %10llx",
				fileSysCBp->nfp_fileHandle.volId.whole_thing,
				fileSysCBp->nfp_fileHandle.inum,
				fileSysCBp->nfp_fileHandle.mntPnt);

			printf("     %20s)\n\n",
				eee_humandate(fileSysCBp->atime));
		}
	}

	mnt_displayStat();
}


/*-----------------------------------------------------------------
 * Name:   	nfs_displayStat 
 * 
 * Description:
 *	Display NFS statistics.
 *
 * Created by:  Joseph Vu
 * 
 * Date Created:        5/28/01
 * 
 *-----------------------------------------------------------------
 */
void nfs_displayStat()
{
//	nfp_nodePtr_t		nodep;
//	nfp_list_t 			*nodeListp;
//	nfp_fileSysCBPtr_t 	fileSysCBp;
//	nfs_clientInfo_t    *clientInfop;
//	nfp_clientCBPtr_t   clientCBp;
	nfs_pktCounters_t	*rxPktStatp;
	nfs_pktCounters_t	*txPktStatp;
	uint64				total_rx;
	uint64				total_tx;
	uint64				total_nfsV2RxPkt;
	uint64				total_nfsV3RxPkt;
	uint64				total_nfsV2TxPkt;
	uint64				total_nfsV3TxPkt;
	uint64				totalRetransmitCalls;
//  uint32              count;
//	uint32				clientIpAddr;
	uint64				totalReq, totalResp;



	total_rx = 0;
	total_tx = 0;
	total_nfsV2RxPkt = 0;
	total_nfsV3RxPkt = 0;
	
	rxPktStatp = &nfs_globalCfgp->nfs_rxPktCounters;
	txPktStatp = &nfs_globalCfgp->nfs_xmtPktCounters;

	//total_rx = nfs_totalPktCnt(rxPktStatp, NFS_VER3);
	total_rx = rxPktStatp->totalUdpCalls + rxPktStatp->totalTcpCalls;

	printf ("Boot time       %s (sec: %d)\n",
            eee_humandate(nfs_bootTimeInSecs),nfs_bootTimeInSecs);
	
	printf("Total NFS UDP calls: %lld\n\n",rxPktStatp->totalUdpCalls);
	printf("Total NFS UDP bad calls: %lld\n\n",rxPktStatp->totalBadUdpCalls);
	printf("Total NFS TCP calls: %lld\n\n",rxPktStatp->totalTcpCalls);
	printf("Total NFS TCP bad calls: %lld\n\n",rxPktStatp->totalBadTcpCalls);
	printf("Total NFS calls: %lld\n\n", total_rx);	
	printf("Total NFS bad calls: %lld\n",
		rxPktStatp->totalBadUdpCalls+rxPktStatp->totalBadTcpCalls);
		
	rxPktStatp = &nfs_globalCfgp->nfs_v2RxPktCounters;
	txPktStatp = &nfs_globalCfgp->nfs_v2XmtPktCounters;
	total_nfsV2RxPkt =   nfs_totalPktCnt(rxPktStatp, NFS_VER2);

	
    printf("Client NFS V2 call counters\n");
    printf("----------------------------\n");
    printf("Null:       %lld\n", rxPktStatp->null);
    printf("GetAttr:    %lld\n", rxPktStatp->getAttr);
    printf("SetAttr:    %lld\n", rxPktStatp->setAttr);
    printf("LookUp:     %lld\n",rxPktStatp->lookUp);
    printf("Readlink:   %lld\n", rxPktStatp->readLink);
    printf("Read:       %lld\n", rxPktStatp->read);
    printf("Write:      %lld\n", rxPktStatp->write);
    printf("Create:     %lld\n", rxPktStatp->create);
    printf("Mkdir:      %lld\n", rxPktStatp->mkdir);
    printf("Symlink:    %lld\n", rxPktStatp->symlink);
    printf("Remove:     %lld\n", rxPktStatp->remove);
    printf("Rmdir:      %lld\n", rxPktStatp->rmdir);
    printf("Rename:     %lld\n", rxPktStatp->rename);
    printf("Link:       %lld\n", rxPktStatp->link);
    printf("Readdir:    %lld\n", rxPktStatp->readdir);
    printf("Statfs:     %lld\n", rxPktStatp->statfs);
    printf("Total client NFS V2 calls: %lld\n\n",total_nfsV2RxPkt);

    total_nfsV2TxPkt =   nfs_totalPktCnt(txPktStatp, NFS_VER2);
    printf("Client NFS V2 Replies Counters\n");
    printf("----------------------------\n");
    printf("Null:       %lld\n", txPktStatp->null);
    printf("GetAttr:    %lld\n", txPktStatp->getAttr);
    printf("SetAttr:    %lld\n", txPktStatp->setAttr);
    printf("LookUp:     %lld\n",txPktStatp->lookUp);
    printf("Readlink:   %lld\n", txPktStatp->readLink);
    printf("Read:       %lld\n", txPktStatp->read);
    printf("Write:      %lld\n", txPktStatp->write);
    printf("Create:     %lld\n", txPktStatp->create);
    printf("Mkdir:      %lld\n", txPktStatp->mkdir);
    printf("Symlink:    %lld\n", txPktStatp->symlink);
    printf("Remove:     %lld\n", txPktStatp->remove);
    printf("Rmdir:      %lld\n", txPktStatp->rmdir);
    printf("Rename:     %lld\n", txPktStatp->rename);
    printf("Link:       %lld\n", txPktStatp->link);
    printf("Readdir:    %lld\n", txPktStatp->readdir);
    printf("Statfs:     %lld\n", txPktStatp->statfs);
    printf("Total client NFS V2 replies: %lld\n\n",total_nfsV2TxPkt);

    //NFS V3 client
    rxPktStatp = &nfs_globalCfgp->nfs_v3RxPktCounters;
    txPktStatp = &nfs_globalCfgp->nfs_v3XmtPktCounters;
    total_nfsV3RxPkt =   nfs_totalPktCnt(rxPktStatp, NFS_VER3);

    printf("Client NFS V3 call counters\n");
    printf("----------------------------\n");
    printf("Null:       %lld\n", rxPktStatp->null);
    printf("GetAttr:    %lld\n", rxPktStatp->getAttr);
    printf("SetAttr:    %lld\n", rxPktStatp->setAttr);
    printf("LookUp:     %lld\n",rxPktStatp->lookUp);
    printf("Access:     %lld\n",rxPktStatp->access);
    printf("Readlink:   %lld\n", rxPktStatp->readLink);
    printf("Read:       %lld\n", rxPktStatp->read);
    printf("Write:      %lld\n", rxPktStatp->write);
    printf("Create:     %lld\n", rxPktStatp->create);
    printf("Mkdir:      %lld\n", rxPktStatp->mkdir);
    printf("Symlink:    %lld\n", rxPktStatp->symlink);
    printf("Mknod:      %lld\n", rxPktStatp->mknod);
    printf("Remove:     %lld\n", rxPktStatp->remove);
    printf("Rmdir:      %lld\n", rxPktStatp->rmdir);
    printf("Rename:     %lld\n", rxPktStatp->rename);
    printf("Link:       %lld\n", rxPktStatp->link);
    printf("Readdir:    %lld\n", rxPktStatp->readdir);
    printf("Readdirplus:%lld\n", rxPktStatp->readdirplus);
    printf("Fsstat:     %lld\n", rxPktStatp->fsstat);
    printf("Fsinfo:     %lld\n", rxPktStatp->fsinfo);
    printf("Pathconf:   %lld\n", rxPktStatp->pathconf);
    printf("Commit:     %lld\n", rxPktStatp->commit);
    printf("Total client NFS V3 calls: %lld\n\n",total_nfsV3RxPkt);

    total_nfsV3TxPkt =   nfs_totalPktCnt(txPktStatp, NFS_VER3);

    printf("Client NFS V3 Replies Counters\n");
    printf("----------------------------\n");
    printf("Null:       %lld\n", txPktStatp->null);
    printf("GetAttr:    %lld\n", txPktStatp->getAttr);
    printf("SetAttr:    %lld\n", txPktStatp->setAttr);
    printf("LookUp:     %lld\n",txPktStatp->lookUp);
    printf("Access:     %lld\n",txPktStatp->access);
    printf("Readlink:   %lld\n", txPktStatp->readLink);
    printf("Read:       %lld\n", txPktStatp->read);
    printf("Write:      %lld\n", txPktStatp->write);
    printf("Create:     %lld\n", txPktStatp->create);
    printf("Mkdir:      %lld\n", txPktStatp->mkdir);
    printf("Symlink:    %lld\n", txPktStatp->symlink);
    printf("Mknod:      %lld\n", txPktStatp->mknod);
    printf("Remove:     %lld\n", txPktStatp->remove);
    printf("Rmdir:      %lld\n", txPktStatp->rmdir);
    printf("Rename:     %lld\n", txPktStatp->rename);
    printf("Link:       %lld\n", txPktStatp->link);
    printf("Readdir:    %lld\n", txPktStatp->readdir);
    printf("Readdirplus:%lld\n", txPktStatp->readdirplus);
    printf("Fsstat:     %lld\n", txPktStatp->fsstat);
    printf("Fsinfo:     %lld\n", txPktStatp->fsinfo);
    printf("Pathconf:   %lld\n", txPktStatp->pathconf);
    printf("Commit:     %lld\n", txPktStatp->commit);
    printf("Total client NFS V3 replies: %lld\n\n",total_nfsV3TxPkt);

	printf("NFS Error counters\n");
	printf("----------------------------\n");
	printf("NFS OK:                   %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_ok);
	printf("NFS Perm (Not owner):     %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errPerm);
	printf("NFS no such file/dir:     %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNoEnt);
	printf("NFS Err IO:               %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errIO);
	printf("NFS Err XIO:              %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNxio);
	printf("NFS Err Access:           %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errAccess);
	printf("NFS Err file exist:       %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errFileExist);
	printf("NFS Err name too long:    %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong);
	printf("NFS Err no such device:   %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNoDev);
	printf("NFS Err Not Dir:          %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNotDir);
	printf("NFS Err Is Dir:           %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errIsDir);
	printf("NFS Err Invalid value:    %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errInVal);
	printf("NFS Err File Too Big:     %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errFileBig);
	printf("NFS Err No Space:         %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNoSpc);
	printf("NFS Err Read-Only FS:     %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errRofs);
	printf("NFS Err Too many links:   %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errMlink);
	printf("NFS Err Dir Not Empty:    %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNotEmpty);
	printf("NFS Err Stale Fhandle:    %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errStaleFhandle);
	printf("NFS Err Bad Fhandle:      %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errBadHandle);
	printf("NFS Err Quota Exceed:     %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errDquota);
	printf("NFS Err Remote:           %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errRemote);
	printf("NFS Err Flush:            %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errFlush);
	printf("NFS Err Not Sync:         %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNotSync);
	printf("NFS Err Bad Cookie:       %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errBadCookie);
	printf("NFS Err Not Support:      %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errNotSupp);
	printf("NFS Err Generic:          %lld\n", nfs_globalCfgp->nfs_statCounters.nfs_errGeneric);


	printf("\nNFS debug counters\n");
	printf("----------------------------\n");
	printf("TPL Xmt Alloc Pkts:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.tpl_xmtPktAlloc);
	printf("TPL Xmt Free Pkts:           %lld\n", nfs_globalCfgp->nfs_debugPktCounters.tpl_xmtPktFree);
	printf("TPL Xmt Pkts:                %lld\n", nfs_globalCfgp->nfs_debugPktCounters.tpl_xmtPktSend);
	printf("TPL Xmt Pkts Completed:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.tpl_xmtPktCmpInd);
	printf("TPL Xmt Pkts Completed Err:  %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_tplNoRespRcvdErr);
	printf("TPL Xmt No Pkt Resources:    %lld\n", nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes);
	printf("TPL NFS rcv'd invalid Pkts:  %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_invalidTplPktRcvd);
	printf("NFS Read Xmts Completed:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_readXmtCmpInd);
	printf("NFS Read Xmts Completed (2): %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_readXmtCmpInd2);
	printf("NFS Call to free resources 1:  %d\n", nfs_callmsgFreeAll);
	printf("NFS Call to free resources 2:  %d\n", nfs_callmsgFreeAll2);
	printf("Num File System responses dropped: %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_dropFsResp);
	printf("Num File System Res/volume err:%lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_fsResourceErr);
	printf("NFS Alloc NFP Req Control Block:%lld\n",
			nfs_globalCfgp->nfs_debugPktCounters.nfs_allocReqCB);
	printf("NFS Free NFP Req Control Block:%lld\n",
			nfs_globalCfgp->nfs_debugPktCounters.nfs_freeReqCB);
	printf("Num of NFS Req timout waiting response from the file system count: %lld\n",
			nfs_globalCfgp->nfs_debugPktCounters.nfs_fsNoRespRcvdErr);
	printf("Num of Rx NFS reqs, but no Mount point found (Get By VolId) : %lld\n", 
			nfs_globalCfgp->nfs_debugPktCounters.nfs_fileSysNotMnt);
	printf("Num of Rx NFS reqs, Get By VolId Err : %lld\n", 
			nfs_globalCfgp->nfs_debugPktCounters.nfs_getFileSysByIdErr);
	printf("Num of Rx NFS reqs on an invalid interface : %lld\n", 
			nfs_globalCfgp->nfs_debugPktCounters.nfs_rxPktIfErr);
	printf("Num of Rx NFS reqs with NFS share err  : %lld\n", 
			nfs_globalCfgp->nfs_debugPktCounters.nfs_rxPktNfsShareErr);
	printf("Rx Invalid Responses from the file system :  %lld\n", 
			nfs_globalCfgp->nfs_debugPktCounters.nfs_rxInvalidFSResp);
	printf("Rx Invalid Responses:                        %lld\n", 
			nfs_globalCfgp->nfs_debugPktCounters.nfs_rxInvalidResp);
	printf("TPL (RX) bad desc error:     %lld\n", 
					nfs_globalCfgp->nfs_debugPktCounters.tpl_badDesc);
	printf("Alloc EEE buffers send to FS:    %lld\n", 
					nfs_globalCfgp->nfs_debugPktCounters.nfs_eeeRamAlloc1);
	printf("Free EEE buffers send to FS:     %lld\n", 
					nfs_globalCfgp->nfs_debugPktCounters.nfs_eeeRamDeAlloc1);
	printf("Alloc EEE buffers for FS read Xmit compl:   %lld\n", 
					nfs_globalCfgp->nfs_debugPktCounters.nfs_eeeRamAlloc2);
	printf("Free EEE buffers for FS read Xmit compl:   %lld\n", 
					nfs_globalCfgp->nfs_debugPktCounters.nfs_eeeRamDeAlloc2);
	
   printf("Alloc EEE buffers for RX RPC hdr:           %lld\n",
                nfs_globalCfgp->nfs_debugPktCounters.nfs_eeeRamAlloc3);
    printf("Free EEE buffers for RX PRC hdr:           %lld\n",
                nfs_globalCfgp->nfs_debugPktCounters.nfs_eeeRamDeAlloc3);
	//nfs_globalCfgp->nfs_debugPktCounters.nfs_tplUDPConnActTimerErr++;	
    printf("Num of Failure for starting TPL/UDP conn timer:%lld\n",
                nfs_globalCfgp->nfs_debugPktCounters.nfs_tplUDPConnActTimerErr);
#ifdef	NOTUSED
    printf("Current NFP/NFS requests threshold:        %d\n",
                nfs_globalCfgp->nfs_maxReqsThreshold);
    printf("Num of dropped rcv'd packets-reach max requests threshold:    %lld\n",
                nfs_globalCfgp->nfs_debugPktCounters.nfs_maxReqsDropPktErr);

#endif
   printf("nfs tmp debug 1:           %d\n",nfs_tmpDebug1);
   printf("nfs tmp debug 2:           %d\n",nfs_tmpDebug2);
	
	printf("NFS retransmit calls received from clients\n");
    printf("---------------------------------------------\n");
    printf("Null:       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.null);
    printf("GetAttr:    %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.getAttr);
    printf("SetAttr:    %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.setAttr);
    printf("LookUp:     %lld\n",nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.lookUp);
    printf("Access:     %lld\n",nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.access);
    printf("Readlink:   %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.readLink);
    printf("Read:       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.read);
    printf("Write:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.write);
    printf("Create:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.create);
    printf("Mkdir:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.mkdir);
    printf("Symlink:    %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.symlink);
    printf("Mknod:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.mknod);
    printf("Remove:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.remove);
    printf("Rmdir:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.rmdir);
    printf("Rename:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.rename);
    printf("Link:       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.link);
    printf("Readdir:    %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.readdir);
    printf("Readdirplus:%lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.readdirplus);
    printf("Fsstat:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.fsstat);
    printf("Fsinfo:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.fsinfo);
    printf("Pathconf:   %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.pathconf);
    printf("Commit:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters.commit);

	rxPktStatp = &nfs_globalCfgp->nfs_debugPktCounters.nfs_PktRxRetriesCounters;
	totalRetransmitCalls  = rxPktStatp->getAttr+ rxPktStatp->setAttr+ rxPktStatp->lookUp+ 
		rxPktStatp->access+ rxPktStatp->readLink+
		rxPktStatp->read + rxPktStatp->write+ rxPktStatp->create + rxPktStatp->mkdir + rxPktStatp->symlink +
		rxPktStatp->mknod + rxPktStatp->remove + rxPktStatp->rmdir + rxPktStatp->rename + rxPktStatp->link +
		rxPktStatp->readdir + rxPktStatp->readdirplus + rxPktStatp->statfs + rxPktStatp->fsstat + rxPktStatp->fsinfo +
		rxPktStatp->commit;	
    printf("Total client retransmit calls: %lld\n\n",totalRetransmitCalls);

	printf("NFS calls/replies sent/received with the File System\n");
	printf("------------------------------------------------------\n");
	printf("NFS getattr send to FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.getAttr);
	printf("NFS getattr Rcv'd from FS:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.getAttr);
	printf("NFS setattr send to FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.setAttr);
	printf("NFS setattr Rcv'd from FS:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.setAttr);
	printf("NFS lookup  send to FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.lookUp);
	printf("NFS lookup  Rcv'd from FS:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.lookUp);
	printf("NFS access  send to FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.access);
	printf("NFS access  Rcv'd from FS:      %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.access);
	printf("NFS readlink send to FS:        %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readLink);
	printf("NFS readlink Rcv'd from FS:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.readLink);
	printf("NFS read send to FS:            %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.read);
	printf("NFS read Rcv'd from FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.read);
	printf("NFS write send to FS:           %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.write);
	printf("NFS write Rcv'd from FS         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.write);
	printf("NFS create send to FS:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.create);
	printf("NFS create Rcv'd from FS        %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.create);
	printf("NFS mkdir send to FS:           %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.mkdir);
	printf("NFS mkdir Rcv'd from FS         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.mkdir);
	printf("NFS symlink send to FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.symlink);
	printf("NFS symlink Rcv'd from FS       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.symlink);
	printf("NFS mknod  send to FS:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.mknod);
	printf("NFS mknod Rcv'd from FS         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.mknod);
	printf("NFS remove  send to FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.remove);
	printf("NFS remove Rcv'd from FS:       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.remove);
	printf("NFS rmdir send to FS:           %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.rmdir);
	printf("NFS rmdir Rcv'd from FS:        %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.rmdir);
	printf("NFS rename send to FS:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.rename);
	printf("NFS rename Rcv'd from FS:       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.rename);
	printf("NFS link send to FS:            %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.link);
	printf("NFS link Rcv'd from FS          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.link);
	printf("NFS readdir send to FS:         %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readdir);
	printf("NFS readir  Rcv'd from FS       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.readdir);
	printf("NFS readdirplus send to FS:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readdirplus);
	printf("NFS readirplus  Rcv'd from FS:  %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.readdirplus);
	printf("NFS statfs send to FS:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.statfs);
	printf("NFS statfs Rcv'd from FS:       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.statfs);
	printf("NFS fsstat send to FS:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.fsstat);
	printf("NFS fsstat Rcv'd from FS        %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.fsstat);
	printf("NFS fsinfo send to FS:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.fsinfo);
	printf("NFS fsinfo Rcv'd from FS        %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.fsinfo);
	printf("NFS pathconf send to FS:        %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.pathconf);
	printf("NFS pathconf Rcv'd from FS:     %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.pathconf);
	printf("NFS commit send to FS:          %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.commit);
	printf("NFS commit Rcv'd from FS:       %lld\n", nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.commit);

    txPktStatp = &nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters;
    rxPktStatp = &nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters;

	totalReq = txPktStatp->getAttr+ txPktStatp->setAttr+ txPktStatp->lookUp+ txPktStatp->access+ txPktStatp->readLink+
				txPktStatp->read + txPktStatp->write+ txPktStatp->create + txPktStatp->mkdir + txPktStatp->symlink +
				txPktStatp->mknod + txPktStatp->remove + txPktStatp->rmdir + txPktStatp->rename + txPktStatp->link +
				txPktStatp->readdir + txPktStatp->readdirplus + txPktStatp->statfs + txPktStatp->fsstat + txPktStatp->fsinfo +
				txPktStatp->commit;	

	totalResp = rxPktStatp->getAttr+ rxPktStatp->setAttr+ rxPktStatp->lookUp+ rxPktStatp->access+ rxPktStatp->readLink+
				rxPktStatp->read + rxPktStatp->write+ rxPktStatp->create + rxPktStatp->mkdir + rxPktStatp->symlink +
				rxPktStatp->mknod + rxPktStatp->remove + rxPktStatp->rmdir + rxPktStatp->rename + rxPktStatp->link +
				rxPktStatp->readdir + rxPktStatp->readdirplus + rxPktStatp->statfs + rxPktStatp->fsstat + rxPktStatp->fsinfo +
				rxPktStatp->commit;	
	printf("\n");
	printf("Total NFS Requests send to FS:                  %lld\n", totalReq);
	printf("Total NFS Replies rcv'd from FS:                %lld\n", totalResp);

    printf("Total MNT Requests send to file system:         %lld\n",
                mnt_globalCfgp->mnt_debugCounters.mnt_totalNumOfReqXmtToFs);
    printf("Total MNT Responses rcv'd from file system:     %lld\n",
                mnt_globalCfgp->mnt_debugCounters.mnt_totalNumOfRespRcvdFromFs);

	printf("Total MOUNT/NFS Requests send to FS:            %lld\n", 
		nfs_globalCfgp->nfs_totalNumOfReqXmtToFs);
	printf("Total MOUNT/NFS Replies rcv'd from FS:          %lld\n", 
		nfs_globalCfgp->nfs_totalNumOfRespRcvdFromFs);	
	
    printf("Current NFP/NFS requests threshold:             %d (%s)\n",
		nfs_globalCfgp->nfs_maxReqsThreshold,
		(nfs_globalCfgp->nfs_maxReqsThreshold == 0)? "Disabled" : "Enabled");
    printf("Number of FS outstanding requests:              %lld\n",
          (nfs_globalCfgp->nfs_totalNumOfReqXmtToFs - nfs_globalCfgp->nfs_totalNumOfRespRcvdFromFs));
    printf("Number of NFP/NFS req control blocks in used:   %d\n",
          (NFP_MAX_REQUEST - nfp_globalCfgp->reqCbCfgList.freeEntriesCnt));
    printf("Num of dropped rcv'd packets-reach max requests threshold:    %lld\n",
                nfs_globalCfgp->nfs_debugPktCounters.nfs_maxReqsDropPktErr);

		
}


boolean	nfsv2_writeReqType[NFS_MAX_V2PROC] = {
		FALSE,		/*0: NULL*/
		FALSE,		/*1: GETATTR*/
		FALSE,		/*2: SETATTR*/
		FALSE,		/*3: ROOT: Obsolete*/
		FALSE,		/*4: LOOKUP*/
		FALSE,		/*5: READLINK*/
		FALSE,		/*6: READ*/
		FALSE,		/*7: WRITECACHE: Not Used*/
		TRUE,		/*8: WRITE*/
		TRUE,		/*9: CREATE*/
		TRUE,		/*10: REMOVE*/
		TRUE,		/*11: RENAME*/
		TRUE,		/*12: LINK*/
		TRUE,		/*13: SYMLINK*/
		TRUE,		/*14: MKDIR*/
		TRUE,		/*15: RMDIR*/
		FALSE,		/*16: READDIR*/
		FALSE		/*17: STATFS*/
};

boolean	nfsv3_writeReqType[NFS_MAX_V3PROC] = {
		FALSE,		/*0: NULL*/
		FALSE,		/*1: GETATTR*/
		FALSE,		/*2: SETATTR*/
		FALSE,		/*3: LOOKUP*/
		FALSE,		/*4: ACCESS*/
		FALSE,		/*5: READLINK*/
		FALSE,		/*6: READ*/
		TRUE,		/*7: WRITE*/
		TRUE,		/*8: CREATE*/
		TRUE,		/*9: MKDIR*/
		TRUE,		/*10: SYMLINK*/
		TRUE,		/*11: MKNOD*/
		TRUE,		/*12: REMOVE*/
		TRUE,		/*13: RMDIR*/
		TRUE,		/*14: RENAME*/
		TRUE,		/*15: LINK*/
		FALSE,		/*16: READDIR*/
		FALSE,		/*17: READDIRPLUS*/
		FALSE,		/*18: FSSTAT*/
		FALSE,		/*19: FSINFO*/
		FALSE,		/*20: PATHCONF*/
		TRUE		/*21: COMMIT*/
};


/*-----------------------------------------------------------------
 * Name:   	nfs_isWriteReqType 
 *
 * Description:
 *		Check if the request is NFS write type request.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_isWriteReqType( int32 nfs_version, int32 procNum )
{
	if (nfs_version == NFS_VER3)
		return (nfsv3_writeReqType[procNum]);
	else
		return (nfsv2_writeReqType[procNum]);
}

/*-----------------------------------------------------------------
 * Name:       	nfs_checkClientPermission 
 *
 * Description:
 *		Validate current request against client permission pertaining
 *		to current volume.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_checkClientPermission(nfs_request_t *nfsReqp)
{
	pkt_desc_t 			*descp;
	nfp_clientCBPtr_t 	clientCBp;
	rpc_authSysInfo_t   *authSysInfop;
	nfp_fileSysCBPtr_t	fileSysCBp;
	rpcCallHdr_t        *rpcCallHdrp;
	char				*nfs_msgp;
	int32				optionMask;
	int32				length;
	int32				error;

	error = NFSERR_SERVERFAULT;
	if (nfsReqp == NULL)
		return (error);


	rpcCallHdrp = nfsReqp->rpcCallHdrp;
	nfs_msgp = nfsReqp->procStart;

	//For NFS V2 client, set default
	//error to NFSERR_ACCES 
	if (nfsReqp->rpcCallHdrp && 
		(nfsReqp->rpcCallHdrp->progVer < NFS_VER3) )
		error = NFSERR_ACCES;

	/*
	 *Extract Filehandle
	 */
	length = nfs_extractFhandle(rpcCallHdrp->progVer, 
								nfs_msgp, &nfsReqp->fileHandle);
	if (length == NFX_ERR)
	{
        //return (NFX_ERR);
        return (NFSERR_BADHANDLE);
    }
#ifdef NFS_DBG_PRINTF_ON
	printf("%s:volId: %lld inode:%lld\n", 
		__FUNCTION__, nfsReqp->fileHandle.volId, nfsReqp->fileHandle.inum);
#endif


	authSysInfop = &rpcCallHdrp->authFlavor.authSys;

	clientCBp =  (nfp_clientCBPtr_t)nfsReqp->clientCBp;

	fileSysCBp =  mnt_findFileSysCB(clientCBp,  &nfsReqp->fileHandle);
	if ( (fileSysCBp == NULL) || 
		 ( fileSysCBp && (fileSysCBp->nfp_flag & NFP_SHARE_INIT_STATE)) )
	{
		nfs_globalCfgp->nfs_debugPktCounters.nfs_fileSysNotMnt++;
		if (mnt_getMountdInfoReqById(nfsReqp, &nfsReqp->fileHandle) != NFS_OK)
		{
			nfs_globalCfgp->nfs_debugPktCounters.nfs_getFileSysByIdErr++;
			return (error);	
		}
		return (NFS_FILESYS_NOT_READY);
	}
	//update access time for this mount point
	fileSysCBp->atime = nfs_utilGetTimeOfDayInSecs();	

	/*
	 *Check if NFS pkt Received on the same interface as the 
	 *orginal Mount request
	 */
	descp = nfsReqp->rxPktDesc;
	if (descp && fileSysCBp->laddr && 
		(descp->laddr != fileSysCBp->laddr))
	{
#ifdef NFS_DBG_PRINTF_ON
		printf("%s: Rcvd NFS request on the Invalid interface (0x%08x)!\n", 
				__FUNCTION__, descp->laddr);
#endif
		nfs_globalCfgp->nfs_debugPktCounters.nfs_rxPktIfErr++;
		return (error);
	}
	/*
	 *Check if the export is disabled or not
	 */
	if (!(fileSysCBp->nfp_flag & NFP_SHARE_ENABLED))
	{
		nfs_globalCfgp->nfs_debugPktCounters.nfs_rxPktNfsShareErr++;
		mnt_globalCfgp->mnt_statCounters.mnt_errFhandle++;
		return (NFSERR_STALE);
	}
	/*
	 *Check client permission against exports
	 */
	//optionMask =  fileSysCBp->protocol.nfp_nfsFileSysInfo.exportOptionMask;
	//Get client Export option mask
	optionMask = mnt_getExportOptMask(fileSysCBp, clientCBp->nfp_ipAddr);

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: fileSysCBp: %p optionMask: %08x\n", 
			__FUNCTION__, fileSysCBp, optionMask);
#endif

	if (!optionMask)
	{
		/*
		 *TED1284: Filehandle is longer valid. 
		 *return stale handle (70) error;
		 */
		mnt_globalCfgp->mnt_statCounters.mnt_errFhandle++;
		return (NFSERR_STALE);
	}

	if (nfs_isWriteReqType(rpcCallHdrp->progVer,
							rpcCallHdrp->rpc_proc) == FALSE)
	{
		/*
		 *Read-only command.  
		 */
		if (!(optionMask & MOUNTD_RO_OPT) && 
			!(optionMask & MOUNTD_RW_OPT) )	
		{
#ifdef NFS_DBG_PRINTF_ON
			printf("%s:  client [0x%08x] has no read permission!\n",
				__FUNCTION__, descp->raddr);
#endif
			mnt_globalCfgp->mnt_statCounters.mnt_errAccess++;
			return (NFSERR_ACCES);
		}
	}
	else
	{
		/*
		 *Check if client has write permission
		 */
		if (!(optionMask & MOUNTD_RW_OPT))
		{
#ifdef NFS_DBG_PRINTF_ON
			printf("%s:  client [0x%08x] has no write permission!\n",
				__FUNCTION__, descp->raddr);
#endif
			mnt_globalCfgp->mnt_statCounters.mnt_errAccess++;
			return (NFSERR_ACCES);
		}
	}
	if (optionMask & MOUNTD_ROOT_SQUASH_OPT)
	{
		/*
		 *root_squash.
		 *map UID/GUI to nobody
		 */
		if (authSysInfop->userId == 0)
		{
			authSysInfop->userId = RPC_UID_NOBODY;
		}
		if (authSysInfop->groupId == 0)
		{
			authSysInfop->groupId = RPC_UID_NOBODY;
		}
#ifdef NFS_DBG_PRINTF_ON
		printf("%s: root_squash: authSysInfop: %p UID: %d GID: %d\n",
			__FUNCTION__, 
			authSysInfop, authSysInfop->userId, authSysInfop->groupId);
#endif
		
	}

	return (NFS_OK);
}

int32 nfs_getNumAttrFields(int32 nfsv3_proc)
{
	switch (nfsv3_proc)
	{
    	case NFS_V3_PROC_NULL:
		case NFS_V3_PROC_GETATTR:
			return (0);
		case NFS_V3_PROC_LOOKUP:
		case NFS_V3_PROC_ACCESS:	
		case NFS_V3_PROC_READLINK:
		case NFS_V3_PROC_READ:
		case NFS_V3_PROC_READDIR:
		case NFS_V3_PROC_READDIRPLUS:
		case NFS_V3_PROC_FSSTAT:
		case NFS_V3_PROC_FSINFO:
		case NFS_V3_PROC_PATHCONF:
			return (1);
		case NFS_V3_PROC_SETATTR:
		case NFS_V3_PROC_WRITE:
		case NFS_V3_PROC_CREATE:
		case NFS_V3_PROC_MKDIR:
		case NFS_V3_PROC_MKNOD:
		case NFS_V3_PROC_SYMLINK:
		case NFS_V3_PROC_REMOVE:
		case NFS_V3_PROC_RMDIR:
		case NFS_V3_PROC_COMMIT:
			return (2);
		case NFS_V3_PROC_LINK:
			return (3);
		case NFS_V3_PROC_RENAME:
			return (4);
		default:
			return (0);
	}
	return (0);
}
/*-----------------------------------------------------------------
 * Name:       	nfs_checkMntPointActivity 
 *
 * Description:
 * Check for activities on File System Control Blocks (mount points). 
 * If there are no activity after certain time period,  
 * the File System Control Block will be removed. 
 *  
 * Created by:  Joseph Vu
 *  
 * Date Created:        4/25/02
 *
 *-----------------------------------------------------------------
 */ 
void nfs_checkMntPointActivity (void)
{
	nfp_nodePtr_t			nodep;
	nfp_list_t				*nodeListp;
	nfp_clientCBPtr_t   	clientCBp;
    nfp_fileSysCBPtr_t      fileSysCBp;
    int32                   count;
    uint32                  currTimeInSecs;
    uint32                  deltaSecs;
	int32					nodeCount;

    count = 0;
    currTimeInSecs = nfs_utilGetTimeOfDayInSecs();

	if (currTimeInSecs < 0)
	{
		return;
	}
    for (fileSysCBp = (nfp_fileSysCBPtr_t)nfp_globalCfgp->fileSysCbCfgList.listBasep;
        ((count < NFP_MAX_FILESYS) && (fileSysCBp != NULL));
        count++, fileSysCBp++)
    {
        //check if MNT point is in used
        if ( (!(fileSysCBp->nfp_flag & NFP_ENTRY_INUSED)) ||
			 (fileSysCBp->atime < 0) )
            continue;
        deltaSecs = currTimeInSecs - fileSysCBp->atime;

        //If no activy on the mount point for
        //certain time period
        //if (deltaSecs > (SEC_PER_MIN * 10))

		if (deltaSecs > NFS_MNT_PNT_INACT_PERIOD)
        {
            //mark for delete
            fileSysCBp->nfp_flag |= NFP_ENTRY_MARK_FOR_DEL;
        }
    }

    //Check the File System CB (Mnt point)
	//need to be remove from the clientCB list 
    count = 0;
    for (clientCBp= (nfp_clientCBPtr_t)nfp_globalCfgp->clientCbCfgList.listBasep;
        ((count < NFP_MAX_CLIENT) && (clientCBp != NULL)); count++, clientCBp++)
    {
        if ((!(clientCBp->nfp_flag & NFP_ENTRY_INUSED)) ||
            (!clientCBp->nfp_ipAddr) )
            continue;
        nodeListp = &clientCBp->nfp_fileSysList;
        nodeCount = nodeListp->nfp_reqCnt;
                                    //It should be set to NFP_MAX_FILESYS
        for (nodep = nodeListp->nfp_headp; (nodep != NULL && nodeCount > 0);
            nodep = nodep->nfp_nextp, nodeCount--)
        {
            fileSysCBp = (nfp_fileSysCBPtr_t)nodep->nfp_entryCBp;
            if ( (nfp_validateFileSysCB(fileSysCBp) != NFX_OK) ||
                (!(fileSysCBp->nfp_flag & NFP_ENTRY_MARK_FOR_DEL)) )
                continue;

            nfs_delFileSysCBFromList(fileSysCBp, clientCBp);
        }
    }

	count = 0;
    for (fileSysCBp = (nfp_fileSysCBPtr_t)nfp_globalCfgp->fileSysCbCfgList.listBasep;
        ((count < NFP_MAX_FILESYS) && (fileSysCBp != NULL));
        count++, fileSysCBp++)
    {
        //check if MNT point is mark for delete
        if ( (nfp_validateFileSysCB(fileSysCBp) != NFX_OK) ||
            (!(fileSysCBp->nfp_flag & NFP_ENTRY_MARK_FOR_DEL)) )
            continue;
        nfp_deallocFileSysCB(fileSysCBp);
    }

    return;
}


/*-----------------------------------------------------------------
 * Name:       	nfs_errMsg 
 *
 * Description:
 *	Handle request error messages.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
void nfs_errMsg(uint32 xid, uint32 procNum,  int32 error)
{

	switch (error)
	{
		case NFSERR_BADHANDLE:
			printf("RX: Invalid filehandle from Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break; 	

		case NFSERR_FILENAME_TLONG:
			printf("RX: Filename too long from Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFSERR_ADD_FILECB:
			printf("RX: Failure to add FileCB from Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;

		case NFSERR_MAP_FS_MSG:
			printf("RX: Failure to map to File Sys. Msg from Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFSERR_SENDMSG_TO_FS:
			printf("RX: Failure to send msg to File System. Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFS_RPC_MISMATCH:
			printf("RX: RPC version number mismatch.  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFS_RPC_PROG_MISMATCH:
			printf("RX: Program version number mismatch.  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFS_RPC_PROC_UNAVAIL:
			printf("RX: Procedure number is unavailable.  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFSERR_NOT_SYNC:
			printf("RX: Update synch mismatch was detected during Setattr.  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFSERR_SETATTR:
			printf("RX: Failure to retrieve attributes (Setattr) .  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFSERR_BAD_COOKIE:
			printf("RX: Failure to retrieve READDIR/READDIRPLUS cookie field .  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		case NFSERR_BADTYPE:
			printf("RX: Invalid Special Device ftype.  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;
		default:
			printf("RX: Unknown Error.  Req XID: %d Proc Num: %d\n",
				xid, procNum);
			break;			
	}


}

int32 nfs_removeCacheEntry(nfp_fileCBPtr_t	fileCBp)
{
	nfs_fileInfo_t      *nfs_fileInfop;

	nfs_fileInfop = &fileCBp->protocol.nfp_nfsFileInfo;
	nfs_fileInfop->nfs_flag = 0;

	if ((nfs_fileInfop->nfs_nameLen != 0) &&
		(nfs_fileInfop->nfs_namep) )
	{
		//printf("nfs_removeCacheEntry: 1: delloc nfs_namep: %s len: %d\n", 
		//						nfs_fileInfop->nfs_namep, nfs_fileInfop->nfs_nameLen);

#ifdef  NFS_USE_SHMEM
		eee_deallocateBuffer((void *)nfs_fileInfop->nfs_namep);
#else
		eee_ramDealloc((uchar8 *)nfs_fileInfop->nfs_namep);
#endif
		nfs_fileInfop->nfs_nameLen = 0;
		nfs_fileInfop->nfs_namep = NULL;

	}
	if (nfs_fileInfop && nfs_fileInfop->nfs_fileNamep)
	{
		//printf("nfs_removeCacheEntry: 2: delloc nfs_namep: %s len: %d\n", 
		//						nfs_fileInfop->nfs_fileNamep, nfs_fileInfop->nfs_filenameLen);
#ifdef  NFS_USE_SHMEM
		eee_deallocateBuffer(nfs_fileInfop->nfs_fileNamep);
#else
		eee_ramDealloc((uchar8 *)nfs_fileInfop->nfs_fileNamep);
#endif
		nfs_fileInfop->nfs_filenameLen = 0;
		nfs_fileInfop->nfs_fileNamep = NULL;

	}
	
	nfs_delFileCBFromList (fileCBp,
		(nfp_fileCBPtr_t)fileCBp->nfp_parentDirFileCBp, NULL);

	return (NFX_OK);
}

int32 nfs_removeAllCacheEntries(nfp_list_t *fileListp)
{
	nfp_nodePtr_t           delNodep;
	nfp_fileCBPtr_t         fileCBp;
	int32                   count; 

	if (fileListp == NULL)
		return (NFX_ERR);
	count = fileListp->nfp_reqCnt;
	for (delNodep = fileListp->nfp_headp;
		((delNodep != NULL) && (count > 0));
		delNodep = delNodep->nfp_nextp, count--)
	{
		fileCBp = (nfp_fileCBPtr_t)delNodep->nfp_entryCBp;
		if (nfp_validateFileCB(fileCBp) != NFX_OK)
			continue;
		nfs_removeCacheEntry(fileCBp);
	}
    
    return (NFX_OK);
}
/*-----------------------------------------------------------------
 * Name:       	nfs_clearCache 
 *
 * Description:
 *  Clear NFS cache. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
void nfs_clearCache(void)
{
	int32				count;
	nfp_fileCBPtr_t 	fileCBp;
	nfs_fileInfo_t      *nfs_fileInfop;

	if (!cacheEnable)
		return;

	count = 0;
	for (fileCBp = (nfp_fileCBPtr_t)nfp_globalCfgp->fileCbCfgList.listBasep;
		((count < NFP_MAX_FILE) && (fileCBp != NULL)); count++, fileCBp++)
	{
		if (!(fileCBp->nfp_flag & NFP_ENTRY_INUSED))
			continue;
		if (fileCBp->nfp_dirCacheFileList.nfp_reqCnt != 0)
		{
			//nfsDelAllFileCBFromDir(&fileCBp->nfp_dirCacheFileList);	
			nfs_removeAllCacheEntries(&fileCBp->nfp_dirCacheFileList);
		}
		nfs_fileInfop = &fileCBp->protocol.nfp_nfsFileInfo;
		nfs_fileInfop->nfs_flag = 0;
		
		if ((nfs_fileInfop->nfs_nameLen != 0) && 
			(nfs_fileInfop->nfs_namep) )
		{
			//printf("nfs_clearCache 2: de-alloc cache name\n");
#ifdef  NFS_USE_SHMEM
			eee_deallocateBuffer((void *)nfs_fileInfop->nfs_namep);
#else
			eee_ramDealloc((uchar8 *)nfs_fileInfop->nfs_namep);
#endif
			nfs_fileInfop->nfs_nameLen = 0;
			nfs_fileInfop->nfs_namep = NULL;

		}
		//Check if cache entry need to deleted
		if ( nfs_fileInfop->fs_attr.atime.secs == nfs_fileInfop->cacheTime.secs)
		{
			//printf("nfs_clearCache: delele fileCBp: %p\n", fileCBp);
			nfs_delFileCBFromList (fileCBp, 
				(nfp_fileCBPtr_t)fileCBp->nfp_parentDirFileCBp,
				NULL);
			nfp_delFileCB (&fileCBp->nfp_fileHandle);
		}
    }
}

int32 nfs_trace(
        int32   argc,
        char8   * argv[]
)
{
	if (argc != 3)
		return (NFX_ERR);

	if (strcmp(argv[1], "nfs") != 0)	
	{
		return (NFX_ERR);
	}

	if (strcmp(argv[2], "on") == 0)
	{
		trace_enable = 1;
		printf("NFS trace has been turned on.\n");
	}
	else if (strcmp(argv[2], "off") == 0)
	{
		trace_enable = 0;
		printf("NFS trace has been turned off\n");
	}
	else
	{
		printf("Invalid option\n");
		return (NFX_ERR);
	}


	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	nfs_findPconnOnClientList 
 *
 * Description:
 *	Returns the index of the found TPL Pconn (Transport Layer Connection)
 *	of the Client Control Block list.  Othersize, return -1 or NFX_ERR if
 *  if not found.
 *		
 * Created by:  Joseph Vu
 *
 * Date Created:        3/01/02
 *
 *-----------------------------------------------------------------
 */
#ifdef	NOTUSED
int32	//index found
nfs_findPconnOnClientList(
	nfp_clientCBPtr_t   clientCBp, 
	void *pConnId
)
{
	int		i;

	for (i =0; (i < clientCBp->connIdIndex && i < CLIENT_MAX_NUM_CONN_ID); i++)
	{
		if ((clientCBp->connId[i]) && (clientCBp->connId[i] == (uint32)pConnId))
		{
			NFS_DBG_PRINTF("nfs_tplDelConnInd: index: %d del conn  %p\n", i, clientCBp->connId[i]);
            //clientCBp->connId[i] = NULL;
			return (i);
        }
    }

	return (NFX_ERR);

}

void nfs_checkNfsTplPconn()
{
	nfp_fileSysCBPtr_t  fileSysCBp;
	int32				count;
	count = 0;

	//printf("nfs_checkNfsTplPconn: Enter\n");	
    for (fileSysCBp = (nfp_fileSysCBPtr_t)nfp_globalCfgp->fileSysCbCfgList.listBasep;
        ((count < NFP_MAX_FILESYS) && (fileSysCBp != NULL)); 
			count++, fileSysCBp++)    
	{
        if (nfp_validateFileSysCB(fileSysCBp) != NFX_OK) 
            continue;
		mnt_pntCleanUp(fileSysCBp);
	}

	return;
}

#endif	//NOTUSED

//------------------------------------------------------------------------------------------------------
//For debugging only
#ifdef	NOTUSED
int32 nfs_display_dirfiles(nfp_fileCBPtr_t dirFileCBp)
{
	int32			nodeCount;
    int32           count;
    nfp_fileCBPtr_t currp;
	nfp_fileCBPtr_t	fileCBp;
	nfp_list_t 		*nodeListp;
	nfp_nodePtr_t	nodep;
	

	nodeListp = &dirFileCBp->nfp_dirFileList;
	nodeCount = nodeListp->nfp_reqCnt;
	printf("parent fileCBp: %p volId: %lld inode: %lld\n", 
		dirFileCBp, dirFileCBp->nfp_fileHandle.volId.whole_thing, 
		dirFileCBp->nfp_fileHandle.inum);
	printf("name: %s len: %d\n",
		dirFileCBp->protocol.nfp_nfsFileInfo.nfs_fileNamep,
		dirFileCBp->protocol.nfp_nfsFileInfo.nfs_filenameLen);	

	if (nodeCount)
	{
		printf("nfp_dirFileList: cnt: %d\n", 
				dirFileCBp->nfp_dirFileList.nfp_reqCnt);
			
		for (nodep = nodeListp->nfp_headp;
			((nodep != NULL) && (nodeCount > 0));
			nodep = nodep->nfp_nextp, nodeCount--)
		{
			fileCBp = (nfp_fileCBPtr_t)nodep->nfp_entryCBp;
			if (fileCBp == NULL)
				continue;
			printf("nodeCount: %d filename: %s nameLen: %d fileCBp: %p\n",
				nodeCount, fileCBp->protocol.nfp_nfsFileInfo.nfs_fileNamep,
				fileCBp->protocol.nfp_nfsFileInfo.nfs_filenameLen, fileCBp);	
		}
		

	}
    
}

int32 nfs_display_fileCB(void)
{
	int32			nodeCount;
    int32           count;
    nfp_fileCBPtr_t currp;
	nfp_fileCBPtr_t	fileCBp;
	nfp_list_t 		*nodeListp;
	nfp_nodePtr_t	nodep;
	

    count = 0;
    for (currp = (nfp_fileCBPtr_t)nfp_globalCfgp->fileCbCfgList.listBasep;
        ((count < NFP_MAX_FILE) && (currp != NULL)); count++, currp++)
    {
        if (!(currp->nfp_flag & NFP_ENTRY_INUSED))
            continue;
        //if (currp->nfp_fileHandle.inum == searchInode)
		//            return (currp);
		nodeListp = &currp->nfp_dirFileList;
		nodeCount = nodeListp->nfp_reqCnt;
		printf("parent fileCBp: %p volId: %lld inode: %lld\n", 
			currp, currp->nfp_fileHandle.volId.whole_thing, 
			currp->nfp_fileHandle.inum);
		printf("name: %s len: %d\n",
			currp->protocol.nfp_nfsFileInfo.nfs_fileNamep,
			currp->protocol.nfp_nfsFileInfo.nfs_filenameLen);	

		if (nodeCount)
		{
			printf("nfp_dirFileList: cnt: %d\n", 
					currp->nfp_dirFileList.nfp_reqCnt);
			
			for (nodep = nodeListp->nfp_headp;
				((nodep != NULL) && (nodeCount > 0));
				nodep = nodep->nfp_nextp, nodeCount--)
			{
				fileCBp = (nfp_fileCBPtr_t)nodep->nfp_entryCBp;
				if (fileCBp == NULL)
					continue;
				printf("nodeCount: %d filename: %s nameLen: %d\n",
					nodeCount, fileCBp->protocol.nfp_nfsFileInfo.nfs_fileNamep,
					fileCBp->protocol.nfp_nfsFileInfo.nfs_filenameLen);	
			}
			

		}
    }

	return (NFX_OK);
}
#endif	//NOTUSED

int32
nfs_dbgCod( nfs_request_t	*nfsReqp, 
			uint32 			numOfBytes)
{
    int32           cod_totalNumOfBytes;
    int32           codIdx;
    uint32          clientIpAddr;
	pkt_desc_t		*descp;
    nfx_cod_entry_t *tempCod;
    nfp_clientCBPtr_t   clientCBp;


	if ((nfsReqp->nbrUsedCod > NFX_MAX_COD_MSG) && (numOfBytes > 0))
	{
		printf("%s: Too many CODs=%d numBytes=%d xid=0x%08x!\n", 
				__FUNCTION__,
				nfsReqp->nbrUsedCod, 
				numOfBytes, 
				nfsReqp->rpcCallHdrp->rpc_xid);

		goto cod_err;
	}

	if ( (nfsReqp->nbrUsedCod != 0) && (numOfBytes <= 0) )
	{
		printf("%s: Invalid COD number=%d numBytes=%d xid=0x%08x!\n", 
				__FUNCTION__,
				nfsReqp->nbrUsedCod, 
				numOfBytes, 
				nfsReqp->rpcCallHdrp->rpc_xid);

		goto cod_err;

	}

	if (numOfBytes)
	{
		cod_totalNumOfBytes = 0;
		tempCod = &nfsReqp->cod[0];
        for (codIdx =0; codIdx < nfsReqp->nbrUsedCod; codIdx++)
        {
            cod_totalNumOfBytes +=  tempCod->length;
            tempCod++;
        }
		if (numOfBytes != cod_totalNumOfBytes)
		{
			printf("%s: numOfBytes: %d cod_totalNumOfBytes: %d numOfCod: %d\n",
					__FUNCTION__,
					numOfBytes,
					cod_totalNumOfBytes,
					nfsReqp->nbrUsedCod);

			goto cod_err;
		}
	}

	return (NFX_OK);

cod_err:

    clientCBp = nfsReqp->clientCBp;

    if (clientCBp)
	{
		descp = (pkt_desc_t *)nfsReqp->rxPktDesc;

        clientIpAddr = ntohl(clientCBp->nfp_ipAddr);
		printf("%s: clientIP: %d.%d.%d.%d NFS_VER: %d TPL PROT: %s\n",
				__FUNCTION__,
				((clientIpAddr >> 24) & 0xff),
				((clientIpAddr >> 16) & 0xff),
				((clientIpAddr >> 8) & 0xff),
				(clientIpAddr & 0xff),	
				nfsReqp->rpcCallHdrp->progVer,  
				(descp->tplProt == TPL_PROT_UDP)? "UDP" : "TCP");
	}
	return (NFX_ERR);
}
