/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:			nlm-utils.c
 *
 * RCS:  $Id: nlm-utils.c,v 1.6 2002/11/06 03:35:39 joseph Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: This file contains NLM utility functions.
 *
 * Created by:		Xiaobing Zhang
 *
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef	NFP_TXRX
#include <nfx-incl.h>
#include <exports/vs-mem-api.h>
#include <exports/lock-api.h>
#include <exports/nfs-api.h>
#include "nlm-prototypes.h"

#define NLM_DBG_PRINTF if(nlm_traceEnable) printf

extern nfp_nodePtr_t nfs_addNode(void *CtrlBlkp, int32 nodeType);
extern int32 nfs_addFileCBToList(nfp_fileCBPtr_t	fileCBp,
								 fs_Handle			*fhandlep,
								 nfp_clientCBPtr_t	clientCBp);
extern int32 nfs_delCBFromList(nfp_list_t *nodeListp,
							   void *CtrlBlkp, int32 nodeType);


#define NLM_DBG_FUNC_PRINTF if (nlm_traceDebugEnable) printf
//#define	NLM_DEBUG

/*-----------------------------------------------------------------
 * Name:		nlm_getFileCB
 *  
 * Description: Get a File Control Block (FCB) given file handle.
 *              If there is no FCB related to the file handle,
 *              create one.
 *
 * Created by:	Xiaobing Zhang
 *  
 * Date Created:	07/16/2002
 *
 *-----------------------------------------------------------------
 */
nfp_fileCB_t *
nlm_getFileCB(fs_Handle *fileHandlep, nfp_clientCB_t *clientCBp)
{
	nfp_fileCB_t	*fileCBp;

	if ((fileHandlep == NULL) || (clientCBp == NULL))
		return (NULL);

	fileCBp = nfp_getFileCB(fileHandlep);
	if (fileCBp == NULL)
	{
		fileCBp = nfp_allocFileCB();
		if (fileCBp)
			nfs_addFileCBToList(fileCBp, fileHandlep, clientCBp);
	}

	return (fileCBp);
}

/*-----------------------------------------------------------------
 * Name:		nlm_regionCmp
 *  
 * Description: Compare positions of two locked regions. The two regions
 *              may be overlapping, or one region has lower address in the
 *              file than the other.
 *    Here are all the possible cases for overlapping:
 *    (1).                           (2).
 *      start1      end1                   start1      end1
 *        *----------*                       *----------*
 *           *----------*                *----------*
 *         start2      end2            start2      end2
 *
 *    (3).                           (4).
 *      start1      end1                 start1  end1
 *        *----------*                       *----*
 *           *----*                       *----------*
 *       start2  end2                  start2      end2
 *
 *    (5).                           (6).
 *      start1  end1                      start1  end1
 *          *----*                            *----*
 *          *----------*                *----------*
 *        start2      end2             start2      end2
 *
 *    (7).                           (8).
 *        start1      end1             start1      end1
 *          *----------*                *----------*
 *          *----*                            *----*
 *      start2  end2                      start2  end2
 *
 *    (9).
 *      start1      end1
 *        *----------*
 *        *----------*
 *      start2      end2
 *
 *    (10).                           (11).
 *      start1  end1                          start1  end1
 *          *----*                                *----*
 *               *----*                      *----*
 *           start2  end2                start2  end2
 *
 *    The common feature of all the cases is 
 *     (start1 <= end2) && (end1 >= start2)
 *
 * Created by:	Xiaobing Zhang 
 *
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NOT_USED
int32			/*-1, if lock1's region has lower address than lock2's region
				  0, if lock1's region and lock2's region overlap
				  1, if lock1's region has higher address than lock2's region*/
nlm_regionCmp(nlm_lockRec_t *lock1p,
			  nlm_lockRec_t *lock2p)
{
	uint64	start1, start2, end1, end2;
	boolean	eof1, eof2;

	if ((lock1p == NULL) || (lock2p == NULL))
	{
		NLM_DBG_PRINTF("nlm_regionCmp: NULL pointer\n");
		return (-2);
	}

	start1 = lock1p->nlm_l_offset;	/*Start of locked region 1*/
	end1 = lock1p->nlm_l_end;		/*End of locked region 1*/
	eof1 = lock1p->nlm_l_eof;

	start2 = lock2p->nlm_l_offset;	/*Start of locked region 2*/
	end2 = lock2p->nlm_l_end;		/*End of locked region 2*/
	eof2 = lock2p->nlm_l_eof;

	if ((eof1 != TRUE) && (start2 > end1))
		return (-1);
	if ((eof2 != TRUE) && (start1 > end2))
		return (1);
	
	//Check for all overlap cases
	if ((start1 == start2) && (end1 == end2) )
		//overlap == lock
		return (0);
	if ((start1 <= start2) && (eof2 != TRUE) &&
		((end1 >= end2) || (eof1 == TRUE)))
		//overlaps contains lock
		return (0);
	if ((start2 <= start1) &&
		(eof2 == TRUE ||
		 (eof1 != TRUE && end2 >= end1)))
		//lock contains overlap
		return (0);
	if ((start1 < start2 ) &&
		((end1 >= start2) || (eof1 == TRUE)))
		//overlaps starts before lock
		return (0);
	if ((start1 > start2) &&
		(eof2 != TRUE) &&
		((end1 > end2) || (eof1 == TRUE)))
		//overlaps ends after lock
		return (0);

	return (-1);
}

/*-----------------------------------------------------------------
 * Name:		nlm_addLockCBToCurrList
 *
 * Description: Add a Lock Control Block (LCB) into lock list related 
 *              to a file control block.
 *              The lock list is arragned such that the locked region of
 *              one node occupies higher address than the locked region
 *              of the node prior to it in the list. For non-exclusive 
 *              overlapped regions, the order doesn't matter.
 *  
 * Created by:	Xiaobing Zhang
 *  
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
int32			/*NFX_OK:  Entry has been added*/
				/*NFX_ERR: Failure to add entry*/
nlm_addLockCBToCurrList(nfp_fileCB_t *fileCBp,
						nfp_node_t *currLockNodep, /*Add in front of this LCB*/
						nfp_lockCB_t *addLockCBp)
{
	nfp_list_t		*lockListp;
	nfp_nodePtr_t	addNodep;

	if ((fileCBp == NULL) || (addLockCBp == NULL))
        return (NFX_ERR);

	lockListp = &fileCBp->nfp_lockList;
    addNodep = nfs_addNode((void *)addLockCBp, NFP_LOCKCB_NODE);
    if (addNodep == NULL)
        return (NFX_ERR);


    if (lockListp->nfp_headp == NULL)
    {
        /*
		 *Add to empty list
		 */
        lockListp->nfp_headp = addNodep;
        lockListp->nfp_tailp = addNodep;

    }
	else if (currLockNodep == NULL)
	{
		/*
		 *Add to the end of the list
		 */
		lockListp->nfp_tailp->nfp_nextp = addNodep;
		addNodep->nfp_prevp = lockListp->nfp_tailp;
		lockListp->nfp_tailp = addNodep;

	}
	else
	{
        /*
		 *Insert in front of currLockNodep
		 */
		addNodep->nfp_nextp = currLockNodep;
		addNodep->nfp_prevp = currLockNodep->nfp_prevp;
		if (lockListp->nfp_headp == currLockNodep)
		{
			lockListp->nfp_headp = addNodep;
		}
		else
		{
			currLockNodep->nfp_prevp->nfp_nextp = addNodep;
		}
		currLockNodep->nfp_prevp = addNodep;

	}

    lockListp->nfp_reqCnt++;

    return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:	nlm_addLockRecToCurrList		
 *
 * Description: Add a Lock record into current list
 *              related to a lock owner Lock Control 
 *              Block (LCB).
 *  
 * Created by:	Joseph Vu	
 *  
 * Date Created:	10/20/2002
 *
 *-----------------------------------------------------------------
 */

nlm_lockRec_t *
nlm_addLockRecToCurrList(nfp_clientCBPtr_t	clientCBp,
						nfp_fileCBPtr_t		fileCBp, 
						nlm_lockInfo_t		*nlm_lockInfop,
						nlm_lockRec_t 		*next_lockRecp,
						int32				newLockOwner,
						fs_Handle 			*fhandlep)
{
	nfp_lockCB_t	*lockCBp;
	nlm_lockRec_t	*lockRecp;
	int32			rc;

	lockRecp = NULL;

	if (!newLockOwner)
	{
		lockCBp = nlm_getLockCBOwner(	fileCBp,
										nlm_lockInfop);

		if (lockCBp)
		{
			/*------------------------------------------
	 	 	 *We had seen this lock owner before,
	 	 	 *Create new lock record and add to the
	 	 	 *lock owner pending list
 	 	 	 *-------------------------------------------
 	 	 	 */

			lockRecp = 	nlm_AllocLockRec(NLM_LOCKREC_CURR_LIST);
			if (lockRecp == NULL)
				return (NULL);	

			lockRecp->nlm_l_offset = nlm_lockInfop->lockRec.nlm_l_offset;
			lockRecp->nlm_l_len = nlm_lockInfop->lockRec.nlm_l_len;
			NLM_GET_L_REGION_END(lockRecp->nlm_l_end,
							lockRecp->nlm_l_offset,
							lockRecp->nlm_l_len,
							lockRecp->nlm_l_eof);   
			lockRecp->nlm_exclusive = 
				nlm_lockInfop->lockRec.nlm_exclusive;


			goto nlm_addLockRecToCurrListExit;
		}

	}

	lockCBp = nfp_allocLockCB();
	if (lockCBp == NULL)
	{
		return (NULL);
	}

	lockRecp = 	nlm_AllocLockRec(NLM_LOCKREC_CURR_LIST);
	if (lockRecp == NULL)
	{
		nfp_deallocLockCB(lockCBp);
		return (NULL);						
	}
			
	rc = nlm_saveLockInfo(	lockCBp,
							lockRecp,
							nlm_lockInfop);
	if (rc != NFX_OK)
	{
		nlm_deallocLockRec(lockRecp);
		nfp_deallocLockCB(lockCBp);
		return (NULL);	
	}

	lockCBp->nfp_fileCbp = fileCBp;	
	lockCBp->nfp_clientCbp = clientCBp;

	if (newLockOwner && fhandlep)
	{
		rc = nlm_addNewLock (fileCBp, 
							lockCBp, 
							NULL,		
							clientCBp, 
				   			fhandlep);
	}
	else
	{
		rc = nlm_addNewLock (fileCBp, 
							lockCBp, 
							NULL,		
							clientCBp, 
				   			&fileCBp->nfp_fileHandle);
	}
	if (rc != NFX_OK)
	{
		nlm_deallocLockRec(lockRecp);
		nfp_deallocLockCB(lockCBp);
		return (NULL);
	}
	

nlm_addLockRecToCurrListExit:

	/*---------------------------------
 	 *Add lockRec into the list
 	 *---------------------------------
 	 */						
	if (nfp_validateLockCB(lockCBp) != NFX_OK)
	{
		panic("%s: Bad lockCBp: %p\n", __FUNCTION__, lockCBp);		
	}

	if (next_lockRecp &&
		(next_lockRecp->flags & NLM_LOCKREC_CURR_LIST) )
	{
		DLL_INSERT_BEFORE(&lockCBp->lockOwnerList.list,
						next_lockRecp, lockRecp, lockList);
	}
	else
	{
		DLL_INSERT_TAIL(&lockCBp->lockOwnerList.list,
						lockRecp, lockList);
	}

	lockCBp->lockOwnerList.numOfEntries++;

	lockRecp->lockCBp = lockCBp;

	return (lockRecp);

}
#else
#if 0
int32			/*-1, if lock1's region has lower address than lock2's region
				  0, if lock1's region and lock2's region overlap
				  1, if lock1's region has higher address than lock2's region*/
nlm_regionCmp(nlm_lockInfo_t *lock1p,
			  nlm_lockInfo_t *lock2p)
{
	uint64	start1, start2, end1, end2;
	boolean	eof1, eof2;

	if ((lock1p == NULL) || (lock2p == NULL))
	{
		NLM_DBG_PRINTF("nlm_regionCmp: NULL pointer\n");
		return (-2);
	}

	start1 = lock1p->nlm_l_offset;	/*Start of locked region 1*/
	end1 = lock1p->nlm_l_end;		/*End of locked region 1*/
	eof1 = lock1p->nlm_l_eof;

	start2 = lock2p->nlm_l_offset;	/*Start of locked region 2*/
	end2 = lock2p->nlm_l_end;		/*End of locked region 2*/
	eof2 = lock2p->nlm_l_eof;
#ifdef	NOTUSED
	/*Normal comparison*/
	if ((start1 <= end2) && (end1 >= start2))
		return (0);
	else if (end1 < start2)
		return (-1);
	else /*start1 > end2*/
		return (1);
#endif	//NOTUSED

	if ((eof1 != TRUE) && (start2 > end1))
		return (-1);
	if ((eof2 != TRUE) && (start1 > end2))
		return (1);
	
	//Check for all overlap cases
	if ((start1 == start2) && (end1 == end2) )
		//overlap == lock
		return (0);
	if ((start1 <= start2) && (eof2 != TRUE) &&
		((end1 >= end2) || (eof1 == TRUE)))
		//overlaps contains lock
		return (0);
	if ((start2 <= start1) &&
		(eof2 == TRUE ||
		 (eof1 != TRUE && end2 >= end1)))
		//lock contains overlap
		return (0);
	if ((start1 < start2 ) &&
		((end1 >= start2) || (eof1 == TRUE)))
		//overlaps starts before lock
		return (0);
	if ((start1 > start2) &&
		(eof2 != TRUE) &&
		((end1 > end2) || (eof1 == TRUE)))
		//overlaps ends after lock
		return (0);

	return (-1);
}
#endif
#endif // NOT_USED


/*-----------------------------------------------------------------
 * Name:		nlm_addLockRecToPendingList	
 *
 * Description: Add a Lock record into pending list
 *              related to a LCB.
 *  
 * Created by:	Joseph Vu	
 *  
 * Date Created:	10/20/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NOT_USED
nlm_lockRec_t *
nlm_addLockRecToPendingList(nfp_clientCBPtr_t	clientCBp,
							nfp_fileCBPtr_t		fileCBp, 
							nlm_lockInfo_t		*nlm_lockInfop,
							nlm_lockRec_t 		*next_lockRecp)
{
	nfp_lockCB_t			*lockCBp;
	nlm_lockRec_t			*lockRecp;
	int32					rc;


	lockRecp = NULL;
	lockCBp = nlm_getLockCBOwner(	fileCBp,
									nlm_lockInfop);

	if (lockCBp)
	{

		/*------------------------------------------
	 	 *We had seen this lock owner before,
	 	 *Create new lock record and add to the
	 	 *lock owner pending list
 	 	 *-------------------------------------------
 	 	 */

		lockRecp = 	nlm_AllocLockRec(NLM_LOCKRED_PENDING_LIST);
		if (lockRecp == NULL)
			return (NULL);	

		lockRecp->nlm_l_offset = nlm_lockInfop->lockRec.nlm_l_offset;
		lockRecp->nlm_l_len = nlm_lockInfop->lockRec.nlm_l_len;
		NLM_GET_L_REGION_END(lockRecp->nlm_l_end,
							lockRecp->nlm_l_offset,
							lockRecp->nlm_l_len,
							lockRecp->nlm_l_eof);   
		lockRecp->nlm_exclusive = 
				nlm_lockInfop->lockRec.nlm_exclusive;


		goto nlm_addLockRecToPendingListExit;
	}


	lockCBp = nfp_allocLockCB();
	if (lockCBp == NULL)
	{
		return (NULL);
	}

	lockRecp = 	nlm_AllocLockRec(NLM_LOCKRED_PENDING_LIST);
	if (lockRecp == NULL)
	{
		nfp_deallocLockCB(lockCBp);
		return (NULL);						
	}
			
	rc = nlm_saveLockInfo(	lockCBp,
							lockRecp,
							nlm_lockInfop);
	if (rc != NFX_OK)
	{
		nlm_deallocLockRec(lockRecp);
		nfp_deallocLockCB(lockCBp);
		return (NULL);	
	}

	lockCBp->nfp_fileCbp = fileCBp;	
	lockCBp->nfp_clientCbp = clientCBp;

	rc = nlm_addNewLock (fileCBp, 
						lockCBp, 
						NULL,		
						clientCBp, 
			   			&fileCBp->nfp_fileHandle);
	
	if (rc != NFX_OK)
	{
		nlm_deallocLockRec(lockRecp);
		nfp_deallocLockCB(lockCBp);
		return (NULL);
	}
	

nlm_addLockRecToPendingListExit:

	/*---------------------------------
 	 *Add lockRec into the list
 	 *---------------------------------
 	 */	

	if ( (next_lockRecp) && 
		 (next_lockRecp->flags & NLM_LOCKRED_PENDING_LIST) )
	{

		DLL_INSERT_BEFORE(&lockCBp->lockOwnerPendingList.list, 
						next_lockRecp, lockRecp, lockList );

	}
	else
	{
		DLL_INSERT_TAIL(&lockCBp->lockOwnerPendingList.list,
						lockRecp, lockList);
	}			
	lockCBp->lockOwnerPendingList.numOfEntries++;	
	lockRecp->lockCBp = lockCBp;


	return (lockRecp);
}
#else
#if 0
int32
nlm_addLockRecToPendingList(nfp_fileCBPtr_t fileCBp,
                            nlm_lockInfo_t *nlm_lockInfop,
                            nlm_request_t *nlmReq)
{
    nlm_lockRecOwnerList_t *pendQ = NULL;
    nlm_pendingLockNode_t  *lockNode = NULL;

    lockNode = nlm_AllocLockRec();
    if (lockNode == NULL)
        return NFX_ERR;

    NLM_GET_L_REGION_END(nlm_lockInfop->nlm_l_end,
                         nlm_lockInfop->nlm_l_offset,
                         nlm_lockInfop->nlm_l_len,
                         nlm_lockInfop->nlm_l_eof);
    lockNode->nlm_lockInfop = nlm_lockInfop;
    lockNode->nlmReq = nlmReq;
    pendQ = &fileCBp->lockOwnerPendingList;

	DLL_INSERT_TAIL(&pendQ->list, lockNode, lockList);
    pendQ->numOfEntries++;

    return (NFX_OK);
}
#endif
#endif


/*-----------------------------------------------------------------
 * Name:		nlm_delLockCB
 *
 * Description: Delete a Lock Control Block (LCB) from either current
 *              or pending lock list related to a file control block.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
int32			/*NFX_OK:  Entry has been deleted*/
				/*NFX_ERR: Failure to delete entry*/
nlm_delLockCB(nfp_lockCB_t	*lockCBp)
{
	nlm_lockInfo_t	*lockp;

	if (lockCBp == NULL)
		return (NFX_ERR);

	lockp = &(lockCBp->protocol.nfp_nlmLock);

	if (lockp->nlm_callerNameLen > 0)
	{
		/*Free nlm_callerName*/
		eee_ramDealloc(lockp->nlm_callerName);
		nlm_globalCfgp->nlm_debugCounters.nlm_freeCallerName++;
		lockp->nlm_callerNameLen = 0;
	}

	if (lockp->nlm_ohLen > 0)
	{
		/*Free nlm_oh*/
		eee_ramDealloc(lockp->nlm_oh);
		nlm_globalCfgp->nlm_debugCounters.nlm_freeOh++;
		lockp->nlm_ohLen = 0;
	}
//	lockCBp->nfp_waitForLockCBp = NULL;
//	lockCBp->nfp_pendingLockCBp = NULL;	
	nfp_deallocLockCB(lockCBp);

    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		 nlm_rmLockCBFromList
 *
 * Description: Remove a Lock Control Block (LCB) from either current
 *              or pending lock list related to a file control block.
 *  
 * Created by:	Xiaobing Zhang
 *  
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NOT_USED
int32			/*NFX_OK:  Entry has been removed*/
				/*NFX_ERR: Failure to remove entry*/
nlm_rmLockCBFromList(nfp_lockCB_t	*rmLockCBp,
					 boolean		deleteIt)	/*Delete the lockCB or not*/
{
	nfp_list_t		*lockListp;
	uint32			count;
	nfp_nodePtr_t   nodep;
	nfp_lockCB_t	*lockCBp;
	boolean			removed = FALSE;

	if (rmLockCBp == NULL)
		return (NFX_ERR);
	lockListp = &(rmLockCBp->nfp_fileCbp->nfp_lockList);

    count = lockListp->nfp_reqCnt;
	for (nodep = lockListp->nfp_headp; (nodep != NULL && count > 0);
	 	 nodep = nodep->nfp_nextp, count--)
	{
		lockCBp = (nfp_lockCB_t *)nodep->nfp_entryCBp;
		if (lockCBp == rmLockCBp)
		{
			if (lockListp->nfp_reqCnt == 1)
			{
				/*Remove the only node*/
				lockListp->nfp_headp = NULL;
				lockListp->nfp_tailp = NULL;
			}
			else if (lockListp->nfp_headp == nodep)
			{
				/*Remove the first node*/
				lockListp->nfp_headp = nodep->nfp_nextp;
				nodep->nfp_nextp->nfp_prevp = NULL;
			}
			else if (lockListp->nfp_tailp == nodep)
			{
				/*Remove the last node*/
				lockListp->nfp_tailp = nodep->nfp_prevp;
				nodep->nfp_prevp->nfp_nextp = NULL;
			}
			else
			{
				/*Remove the node in the middle*/
				nodep->nfp_prevp->nfp_nextp = nodep->nfp_nextp;
				nodep->nfp_nextp->nfp_prevp = nodep->nfp_prevp;
			}

			if (deleteIt == TRUE)
			{
				nlm_delLockCB(rmLockCBp);
				nfp_deallocNode(nodep);

			}

			removed = TRUE;
			break;
		}
	}
	
	if (removed == TRUE)
	{
    	lockListp->nfp_reqCnt--;

    	return (NFX_OK);
	}
	else
	{
    	return (NFX_ERR);
	}
}

/*-----------------------------------------------------------------
 * Name:		nlm_delLockCBList
 *
 * Description: Delete a Lock Control Block (LCB) list related to 
 *              a file control block. The pending lock lists are
 *              also deleted.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
int32			/*NFX_OK:  List has been deleted*/
				/*NFX_ERR: Failure to delete list*/
nlm_delLockCBList(nfp_list_t	*lockListp)
{
	nfp_nodePtr_t   nodep, tmpnodep;
	nfp_lockCB_t	*lockCBp;

	if (lockListp == NULL)
		return (NFX_ERR);

	NLM_DBG_PRINTF("nlm_delLockCBList: lockListp->nfp_reqCnt = %d\n",
						lockListp->nfp_reqCnt);

	nodep = lockListp->nfp_headp;
	while(nodep)
	{
		tmpnodep = nodep;
		nodep = nodep->nfp_nextp;

		/*Delete pending list if there is one*/
		lockCBp = (nfp_lockCB_t *)tmpnodep->nfp_entryCBp;
		nlm_delLockCB((nfp_lockCB_t *)tmpnodep->nfp_entryCBp);
		nfp_deallocNode(tmpnodep);
	}

	lockListp->nfp_headp = NULL;
	lockListp->nfp_tailp = NULL;
	lockListp->nfp_reqCnt = 0;

    return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:	nlm_getLockCBOwner		
 *  
 * Description: Get a Lock Control Block (LCB) owner in the current
 * lock list related to a file control block. 
 *
 * Created by:	Xiaobing Zhang 
 *
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
nfp_lockCB_t *	
nlm_getLockCBOwner(	nfp_fileCB_t	*fileCBp,
					nlm_lockInfo_t	*alockp)
{
	nfp_nodePtr_t	nodep;
	nfp_list_t		*lockListp;
	nfp_lockCB_t	*currLockCBp;
	nlm_lockInfo_t	*currLockp;
	uint32			count;

	lockListp = &fileCBp->nfp_lockList;
	count = lockListp->nfp_reqCnt;
	for (nodep = lockListp->nfp_headp; (nodep != NULL && count > 0);
		 nodep = nodep->nfp_nextp, count--)
	{
		currLockCBp = (nfp_lockCB_t *)nodep->nfp_entryCBp;
		currLockp = &(currLockCBp->protocol.nfp_nlmLock);
		/*--------------------------------------
		 *Look for lock owner with same OH &
		 *same svid (process ID)
		 *--------------------------------------
		 */
		if ((currLockp->nlm_ohLen != alockp->nlm_ohLen) ||
			(memcmp(currLockp->nlm_oh, alockp->nlm_oh, 
					alockp->nlm_ohLen) != 0) ||
				(currLockp->nlm_svid != alockp->nlm_svid) )
				continue;

		
		return (currLockCBp);
	}

	return (NULL);
}
#endif // NOT_USED

/*-----------------------------------------------------------------
 * Name:	nlm_getLockRecInPendingList	
 *  
 * Description: Get a Lock record in pending lock list 
 * of the Lock owner control block (LCB).  The search keys are some 
 * fields in argument "alockp". Argument "searchMask" 
 * defines what these fields are. By default, we always
 * look at nlm_l_offset and nlm_l_end. The locked region
 * of found LCB may have overlap or exact match with that 
 * of the "alockp". The returned next_lockRecp points to
 * lock record (lockRecp) whose locked region 
 * has higher address than that of "alockp"
 *
 * Created by:	Xiaobing Zhang 
 * Modified by:	Joseph Vu		10/02/02	Handle pending locks 
 * Modified by:	Joseph Vu		10/24/02	Consolidated lockownerlist
 *											and lockownerPendingList
 *											under LockCB 
 *
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NOT_USED
nlm_lockRec_t *	/*Pointer to found LockRec or NULL if failed*/
nlm_getLockRecInPendingList(nfp_fileCB_t	*fileCBp,
							nlm_lockInfo_t	*alockp, 
							short16			searchMask,
							nlm_lockRec_t	**next_lockRecp)
{
	nfp_list_t					*lockListp;
	nfp_nodePtr_t   			nodep;
	nfp_lockCB_t				*currLockCBp;
	nlm_lockInfo_t				*currLockp;
	nlm_lockRec_t				*currLockRecp, *currLockRecNextp;
	int32						overlap;
	uint32						count;


	if ((fileCBp == NULL) || (alockp == NULL))
		return (NULL);

	lockListp = &fileCBp->nfp_lockList;

	*next_lockRecp = NULL;

	/*------------------------------
	 *Set end region 
	 *-------------------------------
	 */
	NLM_GET_L_REGION_END(alockp->lockRec.nlm_l_end,
                    	alockp->lockRec.nlm_l_offset,
                    	alockp->lockRec.nlm_l_len,
						alockp->lockRec.nlm_l_eof);


    count = lockListp->nfp_reqCnt;
	for (nodep = lockListp->nfp_headp; (nodep != NULL && count > 0);
		 nodep = nodep->nfp_nextp, count--)
	{
		currLockCBp = (nfp_lockCB_t *)nodep->nfp_entryCBp;
		if (!(currLockCBp->nfp_flag & NFP_ENTRY_INUSED))
			continue;

		currLockp = &(currLockCBp->protocol.nfp_nlmLock);

		DLL_ITERATE_NEXT(currLockRecp, currLockRecNextp,
					&currLockCBp->lockOwnerPendingList.list,
				   	lockList)
		{
			if ((currLockRecp == NULL) || (currLockRecp == DLL_NIL) )
				break;

			if (nlm_validateLockRec(currLockRecp) != NFX_OK)
			{
#ifdef	NLM_VALIDATE_LOCK
				panic ("%s: Invalid currLockRecp %p\n",
						   	__FUNCTION__,currLockRecp);
#endif
				DLL_REMOVE(&currLockCBp->lockOwnerPendingList.list,
						currLockRecp, lockList);
				currLockCBp->lockOwnerPendingList.numOfEntries--;
				nlm_deallocLockRec(currLockRecp);
			}


			/*------------------------------
	 	 	 *Compare 2 locked regions
	 	 	 *-------------------------------
	 	 	 */
			overlap = nlm_regionCmp(currLockRecp, &alockp->lockRec);
			if (overlap == 1)
			{
				/*No overlapping region found*/
				*next_lockRecp = currLockRecp;
				return (NULL);
			}
			if (overlap == -1)
			{
				/*Need to check next lock record*/
				continue;
			}
		
			/*--------------------------
		 	 *Found overlapping region
		 	 *--------------------------
		 	 */

			if (searchMask & NLM_SEARCH_OH)
			{
				/*Check lock owner*/
				if ((currLockp->nlm_ohLen != alockp->nlm_ohLen) ||
				(memcmp(currLockp->nlm_oh, alockp->nlm_oh, 
									alockp->nlm_ohLen) != 0))
			{

				continue;
			}
			}
			if (searchMask & NLM_SEARCH_SVID)
			{
				/*Check process ID*/
				if (currLockp->nlm_svid != alockp->nlm_svid)
				{

					continue;
				}
			}
			if (searchMask & NLM_SEARCH_REGION_MATCH)
			{
				/*Check if regions exactly match*/
				if ((currLockp->lockRec.nlm_l_offset != 
									alockp->lockRec.nlm_l_offset) ||
					(currLockp->lockRec.nlm_l_end != 
					 				alockp->lockRec.nlm_l_end))
				{
					continue;
				}
			}

			if (searchMask & NLM_SEARCH_EXCLUSIVE)
			{
				/*Check exclusive flag*/
				if (currLockp->lockRec.nlm_exclusive != TRUE)
				{
					continue;
				}
			}
#ifdef	NOTUSED
			if ((currLockRecp->lockList.next == NULL) ||
				 (currLockRecp->lockList.next == DLL_NIL))
			{
				*next_lockRecp = currLockRecp;
			}
			else
			{
				*next_lockRecp = currLockRecp->lockList.next;
			}
#endif	
			*next_lockRecp = currLockRecp;

			return (currLockRecp);

		}
	}

	return(NULL);
}
#else // NOT_USED
#if 0
nlm_pendingLockNode_t * 
nlm_getLockRecInPendingList(nfp_fileCB_t   *fileCBp,
							nlm_lockInfo_t *alockp, 
							nlm_request_t  *nlmReqp)
{
    nlm_pendingLockNode_t *curr = NULL, *next;
    nlm_lockRecOwnerList_t *pendQ = NULL;
    nlm_lockInfo_t *lock;
    int32 found = 0;

    if (!fileCBp || !alockp || !nlmReqp)
        return NULL;

    NLM_GET_L_REGION_END(alockp->nlm_l_end, alockp->nlm_l_offset,
                         alockp->nlm_l_len, alockp->nlm_l_eof);
    pendQ = &fileCBp->lockOwnerPendingList;
    DLL_ITERATE_NEXT(curr, next, &pendQ->list, lockList)
    {
        if ((curr == NULL) || (curr == DLL_NIL))
            break;

        if (nlm_validateLockRec(curr) != NFX_OK)
        {
#ifdef  NLM_VALIDATE_LOCK
            panic ("%s: Invalid currLockRecp %p\n", __FUNCTION__, curr);
#endif
            DLL_REMOVE(&pendQ->list, curr, lockList);
            pendQ->numOfEntries--;
            nlm_deallocLockRec(curr);
            continue;
        }
        lock = (nlm_lockInfo_t *)curr->nlm_lockInfop;
#if 0
        /*
         * Compare 2 locked regions
         */
        result = nlm_regionCmp(&curr->nlm_lockInfop, alockp);
        if (result)
            /* Lock does not match */
            continue;
#endif

        /* Check lock owner */
        if ((lock->nlm_ohLen != alockp->nlm_ohLen) ||
            ((memcmp(lock->nlm_oh, alockp->nlm_oh, alockp->nlm_ohLen) != 0)))
            continue;

        /* Check process ID */
        if (lock->nlm_svid != alockp->nlm_svid)
            continue;

        /* Check if the offset and length match */
        if ((lock->nlm_l_offset != alockp->nlm_l_offset) ||
            (lock->nlm_l_end != alockp->nlm_l_end))
            continue;

        /* Check exclusive flag */
        if (lock->nlm_exclusive != alockp->nlm_exclusive)
            continue;

        found = 1;
        break;
    }

    return ((found)?curr:NULL);
}
#endif
#endif // NOT_USED


/*-----------------------------------------------------------------
 * Name:			
 *  
 * Description: Get a Lock record in active lock list 
 * of the Lock owner control block (LCB).  The search keys are some 
 * fields in argument "alockp". Argument "searchMask" 
 * defines what these fields are. By default, we always
 * look at nlm_l_offset and nlm_l_end. The locked region
 * of found LCB may have overlap or exact match with that 
 * of the "alockp". The returned next_lockRecp points to
 * Lock record whose locked region has higher address than that
 * of "alockp"
 *
 * Created by:	Xiaobing Zhang 
 * Modified by:	Joseph Vu		10/02/02	Handle pending locks 
 * Modified by:	Joseph Vu		10/24/02	Consolidated lockownerlist
 *											and lockownerPendingList
 *											under LockCB 
 *
 * Date Created:	07/15/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NOT_USED
nlm_lockRec_t *	/*Pointer to found LockRec or NULL if failed*/
nlm_getLockRecInCurrList(nfp_fileCB_t	*fileCBp,
						nlm_lockInfo_t	*alockp, 
						short16			searchMask,
						nlm_lockRec_t	**next_lockRecp
)
{
	nfp_list_t					*lockListp;
	nfp_nodePtr_t   			nodep;
	nfp_lockCB_t				*currLockCBp;
	nlm_lockInfo_t				*currLockp;
	nlm_lockRec_t				*currLockRecp, *currLockRecNextp;
	int32						overlap;
	uint32						count;

	if ((fileCBp == NULL) || (alockp == NULL))
		return (NULL);

	lockListp = &fileCBp->nfp_lockList;

	*next_lockRecp = NULL;

	/*------------------------------
	 *Set end region 
	 *-------------------------------
	 */
	NLM_GET_L_REGION_END(alockp->lockRec.nlm_l_end,
                    	alockp->lockRec.nlm_l_offset,
                    	alockp->lockRec.nlm_l_len,
						alockp->lockRec.nlm_l_eof);


    count = lockListp->nfp_reqCnt;
	for (nodep = lockListp->nfp_headp; (nodep != NULL && count > 0);
		 nodep = nodep->nfp_nextp, count--)
	{
		currLockCBp = (nfp_lockCB_t *)nodep->nfp_entryCBp;
		if (!(currLockCBp->nfp_flag & NFP_ENTRY_INUSED))
			continue;

		currLockp = &(currLockCBp->protocol.nfp_nlmLock);

		DLL_ITERATE_NEXT(currLockRecp, currLockRecNextp,
					&currLockCBp->lockOwnerList.list,
				   	lockList)
		{
			if ((currLockRecp == NULL) || (currLockRecp == DLL_NIL) )
				break;

			if (nlm_validateLockRec(currLockRecp) != NFX_OK)
			{
#ifdef	NLM_VALIDATE_LOCK
				panic ("%s: Invalid currLockRecp %p\n",
						   	__FUNCTION__,currLockRecp);
#endif
				DLL_REMOVE(&currLockCBp->lockOwnerList.list,
						currLockRecp, lockList);
				currLockCBp->lockOwnerList.numOfEntries--;
				nlm_deallocLockRec(currLockRecp);
			}


			/*------------------------------
	 	 	 *Compare 2 locked regions
	 	 	 *-------------------------------
	 	 	 */
			overlap = nlm_regionCmp(currLockRecp, &alockp->lockRec);
			if (overlap == 1)
			{
				/*No overlapping region found*/
				*next_lockRecp = currLockRecp;
				return (NULL);
			}
			if (overlap == -1)
			{
				/*Need to check next lock record*/
				continue;
			}
		
			/*--------------------------
		 	 *Found overlapping region
		 	 *--------------------------
		 	 */

			if (searchMask & NLM_SEARCH_OH)
			{
				/*Check lock owner*/
				if ((currLockp->nlm_ohLen != alockp->nlm_ohLen) ||
				(memcmp(currLockp->nlm_oh, alockp->nlm_oh, 
									alockp->nlm_ohLen) != 0))
			{

				continue;
			}
			}
			if (searchMask & NLM_SEARCH_SVID)
			{
				/*Check process ID*/
				if (currLockp->nlm_svid != alockp->nlm_svid)
				{
					continue;
				}
			}
			if (searchMask & NLM_SEARCH_REGION_MATCH)
			{
				/*Check if regions exactly match*/
				if ((currLockp->lockRec.nlm_l_offset != 
									alockp->lockRec.nlm_l_offset) ||
					(currLockp->lockRec.nlm_l_end != 
					 				alockp->lockRec.nlm_l_end))
				{
					continue;
				}
			}

			if (searchMask & NLM_SEARCH_EXCLUSIVE)
			{
				/*Check exclusive flag*/
				if (currLockp->lockRec.nlm_exclusive != TRUE)
				{
					continue;
				}
			}
#ifdef	NOTUSED
			if ((currLockRecp->lockList.next == NULL) ||	//1
				 (currLockRecp->lockList.next == DLL_NIL))
			{
				*next_lockRecp = currLockRecp;
			}
			else
			{
				*next_lockRecp = currLockRecp->lockList.next;
			}
#endif	//NOTUSED
			*next_lockRecp = currLockRecp;

			return (currLockRecp);

		}
	}

	return(NULL);
}
#endif // NOT_USED

/*-----------------------------------------------------------------
 * Name:		nlm_delReqCBFromList
 *
 * Description: Delete NLM Request Control Block (RCB) from RCB list 
 *              in the Client Control Block.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
int32		/*NFX_OK or NFX_ERR*/
nlm_delReqCBFromList(nlm_request_t *nlmReqp)
{
	nfp_clientCB_t		*clientCBp;
	nfp_list_t			*nlmReqListp;

	if (nlmReqp == NULL)
		return (NFX_ERR);

	clientCBp = (nfp_clientCB_t *)nlmReqp->clientCBp;
	if (clientCBp == NULL)
	{
		NLM_DBG_PRINTF("nlm_delReqCBFromList: No clientCB!\n");
		return (NFX_ERR);
	}
	nlmReqListp = &clientCBp->nfp_nlmReqList;
	nlm_globalCfgp->nlm_debugCounters.numOfDelNlmReqCbFromList++;
	return(nfs_delCBFromList(nlmReqListp, (void *)nlmReqp, NFP_NLMREQCB_NODE));
}

/*-----------------------------------------------------------------
 * Name:		nlm_delReqCB
 *
 * Description: Delete and de-allocate all memory associate with 
 *              this NLM Request Control BLock (RCB) list.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
int32		/*NFX_OK or NFX_ERR*/
nlm_delReqCB(nfp_list_t *nlmReqListp)
{
	nfp_nodePtr_t		currp;
	uint32				count;
	nlm_request_t		*nlmReqp;

	if (nlmReqListp == NULL)
		return (NFX_ERR);

	if (nlmReqListp->nfp_headp == NULL)
		return (NFX_ERR);
	count = nlmReqListp->nfp_reqCnt;

	for (currp = nlmReqListp->nfp_headp; ((currp != NULL) && (count > 0));
		 currp = currp->nfp_nextp, count--)
	{
		nlmReqp = (nlm_request_t *)currp->nfp_entryCBp;
		if (nlmReqp == NULL)
			continue;

		/* 
		 *Only delete completed requests.
		 *Pending request will be deleted upon tx complete.
		 */
		if (nlmReqp->reqState != NFS_REQ_DONE)
			continue; 

		NLM_DBG_PRINTF("nlm_delReqCB: nlmReqp: %p count: %d\n", nlmReqp, count);

		/*
		 *Free all memory corresponding to this request
		 */
		//nlm_msgFreeAll(nlmReqp, TRUE);
		nlm_freeAll(nlmReqp, TRUE);

	}
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:   nlm_utilMapToCompNFSVer     	 
 *
 * Description:
 *    Map NLM version to a compatible NFS version.
 *
 *    NFS and NLM protocol compatibility
 *
 *              +---------+---------+
 *              |   NFS   |   NLM   |
 *              | Version | Version |
 *              +===================+
 *              |    2    |   1,3   |
 *              +---------+---------+
 *              |    3    |    4    |
 *              +---------+---------+
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        6/28/02
 *
 *-----------------------------------------------------------------
 */
int32 nlm_utilMapToCompNFSVer(int32 nlm_ver)
{

	switch (nlm_ver)
	{
		case NLM_VER1:
		case NLM_VER3:
			return (NFS_VER2);
		default:
			//case NLM_VER4:
			return (NFS_VER3);
	}

	return (NFS_VER3);
}


/*-----------------------------------------------------------------
 * Name:nlm_findReqCBByFhandle	 
 *
 * Description:
 * Find a NLM Request Control Block with a correponding 
 * filehandle. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */

nlm_request_t *nlm_findReqCBByFhandle(
	nfp_clientCBPtr_t clientCBp,    //Client control block
	fs_Handle *fhandlep				//search filehandle
)
{
#ifdef	NOTUSED
	nlm_request_t		*nlmReqp;
	nfp_nodePtr_t		reqp;
	int32				reqCnt;
	nfp_list_t			*nodeListp;
	fs_Handle 			*entryFileHandlep;

	nodeListp = &clientCBp->nfp_nlmReqList;
	reqCnt = nodeListp->nfp_reqCnt;
	NLM_DBG_PRINTF("%s: reqCnt: %d\n", __FUNCTION__, reqCnt);
	for (reqp = nodeListp->nfp_headp; (reqp != NULL && reqCnt > 0);
		reqp = reqp->nfp_nextp, reqCnt--)
	{		
		if (nfp_validateNode(reqp) != NFX_OK)
			continue;
		nlmReqp = (nlm_request_t *)reqp->nfp_entryCBp;
		if (nfp_validateNlmReqCB(nlmReqp) != NFX_OK)
			continue;
	
		entryFileHandlep = &nlmReqp->alock.fileHandle;
		if ( (entryFileHandlep->volId.whole_thing == 
			fhandlep->volId.whole_thing) &&
			(entryFileHandlep->inum == fhandlep->inum ) &&
			(entryFileHandlep->mntPnt == fhandlep->mntPnt) )
		{

			return (nlmReqp);
		}
	}
#endif	//NOTUSED

	return (NULL);
}


/*-----------------------------------------------------------------
 * Name:	nlm_findReqCB
 *
 * Description:
 * Find a NLM Request Control Block with a correponding 
 * transaction ID & procedure number. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */

nlm_request_t *nlm_findReqCB(
	nfp_clientCBPtr_t clientCBp,    //Client control block
	uint32 xid,						//transaction ID
	uint32 procNum					//procedure number
)
{
	nlm_request_t		*nlmReqp;
	nfp_nodePtr_t		reqp;
	int32				reqCnt;
	nfp_list_t			*nodeListp;

	nodeListp = &clientCBp->nfp_nlmReqList;
	reqCnt = nodeListp->nfp_reqCnt;
	NLM_DBG_PRINTF("%s: reqCnt: %d\n", __FUNCTION__, reqCnt);
	for (reqp = nodeListp->nfp_headp; (reqp != NULL && reqCnt > 0);
		reqp = reqp->nfp_nextp, reqCnt--)
	{		
		if (nfp_validateNode(reqp) != NFX_OK)
			continue;
		nlmReqp = (nlm_request_t *)reqp->nfp_entryCBp;
		if (nfp_validateNlmReqCB(nlmReqp) != NFX_OK)
			continue;
		if ( (nlmReqp->rpcCallHdrp) &&
			(nlmReqp->rpcCallHdrp->rpc_xid == xid) &&
			(nlmReqp->rpcCallHdrp->rpc_proc == procNum))
		{
			return (nlmReqp);
		}
	}

	return (NULL);
}

nlm_request_t *nlm_findNlmReply (
	nfp_clientCBPtr_t clientCBp,    //Client control block
	uint32 replyXid					//transaction ID
)
{
	nlm_request_t		*nlmReqp;
	nfp_nodePtr_t		reqp;
	int32				reqCnt;
	nfp_list_t			*nodeListp;

	nodeListp = &clientCBp->nfp_nlmReqList;
	reqCnt = nodeListp->nfp_reqCnt;
	NLM_DBG_PRINTF("%s: reqCnt: %d\n", __FUNCTION__, reqCnt);

	for (reqp = nodeListp->nfp_headp; (reqp != NULL && reqCnt > 0);
		reqp = reqp->nfp_nextp, reqCnt--)
	{		
		if (nfp_validateNode(reqp) != NFX_OK)
			continue;
		nlmReqp = (nlm_request_t *)reqp->nfp_entryCBp;
		if (nfp_validateNlmReqCB(nlmReqp) != NFX_OK)
			continue;

		if (nlmReqp->nlm_txCmdInfo.cmdXid == replyXid)
			return (nlmReqp);
	}

	return (NULL);
}

#ifndef NOT_USED
int32
nlm_convertToLockReq(nlm_lockInfo_t *lockInfo, nlm_request_t *nlmReqp,
                     lock_request_t *lockReq)
{
    lockReq->lockType = (lockInfo->nlm_exclusive)?LOCK_TYPE_EXCL
														:LOCK_TYPE_SHAR;
    lockReq->ownerType                    = LOCK_OWNER_NLM;
    lockReq->owner.nlmOwner.clientState   = lockInfo->nlm_state;
    lockReq->owner.nlmOwner.pid           = lockInfo->nlm_svid;
    lockReq->owner.nlmOwner.callerNameLen = lockInfo->nlm_callerNameLen;
    lockReq->owner.nlmOwner.ownerNameLen  = lockInfo->nlm_ohLen;
    lockReq->owner.nlmOwner.clientIpAddr  =
            ((nfp_clientCB_t *)(nlmReqp->clientCBp))->nfp_ipAddr;

    if (lockInfo->nlm_callerNameLen)
        memcpy(lockReq->owner.nlmOwner.callerName, lockInfo->nlm_callerName,
               lockInfo->nlm_callerNameLen);
    if (lockInfo->nlm_ohLen)
        memcpy(lockReq->owner.nlmOwner.ownerName, lockInfo->nlm_oh,
               lockInfo->nlm_ohLen);
    (nlmReqp->block)?lock_getIndefiniteTimeout(&lockReq->timeout)
                    :lock_getZeroTimeout(&lockReq->timeout);
    lockReq->offset                       = lockInfo->nlm_l_offset;
    lockReq->len                          = lockInfo->nlm_l_len;
    if (lockReq->len == 0)
        lockReq->len                      = LOCK_MAX_OFFSET-1-lockReq->offset+1;

    return NFX_OK;
}

int32
nlm_procTestReq(nfp_fileCB_t *fileCBp, nlm_lockInfo_t *lockInfo,
                nlm_request_t *nlmReqp)
{
    lock_request_t lockReq;
    int32          rc, ret;
    lock_testResult_t *testResult;
    char callerName[NLM_MAX_NAME];
    char oh[NLM_MAX_NAME];

    bzero(&lockReq, sizeof(lock_request_t));
    lockReq.owner.nlmOwner.callerName = &callerName[0];
    lockReq.owner.nlmOwner.ownerName  = &oh[0];
    rc = nlm_convertToLockReq(lockInfo, nlmReqp, &lockReq);
    if (rc != NFX_OK)
        return NLM_FAILED;
    // lockReq.ulPrivate = NULL; // We don't need this.

    rc = lock_testLockReq(fileCBp->fileLockIdx, &lockReq, &testResult);
    // printf ("Test lock rc = %d\n", rc);
    if (rc == NFX_OK)
    {
        ret = (testResult->lockState == LOCK_STATE_GRANTED)?NLM_STAT_GRANTED
                                                           :NLM_ERR_DENIED;
    }
    else
        ret = NLM_FAILED;
    // printf ("Test lock ret = %d\n", ret);

    return ret;
}

int32
nlm_procLockReq(nlm_lockInfo_t *lockInfo, nlm_request_t *nlmReqp)
{
    lock_request_t lockReq;
    nfp_fileCB_t   *fileCBp = (nfp_fileCB_t *)nlmReqp->fileCBp;
    int32          rc;
    nfp_requestCB_t *reqCBp;
    nlm_lockInfo_t *saveInfo;

    rc = nlm_saveLockInfo(lockInfo, nlmReqp, &reqCBp);
    if (rc == NFX_ERR)
        return LOCK_STATE_DENIED;
    saveInfo = &((nlm_request_t *)vs_mem_idxToPtr(reqCBp->protocol.vs_nlm_idx))->nlm_lockInformation;
    // saveInfo = &reqCBp->nlm_lockInformation;

    bzero(&lockReq, sizeof(lock_request_t));
    lockReq.owner.nlmOwner.callerName = saveInfo->nlm_callerName;
    lockReq.owner.nlmOwner.ownerName  = saveInfo->nlm_oh;
    rc = nlm_convertToLockReq(lockInfo, nlmReqp, &lockReq);
    if (rc != NFX_OK)
    {
        eee_ramDealloc(saveInfo->nlm_callerName);
        eee_ramDealloc(saveInfo->nlm_oh);
        nfp_deallocReqCB(reqCBp);
        return LOCK_STATE_DENIED;
    }
    // nlmReqp->lockInfop = saveInfo;
    // lockReq.ulPrivate = ((nfp_requestCBPtr_t)nlmReqp->reqCBp)->protocol.vs_nlm_idx;
    lockReq.ulPrivate = reqCBp->protocol.vs_nlm_idx;

    // printf ("%s : nlm_callnerName : %d\n", __FUNCTION__, saveInfo->nlm_callerNameLen);

    rc = lock_byteRangeLockReq(fileCBp->fileLockIdx, &lockReq);
    // printf ("Lock rc = %d\n", rc);
    if (rc == LOCK_STATE_GRANTED)
    {
        eee_ramDealloc(saveInfo->nlm_callerName);
        eee_ramDealloc(saveInfo->nlm_oh);
        nfp_deallocReqCB(reqCBp);
    }

    return rc;
}

int32
nlm_procCancellockReq(nlm_lockInfo_t *lockInfo, nlm_request_t *nlmReqp)
{
    lock_request_t lockReq;
    nfp_fileCB_t   *fileCBp = (nfp_fileCB_t *)nlmReqp->fileCBp;
    int32          rc;
    char callerName[NLM_MAX_NAME];
    char oh[NLM_MAX_NAME];

    bzero(&lockReq, sizeof(lock_request_t));
    lockReq.owner.nlmOwner.callerName = &callerName[0];
    lockReq.owner.nlmOwner.ownerName  = &oh[0];
    rc = nlm_convertToLockReq(lockInfo, nlmReqp, &lockReq);
    if (rc != NFX_OK)
        return NFX_ERR;
    // lockReq.ulPrivate = NULL; // We do not need the cookie here.

    rc = lock_cancelLockReq(fileCBp->fileLockIdx, &lockReq);
    // printf ("Cancel rc = %d\n", rc);
    /*
     * TODO : if no lock found, then we need to return an error
     */

    return rc;
}

int32
nlm_procUnlockReq(nlm_lockInfo_t *lockInfo, nlm_request_t *nlmReqp)
{
    lock_request_t lockReq;
    nfp_fileCB_t   *fileCBp = (nfp_fileCB_t *)nlmReqp->fileCBp;
    int32          rc, rc1, rc2;
    char callerName[NLM_MAX_NAME];
    char oh[NLM_MAX_NAME];

    bzero(&lockReq, sizeof(lock_request_t));
    lockReq.owner.nlmOwner.callerName = &callerName[0];
    lockReq.owner.nlmOwner.ownerName  = &oh[0];
    rc = nlm_convertToLockReq(lockInfo, nlmReqp, &lockReq);
    if (rc != NFX_OK)
        return NFX_ERR;
    // lockReq.ulPrivate = NULL; // We do not need the cookie here.
    // HACK!!!
    lockReq.lockType = LOCK_TYPE_EXCL;
    rc1 = lock_unlockSingleReq(fileCBp->fileLockIdx, &lockReq);
    // printf ("Unlock EXCL rc = %d\n", rc1);
    lockReq.lockType = LOCK_TYPE_SHAR;
    rc2 = lock_unlockSingleReq(fileCBp->fileLockIdx, &lockReq);
    // printf ("Unlock SHARE rc = %d\n", rc2);

    rc = NFX_ERR;
    if (rc1 == NFX_OK || rc2 == NFX_OK)
        rc = NFX_OK;
    return rc;
}

void
nlm_procGrantedCallback(vs_mem_index_t private)
{
    nlm_request_t *nlmReqp;
#if 0
    nfp_fileCB_t *fileCBp;
    nfp_requestCB_t *new_reqCBp;
    nlm_request_t *new_nlmReqp;
#endif
    int32 rc;

    nlmReqp = (nlm_request_t *)vs_mem_idxToPtr(private);
#if 0
    fileCBp = (nfp_fileCB_t *)nlmReqp->fileCBp;
    new_nlmReqp = (nlm_request_t *)nlmReqp->reqCBp;
#endif

    /*
     * call back to block client
     * Send a Portmapper GetPort to the client to find the client NLM port
     * This could have changed since the last RPC call!!
     */
#if 0
    new_reqCBp = nfp_allocReqCB ();
    if (new_reqCBp == NULL)
        return;
    nlm_globalCfgp->nlm_debugCounters.nlm_allocReqCB++;
    new_nlmReqp = (nlm_request_t *)vs_mem_idxToPtr(new_reqCBp->protocol.vs_nlm_idx);
    new_nlmReqp->reqCBp      = new_reqCBp;
    new_nlmReqp->tplConnp    = NULL;
    new_nlmReqp->rxPktDescp  = NULL;
    new_nlmReqp->txPktDescp  = NULL;
    new_nlmReqp->clientCBp   = (void *)nlmReqp->clientCBp;
    new_nlmReqp->fileCBp     = fileCBp;
    new_nlmReqp->rpcCallHdrp = &new_nlmReqp->rpcHdr;
    // new_nlmReqp->lockInfop   = (void *)lockReq; // Used in grantedreply.
    memcpy(&new_nlmReqp->rpcHdr, nlmReqp->rpcCallHdrp,
           sizeof(rpcCallHdr_t));
    new_nlmReqp->nlm_txCmdInfo.laddr =
        ((pkt_desc_t *)nlmReqp->rxPktDescp)->laddr;
#endif
    rc = nlm_getportNlm(nlmReqp, NULL);
    if (rc != NFX_OK)
        (void)nlm_procUnlockReq(nlmReqp->lockInfop, nlmReqp);
    // printf ("%s : nlm_callnerName : %d\n", __FUNCTION__, ((nlm_lockInfo_t *)nlmReqp->lockInfop)->nlm_callerNameLen);
    return;
}
#endif // NOT_USED

#endif	//NFP_TXRX
