/*
 * Name:  nvram-api.h
 *
 * RCS:  $Id: nvram-api.h,v 1.5 2002/03/30 19:53:04 bobonfrogger Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:  Generic NVRAM client api definition.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#ifndef _NVRAM_API_H
#define _NVRAM_API_H

typedef uint32 nvram_handle_t;
typedef uint64 nvram_cookie_t;
typedef uint32 nvram_clientId_t;

typedef struct nvram_rw
{
    nvram_handle_t  handle;             // Handle to NVRAM segment.
    int32           numVecs;            // Number of I/O vectors.
    uchar8          **datap;            // Ptr to data ptrs of each vector.
    int32           *lenp;              // Ptr to len of each vector.
    uint32          offset;             // Offset in segment to read/write.
    int32           doWrap;             // Wrap around within segment?
    int32           wrapReserve;        // Reserve bytes at beginning of
                                        // segment when wrapping.
} nvram_rw_t;

#define NVRAM_RW_MAX    (8192 + 512)  // 8K payload + nvram headers should fit
                                      // in jumbo frame payload.

#define NVRAM_INVALID_HANDLE ((nvram_handle_t) ~0)

#define NVRAM_CLIENT_NONE     0
#define NVRAM_CLIENT_EFS      1   // Filesystem
#define NVRAM_CLIENT_EVM      2   // Volume management
#define NVRAM_CLIENT_LCKMGR   3   // NFS lock manager
#define NVRAM_CLIENT_INTERNAL 4   // Internal for the NVRAM (free blocks) 





/*-----------------------------------------------------------------------------
 * nvram_init
 *
 * Description:  Init the complete NVRAM
 *               In fact, after restart/crash try to recover
 *               and check the validity of the NVRAM.
 *               Make it ready to use: e.g: replay the journal
 * 
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  8/20/2001
 */
int32
nvram_init(
    void
);


/*-----------------------------------------------------------------------------
 * nvram_format
 *
 * Description:  Format and reinitilize the complete NVRAM
 *               All previously data is erazed
 *               In fact, after restart/crash try to recover
 *               and check the validity of the NVRAM.
 *               Make it ready to use: e.g: replay the journal
 * 
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  8/20/2001
 */
int32
nvram_format (
    uint32      mode,
#define NVRAM_MODE_STANDALONE  1
#define NVRAM_MODE_CLUSTERED   2

    uint32      nvramSize,      // Total size of NVRAM in K bytes (1024)
                                // Initialized by SSC
                                // For the moment:
#define NVRAM_NVRAM_SIZE    (8 * 1024 + 30) // == 2M + overhead (estimate for now)

    uint32      segments,       // Total number of segments
                                // Initialized by SSC
                                // Can only change with reformat
                                // For the moment:
#define NVRAM_SEGMENTS_MAX  512

    uint32      blockSize       // Size of blocks in bytes
                                // Initialized by SSC
                                // Can only change with reformat
                                // For the moment:
#define NVRAM_BLOCK_SIZE    (128 * 1024)

);


/*-----------------------------------------------------------------------------
 * nvram_allocate
 *
 * Description:  Allocate an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_allocate(
    nvram_clientId_t    clientId,
    nvram_cookie_t      cookie,
    int32               size,
    nvram_handle_t      *handlep
);

/*-----------------------------------------------------------------------------
 * nvram_deallocate
 *
 * Description:  Deallocate an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_deallocate(
    nvram_clientId_t    clientId,
    nvram_handle_t      handlep
);

/*-----------------------------------------------------------------------------
 * nvram_find
 *
 * Description:  Find a previously allocated NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_find(
    nvram_clientId_t    clientId,
    nvram_cookie_t      cookie,
    nvram_handle_t      *handlep
);

/*-----------------------------------------------------------------------------
 * nvram_read
 *
 * Description:   Read from an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_read(
    nvram_rw_t    *req
);

/*-----------------------------------------------------------------------------
 * nvram_write
 *
 * Description:   Write to an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_write(
    nvram_rw_t    *req
);

#endif /* _NVRAM_API_H */


