/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : opt-cmd.c
 *
 * Description  : 'options' command
 * Entry module of the 'options' command.
 *
 * Notes/TODO   :
 * When requestions for all the options, sendAgileMsg() on the registry manager
 * returns the error "sendto error: Message too long".
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <ctype.h>
#include "opt.h"
#include "opt-api.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

extern opt_cmd_t    cmds[]; // defined in opt-routines.c
struct timeval waitRegmTo = {10, 0};

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static void
opt_printUsage(void);

static void
opt_tolower(char * string);

static int
opt_parseCmd(int        argc,
             char **    argv);

static void
opt_printOption(opt_option_t * option);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : opt_printUsage()
 *
 * Description  : Print the correct usage of this program.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
opt_printUsage(void)
    
{
    fprintf(stderr, "Usage:\n");
    fprintf(stderr, " options\n");
    fprintf(stderr, " options <option>\n");
    fprintf(stderr, " options <option> <value>\n");
}

/*------------------------------------------------------------------------------
 * Name         : opt_tolower()
 *
 * Description  : Convert a string to lower case.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
opt_tolower(char * string)
{
    int i = 0;

    while (string[i] != 0)
    {
        string[i] = tolower(string[i]);
        i++;
    } // while
    return;
}

/*------------------------------------------------------------------------------
 * Name         : opt_printOption()
 *
 * Description  : Print requested options and their values.
 *
 * - get the options from the registry manager
 * - display the result to the user
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
opt_printOption(opt_option_t * option)
{
    opt_msgType_t   reqType = OPT_GET_OPTION_REQ;
    opt_msg_t       req;
    opt_msg_t       rsp;
    opt_msgHdr_t *  hdr;
    void *          bufp = NULL;
    int             nbEntry = 0;
    int             i;

    OPT_FCT_PRINT(("> opt_printOption()\n"));

    /* Fill the request.
     */
    bufp = &req.buf;
    bzero(bufp,
          (sizeof(opt_msgHdr_t) +
           sizeof(opt_getOptionReq_t)));
    
    memcpy(bufp, (void *) &reqType, sizeof(opt_msgType_t));
    bufp += sizeof(opt_msgType_t);
    // userInfo is not used
    bufp += sizeof(opt_userInfo_t);
    memcpy(bufp, (void *) option, sizeof(opt_option_t));
    bufp += sizeof(opt_option_t);

    req.len = bufp - (void *) req.buf;

    /* Send the request to the registry manager located on the SSC.
     */
    rsp.len = sendAgileMsg(-1,
                           req.buf, req.len,
                           "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &waitRegmTo,
                           rsp.buf, sizeof(rsp.buf));
    
    /* Receive response from the registry manager.
     */
    OPT_DBG_PRINT(("nb bytes received = %d\n", rsp.len));
    if (rsp.len == -1)
    {
		perror("sendAgileMsg failed");
		return;
    }

    /* Read the message header and check if it's expected response.
     */
    bufp = &rsp.buf;
    hdr = (opt_msgHdr_t *) rsp.buf;
    bufp += sizeof(opt_msgHdr_t);
    if (hdr->msgType != OPT_GET_OPTION_RSP)
    {
        OPT_ERR_PRINT(("Message received from REGM is not the expected one!\n"));
		return;
    }
    
    /* Read the response and print its content.
     */
    memcpy((void *) &nbEntry, bufp, sizeof(int));
    bufp += sizeof(int);

    if (nbEntry == 0)
    {
        printf("Option %s not found!\n", (char *) option);
    }
    else
    {
        OPT_DBG_PRINT(("nb entries returned = %d\n", nbEntry));
        printf("%-45.45s\t%s\n", "Option", "Value");
        printf("%-45.45s\t%s\n", "------", "-----");
    }
    
    for (i = 0; i < nbEntry; i++)
    {
        opt_entry_t * entry;

        entry = (opt_entry_t *) bufp;
        bufp += sizeof(opt_entry_t);

        printf("%-45.45s\t%s\n", entry->option, entry->value);
    } // for
    
    OPT_FCT_PRINT(("< opt_printOption()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : opt_parseCmd()
 *
 * Description  : Parse the 'options' command.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_parseCmd(int        argc,
             char **    argv)
    
{
    char *          myName;
	opt_option_t    cmd;
	opt_option_t    option;
	opt_value_t     value;
	int i;

    OPT_FCT_PRINT(("> opt_parseCmd()\n"));
    
    /* Initialize the name of this application.
     * Discard path if any.
     */
	myName = argv[0];
	if (rindex(myName, '/'))
    {
		strcpy(myName, rindex(myName, '/') + 1);
	}

    /* If number of argument is incorrect, print usage.
     */
    if (argc > 3)
    {
        OPT_DBG_PRINT(("Too many arguments!\n"));
        opt_printUsage();
        return(NFX_ERR);
    }

    /* If no options are specified, 
     * print the current value of all available options.
     */
    if (argc == 1)
    {
        strcpy(option, "options.*");
        opt_printOption(&option);
        return(NFX_OK);
    }

    /* An option is specified.
     * Initialize the option name (add 'options.' prefix).
     */
	sprintf(option, "options.%s", argv[1]);
    
    /* An option is specified.
     * Get the command name.
     */
    strcpy(cmd, argv[1]);
    if (index(cmd, '.'))
    {
        *index(cmd, '.') = 0;
    }
    
    /* If an option is specified without value,
     * print the current value of that option.
     */
    if (argc == 2)
    {
        opt_printOption(&option);
        return(NFX_OK);
    }
    
    /* A value is specified.
     * Initialize the value.
     */
	strcpy(value, argv[2]);
    
    /* If an option is specified with a value,
     * execute the option.
     */    
	for (i = 0; cmds[i].name != NULL; i++)
    {
		if (!strcmp(cmds[i].name, cmd))
        {
            cmds[i].cmd(option, value);
			return(NFX_OK);
		}
    } // for
    
	printf("Invalid option %s\n", cmd);

    OPT_FCT_PRINT(("< opt_parseCmd()\n"));
	return(NFX_ERR);
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : main
 *
 * Description  : Options entry point.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int
main(int        argc,
     char **    argv)
    
{
    int i;
    
    OPT_FCT_PRINT(("> main()\n"));
    
    /* To be case insensitive, convert the command in lower cases.
     */
    for (i = 0; i < argc; i++)
    {
        opt_tolower(argv[i]);
    } // for
            
    /* Parse the command.
     */
    opt_parseCmd(argc, argv);

    OPT_FCT_PRINT(("< main()\n"));
    return 0;
}

/*----------------------------------------------------------------------------*/
