/* vi:ts=4
 *------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : opt-regm.c
 *
 * Description  : Registry manager (REGM).
 * Entry module of the registry manager.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include "opt.h"
#include "opt-api.h"

/*------------------------------------------------------------------------------
 * Type definitions.
 *------------------------------------------------------------------------------
 */

/* Registry binary database.
 */
typedef struct opt_db {
	char		*option;
	char		*value;
	struct opt_db *prev, *next;
} opt_db_t;

typedef struct {
    int         nbEntry;
    opt_db_t	*db;
} opt_database_t;

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

/* Registry manager's file descriptor.
 */
static int fd;

/* Registry database.
 */
static opt_database_t   regDB;

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : opt_terminateRegistryMgr()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
opt_terminateRegistryMgr(int sig)

{
    OPT_FCT_PRINT(("> opt_terminateRegistryMgr()\n"));

    /* Unregister REGM.
     */
    unRegAgileApp(fd);
    
    OPT_FCT_PRINT(("< opt_terminateRegistryMgr()\n"));
    exit(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : opt_initRegistryMgr()
 *
 * Description  : Initialize the registry manager.
 *
 * Created by   : Isabelle
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_initRegistryMgr(void)

{
    OPT_FCT_PRINT(("> opt_initRegistryMgr()\n"));
    
    /* Register to receive messages.
     */
	fd = regAgileApp("regm");
	if (fd < 0)
    {
        perror("regAgileApp() failed");
		return(NFX_ERR);
	}

    /* Sets the procedure to follow in case of request for program termination
     * or CTRL-C interrupt.
     */
    signal(SIGTERM, opt_terminateRegistryMgr);
    signal(SIGINT, opt_terminateRegistryMgr);
   
    OPT_FCT_PRINT(("< opt_initRegistryMgr()\n"));
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : opt_initDatabase()
 *
 * Description  : Initialize the registry database.
 *
 * . Read the ascii configuration file.
 * . Fill the binary database.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_initDatabase(char * fileName)

{
    int     rc = NFX_OK;
    FILE *  fd = NULL;

    OPT_FCT_PRINT(("> opt_initDatabase()\n"));

    /* Reset the registry database.
     */
    bzero((void *) &regDB, sizeof(opt_database_t));
    
    /* Open in read only mode the file.
     */
    fd = fopen(fileName, "r");
    if (fd != NULL)
    {
        char    *value;
		char 	*opt;
		size_t	len;
        int     nbOption = 0;
		opt_db_t *newObj;

        /* Read the line of the file one by one and
         * store the option string and its value into the database.
         */
        while ((opt = fgetln (fd, &len)))
        {
			opt[len-1] = '\0';

            OPT_DBG_PRINT(("line read: %s\n", opt));
            if (opt[0] != '#') // ignore comments
            {          
                /* Read the option and its value.
                 * Syntax of an entry in the file : <option>=<value>
                 */
                if ((value = index(opt, '=')))
                {
					*value = '\0';
					value++;
					newObj = malloc (sizeof(opt_db_t));
					if (!newObj) {
						OPT_ERR_PRINT (("Unable to allocate DB entry\n"));
						rc = NFX_ERR;
						break;
					}
					newObj->next = regDB.db;
					newObj->prev = NULL;
					if (regDB.db) {
						regDB.db->prev = newObj;
					}
					regDB.db = newObj;

					newObj->option = strdup (opt);
					newObj->value = strdup (value);
					if (!newObj->option || !newObj->value) {
						OPT_ERR_PRINT (("Unable to allocate DB strings\n"));
						rc = NFX_ERR;
						break;
					}
                    
                    nbOption++;
                }
                else
                {
                    OPT_ERR_PRINT(("Syntax error in %s near '%s'\n",
                                  fileName, opt));
					continue;
                }
            }
        } // while
        regDB.nbEntry = nbOption;

        /* Close file.
         */
        fclose(fd);
    }
    OPT_FCT_PRINT(("< opt_initDatabase()\n"));
    return(rc);
}


/*------------------------------------------------------------------------------
 * Name         : opt_writeFile()
 *
 * Description  : Write the registry database into the registry file.
 *
 * . Backup registry file.
 * . Read the binary database.
 * . Fill the ascii configuration file.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_writeFile(char * fileName)

{
    int     rc = NFX_OK;
    FILE	*fd = NULL;
	char    *bakFile;
	sigset_t block_these;
	sigset_t old_mask;
	char	*newName;
	int		newFd;
	int		ofd, bfd;

    OPT_FCT_PRINT(("> opt_writeFile()\n"));

/*
	Let's make sure we're not interrupted while we dump out our database
 */
	sigemptyset (&block_these);
	sigaddset (&block_these, SIGINT);
	sigaddset (&block_these, SIGTERM);
	sigprocmask (SIG_BLOCK, &block_these, &old_mask);

    /* Backup registry file.
     */

	asprintf (&bakFile, "%s.bak", fileName);
	ofd = open (fileName, O_RDONLY, 0);
	if (ofd >= 0) {		// It was there
		char buff[1024];
		int nrd;

		bfd = open (bakFile, O_RDWR | O_CREAT | O_TRUNC, 0666);
		if (bfd >= 0) {
			while ((nrd = read (ofd, buff, sizeof(buff))) > 0) {
				write (bfd, buff, nrd);
			}
			fsync (bfd);
			close (bfd);
		}
		else {
			perror ("Cannot make backup of registry db");
			exit (1);	// Basically assert and panic
		}
		close (ofd);
	}
    
    /* Overwrite the file.
     */
	asprintf (&newName, "%s.new", fileName);
	newFd = open (newName, O_CREAT | O_TRUNC | O_RDWR, 0666);
    fd = fdopen(newFd, "w");
    if (fd != NULL)
    {
        time_t		currTime;
		opt_db_t	*entry;

        /* Write the current time in the file.
         */
        time(&currTime);      
        fprintf(fd, "# Last update: %s", ctime(&currTime));

        /* Write each database entry in the file.
         */
        for (entry = regDB.db; entry; entry = entry->next)
        {
            fprintf(fd, "%s=%s\n", entry->option, entry->value);
        }

        /* Close file.
         */
        fflush(fd);

		fsync (newFd);	// write it to disk

		fclose (fd);

		rename (newName, fileName);
    }
    else
    {
        perror("fopen() failed\n");
        rc = NFX_ERR;
    }

	close (newFd);
	free (bakFile);
	free (newName);
    
	sigprocmask (SIG_SETMASK, &old_mask, NULL);
    OPT_FCT_PRINT(("< opt_writeFile()\n"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : opt_getOption()
 *
 * Description  :
 * Get list of options and their values matching the given option prefix.
 * The option string may contain a '*' at the end but this isn't mandatory.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_getOption(char				*optionPrefix,
              int *             nbResult,
              opt_db_t    **    result)

{
    int             rc = NFX_OK;
    char    		*matchStr;
    size_t          matchLen;
	int				wildcard;
	opt_db_t		*entry;
    
    OPT_FCT_PRINT(("> opt_getOption()\n"));

    /* Initialize the matching string.
     */
    matchStr = strdup(optionPrefix);
	wildcard = 0;
	if (index(matchStr, '*') != NULL)
    {
        *index(matchStr, '*') = '\0';
		wildcard = 1;
	}
    matchLen = strlen(matchStr);
    OPT_DBG_PRINT(("matchStr = %s\n", matchStr));
       
    /* Go through the database and get list of options and their values
     * matching the given option prefix.
     */
    *nbResult = 0;
    for (entry = regDB.db; entry; entry = entry->next)
    {
		if (wildcard) 
		{
        	if (!strncmp(entry->option, matchStr, matchLen))
        	{
            	result[*nbResult] = entry;
            	(*nbResult)++;
        	}
        }
		else 
		{
			if (!strcmp(entry->option, matchStr))
			{
				result[*nbResult] = entry;
				(*nbResult)++;
			}
		}
    } // for
        
    OPT_FCT_PRINT(("< opt_getOption()\n"));
	free (matchStr);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : opt_processGetOptionReq()
 *
 * Description  : Process the 'get option' request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_processGetOptionReq(opt_getOptionReq_t *    req,
                        opt_senderInfo_t *      sender,
                        opt_userInfo_t          userInfo)

{
    int             rc = NFX_OK;
    int             i;
    opt_db_t 		*entryFound[MAX_ENTRY];
    int             nbEntry;
    opt_msg_t       rsp;
    char *          bufp = rsp.buf;
    opt_msgType_t   msgType = OPT_GET_OPTION_RSP;
	opt_entry_t		*daEntry;
    
    OPT_FCT_PRINT(("> opt_processGetOptionReq()\n"));

    /* Get all entries matching the request.
     */
    opt_getOption(req->option,
                  &nbEntry,
                  entryFound);
    
    /* Initialize the header of the response.
     */
    memcpy((void *) bufp, (void *) &msgType, sizeof(msgType));
    bufp += sizeof(msgType);
    memcpy((void *) bufp, (void *) &userInfo, sizeof(userInfo));
    bufp += sizeof(userInfo);

    /* Fill response with entries found.
     */
    memcpy((void *) bufp, (void *) &nbEntry, sizeof(nbEntry));
    bufp += sizeof(nbEntry);
    
	daEntry = (opt_entry_t *)bufp;
    for (i = 0; i < nbEntry; i++, bufp += sizeof (opt_entry_t))
    {
		strncpy (daEntry[i].option, entryFound[i]->option, OPT_MAX_OPTION_LEN);
		strncpy (daEntry[i].value, entryFound[i]->value, MAX_VALUE_LEN);
    }
    
    rsp.len = bufp - rsp.buf;
    
    /* Send the response to whomever made the request
     */
    if (sendAgileMsg(-1,
                     rsp.buf, rsp.len,
                     sender->name, sender->cpu, sender->slot, sender->plane,
                     NULL, NULL, 0) != 0)
    {
        OPT_ERR_PRINT(("sendAgileMsg() failed\n"));
        rc = NFX_ERR;
    }
        
    OPT_FCT_PRINT(("< opt_processGetOptionReq()\n"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : opt_processSetOptionReq()
 *
 * Description  : Process the 'set option' request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_processSetOptionReq(opt_setOptionReq_t *    req,
                        opt_senderInfo_t *      sender,
                        opt_userInfo_t          userInfo)

{
    int             rc = NFX_OK;
    opt_db_t    *   entryFound[MAX_ENTRY];
    int             nbEntry;
    opt_msg_t       rsp;
    void *          bufp = &rsp.buf;
    opt_msgType_t   msgType = OPT_SET_OPTION_RSP;
    int             result = NFX_OK;
	opt_db_t		*newEntry;
    
    OPT_FCT_PRINT(("> opt_processSetOptionReq()\n"));

    /* Get all entries matching the request.
     */
    opt_getOption(req->entry.option,
                  &nbEntry,
                  entryFound);

    /* Check how many entries are matching the given option.
     * Accordingly, update or add the database.
     * Write the modification in the file.
     */
    if (nbEntry > 1)
    {
        OPT_ERR_PRINT(("Option %s is not unique!\n", req->entry.option));
        result = NFX_ERR;
    }
    else if ((nbEntry == 0) && (req->del == TRUE))
    {
        OPT_DBG_PRINT(("Option %s doesn't exist!\n", req->entry.option));
        OPT_DBG_PRINT(("Cannot be removed!\n"));
        result = NFX_ERR;        
    }
    else
    {
        if (nbEntry == 0)
        {
            // add new option into database
			newEntry = malloc (sizeof(opt_db_t));
			if (!newEntry) {
				OPT_ERR_PRINT (("Unable to allocate a new database entry\n"));
				result = NFX_ERR;
			}
			else {
				newEntry->option = strdup (req->entry.option);
				newEntry->value = strdup (req->entry.value);
				if (!newEntry->option || !newEntry->value) {
					free (newEntry->option);
					free (newEntry->value);
					free (newEntry);
					OPT_ERR_PRINT (("Unable to allocate new db strings\n"));
					result = NFX_ERR;
				}
				else {
					newEntry->prev = NULL;
					newEntry->next = regDB.db;
					if (regDB.db) {
						newEntry->next->prev = newEntry;
					}
					regDB.db = newEntry;
					regDB.nbEntry++;
				}
			}
        }
        else if ((nbEntry == 1) && (req->del == FALSE))
        {
            // write new value into database
			free (entryFound[0]->value);
            entryFound[0]->value = strdup(req->entry.value);
        }
        else if ((nbEntry == 1) && (req->del == TRUE))
        {
            // remove option from database
			if (entryFound[0]->prev) {
				entryFound[0]->prev->next = entryFound[0]->next;
			}
			else {
				regDB.db = entryFound[0]->next;
			}
			if (entryFound[0]->next) {
				entryFound[0]->next->prev = entryFound[0]->prev;
			}
			regDB.nbEntry--;
			free (entryFound[0]->option);
			free (entryFound[0]->value);
			free (entryFound[0]);
        }
        // update file
        opt_writeFile(NFX_CONFIG_FILE);
    }
    
    /* Initialize the header of the response.
     */
    memcpy((void *) bufp, (void *) &msgType, sizeof(msgType));
    bufp += sizeof(msgType);
    memcpy((void *) bufp, (void *) &userInfo, sizeof(userInfo));
    bufp += sizeof(userInfo);

    /* Fill response with the result.
     */
    memcpy((void *) bufp, (void *) &result, sizeof(result));
    bufp += sizeof(result);
    
    rsp.len = bufp - (void *) &(rsp.buf);
    
    /* Send the buffer to the destination located on the SSC.
     */
    if (sendAgileMsg(-1,
                     rsp.buf, rsp.len,
                     sender->name, sender->cpu, sender->slot, sender->plane,
                     NULL, NULL, 0) != 0)
    {
        OPT_ERR_PRINT(("sendAgileMsg() failed\n"));
        rc = NFX_ERR;
    }
        
    OPT_FCT_PRINT(("< opt_processSetOptionReq()\n"));
    return(rc);
}

/*-----------------------------------------------------------------
 * Name: opt_processBigOptGetReq
 *
 * Description: Get option/value pairs based upon what they sent us
 *    responding with the variable string message OPT_GET_BIGOPT_RSP
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 15 July 2002
 *
 *-----------------------------------------------------------------
 */
static void
opt_processBigOptGetReq (regmBigGetReq_t	*rq, 
                         opt_senderInfo_t	*sender,
                         opt_userInfo_t		userInfo)
{
	opt_db_t		**entryFound;
	int 			numFound;
	int				stringSize;
	int				i;
	char			*putHere;
	regmBigGetRsp_t *rsp;
	int				rspSize;

	entryFound = malloc (sizeof(opt_db_t *) * regDB.nbEntry);

	opt_getOption (rq->opt, &numFound, entryFound);
/*
	We'll need to figure out how much message buffer we'll need
 */
	rspSize = sizeof(regmBigGetRsp_t);

	if (numFound > 0) {

		for (i = 0, stringSize = 0; i < numFound; i++) {
			stringSize += strlen (entryFound[i]->option) + 1;
			stringSize += strlen (entryFound[i]->value) + 1;
		}
		rspSize += stringSize;
	}
/*
	Allocate message buffer and fill out the message header
 */
	rsp = malloc (rspSize);
	rsp->hdr.msgType = OPT_GET_BIGOPT_RSP;
	rsp->hdr.userInfo = userInfo;
	rsp->nbEntry = numFound;
/*
	Copy the option/value strings to the message
 */
	putHere = rsp->opt_val;
	for (i = 0; i < numFound; i++) {
		strcpy (putHere, entryFound[i]->option);
		putHere += strlen (entryFound[i]->option) + 1;
		strcpy (putHere, entryFound[i]->value);
		putHere += strlen (entryFound[i]->value) + 1;
	}

	free (entryFound);

    if (sendAgileMsg(-1,
                     rsp, rspSize, 
                     sender->name, sender->cpu, sender->slot, sender->plane,
                     NULL, NULL, 0) != 0)
    {
        OPT_ERR_PRINT(("sendAgileMsg() failed\n"));
    }

	free (rsp);
}

/*-----------------------------------------------------------------
 * Name: opt_processBigOptSetReq
 *
 * Description: Set an option/value pair given the variable length
 *    strings.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 15 July 2002
 *
 *-----------------------------------------------------------------
 */
static void
opt_processBigOptSetReq	(regmBigSetReq_t	*rq, 
                         opt_senderInfo_t	*sender,
                         opt_userInfo_t		userInfo)
{
    int             nbEntry;
	opt_db_t		**entryFound;
    int             result = NFX_OK;
	regmBigSetRsp_t	rsp;
	opt_db_t		*newEntry;

	entryFound = malloc (sizeof(opt_db_t *) * regDB.nbEntry);

	opt_getOption (rq->opt_val, &nbEntry, entryFound);

    /* Check how many entries are matching the given option.
     * Accordingly, update or add the database.
     * Write the modification in the file.
     */
    if (nbEntry > 1)
    {
        OPT_ERR_PRINT(("Option %s is not unique!\n", rq->opt_val));
		result = NFX_ERR;
    }
    else if ((nbEntry == 0) && (rq->del == TRUE))
    {
        OPT_DBG_PRINT(("Option %s doesn't exist!\n", rq->opt_val));
        OPT_DBG_PRINT(("Cannot be removed!\n"));
		result = NFX_ERR;
    }
    else
    {
		char *valStr = rq->opt_val + strlen (rq->opt_val) + 1;

        if (nbEntry == 0)
        {
            // add new option into database
			newEntry = malloc (sizeof(opt_db_t));
			if (!newEntry) {
				OPT_ERR_PRINT (("Unable to allocate a new database entry\n"));
				result = NFX_ERR;
			}
			else {
				newEntry->option = strdup (rq->opt_val);
				newEntry->value = strdup (valStr);

				if (!newEntry->option || !newEntry->value) {
					free (newEntry->option);
					free (newEntry->value);
					free (newEntry);
					OPT_ERR_PRINT (("Unable to allocate new db strings\n"));
					result = NFX_ERR;
				}
				else {
					newEntry->prev = NULL;
					newEntry->next = regDB.db;
					if (regDB.db) {
						newEntry->next->prev = newEntry;
					}
					regDB.db = newEntry;
					regDB.nbEntry++;
				}
			}
		}
        else if ((nbEntry == 1) && (rq->del == FALSE))
        {
            // write new value into database
			free (entryFound[0]->value);
            entryFound[0]->value = strdup(valStr);
        }
        else if ((nbEntry == 1) && (rq->del == TRUE))
        {
            // remove option from database
			if (entryFound[0]->prev) {
				entryFound[0]->prev->next = entryFound[0]->next;
			}
			else {
				regDB.db = entryFound[0]->next;
			}
			if (entryFound[0]->next) {
				entryFound[0]->next->prev = entryFound[0]->prev;
			}
			regDB.nbEntry--;
			free (entryFound[0]->option);	// Don't forget to give back mem
			free (entryFound[0]->value);
			free (entryFound[0]);
        }
        // update file
        opt_writeFile(NFX_CONFIG_FILE);
	}
	free (entryFound);
/*
	Tell the sender how things went
 */
	rsp.hdr.msgType = OPT_SET_BIGOPT_RSP;
	rsp.hdr.userInfo = userInfo;
	rsp.setRsp.result = result;

	if (sendAgileMsg(-1, &rsp, sizeof(rsp),
	                 sender->name, sender->cpu, sender->slot, sender->plane,
	                 NULL, NULL, 0) != 0)
	{
		OPT_ERR_PRINT(("sendAgileMsg() failed\n"));
	}
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : main
 *
 * Description  : Registry manager entry point.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#define BIGBUFSIZE	(64 * 1024)		// Better be big enough
extern int
main(void)
    
{
	char *bigbuf = malloc (BIGBUFSIZE);

    OPT_FCT_PRINT(("> main()"));
	if (!bigbuf) {
		OPT_ERR_PRINT(("Unable to allocate %d bytes for I/O buf\n", 
		               BIGBUFSIZE));
		exit (1);
	}

    /* Initialize the registry manager.
     */
    if (opt_initRegistryMgr() != NFX_OK) {
		exit (1);
	}

    /* Initialize the registry database.
     */
    if (opt_initDatabase(NFX_CONFIG_FILE) != NFX_OK) {
		exit (1);
	}

	/* Tell MCP we're ready to roll.
	 */
	kill (getppid(), SIGUSR1);

    /* Receive messages and process them.
     */
	while (TRUE)
    {
        opt_senderInfo_t    sender;
		int					msg_len;

        /* Wait for messages...
         */
		msg_len = recvAgileMsg(fd,
                               bigbuf, BIGBUFSIZE,
                               NULL,
                               sender.name,
                               &sender.cpu,
                               &sender.slot,
                               &sender.plane);
        
        /* Something has been received, check it.
         */
		if (msg_len <= 0)
        {
			printf("recvAgileMsg() failed\n");
		} 
		else
        {
            opt_msgHdr_t *  hdr;

            hdr = (opt_msgHdr_t *) bigbuf;
            OPT_DBG_PRINT(("msg type is %d\n", hdr->msgType));
            
            /* Process the message according to its type.
             */
            switch (hdr->msgType)
            {
            case OPT_GET_OPTION_REQ:
            {
                opt_getOptionReq_t *    req;
                
                OPT_DBG_PRINT(("OPT_GET_OPTION_REQ\n"));
                req = (opt_getOptionReq_t *) (bigbuf + sizeof(opt_msgHdr_t));
                opt_processGetOptionReq(req, &sender, hdr->userInfo);
                break;
            }
                
            case OPT_SET_OPTION_REQ:
            {
                opt_setOptionReq_t *    req;
                
                OPT_DBG_PRINT(("OPT_SET_OPTION_REQ\n"));
                req = (opt_setOptionReq_t *) (bigbuf + sizeof(opt_msgHdr_t));
                opt_processSetOptionReq(req, &sender, hdr->userInfo);
                break;
            }
            
            case OPT_GET_BIGOPT_REQ:
			{
				regmBigGetReq_t *req = (regmBigGetReq_t *)bigbuf;

				opt_processBigOptGetReq (req, &sender, hdr->userInfo);
				break;
			}

            case OPT_SET_BIGOPT_REQ:
			{
				regmBigSetReq_t *req = (regmBigSetReq_t *)bigbuf;

				opt_processBigOptSetReq (req, &sender, hdr->userInfo);
				break;
			}

            default:
            {
                printf("REGM: Unknown msg type (%d)\n", hdr->msgType);
                break;
            }
            } /* switch */
        }
    } // while

    OPT_FCT_PRINT(("< main()"));
}

/*----------------------------------------------------------------------------*/
