/*
 *-----------------------------------------------------------------
 *
 * Name:	pci-bobcat.c
 *
 * Copyright (C) 2004, OnStor Inc.,
 *
 * Description: PCI initialization routines for BOBCAT platform.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 05/21/04
 *
 *-----------------------------------------------------------------
 */
#if	defined(R9000)

#include <nfx-incl.h>
#include <sys/gt64440r.h>
#include <exports/hwMap.h>
#include <exports/pci-api.h>
#include "pci.h"

#define PCI1_CF_DEV_INTEL 0x11001013
/*-----------------------------------------------------------------
 * Name:	pcix_cfg_check_status()
 *
 * Description: Check status of PCIX config cycle.
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * PCI_ERR_MASTER_ABORT
								 * PCI_ERR_TARGET_ABORT
								 * PCI_ERR_PARITY_ERROR
								 */
pcix_cfg_check_status(
	uint32	bus,
	uint32	dev
)
{
	uint32	addr;
	uint32	stat;

	addr = WORD_SWAP(	(bus << PCI_CFG_BUS_SHIFT) |
						(dev << PCI_CFG_DEV_SHIFT) |
						(0 << PCI_CFG_FUNC_SHIFT) |
						(PCI0_COMMAND_STATUS_REG) |
						PCI_CFG_ENABLE);


	GT_WRITE(PCI_0CONFIGURATION_ADDRESS, addr);
	delay_us(1000);
	stat = GT_READ(PCI_0CONFIGURATION_DATA_VIRTUAL_REGISTER);

	if (stat & PCI_CFG_STAT_MABORT) {
		return PCI_ERR_MASTER_ABORT;
	}
	if (stat & PCI_CFG_STAT_TABORT) {
		return PCI_ERR_TARGET_ABORT;
	}
	if (stat & PCI_CFG_STAT_PERR) {
		return PCI_ERR_PARITY_ERROR;
	}

	return NFX_OK;

}	/* pcix_cfg_check_status() */

/*-----------------------------------------------------------------
 * Name:	pcix_cfg_read()
 *
 * Description:	Generate a PCIX read configuration cycle to GT64440.
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK if data was read successfully
							 * PCI_ERR_BAD_PARAM
							 * PCI_ERR_MASTER_ABORT
							 * PCI_ERR_TARGET_ABORT
							 * PCI_ERR_PARITY_ERROR
							 */
pcix_cfg_read(
	uint32	bus,
	uint32	dev,
	uint32	func,
	uint32	reg,
	uint32	* data
)
{
	uint32	addr;


	if ((bus != PCIX_0_PRI_BUS) || 
		(dev != PCIX_0_GT_DEV) ||
		(func >= PCI_NUM_FUNCS) ||
		(reg >= PCI_NUM_REGS))
		return PCI_ERR_BAD_PARAM;

	addr = WORD_SWAP(	(bus << PCI_CFG_BUS_SHIFT) |
						(dev << PCI_CFG_DEV_SHIFT) |
						(func << PCI_CFG_FUNC_SHIFT) |
						(reg) |
						PCI_CFG_ENABLE);

	GT_WRITE(PCI_0CONFIGURATION_ADDRESS, addr);
	delay_us(1000);
	*data = GT_READ(PCI_0CONFIGURATION_DATA_VIRTUAL_REGISTER);

/*
	printf("[pcix_cfg_read], bus %d, dev %d, reg 0x%x\n", bus, dev, reg);
	printf("[pcix_cfg_read], addr_reg=%p data_reg=%p addr=0x%x data=0x%x\n", 
			addr_reg, data_reg, addr, *data);
*/
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:	pcix_cfg_write()
 *
 * Description:	Generate a PCIX write configuration cycle to GT64440.
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK if data was written successfully
							 * PCI_ERR_BAD_PARAM
							 * PCI_ERR_MASTER_ABORT
							 * PCI_ERR_TARGET_ABORT
							 * PCI_ERR_PARITY_ERROR
							 */
pcix_cfg_write(
	uint32	bus,
	uint32	dev,
	uint32	func,
	uint32	reg,
	uint32	data
)
{
	uint32	addr;


	if ((bus != PCIX_0_PRI_BUS) || 
		(dev != PCIX_0_GT_DEV) ||
		(func >= PCI_NUM_FUNCS) ||
		(reg >= PCI_NUM_REGS))
		return PCI_ERR_BAD_PARAM;

	addr = WORD_SWAP(	(bus << PCI_CFG_BUS_SHIFT) |
						(dev << PCI_CFG_DEV_SHIFT) |
						(func << PCI_CFG_FUNC_SHIFT) |
						(reg) | 
						PCI_CFG_ENABLE);

	GT_WRITE(PCI_0CONFIGURATION_ADDRESS, addr);
	delay_us(1000);
	GT_WRITE(PCI_0CONFIGURATION_DATA_VIRTUAL_REGISTER, data);

/*
	printf("[pcix_cfg_write], bus %d, dev %d, reg 0x%x\n", bus, dev, reg);
	printf("[pcix_cfg_write], addr_reg=%p data_reg=%p addr=0x%x data=0x%x\n", 
			addr_reg, data_reg, addr, data);
*/

	return NFX_OK;

}	

#define PCI_X_STATUS_DEV_SHIFT		3
#define PCI_X_STATUS_BUS_SHIFT		8
#define PCI_X_STATUS_DEV_MASK		0x000000F8
#define PCI_X_STATUS_BUS_MASK		0x0000FF00

/*-----------------------------------------------------------------
 * Name:	pcix_remap_bus_dev()
 *
 * Description:	Remaps PCI-X Bus & Device to new values.
 *
 *-----------------------------------------------------------------
 */
void
pcix_remap_bus_dev(uint32 bus, uint32 dev)
{
	uint32	data;
	int32	rc;

	/* 
	 * Read PCI-X Status register for Bus & Device 
	 */
	pcix_cfg_read( PCIX_0_PRI_BUS, PCIX_0_GT_DEV, 0, PCI_X_STATUS, &data );
	if ((rc = pcix_cfg_check_status(PCIX_0_PRI_BUS, PCIX_0_GT_DEV)) != NFX_OK)
	{
		printf("%s cfg read failed, rc=0x%x\n",__FUNCTION__, rc);
		return;
	}

	/* 
	 * Set PCI-X Status register with new Bus & Device 
	 */
	data &= ~(PCI_X_STATUS_BUS_MASK | PCI_X_STATUS_DEV_MASK);
	data |=  (bus << PCI_X_STATUS_BUS_SHIFT) | (dev << PCI_X_STATUS_DEV_SHIFT);
	pcix_cfg_write( PCIX_0_PRI_BUS, PCIX_0_GT_DEV, 0, PCI_X_STATUS, data );

	if ((rc = pcix_cfg_check_status(bus, dev)) != NFX_OK)
	{
		printf("%s cfg write failed, rc=0x%x\n",__FUNCTION__, rc);
		return;
	}
}

void
pci_init_bobcat()
{ 
    uint32 data = 0;

	/*
	 *  GT64440 PCI0 CS0, CS1 BARs, internal register base
	 */
	pci_cfg_write(PCI0_PRI_BUS, PCI0_GT_DEV, 0, 0x10, HW_GT_MEM_0_BASE|0x0c);
	pci_cfg_write(PCI0_PRI_BUS, PCI0_GT_DEV, 0, 0x18, HW_GT_MEM_1_BASE|0x0c);	
	pci_cfg_write(PCI0_PRI_BUS, PCI0_GT_DEV, 0, 0x20, HW_GALILEO_BASE|0x04);	

	/*
	 *  QLogic Port 0 master cycle & BAR
	 */
	pci_cfg_write(PCI0_PRI_BUS, PCI0_QLOGIC_DEV, 0, 0x04, 0x02a00006);
	pci_cfg_write(PCI0_PRI_BUS, PCI0_QLOGIC_DEV, 0, 0x14, HW_QLOGIC_0_BASE);	

	/*
	 *  QLogic Port 1 master cycle & BAR
	 */
	pci_cfg_write(PCI0_PRI_BUS, PCI0_QLOGIC_DEV, 1, 0x04, 0x02b00006);
	pci_cfg_write(PCI0_PRI_BUS, PCI0_QLOGIC_DEV, 1, 0x14, HW_QLOGIC_1_BASE|0x01);	

	/*
	 *  GT64440 PCI1 CS0, CS1 BARs, internal register & SRAM bases
	 */
	pci_cfg_write(PCI1_PRI_BUS, PCI1_GT_DEV, 0, 0x10, HW_GT_MEM_0_BASE|0x0c);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_GT_DEV, 0, 0x18, HW_GT_MEM_1_BASE|0x0c);	
	pci_cfg_write(PCI1_PRI_BUS, PCI1_GT_DEV, 0, 0x20, HW_GALILEO_BASE|0x04);	
	pci_cfg_write(PCI1_PRI_BUS, PCI1_GT_DEV, 2, 0x20, HW_BMFPGA_BASE);	
	pci_cfg_write(PCI1_PRI_BUS, PCI1_GT_DEV, 1, 0x20, HW_OC_SRAM_BASE|0x04);	

	/*
	 *  TXRX PCI master cycle, BAR, Mailbox BARs
	 */
	pci_cfg_write(PCI1_PRI_BUS, PCI1_TXRX_DEV, 0, 0x04, 0x02a00006);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_TXRX_DEV, 0, 0x10, HW_TXRX_SHMEM_BASE);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_TXRX_DEV, 0, 0x18, HW_TXRX_MBOX0_BASE);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_TXRX_DEV, 0, 0x1c, HW_TXRX_MBOX1_BASE);

	/*
	 *  FP PCI master cycle, BAR, Mailbox BARs
	 */
	pci_cfg_write(PCI1_PRI_BUS, PCI1_FP_DEV, 0, 0x04, 0x02a00006);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_FP_DEV, 0, 0x10, HW_FP_SHMEM_BASE);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_FP_DEV, 0, 0x18, HW_FP_MBOX0_BASE);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_FP_DEV, 0, 0x1c, HW_FP_MBOX1_BASE);

	/*
	 *  Bridge PCI master & io cycle, 
	 *  secondary bus num., mem. BAR, IO BAR
	 */
	pci_cfg_write(PCI1_PRI_BUS, PCI1_BR_DEV, 0, 0x04, 0x02a00007);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_BR_DEV, 0, 0x18, 0x00ff0100);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_BR_DEV, 0, 0x20, 0x19f00080);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_BR_DEV, 0, 0x30, 0x007f0000);

	/*
	 *  Port 0 DP83816 PCI master cycle, BAR
	 */
	pci_cfg_write(PCI1_SEC_BUS, PCI1_DP0_DEV, 0, 0x04, 0x02a00006);
	pci_cfg_write(PCI1_SEC_BUS, PCI1_DP0_DEV, 0, 0x14, HW_DP83816_0_BASE);

	/*
	 *  Port 1 DP83816 PCI master cycle, BAR
	 */
	pci_cfg_write(PCI1_SEC_BUS, PCI1_DP1_DEV, 0, 0x04, 0x02a00006);
	pci_cfg_write(PCI1_SEC_BUS, PCI1_DP1_DEV, 0, 0x14, HW_DP83816_1_BASE);

        /*
         * This is for new revision of Marvell (B1)
         *  It has a device id of 6485 which is not supported in bsd image
         *  so we are setting it to the old ID of 6480
         * pciwrite 0 1 0 0 648011ab
         * pciwrite 8 3 0 0 648011ab
         */
	pci_cfg_write(PCI0_PRI_BUS, PCI0_GT_DEV, 0, 0, 0x648011ab);
	pci_cfg_write(PCI1_PRI_BUS, PCI1_GT_DEV, 0, 0, 0x648011ab);

	/*
	 *  CF Controller I/O cycle
	 */
 // RoHS changes ---        
	// Check which device is installed
	pci_cfg_read(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x0, &data);
	if (data == PCI1_CF_DEV_INTEL)
		// Intel part
		// Enable master and i/o cycles on function 0 and 1
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x04, 0x02a00003);
						      
	else {
		// TI part
							   
		// Enable master and i/o cycles on function 0 and 1
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x04, 0x02a00003);
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 1, 0x04, 0x02a00003);

		// Set-up I/O BAR on function 0 and 1
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x2c, 0x400);
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x30, 0x100);

		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 1, 0x2c, 0x800);
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 1, 0x30, 0x100);

		// Set-up ExCA BAR -- only needed on function 0
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x44, 0x10000);

                // Set-up to enable IRQ3 & IRQ4 Interrupt
// per Brain's e-mail   back to original register values
//                pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x8c, 0x1022);
                pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x8c, 0x22);

		// Set-up system control register
// per Brain's e-mail   back to original register values
//                pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x80, 0x28449040);
		pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x80, 0x8449040);

                pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 0, 0x90, 0x406000c0);
                pci_cfg_write(PCI1_SEC_BUS, PCI1_CF_DEV, 1, 0x90, 0x406000c0);

		// Legacy support for 82365SL
		// Change ExCA id register to 0x82 instead of 0x84
                *(volatile uint8 *)0xb8010000 = 0;
                *(volatile uint8 *)0xb8010001 = 0x82;

                *(volatile uint8 *)0xb8010000 = 0x40;
                *(volatile uint8 *)0xb8010001 = 0x82;

	}

}	/* pci_init_bobcat() */

#endif /* R9000 */

