/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: pkt-alloc.c
 *
 * RCS:  $Id: pkt-alloc.c,v 1.12 2002/11/01 00:23:42 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains Packet Descriptor allocatiion routines
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 6/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/eee-api.h>
#include <exports/tpl-api.h>

#ifdef NOTUSED
/*-----------------------------------------------------------------
 * Name:        pkt_allocBuff
 *
 * Description: Allocates a a free Pkt buffer only.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
char *
pkt_allocBuff( void )
{
	pkt_buffRing_t	*pRing;
    char			*pBuff;
	uint32			*pToken;

	pRing = &pPktCfg->freeBuffRing;

	/*-------------------------------
	 * Make sure we have free buffer.
	 *-------------------------------
	 */
	if (pRing->rdAddr == pRing->wrAddr)
	{
		pPktCfg->stats.cntNoBuffErr++;
		return NULL;
	}

	/*-------------------------------
	 * Allocate from read ptr.
	 *-------------------------------
	 */
    pBuff = (*pRing->rdAddr);
	pToken = (uint32 *) ((uint32) pBuff & ~(PKT_BUFF_SIZE-1));
	*pToken = PKT_TOKEN_USED;

	/*-------------------------------
	 * Update free buff. ring read ptr.
	 * Handle wrap around.
	 *-------------------------------
	 */
	if (pRing->rdAddr == pRing->endAddr)
		pRing->rdAddr = pRing->startAddr;
	else
		pRing->rdAddr++;

	pRing->avail--;
	pPktCfg->stats.cntAllocBuff++;
    return (pBuff);
}
#endif

/*-----------------------------------------------------------------
 * Name:        pkt_allocDesc
 *
 * Description: Allocates a free Pkt Desc only (no buffer).
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
pkt_desc_t *
pkt_allocDesc( char *filename, int32 linenum )
{
    pkt_desc_t 		*pDesc;
    pkt_descRing_t 	*pRing;

	pRing = &(pPktCfg->freeDescRing);

	/*-------------------------------
	 * Make sure we have free Desc.
	 *-------------------------------
	 */
	if (PKT_RING_IS_EMPTY( pRing ))
	{
		pPktCfg->stats.cntNoPktDescErr++;
		return NULL;
	}

	/*-------------------------------
	 * Allocate from read ptr.
	 *-------------------------------
	 */
    pDesc = (*pRing->rdAddr);

	/*-------------------------------
	 * Update free desc. ring read ptr.
	 * Handle wrap around.
	 *-------------------------------
	 */
	PKT_RING_UPDATE_RD_PTR(pRing)

	pDesc->allocFile = filename;
	pDesc->allocLine = linenum;
	pDesc->freeFile  = NULL;;
	pDesc->freeLine  = 0;
	pDesc->flags     = PKT_DESC_INUSE;
	pDesc->pNext     = NULL;

	pRing->avail--;
	pPktCfg->stats.cntAllocDesc++;

	if (pRing->avail < pRing->lowWaterMark)
		pRing->lowWaterMark = pRing->avail; 
	

    return (pDesc);
}

/*-----------------------------------------------------------------
 * Name:        pkt_allocPkt
 *
 * Description: Allocates a single free Pkt Desc & Buffer.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
pkt_desc_t *
pkt_allocPkt(ushort16  pktUsr, int type, char *filename, int32 linenum)
{
    pkt_desc_t	*pDesc;
    eee_desc_t	*eDescp;
    char		*eBuffp;

#if defined(NFX_SMP)
	/* if ACPU, allocate from TPL_IPCQ_PKT_ALLOC */
	if (!CM_IS_TXRX_NCPU)
		return pkt_AcpuAllocPkt(pktUsr, type, filename, linenum);
#endif /* NFX_SMP */

	/*-------------------------------
	 * Make sure both packet desc. &
	 * buffer are not NULL.
	 *-------------------------------
	 */
	pDesc = pkt_allocDesc(filename, linenum );
	if (pDesc == NULL)
		return (NULL);

	/*-------------------------------
	 * Allocate a EEE packet.
	 *-------------------------------
	 */
    eBuffp = eee_allocateBuffer(EEE_SHARED_POOL, type, 0);
	if (eBuffp == NULL)
	{
		pPktCfg->stats.cntNoBuffErr++;
		PKT_FREE_DESC( pDesc );
		return (NULL);
	}

	eDescp = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
	if (eDescp == NULL)
	{
		pPktCfg->stats.cntNoEeeDescErr++;
		eee_deallocateBuffer( eBuffp );
		PKT_FREE_DESC( pDesc );
		return (NULL);
	}
	
	/* su, 06/22/02.
	 * increment buf counter
	 */
	pPktCfg->stats.cntAllocBuff++;

	bzero( eDescp, sizeof(eee_desc_t));


	/*-------------------------------
	 * Setup EEE descriptor & buffer.
	 *-------------------------------
	 */
	eDescp->bd[0].control = 0;
	eDescp->bd[1].buf     = eBuffp;
	if (type == EEE_BUF_ULTRA)
		eDescp->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_ULTRA;
	else
		eDescp->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_LARGE;
	eDescp->hdr.dest_port = 0;
	eDescp->hdr.src_port  = 0;
	eDescp->hdr.control   = (( 1 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);

	/*-------------------------------
	 * Put free buffer into pkt desc.
	 *-------------------------------
	 */
	pDesc->pHdr  = eBuffp + PKT_BUFF_INDENT;
	pDesc->flags = PKT_DESC_INUSE | (pktUsr & PKT_USR_MASK);
	pDesc->eDescp= eDescp;
	pDesc->pNext = NULL;

	/*-------------------------------
	 * Return packet desc.
	 *-------------------------------
	 */
	pPktCfg->stats.cntAllocPkt++;
    return (pDesc);
}

/*-----------------------------------------------------------------
 * Name:        pkt_allocChainPkt
 *
 * Description: Allocates a chained list of free Pkt Desc & Buffer.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
pkt_desc_t *
pkt_allocChainPkt(	ushort16  	pktUsr,
					int32 		count, 
					pkt_desc_t	**ppLast)
{
    pkt_desc_t 		*firstPktDesc;
    pkt_desc_t 		*pDesc;


	/*-------------------------------
	 * Make sure we have enough free
	 * desc. & buffers.
	 *-------------------------------
	 */
	if ((count >= pPktCfg->freeDescRing.avail) ||
		(count >= pPktCfg->freeBuffRing.avail))
		return NULL;

	/*-------------------------------
	 * Allocate first pkt.
	 *-------------------------------
	 */
	pDesc = PKT_ALLOC_PACKET( pktUsr );
	if (pDesc == NULL)
		return (NULL);

	/*-------------------------------
	 * Remember first packet desc.
	 *-------------------------------
	 */
	firstPktDesc = pDesc;

	while (--count)
	{
		pDesc->pNext = PKT_ALLOC_PACKET( pktUsr );
		pDesc = pDesc->pNext;
	}

	/*-------------------------------
	 * Null terminate last desc. in list.
	 *-------------------------------
	 */
	pDesc->pNext = NULL;

	/*-------------------------------
	 * Return first & last packet desc.
	 *-------------------------------
	 */
	*ppLast = pDesc;
	pPktCfg->stats.cntAllocChainPkt++;
    return (firstPktDesc);
}

