/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: pkt.h
 *
 * RCS:  $Id: pkt.h,v 1.36 2002/11/02 02:08:57 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains Packet Processing layer related data structure
 *              definitions.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 6/01/2001
 *
 *-----------------------------------------------------------------
 */

#ifndef _PKT_H
#define _PKT_H

#ifndef PROM
#define PKT_USE_TIME_STAMP		1
#endif

#define PKT_VALIDATE_ENABLE		1
#define PKT_DEBUG				1

#define PKT_LOGERR(_code, _desc)  pkt_logErr(__FILE__, __LINE__, _code, _desc)
#define PKT_PANIC( )			{ *(char *)0 = 1; }

#define PKT_FREE_PKT			0
#define PKT_HOLD_PKT			1

#define PKT_BUFF_INDENT			0x0080
#define PKT_BUFF_SIZE			0x0800 /* make it the same as EEE */

#define PKT_TOKEN_FREE			0x46524545		/* "FREE" */
#define PKT_TOKEN_USED			0x55534544		/* "USED" */

#define PKT_CHK_DUP				1

#ifdef  PROM
#define PKT_MAX_DESC_CNT		(1024)		/* Max. PKT descriptors */
#define PKT_MAX_BUFF_CNT		1024		/* Max. PKT buffers */
#else
#define PKT_MAX_DESC_CNT		(10*1024)	/* Max. PKT descriptors */
#define PKT_MAX_BUFF_CNT		2048		/* Max. PKT buffers */
#endif

/* Su, 08/31/02.
 * watermarks to allocate pkt per tpl conn.
 */
#define PKT_LOW_DESC_MARK		PKT_MAX_DESC_CNT/2

/* size less than 3 does not work */
#define PKT_CONN_UDP_CNT		64			/* pkt per udp connection */
#define PKT_CONN_TCP_CNT		64			/* pkt per tcp connection */
					
#define PKT_IP_VER				0x4			/* IP Version */
#define PKT_TYPE_IP				0x0800		/* IP protocol */
#define PKT_TYPE_TPID			0x8100		/* VLAN Tag Protocol ID */

#define PKT_ETHER_HDR_SIZE		14			/* Ether Type MAC hdr size */
#define PKT_IEEE_HDR_SIZE		22			/* 802.3 Type MAC hdr size */
#define PKT_VLAN_HDR_SIZE		18			/* VLAN  Type MAC hdr size */

#define PKT_IP_HDR_SIZE			20			/* IP hdr size (no option) */
#define PKT_UDP_HDR_SIZE		8			/* UDP hdr size */
#define PKT_TCP_HDR_SIZE		20			/* TCP hdr size (no option) */

#define PKT_HDR_INDENT			128			/* pkt header indent */

#define PKT_LINK_TYPE_ETHER		1
#define PKT_LINK_TYPE_IEEE		2
#define PKT_LINK_TYPE_VLAN		3


/*------------------------------------
 * Pkt Time Stamp related defines
 *------------------------------------
 */
#define PKT_TIME_IDX_LUC_RX             0
#define PKT_TIME_IDX_TPL_RX             1
#define PKT_TIME_IDX_NFS_RX             2
#define PKT_TIME_IDX_FS_REQ             3
#define PKT_TIME_IDX_FS_RSP             4    
#define PKT_TIME_IDX_TPL_TX             5   
#define PKT_TIME_IDX_LUC_TX             6
#define PKT_TIME_IDX_LUC_TX_CMP         7

#define PKT_MAX_TIME_IDX                (PKT_TIME_IDX_LUC_TX_CMP+1)


/*------------------------------------
 * Pkt errors
 *------------------------------------
 */
#define PKT_ERR_BAD_BUFF		(PKT_ERROR_BASE)
#define PKT_ERR_BAD_DESC		(PKT_ERR_BAD_BUFF - 1)
#define PKT_ERR_DUP_BUFF		(PKT_ERR_BAD_DESC - 1)
#define PKT_ERR_DUP_DESC		(PKT_ERR_DUP_BUFF - 1)
#define PKT_ERR_NO_BUFF			(PKT_ERR_DUP_DESC - 1)
#define PKT_ERR_NO_DESC			(PKT_ERR_NO_BUFF  - 1)
#define PKT_ERR_BUFF_RING_FULL	(PKT_ERR_NO_DESC  - 1)
#define PKT_ERR_DESC_RING_FULL	(PKT_ERR_BUFF_RING_FULL - 1)
#define PKT_ERR_MAX				(PKT_ERR_DESC_RING_FULL )

/*------------------------------------
 * Ether type MAC header.
 *------------------------------------
 */
typedef struct
{
	ushort16	dstMac[3];
	ushort16	srcMac[3];
	ushort16	type;
} pkt_ethHdr_t;

/*------------------------------------
 * IEEE 802.3 MAC header.
 *------------------------------------
 */
typedef struct
{
	ushort16	srcMac[3];
	ushort16	dstMac[3];
	ushort16	len;
	byte		dsap;
	byte		ssap;
	byte		ctrl;
	byte		oui[3];
	ushort16	type;
} pkt_ieeeHdr_t;

/*------------------------------------
 * VLAN MAC header.
 *------------------------------------
 */
typedef struct
{
	ushort16	srcMac[3];
	ushort16	dstMac[3];
	ushort16	tpid;
	ushort16	vid;
	ushort16	type;
} pkt_vlanHdr_t;

/*------------------------------------------------------------------------
 * Packet Descriptor data structure.
 *
 *
 * The following diagram illustrates how a packet is described using
 * a PKT Descriptor:
 *
 *
 *      |<-----------------------    pktSize  --------------------->|     
 *      |                       |<-- hdrSize -->|<---  dataSize --->|     
 *   +--+-----+------+----------+---------------+-------------------+
 *   |  | MAC |  IP  |  TCP/UDP |    APP HDR    |     APP DATA      |
 *   +--+-----+------+----------+---------------+-------------------+
 *   ^  ^     ^      ^          ^               ^                    
 *   |  |     |      |          |               |                     
 *   |  |     |      |          |               +----- cod            
 *   |  |     |      |          +--------------------- pHdr
 *   |  |     |      +-------------------------------- pTplHdr
 *   |  |     +--------------------------------------- pIpHdr
 *   |  +--------------------------------------------- pMac
 *   |                                           +---- eDescp
 *   |                                           |
 *   |                                           |
 *   |                                           +---> EEE Descriptor
 *   +------------------------------------------------ bd.buf
 *
 *
 *   Field     Description
 *   -----------------------------------------------------------
 *   eDescp  - points to EEE descriptor associated with this PKT.
 *   pMac    - points to start of MAC address.
 *   pIpHdr  - points to start of IP header.
 *   pTplHdr - points to start of TCP/UDP header.
 *   pHdr    - points to start of Application Header (e.g. RPC, NetBIOS etc).
 *   cod     - identifies USER DATA (e.g. NFS Write data) in GMC.
 *             (0 if no user data)
 *   pktSize - counts pkt size, including header & USER DATA in # of bytes.
 *   hdrSize - counts app. header size (excluding USER DATA) in # of bytes.
 *   dataSize- counts size of USER DATA (COD) in # of bytes.
 *   offset  - counts offset within the COD.
 *
 * The following table shows where each field is set on receive, on
 * transmit slow-path & fast-path:
 *
 *   Field     RX             TX Slow Path       Tx Fast Path
 *   ----------------------------------------------------------------
 *   pMac    - luc_rcvPkt()   TBD                tpl_addConnInd()
 *   pIpHdr  - luc_rcvPkt()   bsd_XmtPkt()       tpl_addConnInd()   
 *   pTplHdr - luc_rcvPkt()   bsd_XmtPkt()       tpl_addConnInd()   
 *   pHdr    - luc_rcvPkt()   bsd_XmtPkt()       tpl_addConnInd()   
 *   pktSize - luc_rcvPkt()   bsd_XmtPkt()       tpl_xmtPkt();
 *   hdrSize - luc_rcvPkt()   bsd_XmtPkt()       App. (e.g. NFS)
 *   cod     - luc_rcvPkt()   N/A                App. (e.g. NFS)
 *   dataSize- luc_rcvPkt()   N/A                App. (e.g. NFS)
 *   offset  - luc_rcvPkt()   N/A                App. (e.g. NFS)
 *   appInfo - N/A            N/A                App. (e.g. NFS)
 *   pTplConn- tpl_rcvPkt()   N/A                tpl_addConnInd()
 *   eDescp  - luc_rcvPkt()   tpl_allocPkt()     tpl_allocPkt()
 *
 *   laddr   - luc_rcvPkt()   App. (eg TFTP)     N/A
 *   raddr   - luc_rcvPkt()   App. (eg TFTP)     N/A
 *   lport   - luc_rcvPkt()   App. (eg TFTP)     N/A
 *   rport   - luc_rcvPkt()   App. (eg TFTP)     N/A
 *   tplProt - luc_rcvPkt()   N/A                N/A
 *   protocol- luc_rcvPkt()   N/A                N/A
 *
 *
 *------------------------------------------------------------------------
 */
typedef struct _pkt_desc
{
	uint32		flags;						/* Pkt desc. flags */
#define PKT_IP_CSUM_ENA			0x0001		/* Enable IP Chk Sum */
#define PKT_UDP_CSUM_ENA		0x0002		/* Enable UDP Chk Sum */
#define PKT_TCP_CSUM_ENA		0x0004		/* Enable TCP Chk Sum */
#define PKT_FAST_PATH			0x0008		/* PKT to/from Fast Path */
#define PKT_USR_MASK			0x00F0		/* PKT owner mask */
#define PKT_USR_TPL				0x0010		/* TPL owns pkt */
#define PKT_USR_PKT				0x0020		/* PKT owns pkt */
#define PKT_USR_TPL_CONN		0x0040		/* PKT belongs to tpl conn. */
#define PKT_USR_TPL_TX		    0x0080      /* PKT is sent, but no tx cmpl*/
#define PKT_USR_TCP_TX_Q		0x0100      /* PKT is in tcp send queue */
#define PKT_USR_TCP_TX_ACK    	0x0200      /* PKT is a TCP TX ACK */
#define PKT_USR_LUC_TX    		0x0400      /* PKT is in LUC driver TX */
#define PKT_RPC_MARKER			0x0800		/* PKT use RPC marker  */
#define PKT_USR_APPL_OWN		0x1000		/* PKT owned by appl, need tpl_xmt_cmp */
#define PKT_IP_FIRST_FRAG		0x2000		/* PKT first fragment, used for chksum calc. */
#define PKT_IP_FRAGMENT			0x4000		/* PKT is an IP fragment */
#define PKT_LUC_THRU			0x8000
#define PKT_DESC_INUSE			0x80000000  /* PKT is in use */

	ushort16	link;
	ushort16	protocol;					/* IP, ARP etc */
	uint32	    pktSize;					/* Packet size */
	uint32	    hdrSize;					/* header size */
	uint32	    dataSize;					/* COD data size */
	ushort16	tplProt;					/* Transport layer (TCP/UDP etc)*/
	ushort16	padSize;					/* padding size */
	uint32		chksum;
	char		*pMac;						/* Start of Mac Header */
	char		*pIpHdr;					/* Ip header start */
	char		*pTplHdr;					/* TCP/UDP header start */
	char		*pHdr;						/* Application Header */
	nfx_cod_t	cod;						/* COD */
	nfx_cod_entry_t		codEntry[NFX_MAX_COD_MSG];		/* COD entries */
	int32		numOfCod;					/* # of CODs */
	int32		offset;						/* COD Offset */
	int32		refCnt;						/* Ref. Count */
	void		*pTplConn;					/* Connection CB */
	uint32		tplConnId;					/* Id of tpl connection */
	void		*appInfo;					/* Cookie for App. layer */
	void		*eDescp;					/* EEE descriptor ptr */
	void		*gmcDescp;					/* GMC Data EEE desc. ptr */
	int32       camIdx;                     /* cam index, hardware only */   

	uint32		laddr;						/* local  ip addr */
	uint32		raddr;						/* remote ip addr */
	ushort16	lport;						/* local  port */
	ushort16	rport;						/* remote port */
	char		*allocFile;
	int32		allocLine;
	char		*freeFile;
	int32		freeLine;
	struct _pkt_desc 	*pTplNext;			/* TPL TX chain */
	struct _pkt_desc 	*pRefDesc;			/* Ref. Desc used by TPL Chain */
	struct _pkt_desc 	*pNext;				/* next pkt desc in chain */
	void		*pTime;						/* ptr to time stamp */
	uint32		sbmSeq;
} pkt_desc_t;

typedef struct 	_pkt_descRing
{
	pkt_desc_t	**startAddr;
	pkt_desc_t	**endAddr;
	pkt_desc_t	**rdAddr;
	pkt_desc_t	**wrAddr;
	int32		avail;
	int32		lowWaterMark;
} pkt_descRing_t;


#define PKT_RING_IS_EMPTY( _ringPtr )					\
	((_ringPtr->rdAddr) == (_ringPtr->wrAddr)) 			\

#define PKT_RING_IS_FULL( _ringPtr )					\
	(((_ringPtr->rdAddr) == (_ringPtr->wrAddr+1)) 	||	\
	 ((_ringPtr->rdAddr  == _ringPtr->startAddr)	&&	\
	 ((_ringPtr->wrAddr) == _ringPtr->endAddr)))		\

#define PKT_RING_UPDATE_RD_PTR( _ringPtr )				\
{														\
	if (_ringPtr->rdAddr == _ringPtr->endAddr)			\
		_ringPtr->rdAddr =  _ringPtr->startAddr;		\
	else												\
		_ringPtr->rdAddr++; 							\
}														\

#define PKT_RING_UPDATE_WR_PTR( _ringPtr )				\
{														\
	if (_ringPtr->wrAddr == _ringPtr->endAddr)			\
		_ringPtr->wrAddr =  _ringPtr->startAddr;		\
	else												\
		_ringPtr->wrAddr++; 							\
}														\

typedef struct 	_pkt_buffRing
{
	char		**startAddr;
	char		**endAddr;
	char		**rdAddr;
	char		**wrAddr;
	int32		avail;
} pkt_buffRing_t;


typedef struct	_pkt_stats
{
	uint32		cntBadCodLen;
	uint32		cntBadCopyLen;
	uint32		cntBadBuffErr;
	uint32		cntBadDescErr;
	uint32		cntDupBuffErr;
	uint32		cntDupDescErr;
	uint32		cntNoBuffErr;
	uint32		cntNoPktDescErr;
	uint32		cntNoEeeDescErr;
	uint32		cntBuffRingFull;
	uint32		cntDescRingFull;
	uint32		cntAllocBuff;
	uint32		cntAllocDesc;
	uint32		cntAllocPkt;
	uint32		cntAllocChainPkt;
	uint32		cntFreeBuff;
	uint32		cntFreeDesc;
	uint32		cntFreePkt;
	uint32		cntFreeChainPkt;

}	pkt_stats_t;

typedef struct	_pkt_cfg
{
	pkt_desc_t			*descStart;
	pkt_desc_t			*descEnd;
	char				*buffStart;
	char				*buffEnd;
	pkt_stats_t			stats;
	pkt_buffRing_t		freeBuffRing;
	pkt_descRing_t		freeDescRing;

}	pkt_cfg_t;

extern pkt_cfg_t		*pPktCfg;

#endif /* _PKT_H */
