/*
 *-----------------------------------------------------------------
 *
 * Name:	queue-api.h
 *
 * RCS:  $Id: queue-api.c,v 1.9 2002/10/28 21:01:46 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API function prototypes for general queue operations
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/queue-api.h>
#include "queue.h"
#include <sys/machine-dep.h>

queues_t	queues;
extern int q_isInited;

#ifdef NFX_QUEUE_DEBUG
/*-----------------------------------------------------------------
 * Name:	q_checkQueue()
 *
 * Description:	Sanity checks a queue
 *
 * Created by:	Jonathan Goldick
 *
 * Date Created:	10/19/02
 *
 *-----------------------------------------------------------------
 */
static void
q_checkQueue(
	q_headNodePtr_t	head_ptr
)
{
	NFX_ASSERT(head_ptr->in_use);

	if (head_ptr->size) {
		q_nodePtr_t	q_ptr,prev = NULL;
		uint32 count = 0;

		NFX_ASSERT(head_ptr->next != NULL);
		NFX_ASSERT(head_ptr->prev != NULL);

		q_ptr = head_ptr->next;
		while (q_ptr) {
			++count;
			if (count == 1) {
				// The first entry in the queue always has a NULL prev pointer
				NFX_ASSERT(q_ptr->prev == NULL);
			} else {
				NFX_ASSERT(q_ptr->prev == prev);
				NFX_ASSERT(q_ptr == prev->next);
			}
			prev = q_ptr;
			q_ptr = q_ptr->next;
			NFX_ASSERT(count <= head_ptr->size);
		}
		NFX_ASSERT(count == head_ptr->size);
	}
	else {
		NFX_ASSERT(head_ptr->next == NULL);
		NFX_ASSERT(head_ptr->prev == NULL);
	}
	return;

}	/* q_checkQueue() */
#endif /* NFX_QUEUE_DEBUG */

/*-----------------------------------------------------------------
 * Name:	q_createQueue()
 *
 * Description:	Create a new queue.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
q_headNodePtr_t					/* returns the queue head pointer or NULL
								 * if a queue couldn't be created
								 */
q_createQueue(
	uint32 item1Size,				/* Size of objects. */
	uint32 item2Size,				/* Size of objects. */
	queue_compare_func compare_func1,/* Comparision routine for item1. */
	queue_compare_func compare_func2/* Comparision routine for item2. */
)
{
	q_headNodePtr_t	head_ptr;
	uint32		i;

/*
	printf("q_createQueue\n");
*/
    if (!q_isInited)
        return NULL;

	if (queues.next_free_queue == Q_NO_FREE_QUEUE_INDEX)
		return NULL;
	else {
		if (queues.next_free_queue >= Q_MAX_QUEUES)
			return NULL;
	}

	head_ptr = &(queues.head_node[queues.next_free_queue]);

	if (((item1Size != sizeof(uint64)) && (item1Size != sizeof(uint32))) ||
		((item2Size != sizeof(uint64)) && (item2Size != sizeof(uint32)))) {
		// Only support pointers and simple integers to date.
		return NULL;
	}

	/*
	 * Get next free queue index
	 */
	for (i=queues.next_free_queue+1; i<Q_MAX_QUEUES; i++) {
		if (queues.head_node[i].in_use == 0) {
			queues.next_free_queue = i;
			break;
		}
	}
	if (i == Q_MAX_QUEUES)
		queues.next_free_queue = Q_NO_FREE_QUEUE_INDEX;

	queues.num_queues_avail--;

	head_ptr->read_ptr = 0;
	head_ptr->size = 0;
	head_ptr->in_use = 1;
	head_ptr->next = 0;
	head_ptr->prev = 0;
	head_ptr->item1Size = item1Size;
	head_ptr->item2Size = item2Size;
	head_ptr->compare_func1 = compare_func1;
	head_ptr->compare_func2 = compare_func2;
    head_ptr->owner = (void *)get_ra();

	return (head_ptr);

}	/* q_createQueue() */

/*-----------------------------------------------------------------
 * Name:	q_destroyQueue()
 *
 * Description:	Destroy a queue.  Further accesses to the queue will fail.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
void
q_destroyQueue(
	q_headNodePtr_t	head_ptr		/* queue head pointer */
)
{
	q_nodePtr_t	q_ptr, temp;
	uint32	index;

/*
	printf("q_destroyQueue\n");
*/

	index = head_ptr->index;

	/*
	 * Free queue nodes in the queue
	 */
	q_ptr = head_ptr->next;
	while (q_ptr) {
		temp = q_ptr;
		q_ptr = q_ptr->next;
		/*
		 * Push node back onto free list
		 */
		temp->next = queues.next_free_node;
		temp->prev = 0;
		temp->item1 = temp->item2 = 0;
		queues.next_free_node = temp;
		queues.num_nodes_avail++;
	}

	/*
	 * Indicate the queue is now available for a new create command.
	 */
	queues.head_node[index].in_use = 0;
	queues.head_node[index].read_ptr = 0;
	queues.head_node[index].size = 0;
	queues.head_node[index].next = 0;
	queues.head_node[index].prev = 0;

	if (index < queues.next_free_queue)
		queues.next_free_queue = index;

	queues.num_queues_avail++;

}	/* q_destroyQueue() */

/*-----------------------------------------------------------------
 * Name:	q_enQueue()
 *
 * Description:	Adds a new node to the 'end' of the queue.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK,
								 * Q_ERR_NO_AVAILABLE_NODE
								 * Q_ERR_INVALID_QUEUE
								 */
q_enQueue(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	q_nodePtr_t	new_q_node;

	if (!(head_ptr->in_use))
		return Q_ERR_INVALID_QUEUE;
		
	if ((new_q_node = q_getFreeNode()) == NULL)
		return Q_ERR_NO_AVAILABLE_NODE;

	if (head_ptr->item1Size == sizeof(uint64)) {
		new_q_node->item1 = *((uint64 *)item1);
	} else {
		new_q_node->item1 = *((uint32 *)item1);
	}
	if (head_ptr->item2Size == sizeof(uint64)) {
		new_q_node->item2 = *((uint64 *)item2);
	} else {
		new_q_node->item2 = *((uint32 *)item2);
	}

	if (head_ptr->next)
		head_ptr->prev->next = new_q_node;
	else
		head_ptr->next = new_q_node;

	new_q_node->prev = head_ptr->prev;
	head_ptr->prev = new_q_node;
	new_q_node->next = 0;

	queues.head_node[head_ptr->index].size++;

	return NFX_OK;

}	/* q_enQueue() */

/*-----------------------------------------------------------------
 * Name:	q_deQueue()
 *
 * Description:	removes top entry from the queue
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 * Q_ERR_INVALID_QUEUE
								 */
q_deQueue(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	q_nodePtr_t	temp;

	if (!(head_ptr->in_use))
		return Q_ERR_INVALID_QUEUE;

	if (head_ptr->next) {
		if (head_ptr->item1Size == sizeof(uint64)) {
			*((uint64 *)item1) = head_ptr->next->item1;
		} else {
			*((uint32 *)item1) = (uint32)head_ptr->next->item1;
		}
		if (head_ptr->item2Size == sizeof(uint64)) {
			*((uint64 *)item2) = head_ptr->next->item2;
		} else {
			*((uint32 *)item2) = (uint32)head_ptr->next->item2;
		}

		temp = head_ptr->next;
		head_ptr->next = head_ptr->next->next;
		if (head_ptr->next)
			head_ptr->next->prev = 0;
		else
			head_ptr->prev = 0;

		/*
		 * Push node back onto free list
		 */
		temp->next = queues.next_free_node;
		temp->prev = 0;
		temp->item1 = temp->item2 = 0;
		queues.next_free_node = temp;
		queues.num_nodes_avail++;
	}
	else {
		return Q_ERR_EMPTY_QUEUE;
	}

	head_ptr->size--;

	return NFX_OK;

}	/* q_deQueue() */

#define q_isQueueEmpty(_q_ptr)		(((_q_ptr)->size == 0) ? 1 : 0)

/*-----------------------------------------------------------------
 * Name:	q_resetReadPointer()
 *
 * Description:	The internal pointer can be reset back to the head of
 *				the queue with this function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_QUEUE_EMPTY
								 */
q_resetReadPointer(
	q_headNodePtr_t	head_ptr		/* queue head pointer */
)
{
	if (head_ptr->size)
		head_ptr->read_ptr = head_ptr->next;
	else {
		head_ptr->read_ptr = NULL;
		return Q_ERR_EMPTY_QUEUE;
	}

	return NFX_OK;

}	/* q_resetReadPointer() */

/*-----------------------------------------------------------------
 * Name:	q_previewQueueItem()
 *
 * Description:	Reads the user data, but does not alter the read pointer
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/25/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_previewQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	if (head_ptr->read_ptr) {
		if (head_ptr->item1Size == sizeof(uint64)) {
			*((uint64 *)item1) = head_ptr->read_ptr->item1;
		} else {
			*((uint32 *)item1) = (uint32)head_ptr->read_ptr->item1;
		}
		if (head_ptr->item2Size == sizeof(uint64)) {
			*((uint64 *)item2) = head_ptr->read_ptr->item2;
		} else {
			*((uint32 *)item2) = (uint32)head_ptr->read_ptr->item2;
		}
	}
	else
		return Q_ERR_EMPTY_QUEUE;

	return NFX_OK;

}	/* q_previewQueueItems() */

/*-----------------------------------------------------------------
 * Name:	q_findQueueItem()
 *
 * Description:	This function searches the list for a specific item.
 *				If Q_FIND_ONE is used, we search for item1 and return
 *				the found value in item2.  The opposite is done for 
 *				Q_FIND_TWO.  Q_FIND_BOTH searches for both and doesn't
 *				update either item.  The internal read pointer will be
 *				advanced.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_ITEM_NOT_PRESENT
								 */
q_findQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2,			/* pointer to user data */
	uint32			control			/* Q_FIND_ONE, Q_FIND_TWO, Q_FIND_BOTH */
)
{
	uint64  	i1, i2;
	q_nodePtr_t	q_ptr;

	if (q_resetReadPointer(head_ptr) != NFX_OK)
		return Q_ERR_ITEM_NOT_PRESENT;

	for (q_ptr=head_ptr->read_ptr; head_ptr->read_ptr != NULL;
		 head_ptr->read_ptr = head_ptr->read_ptr->next) {
		i1 = i2 = 0;
		q_previewQueueItem(head_ptr, (void *)&i1, (void *)&i2);
		switch (control) {
			case Q_FIND_ONE:
				if ((*head_ptr->compare_func1)(item1, (void *)&i1) == 0) {
					if (head_ptr->item2Size == sizeof(uint64)) {
						*((uint64 *)item2) = i2;
					} else {
						*((uint32 *)item2) = (uint32)i2;
					}
					return NFX_OK;
				}
				break;

			case Q_FIND_TWO:
				if ((*head_ptr->compare_func2)(item2, (void *)&i2) == 0) {
					if (head_ptr->item1Size == sizeof(uint64)) {
						*((uint64 *)item1) = i1;
					} else {
						*((uint32 *)item1) = (uint32)i1;
					}
					return NFX_OK;
				}
				break;

			case Q_FIND_BOTH:
				if (((*head_ptr->compare_func1)(item1, (void *)&i1) == 0) &&
					((*head_ptr->compare_func2)(item2, (void *)&i2) == 0)) {
					return NFX_OK;
				}
				break;

			default:
				return Q_ERR_ITEM_NOT_PRESENT;
		}
	}

	return Q_ERR_ITEM_NOT_PRESENT;

}	/* q_findQueueItem() */

/*-----------------------------------------------------------------
 * Name:	q_readQueueItem()
 *
 * Description:	The user data within a found node can be read using this
 *				function.  The internal read pointer will be advanced.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_readQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	if (head_ptr->read_ptr) {
		if (head_ptr->item1Size == sizeof(uint64)) {
			*((uint64 *)item1) = head_ptr->read_ptr->item1;
		} else {
			*((uint32 *)item1) = (uint32)head_ptr->read_ptr->item1;
		}
		if (head_ptr->item2Size == sizeof(uint64)) {
			*((uint64 *)item2) = head_ptr->read_ptr->item2;
		} else {
			*((uint32 *)item2) = (uint32)head_ptr->read_ptr->item2;
		}
		head_ptr->read_ptr = head_ptr->read_ptr->next;
	}
	else
		return Q_ERR_EMPTY_QUEUE;

	return NFX_OK;

}	/* q_readQueueItem() */

/*-----------------------------------------------------------------
 * Name:	q_backupQueueReadPointer()
 *
 * Description:	back track one item in the queue
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_backupQueueReadPointer(
	q_headNodePtr_t	head_ptr
)
{

	return NFX_OK;

}	/* q_backupQueueReadPointer() */

/*-----------------------------------------------------------------
 * Name:	q_readNextQueueItem()
 *
 * Description:	The current read pointer can be advanced and the user data
 *				from the next node read.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_readNextQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	if (head_ptr->read_ptr) {
		head_ptr->read_ptr = head_ptr->read_ptr->next;
		if (head_ptr->read_ptr) {
			if (head_ptr->item1Size == sizeof(uint64)) {
				*((uint64 *)item1) = head_ptr->read_ptr->item1;
			} else {
				*((uint32 *)item1) = (uint32)head_ptr->read_ptr->item1;
			}
			if (head_ptr->item2Size == sizeof(uint64)) {
				*((uint64 *)item2) = head_ptr->read_ptr->item2;
			} else {
				*((uint32 *)item2) = (uint32)head_ptr->read_ptr->item2;
			}
		}
		else
			return Q_ERR_EMPTY_QUEUE;
	}
	else
		return Q_ERR_EMPTY_QUEUE;

	return NFX_OK;

}	/* q_readNextQueueItem() */

/*-----------------------------------------------------------------
 * Name:	q_deleteQueueItem()
 *
 * Description:	This function deletes the node pointed to be the read pointer.
 *				The read pointer was set via the findQueueItem function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_deleteQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	q_nodePtr_t	temp;

	if (head_ptr->read_ptr) {
		if (item1) {
			if (head_ptr->item1Size == sizeof(uint64)) {
				*((uint64 *)item1) = head_ptr->read_ptr->item1;
			} else {
				*((uint32 *)item1) = (uint32)head_ptr->read_ptr->item1;
			}
		}
		if (item2) {
			if (head_ptr->item2Size == sizeof(uint64)) {
				*((uint64 *)item2) = head_ptr->read_ptr->item2;
			} else {
				*((uint32 *)item2) = (uint32)head_ptr->read_ptr->item2;
			}
		}
		temp = head_ptr->read_ptr;
		if ((temp->next == NULL) && (temp->prev == NULL)) {
			head_ptr->next = NULL;
			head_ptr->prev = NULL;
			head_ptr->read_ptr = NULL;
		}
		else if (temp->prev == NULL) {
			head_ptr->next = temp->next;
			temp->next->prev = NULL;
			head_ptr->read_ptr = temp->next;
		}
		else if (temp->next == NULL) {
			temp->prev->next = NULL;
			head_ptr->prev = temp->prev;
			head_ptr->read_ptr = temp->prev;
		}
		else {
			temp->prev->next = temp->next;
			temp->next->prev = temp->prev;
			head_ptr->read_ptr = temp->next;
		}

		head_ptr->size--;
		/*
		 * Push node back onto free list
		 */
		temp->next = queues.next_free_node;
		temp->prev = 0;
		temp->item1 = temp->item2 = 0;
		queues.next_free_node = temp;
		queues.num_nodes_avail++;
	}
	else
		return Q_ERR_EMPTY_QUEUE;

	return NFX_OK;

}	/* q_deleteQueueItem() */

/*-----------------------------------------------------------------
 * Name:	q_changeQueueItem()
 *
 * Description:	After finding a node with a specific item, the data inside
 *				that node can be changed.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_changeQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	if (head_ptr->read_ptr) {
		if (head_ptr->item1Size == sizeof(uint64)) {
			head_ptr->read_ptr->item1 = *((uint64 *)item1);
		} else {
			head_ptr->read_ptr->item1 = *((uint32 *)item1);
		}
		if (head_ptr->item2Size == sizeof(uint64)) {
			head_ptr->read_ptr->item2 = *((uint64 *)item2);
		} else {
			head_ptr->read_ptr->item2 = *((uint32 *)item2);
		}
	}
	else
		return Q_ERR_EMPTY_QUEUE;

	return NFX_OK;

}	/* q_changeQueueItem() */

/*-----------------------------------------------------------------
 * Name:	q_insertBeforeItem()
 *
 * Description:	A new node can be inserted into the list before a found node.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_NO_AVAILABLE_NODE
								 */
q_insertBeforeItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	q_nodePtr_t	q_ptr;
	q_nodePtr_t	new;

	if (queues.num_nodes_avail) {
		new = queues.next_free_node;
		queues.next_free_node = queues.next_free_node->next;
		queues.num_nodes_avail--;

		if (head_ptr->read_ptr) {
			q_ptr = head_ptr->read_ptr;

			if (q_ptr->prev) {
				/*
				 * not first in queue
				 */
				q_ptr->prev->next = new;
				new->next = q_ptr;
				new->prev = q_ptr->prev;
				q_ptr->prev = new;
				head_ptr->read_ptr = new;
			}
			else {
				/*
				 * First in queue
				 */
				new->prev = NULL;
				q_ptr->prev = new;
				new->next = head_ptr->next;
				head_ptr->next = new;
				head_ptr->read_ptr = new;
			}
		}
		else {
			if (head_ptr->prev == NULL) {
				/*
				 * empty queue, add new entry
				 */
				head_ptr->next = new;
				head_ptr->prev = new;
				new->next = NULL;
				new->prev = NULL;
				head_ptr->read_ptr = new;
			}
			else {
				/*
				 * read_pointer is at end of queue
				 */
				head_ptr->prev->next = new;
				new->prev = head_ptr->prev;
				head_ptr->prev = new;
				new->next = NULL;
				head_ptr->read_ptr = NULL;
			}
		}

		if (head_ptr->item1Size == sizeof(uint64)) {
			new->item1 = *((uint64 *)item1);
		} else {
			new->item1 = *((uint32 *)item1);
		}
		if (head_ptr->item2Size == sizeof(uint64)) {
			new->item2 = *((uint64 *)item2);
		} else {
			new->item2 = *((uint32 *)item2);
		}

		head_ptr->size++;

		return NFX_OK;
	}
	else
		return Q_ERR_NO_AVAILABLE_NODE;

}	/* q_insertBeforeItem() */
	
/*-----------------------------------------------------------------
 * Name:	q_insertAfterItem()
 *
 * Description:	A new node can be inserted into the list after a found node.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_NO_AVAILABLE_NODE
								 */
q_insertAfterItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
)
{
	q_nodePtr_t	q_ptr;
	q_nodePtr_t	new;

	if (queues.num_nodes_avail) {
		new = queues.next_free_node;
		queues.next_free_node = queues.next_free_node->next;
		queues.num_nodes_avail--;

		if (head_ptr->read_ptr) {
			q_ptr = head_ptr->read_ptr;

			if (q_ptr->next) {
				/*
				 * Not last in queue
				 */
				new->next = q_ptr->next;
				q_ptr->next->prev = new;
				q_ptr->next = new;
				new->prev = q_ptr;
				head_ptr->read_ptr = new;
			}
			else {
				/*
				 * Last in queue
				 */
				q_ptr->next = new;
				new->prev = q_ptr;
				new->next = NULL;
				head_ptr->prev = new;
				head_ptr->read_ptr = new;
			}
		}
		else {
			if (head_ptr->prev == NULL) {
				/*
				 * empty queue, add new entry
				 */
				head_ptr->next = new;
				head_ptr->prev = new;
				new->next = NULL;
				new->prev = NULL;
				head_ptr->read_ptr = new;
			}
			else {
				/*
				 * read_pointer is at end of queue
				 */
				head_ptr->prev->next = new;
				new->prev = head_ptr->prev;
				head_ptr->prev = new;
				new->next = NULL;
				head_ptr->read_ptr = NULL;
			}
		}

		if (head_ptr->item1Size == sizeof(uint64)) {
			new->item1 = *((uint64 *)item1);
		} else {
			new->item1 = *((uint32 *)item1);
		}
		if (head_ptr->item2Size == sizeof(uint64)) {
			new->item2 = *((uint64 *)item2);
		} else {
			new->item2 = *((uint32 *)item2);
		}

		head_ptr->size++;

		return NFX_OK;
	}
	else
		return Q_ERR_NO_AVAILABLE_NODE;

}	/* q_insertAferItem() */
	
/*-----------------------------------------------------------------
 * Name:	q_insertItemInOrder()
 *
 * Description:	This is essentially a combination of q_findQueueItem() and
 *				q_insertBeforeItem().  Either item1 or item2 can be used
 *				as the comparison key, BUT NOT BOTH.
 *				The caller should supply a comparison function which
 *				will be called to compare the node item with the stored item.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_NO_AVAILABLE_NODE
								 */
q_insertItemInOrder(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2,			/* pointer to user data */
	uint32			control			/* Q_FIND_ONE or Q_FIND_TWO
									 *			and
									 * Q_INSERT_ORDER_ASCENDING or
									 * Q_INSERT_ORDER_DESCENDING
									 */
)
{
	q_nodePtr_t	next;
	uint32		found = 0;
	uint32		ascend = 0;
	int32		rc;

	/*
	 * If no entry in list, add new entry
	 */
	if (q_resetReadPointer(head_ptr) == Q_ERR_EMPTY_QUEUE)
		return (q_insertAfterItem(head_ptr, item1, item2));

	/*
	 * If list has entries, search for where to insert new entry
	 */
	if (control & Q_INSERT_ORDER_ASCENDING)
		ascend = 1;

	if (ascend) {
		while (!found) {
			next = head_ptr->read_ptr->next;

			/*
			 * If node data is == or > new data, then insert new node
			 */
			if (control & Q_FIND_ONE) {
				rc = (*head_ptr->compare_func1)((void *)&head_ptr->read_ptr->item1,
												item1);
			}
			else {
				rc = (*head_ptr->compare_func2)((void *)&head_ptr->read_ptr->item2,
												item2);
			}
			if (rc >= 0)
				return (q_insertBeforeItem(head_ptr, item1, item2));

			/*
			 * If we checked the last item in the list, add new node at end
			 */
			if (next == NULL)
				return (q_insertAfterItem(head_ptr, item1, item2));

			head_ptr->read_ptr = next;
		}
	}
	else /* descending order */ {
		while (!found) {
			next = head_ptr->read_ptr->next;

			/*
			 * If node data is == or > new data, then insert new node
			 */
			if (control & Q_FIND_ONE) {
				rc = (*head_ptr->compare_func1)((void *)&head_ptr->read_ptr->item1,
												item1);
			}
			else {
				rc = (*head_ptr->compare_func2)((void *)&head_ptr->read_ptr->item2,
												item2);
			}
			if (rc <= 0)
				return (q_insertBeforeItem(head_ptr, item1, item2));

			/*
			 * If we checked the last item in the list, add new node at end
			 */
			if (next == NULL)
				return (q_insertAfterItem(head_ptr, item1, item2));

			head_ptr->read_ptr = next;
		}
	}

	return Q_ERR_NO_AVAILABLE_NODE;
}	/* q_insertItemInOrder() */

