/*
 *------------------------------------------------------------------------------
 *
 * Name:	raid-device.c
 *
 * RCS:		$Id: raid-device.c,v 1.5 2002/11/09 00:09:03 nitinb Exp $
 *
 * Copyright (C) 2001, 2002 Agile Storage Inc.
 *
 * Description:	RAID config manager device discover and management.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 4/16/02
 *
 *------------------------------------------------------------------------------
 */

#include "raid-common.h"
#include "raid-device.h"
#include "raid-xyratex.h"
#include <sys/time.h>
#include <signal.h>

#define RAID_DEVICE_INITIAL_POLL_SECS		1
#define RAID_DEVICE_INTERVAL_POLL_SECS		10
#define RAID_SDM_TIMEOUT			5

typedef struct _raid_timer_state_ {
	int			timer_sighandler_registered;
	int			timer_running;
	int			timer_popped;
	struct itimerval	timer_val;
	int			timer_sig;
	void (*timer_sighandler)(int);
	void (*timer_start_callout)(struct _raid_timer_state_ *);
	void (*timer_stop_callout)(struct _raid_timer_state_ *);
	void (*timer_popped_callout)(struct _raid_timer_state_ *);
} raid_timer_state_t;

void raid_device_timer_sighandler(int);
void raid_device_timer_stop(raid_timer_state_t *);
void raid_device_timer_start(raid_timer_state_t *);
void raid_device_timer_popped(raid_timer_state_t *);
void raid_compress_trailing_spaces(char *);

raid_timer_state_t * raid_device_timer = NULL;
uint32		     raid_device_poll_count = 0;
raid_dev_t *	     raid_devs = NULL;

void
raid_device_management_init(void)
{
	/* Initialize the device management timer structure */
	raid_device_timer = (raid_timer_state_t *)malloc(sizeof(raid_timer_state_t));
	if (raid_device_timer == NULL)
		panic("malloc failed in raid_device_management_init: %s\n", strerror(errno));
	
	bzero(raid_device_timer, sizeof(raid_timer_state_t));

	raid_device_timer->timer_val.it_interval.tv_sec = RAID_DEVICE_INTERVAL_POLL_SECS;
	raid_device_timer->timer_val.it_value.tv_sec = RAID_DEVICE_INITIAL_POLL_SECS;
	raid_device_timer->timer_sig = SIGALRM;
	raid_device_timer->timer_sighandler = raid_device_timer_sighandler;
	raid_device_timer->timer_start_callout = raid_device_timer_start;
	raid_device_timer->timer_stop_callout = raid_device_timer_stop;
	raid_device_timer->timer_popped_callout = raid_device_timer_popped;

	/* Start the timer */
	(*raid_device_timer->timer_start_callout)(raid_device_timer);
}

void
raid_device_management_poll(void)
{
	if (raid_device_timer->timer_popped) {
		--raid_device_timer->timer_popped;
		(*raid_device_timer->timer_popped_callout)(raid_device_timer);
		/* Restart timer */
		(*raid_device_timer->timer_start_callout)(raid_device_timer);
	}
}

void  
raid_device_timer_sighandler(int sig)
{
	++raid_device_timer->timer_popped;
}

void
raid_device_timer_stop(raid_timer_state_t *timer_state)
{
}

void
raid_device_timer_start(raid_timer_state_t *timer_state)
{
	if (!timer_state->timer_sighandler_registered) {
		if (signal(raid_device_timer->timer_sig, raid_device_timer->timer_sighandler) == SIG_ERR)
			panic("signal failed in raid_device_timer_start.\n");
		raid_device_timer->timer_sighandler_registered = 1;
	}

	if (!timer_state->timer_running) {
		if (setitimer(ITIMER_REAL, &raid_device_timer->timer_val, NULL) < 0)
			panic("setitimer failed in raid_device_timer_start: %s\n", strerror(errno));
		timer_state->timer_running = 1;
	}
}

void
raid_device_timer_popped(raid_timer_state_t *timer_state)
{
	int			i;
	int			rc;
	char			respbuff[65536];
	sdm_devInfoReqMsg_t	dev_info_req;
	sdm_devInfoRspMsg_t	dev_info_rsp;
	sdm_devListReqMsg_t	dev_list_req;
	sdm_devListRspMsg_t *	dev_list_rsp;
	raid_dev_t *		dev;
	struct timeval		to;

	++raid_device_poll_count;

	RAID_DBG(10, ("raid_device_timer_popped %d\n", raid_device_poll_count));

	/* 
	 * Step 1: Look for existing raid devices
	 */
	to.tv_sec = RAID_SDM_TIMEOUT;
	to.tv_usec = 0;

	bzero(&dev_list_req, sizeof(sdm_devListReqMsg_t));
	dev_list_req.hdr.mid = SDM_MID_DEV_LIST_REQ;
	dev_list_req.type = SDM_DEVLIST_ALL_DISK;
	dev_list_req.port = SDM_DEVLIST_PORTALL;
	rc = sendAgileMsg(-1, 
			  &dev_list_req, sizeof(sdm_devListReqMsg_t),
			  "sdm", 0, 0, 0, &to,
			  respbuff, sizeof(respbuff));
	if (rc <= 0) {
		RAID_DBG(2, ("%s failed in %s: %s\n",
			     "sendAgileMsg", "raid_device_management_thread", strerror(errno)));
		goto sdm_failure;
	}
		
	dev_list_rsp = (sdm_devListRspMsg_t *)respbuff;

	if (dev_list_rsp->hdr.mid != SDM_MID_DEV_LIST_RSP ) {
		RAID_DBG(2, ("Invalid Response Received in %s: %x\n", 
			     "raid_device_management_thread", dev_list_rsp->hdr.mid));
		goto sdm_failure;
	}

	for (i = 0; i < dev_list_rsp->numEntries; i++) {
		if ((dev = raid_device_lookup_by_id(dev_list_rsp->idList[i])) == NULL) {
			/* New device -- go get SDM info for this device */
			dev_info_req.hdr.mid = SDM_MID_DEV_INFO_REQ;
			dev_info_req.id = dev_list_rsp->idList[i];
			bzero(&dev_info_rsp, sizeof(sdm_devInfoRspMsg_t));
			rc = sendAgileMsg(-1, 
					  &dev_info_req, sizeof(sdm_devInfoReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  &dev_info_rsp, sizeof(sdm_devInfoRspMsg_t));
			if (rc < 0) {
				RAID_DBG(2, ("%s failed in %s: %s\n",
					     "sendAgileMsg", "raid_device_management_thread", strerror(errno)));
				goto sdm_failure;
			}
			if (dev_info_rsp.hdr.mid != SDM_MID_DEV_INFO_RSP ) {
				RAID_DBG(2, ("Invalid Response Received in %s: %x\n",
					     "raid_device_management_thread", dev_info_rsp.hdr.mid));
				goto sdm_failure;
			}
			if (dev_info_rsp.rc == SDM_FAIL ) {
				RAID_DBG(2, ("%s failed in %s: Device Not Found %llx\n", 
					     "SDM_MID_DEV_INFO_REQ", "raid_device_management_thread", dev_info_req.id));
				goto sdm_failure;
			}

			/* Filter only those devices we want -- only MYLEX for now XXX */
			if (memcmp(dev_info_rsp.st.vendor, "MYLEX", strlen("MYLEX")) == 0 &&
			    dev_info_rsp.lun == 0) {
				RAID_DBG(1, ("Adding new device: id %016llx, WWN %016llx, LUN %d, vendor '%s', model '%s', name '%s'\n", 
					     dev_info_rsp.id,
					     dev_info_rsp.wwn, dev_info_rsp.lun,
					     dev_info_rsp.st.vendor, dev_info_rsp.st.model, dev_info_rsp.st.name));
				dev = raid_device_new(&dev_info_rsp);
				
				/* Start device state machine */
				raid_sm_create(dev);

				dev->poll_ref_number = raid_device_poll_count;
			}
			else {
				RAID_DBG(10, ("Rejecting device: id %016llx, WWN %016llx, LUN %d, vendor '%s', model '%s'\n", 
					     dev_info_rsp.id,
					     dev_info_rsp.wwn, dev_info_rsp.lun,
					     dev_info_rsp.st.vendor, dev_info_rsp.st.model));
			}
		}
		else 
			dev->poll_ref_number = raid_device_poll_count;
	} /* for (i < dev_list_rsp->numEntries) */

	/* 
	 * Step 2: Remove non-existing raid devices
	 */
	if ((dev = raid_devs)) {
		raid_dev_t *	next;
		do {
			next = dev->next;
			if (dev->poll_ref_number != raid_device_poll_count) {
				RAID_DBG(1, ("Deleting old raid device: id %016llx, vendor '%s', model '%s'\n", 
					     dev->id,
					     dev->vendor_name,
					     dev->model_name));
				raid_sm_run(dev, NULL, RAID_EVENT_DEVICE_REMOVED);
			}
			dev = next;
		} while (raid_devs && dev != raid_devs);
	}

 sdm_failure:
}

raid_dev_t *
raid_device_new(sdm_devInfoRspMsg_t *dev_info)
{
	raid_dev_t *	dev;

	dev = (raid_dev_t *)malloc(sizeof(raid_dev_t));
	if (dev == NULL)
		panic("malloc failed in raid_device_new: %s\n", strerror(errno));
	bzero(dev, sizeof(raid_dev_t));

	dev->id = dev_info->id;
	dev->wwn = dev_info->wwn;
	dev->lun = dev_info->lun;
	memcpy(dev->vendor_name, dev_info->st.vendor, SDM_VENDOR_NAME_LEN);
	memcpy(dev->model_name, dev_info->st.model, SDM_MODEL_NAME_LEN);
	memcpy(dev->device_name, dev_info->st.name, SDM_DEV_NAME_LEN);
	raid_compress_trailing_spaces(dev->vendor_name);
	raid_compress_trailing_spaces(dev->model_name);

	/* XXX Assume MYLEX for now */
	dev->dev_cmd_transformer = xyratex_transformer;

	/* Insert into list */
	if (raid_devs == NULL) {
		dev->next = dev->prev = dev;
		raid_devs = dev;
	}
	else {
		dev->prev = raid_devs->prev;
		dev->next = raid_devs->prev->next;
		raid_devs->prev->next = dev;
		raid_devs->prev = dev;
	}

	return(dev);
}

raid_dev_t *
raid_device_lookup_by_id(scsi_id_t id)
{
	raid_dev_t *    dev;

	if ((dev = raid_devs)) {
		do {
			if (id == dev->id)
				return(dev);
			dev = dev->next;
		} while (dev != raid_devs);
	}
	return(NULL);
}

raid_dev_t *
raid_device_lookup_by_devname(char *devname)
{
	raid_dev_t *    dev;

	if ((dev = raid_devs)) {
		do {
			if (strcmp(devname, dev->device_name) == 0)
				return(dev);
			dev = dev->next;
		} while (dev != raid_devs);
	}
	return(NULL);
}

void
raid_device_destroy(raid_dev_t *dev)
{
	dev->prev->next = dev->next;
	dev->next->prev = dev->prev;

	if (raid_devs == dev) {
		if (dev->next == dev)
			raid_devs = NULL;
		else 
			raid_devs = dev->next;
	}

	free(dev);
}

void
raid_compress_trailing_spaces(char *str)
{
	char *p;
	
	if (strlen(str) == 0)
		return;
	p = &str[strlen(str) - 1];
	while (strlen(str) > 0 && *p == ' ')
		*p-- = '\0';
}
