/*
 *-----------------------------------------------------------------
 *
 * Name:	pte.c
 *
 * RCS:  $Id: pte.c,v 1.22 2002/11/12 18:59:04 janv Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: Page Table functions for TLB refill
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */

#ifdef LARGE_MEMORY

#include <nfx-incl.h>
#include <sys/mips.h>
#include <exports/pte-api.h>
#include <exports/mem-api.h>
#ifdef USE_THREAD_STACKS
#include <exports/esm-api.h>
#endif

pte_t	lm_table[MAX_PTE_TABLE_ENTRIES];
pte_t	sm_txrx_table[MAX_PTE_TABLE_ENTRIES];
pte_t	sm_fp_table[MAX_PTE_TABLE_ENTRIES];
pte_t	sac_table[MAX_PTE_TABLE_ENTRIES];
pte_t	thread_stack_table[MAX_PTE_TABLE_ENTRIES];
pte_t	sm_txrx_cached_table[MAX_PTE_TABLE_ENTRIES];
pte_t	sm_fp_cached_table[MAX_PTE_TABLE_ENTRIES];

uint64		mmu_region_size[MMU_NUM_TABLES+1];
address_t	mmu_phys_offset_start[MMU_NUM_TABLES+1];/* in contig. memory space */
address_t	mmu_phys_offset_end[MMU_NUM_TABLES+1];	/* in contig. memory space */
address_t	mmu_phys_start[MMU_NUM_TABLES+1];		/* actual physical address */
address_t	mmu_phys_end[MMU_NUM_TABLES+1];		/* actual physical address */
address_t	mmu_virt_start[MMU_NUM_TABLES+1];
address_t	mmu_virt_end[MMU_NUM_TABLES+1];
pte_t		* mmu_tptr[MMU_NUM_TABLES+1];


uint64	memorysize = 0;
uint32	memorysize_MB = 0;

/*-----------------------------------------------------------------
 * Name:	pte_setup
 *
 * Description: Set up Page Table for specified 'region'.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */
void
pte_setup(
	uint32		region,			/*	MMU_MALLOC_REGION,
									MMU_SM_TXRX_REGION,
									MMU_SAC_REGION,
									MMU_SM_FP_REGION,
									MMU_THREAD_STACK_REGION
								 */
	uint64		phys_offset,	/* offset from 0, assuming contiguous memory */
	address_t	virt,			/* virtual address */
	uint64		phys_size,		/* of entire region, in bytes */
	uint64		virt_size,		/* of entire region, in bytes */
	uint32		pgsize			/* 8K,32K,128K,512K,2M,8M,32M */
)
{
	pte_t		* tptr = 0;
	uint64		calg = 0;
	address_t	base_addr;
	uint64		paddr, pa;

printf("pte_setup, region = %d, phys_offset = %p, virt = %p, psize = %llx, vsize = %llx, pgsize = %x\n",
region, (void *)phys_offset, (void *)virt, phys_size, virt_size, pgsize);
	if (region > MMU_NUM_TABLES)
		panic("Invalid pte_setup region, %d\n", region);

	if (virt & (PTE_ENTRY_SIZE-1))
		panic("Misaligned virtual address, %p, region %d\n", virt, region);

	if (phys_offset & (PTE_ENTRY_SIZE-1))
		panic("Misaligned physical offset, %p, region %d\n",
													phys_offset, region);

	if (phys_size >= PTE_MAX_MEMORY)
		panic("Size too big for PTE, 0x%lx, region %d\n", phys_size, region);

	if (phys_size & (PTE_ENTRY_SIZE-1))
		panic("Misaligned size, 0x%lx, region %d\n", phys_size, region);

	base_addr = 0;
	switch (region) {
		case MMU_MALLOC_REGION:
			tptr = lm_table;
			calg = CFG_C_COHERENTXCLW;
			break;
		case MMU_SM_TXRX_REGION:
			tptr = sm_txrx_table;
#ifdef NFP_TXRX
			calg = CFG_C_COHERENTXCLW;
#else
			calg = CFG_C_UNCACHEDACCEL;
			base_addr = 0xE000000000;
#endif
			break;
		case MMU_SM_FP_REGION:
			tptr = sm_fp_table;
#ifdef NFP_FP
			calg = CFG_C_COHERENTXCLW;
#else
			calg = CFG_C_UNCACHEDACCEL;
			base_addr = 0xE000000000;
#endif
			break;
#ifdef NFP_FP
		case MMU_SM_TXRX_CACHED_REGION:
			tptr = sm_txrx_cached_table;
			calg = CFG_C_NONCOHERENT;
			base_addr = 0xE000000000;
			break;
#endif
#ifdef NFP_TXRX
		case MMU_SM_FP_CACHED_REGION:
			tptr = sm_fp_cached_table;
			calg = CFG_C_NONCOHERENT;
			base_addr = 0xE000000000;
			break;
#endif
		case MMU_SAC_REGION:
			tptr = sac_table;
#ifdef NFP_TXRX
			calg = CFG_C_COHERENTXCLW;
#else
			calg = CFG_C_UNCACHED;
			base_addr = 0xE000000000;
#endif
			break;
		case MMU_THREADSTACK_REGION:
			tptr = thread_stack_table;
			calg = CFG_C_COHERENTXCLW;
			break;
		default:
			panic("pte_setup, invalid region %d\n", region);
	}

	mmu_region_size[region] = virt_size;
	mmu_phys_offset_start[region] = phys_offset+base_addr;
	mmu_phys_offset_end[region] = phys_offset+base_addr+phys_size;
	mmu_virt_start[region] = virt;
	mmu_virt_end[region] = virt+virt_size;
	mmu_tptr[region] = tptr;

	for (paddr = phys_offset; phys_size > 0;
								paddr += pgsize, phys_size -= pgsize, tptr++) {
		if (paddr < 0x10000000)
			pa = paddr;
		else if (paddr < 0x30000000)
			pa = (paddr - 0x10000000) + 0x80000000;
		else if (paddr < 0x40000000)
			pa = (paddr - 0x30000000) + 0xC0000000;
		else if (paddr < 0x8000000000)
			pa = (paddr - 0x40000000) + 0x100000000;
		else
			pa = paddr;

		pa += base_addr;

		tptr->tlb_lo0_phys = ((pa & TLBLO_ADDR_MASK) >> TLBLO_ADDR_SHIFT) |
									(TLBLO_G | TLBLO_V | TLBLO_D) | (calg << 3);
		tptr->tlb_lo1_phys = (((pa+pgsize/2) & TLBLO_ADDR_MASK) >>
														TLBLO_ADDR_SHIFT) |
									(TLBLO_G | TLBLO_V | TLBLO_D) | (calg << 3);

		if (pgsize == 0x2000)
			tptr->pgsize = TLBMASK_MASK4Kb;
		else if (pgsize == 0x8000)
			tptr->pgsize = TLBMASK_MASK16kb;
		else if (pgsize == 0x20000)
			tptr->pgsize = TLBMASK_MASK64kb;
		else if (pgsize == 0x80000)
			tptr->pgsize = TLBMASK_MASK256kb;
		else if (pgsize == 0x200000)
			tptr->pgsize = TLBMASK_MASK1Mb;
		else if (pgsize == 0x800000)
			tptr->pgsize = TLBMASK_MASK4Mb;
		else if (pgsize == 0x2000000)
			tptr->pgsize = TLBMASK_MASK16Mb;
		else if (pgsize == 0x8000000)
			tptr->pgsize = TLBMASK_MASK64Mb;
		else if (pgsize == 0x4000) {
			tptr->pgsize = TLBMASK_MASK16kb;
			tptr->tlb_lo1_phys = 0;
		}
	}

}

#if 0
/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */
void
pte_refill()
{
	uint64	xc, region, index;
	uint64	tlblo0, tlblo1, pgsize;
	pte_t	* tptr;
	
	xc = _mips_dmfc0(C0_XCTEXT);

printf("pte_refill, xc = 0x%lx\n", xc);

	region = (xc >> 27) & 0xF;		/* which pte table */
	index = (xc >> 14) & 0x1FF0;	/* which 8 MB page */

printf("region = %d, index = %d\n", region, index);
	if ((region == 0) || (region > MMU_NUM_TABLES))
		panic("refill exception, xc = %p, region = %d\n", xc, region);

	if (index >= mmu_region_size[region])
		panic("refill exception, xc = %p, index = 0x%lx\n", xc, index);

	tptr = mmu_tptr[region];

	tlblo0 = tptr[index].tlb_lo0_phys;
	tlblo1 = tptr[index].tlb_lo1_phys;
	pgsize = tptr[index].pgsize;

	__asm__ __volatile__ (
	"     dmtc0 %0,$2 \n"
	"     dmtc0 %1,$3 \n"
	"     mtc0 %2,$5 \n"
	"     nop \n"
	"     nop \n"
	"     nop \n"
	"     nop \n"
	"     tlbwr \n"
	"     nop \n"
	"     nop \n"
	"     nop \n"
	"     nop \n"
	:
	: "r" (tlblo0), "r" (tlblo1), "r" (pgsize)
	);

}
#endif

#define MB	(1024 * 1024)
#define MAX_256MB		16      /* 4 GB max = 256 * 16 */
#define DIVISOR			16      /* used in sac size calculations */

/*-----------------------------------------------------------------
 * Name:	pte_init()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */
void
pte_init()
{

	uint32	sr;
	uint32	init_sac_size = 0;
	uint32	init_lm_size = 0;
	uint32	init_sm_size = 0;
	address_t	lm_base_offset = 0;
	address_t	sm_base_offset = 0;
	address_t	sac_base_offset = 0;
#ifdef USE_THREAD_STACKS
	uint32		init_threadstack_size = 0;
	address_t	threadstack_base_offset = 0;
#endif

	sr = _mips_mfc0(C0_SR);
	sr |= SR_UX;
	_mips_mtc0(C0_SR, sr);

#ifdef NFP_TXRX
	/*
	 * Use TLB for:
	 *		local memory
	 *		txrx shared memory
	 *		fp shared memory (HT)
	 *		sac
	 *
	 *	The following sizes will be determined here:
	 *		local memory size
	 *		shared memory size
	 *		sac size
	 *
	 *	Algorithm for determining SAC size:
	 *		A different percentage of each 256 MB is allocated for SAC usage.
	 *		sac_size = 0;
	 *		for (dimm256=0; dimm256 < 8; dimm256++)
	 *			sac_size += (256/DIVISOR)*sac_ratio[dimm256];
	 *
	 *	Algorithm for determining SM size:
	 *		memsize = memorysize - (_end - K0BASE)
	 *		sm_size = (memsize-sac_size)/(malloc_to_sm_ratio+1);
	 *
	 *	Algorithm for determing local memory (malloc) size:
	 *		memsize = memorysize - (_end - K0BASE)
	 *		lm_size =  (memsize-sac_size)/(malloc_to_sm_ratio+1)*
	 *														alloc_to_sm_ratio;
	 *
	 */
	extern uchar8	_end[];
	uint64	memsize;
	uint32	memorysize_256mb, X;
#ifndef SBM_REV2
	uint32	sac_ratio[MAX_256MB] = {
									4,
									8,
									9,
									10,
									12,
									13,
									14,
									15,
									15,
									15,
									15,
									15,
									15,
									15,
									15,
									15};
	uint32	malloc_to_sm_ratio[MAX_256MB] = {
												1,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2,
												2};
#else
	uint32	sm_to_malloc_ratio[MAX_256MB] = {
												2,
												2,
												3,
												3,
												4,
												4,
												5,
												5,
												6,
												6,
												7,
												7,
												8,
												9,
												10,
												10};
#endif

	extern uint64 get_memorysize();

	memorysize = get_memorysize();
	memorysize_256mb = memorysize / (256*MB);
	memorysize_MB = memorysize / MB;
	X =  (((address_t)_end - (address_t)K0BASE) + PTE_ENTRY_MASK) & ~PTE_ENTRY_MASK;
	memsize = memorysize - X;

#ifndef SBM_REV2
	init_sac_size = 48*MB;
	for (dimm256 = 1; dimm256 < memorysize_256mb; dimm256++)
		init_sac_size += ((256*MB)/DIVISOR * sac_ratio[dimm256]);
	init_sac_size += ((PTE_ENTRY_SIZE/2)-1);
	init_sac_size &= ~PTE_ENTRY_MASK;
#else
	init_sac_size = 0;
#endif

#ifndef SBM_REV2
	init_sm_size = (memsize - init_sac_size) /
									(malloc_to_sm_ratio[memorysize_256mb-1]+1);
#else
	init_sm_size = (memsize - init_sac_size) /
									(sm_to_malloc_ratio[memorysize_256mb-1]+1) *
									sm_to_malloc_ratio[memorysize_256mb-1];
#endif
	init_sm_size += ((PTE_ENTRY_SIZE/2)-1);
	init_sm_size &= ~PTE_ENTRY_MASK;

	init_lm_size = memsize - init_sac_size - init_sm_size;

	lm_base_offset = X;
	sm_base_offset = lm_base_offset + init_lm_size;
	sac_base_offset = sm_base_offset + init_sm_size;

//	tlb_init();		/* init all TLB entries to Invalid */

	pte_setup(	MMU_MALLOC_REGION,
				lm_base_offset,
				0x1000000000,
				init_lm_size,
				init_lm_size,
				PTE_ENTRY_SIZE);
	pte_setup(	MMU_SM_TXRX_REGION,
				sm_base_offset,
				0x2000000000,
				init_sm_size,
				init_sm_size,
				PTE_ENTRY_SIZE);
#ifndef SBM_REV2
	pte_setup(	MMU_SAC_REGION,
				sac_base_offset,
				0x3000000000,
				init_sac_size,
				init_sac_size,
				PTE_ENTRY_SIZE);
#endif
#endif

#ifdef NFP_FP
	/*
	 * Use TLB for:
	 *		local memory
	 *		fp shared memory
	 *		txrx shared memory (HT)
	 *		well-known address (wka)
	 *		thread stacks
	 */
	extern uchar8	_end[];
	uint32	memsize;
	uint32	memorysize_256mb, X;
	uint32	sm_to_malloc_ratio[MAX_256MB] = {
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1,
												1};

	memorysize = get_memorysize();
	memorysize_256mb = memorysize / (256*MB);
	memorysize_MB = memorysize / MB;
	X =  (((address_t)_end - (address_t)K0BASE) + PTE_ENTRY_MASK) & ~PTE_ENTRY_MASK;
#ifdef USE_THREAD_STACKS
	init_threadstack_size = ESM_NUM_STACKS * ESM_STACK_SIZE;
	threadstack_base_offset = X;
	X += init_threadstack_size;
#endif
	memsize = memorysize - X;

	init_sm_size = (memsize - init_sac_size) /
									(sm_to_malloc_ratio[memorysize_256mb-1]+1) *
									sm_to_malloc_ratio[memorysize_256mb-1];
	init_sm_size += ((PTE_ENTRY_SIZE/2)-1);
	init_sm_size &= ~PTE_ENTRY_MASK;

	init_lm_size = memsize - init_sac_size - init_sm_size;

	lm_base_offset = X;
	sm_base_offset = lm_base_offset + init_lm_size;
	sac_base_offset = sm_base_offset + init_sm_size;

//	tlb_init();		/* init all TLB entries to Invalid */

	pte_setup(	MMU_MALLOC_REGION,
				lm_base_offset,
				0x1000000000,
				init_lm_size,
				init_lm_size,
				PTE_ENTRY_SIZE);
	pte_setup(	MMU_SM_FP_REGION,
				sm_base_offset,
				0x4000000000,
				init_sm_size,
				init_sm_size,
				PTE_ENTRY_SIZE);

#ifdef USE_THREAD_STACKS
    {
    uint64  vsize = ESM_NUM_STACKS;
    vsize *= PTE_ENTRY_SIZE;
	pte_setup(	MMU_THREADSTACK_REGION,
				threadstack_base_offset,
				0x5000000000,
				ESM_NUM_STACKS * ESM_STACK_SIZE,        /* phys_size */
				vsize,
				ESM_STACK_SIZE);
    }
#endif

#endif
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description: zero the mapped regions to enable ECC
 *
 * Created by:	John Hahn
 *
 * Date Created:	9/4/02
 *
 *-----------------------------------------------------------------
 */
void
pte_zero_regions(void)
{
    extern void memz(address_t, address_t);

    watchdog_disable();
    printf("zeroing MALLOC_REGION = %d, size = %lx\n",
            MMU_MALLOC_REGION, get_memory_phys_region_size(MMU_MALLOC_REGION));
    memz(0x1000000000,
         0x1000000000 + get_memory_phys_region_size(MMU_MALLOC_REGION));
#ifdef NFP_TXRX
    printf("zeroing SM_TXRX_REGION = %d, size = %lx\n",
            MMU_SM_TXRX_REGION, get_memory_phys_region_size(MMU_SM_TXRX_REGION));
    memz(0x2000000000,
         0x2000000000 + get_memory_phys_region_size(MMU_SM_TXRX_REGION));
#if 0
    printf("zeroing SAC_REGION = %d, size = %lx\n",
            MMU_SAC_REGION, get_memory_phys_region_size(MMU_SAC_REGION));
    memz(0x3000000000,
         0x3000000000 + get_memory_phys_region_size(MMU_SAC_REGION));
#endif
#endif /* NFP_TXRX */
#ifdef NFP_FP
    printf("zeroing SM_FP_REGION = %d, size = %lx\n",
            MMU_SM_FP_REGION, get_memory_phys_region_size(MMU_SM_FP_REGION));
    memz(0x4000000000,
         0x4000000000 + get_memory_phys_region_size(MMU_SM_FP_REGION));
#ifdef USE_THREAD_STACKS
    printf("zeroing THREADSTACK_REGION = %d, size = %lx\n",
            MMU_THREADSTACK_REGION,
            get_memory_phys_region_size(MMU_THREADSTACK_REGION));
    memz(PHYS_TO_K0(get_memory_phys_base(MMU_THREADSTACK_REGION)),
         PHYS_TO_K0(get_memory_phys_base(MMU_THREADSTACK_REGION)) +
         get_memory_phys_region_size(MMU_THREADSTACK_REGION));
         //0x5000000000,
         //0x5000000000 + get_memory_phys_region_size(MMU_THREADSTACK_REGION));
#endif /* USE_THREAD_STACKS */
#endif /* NFP_FP */
    watchdog_enable(8);
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
address_t
phys_offset_to_phys_addr(
	address_t	offset
)
{
	address_t	phys;

	if (offset >= 0xE000000000)
		phys = offset;
	else if (offset >= 0x40000000)
		phys = 0x100000000 + offset - 0x40000000;
	else if (offset >= 0x30000000)
		phys = 0xC0000000 + offset - 0x30000000;
	else if (offset >= 0x10000000)
		phys = 0x80000000 + offset - 0x10000000;
	else
		phys = offset;

//printf("phys_offset_to_phys_addr, offset = %p, phys = %p\n", offset, phys);
	return (phys);
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
address_t
phys_addr_to_phys_offset(
	address_t	phys
)
{
	address_t	phys_offset;

	if (phys >= 0x100000000)
		phys_offset = phys - 0x100000000 + 0x40000000;
	else if (phys >= 0xC0000000)
		phys_offset = phys - 0xC0000000 + 0x30000000;
	else if (phys >= 0x80000000)
		phys_offset = phys - 0x80000000 + 0x10000000;
	else
		phys_offset = phys;

//printf("phys_addr_to_phys_offset, phys = %p, offset = %p\n", phys, phys_offset);
	return (phys_offset);
}

/*-----------------------------------------------------------------
 * Name:	get_memory_region_size()
 *
 * Description:	returns size of requested memory region, in bytes
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
uint64
get_memory_region_size(
	uint32	region_id
)
{
	if ((region_id) && (region_id <= MMU_NUM_TABLES))
		return (mmu_region_size[region_id]);

	panic("get_memory_region_size, region_id = %d, ra = %p\n",
														region_id, get_ra());
    return 0;
}

/*-----------------------------------------------------------------
 * Name:	get_memory_virt_base()
 *
 * Description:	returns virtual base address of requested region
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
address_t
get_memory_virt_base(
	uint32	region_id
)
{
	if ((region_id) && (region_id <= MMU_NUM_TABLES))
		return (mmu_virt_start[region_id]);

	panic("get_memory_virt_base, region_id = %d, ra = %p\n",
														region_id, get_ra());
    return 0;
}

/*-----------------------------------------------------------------
 * Name:	get_memory_virt_end()
 *
 * Description:	returns virtual end address of requested region
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/20/02
 *
 *-----------------------------------------------------------------
 */
address_t
get_memory_virt_end(
	uint32	region_id
)
{
	if ((region_id) && (region_id <= MMU_NUM_TABLES))
		return (mmu_virt_end[region_id]);

	panic("get_memory_virt_end, region_id = %d, ra = %p\n",
														region_id, get_ra());
	return 0;
}

/*-----------------------------------------------------------------
 * Name:	get_memory_phys_base()
 *
 * Description:	calculates phys offset, assuming contiguous memory space,
 *				then finds real physical base address, based on sibyte's
 *				memory map.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
address_t
get_memory_phys_base(
	uint32	region_id
)
{
	address_t	phys_offset;

	if ((region_id) && (region_id <= MMU_NUM_TABLES)) {
		phys_offset = mmu_phys_offset_start[region_id];

		return (phys_offset_to_phys_addr(phys_offset));
	}

	panic("get_memory_phys_base, region_id = %d, ra = %p\n",
														region_id, get_ra());
    return 0;
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
address_t
get_memory_phys_offset(
	uint32	region_id
)
{
	if ((region_id) && (region_id <= MMU_NUM_TABLES))
		return (mmu_phys_offset_start[region_id]);

	panic("get_memory_phys_offset, region_id = %d, ra = %p\n",
														region_id, get_ra());
    return 0;
}


/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	John Hahn
 *
 * Date Created:	9/4/02
 *
 *-----------------------------------------------------------------
 */
address_t
get_memory_phys_region_size(
	uint32	region_id
)
{
	if ((region_id) && (region_id <= MMU_NUM_TABLES))
		return (mmu_phys_offset_end[region_id] -
                mmu_phys_offset_start[region_id]);

	panic("get_memory_phys_region_size, region_id = %d, ra = %p\n",
														region_id, get_ra());
    return 0;
}
#endif	/* LARGE_MEMORY */

