/*
 *-----------------------------------------------------------------
 *
 * Name:	sbm-gmc.h
 *
 * RCS:  $Id: sbm-gmc.h,v 1.6 2001/09/10 17:07:59 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	header file for MCPU <--> GMC communication
 *
 * Created by:	Rick Lund
 *
 * Date Created:	03/29/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _SBM_GMC_H
#define _SBM_GMC_H

/*
 * GMC Memory-mapped Registers
 *
 * Some are read-only (GMC supplying internal queue addresses, etc).
 * Read-only registers begin with "gmc_".
 * Some are write-only (MCPU telling GMC about host memory addresses, etc).
 */
typedef struct gmc_regs {
	uint32	gmc_sac_size;				/* how much memory is installed */
	uint32	gmc_sac_block_size;			/* GMC compile-time configuration */

	uint32	hmt_base_addr;				/* in MCPU memory */

	uint32	gmc_ifq_base_addr;			/* internal GMC memory */
	uint32	gmc_ifq_size;				/* size of GMC queue (in words) */
	uint32	* ifq_credit_addr;			/* where should the GMC write credit */

	uint32	bdl_base_addr;				/* one BDL entry for each sac block */

	uint32	gmc_fbq_base_addr;			/* internal GMC memory */
	uint32	gmc_fbq_size;				/* size of GMC queue (in words) */
	uint32	* fbq_credit_addr;			/* where should the GMC write credit */

	uint32	dsokq_base_addr;			/* in MCPU memory */
	uint32	dsokq_size;					/* how many DSOKQ entries are there */

	uint32	gmc_txq_luc_base_addr;		/* internal GMC memory */
	uint32	gmc_txq_luc_size;			/* size of GMC queue (in words) */
	uint32	* txq_luc_credit_addr;		/* where should the GMC write credit */

	uint32	gmc_txq_snim_base_addr;		/* internal GMC memory */
	uint32	gmc_txq_snim_size;			/* size of GMC queue (in words) */
	uint32	* txq_snim_credit_addr;		/* where should the GMC write credit */

	uint32	gmc_copyq_base_addr;		/* internal GMC memory */
	uint32	gmc_copyq_size;				/* size of GMC queue (in words) */
	uint32	* copyq_credit_addr;		/* where should the GMC write credit */

	uint32	gmc_metaq_base_addr;		/* internal GMC memory */
	uint32	gmc_metaq_size;				/* size of GMC queue (in words) */
	uint32	* metaq_credit_addr;		/* where should the GMC write credit */

	uchar8	own_mac_addr[6];
	uchar8	res1[2];					/* for alignment */
	uchar8	cluster_mac_addr[6];
	uchar8	res2[2];
} gmc_reg_t;

/*-----------------------------------------------------------------*/
/*
 * The HMT is the 'Handle Mapping Table'.  Each entry in the HMT is
 * a list of buffer descriptors, making up a single COD.
 * The overall COD length, as well as the first BD index is stored in the HMT
 * entry.
 */
typedef struct hmt_entry {
	uint32	bdl_index;
#define HMT_ENTRY_IN_USE		0x80000000
#define HMT_ENTRY_FREE			0x00000000
	uint32	COD_len;
} sbm_hmtEntry_t;

/*-----------------------------------------------------------------*/
/*
 * Each BDL entry maps to a SAC block.  The data inside that block can be
 * at an offset, and can have arbitrary length.  The BDL is an array with
 * next pointers that operate as a linked list.  This way, the BDL
 * entries are the RxDesc as well as the block descriptors associated with
 * the HMT.
 *
 * There is one BDL entry for each SAC block, and it is a one-to-one
 * mapping between them (i.e. index 0 in the BDL describes SAC block #0).
 */
typedef struct bdl_entry {
	ushort16	len;
	ushort16	offset;
	uint32		next_bdl_index;
#define BDL_EOC				0xFFFFFFFF
#define BDL_ENTRY_IN_USE	0x80000000
#define BDL_ENTRY_FREE		0x00000000
} sbm_bdlEntry_t;

/*-----------------------------------------------------------------*/
/*
 * Each FBQ entry is an available index into the BDL.  It is therefore also
 * an available SAC block.
 */
typedef uint32		sbm_fbqEntry_t;
#define	FBQ_GMC_OWN		0x80000000
#define FBQ_CPU_OWN		0x00000000

typedef struct fbq {
	sbm_fbqEntry_t	*start;
	sbm_fbqEntry_t	*end;
	sbm_fbqEntry_t	*head;
	sbm_fbqEntry_t	*tail;
} sbm_fbq_t;

/*-----------------------------------------------------------------*/
/*
 * Each IFQ entry is an available HMT index.
 */
typedef uint32		sbm_ifqEntry_t;
#define	IFQ_GMC_OWN		0x80000000
#define IFQ_CPU_OWN		0x00000000

typedef struct ifq {
	sbm_ifqEntry_t	*start;
	sbm_ifqEntry_t	*end;
	sbm_ifqEntry_t	*head;
	sbm_ifqEntry_t	*tail;
} sbm_ifq_t;


/*-----------------------------------------------------------------*/
/*
 * These are the commands formed by the MCPU and posted to the GMC in
 * the internal GMC Command Queues.
 */
#define TXQ_GMC_OWN					0x80000000
#define TXQ_CPU_OWN					0x00000000
#define TXQ_ACK_REQ					0x40000000
#define TXQ_COMMAND_MASK			0x3F000000
#define TXQ_CMD_SHIFT				24
#define TXQ_CMD_SEND_READ_REQ		(0 << TXQ_CMD_SHIFT)
#define TXQ_CMD_SEND_DATA_TO_LUC	(1 << TXQ_CMD_SHIFT)
#define TXQ_CMD_SEND_DATA_TO_SNIM	(2 << TXQ_CMD_SHIFT)
#define TXQ_CMD_RETRIEVE_SAC_DATA	(3 << TXQ_CMD_SHIFT)
#define TXQ_CMD_COPY_SAC_DATA		(4 << TXQ_CMD_SHIFT)
#define TXQ_CMD_XOR_SAC_BLOCKS		(5 << TXQ_CMD_SHIFT)
#define TXQ_CACHELINE_MASK			0x00FF0000
#define TXQ_CACHELINE_SHIFT			16
#define TXQ_BUFCNT_MASK				0x0000FFFF

typedef struct txq_bd {
	ushort16	length;
	ushort16	offset;
	uint32		sac_index;
} sbm_txq_bd_t;

typedef struct txq_entry_send_read_req {
	uchar8					DA[6];
	ushort16				vlan;
	uint32					dest_port;
	uint32					source_port;
	uint32					TID;
	uint32					message_len;
	uint32					res;
	uint32					control;
} sbm_txqSendReadReq_t;

typedef struct txq_entry_send_data_to_luc {
	uchar8					DA[6];
	ushort16				vlan;
	uint32					luc_port;
	uint32					dest_port;		/* for linux sim */
	uint32					TID;
	uint32					message_len;
	uint32					res;
	uint32					control;
} sbm_txqSendDataToLUC_t;

typedef struct txq_entry_send_data_to_snim {
	uchar8					DA[6];
	ushort16				vlan;
	uint32					dest_port;
	uint32					source_port;
	uint32					TID;
	uint32					message_len;
	uint32					res;
	uint32					control;
} sbm_txqSendDataToSNIM_t;

typedef struct txq_entry_retrieve_sac_data {
	uint32					PCI_mem_addr;
	uint32					ACK_addr;
	uint32					TID;
	uint32					res[4];
	uint32					control;
} sbm_txqRetrieveSACData_t;

typedef struct txq_entry_copy_sac_data {
	uint32					src_SAC_block_index;
	uint32					src_offset;
	uint32					dest_SAC_block_index;
	uint32					dest_offset;
	uint32					num_bytes;
	uint32					ACK_addr;
	uint32					TID;
	uint32					control;
} sbm_txqCopySACData_t;

typedef struct txq_entry_xor_sac_block {
	uint32					num_src_lists;
	uint32					num_bytes;
	uint32					ACK_addr;
	uint32					TID;
	uint32					res[3];
	uint32					control;
} sbm_txqXorSACBlock_t;


/*-----------------------------------------------------------------*/
/*
 * A GMC command queue is a queue containing any of the previous command types.
 * There are 2 GMC command queues, one for data sending, one for data
 * copying and xor.
 */
typedef union txqe {
	sbm_txqSendReadReq_t		read_req;
	sbm_txqSendDataToLUC_t		send_to_luc_req;
	sbm_txqSendDataToSNIM_t		send_to_snim_req;
	sbm_txqRetrieveSACData_t	retrieve_req;
} sbm_txqQueueEntry_t;

typedef union copyqe {
	sbm_txqCopySACData_t	copyqe;
	sbm_txqXorSACBlock_t	xorqe;
} sbm_copyqQueueEntry_t;

typedef struct txq {
	sbm_txqQueueEntry_t		* start;
	sbm_txqQueueEntry_t		* end;
	sbm_txqQueueEntry_t		* head;
	sbm_txqQueueEntry_t		* tail;
	uint32					num_avail_cl;
} sbm_gmcTxq_t;

typedef struct copyq {
	sbm_copyqQueueEntry_t	* start;
	sbm_copyqQueueEntry_t	* end;
	sbm_copyqQueueEntry_t	* head;
	sbm_copyqQueueEntry_t	* tail;
	uint32					num_avail_cl;
} sbm_gmcCopyq_t;


/*-----------------------------------------------------------------*/
/*
 * When the GMC completes a new COD store, it must add an entry into the
 * appropriate DSOKQ.
 */
typedef struct dsokq_ack_entry {
	nfx_cod_t	COD;
	uint32		HMT_index;
} sbm_dsokqACKEntry_t;

typedef struct dsokq_nack_entry {
	uint32		err_param1;
	uint32		err_param2;
	uint32		err_param3;
} sbm_dsokqNACKEntry_t;

typedef struct dsokq_entry {
	union {
		sbm_dsokqACKEntry_t		ack;
		sbm_dsokqNACKEntry_t	nack;
	} status;
	uint32		control;
#define DSOKQ_CONT_OWN					0x80000000
#define DSOKQ_CONT_NACK					0x10000000
#define DSOKQ_CONT_APP_MASK				0x0FF00000
#define DSOKQ_CONT_APP_SHIFT			20
#define DSOKQ_CONT_ERROR_MASK			0x000F0000
#define DSOKQ_CONT_NUM_SAC_BLOCKS_MASK	0x0000FFFF
} sbm_dsokqEntry_t;

typedef struct dsokq {
	sbm_dsokqEntry_t	* start;
	sbm_dsokqEntry_t	* end;
	sbm_dsokqEntry_t	* curr;
} sbm_dsokqQueue_t;


/*-----------------------------------------------------------------*/
/*
 * Metadata retrieval ACK and Copy SAC block ACK message format, written
 * by the GMC into MCPU memory.
 */
typedef struct ack_msg {
	uint32		TID;
	uint32		param1;
	uint32		param2;
	uint32		control;
#define ACK_CPU_OWN					0x80000000
#define ACK_GMC_OWN					0x00000000
#define ACK_ERROR_MASK				0x7FFFFFFF
} sbm_gmcACKMsg_t;

#endif	/* _SBM_GMC_H */

