/*  FileName: devI_sm_funcs.c
 *
 *  Description: Contains action functions for the device init state machine.
 *
 *	Created:	18 May 2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_create
 *
 * Description:     Device init state machine initialization function.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 18, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int32
scsi_devI_create(esm_handle_t	handle,		/* handle for this sm inst    */
				 int32			nextState,	/* next state from table	  */
				 esm_event_t	thisEvent,	/* event driving this action  */
				 void			*userInfo)	/* ctrl blk linked to sm  	  */
{
	int32		nState;

	scsi_desc_t *sd;
	scsi_dev_t *pdev;

	scsi_debug("[scsi_devI_create] Enter\n", SCSI_DBL_3);

	pdev = (scsi_dev_t *)userInfo;
	sd = (scsi_desc_t *)scsi_allocDesc();

	if( sd == (scsi_desc_t *)NULL )
	{
		scsi_err( SCSI_ERR_SD_ALLOC | SCSI_ERR_LEVEL_0, 
				  "[scsi_devI_create]");
		return SCSI_ERR_SD_ALLOC;
	}

	pdev->flags.fInitMode = 1;
	scsi_sendInquiry(pdev, sd); 

	scsi_debug("[scsi_devI_create] Leave\n", SCSI_DBL_3);	
	return(nextState);
} /* scsi_devI_create */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_procInq
 *
 * Description:     Called when a INQUIRY response event is queued to the 
 *					state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 18, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devI_procInq(esm_handle_t	handle,		/* handle for this sm inst    */
					 int32			nextState,	/* next state from table	  */
					 esm_event_t	thisEvent,	/* event driving this action  */
					 void			*userInfo)	/* ctrl blk linked to sm  */
				
{
	return scsi_dev_procInq( nextState, 
							 (scsi_dev_t *)userInfo, 
							 (scsi_desc_t *)thisEvent.p1.u.p); 
} /* scsi_devI_procInq */


/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_procRLun
 *
 * Description:     Called when a REPORT_LUN response event is queued 
 *					to the state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
scsi_devI_procRLun(esm_handle_t	handle,		/* handle for this sm inst    */
				  int32			nextState,	/* next state from table	  */
				  esm_event_t	thisEvent,	/* event driving this action  */
				  void			*userInfo)	/* ctrl blk linked to sm  */
				
{
	scsi_dev_t *pdev;

	pdev = (scsi_dev_t *)userInfo;
	if( !strncmp(pdev->vendor, "HITACHI", 7) ) {
		return hds9910_procRLun( nextState,
								 (scsi_dev_t *)userInfo, 
								 (scsi_desc_t *)thisEvent.p1.u.p); 
	}
	else {
		return scsi_dev_procRLun( nextState,
								  (scsi_dev_t *)userInfo, 
								  (scsi_desc_t *)thisEvent.p1.u.p); 
	}
} /* scsi_devInit_procRLun */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_procRCap
 *
 * Description:     Called when a READ_CAPACITY response event is queued 
 *					to the state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
scsi_devI_procRCap(esm_handle_t     handle,		/* handle for this sm inst    */
					  int32         nextState,	/* next state from table	  */
					  esm_event_t	thisEvent,	/* event driving this action  */
					  void          *userInfo)	/* ctrl blk linked to sm  */
				
{
	return scsi_dev_procRCap( nextState,
							  (scsi_dev_t *)userInfo, 
							  (scsi_desc_t *)thisEvent.p1.u.p);
} /* scsi_devI_procRCap */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_procAck
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devI_procAck(esm_handle_t handle,	  /* handle for this sm instance  */
				  int32		  nextState,  /* next state from table	      */
				  esm_event_t thisEvent,  /* event driving this action    */
				  void		  *userInfo)  /* ctrl blk linked to sm	      */
{
	scsi_dev_t *pdev;

	scsi_debug("[scsi_devI_procAck] Enter", SCSI_DBL_3);

	pdev = (scsi_dev_t *)userInfo;

	/* Initialization steps are complete. Terminate this state machine
	 * instance and place device into the close mode.
	 */
	pdev->flags.fInitMode  = 0;
	pdev->flags.fOpenMode  = 0;
	pdev->flags.fCloseMode = 1;

	esm_destroy( pdev->smHandle );
	pdev->smHandle = 0;
	scsi_debug("[scsi_devI_procAck] Leave", SCSI_DBL_3);
	return nextState;
} /* scsi_devI_procAck */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_ignoreEvt
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devI_ignoreEvt(esm_handle_t handle,	   
					int32		nextState, /* next state from table	      */
					esm_event_t	thisEvent, /* event driving this action   */
					void		*userInfo)  /* ctrl blk linked to sm	      */
{
	scsi_debug("[scsi_devI_ignoreEvt]", SCSI_DBL_3);
	return nextState;
} /* scsi_devI_ingnoreEvt */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_bad_state
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devI_bad_state(esm_handle_t handle,	   
					int32		nextState, /* next state from table	      */
					esm_event_t	thisEvent, /* event driving this action   */
					void		*userInfo)  /* ctrl blk linked to sm	      */
{
	scsi_debug("[scsi_devI_bad_state]", SCSI_DBL_3);
	return nextState;
} /* scsi_devI_bad_state */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devI_procDel
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devI_procDel(esm_handle_t handle,	   
					int32		nextState, /* next state from table	      */
					esm_event_t	thisEvent, /* event driving this action   */
					void		*userInfo)  /* ctrl blk linked to sm	      */
{
	int i;
	int32  nState;
	uint32 cHandle, rspAddr;
	scsi_dev_t	*pdev;

	pdev 	= (scsi_dev_t *)userInfo;

	/* Blow away pending requests. */
	scsi_flushReqQ(pdev);
		
	// Set flag that will cause this state machine instance to be deleted
	// the next time the event queue for this sm is serviced.
	esm_destroy(pdev->smHandle);

	// Delete the device control block.
	scsi_deallocDevice( pdev );

	return nextState;
} /* scsi_devI_procDel */
