/*
 *------------------------------------------------------------------------------
 *
 * Name:	scsi-checksum.c
 *
 * RCS:		$Id: scsi-checksum.c,v 1.7 2002/09/05 01:48:45 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Implementation of SCSI level checksum generation and
 *              verification.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 9/18/2001
 *
 *------------------------------------------------------------------------------
 */

#include "scsi.h"
#include "../sm-eee/eee.h"		/* XXX */
#include <exports/wwn.h>

typedef struct {
	uint32		checksum;
	uint32		lba;
	scsi_id_t	wwn;
} scsi_checksum_t;
NFX_ASSERT_STATIC(sizeof(scsi_checksum_t) == ECSUM_LEN);

static inline ushort16 
scsi_computed_checksum_partial(byte *buf, int len, ushort16 csum)
{
	int			wordcnt;
	uint32		unfolded_csum;
	uint32 *	p;

	/* XXX -- for now assume no odd-byte alignment */ 
	VERIFY(!((address_t)buf & 1)); 
	/* XXX -- for now assume no odd-byte counts */
	VERIFY(!(len & 1)); 

	/* Bring to word alignment */
	if ((address_t)buf & 2) {
		csum ^= *(ushort16 *)buf;
		buf += 2;
		len -= 2;
	}

	/* Advance by 32-bit words */
	wordcnt = len/4;
	len -= wordcnt * 4;
	p = (uint32 *)buf;
	unfolded_csum = 0;
	while (wordcnt--)
		unfolded_csum ^= *p++;
	buf = (byte *)p;

	/* Fold checksum */
	csum ^= unfolded_csum >> 16;
	csum ^= unfolded_csum & 0xFFFF;

	if (len) 
		csum ^= *(ushort16 *)buf;

	return(csum);
}

void
scsi_insert_embedded_checksum(scsi_dev_t *pdev, scsi_desc_t *sd, uint32 csum)
{
	uint32				bytecnt;
	uint32				bufcnt;
	uint32				blockcnt;
	eee_desc_t *		edesc;
	eee_descBuf_t *		bd;
	eee_descBuf_t *		cs_bd;

	int					i;
	byte *				data_p;
	scsi_checksum_t *	cs_p;
	int					lba;
	int					bytes_in_block;
	int					bytes_in_buf;
	int					bytes_in_chunk;

	edesc = sd->eDesc;
	while (edesc) {
		bytecnt = edesc->hdr.control & EDESC_TOTAL_LEN;
		bufcnt = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
		VERIFY(pdev->blockSize == BLOCKSIZE_520);
		VERIFY((bytecnt % BLOCKSIZE_512) == 0);
		VERIFY((bytecnt % ECSUM_BLOCKSIZE) == 0);

		blockcnt = bytecnt / ECSUM_BLOCKSIZE;

		bd = &edesc->bd[0];
		if (!(bd->control & EDESC_BD_CMD_IN_USE))
			++bd;
		/* Allocate space for checksum(s) */
		cs_bd = bd + bufcnt - 1;
		/* Space in last buffer? */
		if (cs_bd->len + blockcnt * ECSUM_LEN > EEE_BUFSIZE_ULTRA) {
			++cs_bd;
			VERIFY(cs_bd - &edesc->bd[0] < EEE_MAX_BUFS);
			cs_bd->control = EDESC_BD_CMD_IN_USE | EDESC_BD_ULTRA; 
			cs_bd->buf = eee_allocateBuffer(EEE_SHARED_POOL, cs_bd->control & EDESC_BD_BUFSIZE_MASK, 0);
			VERIFY(cs_bd->buf);
			cs_p = (scsi_checksum_t *)cs_bd->buf;
			cs_bd->len = blockcnt * ECSUM_LEN;
			++bufcnt;
			bytecnt += cs_bd->len;
		}
		else {
			cs_p = (scsi_checksum_t *)(cs_bd->buf + cs_bd->len);
			cs_bd->len += blockcnt * ECSUM_LEN;
			bytecnt += blockcnt * ECSUM_LEN;
		}

		/* Compute checksum(s) */
		if (pdev->flags.fAllowEmbedded_CS) {
			data_p = bd->buf;
			lba = sd->lba;
			bytes_in_buf = bd->len;
			for (i = 0; i < blockcnt; ++i, ++cs_p, lba += ECSUM_BLOCK_SPAN) {
				cs_p->lba = lba;
				cs_p->checksum = (lba >> 16) ^ (lba & 0xFFFF);
				WWN_SET(&cs_p->wwn, &pdev->id);
				bytes_in_block = ECSUM_BLOCKSIZE;
				while (bytes_in_block) {
					bytes_in_chunk = MIN(bytes_in_block, bytes_in_buf);
					cs_p->checksum = scsi_computed_checksum_partial(data_p, 
																	bytes_in_chunk, 
																	(ushort16)cs_p->checksum);
					bytes_in_block -= bytes_in_chunk;
					bytes_in_buf -= bytes_in_chunk;
					data_p += bytes_in_chunk;
					if (bytes_in_buf == 0) {
						++bd;
						data_p = bd->buf;
						bytes_in_buf = bd->len;
					}
				}
			}
		}
		
		/* Adjust e-desc to include checksum */
		edesc->hdr.control &= ~(EDESC_BUFCNT_MASK | EDESC_TOTAL_LEN);
		edesc->hdr.control |= ((bufcnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
		edesc->hdr.control |= (bytecnt & EDESC_TOTAL_LEN);

		if (edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
			edesc = (eee_desc_t *)edesc->bd[EEE_MAX_BUFS-1].buf;
			VERIFY(edesc);
		}
		else 
			edesc = NULL;

	} /* while (edesc) */
}

int 
scsi_verify_embedded_checksum(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32				bytecnt;
	uint32				bufcnt;
	uint32				blockcnt;
	eee_desc_t *		edesc;
	eee_descBuf_t *		bd;
	eee_descBuf_t *		cs_bd;

	int					i;
	byte *				data_p;
	scsi_checksum_t *	cs_p;
	int					lba;
	int					bytes_in_block;
	int					bytes_in_buf;
	int					bytes_in_chunk;

	int					mismatch = 0;

	edesc = sd->eDesc;
	while (edesc) {
		bytecnt = edesc->hdr.control & EDESC_TOTAL_LEN;
		bufcnt = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
		VERIFY(pdev->blockSize == BLOCKSIZE_520);
		VERIFY((bytecnt % BLOCKSIZE_520) == 0);
		VERIFY((bytecnt % (ECSUM_BLOCKSIZE + ECSUM_LEN)) == 0);

		blockcnt = bytecnt / (ECSUM_BLOCKSIZE + ECSUM_LEN);

		bd = &edesc->bd[0];
		if (!(bd->control & EDESC_BD_CMD_IN_USE))
			++bd;
		cs_bd = bd + bufcnt - 1;
		VERIFY(cs_bd->control & EDESC_BD_CMD_IN_USE);
		VERIFY(cs_bd->buf);
	
		/* Verify checksum(s) */
		if (pdev->flags.fAllowEmbedded_CS) {
			data_p = bd->buf;
			cs_p = (scsi_checksum_t *)(cs_bd->buf + cs_bd->len - blockcnt * ECSUM_LEN);
			lba = sd->lba;
			bytes_in_buf = bd->len;
			for (i = 0; i < blockcnt; ++i, ++cs_p, lba += ECSUM_BLOCK_SPAN) {
				bytes_in_block = ECSUM_BLOCKSIZE;
				while (bytes_in_block) {
					bytes_in_chunk = MIN(bytes_in_block, bytes_in_buf);
					cs_p->checksum = scsi_computed_checksum_partial(data_p, 
																	bytes_in_chunk, 
																	(ushort16)cs_p->checksum);
					bytes_in_block -= bytes_in_chunk;
					bytes_in_buf -= bytes_in_chunk;
					data_p += bytes_in_chunk;
					if (bytes_in_buf == 0) {
						++bd;
						data_p = bd->buf;
						bytes_in_buf = bd->len;
					}
				}
				if (cs_p->lba != lba ||
				    WWN_EQUAL(&cs_p->wwn, &pdev->id) != 1 ||
				    cs_p->checksum != ((lba >> 16) ^ (lba & 0xFFFF))) {
					char s1[32], s2[32];
					++mismatch;
					printf("Checksum mismatch on device %016llx, lba %d: "
					       "expected %d/%s/%x; "
					       "got %d/%s/%x\n",
					       pdev->id, lba,
					       lba, wwn_to_str((wwn_t *)&pdev->id, s1), ((lba >> 16) ^ (lba & 0xFFFF)),
					       cs_p->lba, wwn_to_str((wwn_t *)&cs_p->wwn, s2), cs_p->checksum);
				}
			}
		}
		
		/* Adjust e-desc to exclude checksum */
		bytecnt -= blockcnt * ECSUM_LEN;
		cs_bd->len -= blockcnt * ECSUM_LEN;
		if (cs_bd->len == 0) {
			--bufcnt;
			eee_deallocateBuffer(cs_bd->buf);
			cs_bd->buf = NULL;
			cs_bd->len = 0;
			cs_bd->control &= ~EDESC_BD_CMD_IN_USE;
		}

		edesc->hdr.control &= ~(EDESC_BUFCNT_MASK | EDESC_TOTAL_LEN);
		edesc->hdr.control |= ((bufcnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
		edesc->hdr.control |= (bytecnt & EDESC_TOTAL_LEN);

		if (edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
			edesc = (eee_desc_t *)edesc->bd[EEE_MAX_BUFS-1].buf;
			VERIFY(edesc);
		}
		else 
			edesc = NULL;

	} /* while (edesc) */
	return(mismatch);
}

void
scsi_insert_block_checksum(uint32 lba, uint32 lbCnt, uint32 csum, eee_desc_t *edesc)
{
	eee_descBuf_t *		bd;
	uint32				bytecnt;
	uint32				bufcnt;
	scsi_cs_t *			cs_p;

	/*
	 * Verify that if checksum is valid, that the I/O is aligned and
	 * has the right size.
	 */
	VERIFY(csum == 0 || 
		   (lbCnt == (BCSUM_BLOCKSIZE / BLOCKSIZE_512) &&
			(lba % (BCSUM_BLOCKSIZE / BLOCKSIZE_512)) == 0));

	/*
	 * XXX VERIFY that no single I/O spans a checksum block. This
	 * requirement may need to change in the future.  
	 */
	VERIFY(lba / (BCSUM_BLOCKSIZE / BLOCKSIZE_512) == 
		   (lba + lbCnt - 1) / (BCSUM_BLOCKSIZE / BLOCKSIZE_512));

	if ((lba + lbCnt) % (BCSUM_BLOCKSIZE / BLOCKSIZE_512) == 0) {
		while (edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
			edesc = (eee_desc_t *)edesc->bd[EEE_MAX_BUFS-1].buf;

		bytecnt = edesc->hdr.control & EDESC_TOTAL_LEN;
		bufcnt = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;

		bd = &edesc->bd[0];
		if (!(bd->control & EDESC_BD_CMD_IN_USE))
			++bd;
		bd += bufcnt;
		VERIFY(EEE_BUFSIZE_SMALL == BLOCKSIZE_512);
		bd->control = EDESC_BD_CMD_IN_USE | EDESC_BD_SMALL;
		bd->buf = eee_allocateBuffer(EEE_SHARED_POOL, bd->control & EDESC_BD_BUFSIZE_MASK, 0);
		VERIFY(bd->buf);
		bd->len = BLOCKSIZE_512;

		/* Adjust e-desc to include checksum */
		++bufcnt;
		bytecnt += BLOCKSIZE_512;
		edesc->hdr.control &= ~(EDESC_BUFCNT_MASK | EDESC_TOTAL_LEN);
		edesc->hdr.control |= ((bufcnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
		edesc->hdr.control |= (bytecnt & EDESC_TOTAL_LEN);

		cs_p = (scsi_cs_t *)bd->buf;
		if (csum) {
			cs_p->magic = SCSI_CS_MAGIC_REV_1;
			cs_p->checkSum = csum;
			cs_p->blockNo = lba;
		}
		else {
			cs_p->magic = SCSI_CS_MAGIC_INVALID;
			cs_p->checkSum = -1;
			cs_p->blockNo = -1;
		}		

		CACHE_FLUSH(cs_p, cs_p + sizeof(scsi_cs_t));

#ifdef RADEK_DEBUG	
		printf("scsi_insert_block_checksum: magic 0x%x, checkSum 0x%x, lba 0x%x\n", 
			   cs_p->magic, cs_p->checkSum, cs_p->blockNo);
#endif
	}
}

void
scsi_extract_block_checksum(uint32 lba, uint32 lbCnt, eee_desc_t *edesc, uint32 *csum, uint32 *blockno)
{
	eee_desc_t *		prev_edesc;
	eee_descBuf_t *		bd;
	uint32				bytecnt;
	uint32				bufcnt;
	scsi_cs_t *			cs_p;

	*csum = 0;
	*blockno = -1;

	/*
	 * XXX VERIFY that no single I/O spans a checksum block. This
	 * requirement may need to change in the future.  
	 */
	VERIFY(lba / (BCSUM_BLOCKSIZE / BLOCKSIZE_512) == 
	       (lba + lbCnt - 1) / (BCSUM_BLOCKSIZE / BLOCKSIZE_512));

	if ((lba + lbCnt) % (BCSUM_BLOCKSIZE / BLOCKSIZE_512) == 0) {
		prev_edesc = NULL;
		while (edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
			prev_edesc = edesc;
			edesc = (eee_desc_t *)edesc->bd[EEE_MAX_BUFS-1].buf;
		}

		bytecnt = edesc->hdr.control & EDESC_TOTAL_LEN;
		bufcnt = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;

		VERIFY(bytecnt >= BLOCKSIZE_512);

		/* 
		 * Is checksum block contained completely in last edesc ? 
		 */
		if (bytecnt == BLOCKSIZE_512) {
			bd = &edesc->bd[0];
			if (!(bd->control & EDESC_BD_CMD_IN_USE))
				++bd;
			bd += bufcnt - 1;
			VERIFY(bd->len == BLOCKSIZE_512);
			cs_p = (scsi_cs_t *)bd->buf;
#ifdef RADEK_DEBUG	
			printf("scsi_extract_block_checksum: magic 0x%x, checkSum 0x%x, lba 0x%x\n", 
			       cs_p->magic, cs_p->checkSum, cs_p->blockNo);
#endif
			if (cs_p->magic == SCSI_CS_MAGIC_REV_1) {
				*csum = cs_p->checkSum;
				*blockno = cs_p->blockNo;
			}
			VERIFY(prev_edesc);
			eee_freePacket(edesc);
			edesc = prev_edesc;
			edesc->hdr.control &= ~EDESC_COMMAND_CHAIN_LAST_BUFFER;
			edesc->bd[EEE_MAX_BUFS-1].buf = NULL;
		}
		else {
			bd = &edesc->bd[0];
			if (!(bd->control & EDESC_BD_CMD_IN_USE))
				++bd;
			bd += bufcnt - 1;
			if (bd->len == BLOCKSIZE_512) {
				cs_p = (scsi_cs_t *)bd->buf;
#ifdef RADEK_DEBUG	
				printf("scsi_extract_block_checksum: magic 0x%x, checkSum 0x%x, lba 0x%x\n", 
				       cs_p->magic, cs_p->checkSum, cs_p->blockNo);
#endif
				if (cs_p->magic == SCSI_CS_MAGIC_REV_1) {
					*csum = cs_p->checkSum;
					*blockno = cs_p->blockNo;
				}
				eee_deallocateBuffer(bd->buf);
				bd->control = 0;
				bd->buf = NULL;
				bd->len = 0;
				--bufcnt;
				bytecnt -= BLOCKSIZE_512;
			}
			else {
				cs_p = (scsi_cs_t *)(bd->buf + bd->len - BLOCKSIZE_512);
#ifdef RADEK_DEBUG	
				printf("scsi_extract_block_checksum: magic 0x%x, checkSum 0x%x, lba 0x%x\n", 
				       cs_p->magic, cs_p->checkSum, cs_p->blockNo);
#endif
				if (cs_p->magic == SCSI_CS_MAGIC_REV_1) {
					*csum = cs_p->checkSum;
					*blockno = cs_p->blockNo;
				}
				bd->len -= BLOCKSIZE_512;
				bytecnt -= BLOCKSIZE_512;
			}

			/* Adjust e-desc to exclude checksum */
			edesc->hdr.control &= ~(EDESC_BUFCNT_MASK | EDESC_TOTAL_LEN);
			edesc->hdr.control |= ((bufcnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
			edesc->hdr.control |= (bytecnt & EDESC_TOTAL_LEN);
		}
	}
}

uint32
scsi_compute_checksum(eee_desc_t *edesc)
{
	eee_descBuf_t * bd;
	uint32		bufcnt;
	int32		i;
	uint32		chkSum = 0;

	bufcnt = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;  
	bd = &edesc->bd[0];
	if (!(bd->control & EDESC_BD_CMD_IN_USE))
		++bd;
	for (i = 0; i < bufcnt; i++, ++bd)
	        chkSum = uint32_checksum(bd->buf, bd->len, chkSum);
	return chkSum;
}

uint32
scsi_add_checksums(uint32 cs1, uint32 cs2)
{
	return(uint32_checksum_add(cs1, cs2));
}

int
scsi_map_logical_to_physical(uint32 l_lba, uint32 l_lbCnt, uint32 *p_lba, uint32 *p_lbCnt)
{
	uint32		start_pba;
	uint32		end_pba;

	start_pba = l_lba + l_lba/BCSUM_BLOCK_SPAN;
	end_pba = l_lba + l_lbCnt + (l_lba + l_lbCnt)/BCSUM_BLOCK_SPAN;
	*p_lba = start_pba;
	*p_lbCnt = end_pba - start_pba;

	return(NFX_OK);
}

int
scsi_map_physical_to_logical(uint32 p_lba, uint32 p_lbCnt, uint32 *l_lba, uint32 *l_lbCnt)
{
	uint32		cs_blocks_start;
	uint32		cs_blocks_end;

	cs_blocks_start = p_lba / (BCSUM_BLOCK_SPAN + 1);
	cs_blocks_end = (p_lba + p_lbCnt) / (BCSUM_BLOCK_SPAN + 1);

	*l_lbCnt = p_lbCnt - (cs_blocks_end - cs_blocks_start);
	*l_lba = p_lba - cs_blocks_start;

	return(NFX_OK);
}

void 
scsi_add_padding(eee_desc_t *edesc, uint32 pad_block_cnt)
{
	eee_descBuf_t *		bd;
	uint32				bytecnt;
	uint32				bufcnt;

	while (edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
		edesc = (eee_desc_t *)edesc->bd[EEE_MAX_BUFS-1].buf;

	bytecnt = edesc->hdr.control & EDESC_TOTAL_LEN;
	bufcnt = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;

	bd = &edesc->bd[0];
	if (!(bd->control & EDESC_BD_CMD_IN_USE))
		++bd;
	bd += bufcnt;
	bd->control = EDESC_BD_CMD_IN_USE | EDESC_BD_CMD_NO_DEALLOC | EDESC_BD_ULTRA;
	bd->buf = scsiCtrl->writePadBuffer;
	VERIFY(bd->buf);
	bd->len = pad_block_cnt * BLOCKSIZE_512;

	/* Adjust e-desc to include pad */
	++bufcnt;
	bytecnt += pad_block_cnt * BLOCKSIZE_512;
	edesc->hdr.control &= ~(EDESC_BUFCNT_MASK | EDESC_TOTAL_LEN);
	edesc->hdr.control |= ((bufcnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
	edesc->hdr.control |= (bytecnt & EDESC_TOTAL_LEN);
}
