/*  
 *-----------------------------------------------------------------------------
 * FileName   : scsi-dev-init.c
 *
 * Description: 
 *
 * Created    : February 14, 2002
 *
 * Author     : Frank Salzmann
 *
 * RCS        : $Id: scsi-dev-init.c,v 1.6 2002/11/19 20:38:29 nitinb Exp $
 *
 * Copyright (c)  Agile Storage, Inc. 2002
 *-----------------------------------------------------------------------------
 */
#include "scsi.h"
#include "../sm-sdm/sdm-api.h"

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_newDevInit
 *
 * Description: 	This function is called	when a new storage device 
 *					is discovered. This is where the device control structure
 *					and state machine instance are created.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 12, 2002
 *
 * Returns:		    NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int
scsi_newDevInit( uint32   tType,		// SCSI_LL_FC | SCSI_LL_IP
				ushort16 llHandle,	// Lower layer device handle
				uint64 	 devWWN,
				uint64   chnWWN,
				char *   chnName,	// Egress switch, egress port wwn.
				uint32   lun)
{
	scsi_id_t    id;
	scsi_desc_t *sd;
	scsi_dev_t  *pdev;
	
	SCSI_FUNC_TRACE("[scsi_newDevInit] Enter");

	//printf("tType: %x, llHdl: %x, devWWN: %llx, chnWWN: %llx, chName: %s, lun: %d\n", 
	//				  tType, llHandle, devWWN, chnWWN, chnName, lun);

	//-- Set the DEV_ID equal to the WWN for the duration of the 
	//-- initialization process.
	id = devWWN;

	//-- See if the device already exists.
	if(scsi_devCheckExist(devWWN, lun, llHandle)) 
		return(NFX_OK);

	//-- Allocate a new device control block
	pdev = (scsi_dev_t *)scsi_allocDevice(lun, llHandle);
	if( !pdev )
		return(NFX_ERR);

	//-- Save WWN
	pdev->id       = id;
	pdev->wwn      = devWWN;
	pdev->lun      = lun;
	pdev->chnWWN   = chnWWN;
	
	//-- Send SCSI Inquiry command to the device.
	sd = (scsi_desc_t *)scsi_allocDesc();
    if( sd == (scsi_desc_t *)NULL )
    {
        scsi_err( SCSI_ERR_SD_ALLOC | SCSI_ERR_LEVEL_0, "[scsi_newDevInit]");
        return SCSI_ERR_SD_ALLOC;
    }
    pdev->flags.fInitMode = 1;
    scsi_sendInquiry(pdev, sd);

	SCSI_FUNC_TRACE("[scsi_newDevInit] Leave");
	return( NFX_OK );
} // scsi_newDevInit

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devInitModeProc
 *
 * Description: 	This function is called by scsi_llReceive when 
 *                  the initial SCSI INQ response is received from 
 *                  a newly discovered device. Here we will package
 *                  up the response and send a newDev event message
 *                  up to the SDM application on the SSC.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 12, 2002
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_devInitModeProc(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uchar8          *buf;
	sdm_newDevMsg_t *o_msg;
	char			sdm_name[10];

	scsi_debug("[scsi_devInitModeProc] Enter", SCSI_DBL_3);

    // If device returns an inquiry error then assume
    // no LUN is  present so abort the whole process
    // and remove the device.
	if( scsi_parseInqRsp( pdev, sd ) != SCSI_OK ) {
		scsi_deallocDesc(sd);
		scsi_deallocDevice(pdev);
		return(SCSI_ERR_GENERIC); 
	}

	//-- Create state machine instance so we can relay cdb requests from 
	//-- the SDM.
	scsi_openDevice(pdev->path);

	o_msg = (sdm_newDevMsg_t *)eee_ramAlloc(sizeof(sdm_newDevMsg_t));
	bzero(o_msg, sizeof(sdm_newDevMsg_t));

	o_msg->hdr.mid  = SDM_MID_NEW_DEV;
	o_msg->id       = pdev->id;
	o_msg->wwn      = pdev->wwn;
	o_msg->chn_wwn  = pdev->chnWWN;
	o_msg->path     = pdev->path;
	o_msg->llHandle = pdev->llHandle;

	buf = sd->eDesc->bd[1].buf;  
	memcpy(o_msg->inqRsp, buf, 255);
	memcpy(pdev->inqRsp, buf, 255);			// Keep copy of inq data.

	sprintf(sdm_name, "sdmfc%d", nfxMyCPUId); 
	scsi_sendToSscApp((uchar8 *)o_msg, 
					  sizeof(sdm_newDevMsg_t),  
					  sdm_name);
	scsi_deallocDesc(sd);

	scsi_debug("[scsi_devInitModeProc] Leave", SCSI_DBL_3);
	return NFX_OK;
} /* scsi_devInitModeProc */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendNewDevMsg
 *
 * Description: 	
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 08, 2002
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_sendNewDevMsg(scsi_dev_t *pdev)
{
	sdm_newDevMsg_t *o_msg;
	char			sdm_name[10];

	scsi_debug("[scsi_sendNewDevMsg] Enter", SCSI_DBL_3);

	o_msg = (sdm_newDevMsg_t *)eee_ramAlloc(sizeof(sdm_newDevMsg_t));
	bzero(o_msg, sizeof(sdm_newDevMsg_t));

	o_msg->hdr.mid  = SDM_MID_NEW_DEV;
	o_msg->id       = pdev->id;
	o_msg->wwn      = pdev->wwn;
	o_msg->chn_wwn  = pdev->chnWWN;
	o_msg->path     = pdev->path;
	o_msg->llHandle = pdev->llHandle;

	memcpy(o_msg->inqRsp, pdev->inqRsp, 255);

	sprintf(sdm_name, "sdmfc%d", nfxMyCPUId); 
	scsi_sendToSscApp((uchar8 *)o_msg, 
					  sizeof(sdm_newDevMsg_t),  
					  sdm_name);

	scsi_debug("[scsi_sendNewDevMsg] Leave", SCSI_DBL_3);
	return NFX_OK;
} /* scsi_sendNewDevMsg */
