/*  
 *------------------------------------------------------------------------------
 * File Name  :	scsi-dev.h
 *
 *  Description:	Contains defines, macros and structures for managing 
 *					SCSI devices. 
 *
 *  Created    :    20 February 2001
 *
 *  Author	   :    Frank Salzmann
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *------------------------------------------------------------------------------
 */


#ifndef SCSI_DEV_H_INCLUDED

#define SCSI_DEV_H_INCLUDED

#include <exports/scsi-dev-cmds.h>
#include <exports/scsi-api.h>
#include <exports/esm-api.h>

#define SCSI_MAX_DEV_PER_PORT	512       
#define SCSI_MAX_LUN_PER_DEV	255
#define SCSI_MAX_LUN_PROBE_CNT	32
#define SCSI_MAX_CMD_PER_DEV	16		  /* Maximum device commands that */
                                          /* can be outstanding at once.  */
#define SCSI_MAX_MP_PER_DEV		16		  /* Maximum MP state machine simultaneously */
										  /* active per device.                      */

#define SCSI_MAX_LBA_6			0x001FFFFF
#define SCSI_MAX_LBCNT_6		0xFF
#define SCSI_MAX_LBA_10			0xFFFFFFFF
#define SCSI_MAX_LBCNT_10		0xFFFF

/* Queue parameters */
#define SCSI_DISPATCH_CNT_DISK_DEFAULT	2
#define SCSI_GENIO_HIWATER_MARK			32
#if 0
#define SCSI_GENIO_TIMER_DURATION		10		/* 10 mSec */
#else
#define SCSI_GENIO_TIMER_DURATION		0		/* disabled */
#endif
#define SCSI_MAX_CHAIN_LEN				32

struct _scsi_hugeio_state_t;

//----------------------------------
// SCSI Device Configuration Entry
//----------------------------------
typedef struct _scsi_dev_entry
{
	scsi_id_t 	id;						// Logical Device Id				
	uint64		wwn;					// WWN as reported by Lower layer	
	uint64		chnWWN;					// Egress Switch WWN as reported by FC.
	scsi_path_t path;					// Logical Device Path				
	uint32      lun;
	ushort16	llHandle;				// Handle LL uses to address device 
	uchar8 		type;					// DISK, CDROM, TAPE 				
	uchar8 		name[SCSI_DEV_NAME_LEN];// ASCII Device Name				
	uint32      raidLevel;
	uint32 		blocks;					// Number of logical blocks -- DISK ONLY
	uint32 		blockSize;				// Size in bytes of a logical block -- DISK ONLY
	int32		blockno;				// Current block number position -- TAPE ONLY
	int64		byteno;					// Current byte position -- TAPE ONLY
	int32		fileno;					// Current file number position -- TAPE ONLY
	uint32		bom:1,
				eom:1,
				unused:30;				// Position flags -- TAPE ONLY
	uchar8 		ansiVersion;			// SCSI-1, SCSI-2, SCSI-3, CCS		
	char   		vendor[9];				// Manufacturers id string			
	char   		model[17];				// Model number string				
	char   		productId[17];				// Model number string				
	char        sn[32];                 // Serial number string
	scsi_cmd_t  *cmdTable;				// Points to table of cmds 			
	                                	// supported by this device.		
	/*
	 * Flags and Operational Info
	 */
	struct {
		uint32 fTaggedAgile	:1;		/* We have formated and tagged this dev	*/
		uint32 fDeviceReady	:1; 	/* Device is ready to be used.			*/
		uint32 fDeviceBusy 	:1; 	/* Device is busy with another command  */
		uint32 fReadOnly	:1; 	/* Is read-only or write-protected		*/
		uint32 fRmb			:1;		/* Have removable media?				*/
		uint32 fMediaPresent:1;		/* Is removable media present			*/
		uint32 fRW6			:1;		/* Use 6 byte CDB for read/write		*/
		uint32 fOpenMode	:1;
		uint32 fInitMode	:1;
		uint32 fCloseMode	:1;
	    uint32 fAllowChain  :1;		/* Allow I/O coallescing/compaction     */
		uint32 fAllowSort   :1;		/* Allow I/O reordering/sorting         */
	    uint32 fAllowEmbedded_CS:1;	/* Allow Embedded checksum gen'/verf'   */
	    uint32 fAllowBlock_CS:1;	/* Allow Block checksum gen'/verf'      */
		uint32 fLocalCSVerify:1;	/* Perform local packet/IO checksum verification */
		uint32 fSimEarlyEOM :1;	    /* Simulate early EOM condition -- TAPE */
		uint32 unused    	:16; 
	} flags;

	/* 
	 * Mapping Info
	 */
	uint32		 offset;			/* Physical offset to first logical block */
	uint32		 partitionCount;	/* Number of partitions                 */
	scsi_partInfo_t partitions[SCSI_MAPPING_MAX_PARTITIONS];

	uint32		 nextEvent;			/* Next event that statemachine expects	*/
	esm_handle_t smHandle;			/* State Machine handle					*/
	uint32 		 curBlock;			/* Current Block address				*/
	uint32       maxDispatchCnt;	/* max num of cmds device will accept 	*/
	uint32       dispatchCnt;	    /* number of cmds sent to device		*/
	scsi_desc_t  *dispatchQ[SCSI_MAX_CMD_PER_DEV];
	scsi_mp_desc_t *activeMP[SCSI_MAX_MP_PER_DEV];
	int32 		 closeTmr;
	uint64		 tape_iocookie;

	/* Request queues */
	void *		 genio_q;
	int32		 genio_q_hiwater_mark;
	int32		 genio_q_immediate_schedule_count;
	void *		 hiprio_q;
	void *		 other_q;
	void *		 tape_io_q;
#define MAX_NOREORDER_BARRIER_COUNT 1
	uint32		 noreorder_barrier;
	int32		 write_gate_timer_handle;
	int32		 write_gate_timer_running;

	/* Huge I/O info */
	struct _scsi_hugeio_state_t	*hugeio_state;

	/* Some Device Statistics */
	uint32 readReqCnt;
	uint64 readByteCnt;
	uint64 readByteLast;
	uint32 readBytePerSecCur;
	uint32 readBytePerSecMax;
	uint32 readRspCnt;
	uint32 readRspLast;
	uint32 readIoPerSecCur;
	uint32 readIoPerSecMax;
	uint32 readCSECnt;
	uint32 readBECnt;
	uint32 writeSub8kCnt;
	uint32 writeSub8kPaddedCnt;
	uint32 writeReqCnt;
	uint64 writeByteCnt;
	uint64 writeByteLast;
	uint32 writeBytePerSecCur;
	uint32 writeBytePerSecMax;
	uint32 writeRspCnt;
	uint32 writeRspLast;
	uint32 writeIoPerSecCur;
	uint32 writeIoPerSecMax;
	uint32 writeCSECnt;	
	uint32 relayReqCnt;
	uint32 relayRspCnt;
	uint32 devErrCnt;
	uint32 nacksReceived;
	uint32 nacksTransmitted;
	uint32 r_chain_stats[SCSI_COALLESCE_STATS_BUCKETS+1];
	uint32 w_chain_stats[SCSI_COALLESCE_STATS_BUCKETS+1];

	char		inqRsp[256];		// Place to cach the inquiry data.
} scsi_dev_t;

typedef int (*scsi_hugeio_callback_t)(scsi_desc_t *, scsi_dev_t *, int);

// structure for storing state information about ongoing huge read/write operation
typedef struct _scsi_hugeio_state_t {
	scsi_dev_t			*pdev;			// device in question
	uint32				op;				// operation in progress and op. flags
	uint32				lba;			// logical block offset
	uint32				byte_count;		// total bytes of read/write requested
	uint32				rep_count;		// number of times to do operation
	scsi_hugeio_callback_t	callback;	// read/write completion function
	void*				cookie;			// private data
} scsi_hugeio_state_t;


#endif  /* SCSI_DEV_H_INCLUDED */
