/*  FileName: scsi-disk-cmds.c
 *
 *  Description: Contains routines for issuing commands to SCSI disk devices
 *				 and processing responses.
 *
 *	Created:	12 February	2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"

scsi_cmd_t scsi_diskCmds[8];

/*
 * Disk Command Function Prototypes
 */
int scsi_cmnRelayCDB(scsi_dev_t *, scsi_desc_t *);
int scsi_diskReadData(scsi_dev_t *, scsi_desc_t *);
int scsi_diskWriteData(scsi_dev_t *, scsi_desc_t *);
int scsi_diskReadDataErr(scsi_dev_t *, scsi_desc_t *);
int scsi_diskWriteDataErr(scsi_dev_t *, scsi_desc_t *);

/*
 *-------------------------------------------------------------
 * Function Name:	scsi_diskInitCmds
 *
 * Description: Reads a block of data 
 *
 * Returns: NFX_OK | NFX_ERR 
 *
 *-------------------------------------------------------------
 */
int
scsi_diskInitCmds() 
{
	memset( scsi_diskCmds, 0, sizeof(scsi_diskCmds));

	scsi_diskCmds[SD_OPCODE_RELAY].req_func = scsi_cmnRelayCDB;
	scsi_diskCmds[SD_OPCODE_RELAY].err_func = NULL;
	scsi_diskCmds[SD_OPCODE_RELAY].name = "Relay CDB";

	scsi_diskCmds[SD_OPCODE_READ].req_func  = scsi_diskReadData;
	scsi_diskCmds[SD_OPCODE_READ].err_func  = scsi_diskReadDataErr;
	scsi_diskCmds[SD_OPCODE_READ].name  = "Read Disk Data";
	
	scsi_diskCmds[SD_OPCODE_WRITE].req_func = scsi_diskWriteData;
	scsi_diskCmds[SD_OPCODE_WRITE].err_func = scsi_diskWriteDataErr;
	scsi_diskCmds[SD_OPCODE_WRITE].name = "Write Disk Data";

	return NFX_OK;
} /* scsi_diskInitCmds */

/*
 *-------------------------------------------------------------
 * Function Name:	scsi_diskReadData
 *
 * Description: Reads a block of disk data.
 *
 * Returns: NFX_OK | SCSI_MSG_ECODE_OO_RANGE 
 *-------------------------------------------------------------
 */
int
scsi_diskReadData( scsi_dev_t *pDev, scsi_desc_t *sd ) 
{
	uint32 lba;				
	uint32 sCnt;

	lba  = sd->lba;
	sCnt = sd->chain ? sd->lbCntChained : sd->lbCnt;

	/* Don't overwrite if chain */
	if (!sd->chain) {
			sd->retryCnt      = 3;
			sd->timeOutCnt    = 5;
			sd->maxTimeToLive = 120;
			VERIFY(pDev->blockSize);
			sd->expectDataLen = (sCnt * pDev->blockSize);
			sd->status |= SD_DEV_DATA_IN_EXPECT; 
	}
	
	if( !pDev->flags.fRW6) {
		if((sCnt > SCSI_MAX_LBCNT_10)    || 
		   (lba > SCSI_MAX_LBA_10)       || 
		   (sCnt > (SCSI_MAX_LBA_10 - lba))) {
			return SCSI_MSG_ECODE_OO_RANGE;
		}
		sd->cdb[0] = SCSI_CMD_READ_10;
		sd->cdb[1] = 0;
		sd->cdb[2] = (lba >> 24) & 0xff;	/* LBA MSB      */
		sd->cdb[3] = (lba >> 16) & 0xff;	
		sd->cdb[4] = (lba >> 8)  & 0xff;	
		sd->cdb[5] = lba & 0xff;			/* LBA LSB      */ 
		sd->cdb[7] = (sCnt >> 8) & 0xFF;	/* LB Count MSB */
		sd->cdb[8] = sCnt & 0xFF;			/* LB Count	LSB */
	} else {
		if((sCnt > 256)           || 
		   (lba > SCSI_MAX_LBA_6) || 
		   (sCnt > (SCSI_MAX_LBA_6 - lba))) {
			return SCSI_MSG_ECODE_OO_RANGE;
		}
		sd->cdb[0] = SCSI_CMD_READ_6;
		sd->cdb[1] = ((lba >> 16) & 0x1F);	// Block Addr MSB
		sd->cdb[2] = ((lba >> 8) & 0xFF);
		sd->cdb[3] = (lba & 0xFF);			// Block Addr LSB	    
		sd->cdb[4] = sCnt & 0xFF;			// LB Count	
	}

	return NFX_OK;
} /* scsi_readBlkData */


/*
 *-------------------------------------------------------------
 * Function Name:	scsi_diskWriteData
 *
 * Description: Creates CDB with write block command.
 *
 * Returns: NFX_OK | SCSI_MSG_ECODE_OO_RANGE 
 *-------------------------------------------------------------
 */
int
scsi_diskWriteData( scsi_dev_t *pDev, scsi_desc_t *sd ) 
{
	uint32 lba;				
	uint32 sCnt;

	lba  = sd->lba;
	sCnt = sd->chain ? sd->lbCntChained : sd->lbCnt;

	/* Don't overwrite if chain */
	if (!sd->chain) {
			sd->timeOutCnt = 5;
			sd->maxTimeToLive = 120;
			sd->retryCnt = 3;
			sd->status |= SD_DEV_DATA_OUT_EXPECT; 
	}

	/*
	 * Is the Read Only Bit set?
	 */
	if( pDev->flags.fReadOnly )
		return SCSI_MSG_ECODE_READ_ONLY;

	/*
	 * Look for media if removable.
	 */
	if(pDev->flags.fRmb && !(pDev->flags.fMediaPresent))
		return SCSI_MSG_ECODE_NO_MEDIA;

	if( !pDev->flags.fRW6) {
		if((sCnt > SCSI_MAX_LBCNT_10)    || 
		   (lba > SCSI_MAX_LBA_10)       || 
		   (sCnt > (SCSI_MAX_LBA_10 - lba))) {
			return SCSI_MSG_ECODE_OO_RANGE;
		}
		sd->cdb[0] = SCSI_CMD_WRITE_10;
		sd->cdb[1] = 0;
		sd->cdb[2] = (lba >> 24) & 0xff;	/* LBA MSB      */
		sd->cdb[3] = (lba >> 16) & 0xff;	
		sd->cdb[4] = (lba >> 8)  & 0xff;	
		sd->cdb[5] = lba & 0xff;			/* LBA LSB      */ 
		sd->cdb[7] = (sCnt >> 8) & 0xFF;	/* LB Count MSB */
		sd->cdb[8] = sCnt & 0xFF;			/* LB Count	LSB */
	} else {
		if((sCnt > 256)           || 
		   (lba > SCSI_MAX_LBA_6) || 
		   (sCnt > (SCSI_MAX_LBA_6 - lba))) {
			return SCSI_MSG_ECODE_OO_RANGE;
		}
		sd->cdb[0] = SCSI_CMD_WRITE_6;
		sd->cdb[1] = ((lba >> 16) & 0x1F);	// Block Addr MSB	    
		sd->cdb[2] = ((lba >> 8) & 0xFF);
		sd->cdb[3] = (lba & 0xFF);			// Block Addr LSB	    
		sd->cdb[4] = sCnt & 0xFF;			// LB Count	
	}
	return NFX_OK;
} /* scsi_diskWriteData */

int 
scsi_diskReadDataErr(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32			dev_status = (sd->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
	uint32			drv_status = (sd->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT;
	uint32			ecode = SCSI_MSG_ECODE_OK;
	scsi_sense_t *	sense;

	if (drv_status) {
		if (sd->eDesc)
			sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
	    return(SCSI_MSG_ECODE_PATH_FAILED);
	}
	else
	if (dev_status != SCSI_DEV_STATUS_GOOD) {
	    switch (dev_status) {
	    case SCSI_DEV_STATUS_CHECK:
			sense = (scsi_sense_t *)sd->senseData;
			switch (sense->senseKey & SCSI_SENSE_KEY_MASK) {
			case SCSI_SENSE_KEY_NO_SENSE:
			case SCSI_SENSE_KEY_RECOVERED_ERROR:
				goto read_ok;

			case SCSI_SENSE_KEY_MEDIUM_ERROR:
				if (sd->eDesc)
					sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_MEDIUM_ERROR);
				
			case SCSI_SENSE_KEY_NOT_READY:
			case SCSI_SENSE_KEY_HARDWARE_ERROR:
			case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
			case SCSI_SENSE_KEY_UNIT_ATTENTION:
			case SCSI_SENSE_KEY_DATA_PROTECT:
			case SCSI_SENSE_KEY_BLANK_CHECK:
			case SCSI_SENSE_KEY_VENDOR_SPECIFIC:
			case SCSI_SENSE_KEY_COPY_ABORTED:
			case SCSI_SENSE_KEY_ABORTED_COMMAND:
			case SCSI_SENSE_KEY_OBSOLETE:
			case SCSI_SENSE_KEY_VOLUME_OVERFLOW:
			case SCSI_SENSE_KEY_MISCOMPARE:
			case SCSI_SENSE_KEY_RESERVED:
			default:
				if (sd->eDesc)
					sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_DEV_IO_ERROR);
			}
			
	    default:
			if (sd->eDesc)
				sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
			return(SCSI_MSG_ECODE_DEV_IO_ERROR);
	    }
	}

 read_ok:
	if (pdev->flags.fAllowEmbedded_CS && !(sd->control & SD_FLAG_PHY_OP)) {
	    if (scsi_verify_embedded_checksum(pdev, sd)) {
			printf("NOTICE: Checksum error on read from device %016llx; lba %d\n",
				   pdev->id, sd->lba);
#if 0				/* XXX For now, ignore checksum error */
			ecode = SCSI_MSG_ECODE_READ_CHECKSUM_ERR;
#endif
	    }
	}

	return(ecode);
}

int 
scsi_diskWriteDataErr(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32			dev_status = (sd->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
	uint32			drv_status = (sd->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT;
	uint32			ecode = SCSI_MSG_ECODE_OK;
	scsi_sense_t *	sense;

	if (drv_status) {
	    sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
	    return(SCSI_MSG_ECODE_PATH_FAILED);
	}
	else
	if (dev_status != SCSI_DEV_STATUS_GOOD) {
	    switch (dev_status) {
	    case SCSI_DEV_STATUS_CHECK:
			sense = (scsi_sense_t *)sd->senseData;
			switch (sense->senseKey & SCSI_SENSE_KEY_MASK) {
			case SCSI_SENSE_KEY_NO_SENSE:
			case SCSI_SENSE_KEY_RECOVERED_ERROR:
				goto write_ok;

			case SCSI_SENSE_KEY_MEDIUM_ERROR:
				sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_MEDIUM_ERROR);
				
			case SCSI_SENSE_KEY_NOT_READY:
			case SCSI_SENSE_KEY_HARDWARE_ERROR:
			case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
			case SCSI_SENSE_KEY_UNIT_ATTENTION:
			case SCSI_SENSE_KEY_DATA_PROTECT:
			case SCSI_SENSE_KEY_BLANK_CHECK:
			case SCSI_SENSE_KEY_VENDOR_SPECIFIC:
			case SCSI_SENSE_KEY_COPY_ABORTED:
			case SCSI_SENSE_KEY_ABORTED_COMMAND:
			case SCSI_SENSE_KEY_OBSOLETE:
			case SCSI_SENSE_KEY_VOLUME_OVERFLOW:
			case SCSI_SENSE_KEY_MISCOMPARE:
			case SCSI_SENSE_KEY_RESERVED:
			default:
				sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_DEV_IO_ERROR);
			}
			
	    default:
			sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
			return(SCSI_MSG_ECODE_DEV_IO_ERROR);
	    }
	}

 write_ok:

	return(ecode);
}
