/*  
 *------------------------------------------------------------------------------
 *  FileName   : scsi-hugeio.c
 *
 *  Description: Large Fibre Channel IO functions
 *
 *	Created    : 15 Oct 2002
 *
 *  Author     : Nitin Bahadur
 *
 *  RCS        : $Id: scsi-hugeio.c,v 1.2 2002/11/23 00:30:31 nitinb Exp $
 *
 *  Copyright (c)  Agile Storage, Inc. 2002
 *------------------------------------------------------------------------------
 */			 

#include "scsi.h"
#include <exports/malloc-api.h>

#define SD_FLAG_STOP_IO		0x00008000

eee_desc_t*	scsi_allocIOmemory(uint32);
void		scsi_set_edesc(eee_desc_t *, char c);
static int	scsi_compare_edesc(eee_desc_t *, eee_desc_t *);
static void	scsi_dump_data(char *, uint32);

eee_desc_t*
scsi_allocIOmemory(uint32 byte_count)
{
	static uint32	packet_len = 512 * 16;
	eee_descBuf_t * bd;
	int		alloced_count = 0;
	int		buf_count;
	int		buf_byte_count;
	eee_desc_t *	ed;
	eee_desc_t *	first_ed = NULL;
	int32	total_byte_count;

	while (alloced_count < byte_count) {
		if (first_ed == NULL) {
			first_ed = ed = eee_allocateDesc(EEE_SHARED_POOL, 0);
		}
		else {
			VERIFY(!(ed->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE));
			ed->bd[EEE_MAX_BUFS-1].buf = (uchar8 *)eee_allocateDesc(EEE_SHARED_POOL, 0);
			ed->hdr.control |= EDESC_COMMAND_CHAIN_LAST_BUFFER;
			ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
		}
		VERIFY(ed);
		bzero(ed, sizeof(eee_desc_t));

		buf_byte_count = 0;
		buf_count = 0;
		bd = &ed->bd[1];
		while (buf_count < (EEE_MAX_BUFS - 2) && alloced_count < byte_count) {
			bd->control = EDESC_BD_CMD_IN_USE;
			bd->len = packet_len;
			bd->len = MIN(bd->len, EEE_BUFSIZE_ULTRA);
			bd->len = MIN(bd->len, byte_count);
			bd->control |= EDESC_BD_ULTRA;
			bd->buf = eee_allocateBuffer(EEE_SHARED_POOL, bd->control & EDESC_BD_BUFSIZE_MASK, 0);
			VERIFY(bd->buf);
			alloced_count += bd->len;
			buf_byte_count += bd->len;
			++bd;
			++buf_count;
		}
		ed->hdr.control = ((buf_count << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
		ed->hdr.control |= (buf_byte_count & EDESC_TOTAL_LEN);

		total_byte_count = ed->hdr.control & EDESC_TOTAL_LEN;
		ed->hdr.control = (ed->hdr.control & ~EDESC_TOTAL_LEN) | (total_byte_count & EDESC_TOTAL_LEN);
	}

	return first_ed;
} // scsi_allocIOmemory


int
scsi_start_huge_io(scsi_dev_t *pdev, uint32 lba, uint32 size, uint32 rep_count, int op,
					scsi_hugeio_callback_t callback, void *cookie)
{
	scsi_desc_t*		sd;
 	scsi_hugeio_state_t *st;
    esm_event_t			esmEvent;

	st = (scsi_hugeio_state_t*)eee_ramAlloc(sizeof(scsi_hugeio_state_t));
	if(st == NULL) {
		printf("Failed to allocate memory for scsi_hugeio_state\n");
		return 1;
	}

	st->pdev = pdev;
	st->op = op| SD_FLAG_HUGE_IO;
	st->lba = lba;
	st->byte_count = size;
	st->rep_count = rep_count;
	st->callback = callback;
	st->cookie = cookie;

	VERIFY(pdev->hugeio_state == NULL);
	pdev->hugeio_state = st;

	/*
	 * Allocate SCSI descriptor
	 */
	sd = (scsi_desc_t *)scsi_allocDesc();
	if (sd == (scsi_desc_t *)NULL) {
		printf("scsi_procReadReq: Failed to allocate SCSI desc.\n");
		return 1;
	}
	memset((void *)sd, 0, sizeof(scsi_desc_t));

	// set SCSI descriptor
	sd->pdev    = pdev;
	sd->path    = pdev->path;

	sd->control = op | SD_FLAG_PHY_OP | SD_FLAG_HUGE_IO | SD_FLAG_PHY_OP;
	sd->lba = lba;
	sd->lbCnt = size / BLOCKSIZE_512;

	if(op == SD_OPCODE_READ) {
		pdev->readReqCnt++;
		pdev->readByteCnt += (pdev->blockSize * sd->lbCnt);
	}
	else if(op == SD_OPCODE_WRITE) {
		pdev->writeReqCnt++;
		pdev->writeByteCnt += size;
	}

	sd->eDesc = scsi_allocIOmemory(size);

	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = pdev->id;
		if(op == SD_OPCODE_READ)  
			scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_READ_REQ;
		else if(op == SD_OPCODE_WRITE)
			scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_WRITE_REQ;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = 0;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;
	}

	/* 
	 * Send event to device state machine. 	
	 */
    esmEvent.event = SCSI_DEVO_EVENT_REQ;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pdev->smHandle, &esmEvent);

	return NFX_OK;
} // scsi_start_huge_io

int
scsi_stop_huge_io(scsi_dev_t *pdev)
{
	if(!pdev->hugeio_state) {
		printf("No Huge-IO operation in progress\n");
	} else {
		scsi_hugeio_state_t *st = pdev->hugeio_state;
		st->op |= SD_FLAG_STOP_IO;
		printf("Huge-IO operation marked for stop\n");
	}

	return 0;
} // scsi_stop_huge_io

int
scsi_huge_io_done(scsi_desc_t *sd, scsi_dev_t *pdev, int eCode)
{
	scsi_hugeio_state_t *st = pdev->hugeio_state;
	esm_event_t			esmEvent;

	VERIFY(st);

	// check if operation has been marked to be cancelled
	if(eCode) {
		printf("Operation on 0x%llx:%d failed: SCSI error = %d\n", pdev->wwn, pdev->lun, eCode);
		st->rep_count = 0;
	}
	else if(st->op & SD_FLAG_STOP_IO) 
		st->rep_count = 0;
	else
		st->rep_count--;

	// if no more repetitions left, then simply free descriptors
	if(st->rep_count == 0) {
		pdev = st->pdev;
		pdev->hugeio_state = NULL;	
		scsi_deallocDesc(sd);
		eee_ramDealloc(st);
		//scsi_deallocIOmemory();
		printf("IO operation for 0x%llx:%d completed\n", pdev->wwn, pdev->lun);
		return NFX_OK;
	}
	
	sd->lba += sd->lbCnt;
	
	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = pdev->id;
		if((sd->control & SD_OPCODE_MASK) == SD_OPCODE_READ)  
			scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_READ_REQ;
		else if((sd->control & SD_OPCODE_MASK) == SD_OPCODE_WRITE)
			scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_WRITE_REQ;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = 0;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;
	}

	/* 
	 * Send event to device state machine. 	
	 */
    esmEvent.event = SCSI_DEVO_EVENT_REQ;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pdev->smHandle, &esmEvent);

	return NFX_OK;
} // scsi_huge_io_done


static eee_desc_t	*write_desc = NULL;

int
scsi_huge_io_done2(scsi_desc_t *sd, scsi_dev_t *pdev, int eCode)
{
	esm_event_t	esmEvent;
	scsi_hugeio_state_t *st = pdev->hugeio_state;

	// check if write/read completed successfully
	if(eCode) {
		printf("Operation on 0x%llx:%d failed: SCSI error = %d\n", pdev->wwn, pdev->lun, eCode);
		scsi_deallocDesc(sd);
		eee_ramDealloc(st);
		if(write_desc != sd->eDesc)
			eee_freePacket(write_desc);
		write_desc = NULL;
		pdev->hugeio_state = NULL;
		return 1;
	}

	switch(st->op & SD_OPCODE_MASK) {
		case SD_OPCODE_WRITE:

			// now fire off the read operation
			st->op = SD_OPCODE_READ | SD_FLAG_HUGE_IO;

			memset((void *)sd, 0, sizeof(scsi_desc_t));
			sd->pdev    = pdev;
			sd->path    = pdev->path;
			sd->control = SD_OPCODE_READ | SD_FLAG_HUGE_IO | SD_FLAG_PHY_OP;
			sd->lba 	= st->lba;
			sd->lbCnt 	= st->byte_count / BLOCKSIZE_512;
			sd->eDesc 	= scsi_allocIOmemory(st->byte_count);

			/* 
	 		 * Send event to device state machine. 	
	 		 */
    		esmEvent.event = SCSI_DEVO_EVENT_REQ;
    		esmEvent.p1.u.p = sd;
    		esmEvent.p2.u.p = NULL;
    		esm_deliverEvent(pdev->smHandle, &esmEvent);

			break;

		case SD_OPCODE_READ:
			// check if read and write match
			if(scsi_compare_edesc(write_desc, sd->eDesc)) {
				printf("Read data does not match written data !\n");		
			} else {
				printf("Write verified successfuly !\n");
			}
			eee_freePacket(write_desc);
			scsi_deallocDesc(sd);
			eee_ramDealloc(st);
			write_desc = NULL;
			pdev->hugeio_state = NULL;	
			break;
	}

	return NFX_OK;

}

int 
scsi_do_write_verify(scsi_dev_t *pdev, uint32 lba, uint32 size)
{
	scsi_desc_t*		sd;
 	scsi_hugeio_state_t *st;
    esm_event_t			esmEvent;

	if(write_desc) {
		printf("Verify already in progress, can't run multiple at a time\n");
		return 1;
	}
	
	st = (scsi_hugeio_state_t*)eee_ramAlloc(sizeof(scsi_hugeio_state_t));
	if(st == NULL) {
		printf("Failed to allocate memory for scsi_hugeio_state\n");
		return 1;
	}

	st->pdev = pdev;
	st->op = SD_OPCODE_WRITE | SD_FLAG_HUGE_IO | SD_FLAG_PHY_OP;
	st->lba = lba;
	st->byte_count = size;
	st->rep_count = 1;
	st->callback = scsi_huge_io_done2;

	VERIFY(pdev->hugeio_state == NULL);
	pdev->hugeio_state = st;

	/*
	 * Allocate SCSI descriptor
	 */
	sd = (scsi_desc_t *)scsi_allocDesc();
	if (sd == (scsi_desc_t *)NULL) {
		printf("scsi_procReadReq: Failed to allocate SCSI desc.\n");
		return 1;
	}
	memset((void *)sd, 0, sizeof(scsi_desc_t));

	// set SCSI descriptor
	sd->pdev    = pdev;
	sd->path    = pdev->path;

	sd->control = SD_OPCODE_WRITE | SD_FLAG_PHY_OP | SD_FLAG_HUGE_IO;
	sd->lba = lba;
	sd->lbCnt = size / BLOCKSIZE_512;

	sd->eDesc = write_desc = scsi_allocIOmemory(size);
	scsi_set_edesc(sd->eDesc, 3);

	/* 
	 * Send event to device state machine. 	
	 */
    esmEvent.event = SCSI_DEVO_EVENT_REQ;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pdev->smHandle, &esmEvent);


	return NFX_OK;
} // scsi_do_write_verify

void
scsi_set_edesc(eee_desc_t *eDesc, char c)
{
	eee_desc_t *ed = eDesc;
	eee_descBuf_t * bd;
	int buf_count = 0;
	int i;
	
	while(ed) {
		buf_count = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
		for(i=0; i < buf_count; i++) {
			bd = &ed->bd[i+1];
			memset((void *)bd->buf, (int)c, bd->len);
		}

		CACHE_FLUSH_PACKET(ed);

		if(ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
			ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
		else
			ed = NULL;
	}

} // scsi_set_edesc


int
scsi_compare_edesc(eee_desc_t *e1, eee_desc_t *e2)
{
	int i = 0;
	int buf_count1, buf_count2;
	eee_descBuf_t *bd1, *bd2;
	eee_desc_t *ed1, *ed2;

	ed1 = e1;
	ed2 = e2;
	while(ed1 && ed2) {

		buf_count1 = (ed1->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
		buf_count2 = (ed2->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;

		VERIFY(buf_count1 == buf_count2);

		for(i=0; i < buf_count1; i++) {
			bd1 = &ed1->bd[i+1];
			bd2 = &ed2->bd[i+1];

			VERIFY(bd1->len == bd2->len);

			if(memcmp(bd1->buf, bd2->buf, bd1->len) != 0) {
				//scsi_dump_data(bd1->buf, bd1->len);
				//scsi_dump_data(bd2->buf, bd1->len);
				return 1;
			}
		}
		if(ed1->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
			ed1 = (eee_desc_t *)ed1->bd[EEE_MAX_BUFS-1].buf;
		else
			ed1 = NULL;
		if(ed2->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
			ed2 = (eee_desc_t *)ed2->bd[EEE_MAX_BUFS-1].buf;
		else
			ed2 = NULL;
	}

	if(!ed1 && !ed2)
		return 0;

	return 1;
} // scsi_compare_edesc

void
scsi_dump_data(char *buf, uint32 len)
{
	int i;
	
	if(!len) return;
	
	for(i=0; i < len; i++) {
		if( i%16==0 ) printf("\n");
		printf("%2x ", buf[i]);
	}
}

int
scsi_huge_io_done_ssc(scsi_desc_t *sd, scsi_dev_t *pdev, int eCode)
{
	scsi_hugeio_state_t *st = pdev->hugeio_state;
	esm_event_t			esmEvent;

	VERIFY(st);

	// check if operation has been marked to be cancelled
	if(eCode) {
		st->rep_count = 0;
	}
	else if(st->op & SD_FLAG_STOP_IO) 
		st->rep_count = 0;
	else
		st->rep_count--;

	// if no more repetitions left, then simply free descriptors
	if(st->rep_count == 0) {
		eee_desc_t				*edesc;
		scsi_hugeioRspMsg_t 	*rsp;
		pdev = st->pdev;
		pdev->hugeio_state = NULL;	
		scsi_deallocDesc(sd);

		VERIFY(st->cookie);
		edesc = (eee_desc_t *)st->cookie;

		rsp = (scsi_hugeioRspMsg_t *)eee_ramAlloc(sizeof(scsi_hugeioRspMsg_t));
		memset((void *)rsp, 0, sizeof(scsi_hugeioRspMsg_t));
		rsp->hdr.mid = SCSI_MID_HUGEIO_RSP;
		rsp->wwn = pdev->wwn;
		rsp->lun = pdev->lun;
		if(eCode) rsp->eCode = eCode;

		// send back response message
		scsi_sendMessage((uchar8*)rsp, sizeof(scsi_hugeioRspMsg_t), edesc->hdr.src_port, edesc->hdr.dest_port);

		eee_ramDealloc(st);
		eee_freePacket(edesc);
		return NFX_OK;
	}
	
	sd->lba += sd->lbCnt;
	
	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = pdev->id;
		if((sd->control & SD_OPCODE_MASK) == SD_OPCODE_READ)  
			scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_READ_REQ;
		else if((sd->control & SD_OPCODE_MASK) == SD_OPCODE_WRITE)
			scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_WRITE_REQ;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = 0;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;
	}

	/* 
	 * Send event to device state machine. 	
	 */
    esmEvent.event = SCSI_DEVO_EVENT_REQ;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pdev->smHandle, &esmEvent);

	return NFX_OK;
} // scsi_huge_io_done_ssc
