/*  File Name  :	scsi-msg.h
 *
 *  Description:    Contains macros, definitions, types and prototypes 
 *					describing the SCSI module messaging interface.
 *
 *  Created    :    10 July 2001
 *
 *  Author	   :    Frank Salzmann
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#ifndef SCSI_MSG_H_INCLUDED
#define SCSI_MSG_H_INCLUDED

#define SCSI_MAX_MSG_SIZE		2048
#define SCSI_DEV_NAME_LEN	    128  // ***** sync with SDM_DEV_NAME_LEN ****

/*
 *=============================
 * SCSI Message Types 
 *=============================
 */
typedef enum {
	SCSI_MID_READ_REQ = SCSI_MSG_BASE,  //-- 0x0400
	SCSI_MID_READ_RSP,					//-- 0x0401
	SCSI_MID_WRITE_REQ,					//-- 0x0402
	SCSI_MID_WRITE_RSP,					//-- 0x0403
	SCSI_MID_RELAY_REQ,					//-- 0x0404
	SCSI_MID_RELAY_RSP,					//-- 0x0405
	SCSI_MID_OPEN_REQ,					//-- 0x0406
	SCSI_MID_OPEN_RSP,					//-- 0x0407
	SCSI_MID_CLOSE_REQ,					//-- 0x0408
	SCSI_MID_CLOSE_RSP,					//-- 0x0409
	SCSI_MID_ERR_RSP,					//-- 0x040A
	SCSI_MID_SDM_ACK,					//-- 0x040B
	SCSI_MID_SYNC_REQ,					//-- 0x040C
	SCSI_MID_DEV_RESET_REQ,				//-- 0x040D
	SCSI_MID_DEV_RESET_RSP,				//-- 0x040E
	SCSI_MID_DEV_STATUS_REQ,			//-- 0x040F
	SCSI_MID_DEV_STATUS_RSP,			//-- 0x0410
	SCSI_MID_IOCTL_REQ,					//-- 0x0411
	SCSI_MID_IOCTL_RSP,					//-- 0x0412
	SCSI_MID_GET_STATS_REQ,				//-- 0x0413
	SCSI_MID_GET_STATS_RSP,				//-- 0x0414
	SCSI_MID_DELETE_REQ,				//-- 0x0415
	SCSI_MID_ADD_REQ,					//-- 0x0416
	SCSI_MID_ADD_RSP,					//-- 0x0417
	SCSI_MID_DEV_CONTROL_REQ,			//-- 0x0418
	SCSI_MID_DEV_CONTROL_RSP,			//-- 0x0419
	SCSI_MID_DEL_DEV_REQ,				//-- 0x041A
	SCSI_MID_CLR_STATS_REQ,				//-- 0x041B
	SCSI_MID_ACK,						//-- 0x041C
	SCSI_MID_NULL_RSP,					//-- 0x041D
	SCSI_MID_HUGEIO_REQ,				//-- 0x041E
	SCSI_MID_HUGEIO_RSP					//-- 0x041F	
} scsi_msgType_t;

/*
 *---------------------
 * SCSI Message Header
 *---------------------
 */
typedef struct _scsi_msg_hdr
{
	scsi_msgType_t	mid;		/* Message Id					*/
	uint32		    cHandle;	/* Client transaction handle	*/
	scsi_path_t	    path;		/* Logical Device Path			*/
} scsi_msgHdr_t;

/* 
 *---------------------------------
 * Response Status Word Definitions
 *---------------------------------
 */
#define SCSI_MSG_ECODE_OK						0
#define SCSI_MSG_ECODE_NOSUCH_DEV				1
#define SCSI_MSG_ECODE_DEV_CLOSED				2	
#define SCSI_MSG_ECODE_NO_RESOURCE				3
#define SCSI_MSG_ECODE_INVALID_MSG				4
#define SCSI_MSG_ECODE_ILLEGAL_REQUEST			5
#define SCSI_MSG_ECODE_NO_MEDIA					6
#define SCSI_MSG_ECODE_READ_ONLY				7
#define SCSI_MSG_ECODE_OO_RANGE					8
#define SCSI_MSG_ECODE_NOT_SUPPORTED			9
#define SCSI_MSG_ECODE_PATH_FAILED				10
#define SCSI_MSG_ECODE_DEV_IO_ERROR				11	
#define SCSI_MSG_ECODE_READ_CHECKSUM_ERR		12		/* For disk read only */
#define SCSI_MSG_ECODE_MEDIUM_ERROR				13
#define SCSI_MSG_ECODE_ILI						14		/* For tape read only */
#define SCSI_MSG_ECODE_FM_ENCOUNTERED			15		/* For tape read only */
#define SCSI_MSG_ECODE_END_OF_MEDIUM			16		/* For tape read/write only */
#define SCSI_MSG_ECODE_INVALID_IOCTL_CODE		17
#define SCSI_MSG_ECODE_INVALID_IOCTL_PARMS		18
#define SCSI_MSG_ECODE_IOCTL_FAILED				19
#define SCSI_MSG_ECODE_DATA_CHECKSUM_ERR		20
#define SCSI_MSG_ECODE_TIMEOUT					21
#define SCSI_MSG_ECODE_INVALID_LEN				22
#define SCSI_MSG_ECODE_INVALID_TXN_SEQ			23
#define SCSI_MSG_ECODE_WRITE_GATE_CLOSED		24

extern char scsi_msg_ecode_string[][40];

extern char* scsi_msg_ecode_to_string(int);

/*
 *---------------------------------
 * READ/WRITE request attributes
 *---------------------------------
 */
#define SCSI_MSG_ATTR_PRI_MASK				0x000000000000000FLL
#define SCSI_MSG_ATTR_GENERAL_PRI			0x0000000000000000LL /* "Normal" priority					*/
#define SCSI_MSG_ATTR_HI_PRI				0x0000000000000001LL /* "High" priority						*/
#define SCSI_MSG_ATTR_NO_REORDER			0x0000000000000010LL /* Supress sorting/reordering this I/O */
#define SCSI_MSG_ATTR_LOOPBACK				0x0000000000000020LL
#define SCSI_MSG_ATTR_PAD_WRITE				0x0000000000000040LL /* Pad write up till checksum block	*/
#define SCSI_MSG_ATTR_EXPRESS				0x0000000000000080LL /* Supress retries and waiting			*/
#define SCSI_MSG_ATTR_KICK_WRITE_GATE		0x0000000000000100LL /* Arm/rearm write gate				*/
#define SCSI_MSG_ATTR_IGNORE_WRITE_GATE		0x0000000000000200LL /* Ignore closed write gate			*/
#define SCSI_MSG_ATTR_PHY_OP				0x0000000000000400LL /* do physical i/o						*/
#define SCSI_MSG_ATTR_NO_GUAR_DELIVERY		0x0000000000000800LL /* No guaranteed delivery/retry        */

#define SCSI_DEFAULT_WRITE_GATE_PERIOD (NFX_WRITE_GATE_PERIOD*1000) /* Write gate closes after 10 seconds  */
																 /* if not kicked open.					*/

/*
 *------------------------
 * SCSI Error Rsp Msg
 *------------------------
 */
typedef struct _scsi_errRsp_msg
{
	scsi_msgHdr_t hdr;			// SCSI_MID_ERR_RSP
	uint32 eCode;
} scsi_errRspMsg_t;

/*
 *-------------------------
 * SDM Acknowledge Message
 *-------------------------
 */
typedef struct _scsi_sdmAck_msg {
	scsi_msgHdr_t hdr;			// SCSI_MID_SDM_ACK
} scsi_sdmAckMsg_t;

/*
 *-------------------------
 * Device Read Messages
 *-------------------------
 */
typedef struct _scsi_readReq_msg
{
	scsi_msgHdr_t hdr;			/* SCSI_MID_READ_REQ                        */
	uint32		  transSeqId;	/* transaction:sequence id					*/
	uint32		  blkOffset;	/* Block offset to start reading from 		*/
	uint32		  blkCnt;		/* Number of data blocks to read			*/
	uint32		  attributes;	/* See SCSI_MSG_ATTR_*                      */
	uint32		  dataDest;		/* Where to send data once it has been 		*/
	uint64		  iocookie;		/* For tape reads only						*/
	uint32        pad[6];       /* cache line alignment						*/       
} scsi_readReqMsg_t;

typedef struct _scsi_readRsp_msg
{
	scsi_msgHdr_t hdr;		 	/* SCSI_MID_READ_RSP						*/
	uint32		  transSeqId;	/* transaction:sequence id					*/
	uint32        eCode;
	uint32        blkOffset;
	uint32        blkCnt;
	uint32        dataLen;
	uint32		  queueDepth;
	uint32		  checkSum;
	uint64        codHandle;	/* Must be aligned on a 64 byte boundry     */
	uint64		  dataOffset;	/* offset of data in a sequence				*/
} scsi_readRspMsg_t;

/*
 *--------------------------
 * Device Write Messages
 *--------------------------
 */
typedef struct _scsi_writeReq_msg
{
	scsi_msgHdr_t hdr;			/* SCSI_MID_WRITE_REQ						*/
	uint32		  transSeqId;	/* transaction:sequence id					*/
	uint32		  blkOffset;	/* Block offset to start writing to			*/
	uint32		  blkCnt;		/* Number of data blocks to write			*/
	uint32		  attributes;	/* See SCSI_MSG_ATTR_*						*/
	uint32        checkSum;
	uint64		  iocookie;		/* For tape writs only 						*/
	uint64        pad[3];		/* cache line alignment						*/    
	/* Write Data To Follow */
} scsi_writeReqMsg_t;

typedef struct _scsi_writeRsp_msg
{
	scsi_msgHdr_t hdr;			/* SCSI_MID_WRITE_RSP						*/
	uint32		  transSeqId;	/* transaction:sequence id					*/
	uint32		  eCode;
	uint32        blkOffset;
	uint32        blkCnt;
	uint32		  queueDepth;
} scsi_writeRspMsg_t;

/*
 *-------------------------------
 * Multi-packet ACK/NACK messages
 *-------------------------------
 */
typedef struct _scsi_ack_msg
{
	scsi_msgHdr_t hdr;			/* SCSI_MID_ACK								*/
	uint32		  nack;			/* == 0, ACK: == 1, NACK					*/
	uint64        pad2[6];		/* cache line alignment						*/    
} scsi_ackMsg_t;

/* Definition of Transaction:Sequence ID */
/* Note, a value of zero implies a single-packet sequence. */
#define SCSI_MSG_TRANSSEQID_TRANSACTION_ID_MASK		 0xFFFF0000
#define SCSI_MSG_TRANSSEQID_SEQUENCE_MAX_MASK		 0x0000FF00
#define SCSI_MSG_TRANSSEQID_SEQUENCE_ID_MASK		 0x000000FF
#define SCSI_MSG_TRANSSEQID_TRANSACTION_ID_SHIFT	 16
#define SCSI_MSG_TRANSSEQID_SEQUENCE_MAX_SHIFT		 8
#define SCSI_MSG_TRANSSEQID_SEQUENCE_ID_SHIFT		 0
#define SCSI_MSG_TRANSSEQID_PACK(_t_, _m_, _s_) \
		((((_t_) << SCSI_MSG_TRANSSEQID_TRANSACTION_ID_SHIFT) & SCSI_MSG_TRANSSEQID_TRANSACTION_ID_MASK) | \
		 (((_m_) << SCSI_MSG_TRANSSEQID_SEQUENCE_MAX_SHIFT)   & SCSI_MSG_TRANSSEQID_SEQUENCE_MAX_MASK) | \
		 (((_s_) << SCSI_MSG_TRANSSEQID_SEQUENCE_ID_SHIFT)    & SCSI_MSG_TRANSSEQID_SEQUENCE_ID_MASK))
#define SCSI_TRANSACTION_ID(_p_) \
		(((_p_) & SCSI_MSG_TRANSSEQID_TRANSACTION_ID_MASK) >> SCSI_MSG_TRANSSEQID_TRANSACTION_ID_SHIFT)
#define SCSI_SEQUENCE_MAX(_p_) \
		(((_p_) & SCSI_MSG_TRANSSEQID_SEQUENCE_MAX_MASK)   >> SCSI_MSG_TRANSSEQID_SEQUENCE_MAX_SHIFT)
#define SCSI_SEQUENCE_ID(_p_) \
		(((_p_) & SCSI_MSG_TRANSSEQID_SEQUENCE_ID_MASK)    >> SCSI_MSG_TRANSSEQID_SEQUENCE_ID_SHIFT)

/*
 *----------------------------------
 * Multi-packet SCSI message state machine
 *----------------------------------
 */
typedef struct {
	uint32	    tag;
	void *		pdev;
	int			no_guaranteed_delivery;
	uint32	    transactionId;
	uint32		maxSequenceId;
	uint32		lastSequenceId;
	eee_desc_t *firstEDesc;
	eee_desc_t *lastEDesc;
	uint32	    accumulated_checksum;
    uint32      totalpktlen;
    uint32      totalBlks;
	int32		ack_timer_handle;		/* ACK timer handle */
	int32		exch_timer_handle;		/* EXCH timer handle */
	uint32		eCode;					/* Read response error code */
	int32		nack_received;			/* Schedule read retransmission */
	void *		private;				/* Pointer to read scsi desc */
} scsi_mp_desc_t;

/*
 *----------------------
 * Relay CDB Messages
 *----------------------
 */
typedef struct _scsi_relayCdb_msg {
	scsi_msgHdr_t hdr;			  // SCSI_MID_RELAY_REQ
	uint32		  transSeqId;	  // transaction:sequence id 
	uint32        lun;			
	uchar8		  cdb[16];
	uint32        retryCnt;       // Number of times retry attempted  
	uint32        timeOutCnt;     // Number of seconds before timeout 
	uint32        maxTimeToLive;  // Num secs desc can live   120sec 
	int32         dataInLen;	  // Specifies len of data in
	int32		  dataLen;		  // non-zero if data OUT; zero if data IN.
	char		  data[1];		  // Variable length data 
} __attribute__ ((packed)) scsi_relayReqMsg_t;

typedef struct _scsi_relayRsp_msg {
	scsi_msgHdr_t hdr;			// SCSI_MID_RELAY_RSP
	uint32		  transSeqId;	// transaction:sequence id 
	uint32		  status;		// Status byte ret by HAL & device 
	uchar8		  cdb[16];
	uchar8		  senseData[32];
	int32		  dataLen;
	uchar8		  data[1];		// Variable length data 
} __attribute__ ((packed)) scsi_relayRspMsg_t;

/*
 *----------------------
 * Device OPEN Messages
 *----------------------
 */
typedef struct _scsi_openReq_msg {
    scsi_msgHdr_t hdr;			// SCSI_MID_OPEN_REQ
} scsi_openReqMsg_t;


typedef struct _scsi_openRsp_msg {
    scsi_msgHdr_t hdr;			// SCSI_MID_OPEN_RSP
	uint32        eCode;
} scsi_openRspMsg_t;


/*
 *----------------------
 * Device CLOSE Message 
 *----------------------
 */
typedef struct _scsi_closeReq_msg {
    scsi_msgHdr_t hdr;			// SCSI_MID_CLOSE_REQ
} scsi_closeReqMsg_t;

typedef struct _scsi_closeRsp_msg {
    scsi_msgHdr_t hdr;			// SCSI_MID_CLOSE_RSP
	uint32        eCode;
} scsi_closeRspMsg_t;

/*
 *----------------------
 * Device CLOSE Message 
 *----------------------
 */
typedef struct _scsi_deleteReq_msg {
    scsi_msgHdr_t hdr;			// SCSI_MID_DELETE_REQ
} scsi_deleteReqMsg_t;

typedef struct _scsi_deleteRsp_msg {
    scsi_msgHdr_t hdr;			// SCSI_MID_DELETE_RSP
	uint32        eCode;
} scsi_deleteRspMsg_t;

/*
 *----------------------
 * Device ADD Message 
 *----------------------
 */
#define SCSI_MAPPING_MAX_PARTITIONS		16
#define SCSI_MAPPING_OFFSET_BLOCKS		32

typedef struct {
	uint32				partBegin;
	uint32				partSize;
} scsi_partInfo_t;

typedef struct {
#define SCSI_MAPPING_FLAGS_BCS_ENABLED		0x00000001
	uint32	            flags;
	uint32				offset;
	uint32				partitionCount;
	scsi_partInfo_t		partitions[SCSI_MAPPING_MAX_PARTITIONS];
} scsi_MappingInfo_t;

typedef struct _scsi_addReq_msg {
    scsi_msgHdr_t		hdr;			// SCSI_MID_ADD_REQ
	uint32				devType;
	scsi_id_t			id;
	uint64				wwn;
	uint64				chn_wwn;
	uint32				lun;
	uint32				llHandle;
	uint32				blockCnt;
	uint32				blockSize;
	uint32				raidLevel;
	uint32				sdmState;
	char				vendor[9];
	char				model[17];
	char				productId[17];
	char				sn[32];
	char				name[SCSI_DEV_NAME_LEN];
	scsi_MappingInfo_t	map;
} scsi_addReqMsg_t;

typedef struct _scsi_addRsp_msg {
    scsi_msgHdr_t hdr;			// SCSI_MID_ADD_RSP
	uint32        eCode;
} scsi_addRspMsg_t;

/*
 *----------------------
 * Device DELETE Message
 *----------------------  
 */
typedef struct _scsi_delDevReq_msg {
	scsi_msgHdr_t	hdr;		// SCSI_MID_DEL_DEV_REQ
	scsi_id_t		id;
} scsi_delDevReqMsg_t;


/*
 *--------------------------------------------------------------
 * Device Sync Request Message
 *--------------------------------------------------------------
 *
 * When sent to the SCSI application it will result in the
 * SCSI application sending sdm_addDevMsg_t to the SDM for
 * each device in the SCSI application device list. The purpose
 * of this message is to allow the SDM to resychronize with
 * the SCSI application following a shutdown of the SDM.
 *--------------------------------------------------------------
 */
typedef struct _scsi_syncReq_msg
{
	scsi_msgHdr_t hdr;					// SCSI_MID_SYNC_REQ
} scsi_syncReqMsg_t;

/*
 *---------------------------------
 * Device Reset messages
 *---------------------------------
 */
typedef struct _scsi_devResetReqMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_RESET_REQ
} scsi_devResetReqMsg_t;

typedef struct _scsi_devResetRspMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_RESET_RSP
	uint32		  eCode;					
} scsi_devResetRspMsg_t;

/*
 *---------------------------------
 * Device Status messages
 *---------------------------------
 */
typedef struct _scsi_devStatusReqMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_STATUS_REQ
} scsi_devStatusReqMsg_t;

typedef struct _scsi_devStatusRspMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_STATUS_RSP
	uint32		  eCode; 
	uint64		  scsi_id;
	uint32		  tape_position_eom;
	uint32		  tape_position_bom;
	int32		  tape_position_blockno;
	int32		  tape_position_fileno;
} scsi_devStatusRspMsg_t;

/*
 *---------------------------------
 * IOCTL messages
 *---------------------------------
 */
#define SCSI_IOCTL_MAX_ARGS		16
typedef struct _scsi_ioctlReqMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_IOCTL_REQ
	uint32		  code;
	uint32		  arg_count;
	uint64		  args[SCSI_IOCTL_MAX_ARGS];
} scsi_ioctlReqMsg_t;

typedef struct _scsi_ioctlRspMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_IOCTL_RSP
	uint32		  eCode;					
    int32		  dataLen;
    uint64		  data[1];         		// Variable length data 
} scsi_ioctlRspMsg_t;


//---------------------------------
//-- Device Stats messages
//---------------------------------
typedef struct _scsi_devStatsReqMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_STATS_REQ
} scsi_devStatsReqMsg_t;

typedef struct _scsi_devStatsRspMsg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_STATS_RSP
	uint32		  eCode;					
	uint32        writeReqCnt;
	uint32        writeRspCnt;
	uint32        writeCSECnt;
	uint32		  readReqCnt;
	uint32		  readRspCnt;
	uint32        readCSECnt;
	uint32        readBECnt;
	uint32		  relayReqCnt;
	uint32		  relayRspCnt;
	uint32		  devErrCnt;
} scsi_devStatsRspMsg_t;

/*
 *----------------------
 * Device Control Messages
 *----------------------
 */
#define SCSI_DEV_CONTROL_ENABLE_MASK_SORTING		0x00000001
#define SCSI_DEV_CONTROL_ENABLE_MASK_COALLESCING	0x00000002
#define SCSI_DEV_CONTROL_ENABLE_MASK_CS_VERIFY		0x00000004

#define SCSI_DEV_CONTROL_ENABLE_SORTING				0x00000001
#define SCSI_DEV_CONTROL_DISABLE_SORTING			0x00000000
#define SCSI_DEV_CONTROL_ENABLE_COALLESCING			0x00000002
#define SCSI_DEV_CONTROL_DISABLE_COALLESCING		0x00000000
#define SCSI_DEV_CONTROL_ENABLE_CS_VERIFY			0x00000004
#define SCSI_DEV_CONTROL_DISABLE_CS_VERIFY			0x00000000

typedef struct _scsi_control_msg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_CONTROL_REQ
	uint32		  control_enable_mask;	// Gate
	uint32		  control_mask;			// Enable/Disable
} scsi_controlReqMsg_t;

typedef struct _scsi_controlRsp_msg {
	scsi_msgHdr_t hdr;					// SCSI_MID_DEV_CONTROL_RSP
	uint32		  status;				// Status byte ret by HAL & device 
} scsi_controlRspMsg_t;


/*
 * -----------------------
 * Clear stats message
 * ----------------------- 
 */
#define SCSI_STATS_TYPE_SCSI	0x00000001
#define SCSI_STATS_TYPE_DEV		0x00000002
#define SCSI_STATS_TYPE_DEVALL	0x00000004

typedef struct _scsi_clearStatsReq_msg {
	scsi_msgHdr_t	hdr;			// SCSI_MID_CLR_STATS_REQ
	uint32			type;
	scsi_id_t		id;
} scsi_clearStatsReqMsg_t;

/*
 * -----------------------
 * Get Stats message
 * -----------------------
 */
union _scsi_dev_stats {
	scsi_stats_t	scsi_stats;
	dev_stats_t		dev_stats;
};

typedef struct _scsi_getStatsReq_msg {
	scsi_msgHdr_t	hdr;			// SCSI_MID_GET_STATS_REQ
	uint32			type;
	scsi_id_t		id;
} scsi_getStatsReqMsg_t;

typedef struct _scsi_getStatsRsp_msg {
	scsi_msgHdr_t	hdr;			// SCSI_MID_GET_STATS_RSP
	uint32			type;
	uint32			eCode;
	union _scsi_dev_stats	stats;
} scsi_getStatsRspMsg_t;

/*
 * --------------------
 * Large IO message
 * --------------------
 */

#define SCSI_HUGEIO_READ		1
#define SCSI_HUGEIO_WRITE		2

typedef struct _scsi_hugeioReq_msg {
	scsi_msgHdr_t	hdr;			// SCSI_MID_HUGEIO_REQ
	int				op;				// read or write
	uint64			wwn;			// device wwn
	uint32			lun;			// device lun
	uint32			lba;			// start block offset
	uint32			size;			// read/write size
	uint32			count;			// number of times to do the op.
} scsi_hugeioReqMsg_t;

typedef struct _scsi_hugeioRsp_msg {
	scsi_msgHdr_t	hdr;			// SCSI_MID_HUGEIO_RSP
	uint64			wwn;			// device wwn
	uint32			lun;			// device lun
	uint32			eCode;			// return code
} scsi_hugeioRspMsg_t;

#endif  /* SCSI_MSG_H_INCLUDED */

