/*  FileName: scsi-sm-funcs.c
 *
 *  Description: Contains action functions for the device state machines.
 *
 *	Created:	18 May 2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_dev_procInq
 *
 * Description:     Called when a INQUIRY response event is queued to the 
 *					state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 18, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_dev_procInq( int32		   nextState,	/* next state from table	  */
				  scsi_dev_t*  pdev,
				  scsi_desc_t* sd) 	
				
{
	int			rc;
	int32		nState;

	scsi_debug("[scsi_dev_procInq] Enter", SCSI_DBL_3);

	/*
	 * Call function to parse the INQUIRY response.
	 */
	rc = scsi_parseInqRsp(pdev,sd);
	if( rc == SCSI_ERR_GENERIC ) {

		/*
		 * Dealloc the sd and if present the eDesc
		 */
		scsi_deallocDesc(sd);
		esm_destroy( pdev->smHandle );
		pdev->smHandle = 0;
		scsi_deallocDevice( pdev );
		return(nextState);
	}

#if 0 //-- Inialization is now done by SDM.
	/*
	 * Release inquiry response packet descriptor.
	 */
	if( sd->eDesc)
		eee_freePacket(sd->eDesc);
	sd->eDesc = 0;

	/*
	 * Get SCSI descriptor ready for reuse.
	 */
	sd->status     = 0;
	sd->retryCnt   = 0;
	sd->timeOutCnt = 0;
	memset(&sd->cdb, 0, 16);

	/* 
	 * Perform next initialization step based on device type.
	 */
	//printf("scsi_dev_procInq: vendor %s, product %s, path 0x%x, scsi_id 0x%016llx, SCSI Type %d\n",
	//	   pdev->vendor, pdev->model, pdev->path, pdev->id, pdev->type);
	switch(pdev->type)
	{
		case SCSI_DEVTYPE_DISK:
			if (SCSI_ID_GET_LUN(pdev->id) == 0) {
				scsi_sendReportLun(pdev, sd);
				nState = SCSI_DEVI_STATE_RLUN_WAIT;
			}
			else {
				scsi_sendReadCapacity(pdev, sd);
				nState = SCSI_DEVI_STATE_RCAP_WAIT;
			}
			break;

		case SCSI_DEVTYPE_TAPE: 
			pdev->blocks = 0;
			pdev->blockSize = 0;
			pdev->blockno = -1;
			pdev->byteno = -1;
			pdev->fileno = -1;
			pdev->eom = pdev->bom = 0;
			scsi_sendAddDevMsg(pdev);
			nState = SCSI_DEVI_STATE_ACK_WAIT;
			break;
			
		case SCSI_DEVTYPE_MEDIA_CHANGER:
			/* Look for LUNs sitting behind stacker */
			for (i=1; i < 16; i++) {
				scsi_newDevice( SCSI_LLNOTIFY_FC,
								pdev->llHandle,
								pdev->wwn,
								0, 0, i, "Null");
			}
			pdev->blocks = 0;
			pdev->blockSize = 0;
			scsi_sendAddDevMsg(pdev);
			nState = SCSI_DEVI_STATE_ACK_WAIT;
			break;

		/* Assume media is not installed. Don't do anything more now */
		case SCSI_DEVTYPE_CDROM:
		case SCSI_DEVTYPE_WORM:
		case SCSI_DEVTYPE_OPTICAL_DISK:

		/* Device Types Not Supported */
		case SCSI_DEVTYPE_RAID:
		case SCSI_DEVTYPE_PRINTER:
		case SCSI_DEVTYPE_PROCESSOR:
		case SCSI_DEVTYPE_COMM_DEVICE:
		case SCSI_DEVTYPE_SES:
		case SCSI_DEVTYPE_UNKNOWN:
		default:
        {
			// Set flag to delete this state machine instance.
			// the next time the event queue for this sm is serviced.
			esm_destroy(pdev->smHandle);

			// Delete the device control block.
			scsi_deallocDevice( pdev );

			nState = nextState;
		}
	}// End Switch 
	scsi_debug("[scsi_dev_procInq] Leave", SCSI_DBL_3);
#else
	scsi_deallocDesc(sd);
#endif


	return nState;
} /* scsi_dev_procInq */


/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_dev_procRLun
 *
 * Description:     Called when a REPORT_LUN response event is queued 
 *					to the state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
scsi_dev_procRLun(int32		   nextState,	/* next state from table	  */
				  scsi_dev_t*  pdev,
				  scsi_desc_t* sd)
{
#if 0 //-- Device initialization now performed by the SDM
	int32		i, lunCnt, *luns;
	int32		nState, myLunId, cmdStatus;
	scsi_sense_t *sense;

	scsi_debug("[scsi_dev_procRLun] Enter", SCSI_DBL_3);

	/* 
	 * Check command completion status.
	 */
	cmdStatus = sd->status & SD_DEV_STATUS_MASK;
	switch( cmdStatus )
	{
		case SCSI_DEV_STATUS_CHECK:
		case SCSI_DEV_STATUS_CMD_TERMINATED:
			if( sd->status & SD_DEV_SENSE_VALID ) {

				sense = (scsi_sense_t *)sd->senseData;

				/* Look at sense data */
				switch( sense->senseKey )
				{
					case SCSI_SENSE_KEY_NOT_READY:
					case SCSI_SENSE_KEY_HARDWARE_ERROR:
						/* 
						 * Fatal Error unable to access device -- destroy device instance 
						 */
						esm_destroy(pdev->smHandle);
						scsi_deallocDevice( pdev );
						scsi_deallocDesc( sd );
						break;

					case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
						/* 
						 * Device does not support this command, so
						 * probe for n LUN's and see what comes
						 * back.
						 */ 
						for(i=1; i < SCSI_MAX_LUN_PROBE_CNT; i++)
							scsi_newDevice( SCSI_LLNOTIFY_FC,
											pdev->llHandle,
											pdev->wwn,
											0, 0, i, "Null");
						/* fall thru to add "primary" device */
						
					case SCSI_SENSE_KEY_NO_SENSE:
					case SCSI_SENSE_KEY_RECOVERED_ERROR:
					default:
						/*
						 * Release report luns response packet descriptor.
						 */
						if( sd->eDesc)
							eee_freePacket(sd->eDesc);
						sd->eDesc = 0;

						/*
						 * Get SCSI descriptor ready for reuse.
						 */
						sd->status     = 0;
						sd->retryCnt   = 0;
						sd->timeOutCnt = 0;
						memset(&sd->cdb, 0, 16);

						scsi_sendReadCapacity(pdev, sd);
						nState = SCSI_DEVI_STATE_RCAP_WAIT;
						break;
				}
			}
			break;

		case SCSI_DEV_STATUS_GOOD:
		case SCSI_DEV_STATUS_CONDITION_MET:
			/* Parse the REPORT_LUN response. */
			lunCnt = 0;
			luns   = (int32 *)NULL;
			scsi_parseReportLunRsp(pdev,sd, &lunCnt, &luns);

			/* For each LUN that is reported create SM instance. */
			myLunId = SCSI_ID_GET_LUN(pdev->id);
			for(i=0; i < lunCnt; i++) {

				if( *(luns+i) == myLunId )
					continue;

				scsi_newDevice( SCSI_LLNOTIFY_FC,
								pdev->llHandle,
								pdev->wwn,
								(uint64)0,
								0,
								*(luns+i), 
								"Null");
			}

			if( luns && (lunCnt > 0))
				eee_ramDealloc(luns);

			/*
			 * Release report luns response packet descriptor.
			 */
			if( sd->eDesc)
				eee_freePacket(sd->eDesc);
			sd->eDesc = 0;

			/*
			 * Get SCSI descriptor ready for reuse.
			 */
			sd->status     = 0;
			sd->retryCnt   = 0;
			sd->timeOutCnt = 0;
			memset(&sd->cdb, 0, 16);

			scsi_sendReadCapacity(pdev, sd);
			nState = SCSI_DEVI_STATE_RCAP_WAIT;
			break;

		default:
			scsi_deallocDesc( sd );
	}
	scsi_debug("[scsi_dev_procRLun] Leave", SCSI_DBL_3);
	return nState;
#else
	return 0;
#endif
	
} /* scsi_dev_procRLun */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_dev_procRCap
 *
 * Description:     Called when a READ_CAPACITY response event is queued 
 *					to the state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
scsi_dev_procRCap(int32		   nextState,	/* next state from table	  */
				  scsi_dev_t*  pdev,
				  scsi_desc_t* sd)
{
#if 0 //-- Device initialization now performed by SDM.
	int32		cmdStatus, lba, lbSize, myLunId;
	int32		nState;
	scsi_sense_t *sense;

	scsi_debug("[scsi_dev_procRCap] Enter", SCSI_DBL_3);

	/* 
	 * Check command completion status.
	 */
	cmdStatus = sd->status & SD_DEV_STATUS_MASK;
	switch( cmdStatus )
	{
		case SCSI_DEV_STATUS_RESERVE_CONFLICT:
			/* Device is reserved by another initiator so
			 * don't try to use it anymore.
			 */
			nState = nextState;
			break;

		case SCSI_DEV_STATUS_CHECK:
		case SCSI_DEV_STATUS_CMD_TERMINATED:
			if( cmdStatus & SD_DEV_SENSE_VALID ) {

				sense = (scsi_sense_t *)sd->senseData;

				/* Look at sense key */
				switch( sense->senseKey )
				{
					case SCSI_SENSE_KEY_NOT_READY:
					case SCSI_SENSE_KEY_HARDWARE_ERROR:
					case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
						/* Fatal Error unable to access device */
						nState = nextState;
						break;

					case SCSI_SENSE_KEY_NO_SENSE:
					case SCSI_SENSE_KEY_RECOVERED_ERROR:
					default:
						/* problem not fatal, fall through */
				}
			}

		case SCSI_DEV_STATUS_GOOD:
		case SCSI_DEV_STATUS_CONDITION_MET:
			/* Parse the READ CAPACITY response. */
			scsi_parseReadCapacityRsp(pdev,sd);
			/* Only advertise devices with "sane" blocksizes; ignore all others */
			if (pdev->blockSize == BLOCKSIZE_512 || pdev->blockSize == BLOCKSIZE_520)
			    scsi_sendAddDevMsg(pdev);
			nState = SCSI_DEVI_STATE_ACK_WAIT;
			break;

		default:

	}
	scsi_deallocDesc( sd );
	scsi_debug("[scsi_dev_procRCap] Leave", SCSI_DBL_3);
	return nState;	
#else
	return 0;
#endif
} /* scsi_dev_procRCap */

