/*  FileName: scsi-dev-cmds.c
 *
 *  Description: Contains support routines for sending and recieving 
 *               scsi device commands.
 *
 *	Created:	12 February	2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"

scsi_cmd_t scsi_tapeCmds[8];

/*
 * Tape Command Function Prototypes
 */
int scsi_cmnRelayCDB(scsi_dev_t *, scsi_desc_t *);
int scsi_tapeReadData(scsi_dev_t *, scsi_desc_t *);
int scsi_tapeWriteData(scsi_dev_t *, scsi_desc_t *);
int scsi_tapeRelayCDBErr(scsi_dev_t *, scsi_desc_t *);
int scsi_tapeReadDataErr(scsi_dev_t *, scsi_desc_t *);
int scsi_tapeWriteDataErr(scsi_dev_t *, scsi_desc_t *);


/*
 *-------------------------------------------------------------
 * Function Name:   scsi_tapeInitCmds
 *
 * Description: Sets up tape specific command table.
 *
 * Returns: NFX_OK | NFX_ERR 
 *
 *-------------------------------------------------------------
 */
int
scsi_tapeInitCmds()
{
    memset( scsi_tapeCmds, 0, sizeof(scsi_tapeCmds));

    scsi_tapeCmds[SD_OPCODE_RELAY].req_func = scsi_cmnRelayCDB;
    scsi_tapeCmds[SD_OPCODE_RELAY].err_func = scsi_tapeRelayCDBErr;
    scsi_tapeCmds[SD_OPCODE_RELAY].name = "Relay CDB";

    scsi_tapeCmds[SD_OPCODE_READ].req_func  = scsi_tapeReadData;
    scsi_tapeCmds[SD_OPCODE_READ].err_func  = scsi_tapeReadDataErr;
    scsi_tapeCmds[SD_OPCODE_READ].name  = "Read Tape Data";

    scsi_tapeCmds[SD_OPCODE_WRITE].req_func = scsi_tapeWriteData;
    scsi_tapeCmds[SD_OPCODE_WRITE].err_func = scsi_tapeWriteDataErr;
    scsi_tapeCmds[SD_OPCODE_WRITE].name = "Write Tape Data";

    return NFX_OK;
} /* scsi_tapeInitCmds */

/*
 *-------------------------------------------------------
 * Function Name:	scsi_tapeReadData
 *
 * Description: Will send the SCSI_READ_BLOCK command to 
 *  			a SCSI device.
 *
 * Returns: NFX_OK | NFX_ERR 
 *
 *-------------------------------------------------------
 */
int
scsi_tapeReadData( scsi_dev_t *pDev, scsi_desc_t *sd ) 
{
	uint32 count;

    count = sd->lbCnt;

	sd->retryCnt      = 1;
	sd->timeOutCnt	  = 60 * 60;
	sd->maxTimeToLive = 60 * 60;
	sd->expectDataLen = count;
	sd->status |= SD_DEV_DATA_IN_EXPECT;

	if (count > SCSI_MAX_LBA_6) {
		return SCSI_MSG_ECODE_OO_RANGE;
	}
	else {
        sd->cdb[0] = SCSI_CMD_READ_6;
        sd->cdb[1] = 0;							// Fixed or Variable len block modes.

        sd->cdb[2] = ((count >> 16) & 0xFF);	// Block Count MSB
        sd->cdb[3] = ((count >> 8)  & 0xFF);  
        sd->cdb[4] = (count & 0xFF);			// Block Count LSB
    }
    return NFX_OK;
} /* scsi_tapeReadData */

/*
 *-------------------------------------------------------
 * Function Name:	scsi_tapeWriteData
 *
 * Description: Creates CDB with write block command.
 *
 * Returns: NFX_OK | NFX_ERR 
 *
 *-------------------------------------------------------
 */
int
scsi_tapeWriteData( scsi_dev_t *pDev, scsi_desc_t *sd ) 
{
	uint32 count;

    count = sd->lbCnt;

    sd->retryCnt      = 1;
	sd->timeOutCnt	  = 60 * 60;
	sd->maxTimeToLive = 60 * 60;
	sd->expectDataLen = count;
	sd->status |= SD_DEV_DATA_OUT_EXPECT;

	if (count > SCSI_MAX_LBA_6) {
		return SCSI_MSG_ECODE_OO_RANGE;
	}
	else {
		 sd->cdb[0] = SCSI_CMD_WRITE_6;
		 sd->cdb[1] = 0;						// Fixed or Variable len block modes.

		 sd->cdb[2] = ((count >> 16) & 0xFF);	// Block Count MSB
		 sd->cdb[3] = ((count >> 8)  & 0xFF);  
		 sd->cdb[4] = (count & 0xFF);			// Block Count LSB
    }
	return NFX_OK;
} /* scsi_tapeWriteData */

int 
scsi_tapeRelayCDBErr(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32			dev_status = (sd->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
	uint32			drv_status = (sd->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT;
	scsi_sense_t *	sense;
	int32			count;
	int32			code;
	int32			residue;

	pdev->eom = pdev->bom = 0;

	switch(sd->cdb[0]) {
	case SCSI_CMD_ERASE:
	case SCSI_CMD_LOAD_UNLOAD:
		pdev->blockno = pdev->byteno = pdev->fileno = -1;
		break;	

	case SCSI_CMD_REWIND_UNIT:
		if (drv_status || dev_status != SCSI_DEV_STATUS_GOOD) {
			pdev->blockno = pdev->byteno = pdev->fileno = -1;
		}
		else {
			pdev->bom = 1;
			pdev->blockno = 0;
			pdev->byteno = 0;
			pdev->fileno = 0;
		}
		break;	

	case SCSI_CMD_SPACE:
		code = sd->cdb[1] & 0x07;
		count = (sd->cdb[2] << 16) | (sd->cdb[3] << 8) | sd->cdb[4];
		if (sd->cdb[2] & 0x80)
			count |= ((int)-1 & ~0xFFFFFF);
		if (drv_status) {
			pdev->blockno = pdev->byteno = pdev->fileno = -1;
		}
		else
		if (dev_status != SCSI_DEV_STATUS_GOOD) {
			switch (dev_status) {
			case SCSI_DEV_STATUS_CHECK:
				sense = (scsi_sense_t *)sd->senseData;
				switch (sense->senseKey & SCSI_SENSE_KEY_MASK) {
				case SCSI_SENSE_KEY_NO_SENSE:
				case SCSI_SENSE_KEY_RECOVERED_ERROR:
					if ((sense->senseKey & SCSI_SENSE_EOM_BIT) && sense->asc == 0 && sense->ascQ == 2)
						pdev->eom = 1;
					if (sense->respCode & SCSI_SENSE_VALID_BIT) 
						residue = (sense->info[0] << 24) | (sense->info[1] << 16) | (sense->info[2] << 8) | sense->info[3];
					else
						residue = 0;

					if (code == 0) {	/* space blocks */
						if (pdev->blockno >= 0) {
							if (sense->senseKey & SCSI_SENSE_FILEMARK_BIT) {
								if (count >= 0) {
									if (pdev->fileno >= 0) {
										pdev->fileno++;
										pdev->blockno = 0;
										pdev->byteno = 0;
									}
								}
								else {
									if (pdev->fileno >= 0) {
										pdev->fileno--;
										pdev->blockno = -1;
										pdev->byteno = -1;
										if (pdev->fileno == 0)
											pdev->bom = 1;
									}
								}
							}
							else 
							if (sense->senseKey & SCSI_SENSE_EOM_BIT) { 
									pdev->fileno = 0;
									pdev->blockno = 0;
									pdev->byteno = 0;
							}
							else {
								pdev->blockno += (count - residue);
								pdev->byteno += (count - residue) * 65536; /* Estimate blocksize */
							}
						}
					}
					else {				/* space filemarks */
						if (pdev->fileno >= 0) {
							pdev->fileno += (count - residue);
							if (pdev->fileno == 0)
								pdev->bom = 1;
							if (count >= 0)
								pdev->blockno = pdev->byteno = 0;
							else 
								pdev->blockno = pdev->byteno = -1;
						}
					}
					break;
					
				default:
					pdev->blockno = pdev->byteno = pdev->fileno = -1;
				}
				break;
				
			default:
				pdev->blockno = pdev->byteno = pdev->fileno = -1;
			}
		}
		else
		if (code == 0) {	/* space blocks */
			if (pdev->blockno >= 0) {
				pdev->blockno += count;
				pdev->byteno += (count - residue) * 65536; /* Estimate blocksize */
			}
		}
		else {				/* space filemarks */
			if (pdev->blockno >= 0) {
				pdev->fileno += count;
				if (pdev->fileno == 0)
					pdev->bom = 1;
				if (count > 0)
					pdev->blockno = 0;
				else
					pdev->blockno = -1;
			}
		}
		break;

	case SCSI_CMD_WRITE_FILEMARKS:
		count = (sd->cdb[2] << 16) | (sd->cdb[3] << 8) | sd->cdb[4];
		if (drv_status) {
			pdev->blockno = pdev->byteno = pdev->fileno = -1;
		}
		else
		if (dev_status != SCSI_DEV_STATUS_GOOD) {
			switch (dev_status) {
			case SCSI_DEV_STATUS_CHECK:
				sense = (scsi_sense_t *)sd->senseData;
				switch (sense->senseKey & SCSI_SENSE_KEY_MASK) {
				case SCSI_SENSE_KEY_NO_SENSE:
				case SCSI_SENSE_KEY_RECOVERED_ERROR:
					if (pdev->fileno >= 0) {
						if ((sense->senseKey & SCSI_SENSE_EOM_BIT) && sense->asc == 0 && sense->ascQ == 2)
							pdev->eom = 1;
						if (sense->respCode & SCSI_SENSE_VALID_BIT) 
							residue = (sense->info[0] << 24) | (sense->info[1] << 16) | (sense->info[2] << 8) | sense->info[3];
						else
							residue = 0;
						pdev->fileno += (count - residue);
						pdev->blockno = pdev->byteno = 0;
					}
					break;

				default:
					pdev->blockno = pdev->byteno = pdev->fileno = -1;
				}
				break;
				
			default:
				pdev->blockno = pdev->byteno = pdev->fileno = -1;
			}
		}
		else
		if (pdev->fileno >= 0) {
			pdev->fileno += count;
			pdev->blockno = pdev->byteno = 0;
		}
		break;
	}

	return(SCSI_MSG_ECODE_OK);
}

int 
scsi_tapeReadDataErr(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32			dev_status = (sd->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
	uint32			drv_status = (sd->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT;
	uint32			ecode = SCSI_MSG_ECODE_OK;
	scsi_sense_t *	sense;

	pdev->eom = pdev->bom = 0;

	if (drv_status) {
		pdev->fileno = pdev->byteno = pdev->blockno = -1;
		if (sd->eDesc)
			sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
	    return(SCSI_MSG_ECODE_PATH_FAILED);
	}
	else
	if (dev_status != SCSI_DEV_STATUS_GOOD) {
	    switch (dev_status) {
	    case SCSI_DEV_STATUS_CHECK:
			sense = (scsi_sense_t *)sd->senseData;
			switch (sense->senseKey & SCSI_SENSE_KEY_MASK) {
			case SCSI_SENSE_KEY_NO_SENSE:
				if (sense->senseKey & SCSI_SENSE_ILI_BIT) {
					/* XXX Ignore residue for now */
					ecode = SCSI_MSG_ECODE_ILI;
				}
				else
				if ((sense->senseKey & SCSI_SENSE_EOM_BIT) && sense->asc == 0 && sense->ascQ == 2) {
					ecode = SCSI_MSG_ECODE_END_OF_MEDIUM;
				}
				else
				if (sense->senseKey & SCSI_SENSE_FILEMARK_BIT) {
					ecode = SCSI_MSG_ECODE_FM_ENCOUNTERED;
				}
				goto read_ok;

			case SCSI_SENSE_KEY_RECOVERED_ERROR:
				goto read_ok;

			case SCSI_SENSE_KEY_MEDIUM_ERROR:
				pdev->fileno = pdev->byteno = pdev->blockno = -1;
				if (sd->eDesc)
					sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_MEDIUM_ERROR);

			case SCSI_SENSE_KEY_BLANK_CHECK:
			case SCSI_SENSE_KEY_VOLUME_OVERFLOW:
				if (sd->eDesc)
					sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_END_OF_MEDIUM);
				
			case SCSI_SENSE_KEY_NOT_READY:
			case SCSI_SENSE_KEY_HARDWARE_ERROR:
			case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
			case SCSI_SENSE_KEY_UNIT_ATTENTION:
			case SCSI_SENSE_KEY_DATA_PROTECT:
			case SCSI_SENSE_KEY_VENDOR_SPECIFIC:
			case SCSI_SENSE_KEY_COPY_ABORTED:
			case SCSI_SENSE_KEY_ABORTED_COMMAND:
			case SCSI_SENSE_KEY_OBSOLETE:
			case SCSI_SENSE_KEY_MISCOMPARE:
			case SCSI_SENSE_KEY_RESERVED:
			default:
				pdev->fileno = pdev->byteno = pdev->blockno = -1;
				if (sd->eDesc)
					sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_DEV_IO_ERROR);
			}
			
	    default:
			pdev->fileno = pdev->byteno = pdev->blockno = -1;
			if (sd->eDesc)
				sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
			return(SCSI_MSG_ECODE_DEV_IO_ERROR);
	    }
	}

 read_ok:
	if (ecode == SCSI_MSG_ECODE_END_OF_MEDIUM)
		pdev->eom = 1;
	if (ecode == SCSI_MSG_ECODE_FM_ENCOUNTERED) {
		if (pdev->fileno >= 0) {
			pdev->fileno++;
			pdev->blockno = pdev->byteno = 0;
		}
	}
	else
	if (pdev->blockno >= 0) {
		pdev->blockno++;
		pdev->byteno += 65536; /* Estimate blocksize */
	}

	return(ecode);
}

int 
scsi_tapeWriteDataErr(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32			dev_status = (sd->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
	uint32			drv_status = (sd->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT;
	uint32			ecode = SCSI_MSG_ECODE_OK;
	scsi_sense_t *	sense;

	pdev->eom = pdev->bom = 0;

	if (drv_status) {
		pdev->fileno = pdev->byteno = pdev->blockno = -1;
		sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
	    return(SCSI_MSG_ECODE_PATH_FAILED);
	}
	else
	if (dev_status != SCSI_DEV_STATUS_GOOD) {
	    switch (dev_status) {
	    case SCSI_DEV_STATUS_CHECK:
			sense = (scsi_sense_t *)sd->senseData;
			switch (sense->senseKey & SCSI_SENSE_KEY_MASK) {
			case SCSI_SENSE_KEY_NO_SENSE:
				if ((sense->senseKey & SCSI_SENSE_EOM_BIT) && sense->asc == 0 && sense->ascQ == 2)
				  pdev->eom = 1;
				goto write_ok;

			case SCSI_SENSE_KEY_RECOVERED_ERROR:
				goto write_ok;

			case SCSI_SENSE_KEY_MEDIUM_ERROR:
				pdev->fileno = pdev->byteno = pdev->blockno = -1;
				sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_MEDIUM_ERROR);
				
			case SCSI_SENSE_KEY_NOT_READY:
			case SCSI_SENSE_KEY_HARDWARE_ERROR:
			case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
			case SCSI_SENSE_KEY_UNIT_ATTENTION:
			case SCSI_SENSE_KEY_DATA_PROTECT:
			case SCSI_SENSE_KEY_BLANK_CHECK:
			case SCSI_SENSE_KEY_VENDOR_SPECIFIC:
			case SCSI_SENSE_KEY_COPY_ABORTED:
			case SCSI_SENSE_KEY_ABORTED_COMMAND:
			case SCSI_SENSE_KEY_OBSOLETE:
			case SCSI_SENSE_KEY_VOLUME_OVERFLOW:
			case SCSI_SENSE_KEY_MISCOMPARE:
			case SCSI_SENSE_KEY_RESERVED:
			default:
				pdev->fileno = pdev->byteno = pdev->blockno = -1;
				sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
				return(SCSI_MSG_ECODE_DEV_IO_ERROR);
			}
			
	    default:
			pdev->fileno = pdev->byteno = pdev->blockno = -1;
			sd->eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
			return(SCSI_MSG_ECODE_DEV_IO_ERROR);
	    }
	}

 write_ok:
	if (pdev->blockno >= 0) {
		pdev->blockno++;
		pdev->byteno += 65536; /* Estimate blocksize */
	}

	/* Simulating end-of-medium after 500MB written */
	if (pdev->flags.fSimEarlyEOM && pdev->byteno >= 500 * 1024 * 1024)
	  pdev->eom = 1;
	
	return(ecode);
}

