/*
 * ----------------------------------------------------------------------------
 *  FileName: sdm-agent.c
 *
 *  Description:  Contains SDM agent functions.
 *
 *  Created:    21 August 2001
 *
 *  Author:     Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 *
 *  CVS Version:  $Id: sdm-agent.c,v 1.8 2002/08/29 21:54:33 frank Exp $
 * ----------------------------------------------------------------------------
 */


#ifdef LINUX_TEST
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/cm-api.h>
#include <exports/timer-api.h>
#include <exports/esm-api.h>
#include "../sm-scsi/scsi-api.h"
#include "../sm-sdm/sdm-api.h"
#include "sdm-agent-api.h"

// 
// SDMA Global Variables
//
int32	        sdma_app_id;
uint32	        scsi_app_id;
uint32          sdma_port_addr;
sdma_hEntry_t sdma_h_table[SDM_SLOT_MAX][NFX_FCNIM_NUM_CPUS][SDMA_MAX_HANDLES];

int
sdma_procHandleReqMsg(eee_desc_t *edesc, sdm_sdmaHandleReqMsg_t *i_msg);

/*
 *------------------------------------------------------------------------------
 * Function Name:   sdma_handleToPath
 *
 * Description:     Initializes the SDM agent.
 *
 * Created by:      Frank Salzmann
 *
 * Data Created:    August 20, 2001
 *
 *------------------------------------------------------------------------------
 */
int
sdma_init()
{
	int    rc;

	// Initialize device handle table.
	// Table will be indexed by slot/cpu/handle.
	memset(sdma_h_table, 0, sizeof(sdma_h_table));
	scsi_app_id = eee_getApplicationId("scsi");

	// Setup local SDM Agent messaging.
	sdma_app_id = eee_allocateApplicationId("sdm_agent");
    if( sdma_app_id == NFX_ERR )
    {
		printf("SDMA: Allocate app id failed.");
        return NFX_ERR;
    }

    sdma_port_addr = PACK_UNICAST_PORT_ADDRESS(
									   nfxMySlotId,
									   nfxMyCPUId,
									   sdma_app_id,
									   EEE_CONTROL,
									   NFX_COS_HIGH,
										0);

    rc = eee_registerReceiveFunc( sdma_app_id,      /* App Id        */
                                  sdma_receiveMsg,  /* Ucast Rx Func */
                                  NULL);            /* Mcast Rx Func */
    if( rc == NFX_ERR )
    {
        printf("SDMA: Register application rx func failed.");
        return NFX_ERR;
    }

	return(NFX_OK);
} // sdma_init


/*
 *------------------------------------------------------------------------------
 * Function Name:   sdma_handleToPath
 *
 * Description:     Given a handle, this function will map it to a scsi_path_t 
 *					and generate eee destination port address.	
 *
 * Created by:      Frank Salzmann
 *
 * Data Created:    August 20, 2001
 *
 *------------------------------------------------------------------------------
 */
int
sdma_handleToPath(uint32 handle, scsi_path_t *path, uint32 *scsi_port)
{
	int rc;
	uint32 slot, port, index;
	sdma_hEntry_t *entry;
	
	slot  = SDM_HANDLE_GET_SLOT(handle);
	port  = SDM_HANDLE_GET_PORT(handle);
	index = SDM_HANDLE_GET_INDEX(handle);

	entry = &sdma_h_table[slot][port][index];

	if( entry->state & SDMA_VALID_ENTRY ) {
		*path = entry->path;

		*scsi_port = PACK_UNICAST_PORT_ADDRESS(
									   SCSI_PATH_GET_SLOT(entry->path),
									   SCSI_PATH_GET_PORT(entry->path),
									   scsi_app_id,
									   EEE_USER_DATA,
									   NFX_COS_HIGH,
										0);
		rc = NFX_OK;
	}
	else {
		rc = -1;
	}

	return(rc);
} // sdma_handleToPath


/*
 *------------------------------------------------------------------------------
 * Function Name:   sdma_receiveMsg
 *
 * Description:     Receives messages sent to the sdm agent from the sdm daemon.
 *
 * Created by:      Frank Salzmann
 *
 * Data Created:    August 20, 2001
 *
 *------------------------------------------------------------------------------
 */
void
sdma_receiveMsg( eee_desc_t *edesc )
{
	uint32        mid;
    sdm_msgHdr_t *msg;
    eee_descBuf_t *bd;

    // printf("[sdma_receiveMsg] Enter"); 

    /*
     * Point to request message portion of data buffer.
     */
    if( !edesc ) {
        printf("Null eDesc");
        return;
    }
   
    bd = &edesc->bd[1];
    msg = (sdm_msgHdr_t *)(bd->buf);
    mid = msg->mid;

    switch( mid )
    {
        case SDM_MID_SDMA_SET_PATH_REQ:
			sdma_procSetPathMsg( edesc, msg );
            break;

        case SDM_MID_SDMA_REM_PATH_REQ:
			sdma_procRemPathMsg( edesc, msg );
            break;

		case SDM_MID_SDMA_HANDLE_REQ:
			sdma_procHandleReqMsg( edesc,(sdm_sdmaHandleReqMsg_t *)msg );
            break;

		default: // Invalid Message
			printf("[scsi_receiveMsg] Invalid Message: %x\n", mid);
	}

	eee_freePacket(edesc);
    // printf("[sdma_receiveMsg] Enter"); 
	return;
} // sdma_receiveMsg

/*
 *------------------------------------------------------------------------------
 * Function Name:   sdma_procSetPathMsg
 *
 * Description:     Handler function for the sdm_setPathReqMsg.
 *
 * Created by:      Frank Salzmann
 *
 * Data Created:    August 20, 2001
 *
 *------------------------------------------------------------------------------
 */
int
sdma_procSetPathMsg( eee_desc_t *edesc, sdm_setPathReqMsg_t *i_msg )
{
	sdma_hEntry_t *entry;
	uint32 handle, slot, port, index;

	handle      = i_msg->handle;

	slot  = SDM_HANDLE_GET_SLOT(handle);
	port  = SDM_HANDLE_GET_PORT(handle);
	index = SDM_HANDLE_GET_INDEX(handle);

	entry = &sdma_h_table[slot][port][index];
	entry->id    = i_msg->id;
	entry->path  = i_msg->path;
	entry->state = SDMA_VALID_ENTRY;
	//printf("[SDMA] Set Handle: %x to Path: %x\n", handle, i_msg->path);
	return 0;
} // sdma_procSetPathMsg

/*
 *------------------------------------------------------------------------------
 * Function Name:   sdma_procRemPathMsg
 *
 * Description:     Removes a scsi_path from the handle table.
 *
 * Created by:      Frank Salzmann
 *
 * Data Created:    January 31, 2002
 *
 *------------------------------------------------------------------------------
 */
int
sdma_procRemPathMsg( eee_desc_t *edesc, sdm_remPathReqMsg_t *i_msg )
{
	sdma_hEntry_t *entry;
	uint32 handle, slot, port, index;

	handle      = i_msg->handle;

	slot  = SDM_HANDLE_GET_SLOT(handle);
	port  = SDM_HANDLE_GET_PORT(handle);
	index = SDM_HANDLE_GET_INDEX(handle);

	entry = &sdma_h_table[slot][port][index];
	entry->id    = 0;
	entry->path  = 0;
	entry->state = 0;
	//printf("[SDMA] Rem Handle: %x\n", handle);
	return 0;
} // sdma_procRemPathMsg

/*
 *------------------------------------------------------------------------------
 * Function Name:   sdma_viewHandleTable
 *
 * Description:     UI interface routine for displaying the contents of the
 *					SDM agent handle table.
 *
 * Created by:      Frank Salzmann
 *
 * Data Created:    August 21, 2001
 *
 *------------------------------------------------------------------------------
 */
int
sdma_uiViewHandleTable( int32 argc, char8 *argv[] )
{
	
	uint32 slot, port;

	printf("\n");
	printf("SDMA HANDLE TABLE ENTRIES: \n");
	printf("---------------------------\n");
	for(slot=2, port=0; port < 4; port++)
		sdma_handleTableDisplay( slot, port );

	return NFX_OK;
} // sdma_uiViewHandleTable

/*
 *------------------------------------------------------------------------------
 * Function Name:   sdma_handleTableDisplay
 *
 * Description:     Shows valid entries in the SDM agent handle table.
 *
 * Created by:      Frank Salzmann
 *
 * Data Created:    August 21, 2001
 *
 *------------------------------------------------------------------------------
 */
int
sdma_handleTableDisplay( int32 slot, int32 port )
{
	uint32 i, j, handle;
	sdma_hEntry_t *entry;

	j=0;
	for( i=0; i < SDMA_MAX_HANDLES; i++ ) {
		entry = &sdma_h_table[slot][port][i];
		if( entry->state & SDMA_VALID_ENTRY ) {
			handle = 0;
			SDM_HANDLE_SET_FP_SLOT(nfxMySlotId, handle);
			SDM_HANDLE_SET_SLOT(slot, handle);
			SDM_HANDLE_SET_PORT(port, handle);
			SDM_HANDLE_SET_INDEX(i, handle);
			printf("%03d). Handle: %x ... Path: %x ... Id: %016llx\n", 
				   j, handle, entry->path, entry->id);
			j++;
		}
	}
	printf("\n");
	return NFX_OK;
} // sdma_handleTableDisplay

//------------------------------------------------------------------------------
// Function Name:   sdma_handleTableDisplay
//
// Description:     Returns handle for scsi_id/path.
//
// Created by:      Frank Salzmann
//
// Data Created:    May 08, 2002
//
//------------------------------------------------------------------------------
int
sdma_procHandleReqMsg( eee_desc_t *edesc, sdm_sdmaHandleReqMsg_t *i_msg )
{
	uint32 i, slot, port, handle, mlen; 
	sdma_hEntry_t *h_entry; 
	sdm_sdmaHandleRspMsg_t o_msg;

#if 0
	printf("sdma_procHandleReqMsg: id: %llx, path: %x \n", i_msg->id,
		                                                   i_msg->path);
	sdma_printBuf((char *)i_msg, sizeof(sdm_sdmaHandleReqMsg_t));
	printf("\n");
#endif

	//-- Initialize handle to zero.
	handle = 0;

	//-- First check if handle already exists for device/path.
	for( slot=0; slot < SDM_SLOT_MAX; slot++ ) {
		for( port=0; port < NFX_FCNIM_NUM_CPUS; port++ ) {
			for( i=0; i < SDMA_MAX_HANDLES; i++ ) {
				h_entry = &sdma_h_table[slot][port][i];
				if( h_entry->id == i_msg->id ) {
					SDM_HANDLE_SET_FP_SLOT(nfxMySlotId, handle);
					SDM_HANDLE_SET_SLOT(slot, handle);
					SDM_HANDLE_SET_PORT(port, handle);
					SDM_HANDLE_SET_INDEX(i, handle);
					goto sdma_procHandleReqMsg_hdone;
				}
			}
		}
	}

 sdma_procHandleReqMsg_hdone:

	//-- No existing handle was found for this device/path so
	//-- allocate one.
	if( handle == 0 ) 
		handle = sdma_allocDevHandle(i_msg->id, i_msg->path);
	else
		h_entry->path = i_msg->path; //-- Set path to that requested by the open req.

	//-- Setup response message.
	mlen = sizeof(sdm_sdmaHandleRspMsg_t);
	memset(&o_msg, 0, mlen);
	o_msg.hdr.mid     = SDM_MID_SDMA_HANDLE_RSP;
	o_msg.hdr.chandle = i_msg->hdr.chandle;
	o_msg.handle      = handle;
	o_msg.path        = i_msg->path;
	if( handle )
		o_msg.rCode   = SDM_OK;
	else
		o_msg.rCode   = SDM_FAIL;

	//-- Send response back to SDM.
	eee_sendMessage((uchar8 *)&o_msg, mlen, 
					edesc->hdr.src_port, sdma_port_addr);

	return NFX_OK;
} // sdma_procHandleReqMsg

//------------------------------------------------------------------------------
// Function Name:   sdma_allocDevHandle
//
// Description:     Allocates a device handle 
//
// Created by:      Frank Salzmann
//
// Data Created:    May 08, 2002
//------------------------------------------------------------------------------
int
sdma_allocDevHandle(scsi_id_t id, scsi_path_t path)
{
	uint32 i;
	uint32 slot, port, handle;
	sdma_hEntry_t *h_entry; 

#if 0
	printf("sdma_allocDevHandle: id: %llx, path: %x, ", id, path);
#endif

	slot = SCSI_PATH_GET_SLOT(path);
	port = SCSI_PATH_GET_PORT(path);
	
	handle = 0;
	for(i=0; i < SDMA_MAX_HANDLES; i++ ) { 
		//-- Look for empy space.
		h_entry = &sdma_h_table[slot][port][i];
		if( h_entry->state == 0 ) {
			SDM_HANDLE_SET_FP_SLOT(nfxMySlotId, handle);
			SDM_HANDLE_SET_SLOT(slot,handle);
			SDM_HANDLE_SET_PORT(port,handle);
			SDM_HANDLE_SET_INDEX(i, handle);
			h_entry->id    = id;
			h_entry->path  = path;
			h_entry->state = SDMA_VALID_ENTRY;
			break;
		}
	}
#if 0
	printf("Handle: %x\n", handle);
#endif
	return(handle);
} // sdma_allocDevHandle

//------------------------------------------------------------------------------
// Function Name:   sdma_printBuf
// 
// Description:     
//
// Data Created:    May 09, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdma_printBuf(char *buf, int len)
{
    int i, j;
    char ch;
    char *data;
 
    data = buf;
    for(i=0; i < len; i+=16) {
        printf("\n%04x:\t", i);
        for( j=i; j < (i+16); j++ )
            printf("%02x ", data[j] & 0xff);
        printf("    ");
        for( j=i; j < (i+16); j++ ) {
            ch = data[j] & 0xff;
            if( (ch > 0x20) && (ch < 0x7F))
                printf("%c", ch);
            else
            if( ch == 0 )
                printf("%c",'.');
            else
                printf("%c",'_');
        }
    }
    printf("\n\n");
    return 0;
} // sdma_printBuf
