/*------------------------------------------------------------------------------
 * RCS:  $Id: spm-pool.c,v 1.18 2002/11/07 00:13:34 nitinj Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : spm-pool.c
 *
 * Description  : This file contains the methods used to manipulate
 *                the storage pool databases. There is a single
 *                instance of these on each SSC.
 *
 * Created by   : Greg Seibert (in EVM)
 * Date Created : 02-JUL-2001
 *
 * Modified by  : Isabelle Cnudde
 * Date modified: 15-FEB-2002
 *
 * Modified by  : Nitin Jain
 * Date modified: 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "spm.h"

/*
 * Global variables.
 */

#define POOL_ID_SEQ_NAME "spoolseq"
static int pool_seq;

/*
 * poolDb uses the poolId as key, record is spm_poolDbCb_t
 */
DB *poolDb;        // Storage pool database
uint32 poolEntries;       // Counter to tell us how many pool entries

typedef struct
{
    int32  id;
    uint64 sid;
} spm_poolScsiList_t;

/*
 * Local function prototypes.
 */

static int32 next_sPool_id(void);

static uint64 spm_changeLunState(scsi_id_t sid, DB *fromDb, DB *toDb);

/*
 * Local functions.
 */

/*
 *------------------------------------------------------------------------------
 * Name         : next_sPool_id
 *
 * Description  : Get the next Storage Pool id in sequence
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 26-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

static int32 next_sPool_id(void)
{
    pool_seq++;

    return (pool_seq - 1);
}

/*
 * Global functions.
 */

/*
 *------------------------------------------------------------------------------
 * Name         : init_poolDb
 *
 * Description  : Open/Initialize the Storage Pool configuration database
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 02-MAY-2001
 *
 *------------------------------------------------------------------------------
 */

int32 init_poolDb(void)
{
    pool_seq = 1;  // Initialize pool seq. always start from 1

    poolDb  = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
    if (!poolDb)
    {
        SPM_ERR_PRINT(("Error opening local Storage Pool %s", strerror(errno)));
        return NFX_ERR;
    }
    poolEntries = 0;

    return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 * Name: clean_poolDb
 *
 * Description: Clean poolDb database
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 04-APR-2002
 *
 *------------------------------------------------------------------------------
 */

void clean_poolDb(void)
{
    DBT key;
    DBT rec;
    spm_poolDbCb_t *pCfg = NULL;
    int32  stat;

    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg));
        if (pCfg && pCfg->magic == SPM_MAGIC_STRING)
        {
            if (pCfg->usedDb)
                (pCfg->usedDb)->close(pCfg->usedDb);
            if (pCfg->freeDb)
                (pCfg->freeDb)->close(pCfg->freeDb);
            if (pCfg->foreDb)
                (pCfg->foreDb)->close(pCfg->foreDb);
            if (pCfg->outcDb)
                (pCfg->outcDb)->close(pCfg->outcDb);
            SPM_FREE(pCfg);
        }
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }
    poolDb->close(poolDb);
    poolDb = NULL;

    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_moveLunState
 *
 * Description  : Move LUN state
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_moveLunState(int32 poolId, scsi_id_t sid, spm_lunState_t fromState,
                       spm_lunState_t toState)
{
    DBT  key;
    DBT  rec;
    spm_poolDbCb_t *pCfg;
    int32 done = 0;
    uint64 size = 0;
    spm_lunDbCb_t *lCfg = NULL;
    
    SPM_FCT_PRINT(("spm_moveLunState..."));
    SPM_DBG_PRINT(("\t   Pool-Id : %d", poolId));
    SPM_DBG_PRINT(("\t   SCSI-Id : 0x%llx", sid));
    SPM_DBG_PRINT(("\tFrom State : %d", fromState));
    SPM_DBG_PRINT(("\t  To State : %d", toState));

    if (fromState == toState)
        return 1; // We are already in same state. SO say done!!!
    key.data = &poolId;
    key.size = sizeof(int32);
    if (!(poolDb->get(poolDb, &key, &rec, 0)))
    {
        DB  *fromDb = NULL;
        DB  *toDb = NULL;

        memcpy(&pCfg, rec.data, sizeof(pCfg));
        if (!pCfg)
            return done;
        if (pCfg->magic != SPM_MAGIC_STRING)
            return done;

        switch (fromState)
        {
            case SPM_LUN_FREE:
                if (pCfg->freeDb)
                    fromDb = pCfg->freeDb;
                break;
            case SPM_LUN_USED:
                if (pCfg->usedDb)
                    fromDb = pCfg->usedDb;
                break;
            case SPM_LUN_FOREIGN:
                if (pCfg->foreDb)
                    fromDb = pCfg->foreDb;
                break;
            case SPM_LUN_OC:
                if (pCfg->outcDb)
                    fromDb = pCfg->outcDb;
                break;
        }
        switch (toState)
        {
            case SPM_LUN_FREE:
                if (!pCfg->freeLun)
                {
                    pCfg->freeDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                           0666, DB_BTREE, NULL);
                    if (!pCfg->freeDb)
                    {
                        SPM_DBG_PRINT(("spm_moveLunState : Memory failure"));
                        return done;
                    }
                }
                toDb = pCfg->freeDb;
                break;
            case SPM_LUN_USED:
                if (!pCfg->usedLun)
                {
                    pCfg->usedDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                           0666, DB_BTREE, NULL);
                    if (!pCfg->usedDb)
                    {
                        SPM_DBG_PRINT(("spm_moveLunState : Memory failure"));
                        return done;
                    }
                }
                toDb = pCfg->usedDb;
                break;
            case SPM_LUN_FOREIGN:
                if (!pCfg->foreLun)
                {
                    pCfg->foreDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                           0666, DB_BTREE, NULL);
                    if (!pCfg->foreDb)
                    {
                        SPM_DBG_PRINT(("spm_moveLunState : Memory failure"));
                        return done;
                    }
                }
                toDb = pCfg->foreDb;
                break;
            case SPM_LUN_OC:
                if (!pCfg->outcLun)
                {
                    pCfg->outcDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                    0666, DB_BTREE, NULL);
                    if (!pCfg->outcDb)
                    {
                        SPM_DBG_PRINT(("spm_moveLunState : Memory failure"));
                        return done;
                    }
                }
                toDb = pCfg->outcDb;
                break;
        }
        size = spm_changeLunState(sid, fromDb, toDb);
        if (size)
        {
            done = 1;
            switch (toState)
            {
                case SPM_LUN_FREE:
                    pCfg->freeSize += size;
                    pCfg->freeLun++;
                    break;
                case SPM_LUN_USED:
                    pCfg->usedSize += size;
                    pCfg->usedLun++;
                    break;
                case SPM_LUN_FOREIGN:
                    pCfg->foreSize += size;
                    pCfg->foreLun++;
                    break;
                case SPM_LUN_OC:
                    pCfg->outcSize += size;
                    pCfg->outcLun++;
                    break;
            }
            switch (fromState)
            {
                case SPM_LUN_FREE :
                    pCfg->freeSize -= size;
                    pCfg->freeLun--;
                    if (!pCfg->freeLun)
                    {
                        (pCfg->freeDb)->close(pCfg->freeDb);
                        pCfg->freeDb = NULL;
                    }
                    break;
                case SPM_LUN_USED :
                    pCfg->usedSize -= size;
                    pCfg->usedLun--;
                    if (!pCfg->usedLun)
                    {
                        (pCfg->usedDb)->close(pCfg->usedDb);
                        pCfg->usedDb = NULL;
                    }
                    break;
                case SPM_LUN_FOREIGN :
                    pCfg->foreSize -= size;
                    pCfg->foreLun--;
                    if (!pCfg->foreLun)
                    {
                        (pCfg->foreDb)->close(pCfg->foreDb);
                        pCfg->foreDb = NULL;
                    }
                    break;
                case SPM_LUN_OC :
                    pCfg->outcSize -= size;
                    pCfg->outcLun--;
                    if (!pCfg->outcLun)
                    {
                        (pCfg->outcDb)->close(pCfg->outcDb);
                        pCfg->outcDb = NULL;
                    }
                    break;
            }
            spm_getScsiDb_rec(sid, &lCfg);
            if (lCfg)
                spm_sendLunChangeStateInd(lCfg->dev, fromState, toState);
        }
    }

    SPM_FCT_PRINT(("...spm_moveLunState"));
    return done;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_changeLunState
 *
 * Description  : Change LUN state
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static uint64 spm_changeLunState(scsi_id_t sid, DB *fromDb, DB *toDb)
{
    DBT  sKey;
    DBT  sRec;
    spm_lunNodeInPool_t lnCfg;

    uint64 size = 0;

    SPM_FCT_PRINT(("spm_changeLunState..."));
    SPM_DBG_PRINT(("\tscsi-id to change state : 0x%llx", sid));

    if (fromDb == toDb)
    {
        SPM_DBG_PRINT(("\tFrom and To state are same"));
        return 0;
    }

    sKey.data = &sid;
    sKey.size = sizeof(scsi_id_t);
    if (fromDb && toDb && !fromDb->get(fromDb, &sKey, &sRec, 0))
    {
        memcpy(&lnCfg, (spm_lunNodeInPool_t *)sRec.data,
                              sizeof(spm_lunNodeInPool_t));
        fromDb->del(fromDb, &sKey, 0);
        sRec.data = &lnCfg;
        sRec.size = sizeof(spm_lunNodeInPool_t);
        toDb->put(toDb, &sKey, &sRec, R_NOOVERWRITE);
        size = lnCfg.size;
        // Send indication to SRM
    }

    SPM_FCT_PRINT(("...spm_changeLunState"));
    return size;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getLunState
 *
 * Description  : Get LUN state
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getLunState(int32 poolId, scsi_id_t sid)
{
    DBT  key;
    DBT  rec;
    spm_poolDbCb_t      *pCfg;

    SPM_FCT_PRINT(("spm_getLunState..."));
    SPM_DBG_PRINT(("\tPool-ID : %d", poolId));
    SPM_DBG_PRINT(("\tSCSI-ID : 0x%llx", sid));
    key.data = &poolId;
    key.size = sizeof(int32);
    if (poolDb->get(poolDb, &key, &rec, 0) == 0)
    {
        DBT  skey;

        memcpy(&pCfg, rec.data, sizeof(pCfg));   // Copy pointer
        if (!pCfg)
            return 0;
        if (pCfg->magic != SPM_MAGIC_STRING)
            return 0;
        skey.data = &sid;
        skey.size = sizeof(sid);
        if (pCfg->freeLun)
            if (!((pCfg->freeDb)->get((pCfg->freeDb), &skey, &rec, 0)))
                return SPM_LUN_FREE;
        if (pCfg->usedLun)
            if (!((pCfg->usedDb)->get((pCfg->usedDb), &skey, &rec, 0)))
                return SPM_LUN_USED;
        if (pCfg->foreLun)
            if (!((pCfg->foreDb)->get((pCfg->foreDb), &skey, &rec, 0)))
                return SPM_LUN_FOREIGN;
        if (pCfg->outcLun)
            if (!((pCfg->outcDb)->get((pCfg->outcDb), &skey, &rec, 0)))
                return SPM_LUN_OC;
    }

    SPM_FCT_PRINT(("...spm_getLunState"));
    return 0;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_pool_id_list
 *
 * Description  : Give me a list of all the storage pool ids known
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 26-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_pool_id_list(spm_poolNode_t **ids, char *err)
{
    int32 num = 0, stat;
    DBT key;
    DBT rec;
    spm_poolNode_t lclIdList[50];
    spm_poolDbCb_t *pCfg;
    int32   idCnt;
    spm_poolNode_t *new;

    /*
     * Sequentially go through the pool config db recording the pool ids
     */

    SPM_FCT_PRINT(("spm_get_pool_id_list..."));
    *ids = NULL;
    idCnt = 0;
    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg));  // Copy the pointer
        if (!pCfg)
        {
            if (*ids)
            {
                SPM_FREE(*ids);
                *ids = NULL;
            }
            return -1;
        }
        if (pCfg->magic != SPM_MAGIC_STRING)
        {
            if (*ids)
            {
                SPM_FREE(*ids);
                *ids = NULL;
            }
            return -1;
        }

        if (num == 50)
        {
            SPM_ALLOC_ZERO(new, spm_poolNode_t,
                                (sizeof(spm_poolNode_t)*(idCnt+num)));
            if (*ids != NULL)
            {
                memcpy(new, *ids, sizeof(spm_poolNode_t)*idCnt);
                memcpy(new+idCnt, lclIdList, sizeof(spm_poolNode_t)*num);
                SPM_FREE(*ids);
            }
            else
                memcpy(new, lclIdList, sizeof(spm_poolNode_t)*num);
            *ids = new;
            idCnt += num;
            num = 0;
            memset (lclIdList, 0, sizeof(spm_poolNode_t)*50);
        }

        SPM_DBG_PRINT(("Pool ID : %d", pCfg->id));
        strcpy(lclIdList[num].ctrl, pCfg->attr.ctrl);
        strcpy(lclIdList[num].make, pCfg->attr.name);
        strcpy(lclIdList[num].model, pCfg->attr.model);
        lclIdList[num].rL          = pCfg->attr.raidLevel;
        lclIdList[num].usedSize    = pCfg->usedSize;
        lclIdList[num].freeSize    = pCfg->freeSize;
        lclIdList[num].ocSize      = pCfg->outcSize;
        lclIdList[num].foreignSize = pCfg->foreSize;
        lclIdList[num].type        = pCfg->attr.type;
        num++;
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }
    
    /*
     * If there was an error or we reached max before coming to the end,
     * it is a problem.
     */
    if (stat < 0)
    {
        if (err)
        {
            sprintf (err, "Error reading pool database %s", strerror(errno));
            SPM_DBG_PRINT(("%s", err));
            idCnt = -1;
            if (*ids)
            {
                SPM_FREE(*ids);
                *ids = NULL;
            }
        }
    }
    else if (stat && (num != 0))
    {
        /*
         * No more records in the DB.
         */
        if (idCnt == 0)
        {
            SPM_ALLOC_ZERO(new, spm_poolNode_t, (sizeof(spm_poolNode_t)*num));
            memcpy(new, lclIdList, sizeof(spm_poolNode_t)*num);
            *ids = new;
            idCnt += num;
        }
        else if (num != 0)
        {
            SPM_ALLOC_ZERO(new, spm_poolNode_t,
                                (sizeof(spm_poolNode_t)*(idCnt+num)));
            memcpy(new, *ids, sizeof(spm_poolNode_t)*idCnt);
            memcpy(new+idCnt, lclIdList, sizeof(spm_poolNode_t)*num);
            SPM_FREE(*ids);
            *ids = new;
            idCnt += num;
        }
    }

    SPM_FCT_PRINT(("...spm_get_pool_id_list"));
    return idCnt;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_pool_list_debug
 *
 * Description  : Give me a list of all the storage pool known
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_pool_list_debug(spm_storageAttrib_t **poolList, char *err)
{
    spm_storageAttrib_t  *new, strgList[50];
    spm_poolDbCb_t       *pCfg;
    DBT                  key;
    DBT                  rec;
    int32                stat;
    int32                num = 0, numRec = 0;

    SPM_FCT_PRINT(("spm_get_pool_list_debug..."));
    bzero(strgList, sizeof(spm_storageAttrib_t)*50);
    *poolList = NULL;
    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
        if (!pCfg)
        {
            if (*poolList)
            {
                SPM_FREE(*poolList);
                *poolList = NULL;
            }
            return -1;
        }
        if (pCfg->magic != SPM_MAGIC_STRING)
        {
            if (*poolList)
            {
                SPM_FREE(*poolList);
                *poolList = NULL;
            }
            return -1;
        }

        strgList[num].poolId    = pCfg->id;
        strgList[num].raidLevel = pCfg->attr.raidLevel;
        strgList[num].usedSize  = pCfg->usedSize;
        strgList[num].usedLun   = pCfg->usedLun;
        strgList[num].freeSize  = pCfg->freeSize;
        strgList[num].freeLun   = pCfg->freeLun;
        strgList[num].foreSize  = pCfg->foreSize;
        strgList[num].foreLun   = pCfg->foreLun;
        strgList[num].outcSize  = pCfg->outcSize;
        strgList[num].outcLun   = pCfg->outcLun;
        strcpy(strgList[num].name, pCfg->attr.name);
        strcpy(strgList[num].model, pCfg->attr.model);
        strcpy(strgList[num].ctrl, pCfg->attr.ctrl);
        num++;
        if (num == 50)
        {
            if (*poolList)
            {
                SPM_ALLOC_ZERO(new, spm_storageAttrib_t,
                                 sizeof(spm_storageAttrib_t)*(numRec+num));
                memcpy(new, *poolList, sizeof(spm_storageAttrib_t)*numRec);
                memcpy((new+numRec), strgList, sizeof(spm_storageAttrib_t)*num);
                SPM_FREE(*poolList);
            }
            else
            {
                SPM_ALLOC_ZERO(new, spm_storageAttrib_t,
                                 sizeof(spm_storageAttrib_t)*num);
                memcpy(new, strgList, sizeof(spm_storageAttrib_t)*num);
            }
            numRec += num;
            bzero(strgList, sizeof(spm_storageAttrib_t)*num);
            num = 0;
            *poolList = new;
        }
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }
    if (stat < 0)
    {
        if (err)
        {
            sprintf (err, "Error reading pool database %s", strerror(errno));
            SPM_DBG_PRINT(("%s", err));
            numRec = -1;
            if (*poolList)
            {
                SPM_FREE(*poolList);
                *poolList = NULL;
            }
        }
    }
    else if (stat && (num != 0))
    {
        if (*poolList)
        {
            SPM_ALLOC_ZERO(new, spm_storageAttrib_t,
                                sizeof(spm_storageAttrib_t)*(numRec+num));
            memcpy(new, *poolList, sizeof(spm_storageAttrib_t)*numRec);
            memcpy((new+numRec), strgList, sizeof(spm_storageAttrib_t)*num);
            SPM_FREE(*poolList);
        }
        else
        {
            SPM_ALLOC_ZERO(new, spm_storageAttrib_t,
                                sizeof(spm_storageAttrib_t)*num);
            memcpy(new, strgList, sizeof(spm_storageAttrib_t)*num);
        }
        numRec += num;
        *poolList = new;
    }

    SPM_FCT_PRINT(("...spm_get_pool_list_debug"));
    return numRec;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_num_ctrls
 *
 * Description  : Give number of controllers in the cluster
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 11-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_num_ctrls(char *err)
{
    char                 *new;
    spm_poolDbCb_t       *pCfg;
    char                 cList[50][SDM_DEV_NAME_LEN], *cL = NULL;
    DBT                  key;
    DBT                  rec;
    int32                stat;
    int32                num = 0, numRec = 0, dup, i;

    SPM_FCT_PRINT(("spm_get_num_ctrls..."));
    bzero(cList, SDM_DEV_NAME_LEN*50);
    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
        if (!pCfg)
        {
            if (cL)
            {
                SPM_FREE(cL);
                cL = NULL;
            }
            return -1;
        }
        if (pCfg->magic != SPM_MAGIC_STRING)
        {
            if (cL)
            {
                SPM_FREE(cL);
                cL = NULL;
            }
            return -1;
        }

        for (i = 0, dup = 0; !dup && i < (num + numRec); i++)
        {
            if (i < numRec)
                dup = !(strcmp((cL+i), pCfg->attr.ctrl));
            else
                dup = !(strcmp(cList[i-numRec], pCfg->attr.ctrl));
        }
        if (!dup) strcpy(cList[num++], pCfg->attr.ctrl);

        if (num == 50)
        {
            if (cL)
            {
                SPM_ALLOC_ZERO(new, char, SDM_DEV_NAME_LEN*(numRec+num));
                memcpy(new, cL, SDM_DEV_NAME_LEN*numRec);
                memcpy((new+numRec), cList, SDM_DEV_NAME_LEN*num);
                SPM_FREE(cL);
            }
            else
            {
                SPM_ALLOC_ZERO(new, char, SDM_DEV_NAME_LEN*num);
                memcpy(new, cList, SDM_DEV_NAME_LEN*num);
            }
            numRec += num;
            bzero(cList, SDM_DEV_NAME_LEN*num);
            num = 0;
            cL = new;
        }
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }
    SPM_FREE(cL);
    numRec += num;

    SPM_FCT_PRINT(("...spm_get_num_ctrls"));
    return numRec;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getCtrlList
 *
 * Description  : Give list of controllers in the cluster
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 11-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getCtrlList(char **ctrlList, char *lastCtrl, int32 *offset, char *err)
{
    char                 *new;
    spm_poolDbCb_t       *pCfg;
    char                 cList[50][SDM_DEV_NAME_LEN], *cL = NULL;
    DBT                  key;
    DBT                  rec;
    int32                stat, found;
    int32                num = 0, numRec = 0, dup, i;

    SPM_FCT_PRINT(("spm_getCtrlList..."));
    bzero(cList, SDM_DEV_NAME_LEN*50);
    *ctrlList = NULL;
    *offset = 0;
    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
        if (!pCfg)
        {
            if (cL)
            {
                SPM_FREE(cL);
                cL = NULL;
            }
            return -1;
        }
        if (pCfg->magic != SPM_MAGIC_STRING)
        {
            if (cL)
            {
                SPM_FREE(cL);
                cL = NULL;
            }
            return -1;
        }

        for (i = 0, dup = 0; !dup && i < (num + numRec); i++)
        {
            if (i < numRec)
                dup = !(strcmp((cL+i), pCfg->attr.ctrl));
            else
                dup = !(strcmp(cList[i-numRec], pCfg->attr.ctrl));
        }
        if (!dup) strcpy(cList[num++], pCfg->attr.ctrl);

        if (num == 50)
        {
            if (cL)
            {
                SPM_ALLOC_ZERO(new, char, SDM_DEV_NAME_LEN*(numRec+num));
                memcpy(new, cL, SDM_DEV_NAME_LEN*numRec);
                memcpy((new+numRec), cList, SDM_DEV_NAME_LEN*num);
                SPM_FREE(cL);
            }
            else
            {
                SPM_ALLOC_ZERO(new, char, SDM_DEV_NAME_LEN*num);
                memcpy(new, cList, SDM_DEV_NAME_LEN*num);
            }
            numRec += num;
            bzero(cList, SDM_DEV_NAME_LEN*num);
            num = 0;
            cL = new;
        }
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }

    if (stat && (num != 0))
    {
        if (cL)
        {
            SPM_ALLOC_ZERO(new, char, SDM_DEV_NAME_LEN*(numRec+num));
            memcpy(new, cL, SDM_DEV_NAME_LEN*numRec);
            memcpy((new+numRec), cList, SDM_DEV_NAME_LEN*num);
            SPM_FREE(cL);
        }
        else
        {
            SPM_ALLOC_ZERO(new, char, SDM_DEV_NAME_LEN*num);
            memcpy(new, cList, SDM_DEV_NAME_LEN*num);
        }
        numRec += num;
        cL = new;
    }

    found = 0;
    *ctrlList = cL;
    for (i = 0; i < numRec; i++)
    {
        SPM_DBG_PRINT(("Ctrl : %s", (cL+(i*SDM_DEV_NAME_LEN))));
        if (!strcasecmp((cL+(i*SDM_DEV_NAME_LEN)), lastCtrl))
        {
            found = 1;
            continue;
        }
        if (found)
        {
            *offset = i;
            // numRec -= i;
            break;
        }
    }

    SPM_DBG_PRINT(("Offset : %d", *offset));
    SPM_DBG_PRINT(("numReq : %d", numRec));
    /* Calling routine will free cL */
    SPM_FCT_PRINT(("...spm_getCtrlList"));
    return numRec;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_ctrl_node_list
 *
 * Description  : Give list of nodes for the controller
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 11-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_ctrl_node_list(char **nodeList, char *ctrl, spm_dev_t *type,
                             char *err)
{
    spm_poolDbCb_t       *pCfg;
    DBT                  key;
    DBT                  rec;
    int32                stat, cnt, ndCnt;
    spm_nodeList_t       *ndList;
    spm_lunsInPool_t     *lunList;
    int32                _ndCnt, i, j;

    SPM_FCT_PRINT(("spm_get_ctrl_node_list..."));
    *nodeList = NULL;

    spm_getNfxDbNodeList(&ndList);
    if (!ndList) return -1;
    if (ndList->cnt == 0)
    {
        SPM_FREE(ndList);
        return 0;
    }
    ndCnt = 0;
    _ndCnt = ndList->cnt;
    SPM_ALLOC_ZERO(*nodeList, char , SPM_MAX_NODE_NAME_LEN*_ndCnt);

    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
        if (!pCfg)
        {
            SPM_FREE(*nodeList);
            *nodeList = NULL;
            return -1;
        }
        if (pCfg->magic != SPM_MAGIC_STRING)
        {
            SPM_FREE(*nodeList);
            *nodeList = NULL;
            return -1;
        }

        if (strcmp(ctrl, pCfg->attr.ctrl))
        {
            stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
            continue;
        }
        *type = pCfg->attr.type;

        if ((cnt = spm_get_lun_list_debug(&lunList, SPM_LUN_ALL_LIST,
                                          pCfg->id, err)) > 0)
        {
            for (i = 0; i < cnt && ndCnt != _ndCnt; i++)
            {
                for (j = 0; j < _ndCnt; j++)
                {
                    if (strlen(ndList->node[j]) &&
                        spm_checkLunInNfxDb(ndList->node[j],
                                            lunList[i].scsiId))
                    {
                        memcpy((*nodeList+ndCnt), ndList->node[j],
                                                  SPM_MAX_NODE_NAME_LEN);
                        bzero(ndList->node[j], SPM_MAX_NODE_NAME_LEN);
                        ndCnt++;
                    }
                }
            }

            SPM_FREE(lunList);
        }
        if (_ndCnt == ndCnt) break;
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }

    SPM_FREE(ndList);
    SPM_FCT_PRINT(("...spm_get_ctrl_node_list"));

    return ndCnt;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_pool_list_debug
 *
 * Description  : Give me a list of all the storage pool known
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_lun_list_debug(spm_lunsInPool_t **lunList, spm_lunType_t type,
                             int32 poolId, char *rtnstr)
{
    int32          num = 0;
    spm_poolDbCb_t *pCfg;
    int32          stat;
    DBT            key;
    DBT            rec;
    uint32         totLun;

    SPM_FCT_PRINT(("spm_get_lun_list_debug..."));
    SPM_DBG_PRINT(("\tPool-ID : %d", poolId));
    SPM_DBG_PRINT(("\t   Type : %d", type));
    *lunList = NULL;
    key.data = &poolId;
    key.size = sizeof(int32); 
    stat = poolDb->get(poolDb, &key, &rec, 0);
    if (!stat)
    {
        SPM_DBG_PRINT(("Found Pool"));
        memcpy(&pCfg, rec.data, sizeof(pCfg));  // Copy pointer only
        if (!pCfg)
            return num;
        if (pCfg->magic != SPM_MAGIC_STRING)
            return num;

        totLun = pCfg->usedLun + pCfg->freeLun +
                 pCfg->foreLun + pCfg->outcLun;
        SPM_ALLOC_ZERO(*lunList, spm_lunsInPool_t,
                                 sizeof(spm_lunsInPool_t)*totLun);

        if (pCfg->usedLun &&
            (type == SPM_LUN_ALL_LIST || type == SPM_LUN_USED_LIST))
        {
            DBT  lkey;
            DBT  lrec;
            int32 rc;
            spm_lunNodeInPool_t  lCfg;

            rc = (pCfg->usedDb)->seq((pCfg->usedDb), &lkey, &lrec, R_FIRST);
            while (!rc)
            {
                memcpy(&lCfg, lrec.data, sizeof(lCfg));
                (*lunList+num)->scsiId    = lCfg.sid;
                (*lunList+num)->nfxRefCnt = lCfg.nfxRefCnt;
                (*lunList+num)->lunSize   = lCfg.size;
                (*lunList+num)->lunState  = SPM_LUN_USED;
                num++;
                rc = (pCfg->usedDb)->seq((pCfg->usedDb), &lkey, &lrec, R_NEXT);
            }
        }
        if (pCfg->freeLun &&
            (type == SPM_LUN_ALL_LIST || type == SPM_LUN_FREE_LIST))
        {
            DBT  lkey;
            DBT  lrec;
            int32 rc;
            spm_lunNodeInPool_t  lCfg;

            rc = (pCfg->freeDb)->seq((pCfg->freeDb), &lkey, &lrec, R_FIRST);
            while (!rc)
            {
                memcpy(&lCfg, lrec.data, sizeof(lCfg));
                (*lunList+num)->scsiId    = lCfg.sid;
                (*lunList+num)->nfxRefCnt = lCfg.nfxRefCnt;
                (*lunList+num)->lunSize   = lCfg.size;
                (*lunList+num)->lunState  = SPM_LUN_FREE;
                num++;
                rc = (pCfg->freeDb)->seq((pCfg->freeDb), &lkey, &lrec, R_NEXT);
            }
        }
        if (pCfg->foreLun &&
            (type == SPM_LUN_ALL_LIST || type == SPM_LUN_FOREIGN_LIST))
        {
            DBT  lkey;
            DBT  lrec;
            int32 rc;
            spm_lunNodeInPool_t  lCfg;

            rc = (pCfg->foreDb)->seq((pCfg->foreDb), &lkey, &lrec, R_FIRST);
            while (!rc)
            {
                memcpy(&lCfg, lrec.data, sizeof(lCfg));
                (*lunList+num)->scsiId    = lCfg.sid;
                (*lunList+num)->nfxRefCnt = lCfg.nfxRefCnt;
                (*lunList+num)->lunSize   = lCfg.size;
                (*lunList+num)->lunState  = SPM_LUN_FOREIGN;
                num++;
                rc = (pCfg->foreDb)->seq((pCfg->foreDb), &lkey, &lrec, R_NEXT);
            }
        }
        if (pCfg->outcLun &&
            (type == SPM_LUN_ALL_LIST || type == SPM_LUN_OUTCLUSTER_LIST))
        {
            DBT  lkey;
            DBT  lrec;
            int32 rc;
            spm_lunNodeInPool_t  lCfg;

            rc = (pCfg->outcDb)->seq((pCfg->outcDb), &lkey, &lrec, R_FIRST);
            while (!rc)
            {
                memcpy(&lCfg, lrec.data, sizeof(lCfg));
                (*lunList+num)->scsiId    = lCfg.sid;
                (*lunList+num)->nfxRefCnt = lCfg.nfxRefCnt;
                (*lunList+num)->lunSize   = lCfg.size;
                (*lunList+num)->lunState  = SPM_LUN_OC;
                num++;
                rc = (pCfg->outcDb)->seq((pCfg->outcDb), &lkey, &lrec, R_NEXT);
            }
        }

    }
    return num;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_compare_pool_attr
 *
 * Description  : Compare pool attributes
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_compare_pool_attr(spm_poolAttr_t *poolAttr,
                            spm_poolDbCb_t *poolCfg)
{
    SPM_FCT_PRINT(("spm_compare_pool_attr..."));

    if ((poolAttr->mask & SPM_POOL_ATTR_TYPE_MASK) &&
        (poolAttr->type != poolCfg->attr.type))
        // not correct type
        return 0;
    if ((poolAttr->mask & SPM_POOL_ATTR_RAID_MASK) &&
        (poolAttr->raidLevel != poolCfg->attr.raidLevel))
        // not the correct raid level
        return 0;
    if ((poolAttr->mask & SPM_POOL_ATTR_CTRL_MASK) &&
        strlen(poolAttr->ctrl) &&
        strncasecmp(poolAttr->ctrl, poolCfg->attr.ctrl,
                                     strlen(poolAttr->ctrl)) != 0)
        // not the correct controller
        return 0;
    if ((poolAttr->mask & SPM_POOL_ATTR_MAKE_MASK) &&
        strlen(poolAttr->name) &&
        strncasecmp(poolAttr->name, poolCfg->attr.name,
                                    strlen(poolAttr->name)) != 0)
        // not the correct vendor
        return 0;
    if ((poolAttr->mask & SPM_POOL_ATTR_MODEL_MASK) &&
        strlen(poolAttr->model) &&
        strncasecmp(poolAttr->model, poolCfg->attr.model,
                                     strlen(poolAttr->model)) != 0)
        // not the correct model
        return 0;

    SPM_FCT_PRINT(("...spm_compare_pool_attr"));
    return 1; // Match
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_pool_by_attrs
 *
 * Description  : check if pool exists with given LUN attributes
 *
 * Created by   : Lan Chin
 *
 * Date Created : 07-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_pool_by_attrs(
    spm_poolAttr_t *attr,     // need to match these attributes
    int32          lunsAvail, // Return only if LUNs available
    int32          startId,   // Get the next matching pool from this ID
    char           *err)
{
    int32           stat;
    DBT             key;
    DBT             rec;
    spm_poolDbCb_t  *pCfg;
    int32           skipPool = 1;
    int32           poolId = -1;

    SPM_FCT_PRINT(("spm_get_pool_by_attrs..."));
    if (!attr)
        return poolId;
    SPM_DBG_PRINT(("Match following attrib"));
    SPM_DBG_PRINT(("\traid level : %d", attr->raidLevel));
    SPM_DBG_PRINT(("\t      CTRL : %s", attr->ctrl));
    SPM_DBG_PRINT(("\t    Vendor : %s", attr->name));
    SPM_DBG_PRINT(("\t     Model : %s", attr->model));
    SPM_DBG_PRINT(("\t     Type  : %d", attr->type));

    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg));
        if (!pCfg)
            return poolId;
        if (pCfg->magic != SPM_MAGIC_STRING)
            return poolId;

        SPM_DBG_PRINT(("\tpool [%d] -> raid level = %d\tCtrl = %s",
                       pCfg->id, pCfg->attr.raidLevel, pCfg->attr.ctrl));
        SPM_DBG_PRINT(("\t\t\tvendor %s %s",
                      (strlen(pCfg->attr.name)?pCfg->attr.name:"no vendor"),
                      (strlen(pCfg->attr.model)?pCfg->attr.model:"no model")));
        SPM_DBG_PRINT(("\t\t\ttype %d", pCfg->attr.type));

        if (skipPool && startId > 0)
        {
            if (pCfg->id == startId)
                skipPool = 0;
            stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
            continue;
        }

        if (!spm_compare_pool_attr(attr, pCfg))
        {
            stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
            continue;
        }
        /*
         * check if this storage pool ran out of LUNs
         * if so, continue searching for another pool with same attributes
         */
        if (lunsAvail && !pCfg->freeLun)
        {
            stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
            continue;
        }
        /*
         * pass all tests, this pool will do
         * return poolId
         */
        poolId = pCfg->id;
        SPM_DBG_PRINT(("  found match with id %d", pCfg->id));
        return poolId;
    } /* while */

    SPM_DBG_PRINT(("No matching pool"));
    strcpy(err, "No pools found matching given attr");

    SPM_FCT_PRINT(("...spm_get_pool_by_attrs"));
    return poolId;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_new_pool
 *
 * Description  : Create a new storage pool with these attributes
 *
 * Created by   : Lan Chin
 *
 *  New feature to auto create storage pools based on attributes:
 *  (attributes available from SDM)
 *      1. lun size
 *      2. raid level
 *      3. vendor name
 *      4. model name
 *
 * Date Created : 08-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

int32 spm_new_pool(spm_poolAttr_t *attr, // create using these attributes
                   int32 *id,            // created pool's id
                   char *err)
{
    DBT                key;
    DBT                rec;
    int32              stat;
    spm_poolDbCb_t     *pCfg;

    /*
     * create the new storage pool
     */
    *id = -1;
    SPM_FCT_PRINT(("spm_new_pool..."));
    if (!attr)
        return NFX_ERR;
    SPM_DBG_PRINT(("Create pool with attrib"));
    SPM_DBG_PRINT(("\traid level : %d", attr->raidLevel));
    SPM_DBG_PRINT(("\t      CTRL : %s", attr->ctrl));
    SPM_DBG_PRINT(("\t    Vendor : %s", attr->name));
    SPM_DBG_PRINT(("\t     Model : %s", attr->model));
    *id = next_sPool_id();

    SPM_ALLOC_ZERO(pCfg, spm_poolDbCb_t, sizeof(spm_poolDbCb_t));

    pCfg->magic = SPM_MAGIC_STRING;
    pCfg->id = *id;
    memcpy(&pCfg->attr, attr, sizeof(spm_poolAttr_t));

    key.data = id;
    key.size = sizeof(int32);
    rec.data = &pCfg;
    rec.size = sizeof(pCfg);  // Store pointer
    stat = poolDb->put(poolDb, &key, &rec, 0);
    if (stat < 0)
    {
        sprintf(err, "error writing pool db: %s", strerror(errno));
        return NFX_ERR;
    }
    SPM_DBG_PRINT(("New Pool ID - %d", *id));
    poolEntries++;

    SPM_FCT_PRINT(("...spm_new_pool"));
    return NFX_OK;
}


/*
 *------------------------------------------------------------------------------
 *  Name         : spm_del_pool
 *
 *  Description  : Delete pool
 *
 *  Created by   : Nitin Jain
 *
 *  Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_del_pool(int32 poolId)
{
    DBT  key;
    DBT  rec;
    spm_poolDbCb_t *pCfg = NULL;

    SPM_FCT_PRINT(("spm_del_pool..."));
    SPM_DBG_PRINT(("\tPool-Id : %d", poolId));

    key.data = &poolId;
    key.size = sizeof(int32);
    if (poolDb->get(poolDb, &key, &rec, 0) == 0)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg));
        if (!pCfg)
            return;
        if (pCfg->magic != SPM_MAGIC_STRING)
            return;

        if (pCfg->usedLun)
            (pCfg->usedDb)->close(pCfg->usedDb);
        if (pCfg->freeLun)
            (pCfg->freeDb)->close(pCfg->freeDb);
        if (pCfg->foreLun)
            (pCfg->foreDb)->close(pCfg->foreDb);
        if (pCfg->outcLun)
            (pCfg->outcDb)->close(pCfg->outcDb);
        poolDb->del(poolDb, &key, 0);
        poolEntries--;
        SPM_FREE(pCfg);
    }

    SPM_FCT_PRINT(("...spm_del_pool"));
    return;
}


/*
 *------------------------------------------------------------------------------
 *  Name         : spm_add_lunTo_poolDb
 *
 *  Description  : Add a new lun to the storage lun data base.
 *
 *  Created by   : Nitin Jain
 *
 *  Date Created : 21-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_add_lunTo_poolDb(uint32 poolId, scsi_id_t sid, uchar8 addLun,
                           uint64 lSize, spm_lunState_t state, char *err)
{
    DBT key;
    DBT rec;
    DBT sKey;
    DBT sRec;
    spm_poolDbCb_t      *pCfg;
    spm_lunNodeInPool_t lCfg;

    SPM_FCT_PRINT(("spm_add_lunTo_poolDb..."));
    SPM_DBG_PRINT(("\tPool ID : %d", poolId));
    SPM_DBG_PRINT(("\tSCSI-ID : 0x%llx", sid));
    SPM_DBG_PRINT(("\t   Size : %lld", lSize));
    SPM_DBG_PRINT(("\t  State : %d", state));
    key.data = &poolId;
    key.size = sizeof(int32);
    if (poolDb->get(poolDb, &key, &rec, 0))
    {
        sprintf(err, "Invalid spool-Id in request (%d)", poolId);
        return NFX_ERR;
    }

    memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy the pointer
    if (!pCfg)
        return NFX_ERR;
    if (pCfg->magic != SPM_MAGIC_STRING)
        return NFX_ERR;

    if (addLun) /* This is a new LUN. Added it */
    {
        DB  *tmpDb = NULL;

        SPM_DBG_PRINT(("Add new LUN in DB"));

        if (state == SPM_LUN_FREE)
        {
            if (!pCfg->freeLun)
            {
                pCfg->freeDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                0666, DB_BTREE, NULL);
                if (!pCfg->freeDb)
                {
                    SPM_DBG_PRINT(("Cannot add LUN. Memory failure"));
                    sprintf(err, "Cannot add LUN. Memory failure");
                    return NFX_ERR;
                }
            }
            tmpDb = pCfg->freeDb;
            pCfg->freeSize += lSize;
            pCfg->freeLun++;
        }
        else if (state == SPM_LUN_FOREIGN)
        {
            if (!pCfg->foreLun)
            {
                pCfg->foreDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                0666, DB_BTREE, NULL);
                if (!pCfg->foreDb)
                {
                    SPM_DBG_PRINT(("Cannot add LUN. Memory failure"));
                    sprintf(err, "Cannot add LUN. Memory failure");
                    return NFX_ERR;
                }
            }
            tmpDb = pCfg->foreDb;
            pCfg->foreSize += lSize;
            pCfg->foreLun++;
        }
        else if (state == SPM_LUN_OC)
        {
            if (!pCfg->outcLun)
            {
                pCfg->outcDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                0666, DB_BTREE, NULL);
                if (!pCfg->outcDb)
                {
                    SPM_DBG_PRINT(("Cannot add LUN. Memory failure"));
                    sprintf(err, "Cannot add LUN. Memory failure");
                    return NFX_ERR;
                }
            }
            tmpDb = pCfg->outcDb;
            pCfg->outcSize += lSize;
            pCfg->outcLun++;
        }
        else
        {
            if (!pCfg->usedLun)
            {
                pCfg->usedDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                0666, DB_BTREE, NULL);
                if (!pCfg->usedDb)
                {
                    SPM_DBG_PRINT(("Cannot add LUN. Memory failure"));
                    sprintf(err, "Cannot add LUN. Memory failure");
                    return NFX_ERR;
                }
            }
            tmpDb = pCfg->usedDb;
            pCfg->usedSize += lSize;
            pCfg->usedLun++;
        }
        bzero(&lCfg, sizeof(lCfg));
        lCfg.sid = sid;
        lCfg.nfxRefCnt = 1;
        lCfg.size = lSize;
        sKey.data = &sid;
        sKey.size = sizeof(sid);
        sRec.data = &lCfg;
        sRec.size = sizeof(lCfg);
        if (tmpDb->put(tmpDb, &sKey, &sRec, R_NOOVERWRITE))
        {
            SPM_DBG_PRINT(("Cannot add LUN."));
            sprintf(err, "Cannot add LUN.");
            return NFX_ERR;
        }
    }
    else /* We know the LUN. Just increment the ref Count */
    {
        DB  *tmpDb = NULL;
        int32 fnd = 0, currState;

        sKey.data = &sid;
        sKey.size = sizeof(sid);

        if (pCfg->usedDb && !fnd)
        {
            tmpDb = pCfg->usedDb;
            if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
            {
                SPM_DBG_PRINT(("Found in used Db"));
                memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data,
                       sizeof(spm_lunNodeInPool_t));
                fnd = 1;
                currState = SPM_LUN_USED;
            }
        }
        if (pCfg->freeLun && !fnd)
        {
            tmpDb = pCfg->freeDb;
            if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
            {
                SPM_DBG_PRINT(("Found in free Db"));
                memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data,
                       sizeof(spm_lunNodeInPool_t));
                fnd = 1;
                currState = SPM_LUN_FREE;
            }
        }
        if (pCfg->foreLun && !fnd)
        {
            tmpDb = pCfg->foreDb;
            if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
            {
                SPM_DBG_PRINT(("Found in foreign Db"));
                memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data,
                       sizeof(spm_lunNodeInPool_t));
                fnd = 1;
                currState = SPM_LUN_FOREIGN;
            }
        }
        if (pCfg->outcLun && !fnd)
        {
            tmpDb = pCfg->outcDb;
            if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
            {
                SPM_DBG_PRINT(("Found in out cluster Db"));
                memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data,
                       sizeof(spm_lunNodeInPool_t));
                fnd = 1;
                currState = SPM_LUN_OC;
            }
        }
        if (fnd)
        {
            SPM_DBG_PRINT(("Update Db"));
            lCfg.nfxRefCnt++;
            sRec.data = &lCfg;
            sRec.size = sizeof(spm_lunNodeInPool_t);
            tmpDb->put(tmpDb, &sKey, &sRec, 0);
        }
        if (currState != state)
        {
            /* Should be change the state of a used LUN !!! */
            /* This is to avoid nodes using the same LUN. */
            /* USED is the most critical state, so we have to be careful */
            if (state == SPM_LUN_USED || currState != SPM_LUN_USED)
                spm_moveLunState(poolId, sid, currState, state);
        }
    }

    SPM_FCT_PRINT(("...spm_add_lunTo_poolDb"));
    return NFX_OK;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_delLunFrm_poolDb
 *
 * Description  : Remove LUN from pool
 *                Returns # - nfxRefCnt
 *                       -1 - err
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_delLunFrm_poolDb(uint32 poolId, scsi_id_t sid, char *err)
{
    DBT key;
    DBT rec;
    DBT sKey;
    DBT sRec;
    spm_poolDbCb_t *pCfg = NULL;
    DB  *tmpDb = NULL;
    int32  retCode = 0, found = 0;
    spm_lunNodeInPool_t lCfg;
    spm_lunState_t state;

    SPM_FCT_PRINT(("spm_delLunFrm_poolDb..."));
    SPM_DBG_PRINT(("\tPool ID : %d", poolId));
    SPM_DBG_PRINT(("\tSCSI-ID : 0x%llx", sid));
    SPM_DBG_PRINT(("\t  State : %d", state));

    key.data = &poolId;
    key.size = sizeof(uint32);
    if (poolDb->get(poolDb, &key, &rec, 0))
    {
        sprintf(err, "Invalid Pool ID in request (%d)", poolId);
        return -1;
    }

    memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
    if (!pCfg)
        return -1;
    if (pCfg->magic != SPM_MAGIC_STRING)
        return -1;

    sKey.data = &sid;
    sKey.size = sizeof(sid);
    if (pCfg->freeLun && !found)
    {
        tmpDb = pCfg->freeDb;
        if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
        {
            memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data, sizeof(lCfg));
            found = 1;
            state = SPM_LUN_FREE;
        }
    }
    if (pCfg->usedLun && !found)
    {
        tmpDb = pCfg->usedDb;
        if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
        {
            memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data, sizeof(lCfg));
            found = 1;
            state = SPM_LUN_USED;
        }
    }
    if (pCfg->foreLun && !found)
    {
        tmpDb = pCfg->foreDb;
        if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
        {
            memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data, sizeof(lCfg));
            found = 1;
            state = SPM_LUN_FOREIGN;
        }
    }
    if (pCfg->outcLun && !found)
    {
        tmpDb = pCfg->outcDb;
        if (tmpDb->get(tmpDb, &sKey, &sRec, 0) == 0)
        {
            memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data, sizeof(lCfg));
            found = 1;
            state = SPM_LUN_OC;
        }
    }

    if(found)
    {
        retCode = --lCfg.nfxRefCnt;
        if (lCfg.nfxRefCnt)
        {
            /* Update database */
            sRec.data = &lCfg;
            sRec.size = sizeof(lCfg);
            tmpDb->put(tmpDb, &sKey, &sRec, 0);
        }
        else
        {
            /* Removed LUN */
            tmpDb->del(tmpDb, &sKey, 0);
            switch (state)
            {
                case SPM_LUN_FREE :
                    pCfg->freeSize -= lCfg.size;
                    pCfg->freeLun--;
                    if (!pCfg->freeLun)
                    {
                        (pCfg->freeDb)->close(pCfg->freeDb);
                        pCfg->freeDb = NULL;
                    }
                    break;
                case SPM_LUN_USED :
                    pCfg->usedSize -= lCfg.size;
                    pCfg->usedLun--;
                    if (!pCfg->usedLun)
                    {
                        (pCfg->usedDb)->close(pCfg->usedDb);
                        pCfg->usedDb = NULL;
                    }
                    break;
                case SPM_LUN_FOREIGN :
                    pCfg->foreSize -= lCfg.size;
                    pCfg->foreLun--;
                    if (!pCfg->foreLun)
                    {
                        (pCfg->foreDb)->close(pCfg->foreDb);
                        pCfg->foreDb = NULL;
                    }
                    break;
                case SPM_LUN_OC :
                    pCfg->outcSize -= lCfg.size;
                    pCfg->outcLun--;
                    if (!pCfg->outcLun)
                    {
                        (pCfg->outcDb)->close(pCfg->outcDb);
                        pCfg->outcDb = NULL;
                    }
                    break;
            }
            if (!(pCfg->usedLun + pCfg->freeLun +
                  pCfg->foreLun + pCfg->outcLun))
                spm_del_pool(pCfg->id);
        }
    }

    SPM_FCT_PRINT(("...spm_delLunFrm_poolDb"));
    return retCode;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getFree_lunList
 *
 * Description  : Return a list of free LUN's that match the storage.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 21-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getFree_lunList(uint64 reqSize, evm_strg_attributes_t *lunAttr,
                          char *node, scsi_id_t **scsiList, int64 *listCnt,
                          uint64 *allocSize, uint32 txnId)
{
    int32               rc = NFX_ERR;
    int32               idCnt, stat;
    int64               totalSize, maxLunSize, minLunSize;
    int32               gotIt = 0;
    DBT                 rec, sRec;
    DBT                 key, sKey;
    spm_poolDbCb_t      *pCfg;
    spm_poolScsiList_t  sid[100], *sList = NULL;
    int32               sidCnt = 0, sCnt = 0;
    spm_lunDbCb_t       *lunCfg = NULL;
    int32               pSetCnt, setCnt;
    spm_poolCmmNode_t   **pSet;
    spm_allocStatus_t   spmRc = SPM_ALLOC_OK;
    scsi_id_t           *lst = NULL;
    spm_lunNodeInPool_t lCfg;

    /*
     * Check if a matching storage pool exists
     */
    SPM_FCT_PRINT(("spm_getFree_lunList..."));
    *scsiList = NULL;
    *listCnt = 0;
    *allocSize = 0;
    if (!lunAttr)
        return SPM_ALLOC_FAILED;
    SPM_DBG_PRINT(("\treqSize : %lld", reqSize));
    SPM_DBG_PRINT(("\t   Node : %s", node));
    SPM_DBG_PRINT(("\t# Vendor: %d", lunAttr->num));

    spm_createPoolSetsFromStrg(reqSize, lunAttr, &pSet, &pSetCnt);

    if (pSetCnt <= 0 || !pSet)
    {
        if (pSet)
            spm_destroyCmmNode(pSet, pSetCnt);
        return SPM_ALLOC_PENDING;
    }

    /*
     * Need to handle the max data size
     */
    if (lunAttr->minLunSize == EVM_DONT_CARE)
        minLunSize = 0;
    else
        minLunSize = lunAttr->minLunSize;

    if (lunAttr->maxLunSize == EVM_DONT_CARE)
        maxLunSize = 0;
    else
        maxLunSize = lunAttr->maxLunSize;

    for (setCnt = 0; !gotIt && setCnt < pSetCnt; setCnt++)
    {
        spm_poolCmmNode_t *_set;

        totalSize = 0;
        idCnt = 0;
        _set = pSet[setCnt];
        SPM_DBG_PRINT(("Free ind. for set %d %s", _set->setId, _set->data));
        if (sList)
            SPM_FREE(sList);
        bzero(sid, sizeof(sid));
        sidCnt = 0;
        sCnt = 0;
        while (idCnt < _set->s->num)
        {
            SPM_DBG_PRINT(("Match ID %d", _set->s->id[idCnt].id));
            key.data = &_set->s->id[idCnt].id;
            key.size = sizeof(int32);
            rc = poolDb->get(poolDb, &key, &rec, 0);
            if (!rc)
            {
                memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
                if (!pCfg)
                {
                    if (sList)
                        SPM_FREE(sList);
                    return SPM_ALLOC_FAILED;
                }
                if (pCfg->magic != SPM_MAGIC_STRING)
                {
                    if (sList)
                        SPM_FREE(sList);
                    return SPM_ALLOC_FAILED;
                }
                if (pCfg->freeSize)
                {
                    DB   *tmpDb = pCfg->freeDb;
                    DBT  sKey;
                    DBT  sRec;
                    int32 stat;

                    stat = tmpDb->seq(tmpDb, &sKey, &sRec, R_FIRST);
                    while (!stat)
                    {
                        memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data,
                                               sizeof(spm_lunNodeInPool_t));
                        if ((minLunSize > 0) && (minLunSize > lCfg.size))
                        {
                            stat = tmpDb->seq(tmpDb, &sKey, &sRec, R_NEXT);
                            continue;
                        }
                        if ((maxLunSize > 0) && (maxLunSize < lCfg.size))
                        {
                            stat = tmpDb->seq(tmpDb, &sKey, &sRec, R_NEXT);
                            continue;
                        }
                        if (lCfg.resId)
                        {
                            /*
                             * Don't use the res LUNs
                             */
                            stat = tmpDb->seq(tmpDb, &sKey, &sRec, R_NEXT);
                            continue;
                        }
                        if (spm_checkLunInNfxDb(node, lCfg.sid))
                        {
                            sid[sidCnt].id = pCfg->id;
                            sid[sidCnt].sid = lCfg.sid;
                            sidCnt++;
                            SPM_DBG_PRINT(("\tFree (0x%llx)\tSize (%lld)",
                                                       lCfg.sid, lCfg.size));
                            totalSize += lCfg.size;
                            if (totalSize >= reqSize)
                                break;
                        }
                        if (sidCnt == 100)
                        {
                            if(sList)
                            {
                                spm_poolScsiList_t *new;

                                SPM_ALLOC_ZERO(new, spm_poolScsiList_t,
                                     sizeof(spm_poolScsiList_t)*(sCnt+sidCnt));
                                memcpy(new, sList,
                                         sizeof(spm_poolScsiList_t)*sCnt);
                                memcpy((new+sCnt), sid,
                                         sizeof(spm_poolScsiList_t)*sidCnt);
                                SPM_FREE(sList);
                                sList = new;
                            }
                            else
                            {
                                SPM_ALLOC_ZERO(sList, spm_poolScsiList_t,
                                             sizeof(spm_poolScsiList_t)*sidCnt);
                                memcpy(sList, sid,
                                         sizeof(spm_poolScsiList_t)*sidCnt);
                            }
                            sCnt += sidCnt;
                            bzero(sid, sizeof(spm_poolScsiList_t)*sidCnt);
                            sidCnt = 0;
                        }
                        stat = tmpDb->seq(tmpDb, &sKey, &sRec, R_NEXT);
                    }
                }
            }
            if (totalSize && totalSize >= reqSize)
            {
                gotIt = 1;
                break;
            }
            idCnt++;
        }
    } // for
    spm_destroyCmmNode(pSet, pSetCnt);


    SPM_ALLOC_ZERO(lst, scsi_id_t, sizeof(scsi_id_t)*(sCnt+sidCnt));
    idCnt = 0;
    while (idCnt < (sidCnt+sCnt))
    {
        DB *freeDb = NULL;
        DB *usedDb = NULL;
        scsi_id_t tsid;
        int32 tid;
        uint64 size;

        lunCfg = NULL;
        if (idCnt < sCnt)
            tid = sList[idCnt].id;
        else
            tid = sid[idCnt - sCnt].id;

        key.data = &tid;
        key.size = sizeof(int32);

        rc = poolDb->get(poolDb, &key, &rec, 0);
        /* We should always have this record in poolDb */
        NFX_ASSERT(rc == 0);

        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
        if (!pCfg)
        {
            if (sList)
                SPM_FREE(sList);
            SPM_FREE(lst);
            return SPM_ALLOC_FAILED;
        }
        if (pCfg->magic != SPM_MAGIC_STRING)
        {
            if (sList)
                SPM_FREE(sList);
            SPM_FREE(lst);
            return SPM_ALLOC_FAILED;
        }
        freeDb = pCfg->freeDb;

        if (idCnt < sCnt)
            tsid = sList[idCnt].sid;
        else
            tsid = sid[idCnt - sCnt].sid;

        if (gotIt)
        {
            /* We have got what we need so change the status for LUNs */
            if (!pCfg->usedLun)
            {
                pCfg->usedDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                0666, DB_BTREE, NULL);
                if (!pCfg->usedDb)
                {
                    SPM_DBG_PRINT(("spm_getFree_lunList : Memory failure"));
                    SPM_FREE(lst);
                    return SPM_ALLOC_FAILED;
                }
            }
            usedDb = pCfg->usedDb;

            size = spm_changeLunState(tsid, freeDb, usedDb);
            spm_getScsiDb_rec(tsid, &lunCfg);
            if (lunCfg)
                spm_sendLunChangeStateInd(lunCfg->dev, SPM_LUN_FREE,
                                          SPM_LUN_USED);
            pCfg->usedSize += size;
            pCfg->usedLun++;
            pCfg->freeSize -= size;
            pCfg->freeLun--;
            if (!pCfg->freeLun)
            {
                freeDb->close(freeDb);
                pCfg->freeDb = NULL;
            }
        }
        else
        {
            /* We have NOT got what we need so just mark the LUN for SRM txn */
            sKey.data = &tsid;
            sKey.size = sizeof(uint64);
            stat = freeDb->get(freeDb, &sKey, &sRec, 0);
            if (!stat)
            {
                SPM_DBG_PRINT(("\tReserved LUN : 0x%llx", tsid));
                memcpy(&lCfg, (spm_lunNodeInPool_t *)sRec.data,
                                               sizeof(spm_lunNodeInPool_t));
                pCfg->freeSize -= lCfg.size;
                // pCfg->freeLun--;
                lCfg.resId = txnId;
                sRec.data = &lCfg;
                sRec.size = sizeof(spm_lunNodeInPool_t);
                freeDb->put(freeDb, &sKey, &sRec, 0);
            }
        }
        *(lst+idCnt) = tsid;
        idCnt++;
    }
    *listCnt = idCnt;
    *scsiList = lst;
    *allocSize = totalSize;
    if (sList)
        SPM_FREE(sList);
    SPM_DBG_PRINT(("List of SCSI-ID's..."));
    for (idCnt = 0; idCnt < *listCnt && lst; idCnt++)
       SPM_DBG_PRINT(("\tId : 0x%llx", lst[idCnt]));
    if (!gotIt)
       spmRc = SPM_ALLOC_PENDING;

    SPM_FCT_PRINT(("...spm_getFree_lunList"));
    return spmRc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_checkFreeLunList
 *
 * Description  : Check luns that are seen by the same node and have same raid
 *                level
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 05-AUG-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_checkFreeLunList(spm_MsgAllocLunReq_t *req,
                           scsi_id_t **scsiList, int64 *listCnt)
{
    spm_lunDbCb_t *lCfg;
    evm_raid_t    rL;
    uint32 cnt;
    spm_poolScsiList_t     sid[100], *sList = NULL;
    int32         sidCnt = 0, sCnt = 0, idCnt = 0, rc;
    scsi_id_t     *tmp;

    SPM_FCT_PRINT(("spm_checkFreeLunList..."));

    *scsiList = NULL;
    *listCnt = 0;

    spm_getLunDb_rec(req->lunList[0], &lCfg);
    if (!lCfg)
        return SPM_ALLOC_FAILED;
    if (!spm_checkLunInNfxDb(req->hdr.node, lCfg->sid))
        return SPM_ALLOC_FAILED;
    if (spm_getLunState(lCfg->poolId, lCfg->sid) != SPM_LUN_FREE)
        return SPM_ALLOC_FAILED;

    rL = lCfg->rL;
    bzero(sid, sizeof(spm_poolScsiList_t)*100);
    sid[sidCnt].sid = lCfg->sid;
    sid[sidCnt++].id = lCfg->poolId;
    for (cnt = 1; cnt < req->nbLun; cnt++)
    {
        spm_getLunDb_rec(req->lunList[cnt], &lCfg);
        if (!lCfg)
        {
            if (sList)
                SPM_FREE(sList);
            return SPM_ALLOC_FAILED;
        }
        if (!spm_checkLunInNfxDb(req->hdr.node, lCfg->sid))
        {
            if (sList)
                SPM_FREE(sList);
            return SPM_ALLOC_FAILED;
        }
        if (spm_getLunState(lCfg->poolId, lCfg->sid) != SPM_LUN_FREE)
        {
            if (sList)
                SPM_FREE(sList);
            return SPM_ALLOC_FAILED;
        }
        if (lCfg->rL != rL)
        {
            if (sList)
                SPM_FREE(sList);
            return SPM_ALLOC_FAILED;
        }

        if (sidCnt == 100)
        {
            if (sCnt)
            {
                spm_poolScsiList_t *tmp;
                SPM_ALLOC_ZERO(tmp, spm_poolScsiList_t,
                                    sizeof(spm_poolScsiList_t)*(sCnt+100));
                memcpy(tmp, sList, sizeof(spm_poolScsiList_t)*sCnt);
                memcpy((tmp+sCnt), sid, sizeof(spm_poolScsiList_t)*100);
                SPM_FREE(sList);
                sList = tmp;
                sCnt += 100;
            }
            else
            {
                SPM_ALLOC_ZERO(sList, spm_poolScsiList_t,
                                      sizeof(spm_poolScsiList_t)*100);
                memcpy(sList, sid, sizeof(spm_poolScsiList_t)*100);
                sCnt = 100;
            }
            bzero(sid, sizeof(spm_poolScsiList_t)*100);
            sidCnt = 0;
        }
        sid[sidCnt].sid = lCfg->sid;
        sid[sidCnt++].id = lCfg->poolId;
    }

    /* If we reach here, every thing was fine. */
    SPM_ALLOC_ZERO(tmp, scsi_id_t, sizeof(scsi_id_t)*(sCnt+sidCnt));
    idCnt = 0;
    while (idCnt < (sidCnt+sCnt))
    {
        spm_poolDbCb_t      *pCfg;
        DB *freeDb = NULL;
        DB *usedDb = NULL;
        scsi_id_t tsid;
        int32 tid;
        uint64 size;
        DBT    rec, key;

        lCfg = NULL;
        if (idCnt < sCnt)
            tid = sList[idCnt].id;
        else
            tid = sid[idCnt - sCnt].id;

        key.data = &tid;
        key.size = sizeof(int32);

        rc = poolDb->get(poolDb, &key, &rec, 0);
        /* We should always have this record in poolDb */
        NFX_ASSERT(rc == 0);

        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
        if (!pCfg)
        {
            if (sList)
                SPM_FREE(sList);
            SPM_FREE(tmp);
            *scsiList = NULL;
            *listCnt = 0;
            return SPM_ALLOC_FAILED;
        }
        if (pCfg->magic != SPM_MAGIC_STRING)
        {
            if (sList)
                SPM_FREE(sList);
            SPM_FREE(tmp);
            *scsiList = NULL;
            *listCnt = 0;
            return SPM_ALLOC_FAILED;
        }
        freeDb = pCfg->freeDb;

        if (!pCfg->usedLun)
        {
            pCfg->usedDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                            0666, DB_BTREE, NULL);
            if (!pCfg->usedDb)
            {
                SPM_DBG_PRINT(("spm_checkFreeLunList : Memory failure"));
                SPM_FREE(tmp);
                *scsiList = NULL;
                *listCnt = 0;
                return SPM_ALLOC_FAILED;
            }
        }
        usedDb = pCfg->usedDb;

        if (idCnt < sCnt)
            tsid = sList[idCnt].sid;
        else
            tsid = sid[idCnt - sCnt].sid;

        size = spm_changeLunState(tsid, freeDb, usedDb);
        spm_getScsiDb_rec(tsid, &lCfg);
        if (lCfg)
            spm_sendLunChangeStateInd(lCfg->dev, SPM_LUN_FREE,
                                      SPM_LUN_USED);
        pCfg->usedSize += size;
        pCfg->usedLun++;
        pCfg->freeSize -= size;
        pCfg->freeLun--;
        if (!pCfg->freeLun)
        {
            freeDb->close(freeDb);
            pCfg->freeDb = NULL;
        }
        *(tmp+idCnt) = tsid;
        idCnt++;
    }
    *listCnt = idCnt;
    *scsiList = tmp;
    SPM_DBG_PRINT(("List of SCSI-ID's..."));
    for (idCnt = 0; idCnt < *listCnt && tmp; idCnt++)
       SPM_DBG_PRINT(("\tId : 0x%llx", tmp[idCnt]));

    SPM_FCT_PRINT(("...spm_checkFreeLunList"));

    return SPM_ALLOC_OK;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_free_lunList
 *
 * Description  : free LUN's in the list.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 21-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_free_lunList(uint32 nbLun, scsi_id_t *scsiList)
{
    DBT                 key;
    DBT                 rec;
    int32               idCnt;
    spm_poolDbCb_t      *pCfg;
    int32               rc;
    int32               poolId;
    spm_lunDbCb_t       *lunCfg;

    SPM_FCT_PRINT(("spm_free_lunList..."));
    idCnt = 0;
    while (idCnt < nbLun)
    {
        SPM_DBG_PRINT(("\tSCSI-ID : 0x%llx", scsiList[idCnt]));
        poolId = spm_getPoolIdForLun((char *)NULL, scsiList[idCnt]);
        if (poolId > 0)
        {
            if (spm_getLunState(poolId, scsiList[idCnt]) == SPM_LUN_USED)
            {
                DB   *freeDb = NULL;
                DB   *usedDb = NULL;
                uint64 size = 0;

                lunCfg = NULL;
                pCfg = NULL;
                SPM_DBG_PRINT(("Free Lun %llx", scsiList[idCnt]));
                key.data = &poolId;
                key.size = sizeof(int32);
                rc = poolDb->get(poolDb, &key, &rec, 0);
                /* We should always have this record in poolDb */
                NFX_ASSERT(rc == 0);

                memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
                if (!pCfg)
                {
                    idCnt++;
                    continue;
                }
                if (pCfg->magic != SPM_MAGIC_STRING)
                {
                    idCnt++;
                    continue;
                }
                usedDb = pCfg->usedDb;

                if (!pCfg->freeLun)
                {
                    pCfg->freeDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                           0666, DB_BTREE, NULL);
                    if (!pCfg->freeDb)
                    {
                        SPM_DBG_PRINT(("spm_free_lunList : Memory failure"));
                        return NFX_ERR;
                    }
                }
                freeDb = pCfg->freeDb;
                size = spm_changeLunState(scsiList[idCnt], usedDb, freeDb);
                spm_getScsiDb_rec(scsiList[idCnt], &lunCfg);
                if (lunCfg)
                    spm_sendLunChangeStateInd(lunCfg->dev, SPM_LUN_USED,
                                              SPM_LUN_FREE);

                pCfg->freeSize += size;
                pCfg->freeLun++;
                pCfg->usedSize -= size;
                pCfg->usedLun--;
                if (!pCfg->usedLun)
                {
                    usedDb->close(usedDb);
                    pCfg->usedDb = NULL;
                }
            }
        }
        idCnt++;
    }

    SPM_FCT_PRINT(("...spm_free_lunList"));
    return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_poolNum
 *
 * Description  : Get number of pools
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

uint32 spm_get_poolNum()
{
    return poolEntries;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_poolLunNum
 *
 * Description  : Get number of LUN's of a given controller and Raid Level
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_poolLunNum(char *ctrl, int32 rL, spm_lunType_t lType,
                         uint32 doa, char *err)
{
    DBT            key;
    DBT            rec;
    int32          stat;
    int32          lunCnt = 0;
    spm_poolDbCb_t *pCfg;
    spm_poolAttr_t attr;

    SPM_FCT_PRINT(("spm_get_poolLunNum..."));

    bzero(&attr, sizeof(attr));
    attr.mask = SPM_POOL_ATTR_CTRL_MASK;
    strcpy(attr.ctrl, ctrl);
    if (rL != EVM_RAID_ANY)
        attr.mask |= SPM_POOL_ATTR_RAID_MASK;
    attr.raidLevel = rL;

    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer.
        if (pCfg && (pCfg->magic == SPM_MAGIC_STRING))
        {
            switch (doa)
            {
                case 0:
                    /* Disk only */
                    if (pCfg->attr.type != SPM_DEV_DISK)
                    {
                       stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
                       continue;
                    }
                    break;
                case 1:
                    /* All */
                    break;
                case 2:
                    /* non disk only */
                    if (pCfg->attr.type == SPM_DEV_DISK)
                    {
                       stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
                       continue;
                    }
                    break;
            }

            if (spm_compare_pool_attr(&attr, pCfg))
            {
                switch (lType)
                {
                    case SPM_LUN_ALL_LIST:
                        lunCnt += pCfg->freeLun + pCfg->outcLun +
                                  pCfg->foreLun+pCfg->usedLun;
                        break;
                    case SPM_LUN_FREE_LIST:
                        lunCnt += pCfg->freeLun;
                        break;
                    case SPM_LUN_OUTCLUSTER_LIST:
                        lunCnt += pCfg->outcLun;
                        break;
                    case SPM_LUN_FOREIGN_LIST:
                        lunCnt += pCfg->foreLun;
                        break;
                    case SPM_LUN_USED_LIST:
                        lunCnt += pCfg->usedLun;
                        break;
                }
            }
        }
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }

    SPM_FCT_PRINT(("...spm_get_poolLunNum"));

    return lunCnt;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_get_poolLunList
 *
 * Description  : Get list of LUN's of a given controller and Raid Level
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_get_poolLunList(char *ctrl, int32 rL, spm_poolLunList_t **lunList,
                          spm_lunType_t lType, scsi_id_t s_sid, uint32 reqNum,
                          uint32 doa, char *tNode, spm_poolNode_t **poolNode,
                          char *err)
{
    DBT                key;
    DBT                rec;
    int32              stat;
    int32              lunCnt = 0, num = 0;
    spm_poolDbCb_t     *pCfg;
    spm_poolAttr_t     attr;
    int32              skip = 1;
    spm_poolLunList_t  lList[50];
    evm_lunState_t     state;

    SPM_FCT_PRINT(("spm_get_poolLunList..."));

    bzero(&attr, sizeof(attr));
    bzero(lList, sizeof(spm_poolLunList_t)*50);
    attr.mask = SPM_POOL_ATTR_CTRL_MASK;
    strcpy(attr.ctrl, ctrl);
    if (rL != EVM_RAID_ANY)
        attr.mask |= SPM_POOL_ATTR_RAID_MASK;
    attr.raidLevel = rL;

    *poolNode = NULL;

    stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer.
        if (pCfg && (pCfg->magic == SPM_MAGIC_STRING))
        {
            switch (doa)
            {
                case 0:
                    /* Disk only */
                    if (pCfg->attr.type != SPM_DEV_DISK)
                        goto get_next;
                    break;
                case 1:
                    /* All */
                    break;
                case 2:
                    /* non disk only */
                    if (pCfg->attr.type == SPM_DEV_DISK)
                        goto get_next;
                    break;
            }

        }
        else
            goto get_next;
        
        if (spm_compare_pool_attr(&attr, pCfg))
        {
            DBT            skey;
            DBT            srec;
            spm_lunDbCb_t  *lCfg;
            spm_lunNodeInPool_t lKey;
            DB             *tmpDb = NULL;
            int32          rc, aCnt;

            switch (lType)
            {
                case SPM_LUN_FREE_LIST:
                    state = EVM_FREE_LUN;
                    if (pCfg->freeLun)
                        tmpDb = pCfg->freeDb;
                    break;
                case SPM_LUN_OUTCLUSTER_LIST:
                    state = EVM_OC_FREE_LUN;
                    if (pCfg->outcLun)
                        tmpDb = pCfg->outcDb;
                    break;
                case SPM_LUN_FOREIGN_LIST:
                    state = EVM_FOREIGN_LUN;
                    if (pCfg->foreLun)
                        tmpDb = pCfg->foreDb;
                    break;
                case SPM_LUN_USED_LIST:
                    state = EVM_USED_LUN;
                    if (pCfg->usedLun)
                        tmpDb = pCfg->usedDb;
                    break;
                case SPM_LUN_ALL_LIST:
                    break;
            }

            aCnt = 0;
            while (aCnt < 4)
            {
                if (lType != SPM_LUN_ALL_LIST)
                    aCnt = 4;
                else
                {
                    tmpDb = NULL;
                    if (aCnt == 0)
                    {
                        state = EVM_FREE_LUN;
                        if ( pCfg->freeLun)
                            tmpDb = pCfg->freeDb;
                    }
                    else if(aCnt == 1)
                    {
                        state = EVM_OC_FREE_LUN;
                        if (pCfg->outcLun)
                            tmpDb = pCfg->outcDb;
                    }
                    else if(aCnt == 2)
                    {
                        state = EVM_FOREIGN_LUN;
                        if (pCfg->foreDb)
                            tmpDb = pCfg->foreDb;
                    }
                    else if(aCnt == 3)
                    {
                        state = EVM_USED_LUN;
                        if (pCfg->usedDb)
                            tmpDb = pCfg->usedDb;
                    }
                    aCnt++;
                }

                if (!tmpDb)
                {
                    // stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
                    // aCnt = 4;
                    continue;
                }

                rc = tmpDb->seq(tmpDb, &skey, &srec, R_FIRST);
                while (!rc)
                {
                    bzero(&lKey, sizeof(spm_lunNodeInPool_t));
                    memcpy(&lKey, (spm_lunNodeInPool_t *)srec.data,
                                   sizeof(spm_lunNodeInPool_t));
                    spm_getScsiDb_rec(lKey.sid, &lCfg);
                    if (lCfg != NULL)
                    {
                        if (s_sid != 0 && skip)
                        {
                            if (lCfg->sid != s_sid ||
                                (tNode &&
                                !spm_checkLunInNfxDb(tNode, lCfg->sid)))
                            {
                                rc = tmpDb->seq(tmpDb, &skey, &srec, R_NEXT);
                                continue;
                            }
                            else
                            {
                                /* We found the match. Start from next one */
                                rc = tmpDb->seq(tmpDb, &skey, &srec, R_NEXT);
                                skip = 0;
                                SPM_ALLOC_ZERO(*poolNode, spm_poolNode_t,
                                                      sizeof(spm_poolNode_t));
                                strcpy((*poolNode)->ctrl, pCfg->attr.ctrl);
                                (*poolNode)->rL = rL;
                                strcpy((*poolNode)->make, lCfg->make);
                                strcpy((*poolNode)->model, lCfg->model);
                                continue;
                            }
                        }
                        if (reqNum > 0 && ((lunCnt+num) == reqNum))
                            break;
                        if (tNode && !spm_checkLunInNfxDb(tNode, lCfg->sid))
                        {
                            rc = tmpDb->seq(tmpDb, &skey, &srec, R_NEXT);
                            continue;
                        }
                        if (skip)
                        {
                            SPM_ALLOC_ZERO(*poolNode, spm_poolNode_t,
                                                      sizeof(spm_poolNode_t));
                            strcpy((*poolNode)->ctrl, pCfg->attr.ctrl);
                            (*poolNode)->rL = rL;
                            strcpy((*poolNode)->make, lCfg->make);
                            strcpy((*poolNode)->model, lCfg->model);
                            skip = 0;
                        }
                        if (num == 50)
                        {
                            spm_poolLunList_t *new;

                            if (*lunList == NULL)
                            {
                                SPM_ALLOC_ZERO(new, spm_poolLunList_t,
                                                sizeof(spm_poolLunList_t)*num);
                                memcpy(new, lList,
                                                sizeof(spm_poolLunList_t)*num);
                            }
                            else
                            {
                                SPM_ALLOC_ZERO(new, spm_poolLunList_t,
                                        sizeof(spm_poolLunList_t)*(num+lunCnt));
                                memcpy(new, *lunList,
                                        sizeof(spm_poolLunList_t)*lunCnt);
                                memcpy((new+lunCnt), lList,
                                        sizeof(spm_poolLunList_t)*num);
                                SPM_FREE(*lunList);
                            }
                            *lunList = new;
                            lunCnt += num;
                            num = 0;
                            bzero(lList, sizeof(spm_poolLunList_t)*50);
                        }
                        if (lCfg->dev)
                            strcpy(lList[num].dev, lCfg->dev);
                        if (lCfg->cluster)
                            strcpy(lList[num].lunCluster, lCfg->cluster);
                        lList[num].lun  = lCfg->lun;
                        lList[num].size  = lCfg->size;
                        lList[num].state = state;
                        lList[num].sid   = lCfg->sid;
                        lList[num].rL    = lCfg->rL;
                        lList[num].type  = pCfg->attr.type;
                        strcpy(lList[num].model, lCfg->model);
                        num++;
                    }
                    rc = tmpDb->seq(tmpDb, &skey, &srec, R_NEXT);
                }
            }
            if (reqNum > 0 && ((lunCnt+num) == reqNum))
                break;
        }
get_next:
        stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
    }
    if (num > 0)
    {
        spm_poolLunList_t *new;

        if (*lunList == NULL)
        {
            SPM_ALLOC_ZERO(new, spm_poolLunList_t,
                                sizeof(spm_poolLunList_t)*num);
            memcpy(new, lList, sizeof(spm_poolLunList_t)*num);
        }
        else
        {
            SPM_ALLOC_ZERO(new, spm_poolLunList_t,
                                sizeof(spm_poolLunList_t)*(num+lunCnt));
            memcpy(new, *lunList, sizeof(spm_poolLunList_t)*lunCnt);
            memcpy((new+lunCnt), lList, sizeof(spm_poolLunList_t)*num);
            SPM_FREE(*lunList);
        }
        *lunList = new;
        lunCnt += num;
    }

    SPM_FCT_PRINT(("...spm_get_poolLunList"));
    return (lunCnt);
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getSpmDevPortDetail
 *
 * Description  : Get list of LUN's of a given controller and Raid Level
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 13-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getSpmDevPortDetail(int32 idCnt, scsi_id_t *sList,
                              int32 numReq, scsi_id_t lastId,
                              spm_poolLunList_t *lunList)
{
    int32              lunCnt = 0, num = 0;
    spm_lunDbCb_t      *lCfg;
    uint32             state;

    SPM_FCT_PRINT(("spm_getSpmDevPortDetail..."));
    SPM_DBG_PRINT(("\t# Req. : %d", numReq));
    SPM_DBG_PRINT(("\tLastID : 0x%llx", lastId));

    for (num = 0; num < idCnt; num++)
    {
        if (numReq)
        {
            if (numReq > lunCnt)
            {
                if (lastId)
                { 
                    if(lastId == sList[num])
                        lastId = 0;
                    continue;
                }
            }
            else
                break;
        }
        spm_getScsiDb_rec(sList[num], &lCfg);
        if (lCfg)
        {
            if (lCfg->dev)
                strcpy(lunList[lunCnt].dev, lCfg->dev);
            if (lCfg->cluster)
                strcpy(lunList[lunCnt].lunCluster, lCfg->cluster);
            lunList[lunCnt].lun   = lCfg->lun;
            lunList[lunCnt].size  = lCfg->size;
            state = spm_getLunState(lCfg->poolId, lCfg->sid);
            if (state == SPM_LUN_FREE)
                lunList[lunCnt].state = EVM_FREE_LUN;
            else if (state == SPM_LUN_USED)
                lunList[lunCnt].state = EVM_USED_LUN;
            else if (state == SPM_LUN_FOREIGN)
                lunList[lunCnt].state = EVM_FOREIGN_LUN;
            else if (state == SPM_LUN_OC)
                lunList[lunCnt].state = EVM_OC_FREE_LUN;
            lunList[lunCnt].sid   = lCfg->sid;
            lunList[lunCnt].rL    = lCfg->rL;
            lunList[lunCnt].type  = lCfg->devType;
            lunCnt++;
        }
    }

    SPM_FCT_PRINT(("...spm_getSpmDevPortDetail"));
    return (lunCnt);
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_unReserveLun
 *
 * Description  : Unreseve the list of scsi-ids. These are caused by pending req
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 13-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_unReserveLun(uint32 idCnt, scsi_id_t *list)
{
    uint32         cnt;
    DB             *freeDb;
    spm_poolDbCb_t *pCfg = NULL;
    spm_lunDbCb_t  *lCfg = NULL;
    DBT            key, sKey;
    DBT            rec, sRec;
    uint32         stat;
    uint64         sid;
    spm_lunNodeInPool_t poolNode;

    SPM_FCT_PRINT(("spm_unReserveLun..."));

    for (cnt = 0; list && cnt < idCnt; cnt++)
    {
        sid = list[cnt];
        spm_getScsiDb_rec(sid, &lCfg);
        if (lCfg)
        {
            key.data = &lCfg->poolId;
            key.size = sizeof (lCfg->poolId);
            stat = poolDb->get(poolDb, &key, &rec, 0);
            if (!stat)
            {
                memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer.
                if (pCfg && (pCfg->magic == SPM_MAGIC_STRING))
                {
                    freeDb = pCfg->freeDb;
                    sKey.data = &sid;
                    sKey.size = sizeof(sid);
                    stat = freeDb->get(freeDb, &sKey, &sRec, 0);
                    if (!stat)
                    {
                        memcpy(&poolNode, (spm_lunNodeInPool_t *)sRec.data,
                               sizeof(poolNode));
                        SPM_DBG_PRINT(("\tUn reserve completed"));
                        poolNode.resId = 0;
                        pCfg->freeSize += poolNode.size;
                        sRec.data = &poolNode;
                        sRec.size = sizeof(spm_lunNodeInPool_t);
                        freeDb->put(freeDb, &sKey, &sRec, 0);
                    }
                } 
            }
        }
    }

    SPM_FCT_PRINT(("...spm_unReserveLun"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_unReserveDevice
 *
 * Description  : Unreseve the list of Device. These are caused by pending req
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 13-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_unReserveDevice(uint32 idCnt, char *list)
{
    uint32         cnt;
    spm_lunDbCb_t  *lCfg = NULL;

    SPM_FCT_PRINT(("spm_unReserveDevice..."));

    for (cnt = 0; cnt < idCnt; cnt++)
    {
        spm_getLunDb_rec((list+(cnt*(SDM_DEV_NAME_LEN+8))), &lCfg);
        if (lCfg)
        {
            SPM_DBG_PRINT(("\tDevice unreserved : %s", (list+(cnt*(SDM_DEV_NAME_LEN+8)))));
            spm_unReserveLun(1, &lCfg->sid);
        }
    }
    SPM_FCT_PRINT(("...spm_unReserveDevice"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_reserveLun
 *
 * Description  : reserve the LUN. To co-relate the label requests with pending
 *                vol grow req
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 13-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_reserveLun(scsi_id_t sid, uint32 resId)
{
    DB             *freeDb;
    spm_poolDbCb_t *pCfg = NULL;
    spm_lunDbCb_t  *lCfg = NULL;
    DBT            key, sKey;
    DBT            rec, sRec;
    uint32         stat;
    spm_lunNodeInPool_t poolNode;

    SPM_FCT_PRINT(("spm_reserveLun..."));

    SPM_DBG_PRINT(("\tSCSI-ID : 0x%llx", sid));
    SPM_DBG_PRINT(("\tRes ID  : %d", resId));
    spm_getScsiDb_rec(sid, &lCfg);
    if (lCfg)
    {
        SPM_DBG_PRINT(("\tFound LUN : poolId %d", lCfg->poolId));
        key.data = &lCfg->poolId;
        key.size = sizeof (lCfg->poolId);
        stat = poolDb->get(poolDb, &key, &rec, 0);
        if (!stat)
        {
            SPM_DBG_PRINT(("\tFound pool 1"));
            memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer.
            if (pCfg && (pCfg->magic == SPM_MAGIC_STRING))
            {
                SPM_DBG_PRINT(("\tPool OK"));
                freeDb = pCfg->freeDb;
                sKey.data = &sid;
                sKey.size = sizeof(sid);
                stat = freeDb->get(freeDb, &sKey, &sRec, 0);
                if (!stat)
                {
                    memcpy(&poolNode, (spm_lunNodeInPool_t *)sRec.data,
                           sizeof(spm_lunNodeInPool_t));
                    SPM_DBG_PRINT(("\tMarked reserved"));
                    poolNode.resId = resId;
                    pCfg->freeSize -= poolNode.size;
                    sRec.data = &poolNode;
                    sRec.size = sizeof(spm_lunNodeInPool_t);
                    freeDb->put(freeDb, &sKey, &sRec, 0);
                }
            }
        }
    }

    SPM_FCT_PRINT(("...spm_reserveLun"));
    return;
}

/*----------------------------------------------------------------------------*/
