/*------------------------------------------------------------------------------
 * RCS:
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : spm-util.c
 *
 * Description  : This file contains the methods used to manipulate
 *                sets for SPM alloc requests.
 *
 * Created by   : Nitin Jain
 * Date Created : 19-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

#include "spm.h"

/*
 * Global variables.
 */

extern DB *poolDb;
extern uint32 poolEntries;

char *raidString[] =
{
    "Raid-0", // EVM_RAID0 = 0,      // stripping
    "Raid-1", // EVM_RAID1 = 1,      // mirror
    "Raid-01", // EVM_RAID01 = 2,     // stripe + mirror
    "Raid-3", // EVM_RAID3 = 3,      // raid 3
    "",
    "Raid-5", // EVM_RAID5 = 5,      // raid 5
    "Plain", // EVM_RAID_NONE = 6,  // no raid level, ie plain
    "Any" // EVM_RAID_ANY = 0xffff
};

/*
 * Local function prototypes.
 */

static spm_poolSetNode_t *spm_createPoolSetNode(int32 maxEl);

static void spm_destroyPoolSetNode(spm_poolSetNode_t *s);

static int32 spm_getPoolSetNode(spm_poolSetNode_t *s, int32 *id,
                         spm_poolDbCb_t **cfg, int32 next);

static void spm_insertPoolSetNodeId(spm_poolSetNode_t *s, int32 id,
                             spm_poolDbCb_t *cfg);

static spm_poolCmmNode_t *spm_createCmmNode(spm_poolCmmNode_t **cs, char *name,
                                     int32 num, int32 pSetId);

static spm_poolCmmNode_t *spm_getCmmNode(spm_poolCmmNode_t **cs, char *name,
                                  int32 num, int32 pSetId);

/* Functions for spm_poolSetNode_t */

static spm_poolSetNode_t *spm_createPoolSetNode(int32 maxEl)
{
    spm_poolSetNode_t *s;

    SPM_FCT_PRINT(("spm_createPoolSetNode..."));
    SPM_FCT_PRINT(("\tElements (%d)", maxEl));
    if (!maxEl) return NULL;

    SPM_ALLOC_ZERO(s, spm_poolSetNode_t, sizeof(spm_poolSetNode_t));
    SPM_ALLOC_ZERO(s->id, spm_poolIdPtr_t, sizeof(spm_poolIdPtr_t)*maxEl);

    SPM_FCT_PRINT(("...spm_createPoolSetNode"));
    return s;
}

static void spm_destroyPoolSetNode(spm_poolSetNode_t *s)
{
    SPM_FCT_PRINT(("spm_destroyPoolSetNode..."));

    if (!s) return;
    SPM_FREE(s->id);
    SPM_FREE(s);
    SPM_FCT_PRINT(("...spm_destroyPoolSetNode"));

    return;
}

static int32 spm_getPoolSetNode(spm_poolSetNode_t *s, int32 *id,
                         spm_poolDbCb_t **cfg, int32 next)
{
    *cfg = NULL;
    *id = 0;

    SPM_FCT_PRINT(("spm_getPoolSetNode..."));
    SPM_DBG_PRINT(("\t Next (%d)", next));

    if (!s) return -1;
    if (next == s->num) return 0;
    else
    {
        *id = s->id[next].id;
        *cfg = s->id[next].cfg;
    }
    SPM_DBG_PRINT(("\tid (%d)", *id));
    SPM_FCT_PRINT(("...spm_getPoolSetNode"));
    return (++next);
}

static void spm_insertPoolSetNodeId(spm_poolSetNode_t *s, int32 id,
                             spm_poolDbCb_t *cfg)
{
    int32 i = 0;

    SPM_FCT_PRINT(("spm_insertPoolSetNodeId..."));

    if (!s) return;

    while (i < s->num)
    {
        if (s->id[i].id == id)
            return;
        i++;
    }
    s->id[s->num].id = id;
    s->id[s->num].cfg = cfg;
    s->num++;
    SPM_DBG_PRINT(("\tid (%d) num (%d)", id, s->num));

    SPM_FCT_PRINT(("...spm_insertPoolSetNodeId"));
    return;
}


/* Functions for spm_poolCmmNode_t */

static spm_poolCmmNode_t *spm_createCmmNode(spm_poolCmmNode_t **cs, char *name,
                                     int32 num, int32 pSetId)
{
    SPM_FCT_PRINT(("spm_createCmmNode..."));
    SPM_DBG_PRINT(("\tName (%s) Num (%d), pSetId (%d)", name, num, pSetId));

    if (!cs) return NULL;

    SPM_ALLOC_ZERO(cs[num], spm_poolCmmNode_t, sizeof(spm_poolCmmNode_t));
    if (name)
        strcpy(cs[num]->data, name);
    cs[num]->prevSetId = pSetId;
    cs[num]->setId = num;

    SPM_FCT_PRINT(("...spm_createCmmNode"));
    return cs[num];
}

//static void spm_destroyCmmNode(spm_poolCmmNode_t **cs, int32 num)
void spm_destroyCmmNode(spm_poolCmmNode_t **cs, int32 num)
{
    int32 i = 0;

    SPM_FCT_PRINT(("spm_destroyCmmNode..."));
    SPM_DBG_PRINT(("\tNum  : %d", num));

    if (!cs) return;

    while (i < num)
    {
        SPM_DBG_PRINT(("\tData : %s", cs[i]->data));
        spm_destroyPoolSetNode(cs[i]->s);
        SPM_FREE(cs[i]);
        i++;
    }
    SPM_FREE(cs);

    SPM_FCT_PRINT(("...spm_destroyCmmNode"));
    return;
}

static spm_poolCmmNode_t *spm_getCmmNode(spm_poolCmmNode_t **cs, char *name,
                                  int32 num, int32 pSetId)
{
    int32 i = 0;

    SPM_FCT_PRINT(("spm_getCmmNode..."));
    SPM_DBG_PRINT(("\tName (%s) Num (%d), pSetId (%d)", name, num, pSetId));

    if (!cs) return NULL;

    while (i < num)
    {
        if (name)
        {
            if (cs[i]->data && strlen(cs[i]->data))
               if (!strcasecmp(cs[i]->data, name) && cs[i]->prevSetId == pSetId)
                    return cs[i];
        }
        else
        {
            if (!cs[i]->data && cs[i]->prevSetId == pSetId) return cs[i];
        }
        i++;
    }

    SPM_FCT_PRINT(("...spm_getCmmNode"));
    return NULL;
}

/* create sets of pools that match the given criteria of a search */

int32 spm_createPoolSetsFromStrg(uint64 size, const evm_strg_attributes_t *const attr,
                                 spm_poolCmmNode_t ***idSet, int32 *const idCnt)
{
    spm_poolSetNode_t *pSet;
    spm_poolCmmNode_t **cSet, **vSet, **mSet, **rSet, *tSet;
    int32 pCnt, cCnt, vCnt, mCnt, rCnt, cnt, next;
    spm_poolAttr_t poolAttr;
    DBT key, rec;
    int32 stat, id;
    spm_poolDbCb_t *pCfg;
    uint64 pSize = 0;
    uint32 raidIdx = 0, pLuns = 0;

    SPM_FCT_PRINT(("spm_createPoolSetsFromStrg..."));
    *idSet = NULL;
    *idCnt = 0;

    if (!poolEntries)
    {
        SPM_ERR_PRINT(("There are no luns!!!"));
        return 0;
    }

    bzero (&poolAttr, sizeof(spm_poolAttr_t));
    poolAttr.mask = SPM_POOL_ATTR_ALL_MASK;
    poolAttr.raidLevel = attr->raidLevel;
    poolAttr.type = SPM_DEV_DISK;

    if (attr->raidLevel == EVM_RAID_ANY)
        poolAttr.mask &= ~SPM_POOL_ATTR_RAID_MASK;
    if (strlen(attr->raidCtrl))
        strcpy(poolAttr.ctrl, attr->raidCtrl);
    else
        poolAttr.mask &= ~SPM_POOL_ATTR_CTRL_MASK;

    /* Create pool set for the given strg attr */
    cnt = 0;
    pCnt = 0;
    pSet = spm_createPoolSetNode(poolEntries);
    if (pSet)
    {
        do {
            if (strlen(attr->vendorInfo[cnt].vendor))
            {
                poolAttr.mask |= SPM_POOL_ATTR_MAKE_MASK;
                strcpy(poolAttr.name, attr->vendorInfo[cnt].vendor);
                SPM_DBG_PRINT(("\tvendor : %s", poolAttr.name));
            }
            else
                poolAttr.mask &= ~SPM_POOL_ATTR_MAKE_MASK;

            if (strlen(attr->vendorInfo[cnt].model))
            {
                poolAttr.mask |= SPM_POOL_ATTR_MODEL_MASK;
                strcpy(poolAttr.model, attr->vendorInfo[cnt].model);
                SPM_DBG_PRINT(("\tmodel  : %s", poolAttr.model));
            }
            else
                poolAttr.mask &= ~SPM_POOL_ATTR_MODEL_MASK;

            stat = poolDb->seq(poolDb, &key, &rec, R_FIRST);
            while (!stat)
            {
                memcpy(&pCfg, rec.data, sizeof(pCfg)); // Copy pointer
                if (!pCfg)
                {
                    stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
                    continue;
                }
                if (pCfg->magic != SPM_MAGIC_STRING)
                {
                    stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
                    continue;
                }

                if (pCfg->freeLun)
                    if (spm_compare_pool_attr(&poolAttr, pCfg))
                    {
                        pLuns += pCfg->freeLun;
                        pSize += pCfg->freeSize;
                        spm_insertPoolSetNodeId(pSet, pCfg->id, pCfg);
                    }
                stat = poolDb->seq(poolDb, &key, &rec, R_NEXT);
            }
            cnt++;
        } while(cnt < attr->num);
        pCnt = pSet->num;
    }
    else
    {
        SPM_DBG_PRINT(("No luns available"));
        SPM_FCT_PRINT(("...spm_createPoolSetsFromStrg"));
        return 0;
    }

    if (!pCnt)
    {
        SPM_DBG_PRINT(("No matching attributes found"));
        spm_destroyPoolSetNode(pSet);
        SPM_FCT_PRINT(("...spm_createPoolSetsFromStrg"));
        return 0;
    }

    SPM_DBG_PRINT(("Value set : "));
    for (cnt = 0, next = 0; cnt < pCnt; cnt++)
        SPM_DBG_PRINT(("%d ", pSet->id[cnt].id));

    /* Create pool set for CTRL */
    cnt = 0;
    cCnt = 0;
    if (attr->flags & EVM_SAME_RAIDCTRL)
    {
        SPM_ALLOC_ZERO(cSet, spm_poolCmmNode_t *,
                             sizeof(spm_poolCmmNode_t *)*poolEntries);
        next = 0;
        while (cnt < pCnt)
        {
            next = spm_getPoolSetNode(pSet, &id, &pCfg, next);
            if (pCfg)
            {
                tSet = spm_getCmmNode(cSet, pCfg->attr.ctrl, cCnt, 0);
                if (!tSet)
                {
                    tSet = spm_createCmmNode(cSet, pCfg->attr.ctrl, cCnt, 0);
                    cCnt++;
                    tSet->s = spm_createPoolSetNode(pCnt);
                }
                spm_insertPoolSetNodeId(tSet->s, pCfg->id, pCfg);
                tSet->numLuns += pCfg->freeLun;
                tSet->size += pCfg->freeSize;
            }
            cnt++;
        }
        spm_destroyPoolSetNode(pSet);
    }
    else
    {
        SPM_ALLOC_ZERO(cSet, spm_poolCmmNode_t *, sizeof(spm_poolCmmNode_t *));
        tSet = spm_createCmmNode(cSet, NULL, cCnt, 0);
        cCnt = 1;
        tSet->s = pSet;
        tSet->numLuns = pLuns;
        tSet->size = pSize;
    }

    SPM_DBG_PRINT(("CTRL set :"));
    for (cnt = 0; cnt < cCnt; cnt++)
    {
        spm_poolCmmNode_t *_cSet;
        int32 i;

        _cSet = cSet[cnt];
        SPM_DBG_PRINT(("\tSet %d %s", _cSet->setId, _cSet->data));
        for (i = 0; i < _cSet->s->num; i++)
            SPM_DBG_PRINT(("%d ", _cSet->s->id[i].id));
    }

    /* Create pool set for Vendor */
    cnt = 0;
    vCnt = 0;
    if (attr->flags & EVM_SAME_VENDOR)
    {
        spm_poolCmmNode_t *_cSet;
        int32 i;

        SPM_ALLOC_ZERO(vSet, spm_poolCmmNode_t *,
                             sizeof(spm_poolCmmNode_t *)*poolEntries);
        while (cnt < cCnt)
        {
            _cSet = cSet[cnt];
            i = 0;
            next = 0;
            while (i < _cSet->s->num)
            {
                next = spm_getPoolSetNode(_cSet->s, &id, &pCfg, next);
                if (pCfg)
                {
                    tSet = spm_getCmmNode(vSet, pCfg->attr.name, vCnt,
                                          _cSet->setId);
                    if (!tSet)
                    {
                        tSet = spm_createCmmNode(vSet, pCfg->attr.name,
                                                 vCnt, _cSet->setId);
                        vCnt++;
                        tSet->s = spm_createPoolSetNode(pCnt);
                    }
                    spm_insertPoolSetNodeId(tSet->s, pCfg->id, pCfg);
                    tSet->numLuns += pCfg->freeLun;
                    tSet->size += pCfg->freeSize;
                }
                i++;
            }
            cnt++;
        }
        spm_destroyCmmNode(cSet, cCnt);
    }
    else
    {
        vCnt = cCnt;
        vSet = cSet;
    }

    SPM_DBG_PRINT(("Vendor set :"));
    for (cnt = 0; cnt < vCnt; cnt++)
    {
        spm_poolCmmNode_t *_vSet;
        int32 i;

        _vSet = vSet[cnt];
        SPM_DBG_PRINT(("\tSet %d %s", _vSet->setId, _vSet->data));
        for (i = 0; i < _vSet->s->num; i++)
            SPM_DBG_PRINT(("%d ", _vSet->s->id[i].id));
    }

    /* Create pool set for Model */
    cnt = 0;
    mCnt = 0;
    if (attr->flags & EVM_SAME_MODEL)
    {
        spm_poolCmmNode_t *_vSet;
        int32 i;

        SPM_ALLOC_ZERO(mSet, spm_poolCmmNode_t *,
                             sizeof(spm_poolCmmNode_t *)*poolEntries);
        while (cnt < vCnt)
        {
            _vSet = vSet[cnt];
            i = 0;
            next = 0;
            while (i < _vSet->s->num)
            {
                next = spm_getPoolSetNode(_vSet->s, &id, &pCfg, next);
                if (pCfg)
                {
                    tSet = spm_getCmmNode(mSet, pCfg->attr.model, mCnt,
                                          _vSet->setId);
                    if (!tSet)
                    {
                        tSet = spm_createCmmNode(mSet, pCfg->attr.model,
                                                 mCnt, _vSet->setId);
                        mCnt++;
                        tSet->s = spm_createPoolSetNode(pCnt);
                    }
                    spm_insertPoolSetNodeId(tSet->s, pCfg->id, pCfg);
                    tSet->numLuns += pCfg->freeLun;
                    tSet->size += pCfg->freeSize;
                }
                i++;
            }
            cnt++;
        }
        spm_destroyCmmNode(vSet, vCnt);
    }
    else
    {
        mCnt = vCnt;
        mSet = vSet;
    }

    SPM_DBG_PRINT(("Model set :"));
    for (cnt = 0; cnt < mCnt; cnt++)
    {
        spm_poolCmmNode_t *_mSet;
        int32 i;
    
        _mSet = mSet[cnt];
        SPM_DBG_PRINT(("\tSet %d %s", _mSet->setId, _mSet->data));
        for (i = 0; i < _mSet->s->num; i++)
            SPM_DBG_PRINT(("%d ", _mSet->s->id[i].id));
    }

    cnt = 0;
    rCnt = 0;
    rSet = NULL;
    if (attr->raidLevel == EVM_RAID_ANY)
    {
        /* Sort the sets in sets of RLs */
        spm_poolCmmNode_t *_mSet;
        int32 i;

        SPM_ALLOC_ZERO(rSet, spm_poolCmmNode_t *,
                             sizeof(spm_poolCmmNode_t)*poolEntries);
        while (cnt < mCnt)
        {
            _mSet = mSet[cnt];
            i = 0;
            next = 0;
            while (i < _mSet->s->num)
            {
                next = spm_getPoolSetNode(_mSet->s, &id, &pCfg, next);
                if (pCfg)
                {
                    raidIdx = (pCfg->attr.raidLevel == EVM_RAID_ANY)
                                                 ?7:pCfg->attr.raidLevel;
                    tSet = spm_getCmmNode(rSet, raidString[raidIdx], rCnt,
                                          _mSet->setId);
                    if (!tSet)
                    {
                        tSet = spm_createCmmNode(rSet, raidString[raidIdx],
                                                 rCnt, _mSet->setId);
                        rCnt++;
                        tSet->s = spm_createPoolSetNode(pCnt);
                    }
                    spm_insertPoolSetNodeId(tSet->s, pCfg->id, pCfg);
                    tSet->numLuns += pCfg->freeLun;
                    tSet->size += pCfg->freeSize;
                }
                i++;
            }
            cnt++;
        }

        SPM_DBG_PRINT(("Raid Level set :"));
        for (cnt = 0; cnt < rCnt; cnt++)
        {
            spm_poolCmmNode_t *_rSet;
            int32 i;

            _rSet = rSet[cnt]; 
            SPM_DBG_PRINT(("\tSet %d %s", _rSet->setId, _rSet->data));
            for (next = 0, i = 0; i < _rSet->s->num; i++)
                SPM_DBG_PRINT(("\tID : %d", _rSet->s->id[i].id));
        }
        *idSet = rSet;
        *idCnt = rCnt;
        spm_destroyCmmNode(mSet, mCnt);
    }
    else
    {
        *idSet = mSet;
        *idCnt = mCnt;
    }

    SPM_FCT_PRINT(("...spm_createPoolSetsFromStrg"));
    return 1;
}
