/*
 *------------------------------------------------------------------------------
 *
 * Name:	mc-ui.c
 *
 * RCS:		$Id: mc-ui.c,v 1.5 2002/07/24 04:11:01 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	MC driver UI.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 12/27/01
 *
 *------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
//#include <string.h> - Isabelle
#include <sys/types.h>
#include <sys/time.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include "tape-api.h"

#define MC_CLIENT_TAG 0x12345678

int send_tape_message(char *, int, char *, int *, int);
char *tape_ecode_to_string(uint32);

int32 mc_ui_devname_to_handle(char *, uint32 *);

int mc_ui_devlist(int, char **, ...);
int mc_ui_open(int, char **, ...);
int mc_ui_close(int, char **, ...);
int mc_ui_dbg_level(int, char **, ...);

typedef struct {
	char *	name;
	int	parm_cnt;	/* Min command-line parameter count */
	int	implicit_arg1;	/* Implicit argument, derived from command line parms */
	int	implicit_arg2;	/* Implicit argument, derived from command line parms */
	int (*func)(int, char **, ...);
	char *	usage;
} mc_ui_node_t;

mc_ui_node_t mc_ui_root[] = {
	{ "devlist",	      0,  0, 0,					mc_ui_devlist,
	  "devlist" },
	{ "open",	      1,  0, 0,					mc_ui_open, 
	  "open <device-name>" },
	{ "close",	      1,  0, 0,					mc_ui_close, 
	  "close <device-name>" },
	{ "dbg_level",	      1,  0, 0,					mc_ui_dbg_level,
	  "dbg_level <value> -- set driver debug level" },
	{ NULL,		      0,  0, 0,					NULL,			
	  NULL }
};
	
static void 
print_usage(char *command)
{
	int i;

	for (i = 0; mc_ui_root[i].name; ++i) {
		fprintf(stderr, "%s %s %s\n",
			(i == 0 ? "Usage:" : "      "),
			command, mc_ui_root[i].usage);
	}
}

int32
mc_ui_devname_to_handle(char *devname, uint32 *handle)
{
	int			    rc;
	char			    req_msg[1024];
	char			    rsp_msg[1024];
	int			    req_len = 0;
	int			    rsp_len = 0;
	mc_get_handle_req_msg_t *   req;
	mc_get_handle_rsp_msg_t *   rsp;

	req = (mc_get_handle_req_msg_t *)req_msg;
	req->hdr.mid = MC_MID_GET_HANDLE_REQ;
	req->hdr.cookie = 1;
	strcpy(req->devicename, devname);
			
	req_len = sizeof(mc_get_handle_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 5);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		return(NFX_ERR);
	}

	rsp = (mc_get_handle_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != MC_MID_GET_HANDLE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       MC_MID_GET_HANDLE_RSP, rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(mc_get_handle_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, 
		       sizeof(mc_get_handle_rsp_msg_t), 
		       rsp_len);
		return(NFX_ERR);
	}
	if (rsp->ecode == TAPE_ECODE_SUCCESS)
		*handle = rsp->handle;
	return(rsp->ecode);
}


int
mc_ui_devlist(int argc, char **argv, ...)
{
	int			     rc;
	char			     req_msg[1024];
	char			     rsp_msg[16 * 1024];
	int			     req_len = 0;
	int			     rsp_len = 0;
	mc_get_devlist_req_msg_t *   devlist_req;
	mc_get_devlist_rsp_msg_t *   devlist_rsp;
	int			     i;
	mc_dev_info_t *		     dev_info;

	devlist_req = (mc_get_devlist_req_msg_t *)req_msg;
	devlist_req->hdr.mid = MC_MID_GET_DEVLIST_REQ;
	devlist_req->hdr.cookie = 1;
			
	req_len = sizeof(mc_get_devlist_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 5);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", devlist_req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", devlist_req->hdr.mid);
		return(NFX_ERR);
	}

	devlist_rsp = (mc_get_devlist_rsp_msg_t *)rsp_msg;
	if (devlist_rsp->hdr.mid != MC_MID_GET_DEVLIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       MC_MID_GET_DEVLIST_RSP, devlist_rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(mc_get_devlist_rsp_msg_t) + sizeof(mc_dev_info_t) * (devlist_rsp->dev_count - 1)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       devlist_req->hdr.mid, 
		       sizeof(mc_get_devlist_req_msg_t) + sizeof(mc_dev_info_t) * (devlist_rsp->dev_count - 1),
		       rsp_len);
		return(NFX_ERR);
	}
	if (devlist_rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("mc get_devlist failed: ecode %s(0x%x)\n", 
		       tape_ecode_to_string(devlist_rsp->ecode), devlist_rsp->ecode);
		return(NFX_ERR);
	}

	printf("%d devices found.\n", devlist_rsp->dev_count);

	for (i = 0, dev_info = devlist_rsp->info; i < devlist_rsp->dev_count; ++i, ++dev_info) {
		printf("[%2d]: devicename %s\n", i, dev_info->devicename);
		printf("      physical devicename %s\n", dev_info->physical_devicename);
		printf("      vendor '%s', product '%s'\n", dev_info->scsi_vendor_id, dev_info->scsi_product_id);
		printf("      scsi ID 0x%016llx\n", dev_info->scsi_id);
		printf("\n");
	} 
	return(NFX_OK);
}

int
mc_ui_open(int argc, char **argv, ...)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	mc_open_req_msg_t *	open_req;
	mc_open_rsp_msg_t *	open_rsp;

	open_req = (mc_open_req_msg_t *)req_msg;
	open_req->hdr.mid = MC_MID_OPEN_REQ;
	open_req->hdr.cookie = 1;
	strcpy(open_req->devicename, argv[1]);
	open_req->client_tag = MC_CLIENT_TAG;
			
	req_len = sizeof(mc_open_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 5);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", open_req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", open_req->hdr.mid);
		return(NFX_ERR);
	}

	open_rsp = (mc_open_rsp_msg_t *)rsp_msg;
	if (open_rsp->hdr.mid != MC_MID_OPEN_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       MC_MID_OPEN_RSP, open_rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(mc_open_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       open_req->hdr.mid, sizeof(mc_open_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (open_rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("mc open failed: ecode %s(0x%x)\n", 
		       tape_ecode_to_string(open_rsp->ecode), open_rsp->ecode);
		return(NFX_ERR);
	}

	printf("Handle               : 0x%x\n", open_rsp->handle);
	printf("SCSI ID              : %016llx\n", open_rsp->scsi_id);
	printf("WWN                  : %016llx\n", open_rsp->wwn);
	printf("LUN                  : %d\n", open_rsp->lun);
	printf("Device name          : %s\n", open_rsp->devicename);
	printf("Physical device name : %s\n", open_rsp->physical_devicename);
	printf("Client tag           : 0x%x\n", open_rsp->client_tag);
	return(NFX_OK);
}

int
mc_ui_close(int argc, char **argv, ...)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	mc_close_req_msg_t *	close_req;
	mc_close_rsp_msg_t *	close_rsp;

	close_req = (mc_close_req_msg_t *)req_msg;
	close_req->hdr.mid = MC_MID_CLOSE_REQ;
	close_req->hdr.cookie = 1;
	strcpy(close_req->devicename, argv[1]);
	close_req->client_tag = MC_CLIENT_TAG;
			
	req_len = sizeof(mc_close_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 5);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", close_req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", close_req->hdr.mid);
		return(NFX_ERR);
	}

	close_rsp = (mc_close_rsp_msg_t *)rsp_msg;
	if (close_rsp->hdr.mid != MC_MID_CLOSE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       MC_MID_CLOSE_RSP, close_rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(mc_close_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       close_req->hdr.mid, sizeof(mc_close_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (close_rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("mc close failed: ecode %s(0x%x)\n", 
		       tape_ecode_to_string(close_rsp->ecode), close_rsp->ecode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
mc_ui_dbg_level(int argc, char **argv, ...)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_set_debug_level_req_msg_t *req;
	tape_set_debug_level_rsp_msg_t *rsp;

	req = (tape_set_debug_level_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_SET_DEBUG_LEVEL_REQ;
	req->hdr.cookie = 1;
	req->debug_level = strtoul(argv[1], NULL, 0);
			
	req_len = sizeof(tape_set_debug_level_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 5);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		return(NFX_ERR);
	}

	rsp = (tape_set_debug_level_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_SET_DEBUG_LEVEL_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_SET_DEBUG_LEVEL_RSP, rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_set_debug_level_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, sizeof(tape_set_debug_level_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	return(NFX_OK);
}

int 
main(int argc, char **argv)
{
	int	i;
	int	rc;

	if (argc < 2)
		goto arg_error;

	for (i = 0; mc_ui_root[i].name; ++i) {
		if (strcmp(mc_ui_root[i].name, argv[1]) == 0) {
			if (argc-2 < mc_ui_root[i].parm_cnt) 
				goto arg_error;
			else {
				rc = mc_ui_root[i].func(argc-1, argv+1, 
							mc_ui_root[i].implicit_arg1, 
							mc_ui_root[i].implicit_arg2);
				return(rc == NFX_OK ? 0 : 1);
			}
		}
	}

arg_error:
	print_usage(argv[0]);
	exit(1);
}

int
send_tape_message(char *req_buf, int req_len, char *resp_buf, int *resp_len, int timeout)
{
	int		stat;
	struct timeval	to;

	to.tv_sec = timeout;
	to.tv_usec = 0;
	
	stat = sendAgileMsg(-1, 
			    req_buf, req_len,
			    "tape-driver", 0, 0, 0, &to,
			    resp_buf, *resp_len);
	if (stat > 0)
		*resp_len = stat;
	return(stat);
}

char *
tape_ecode_to_string(uint32 ecode)
{
	switch(ecode) {
	case TAPE_ECODE_SUCCESS:			return("SUCCESS");
	case TAPE_ECODE_DEVICE_BUSY:			return("DEVICE_BUSY");
	case TAPE_ECODE_NO_DEVICE:			return("NO_DEVICE");
	case TAPE_ECODE_NO_TAPE_LOADED:			return("NO_TAPE_LOADED");
	case TAPE_ECODE_ILI:				return("ILI");
	case TAPE_ECODE_DATA_SIZE_MISMATCH:		return("DATA_SIZE_MISMATCH");
	case TAPE_ECODE_MEDIUM_ERR:			return("MEDIUM_ERR");
	case TAPE_ECODE_EOM_ENCOUNTERED:		return("EOM_ENCOUNTERED");
	case TAPE_ECODE_EOD_ENCOUNTERED:		return("EOD_ENCOUNTERED");
	case TAPE_ECODE_FM_ENCOUNTERED:			return("FM_ENCOUNTERED");
	case TAPE_ECODE_IO_ERROR:			return("IO_ERROR");
	case TAPE_ECODE_DEVICE_NOT_OPEN:		return("DEVICE_NOT_OPENED");
	case TAPE_ECODE_DEVICE_NOT_OPEN_BY_CLIENT:	return("DEVICE_NOT_OPENED_BY_CLIENT");
	default:				return("???");
	}
}
