/*
 *------------------------------------------------------------------------------
 *
 * Name:	tape-discovery.c
 *
 * RCS:		$Id: tape-discovery.c,v 1.16 2002/11/09 00:08:36 nitinb Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Tape driver device discover and management.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 10/1/01
 *
 *------------------------------------------------------------------------------
 */

#include "tape-common.h"
#include "tape-discovery.h"
#include "tape-msg.h"
#include <sys/time.h>
#include <signal.h>

#define TAPE_DEVICE_INITIAL_POLL_SECS		1
#define TAPE_DEVICE_INTERVAL_POLL_SECS		10
#define TAPE_SDM_TIMEOUT			5

typedef struct _tape_timer_state_ {
	int			timer_sighandler_registered;
	int			timer_running;
	int			timer_popped;
	struct itimerval	timer_val;
	int			timer_sig;
	void (*timer_sighandler)(int);
	void (*timer_start_callout)(struct _tape_timer_state_ *);
	void (*timer_stop_callout)(struct _tape_timer_state_ *);
	void (*timer_popped_callout)(struct _tape_timer_state_ *);
} tape_timer_state_t;

void tape_device_timer_sighandler(int);
void tape_device_timer_stop(tape_timer_state_t *);
void tape_device_timer_start(tape_timer_state_t *);
void tape_device_timer_popped(tape_timer_state_t *);
void tape_compress_trailing_spaces(char *);

tape_timer_state_t * tape_device_timer = NULL;
uint32		     tape_device_poll_count = 0;
tape_dev_t *	     tape_devs = NULL;
ushort16	     tape_index = 1;

void
tape_device_management_init(void)
{
	/* Initialize the device management timer structure */
	tape_device_timer = (tape_timer_state_t *)malloc(sizeof(tape_timer_state_t));
	if (tape_device_timer == NULL)
		panic("malloc failed in tape_device_management_init: %s\n", strerror(errno));
	
	bzero(tape_device_timer, sizeof(tape_timer_state_t));

	tape_device_timer->timer_val.it_interval.tv_sec = TAPE_DEVICE_INTERVAL_POLL_SECS;
	tape_device_timer->timer_val.it_value.tv_sec = TAPE_DEVICE_INITIAL_POLL_SECS;
	tape_device_timer->timer_sig = SIGALRM;
	tape_device_timer->timer_sighandler = tape_device_timer_sighandler;
	tape_device_timer->timer_start_callout = tape_device_timer_start;
	tape_device_timer->timer_stop_callout = tape_device_timer_stop;
	tape_device_timer->timer_popped_callout = tape_device_timer_popped;

	/* Start the timer */
	(*tape_device_timer->timer_start_callout)(tape_device_timer);
}

void
tape_device_management_poll(void)
{
	if (tape_device_timer->timer_popped) {
		--tape_device_timer->timer_popped;
		(*tape_device_timer->timer_popped_callout)(tape_device_timer);
		/* Restart timer */
		// (*tape_device_timer->timer_start_callout)(tape_device_timer); // Really don't need this!!
	}
}

void  
tape_device_timer_sighandler(int sig)
{
	++tape_device_timer->timer_popped;
}

void
tape_device_timer_stop(tape_timer_state_t *timer_state)
{
}

void
tape_device_timer_start(tape_timer_state_t *timer_state)
{
	if (!timer_state->timer_sighandler_registered) {
		if (signal(tape_device_timer->timer_sig, tape_device_timer->timer_sighandler) == SIG_ERR)
			panic("signal failed in tape_device_timer_start.\n");
		tape_device_timer->timer_sighandler_registered = 1;
	}

	if (!timer_state->timer_running) {
		if (setitimer(ITIMER_REAL, &tape_device_timer->timer_val, NULL) < 0)
			panic("setitimer failed in tape_device_timer_start: %s\n", strerror(errno));
		timer_state->timer_running = 1;
	}
}

void
tape_device_timer_popped(tape_timer_state_t *timer_state)
{
	int			i;
	int			rc;
	char			respbuff[1024];
	sdm_devInfoReqMsg_t	dev_info_req;
	sdm_devInfoRspMsg_t	dev_info_rsp;
	sdm_devListReqMsg_t	dev_list_req;
	sdm_devListRspMsg_t *	dev_list_rsp;
	tape_dev_t *		dev;
	struct timeval		to;

	++tape_device_poll_count;

	TP_DBG(10, ("tape_device_timer_popped %d\n", tape_device_poll_count));

	/* 
	 * Step 1: Look for existing tape devices
	 */
	to.tv_sec = TAPE_SDM_TIMEOUT;
	to.tv_usec = 0;

	bzero(&dev_list_req, sizeof(sdm_devListReqMsg_t));
	dev_list_req.hdr.mid = SDM_MID_DEV_LIST_REQ;
	dev_list_req.type = SDM_DEVLIST_TAPE;
	dev_list_req.port = SDM_DEVLIST_PORTALL;
	rc = sendAgileMsg(-1, 
			  &dev_list_req, sizeof(sdm_devListReqMsg_t),
			  "sdm", 0, 0, 0, &to,
			  respbuff, sizeof(respbuff));
	if (rc <= 0) {
		TP_DBG(2, ("%s failed in %s: %s\n",
			   "sendAgileMsg", "tape_device_management_thread", strerror(errno)));
		goto sdm_failure;
	}
		
	dev_list_rsp = (sdm_devListRspMsg_t *)respbuff;

	if (dev_list_rsp->hdr.mid != SDM_MID_DEV_LIST_RSP ) {
		TP_DBG(2, ("Invalid Response Received in %s: %x\n", 
			   "tape_device_management_thread", dev_list_rsp->hdr.mid));
		goto sdm_failure;
	}

	for (i = 0; i < dev_list_rsp->numEntries; i++) {
		if ((dev = tape_device_lookup_by_id(dev_list_rsp->idList[i])) == NULL) {
			/* New device -- go get SDM info for this device */
			dev_info_req.hdr.mid = SDM_MID_DEV_INFO_REQ;
			dev_info_req.id = dev_list_rsp->idList[i];
			bzero(&dev_info_rsp, sizeof(sdm_devInfoRspMsg_t));
			rc = sendAgileMsg(-1, 
					  &dev_info_req, sizeof(sdm_devInfoReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  &dev_info_rsp, sizeof(sdm_devInfoRspMsg_t));
			if (rc < 0) {
				TP_DBG(2, ("%s failed in %s: %s\n",
					   "sendAgileMsg", "tape_device_management_thread", strerror(errno)));
				goto sdm_failure;
			}
			if (dev_info_rsp.hdr.mid != SDM_MID_DEV_INFO_RSP ) {
				TP_DBG(2, ("Invalid Response Received in %s: %x\n",
					   "tape_device_management_thread", dev_info_rsp.hdr.mid));
				goto sdm_failure;
			}
			if (dev_info_rsp.rc == SDM_FAIL ) {
				TP_DBG(2, ("%s failed in %s: Device Not Found %llx\n", 
					   "SDM_MID_DEV_INFO_REQ", "tape_device_management_thread", dev_info_req.id));
				goto sdm_failure;
			}
			
			TP_DBG(1, ("Adding new tape device: id %016llx, WWN %016llx, LUN %d, vendor '%s', model '%s'\n", 
				   dev_info_rsp.id,
				   dev_info_rsp.wwn, dev_info_rsp.lun,
				   dev_info_rsp.st.vendor, dev_info_rsp.st.model));

			dev = tape_device_new(dev_info_rsp.id,
					      dev_info_rsp.wwn, dev_info_rsp.lun,
					      dev_info_rsp.st.vendor, dev_info_rsp.st.model);
			dev->state = TAPE_STATE_CLOSED;

			/* Start device state machine */
			tape_sm_create(dev);
		}

		dev->poll_ref_number = tape_device_poll_count;
	} /* for (i < dev_list_rsp->numEntries) */

	/* 
	 * Step 2: Remove non-existing tape devices
	 */
	if ((dev = tape_devs)) {
		tape_dev_t *	next;
		do {
			next = dev->next;
			if (dev->mc == 0 && dev->poll_ref_number != tape_device_poll_count) {
				TP_DBG(1, ("Deleting old tape device: id %016llx, vendor '%s', model '%s'\n", 
					   dev->id,
					   dev->vendor_name,
					   dev->model_name));
				tape_sm_run(dev, NULL, TAPE_EVENT_DEVICE_REMOVED);
			}
			dev = next;
		} while (tape_devs && dev != tape_devs);
	}

	/* 
	 * Step 1: Look for existing media-changer devices
	 */
	to.tv_sec = TAPE_SDM_TIMEOUT;
	to.tv_usec = 0;

	bzero(&dev_list_req, sizeof(sdm_devListReqMsg_t));
	dev_list_req.hdr.mid = SDM_MID_DEV_LIST_REQ;
	dev_list_req.type = SDM_DEVLIST_STACKER;
	dev_list_req.port = SDM_DEVLIST_PORTALL;
	rc = sendAgileMsg(-1, 
			  &dev_list_req, sizeof(sdm_devListReqMsg_t),
			  "sdm", 0, 0, 0, &to,
			  respbuff, sizeof(respbuff));
	if (rc <= 0) {
		TP_DBG(2, ("%s failed in %s: %s\n",
			   "sendAgileMsg", "tape_device_management_thread", strerror(errno)));
		goto sdm_failure;
	}
		
	dev_list_rsp = (sdm_devListRspMsg_t *)respbuff;

	if (dev_list_rsp->hdr.mid != SDM_MID_DEV_LIST_RSP ) {
		TP_DBG(2, ("Invalid Response Received in %s: %x\n", 
			   "tape_device_management_thread", dev_list_rsp->hdr.mid));
		goto sdm_failure;
	}

	for (i = 0; i < dev_list_rsp->numEntries; i++) {
		if ((dev = tape_device_lookup_by_id(dev_list_rsp->idList[i])) == NULL) {
			/* New device -- go get SDM info for this device */
			dev_info_req.hdr.mid = SDM_MID_DEV_INFO_REQ;
			dev_info_req.id = dev_list_rsp->idList[i];
			bzero(&dev_info_rsp, sizeof(sdm_devInfoRspMsg_t));
			rc = sendAgileMsg(-1, 
					  &dev_info_req, sizeof(sdm_devInfoReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  &dev_info_rsp, sizeof(sdm_devInfoRspMsg_t));
			if (rc < 0) {
				TP_DBG(2, ("%s failed in %s: %s\n",
					   "sendAgileMsg", "tape_device_management_thread", strerror(errno)));
				goto sdm_failure;
			}
			if (dev_info_rsp.hdr.mid != SDM_MID_DEV_INFO_RSP ) {
				TP_DBG(2, ("Invalid Response Received in %s: %x\n",
					   "tape_device_management_thread", dev_info_rsp.hdr.mid));
				goto sdm_failure;
			}
			if (dev_info_rsp.rc == SDM_FAIL ) {
				TP_DBG(2, ("%s failed in %s: Device Not Found %llx\n", 
					   "SDM_MID_DEV_INFO_REQ", "tape_device_management_thread", dev_info_req.id));
				goto sdm_failure;
			}

			TP_DBG(1, ("Adding new media-changer: id %016llx, WWN %016llx, LUN %d, vendor '%s', model '%s'\n", 
				   dev_info_rsp.id,
				   dev_info_rsp.wwn, dev_info_rsp.lun,
				   dev_info_rsp.st.vendor, dev_info_rsp.st.model));

			dev = tape_device_new(dev_info_rsp.id,
					      dev_info_rsp.wwn, dev_info_rsp.lun,
					      dev_info_rsp.st.vendor, dev_info_rsp.st.model);
			dev->mc = 1;
			dev->state = TAPE_STATE_CLOSED;

			/* Start device state machine */
			tape_sm_create(dev);
		}

		dev->poll_ref_number = tape_device_poll_count;
	} /* for (i < dev_list_rsp->numEntries) */

	/* 
	 * Step 2: Remove non-existing media-changer devices
	 */
	if ((dev = tape_devs)) {
		tape_dev_t *	next;
		do {
			next = dev->next;
			if (dev->mc == 1 && dev->poll_ref_number != tape_device_poll_count) {
				TP_DBG(1, ("Deleting old media-changer device: id %016llx, vendor '%s', model '%s'\n", 
					   dev->id,
					   dev->vendor_name,
					   dev->model_name));
				tape_sm_run(dev, NULL, TAPE_EVENT_DEVICE_REMOVED);
			}
			dev = next;
		} while (tape_devs && dev != tape_devs);
	}

 sdm_failure:
}

tape_dev_t *
tape_device_new(scsi_id_t id, uint64 wwn, uint32 lun, char *vendor_name, char *model_name)
{
	tape_dev_t *	dev;
	int32		rc;

	dev = (tape_dev_t *)malloc(sizeof(tape_dev_t));
	if (dev == NULL)
		panic("malloc failed in tape_device_new: %s\n", strerror(errno));
	bzero(dev, sizeof(tape_dev_t));

	dev->id = id;
	dev->wwn = wwn;
	dev->lun = lun;
	memcpy(dev->vendor_name, vendor_name, SDM_VENDOR_NAME_LEN);
	memcpy(dev->model_name, model_name, SDM_MODEL_NAME_LEN);
	tape_compress_trailing_spaces(dev->vendor_name);
	tape_compress_trailing_spaces(dev->model_name);
	dev->cap = tape_find_cap_entry(dev->vendor_name, dev->model_name);
	dev->index = tape_index;
	if (tape_index++ == 0)
		tape_index = 1;

	rc = tape_get_alias(dev->id, &dev->alias);
	if (rc != NFX_OK)
		dev->alias = -1;

	/* Insert into list */
	if (tape_devs == NULL) {
		dev->next = dev->prev = dev;
		tape_devs = dev;
	}
	else {
		dev->prev = tape_devs->prev;
		dev->next = tape_devs->prev->next;
		tape_devs->prev->next = dev;
		tape_devs->prev = dev;
	}

	return(dev);
}

tape_dev_t *
tape_device_lookup_by_id(scsi_id_t id)
{
	tape_dev_t *    dev;

	if ((dev = tape_devs)) {
		do {
			if (id == dev->id)
				return(dev);
			dev = dev->next;
		} while (dev != tape_devs);
	}
	return(NULL);
}

tape_dev_t *
tape_device_lookup_by_client_tag(uint32 ctag, char *type)
{
    tape_dev_t *dev;
    uint32     devType;

    devType = (strcasecmp(type, "tape") == 0) ? 0 : 1;
    if ((dev = tape_devs)) {
        do {
            if (ctag == dev->open_client_tag &&
                devType == dev->mc)
                return(dev);
            dev = dev->next;
        } while (dev != tape_devs);
    }
    return(NULL);
}

tape_dev_t *
tape_device_lookup_by_alias(int32 alias)
{
	tape_dev_t *    dev;

	if ((dev = tape_devs)) {
		do {
			if (alias == dev->alias)
				return(dev);
			dev = dev->next;
		} while (dev != tape_devs);
	}
	return(NULL);
}

tape_dev_t *
tape_device_lookup_by_handle(uint32 handle)
{
	tape_dev_t *    dev;

	if ((dev = tape_devs)) {
		do {
			if (handle == dev->open_handle)
				return(dev);
			dev = dev->next;
		} while (dev != tape_devs);
	}
	return(NULL);
}

tape_dev_t *
tape_device_lookup_by_index(ushort16 index)
{
	tape_dev_t *    dev;

	if ((dev = tape_devs)) {
		do {
			if (index == dev->index)
				return(dev);
			dev = dev->next;
		} while (dev != tape_devs);
	}
	return(NULL);
}

void
tape_device_destroy(tape_dev_t *dev)
{
	dev->prev->next = dev->next;
	dev->next->prev = dev->prev;

	if (tape_devs == dev) {
		if (dev->next == dev)
			tape_devs = NULL;
		else 
			tape_devs = dev->next;
	}

	free(dev);
}

void
tape_compress_trailing_spaces(char *str)
{
	char *p;
	
	if (strlen(str) == 0)
		return;
	p = &str[strlen(str) - 1];
	while (strlen(str) > 0 && *p == ' ')
		*p-- = '\0';
}
