/*
 * Simple test code for threaded state machine.
 */

#include <nfx-incl.h>
#include <exports/esm-api.h>
#ifdef NOTDEF // use as needed
#include <exports/file-api.h>
#endif
#include "threadtest.h"

#define NUM_THREADS  4

static int32 threadtest_u[NUM_THREADS];
static void *threadtest_p1[NUM_THREADS];
static void *threadtest_p2[NUM_THREADS];

int32
threadtest_func()
{
    esm_event_t     esmEvent;
    esm_handle_t    handle[NUM_THREADS];
    int32           rc;
    int32           i;

    for (i = 0; i < NUM_THREADS; i++)
    {
        rc = threadtest_sm_create(&handle[i], &threadtest_u[i]);
        if (rc != NFX_OK)
        {
            printf("do_threadtest: failed to create esm (%d)\n", rc);
            return NFX_ERR;
        }

        esmEvent.event = THREADTEST_EVENT_START;
        esmEvent.p1.u.w = i;
        esmEvent.p2.u.p = NULL;
        rc = esm_deliverEvent(handle[i], &esmEvent);
        if (rc != NFX_OK)
        {
            printf("do_threadtest: failed to deliver start event (%d)\n", rc);
            return NFX_ERR;
        }
    }

    return NFX_OK;
}

int32
threadtest_start(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *smCookie)
{
    esm_event_t event;
    int32       rc, i = thisEvent.p1.u.w;
    int32       *u = smCookie;

    if (u != &threadtest_u[i])
    {
        printf("threadtest_start: wrong smCookie (%p/%p/%d\n",
            u, &threadtest_u[i], i);
        goto done;
    }

    // Enqueue an event so yield will return to us.
    event.event = THREADTEST_EVENT_DONE;
    event.p1.u.p = &threadtest_p1[i];
    event.p2.u.p = &threadtest_p2[i];
    rc = esm_deliverEvent(handle, &event);

    printf("threadtest: calling esm_yield\n");
    esm_yield(&event, 0);

    printf("threadtest: yield returned\n");
    if (event.event != THREADTEST_EVENT_DONE)
    {
        printf("threadtest: incorrect event (%d/%d)\n",
            THREADTEST_EVENT_DONE, event.event);
    }
    if (event.p1.u.p != &threadtest_p1[i])
    {
        printf("threadtest: incorrect p1 (%p/%p)\n",
            &threadtest_p1[i], event.p1.u.p);
    }
    if (event.p2.u.p != &threadtest_p2[i])
    {
        printf("threadtest: incorrect p2 (%p/%p)\n",
            &threadtest_p2[i], event.p2.u.p);
    }

    printf("threadtest: thread %d of %d start succeeded\n",
            i+1, NUM_THREADS);

#ifdef NOTDEF // use as needed
{
    int fd;
    char fileName[80];
    int32 dat[512], dati;
    uint32 offset;

    sprintf(fileName, "vol1:threadTest_%d", i);

    fd = file_create(fileName, 0644);
    if (fd < 0)
    {
        printf("threadtest: couldn't create %s\n", fileName);
        return nextState;
    }

    for (offset = 0; offset < (1024 * 1024 * 4); offset += sizeof dat)
    {
        for (dati = 0; dati < 512; dati++)
        {
            dat[dati] = dati + (offset / 4);
        }
    
        rc = file_write(fd, (void *) dat, sizeof dat);
        if (rc != sizeof dat)
        {
            printf("threadtest: short write (%ld/%d)\n",
                sizeof dat, rc);
            file_close(fd);
            return nextState;
        }
    }

    printf("threadtest: thread %d of %d file_write succeeded\n",
            i+1, NUM_THREADS);

    file_lseek(fd, FILE_SEEK_SET, 0);

    for (offset = 0; offset < (1024 * 1024 * 4); offset += sizeof dat)
    {
        for (dati = 0; dati < 512; dati++)
        {
            dat[dati] = 0x55555555;
        }
    
        rc = file_read(fd, (void *) dat, sizeof dat);
        if (rc != sizeof dat)
        {
            printf("threadtest: short read (%ld/%d)\n",
                sizeof dat, rc);
            file_close(fd);
            return nextState;
        }

        for (dati = 0; dati < 512; dati++)
        {
            if (dat[dati] != dati + (offset / 4))
            {
                printf("threadtest: read err on %s\n",
                    fileName);
                printf("  offset %d exp 0x%x got 0x%x\n",
                    offset + dati, dati + (offset / 4), dat[dati]);
                file_close(fd);
                return nextState;
            }
        }
    }

    printf("threadtest: thread %d of %d file_read succeeded\n",
            i+1, NUM_THREADS);

    file_close(fd);
}
#endif // NOTDEF

done:
    rc = threadtest_sm_destroy(handle);
    if (rc != NFX_OK)
    {
        printf("threadtest_start: destroy failed (%d)\n", rc);
    }

    return nextState;
}

int32
threadtest_create(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *smCookie)
{
    printf("threadtest_create called\n");

    return nextState;

} /* threadtest_create */


int32
threadtest_bad_state(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *smCookie)
{
    printf("threadtest_bad_state called\n");

    return nextState;

} /* threadtest_bad_state */
