/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tftp-api.c
 *
 * RCS:  $Id: tftp-api.c,v 1.5 2002/03/26 20:57:42 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains TFTP API routines.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 09/15/2001
 *
 *-----------------------------------------------------------------
 */
#include    <nfx-incl.h>

#include    <exports/pkt-api.h>
#include    <exports/luc-api.h>
#include    <exports/tpl-api.h>
#include    "../sm-tpl/tpl.h"
#include    <exports/timer-api.h>
#include    <exports/tftp-api.h>
#include    "tftp.h"


tftp_cb_t	*TftpCbPtr;

extern uint32 	myIpaddr;
extern ushort16 nfsPort;

/*-----------------------------------------------------------------
 * Name:        tftp_init
 *
 * Description: Initialize TFTP module related data structure.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
void tftp_init( void )
{

	/*-------------------------------
	 * Allocate TFTP Control Block
	 *-------------------------------
	 */
	TftpCbPtr = (tftp_cb_t *)eee_ramAlloc( sizeof (tftp_cb_t) );
	if (TftpCbPtr == NULL)
	{
		TFTP_LOGERR(TFTP_ERR_NO_RAM_RSRC, TftpCbPtr);	
	}

	memset(TftpCbPtr, 0, sizeof(tftp_cb_t));

	/*-------------------------------
	 * Allocate TFTP Retry Timer
	 *-------------------------------
	 */
	TftpCbPtr->tmr_id = eee_timerCreate( TFTP_RETRY_TIME, EEE_TIMER_ONCE);
	if (TftpCbPtr->tmr_id == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		TFTP_LOGERR(TFTP_ERR_NO_TIMER, TftpCbPtr->tmr_id);	
	}

}

/*-----------------------------------------------------------------
 * Name:        tftp_open
 *
 * Description: Open an TFTP connection for file xfer.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
void *  /* NULL if failed, or ptr to TftpCB */
tftp_open(char    *filename,                        /* filename */
          uint32  srvrIp,                           /* tftp IP address */
          int32   (*ul_dataInd)(char  *pData,       /* RxData callback */
                                int32  size ),
          int32   (*ul_eventInd)(void *tftpCb,      /* Event callback */
                                int32 event) )
{

	tftp_cb_t	*tftpCb;
    uint32      ipaddr;
    
	/*-------------------------------
	 * Use BP IP addr when loading from SSC. 
	 *-------------------------------
	 */
    ipaddr = ((srvrIp & LUC_BP_NETMASK) == LUC_BP_SUBNET) ?
                bp_getIpAddr() : myIpaddr;

	tftpCb = TftpCbPtr;

	/*-------------------------------
	 * Make sure TFTP is initialized.
	 *-------------------------------
	 */
	if (tftpCb == NULL)
	{
		TFTP_LOGERR(TFTP_ERR_NOT_INIT, 0);	
		return NULL;
	}

	/*-------------------------------
	 * App upcalls MUST be present.
	 *-------------------------------
	 */
	if ((ul_dataInd == NULL) || (ul_eventInd == NULL))
	{
		TFTP_LOGERR(TFTP_ERR_NO_FUNC, 0);	
		return NULL;
	}

	/*-------------------------------
	 * Check filename size.
	 *-------------------------------
	 */
	if ( strlen(filename) > (TFTP_MAX_FILENAME - 1))
	{
		return NULL;
	}

	/*-------------------------------
	 * Copy xfer filename.
	 *-------------------------------
	 */
	strcpy(tftpCb->filename, filename);

	/*-------------------------------
	 * Setup Local/remote IP & ports.
	 *-------------------------------
	 */
	tftpCb->laddr	 = ipaddr;
	tftpCb->lport	 = nfsPort-2;
	tftpCb->raddr	 = srvrIp;
	tftpCb->rport	 = TFTP_SRVR_PORT;

	/*-------------------------------
	 * Save app. upcalls.
	 *-------------------------------
	 */
	tftpCb->dataInd  = ul_dataInd;
	tftpCb->eventInd = ul_eventInd;
	tftpCb->flags    = 0;
	tftpCb->block    = 0;
	tftpCb->retry	 = 0;

	/*-------------------------------
	 * Bind with TPL once only.
	 *-------------------------------
	 */
	if (tftpCb->pTplBind == NULL)
	{
#ifdef LINUX_TEST
		if (tftp_simOpen( tftpCb ) < 0)
			return NULL;
#endif

		if (tftp_tplBindReq( tftpCb ) != NFX_OK)
			return NULL;
	}

	/*-------------------------------
	 * Send a read request (RRQ).
	 *-------------------------------
	 */
	tftp_sendRrq( tftpCb );

	return (tftpCb);
}


/*-----------------------------------------------------------------
 * Name:        tftp_close
 *
 * Description: Close & release TFTP related resource.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
void tftp_close( void *tftpCb )
{
	if (((tftp_cb_t*)tftpCb)->pTplConn)
		tpl_delConnReq(((tftp_cb_t *)tftpCb)->pTplConn);
/*
	bsd_CloseSock( ((tftp_cb_t *)tftpCb)->sock);
*/
}

