/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-free.c
 *
 * RCS:  $Id: tpl-free.c,v 1.20 2002/10/22 22:36:39 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer PKT free routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 07/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/tpl-api.h>
#include "tpl.h"

/*-----------------------------------------------------------------
 * Name:  		tpl_freePkt
 *
 * Description: Free pkt allocated using tpl_allocPkt() back to
 *				free pkt list associated with connection.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 *-----------------------------------------------------------------
 */
int32 		/* NFX_OK or error */
tpl_freePkt(void *pConnId, pkt_desc_t *pDesc, char *filename, int32 linenum)
{
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;
	pkt_desc_t		*nextPkt;

	pConn  = (tpl_connCb_t *)pConnId;
	pStats = &pTplCfg->tplStats;

	while (pDesc)
	{
		nextPkt = pDesc->pTplNext;

#ifdef TPL_DBG
		if (nextPkt)
			printf("%s: pDesc=%x, nextPkt=%x, refCnt=%d\n", 
			__FUNCTION__, pDesc, nextPkt, pDesc->refCnt);
#endif

		/*-------------------------------
		 * Make sure connCb is still vaild.
		 *-------------------------------
		 */
		if (!pConn || !(pConn->flags & TPL_CONN_INUSE) ||
			!(pConn->connId == pDesc->tplConnId))
		{
			/*-------------------------------
			 * Conn CB is gone. Free back to 
			 * gen. pool & incr err count.
			 *-------------------------------
			 */
			pStats->cntBadConnErr++;
			TPL_LOGERR( TPL_ERR_BAD_CONN, pDesc );

			goto freeErrPkt;
		}

		/*-------------------------------
		 * Make sure desc is vaild.
		 *-------------------------------
		 */
		if (!pDesc || !(pDesc->flags & PKT_DESC_INUSE))
		{
			pStats->cntBadPktErr++;

			TPL_LOGERR( TPL_ERR_BAD_PKT, pDesc );
			goto freeErrPkt;
		}

		/*-------------------------------
		 * Don't free pkt if refCnt > 0.
		 * Pkt is referenced by IP fragments.
		 *-------------------------------
		 */
	    if (--pDesc->refCnt > 0)
		{
			pStats->cntRefCnt++;
	        goto skipRefPkt;
		}

		/*-------------------------------
		 * Pkt was in driver sent. 
		 * Skip it.
		 *-------------------------------
		 */
		if (pDesc->flags & PKT_USR_LUC_TX)
		{
			pStats->cntInTcpXmt++;
	        goto skipRefPkt;
		}

		/*-------------------------------
		 * Pkt still in TCP TX Q, skip it.
		 *-------------------------------
		 */
	    if (pDesc->flags & PKT_USR_TCP_TX_Q)
		{
			pStats->cntInTcpTxQ++;
	        goto skipRefPkt;
		}

	    pDesc->cod   = 0;

		/*-------------------------------
		 * Make sure PKT is own by TPL
		 *-------------------------------
		 */
		if (!(pDesc->flags & PKT_USR_TPL))
		{
			/*-------------------------------
			 * Not owner. Free back to gen. pool. 
			 *-------------------------------
			 */
			pStats->cntBadPktErr++;
			TPL_LOGERR( TPL_ERR_BAD_PKT, pDesc );

			goto freeErrPkt;
		}


		/* Su, 08/31/02.
		 * free pkt back to connection
		 */
		if (pDesc->flags & PKT_USR_TPL_CONN) {
			
			pkt_descRing_t	*pRing;

			/*-------------------------------
			 * If rpc packet, reset pHdr.
			 *-------------------------------
			 */
			if ((pConn->prot == TPL_PROT_TCP) &&
				(pConn->flags & TPL_CONN_USE_RPC))
				pDesc->pHdr += 4;

			/* clear IP offset bits that may have been set */
			*(ushort16*)(pDesc->pIpHdr + 6) = 0;
			
			/*
			 *---------------------------------------------
			 * If UDP, pHdr may have been moved ahead, reset
			 *----------------------------------------------
			 */
			if ((pConn->prot == TPL_PROT_UDP) && 
				(pDesc->pTplHdr == pDesc->pHdr)) {
				pDesc->pHdr	+= PKT_UDP_HDR_SIZE;
				*(ushort16*)pDesc->pTplHdr		= pConn->lport;
				*(ushort16*)(pDesc->pTplHdr+2)  = pConn->rport;
			}

#if defined(NFX_SMP)
			/* sanity chk, make sure that pkt of a connection
			 * is not freed on NCPU. This ensures pConn->txPktRing
			 * can only be manipulated on ACPU
			 */
			if (CM_IS_TXRX_NCPU)
				panic("%s: connection pkt freed on ncpu\n", __FUNCTION__);

#endif /* NFX_SMP */			

			pRing = &(pConn->txPktRing);

			/* sanity check */
			if ((((pRing->wrAddr+1) == (pRing->rdAddr)) ||
				((pRing->rdAddr == pRing->startAddr) &&
				 (pRing->wrAddr == pRing->endAddr))) && pRing->avail) {
				//printf("panic avail %d, pdesc %lx\n", 
				//	   pRing->avail, pDesc);
				pkt_showDescRing(pRing);
				TPL_LOGERR(TPL_ERR_RING_FULL, pConn);
			}
			

			/* reset flags */
			pDesc->flags   &=  (PKT_DESC_INUSE | PKT_USR_TPL | 
								PKT_FAST_PATH | PKT_USR_TPL_CONN);
			pDesc->refCnt   = 0;
			pDesc->appInfo  = 0;
			pDesc->pRefDesc = 0;
			pDesc->pTplNext = 0;
			pDesc->dataSize = 0;
			pDesc->numOfCod = 0;
			pDesc->padSize  = 0;
			*pRing->wrAddr  = pDesc;


#ifdef PKT_USE_TIME_STAMP
        pkt_freeTimeStamp(pDesc->pTime);
        pDesc->pTime = NULL;
#endif

			/*-------------------------------
			 * Update wr ptr & handle wrap.
			 *-------------------------------
			 */
			if (pRing->wrAddr == pRing->endAddr)
				pRing->wrAddr = pRing->startAddr;
			else
				pRing->wrAddr++; 


			pRing->avail++;
			pStats->cntFreePkt++;
			goto skipRefPkt;
		}
		
freeErrPkt:

//  pDesc->flags    = 0;
	    pDesc->refCnt   = 0;
	    pDesc->appInfo  = 0;
	    pDesc->pRefDesc = 0;
	  	pDesc->pTplNext = 0;
//  pDesc->pCodEntry= 0;
	    pDesc->dataSize = 0;
	    pDesc->numOfCod = 0;

		
	    pDesc->freeFile = filename;
	    pDesc->freeLine = linenum;
	    pDesc->freeLine = linenum;


		PKT_FREE_PACKET(pDesc);

		pStats->cntFreePkt++;

skipRefPkt:
		pDesc = nextPkt;
	} 

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:  		tpl_freePktByApp
 *
 * Description: consult application before free pkt descp
 *
 * Created by : Su Wang
 * 
 * Date Created: 1/04/02
 *-----------------------------------------------------------------
 */
int32 		/* NFX_OK or error */
tpl_freePktByApp( void *pConnId, pkt_desc_t *pDesc )
{
	tpl_bindCb_t	*pBind;
	tpl_connCb_t	*pConn;

	pConn  = (tpl_connCb_t *)pConnId;
	pBind = pConn->pBindCb;
  
#ifdef TPL_DBG
	printf("%s: UPCALL pDesc=%x refDesc=%x, refCnt=%d\n", 
		__FUNCTION__, pDesc, pDesc->pRefDesc, 
			(pDesc->pRefDesc) ? pDesc->pRefDesc->refCnt : pDesc->refCnt);
#endif

	if (pDesc->pRefDesc)
	{
		if (--pDesc->pRefDesc->refCnt > 0)
			return (NFX_OK);

		pDesc = pDesc->pRefDesc;

	}

	/*-------------------------------
	 * Upcall app. for xmt complete.
	 *-------------------------------
	 */
	if ((pBind->ul_xmtCmpInd)(pConn->pAppHdl, pDesc) == TPL_FREE_PKT)
		return TPL_FREE_PACKET(pConnId, pDesc);
	else							/* application holding it */
		return NFX_OK;
}
