/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-loop.c
 *
 * RCS:  $Id: tpl-loop.c,v 1.7 2002/10/17 01:38:56 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer Loopback test related code.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 11/09/2001
 *
 *-----------------------------------------------------------------
 */

#include    <nfx-incl.h>

#include    <exports/pkt-api.h>
#include    <exports/luc-api.h>
#include    <exports/tpl-api.h>

/*-------------------------------
 * TPL loopback header.
 *-------------------------------
 */
typedef struct _tpl_loopHdr_t
{
	uint32 	seqNum;
	uint32	msgLen;
} tpl_loopHdr_t;

/*-------------------------------
 * TPL loopback application CB.
 *-------------------------------
 */
typedef struct _tpl_loopAppCb_t
{
	tpl_bindCb_t	*pBind;
	void			*pConn;
	uint32			rxSeq;
	uint32			txSeq;
} tpl_loopAppCb_t;

/*-------------------------------
 * TPL loopback CB.
 *-------------------------------
 */
typedef struct _tpl_loopCb_t
{
	tpl_loopAppCb_t	srvrApp;	/* server app */
	tpl_loopAppCb_t	clntApp;	/* client app */
	int32			inited;
	int32			pktCnt;
	int32			pktLen;
	uint32			rxPkt;
	uint32			txPkt;
} tpl_loopCb_t;

static tpl_loopCb_t loopCb = {0};

void tpl_loopXmt( tpl_loopAppCb_t *appCb );

/*-----------------------------------------------------------------
 * Name:        tpl_loopRcvPktInd
 *
 * Description: Receive TPL loopback packet.
 *  
 * Created by : Shaun Hui
 *
 * Date Created: 11/09/01
 *-----------------------------------------------------------------
 */
static int32
tpl_loopRcvPktInd( void *pHdl, pkt_desc_t *pDesc)
{
	tpl_loopAppCb_t 	*appCb;

	appCb = pHdl;
	
	if (++loopCb.rxPkt < loopCb.pktCnt )
		tpl_loopXmt(  appCb );
	else
		printf("TPL LOOP DONE\n");

	return TPL_FREE_PKT;
}

/*-----------------------------------------------------------------
 * Name:        tpl_loopXmtCmpInd
 *
 * Description: Handle TPL loopback tranmsit complete.
 *  
 * Created by : Shaun Hui
 *
 * Date Created: 11/09/01
 *-----------------------------------------------------------------
 */
static int32
tpl_loopXmtCmpInd( void *pHdl, pkt_desc_t *pDesc)
{
	return TPL_FREE_PKT;
}

/*-----------------------------------------------------------------
 * Name:        tpl_loopAddConnInd
 *
 * Description: Handle TPL add connection.
 *  
 * Created by : Shaun Hui
 *
 * Date Created: 11/09/01
 *-----------------------------------------------------------------
 */
static int32
tpl_loopAddConnInd( tpl_bindCb_t    *pBind,
                    void            *pConnHdl, 
                    void            **pAppHdl,
                    uint32          raddr,
                    ushort16        rport )
{
	tpl_loopAppCb_t	*appCb;

	appCb = (pBind == loopCb.srvrApp.pBind) ?  
			&loopCb.srvrApp : &loopCb.clntApp;

	appCb->pConn = pConnHdl;
	*pAppHdl     = appCb;
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:        tpl_loopDelConnInd
 *
 * Description: Handle TPL del connection.
 *  
 * Created by : Shaun Hui
 *
 * Date Created: 11/09/01
 *-----------------------------------------------------------------
 */
static int32
tpl_loopDelConnInd( void *pHdl, void *pConn)
{
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:        tpl_loopXmt
 *
 * Description: Transmit TPL loopback packet.
 *  
 * Created by : Shaun Hui
 *
 * Date Created: 11/09/01
 *-----------------------------------------------------------------
 */
void tpl_loopXmt( tpl_loopAppCb_t *appCb )
{
	pkt_desc_t		*pDesc;
	tpl_loopHdr_t	*pLoopHdr;

	pDesc = TPL_ALLOC_PACKET(appCb->pConn, 0);
	if (!pDesc)
	{
		printf("TPL_LOOP TEST: NO PKT!\n");
		return;
	}

	pLoopHdr = (tpl_loopHdr_t *)pDesc->pHdr;
	pLoopHdr->seqNum = appCb->txSeq++;
	pLoopHdr->msgLen = loopCb.pktLen;

	/*-------------------------------
	 * Exclude MAC, IP & UDP headers.
	 *-------------------------------
	 */
	pDesc->hdrSize = loopCb.pktLen - PKT_ETHER_HDR_SIZE - 
					 PKT_IP_HDR_SIZE - PKT_UDP_HDR_SIZE;

	tpl_xmtPkt(appCb->pConn, pDesc);

	loopCb.txPkt++;
}

extern uint32 	myIpaddr;
extern ushort16 nfsPort;
extern int		loop_srvr_sock;
extern int		loop_clnt_sock;


int32 tpl_loopTestStart( int32 pktLen, int32 pktCnt )
{
	uint32 			loopAddr; 
	int32			rc = NFX_OK;
	tpl_bindCb_t	*pBindRsp;
	tpl_bindCb_t	loopBindReq;

	loopCb.pktLen = pktLen;
	loopCb.pktCnt = pktCnt;
	loopCb.txPkt  = 0;
	loopCb.rxPkt  = 0;
	loopAddr = myIpaddr;

	if (loopCb.inited )
		goto tpl_loopSend;

	/*-------------------------------
	 * Setup loopback server app.
	 *-------------------------------
	 */
    memset( &loopBindReq, 0, sizeof(tpl_bindCb_t));

	loopBindReq.laddr 		  = htonl(loopAddr);
	loopBindReq.lport 		  = htons(nfsPort-3);
	loopBindReq.raddr 		  = 0;
	loopBindReq.rport 		  = 0;
	loopBindReq.tplProt 	  = TPL_PROT_UDP;
	loopBindReq.ul_addConnInd = tpl_loopAddConnInd;
	loopBindReq.ul_delConnInd = tpl_loopDelConnInd;
	loopBindReq.ul_rcvPktInd  = tpl_loopRcvPktInd;
	loopBindReq.ul_xmtCmpInd  = tpl_loopXmtCmpInd;

	rc = tpl_bindReq(&loopBindReq, &pBindRsp, TPL_BIND_SERVER);
	if (rc != NFX_OK)
	{
		return (rc);
	}
	loopCb.srvrApp.pBind = pBindRsp;
#ifdef LINUX_TEST
	loop_srvr_sock = tpl_loopSimOpen(myIpaddr, nfsPort-3 );
#endif
	
	/*-------------------------------
	 * Setup loopback client  app.
	 *-------------------------------
	 */
    memset( &loopBindReq, 0, sizeof(tpl_bindCb_t));

	loopBindReq.laddr 	 	  = htonl(loopAddr);
	loopBindReq.lport 	 	  = htons(nfsPort-4);
	loopBindReq.raddr	 	  = htonl(loopAddr);
	loopBindReq.rport 	 	  = htons(nfsPort-3);
	loopBindReq.tplProt 	  = TPL_PROT_UDP;
	loopBindReq.ul_addConnInd = tpl_loopAddConnInd;
	loopBindReq.ul_delConnInd = tpl_loopDelConnInd;
	loopBindReq.ul_rcvPktInd  = tpl_loopRcvPktInd;
	loopBindReq.ul_xmtCmpInd  = tpl_loopXmtCmpInd;

	rc = tpl_bindReq(&loopBindReq, &pBindRsp, TPL_BIND_CLIENT);
	if (rc != NFX_OK)
	{
		return (rc);
	}
	loopCb.clntApp.pBind = pBindRsp;
#ifdef LINUX_TEST
	loop_clnt_sock = tpl_loopSimOpen(myIpaddr, nfsPort-4 );
#endif
	loopCb.inited = 1;

tpl_loopSend:
	/*-------------------------------
	 * Send a batch loopback pkts.
	 *-------------------------------
	 */
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);
	tpl_loopXmt(  &loopCb.clntApp);

	return rc;
}

#ifdef LINUX_TEST

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/udp.h>


/*-----------------------------------------------------------------
 * Name:        tpl_loopSimOpen
 *
 * Description: Open linux sockets in simulated env. for TPL Loopback.
 *  
 * Created by : Shaun Hui
 *
 * Date Created: 11/09/01
 *-----------------------------------------------------------------
 */
int
tpl_loopSimOpen( uint32 laddr, ushort16 lport )
{
	int					sock;
    struct sockaddr_in  sock_addr;

    /*-------------------------------
     * Get TFTP UDP socket.
     *-------------------------------
     */
    sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (sock < 0)
    {
        printf ("can't open UDP socket: %s\n", sys_errlist[errno]);
        return -1;
    }

    /*-------------------------------
     * Make it non-blocking.
     *-------------------------------
     */
    if (fcntl(sock, F_SETFL, FNDELAY) < 0)
    {
        printf ("can't fcntl socket: %s\n", sys_errlist[errno]);
        return -1;
    }

    /*-------------------------------
     * Bind the socket.
     *-------------------------------
     */
    bzero( (char *) &sock_addr, sizeof(sock_addr) );
    sock_addr.sin_family      = AF_INET;
    sock_addr.sin_addr.s_addr = htonl(laddr);
    sock_addr.sin_port        = htons(lport);

    if (bind(sock, (struct sockaddr *)&sock_addr, sizeof(sock_addr)) < 0)
    {
        printf ("can't bind socket: %s\n", errno, sys_errlist[errno]);
        return -1;
    }

    return (sock);
}
#endif

