
/*
 *-----------------------------------------------------------------
 *
 * Name:	tssccomm.c
 *
 * RCS:  $Id: tssccomm.c,v 1.7 2002/10/22 01:30:43 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Communication between Toshiba and SSC
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	01/31/2002
 *
 *-----------------------------------------------------------------
 */


#if defined(__OpenBSD__)
#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <net/if.h>
#endif

#include <nfx-incl.h>
#include <exports/eee-desc.h>
#if !defined(__OpenBSD__)
#include <exports/libc.h>
#endif
#include "tssccomm.h"


#if !defined(__OpenBSD__)
extern void panic(char *, ...);
#endif

#if defined(__OpenBSD__)
#include <sys/types.h>
#include <sys/systm.h>
#endif

#if defined(__OpenBSD__)
#if defined(DEBUG)
extern int debug_ethereee;
#define debug(x) { if (debug_ethereee) printf x; }
#else
#define debug(x)
#endif
#else
#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif
#endif

/*
 * tsc_ssc_tx_ring_addr - address of tx ring on ssc
 */
tsc_message_t *
tsc_ssc_tx_ring_addr()
{
	return (tsc_message_t *)0x8fed0000;
}


/*
 * tsc_ssc_rx_ring_addr - address of rx ring on ssc
 */
tsc_message_t *
tsc_ssc_rx_ring_addr()
{
	return tsc_ssc_tx_ring_addr() + TSC_RING_LENGTH + 10;
}


/*
 * tsc_make_tx_message - make message to be send
 */
void
tsc_make_tx_message(tsc_message_t *message, uint32 flags, uint32 bytecount,
					data_descriptor_t *data, uint32 offset)
{
	message->header.flags = flags;
	message->header.bytecount = bytecount;
	tsc_copy_message_data(tsc_message_data(message), data, offset, bytecount);
}


/*
 * tsc_receiver_owns -
 */
int
tsc_receiver_owns(tsc_message_t *message)
{
	return message->header.flags & TSC_MESSAGE_FLAG_RECEIVER_OWNS;
}


/*
 * tsc_message_data - return the pointer to message data
 */
char *
tsc_message_data(tsc_message_t *message)
{
	return (char *)message + message->header.data_offset;
}


/*
 * tsc_receiver_cleanup - cleanup after receiving a message
 */
void
tsc_receiver_cleanup(tsc_message_t *message)
{
	message->header.flags &= ~TSC_MESSAGE_FLAG_RECEIVER_OWNS;
}


/*
 * tsc_is_first_message - return true if it is the first fragment in packet
 */
int
tsc_is_first_message(tsc_message_t *message)
{
	return message->header.flags & TSC_MESSAGE_FLAG_FIRST;
}


/*
 * tsc_is_last_message - return true if it is the last fragment in packet
 */
int
tsc_is_last_message(tsc_message_t *message)
{
	return message->header.flags & TSC_MESSAGE_FLAG_LAST;
}


/*
 * tsc_next_message_index -
 */
int
tsc_next_message_index(tsc_message_t *message)
{
	return message->header.next_index;
}


/*
 * tsc_message_length -
 */
int
tsc_message_length(tsc_message_t *message)
{
	return message->header.bytecount;
}


/*
 * tsc_next_tx_message
 */
tsc_message_t *
tsc_next_tx_message(struct tsc_sc *sc, tsc_message_t *message)
{
	return sc->first_tx_message + tsc_next_message_index(message);
}


/*
 * tsc_last_rx_message - message after the last received
 */
tsc_message_t *
tsc_last_rx_message(struct tsc_sc *sc)
{
	return sc->rx_message;
}


/*
 * tsc_set_message_flags -
 */
void
tsc_set_message_flags(tsc_message_t *message, uint32 flags)
{
	message->header.flags = flags;
}


/*
 * tsc_set_last_rx_message -
 */
void
tsc_set_last_rx_message(struct tsc_sc *sc, tsc_message_t *message)
{
	sc->rx_message = message;
}


/*
 * tsc_next_rx_message -
 */
tsc_message_t *
tsc_next_rx_message(struct tsc_sc *sc, tsc_message_t *message)
{
	return sc->first_rx_message + tsc_next_message_index(message);
}


/*
 * tsc_calculate_packet_length - calculate the packet length
 */
int
tsc_calculate_packet_length(struct tsc_sc *sc,
							 tsc_message_t *start_message)
{
	int len = 0;
	tsc_message_t *message;
	
	assert(tsc_receiver_owns(start_message)
		   && tsc_is_first_message(start_message));

	for (message = start_message;
		 tsc_receiver_owns(message) && !tsc_is_last_message(message);
		 message = tsc_next_rx_message(sc, message)) {
		len += tsc_message_length(message);
	}

	assert(tsc_is_last_message(message));

	len += tsc_message_length(message);

	return len;
}


/*
 * tsc_rx_cleanup_packet - cleanup after receiving packet
 */
void
tsc_rx_cleanup_packet(struct tsc_sc *sc, tsc_message_t *start_message)
{
	tsc_message_t *message;
	
	assert(tsc_receiver_owns(start_message) &&
		   tsc_is_first_message(start_message));

	for (message = start_message;
		 tsc_receiver_owns(message) && !tsc_is_last_message(message);
		 message = tsc_next_rx_message(sc, message)) {
		tsc_receiver_cleanup(message);
	}

	assert(tsc_is_last_message(message));

	tsc_receiver_cleanup(message);

	tsc_set_last_rx_message(sc, tsc_next_rx_message(sc, message));
}


/*
 * tsc_first_message_flags - flags to set for the first message in packet
 */
const uint32
tsc_first_message_flags()
{
	return TSC_MESSAGE_FLAG_RECEIVER_OWNS | TSC_MESSAGE_FLAG_FIRST;
}

/*
 * tsc_first_and_last_message_flags - flags to set for the first message
 * in packet which is also the last
 */
const uint32
tsc_first_and_last_message_flags()
{
	return TSC_MESSAGE_FLAG_RECEIVER_OWNS | TSC_MESSAGE_FLAG_FIRST
		| TSC_MESSAGE_FLAG_LAST;
}


/*
 * tsc_last_tx_message - message after the last transmitted
 */
tsc_message_t *
tsc_last_tx_message(struct tsc_sc *sc)
{
	return sc->tx_message;
}


/*
 * tsc_set_last_tx_message -
 */
void
tsc_set_last_tx_message(struct tsc_sc *sc, tsc_message_t *message)
{
	sc->tx_message = message;
}


/*
 * tsc_check_buffers_availability - check if there are enough free message
 * buffers for packet of length specified by the argument length
 */
int
tsc_check_buffers_availability(struct tsc_sc *sc, int length)
{
	tsc_message_t *message = tsc_last_tx_message(sc);
	int available_length = 0;

	debug(("tsc_check_buffers: idx=%d length=%d\n",
		   (unsigned int)(message - sc->first_tx_message), length));

	if (length <= TSC_MESSAGE_DATA_SIZE * TSC_RING_LENGTH) {
		while (!tsc_receiver_owns(message) && available_length < length) {
			available_length += TSC_MESSAGE_DATA_SIZE;
			message = tsc_next_rx_message(sc, message);
		}
	}

	return available_length >= length;
}


/*
 * tsc_txput_packet - insert the packet to the ring
 */
int
tsc_txput_packet(struct tsc_sc *sc, data_descriptor_t *data)
{
	int rc = 0;
	uint32 flags = TSC_MESSAGE_FLAG_RECEIVER_OWNS;
	tsc_message_t *first_message = tsc_last_tx_message(sc);
	tsc_message_t *last_message = tsc_last_tx_message(sc);
	int bytecount = tsc_data_length(data);
	int offset = 0;

	debug(("tsc_txput_packet\n"));

	while (bytecount > TSC_MESSAGE_DATA_SIZE &&
		   !tsc_receiver_owns(last_message)) {
		tsc_make_tx_message(last_message, flags, TSC_MESSAGE_DATA_SIZE,
						data, offset);
		bytecount -= TSC_MESSAGE_DATA_SIZE;
		offset += TSC_MESSAGE_DATA_SIZE;
		last_message = tsc_next_tx_message(sc, last_message);
	}

	flags |= TSC_MESSAGE_FLAG_LAST;
	if (!tsc_receiver_owns(last_message)) {
		tsc_make_tx_message(last_message, flags, bytecount, data, offset);

		/* When the message is ready, set the flag for the first message */
		tsc_set_message_flags(first_message,
							  last_message == first_message ?
							  tsc_first_and_last_message_flags() :
							  tsc_first_message_flags());

		/* Adjust the pointer to last message send */
		tsc_set_last_tx_message(sc, tsc_next_tx_message(sc, last_message));

		/* All messages put to tx ring successfully */
		rc = 1;
	}
	else {
		tsc_message_t *cleanup_message;
		for (cleanup_message = first_message;
			 cleanup_message != last_message;
			 cleanup_message = tsc_next_tx_message(sc, cleanup_message)) {
			tsc_receiver_cleanup(cleanup_message);
		}
	}

	return rc;
}


/*
 * tsc_rxpacket_ready - return true if there is a packet available for
 * receiver
 */
int
tsc_rxpacket_ready(struct tsc_sc *sc)
{
	tsc_message_t *message = tsc_last_rx_message(sc);
	return tsc_receiver_owns(message) && tsc_is_first_message(message);
}


data_descriptor_t *
tsc_do_rxmessage(struct tsc_sc *sc)
{
	assert(tsc_rxpacket_ready(sc));

	return tsc_make_data_descriptor(sc, tsc_last_rx_message(sc));
}


/*
 * tsc_copy_from_packet - copy the data from the packet into buffer
 * specified by the argument buf. The argument start_message_ptr points
 * to the first message in the packet, offset_ptr to offset in this message.
 * On return the start_message_ptr and offset_ptr are adjusted to point
 * to the next byte after the last copied.
 */
void
tsc_copy_from_packet(char *buf, struct tsc_sc *sc,
				  tsc_message_t **start_message_ptr,
				  int *offset_ptr, int len)
{
	int offset = *offset_ptr;
	tsc_message_t *message = *start_message_ptr;
	char *cp = buf;

	assert(offset < tsc_message_length(message));
	
	while (len) {
		int copy_len = MIN(tsc_message_length(message) - offset, len);
		memcpy(cp, tsc_message_data(message) + offset, copy_len);
		assert(copy_len + offset <= tsc_message_length(message));
		if (copy_len + offset == tsc_message_length(message)) {
			offset = 0;
			message = tsc_next_rx_message(sc, message);
		}
		else
			offset += copy_len;
		len -= copy_len;
	}

	*start_message_ptr = message;
	*offset_ptr = offset;
}


/*
 * tsc_init_tx_message - initialize message in tx ring
 */
void
tsc_init_tx_message(tsc_message_t *message, int next_idx)
{
	tsc_set_message_flags(message, 0);
	message->header.next_index = next_idx;
	message->header.bytecount = 0;
	message->header.data_offset = &message->data[0] - (char *)message;
}

/*
 * tsc_init_rx_ring - initialize transmission ring, start is the first
 * message in the ring, len is number of messages
 */
void
tsc_init_tx_ring(tsc_sc_t *sc, tsc_message_t *start, int len)
{
	tsc_message_t *message;
	tsc_message_t *last;
	
	debug(("tsc_init_tx_ring: %x\n", start));

	sc->first_tx_message = start;
	sc->tx_message = start;
	last = sc->first_tx_message + len - 1;
	for (message = start; message < last; ++message) {
		tsc_init_tx_message(message, (message - sc->first_tx_message) + 1);
	}
	tsc_init_tx_message(message, 0);
}


/*
 * tsc_reinit_tx_ring - reinitialize tx ring after the system restart
 */
void
tsc_reinit_tx_ring(tsc_sc_t *sc, tsc_message_t *start,
				   tsc_message_t *last, int len)
{
	debug(("tsc_reinit_tx_ring: %x, %x\n", start, last));
	sc->first_tx_message = start;
	sc->tx_message = last;
}


/*
 * tsc_init_rx_ring - initialize receive ring
 */
void
tsc_init_rx_ring(tsc_sc_t *sc, tsc_message_t *start, int len)
{
	debug(("tsc_init_rx_ring: %x\n", start));
	sc->first_rx_message = start;
	sc->rx_message = start;
}


/*
 * tsc_reinit_rx_ring - reinitialize receive ring
 */
void
tsc_reinit_rx_ring(tsc_sc_t *sc, tsc_message_t *start,
				   tsc_message_t *last, int len)
{
	debug(("tsc_reinit_rx_ring: %x, %x\n", start, last));
	sc->first_rx_message = start;
	sc->rx_message = last;
}


/*
 * tsc_print_message - print the message
 */
void
tsc_print_message(tsc_message_t *message)
{
	int i;
	
	printf("f=%x cnt=%d n=%d", message->header.flags,
		   message->header.bytecount, message->header.next_index);
	for (i = 0; i < 20; ++i)
		printf(" %02x", *(tsc_message_data(message) + i));
	printf("\n");
}


/*
 * tsc_print_txqueue - print transmit ring
 */
void
tsc_print_txqueue(tsc_sc_t *sc)
{
	tsc_message_t *message;
	printf("tsc_txqueue at %p\n", sc->first_tx_message);
	tsc_print_message(sc->first_tx_message);
	for (message = tsc_next_tx_message(sc, sc->first_tx_message);
		 message != sc->first_tx_message;
		 message = tsc_next_tx_message(sc, message)) {
		tsc_print_message(message);
	}
}


/*
 * tsc_print_rxqueue - print receive ring
 */
void
tsc_print_rxqueue(tsc_sc_t *sc)
{
	tsc_message_t *message;
	printf("tsc_rxqueue at %p\n", sc->first_rx_message);
	tsc_print_message(sc->first_rx_message);
	for (message = tsc_next_rx_message(sc, sc->first_rx_message);
		 message != sc->first_rx_message;
		 message = tsc_next_rx_message(sc, message)) {
		tsc_print_message(message);
	}
}
