/*------------------------------------------------------------------------------
 * RCS:  $Id: utf-case.c,v 1.2 2002/04/12 14:57:31 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : utf-case.c
 *
 * Description  : Unicode Transformation Format (UTF).
 *                Case Mapping.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-APR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "utf-api.h"
#include "utf-unicodeData.h"

/*------------------------------------------------------------------------------
 * Local functions prototypes.
 *------------------------------------------------------------------------------
 */

static int
utf_findChar(utf32 Code);

static void
utf_printfUTF32str(utf32 *  source,
                   uint32   sourceLen);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

static int
utf_findChar(utf32 Code)
{
    int B = 0;
    int E = utf_caseMapTableLength-1;    
    int i;
    
    while (E>B)
    {
        i = (B+E)/2;
        if (utf_caseMapTable[i].codePoint<Code)
        {
            B=i+1;
        }
        else if (utf_caseMapTable[i].codePoint>Code)
        {
            E=i-1;
        }
        else
        {
            // Found
            return i;
        }
    } // while

    if (utf_caseMapTable[B].codePoint==Code)
    {
        return B;
    } else
    {
        // NOT FOUND
        return -1;
    }
}

/*----------------------------------------------------------------------------*/

static void
utf_printfUTF32str(utf32 *  source,
                   uint32   sourceLen)

{
    int i;
    
    for (i = 0; i < sourceLen; i++)
    {
        printf("%x-", source[i]);
    } // for
    printf("\n");
    return;
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/* Convert UTF32 character to upper case.
 */
utf32
utf_toUpper(utf32   c)

{
    utf32   upper = c;
    int     idx = utf_findChar(c);
    
    if (idx != -1)
    {
        upper = utf_caseMapTable[idx].upperCase;
    }
    return(upper);
}

/* Convert UTF32 character to lower case.
 * Note XXX
 */
utf32
utf_toLower(utf32   c)

{
    utf32   lower = c;
    int     idx = utf_findChar(c);
    
    if (idx != -1)
    {
        lower = utf_caseMapTable[idx].lowerCase;
    }
    return(lower);
}

/* Convert UTF32 atring to upper case.
 */
void
utf_UTF32strToUpper(const utf32 *   source,
                    const int       sourceLen,
                    utf32 *         target)

{
    int i;

    for (i = 0; i < sourceLen; i++)
    {
        target[i] = utf_toUpper(source[i]);
    } // for
    
    return;
}

/* Return TRUE if the string is entirely composed of Ascii characters.
 */
boolean
utf_isAsciiString(utf8 *    source,
                  int       len)
{
    byte *  b = source;
    int     i;
    
    for (i = 0; i < len; i++)
    {
        if ((*b) & 0x80)
        {
            return(FALSE);
        }
        b++;
    } // for
        
    return(TRUE);
}

/* Convert UTF8 string to upper case.
 * To do so:
 *    1. convert UTF8 string to UTF32
 *    2. uppercase UTF32 string
 *    3. convert back UTF32 string to UTF8
 * If the string is only omposed of ASCII characters,
 * then skip the conversion to UTF32.
 */
void
utf_UTF8strToUpper(utf8 *   source,     // input
                   uint32   sourceLen,  // input
                   utf8 **  target,     // output
                   uint32 * targetLen)  // output

{
#ifdef ISABELLE // debug only!!!
    printf("source = %s\n", source);
#endif // ISABELLE
        
    /* If the string is only omposed of ASCII characters,
     * then skip the conversion to UTF32.
     */
    if (utf_isAsciiString(source, sourceLen) == TRUE)
    {
        int     i;
        
#ifdef ISABELLE // debug only!!!
        printf("source is an ascii string.\n");
#endif // ISABELLE
        *targetLen = sourceLen;
        *target = (utf8 *) eee_ramAllocZero(((*targetLen) * sizeof(utf8)) + 1);
        for (i = 0; i < *targetLen; i++)
        {
            (*target)[i] = toupper(source[i]);
        } // for
    }
    else
    {
        utf_conversionResult_t  rc;
        utf32 * str1 = 0;
        int     str1Len = 2 * sourceLen; // best guess!
        utf32 * str2 = 0;
        int     str2Len = 0;

        str1 = (utf32 *) eee_ramAllocZero(str1Len * sizeof(utf32));
        
        rc = utf_convertUTF8toUTF32(source, sourceLen,
                                    str1, &str1Len);
        if (rc != UTF_CONVERSION_OK)
        {
            printf("utf_convertUTF8toUTF32() failed = %d\n", rc);
            return;
        }
#ifdef ISABELLE // debug only!!!
        printf("str1 = ");
        utf_printfUTF32str(str1, str1Len);
#endif // ISABELLE

        str2Len = str1Len;
        str2 = (utf32 *) eee_ramAllocZero(str2Len * sizeof(utf32));
        
        utf_UTF32strToUpper(str1, str1Len, str2);
#ifdef ISABELLE // debug only!!!
        printf("str2 = ");
        utf_printfUTF32str(str2, str2Len);
#endif // ISABELLE

        *targetLen = sourceLen; // is that always true???
        *target = (utf8 *) eee_ramAllocZero((*targetLen) * sizeof(utf8));
        rc = utf_convertUTF32toUTF8(str2, str2Len,
                                    *target, targetLen,
                                    UTF_STRICT_CONVERSION);
        if (rc != UTF_CONVERSION_OK)
        {
            printf("utf_convertUTF32toUTF8() failed = %d\n", rc);
            return;
        }
    }
    
#ifdef ISABELLE // debug only!!!
    printf("target = %s\n", *target);
#endif // ISABELLE
    return;
}

/*----------------------------------------------------------------------------*/

#ifdef NO_SUCH_DEF
/* Test program.
 */
int main(void)
{
    utf8 *  utf8_str1 = 0;
    utf8 *  utf8_str2 = 0;
    uint32  utf8_str2_len = 0;
    int     rc;
    int     i;

    utf8_str1 = (utf8 *) eee_ramAllocZero(50 * sizeof(utf8));
    
    strcpy((char *) utf8_str1, "IsaBeLle can cOunt 123");
    printf("utf8_str1 = %s\n", utf8_str1);
    utf_UTF8strToUpper(utf8_str1, strlen(utf8_str1),
                       &utf8_str2, &utf8_str2_len);
    printf("utf8_str2 = %s\n", utf8_str2);

    eee_ramDealloc(utf8_str1);
    eee_ramDealloc(utf8_str2);

    exit(0);
}
#endif // NO_SUCH_DEF

/*----------------------------------------------------------------------------*/
