/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:         vs-mem-api.h
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description:  Virtual server memory management API function definitions.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/09/2002
 *
 *-----------------------------------------------------------------
 */

#ifndef _VS_MEM_API_H
#define _VS_MEM_API_H

#include <nfx-incl.h>
#include <exports/malloc-api.h>

/*-------------
 * Memory Index
 *-------------
 */
 
/* Virtual server index */
typedef uint8    vs_mem_srvr_idx_t;
 
#define VS_MEM_SRVR_IDX_ANY 0    /*FIXME: TBD*/
 
/* Data types index number. Part of vs_mem_index_t. */
typedef enum vs_mem_dataType {
    vs_mem_dt_cifsLockOwner    = 0,      /*lock_cifsLockOwner_t*/
    vs_mem_dt_nlmRecOnVolCB    = 1,      /*lock_nlmRecOnVolCB_t*/
    vs_mem_dt_nlmLockClient    = 2,      /*lock_nlmLockClient_t*/
    vs_mem_dt_nlmLockOwner     = 3,      /*lock_nlmLockOwner_t*/
    vs_mem_dt_cifsLockInfo     = 4,      /*lock_cifsLockInfo_t*/
    vs_mem_dt_nlmLockInfo      = 5,      /*lock_nlmLockInfo_t*/
    vs_mem_dt_exLock           = 6,      /*lock_exLock_t*/
    vs_mem_dt_shLock           = 7,      /*lock_shLock_t*/
    vs_mem_dt_shNode           = 8,      /*lock_shNode_t*/
    vs_mem_dt_pendingLock      = 9,      /*lock_pendingLock_t*/
    vs_mem_dt_fileLock         = 10,     /*lock_fileLock_t*/
	vs_mem_dt_lockDependency   = 11,     /*lock_dependency_t*/
	vs_mem_dt_lockOwnerTimeout = 12,     /*lock_ownerTimeout_t*/
	vs_mem_dt_nlm_request_t    = 13,     /*nlm_request_t*/
    vs_mem_dt_max              = 14,     /*vs_mem_dt_max <= 0xFFFF*/
} vs_mem_dataType_t;
 
typedef struct vs_mem_index {
    /*Virtual server index, 0 - 255*/
    vs_mem_srvr_idx_t srvrIdx;
#define VS_MEM_MAX_SRVR_IDX 256       /*Max number of virtual servers*/

#ifdef VS_MEM_DEBUG
    /*Generation number used to check if the index has already been
	 *reused or not. This can be turned on for debugging. Two bits are
	 *probably enough to catch it when it happens.*/
	uint8             genNb : 2;
	uint8             spare : 6;
#else
    uint8             spare;
#endif /*VS_MEM_DEBUG*/

    /*Data structure type index, 0 - 65535*/
    vs_mem_dataType_t dataType : 16;

    /*Index of the data structure of a particular type for a particular
      virtual server in the memory. 1 - 0xFFFFFFFF*/
    uint32            idx;
#define VS_MEM_MAX_INDEX  0xFFFFFFFF  /*Max index value of one data
                                        structure type on a virtual server*/
} vs_mem_index_t;


/*-----
 * APIs
 *-----
 */

/* Set an index to zero.
 * Argument:
 *     Index to a memory unit.
 * Return value:
 *     None.
 */
#define VS_MEM_IDX_SET_NULL(_idx_)                                          \
    {                                                                       \
        (_idx_).srvrIdx = 0;                                                \
        (_idx_).dataType = 0;                                               \
        (_idx_).idx = 0;                                                    \
    }

/* Check if an index is zero or not.
 * Argument:
 *     Index to a memory unit.
 * Return value:
 *     TRUE - If index is 0.
 *     FALSE - If index is not 0.
 */
#define VS_MEM_IDX_IS_NULL(_idx_)                                           \
    ((_idx_).idx == 0)

/* Check if an index is valid or not.
 * Argument:
 *     Index to a memory unit.
 * Return value:
 *     TRUE - If index to the memory unit is valid.
 *     FALSE - If index to the memory unit is invalid.
 */
#define VS_MEM_IDX_IS_VALID(_idx_)                                          \
        (vs_mem_idxToPtr(_idx_) != NULL)

/* Check if two indices are equal or not.
 * Argument:
 *     Indices to two memory units.
 * Return value:
 *     TRUE - If two indices are equal.
 *     FALSE - If two indices are not equal.
 */
#define VS_MEM_IDX_IS_EQUAL(_idx1_, _idx2_)                                 \
    (((_idx1_).srvrIdx == (_idx2_).srvrIdx) &&                              \
     ((_idx1_).dataType == (_idx2_).dataType) &&                            \
     ((_idx1_).idx == (_idx2_).idx))                                        \

/* Get data type index.
 * Argument:
 *     Index to a memory unit.
 * Return value:
 *     dataType part of mem_index_t structure.
 */
#define VS_MEM_IDX_GET_DATA_TYPE(_idx_)                                     \
    ((_idx_).dataType)


/*-----------------------------------------------------------------
 * Name:         vs_mem_alloc
 *
 * Description:
 *
 *     Allocate a free memory unit of a given data type.
 *
 * Arguments:    
 *
 *     srvrIdx - Virtual server index.
 *
 *     dataType - Data type index.
 *
 *     index - This is an output. It it the index to the allocated
 *             memory unit.
 *
 * Return Value:
 *
 *    Pointer to the allocated data structure - If succeed.
 *
 *    NULL - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void *
vs_mem_alloc(vs_mem_srvr_idx_t srvrIdx,
             vs_mem_dataType_t dataType,
             vs_mem_index_t    *index);

/*-----------------------------------------------------------------
 * Name:         vs_mem_free
 *
 * Description:
 *
 *     Free a memory unit.
 *
 * Arguments:    
 *
 *     index - The index to the memory unit to be freed.
 *
 * Return Value:
 *
 *    NFX_OK - If succeed.
 *
 *    NFX_ERR - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
int32
vs_mem_free(vs_mem_index_t index);

/*-----------------------------------------------------------------
 * Name:         vs_mem_idxToPtr
 *
 * Description:
 *
 *     Find a memory unit.
 *
 * Arguments:    
 *
 *     index - The index to the memory unit to be found.
 *
 * Return Value:
 *
 *    Pointer to the data structure - If found.
 *
 *    NULL - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void *
vs_mem_idxToPtr(vs_mem_index_t index);

/*-----------------------------------------------------------------
 * Name:         vs_mem_ptrToIdx
 *
 * Description:
 *
 *     Obtain the index from a given data structure pointer.
 *
 * Arguments:    
 *
 *     srvrIdx - Virtual server index.
 *
 *     dataType - Data type index.
 *
 *     dataStructp - Pointer to the data structure.
 *
 * Return Value:
 *
 *    A valid index to the given data structure - If found.
 *
 *    A zero index - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/20/2002
 *
 *-----------------------------------------------------------------
 */
vs_mem_index_t
vs_mem_ptrToIdx(vs_mem_srvr_idx_t  srvrIdx,
                vs_mem_dataType_t  dataType,
                void               *dataStructp);

/*-----------------------------------------------------------------
 * Name:         vs_mem_init
 *
 * Description:
 *
 *     Initialize virtual server table.
 *
 * Arguments:    
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void
vs_mem_init(void);

/*-----------------------------------------------------------------
 * Name:         vs_memSetDataSize
 *
 * Description:
 *
 *     Initialize data size.
 *
 * Arguments:    
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void
vs_memSetDataSize(uint32 type, uint32 size);

#endif /* _VS_MEM_API_H */

