/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ccm-nodefail.c
 *
 * RCS:  $Id: ccm-nodefail.c,v 1.6 2002/08/16 21:35:01 gregs Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: This is where we go when a node fails
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 10 June 2002
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "exports/anpssc-api.h"
#include "exports/ipmd-lib.h"
#include "exports/elog-api.h"

#include "exports/ccm-api.h"
#include "exports/ncm-api.h"
#include "ccm-loc.h"

typedef struct ip_lst {
	uint32		addr;
	uint32		mask;
	char		*port;
	struct ip_lst *next;
} ip_lst_t;

typedef struct cand {
	char		*cand_node;
	ip_lst_t	*ip_cfg;
	struct cand	*next;
} cand_t;

typedef struct serv {
	char		*lv;
	cand_t		*candidates;
	struct serv	*next;
} serv_t;


static void
eval_candidates (serv_t *this_srv, char **slaves)
{
	char **sl;
	cand_t *its_one;
	char *lv_str;
	char *node_name;
	char *name_start;
	int sz_node_name;
/*
	If the node is able to serve a volume:
	  sl[0] needs to be "slave.NODENAME.this_srv->lv"
	  sl[1] needs to be "visible"
 */
	for (sl = slaves; *sl; sl += 2) {

		lv_str = rindex (sl[0], '.') + 1;

		if (!strcasecmp (this_srv->lv, lv_str) &&
		    !strcasecmp (sl[1], "visible")) { // It sees the LV
		/*
			So push it on the list
		 */
			its_one = malloc (sizeof(cand_t));
			memset (its_one, 0, sizeof(cand_t));
			its_one->next = this_srv->candidates;
			this_srv->candidates = its_one;
		/*
			Duplicate node name from "slave.NODENAME.lv" in sl[0]
		 */
			name_start = sl[0] + strlen ("slave.");
			sz_node_name = lv_str - name_start;
			node_name = malloc (sz_node_name);
			memcpy (node_name, name_start, sz_node_name);
			node_name[sz_node_name - 1] = '\0';

			its_one->cand_node = node_name;
		}
	}
}

static ip_lst_t *
get_node_ip (char *node, char **ip_cfg)
{
	ip_lst_t *lst = NULL;
	ip_lst_t *n;
	char **look;
	char *node_srch;
	int len_srch;
/*
	We're going to find ALL the IP addresses assigned to "node"
 */
	if (ip_cfg) {
		asprintf (&node_srch, "%s:", node);
		len_srch = strlen (node_srch);
/*
	look[0] will be "cluster.ip.ip.addr.in.dotnotation"
	look[1] will be "nodename:port/mask"
	Just a reminder
 */
		for (look = ip_cfg; *look; look += 2) {

			if (!strncasecmp (node_srch, look[1], len_srch)) {
		/*
			It's an IP address assigned to a port on "node" so let's
			collect it.
		 */
				n = malloc (sizeof(ip_lst_t));
				n->next = lst;
				lst = n;

				n->addr = inet_addr (look[0] + strlen ("cluster.ip."));
				n->mask = strtol (index (look[1], '/') + 1, NULL, 10);
				n->port = strdup (index (look[1], ':') + 1);
				*index (n->port, '/') = '\0';
			}
		}
		free (node_srch);
	}

	return lst;
}

static cand_t *
pick_candidate (serv_t *services, ip_lst_t *ipaddrs)
{
	serv_t *nxt_svc;
	cand_t *nxt_vic;
	cand_t *prv_vic;
	cand_t *look_node;
	ip_lst_t *nxt_addr;
/*
	We need to find the set of nodes that show up on all services
 */
	for (nxt_svc = services->next; nxt_svc; nxt_svc = nxt_svc->next) {
/*
	Remove those nodes from services->candidates that don't show up on
	nxt_svc->candidates.
 */
		for (prv_vic = NULL, nxt_vic = services->candidates; nxt_vic;) {

			for (look_node = nxt_svc->candidates; look_node; 
			     look_node = look_node->next) {
				if (!strcasecmp (nxt_vic->cand_node, look_node->cand_node)) {
					break;
				}
			}

			if (!look_node) {	// Didn't find it so remove it as a candidate
				if (prv_vic) {
					prv_vic->next = nxt_vic->next;
				}
				else {
					services->candidates = nxt_vic->next;
				}
				free (nxt_vic->cand_node); // And free its heap
				look_node = nxt_vic->next;
				free (nxt_vic);
				nxt_vic = look_node;
			}
			else {
				prv_vic = nxt_vic;
				nxt_vic = nxt_vic->next;
			}
		}
	}
/*
	If we don't have a node than can serve the failing node's services,
	then don't bother checking network connectivity.
 */
	if (!services->candidates) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0,
		       ("Unable to locate node that can serve all volumes"));
	}
/*
	Now eliminate those nodes without network connectivity to the failed
	node. In the process we'll collect the candidate's port name.
 */
	else {
#undef REALSTUFF
#ifdef REALSTUFF
#define LSTLEN 10

		for (nxt_addr = ipaddrs; nxt_addr; nxt_addr = nxt_addr->next) {
			for (prv_vic = NULL, nxt_vic = services->candidates; nxt_vic;) {
				ipm_arp_t arplist[LSTLEN];
				int rtncnt, need;
				int stat;

				need = 0;
				rtncnt = 0;
				stat = ipmd_GetL2ConnEntCluster (nxt_addr->addr, arplist, 
				          sizeof(arplist), &rtncnt, &need, nxt_vic->cand_node);

				if ((stat == IPM_GET_OK) && (rtncnt > 0)) {
// Record which port it's on
					ip_lst_t *newone = malloc (sizeof(ip_lst_t));
					newone->next = nxt_vic->ip_cfg;
					nxt_vic->ip_cfg = newone;
					newone->port = 
					          strdup (index(arplist[0].arp_ifname, ':') + 1);
					newone->addr = nxt_addr->addr;
					newone->mask = nxt_addr->mask;
					prv_vic = nxt_vic;
					nxt_vic = nxt_vic->next;
				}
				else {
					if (prv_vic) {
						prv_vic->next = nxt_vic->next;
					}
					else {
						services->candidates = nxt_vic->next;
					}
					free (nxt_vic->cand_node); // And free its heap
					look_node = nxt_vic->next;
					free (nxt_vic);
					nxt_vic = look_node;
				}
			}
		}

		if (!services->candidates) {
			E_LOG (class_1, err_s, myappid, 0, 0, 0,
			       ("Unable to locate node with all network connectivity"));
		}
#else
// NOTICE:
// This just forces the ports to be hooked up in parallel 0 to 0, 1 to 1, etc
		services->candidates->ip_cfg = NULL;
		for (nxt_addr = ipaddrs; nxt_addr; nxt_addr = nxt_addr->next) {
			ip_lst_t *t_addr;
			t_addr = malloc (sizeof(ip_lst_t));
			memcpy (t_addr, nxt_addr, sizeof(ip_lst_t));
			t_addr->port = strdup (nxt_addr->port);
			t_addr->next = services->candidates->ip_cfg;
			services->candidates->ip_cfg = t_addr;
		}
#endif
	}
/*
	Take the top one on the list
 */
	return services->candidates;
}

static void
force_ip (char *old, cand_t *node)
{
	ip_lst_t *rstr_this;
	char *port_info;

	for (rstr_this = node->ip_cfg; rstr_this; rstr_this = rstr_this->next) {

		if (ipmd_AddIpCluster (rstr_this->port, rstr_this->addr, 
		                       rstr_this->mask, NULL, 0, node->cand_node)) {
			E_LOG (class_1, alert_s, myappid, 0, 0, 0,
			       ("force_ip ipmd_AddIpCluster failed on %s for %x/%x", 
			        rstr_this->port, rstr_this->addr, rstr_this->mask));
		}
		restore_rte (rstr_this->addr, rstr_this->mask, node->cand_node);

		asprintf (&port_info, "%s/%d", rstr_this->port, rstr_this->mask);

		record_ip_taken (rstr_this->addr, old, node->cand_node, port_info);

		free (port_info);
	}
}

extern void
node_failed (char *node)
{
	char **clust_ip;
	char **oper_state;
	char **clus_serves;
	char *node_online;
	char **look;
	serv_t *serv_lst = NULL;
	serv_t *srvc;
	cand_t *new_server;
	ip_lst_t *ip_addrs;

	E_LOG (class_1, err_s, myappid, 0, 0, 0,
	       ("Node %s has failed", node));
/*
	Collect the set of services this node has online
 */
	clus_serves = get_cfg_val ("cluster.lv.*");
/*
	In the unlikely event the cluster isn't serving any volumes, we're done.\
 */
	if (!clus_serves) {
		return;
	}

	asprintf (&node_online, "online:%s", node);
/*
	We need the operational state of all the nodes in the cluster for
	determining which nodes are able to take over the failed node's
	volumes (they're visible).
 */
	oper_state = get_cfg_val ("slave.*");

	if (!oper_state) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0,
		       ("No volumes visible anywhere"));
		free_cfg_val (clus_serves);
		return;
	}

	for (look = clus_serves; *look; look += 2) {

		if (!strcasecmp (node_online, look[1])) {
			srvc = malloc (sizeof(serv_t));
			memset (srvc, 0, sizeof(serv_t));
			srvc->next = serv_lst;
			serv_lst = srvc;
/*
	Find suitable replacement nodes for this service
 */
			serv_lst->lv = strdup (look[0] + strlen ("cluster.lv."));
			eval_candidates (serv_lst, oper_state);
		}
	}
	free (node_online);	// All done with this
/*
	Only look for a candidate to take over if the failed node has
	services assigned to it.
 */
	if (serv_lst) {
/*
	Collect the IP addresses the new node will have to be able to assume
 */
		clust_ip = get_cfg_val ("cluster.ip.*");
		ip_addrs = get_node_ip (node, clust_ip);
		free_cfg_val (clust_ip);
/*
	Find the one node that shows up in all the serv_lst->candidates
	(and can reach all the networks the failed node could)
 */
		new_server = pick_candidate (serv_lst, ip_addrs);
/*
	Can only takeover if there is another node able to handle all volumes
	that has all the network connectivity.
 */
		if (!new_server) {
			E_LOG (class_1, err_s, myappid, 0, 0, 0,
			       ("Unable to find a node to takeover for %s", node));
		}
/*
	Command the new node to forcably take over all the services of
	the just failed node.
 */
		else {
			for (srvc = serv_lst; srvc; srvc = srvc->next) {
				char *lvnfs;
				char **nfscfg;
				evm_volId_t lvid;

				asprintf (&lvnfs, "cluster.nfs.%s", srvc->lv);
				nfscfg = get_cfg_val (lvnfs);
				if (nfscfg) {
					force_takeover (new_server->cand_node, srvc->lv, nfscfg[1]);
					lvid.whole_thing = strtouq (srvc->lv, NULL, 16);
					record_lv_taken (lvid, node, new_server->cand_node);
				}
				free_cfg_val (nfscfg);
				free (lvnfs);
			}
/*
	Force the new node to assume the failed node's IP addresses.
    The reason we do this after the LV forced takeovers is to ensure
	an orphaned node (lost both mgmt ports or SSC) notices it's getting its
	LVs taken and will remove its GbE IP config.
	Or so that's the theory.
 */
			force_ip (node, new_server);
		}
/*
	Clean up all the crap we just malloc()ed
 */
		while (ip_addrs) {
			ip_lst_t *tmp;
			tmp = ip_addrs;
			ip_addrs = tmp->next;
			free (tmp->port);
			free (tmp);
		}

		for (srvc = serv_lst; serv_lst; srvc = serv_lst) {
			cand_t *whack;

			for (whack = srvc->candidates; srvc->candidates; 
			     whack = srvc->candidates) {
				free (whack->cand_node);
				while (srvc->candidates->ip_cfg) {
					ip_lst_t *tmp;
					tmp = srvc->candidates->ip_cfg;
					srvc->candidates->ip_cfg = srvc->candidates->ip_cfg->next;
					free (tmp->port);
					free (tmp);
				}
				srvc->candidates = whack->next;
				free (whack);
			}

			free (srvc->lv);
			serv_lst = srvc->next;
			free (srvc);
		}

	}
	free_cfg_val (clus_serves);
	free_cfg_val (oper_state);
}
