/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: Node shutdown handler
 *
 * RCS:  $Id: shut_hand.c,v 1.17 2002/10/04 02:17:06 gregs Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: These functions take care of sending the messages to do an
 *              orderly shutdown of a node.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 22 Mar 2002
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>

#include "nfx-defs.h"
#include "exports/anpssc-api.h"
#include "exports/evm-api.h"
#include "exports/ccm-api.h"
#include "exports/ncm-api.h"
#include "exports/elog-api.h"
#include "exports/ea-api.h"
#include "exports/rpc.h"    // Because mount.h needs it that's why
#include "exports/mount.h"
#include "exports/elog-api.h"
#include "exports/ipmd-lib.h"
#include "exports/cfgrules-api.h"

#include "ccm-loc.h"

/*
	How to shutdown a node:
	1) Figure out which LVs the node has exported/has online.
	2) "Unexport" those LVs. Wait for response from NFS before
	   proceeding to the next steps.
	3) Shutdown the IP addresses served by that node.
	4) Dismount the file systems.
	5) Take the LVs offline.

	Sequence of the state machine: 2) occurs before 3) and 4) start. 3) is
	a precursor to 5). Since 2) 4), and 5) can be done asynchronously, the
	message to initiate 4) is sent before calling ipm library to cause 3)
	which is synchronous.

	We need to make sure that ALL 2) complete before 3). Trigger event for
	3) is that all unshare requests have been received. Trigger event for 5)
	is 4).
 */

typedef enum {
	unexp_sent,
	dismount_sent,
	offline_sent
} shutdown_state_t;

#define SHUT_MARK 0xbabebabe

typedef struct LV_shut {
	evm_volId_t			lv_id;
	uint32				mark;				// To make sure the pointer's cool
	shutdown_state_t	wassup;
	char 				*lvName;			// The name is what EA needs
	struct node_shut	*node;
	struct LV_shut		*prev;
	struct LV_shut		*next;
} LV_shut_t;

typedef struct ip_shut {
	char				if_name[10];		// fpN.M
	uint32				ip_addr;
	struct node_shut	*node;
	struct ip_shut		*next;
} ip_shut_t;

typedef struct node_shut {
	ip_shut_t			*ip_lst;			// List of IP addresses to be shut
	LV_shut_t			*lv_unexp_lst;		// LVs waiting for unexport resp
	LV_shut_t			*lv_dism_lst;		// LVs waiting for dismount resp
	char				name[256];
	struct node_shut	*next;
	ccmShutdown_Msg_t	req;				// Shutdown req msg
	int					sock;
	char				app[256];			// From whence the req came
	int					CPU;
	int 				slot;
	int					plane;
} node_shut_t;

static node_shut_t *nodes_shutting_down = NULL;

extern void
node_up (int appsock, ccmLclCfg_t *lcfg)
{
	char **cfg_obj;
	node_shut_t *look, *prev;
	char *obj_name;
	char *obj_val;
	int is_lv_obj;
	int lv_online_ok;

	E_LOG (class_1, info_s, myappid, 0, 0, 0,
	       ("ccm got new node info for %s", lcfg->name));
/*
	We could have known this node before and had shut it down. That
	being the case, we'll remove it from the shutting down list now.
 */
	for (prev=NULL, look = nodes_shutting_down; look; 
	     prev = look, look = look->next) {

		if (!strcasecmp (lcfg->name, look->name)) {
			if (prev) {
				prev->next = look->next;
			}
			else {
				nodes_shutting_down = look->next;
			}
			free (look);
			break;
		}
	}
/*
 	Is it a real update or the NULL indicating the node doesn't have any
	operational state to share?
 */
	if (strlen (lcfg->cfg_pair)) {
		obj_name = lcfg->cfg_pair;
		obj_val = lcfg->cfg_pair + strlen (lcfg->cfg_pair) + 1;
/*
 * In the case the other node is telling us of a volume it shows as assigned
 * to it, we need to make sure it still can be placed online there.
 */
		is_lv_obj = !strncasecmp (obj_name, "cluster.lv.", 11);
		if (is_lv_obj) {
			evm_volId_t lvid;

			lvid.whole_thing = strtoq (&obj_name[11], NULL, 16);

			lv_online_ok = ok_to_online (lvid, lcfg->name);
		}
/*
 * Now we can update its configuration so long as its not an invalid
 * lv assignment. Always add if !is_lv_obj.
 */
		if (!is_lv_obj || lv_online_ok) {
			cfg_obj = get_cfg_val (obj_name);
			if (!cfg_obj) {
				cfg_obj = malloc (3 * sizeof(char *));
				cfg_obj[0] = strdup (obj_name);
				cfg_obj[1] = strdup (obj_val);
				cfg_obj[2] = NULL;
				put_cfg_val (cfg_obj);
				free_cfg_val (cfg_obj);
			}
		}
/*
	If it's a recorded slave visible object, let's fake the visible event
 */
		if (!strncasecmp ("slave.", obj_name, 6) && 
		    !strcasecmp ("visible", obj_val)) {

			evm_volId_t lv_id;
			char *lvstr;

			lvstr = strrchr (obj_name, '.') + 1;	// slave.nodeName.lvstr
			
			lv_id.whole_thing = strtoq (lvstr, NULL, 16);
	
			lvVis (appsock, lcfg->name, lv_id);
		}
	
	}
}

extern int
is_shutting_down (char *node)
{
	node_shut_t *look;

	for (look = nodes_shutting_down; look; look = look->next) {
		if (!strcasecmp (node, look->name)) {
			return 1;
		}
	}
	return 0;
}

static void
delIpaddrs (ip_shut_t *ips)
{
	ip_shut_t *whackthis;
	char *nodeName = NULL;

	if (ips) {
		nodeName = ips->node->name;
	}

	while (ips) {

		if (ipmd_DelIpCluster (ips->if_name, ips->ip_addr, NULL, 0,
		    ips->node->name)) {
			E_LOG (class_1, err_s, myappid, 0, 0, 0,
			       ("Error deleting IP address %x from %s on %s", 
			        ips->ip_addr, ips->if_name, ips->node->name));
		}

		whackthis = ips;
		ips = ips->next;
		free (whackthis);
	}
/*
	Su said we need to do this too in order to close all TCP connections
 */
	if (nodeName) {
		ipmd_ShutDownCluster ("Su said we", nodeName);
		ipmd_RestartCluster ("need to do this", nodeName);
	}
}

extern void 
shutdown_node (int sock, char *app, int CPU, int Slot, int Plane, 
               ccmShutdown_Msg_t *msg)
{
	uint64 msgbuf[1024];
	ncmMsg_t *ncm = (ncmMsg_t *)msgbuf;
	mnt_infoReq_t *unexp = (mnt_infoReq_t *)ncm->appMsg;
	node_shut_t *new;
	char **cfg;
	char **look;
	char *portspec;
	char *maskspec;
	char *nodespec;
	char *nfs_export;
	char *lvstr;
	ip_shut_t *newip;
	LV_shut_t *newlv, **lastlv;
	int rc;
	int sz;
/*
	First thing to do is collect the list of things this node serves and
	send the unexport messages for all.
	This object will get free()ed when everything is closed up in shut_reply().
 */
	new = malloc (sizeof(node_shut_t));
	memset (new, 0, sizeof(node_shut_t));
	new->sock = sock;
	new->CPU = CPU;
	new->slot = Slot;
	new->plane = Plane;
	strcpy (new->app, app);
	strcpy (new->name, msg->fromNode);
	memcpy (&(new->req), msg, sizeof(ccmShutdown_Msg_t));

	E_LOG (class_4, debug_s, myappid, 0, 0, 0,
	       ("Got shutdown req for %s", msg->fromNode));

	new->next = nodes_shutting_down;
	nodes_shutting_down = new;

	cfg = get_cfg_val ("cluster.nfs.*");
/*
	If nothing is exported, nothing should be mounted or online. If so,
	then automatic clean shutdown can't be done. The assumption is that
	a file system is mounted only when shared.
 */	
	if (!cfg) {
		shut_reply (NULL);
	}
	else {
/*
	Make a list of all the IP addresses this node is serving.
	Keys "cluster.ip.*" and match the value with the node name.
 */
		free_cfg_val (cfg);
		cfg = get_cfg_val ("cluster.ip.*");

		if (!cfg) {
			E_LOG (class_4, err_s, myappid, 0, 0, 0,
			       ("Danger: There are cluster.nfs but no cluster.ip items"));
		}
/*
	cluster.ip.* values are: "nodename:portspec/maskspec". We match "nodename"
	to determine if this is an IP address served by the node shutting down.
 */
		else {
			nodespec = malloc (strlen(msg->fromNode) + 2);
			strcpy (nodespec, msg->fromNode);
			strcat (nodespec, ":");

			for (look = cfg; *look; look +=2) {
				maskspec = look[1];
				portspec = strsep (&maskspec, "/");
/*
		If it happens to be on this node, stack it on ip_lst
 */
				if (!strncasecmp (nodespec, portspec, strlen(nodespec))) {

					newip = malloc(sizeof(ip_shut_t));

					strcpy (newip->if_name, strchr(portspec, ':') + 1);
					newip->ip_addr = 
					              inet_addr (look[0] + strlen("cluster.ip."));
					newip->next = new->ip_lst;
					newip->node = new;
					new->ip_lst = newip;
				}
			}
			free (nodespec);
			free_cfg_val (cfg);
		}
/*
	Make a list of the LVs this node has exported.
	Keys "cluster.lv.*" and match the value 'online:"nodename"'
 */
		cfg = get_cfg_val ("cluster.lv.*");

		if (!cfg) {
			E_LOG (class_4, err_s, myappid, 0, 0, 0,
			       ("Danger: There are cluster.nfs but no cluster.lv items"));
		}
		else {
			nfs_export = malloc (strlen ("cluster.nfs.") + 17);
			nodespec = malloc (strlen ("online:") + strlen(msg->fromNode) + 1);
			strcpy (nodespec, "online:");
			strcat (nodespec, msg->fromNode);
/*
	Set up NCM header for all the messages going to mountd
 */
			memset (ncm, 0, sizeof (ncmMsg_t));
			ncm->msgType = NCM_SEND;
			strcpy (ncm->appName, "mountd");
			strcpy (ncm->nodeName, msg->fromNode);
			ncm->rqResp = 1;
/*
	Collect the export lists for all LVs this node has online
 */
			for (lastlv = &(new->lv_unexp_lst), look = cfg; *look; look +=2) {
/*
		If it happens to be on this node, put it on lv_unexp_lst and send
		the unexport message to NFS.
 */
				if (!strcasecmp (nodespec, look[1])) {
					char **nfscfg;
/*
		Get the export path for the unexport message
 */
					lvstr = look[0] + strlen ("cluster.lv.");
					sprintf (nfs_export, "cluster.nfs.%s", lvstr);
					nfscfg = get_cfg_val (nfs_export);
					if (!nfscfg) {
						E_LOG (class_4, err_s, myappid, 0, 0, 0,
				              ("There was %s but no %s", look[0], nfs_export));
						continue;
					}
					else {
/*
		We have to unexport ALL paths to this LV
 */
						char *path, *next_path;
						next_path = nfscfg[1];
						do {
							char *laddr;
/*
			First we need a tracking object to link to the node's object
 */
							newlv = malloc(sizeof(LV_shut_t));
							newlv->lv_id.whole_thing = 
							                         strtoq (lvstr, NULL, 16);
							newlv->mark = SHUT_MARK;
							newlv->wassup = unexp_sent;
							newlv->node = new;

							newlv->next = NULL;
							newlv->prev = *lastlv;
							*lastlv = newlv;
							lastlv = &(newlv->next);
/*
		Each export path specifcation for this LV being separated by a 
		semicolon while the path part is separated from the options part
		by our friend the vertical bar.
 */
							path = strsep (&next_path, ";");
							laddr = strsep (&path, ":");
							*index (path, '|') = '\0';

							memset (unexp, 0, sizeof (mnt_infoReq_t));
							unexp->msgHdr.msgType = MNT_UNEXPORT_REQ;
							unexp->msgHdr.private = (uint32)newlv;
							unexp->nameLen = strlen (path);
							memcpy (unexp->pathName, path, unexp->nameLen);
							unexp->localIpAddr = strtol (laddr, NULL, 16);

							sz = (char *)unexp - (char *)ncm + 
							     sizeof (mnt_infoReq_t) + unexp->nameLen + 1;

							E_LOG (class_4, debug_s, myappid, 0, 0, 0,
							       ("TO unexp %s on %s", path, msg->fromNode));

							rc = sendAgileMsg (sock, ncm, sz, 
							         "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							         NULL, NULL, NULL);

							if (rc < 0) {
								E_LOG (class_4, err_s, myappid, 0, 0, 0,
								       ("Error sending unshare via ncm"));
							}
/*
		We need to get the LV name in order for EA to be able to dismount
		it. The export path is always: "addr:/VOLUMENAME/somepath/...
 */
							path = strchr (path, '/');
							path++;		// past leading '/'
							if (strchr (path, '/')) {
								*strchr (path, '/') = '\0';
							}
							newlv->lvName = strdup (path);
						} while (next_path);
					}
				}
			}
			free (nfs_export);
			free (nodespec);
			free_cfg_val (cfg);
		}

		if (!new->lv_unexp_lst) {	// Nothing to wait for
			delIpaddrs (new->ip_lst);
			shut_reply (NULL);
		}
	}

	return;
}

static int
send_dismount(LV_shut_t *lv)
{
	uint64 msgbuf[1024];
	ncmMsg_t *ncm = (ncmMsg_t *)msgbuf;
	ea_movOutFsysReq_t *mo = (ea_movOutFsysReq_t *)ncm->appMsg;
	LV_shut_t *look;
	int sz;
	int rc;
/*
	Let's make sure this LV is not on the node's lv_unexp_lst anywhere else
	before we dismount the file system.
 */
	for (look = lv->node->lv_unexp_lst; look; look = look->next) {
		if (lv->lv_id.whole_thing == look->lv_id.whole_thing) {
			return 0;
		}
	}
/*
	Looks to be the last unexport of this LV. Let's tell EA to unmount
	the file system.
 */
	memset (ncm, 0, sizeof (ncmMsg_t));
	ncm->msgType = NCM_SEND;
	strcpy (ncm->appName, "ea");
	strcpy (ncm->nodeName, lv->node->name);
	ncm->rqResp = 1;
	
	memset (mo, 0, sizeof(ea_movOutFsysReq_t));
	mo->hdr.msgType = EA_MOVOUT_FSYS_REQ;
	mo->hdr.userInfo = (uint32)lv;
	strcpy (mo->hdr.node, lv->node->name);
	mo->fsysId.volId = lv->lv_id;
	strcpy (mo->fsysId.fsysName, lv->lvName);

	E_LOG (class_4, debug_s, myappid, 0, 0, 0,
	       ("moveout req for %s/%016llx at %s", mo->fsysId.fsysName,
	        mo->fsysId.volId.whole_thing, mo->hdr.node));

	sz = (char *)mo - (char *)ncm + sizeof (ea_movOutFsysReq_t);

	rc = sendAgileMsg (lv->node->sock, ncm, sz, 
	                   "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	                   NULL, NULL, NULL);

	if (rc < 0) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0,
		       ("Error sending FS move out via ncm"));
	}

	return 1;
}

static node_shut_t *
this_unshared (mnt_infoResp_t *msg)
{
	LV_shut_t *lv;
	node_shut_t *this_node;
/*
	Make sure the cookie's pointing to a proper LV_shut before we
	do anything else.
 */
	lv = (LV_shut_t *)(uint32)(msg->msgHdr.private);
	if (lv->mark != SHUT_MARK) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0,
		       ("Unshare response not to our message"));
		return NULL;
	}
/*
	Unlink from lv_unexp_lst.
	Send dismount message if it's the last share of this LV and put it on
	lv_dism_lst. 
 */
	this_node = lv->node;
	if (lv->next) {
		lv->next->prev = lv->prev;
	}
	if (lv->prev) {
		lv->prev->next = lv->next;
	}
	else {
		lv->node->lv_unexp_lst = lv->next;
	}

	if (send_dismount (lv)) {		// Last share so push on dismount wait list
		lv->wassup = dismount_sent;
		lv->prev = NULL;
		lv->next = lv->node->lv_dism_lst;
		lv->node->lv_dism_lst = lv;
		if (lv->next) {
			lv->next->prev = lv;
		}
	}
	else {	// Still shared (implies lv->node->lv_unexp_lst)
		free (lv->lvName);
		free (lv);
	}
/*
	Once this the node's unshare list is NULL, call ipm to delete interfaces.
 */
	if (!this_node->lv_unexp_lst) {
		E_LOG (class_4, debug_s, myappid, 0, 0, 0,
		       ("Done with share removal on %s", this_node->name));
		delIpaddrs (lv->node->ip_lst);
	}
	return this_node;
}

static node_shut_t *
this_unmounted (ea_movOutFsysRsp_t *msg)
{
	LV_shut_t *lv;
	node_shut_t *da_node;
	char *slave_cfg[3];
/*
	Make sure the cookie's pointing to a proper LV_shut before we
	do anything else.
 */
	lv = (LV_shut_t *)(msg->hdr.userInfo);
	if (lv->mark != SHUT_MARK) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0,
		       ("Move out response not to our message"));
		return NULL;
	}
	E_LOG (class_4, debug_s, myappid, 0, 0, 0,
	       ("Dismount response for %s", lv->node->name));
/*
	All done, remove it from lv_dism_lst
 */
	if (lv->next) {
		lv->next->prev = lv->prev;
	}
	if (lv->prev) {
		lv->prev->next = lv->next;
	}
	else {
		lv->node->lv_dism_lst = lv->next;
	}
/*
	We need to rewrite the slave's status as this LV is now just plain
	old visible.
 */
	asprintf (&slave_cfg[0], "slave.%s.%016llx", lv->node->name, 
	                                             lv->lv_id.whole_thing);
	slave_cfg[1] = "visible";
	slave_cfg[2] = NULL;

	put_cfg_val (slave_cfg);
	free (slave_cfg[0]);
/*
	Return what you used
 */
	da_node = lv->node;
	free (lv->lvName);
	free (lv);

	return da_node;
}

extern int 
shut_reply (ncmMsg_t *buf)
{
	int rc;
	node_shut_t *this_node;
/*
	Only time we're passed a NULL message is when the shutdown is on a node
	that has no LVs currently exported/mounted/online.
 */
	if (!buf) {
		this_node = nodes_shutting_down;
	}
	else {
		if (buf->msgType != NCM_RESP) {
			E_LOG (class_1, err_s, myappid, 0, 0, 0,
				   ("ccmd:shut_reply expected NCM_RESP"));
			return 1;
		}
/*
	Process this response to what we've sent. We stash the pointer to the
	specific *shut_t in the private cookie in the request.
 */
		switch (*(int *)(buf->appMsg)) {
			case MNT_INFO_RESP:		// Unshared now so unmount it
				this_node = this_unshared ((mnt_infoResp_t *)(buf->appMsg));
				break;

			case EA_MOVOUT_FSYS_RSP:	// It's unmounted now
				this_node = 
				         this_unmounted ((ea_movOutFsysRsp_t *)(buf->appMsg));
				break;

			default:
				return 1;			// Didn't get anything we expected
		}
	}
/*
	Node is clean when nothing is left to do so let it shut down.
 */	
	if (this_node && !this_node->lv_unexp_lst && !this_node->lv_dism_lst) {
		this_node->req.done = 1;
/*
	After receiving this response, the sender is assured the node is
	no longer providing any services.
 */
		E_LOG (class_1, info_s, myappid, 0, 0, 0,
			   ("ccmd: Node %s Shutdown complete", this_node->name));

		rc = sendAgileMsg (this_node->sock, &(this_node->req), 
		                   sizeof (ccmShutdown_Msg_t), 
		                   this_node->app, this_node->CPU, this_node->slot, 
		                   this_node->plane, NULL, NULL, 0);

	}

	return 0;
}
