/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	lnm-local.h
 *
 * RCS:  $Id: lnm-local.h,v 1.11 2002/08/26 21:14:19 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Local structures for the nodes and controller
 *
 * Created by:	Pankaj Gupta 
 *
 * Date Created:	11/05/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _LNM_LOCAL_H
#define _LNM_LOCAL_H

#include <exports/eee-desc.h>
#include <assert.h>

#define	CLUSTER_TIMER_SECONDS	5

#define LISTEN_STATE_INTERVAL	6

extern uint32 lnmDebugFlag;
#define LNM_DBG_PRINT if (lnmDebugFlag) printf

#define LNM_ERR_PRINT printf

extern int myAppId;
extern int nfxMyCPUId, nfxMyChassisId, nfxMySlotId;

extern time_t now;
extern time_t node_timeout;

/*
 * Each node has this struct for maintaining it's
 * info. Info from here is extracted and sent in
 * Node Hello Message. Cluster controllers
 * maintain a table of cluster node structure
 * to track cluster status
 */

typedef enum {
	member_mode,
	CC_mode,
	standalone_mode
} ccmode_t;

typedef struct {

	ccmode_t	CCMode;	
	uint32		useFlag;
#define NODE_ENTRY_ALLOC		0x01
#define NODE_ENTRY_ADDED		0x02

	uchar8		nodeState;
#define CN_STATE_INIT 			0
#define CN_STATE_LISTEN 		1
#define CN_STATE_FIND_PCC 		2
#define CN_STATE_CC				3	
#define CN_STATE_NON_CC			4	
#define CN_STATE_PCC			5	
#define CN_STATE_SCC			6	
#define CN_STATE_DOWN			7	

	uchar8		preferredPCCMode;
	uchar8		clusterState;
	uchar8		unmanagedOrphanAdoptMode;
	uchar8		nodeTimeoutForcedTakeoverMode;
	uint32 		deadWhen;
	uint32		nodeIpa[2];
	uint32		nodeBroadcast[2];	// Really only used for local node
	uint32		nodeChassisId;
	char		nodeName[LNM_MAX_NODE_NAME];

#ifdef LINUX_TEST
	uint32	nodePortOffset;
#endif

} cluster_nodeInfo_t;

extern cluster_nodeInfo_t	clusterNodeInfoArray[MAX_NODE_PER_CLUSTER];

/*
 * Cluster Control message structure
 * for inter node communication
 */

typedef struct {

	cluster_msgType_t 	msgType;

	uchar8	clusterName[LNM_MAX_CLUSTER_NAME];
	uint32 	srcChassisId;
	uint32 	srcIpa[2];
	uint32 	dstIpa;

	/*
	 * Different message formats
	 */

	union { 

		/* 
		 * Initial Hello Message 
		 */
		struct initHelloMsg {
			uint32 			nodeState;
			uchar8			preferredPCCMode;
		} initHelloMsg;

		/* 
		 * Node Hello Message 
		 */
		struct nodeHelloMsg {
			uint32 			nodeState;
			uchar8			CCMode;	
			uchar8			preferredPCCMode;
			uchar8			reportedVolType;
			uchar8			fpNimModState;
			uchar8			fcNimModState;
			uchar8			sscModState;
			char			nodeName[LNM_MAX_NODE_NAME];
		} nodeHelloMsg;


		/*
		 * Assign SCC Request
		 */
		struct assignSCCReq {
			uint32 			pccChassisId;
			uint32			pccIpa[2];
			uint32			sccChassisId;
			uint32			sccIpa[2];

			uchar8			preferredPCCMode;
			uchar8			clusterState;
			uchar8			unmanagedOrphanAdoptMode;
			uchar8			nodeTimeoutForcedTakeoverMode;
		} assignSCCReq;

		/* 
		 * Find PCC Request
		 */
		struct findPCCReq {
			uint32 			nodeState;
			uchar8			preferredPCCMode;
			char			myName[LNM_MAX_NODE_NAME];
		} findPCCReq;

	} msg;
} cluster_controlMsg_t;

/*
 * Global Volume Table entry format
 */
typedef struct {

	uint32 			volId;
	uint32			volOwner;
	uint32			priFailoverPort;
	uint32			secFailoverPort;
	uchar8			volType;
	uchar8			volStatus;
	uchar8			lastActionRequested;

} cluster_globalVolumeTableEntry_t;


/*
 * Function prototypes
 */
cluster_nodeInfo_t*	cluster_findNode(uint32 nodeChassisId);
cluster_nodeInfo_t*	cluster_allocateNode(void);
void 				cluster_deallocateNode(cluster_nodeInfo_t *delNodeEntry);
cluster_nodeInfo_t*	cluster_addNode(cluster_nodeInfo_t *newNodeEntry);
void 				cluster_deleteNode(uint32 nodeChassisId);
void 				cluster_changeNodeState(cluster_nodeInfo_t* nodeEntry, 
										int newState);
void 				cluster_updateNode(cluster_nodeInfo_t *nodeEntry, 
								cluster_controlMsg_t *ccMsg);
cluster_nodeInfo_t*	cluster_findSccCandidate(void);
int 				cluster_nodeIsUp(cluster_nodeInfo_t *nodeEntry);
char*				printIpa(int ipaddr);
void 				get_ip_addr(struct in_addr* in_addr, char* str);
int 				cluster_postEvent(cluster_nodeInfo_t* nodeEntry, 
								int eventId, int eventState,
								void *payload, int paysize);
void 				lnm_stop(int sig);
void 				lnm_processTimer(void);
void 				cluster_sendInitHello(void);
void 				cluster_sendNodeHello(uint32 *dstIpa);
void 				cluster_sendAssignSCCReq(cluster_nodeInfo_t 
											*sccCandidateNode);
void 				cluster_sendFindPCCReq(void);
void 				cluster_txControlMsg(cluster_controlMsg_t* msg, 
									int msgLen, uint32 dstIpa);
void 				cluster_rcvInitHello(cluster_controlMsg_t *ccMsg, 
									int msgLen);
void 				cluster_rcvAssignSCCReq(cluster_controlMsg_t *ccMsg, 
									int msgLen);
void 				cluster_rxControlMsg(cluster_controlMsg_t* msg, 
									int msgLen, int srcIpa);
void 				cluster_rxAgileMsg(char *rcvdBuf, int msgLen, char *sender);
void 				cluster_processCli(int argc, char  *argv[]);
void 				cluster_displayStatus(void);
void 				cluster_printUiUsage(void);

int 				lnm_updateConfig(opt_entry_t * entry);

#endif /* _LNM_LOCAL_H */
