// vi:ts=4
/*-----------------------------------------------------------------
 *
 * Name: mcp.c
 *
 * RCS:  $Id: mcp.c,v 1.14 2002/09/04 01:49:58 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: control process for SSC 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created: 02/28/2002
 *
 *-----------------------------------------------------------------
 */

#include <sys/types.h>
#include <syslog.h>
#include <stdio.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <sys/resource.h>

#define MAX_ARGS 50
#define CANT_RUN 127

#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif

struct mcp_config_entry
{
	char *action;				/* One of "initwait", "donewait" */
	char *command;				/* Command with aguments */
	char **command_args;		/* Parsed arguments */
};


struct mcp_config
{
	int n_entries;				/* Number of entries */
	int n_current;				/* Active entries */
	struct mcp_config_entry *entries; /* Array of config entries */
};

/* For the debug case, we will only allow a child to die a limited number
   of times before we terminate everything.
   Undef this when we're done with the debugging.
 */
#define MAX_DEATHS 10
#undef MAX_DEATHS

struct mcp_proc_info
{
	pid_t child_pid;			/* Child pid */
	int first_time;				/* If zero, append "-r" to arguments */
	int wait_init;				/* If non-zero, currently waiting for
								 initialization to finish */
	int death_count;			/* How many times has this died */
	struct mcp_config_entry *config_entry; 
};


struct mcp_run_state
{
	int n_procs;				
	struct mcp_proc_info *procs;
};


const int initial_size = 20;	/* Initial size of config table */
const int increment = 10;		/* config table growth increment */
const char *default_config_file = "/usr/local/agile/etc/mcptab";

/* Directory to run the daemons in */
const char *run_directory = "/var/run"; 

char *PidFile = "/var/run/mcp.pid";

struct mcp_run_state *RunState;

sigset_t wait_set;				/* Signal mask used for sigsuspend() */
sigset_t got_signals;			/* Pending signals */

/*
 * do_log - print log message
 */
static void
do_log(int priority, char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	vsyslog(priority, fmt, va);
	va_end(va);
}


/*
 * terminate - print log message and exit
 */
static void
terminate()
{
	int i;
	do_log(LOG_CRIT, "terminating: %m");
	if (RunState) {
		for (i = 0; i < RunState->n_procs; ++i) {
			int pid = RunState->procs[i].child_pid;
			if (pid != 0) {
				int status;
				debug(("terminating child %d\n", pid));
				kill(pid, SIGTERM);
				waitpid(pid, &status, 0);
				debug(("child %d terminated\n", pid));
			}
		}
	}
	(void)unlink(PidFile);
	exit(EXIT_FAILURE);
}


/*
 * make_arguments - parse the command line in entry into arguments
 */
static int
make_arguments(char **arguments, struct mcp_config_entry *entry)
{
	char *command = strdup(entry->command);
	char *str;
	int argno = 0;
	if (command == NULL)
		terminate();
	while ((str = strsep(&command, " \t")) != NULL) {
		if (str[0] != 0) {
			arguments[argno] = str;
			++argno;
			if (argno >= MAX_ARGS)
				terminate();
		}
	}
	arguments[argno] = NULL;

	return argno;
}



/*
 * copy_arguments - allocate array of arguments and copy from args
 */
static char **
copy_arguments(char **args, int count)
{
	/* Extra for inserting "-r" for restart runs */
	char **new_args = malloc(sizeof(char *) * (count + 2));
	int idx;
	if (new_args == NULL)
		terminate();
	for (idx = 0; idx < count; ++idx) {
		new_args[idx] = args[idx];
	}
	new_args[idx] = NULL;
	return new_args;
}


/*
 * make_restart_arguments - make arguments array when restarting
 */
static char **
make_restart_arguments(char **args)
{
	int idx = 0;
	while (args[idx] != NULL) {
		++idx;
	}
	args[idx] = "-r";
	args[idx + 1] = NULL;

	return args;
}


/*
 * config_expand - add more entries to config table
 */
static int
config_expand(struct mcp_config *config)
{
	int ok = 0;
	struct mcp_config_entry *new_entries =
		realloc(config->entries, (config->n_entries + increment)
				* sizeof(struct mcp_config_entry));
	if (new_entries) {
		config->entries = new_entries;
		config->n_entries += increment;
		ok = 1;
	}

	return ok;
}


/*
 * config_alloc_entry - allocate a new entry in the config table
 */
static struct mcp_config_entry *
config_alloc_entry(struct mcp_config *config)
{
	struct mcp_config_entry *entry = NULL;
	if (config->n_current < config->n_entries || config_expand(config)) {
		entry = &config->entries[config->n_current++];
	}
	return entry;
}


/*
 * alloc_config - allocate config table
 */
static struct mcp_config *
alloc_config()
{
	struct mcp_config *config = malloc(sizeof(struct mcp_config));
	config->entries = malloc(initial_size * sizeof(struct mcp_config_entry));
	if (config->entries != NULL) {
		config->n_entries = initial_size;
		config->n_current = 0;
	}
	else {
		free(config);
		config = NULL;
	}
	return config;
}


/*
 * print_entry -
 */
static void
print_entry(struct mcp_config_entry *entry)
{
	int i;
	
	debug(("action=%s\n", entry->action));
	debug(("command=%s\n", entry->command));

	for (i = 0; entry->command_args[i] != NULL; ++i) {
		debug(("arg[%d]=%s\n", i, entry->command_args[i]));
	}
}

	
/*
 * fill_entry - fill the config entry from string str
 */
static struct mcp_config_entry *
fill_entry(struct mcp_config_entry *entry, char *str)
{
	char *action;

	action = strsep(&str, ":");
	if (!strcmp(action, "initwait"))
		entry->action = "initwait";
	else if (!strcmp(action, "donewait"))
		entry->action = "donewait";
	else
		entry = NULL;

	if (entry) {
		entry->command = strdup(str);
		if (!entry->command) {
			entry = NULL;
		}
		else {
			char *args[MAX_ARGS];
			int arg_count = make_arguments(args, entry);
			entry->command_args = copy_arguments(args, arg_count);

			print_entry(entry);
		}
	}

	return entry;
}


/*
 * parse_config_line - allocate new config entry and fill it
 */
static struct mcp_config_entry *
parse_config_line(struct mcp_config *config, char *str)
{
	struct mcp_config_entry *new_entry;

	if ((new_entry = config_alloc_entry(config))) {
		new_entry = fill_entry(new_entry, str);
	}

	return new_entry;
}


/*
 * read_config_file - read configuration file f into config table config
 */
static int
read_config_file(FILE *f, struct mcp_config *config)
{
	int ok = 1;

	while (ok && !feof(f)) {
		size_t size;
		char *str = fgetln(f, &size);
		if (str != NULL) {
			/* Skip comments and empty lines */
			if (str[0] != '#' && str[0] != ' ' && str[0] != '\t') {
				if (str[size - 1] != '\n')
					terminate();
				str[size - 1] = 0;
				ok = (int)parse_config_line(config, str);
				str[size - 1] = '\n';
			}
		}
	}

	return ok;
}


/*
 * need_init_wait - return true if need to wait for initialization to finish
 * for this process
 */
static int
need_init_wait(struct mcp_proc_info *info)
{
	return !strcmp(info->config_entry->action, "initwait");
}


/*
 * read_config - allocate config table and read it from file filename
 */
static struct mcp_config *
read_config(const char *filename)
{
	FILE *f = NULL;
	struct mcp_config *config = NULL;

	if ((f = fopen(filename, "r")) && (config = alloc_config())) {
		if (!read_config_file(f, config)) {
			free(config);
			config = NULL;
		}
	}

	if (f != NULL)
		fclose(f);

	return config;
}


/*
 * sigwait_handler - signal handler called when doing sigsuspend
 */
static void
sigwait_handler(int signal)
{
	debug(("got signal %d\n", signal));
	if (signal == SIGTERM) {
		terminate();
	}
	sigaddset(&got_signals, signal);
}



/*
 * install_handlers - install default signal handlers
 */
static void
install_handlers(sigset_t *wait_sigs)
{
	int sig;
	struct sigaction act;
	act.sa_handler = sigwait_handler;
	act.sa_mask = *wait_sigs;
	act.sa_flags = 0;
	for (sig = 1; sig < NSIG; ++sig) {
		if (sigismember(wait_sigs, sig)) {
			int err = sigaction(sig, &act, NULL);
			if (err != 0)
				terminate();
		}
	}
}


/*
 * init_signals - initialize signal masks
 */
static void
init_signals(void)
{
	sigset_t set;

	sigemptyset(&set);
	sigaddset(&set, SIGCHLD);
	sigaddset(&set, SIGUSR1);
	sigaddset(&set, SIGTERM);
	sigprocmask(SIG_BLOCK, &set, &wait_set);

	install_handlers(&set);
}


/*
 * init_proc_info - initalize proc info structure for config entry
 */
static void
init_proc_info(struct mcp_proc_info *info, struct mcp_config_entry *entry)
{
	info->first_time = 1;
	info->wait_init = 0;
	info->child_pid = 0;
	info->death_count = 0;
	info->config_entry = entry;
}


/*
 * init_run_state - initialize process table
 */
static struct mcp_run_state *
init_run_state(struct mcp_config *config)
{
	struct mcp_run_state *run_state = malloc(sizeof(struct mcp_run_state));

	if (run_state) {
		run_state->procs = malloc(config->n_current
								  * sizeof(struct mcp_proc_info));
		if (run_state->procs) {
			int i;
			run_state->n_procs = config->n_current;
			for (i = 0; i < run_state->n_procs; ++i) {
				init_proc_info(&run_state->procs[i], &config->entries[i]);
			}
		}
		else {
			free(run_state);
			run_state = NULL;
		}
	}

	return run_state;
}



/*
 * sigisemptyset - return true if the signal set is empty
 */
static int
sigisemptyset(sigset_t *wait_sigs)
{
	return *wait_sigs == 0;
}


/*
 * get_pendingsignals - return currently pending signals
 */
static sigset_t
get_pendingsignals()
{
	sigset_t set;
	sigpending(&set);
	return set;
}


/*
 * first_pendingsignal - return the first pending signal in the set
 */
static int
first_pendingsignal(sigset_t *set)
{
	int sig;
	for (sig = 1; sig < NSIG; ++sig) {
		if (sigismember(set, sig))
			return sig;
	}
	return 0;
}


/*
 * my_sigwait - kind of sigwait(3)
 */
static int
my_sigwait(sigset_t *wait_sigs, int *sig)
{
	int got_sig;
	if (sigisemptyset(&got_signals)) {
		got_signals = get_pendingsignals() & *wait_sigs;
		if (sigisemptyset(&got_signals)) {
			sigset_t suspend_sigs = ~*wait_sigs;
			int rc;
			rc = sigsuspend(&suspend_sigs);
			assert(rc == -1 && errno == EINTR);
		}
	}

	got_sig = first_pendingsignal(&got_signals);
	assert(got_sig != 0);
	sigdelset(&got_signals, got_sig);
	*sig = got_sig;
	return 0;
}


/*
 * run_child - execv new process
 */
static int
run_child(struct mcp_proc_info *info)
{
	pid_t pid;
	int rc = 1;

	if (info) {
		debug(("running_child: first_time=%d\n", info->first_time));
		print_entry(info->config_entry);

		pid = fork();
		switch (pid) {
		case -1:
			/* Can't fork */
			rc = 0;
			break;
		case 0:
			/* Child */
		{
			sigset_t set;
			sigfillset(&set);
			sigprocmask(SIG_UNBLOCK, &set, 0);
			/* If not first run, run with modified arguments */
			if (execv(info->config_entry->command_args[0],
					  info->first_time ?
					  info->config_entry->command_args :
					  make_restart_arguments(
						  info->config_entry->command_args)))
				do_log (LOG_EMERG, "Unable to exec %s because %m", 
				        info->config_entry->command_args[0]);
				exit(CANT_RUN);
			break;
		}
		default:
			info->child_pid = pid;
			if (need_init_wait(info))
				info->wait_init = 1;
			info->first_time = 0;
			rc = 1;
			break;
		}
	}

	return rc;
}


/*
 * find_info_for_pid - find proc info structure for process id pid
 */
static struct mcp_proc_info *
find_info_for_pid(struct mcp_run_state *run_state, pid_t pid)
{
	int i = 0;
	for (i = 0; i < run_state->n_procs; ++i) {
		struct mcp_proc_info *info = &run_state->procs[i];
		if (info->child_pid == pid)
			return info;
	}
	return NULL;
}


/*
 * find_child_to_run - find the next process to run. Return NULL if
 * there are no processes to run.
 *
 * The processes will be run if:
 * 1) child_pid == 0 && need_init_wait - need to restart this process
 * 2) child_pid == 0 && !need_init_wait && first_time - run process once
 * If there is process with child_pid != 0 and either need_init_wait and
 * waiting for initialization to finish, or waiting for the process to
 * finish, return NULL
 */
static struct mcp_proc_info *
find_child_to_run(struct mcp_run_state *run_state)
{
	int i;
	for (i = 0; i < run_state->n_procs; ++i) {
		struct mcp_proc_info *info = &run_state->procs[i];
		if (info->child_pid == 0 &&
			(need_init_wait(info) || info->first_time)) {
			return info;
		}
		if (info->child_pid != 0 && ((need_init_wait(info) &&
			info->wait_init) || !need_init_wait(info))) {
			return NULL;
		}
	}

	return NULL;
}


/*
 * find_info_waiting - find the process which we wait for initialization
 * to complete
 */
static struct mcp_proc_info *
find_info_waiting(struct mcp_run_state *run_state)
{
	int i;
	for (i = 0; i < run_state->n_procs; ++i) {
		struct mcp_proc_info *info = &run_state->procs[i];
		if (info->wait_init)
			return info;
	}
	return NULL;
}


/*
 * proc_name - return the process name
 */
static const char *
proc_name(struct mcp_proc_info *info)
{
	return info->config_entry->command_args[0];
}


/*
 * update_proc_info_exited -
 */
static void
update_proc_info_exited(struct mcp_proc_info *info)
{
	if (info) {
		info->child_pid = 0;
		info->wait_init = 0;
	}
}


/*
 * update_proc_info_terminated -
 */
static void
update_proc_info_terminated(struct mcp_proc_info *info)
{
	if (info) {
		info->child_pid = 0;
		info->wait_init = 0;
	}
}


/*
 * update_proc_info_init_done -
 */
static void
update_proc_info_init_done(struct mcp_proc_info *info)
{
	if (info) {
		debug(("%s finished initialization\n", proc_name(info)));
		do_log(LOG_LOCAL0 | LOG_INFO,
			   "%s finished initialization\n", proc_name(info));
		info->wait_init = 0;
	}
}


/*
 * childs_status_change - process SIGCHLD
 */
static void
childs_status_change(struct mcp_run_state *run_state)
{
	int status;
	pid_t pid;
	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		struct mcp_proc_info *info = find_info_for_pid(run_state, pid);
		if (WIFEXITED(status)) {
			int pri;
			assert(info != NULL);
			debug(("%s exited with status %d\n", proc_name(info),
				   WEXITSTATUS(status)));
			if (need_init_wait(info))
				pri = LOG_LOCAL0 | LOG_ERR;
			else
				pri = LOG_LOCAL0 | LOG_INFO;
			do_log(pri, "%s exited with status %d", proc_name(info),
				   WEXITSTATUS(status));
			update_proc_info_exited(info);
		}
		else if (WIFSIGNALED(status)) {
			assert(info != NULL);
			debug(("%s terminated due to signal %s\n", proc_name(info),
				   sys_signame[WTERMSIG(status)]));
			do_log(LOG_LOCAL0 | LOG_ERR,
				   "%s terminated due to signal %s", proc_name(info),
				   sys_signame[WTERMSIG(status)]);
			update_proc_info_terminated(info);
		}
		else {
			assert(0);
		}
#if MAX_DEATHS
		info->death_count++;
		if (info->death_count > MAX_DEATHS) {
			do_log (LOG_CRIT, "MCP client %s had died %d times\n", 
			         proc_name(info), info->death_count);
			do_log (LOG_CRIT, "MCP terminating!!!!!!!!!!!!!\n");
			terminate();
		}
#endif
	}
}


/*
 * child_init_done -
 */
static void
child_init_done(struct mcp_run_state *run_state)
{
	update_proc_info_init_done(find_info_waiting(run_state));
}


/*
 * process_signals - wait for the next signal and process it. If there are
 * no signals pending, will block until next signal. Will block only once.
 */
static void
process_signals(struct mcp_run_state *run_state)
{
	sigset_t set;
	int sig;
	int error;
	sigemptyset(&set);
	sigaddset(&set, SIGCHLD);
	sigaddset(&set, SIGUSR1);
	sigaddset(&set, SIGTERM);
	error = my_sigwait(&set, &sig);
	if (error) {
		terminate("sigwait error");
	}
	if (sig == SIGTERM) {
		terminate();
	}
	assert(sig == SIGCHLD || sig == SIGUSR1);
	switch (sig) {
	case SIGCHLD:
		childs_status_change(run_state);
		break;
	case SIGUSR1:
		child_init_done(run_state);
		break;
	}
	if (!sigisemptyset(&got_signals) || sigpending(&set)) {
		/* my_sigwait will not block */
		process_signals(run_state);
	}
}


/*
 * run_procs - run the processes in the config table
 */
static void
run_procs(struct mcp_config *config)
{
	struct mcp_run_state *run_state = init_run_state(config);
	int ok = 1;
	init_signals();
	RunState = run_state;
	run_child(find_child_to_run(run_state));
	while (ok) {
		process_signals(run_state);
		ok = run_child(find_child_to_run(run_state));
	}
}



#define OUR_DATA_LIMIT 5 * 1024 * 1024
#define OUR_STACK_LIMIT 2 * 1024 * 1024


/*
 * set_limits - initalize our resource limits.
 */
static void
set_limits(void)
{
	struct rlimit rl;
	rl.rlim_cur = OUR_DATA_LIMIT;
	rl.rlim_max = RLIM_INFINITY;
	if (setrlimit(RLIMIT_DATA, &rl) != 0) {
		do_log(LOG_WARNING,"Failed to set rlimit for data: %m\n");
	}
 	rl.rlim_cur = OUR_STACK_LIMIT;
	rl.rlim_max = RLIM_INFINITY;
	if (setrlimit(RLIMIT_STACK, &rl) != 0) {
	 	do_log(LOG_WARNING, "Failed to set rlimit for stack: %m\n");
	}
}



int
main(int argc, char **argv)
{
	const char *config_file = NULL;
	FILE *fp;
	
	switch (argc) {
	case 1:
		config_file = default_config_file;
		break;
	case 2:
		config_file = argv[1];
		break;
	default:
		do_log(LOG_LOCAL0 | LOG_ERR, "invalid arguments");
		break;
	}

	if (config_file) {
		struct mcp_config *config = read_config(config_file);

		daemon (1,1);

		if (chdir(run_directory) != 0) {
			do_log(LOG_WARNING, "Can not chdir to %s: %m\n", run_directory);
		}

		fp = fopen(PidFile, "w");
		if (fp != NULL) {
			fprintf(fp, "%d\n", getpid());
			(void) fclose(fp);
		}

		set_limits();

		if (config) {
			run_procs(config);
		}
	}
	
	exit(EXIT_FAILURE);
	return 1;
}
