/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: am_get_our_addr.c
 *
 * RCS:  $Id: am_get_our_addr.c,v 1.3 2002/11/22 20:20:42 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Determines the local address we'll end up using to
 *              connect to the host/address "destName".
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Oct 2001
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <ifaddrs.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/route.h>
#include <net/if_dl.h>
#include <stdio.h>
#include <netdb.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

/*-----------------------------------------------------------------
 * Name: get_our_addr
 *
 * Description: Given a host name or IP address string, determine the
 *              local IP address we will use when we make a connection.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Oct 2001
 *
 *-----------------------------------------------------------------
 */
extern int
get_our_addr (in_addr_t them, char *iface, in_addr_t *us)
{
	int rteSock;
	struct {
		struct	rt_msghdr	hdr;
		char	space[512];
	} rteMsg;		// Route request message buffer
	struct sockaddr_in	*dstSa = (struct sockaddr_in *)rteMsg.space;
	struct sockaddr 	*which;
	struct sockaddr		*gateWay = NULL;	// Which gateway used in this route
	int rtn;
	int i;
	char *lookp;
	struct ifaddrs *ifAddrLst, *addrp;
	in_addr_t hisAddr, myAddr = { 0 }, myMask;

	int pid = getpid();	// Use this to tag our routing socket request
/*
	This code was created by looking at what "route get" did to determine
	the route (using the routing socket) and what ifconfig did to display
	interface's addresses.

	It's pretty grungy because there really isn't any documentation as to
	how one's supposed to collect this kind of information.

	Figure out what the IP address is of the destination network node
	given either its name or IP address.
	We finish up with the destination address, in network form, in 
	dstAddr.sin_addr. The family MUST be AF_INET.
 */
	memset (&rteMsg, 0, sizeof (rteMsg));	// Start off clean

	dstSa->sin_addr.s_addr = them;
	dstSa->sin_family = AF_INET;
	dstSa->sin_len = sizeof (struct sockaddr_in);
/*
	We will need the link record too so add that request
 */
	which = (struct sockaddr *)&(rteMsg.space[dstSa->sin_len]);
	which->sa_family = AF_LINK;
	which->sa_len = sizeof(struct sockaddr_dl);
/*
	To get the route to the given host, we need to open a routing socket
	on which to send the GET request and receive the response.
 */
	rteSock = socket (PF_ROUTE, SOCK_RAW, 0);
	if (rteSock < 0) {
		perror ("Routing socket open error");
		return 0;
	}
/*
	Form the request message and send it
 */
	rteMsg.hdr.rtm_type = RTM_GET;
	rteMsg.hdr.rtm_flags = RTF_UP | RTF_STATIC | RTF_HOST | RTF_GATEWAY;
	rteMsg.hdr.rtm_version = RTM_VERSION;
	rteMsg.hdr.rtm_addrs = RTA_DST | RTA_IFP;
	rteMsg.hdr.rtm_seq = pid;
	rteMsg.hdr.rtm_msglen = sizeof (struct rt_msghdr) + 
	                        sizeof (struct sockaddr_in) + 
	                        sizeof (struct sockaddr_dl);

	if (write (rteSock, &rteMsg, rteMsg.hdr.rtm_msglen) < 0) {
		close (rteSock);
		return -1;
	}
/*
	Get the results
 */
	do {
		rtn = read (rteSock, &rteMsg, sizeof (rteMsg));
	} while ((rtn > 0) && 
	         ((rteMsg.hdr.rtm_seq != pid) || (rteMsg.hdr.rtm_pid != pid)));
	close (rteSock);	// Done with it
/*
	Any errors?
 */
	if (rtn < 0) {
		return -1;
	}
	if (rteMsg.hdr.rtm_errno) {
		errno = rteMsg.hdr.rtm_errno;
		return -1;
	}
/*
	Let's pick out the interface and gateway that we'll be using to talk to
	destName.
	The gateway's address tells us which, of the possibly many addresses,
	that will be used.
 */
	lookp = rteMsg.space;
	for (i = 1; i; i <<= 1) {
		if (i & rteMsg.hdr.rtm_addrs) {
			which = (struct sockaddr *)lookp;

			if (i == RTA_GATEWAY) {
				gateWay = which;
				break;
			}

			lookp += ((which->sa_len + 3) & ~3); // sockaddrs on word boundaries
		}
	}
/*
	Make sure we got the interface and gateway address
 */
	if (!gateWay) {
		errno = ESRCH;
		return -1;
	}
/*
	Get the interface list and go through it to figure out what our 
	local address is.
 */
	if (getifaddrs (&ifAddrLst) < 0) {
		errno = ESRCH;
		return -1;
	}
/*
	We ahve two ways of identfing the interface:
	1) The "gateWay" is AF_LINK when the destination address is directly
	   accessable from one of our interfaces; or
	2) The "gateway" is AF_INET and we'll have to match the interface
	   with its address.
 */
	if (gateWay->sa_family == AF_INET) {
		hisAddr = ((struct sockaddr_in *)gateWay)->sin_addr.s_addr;
	}
	else {
		hisAddr = dstSa->sin_addr.s_addr;
	}

	for (addrp = ifAddrLst; addrp; addrp = addrp->ifa_next) {
		if (addrp->ifa_addr->sa_family == AF_INET) {
			myAddr = ((struct sockaddr_in *)(addrp->ifa_addr))->sin_addr.s_addr;
			myMask = 
			      ((struct sockaddr_in *)(addrp->ifa_netmask))->sin_addr.s_addr;
			if ((myAddr & myMask) == (hisAddr & myMask)) {
				strcpy (iface, addrp->ifa_name);
				break;
			}
		}
	}

	freeifaddrs (ifAddrLst);		// Done with it so release the memory
/*
	That's all folks!!!
 */
	*us = myAddr;
	return 0;
}
