/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_nfs.c,v 1.31 2002/10/02 21:32:06 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#include <exports/evm-api.h>
#include <exports/anpssc-api.h>
#include <exports/ncm-api.h>
#include <exports/rpc.h>
#include <exports/fs-api.h>
#include <exports/mount.h>
#include "exports/ipmd-lib.h"


#define		NFSCMD_HELP "NFS Commands\n"
#define		NFSCMD_ERR	"NFS CMD ERR"
#define		NFSCMD_SHARE_DEL_HELP "nfs share delete PATHNAME [-a A.B.C.D][-n CLUSTERNODE]"
#define		NFSCMD_EXECUTABLE "/usr/local/agile/bin/nfscmd"

#define		NFS_DBG_PRINTF(x) 		
#define		NFSSH_OPT_KEYWORD_LEN	40
#define		NFSSH_OPT_BUFLEN		128	
#define		NFSSH_CLIENT_BUFLEN		512
#define		NFSSH_CLUSTER_BUFLEN	64
#define		NFS_MAX_NAME_LEN		255
#define		NFS_MAX_PATHNAME_LEN	1024
#define		MSG_BUF_LEN				2048
//define default NFS export
#define		DEFAULT_LADDR_STR		"0.0.0.0"
#define		DEFAULT_LADDR			0
#define		EXPORT_OPT_STR			"rw,root_squash=all"
#define		ALL_ADDR				0xffffffff

typedef enum
{
	NFS_EXPORTFS=0,			//exportfs 
	NFS_PORTMAP_REG,		//register ports with portmapper
} nfs_cmdType;

struct 
{
	nfs_cmdType	type;
	char		*cmd;		//cmd name for nfscmd ->argv[1]
} nfs_cmds[] =
{
	{NFS_EXPORTFS, "exportfs"},
	{NFS_PORTMAP_REG, "portmapreg"},
};

char laddr_str[] = {'l','a','d','d','r','='};

void cmd_nfsPrintMntRespStatus(int32 status);
void cmd_nfsPrepareNcmMsg(char				*reqMsgBuf,
						  char				*respMsgBuf,
						  ncmMsg_t			**ncmReq,
						  ncmMsg_t			**ncmRsp,
						  uint32			*ncmReqSize,
						  uint32			*ncmRspSize,
						  mnt_infoReq_t		**mntInfoReqp,
						  mnt_infoResp_t	**mntInfoResp,
						  char				*clusterNode);
int cmd_nfsMntUnexportReq(char *pathName, uint32 localIpAddr, char *clusterNode);
int cmd_nfsMntExportReq(char *pathName, uint32 localIpAddr,
						char *optionp, char *clusterNode);
int cmd_nfsExtractOption(char *optionp, char *optionKeyWordp);
int cmd_nfsCheckOption (char *optionStr);
int nfs_cmdGetVolumeName(char *volNamep, int *volNameLen,
						char *pathNamep, int pathNameLen);
int cmd_nfsValidateOption(char *optionp, int optionLen);
evm_volId_t cmd_nfsGetVolId(char *pathNamep, int32 pathNameLen);
void cmd_nfsBuildOptionStr(char *optStr, int optionMask);
void cmd_nfsDisplayEntries(mnt_infoResp_t *mntInfoResp);

#define BUF_SIZE    1024

/*-----------------------------------------------------------------
 * Name:   	cmd_nfsIsIpIfValid 
 *  
 * Description:  
 *	Check the operation state of input IP interface.
 *
 * Created by: Joseph Vu 
 *
 * Date Created: 17 Sep 2002
 *
 *-----------------------------------------------------------------
 */
static 
int		//1: valid
		//0: Error 
cmd_nfsIsIpIfValid(uint32 addr, char *clusterNodep)
{
	ipm_ip_t    *ipm_ipp;
	uint64      result_buf[BUF_SIZE];
	int			valid;
	int         count=0, need=0;

	valid = 1;

	if (!addr)
		return (valid);

	if (ipmd_GetIpByNameCluster(NULL, (ipm_ip_t*)result_buf,
				BUF_SIZE*sizeof(uint64),
			&count, &need, clusterNodep) != IPM_GET_OK)
	{
		return valid;
	}
	for (ipm_ipp = (ipm_ip_t*)result_buf; count >0; count--, ipm_ipp++) 
	{
		if (addr == ipm_ipp->ip_addr)
			return (valid);
	}
	return (0);
}

/*-----------------------------------------------------------------
 * Name:	cmd_nfsPrintMntRespStatus
 *
 * Description: Print status of response from mountd.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created: 27 Apr 2002
 *
 *-----------------------------------------------------------------
 */
void
cmd_nfsPrintMntRespStatus(int32 status)
{
	switch(status)
	{
		case MOUNTD_ERR_VOL_NOT_FOUND:
			printf("Error: Volume is not found!\n");
			break;

		case MOUNTD_ERR_BAD_VOL_ID:
			printf("Error: Invalid volume ID!\n");
			break;

		case MOUNTD_ERR_PERMISSION_DENIED:
			printf("Error: Permission denied!\n");
			break;

		case MOUNTD_ERR_INVALID_CLIENT_PATH:
			printf("Error: Invalid client:/pathname format!\n");
			break;

		case MOUNTD_ERR_NO_HOST_FOUND:
			printf("Error: Client is not found!\n");
			break;

		case MOUNTD_ERR_NO_EXPORT_FOUND:
			printf("Error: Export is not found! Incorrect path name.\n");
			break;

		case MOUNTD_ERR_CONFLICT_EXPORT_OPT:
			printf("Error: Conflicting in export option permission!\n");
			break;

		case MOUNTD_ERR_INVALID_EXPORT_OPT:
			printf("Error: Invalid export option!\n");
			break;

		case MOUNTD_ERR_DUPLICATE_EXPORT:
			printf("Error: Duplicated export!\n");
			break;

		case MOUNTD_ERR_MORE_GENERIC_EXPORT_FOUND:
			printf("Error: Export with less restriction exists!\n");
			break;

		case MOUNTD_ERR_MORE_SPECIFIC_EXPORT_FOUND:
			printf("Error: Export with more restriction exists!\n");
			break;

		case MOUNTD_ERR_CONFLICT_EXPORT:
			printf("Error: Conflict export!\n");
			break;

		case MOUNTD_ERR_MULTIPLE_INTERFACE:
			printf("Error: Multiple interfaces for a path name are not allowed!\n");
			break;

		case MOUNTD_ERR_ENABLE_FAILURE:
			printf("Error: Share enable failed!\n");
			break;

		case MOUNTD_ERR_DISABLE_FAILURE:
			printf("Error: Share disenable failed!\n");
			break;

		case MOUNTD_ERR_FILESYS_NOT_READY:
			printf("Error: File system is not in ready state!\n");
			break;

		case MOUNTD_ERR_EXPORT_INVALID_0_LOCIP:
			printf("Error: Export is not found! Please provide a specific local IP interface.\n");
			break;

		case MOUNTD_ERR_EXPORT_INVALID_LOCIP:
			printf("Error: Export for this local IP interface is not found!.\n");
			break;

		case MOUNTD_ERR_EXPORT_INVALID_0_CLNT:
			printf("Error: Export is not found! Please provide a specific client.\n");
			break;

		case MOUNTD_ERR_EXPORT_INVALID_CLNT:
			printf("Error: Export for this client is not found!.\n");
			break;

		case MOUNTD_ERR_EXPORT_INVALID_CLNT_LOCIP:
			printf("Error: Export is not found! Incorrect local IP interface and/or client.\n");
			break;

		case MOUNTD_ERR_PATHNAME_NOT_FOUND:
			printf("Error: Path name is not found on this volume!\n");
			break;

		case MOUNTD_ERR_INTERNAL:
			printf("Error: mountd internal error!\n");
			break;

		case MOUNTD_ERR_MAX_NUM_OF_SHARES_REACHED:
			printf("Error: Maximum number of NFS shares (%d) reached!\n",
				MOUNTD_MAX_EXPORTS);
			break;

		default :
			printf("Error: Unknown error!\n");
	}
}

/*-----------------------------------------------------------------
 * Name:	cmd_nfsPrepareNcmMsg
 *
 * Description: Prepare NCM request and response messages
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created: 02 Jul 2002
 *
 *-----------------------------------------------------------------
 */
void cmd_nfsPrepareNcmMsg(char				*reqMsgBuf,
						  char				*respMsgBuf,
						  ncmMsg_t			**ncmReq,
						  ncmMsg_t			**ncmRsp,
						  uint32			*ncmReqSize,
						  uint32			*ncmRspSize,
						  mnt_infoReq_t		**mntInfoReqp,
						  mnt_infoResp_t	**mntInfoResp,
						  char				*clusterNode)
{
	// Prepare NCM request
	*ncmReqSize = sizeof(ncmMsg_t) + sizeof(mnt_infoReq_t)
											- sizeof(long long);
	memset(reqMsgBuf, 0, *ncmReqSize);
	*ncmReq = (ncmMsg_t *)reqMsgBuf;
	(*ncmReq)->msgType = NCM_SEND;
	strcpy((*ncmReq)->appName, MOUNTD_APP_NAME);
	strcpy((*ncmReq)->nodeName, clusterNode);
	(*ncmReq)->rqResp = 1;
	*mntInfoReqp = (mnt_infoReq_t *) ((*ncmReq)->appMsg);

	// Prepare space for NCM response
	*ncmRspSize = sizeof(ncmMsg_t) + sizeof(mnt_infoResp_t)
											- sizeof(long long);
	memset(respMsgBuf, 0, *ncmRspSize);
	*ncmRsp = (ncmMsg_t *)respMsgBuf;
	*mntInfoResp = (mnt_infoResp_t *) ((*ncmRsp)->appMsg);
}

/*-----------------------------------------------------------------
 * Name:	cmd_nfsMntUnexportReq 
 *
 * Description: Send UNEXPORT request to mountd.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created: 23 Apr 2002
 *
 *-----------------------------------------------------------------
 */
int
cmd_nfsMntUnexportReq(char *pathName, uint32 localIpAddr, char *clusterNode)
{
	char				reqMsgBuf[MSG_BUF_LEN];
	char				respMsgBuf[MSG_BUF_LEN];
	ncmMsg_t			*ncmReq;
	ncmMsg_t			*ncmRsp;
	uint32				ncmReqSize;
	uint32				ncmRspSize;
	mnt_infoReq_t		*mntInfoReqp;
	mnt_infoResp_t		*mntInfoResp;
	int					nameLen = 0;
	int					stat;
	struct timeval		timeout = { 10, 0};

	if (pathName && ((nameLen = strlen(pathName)) == 0))
		return (CMD_ERR_FAILED);

	// Prepare NCM request and response
	cmd_nfsPrepareNcmMsg(reqMsgBuf,
						 respMsgBuf,
						 &ncmReq,
						 &ncmRsp,
						 &ncmReqSize,
						 &ncmRspSize,
						 &mntInfoReqp,
						 &mntInfoResp,
						 clusterNode);

	//Build MOUNTD request message
	mntInfoReqp->msgHdr.private = 0;
	mntInfoReqp->msgHdr.msgType = MNT_UNEXPORT_REQ;
	if (pathName)
	{
		mntInfoReqp->nameLen = nameLen;
		strcpy(&mntInfoReqp->pathName[0], pathName);
	}
	mntInfoReqp->localIpAddr = localIpAddr;

	NFS_DBG_PRINTF(("cmd_nfsMntUnexportReq: "
					"pathName: %s nameLen: %d localIpAddr: %08x\n",
					mntInfoReqp->pathName,
					mntInfoReqp->nameLen,
					mntInfoReqp->localIpAddr));

	//Send request
	if (pathName)
		ncmReqSize += nameLen + 1;
	stat = sendAgileMsg (-1, ncmReq, ncmReqSize,
						"ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout, ncmRsp, ncmRspSize);
	if (stat < 0) 
	{
		printf("Error: NFS unexport failed!\n");
		NFS_DBG_PRINTF("cmd_nfsMntUnexportReq: sendAgileMsg failed!\n");
		return (CMD_ERR_FAILED);
	}

	if (mntInfoResp->msgHdr.status != NFX_OK)
	{
		cmd_nfsPrintMntRespStatus(mntInfoResp->msgHdr.status);
		NFS_DBG_PRINTF("cmd_nfsMntUnexportReq: NFS unexport failed!\n");
		return (CMD_ERR_FAILED);
	}

	return (CMD_SUCCESS);
} /* cmd_nfsMntUnexportReq() */

/*-----------------------------------------------------------------
 * Name:	cmd_nfsMntExportReq 
 *
 * Description: Send EXPORT request to mountd.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created: 23 Apr 2002
 *
 *-----------------------------------------------------------------
 */
int
cmd_nfsMntExportReq(char *pathName, uint32 localIpAddr,
					char *optionp, char *clusterNode)
{
	char				reqMsgBuf[MSG_BUF_LEN];
	char				respMsgBuf[MSG_BUF_LEN];
	ncmMsg_t			*ncmReq;
	ncmMsg_t			*ncmRsp;
	uint32				ncmReqSize;
	uint32				ncmRspSize;
	mnt_infoReq_t		*mntInfoReqp;
	mnt_infoResp_t		*mntInfoResp;
	int					nameLen, optLen;
	char				*reqOptionp;
	int					stat;
	struct timeval		timeout = { 10, 0};

	nameLen = strlen(pathName);
	optLen = strlen(optionp);

	if ((nameLen <= 0) || (optLen <= 0))
		return (CMD_ERR_FAILED);

	// Prepare NCM request and response
	cmd_nfsPrepareNcmMsg(reqMsgBuf,
						 respMsgBuf,
						 &ncmReq,
						 &ncmRsp,
						 &ncmReqSize,
						 &ncmRspSize,
						 &mntInfoReqp,
						 &mntInfoResp,
						 clusterNode);

	//Build MOUNTD request message
	mntInfoReqp->msgHdr.private = 0;
	mntInfoReqp->msgHdr.msgType = MNT_EXPORT_REQ;

	mntInfoReqp->nameLen = nameLen;
	strcpy(&mntInfoReqp->pathName[0], pathName);
	mntInfoReqp->optionsLen = optLen;
	reqOptionp = &mntInfoReqp->pathName[0] + nameLen;
	strcpy(reqOptionp, optionp);

	mntInfoReqp->localIpAddr = localIpAddr;

	//Send request
	ncmReqSize += nameLen + optLen + 2;
	stat = sendAgileMsg (-1, ncmReq, ncmReqSize,
						"ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout, ncmRsp, ncmRspSize);
	if (stat < 0) 
	{
		printf("Error: NFS export failed!\n");
		NFS_DBG_PRINTF("cmd_nfsMntExportReq: sendAgileMsg failed!\n");
		return (CMD_ERR_FAILED);
	}

	if (mntInfoResp->msgHdr.status != NFX_OK)
	{
		cmd_nfsPrintMntRespStatus(mntInfoResp->msgHdr.status);
		NFS_DBG_PRINTF("cmd_nfsMntExportReq: NFS export failed!\n");
		return (CMD_ERR_FAILED);
	}

	return (CMD_SUCCESS);
} /* cmd_nfsMntExportReq() */

/*-----------------------------------------------------------------
 * Name:	nfs_shareDel 
 *
 * Description: NFS unexport or delete share.
 *		nfs share del <<client|all>:/pathname | /pathname | all> [-a <A.B.C.D|all>][-n CLUSTERNODE] 
 * Created by:  Joseph Vu
 *
 * Date Created: 11 Feb 2002
 *
 *-----------------------------------------------------------------
 */
DEFUN   (nfs_shareDel,
         nfs_share_del_cmd,
         "nfs share delete PATHNAME [-a A.B.C.D] [-n CLUSTERNODE]",
         NFSCMD_HELP
		 "Unexport \n"
		 "Delete share\n"
         "Directory pathname as [[CLIENTLIST: | all:]/pathname | all]\n"
         "IP address of the unexport interface\n"
         "A.B.C.D|all\n"
		 "Which node\n"
		 "Node name\n")


{
	uint32	localIpAddr;
	char    clusterNode[NFSSH_CLUSTER_BUFLEN];
	char	cmdBuf[NFS_MAX_NAME_LEN + NFS_MAX_PATHNAME_LEN + 3];
    char    *argvp;
	char	*pathNamep, *hostNameListp, *hostNamep;
	char	*cmdBufp;
	int		pathNameLen, hostNameLen;	
	int		hostCount;
	int		c;
	int		opt_num = 0;    /* num. of options */
	int		my_argc = 1;
	char	*ridx;
	int		return_code, final_return_code;


#ifdef    NOTUSED
    {  
        //For debug only
        int     i;
        printf("nfs share del: argc: %d\n", argc);
        for (i=0; i < argc; i++)
        {
            printf("argv[%d]: %s\n", i, argv[i]);
        }
    }      
#endif    //NOTUSED

	strcpy(clusterNode, "local");
	//set default option
	localIpAddr = DEFAULT_LADDR;
	optind = 1;
	while ((c = getopt (argc, argv, "a:n:")) != -1)
	{
		switch (c)
		{
			case 'a':
			{
				struct in_addr	ip_addr;
				if (strcmp(optarg, "all") == 0)
				{
					localIpAddr = ALL_ADDR;
				}
				else
				{
					if (inet_aton(optarg, (struct in_addr*)&ip_addr) == 0)
					{
						printf("Error: Invalid IP address (%s)!\n", optarg);
						return CMD_ERR_FAILED;
					}
					localIpAddr = ip_addr.s_addr;
				}
				opt_num++;
				break;
			}
			case 'n':
			{
				strcpy(clusterNode, optarg);
				opt_num++;
				break;
			}
		}	//switch
	}	//while

	//check parameter number matches
	my_argc += (2 * opt_num);
	if (my_argc != argc) {
		printf("Error: Incorrect number of parameters, argc %d, in_arg %d\n",
				argc, my_argc);
		return  CMD_ERR_FAILED;
	}


	//parse directory pathname in the form
	//of (client:/pathname or /pathname or client1:client2:/pathname)
	//argv[0] = client:/pathname or /pathname or client1:client2:/pathname

	argvp = argv[0];

	//Check if this has only '/pathname'
	if (*argvp == '/')
	{
		if (strcmp(argvp, "/") == 0)
		{
			printf("Error: Invalid path name!\n");
			return  CMD_ERR_FAILED;
		}
		return (cmd_nfsMntUnexportReq(argvp, localIpAddr, clusterNode));
	}

	//Check if this an delete 'all' option
	if (!strcmp(argvp, "all") ) {
		long long msgBuf[4096];
		ncmNodeList_t *msg = (ncmNodeList_t *)msgBuf;
		ncmRqNodeList_t rq;
		int i;

	// Let's find out what "all" means

		rq.msgType = NCM_RQ_NODE_LIST;
		if (sendAgileMsg (-1, &rq, sizeof (rq), 
		              "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
		              NULL, msgBuf, sizeof (msgBuf)) <= 0) {
			return CMD_ERR_FAILED;
		}

	// And then do them all

		for (i = 0; i < msg->numConnNodes; i++) {
			if (cmd_nfsMntUnexportReq(NULL, localIpAddr, msg->nodeNames[i]) !=
			    CMD_SUCCESS) {
				fprintf (stderr, "Error deleting all shares from %s\n",
				msg->nodeNames[i]);        
			}
		}
		return CMD_SUCCESS;
	}


	//Pathname has the form of 'client:/pathname' or 'client1:client2:/pathname'

	pathNamep = argvp;
	hostNameListp = strsep(&pathNamep, "/");

	if (pathNamep)
	{
		pathNameLen = strlen(pathNamep);

		if ((pathNameLen > NFS_MAX_PATHNAME_LEN))
		{
			printf("Error: Pathname is too long!\n");
			return CMD_ERR_FAILED;
		}
	}
	else
	{
		//Pathname is not specified
		printf("Error: Invalid '/pathname' format!\n");
		printf("%s\n", NFSCMD_SHARE_DEL_HELP);
		return CMD_ERR_FAILED;
	}

	ridx = rindex(hostNameListp, ':');
	if (ridx != (hostNameListp + strlen(hostNameListp) - 1))
	{
		//There is no ":" at the end of client name list
		printf("Error: Invalid 'client:' format!\n");
		printf("%s\n", NFSCMD_SHARE_DEL_HELP);
		return CMD_ERR_FAILED;
	}
	else
	{
		//Remove ":" at the end of client name list
		*ridx = 0;
	}

	hostCount = 0;
	while ((hostNamep = strsep(&hostNameListp, ":")))
	{
		hostCount++;

		hostNameLen = strlen(hostNamep);
		if (hostNameLen == 0)
		{
			printf("Error: Invalid 'client:' format!\n");
			printf("%s\n", NFSCMD_SHARE_DEL_HELP);
			return CMD_ERR_FAILED;
		}
		if (hostNameLen > NFS_MAX_NAME_LEN)
		{
			printf("Error: Client name is too long!\n");
			return CMD_ERR_FAILED;
		}

		//Put in form of client:/pathname
		cmdBufp = &cmdBuf[0];
		strcpy(cmdBufp, hostNamep);
		strcat(cmdBufp, ":/");
		strcat(cmdBufp, pathNamep);

		//Send unexport request to mountd
		return_code = cmd_nfsMntUnexportReq(cmdBufp, localIpAddr, clusterNode);
		if (return_code != CMD_SUCCESS)
		{
			printf("Error: NFS failure to delete share (%s)!\n", cmdBufp);
			final_return_code = CMD_ERR_FAILED;
		}
	}
	if (!hostCount)
	{
		printf("Error: Invalid 'client:' format. No host specified!\n");
		printf("%s\n", NFSCMD_SHARE_DEL_HELP);
		return CMD_ERR_FAILED;
	}

	return (final_return_code);
}

/*-----------------------------------------------------------------
 * Name: nfs_shareAdd
 *
 * Description: NFS Export 
 * Example:
 *	nfs share <PATHNAME> [-a INTERFACE] [-o OPTIONS] [-n CLUSTERNODE]
 *	where,
 *	PATHNAME:  Required pathname of directory in the 
 *  starting with /volumeName
 *	INTERFACE:  Optional received configure IP interface, 
 *	if specified MOUNTD 
 *  shall checked incoming mount request is actually received 
 *  from specified interface
 *	OPTIONS:  [rw|ro|root_squash|root]=client|all
 *		rw: read & write
 *		ro: read only
 *		root_squash: uid = 0 will be mapped by NFS server to non-root UID.
 *  client=list (client1:client2,client3) or 'all' allowing all access.
 *	example:
 *  1.  nfs share add /vol1 
 *	2.	nfs share add /vol1 -a 10.2.2.200 -o rw=all 	
 *	3.	nfs share add /vol1/testdir -a 10.2.2.200 -o rw=galaxian -n local
 *	4.	nfs share add /vol1 -a 10.2.2.200 -o rw=pacman:galaxian;rw,no_root_squash=raiders:skywalker -n local
 *	  
 * Created by:  Joseph Vu
 *
 * Date Created: 11 Feb 2002
 *
 *-----------------------------------------------------------------
 */
DEFUN   (nfs_shareAdd,
         nfs_share_add_cmd,
         "nfs share add PATHNAME [-a A.B.C.D] [-o OPTIONS] [-n CLUSTERNODE]",
         NFSCMD_HELP
         "Export\n"
		 "Add share\n"
         "Directory pathname (/vol/pathname)\n"
         "IP address of the export interface\n"
		 "A.B.C.D\n"
		 "Options\n"
         "rw|ro, root_squash|no_root_squash=client | all\n"
		 "Which node\n"
		 "Node name\n")
{
	char	ip[20];
	char	exportOpt[NFSSH_OPT_BUFLEN];
	char	clusterNode[NFSSH_CLUSTER_BUFLEN];
    char    optionBuf[NFSSH_OPT_BUFLEN];
	char	clientBuf[NFS_MAX_NAME_LEN + NFS_MAX_PATHNAME_LEN + 3];
    char    *argvp;
	char	*optionHostListp, *optionHostp, *hostp;
    char    *optionp;
	char	*optionBufp, *clientBufp;
	int		c;
	uint32	localIpAddr;
	int		opt_num = 0;    /* num. of options */
    int		my_argc = 1;
	char	*ridx;
	int		return_code, final_return_code;

#ifdef	NOTUSED
	{	
		//For debug only
		int		i;
		printf("nfs share: argc: %d\n", argc);
		for (i=0; i < argc; i++)
		{
			printf("argv[%d]: %s\n", i, argv[i]);
		}
	}		
#endif	//NOTUSED

	memset(ip, 0, sizeof(ip));
	memset(exportOpt, 0, sizeof(exportOpt));
	strcpy(clusterNode, "local");
	//set default option 
	memcpy(ip, DEFAULT_LADDR_STR, strlen(DEFAULT_LADDR_STR));	
	memcpy(exportOpt, EXPORT_OPT_STR, strlen(EXPORT_OPT_STR));
	localIpAddr = DEFAULT_LADDR;

	optind = 1;
	while ((c = getopt (argc, argv, "a:o:n:")) != -1)
	{
		switch (c)
		{
			case 'a':
			{
				struct in_addr	ip_addr;
				if (inet_aton(optarg, (struct in_addr*)&ip_addr) == 0)
				{
					printf("Error: Invalid IP address (%s)!\n", optarg);
					return CMD_ERR_FAILED;
				} 
				opt_num++;
				memset(ip, 0, sizeof(ip));
				memcpy(ip, optarg, strlen(optarg));
				//printf("nfs share add: -a ip: %s len: %d\n", 
				//		ip, strlen(ip));
				localIpAddr = ip_addr.s_addr;
				break;
			}
			case 'o':
			{
				opt_num++;
				memset(exportOpt, 0, sizeof(exportOpt));
				memcpy(exportOpt, optarg, strlen(optarg));
				//printf("nfs share add: -o exportOpt: %s len: %d\n", 
				//		exportOpt, strlen(exportOpt));
				
				break;
			}
			case 'n':
			{
				opt_num++;
				strcpy(clusterNode, optarg);
				break;
			}
		}   //switch
	}   //while

    //check parameter number matches
	my_argc += (2 * opt_num);
	if (my_argc != argc) {
		printf("Error: Incorrect number of parameters, argc %d, in_arg %d\n",
				argc, my_argc);
		return  CMD_ERR_FAILED;
	}

	//Defect 1480: 
	//Check if local IP interface is valid
	if (localIpAddr)
	{
		int			if_valid;
		uint32		clientIpAddr;

		if_valid = cmd_nfsIsIpIfValid(localIpAddr, clusterNode);
		if (!(if_valid))
		{
			clientIpAddr = ntohl(localIpAddr);
			printf("Error: IP (%d.%d.%d.%d) interface is not valid!\n",
                ((clientIpAddr >> 24) & 0xff),
                ((clientIpAddr >> 16) & 0xff),
                ((clientIpAddr >> 8) & 0xff),
                (clientIpAddr & 0xff) );
			return  CMD_ERR_FAILED;
		}
	}

    //Process option field
    argvp = argv[0];
	if (strlen(argvp) > NFS_MAX_PATHNAME_LEN)
	{
		printf("Error: Pathname is too long\n");
	}
	//Validate volume name
	{
		char			*pathNamep;

		if (*argvp == '/')
		{
			pathNamep = argvp+1;
			if (strcmp(argvp, "/") == 0)
			{
				printf("Error: Invalid path name!\n");
				return  CMD_ERR_FAILED;
			}
		}
		else
			pathNamep = argvp;

		/*
		 * The following code for getting volId doesn't work in clustering:
		volId = cmd_nfsGetVolId(pathNamep, strlen(pathNamep));
		if (volId.whole_thing <= 0)
		{
			printf("Error: Invalid volume name.\n");
			return  CMD_ERR_FAILED;
		}
		*/
	}

	optionHostListp = exportOpt;
	ridx = rindex(optionHostListp, ';');
	if (ridx == (optionHostListp + strlen(optionHostListp) - 1))
	{
		//There is ";" at the end of option host list. Remove the ";".
		*ridx = 0;
	}
	while((optionHostp = strsep(&optionHostListp, ";")))
	{
		//Now optionHostp points to string in form of
		//"rw,ro=client1:client2" or "rw,ro=all"

		optionp = strsep(&optionHostp, "=");
		if (!optionp)
		{
			printf("Error: Invalid option format.\n");
			return CMD_ERR_FAILED;
		}
		if (!optionHostp)
		{
			printf("Error: Invalid option format. Missing host list.\n");
			return CMD_ERR_FAILED;
		}

		if ((strlen(optionp) + strlen(laddr_str) + strlen(ip) + 2) > 
						NFSSH_OPT_BUFLEN)
		{ 
			printf("Error: Option string (%s) is too long\n", optionp);
			return CMD_ERR_FAILED;
		}

		optionBufp = &optionBuf[0];
		memset(optionBufp, 0, sizeof(optionBuf));
		strcpy(optionBufp, optionp);
		strcat(optionBufp, ",");
		strcat(optionBufp, laddr_str);
		strcat(optionBufp, ip);
		//Now optionBuf contains string in form of "rw,ro,laddr=0.0.0.0"

		//validate option string	
		if (cmd_nfsValidateOption(optionBufp, strlen(optionBufp)) != NFX_OK)
		{
			return CMD_ERR_FAILED;
		}	

		//Now optionHostp points to string in form of "client1:client2" or "all"
		ridx = rindex(optionHostp, ':');
		if (ridx == (optionHostp + strlen(optionHostp) - 1))
		{
			//There is ":" at the end of host list. Remove the ":".
			*ridx = 0;
		}
		while((hostp = strsep(&optionHostp, ":")))
		{
			clientBufp = &clientBuf[0];
			memset(clientBufp, 0, sizeof(clientBuf));

			if (strlen(hostp) == 0)
			{
				printf("Error: Invalid 'client:' format!\n");
				printf("%s\n", NFSCMD_SHARE_DEL_HELP);
				return CMD_ERR_FAILED;
			}

			if (strlen(hostp) > NFS_MAX_NAME_LEN)
			{
				printf("Error: Host string (%s) is too long\n", hostp);
				return CMD_ERR_FAILED;
			}

			if (strcmp(hostp, "all"))
			{
				//"all" is not specified
				strcpy(clientBufp, hostp);
				strcat(clientBufp, ":");
			}

			if (*argvp != '/')
			{
				strcat(clientBufp, "/");
			}

			strcat(clientBufp, argvp);
			//Now clientBuf contains string in form of
			//"client1:/pathname" or "/pathname"

			NFS_DBG_PRINTF(("NFS share add: optionBuf: %s\n", &optionBuf[0])); 
			NFS_DBG_PRINTF(("NFS share add: clientBuf: %s\n", &clientBuf[0])); 
			
			//Send export request to mountd
			return_code = cmd_nfsMntExportReq(&clientBuf[0],
											  localIpAddr,
											  optionBufp,
											  clusterNode);
			if (return_code != CMD_SUCCESS)
			{
				printf("Error: NFS failure to add share (%s)!\n", &clientBuf[0]);
				final_return_code = CMD_ERR_FAILED;
			}
		} //while(hostp)
	} //while(optionHostp)

	NFS_DBG_PRINTF(("NFS share add: exit\n"));
	return (final_return_code);
}

/*-----------------------------------------------------------------
 * Name:   	nfs_shareEnable 
 *
 * Description: NFS enable share.
 *      nfs share enable </pathname | all> [-n CLUSTERNODE]
 * Created by:  Joseph Vu
 *  
 * Date Created: 11 Feb 2002
 *
 *-----------------------------------------------------------------
 */
DEFUN   (nfs_shareEnable,
         nfs_share_ena_cmd,
         "nfs share enable PATHNAME [-n CLUSTERNODE]",
         NFSCMD_HELP
         "Export\n"
         "Enable share\n"
         "Directory pathname as [ /vol/pathname | all ]\n"
         "Which node\n"
         "Node name\n")
{
	char				clusterNode[NFSSH_CLUSTER_BUFLEN];
	char				reqMsgBuf[MSG_BUF_LEN];
	char				respMsgBuf[MSG_BUF_LEN];
	ncmMsg_t			*ncmReq;
	ncmMsg_t			*ncmRsp;
	uint32				ncmReqSize;
	uint32				ncmRspSize;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;
	int					stat;
	struct timeval      timeout = { 10, 0};
	int     			opt_num = 0;    //num. of options
	int     			my_argc = 1;
	int					c;

	strcpy(clusterNode, "local");
	optind = 1;
	while ((c = getopt (argc, argv, "n:")) != -1)		
	{
		switch (c)
		{
			case 'n':
				strcpy(clusterNode, optarg);
				opt_num++;
				break;
		}	//switch
	}	//while

	//check parameter number matches
	my_argc += (2 * opt_num);
	if (my_argc != argc) {
		printf("Error: Incorrect number of parameters, argc %d, in_arg %d\n",
				argc, my_argc);
		return  CMD_ERR_FAILED;
	}

	// Prepare NCM request and response
	cmd_nfsPrepareNcmMsg(reqMsgBuf,
						 respMsgBuf,
						 &ncmReq,
						 &ncmRsp,
						 &ncmReqSize,
						 &ncmRspSize,
						 &mntInfoReqp,
						 &mntInfoResp,
						 clusterNode);

	//Build MOUNTD request message
	mntInfoReqp->msgHdr.private = 0;

	if (!strncmp(argv[0], "all", 3) )
	{
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_ENABLED_ALL;
	}
	else
	{
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_ENABLED;
		mntInfoReqp->nameLen = strlen(argv[0]);
		strcpy(&mntInfoReqp->pathName[0], argv[0]);
		ncmReqSize +=  mntInfoReqp->nameLen + 1;

	}

	stat = sendAgileMsg (-1, ncmReq, ncmReqSize,
						"ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout, ncmRsp, ncmRspSize);
	if (stat < 0) 
	{
		printf("Error: NFS failure to enable shares!\n");
		return (CMD_ERR_FAILED);
	}

	if (mntInfoResp->msgHdr.status != NFX_OK)
	{
		cmd_nfsPrintMntRespStatus(mntInfoResp->msgHdr.status);
		return (CMD_ERR_FAILED);
	}

	return CMD_SUCCESS;
}

/*-----------------------------------------------------------------
 * Name:   	nfs_shareDisable 
 *
 * Description: NFS disable share.
 *      nfs share disable </pathname> [-n CLUSTERNODE]
 * Created by:  Joseph Vu
 *
 * Date Created: 11 Feb 2002
 *
 *-----------------------------------------------------------------
 */

DEFUN   (nfs_shareDisable,
         nfs_share_dis_cmd,
         "nfs share disable PATHNAME [-n CLUSTERNODE]",
         NFSCMD_HELP
         "Export\n"
         "Disable share\n"
         "Directory pathname as [ /vol/pathname | all ]\n"
         "Which node\n"
         "Node name\n")
{
	char				clusterNode[NFSSH_CLUSTER_BUFLEN];
	char				reqMsgBuf[MSG_BUF_LEN];
	char				respMsgBuf[MSG_BUF_LEN];
	ncmMsg_t			*ncmReq;
	ncmMsg_t			*ncmRsp;
	uint32				ncmReqSize;
	uint32				ncmRspSize;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;
	int					stat;
	int					c;
	struct timeval      timeout = { 10, 0};
	int     			opt_num = 0;    //num. of options
	int     			my_argc = 1;

	strcpy(clusterNode, "local");
    optind = 1;
	while ((c = getopt (argc, argv, "n:")) != -1)
	{
		switch (c)
		{
			case 'n':
				strcpy(clusterNode, optarg);
				opt_num++;
				break;
		}   //switch
	}   //while

	//check parameter number matches
	my_argc += (2 * opt_num);
	if (my_argc != argc) {
		printf("Error: Incorrect number of parameters, argc %d, in_arg %d\n",
			argc, my_argc);
		return  CMD_ERR_FAILED;
	}

	// Prepare NCM request and response
	cmd_nfsPrepareNcmMsg(reqMsgBuf,
						 respMsgBuf,
						 &ncmReq,
						 &ncmRsp, 
						 &ncmReqSize,
						 &ncmRspSize,
						 &mntInfoReqp,
						 &mntInfoResp,
						 clusterNode);

	//Build MOUNTD request message
	mntInfoReqp->msgHdr.private = 0;

	if (!strncmp(argv[0], "all", 3) )
	{
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_DISABLED_ALL;

	}
	else
	{
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_DISABLED;
		mntInfoReqp->nameLen = strlen(argv[0]);
		strcpy(&mntInfoReqp->pathName[0], argv[0]);
		ncmReqSize +=  mntInfoReqp->nameLen + 1;
	}

	stat = sendAgileMsg (-1, ncmReq, ncmReqSize,
						"ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout, ncmRsp, ncmRspSize);
	if (stat < 0) 
	{
		printf("Error: NFS failure to disable shares!\n");
		return (CMD_ERR_FAILED);
	}

	if (mntInfoResp->msgHdr.status != NFX_OK)
	{
		cmd_nfsPrintMntRespStatus(mntInfoResp->msgHdr.status);
		return (CMD_ERR_FAILED);
	}

	return CMD_SUCCESS;
}

/*-----------------------------------------------------------------
 * Name:    cmd_nfsDisplayEntriesHeader 
 *                                        
 * Description:
 *      NFS display export entries header.
 * 
 * Modified by:     Xiaobing Zhang
 *          
 * Date Created:    08/13/2002
 *
 *-----------------------------------------------------------------
 */
static void
cmd_nfsDisplayEntriesHeader()
{
	printf("Path              State     Export Option      Client         Local IP Interface\n");
	printf("--------------------------------------------------------------------------------\n");
	return;
}


/*-----------------------------------------------------------------
 * Name: nfs_show
 *
 * Description: Show nfs export 
 *		nfs show [VOLUMENAME] [-n CLUSTERNODE]
 * Created by:  Joseph Vu 
 *
 * Date Created: 11 Feb 2002
 *
 *-----------------------------------------------------------------
 */
DEFUN   (nfs_show,
        nfs_show_cmd,
		"nfs show [VOLUMENAME] [-n CLUSTERNODE]",
		NFSCMD_HELP
        "Show nfs exports\n"
		"Volume name\n"
		"Which node\n"
		"Node name\n")
{
	char				clusterNode[NFSSH_CLUSTER_BUFLEN];
	char				reqMsgBuf[MNT_MSG_BUF_MAX];
	char				respMsgBuf[MNT_MSG_BUF_MAX];
	ncmMsg_t			*ncmReq;
	ncmMsg_t			*ncmRsp;
	uint32				ncmReqSize;
	uint32				ncmRspSize;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;
	int					stat;
	struct timeval      timeout = { 10, 0};
	int     			my_argc = 0;
	char				*volNamep = NULL;
	int					totalNumOfEntries = 0;
	boolean				firstTime = TRUE;
	uint32				cookie = 0;

    strcpy(clusterNode, "local");

	//Parse parameters
	if (argc > 0)
	{
		if (*argv[0] != '-')
		{
			//Command contains VOLUMENAME
			volNamep = argv[0];
			if (*volNamep == '/')
				//Eat the leading '/'
				volNamep++;
			my_argc++;
		}

		if ((my_argc + 2) == argc)
		{
			if (strcasecmp(argv[my_argc], "-n") == 0)
			{
				strcpy(clusterNode, argv[my_argc+1]);
				my_argc++;
			}	
			else
			{
				printf("Error: Invalid option.\n");
				return  CMD_ERR_FAILED;
			}
		}
		else if (!(volNamep && argc == 1))
		{
			printf("Error: Invalid option.\n");
			return  CMD_ERR_FAILED;
		}
	}

	// Prepare NCM request and response
	cmd_nfsPrepareNcmMsg(reqMsgBuf,
					 	 respMsgBuf,
					 	 &ncmReq,
					 	 &ncmRsp, 
					 	 &ncmReqSize,
					 	 &ncmRspSize,
					 	 &mntInfoReqp,
					 	 &mntInfoResp,
					 	 clusterNode);

	//Build MOUNTD request message
	if (volNamep == NULL)
	{
		//Show all the volumes
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_SHOWALL;
	}
	else
	{
		//Show one volume
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_SHOW;
		mntInfoReqp->nameLen = strlen(volNamep);
		memcpy(&mntInfoReqp->pathName[0], volNamep, mntInfoReqp->nameLen);
		ncmReqSize +=  mntInfoReqp->nameLen + 1;
	}
	//Set response msg size limit. The total bytes of mount information
	//contained in one response message won't exceed this size.
	mntInfoReqp->respBufSize = sizeof(respMsgBuf) - sizeof(ncmMsg_t);

	while(1)
	{
		mntInfoReqp->cookie = cookie;

		//Send request
		stat = sendAgileMsg (-1, ncmReq, ncmReqSize,
							"ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							&timeout, ncmRsp, sizeof(respMsgBuf));
		if (stat < 0)
		{
			printf("Error: NFS failure to show shares!\n");
			return (CMD_ERR_FAILED);
		}

		if (mntInfoResp->msgHdr.status != NFX_OK)
		{
			cmd_nfsPrintMntRespStatus(mntInfoResp->msgHdr.status);
			return (CMD_ERR_FAILED);
		}

		//Obtain cookie from response message
		cookie = (uint32)mntInfoResp->cookie;

		// Print header
		if (firstTime == TRUE)
			cmd_nfsDisplayEntriesHeader();

		if (mntInfoResp->numOfEntries <= 0)
			break;		//No more export entries.

		//Show shares
		cmd_nfsDisplayEntries(mntInfoResp);

		totalNumOfEntries += mntInfoResp->numOfEntries;

		if (firstTime == TRUE)
			firstTime = FALSE;
	}

	printf("\nNumber of current export entries: %d\n", totalNumOfEntries);

	return CMD_SUCCESS;
}

void
install_nfs (priv_t myPriv, enum node_type node)
{
	switch (myPriv) {
	case diagPriv:
		//install_element(node, &nfs_exportall_ena_cmd);
		//install_element(node, &nfs_unexportall_ena_cmd);
	case fullPriv:
	case changePriv:
	case addPriv:
		install_element(node, &nfs_share_add_cmd);
		install_element(node, &nfs_share_del_cmd);
		install_element(node, &nfs_share_ena_cmd);
		install_element(node, &nfs_share_dis_cmd);

	case viewPriv:
		install_element(node, &nfs_show_cmd);
	}	
}
//move the following utility function to common library


/*-----------------------------------------------------------------
 * Name:  cmd_nfsExtractOption 
 *
 * Description:
 *  Extract  NFS option.
 * Created by:      Joseph Vu
 *
 * Date Created:    04/9/02
 *
 *-----------------------------------------------------------------
 */
int cmd_nfsExtractOption(char *optionp, char *optionKeyWordp)
{
	int		count;

	count = 0;

	while ( (count < NFSSH_OPT_KEYWORD_LEN) && 
			(*optionp != '\0') &&
			(*optionp != ',') )
	{
		*optionKeyWordp++ = *optionp++;
		count++;
	}
	if (*optionp == '\0')
		return (count);
	else
		return (count + 1);

}

/*-----------------------------------------------------------------
 * Name: 	cmd_nfsCheckOption 
 *
 * Description:
 *  Check  NFS option string is one of valid options.
 * Created by:      Joseph Vu
 *
 * Date Created:    04/9/02
 *
 *-----------------------------------------------------------------
 */
int cmd_nfsCheckOption (char *optionStr)
{

	if (strcmp(optionStr, "ro") == 0)
	{
		return (MOUNTD_RO_OPT);
	}
	else if (strcmp(optionStr, "rw") == 0)
	{
		return (MOUNTD_RW_OPT);
	}
	else if (strcmp(optionStr, "no_root_squash") == 0)
	{
		return (MOUNTD_NO_ROOT_SQUASH_OPT);
	}
	else if (strcmp(optionStr, "root_squash") == 0)
	{
		return (MOUNTD_ROOT_SQUASH_OPT);
	}
	else if (!(strncmp(optionStr,laddr_str,
				sizeof(laddr_str))))
	{
		return (MOUNTD_LADDR_OPT);
	}

	//Invalid option
	return (NFX_ERR);
}

/*-----------------------------------------------------------------
 * Name:   cmd_nfsValidateOption 
 *
 * Description:
 *	Validate NFS option string.
 * Created by:      Joseph Vu
 *
 * Date Created:    04/9/02
 *
 *-----------------------------------------------------------------
 */
int cmd_nfsValidateOption(char *optionp, int optionLen)
{
	int		len;
	int		count;
	int		optionMask;
	char	optionKeyWord[NFSSH_OPT_KEYWORD_LEN ];
	int		option;

	if ( (optionLen > NFSSH_OPT_BUFLEN) ||
		 (optionLen <= 0 ) )
		return (NFX_ERR);

	optionMask = 0;
	//sample input optionp = rw,no_root_squash, laddr=A.B.C.D
	count = 0;
	while ( (count < optionLen) && (*optionp != '\0') )
	{
		memset(&optionKeyWord[0], 0, sizeof(optionKeyWord));
		len = cmd_nfsExtractOption (optionp, &optionKeyWord[0]);
		option = cmd_nfsCheckOption(&optionKeyWord[0]);	
		if (option == NFX_ERR)
		{
			printf("Error: Invalid option (%s)\n!", &optionKeyWord[0]);
			return (NFX_ERR);
		}
		optionMask |= option;	
		count += len;
		optionp += len;
	}

	//Check for conflicting permission
	if  ( ((optionMask & MOUNTD_RW_OPT) &&
			(optionMask & MOUNTD_RO_OPT ))  ||
			( (optionMask & MOUNTD_NO_ROOT_SQUASH_OPT) &&
		(optionMask & MOUNTD_ROOT_SQUASH_OPT)) )
	{
		printf("Error: Conflicting in export option permission!\n");
		return (NFX_ERR);
	}	
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        nfs_cmdGetVolumeName
 *
 * Description:
 * Extract volume name from the input pathname.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int
nfs_cmdGetVolumeName(char *volNamep, int *volNameLen,
	char *pathNamep, int pathNameLen)
{
	int	count = 0;
   
	while ((pathNameLen--) && (pathNamep) && (*pathNamep != '/') &&
		  (*pathNamep != ' ') && (*pathNamep != '\t') )
	{
		*volNamep++ = *pathNamep++;
		count++;
	}
	*volNameLen = count;
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	 cmd_nfsGetVolId 
 *
 * Description:
 *		Get volume Id from EVM based for the input pathname.
 *
 * Created by:          Joseph Vu
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */
evm_volId_t cmd_nfsGetVolId(char *pathNamep, int32 pathNameLen)
{
	char			volume[VOL_MAX_NAME];
	int				volumeLen;
	evm_MsgGetId_t	getVolIdmsg;
	char 			respbuff[1024];
	evm_MsgStat_t	*respmsg = (evm_MsgStat_t *)respbuff;
	int 			stat;
	struct timeval	timeout = { 10, 0};

	memset(respbuff, 0, sizeof(respbuff));
	memset (&getVolIdmsg, 0, sizeof(getVolIdmsg)); 
	getVolIdmsg.hdr.msgType = EVM_VOL_GET_ID;

	getVolIdmsg.hdr.clientPrivate = 0;
	//
	//Get volume name to retrieve ID from EVM
	memset(&volume[0], 0, sizeof(volume));
	nfs_cmdGetVolumeName(&volume[0], &volumeLen,
			pathNamep, pathNameLen);

	//memcpy(getVolIdmsg.name, pathNamep, pathNameLen);
	memcpy(getVolIdmsg.name, &volume[0], volumeLen);

	stat = sendAgileMsg (-1, &getVolIdmsg, sizeof(getVolIdmsg),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout,
						respbuff, sizeof(respbuff));
	if (stat < 0)
	{
		respmsg->payload.lVolId.whole_thing = -1;
	}
	/*
	MOUNTD_ELOG(class_1, info_s, mountd_globalInfop->mountd_AppId,
			0, 0, 0,
			(MOUNTD_GET_VOLID_INFO_STR, MOUNTD_STR,
			respmsg->payload.lVolId.whole_thing, 
			pathNamep)); 
	*/
	return respmsg->payload.lVolId;
}

/*-----------------------------------------------------------------
 * Name:    cmd_nfsBuildOptionStr
 *
 * Description:
 *      Given option mask, build the corresponding option string.
 *
 * Created by:      Xiaobing Zhang
 *
 * Date Created:    08/13/2002
 *
 *-----------------------------------------------------------------
 */
void
cmd_nfsBuildOptionStr(char *optStr, int optionMask)
{
	char	*idx;

	if (optionMask & MOUNTD_RW_OPT)
	{
		strcat(optStr, MNT_RW_OPT_STR);
		strcat(optStr, ",");
	}

	if (optionMask & MOUNTD_RO_OPT)
	{
		strcat(optStr, MNT_RO_OPT_STR);
		strcat(optStr, ",");
	}

	if (optionMask & MOUNTD_ROOT_SQUASH_OPT)
	{
		strcat(optStr, MNT_ROOT_SQUASH_STR);
		strcat(optStr, ",");
	}

	if (optionMask & MOUNTD_NO_ROOT_SQUASH_OPT)
	{
		strcat(optStr, MNT_NO_ROOT_SQUASH_STR);
		strcat(optStr, ",");
	}

	idx = rindex(optStr, ',');
	*idx = 0;

	return;
}

/*-----------------------------------------------------------------
 * Name:   	cmd_nfsDisplayEntries 
 *
 * Description:
 *      NFS display export entries.
 * 
 * Created by:      Joseph Vu
 *
 * Date Created:    03/9/02
 *
 *-----------------------------------------------------------------
 */
void cmd_nfsDisplayEntries(mnt_infoResp_t *mntInfoResp)
{
	int				numOfEntries;
	int				i;
	mnt_infoEntry_t *mtnInfoEntryp;
	char            pathName[VOL_MAX_NAME];
	char			option[50];


	numOfEntries = mntInfoResp->numOfEntries;
	mtnInfoEntryp = &mntInfoResp->mtnInfoEntry[0];		
	for (i=0; i < numOfEntries; i++)
	{
		//Print path name
		memset(&pathName[0], 0, sizeof(pathName));
		if (mtnInfoEntryp->pathNameLen <= 0)
		{
			//Error. No export path name! Shouldn't happen.
			continue;
		}
		if (mtnInfoEntryp->pathNameLen >= sizeof(pathName))
		{
			memcpy(&pathName[0],
					&mtnInfoEntryp->pathName[0],
					sizeof(pathName)-1);
			printf("Warning: Export path name is truncated (%s)!\n", pathName);
		}
		else
		{
			memcpy(&pathName[0],
					&mtnInfoEntryp->pathName[0],
					mtnInfoEntryp->pathNameLen);
		}

		if (mtnInfoEntryp->pathNameLen > 16)
		{
			printf("%s\n", pathName);
			printf("%-18s", " ");
		}
		else
			printf("%-18s", pathName);

		//Print state
		printf("%-10s", (mtnInfoEntryp->state == MNT_SHARE_EXPORT_ENABLED)?
							"Enabled" : "Disabled");

		//Print option
		memset(&option[0], 0, sizeof(option));
		cmd_nfsBuildOptionStr(option, mtnInfoEntryp->exportOptionMask);
		printf("%-19s", option);

		//Print client IP address
		if (mtnInfoEntryp->hostAddr == 0)
			printf("%-17s", "any");
		else
		{
			struct in_addr ip_addr;

			ip_addr.s_addr = mtnInfoEntryp->hostAddr;
			printf("%-17s", inet_ntoa(ip_addr));
		}

		// Print local IP interface
		if (mtnInfoEntryp->localIpAddr == 0)
			printf("%-16s\n", "any");
		else
		{
			struct in_addr ip_addr;

			ip_addr.s_addr = mtnInfoEntryp->localIpAddr;
			printf("%-16s\n", inet_ntoa(ip_addr));
		}

		//Advance to next entry
		MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfoEntryp);
	}	
	return;
}

