/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_vif.c
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#include <netinet/if_ether.h>
#include "exports/ipmd-lib.h"

#if defined(LINK_AGG)
extern uint64		result_buf[];

static int	parse_links(char *dest, char *src)
{
	int		count = 0;
	char	*tmpDest = dest;

	bzero(dest, IPM_STR_SIZE);
	while (*src) {
		if (*src == ',') {
			count ++;
			tmpDest = &(dest[count*IPM_STR_SIZE]);
			bzero(tmpDest, IPM_STR_SIZE);
			src ++;
			continue;
		}
		
		*tmpDest = *src;
		tmpDest++, src++;
	}

	if (tmpDest != dest)
		count += 1;
	
	return count;
}


/*-----------------------------------------------------------------
 * Name: add_vif
 *
 * Description: create a virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/14/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (add_vif,
         add_vif_cmd,
		 "vif create  VIFNAME -m (failover|aggregation) -a PORTS [-n NODE]",
		 "virtual interface commands\n"
		 "Create a new virtual interface\n"
		 "virtual interface name, must start with vif\n"
		 "virtual interface mode\n"
		 "fail over mode\n"
		 "port aggregation mode\n"
		 "ports contained in virtual interface\n"
         "port names, separated by ,\n"
		 "Which node\n"
		 "Node name\n")
{
	char	node[64];
	char	links[64];
	char	*result_link;
	int		linkCnt;
	int		c, opt_num=0;
	int32	rc;
	int		my_argc = 3;
	int		flags;

	strcpy(node, "local");
	links[0] = '\0';
	optind = 0;
	while ((c = getopt(argc, argv, "a:n:")) != -1) {
		switch (c) {
		case 'n':
			strcpy(node, optarg);
			opt_num ++;
			break;
		}
	}

	my_argc += 2 * opt_num;

	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	
	/* get links */
	strcpy(links, argv[2]);
	/* parse links from  fp1.1,fp1.2, */
	linkCnt = parse_links((char*)result_buf, links);

	if (linkCnt)
		result_link = (char*)result_buf;
	else
		result_link = NULL;

	if (strncmp(argv[1], "failover", 4) == 0)
		flags = IPM_VIF_MODE_FO;
	else
		flags = IPM_VIF_MODE_LA;
	
	rc = ipmd_CreateVifCluster(argv[0], result_link, linkCnt, flags,
							  NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, links, argv[0], 0, 0);
		return CMD_ERR_FAILED;	
	}
}

/*-----------------------------------------------------------------
 * Name: del_vif
 *
 * Description: delete a virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/14/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (del_vif,
         del_vif_cmd,
		 "vif delete  VIFNAME [-n NODE]",
		 "virtual interface commands\n"
		 "Delete an existing virtual interface\n"
		 "Virtual interface name, must start with vif\n"
		 "Which node\n"
		 "Node name\n")
{
	char	node[64];
	int32	rc;
	int		c, opt_num=0;
	int		my_argc = 1;
	
	strcpy(node, "local");
	optind = 0;
	while ((c = getopt(argc, argv, "n:")) != -1) {
		switch (c) {
		case 'n':
			strcpy(node, optarg);
			opt_num ++;
			break;
		}
	}

	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	rc = ipmd_DeleteVifCluster(argv[0], NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, NULL, argv[0], 0, 0);
		return CMD_ERR_FAILED;	
	}
}

/*-----------------------------------------------------------------
 * Name: add_viflink
 *
 * Description: add links to virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/14/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (add_viflink,
         add_viflink_cmd,
		 "vif addport  VIFNAME -a PORTS [-n NODE]",
		 "virtual interface commands\n"
		 "Add ports to existing virtual interface\n"
		 "virtual interface name, must start with vif\n"
		 "ports\n"
         "port names, separated by ,\n"
		 "Which node\n"
		 "Node name\n")
{
	char	node[64];
	char	links[64];
	char	*result_link;
	int		linkCnt;
	int32	rc;
	int		c, opt_num=0;
	int		my_argc = 2;
	
	strcpy(node, "local");
	optind = 0;
	while ((c = getopt(argc, argv, "n:")) != -1) {
		switch (c) {
		case 'n':
			strcpy(node, optarg);
			opt_num ++;
			break;
		}
	}

	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
		
	strcpy(links, argv[1]);
	/* parse links from  fp1.1,fp1.2, */
	linkCnt = parse_links((char*)result_buf, links);

	if (linkCnt)
		result_link = (char*)result_buf;
	else
		result_link = 0;
	
	rc = ipmd_AddLinkVifCluster(argv[0], result_link, linkCnt, NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, links, argv[0], 0, 0);
		return CMD_ERR_FAILED;	
	}
}

/*-----------------------------------------------------------------
 * Name: del_viflink
 *
 * Description: remove links from virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/14/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (del_viflink,
         del_viflink_cmd,
		 "vif delport  VIFNAME -a PORTS [-n NODE]",
		 "virtual interface commands\n"
		 "Delete ports from existing virtual interface\n"
		 "virtual interface name, must start with vif\n"
		 "ports\n"
         "port names, separated by ,\n"
		 "Which node\n"
		 "Node name\n")
{
	char	node[64];
	char	links[64];
	char	*result_link;
	int		linkCnt;
	int32	rc;
	int		c, opt_num=0;
	int		my_argc = 2;
	
	strcpy(node, "local");
	optind = 0;
	while ((c = getopt(argc, argv, "a:n:")) != -1) {
		switch (c) {
		case 'n':
			strcpy(node, optarg);
			opt_num ++;
			break;
		}
	}

	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	strcpy(links, argv[1]);
	/* parse links from  fp1.1,fp1.2, */
	linkCnt = parse_links((char*)result_buf, links);

	if (linkCnt)
		result_link = (char*)result_buf;
	
	rc = ipmd_RmLinkVifCluster(argv[0], result_link, linkCnt, NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, links, argv[0], 0, 0);
		return CMD_ERR_FAILED;	
	}
}

/*-----------------------------------------------------------------
 * Name: set_vifpref
 *
 * Description: delete a virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/14/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (set_vifpref,
         set_vifpref_cmd,
		 "vif prefer  VIFNAME -a PORT [-n NODE]",
		 "virtual interface commands\n"
		 "Set prefered port of an existing  virtual interface\n"
		 "Virtual interface name, must start with vif\n"
		 "port\n"
         "prefered port name\n"
		 "Which node\n"
		 "Node name\n")
{
	char	node[64];
	int32	rc;
	int		c, opt_num=0;
	int		my_argc = 2;
	
	strcpy(node, "local");
	optind = 0;
	while ((c = getopt(argc, argv, "n:")) != -1) {
		switch (c) {
		case 'n':
			strcpy(node, optarg);
			opt_num ++;
			break;
		}
	}

	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	rc = ipmd_PrefVifCluster(argv[0], argv[1], NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, argv[1], argv[0], 0, 0);
		return CMD_ERR_FAILED;	
	}
}


/*-----------------------------------------------------------------
 * Name: show_vif
 *
 * Description: display virtual interface 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/14/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (show_vif,
         show_vif_cmd,
		 "vif show [VIFNAME] [-n NODE]",
		 "virtual interface commands\n"
		 "Display virtual interface\n"
		 "Virtual interface name, must start with vif\n"
		 "Which node\n"
		 "Node name\n")
{
	char		node[64];
	int32		rc;
	int			i, opt_num=0;
	ipm_vif_t	*ipm_vifp = (ipm_vif_t*)result_buf;
	ipm_if_t	*ipm_ifp;
	int			need=0, count=0;
	int			fail_over;
	int			my_argc = 0;
	char		vifName[IPM_STR_SIZE];

	strcpy(node, "local");
	optind = 0;
	vifName[0] = '\0';

	/* copy option ourselves, 
	 * getopt() does not work
	 */
	for (i=0; i < argc; i++) { 
		
		if (strcasecmp(argv[i], "-n") == 0) {
			strcpy(node, argv[i+1]);
			opt_num ++;
			i++;
		}
		else {
			
//		if (strncasecmp(argv[i], "vif", 3) == 0) {
			strcpy(vifName, argv[i]);
			opt_num ++;
			my_argc--;			/* compensate for no flag */
			
		}
	}
	
	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	rc = ipmd_GetVifByNameCluster(((vifName[0] == '\0') ? NULL:vifName), 
								  ipm_vifp, IPM_SSC_BUFSIZE,
								  &count, &need, node);
	
	if (rc != NFX_OK) {
		ipmd_PrntErrMsg(rc, NULL, vifName, 0, 0);
		return CMD_ERR_FAILED;	
	}

	
	if (vifName[0]) {
		if ((ipm_vifp->vif_flags & IPM_VIF_MODE_FO))
			fail_over = 1;
		else
			fail_over = 0;
	
		printf("Name             : %s\n", ipm_vifp->vif_name);
		printf("Num. of Ports	 : %d\n", ipm_vifp->vif_linkCnt);
		printf("Mode             : %s\n", fail_over?"fail-over":"aggregation");
		printf("IP Cnt           : %d\n", ipm_vifp->vif_ipCnt);
		if (fail_over) {
			printf("Prefered Port    : %s\n", ipm_vifp->vif_prefLink);
			printf("Active Port      : %s\n", ipm_vifp->vif_actLink);
		}
		
		printf("\n");
		
		printf("Ports\n");
		printf("-------------------------------------------------------\n");
		printf("Interface     Admin State   Operat. State		   \n");
		printf("-------------------------------------------------------\n");
		ipm_ifp = (ipm_if_t*)(ipm_vifp + 1);
		while (ipm_vifp->vif_linkCnt --) {
			printf("%-15s%-15s%s\n", ipm_ifp->if_name,
				   (ipm_ifp->if_flags & IFF_UP)?"UP":"DOWN",
				   (ipm_ifp->if_auxFlags & IPM_AUXIF_RUNNING)?"UP":"DOWN");
			ipm_ifp ++;
		}
	}
	else {
		int	i, j;
		char	links[IPM_STR_SIZE*IPM_MAXLINK_VIF];
		
	
		printf("Virtual Interface Table\n");
		printf("-----------------------------------------------------------------\n");
		printf("Name      Mode        IPCnt   PrefPort  ActPort   Ports\n");
		printf("-----------------------------------------------------------------\n");

		for (i = 0; i < count; i ++) {
			/* display summery of vif */
			if ((ipm_vifp->vif_flags & IPM_VIF_MODE_FO))
				fail_over = 1;
			else
				fail_over = 0;

			links[0] = '\0';
			for (j=0; j<ipm_vifp->vif_linkCnt; j++) {
				strcat(links, ipm_vifp->vif_links[j]);
				if (j < (ipm_vifp->vif_linkCnt -1))
					strcat(links, ",");
			}
			
			printf("%-10s%-12s%-8d%-10s%-10s%s\n", ipm_vifp->vif_name,
				   fail_over?"fail-over":"aggregation",
				   ipm_vifp->vif_ipCnt,
				   fail_over?ipm_vifp->vif_prefLink:"N/A",
				   fail_over?ipm_vifp->vif_actLink:"N/A",
				   links);

			ipm_vifp ++;
		}
	}
	
	return CMD_SUCCESS;
}

#endif /* LINK_AGG */

void
install_vif (priv_t myPriv, enum node_type node)
{
#if defined(LINK_AGG)
	switch (myPriv) {
	case diagPriv:
	case fullPriv:
		install_element(node, &add_vif_cmd);
		install_element(node, &del_vif_cmd);
		install_element(node, &add_viflink_cmd);
		install_element(node, &del_viflink_cmd);		
		install_element(node, &set_vifpref_cmd);
	case changePriv:
	case addPriv:
	case viewPriv:
		install_element(node, &show_vif_cmd);
	}
#endif /* LINK_AGG */	
}
