/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ntp_utils.c
 *
 * RCS:  $Id: ntp_utils.c,v 1.9 2002/08/20 21:05:30 maximk Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: Restarts NTP using the Agile configuration.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Apr 2002
 *
 *-----------------------------------------------------------------
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>

#include "exports/opt-api.h"
#include "exports/anpssc-api.h"

#include "timekeep-api.h"

/*
	NTP configuration manipulation things
 */
static FILE *
copy_template (void)
{
	FILE *templ;
	FILE *conf;
	char buf[256];

	templ = fopen (NTPCONFTEMPLATE, "r");
	conf = fopen (NTPCONF, "w");

	if (!templ || !conf) {
		fprintf (stderr, "Unable to rewrite ntp configuration\n");
	}
	else {
		while (!feof (templ)) {
			fgets (buf, sizeof(buf), templ);
			fputs (buf, conf);
		}
	}
	fclose (templ);
	return conf;
}

static void
restart_ntpd (void)
{
	FILE *pidfile = fopen (NTPDPID, "r");
	pid_t ntpd_pid = 0;
	pid_t child_pid;
	int i;
	int ch_stat;

	if (pidfile) {
		fscanf (pidfile, "%d", &ntpd_pid);
		fclose (pidfile);
	}
	if (ntpd_pid) {
		kill (ntpd_pid, SIGTERM);
	}
/*
	Restart ntpd. Need to make sure none of the caller's open file descriptors
	get inherited by the new ntpd child.
 */
	if (!(child_pid = fork())) {
		// Close all FDs except stdin, stdout, and stderr
		for (i = 3; i < FD_SETSIZE; i++) {
			close(i);
		}
		daemon (0, 0);
		execl (NTPDPROG, NTPDPROG, "-p", NTPDPID, NULL);
	}
	else {
		waitpid (child_pid, &ch_stat, 0); // reap your child's death
	}
}

/*-----------------------------------------------------------------
 * Name :	tk_write_ntp_conf()
 *
 * Description: Write ntp configuration files
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/15/02
 *
 *-----------------------------------------------------------------
 */
static void
tk_write_ntp_conf (void)
{
/*
	As a cluster controller or a stand-alone system, we have three choices 
	here:
	1) They've disabled ntp
	2) We're using specified servers
	3) We're a broadcast client

	In the 1) case, we continue to use our local clock as a server source
	as that config is in the template.
	In the 2) case, we append the list of configured servers to the template
	In the 3) case, we append the "broadcast" client directive to the
	template.
*/
	char **opts;
	FILE *conf;
	FILE *tickers;

	conf = copy_template();	// This returns NULL and we're hosed
	if (!conf) {
		fprintf (stderr, "Cannot update ntp configuration\n");
		return;
	}

	opts = get_cfg_val (NTP_MODE);

	tickers = fopen (NTPTICKERS, "w");

	if (opts) {	// Something to append to ntp.conf
		if (!strcasecmp (opts[1], "broadcast")) {
			fputs ("broadcastclient\n", conf);
		}
		else {
			char *nxt;
			char *srv = index (opts[1], ':') + 1;

			nxt = srv;
			while (srv) {
				nxt = strsep (&srv, "|");
				fprintf (conf, "server %s\n", nxt);
				fprintf (tickers, "%s\n", nxt);
			}
		}
		free_cfg_val (opts);
	}
	fprintf (tickers, "127.0.0.1\n");		// We're always there
	fclose (tickers);

	fclose (conf);
}


/*-----------------------------------------------------------------
 * Name: tk_update_ntp_conf
 *
 * Description: Commit the current configuration to the running ntpd
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Apr 2002
 *
 *-----------------------------------------------------------------
 */
extern void
tk_update_ntp_conf (void)
{
/*
    Write ntp.conf and restart ntpd.
 */
	modify_readonly_file("/etc/ntp.conf", tk_write_ntp_conf, NULL);

	restart_ntpd();
	sleep (1);		// Give it time to restart

	return;
}

/*-----------------------------------------------------------------
 * Name: tk_mod_ntp_servers
 *
 * Description: Change the current NTP configuration
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Apr 2002
 *
 *-----------------------------------------------------------------
 */
extern int 
tk_mod_ntp_servers (char *serv_addr, int del)
{
	char **opts;
	char *servers;

	opts = get_cfg_val (NTP_MODE);

	if (opts && !strstr (opts[1], SERVERSTR)) {	// Need to re-write object
		free_cfg_val (opts);
		opts = NULL;
	}

	if (opts) {
		char *in_list = strstr (opts[1], serv_addr);
		char *term;
		char *start_list = index (opts[1], ':') + 1;

		if (del) {
			if (!in_list) {
#if defined (DEBUG)
				fprintf (stderr, "%s not a configured server\n", serv_addr);
#endif
				errno = ENOENT;
				return 1;
			}

			term = index (in_list, '|');

			if ((in_list == start_list) && !term) {	// Last one, delete object
				del_cfg_obj (opts);
				free_cfg_val (opts);
				opts = NULL;
			}
			else if (!term) { // It's at the end -- easy
				in_list[-1] = '\0';
			}
			else {	// It's first or in the middle
				char *nxt = index (in_list, '|') + 1;
				strcpy (in_list, nxt);
			}
		}
		else { // Append if not already there
			if (in_list) {
#if defined (DEBUG)
				fprintf (stderr, "%s server already configured\n", serv_addr);
#endif
				return 0;
			}
			servers = malloc (strlen(opts[1]) + strlen(serv_addr) + 2);
			sprintf (servers, "%s|%s", opts[1], serv_addr);
			free (opts[1]);
			opts[1] = servers;
		}
	}
	else {	// Brand new ntp config
		opts = malloc (3 * sizeof(char *));
		opts[0] = strdup (NTP_MODE);
		opts[1] = malloc (strlen(SERVERSTR) + strlen (serv_addr) + 1);
		sprintf (opts[1], "%s%s", SERVERSTR, serv_addr);
		opts[2] = NULL;
	}

	if (opts) {
		if (put_cfg_val (opts)) {
#if defined (DEBUG)
			fprintf (stderr, "Error setting NTP servers\n");
#endif
			free_cfg_val (opts);
			errno = EIO;
			return 1;
		}
		free_cfg_val (opts);
	}

	return 0;
}

/*-----------------------------------------------------------------
 * Name: tk_get_ntp_cfg
 *
 * Description: Give me the NTP configuration
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Apr 2002
 *
 *-----------------------------------------------------------------
 */
extern char *
tk_get_ntp_cfg (void)
{
	char **opts;
	char *cfg;

	opts = get_cfg_val (NTP_MODE);
	if (opts) {
		cfg = strdup (opts[1]);
		free_cfg_val (opts);
	}
	else {
		cfg = NULL;
	}

	return cfg;
}

/*-----------------------------------------------------------------
 * Name: tk_set_ntp_broadcast
 *
 * Description: Set the configuration to use NTP broadcast messages
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Apr 2002
 *
 *-----------------------------------------------------------------
 */
extern int 
tk_set_ntp_broadcast (void)
{
	char *opts[3];

	opts[0] = NTP_MODE;
	opts[1] = "broadcast";
	opts[2] = NULL;

	return put_cfg_val(opts);
}

/*-----------------------------------------------------------------
 * Name: tk_disable_ntp
 *
 * Description: Completely disable the use of NTP.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Apr 2002
 *
 *-----------------------------------------------------------------
 */
extern void 
tk_disable_ntp (void)
{
	char *opts[2];

	opts[0] = NTP_MODE;
	opts[1] = NULL;

	del_cfg_obj (opts);
}
